/* @(#)FMatrix.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _FMATRIX_HH
#define _FMATRIX_HH 1

#include "IMatrix.hh"
#include "Fractional.hh"
#include "IntegerTricks.hh"

/* ***********************************************************************
 *
 * ***********************************************************************
 */


class Matrix3x3;

class FMatrix{
public:

  /* ***************************
   * construction
   **************************** */

  FMatrix():prows(0),pcols(0),p(0){}

  FMatrix(long rows,long cols):p(0){
    setDimenions(rows,cols);}

  FMatrix(long rows,long cols,const long *m,long denom=1):p(0){
    setDimenions(rows,cols);
    for(long i=0;i<rows*cols;i++){
      p[i]=Fractional(m[i],denom);
    }
  }


  FMatrix(const FMatrix &other)
    :prows(other.prows),pcols(other.pcols),p(0){
    setDimenions(prows,pcols);
    if(p && other.p) 
      for(long i=0;i<prows*pcols;i++) p[i]=other.p[i];
  }

  FMatrix(const IMatrix &other)
    :prows(other.prows),pcols(other.pcols),p(0){
    setDimenions(prows,pcols);
    if(p && other.p) 
      for(long i=0;i<prows*pcols;i++) p[i]=other.p[i];
  }

  FMatrix(const FMatrix &o1,const FMatrix &o2):prows(0),pcols(0),p(0){
    addMatrix(o1);
    addMatrix(o2);
  }
  FMatrix(const FMatrix &o1,const FMatrix &o2,const FMatrix &o3)
    :prows(0),pcols(0),p(0){
    addMatrix(o1);
    addMatrix(o2);
    addMatrix(o3);
  }

  ~FMatrix(){clean();}

  FMatrix &operator=(const FMatrix &other){
    setDimenions(other.prows,other.pcols);
    if(p && other.p) 
      for(long i=0;i<prows*pcols;i++) p[i]=other.p[i];
    return *this;
  }
  FMatrix &operator=(const Fractional &o){
    for(long i=0;i<prows*pcols;i++) p[i]=o;
    return *this;
  }

  static FMatrix diagMatrix(long rows,long cols,const Fractional &d){
    FMatrix pp(rows,cols);
    for(long r=0;r<MyMath::min<long>(pp.prows,pp.pcols);r++) pp.p[r+pp.prows*r]=d;
    return pp;
  }

  static FMatrix vector3(const Fractional &v0,const Fractional &v1,
			 const Fractional &v2){
    FMatrix v(3,1);
    v[0]=v0;
    v[1]=v1;
    v[2]=v2;
    return v;
  };
  static FMatrix vector3(long v0,long v1,long v2,long denom){
    FMatrix v(3,1);
    v.setColumn3(0,v0,v1,v2,denom);
    return v;
  };
  static FMatrix vector2(long v0,long v1,long denom){
    FMatrix v(2,1);
    v[0]=Fractional(v0,denom);
    v[1]=Fractional(v1,denom);
    return v;
  };


  static FMatrix vector3(const long *m){
    FMatrix res(3,1);
    for(long i=0;i<3;i++){
      res.p[i]=m[i];
    }
    return res;
  };


  FMatrix fromColumn(long c)const
  {
    FMatrix m(prows,1);
    for(long i=0;i<prows;i++){
      m.p[i]=p[i+prows*c];
    }
    return m;
  }

  FMatrix &cycleColumnsL(){
    addMatrix(fromColumn(0));
    removeColumns(0,0);
    return *this;
  }
  FMatrix &swapColumns(long c1,long c2){
    for(long r=0;r<prows;r++){
      Fractional d=p[r+prows*c1];
      p[r+prows*c1]=p[r+prows*c2];
      p[r+prows*c2]=d;
    }
    return *this;
  }

  /* ***************************
   * restructuring
   **************************** */

  void setDimenions(long rows,long cols){
    //destroys the array
    clean();
    prows=rows;pcols=cols;
    long n=prows*pcols;
    if(n>0){
      p=new Fractional[prows*pcols];
    }
  }

  void addRows(long n){
    if(n<1) return;
    if(!p){
      setDimenions(n,pcols);
      return;
    }
    //myDebug()<<"addRows"<<n<<matInfoS();
    long prnew=(prows+n);
    Fractional *temp=new Fractional[prnew*pcols];
    for(long c=0;c<pcols;c++){
      for(long i=0;i<prows;i++){
	temp[i+prnew*c]=p[i+prows*c];
      }
      for(long i=prows;i<prnew;i++){
	temp[i+prnew*c]=Fractional();
      }
    }
    swapp(temp);
    delete [] temp;
    prows=prnew;
  }

  void addColumns(long n){
    if(n<1) return;
    if(!p){
      setDimenions(prows,n);
      return;
    }
    //myDebug()<<"addColumnms"<<matInfoS();
    Fractional *temp=new Fractional[prows*(pcols+n)];
    for(long i=0;i<prows*pcols;i++) temp[i]=p[i];
    for(long i=prows*pcols;i<prows*(pcols+n);i++) temp[i]=Fractional();

    swapp(temp);
    delete [] temp;
    pcols+=n;
    //myDebug()<<"addColumnms"<<matInfoS();
  }

  FMatrix &removeColumns(long from,long to){
    if(to<from) return *this;
    if(from>=pcols) return *this;
    long lto=MyMath::min<long>(to,pcols-1);
    //myDebug()<<"removeColumns: "<<from<<lto<<pcols;
    long pcnew=pcols-(lto-from+1);
    //myDebug()<<"removeColumns: pcnew"<<pcnew;
    if(pcnew<1){
      pcols=0;
      clean();
      return *this;
    }
    Fractional *temp=new Fractional[prows*pcnew];
    if(from>0){
      for(long i=0;i<prows*from;i++) temp[i]=p[i];
    }


    for(long i=prows*(lto+1);i<prows*pcols;i++) 
      temp[i+prows*(from-lto-1)]=p[i];

    swapp(temp);
    pcols=pcnew;
    delete [] temp;
    return *this;
  }


  FMatrix &removeColumns(const IMatrix &map){
    // remove column if map(0,c)==0
    if(map.columns()<pcols)
      throw Except("FMatrix::removeColumns: map has not enough columns");


    FMatrix temp(prows,pcols);
    
    long pcnew=0;
    for(long c=0;c<pcols;c++){
      if(map(0,c)!=0){
	for(long r=0;r<prows;r++){
	  temp.p[r+prows*pcnew]=p[r+prows*c];
	}
	pcnew++;
      }
    }
    temp.removeColumns(pcnew,pcols-1);
    swapAll(temp);
    return *this;
  }

  FMatrix &removeRows(long from,long to){
    if(to<from) return *this;
    if(from>=prows) return *this;
    long lto=MyMath::min<long>(to,prows-1);
    //myDebug()<<"removeRows: "<<from<<lto<<prows;
    long prnew=prows-(lto-from+1);
    //myDebug()<<"removeRows: prnew"<<prnew;
    if(prnew<1){
      prows=0;
      clean();
      return *this;
    }
    Fractional *temp=new Fractional[prnew*pcols];

    if(from>0){
      for(long c=0;c<pcols;c++) 
	for(long r=0;r<from;r++) temp[r+prnew*c]=p[r+prows*c];
    }

    for(long c=0;c<pcols;c++) 
      for(long r=lto+1;r<prows;r++) 
	temp[r-lto-1+from+prnew*c]=p[r+prows*c];


    swapp(temp);
    prows=prnew;
    delete [] temp;
    return *this;
  }
  FMatrix &addMatrix(const FMatrix &o){
    // new matrix has the first columns from this and o columns appended
    // rows are adjusted
    long opc=pcols;
    addColumns(o.pcols);
    addRows(o.prows-prows);
    for(long r=0;r<o.prows;r++){
      for(long c=0;c<o.pcols;c++){
	p[r+prows*(c+opc)]=o.p[r+o.prows*c];
      }
    }
    return *this;
  }
  /* ***************************
   * access
   **************************** */

  long rows()const{return prows;}
  long columns()const{return pcols;}
  
  const Fractional *data()const{return p;}
  
  Fractional &operator()(long r,long c){
    checkIndex(r,c);
    return p[r+prows*c];
  }
  const Fractional &operator()(long r,long c)const{
    checkIndex(r,c);
    return p[r+prows*c];
  }
  Fractional &operator[](long r){
    if(pcols!=1)throw Except("FMatrix::operator[] needs pcols==1!");
    checkIndex(r,0);
    return p[r];
  }
  const Fractional &operator[](long r)const{
    if(pcols!=1)throw Except("FMatrix::operator[] needs pcols==1!");
    checkIndex(r,0);
    return p[r];
  }


  void setColumn3(long c,long r0,long r1,long r2,long denom){
    if(prows<3)throw Except("FMatrix::setColumn3 less than 3 rows.");
    if(c+1>pcols)throw Except(Sprintf("FMatrix::setColumn "
				      "less than %ld columns.",c));

    p[0+prows*c]=Fractional(r0,denom);
    p[1+prows*c]=Fractional(r1,denom);
    p[2+prows*c]=Fractional(r2,denom);
  }

  void setColumn(long c,const FMatrix &other){
    if(prows!=other.prows)
      throw Except("FMatrix::setColumn dimensions dont match.");
    if(other.pcols!=1)
      throw Except("FMatrix::setColumn argument must be a vector.");
    if(c+1>pcols)throw Except(Sprintf("FMatrix::setColumn "
				      "less than %ld columns.",c));
    
    for(long r=0;r<prows;r++){
      p[r+prows*c]=other.p[r];
    }
    
  }


  void scaleColumn(long c,const Fractional &s){
    checkIndex(0,c);

    for(long r=0;r<prows;r++) p[r+prows*c]*=s;
  }

  
  void addColumn3(long r0,long r1,long r2,long denom){
    if(prows<3)throw Except("FMatrix::addColumn3 less than 3 rows.");

    addColumns(1);
    
    p[0+prows*(pcols-1)]=Fractional(r0,denom);
    p[1+prows*(pcols-1)]=Fractional(r1,denom);
    p[2+prows*(pcols-1)]=Fractional(r2,denom);
  }

  /* ***************************
   * nummerics
   **************************** */

  bool containsColumn(const FMatrix &o)const{
    if(o.pcols!=1)
      throw Except("FMatrix:: containsColumn(): argument is not a vector!");
    if(prows!=o.prows)return false;
    //myDebug()<<"o="<<o.fromColumn(0).Td().printS();
	  
    for(long c=0;c<pcols;c++){
      //myDebug()<<"col"<<c<<this->fromColumn(c).Td().printS();

      bool eq=true;
      for(long r=0;r<prows;r++){
	if(p[r+prows*c]!=o.p[r]){eq=false;break;}
      }
      //myDebug()<<"eq"<<eq;
      if(eq) return true;
    }
    return false;
  }

  bool isInteger()const{
    for(long i=0;i<prows*pcols;i++){
      if(!p[i].isInteger()) return false;
    }
    return true;
  };

  IMatrix toIMatrix()const{
    if(!isInteger())
      throw Except("FMatrix::toIMatrix: matrix is not integer?"); 

    IMatrix res(prows,pcols);
    for(int i=0;i<prows*pcols;i++){
      res.p[i]=p[i].n();
    }
    return res; 
  }

  IMatrix toIMatrixColumnsWise(IMatrix &solvable)const{
    // return integer matrix and set element of solvable ot zero, if
    // corresponding column is not convertible, solvable must be initialized
    // before calling this function.
    
    IMatrix res(prows,pcols);
    for(int c=0;c<pcols;c++){
      Fractional *pp=p+prows*c;
      long *ppr=res.p+prows*c;
      for(int r=0;r<prows;r++){
	if(pp[r].isInteger()){
	  ppr[r]=pp[r].n();
	}else{
	  ppr[r]=0;
	  solvable(0,c)=0;
	}
      }
    }
    return res; 
  }

  IMatrix round()const{

    IMatrix res(prows,pcols);
    for(int i=0;i<prows*pcols;i++){
      res.p[i]=p[i].round();
    }
    return res; 
  }

  Matrix3x3 toMatrix3x3()const;

  bool moreComplex(const FMatrix &o)const{
    if(prows!=o.prows || pcols!=o.pcols)
      throw Except("FMatrix:: moreComplex(): different dimensions");
    FMatrix v1=this->regularizedInterval(false);
    FMatrix v2=o.regularizedInterval(false);
    //myDebug()<<"FMatrix::moreComplex:"<<v1.printS()<<v2.printS();

    long v1zerocount=0;
    for(long i=0;i<prows*pcols;i++){
      if(v1.p[i]==0) v1zerocount++;
    }
    long v2zerocount=0;
    for(long i=0;i<prows*pcols;i++){
      if(v2.p[i]==0) v2zerocount++;
    }
    //myDebug()<<"FMatrix::moreComplex:"<<v1zerocount<<v2zerocount;
    return (v1zerocount<v2zerocount);
  }

  FMatrix &operator+=(const FMatrix &o){
    if(prows!=o.prows || pcols!=o.pcols)
      throw Except("FMatrix:: operator+=(): different dimensions");
    for(long i=0;i<prows*pcols;i++){
      p[i]+=o.p[i];
    }
    return *this;
  }
  FMatrix &operator-=(const FMatrix &o){
    if(prows!=o.prows || pcols!=o.pcols)
      throw Except("FMatrix:: operator+=(): different dimensions");
    for(long i=0;i<prows*pcols;i++){
      p[i]-=o.p[i];
    }
    return *this;
  }

  FMatrix &operator/=(const Fractional &s){
    for(long i=0;i<prows*pcols;i++){
      p[i]/=s;
    }
    return *this;
  }
  FMatrix &operator*=(long s){
    for(long i=0;i<prows*pcols;i++){
      p[i]*=s;
    }
    return *this;
  }

  FMatrix operator-()const{
    FMatrix res(*this);
    for(long i=0;i<prows*pcols;i++){
      res.p[i]=-res.p[i];
    }
    return res;
  }

  FMatrix operator-(const FMatrix &o)const{
    if(prows!=o.prows || pcols!=o.pcols)
      throw Except("FMatrix:: operator-(): different dimensions");
    FMatrix res(*this);
    for(long i=0;i<prows*pcols;i++){
      res.p[i]-=o.p[i];
    }
    return res;
  }
  FMatrix operator-(const IMatrix &o)const{
    if(prows!=o.prows || pcols!=o.pcols)
      throw Except("FMatrix:: operator-(IMatrix): different dimensions");
    FMatrix res(*this);
    for(long i=0;i<prows*pcols;i++){
      res.p[i]-=o.p[i];
    }
    return res;
  }
  FMatrix operator+(const FMatrix &o)const{
    if(prows!=o.prows || pcols!=o.pcols)
      throw Except("FMatrix:: operator+(): different dimensions");
    FMatrix res(*this);
    for(long i=0;i<pcols*prows;i++){
      res.p[i]+=o.p[i];
    }
    return res;
  }
  FMatrix operator*(const Fractional &d)const{
    FMatrix res(prows,pcols);
    for(long i=0;i<pcols*prows;i++){
      res.p[i]=p[i]*d;
    }
    return res;
  }
  FMatrix operator/(const Fractional &d)const{
    FMatrix res(prows,pcols);
    for(long i=0;i<pcols*prows;i++){
      res.p[i]=p[i]/d;
    }
    return res;
  }

  bool operator==(long o)const{
    for(long i=0;i<pcols*prows;i++){
      if(p[i]!=o)return false;
    }
    return true;
  };

  bool operator!=(long o)const{
    for(long i=0;i<pcols*prows;i++){
      if(p[i]!=o)return true;
    }
    return false;
  };

  bool operator==(const FMatrix &o)const{
    if(prows!=o.prows || pcols!=o.pcols) return false;
    for(long i=0;i<pcols*prows;i++){
      if(p[i]!=o.p[i])return false;
    }
    return true;
  };

  bool operator!=(const IMatrix &o)const{
    if(prows!=o.prows || pcols!=o.pcols)
      return true;
    for(long i=0;i<pcols*prows;i++){
      if(p[i]!=o.p[i])return true;
    }
    return false;
  }
  
  bool operator!=(const FMatrix &o)const{
    if(prows!=o.prows || pcols!=o.pcols)
      return true;
    for(long i=0;i<pcols*prows;i++){
      if(p[i]!=o.p[i])return true;
    }
    return false;
  }
  
  IMatrix areColumnsZero()const{
    // return pcolsx1 matrix, with elements indicating, if the
    // columns is zero
    IMatrix res(1,pcols);
    for(long c=0;c<pcols;c++){
      res(0,c)=1;
      Fractional *pp=p+prows*c;
      for(long r=0;r<prows;r++){
	if(pp[r]!=0){
	  res(0,c)=0; break;
	}
      }
    }
    return res;
  }

  FMatrix &regularizeInterval(bool centered){
    for(long i=0;i<pcols*prows;i++){
      p[i].regularizeInterval(centered);
    }
    return *this;
  }

  FMatrix regularizedInterval(bool centered)const{
    FMatrix v(*this);
    return v.regularizeInterval(centered);
  }

  
  IMatrix decompAVeqB(long &rank,bool detnotone=false){
    // Decompose R such that R*V=(B 0), where B forms a basis of R
    // Return V and modify R such that it equals a square rows x rows B, with 
    // the rank of B returned in rank.
    // The determinant of V is 1 if !detnotone.
    // If detnotone detV>1 and B diagonal.
    IMatrix V(pcols,pcols);
    rank=IntegerTricks::decompAVeqBF(p,V.p,prows,pcols,detnotone); 
    //removeColumns(rank,pcols-1);
    //if(pcols<prows) addColumns(prows-pcols);
    //T();
    return V;
  }

  void  decompGTAVeqD(long &rank,IMatrix &V,IMatrix &GT){
    // Decompose R such that R*V=(B 0), where B forms a basis of R followed by
    // G^T*B=D. Hence G^T*R*V=D. Dimensions are 
    //     R:rows x cols
    //     V:cols x cols
    //     B:rows x cols (effectivly rows x rank)
    //     GT: rows x rows
    // Return V and B in R
    // The rank of B returned in rank.
    // The determinant of V is 1 if !detnotone.
    // If detnotone detV>1 and B diagonal.
    V.setDimenions(pcols,pcols);
    GT.setDimenions(prows,prows);
    if(pcols==0){
      rank=0;
      GT=IMatrix::diagMatrix(prows,prows,1);
      return;
    }
    rank=IntegerTricks::decompAVeqBF(p,V.p,prows,pcols,false);
    T();
    IntegerTricks::decompAVeqBF(p,GT.p,prows,pcols,true);
    GT.T();
    T();
  }
  
  FMatrix &T(){
    FMatrix temp=Td();
    swapAll(temp);
    return *this;
  }

  FMatrix Td()const{
    FMatrix res(pcols,prows);
    for(long r=0;r<prows;r++){
      for(long c=0;c<pcols;c++){
	res.p[c+pcols*r]=p[r+prows*c];
      }
    }
    return res;
  };

  FMatrix inverse()const{Fractional idet;return inverse(idet);};
  FMatrix inverse(Fractional &idet)const;

  FMatrix diagInverseT(long rank)const{
    //invert diagonal up to rank and transpose
    FMatrix res(Td());
    long n=MyMath::min3<long>(rank,res.prows,res.pcols);
    for(long r=0;r<n;r++){
      res.p[r+res.prows*r].inv();
    }
    return res;
  }

  Fractional determinant()const{
    if(prows!=3|| pcols!=3)
      throw Except("FMatrix::determinant: only 3x3 implemented!");
    return 
      p[0]*(p[4]*p[8]-p[5]*p[7])+
      p[1]*(p[5]*p[6]-p[3]*p[8])+
      p[2]*(p[3]*p[7]-p[4]*p[6]);
  }
  
  /* ***************************
   * utilities
   **************************** */


  String matInfoS()const{
    return Sprintf("FMatrix: dimensions: %ld x %ld",prows,pcols);
  }

  String printS(long padding=6,const char *paren="",
		const char *delim=" ",bool transposevectors=false)const{
    long rows=prows;
    long cols=pcols;
    if(transposevectors && cols==1 && rows>1){
      cols=prows;
      rows=pcols;
    }
    String s="";
    for(long i=0;i<rows;i++){
      if(paren[0]!='\0') s+=paren[0];
      for(long j=0;j<cols;j++){
	s+=p[i+rows*j].printS().leftPadd(' ',padding);
	if(j<cols-1)
	  s+=delim;
      }
      if(paren[0]!='\0' && paren[1]!='\0') s+=paren[1];
      if(i<rows-1)s+="\n";
    }
    return s;
  }

protected:
private:
  void clean(){if(p)delete [] p; p=0;}
  void swapp(Fractional *&other){Fractional *c=p; p=other; other=c;}
  void swapAll(FMatrix &other){
    long d=prows;prows=other.prows;other.prows=d;
    d=pcols;pcols=other.pcols;other.pcols=d;
    Fractional *c=p; p=other.p; other.p=c;
  }
  void checkIndex(long r,long c)const{
    if(r<0||r>=prows) throw Except("FMatrix:: index out of bounds.");
    if(c<0||c>=pcols) throw Except("FMatrix:: index out of bounds.");
  };
private:
  long prows,pcols;
  Fractional *p;

  friend FMatrix operator*(const FMatrix &oi,const FMatrix &of);
  friend FMatrix operator*(const IMatrix &oi,const FMatrix &of);
  friend FMatrix operator+(const FMatrix &of,const IMatrix &oi);
  friend FMatrix operator*(const FMatrix &oi,const IMatrix &of);
  // Froebenius norm squared
  friend Fractional Fnorm2(const FMatrix &m);
  friend class IMatrix;

};


inline MyDebug operator<<(MyDebug dbg,const FMatrix &v){
  dbg<<v.printS(6,"||"," ",true);
  return dbg;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
FMatrix operator*(const IMatrix &oi,const FMatrix &of);

inline Fractional Fnorm2(const FMatrix &m)
{
  Fractional d=0;
  for(int i=0;i<m.prows*m.pcols;i++){
    d+=m.p[i]*m.p[i];
  }
  return d;
}

#endif /* _FMATRIX_HH */

