/*************************************************************************
 * @(#)HallSymbol.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   10 Aug 2012
 *************************************************************************
 */

#include "HallSymbol.hh"
#include "Except.hh"
#include "SeitzMatrix.hh"
#include "MyDebug.hh"
#include "SymmetryGroup.hh"
#include "Vec3.hh"

using namespace SpaceGroup;

const char *HallSymbol::latticeSymbols="pabcirstf";
const char *HallSymbol::translationSymbol="12345abcnuvwd";
const char *HallSymbol::axisSymbols="xyz'\"*.,;:";
const long HallSymbol::rotationpart[3][6][9]={
  { // x
    {1,0,0,0,1,0,0,0,1}, //1
    {1,0,0,0,-1,0,0,0,-1}, //2
    {1,0,0,0,0,1,0,-1,-1}, //3
    {1,0,0,0,0,1,0,-1,0}, //4
    {1,0,0,0,1,0,0,0,1}, //5 not used
    {1,0,0,0,1,1,0,-1,0} //6
  },
  { //y 
    {1,0,0,0,1,0,0,0,1}, //1
    {-1,0,0,0,1,0,0,0,-1}, //2
    {-1,0,-1,0,1,0,1,0,0}, //3
    {0,0,-1,0,1,0,1,0,0}, //4
    {1,0,0,0,1,0,0,0,1}, //5 not used
    {0,0,-1,0,1,0,1,0,1} //6
  },
  { //z
    {1,0,0,0,1,0,0,0,1}, //1
    {-1,0,0,0,-1,0,0,0,1}, //2
    {0,1,0,-1,-1,0,0,0,1}, //3
    {0,1,0,-1,0,0,0,0,1}, //4
    {1,0,0,0,1,0,0,0,1}, //5 not used
    {1,1,0,-1,0,0,0,0,1} //6
  },
};
const long HallSymbol::rotationpart2prime[][9]={
  {-1,0,0,0,0,-1,0,-1,0},
  {0,0,-1,0,-1,0,-1,0,0},
  {0,-1,0,-1,0,0,0,0,-1},
};
const long HallSymbol::rotationpart2dprime[][9]={
  {-1,0,0,0,0,1,0,1,0},
  {0,0,1,0,-1,0,1,0,0},
  {0,1,0,1,0,0,0,0,-1},
};

const long HallSymbol::rotationpart3star[9]={0,1,0,0,0,1,1,0,0};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
HallSymbol::HallSymbol(const String &symbol)
  :psymbol(symbol),pisvalid(false),pparsed(false),
   pcrystalsystem(Data::None),pbravais(Data::Simple),pislayergroup(false)
{
  // NSF("HallSymbol::HallSymbol");
  // myDebug()<<nsubfunc<<symbol;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
HallSymbol &HallSymbol::fromSymbol(const String &symbol)
{
  // NSF("HallSymbol::fromSymbol");
  // myDebug()<<nsubfunc<<symbol;
  pparsed=false;
  psymbol=symbol;
  pisvalid=false;
  pcrystalsystem=Data::None;
  pbravais=Data::Simple;
  pislayergroup=false;
  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
const ValueArray<SeitzMatrix> &HallSymbol::operations(bool analyse)
{
  if(!pparsed)parse(pops); 
  if(analyse){
    for(long i=0;i<pops.size();i++) pops[i].analyse(true);
  }
  
  return pops;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String HallSymbol::checkLatticeParameter(LatticeParameters &lp,bool adjust,
					 double tol)
{
  if(pislayergroup)lp.platticetype=SpaceGroup::Slab;
  else lp.platticetype=SpaceGroup::Solid;
  String warnings="";
  if(!pislayergroup &&(pcrystalsystem==Data::Cubic || pcrystalsystem==Data::Tetragonal ||
		       pcrystalsystem==Data::Hexagonal || pcrystalsystem==Data::Trigonal ||
		       pcrystalsystem==Data::Orthorhombic)){
    if(adjust)
      lp.pangles[0]=90;
    else 
      if(fabs(lp.pangles[0]-90)>tol) warnings+="Alpha should be 90 degree.\n";
  }

  if(!pislayergroup && (pcrystalsystem==Data::Cubic || pcrystalsystem==Data::Tetragonal ||
			pcrystalsystem==Data::Hexagonal|| pcrystalsystem==Data::Trigonal||
			pcrystalsystem==Data::Orthorhombic)){
    if(adjust)
      lp.pangles[1]=90;
    else
      if(fabs(lp.pangles[1]-90)>tol) warnings+="Beta should be 90 degree.\n";
  }

  if(pcrystalsystem==Data::Cubic || pcrystalsystem==Data::Tetragonal ||
     pcrystalsystem==Data::Orthorhombic){
    if(adjust)
      lp.pangles[2]=90;
    else
      if(fabs(lp.pangles[2]-90)>tol) warnings+="Gamma should be 90 degree.\n";
  }
  
  if(pcrystalsystem==Data::Cubic || pcrystalsystem==Data::Tetragonal || 
     pcrystalsystem==Data::Rhombohedral  || 
     pcrystalsystem==Data::Hexagonal || pcrystalsystem==Data::Trigonal){
    if(adjust)
      lp.pabc[1]=lp.pabc[0];
    else
      if(fabs(lp.pabc[0]-lp.pabc[1])>tol) warnings+="a0 and b0 should be equal. Will take a0.\n";
  }
  if(pcrystalsystem==Data::Cubic || pcrystalsystem==Data::Rhombohedral){
    if(adjust)
      lp.pabc[2]=lp.pabc[0];
    else
      if(fabs(lp.pabc[0]-lp.pabc[2])>tol) warnings+="a0 and c0 should be equal. Will take a0.\n";
  }
  
  if(pcrystalsystem==Data::Monoclinic){
    if(pmonoorientation==0){
      if(adjust){
	lp.pangles[1]=lp.pangles[2]=90;
      }else{
	if(!pislayergroup){
	  if(fabs(lp.pangles[1]-90)>tol) warnings+="Beta should be 90 degree.\n";
	}
	if(fabs(lp.pangles[2]-90)>tol) warnings+="Gamma should be 90 degree.\n";
      }
    }else if(pmonoorientation==1){
      if(adjust){
	lp.pangles[0]=lp.pangles[2]=90;
      }else{
	if(!pislayergroup){
	  if(fabs(lp.pangles[0]-90)>tol) warnings+="Alpha should be 90 degree.\n";
	}
	if(fabs(lp.pangles[2]-90)>tol) warnings+="Gamma should be 90 degree.\n";
      }
    }else if(pmonoorientation==2){
      if(adjust){
	lp.pangles[0]=lp.pangles[1]=90;
      }else{
	if(!pislayergroup){
	  if(fabs(lp.pangles[0]-90)>tol) warnings+="Alpha should be 90 degree.\n";
	  if(fabs(lp.pangles[1]-90)>tol) warnings+="Beta should be 90 degree.\n";
	}
      }
    }
  }

  if(pcrystalsystem==Data::Rhombohedral){ 
    if(adjust){
      lp.pangles[1]=lp.pangles[2]=lp.pangles[0];
    }else{
      if(fabs(lp.pangles[0]-lp.pangles[1])>tol) 
	warnings+="Beta and Alpha should be equal. Will take alpha.\n";
      if(fabs(lp.pangles[0]-lp.pangles[2])>tol) 
	warnings+="Gamma and Alpha should be equal. Will take alpha.\n";
    }
  }
  if(pcrystalsystem==Data::Hexagonal || pcrystalsystem==Data::Trigonal){
    if(adjust)
      lp.pangles[2]=120;
    else
      if(fabs(lp.pangles[2]-120)>tol) warnings+="Gamma should be 120 degree.\n";
  }



  return warnings;
}

  
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void HallSymbol::parse(ValueArray<SeitzMatrix> &ops)
{
  NSF("void HallSymbol::parse()");
  bool centrosym=false;

  //myDebug()<<"\n\n"<<nsubfunc;
  //myDebug()<<"\tsymbol"<<psymbol;

  String ss=psymbol; 
  ss.stripBlanks().replace("_"," ",0,true);
  String ss0(ss);
  ss.toLower();
  //myDebug()<<"\ts"<<s;
  long lop=-1;
  FMatrix trans(3,1);
  bool havetrans=false;
  if((lop=ss.find("("))>=0){
    if(!parseTranslationPart(ss.subString(lop),trans)){
      //failure to parse V means invalid symbol
      throw Except(Sprintf("Failure in parsing shift part: %s",
			   ss.subString(lop)()));
    }else{
      havetrans=true;
      ss[lop]='\0';
      ss0[lop]='\0';
    }
  }
  //myDebug()<<"after trans s="<<s;
  StringList list0=ss0.split(" ");
  StringList list=ss.split(" ");
  long Llist=list.size();

  //myDebug()<<"list"<<list;

  long icurrent=0;
  if(list[0][icurrent]=='-'){
    centrosym=true;
    icurrent++;
  }
  long ilatsymbol=-1;
  pislayergroup=false;
  for(size_t i=0;i<strlen(latticeSymbols);i++){
    if(list[0][icurrent]==latticeSymbols[i]){
      if(list0[0][icurrent]==list[0][icurrent]) pislayergroup=true;
      ilatsymbol=i;
      icurrent++;
      break;
    }
  }
  if(ilatsymbol<0 || list[0][icurrent]!='\0'){ 
    throw Except("Cannot parse lattice symbol!"); 
  }


  //myDebug()<<"nSeitz:"<<Llist;
  pSeitzSymbol *seitzsymbols=new pSeitzSymbol[Llist];
  ValueArray<SeitzMatrix> gen;
  gen.resize(Llist-1);
  for(int is=1;is<Llist;is++){
    //myDebug()<<"Seitz symbol:"<<list[is];
    if(!parseSeitzSymbol(list[is],seitzsymbols,gen[is-1],is)){
      throw Except(Sprintf("Failure in parsing Seitz symbol: %s",list[is]()));
    }
  }
  if(pcrystalsystem==Data::None) pcrystalsystem=Data::Triclinic;
  //myDebug()<<"make centro:";
  if(centrosym) {
    long N=gen.size();
    for(int i=0;i<N;i++){
      gen.append(SeitzMatrix(-1)*gen.at(i));
    }
  }
    
  if(havetrans) {
    SeitzMatrix t(-trans);
    long N=gen.size();
    for(int i=0;i<N;i++){
      gen[i].apply(t);
    }
  }  


  for(long i=0;i<gen.size();i++){ gen[i].regularize(); }


  //myDebug()<<"\tV="<<trans.Td();
  //myDebug()<<"\tcentrosym"<<centrosym;
  //myDebug()<<"\tlatticesymbol"<<latticeSymbols[ilatsymbol];
  //if(havetrans) myDebug()<<"\tV="<<trans.Td();
  
  
  // for(int i=0;i<gen.size();i++){
  //   myDebug()<<"i="<<i<<" gen="<<gen.at(i).printWwS();
  // }
  
  
  ops.clear();
  ops.append(1);
  
  
  
  
  
  
  //myDebug()<<"groupAddOperation:";
  for(int i=0;i<gen.size();i++){
    SeitzMatrix::groupAddOperation(ops,gen.at(i));
  }
  
  
  //todo  for(long i=0;i<ops.size();i++){ ops[i].regularize(); }
  switch (latticeSymbols[ilatsymbol]){
  case 'a':{
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(0,1,1,2)));
    pbravais=Data::ABaseCentered;    break;
  }
  case 'b':{
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(1,0,1,2)));
    pbravais=Data::BBaseCentered;break;
  }
  case 'c':{
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(1,1,0,2)));
    pbravais=Data::CBaseCentered;    break;
  }
  case 'i':{
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(1,1,1,2)));
    pbravais=Data::BodyCentered;    break;
  }
  case 'f':{
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(0,1,1,2)));
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(1,0,1,2)));
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(1,1,0,2)));
    pbravais=Data::FaceCentered;    break;
  }
  case 'r':{
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(2,1,1,3)));
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(1,2,2,3)));
    pbravais=Data::BRhombohedral;    break;
  }
  case 's':{
    throw Except("unusual S-setting for rhombohedral lattice not implemented!");
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(1,1,2,3)));
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(2,2,1,3)));
    pbravais=Data::BRhombohedral;//todo
    break;
  }
  case 't':{
    throw Except("unusual T-setting for rhombohedral lattice not implemented!");
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(1,2,1,3)));
    SeitzMatrix::groupAddOperation(ops,SeitzMatrix(FMatrix::vector3(2,1,2,3)));
    pbravais=Data::BRhombohedral;//todo
    break;
  }
  }
  
  //   myDebug()<<"final:";
  //   for(int i=0;i<ops.size();i++){
  //     myDebug()<<"i="<<i<<" ops="<<ops.at(i).toXYZSymbol()<<"\n"
  // 	     <<ops.at(i).printS();
  //   }
  for(int i=0;i<ops.size();i++){
    if(!ops.at(i).isInteger())
      throw Except("SeitzMatrix: Rotation matrix is not integer?");
  }
  delete [] seitzsymbols;
  
  pparsed=true;

  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool HallSymbol::parseTranslationPart(const String &V,FMatrix &trans)
{
  NSF("bool HallSymbol::parseTranslationPart(const String &V)");
  String s=V;
  //myDebug()<<nsubfunc<<s;
  long L=s.length();
  if(s[0]!='(')return false;
  s[0]=' ';
  if(s[L-1]!=')')return false;
  s[L-1]='\0';
  if(s.scan("()")>=0)return false;
  s.stripBlanks();
  StringList sl=s.split(" ,");
  //myDebug()<<"sl"<<sl;
  if(sl.size()!=3) return false;
  try{ trans(0,0)=sl[0].convertToLong(); }catch(...){ return false; }
  try{ trans(1,0)=sl[1].convertToLong(); }catch(...){ return false; }
  try{ trans(2,0)=sl[2].convertToLong(); }catch(...){ return false; }
  trans/=12;
  return true;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool HallSymbol::parseSeitzSymbol(const String &ss,pSeitzSymbol *seitzsym,
				  SeitzMatrix &gen,long noOfOp)
{
  NSF("bool HallSymbol::parseSeitzSymbol(const String &ss)");
  //myDebug()<<nsubfunc<<ss;
  
  long index=0;
  if(ss[index]=='-'){ 
    seitzsym[noOfOp].improper=true;
    index++;
  }
  long &norder=seitzsym[noOfOp].norder;
  if(!(String::islong(&norder,ss[index])&& 
       1<=norder && norder<=6 && norder!=5)){ return false; }
  index++;

  char &axissymbol=seitzsym[noOfOp].axissymbol;
  for(size_t i=0;i<strlen(axisSymbols);i++){
    if(ss[index]==axisSymbols[i]){
      axissymbol=axisSymbols[i];
      index++;
      break;
    }
  }
  if(axissymbol==',' || axissymbol=='.') axissymbol='\'';
  if(axissymbol==';' || axissymbol==':') axissymbol='"';

  //if no axis symbol: it is implied   
  if(axissymbol=='\0'){
    if(noOfOp==1){
      axissymbol='z';
    }else if(noOfOp==2){
      if(norder==2){
	long &precorder=seitzsym[noOfOp-1].norder;
	if(precorder==2 || precorder==4){
	  axissymbol='x';
	}else if(precorder %3 ==0){
	  axissymbol='\'';
	}
      }
    }else if(noOfOp==3){
      if(norder==3) axissymbol='*';
    }
  }
  
  ValueArray<char> &transsymbols=seitzsym[noOfOp].transsymbols;
  for(;index<ss.length();index++){
    bool hav=false;
    for(size_t i=0;i<strlen(translationSymbol);i++){
      //myDebug()<<ss[index]<<translationSymbol[i];
      if(ss[index]==translationSymbol[i]){
	transsymbols<<translationSymbol[i];
	hav=true;
	break;
      }
    }
    if(!hav)return false;
  }
  
  //myDebug()<<"improper="<<seitzsym[noOfOp].improper;
  // myDebug()<<"norder,axis="<<noOfOp<<":"<<(seitzsym[noOfOp].improper?'-':'+')
  // 	   <<seitzsym[noOfOp].norder<<seitzsym[noOfOp].axissymbol;
  //myDebug()<<"trans="<<seitzsym[noOfOp].transsymbols<<"\n";

  Fractional fi=(seitzsym[noOfOp].improper?-1:1);

  bool have21=false;
  bool have22=false;
  bool have3=false;
  bool have4=false;
  bool have6=false;
  if(axissymbol=='x' ||axissymbol=='y' || axissymbol=='z'){
    int ia;
    if(axissymbol=='x') ia=0;
    else if(axissymbol=='y') ia=1;
    else if(axissymbol=='z') ia=2;
    else return false;
    gen=FMatrix(3,3,rotationpart[ia][seitzsym[noOfOp].norder-1])*fi;
    if(seitzsym[noOfOp].norder==3) have3=true;
    if(seitzsym[noOfOp].norder==4) have4=true;
    if(seitzsym[noOfOp].norder==6) have6=true;
    if(seitzsym[noOfOp].norder==2){
      if( noOfOp==1) have21=true;
      if( noOfOp==2) have22=true;
      pmonoorientation=ia;
    }
  }else if(norder==2 && axissymbol=='\''){
    if(noOfOp<2) return false;
    char &precaxissymbol=seitzsym[noOfOp-1].axissymbol;
    int ia;
    if(precaxissymbol=='x') ia=0;
    else if(precaxissymbol=='y') ia=1;
    else if(precaxissymbol=='z') ia=2;
    else if(precaxissymbol=='*') ia=2;
    else return false;
    gen=FMatrix(3,3,rotationpart2prime[ia])*fi;
  }else if(norder==2 && axissymbol=='\"'){
    if(noOfOp<2) return false;
    char &precaxissymbol=seitzsym[noOfOp-1].axissymbol;
    int ia;
    if(precaxissymbol=='x') ia=0;
    else if(precaxissymbol=='y') ia=1;
    else if(precaxissymbol=='z') ia=2;
    else if(precaxissymbol=='*') ia=2;
    else return false;
    gen=FMatrix(3,3,rotationpart2dprime[ia])*fi;
  }else if(norder==3 && axissymbol=='*'){
    gen=FMatrix(3,3,rotationpart3star)*fi;
    if(noOfOp==1)
      pcrystalsystem=Data::Rhombohedral;
    else
      pcrystalsystem=Data::Cubic;
  }else if(norder==1){
    gen=FMatrix::diagMatrix(3,3,1)*fi;
  }else{
    throw Except(Sprintf("Unkown combination of order %ld and axis %c",
			 seitzsym[noOfOp].norder,seitzsym[noOfOp].axissymbol));
  }
  //myDebug()<<crystalSystemName();
  
  if(have21) pcrystalsystem=Data::Monoclinic;
  if(have22 && pcrystalsystem==Data::Monoclinic) pcrystalsystem=Data::Orthorhombic;
  if(have3 && pcrystalsystem==Data::None) pcrystalsystem=Data::Trigonal;
  if(have4 && pcrystalsystem!=Data::Cubic) pcrystalsystem=Data::Tetragonal;
  if(have6) pcrystalsystem=Data::Hexagonal;

  //myDebug()<<"gen"<<gen;
  //myDebug()<<"trans"<<transsymbols;
  for(long i=0;i<transsymbols.size();i++){
    pmonotranss+=transsymbols.at(i);
    if(transsymbols.at(i)=='a') gen+=FMatrix::vector3(1,0,0,2);
    if(transsymbols.at(i)=='b') gen+=FMatrix::vector3(0,1,0,2);
    if(transsymbols.at(i)=='c') gen+=FMatrix::vector3(0,0,1,2);
    if(transsymbols.at(i)=='n') gen+=FMatrix::vector3(1,1,1,2);
    if(transsymbols.at(i)=='u') gen+=FMatrix::vector3(1,0,0,4);
    if(transsymbols.at(i)=='v') gen+=FMatrix::vector3(0,1,0,4);
    if(transsymbols.at(i)=='w') gen+=FMatrix::vector3(0,0,1,4);
    if(transsymbols.at(i)=='d') gen+=FMatrix::vector3(1,1,1,4);
    long itr=0;
    switch (transsymbols.at(i)){
    case '1': itr=1;break;
    case '2': itr=2;break;
    case '3': itr=3;break;
    case '4': itr=4;break;
    case '5': itr=5;break;
    }
    if((12%norder)!=0) return false;
    if(itr){
      if(axissymbol=='x')      gen+=FMatrix::vector3(itr,0,0,norder);
      else if(axissymbol=='y') gen+=FMatrix::vector3(0,itr,0,norder);
      else if(axissymbol=='z') gen+=FMatrix::vector3(0,0,itr,norder);
      else
	return false;
    }
  }
  //myDebug()<<"gen"<<gen;

  return true;
}

/*EOF*/ 
