/* @(#)IMatrix.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _IMATRIX_HH
#define _IMATRIX_HH 1

#include "IntegerTricks.hh"

/* ***********************************************************************
 *
 * ***********************************************************************
 */


class FMatrix;

class IMatrix{
public:

  /* ***************************
   * construction
   **************************** */

  IMatrix():prows(0),pcols(0),p(0){}

  IMatrix(long rows,long cols):p(0){
    setDimenions(rows,cols);}

  IMatrix(long rows,long cols,long *data):p(0){
    prows=rows;pcols=cols;
    long n=prows*pcols;
    if(n>0){
      p=new long[prows*pcols];
      for(long i=-1;++i<n;)
	p[i]=data[i];
    }
  }

  IMatrix(long rows,long cols,double *data,double tol=1.0e-6):p(0){
    prows=rows;pcols=cols;
    long n=prows*pcols;
    if(n>0){
      p=new long[prows*pcols];
      for(long i=-1;++i<n;){
	p[i]=lRound(data[i]);
	if(fabs(p[i]-data[i])>tol){
	  throw Except("IntegerTricks::enlarged_lattice_sites:"
		       " input matrix U is not integer\n");
	}
      }
    }
  }


  
  IMatrix(const IMatrix &other)
    :prows(other.prows),pcols(other.pcols),p(0){
    setDimenions(prows,pcols);
    if(p && other.p) memcpy(p,other.p,sizeof(long)*prows*pcols);
  }

  IMatrix(const IMatrix &o1,const IMatrix &o2):prows(0),pcols(0),p(0){
    addMatrix(o1);
    addMatrix(o2);
  }
  IMatrix(const IMatrix &o1,const IMatrix &o2,const IMatrix &o3)
    :prows(0),pcols(0),p(0){
    addMatrix(o1);
    addMatrix(o2);
    addMatrix(o3);
  }
  
  ~IMatrix(){clean();}

  static IMatrix diagMatrix(long rows,long cols,long d){
    IMatrix pp(rows,cols);
    for(long r=0;r<MyMath::min<long>(pp.prows,pp.pcols);r++) pp.p[r+pp.prows*r]=d;
    return pp;
  }

  static IMatrix P(long rows,long cols,long rank){
    IMatrix pp(rows,cols);
    for(long r=0;r<MyMath::min3<long>(pp.prows,pp.pcols,rank);r++) pp.p[r+pp.prows*r]=1;
    return pp;
  }
  static IMatrix Q(long rows,long cols,long rank){
    IMatrix pp(rows,cols);
    long c,r;
    //myDebug()<<"C,R"<<pcols<<prows;
    for(c=pp.pcols-1,r=pp.prows-1; 
	r>=MyMath::max<long>(0,pp.prows-rank) && c>=
	  MyMath::max<long>(0,pp.pcols-rank);c--,r--) {
      //myDebug()<<"c,r"<<c<<r;
      pp.p[r+pp.prows*c]=1;
    }
    return pp;
  }

  static IMatrix vector3(long v0,long v1,long v2){
    IMatrix v(3,1);
    v.setColumn3(0,v0,v1,v2);
    return v;
  };

  IMatrix &operator=(const IMatrix &other){
    setDimenions(other.prows,other.pcols);
    if(p && other.p) 
      for(long i=0;i<prows*pcols;i++) p[i]=other.p[i];
    return *this;
  }
  IMatrix &operator=(long o){
    for(long i=0;i<prows*pcols;i++) p[i]=o;
    return *this;
  }


  /* ***************************
   * restructuring
   **************************** */

  void setDimenions(long rows,long cols){
    //destroys the array
    clean();
    prows=rows;pcols=cols;
    long n=prows*pcols;
    if(n>0){
      p=new long[prows*pcols];
      memset(p,0,sizeof(long)*prows*pcols);
    }
  }


  void addRows(long n){
    if(n<1) return;
    if(!p){
      setDimenions(n,pcols);
      return;
    }
    //myDebug()<<"addRows"<<n<<matInfoS();
    long prnew=(prows+n);
    long *temp=new long[prnew*pcols];
    for(long c=0;c<pcols;c++){
      memcpy(temp+prnew*c,p+prows*c,sizeof(long)*prows);
      memset(temp+prows+prnew*c,0,sizeof(long)*n);
    }
    swapp(temp);
    delete [] temp;
    prows=prnew;
  }

  void addColumns(long n){
    if(n<1) return;
    if(!p){
      setDimenions(prows,n);
      return;
    }
    //myDebug()<<"addColumnms"<<matInfoS();
    long *temp=new long[prows*(pcols+n)];
    memcpy(temp,p,sizeof(long)*prows*pcols);

    memset(temp+prows*pcols,0,sizeof(long)*prows*n);
    swapp(temp);
    delete [] temp;
    pcols+=n;
    //myDebug()<<"addColumnms"<<matInfoS();
  }

  IMatrix &addMatrix(const IMatrix &o){
    // new matrix has the first columns from this and o columns appended
    // rows are adjusted
    long opc=pcols;
    addColumns(o.pcols);
    addRows(o.prows-prows);
    for(long r=0;r<o.prows;r++){
      for(long c=0;c<o.pcols;c++){
	p[r+prows*(c+opc)]=o.p[r+o.prows*c];
      }
    }
    return *this;
  }

  IMatrix fromColumn(long c)const
  {
    IMatrix m(prows,1);
    memcpy(m.p,p+prows*c,sizeof(long)*prows);
    return m;
  }

  IMatrix fromRow(long r)const
  {
    IMatrix m(pcols,1);
    for(long c=0;c<pcols;c++){
      m.p[c]=p[r+prows*c];
    }
    return m;
  }

  IMatrix &cycleColumnsL(){
    addMatrix(fromColumn(0));
    removeColumns(0,0);
    return *this;
  }
  IMatrix &swapColumns(long c1,long c2){
    for(long r=0;r<prows;r++){
      long d=p[r+prows*c1];
      p[r+prows*c1]=p[r+prows*c2];
      p[r+prows*c2]=d;
    }
    return *this;
  }


  /* ***************************
   * access
   **************************** */

  long rows()const{return prows;}
  long columns()const{return pcols;}

  const long *data()const{return p;}

  long &operator()(long r,long c){
    checkIndex(r,c);
    return p[r+prows*c];
  }
  long operator()(long r,long c)const{
    checkIndex(r,c);
    return p[r+prows*c];
  }
  long operator[](long r)const{
    if(pcols!=1)throw Except("IMatrix::operator[] needs pcols==1!");
    checkIndex(r,0);
    return p[r];
  }

  void setColumn3(long c,long r0,long r1,long r2){
    if(prows<3)throw Except("IMatrix::setColumn3 less than 3 rows.");
    if(c+1>pcols)throw Except(Sprintf("IMatrix::setColumn "
				    "less than %ld columns.",c));

    p[0+prows*c]=r0;
    p[1+prows*c]=r1;
    p[2+prows*c]=r2;
  }

  void addColumn3(long r0,long r1,long r2){
    if(prows<3)throw Except("IMatrix::addColumn3 less than 3 rows.");

    addColumns(1);
    
    p[0+prows*(pcols-1)]=r0;
    p[1+prows*(pcols-1)]=r1;
    p[2+prows*(pcols-1)]=r2;
  }

  IMatrix &scaleColumn(long c,long s){
    checkIndex(0,c);
    for(long r=0;r<prows;r++) p[r+prows*c]*=s;
    return *this;
  }

  IMatrix &removeColumns(long from,long to){
    if(to<from) return *this;
    if(from>=pcols) return *this;
    long lto=MyMath::min<long>(to,pcols-1);
    //myDebug()<<"removeColumns: "<<from<<lto;
    long pcnew=pcols-(lto-from+1);
    //myDebug()<<"removeColumns: pcnew"<<pcnew;
    if(pcnew<1){
      pcols=0;
      clean();
      return *this;
    }
    long *temp=new long[prows*pcnew];
    if(from>0)
      memcpy(temp,p,sizeof(long)*prows*(from));

    memcpy(temp+prows*from,p+prows*(lto+1),sizeof(long)*prows*(pcols-lto-1));

    swapp(temp);
    pcols=pcnew;
    delete [] temp;
    return *this;
  }
  IMatrix &removeRows(long from,long to){
    if(to<from) return *this;
    if(from>=prows) return *this;
    long lto=MyMath::min<long>(to,prows-1);
    //myDebug()<<"removeRows: "<<from<<lto<<prows;
    long prnew=prows-(lto-from+1);
    //myDebug()<<"removeRows: prnew"<<prnew;
    if(prnew<1){
      prows=0;
      clean();
      return *this;
    }
    long *temp=new long[prnew*pcols];

    if(from>0){
      for(long c=0;c<pcols;c++) 
	for(long r=0;r<from;r++) temp[r+prnew*c]=p[r+prows*c];
    }

    for(long c=0;c<pcols;c++) 
      for(long r=lto+1;r<prows;r++) 
	temp[r-lto-1+from+prnew*c]=p[r+prows*c];


    swapp(temp);
    prows=prnew;
    delete [] temp;
    return *this;
  }

  /* ***************************
   * nummerics
   **************************** */

  bool containsColumn(const IMatrix &o)const{
    if(o.pcols!=1)
      throw Except("IMatrix:: containsColumn(): argument is not a vector!");
    if(prows!=o.prows)return false;
    //myDebug()<<"o="<<o.fromColumn(0).Td().printS();
	  
    for(long c=0;c<pcols;c++){
      //myDebug()<<"col"<<c<<this->fromColumn(c).Td().printS();

      bool eq=true;
      for(long r=0;r<prows;r++){
	if(p[r+prows*c]!=o.p[r]){eq=false;break;}
      }
      //myDebug()<<"eq"<<eq;
      if(eq) return true;
    }
    return false;
  }
  
  IMatrix &operator-=(const IMatrix &o){
    if(prows!=o.prows || pcols!=o.pcols)
      throw Except("IMatrix:: operator-=(): different dimensions");
    for(long i=0;i<prows*pcols;i++){
      p[i]-=o.p[i];
    }
    return *this;
  }

  IMatrix &operator/=(long o){
    for(long i=0;i<prows*pcols;i++){
      p[i]/=o;
    }
    return *this;
  }
  IMatrix &operator*=(long o){
    for(long i=0;i<prows*pcols;i++){
      p[i]*=o;
    }
    return *this;
  }

  IMatrix operator-()const{
    IMatrix res(*this);
    for(long i=0;i<prows*pcols;i++){
      res.p[i]=-res.p[i];
    }
    return res;
  }

  IMatrix operator*(long d)const{
    IMatrix res(prows,pcols);
    for(long i=0;i<pcols*prows;i++){
      res.p[i]=p[i]*d;
    }
    return res;
  }
  IMatrix operator-(const IMatrix &o)const{
    if(prows!=o.prows || pcols!=o.pcols)
      throw Except("IMatrix:: operator-(): different dimensions");
    IMatrix res(*this);
    for(long i=0;i<prows*pcols;i++){
      res.p[i]-=o.p[i];
    }
    return res;
  }
  IMatrix operator+(const IMatrix &o)const{
    if(prows!=o.prows || pcols!=o.pcols)
      throw Except("IMatrix:: operator+(): different dimensions");
    IMatrix res(*this);
    for(long i=0;i<prows*pcols;i++){
      res.p[i]+=o.p[i];
    }
    return res;
  }

  bool operator==(const FMatrix &o)const;

  bool operator!=(long o)const{
    for(long i=0;i<pcols*prows;i++){
      if(p[i]!=o)return true;
    }
    return false;
  }

  bool operator!=(const IMatrix &o)const{
    if(prows!=o.prows || pcols!=o.pcols) return true;
    for(long i=0;i<pcols*prows;i++){
      if(p[i]!=o.p[i])return true;
    }
    return false;
  }

  // shortest possible integer vector in this direction
  IMatrix &shorten(){
    long D=0;
    for(long c=0;c<pcols;c++){
      for(long r=0;r<prows;r++){
	D=MyMath::gcd(D,p[r+prows*c]);
      }
    }
    (*this)/=D;
    return *this;
  }
  IMatrix &shortenColumns(){
    long N=prows*pcols;
    for(long indc=0;indc<N;indc+=prows){
      long *pp=p+indc;
      long D=0;
      for(long r=0;r<prows;r++){
	D=MyMath::gcd(D,pp[r]);
      }
      for(long r=0;r<prows;r++){
	pp[r]/=D;
      }
    }
    return *this;
  }

  IMatrix decompAVeqB(long &rank,bool detnotone=false){
    // Decompose R such that R*V=(B 0), where B forms a basis of R
    // Return V and B in R
    // The rank of B returned in rank.
    // The determinant of V is 1 if !detnotone.
    // If detnotone detV>1 and B diagonal.
    IMatrix V(pcols,pcols);
    rank=IntegerTricks::decompAVeqB(p,V.p,prows,pcols,detnotone);
    //removeColumns(rank,pcols-1);
    //if(detnotone && pcols<prows) addColumns(prows-pcols);
    //T();
    return V;
  }
  
  void  decompGTAVeqD(long &rank,IMatrix &V,IMatrix &GT){
    // Decompose R such that R*V=(B 0), where B forms a basis of R followed by
    // G^T*B=D. Hence G^T*R*V=D. Dimensions are 
    //     R:rows x cols
    //     V:cols x cols
    //     B:rows x cols (effectivly rows x rank)
    //     GT: rows x rows
    // Return V and B in R
    // The rank of B returned in rank.
    // The determinant of V is 1 if !detnotone.
    // If detnotone detV>1 and B diagonal.
    V.setDimenions(pcols,pcols);
    GT.setDimenions(prows,prows);
    if(pcols==0){
      rank=0;
      GT=IMatrix::diagMatrix(prows,prows,1);
      return;
    }
    rank=IntegerTricks::decompAVeqB(p,V.p,prows,pcols,false);
    T();
    IntegerTricks::decompAVeqB(p,GT.p,prows,pcols,true);
    GT.T();
    T();
  }
  
  IMatrix &T(){
    IMatrix temp=Td();
    swapAll(temp);
    return *this;
  }

  IMatrix Td()const{
    IMatrix res(pcols,prows);
    for(long c=0;c<pcols;c++){
      for(long r=0;r<prows;r++){
	res.p[c+pcols*r]=p[r+prows*c];
      }
    }
    return res;
  };

  IMatrix inversetV(long &idet)const;
  IMatrix inverseTtV(long &idet){
    return inversetV(idet).Td();
  }

  FMatrix diagInverseT(long rank)const;


  long determinant() const 
  { 
    if(prows!=3|| pcols!=3)
      throw Except("IMatrix::determinant: only 3x3 implemented!");
    return (fromColumn(0).cross(fromColumn(1))).dot(fromColumn(2)); 
  }

  long trace(){
    if(prows!=pcols)
      throw Except("IMatrix::trace: only square matrices implemented!");
    long d=0;
    for(long r=0;r<prows;r++){
      d+=p[r+prows*r];
    }
    return d;
  }

  long dot(const IMatrix &other)const
  {
    if(pcols!=1 || other.pcols!=1 ||
       prows!=other.prows)
      throw Except("IMatrix::dot: incompatible dimensions!");

    long d=0;
    for(long r=0;r<prows;r++){
      d+=p[r]*other.p[r];
    }
    return d;
    
  }

  Fractional dot(const FMatrix &other)const;


  // vector product 
  IMatrix cross(const IMatrix &o) const{
    if(pcols!=1|| o.pcols!=1 ||
       prows!=3|| o.prows!=3)
      throw Except("IMatrix::cross: cross product only for 3x1 matrices!");
    IMatrix res(3,1);
    res.setColumn3(0,p[1]*o.p[2]-p[2]*o.p[1],
		   p[2]*o.p[0]-p[0]*o.p[2],
		   p[0]*o.p[1]-p[1]*o.p[0]);
    return res;
  }

  /* ***************************
   * utilities
   **************************** */


  String matInfoS()const{
    return Sprintf("IMatrix: dimensions: %ld x %ld",prows,pcols);
  }

  String printS(const char *fmt="%5ld",bool transposevectors=false)const{
    String s="";
    long rows=prows;
    long cols=pcols;
    if(transposevectors && cols==1 && rows>1){
      cols=prows;
      rows=pcols;
    }
    for(long i=0;i<rows;i++){
      for(long j=0;j<cols;j++){
	s+=Sprintf(fmt,p[i+rows*j])+(j<cols-1?" ":"");
      }
      if(i<rows-1)s+="\n";
    }
    return s;
  }

  
  

protected:
private:

  void clean(){if(p)delete [] p; p=0;}
  void swapp(long *&other){long *c=p; p=other; other=c;}
  void swapAll(IMatrix &other){
    long d=prows;prows=other.prows;other.prows=d;
    d=pcols;pcols=other.pcols;other.pcols=d;
    long *c=p; p=other.p; other.p=c;
  }
  void checkIndex(long r,long c)const{
    if(r<0||r>=prows) throw Except("IMatrix:: index out of bounds.");
    if(c<0||c>=pcols) throw Except("IMatrix:: index out of bounds.");
  };

private:
  long prows,pcols;
  long *p;

  friend class FMatrix;
  friend FMatrix operator*(const IMatrix &oi,const FMatrix &of);
  friend FMatrix operator*(const FMatrix &oi,const IMatrix &of);
  friend FMatrix operator+(const FMatrix &oi,const IMatrix &of);
  friend IMatrix operator*(const IMatrix &oi,const IMatrix &of);
  // Froebenius norm squared
  friend long Fnorm2(const IMatrix &m);
  friend long mhnorm(const IMatrix &v);

};



inline MyDebug operator<<(MyDebug dbg,const IMatrix &v){
  dbg<<v.printS("%5ld",true);
  return dbg;
}


IMatrix operator*(const IMatrix &oi,const IMatrix &of);

inline long Fnorm2(const IMatrix &m)
{
  long d=0;
  for(int i=0;i<m.prows*m.pcols;i++){
    d+=m.p[i]*m.p[i];
  }
  return d;
}

inline long mhnorm(const IMatrix &m){
  long d=0;
  for(int i=0;i<m.prows*m.pcols;i++){
    d+=labs(m.p[i]);
  }
  return d;
}

#endif /* _IMATRIX_HH */

