/* @(#)SeitzMatrix.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _SEITZMATRIX_HH
#define _SEITZMATRIX_HH 1

#include "FMatrix.hh"
#include "Except.hh"
#include "PeriodicHyperPlane.hh"
#include "MyDebug.hh"
#include "SpaceGroupData.hh"

namespace SpaceGroup{
  
  class MainAxisOperation;
  class HallSymbol;
  
  
  /* *********************************************************************
   * Cartesian basis is E^T=(\vec{e}_1,\vec{e}_2,\vec{e}_3)
   * with transformation law
   * \hat{D} E^T=E^T* R_D, with R_D being the representation matrix of \hat{D}.
   * Then a vector \vec{r}=E^T * \bar{r} with \bar{R}^T=(x,y,z)
   * transforms according to 
   * \hat{D} \vec{r}= E^T * R_D * \bar{r}
   * or \bar{r}'=R_D * \bar{r}.
   * The SeitzMatrix is {W|w} with W being some R_D and w a translation
   ***********************************************************************
   */
  
#define ERROR_NOTINTEGER					\
  throw Except("SeitzMatrix: Rotation matrix is not integer?");
  
  class SeitzMatrixInfo{
  public:
    SeitzMatrixInfo():pnorder(1),pm1(1),pp(1),ptau0(0),
		      pnonsymmorphic(true){
      Pmin.setDimenions(3,1);
      Gmin.setDimenions(3,1);
      RA.setDimenions(3,3);
      GA.setDimenions(3,3);
      Delta.setDimenions(3,3);
      pw00.setDimenions(3,1);
      porigin0.setDimenions(3,1);
    }

    // return the symbol with axis,location and such
    String symbol(bool sshort=false)const; 

    String mirrorSignature(const FMatrix &Bcenter,const FMatrix &w00,
			   const FMatrix &orig)const;
    String mirrorSignatures(const FMatrix &Bcenter)const;

    long skrew()const{return skrew(pnorder,ptau0,pm1);}
    
    static long skrew(long norder,const Fractional &tau0,long m){
      long n=labs(norder);
      if(m<0) return (((-tau0+n)/n).regularizeInterval(false)*n).toLong();
      else return tau0.toLong();
    }


    PeriodicHyperPlane toPHPlane(long tauallow,const FMatrix &delta=FMatrix(3,1),
				 bool betweenpairs=false)const{
      PeriodicHyperPlane pl(3);
      if(pnorder>1){
	pl.R=Pmin;
	pl.Delta.addMatrix(Delta.fromColumn(0));
	pl.Delta.addMatrix(Delta.fromColumn(1));
	for(int iip=0;iip<porlist.size();iip++){
	  //myDebug()<<"   origin: "<<iip;
	  Fractional tau0;
	  if(tauallow>=0){
	    tau0=(Gmin.dot(pw0list.at(iip))/pp)
	      .regularizeInterval(false)*labs(pnorder);
	    if(skrew(pnorder,tau0,pm1)!=tauallow) {
	      //myDebug()<<"skip tau0 :"<<tau0;
	      continue;
	    }
	  }
	  //myDebug()<<"take tau0 :"<<tau0;
	  pl.O.addMatrix(porlist.at(iip)+delta);
	}
      }else if(pnorder< -2){
	if(betweenpairs){
	  pl.R=Pmin;
	  pl.Delta.addMatrix(Delta.fromColumn(0));
	  pl.Delta.addMatrix(Delta.fromColumn(1));
	  for(int iip=0;iip<porlist.size();iip++){
	    pl.O.addMatrix(porlist.at(iip)+delta);
	  }
	}else{
	  pl.Delta=Delta;
	  for(int iip=0;iip<porlist.size();iip++){
	    pl.O.addMatrix(pw0list.at(iip)+delta);
	  }
	}
      }else if(pnorder==-1){
	pl.Delta=Delta;
	for(int iip=0;iip<porlist.size();iip++){
 	  pl.O.addMatrix(porlist.at(iip)+delta);
	}
      }else if(pnorder==-2){
	pl.R.addMatrix(RA.fromColumn(0));
	pl.R.addMatrix(RA.fromColumn(1));
	pl.Delta.addMatrix(Delta.fromColumn(0));
	for(int iip=0;iip<porlist.size();iip++){
	  //myDebug()<<"   origin: "<<iip;
	  if(tauallow==0){
	    if(FMatrix(pw0list.at(iip)).regularizeInterval(false)!=0){
	      //myDebug()<<"skip w0 :"<<pw0list.at(iip);
	      continue;
	    }
	  }else if(tauallow==3){
	    if(FMatrix(pw0list.at(iip)).regularizeInterval(false)!=
	       FMatrix::vector3(0,0,1,2)){
	      //myDebug()<<"skip w0 :"<<pw0list.at(iip);
	      continue;
	    }
	  }else if(tauallow==1){
	    if(FMatrix(pw0list.at(iip)).regularizeInterval(false)!=
	       FMatrix::vector3(1,0,0,2)){
	      //myDebug()<<"skip w0 :"<<pw0list.at(iip);
	      continue;
	    }
	  }
	  //myDebug()<<"take w0 :"<<pw0list.at(iip);
	  pl.O.addMatrix(porlist.at(iip)+delta);
	}
      }else{
	throw Except(" SeitzMatrixInfo::toPHPlane: "
		     "Unkown operation type.");
      }
      if(betweenpairs && pl.O.columns()>0){
	pl.normalize();
	FMatrix o(pl.O);
	pl.O.setDimenions(3,0);
	pl.O.addMatrix(o+pl.Delta.fromColumn(0)/2);
	pl.O.addMatrix(o+pl.Delta.fromColumn(1)/2);
	pl.O.addMatrix(o+pl.Delta.fromColumn(0)/2
		       +pl.Delta.fromColumn(1)/2);
      }
      return pl.normalize();
    }
    
    String locationSymbol(const FMatrix &w0,
			  const FMatrix &orig,bool sshort=false)const;

  private:
    // helper functions for symbol
    String addlocdir(long l,char c0,const Fractional &o,bool comma)const;
    String addlocdirshort(const IMatrix &axis)const;

  public:
    long pnorder; // order of operation: positive proper, negative improper
    long pm1; // the power of the rotation, in crystals +-1
    IMatrix Pmin; // real vector in axis direction;
    IMatrix Gmin; // real vector in axis direction;
    long pp; // Pmin*Gmin
    IMatrix RA; /* RA.column(0/1) are perp to axis and RA.column(3)
		 * completes the minimal cell
		 */
    IMatrix GA; /* GA.column(0/1) are perp to axis and GA.column(3)
		 * completes the minimal reciprocal cell
		 */
    FMatrix Delta; /* Delta.column(0) and Delta,.column(1) form
		      the axis-perp periodicity lattice of the 
		      axis, which appear due to tau+R.
		      For mirrors Delta(0) is the periodicity along
		      the axis. For rotoinversion the first two
		      describe the periodicity of the axis and Delta(2)
		      shows the periodicity of the inversion point on the 
		      axis
		   */
    FMatrix pw00; /* the essential nonsymmorphic part of w calculated such
		     that IT notations follow.
		     For rotoinversions it is the inversion center on the axis */
    FMatrix porigin0;   /* the origin of the operation calculated such
			   that IT notations follow */
    ValueArray<FMatrix> pw0list; /* a list of all resulting essential w 
				    usefull for analysis */
    ValueArray<FMatrix> porlist; /* a list of all resulting origins w*/
    Fractional ptau0;/* For rotations pw0 = Pmin * ptau0/pnorder.
		      * When taking the primitive centered lattice as basis, 
		      * the non-trivial rotation tau0 must be integer.
		      * If the centering vectors are taken as generators 
		      * and the basis is the conventional lattice the tau0
		      * of rotations about non principal  axes can  be 
		      * fractionals.
		      */
    bool pnonsymmorphic;
  };
  
  class SeitzMatrix{
  public:
    SeitzMatrix():pactive(true),pinteger(true),panalysed(false),
		  pW(FMatrix::diagMatrix(3,3,1)){ pw.setDimenions(3,1); }
    SeitzMatrix(const SeitzMatrix &o):pactive(o.pactive),pinteger(o.pinteger),
				      panalysed(o.panalysed),pW(o.pW),pw(o.pw),
				      pinfo(o.pinfo){
      if(!pW.isInteger()) ERROR_NOTINTEGER;
    }
    SeitzMatrix(const String &xyzsymbol);
    SeitzMatrix(long diag):pactive(true),pinteger(true),panalysed(false),
			   pW(FMatrix::diagMatrix(3,3,diag)){
      pw.setDimenions(3,1); };
    SeitzMatrix(const FMatrix &m,const FMatrix &t,bool mustbeint=true)
      :pactive(true),pinteger(mustbeint),panalysed(false),pW(m),pw(t){
      if(mustbeint && !pW.isInteger()) ERROR_NOTINTEGER;
    };
    SeitzMatrix(const FMatrix &m,bool mustbeint=true)
      :pactive(true),pinteger(mustbeint),panalysed(false){

      if(m.rows()==3 && m.columns()==3){
	pW=m;
	pw.setDimenions(3,1); 
	if(mustbeint && !pW.isInteger()) ERROR_NOTINTEGER;
      }
      if(m.rows()==3 && m.columns()==1){
	pW=FMatrix::diagMatrix(3,3,1);
	pw=m; 
	pw.regularizeInterval(false);

      }

    };
    
    
    SeitzMatrix &operator=(const SeitzMatrix &other){
      pactive=other.pactive,
      pinteger=other.pinteger;
      panalysed=other.panalysed;
      pW=other.pW;pw=other.pw;
      pinfo=other.pinfo;
      if(pinteger && !pW.isInteger()) ERROR_NOTINTEGER;
      return *this;
    }
    
    
    SeitzMatrix &operator+=(const FMatrix &m){
      panalysed=false;
      if(m.rows()==3 && m.columns()==1){
	pw+=m;
      }
      return *this;
    }
    
    const FMatrix &W()const{return pW;}
    
    
    bool operator==(const SeitzMatrix &other)const{
      return ((pW==other.pW) && (pw-other.pw).regularizeInterval(false)==0 );
    };
    
    SeitzMatrix &operator*=(const SeitzMatrix &other){
      pinteger=(pinteger && other.pinteger);
      pw+=pW*other.pw;
      pW=pW*other.pW;
      pw.regularizeInterval(false);
      if(pinteger && !pW.isInteger()) ERROR_NOTINTEGER;
      panalysed=false;
      return *this;
    };
    SeitzMatrix operator*(const SeitzMatrix &other)const{
      SeitzMatrix res;
      res.pinteger=pinteger && other.pinteger;
      res.pW=pW*other.pW;
      res.pw=pw+pW*other.pw;
      res.pw.regularizeInterval(false);
      if(pinteger && other.pinteger && !res.pW.isInteger()) ERROR_NOTINTEGER;
      res.panalysed=false;
      return res;
    };
    
    SeitzMatrix inverse()const{
      //{a|t}^{-1}={a^{-1}| -a^{-1}*t }  
      Fractional det;
      SeitzMatrix res(pW.inverse(det),pinteger);
      if(pinteger && abs(det)!=1) throw Except("SeitzMatrix abs(det)!=1?");
      res.pw=-(res.pW*pw);
      res.panalysed=false;
      return res;
    };
    
    SeitzMatrix &regularize(){
      pw.regularizeInterval(false);
      panalysed=false;
      return *this;
    }

    
    inline bool isInteger()const{return pW.isInteger();};
    
    // other^{-1}*this*other
    SeitzMatrix &apply(const SeitzMatrix &other){
      pinteger=(pinteger && other.pinteger);
      // myDebug()<<"apply: W=\n"<<this->printWwS();
      // myDebug()<<"apply: op=\n"<<other.printWwS();
      // myDebug()<<"apply: opI=\n"<<other.inverse().printWwS();
      // myDebug()<<"apply: W*op=\n"<<((*this)*other).printWwS();
      // myDebug()<<"apply: opI*W*op=\n"<<(other.inverse()*(*this)*other).printWwS();
      *this=other.inverse()*(*this)*other;
      if(pinteger && !pW.isInteger()) ERROR_NOTINTEGER;
      panalysed=false;
      return *this;
    }
    
    void changeBasis(const FMatrix &U,const FMatrix &Uinv){
      pW=Uinv*pW*U;
      if(!pW.isInteger()) 
	throw Except("SeitzMatrix: basis transformed rotation"
		     " matrix is not integer?");
      pw=Uinv*pw;
      pw.regularizeInterval(false);
      panalysed=false;
      
    }



    void changeOrigin(const FMatrix &neworigin){
      // in current origin gr=(pW|pw)r 
      // transformed r'=r-neworigin
      // gr'+neworigin=(pW|pw)(r'+neworigin) 
      // new operation is (pW|pw+(pW-1)*neworigin)
      // which is equivalent to the affine transformation
      //
      // (pW|pw)'=(1|-neworigin)*(pW|pw)*(1|neworigin)
      //
      //
      pw+=pW*neworigin-neworigin;
      pw.regularizeInterval(false);
      panalysed=false;
    }
    

    void analyse(bool prot=false);

    
    
    String toXYZSymbol()const{
      String res;
      for(int k=0;k<3;k++){
	bool haveone=false;
	if(pW(k,0)!=0){ res+=pW(k,0).printS("x",haveone);haveone=true; }
	if(pW(k,1)!=0){ res+=pW(k,1).printS("y",haveone);haveone=true; }
	if(pW(k,2)!=0){ res+=pW(k,2).printS("z",haveone);haveone=true; }
	if(pw[k]  !=0){ res+=pw[k].printS(haveone);haveone=true; }
	if(k<2)res+=",";
      }
      return res;
    }
    
    String symbol()const{
      if(!panalysed) return "not analysed";
      return pinfo.symbol();
    }
    
    String printWwS()const{
      return Sprintf("\n"
		     "%5s %5s %5s %5s\n"
		     "%5s %5s %5s %5s\n"
		     "%5s %5s %5s %5s\n"
		     "%5ld %5ld %5ld %5ld\n",
		     pW(0,0).printS()(),pW(0,1).printS()(),pW(0,2).printS()(),
		     pw[0].printS()(),
		     pW(1,0).printS()(),pW(1,1).printS()(),pW(1,2).printS()(),
		     pw[1].printS()(),
		     pW(2,0).printS()(),pW(2,1).printS()(),pW(2,2).printS()(),
		     pw[2].printS()(),
		     0L,0L,0L,1L
		     );
    };
    

    static void groupAddOperation(ValueArray<SeitzMatrix> &ops,
				  const SeitzMatrix &op);

    static void analyseGroup(HallSymbol &hallsymbolout,SeitzMatrix &Stransform,
			     String &spgrnum,
			     const ValueArray<SeitzMatrix> &opsin,
			     const String &requiredsetting,
			     bool testmix=false);

  private:

    
    // operations that act on all operations
    
    
    
    
    
    static void changeBasis(ValueArray<SeitzMatrix> &ops,
			    const FMatrix &U,
			    bool analyse,
			    bool prot=false);
    
    static void changeOrigin(ValueArray<SeitzMatrix> &ops,
			     const FMatrix neworigin,
			     bool analyse,
			     bool prot=false);
    
    static SeitzMatrix changeSettingOp(const String &spgrnum,
				       const String &oldsetting, 
				       const String &newsetting);

    static SeitzMatrix changeSetting(ValueArray<SeitzMatrix> &ops,
				     Data::BravaisLattice &pbravais,
				     Data::CrystalSystem &pcrystalsystem,
				     String &spgrnum,
				     FMatrix &Bcenter,
				     FMatrix &Bcenterinv,
				     const String &newsetting,bool analyse);

    FMatrix findHexPlanarBasisFromC3();
    FMatrix findTetPlanarBasisFromC4();

  private:
    bool pactive; // skip this ... makes algorithms faster
    bool pinteger; // must beinteger
    bool panalysed;
    FMatrix pW;
    FMatrix pw;
    SeitzMatrixInfo pinfo;
    friend class MainAxisOperation;
    friend class Tests;
  };
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class Tests{
  public:
    static SeitzMatrix mixIntSeitzRepresentation(ValueArray<SeitzMatrix> &ops,
						 bool mixlat,bool rotlat,
						 bool shift,bool shuffle);
  };
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

/* ***********************************************************************
 *
 * ***********************************************************************
 */

inline MyDebug operator<<(MyDebug dbg,const SpaceGroup::SeitzMatrix &s){
  dbg<<s.printWwS();
  return dbg;
}

#endif /* _SEITZMATRIX_HH */

