/*************************************************************************
 * @(#)BandWeights.cpp
 * author: magru@unas
 * date:   22 Nov 2013
 *************************************************************************
 */

#include "Assorted.hh"
#include "Except.hh"
#include "BandWeights.hh"
#include "MyDebug.hh"
#include "MachineTime.hh"
#include "MyMath.hh"
#include "StringPtr.hh"


/* ***********************************************************************
 *
 * ***********************************************************************
 */
BandWeights::BandWeights()
  :pheaderready(false),
   pnsite(0),pef(0.),pnk(0),pnvdim(0),pnband(0),pnspin(0),pilower(-1),
   piupper(-2),pnweights(0)
{
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
BandWeights::~BandWeights()
{
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BandWeights::writeTemplate(const String &statefile)
{
  if(pathExistence(statefile)==DoesNotExist){
    printf("file '%s' does not exist. Shall I create a template?: [y/n]",
	   statefile());
    String line;
    getLine(stdin,line);line.stripBlanks(true);
    if(line.lowerCase()[0]=='y'){
      FILE *FHO=openforwrite(statefile);
      try{
	
	fprintf(FHO,"# edit the stuff below\n\n"
		"weightinfile +bweights\n"
		"weightoutfile +bwsum\n\n"
		"#ewindow emin emax # is optional: removes all bands, which are \n"
		"                   # completely outside this window\n\n"
		"name WEIGHTSUM1      # WEIGHTSUM1 is the name of the resulting weight sum\n"
		"                    "
		" # containing the orbitals from the three lines below\n" 
		"    atom Fe sites 1..3 5 orbitals 4s 3d \n"
		"    atom O sites  7 8 orbitals 2p-1 2p+1\n"
		"    labels  'O(007)2s+0' \"O(007)2p+0\" fac 2 # here fully qualified "
		"labels are used\n"
		"                                            "
		"# both are multiplied by 2\n\n"
		"# You can have more weight sums, just by "
		"starting a new one with 'name ...'.\n\n"
		"name WEIGHTSUM2\n"
		"    atom ...\n"
		"    labels ...\n"
		"    atom ...\n"
		"    atom ...\n\n"
		"name WEIGHTSUM2\n"
		"    labels ...\n"
		"    labels ...\n"
		"    atom ...\n"
		"    atom ...\n\n"
		"# 'labels' and 'atom' can be mixed as shown above.\n"
		"# 'labels' can have more than one label-argument.\n"
		"# Quotes (\") or single quotes (') can be used to protect spaces "
		"in names/labels.\n\n"
		"# For pseudo-non-relativistic projections the keyword spin up|dn|both\n"
		"# might be nessecary ... you will find out.\n"
		);
	
      }catch(Except &ex){
	fclose(FHO);
	throw ex;
      }catch(...){}
      fclose(FHO);
      throw TerminateExcept(Sprintf("\nDone. Please edit '%s' and re-run!"
				    " Aborting...\n\n",statefile()));
    }else{
      throw TerminateExcept("\nOk, aborting...\n\n");
    }
  

  }

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BandWeights::extractHeaderInfo(const String &infile,long prot)
{
  NSF("BandWeights::processFile");
  pheaderready=false;

  String truncerror;
  truncerror.sprintf("%s: File \'%s\' : truncated?!\n",nsubfunc,infile());

  FILE *FHI=openforread(infile);

  // here we either got thrown out or we have an open file
  
  try{ // catch to close file
    String line;
    

    {
      // read the "band" header
      if( !getLine(FHI,line) ) throw Except(truncerror);
      
      int err=sscanf(line(),"%*s%ld%lg%ld%ld%ld%ld%ld%ld"
		     ,&pnsite,&pef,&pnk,&pnvdim,&pnband,&pnspin,&pilower,&piupper)-8;
      pilower--; piupper--;
      if(err==-2){ /* old file assumed */ pilower=0; piupper=pnband-1; err=0; }
      if(err) throw Except(Sprintf("Header of file '%s' corrupted?\n",infile()));
    }
    
    // read orbital/label header -> for later
    String header;
    if( !getLine(FHI,header) ) throw Except(truncerror);

    // get the number of orbitals from reading one data line
    long lmax=0,start=0;
    {
      if( !getLine(FHI,line) ) throw Except(truncerror);
      {
	StringPtr pp(line); //pp=beginning of line
	long wordlength;
	pnweights=-2;
	do{
	  StringPtr save0(pp);// save0=beginning of whitespace or word
	  pp.skipWhiteSpace();// pp=after whitespace
	  StringPtr savew(pp);// savew=pp=after whitespace possibly word-start
	  if((wordlength=(pp.skipNoneWhiteSpace()-savew))>0){ //a word left
	    pnweights++;
	    if(pnweights==1){
	      start=savew.skipNoneWhiteSpace(".")-line();
	    }
	  }
	  if(pnweights==1) lmax=pp-save0;
	  
	}while(wordlength);
      }
      if(prot>=1) printf("weights=%ld startcolumn=%ld width=%ld\n",pnweights,start,lmax);
    }

    {
      // actually read the header
      if(prot>=10) myDebug()<<header;
      for(long i=-1;++i<pnweights;){
	long index=start+i*lmax;
	if(i+lmax-1>header.length()) break;
	if(pnweights==1){
	  plabels<<header.subString(index).stripBlanks(true);
	}else{
	  plabels<<header.subString(index,index+lmax-1).stripBlanks(true);
	}
      }
      pnweights=plabels.size();
      
      if(prot>=1) myDebug()<<plabels<<plabels.size();
    }
    
    pheaderready=true;
    
  }catch(Except &ex){
    fclose(FHI);
    throw ex;
  }catch(...){}
  
  fclose(FHI);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BandWeights::readStateDefinitions(const String &statefile,
				       String &infile,String &outfile,
				       ValueArray<StateSum> &statesums,
				       ValueArray<double> &ewindow,
				       bool getfilenamesonly,long prot)
{


#define PRNT(txt,v)							\
  {myDebug()								\
      <<(txt)<<"Line"<<pl.value().linenr<<":"				\
      <<StringPtrLexer::tokenNames[pl.value().type]			\
      <<pl.value().v							\
      << "ahead: Line"							\
      <<pl.lookAheadValue().linenr<<":"					\
      <<StringPtrLexer::tokenNames[pl.lookAheadValue().type];}
#define PRNTNOVAL(txt)							\
  {myDebug()								\
      <<(txt)<<"Line"<<pl.value().linenr<<":"				\
      <<StringPtrLexer::tokenNames[pl.value().type]			\
      << "ahead: Line"							\
      <<pl.lookAheadValue().linenr<<":"					\
      <<StringPtrLexer::tokenNames[pl.lookAheadValue().type];}
#define RPRNT(txt)							\
  {myDebug()								\
      <<(txt)<<"Line"<<pl.value().linenr<<":"				\
      <<StringPtrLexer::tokenNames[pl.value().type]			\
      <<pl.value().lvalue<<".."<<pl.value().rvalue			\
      << "ahead: Line"							\
      <<pl.lookAheadValue().linenr<<":"					\
      <<StringPtrLexer::tokenNames[pl.lookAheadValue().type];}
#define THROW(what)							\
  throw Except(Sprintf("File '%s', Line %ld,%ld: expecting %s.\n",	\
		       statefile(),pl.value().linenr,			\
		       pl.value().columnnr,what));
#define THROWAHEAD(what)							\
  throw Except(Sprintf("File '%s', Line %ld,%ld: expecting %s.\n",	\
		       statefile(),pl.lookAheadValue().linenr,		\
		       pl.lookAheadValue().columnnr,what));
#define THROWORB(what)							\
  throw Except(Sprintf("File '%s', Line %ld,%ld: invalid label/orbital(s) '%s'.\n", \
		       statefile(),					\
		       pl.value().linenr,pl.value().columnnr,what));			
#define THROWORBSPEC(what)							\
  throw Except(Sprintf("File '%s', Line %ld,%ld: invalid orbital specification '%s'.\n", \
		       statefile(),					\
		       pl.value().linenr,pl.value().columnnr,what));			
#define THROWORBRES(what,help)						\
  throw Except(Sprintf("File '%s', Line %ld: invalid resulting label/orbital '%s'.%s\n", \
		       statefile(),pl.value().linenr,what,help));
#define THROWUNEXP(stop)						\
  if(pl.lookAheadValue().type==StringPtrLexer::TKeyWord){		\
    throw Except(Sprintf("File '%s', Line %ld,%ld: unexpected keyword '%s'.\n",	\
			 statefile(),					\
			 pl.lookAheadValue().linenr,			\
			 pl.lookAheadValue().columnnr,			\
			 pl.lookAheadValue().sval()));			\
  }else if(pl.lookAheadValue().type==StringPtrLexer::TWord){		\
    throw Except(Sprintf("File '%s', Line %ld,%ld: unexpected input '%s'.\n", \
			 statefile(),					\
			 pl.lookAheadValue().linenr,			\
			 pl.lookAheadValue().columnnr,			\
			 pl.lookAheadValue().sval()));			\
  }else if(pl.lookAheadValue().type==StringPtrLexer::TInteger){		\
    throw Except(Sprintf("File '%s', Line %ld,%ld: unexpected integer '%ld'.\n", \
			 statefile(),					\
			 pl.lookAheadValue().linenr,			\
			 pl.lookAheadValue().columnnr,			\
			 pl.lookAheadValue().lvalue));			\
  }else if(pl.lookAheadValue().type==StringPtrLexer::TDouble){		\
    throw Except(Sprintf("File '%s', Line %ld,%ld: unexpected double '%g'.\n",	\
			 statefile(),					\
			 pl.lookAheadValue().linenr,			\
			 pl.lookAheadValue().columnnr,			\
			 pl.lookAheadValue().dvalue));			\
  }else if(pl.lookAheadValue().type==StringPtrLexer::TNone){		\
    if(stop)								\
      throw Except(Sprintf("File '%s', Line %ld,%ld: unexpected end of file.\n", \
			   statefile(),					\
			   pl.lookAheadValue().linenr,			\
			   pl.lookAheadValue().columnnr,		\
			   StringPtrLexer::tokenNames[pl.lookAheadValue().type])); \
    expect=Nothing;							\
  }else{								\
    throw Except(Sprintf("File '%s', Line %ld,%ld: unexpected input '%s'.\n", \
			 statefile(),					\
			 pl.lookAheadValue().linenr,			\
			 pl.lookAheadValue().columnnr,			\
			 StringPtrLexer::tokenNames[pl.lookAheadValue().type])); \
  }
#define WARNMULTIPLE(which)						\
  warnings+=Sprintf("WARNING:File '%s', Line %ld: (Resulting) label '%s' exists more\n" \
		    " than once "					\
		    "in weight file, first occurance taken.\n",		\
		    statefile(),					\
		    pl.value().linenr,					\
		    which);


  
  if(!getfilenamesonly && !pheaderready) throw Except("No header info available yet. "
				 "Call extractHeaderInfo first.\n");

  FILE *FHI=openforread(statefile);

  statesums.clear();
  

  try{

    
    // first read the whole file: should never be a big thing
    String txt;
    String line;
    while(getLine(FHI,line)){
      txt+=line;
    }

    // then analyse the content

    enum Expectations{Nothing=0,Name=1,Contrib=2,Sites=3,Orbitals=4,Factor=5,Spin=6};
    Expectations expect=Name;
    StringList keywords;
    keywords<<"name"<<"labels"<<"atom"<<"sites"<<"orbitals"<<"fac"<<"spin"
	    <<"weightinfile"<<"weightoutfile"<<"ewindow";
    StringList basekeywords;
    basekeywords<<"name"<<"weightoutfile"<<"weightinfile"<<"ewindow";
    StateSum *statesum=0;
    long jstatesum0=0;
    StringPtr p(txt);
    StringPtrLexer pl(p,keywords,",=");

    if(getfilenamesonly){
      do{
	pl.nextToken();

	if(pl.value().isKeyword("weightinfile")){
	  pl.eatPunctuation("=");
	  if(!pl.lookAheadValue().isStringOrWord()) 
	    THROWAHEAD("a string after 'weightinfile'");
	  pl.nextToken();
	  infile=pl.value().sval;
	  
	  //break;
	}else if(pl.value().isKeyword("weightoutfile")){
	  pl.eatPunctuation("=");
	  if(!pl.lookAheadValue().isStringOrWord()) 
	    THROWAHEAD("a string after 'weightoutfile'");
	  pl.nextToken();
	  outfile=pl.value().sval;
	  
	  //break;
	}
	
      }while(!pl.eof());
	
    }else{
      
      String warnings;
      class StateInfo{
      public:
	void clear(){sites.clear();orbitals.clear();spin=0;};
	long spin;
	String el;
	ValueArray<long> sites;
	ValueArray<String> orbitals;
      } stateinfo;
      
      bool parsedebug=false;
      bool collectlabels;
      do{
	//myDebug("p=<%s...>",String(p(),20).replace("\n","\\n",0,true)());
	pl.nextToken();
	collectlabels=false;
	//myDebug()<<p.value().type;
	if(parsedebug){
	  if(pl.value().type==StringPtrLexer::TString){
	    PRNT("\t\t",sval);
	  }else if(pl.value().type==StringPtrLexer::TComment){
	    PRNT("\t\t",sval);
	  }else if(pl.value().type==StringPtrLexer::TKeyWord){
	    if(pl.value().sval=="name"){
	      PRNT("",sval);
	    }else{
	      PRNT("\t",sval);
	    }
	    
	  }else if(pl.value().type==StringPtrLexer::TRange){
	    RPRNT("\t");
	  }else if(pl.value().type==StringPtrLexer::TInteger){
	    PRNT("\t\t",lvalue);
	  }else if(pl.value().type==StringPtrLexer::TDouble){
	    PRNT("\t\t",dvalue);
	  }else if(pl.value().type==StringPtrLexer::TPunctuation){
	    PRNT("\t\t",sval);
	  }else if(pl.value().type==StringPtrLexer::TWord){
	    PRNT("\t\t",sval);
	  }else if(pl.value().type==StringPtrLexer::TNone){
	    PRNTNOVAL("\t\t");
	  }
	}
	
	//myDebug()<<"expect="<<(long)expect;
	
	switch(expect){
	  //myDebug()<<"switch start";
	case Nothing:
	  //myDebug()<<"case Nothing";
	  break;
	case Name:
	  {
	    //myDebug()<<"case name"<<pl.value().sval;
	    
	    if(pl.value().isKeyword("weightinfile")){
	      pl.eatPunctuation("=");
	      if(!pl.lookAheadValue().isStringOrWord()) 
		THROWAHEAD("a string after 'weightinfile'");
	      pl.nextToken();
	      //myDebug()<<pl.value().sval;
	      if(pl.lookAheadValue().isNone()) expect=Nothing;
	      break;
	    }else if(pl.value().isKeyword("weightoutfile")){
	      pl.eatPunctuation("=");
	      if(!pl.lookAheadValue().isStringOrWord()) 
		THROWAHEAD("a string after 'weightoutfile'");
	      pl.nextToken();
	      //myDebug()<<pl.value().sval;
	      if(pl.lookAheadValue().isNone()) expect=Nothing;
	      
	      break;
	    }else if(pl.value().isKeyword("ewindow")){
	      pl.eatPunctuation("=");
	      if(!pl.lookAheadValue().isNumber())
		THROWAHEAD("two numbers after 'ewindow'");
	      pl.nextToken();
	      double emin=pl.value().dvalue;
	      pl.eatPunctuation(",",true);
	      if(!pl.lookAheadValue().isNumber())
		THROWAHEAD("two numbers after 'ewindow'");
	      pl.nextToken();
	      double emax=pl.value().dvalue;
	      
	      ewindow.clear();
	      ewindow<<emin<<emax;
	      if(pl.lookAheadValue().isNone()) expect=Nothing;
	      break;
	    }
	    if(!pl.value().isKeyword("name")) THROW("name");
	    pl.eatPunctuation("=");
	    if(!pl.lookAheadValue().isStringOrWord()) THROWAHEAD("a string after 'name'");
	    pl.nextToken();
	    statesums<<StateSum(pl.value().sval);
	    statesum=&statesums.last();
	    //printf("\nname=\"%s\"\n",statesum->name());
	    
	    if(pl.lookAheadValue().isKeyword("labels") ||
	       pl.lookAheadValue().isKeyword("atom")){
	      expect=Contrib;
	    }else if(pl.lookAheadValue().isKeywords(basekeywords)){
	      expect=Name;
	    }else
	      THROWUNEXP(false);
	    break;
	  }
	case Factor:
	  {
	    //myDebug()<<"case Factor";
	    if(!pl.value().isKeyword("fac")) THROW("'fac'");
	    
	    pl.eatPunctuation("=");
	    
	    //myDebug()<<"case Factor"<<pl.value().sval;
	    //myDebug()<<"case Factor"<<pl.lookAheadValue().type;
	    if(!pl.lookAheadValue().isNumber())
	      THROWAHEAD("a number after 'fac'");
	    
	    pl.nextToken();
	    //printf(" fac %g\n",pl.value().dvalue);
	    
	    for(long i=jstatesum0;i<statesum->data.size();i++){
	      statesum->data[i].factor=pl.value().dvalue;
	    }
	    
	    if(pl.lookAheadValue().isKeyword("labels") ||
	       pl.lookAheadValue().isKeyword("atom")){
	      expect=Contrib;
	    }else if(pl.lookAheadValue().isKeywords(basekeywords)){
	      expect=Name;
	    }else
	      THROWUNEXP(false);
	    
	    break;
	  }
	case Contrib:
	  {
	    //myDebug()<<"case Contrib";
	    if(pl.value().isKeyword("labels")){
	      jstatesum0=statesum->data.size();
	      //printf("\nlabels");
	      
	      // eatup =
	      pl.eatPunctuation("=");
	      pl.eatPunctuation(",",true);
	      
	      
	      // make sure there is at least one argument
	      if(!pl.lookAheadValue().isStringOrWord())
		THROWAHEAD("string(s) after 'labels'");
	      pl.nextToken();
	      long j=plabels.find(pl.value().sval);
	      if(j<0)THROWORB(pl.value().sval());
	      //printf(" %s g=%ld",pl.value().sval(),j);
	      statesum->data<<State(j);
	      j=plabels.find(pl.value().sval,j+1);
	      if(j>=0) WARNMULTIPLE(pl.value().sval());
	      
	      
	      pl.eatPunctuation(",",true);
	      
	      while(pl.lookAheadValue().isStringOrWord()){
		pl.nextToken();
		long j=plabels.find(pl.value().sval);
		if(j<0)THROWORB(pl.value().sval());
		//printf(" %s g=%ld",pl.value().sval(),j);
		statesum->data<<State(j);
		j=plabels.find(pl.value().sval,j+1);
		if(j>=0) WARNMULTIPLE(pl.value().sval());
		pl.eatPunctuation(",",true);
	      }
	      if(pl.lookAheadValue().isKeyword("labels") ||
		 pl.lookAheadValue().isKeyword("atom")){
		expect=Contrib;
	    }else if(pl.lookAheadValue().isKeywords(basekeywords)){
		expect=Name;
	      }else if(pl.lookAheadValue().isKeyword("fac")){
		expect=Factor;
	      }else
		THROWUNEXP(false);
	      
	    }else if(pl.value().isKeyword("atom")){
	      //myDebug()<<"atom";
	      jstatesum0=statesum->data.size();
	      // eatup =
	      pl.eatPunctuation("=");
	      pl.eatPunctuation(",",true);
	      // make sure there is at least one argument
	      if(!pl.lookAheadValue().isStringOrWord())
		THROWAHEAD("element name after 'atom'");
	      
	      pl.nextToken();
	      //printf("\natom=%s\n",pl.value().sval());
	      
	      stateinfo.clear();
	      stateinfo.el=pl.value().sval();
	      
	      //myDebug()<<pl.lookAheadValue().type;
	      if(pl.lookAheadValue().isKeyword("sites")){
		expect=Sites;
	      }else{
		THROWAHEAD("'sites'");
	      }
	      
	      
	    }else{
	      THROW("'labels' or 'atom'");
	    }
	    
	    break;
	  }
	  
	case Sites:
	  {
	    //myDebug()<<"case sites";
	    if(!pl.value().isKeyword("sites")) THROW("'sites'");
	    pl.eatPunctuation("=");
	    pl.eatPunctuation(",",true);
	    
	    // make sure there is at least one argument
	    if(!pl.lookAheadValue().isIntegerOrRanges())
	      THROWAHEAD("integer(s)/range(s) after 'sites'");
	    pl.nextToken();
	    if(pl.value().type==StringPtrLexer::TInteger){
	      //printf(" %ld",pl.value().lvalue);
	      stateinfo.sites<<pl.value().lvalue;
	    }else{
	      //printf(" %ld..%ld",pl.value().lvalue,pl.value().rvalue);
	      for(long i=pl.value().lvalue;i<=pl.value().rvalue;i++){
		stateinfo.sites<<i;
	      }
	    }
	    
	    pl.eatPunctuation(",",true);
	    while(pl.lookAheadValue().isIntegerOrRanges()){
	      pl.nextToken();
	      if(pl.value().type==StringPtrLexer::TInteger){
		//printf(" %ld",pl.value().lvalue);
		stateinfo.sites<<pl.value().lvalue;
	      }else{
		//printf(" %ld..%ld",pl.value().lvalue,pl.value().rvalue);
		for(long i=pl.value().lvalue;i<=pl.value().rvalue;i++){
		  stateinfo.sites<<i;
		}
	      }
	      pl.eatPunctuation(",",true);
	    }
	    
	    
	    if(pl.lookAheadValue().isKeyword("orbitals")){
	      expect=Orbitals;
	    }else
	      THROWAHEAD("'orbitals'");
	    
	    
	    
	    break;
	  }
	case Orbitals:
	  {
	    
	    //myDebug()<<"case orbitals";
	    if(!pl.value().isKeyword("orbitals")) THROW("'orbitals'");
	    
	    pl.eatPunctuation("=");
	    pl.eatPunctuation(",",true);
	    
	    // make sure there is at least one argument
	    if(!pl.lookAheadValue().isStringOrWord())
	      THROWAHEAD("string(s) after 'orbitals'");
	    pl.nextToken();
	    //printf(" %s",pl.value().sval());
	    bool haveall=false;
	    long nm,l,jj,m;
	    String::StateLabels ret;
	    ret=String::isStateLabel(pl.value().sval(),nm,l,jj,m);
	    if(ret==String::SL_invalid) THROWORBSPEC(pl.value().sval());
	    if(ret==String::SL_all) haveall=true;
	    
	    
	    stateinfo.orbitals<<pl.value().sval;
	    
	    
	    pl.eatPunctuation(",",true);
	    
	    while(pl.lookAheadValue().isStringOrWord()){
	      pl.nextToken();
	      //printf(" %s",pl.value().sval()); 
	      String::StateLabels ret;
	      ret=String::isStateLabel(pl.value().sval(),nm,l,jj,m);
	      if(ret==String::SL_invalid) THROWORBSPEC(pl.value().sval());
	      if(ret==String::SL_all) haveall=true;
	      if(haveall && stateinfo.orbitals.size()>0)
		THROW("only one orbital descriptor if 'orbitals' contains 'all'");
	      stateinfo.orbitals<<pl.value().sval;
	      pl.eatPunctuation(",",true);
	    }
	    
	    collectlabels=true;
	    if(pl.lookAheadValue().isKeyword("fac")){
	      expect=Factor;
	    }else if(pl.lookAheadValue().isKeyword("labels")||
		     pl.lookAheadValue().isKeyword("atom")){
	      expect=Contrib;
	    }else if(pl.lookAheadValue().isKeywords(basekeywords)){
	      expect=Name;
	    }else if(pl.lookAheadValue().isKeyword("spin")){
	      if(haveall){
		THROWAHEAD("no spin keyword after 'orbitals all'");
	      }
	      collectlabels=false;
	      expect=Spin;
	    }else
	      THROWUNEXP(false);
	    break;
	  }
	case Spin:
	  {
	    if(!pl.value().isKeyword("spin")) THROW("'spin'");
	    pl.eatPunctuation("=");
	    // make sure there is  one argument
	    if(!pl.lookAheadValue().isStringOrWord())
	      THROWAHEAD("up/dn/down/both after 'spin'");
	    pl.nextToken();
	    if(pl.value().sval.lowerCase()=="up"){
	      stateinfo.spin=1;
	    }else if(pl.value().sval.lowerCase()=="dn"){
	      stateinfo.spin=2;
	    }else if(pl.value().sval.lowerCase()=="down"){
	      stateinfo.spin=2;
	    }else if(pl.value().sval.lowerCase()=="both"){
	      stateinfo.spin=3;
	    }else{
	      THROW("up/dn/down/both after 'spin'");
	    }
	    
	    collectlabels=true;
	    
	    if(pl.lookAheadValue().isKeyword("fac")){
	      expect=Factor;
	    }else if(pl.lookAheadValue().isKeyword("labels")||
		     pl.lookAheadValue().isKeyword("atom")){
	      expect=Contrib;
	    }else if(pl.lookAheadValue().isKeywords(basekeywords)){
	      expect=Name;
	    }else
	      THROWUNEXP(false);
	    break;
	  }
	  
	}
	
	//myDebug()<<"after switch expect="<<expect<<" collectlabels="<<collectlabels;

	if(collectlabels){
	  for(long i=-1;++i<stateinfo.sites.size();){
	    for(long io=-1;++io<stateinfo.orbitals.size();){
	      long nm,l,jj,m,m0,m1,dm=1;
	      String::StateLabels ret=
		String::isStateLabel(stateinfo.orbitals.at(io),nm,l,jj,m);
	      //myDebug()<<stateinfo.orbitals.at(io)<<ret<<nm<<l<<jj<<m;
	      String sel=(stateinfo.el+" ").subString(0,1);
	      String selsit=Sprintf("%2s(%03.3ld)",sel(),stateinfo.sites.at(i));
	      String selsitnl=selsit;
	      if(ret!=String::SL_all)
		selsitnl+=String::intToOrbital(nm,l);
	      const char *ending="";
	      const char *fending="/2";
	      m0=m1=m;
	      String s1;
	      if(ret==String::SL_all){
	      }else if(ret==String::SL_nl){
		m0=-l;m1=l;
		s1=selsitnl;
	      }else if(ret==String::SL_nlj){
		m0=-jj;m1=jj;dm=2;
		s1=selsitnl+Sprintf("%ld/2",jj);
		ending=fending;
	      }else if(ret==String::SL_nlm){
		s1=selsitnl;
	      }else if(ret==String::SL_nljm){
		s1=selsitnl+Sprintf("%ld/2",jj);
		ending=fending;
	      }
	      //myDebug()<<"sel="<<sel;
	      //myDebug()<<"selsit="<<selsit;
	      //myDebug()<<"selsitnl="<<selsitnl;
	      //myDebug()<<"s1="<<s1;
	      
	      for(long mm=m0;mm<=m1;mm+=dm){
		int ms0=stateinfo.spin,ms1=stateinfo.spin;
		if(stateinfo.spin==3){
		  ms0=1;ms1=2;
		}
		if(ret==String::SL_all){
		  StringList sl=plabels.findStartWith(selsit);
		  if(sl.size()<0) {
		    String sim=plabels.findStartWith(sel).join();
		    THROWORBRES(selsit(),
				(String("\n\tsimilar orbitals in weights file:\n\t")
				 +sim())());
		  }
		  for(int k=0;k<sl.size();k++){
		    long kk=plabels.find(sl.at(k));
		    statesum->data<<State(kk);
		    kk=plabels.find(sl.at(k),kk+1);
		    if(kk>=0) WARNMULTIPLE(sl.at(k)());
		  }
		}else{
		  for(int ms=ms0;ms<=ms1;ms++){
		    const char *sp=(ms==0?"":(ms==1?"up":"dn"));
		    String ss=s1+Sprintf("%+ld%s%s",mm,ending,sp);
		    //myDebug()<<ss;
		    long j=plabels.find(ss);
		    if(j<0){
		      String sim;
		      sim=plabels.findStartWith(ss).join();
		      if(!sim.length()) sim=plabels.findStartWith(s1).join();
		      if(!sim.length()) sim=plabels.findStartWith(selsitnl).join();
		      if(!sim.length()) sim=plabels.findStartWith(selsit).join();
		      if(!sim.length()) sim=plabels.findStartWith(sel).join();
		      THROWORBRES(ss(),(String("\n\tsimilar orbitals in weights file:\n\t")
					+sim())());
		    }
		    statesum->data<<State(j);
		    j=plabels.find(ss,j+1);
		    if(j>=0) WARNMULTIPLE(ss());
		  }// ms
		}//if SL_all
	      }
	      
	    }
	  }
	}
	
      }while(!pl.eof());
      
      if(!warnings.empty()){
	printf("%s",warnings());
      }
    }//not passive      
      
  }catch(Except &ex){
    fclose(FHI);
    throw ex;
  }catch(...){}

  fclose(FHI);

#undef PRNT
#undef RPRNT
#undef THROW
#undef THROWUNEXP
#undef THROWORB
#undef THROWORBSPEC
#undef THROWORBRES
#undef WARNMULTIPLE

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BandWeights::addWeights(const String &infile,const String &outfile,
			     const ValueArray<StateSum> &statesums,
			     const ValueArray<double> &ewindow,
			     long prot)
{
  NSF("BandWeights::addWeights");

  if(!pheaderready) throw Except("No header info available yet. "
				 "Call extractHeaderInfo first.\n");
  

  MachineTime time;
  
  FILE *FHI=openforread(infile);
  FILE *FHO=openforwrite(outfile);

  bool hasNaN=false;
  
  try{
    if(prot==1){
      printf("Start\n");
      //printf("Read from '%s'\n",infile());
      //printf("Write to '%s'\n",outfile());
    }
    
    if(statesums.size()<=0){
      throw TerminateExcept("No orbitals/states resulted from the input. "
			    "Nothing to be done here!\n",TerminateExcept::ExitNormal);
    }

    if(prot==1){
      printf("Selected weight sums:\n");
      
      for(long ist=-1;++ist<statesums.size();){
	const StateSum &ssum=statesums.at(ist);
	const ValueArray<State> &sdata=ssum.data;
	if(sdata.size()<=0){
	  printf("Sum %ld (%s) does not contain any orbitals.\n",ist,
		 ssum.name());
	}else{
	  printf("State %ld, %s\n",ist+1,ssum.name());
	}
	
	for (long i=-1;++i<sdata.size();) {
	  const State &sta=sdata.at(i);
	  printf("%ld: %g*%s\n"
		 ,sta.index+1,sta.factor,plabels.at(sta.index)());
	}
      }
    }



 
    bool haveewindow=false;
    if(prot==1){
      printf("\nStart extracting data ...\n");
      if(ewindow.size()>0){
	printf("\nRestyricting to energy window [%g,%g]\n\n",
	       ewindow.at(0),ewindow.at(1));
	haveewindow=true;
      }

    }


    long subdiv=MyMath::max<long>(1,pnk/(100/10));
    long countMax=MyMath::max<long>(1,pnk/subdiv);

    long imin=0,imax=pnband-1;
    String line;
    getLine(FHI,line);
    getLine(FHI,line);// skip header

    if(haveewindow){


      ValueArray<bool> isbelow(pnband,true);
      ValueArray<bool> isabove(pnband,true);
      

      for (long ik=-1;++ik<pnk;){
	if ( (ik+1)%subdiv ==0 ){
	  if(prot==1){
	    printf("progress %.2f%%\n",100.*(long)((ik+1)/subdiv)/countMax);
	  }
	}
	for (long ms=-1;++ms<pnspin;){
	  for (long in=-1;++in<pnband;){
	    getLine(FHI,line);
	    char *end;
	    const char *p=line();
	    double dk=strtod(p,&end);
	    if(end==p) throw Except("invalid k-data\n");
	    p=end;
	    if((dk!=dk)) { hasNaN=true; dk=0; }
	    double ek=strtod(p,&end);
	    if(end==p) throw Except("invalid e-data\n");
	    if((ek!=ek)) { hasNaN=true; ek=0; }
	    if(ek>ewindow.at(0)) isbelow[in]=false;
	    if(ek<ewindow.at(1)) isabove[in]=false;
	    //myDebug()<<in<<"ek="<<ek<<ewindow.at(0)<<" isbelow="<<isbelow[in];
	  }//bands
	}//spin
	

      }//k-points
      
      for (long in=-1;++in<pnband;){
	if(!isbelow[in]) {imin=in;break;}
      }
      for (long in=pnband;--in>-1;){
	if(!isabove[in]) {imax=in;break;}
      }
      
      
      rewind(FHI);
      getLine(FHI,line);
      getLine(FHI,line);

      printf("\nminimum band in output=%ld\n",imin+1);
      printf("maximum band in output=%ld\n\n",imax+1);
      
    }//haveewindow

    
    subdiv=MyMath::max<long>(1,pnk/(100/5));
    countMax=MyMath::max<long>(1,pnk/subdiv);



    //write header
    fprintf(FHO,"# %5ld%12.7f %10ld %10ld %10ld %10ld %10ld %10ld\n",
	    pnsite,pef,pnk,pnvdim,imax-imin+1,pnspin,imin+1,imax+1); 
    fprintf(FHO,"#       ik           e(k,n)               ");
    for(long is=-1;++is<statesums.size();){
      const ValueArray<State> &sdata=statesums.at(is).data;
      String s;
      if( sdata.size() == 1 ){
	const State &sta=sdata.at(0);
	s=plabels[sta.index];
      }else{
	s=Sprintf("%-17s",statesums.at(is).name());
      }
      fprintf(FHO,"%-17s",s());
    }
    fprintf(FHO,"\n");


    double *wei=new double[pnweights];
    /* read rest of file */


    for (long ik=-1;++ik<pnk;){
      if ( (ik+1)%subdiv ==0 ){
	if(prot==1){
	  printf("progress %.2f%%\n",100.*(long)((ik+1)/subdiv)/countMax);
	}
      }
      for (long ms=-1;++ms<pnspin;){
	for (long in=-1;++in<pnband;){
	  getLine(FHI,line);
	  if(in<imin||in>imax) continue;
	  char *end;
	  const char *p=line();
	  double dk=strtod(p,&end);
	  if(end==p) throw Except("invalid k-data\n");
	  p=end;
	  if((dk!=dk)) { hasNaN=true; dk=0; }
	  double ek=strtod(p,&end);
	  if(end==p) throw Except("invalid e-data\n");
	  if((ek!=ek)) { hasNaN=true; ek=0; }
	  for(long i=-1;++i<pnweights;){
	    p=end; wei[i]=strtod(p,&end);
	    if(end==p) throw Except("invalid w-data\n");
	    if((wei[i]!=wei[i])) { hasNaN=true; wei[i]=0; }
	  }

	  fprintf(FHO,"%16.7f %16.7f",dk,ek);
	  for (long ist=-1;++ist<statesums.size();) {
	    const ValueArray<State> &sdata=statesums.at(ist).data;
	    double w=0.;
	    for (long i=-1;++i<sdata.size();) {
	      const State &sta=sdata.at(i);
	      w+=sta.factor*wei[sta.index];
	    }
	    /* now we have the states weight */
	    fprintf(FHO," %16.7f",w);
	  }
	  fprintf(FHO,"\n");

	}
      }
    }


    
  }catch(Except &ex){
    fclose(FHI);
    fclose(FHO);
    throw ex;
  }catch(TerminateExcept &ex){
    fclose(FHI);
    fclose(FHO);
    throw ex;
  }catch(...){
  }
  fclose(FHI);
  fclose(FHO);

  if(hasNaN) {
    printf("\nWARNING: in file \"%s\": some numbers are NaN! "
	   "Replaced them with 0!\n\n",infile());
  }

  if(prot==1){
    printf("End normal\n");
    time.print("total");
  }

}




/*EOF*/ 
