/*************************************************************************
 * @(#)CoeffBasisInfo.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   01 Nov 2010
 *************************************************************************
 */

#include "Except.hh"
#include "CoeffBasisInfo.hh"
#include "MyMath.hh"
#include "Exception.hh"

using namespace XFPLO;

const String CoeffBasisInfo::defaultEl="??";

/* ***********************************************************************
 *
 * ***********************************************************************
 */
CoeffBasisInfo::CoeffBasisInfo()
  :pnsite(0),pnvdim(0),pFREL(false),pisready(false)
{
  //myDebug() <<"CoeffBasisInfo::CoeffBasisInfo";

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
CoeffBasisInfo::~CoeffBasisInfo()
{
  clean();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void CoeffBasisInfo::clean()
{
  pisready=false;
  pelements.clear();
  pnRv.clear();
  pnnRv.clear();
  plRv.clear();
  pdegRv.clear();
  pjRv.clear();
  pblockstart.clear();
  pblocklength.clear();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void CoeffBasisInfo::getCBIFromLabels(const StringList &labels)
{
  bool prot=false;
  //  myDebug() <<"CoeffBasisInfo::getCBIFromLabels";
  clean();
  pnsite=-1;
  for(int i=0;i<labels.size();i++){
    StringList sl=labels[i].split("()");
    try{
      pnsite=MyMath::max<long>(pnsite,sl[1].convertToLong());
    }catch(Except ex) {
      throw Exception(ex.msg());  
    }
  }
  if(prot)
    (void)printf("nsite=%ld\n",pnsite);

  if(pnsite<0){
    throw Exception("Cannot deduce number of sites from labels");  
  }

  pelements.adjustSize(pnsite);
  pnRv.resize(pnsite);
  pblockstart.resize(pnsite);
  pblocklength.resize(pnsite);
  pnnRv.resize(pnsite);
  plRv.resize(pnsite);
  pdegRv.resize(pnsite);
  pjRv.resize(pnsite);
  
  StringList *orblist=new StringList[pnsite];
  for(int i=0;i<labels.size();i++){
    StringList sl=labels[i].split("()");
    long isa=sl[1].convertToLong()-1;
    pelements[isa]=labels[i].subString(0,1);
    sl[2].stripBlanks();

    long nmain,l,j,m;
    String::StateLabels ret=
      String::isStateLabel(sl[2],nmain,l,j,m);
    
    if(!(ret==String::SL_nlm || ret==String::SL_nljm))
      throw Exception("Invalid state label in coeeficient file?");  

    String orb;
    if(ret==String::SL_nlm)
      orb=sl[2].subString(0,1);
    else if(ret==String::SL_nljm)
      orb.sprintf("%1ld%1c%1ld/2",nmain,String::intToSpdf(l),j);
    
    if(!orblist[isa].contains(orb))
      orblist[isa].append(orb);
  }


  for(long isa=0;isa<pnsite;isa++){
    pnRv[isa]=orblist[isa].size();
    pnnRv[isa].resize(pnRv[isa]);
    plRv[isa].resize(pnRv[isa]);
    pdegRv[isa].resize(pnRv[isa]);
    pjRv[isa].resize(pnRv[isa]);
  }
  if(prot){
    printf("No of labels=%ld\n",labels.size());
  }


  pFREL=false;
  
  for(long isa=0;isa<pnsite;isa++){
    for(long iv=0;iv<pnRv[isa];iv++){
      String orb=orblist[isa][iv];


      long nmain,l,j,m;
      String::StateLabels ret=
	String::isStateLabel(orb,nmain,l,j,m);
      if(ret==String::SL_nl){
	pnnRv[isa][iv]=nmain;
	plRv[isa][iv]=l;
	pjRv[isa][iv]=2*l+2;
	pdegRv[isa][iv]=2*l+1;
	pFREL=false;
      }else if(ret==String::SL_nlj){
	pnnRv[isa][iv]=nmain;
	plRv[isa][iv]=l;
	pjRv[isa][iv]=j;
	pdegRv[isa][iv]=j+1;
	pFREL=true;
      }
    }
  }

  pnvdim=0;
  for(long isa=0;isa<pnsite;isa++){
    pblockstart[isa]=pnvdim;
    long n=0;
    for(long iv=0;iv<pnRv[isa];iv++){
      if(prot){
	if(pFREL){
	  printf("%s(%03ld) %1ld%1c%1ld/2\n",element(isa)(),isa+1
		 ,nnRv(iv,isa),String::intToSpdf(lRv(iv,isa))
		 ,jRv(iv,isa));
	}else{
	  printf("%s(%03ld) %1ld%1c: [%ld] \n",element(isa)(),isa+1
		 ,nnRv(iv,isa),String::intToSpdf(lRv(iv,isa))
		 ,pblockstart[isa]);
	}
      }
      n+=degRv(iv,isa);
    }
    pblocklength[isa]=n;
    pnvdim+=n;
  }
  pisready=true;
 
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void CoeffBasisInfo::setNsite(long n)
{
  pnsite=n;
  pelements.adjustSize(pnsite);
  pnRv.resize(pnsite);
  pblockstart.resize(pnsite);
  pblocklength.resize(pnsite);
  pnnRv.resize(pnsite);
  plRv.resize(pnsite);
  pdegRv.resize(pnsite);
  pjRv.resize(pnsite);

}

/*EOF*/ 
