/*************************************************************************
 * @(#)DMQuantAxis.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   08 Nov 2010
 *************************************************************************
 */

#include "DMQuantAxis.hh"
#include "MyMath.hh"
#include "Except.hh"

//todo: no S/printf if in heavy calculations

/* ***********************************************************************
 *
 * ***********************************************************************
 */
StringList DMQuantAxis::calcAxesTrans(const Vec3 &x,const Vec3 &z,long l
				      ,SqMatrix<Cmplx> &ytr)
{
  bool ok=true;
  String output;
  String errors;
  output+="input: x-axis "+x.printS()+"\n";
  output+="input: z-axis "+z.printS()+"\n\n";

  double d=norm(x);
  if(d==0) {
    errors+="ERROR: input x-axis has zero norm\n\n";
    ok=false;
  }
  Vec3 newx=x/d;
  
  d=norm(z);
  if(d==0){
    errors+="ERROR: input z-axis has zero norm\n\n";
    ok=false;
  }
  Vec3 newz=z/d;

  d=newx*newz;

  if(fabs(d)>1.0e-6){
    errors+="WARNING: x- and z-axis are not orthogonal?!\n"
      "will correct x-axis accordingly -> check data!\n\n";
  }

  // make newx ortho to new z
  newx=newx-newz*d;

  d=newx*newx;
  if(d<=1.0e-12) {
    errors+="ERROR: corrected new x-axis has zero lenght?!\n\n";
    ok=false;
  }
  newx/=sqrt(d);

  Vec3 newy=newz.cross(newx);

  
  output+="resulting x-axis "+newx.printS()+"\n";
  output+="resulting y-axis "+newy.printS()+"\n";
  output+="resulting z-axis "+newz.printS()+"\n\n";
  
  Matrix3x3 R(newx,newy,newz);
  
  output+=" E^T = (e_x,e_y,e_z) \n";
  output+=" E'^T=E^T R \n";
  output+="R=\n"+R.printS()+"\n\n";

  Vec3 eul=Vec3::eulerAngles(newz,newx);

  output+=Sprintf("Euler angles of this transformation: \n"
		  "al: %12.7f\n"
		  "be: %12.7f\n"
		  "ga: %12.7f\n"
		  ,(eul/Pi*180)[0]
		  ,(eul/Pi*180)[1]
		  ,(eul/Pi*180)[2]
		  );

  Matrix3x3 MM=eul.eulerRMatrix();

  double tol=1.0e-7;
  if( norm(MM(0)-R(0))>tol ||
      norm(MM(1)-R(1))>tol ||
      norm(MM(2)-R(2))>tol    ){
    errors+="\nWARNING: \nR from euler anlges=\n";
    errors+=MM.printS();
    errors+="\n\t !! differs from R !!\n\n";
  }


  if(l>=0){
    
    ytr=wignerDRealYlm(eul,false,l,true);
    
    output+="\nY'_{lm}=\\sum_{n} Y_{ln} D^l_{nm}\n\n";
    output+="    ";
    for(int m=-l;m<=l;m++){
      output+=Sprintf("     Y%1ld%+2d       ",l,m);
    }
    output+="\n";
    for(int i=-l;i<=l;i++){
      output+=Sprintf("Y'%1ld%+2d   ",l,i);
      for(int j=-l;j<=l;j++){
	double d=ytr(j+l,i+l).real();
	output+=Sprintf("% -15g ",d);
      }
      output+=Sprintf("\n");
    }
    
    
    // transform global into local occupation matrix
    // Y'=YD

    // Y n Y^{+} = Y' D^{+} n D Y'^{+}
    //
    //  n'  =  D^{+} n  D 
    //  n   =  D     n' D^{+}   

    output+="\nY'=YD\n\n"
      "Y n Y^{+} = Y' D^{+} n D Y'^{+}\n\n"    
      "n'  =  D^{+} n  D \n"
      "n   =  D     n' D^{+}  \n\n";
  }
  
  if(!ok){
    throw Except(errors);
  }

  StringList sl;
  sl.append(output);
  sl.append(errors);
  return sl;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
StringList DMQuantAxis::calcAxesTransJ(const Vec3 &x,const Vec3 &z,long J
				      ,SqMatrix<Cmplx> &Djmm)
{
  bool ok=true;
  String output;
  String errors;
  output+="input: x-axis "+x.printS()+"\n";
  output+="input: z-axis "+z.printS()+"\n\n";

  double d=norm(x);
  if(d==0) {
    errors+="ERROR: input x-axis has zero norm\n\n";
    ok=false;
  }
  Vec3 newx=x/d;
  
  d=norm(z);
  if(d==0){
    errors+="ERROR: input z-axis has zero norm\n\n";
    ok=false;
  }
  Vec3 newz=z/d;

  d=newx*newz;

  if(fabs(d)>1.0e-6){
    errors+="WARNING: x- and z-axis are not orthogonal?!\n"
      "will correct x-axis accordingly -> check data!\n\n";
  }

  // make newx ortho to new z
  newx=newx-newz*d;

  d=newx*newx;
  if(d<=1.0e-12) {
    errors+="ERROR: corrected new x-axis has zero lenght?!\n\n";
    ok=false;
  }
  newx/=sqrt(d);

  Vec3 newy=newz.cross(newx);

  
  output+="resulting x-axis "+newx.printS()+"\n";
  output+="resulting y-axis "+newy.printS()+"\n";
  output+="resulting z-axis "+newz.printS()+"\n\n";
  
  Matrix3x3 R(newx,newy,newz);
  
  output+=" E^T = (e_x,e_y,e_z) \n";
  output+=" E'^T=E^T R \n";
  output+="R=\n"+R.printS()+"\n\n";

  Vec3 eul=Vec3::eulerAngles(newz,newx);
				      

  output+=Sprintf("Euler angles of this transformation: \n"
		  "al: %12.7f\n"
		  "be: %12.7f\n"
		  "ga: %12.7f\n"
		  ,(eul/Pi*180)[0]
		  ,(eul/Pi*180)[1]
		  ,(eul/Pi*180)[2]
		  );

  Matrix3x3 MM=eul.eulerRMatrix();

  double tol=1.0e-7;
  if( norm(MM(0)-R(0))>tol ||
      norm(MM(1)-R(1))>tol ||
      norm(MM(2)-R(2))>tol    ){
    errors+="\nWARNING: \nR from euler anlges=\n";
    errors+=MM.printS();
    errors+="\n\t !! differs from R !!\n\n";
  }


  if(J>=1){
    
    Djmm=wignerDChijm(eul,J,false);
    
    output+="\n|jm>'=\\sum_{n} |jn> D^l_{nm}\n\n";
    output+="         ";
    for(int m=-0;m<=J;m++){
      output+=Sprintf("     |%1ld/2%+2d/2>        ",J,m*2-J);
    }
    output+="\n";
    for(int i=0;i<=J;i++){
      output+=Sprintf("|%1ld/2%+2d/2>'   ",J,i*2-J);
      for(int j=0;j<=J;j++){
	double d=Djmm(j,i).real();
	double di=Djmm(j,i).imag();
	output+=Sprintf("%- 9.5f+i*% -9.5f ",d,di);
      }
      output+=Sprintf("\n");
    }
    
    
    // transform global into local occupation matrix
    // Y'=YD

    // Y n Y^{+} = Y' D^{+} n D Y'^{+}
    //
    //  n'  =  D^{+} n  D 
    //  n   =  D     n' D^{+}   

    output+="\n|>'=|> D\n\n"
      "|> n <| = |>' D^{+} n D <|'\n\n"    
      "n'  =  D^{+} n  D \n"
      "n   =  D     n' D^{+}  \n\n";
  }
  
  if(!ok){
    throw Except(errors);
  }

  StringList sl;
  sl.append(output);
  sl.append(errors);
  return sl;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DMQuantAxis::calcTrans(long l)
{

  calcAxesTrans(xaxis,zaxis,l,ytr);
  //printSqMatrix(ytr);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DMQuantAxis::calcTransJ(long j)
{

  calcAxesTransJ(xaxis,zaxis,j,ytr);
  //printSqMatrix(ytr);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DMQuantAxis::calcTransLJ(long l)
{
  // the combined Dj rotation for the l-shell where
  // j=l-1/2 forms the first block and j=l+1/2 the second

  SqMatrix<Cmplx> Dm,Dp;

  calcAxesTransJ(xaxis,zaxis,2*l-1,Dm);
  calcAxesTransJ(xaxis,zaxis,2*l+1,Dp);

  
  ytr.resize(2*(2*l+1));
  ytr.copyBlock(0,0,Dm,0,0,2*l,2*l);
  ytr.copyBlock(2*l,2*l,Dp,0,0,2*l+2,2*l+2);



  //printSqMatrix(ytr);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DMQuantAxis::calcTransLL(long l)
{
  // the combined Dl rotation for the l-shell spin matrix
  // (2x2 matrix of l-blocks)

  //myDebug()<<"calcTransLL";

  SqMatrix<Cmplx> Dl;

  calcAxesTrans(xaxis,zaxis,l,Dl);

  int deg=2*l+1;
  
  ytr.resize(2*deg);
  ytr.copyBlock(0,0,Dl,0,0,deg,deg);
  ytr.copyBlock(deg,deg,Dl,0,0,deg,deg);



  //printSqMatrix(ytr);
}




/*EOF*/ 
