/* @(#)FString.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _STRINGXX_H
#define _STRINGXX_H 1

#include <stdio.h>
#include <string.h>
#include <limits.h>
#if (defined QXFSF || defined QXFPL)
#include <QString>
#endif

#include "ValueArray.hh"

class StringList;

class String;
String Sprintf(const char *fmt,...);

/* ***********************************************************************
 *
 * ***********************************************************************
 */

class ReadBuffer{
 public:
  ReadBuffer(long buffersize=2048)
    :pbuffersize(buffersize),pbuffer(new char[pbuffersize]),
     pcurrent_pos(pbuffer),plen(0){ pbuffer[0]='\0';  }

  ~ReadBuffer(){delete [] pbuffer;};

  inline const char *currentPos()const{return pcurrent_pos;};
  inline void setCurrentPos(char *p){pcurrent_pos=p;};
  inline void incrCurrentPos(size_t l){pcurrent_pos+=l;plen-=l;};
  inline char *data(){return pbuffer;};


  inline size_t length()const {return plen;};
  inline void setLength(size_t l){plen=l;};

  inline size_t bufferSize()const{return pbuffersize;};

 private:
  size_t pbuffersize;
  char *pbuffer; /* the read buffer */
  char *pcurrent_pos;		/* current position in buffer */
  size_t plen;			/* number of remaing bytes in buffer */

};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
class String{
 public:
  //
  enum StateLabels{SL_invalid=0,SL_all=1,SL_nl=2,SL_nlj=3,SL_nlm=4,SL_nljm=5,
		   SL_allnlm=6};
  enum Strategy{Default=0,DontShrink=1};
  //
  String(char const* v="");
  String(const char *other,size_t len);
  String(long i);
  String(String const& other);
#if (defined QXFSF || defined QXFPL)
  String(const QString &);
#endif
  //
  ~String();
  //
  void setStrategy(Strategy s){pstrategy=s;}// strategy DontShrink means that
  // the memory is not freed in assignment, which avoids repeated reallocs (delete new)
  // It is incompletly implemented

  String const& operator=(char other); // this does not work in 
  //                                               String s='x'; 
  //                                      but in 
  //                                               String s; s='x';
  String const& operator=(const char *other);
  String const& operator=(String const& other);
  String const& operator+=(const char *other);
  String const& operator+=(String const& other);
  String const& operator+=(char other);
  String &ncat(const char *other,size_t len);
  String &ncopy(const char *other,size_t len);

  

  inline long operator==(String const &other)const {
    return cmp(*this,other)==0;};
  inline long operator!=(String const &other)const {
    return cmp(*this,other)!=0;};

  String &sprintf(const char *fmt,...);
  //
  inline long length() const{return plength;};
  inline long internalSize()const{return pinternalsize;}
  String &setLength(unsigned long len);
  //
  String &shorten(long size);
  long empty() const {
    return ( strspn(pvalue," \t") == plength ); };
  inline char &operator[](long index){return pvalue[index];};
  inline const char &operator[](long index)const{return pvalue[index];};
  const char *operator()()const{return pvalue;};  
  char * data(){return pvalue;};
  //
  const char * data()const{return pvalue;};
  //
  long convertToLong() const;
  unsigned long convertToULong() const;
  double convertToFortranReal();
  double convertToReal()const;
  void convertToFortranChar(char *dest,long dest_len)const;
  String & removeTrailingBlanks(bool removenewline=false);
  String & chop(const char *removeset="\n");
  String & removeLeadingBlanks(bool removenewline=false);
  String & stripBlanks(bool removenewline=false){  /* trailing and leading */
    return removeLeadingBlanks(removenewline)
      .removeTrailingBlanks(removenewline); };
  String & toUpper();
  String & toLower();
  String &escapeDecode();
  String &escapeEncode();
  String &expandComma();
  String &interpolateQuotes();
  //
  String &removeSection(long i,long j);
  String &insertChar(long i,char c);
  //
  // delete all occurence of characters, which are in set 
  String removeFromSet(String const &set) const; 
  static String escapeDecode(const char *c);
  static String escapeEncode(const char *c); //todo:remove and check
  static String escapeDecode(String const &);
  static String escapeEncode(String const &); //todo:remove and check
  String lowerCase()const { return String(*this).toLower();};
  String upperCase()const { return String(*this).toUpper(); };
  String capitalized(bool all=false)const;
  String &capitalize(bool all=false);
  //
  String &rightPadd(char c,long length);
  String &leftPadd(char c,long length);
  //
  long contains(String const &set) const;
  // find first occurence of any character of set, or -1
  long scan(String const &set) const;
  // find first occurence of any character not in set, or -1
  long reject(String const &set) const;
  long find(String const &substring,bool fromEnd=false) const;
  bool startsWith(String const &substring)const{return find(substring)==0;}
  String subString(long i)const;
  String subString(long i,long j)const;
  String &replace(const String &substring,const String &with,long startat=0,
		  bool all=false);
  String &set(unsigned long i,char with);
  //
  // split into tokens delimited by any characters in delim
  StringList split(const char *delim) const;
  // split into tokens delimited by the string delim
  StringList breakup(const char *delim) const;
  // introduce line breaks to limit to cols columns
  StringList wrapAround(unsigned long cols,const char *marker="\\") const;
  // 
  ValueArray<long> splitToLongArray(const char *delim) const;
  ValueArray<double> splitToDoubleArray(const char *delim) const;
  //
  static long cmp(String const &s1,String const &s2);
  static long cmp(const char *s1,String const &s2);
  static long cmp(String const &s1,const char *s2);
  static long cmp(const char *s1,const char *s2);
  //
  static long cmp(String const &s1,String const &s2,long l);
  static long cmp(const char *s1,String const &s2,long l);
  static long cmp(String const &s1,const char *s2,long l);
  static long cmp(const char *s1,const char *s2,long l);
  //
  static long islong(long *l,const char *s);
  static long islong(long *l,char c);
  static long islong(long *l,const char *s,long lowerbound,long upperbound);
  static long isSPDF(char s,long maxmoment);
  static long spdfToInt(char s);
  static char intToSpdf(long l);
  void isOrbital(long maxmoment) const;
  void orbitalToInt(long &nmain,long &l,long maxmoment=6)const;
  static StateLabels isStateLabel(const char *s,long &nmain,long &l,long &j,long &m);
  static StateLabels isStateLabel(const String &s,long &nmain,long &l,long &j,long &m){ 
    return isStateLabel(s(),nmain,l,j,m);};
  static String intToOrbital(long nmain,long l);
  static long isPrint(long );
  static long isDigit(long );
  static long isAlpha(long c);
  static void makeRatio(double d,long &m,long &n,long maxn,double tol);
  static String doubleToRatio(double d,long maxn,double tol);
  static String doubleToString(const String &fmt,double d,bool contract=false);
  static String sprintRat(long nom,long den);

  bool getLine(FILE *FH,ReadBuffer &);
  inline char *fgets(FILE *FH){
    return ::fgets(pvalue,pinternalsize,FH);
  }
  //
  String dump() {return Sprintf("pinternalsize=%ld",pinternalsize);}
  void dumpinternals()const;

 public:
  static const String setaz; 
  static const String setAZ; 
  static const String setazAZ; 
 private:
  void swapAll(String& other); // swap two String objects internal data
 private:
  //static const char* classname;
  // we keep a buffer and reallocate if nessecary
  unsigned long pinternalsize;  // at least strlen+1
  unsigned long plength;// the actual string length excluding terminal '\0'
  //                       -> avoid calls to strlen
  char *pvalue;
  Strategy pstrategy;
  //
  //friend String const operator+(String const& a,String const& b);
  friend String Sprintf(const char *fmt,...);
};


inline bool operator<(const String &a,const String &b)
{
  return String::cmp(a,b)<0;
}
inline bool operator>(const String &a,const String &b)
{
  return String::cmp(a,b)>0;
}

inline String operator+( String lhs, const String & rhs ) {
  lhs += rhs;
  return lhs;
}

//String const operator+(long a,String const& b);
//String const operator+(String const& a,long b);

template <class T>
String toBin(T x,bool cut=false)
{
  if(!x and cut) return "0";
  String ret;
  T mask=1<<(sizeof(T)*CHAR_BIT-1);
  bool start=cut;
  for(int i=sizeof(T)*CHAR_BIT;--i>=0;){
    if(start){
      if(x&mask){
	start=false;
      }else{
	mask>>=1;
	continue;
      }
    }
    ret+=(x&mask?'1':'0');
    mask>>=1;
  }
  return ret;
}

String printS(const ValueArray<double> &va,const char *fmt="%g",
 	      const char *delim=" ");
String printS(const ValueArray<long> &va,const char *fmt="%ld",
 	      const char *delim=" ");



/* ***********************************************************************
 *
 * ***********************************************************************
 */
class StringList{
 public:
  StringList();
  StringList(StringList const &);
  StringList(String const &s)
    :psize(0),internalsize(0),parray(0){ append(s);};

  ~StringList(){clear();}
  StringList const& operator=(StringList const &other);
  //
  String &operator[](long index) const;
  const String &at(long index) const;
  //StringList const& operator=(StringList const& other);
  long size() const {return psize;}
  long internalSize()const{return internalsize;}
  void preallocate(long size){clear();grow(size);}
  StringList &append(const String &);
  inline StringList &operator<<(const String &v){
    this->append(v);
    return *this;
  }
  String join(const char *delim=" ",bool delimatend=false,
	      bool delimatfront=false) const;
  String &last(){return *(parray[psize-1]);}
  const String &last() const{return *(parray[psize-1]);}


  void swap(long i,long j);
  //
  StringList const &escapeDecode();
  StringList &interpolateQuotes();
  //
  long allEmpty()const;
  // find first occurance of needle in list starting from index startfrom
  // if fromback search backwards
  // if startfrom is invalid or nothing found return -1;
  long find(String const &needle,long startfrom=0,bool fromback=false)const;
  StringList findStartWith(String const &needle,long startfrom=0,bool fromback=false)const;
  long contains(String const &s)const{return find(s)>=0;};
  long remove(String const &);
  long remove(long );
  void clear();
  //
  void adjustSize(long minsize){
    grow(minsize);
    psize=minsize;
  }
  //
  void grow(long size);
  //
  void dumpinternals()const;
  //
 private:
  //static const char *classname;
  long psize;
  long internalsize;
  String **parray;
};


/* ***********************************************************************
 *
 * ***********************************************************************
 */

#ifdef	__cplusplus
extern "C" {
#endif
  
#ifdef	__cplusplus
}
#endif


#endif /* _STRINGXX_H */

