/* @(#)SqMatrix.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _SQMATRIX_H
#define _SQMATRIX_H 1


#include <stdlib.h>
#include <stdio.h>
#include "Cmplx.hh"
#include "Vec3.hh"
#include "MyDebug.hh"


class String;
class Vec3;


template <class T>
class ValueArray;
class Matrix3x3;

/* ***********************************************************************
 *
 * ***********************************************************************
 */


template <class T>
class SqMatrix
{
 public:
  // ----------------------------------------------------------------
  SqMatrix():diml(0),dimr(0),array(0){
    array=new T[1];
  };
  
  // ----------------------------------------------------------------
  //create square matrix
  SqMatrix(long d):diml(d),dimr(d),array(0){
    array=new T[diml*dimr];
    for(int i=0;i<diml;i++){
      for(int j=0;j<dimr;j++){
	array[i+diml*j]=0;
      }
    }
  };
  // ----------------------------------------------------------------
  SqMatrix(const Matrix3x3 &m)
    :diml(3),dimr(3),array(new T[9])
  {
    for(int j=0;j<3;j++){
      for(int i=0;i<3;i++){
	array[i+3*j]=m(i,j);
      }
    }
  }
  // ----------------------------------------------------------------
  //rect square matrix
  SqMatrix(long dl,long dr):diml(dl),dimr(dr),array(0){
    array=new T[diml*dimr];
    for(int i=0;i<diml;i++){
      for(int j=0;j<dimr;j++){
	array[i+diml*j]=0;
      }
    }
  };
  
  // ----------------------------------------------------------------
  // create square diagonal matrix from vector
  SqMatrix(const ValueArray<T> &v):diml(v.size()),dimr(v.size()){
    array=new T[diml*dimr];
    for(long i=0;i<diml;i++){
      for(long j=0;j<dimr;j++){
	array[i+diml*j]=0;
      }
      array[i+diml*i]=v.at(i);
    }
  }
  
  // ----------------------------------------------------------------
  ~SqMatrix(){
    if(array){
      delete [] array;
      array=0;
    }
  };

  // ----------------------------------------------------------------
  SqMatrix(const SqMatrix &other){
    diml=other.diml;
    dimr=other.dimr;
    array=new T[diml*dimr];
    for(int i=0;i<diml;i++){
      for(int j=0;j<dimr;j++){
	array[i+diml*j]=other.array[i+diml*j];
      }
    }
  };

  // ----------------------------------------------------------------
  const SqMatrix &operator=(const T &value){
    for(int i=0;i<diml;i++){
      for(int j=0;j<dimr;j++){
	array[i+diml*j]=value;
      }
    }
    return *this;
  }

  // ----------------------------------------------------------------
  SqMatrix &operator=(const SqMatrix &other){
    SqMatrix temp(other);
    swap(temp);
    return *this;
  };

  // ----------------------------------------------------------------
  SqMatrix &copyBlock(long r0,long c0,const SqMatrix &a,
		      long ra ,long ca,long dr, long dc)
  { // copy a block of length dr,dc of 'a' starting at (row,col)=(ra,ca)
    // to the block of '*this' sitting at (row,col)=(r0,c0)
    if(c0+dc>dimr){
      printf("SqMatrix::copyBlock: result matrix columns run over bound\n");
      exit(1);
    }
    if(r0+dr>diml){
      printf("SqMatrix::copyBlock: result matrix rows run over bound\n");
      exit(1);
    }
    if(ra+dr>a.diml){
      printf("SqMatrix::copyBlock: other matrix rows run over bound\n");
      exit(1);
    }
    if(ca+dc>a.dimr){
      printf("SqMatrix::copyBlock: other matrix rows run over bound\n");
      exit(1);
    }
    // myDebug()<<"args="<<r0<<c0<<ra<<ca<<dr<<dc;
    for(int j=0;j<dc;j++){
      for(int i=0;i<dr;i++){
	// myDebug()<<"array["<<r0+i<<"+"<<diml<<"*("<<c0+j<<")]=a.array["
	// 	 <<ra+i<<"+"<<a.diml<<"*("<<j+ca<<")]";
	array[r0+i+diml*(c0+j)]=a.array[ra+i+a.diml*(j+ca)];
      }
    }
    return *this;
  }


  // ----------------------------------------------------------------
  SqMatrix<T> &operator+=(const SqMatrix<T> &mb){
    if(diml!=mb.diml || dimr!=mb.dimr){
      printf("SqMatrix::operator+=() different dimension\n");
      exit(1);
    }
    for(int i=0;i<diml;i++){
      for(int j=0;j<dimr;j++){
	array[i+diml*j]+=mb.array[i+diml*j];
      }
    }
    return *this;
  }
  // ----------------------------------------------------------------
  SqMatrix<T> &operator-=(const SqMatrix<T> &mb){
    if(diml!=mb.diml || dimr!=mb.dimr){
      printf("SqMatrix::operator-=() different dimension\n");
      exit(1);
    }
    for(int i=0;i<diml;i++){
      for(int j=0;j<dimr;j++){
	array[i+diml*j]-=mb.array[i+diml*j];
      }
    }
    return *this;
  }
  
  // ----------------------------------------------------------------
  inline const T &operator()(long i,long j) const { 
    if(i<0 || i>diml-1){
      printf("SqMatrix left index out of bound \n");
      exit(1);
    }
    if(j<0 || j>dimr-1){
      printf("SqMatrix right index out of bound \n");
      exit(1);
    }
    return array[i+diml*j];
  }

  // ----------------------------------------------------------------
  inline T &operator()(long i,long j) { 
    if(i<0 || i>diml-1){
      printf("SqMatrix left index out of bound \n");
      exit(1);
    }
    if(j<0 || j>dimr-1){
      printf("SqMatrix right index out of bound \n");
      exit(1);
    }
    return array[i+diml*j];
  }

  // ----------------------------------------------------------------
  void resize(long d){
    diml=dimr=d;
    if(array){
      delete [] array;
      array=0;
    };
    array=new T[diml*dimr];
    for(int i=0;i<diml;i++){
      for(int j=0;j<dimr;j++){
	array[i+diml*j]=0;
      }
    }
  };

  // ----------------------------------------------------------------
  void resize(long dl,long dr){
    diml=dl;
    dimr=dr;
    if(array){
      delete [] array;
      array=0;
    };
    array=new T[diml*dimr];
    for(int i=0;i<diml;i++){
      for(int j=0;j<dimr;j++){
	array[i+diml*j]=0;
      }
    }
  };

  // ----------------------------------------------------------------
  inline long size(long which=0) const{
    if(which<=0 && diml!=dimr){
      printf("cannot use size() on non-square matrix\n");
      exit(1);
    }
    if(which==0 || which ==1){
      return diml;
    }
    if(which==2){
      return dimr;
    }
    printf("invalid argument of SqMatrix::size\n");
    exit(1);
  };

  // ----------------------------------------------------------------
  SqMatrix &removeRows(long i0,long i1){
    if(i0<0 || i0>diml-1){
      printf("SqMatrix::removeRows row index out of bound \n");
      exit(1);
    }
    if(i1<0 || i1>diml-1){
      printf("SqMatrix::removeRows row index out of bound \n");
      exit(1);
    }
    if(i1<i0){
      printf("SqMatrix::removeRows arguments invalid \n");
      exit(1);
    }
    long dn=i1-i0+1;
    SqMatrix temp(diml-dn,dimr);
    
    for(long j=0;j<dimr;j++){
      for(int i=0;i<i0;i++){
	temp.array[i+temp.diml*j]=array[i+diml*j];
      }
      for(int i=i1+1;i<diml;i++){
	temp.array[i-dn+temp.diml*j]=array[i+diml*j];
      }
    }
    swap(temp);
    return *this;

  }

  // ----------------------------------------------------------------
  SqMatrix &removeCols(long j0,long j1){
    if(j0<0 || j0>diml-1){
      printf("SqMatrix::removeCols column index out of bound \n");
      exit(1);
    }
    if(j1<0 || j1>diml-1){
      printf("SqMatrix::removeCols column index out of bound \n");
      exit(1);
    }
    if(j1<j0){
      printf("SqMatrix::removeCols arguments invalid \n");
      exit(1);
    }
    long dn=j1-j0+1;
    
    SqMatrix temp(diml,dimr-dn);
    
    for(long i=0;i<diml;i++){
      for(int j=0;j<j0;j++){
	temp.array[i+diml*j]=array[i+diml*j];
      }
      for(int j=j1+1;j<dimr;j++){
	temp.array[i+diml*(j-dn)]=array[i+diml*j];
      }
    }
    swap(temp);
    return *this;

  }

  // ----------------------------------------------------------------
  SqMatrix operator-()const{
    SqMatrix tmp(diml,dimr);
    for(int i=0;i<diml*dimr;i++){
      tmp.array[i]=-array[i];
    }
    return tmp;
  }

  // ----------------------------------------------------------------
  SqMatrix transposed()const{
    SqMatrix result(dimr,diml);
    for(long i=0;i<diml;i++){
      for(long j=0;j<dimr;j++){
	result.array[j+dimr*i]=array[i+diml*j];
      }
    }
    return result;
  }

  // ----------------------------------------------------------------
  SqMatrix &conjugate(){
    for(long i=0;i<diml*dimr;i++){
      array[i].conjugate();
    }
    return *this;
  }
  
  // ----------------------------------------------------------------
  SqMatrix conjugated()const{
    SqMatrix result(dimr,diml);
    for(long i=0;i<diml*dimr;i++){
      result.array[i]=conjg(array[i]);
    }
    return result;
  }
  
  // ----------------------------------------------------------------
  template <class X>
  SqMatrix &scale(X d){
    for(int i=0;i<diml*dimr;i++){
      array[i]*=d;
    }
    return *this;
  }

  // ----------------------------------------------------------------
  SqMatrix hermiteConjugated()const{
    SqMatrix result(dimr,diml);
    for(long i=0;i<diml;i++){
      for(long j=0;j<dimr;j++){
	result.array[j+dimr*i]=conjg(array[i+diml*j]);
      }
    }
    return result;
  }

  // ----------------------------------------------------------------
  SqMatrix &invert();
  
  // ----------------------------------------------------------------
  SqMatrix inverse()const{ 
    SqMatrix<T> res(*this);
    return res.invert();
  };

  // ----------------------------------------------------------------
  SqMatrix operator*(T d) const{
    SqMatrix tmp(diml,dimr);
    for(int i=0;i<diml*dimr;i++){
      tmp.array[i]=array[i]*d;
    }
    return tmp;
  }
  // ----------------------------------------------------------------
  SqMatrix &operator*=(T d){

    for(int i=0;i<diml*dimr;i++){
      array[i]*=d;
    }
    return *this;
  }
  // ----------------------------------------------------------------
  SqMatrix operator/(T d) const{
    SqMatrix tmp(diml,dimr);
    for(int i=0;i<diml*dimr;i++){
      tmp.array[i]=array[i]/d;
    }
    return tmp;
  }

  // ----------------------------------------------------------------
  SqMatrix operator*(const SqMatrix &a) const{
    if(this->dimr != a.diml){
      fprintf(stderr
	      ,"SqMatrix::operator*: dimension of arguments does not fit\n");
      exit(2);

    }
    SqMatrix result(diml,a.dimr);
    for(long j=0;j<a.dimr;j++){
      for(long i=0;i<diml;i++){
	T r=0;
	for(long k=0;k<dimr;k++){
	  r+=array[i+diml*k]*a.array[k+dimr*j];
	}
	result.array[i+diml*j]=r;
      }
    }
    return result;
  };

  // ----------------------------------------------------------------
  // the complex scalar product of two columns of two SqMatrix
  // limited to a range of indices:
  //
  // dotcol=conjg(a(arow0:arow0+len-1,acol))*b(brow0:brow0+len-1,bcol)
  //
  static T dotcol(const SqMatrix &a,long arow0,long acol,
	const SqMatrix &b,long brow0,long bcol,long len)
  {
    T c=0;
    for(long i=0;i<len;i++){
      c+=conjg(a(arow0++,acol))*b(brow0++,bcol);
    }
    return c;
  }

  // ----------------------------------------------------------------
  ValueArray<T> column(long i)const{
    if(i<0 || i>dimr-1){
      printf("SqMatrix::column: index out of bounds\n");
      exit(2);
    }
    ValueArray<T> res;
    res.resize(diml);
    for(long j=0;j<diml;j++){
      res[j]=array[j+diml*i];
    }
    return res;
  }

  // ----------------------------------------------------------------
  ValueArray<T> row(long i)const{
    if(i<0 || i>diml-1){
      printf("SqMatrix::row: index out of bounds\n");
      exit(2);
    }
    ValueArray<T> res;
    res.resize(dimr);
    for(long j=0;j<dimr;j++){
      res[j]=array[i+diml*j];
    }
    return res;
  }

  // ----------------------------------------------------------------
  void printSInfo(const char *name) const{
    printf("printSInfo: array %s(%ld,%ld) at %p\n",name,diml,dimr,(void*)array);
  };


  // ----------------------------------------------------------------
  SqMatrix &diag(T value){
    for(long i=0;i<diml*dimr;i++){
      array[i]=0;
    }
    long n=(diml<dimr?diml:dimr);
    for(long i=0;i<n;i++){
      array[i+diml*i]=value;
    }
    return *this;
  }

  T trace()const
  {
    long n=(diml<dimr?diml:dimr);
    T res=0;
    for(long i=-1;++i<n;){
      res+=array[i+diml*i];
    }
    return res;
  }


  // ----------------------------------------------------------------
  inline bool isSquare()const
  {
    return (diml==dimr);
  }


  // ----------------------------------------------------------------
  bool isZero(double tol=1.0e-12)const
  {
    for(long i=0;i<diml;i++){
      for(long j=0;j<dimr;j++){
	if(abs(array[i+diml*j])>tol)return false;
      }
    }
    return true;
  }

  // ----------------------------------------------------------------
  double Fnorm2()const
  {
    double d=0.;
    for(long i=-1;++i<diml*dimr;){
      d+=MyMath::toDouble(conjg(array[i])*array[i]);
    }
    return d;
  }


 private:

  // ----------------------------------------------------------------
  void swap(SqMatrix& other)
  {
    T *c=this->array;
    this->array=other.array;
    other.array=c;
    long d=this->diml;
    this->diml=other.diml;
    other.diml=d;
    d=this->dimr;
    this->dimr=other.dimr;
    other.dimr=d;
  }
  
 private:
  long diml,dimr;
  T *array;

  template <class X>
  friend X trace(const SqMatrix<X> &a);

  template <class X>
  friend X trace(const SqMatrix<X> &a,const SqMatrix<X> &b);

};


/* ***********************************************************************
 *
 * ***********************************************************************
 */
template <class T>
SqMatrix<T> operator-(SqMatrix<T> ma,const SqMatrix<T> &mb){
  ma-=mb;
  return ma;
}
template <class T>
SqMatrix<T> operator+(SqMatrix<T> ma,const SqMatrix<T> &mb){
  ma+=mb;
  return ma;
}

template <class T>
T trace(const SqMatrix<T> &a)
{
  long n=(a.diml<a.dimr?a.diml:a.dimr);
  T res=0;
  for(long i=-1;++i<n;){
    res+=a.array[i+a.diml*i];
  }
  return res;
}

template <class T>
T trace(const SqMatrix<T> &a,const SqMatrix<T> &b)
{ // trace(a*b)
  if(a.dimr != b.diml){
    fprintf(stderr
	    ,"SqMatrix::trace(a,b): dimension of arguments does not fit\n");
    exit(2);
    
  }
  long n=(a.diml<b.dimr?a.diml:b.dimr);
  T res=0;
  for(long i=-1;++i<n;){
    for(long j=-1;++j<a.dimr;){
      res+=a.array[i+a.diml*j]*b.array[j+b.diml*i];
    }
  }
  return res;
}

ValueArray<double> operator*(const SqMatrix<double> &m,
			     const ValueArray<double> &v);
ValueArray<double> operator*(const ValueArray<double> &v,
			     const SqMatrix<double> &m);

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void printSqMatrix(const SqMatrix<long> &m,const char*fmt="%10ld ");
void printSqMatrix(const SqMatrix<int> &m,const char*fmt="%10d ");
void printSqMatrix(const SqMatrix<double> &m,const char*fmt="%10g ",
		   const char*prefix="");
void printSqMatrix(const SqMatrix<Cmplx> &m,const char*fmt="(%10g, %10g) ");

/* ***********************************************************************
 *
 * ***********************************************************************
 */

void diagonalizeSymmetric(const SqMatrix<double> &A,ValueArray<double> &e,
			  SqMatrix<double> &Z,String &msg);

template<>
SqMatrix<double> &SqMatrix<double>::invert();
template<>
SqMatrix<Cmplx> &SqMatrix<Cmplx>::invert();

/* ***********************************************************************
 *
 * ***********************************************************************
 */
// 
//  make U such that \cal{Y}=Y U
//  where \cal{Y} are complex spherical harmonics and Y are real ones
//
SqMatrix<Cmplx> makeYlmUMatrix(long l);
SqMatrix<Cmplx> relTOpseudoCmplx(const Vec3 &spinzaxis,long l);
SqMatrix<Cmplx> relTOpseudo(const Vec3 &spinzaxis,long l);
SqMatrix<Cmplx> wignerDRealYlm(const Vec3 &eul,bool inv,long l,bool trans);
SqMatrix<Cmplx> wignerDChijm(const Vec3 &eul,long j,bool trans);

// time reversal for real harmonics block
SqMatrix<Cmplx> timeReversalRealYlm(long l);
// time reversal for spin-real harmonics block (Y_{lm}*\chi_s)
SqMatrix<Cmplx> timeReversalRealYlmChi(long l);
// time reversal for complex harmonics block
SqMatrix<Cmplx> timeReversalYlm(long l);
// time reversal for spin-complex harmonics block (Y_{lm}*\chi_s)
SqMatrix<Cmplx> timeReversalYlmChi(long l);
// time reversal for spherical spinor block
SqMatrix<Cmplx> timeReversalSphericalSpinor(long l,long j);
// time reversal for spherical spinor blocks of a whole l-shell
SqMatrix<Cmplx> timeReversalSphericalSpinorLShell(long l);


// L-matrix in real harmonics basis
SqMatrix<Cmplx> LsYlm(long l,long s);
// L-matrix in spin-real harmonics basis
SqMatrix<Cmplx> LsYlmChi(long l,long s);
// L-matrix in complex harmonics basis
SqMatrix<Cmplx> LxyzYlm(long l,long k);
// L-matrix in spin-complex harmonics basis
SqMatrix<Cmplx> LxyzYlmChi(long l,long k);
// L-matrix in spherical spinor blocks of a whole l-shell
SqMatrix<Cmplx> LxyzSphericalSpinorLShell(long l,long k);

// sigma-matrix in spherical spinor blocks of a whole l-shell
SqMatrix<Cmplx> SigmaxyzSphericalSpinorLShell(long l,long k);


SqMatrix<double> real(const SqMatrix<Cmplx> &);
SqMatrix<double> imag(const SqMatrix<Cmplx> &);


template<class T>
inline MyDebug operator<<(MyDebug dbg, const SqMatrix<T> &m)
{
  FPRINT(dbg.FH1,"[");
  for(long i=-1;++i<m.size(1);){
    dbg<<"[";
    for(long j=-1;++j<m.size(2);){
      dbg<<m(i,j);
      if(j<m.size(2)-1)dbg<<",";
    }
    if(i==m.size(1)-1)
      dbg<<"]]\n";
    else
      dbg<<"]\n";

  }
  return dbg;
}


#ifdef	__cplusplus
extern "C" {
#endif

#ifdef	__cplusplus
}
#endif


#endif /* _SQMATRIX_H */

