/*************************************************************************
 * @(#)SqMatrixInverse.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   14 Dec 2011
 *************************************************************************
 */

#include "Except.hh"
#include "SqMatrix.hh"
#include "ValueArray.hh"


/* ***********************************************************************
 *
 * ***********************************************************************
 */
template<>
SqMatrix<double> &SqMatrix<double>::invert()
{

  if(!isSquare()){
    throw Except("inverse: not square\n");
  }
  long m=size();
  ValueArray<double> q12(m,0),q21(m,0),g12(m,0),g21(m,0);

  SqMatrix<double> &A=*this;
    
  if(A(0,0)==0.0){
    throw Except("inverse: matrix indefinite\n");
  }
  A(0,0)=1.0/A(0,0);
  for(long n=2;n<=m;n++){
    long k=n-1;
    
    for(long j=1;j<=k;j++){
      double d=0;
      for(long l=1;l<=k;l++){
	d+=A(j-1,l-1)*A(l-1,n-1);
      }
      q12[j-1]=d;
    }


    for(long j=1;j<=k;j++){
      double d=0;
      for(long l=1;l<=k;l++){
	d+=A(n-1,l-1)*A(l-1,j-1);
      }
      q21[j-1]=d;
    }
    
    double s22=0.;
    for(long l=1;l<=k;l++){
      s22+=A(n-1,l-1)*q12.at(l-1);
    }
    
    double g22=(A(n-1,n-1)-s22);
    if(g22==0.0){
      throw Except("inverse: matrix indefinite");
    }
    g22=1.0/g22;
    A(n-1,n-1)=g22;
    
    for(long j=1;j<=k;j++){
      g21[j-1]=-g22*q21.at(j-1);
      g12[j-1]=-(q12.at(j-1)*g22);
      A(n-1,j-1)=g21.at(j-1);
      A(j-1,n-1)=g12.at(j-1);
    }
    
    for(long i=1;i<=k;i++){
      for(long j=1;j<=k;j++){
	A(i-1,j-1)=A(i-1,j-1)-(q12.at(i-1)*g21.at(j-1));
      }
    } 
  } 

  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static long idamax(long n,double *dx,long incx=1)
{
  // finds the index of element having max. absolute value.
  // jack dongarra, linpack, 3/11/78.
  // dx must hold enough elements such that n increments by incx still
  // point to a valid address of dx.
  long ret=-1;
  if(n<1) return -1;
  if(n==1) return 0;
  ret=0;
  if(incx>1){
    // code for increment not equal to 1
    long ix=0;
    double dmax=fabs(dx[0]);
    ix+=incx;
    for(long i=0;++i<n;){
      if(fabs(dx[ix]) <= dmax) continue;
      ret=i;
      dmax=fabs(dx[ix]);
      ix+=incx;
    }
    return ret;
  }else{
    // code for increment equal to 1
    
    double dmax=fabs(dx[0]);
    for(long i=0;++i<n;){
      if(fabs(dx[i]) <= dmax) continue;
      ret=i;
      dmax=fabs(dx[i]);
    }
  }
  return ret;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
template<>
SqMatrix<Cmplx> &SqMatrix<Cmplx>::invert()
{
  if(!isSquare()){
    throw Except("inverse: not square\n");
  }
  
  
  long n=size();
  
  SqMatrix<Cmplx> w(*this);

  
  long *ind=new long[n];
  double *aux=new double[n];

  
  for(long i=-1;++i<n;){
    for(long ii=i-1;++ii<n;) aux[ii]=w(i,ii).abs();
    long k=idamax(n-i,aux+i)+i;
    ind[i]=k;
    if(k != i){
      for(long j=-1;++j<n;){
	MyMath::swap(w(j,i),w(j,k));
      }
    }
    Cmplx qq=Cmplx::c1/w(i,i);
    w(i,i)=Cmplx::c1;
    for(long ii=-1;++ii<n;) w(ii,i)=qq*w(ii,i);
    for(long j=-1;++j<n;){
      if(j != i) {
	qq=w(i,j);
	w(i,j)=Cmplx();
	for(long ii=-1;++ii<n;) w(ii,j)-=qq*w(ii,i);
      }
    }
  }
    
  for(long i=n;--i>-1;){
    long j=ind[i];
    if(i != j) {
      for(long k=-1;++k<n;){
	MyMath::swap(w(i,k),w(j,k));
      }
    }
  }    
  *this=w;
      
  delete [] ind;
  delete [] aux;
  
  return *this;
}



/*EOF*/ 
