/* @(#)Wigner3J.hh
 * author: magru@unas
 */

#ifndef _WIGNER3J_HH
#define _WIGNER3J_HH 1


#include "ValueArray.hh"
#include "FString.hh"
#include "MyDebug.hh"
#include "MyMath.hh"

class IPrimeDecompositions;
class APInteger;

// ==================================================================
// a collection of prime numbers
// ==================================================================
class PrimeNumbers{
public:
  PrimeNumbers(){}
  ~PrimeNumbers(){}

  inline long size()const {return pprimes.size();}
  inline long operator[](long i){
    if(i>=pprimes.size()) setMaxPrimeIndex(i);
    return pprimes.at(i);
  }
  void setMaxPrime(long n);
  void setMaxPrimeIndex(long n);
  
  void print();
private:
  ValueArray<long> pprimes;
};

// ==================================================================
// represent integer numbers by prime decomposition, usefull
// for Wigner3j to cancel huge factorials
// ==================================================================
class IPrimeDecomposed{
public:
  IPrimeDecomposed(){
    psize=psizeinternal=1;
    pp=new long[1];
    pp[0]=0;
  }
  IPrimeDecomposed(const IPrimeDecomposed &other):pp(0){
    *this=other;
  }
  ~IPrimeDecomposed(){delete [] pp;}
  IPrimeDecomposed &operator=(const IPrimeDecomposed &other){
    if(pp) delete [] pp;
    psize=other.psize;
    psizeinternal=other.psizeinternal;
    pp=new long[psizeinternal];
    for(long j=-1;++j<psizeinternal;) pp[j]=other.pp[j];
    return *this;
  }
  inline long size()const{return psize;}
  // inline void resize(long size){if(size>pp.size()) pp.resize(size,false,0L);}
  inline long &operator[](long i){return pp[i];}
  inline void setZero(){
    setSize(1);
    for(long j=-1;++j<psizeinternal;) pp[j]=0;
  }
  inline void setOne(){
    setSize(1);
    for(long j=0;++j<psizeinternal;) pp[j]=0;
    pp[0]=1;
  }

  void setSize(long newsize){grow(newsize);psize=newsize;}
  void setMinSize(long newsize){grow(newsize);psize=MyMath::max(newsize,psize);}

  IPrimeDecomposed &operator*=(const IPrimeDecomposed &b);
  IPrimeDecomposed &operator/=(const IPrimeDecomposed &b);
  
private:
  void grow(long newsize){// make sure there is space and set new ones to zero
    if(newsize<=psizeinternal) return;

    long *tmp=new long[newsize];
    for(long i=-1;++i<psizeinternal;)tmp[i]=pp[i];
    for(long i=psizeinternal-1;++i<newsize;)tmp[i]=0;
    MyMath::swap(tmp,pp);
    delete [] tmp;
    psizeinternal=newsize;
  }
private:
  long *pp;
  long psize;
  long psizeinternal;

  //  friend class IPrimeDecompositions;
  friend MyDebug operator<<(MyDebug dbg,IPrimeDecomposed &v);
};

// ==================================================================
//
// ==================================================================
inline MyDebug operator<<(MyDebug dbg,IPrimeDecomposed &v)
{
  dbg<<"[";
  for(long i=-1;++i<v.psize;) dbg<<v.pp[i];
  dbg<<"]";
  return dbg;
}


// ==================================================================
// a cache of prime number decompositions
// ==================================================================
class IPrimeDecompositions{
public:
  IPrimeDecompositions(){}
  ~IPrimeDecompositions(){}
  IPrimeDecomposed &number(long x);

  String printS(IPrimeDecomposed &);
  double toDouble(IPrimeDecomposed &);
  long toLong(IPrimeDecomposed &);
  APInteger toAPInteger(IPrimeDecomposed &);

  void fac(long x,IPrimeDecomposed &);
  //void muleq(IPrimeDecomposed &a,IPrimeDecomposed &b);
  //void diveq(IPrimeDecomposed &a,IPrimeDecomposed &b);
  void add(double &s,IPrimeDecomposed &f,IPrimeDecomposed &a);
  void add(APInteger &s,IPrimeDecomposed &f,IPrimeDecomposed &a);

  static void primeDecompose(long x,IPrimeDecomposed &pa,PrimeNumbers &primes);
  
private:
  PrimeNumbers pprimes;
  ValueArray<IPrimeDecomposed> pcache;
  
};




// ==================================================================
// Collection of stuff to calculate Wigner3j symbols 
// ==================================================================
class Wigner3J{
public:
  Wigner3J();
  ~Wigner3J();

  // ----------------------------------------------------------------
  // input is 2*physical momentum, to allow for half integer
  // | 1/2 1/2  1|
  // | 1/2 1/2 -1| =wigner3J(1,1,1,1,2,-2)
  //
  double wigner3J(long j1,long m1,long j2,long m2,long j3,long m3);
  
  // ----------------------------------------------------------------
  // input is 2*physical momentum, to allow for half integer
  static void printW3J(long j1,long m1,long j2,long m2,
		       long j3,long m3,double w);
  

  // ----------------------------------------------------------------
  // input is 2*physical momentum, to allow for half integer
  bool isPhysical(long j1,long m1,long j2,long m2,long j3,long m3);

  static void Wigner3JTest();


private:
  IPrimeDecomposed wignerPre(long j1,long m1,long j2,long m2,long j3,long m3);
  APInteger wignerSu(long j1,long m1,long j2,long m2,long j3,long m3,
		  IPrimeDecomposed &pre);
protected:
private:
  IPrimeDecompositions pipdcmp;
};



#endif /* _WIGNER3J_HH */

