/* @(#)Calculations.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _CALCULATIONS_H
#define _CALCULATIONS_H 1

#include "FString.hh"
#include "Except.hh"
#include "VarNode.hh"
#include "PtrArray.hh"
#include "ValueArray.hh"
#include "CalcDriver.hh"  
#include "MyDebug.hh"  


class Variables;
class MethodControl;
class PointStatistics;
template <class T>
class SqMatrix;

/* ***********************************************************************
 *
 * ***********************************************************************
 */
class Vartable{
 public:

  Vartable():pvtable(new VarNode()){};
  ~Vartable(){delete pvtable;};

  String replaceVariablesIn(const String &);
  
  void clear(){pvtable->clear();};
  void defineVariable(const String &name,double value){
    if(!pvtable->define(name,value)){
      throw Except("you cannot overwrite predefined constants\n");
    }
  };
  void parse(const String &text){
    CalcDriver calcdriver;
    calcdriver.setTraceParsing(false);
    calcdriver.setTraceScanning(false);
    long ret=calcdriver.parse(pvtable,text());
    if(ret){
      throw Except(calcdriver.errorMsg());
    }
  };

  void print(const String &text=""){
    printf("%s\n%s",text(),pvtable->printS()());
  }
 private:
  VarNode *pvtable;
  friend MyDebug operator<<(MyDebug dbg, const Vartable &v);
};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
class Calculation{
public:
  enum Status{New=1,Created=2,Ready=3,Running=4,Stoped=5,NotUsed=6};
  
  Calculation():pid(-1),pnorder(-1),pdir(""),pstatus(New),psc(false),
		pused(false),petot(0.),piterationprogress(1e30){};
  Calculation(long id,long norder,const ValueArray<double> &values,
	      const String &dir,Status status,bool sc,bool used)
    :pid(id),pnorder(norder),pvalues(values),pdir(dir),pstatus(status),
     psc(sc),pused(used),petot(0.),
     piterationprogress(1e30){};
  
  long ID()const{return pid;};
  void setID(long id){pid=id;};
  
  
  const String &dir()const{return pdir;};
  void setDir(const String &d){pdir=d;};
  
  Status status()const{return pstatus;};
  void setStatus(Status s){pstatus=s;};
  void setStatusFromString(const String &s){
    for(int i=0;i<pmaxstatusinfo;i++){
      if(s==pstatusinfo[i].guiname){
	pstatus=pstatusinfo[i].type;
	return;
      }
    }
    throw Except(Sprintf("unknown status flag '%s'\n",s())); 
  };
  void setSc(bool s){psc=s;};
  void setUsed(bool u){pused=u;};
  bool used()const{return pused;};

  String result(long i)const{
    if(i<presults.size()){
      return presults[i];
    }else{
      return "undef";
    }
  };
  void setResult(long i,const String &s){
    presults.adjustSize(i+1);
    presults[i]=s;
  }
  
  long order()const{return pnorder;};
  void setOrder(long o){pnorder=o;};
  
  double etot()const{return petot;};
  void setEtot(double e){petot=e;};
  
  
  
  double iterationProgress()const{return piterationprogress;};
  void setIterationProgress(double e){piterationprogress=e;};
  
  const String &termination()const{return ptermination;};
  void setTermination(const String &t){ptermination=t;};
  
  
  
  ValueArray<double> &values(){return pvalues;};
  const ValueArray<double> &values()const{return pvalues;};
  void setValues(const ValueArray<double> &v){pvalues=v;};
  
  
  void writeToFile(FILE *FH){
    fprintf(FH,
	    "calculation:\n"
	    "\tid=%ld\n"
	    "\tdir=\"%s\"\n"
	    "\tstatus=\"%s\"\n"
	    "\tsc=%d\n"
	    "\tx=["
	    ,pid,pdir(),statusGuiName(pstatus),psc
	    );
    for(long i=0;i<pvalues.size();i++){
      fprintf(FH,"%30.15e",pvalues.at(i));
      if(i<pvalues.size()-1)
	fprintf(FH,",");
    }
    fprintf(FH,"\t]\n");
    fprintf(FH,"endcalculation:\n");
    
  };
  
  static const char *statusGuiName(Status s){
    for(int i=0;i<pmaxstatusinfo;i++){
      if(s==pstatusinfo[i].type){
	return pstatusinfo[i].guiname;
      }
    }
    return "undef";};
  
  
private:
  struct tsStatusInfo{
    Status type;
    const char *guiname;
  };
  static const tsStatusInfo pstatusinfo[];
  static const long pmaxstatusinfo;
  
  
private:
  long pid;
  long pnorder;
  ValueArray<double> pvalues;
  String pdir;
  Status pstatus;
  bool psc,pused;
  double petot;
  double piterationprogress;
  String ptermination;
  StringList presults;
  friend MyDebug operator<<(MyDebug dbg, const Calculation &c);
};
/* ***********************************************************************
 *
 * ***********************************************************************
 */
class Calculations{
 public:
  
  class Min1DPoint{
  public:
    enum Type{Unknown=0,Guess=1,Old=2};
    Min1DPoint():value(0.),id(-1),type(Unknown){};
    Min1DPoint(double d,const Type &t):value(d),id(-1),type(t){};
    Min1DPoint(long idd,const Type &t):value(0.),id(idd),type(t){};
	double value;
	long id;
	Type type;
  };
  
  Calculations(MethodControl *control):pcontrol(control),
    pstatusfilename("=.status_min"),pdidone(false),pweightpower(2){};

  void clear(){pcalculations.clear();};
  
  void readStatusFromFile();
  void writeStatusToFile(bool tostdout=false);

  // register calculation, create input and return id
  long registerCalculation(long norder,long idsc=-1);
  // launch the calculations
  void run();
  // check each calc and finish the calc if not finished
  void initCheckCalculations(); //todo
  
  bool stopRunningJobs(); //returns true if some where stoped
  void restartJobs(bool onlyunfinished=false);
  void attemptRefresh(bool registernew=false);
  void sortCalculationOrder();
  void deleteUnusedDirs();

  void minimize1D();
  void minimizeSimplex(); 
  void minimizeMulti1D();

  void analyseDirectories();

  void forceFileTransfer();

 private:
  void writePipeFile();
  void executePerjob();

  // if this vars.currentValues() is already defined in some calc
  // return its id
  long valuesAlreadyDefined();

  // return the calculation with this id or 0
  Calculation *getById(long id);
  
  //return the calculation witht the closest set od values;
  String getClosestCalcDir();
  Calculation *getClosestCalc();

  long  getFreshId();

  void setupCurrentVarTable();

  void createFploInput(Calculation *c,const String &from_dir);

  void reportResults();

  void sortCalculation1D();
  
  void findBracket_1D(ValueArray<Min1DPoint> initialPoints,
		      ValueArray<long> &bracket,
		      long idsc=-1);
  void brent(ValueArray<long> bracketids,long idsc=-1);
  void yminInCalcs(long &imin,double &ymin);

  double func(double x,long idsc=-1);

  void createStartSimplex(const ValueArray<double> &start_point_vec,
			  const ValueArray<double> &start_scale_vec);

  void printCalculations();
  void sortCalcsAccordingToY();
  bool calcGeneralExtremum(ValueArray<double> &start_point_vec,
			   ValueArray<double> &start_scale_vec);
  void writeResult(const Calculation *c0);
  void calcInterpolatorGeneral(ValueArray<double> &B,
			       SqMatrix<double> &A,
			       const PointStatistics &pointstats);
  void classifyAllPoints(long &idmin,long &idmax2,long &idmax);
  double func(const ValueArray<double> &x,long idsc=-1);
  
  void respondToQuitRequest();

  String constructDirName(long id,bool asmask=false);


 private:
  MethodControl *pcontrol;
  String pstatusfilename;
  PtrArray<Calculation> pcalculations;
  Vartable pvartable;
  bool pdidone;
  bool pinterpolindef;
  double pweightpower;
  long piterationcounter;
  long pnbetween;
  friend  MyDebug operator<<(MyDebug dbg, const Calculations &p);
};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
inline MyDebug operator<<(MyDebug dbg, const Vartable &v)
{
  dbg<< *(v.pvtable);
  return dbg;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
inline MyDebug operator<<(MyDebug dbg, const Calculation &c)
{
  dbg<<"id:"<<c.pid
     <<"values:"<<c.pvalues
     <<"dir:"<<c.pdir
     <<"status:"<<c.statusGuiName(c.pstatus)
     <<"used:"<<c.pused
     <<"results:"<<c.presults
     <<"iterationprogress:"<<c.piterationprogress
     <<"termination:"<<c.ptermination
     <<"etot:"<<c.petot
     <<"\n";
  return dbg;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
inline MyDebug operator<<(MyDebug dbg, const Calculations &p)
{
  dbg<<"\nstart Calculations\n\n";
  dbg<< "start vartable:\n"<<p.pvartable<< "end vartable\n\n";
  dbg<< "statusfile:"<<p.pstatusfilename<<"\n";
  dbg<< "start Calculations array:\n"<<p.pcalculations<< "end Calculations array\n\n";
  dbg<<"\nend Calculations\n";
  return dbg;

}



#ifdef	__cplusplus
extern "C" {
#endif

#ifdef	__cplusplus
}
#endif


#endif /* _CALCULATIONS_H */

