/* @(#)Variables.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _VARIABLES_H
#define _VARIABLES_H 1

#include <stdio.h>
#include "FString.hh"
#include "ValueArray.hh"
#include "MinEnums.hh"
#include "MyDebug.hh"



/* ***********************************************************************
 *
 * ***********************************************************************
 */

class Value{
 public:
  enum Type{Single=1,Range=2};
  Value():ptype(Single),pvalue(0.){};
  Value(double d):ptype(Single),pvalue(d){};
  Value(double d[3]):ptype(Range),pmin(d[0]),pmax(d[1]),pstepwidth(d[2]){};

  inline Type type()const{return ptype;};
  inline double value()const{return pvalue;};
  inline double min()const{return pmin;};
  inline double max()const{return pmax;};
  inline double stepWidth()const{return pstepwidth;};

  String printS()const{
    if(ptype==Single)
      return Sprintf("%g",pvalue);
    else
      return Sprintf("[%g %g %g]",pmin,pmax,pstepwidth);
  };
  bool check(String &msg)const{
    if(ptype==Range){
      if(pmin>=pmax){
	msg="min>=max";
	return false;
      }
      if(pstepwidth<=0){
	msg="step width<=0";
	return false;
      }
    }
    return true;
  };
    
 private:
  Type ptype;
  double pvalue;
  double pmin,pmax,pstepwidth;

};

/* ***********************************************************************
 * describe a variable
 * ***********************************************************************
 */



class Variable{
 public:
  Variable():pname(""),pxfrmt(""),pmaxstep(0.){
    pbracket[0]=pbracket[1]=pbracket[2]=0.;};
  ~Variable(){};
  
  inline String prefix()const{return pname+"=";};

  inline const String &name()const{return pname;};
  inline void setName(const String & n){pname=n;};
  inline const String &xformat()const{return pxfrmt;};
  inline void setXformat(const String & f){pxfrmt=f;};
  inline const double *bracket()const{return pbracket;};
  inline double *bracket(){return pbracket;};
  inline double maxStep()const{return pmaxstep;};
  inline void setMaxStep(double s){ pmaxstep=s;};

  inline const ValueArray<Value> &values()const{return pvalues;};
  inline ValueArray<Value> &values(){return pvalues;};
  
  ValueArray<double> getValues()const{
    double tol=1.0e-6;
    ValueArray<double> va;  
    for(int i=0;i<pvalues.size();i++){
      const Value &v=pvalues.at(i);
      if(v.type()==Value::Single){
	if(!va.contains(v.value(),1.0e-12)){
	  va.append(v.value());
	}
      }else if(v.type()==Value::Range){
	double d=v.min();
	while(d<=v.max()+tol*v.stepWidth()){
	  if(!va.contains(d,1.0e-12)){
	    va.append(d);
	  }
	  d+=v.stepWidth();
	}
      }
    }
    va.qSort();
    return va;
  };

 private:
  String pname;
  String pxfrmt;
  double pbracket[3];
  double pmaxstep;
  ValueArray<Value> pvalues;
  friend MyDebug operator<<(MyDebug dbg, const Variable &v);
};

/* ***********************************************************************
 * the collections of all variables
 * ***********************************************************************
 */
class Variables{
 public:
  
  Variables(){};
  ~Variables(){};

  inline long size()const{return pvariables.size();};
  inline const Variable &at(long i)const{return pvariables.at(i);};

  inline void clear(){pvariables.clear();};

  void check(MinEnums::Method method);

  void writeToFile(FILE *FH,MinEnums::Method method);
  
  Variable *appendNew();

  void print(FILE *FH,MinEnums::Method method);


  void setCurrentValue(double d){
    pcurrentvalues.resize(1);
    pcurrentvalues[0]=d;
  };

  void setCurrentValue(const ValueArray<double> &va){
    pcurrentvalues=va;
  };

  //
  // return the current values as a double-vector
  //
  inline const ValueArray<double> &currentValues()const{
    return pcurrentvalues;
  };
  inline ValueArray<double> &currentValues(){
    return pcurrentvalues;
  };

  inline String stringify(bool placeholders=false)const{
    return stringify(pcurrentvalues,false,placeholders);
  }
  String stringify(const ValueArray<double> &x,bool valueonly=false,
		   bool placeholders=false)const{
    String s="";
    for(int iv=0;iv<pvariables.size();iv++){
      if(valueonly){
	String fmt=Sprintf("%s",pvariables.at(iv).xformat()());
	if(iv<x.size())
	  s+=Sprintf(fmt(),x.at(iv));
	else
	  s+=Sprintf(fmt(),0.);
      }else{
	if(placeholders){
	  s+=Sprintf("%s=*",pvariables.at(iv).name()());
	}else{
	  String fmt=Sprintf("%%s=%s",pvariables.at(iv).xformat()());
	  if(iv<x.size())
	    s+=Sprintf(fmt(),pvariables.at(iv).name()(),x.at(iv));
	  else
	    s+=Sprintf(fmt(),pvariables.at(iv).name()(),0.);
	}
      }
      if(iv<pvariables.size()-1) s+=" ";
    }
    return s;

  };

 private:
  ValueArray<Variable> pvariables;
  ValueArray<double> pcurrentvalues;

  friend MyDebug operator<<(MyDebug dbg, const Variables &v);

};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
class GridLooper{
 public:
  //
  // initialize the looper, which creates a grid from all varaibles and 
  // their values
  //
  GridLooper(Variables *vs,bool lastrunsfirst=false)
    :pvariables(vs),pcounter(0),pvalues(0),plastrunsfirst(lastrunsfirst){
    pnvar=pvariables->size();
    pcounter=new long[pnvar];
    pvalues=new ValueArray<double>[pnvar];
    pvariables->currentValues().resize(pnvar);
    for(int iv=0;iv<pnvar;iv++) {
      pvalues[iv]=pvariables->at(iv).getValues();
      pvariables->currentValues()[iv]=0.;
      pcounter[iv]=0;
    }
    if(plastrunsfirst)
      pcounter[pnvar-1]=-1;
    else
      pcounter[0]=-1;
  };
  // 
  // cleanup
  //
  ~GridLooper(){
    if(pcounter) delete [] pcounter; pcounter=0;
    if(pvariables) delete [] pvalues; pvalues=0;
  };
  

  //
  // while(nextNtuple()){}; loops over all possible combinations of values
  //
  bool nextNtuple(){
    bool thatstheend=false;
    long start=0;
    long end=pnvar;
    long incr=1;
    if(plastrunsfirst){
      start=pnvar-1;
      incr=-1;
      end=-1;
    }
    long jc=start;
    while((plastrunsfirst && jc>=0)||(!plastrunsfirst && jc<pnvar)){
      pcounter[jc]++;
      if(pcounter[jc]>=pvalues[jc].size()){
	pcounter[jc]=0;
	jc=jc+incr;
      }else{
	break;
      }
      if(jc==end) thatstheend=true;
    }
    if(!thatstheend){
      pvariables->currentValues().resize(pnvar);
      for(int iv=0;iv<pnvar;iv++){
	pvariables->currentValues()[iv]=pvalues[iv].at(pcounter[iv]);
      }
    }
    return !thatstheend;
  };


  //
  // print the current index counter as String
  //
  String printSCounter(){
    String s="counter: ";
    for(int iv=0;iv<pnvar;iv++){
      s+=Sprintf(" %ld",pcounter[iv]);
    }
    return s;
  };
 private:
  Variables *pvariables;
  long pnvar;
  long *pcounter;
  ValueArray<double> *pvalues;
  bool plastrunsfirst;
};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
inline MyDebug operator<<(MyDebug dbg, const Variable &v)
{
  dbg<<v.pname;
  return dbg;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
inline MyDebug operator<<(MyDebug dbg, const Variables &v)
{
  dbg<<"vars"<<v.pvariables<<"values"<<v.pcurrentvalues;
  return dbg;
}
#ifdef	__cplusplus
extern "C" {
#endif

#ifdef	__cplusplus
}
#endif


#endif /* _VARIABLES_H */

