/*************************************************************************
 * @(#)OutGrep.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   28 Sep 2011
 *************************************************************************
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include "OutGrep.hh"
#include "Assorted.hh"
#include "Except.hh"
#include "ValueArray.hh"
#include "MyDebug.hh"


OutGrep::tsModes OutGrep::modes[]={
  {"EE",TotalEnergy,"total energy"},
  {"SSat",AtomSpin,"atom spin"},
  {"SS",TotalSpin,"total spin"},
  {"time",CycleTime,"cycle time"},
  {"term",Termination,"termination"},
  {"it",Iteration,"last deviation"},
  {"S_gros",AtomSpinGros,"atom gross spin"},
  {"S_net",AtomSpinNet,"atom net spin"},
  {"N_gros",AtomOccuGros,"atom gross occupation"},
  {"N_net",AtomOccuNet,"atom net occupation"},
  {"PP",Pressure,"pressure"},
  {"Bfsm",Bfsm,"FSM B-field"},
  {"dBfsm",dBfsm,"change of FSM B-field"},
  {"fit",ForceIteration,"force last deviation"},
  {"EFG",AtomEFG,"atom EFG"},
  {"Lzat",AtomOrbitalMoment,"atom orbital moment"},
};
long OutGrep::maxmodes=sizeof(modes)/sizeof(tsModes);

static const char *unkownmode="unkonw mode";


/* ***********************************************************************
 *
 * ***********************************************************************
 */
class MiniString{
public:

  MiniString():pinternalsize(1),pdata((char*)malloc(sizeof(char)*pinternalsize)){
    *pdata='\0';
  }

  ~MiniString(){if(pdata) free(pdata); pdata=0;pinternalsize=0;}

  inline char* operator()(){return pdata;}

  char *subString(const char *src,long i0,long i1=-1)
  {
    long len;
    if(i1>=0)len=i1-i0+2; else len=strlen(src+i0)+1;
    if(len>pinternalsize){
      pinternalsize=len;
      pdata=(char*)realloc(pdata,sizeof(char)*pinternalsize);
    }
    strncpy(pdata,src+i0,len);
    pdata[len-1]='\0';
    return pdata;
  }

  char *nextWord(const char *src,const char *delims=" \t")
  {
    long len;

    const char *p0=src+strspn(src," \t");
    const char *p1=p0+strcspn(p0,delims);

    len=p1-p0+1;
    if(len>pinternalsize){
      pinternalsize=len;
      pdata=(char*)realloc(pdata,sizeof(char)*pinternalsize);
    }
    strncpy(pdata,p0,len);
    pdata[len-1]='\0';
    return pdata;
  }

private:
  long pinternalsize;
  char *pdata;
};


/* ***********************************************************************
 *
 * ***********************************************************************
 */
static long find(const char *in,const char *what)
{
  const char *p;
  if((p=strstr(in,what))==NULL) return -1;
  return p-in;
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */

String OutGrep::SiteInfo::getElementfromSite(const ValueArray<SiteInfo>&sites,
					     long site)
{
  for(int i=0;i<sites.size();i++){
    if(sites.at(i).site==site){
      return sites.at(i).el;
    }
  }
  return "??";
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
OutGrep::Mode OutGrep::modeFromString(const char *name)
{
  String n=String(name).lowerCase();
  //myDebug()<<"n"<<n;
  for(int i=0;i<maxmodes;i++){
    String mn=String(modes[i].name).lowerCase();
    //myDebug()<<"mn"<<mn;
    if(n.find(mn)==0) return (Mode)modes[i].mode;
    mn=String(modes[i].altname).lowerCase();
    //myDebug()<<"mn"<<mn;
    if(mn.length()>0 && n.find(mn)==0) return (Mode)modes[i].mode;
  }
  throw Except(Sprintf("Mode with name '%s' does not exist\n",name));
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
const char *OutGrep::modeName(Mode mode,bool alternativestring)
{
  for(int i=0;i<maxmodes;i++){
    if(modes[i].mode==mode){ 
      if(alternativestring)
	return modes[i].altname;
      else
	return modes[i].name;
    }
  }
  return unkownmode;

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
String OutGrep::modeString(Mode mode,bool alternativestring)
{
  if(mode==AtomSpin || mode==AtomEFG || mode==AtomOrbitalMoment){
    return Sprintf("%s site %ld",modeName(mode,alternativestring),psite);
  }else if(mode==AtomSpinGros||mode==AtomSpinNet||mode==AtomOccuGros||
	   mode==AtomOccuNet){
    return Sprintf("%s site %ld orbital %ld",
		   modeName(mode,alternativestring),psite,porbital);
  }else{
    return String(modeName(mode,alternativestring));
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
OutGrep::OutGrep()
  :poutfile("out"),pbuffer(0),plines(0),plinessize(0),
   pnlines(0),psite(0),porbital(-1),psilentopen(false)
{
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
OutGrep::OutGrep(const String &outfile)
  :poutfile(outfile),pbuffer(0),plines(0),plinessize(0),
   pnlines(0),psite(0),porbital(-1),psilentopen(false)
{
  
  readFile();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void OutGrep::clear()
{
  if(pbuffer) delete [] pbuffer; pbuffer=0;
  if(plines) free(plines); plines=0;
  plinessize=0;
  pnlines=0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void OutGrep::readFile()
{
  clear();

  FILE *FH;
  
  try{
    FH=openforread(poutfile);
  }catch(Except &ex){
    if(psilentopen) return;
    throw ex;
  }

  struct stat buf;
  if(fstat(fileno(FH),&buf)!=0) throw Except(Sprintf("Cannot stat ",poutfile()));
  
  pbuffer=new char[buf.st_size+1];
  
  size_t result=0;
  if( ((result=fread(pbuffer,(size_t)1,buf.st_size,FH))==0) && ferror(FH) )
    throw  Except("Error reading file \n");
  
  pbuffer[buf.st_size]='\0';

  plinessize=1000;
  plines=(char**)malloc(sizeof(char*)*plinessize);

  char *p=pbuffer;
  pnlines=0;
  char *pn;
  while(*p!='\0'){
    pn=strchr(p,'\n');
    if(pn==NULL){
      pn=p+strlen(p);
    }else{
      *pn='\0';
    }
    if(pnlines>=plinessize){
      plinessize*=2;
      plines=(char**)realloc(plines,sizeof(char*)*plinessize);
    }
    plines[pnlines]=p;
    p=pn;p++;
    pnlines++;
  }
  
  fclose(FH);
  //myDebug()<<"read file";
  
  //for(long i=0;i<pnlines;i++) myDebug()<<plines[i];

  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool OutGrep::grepLast(Mode mode,double &result)
{
  //myDebug()<<"grepLast(Mode mode,double &result)";
  ValueArray<double> x,yd;
  ValueArray<String> ys;
  grepinternal(mode,x,yd,ys,false);
  if(yd.size()>0){
    result=yd.last();
    return true;
  }else{
    result=0.;
    return false;
  }
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool OutGrep::grepLast(Mode mode,String &result)
{
  //myDebug()<<"grepLast(Mode mode,String &result)"<<mode;
  ValueArray<double> x;
  ValueArray<double> yd;
  ValueArray<String> ys;
  grepinternal(mode,x,yd,ys,true);
  if(ys.size()>0){
    result=ys.last();
    return true;
  }else{
    result="";
    return false;
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void OutGrep::grep(Mode mode,ValueArray<double> &x,ValueArray<double> &yd)
{
  ValueArray<String> ys;
  x.resize(0);
  yd.resize(0);
  grepinternal(mode,x,yd,ys,false);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void OutGrep::grep(Mode mode,ValueArray<double> &x,ValueArray<String> &ys)
{
  ValueArray<double> yd;
  x.resize(0);
  ys.resize(0);
  grepinternal(mode,x,yd,ys,true);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void OutGrep::grepinternal(Mode mode,ValueArray<double> &x,
			   ValueArray<double> &yd,ValueArray<String> &ys,
			   bool asstring)
{
  //myDebug() << "grepinternal(ValueArray<double>)";
  //myDebug() << "mode"<<mode<<"site"<<psite;
  long n=0;
  bool startEFG=false;
  int j0;
  double d;
  const char *pl;
  MiniString res;
  char *p,*endptr;
  long i=-1;
  while(++i<pnlines){
    pl=plines[i];
    //printf("i=%ld %s\n",i,pl);
    d=0;
    if(mode==Iteration && find(pl,"last dev")>=0){
      j0=find(pl,"u=");
      if(j0>=0){
 	j0+=2;
      }else if((j0=find(pl,"deviation="))>=0){
	j0+=10;
      }
      if(j0>=0){
	x.append(++n);
	p=res.nextWord(pl+j0);
	if(asstring) ys.append(p); else yd.append(strtod(p,NULL));
      }
    }else if(mode==IterationBest && find(pl,"last deviation=")>=0){
      j0=find(pl,"best=");
      if(j0>=0){
 	j0+=5;
      }
      if(j0>=0){
	x.append(++n);
	p=res.nextWord(pl+j0);
	if(asstring) ys.append(p); else yd.append(strtod(p,NULL));
      }
    }else if(mode==ForceIteration && (j0=find(pl,"|force| last dev"))>=0){
      x.append(++n);
      p=res.nextWord(pl+j0+17);
      if(asstring) ys.append(p); else yd.append(strtod(p,NULL));
      
    }else if(mode==TotalEnergy && (j0=find(pl,"EE:"))>=0){
      x.append(++n);
      p=res.nextWord(pl+j0+3);
      if(asstring) ys.append(p); else yd.append(strtod(p,NULL)); 
      
    }else if(mode==Bfsm && ((j0=find(pl,"Fermi energy:"))>=0)){
      if((j0=find(pl,"Bfsm="))>=0){
	x.append(++n);
	p=res.nextWord(pl+j0+5);
	if(asstring) ys.append(p); else yd.append(strtod(p,NULL)); 
	
      }
    }else if(mode==dBfsm && ((j0=find(pl,"Fermi energy:"))>=0)){
      if((j0=find(pl,"dB="))>=0){
	x.append(++n);
	p=res.nextWord(pl+j0+3," \t;");
	if(asstring) ys.append(p); else yd.append(strtod(p,NULL)); 
	
      }
    }else if(mode==Pressure && ((j0=find(pl,"Pressure:"))>=0)){
      x.append(++n);
      p=res.nextWord(pl+j0+9);
      if(asstring) ys.append(p); else yd.append(strtod(p,NULL)); 
      
    }else if(mode==TotalSpin && ((j0=find(pl,"total gross spin"))>=0)){
      x.append(++n);
      j0=find(pl+j0,":");
      p=res.nextWord(pl+j0+1);
      if(asstring) ys.append(p); else yd.append(strtod(p,NULL)); 
      
    }else if(mode==CycleTime && ((j0=find(pl,"fplo cycle"))>=0)){
      x.append(++n);
      long j1=find(pl+j0+10,"cpu time:");
      p=res.nextWord(pl+j0+j1+19);
      if(asstring) ys.append(p); else yd.append(strtod(p,NULL)); 
      
    }else if(
	     ((mode==AtomSpinGros || mode==AtomSpin) && ((j0=find(pl,"S_gros"))>=0))||
	     ((mode==AtomSpinNet) && ((j0=find(pl,"S_net"))>=0))||
	     (mode==AtomOccuGros && ((j0=find(pl,"N_gros"))>=0))||
	     (mode==AtomOccuNet && ((j0=find(pl,"N_net"))>=0))   ){
      
      StringList sl=String(pl+j0+9).stripBlanks().split(" ");
      long site,orb=porbital;
      if(mode==AtomSpin)orb=0;
      if(String::islong(&site,sl[0]())){
	if(site==psite&&orb>=0){
	  x.append(++n);
	  if(orb<=0 || orb>=sl.size())orb=sl.size()-(mode==AtomOccuGros?2:1);
	  try{
	    d=sl[orb].convertToFortranReal();
	  }catch(...){}
	  if(asstring) ys.append(sl[orb]); else yd.append(d); 
	}
      }
    }else if(mode==AtomOrbitalMoment){
      if ((j0=find(pl,"ONSITE ORBITAL MOMENTS"))>=0){
	long nn=0;
	while(++i<pnlines){
	  pl=plines[i];
	  if(nn==1) break;
	  nn++;
	}
	while(++i<pnlines){
	  pl=plines[i];
	  if(strstr(pl,"|-")==pl) { i--;break;}
	  StringList sl(String(pl).replace("|"," ",0,true).split(" \t"));
	  long site;
	  if(String::islong(&site,sl[1]())&&site==psite){
	    x.append(++n);
	    if(asstring) ys.append(sl[2]); else yd.append(strtod(sl[2](),NULL));
	  }
	}
      }
    }else if(mode==AtomEFG){
      if ((j0=find(pl,"EFG:"))>=0){
	//myDebug()<<pl<<startEFG;
	if((j0=find(pl,"Site"))>=0){
	  //myDebug()<<"SITE"<<pl+j0+4;
	  const char *po=strchr(pl+j0+4,'(');
	  const char *pc=strchr(pl+j0+4,')');
	  if(po!=NULL and pc!=NULL){
	    //myDebug()<<"EL"<<res.subString(pl+j0+4,0,po-pl-j0-5);
	    p=res.subString(po,1,pc-po-1);
	    //myDebug()<<"SI"<<p;
	    long site=strtol(res(),&endptr,10);
	    if(*(res())!='\0' && *endptr=='\0'){
	      //myDebug()<<"valid site"<<site;
	      if(site==psite){
		startEFG=true;
	      }else{
		startEFG=false;
	      }
	    }
	  }
	}
	if(startEFG && (j0=find(pl,"Sum of all onsite orbital contributions"))>=0){
	  startEFG=false;
	}
	if(startEFG && (j0=find(pl,"Vzz ["))>=0){
	  if((j0=find(pl,"eta"))>=0){
	    p=res.nextWord(pl+j0+3);
	    //myDebug()<<"Vzz"<<p;
	    x.append(++n);
	    if(asstring) ys.append(p); else yd.append(strtod(p,NULL));
	  }
	  startEFG=false;
	}
      }
    }else if(mode==Termination &&  (j0=find(pl,"TERMINATION:"))>=0){
      x.append(++n);
      p=res.subString(pl+j0+12,0);
      if(asstring) ys.append(p); 
      else {
	if(find(pl+j0,"Finished :")>=0) yd.append(1); else yd.append(0);
      }
      
    }
  }
       
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

ValueArray<OutGrep::SiteInfo> OutGrep::extractSites()
{
  //myDebug() << "OutGrep::extractSites()";
  ValueArray<SiteInfo> sites;
  int j0;
  const char *pl;
  for(long i=0;i<pnlines;i++){
    pl=plines[i];
    if(((j0=find(pl,"N_gros"))>=0)){
      StringList sl=String(pl+j0+6).stripBlanks().split(" ");
      long site;
      if(String::islong(&site,sl[1]())){
	//myDebug() << site;
	sites.append(SiteInfo(sl[0],site));
      }
    }else if(find(pl,"total gross electron number")>=0){
      break;
    }
  }
  return sites;
}




/*EOF*/ 
