/*************************************************************************
 * @(#)SqMatrixDiagno.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   13 Dec 2011
 *************************************************************************
 */

#include "SqMatrix.hh"
#include "ValueArray.hh"
#include "Except.hh"
#include "MyMath.hh"


/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void rbtridi(long n,SqMatrix<double> &ar,ValueArray<double> &d,
		    ValueArray<double> &e,ValueArray<double> &e2,
		    ValueArray<double> &tau0,
		    ValueArray<double> &tau1
		    )
{
  
  /*  ***********************************************************************
   *- Reduces a real symmetric matrix to a real symmetric tridiagonal
   *- matrix using unitary similarity transformations
   * ----------------------------------------------------------------------
   *i Inputs:
   *i   n     :order of matrix
   *io  ar    :real symmetric input matrix
   *io         only the upper triangle of the matrix need be supplied
   *o Outputs:
   *io  ar    :contains information about the unitary transformations used
   *io         in the reduction (in strict upper triangle)
   *io         the strict lower triangle and the diagonal unaltered
   *o   d     :contains the diagonal elements of the the tridiagonal matrix
   *o   e     :contains the subdiagonal elements of the tridiagonal matrix
   *o          in its last n-1 positions.  e(1) is set to zero.
   *o   e2    :contains the squares of the corresponding elements of e
   *r          e2 may coincide with e if the squares are not needed.
   *o   tau   :contains further information about the transformations
   *r Remarks:
   *r     this subroutine is a translation of a complex analogue of
   *r     the algol procedure tred1, num. math. 11, 181-195(1968)
   *r     by martin, reinsch, and wilkinson.
   *r     handbook for auto. comp., vol.ii-linear algebra, 212-226(1971).
   *r     this version is adapted from the eispack htridi. In contrast
   *r     only the upper triangle of a is used for input and output
   *r     questions and comments should be directed to burton s. garbow,
   *r     mathematics and computer science div, argonne national laboratory
   * ----------------------------------------------------------------------
   */
  long i,j,k,im1,jp1;
  double f,g,h,hh,scale;
  
  tau0[n-1]=1.0;
  
  for(i=1;i<=n;i++){
    d[i-1]=ar(i-1,i-1);
  }
  
  for(i=n;i>=1;i--){
    im1=i-1;
    h=0.0;
    scale=0.0;
    if(im1 >= 1) {
      // -------- scale row
      for(k=1;k<=im1;k++){
	scale=scale+fabs(ar(k-1,i-1)) ;
      }
      if(scale !=  0.0) goto l140;
      tau0[im1-1]=1.0;
    }
    e[i-1]=0.0;
    e2[i-1]=0.0;
    goto l290;
    
  l140:;
    for(k=1;k<=im1;k++){
      ar(k-1,i-1)=ar(k-1,i-1)/scale;
      h=h+ar(k-1,i-1)*ar(k-1,i-1) ;
    }
    
    e2[i-1]=scale*scale*h;
    g=sqrt(h);
    e[i-1]=scale*g;
    f=sqrt(::pow(ar(im1-1,i-1),2));
    // ------- form next diagonal element of matrix t
    if(f != 0.0) {
      tau0[im1-1]=(-ar(im1-1,i-1)*tau0[i-1])/f;
      
      h=h+f*g;
      g=1.0+g/f;
      ar(im1-1,i-1)=g*ar(im1-1,i-1);
      
      if(im1==1) goto l270;
    }else{
      tau0[im1-1]=-tau0[i-1];
      ar(im1-1,i-1)=g;
    }
    f=0.0;
    
    for(j=1;j<=im1;j++){
      g=0.0;
      
      // --------- form element of a*u
      for(k=1;k<=j;k++){
	g=g+ar(k-1,j-1)*ar(k-1,i-1) ;
	
      }
      jp1=j+1;
      if(im1  >= jp1) {
	for(k=jp1;k<=im1;k++){
	  g=g+ar(j-1,k-1)*ar(k-1,i-1) ;
	}
      }
      // --------- form element of p
      e[j-1]=g/h;
      f=f+e[j-1]*ar(j-1,i-1) ;
    }
    
    hh=f/(h+h);
    // ------ form reduced a
    for(j=1;j<=im1;j++){
      f=ar(j-1,i-1);
      g=e[j-1]-hh*f;
      e[j-1]=g;
      
      for(k=1;k<=j;k++){
	ar(k-1,j-1)=ar(k-1,j-1)-f*e[k-1]-g*ar(k-1,i-1) ;
      }
    }
    
  l270:;
    for(k=1;k<=im1;k++){
      ar(k-1,i-1)=scale*ar(k-1,i-1);
    }
    
  l290:;
    hh=d[i-1];
    d[i-1]=ar(i-1,i-1);
    ar(i-1,i-1)=hh;
    tau1[i-1]=scale*sqrt(h);
    
  }
}  


/* ***********************************************************************
 *
 * ***********************************************************************
 */
static long imtql2(ValueArray<double> &d,ValueArray<double> &e,
		   SqMatrix<double> &z)
{
  /* ***********************************************************************
   *
   *     this subroutine is a translation of the algol procedure imtql2,
   *     num. math. 12, 377-383(1968) by martin and wilkinson,
   *     as modified in num. math. 15, 450(1970) by dubrulle.
   *     handbook for auto. comp., vol.ii-linear algebra, 241-248(1971).
   *
   *     this subroutine finds the eigenvalues and eigenvectors
   *     of a symmetric tridiagonal matrix by the implicit ql method.
   *     the eigenvectors of a full symmetric matrix can also
   *     be found if  tred2  has been used to reduce this
   *     full matrix to tridiagonal form.
   *
   *     on input
   *
   *        nm must be set to the row dimension of two-dimensional
   *          array parameters as declared in the calling program
   *          dimension statement.
   *
   *        n is the order of the matrix.
   *
   *        d contains the diagonal elements of the input matrix.
   *
   *        e contains the subdiagonal elements of the input matrix
   *          in its last n-1 positions.  e(1) is arbitrary.
   *
   *        z contains the transformation matrix produced in the
   *          reduction by  tred2, if performed.  if the eigenvectors
   *          of the tridiagonal matrix are desired, z must contain
   *          the identity matrix.
   *
   *      on output
   *
   *        d contains the eigenvalues in ascending order.  if an
   *          error exit is made, the eigenvalues are correct but
   *          unordered for indices 1,2,...,return-value-1.
   *
   *        e has been destroyed.
   *
   *        z contains orthonormal eigenvectors of the symmetric
   *          tridiagonal (or full) matrix.  if an error exit is made,
   *          z contains the eigenvectors associated with the stored
   *          eigenvalues.
   *
   *        return-value is set to
   *          zero       for normal return,
   *          j          if the j-th eigenvalue has not been
   *                     determined after 30 iterations.
   *
   *     calls dcabs2 for  dsqrt(a*a + b*b) .
   *
   *     questions and comments should be directed to burton s. garbow,
   *     mathematics and computer science div, argonne national laboratory
   *
   *     this version dated august 1983.
   *
   *     -----------------------------------------------------------------
   */

  long n=z.size();
  
  long i,j,k,l,m,ii,mml;
  double b,c,f,g,p,r,s,tst1,tst2;
  
  if(n==1) return 0;

  
  for(i=2;i<=n;i++){
    e[i-1-1]=e[i-1];
  }
  
  e[n-1]=0.0;
  
  for(l=1;l<=n;l++){
    j=0;
    //     .......... look for small sub-diagonal element ..........
  l105:;
    for(m=l;m<=n;m++){
      if(m==n) goto l120;
      tst1=fabs(d[m-1])+fabs(d[m+1-1]);
      tst2=tst1+fabs(e[m-1]);
      if(tst2==tst1) goto l120;
    }
    
  l120:;
    
    p=d[l-1];
    if(m==l) goto l240;
    if(j==30) return l;
    j=j+1;
    //     .......... form shift ..........
    g=(d[l+1-1]-p)/(2.0*e[l-1]);
    r=sqrt(g*g+1.0*1.0);
    g=d[m-1]-p+e[l-1]/(g+dSign(r,g));
    s=1.0;
    c=1.0;
    p=0.0;
    mml=m-l;
    //     .......... for i=m-1 step -1 until l do -- ..........
    for(ii=1;ii<=mml;ii++){
      i=m-ii;
      f=s*e[i-1];
      b=c*e[i-1];
      r=sqrt(f*f+g*g);
      e[i+1-1]=r;
      if(r==0.0) goto l210;
      s=f/r;
      c=g/r;
      g=d[i+1-1]-p;
      r=(d[i-1]-g)*s+2.0*c*b;
      p=s*r;
      d[i+1-1]=g+p;
      g=c*r-b;
      //     .......... form vector ..........
      for(k=1;k<=n;k++){
	f=z(k-1,i+1-1);
	z(k-1,i+1-1)=s*z(k-1,i-1)+c*f;
	z(k-1,i-1)=c*z(k-1,i-1)-s*f;
	
      }
      
    }
    
    d[l-1]=d[l-1]-p;
    e[l-1]=g;
    e[m-1]=0.0;
    goto l105;
    //     .......... recover from underflow ..........
  l210:;
    d[i+1-1]=d[i+1-1]-p;
    e[m-1]=0.0;
    goto l105;
  l240:;
  }
  //     .......... order eigenvalues and eigenvectors ..........
  for(ii=2;ii<=n;ii++){
    i=ii-1;
    k=i;
    p=d[i-1];
    
    for(j=ii;j<=n;j++){
      if(d[j-1] >= p) goto l260;
      k=j;
      p=d[j-1];
    l260:;
    }
    
    if(k==i) goto l300;
    d[k-1]=d[i-1];
    d[i-1]=p;
    
    for(j=1;j<=n;j++){
      p=z(j-1,i-1);
      z(j-1,i-1)=z(j-1,k-1);
      z(j-1,k-1)=p;
    }
    
  l300:;
  }
  
  return 0;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

static void rbtribk(long n,const SqMatrix<double> &ar,
		    const ValueArray<double> &tau0,
		    const ValueArray<double> &tau1,
		    long m,SqMatrix<double> &zr)
{
  /* ***********************************************************************
   *- Get eigenvectors of a real symmetric matrix by back transforming
   *- those of the corresponding real symmetric tridiagonal matrix
   * ----------------------------------------------------------------------
   *i Inputs:
   *i   n     :order of the matrix
   *i   ar    :contains information about the unitary transformations used
   *i          in the reduction by rbtridi (in strict upper triangle)
   *i   ai    :contains information about the unitary transformations used
   *i          in the reduction by rbtridi (in full upper triangle)
   *i   tau   :contains further information about the transformations
   *i         (output from rbtridi)
   *i   m     :number of eigenvectors to be back transformed
   *i   zr    :contains the eigenvectors to be back transformed
   *o Outputs:
   *o   zr    :contain  the transformed eigenvectors
   *r Remarks:
   *r    this subroutine is a translation of a complex analogue of
   *r    the algol procedure trbak1, num. math. 11, 181-195(1968)
   *r    by martin, reinsch, and wilkinson.
   *r    handbook for auto. comp., vol.ii-linear algebra, 212-226(1971).
   *r
   *r    this subroutine forms the eigenvectors of a real symmetric
   *r    matrix by back transforming those of the corresponding
   *r    real symmetric tridiagonal matrix determined by  rbtridi.
   *r
   *r    note that the last component of each returned vector
   *r    is real and that vector euclidean norms are preserved.
   *r
   *r    questions and comments should be directed to burton s. garbow,
   *r    mathematics and computer science div, argonne national laboratory
   *r
   *r    this version is adapted from the eispack htridi. In contrast
   *r    only the upper triangle of a is used for input
   * ----------------------------------------------------------------------
   */
  long i,j,k,l;
  double h,s,smin;
  
  if(m==0) return;
  // --- transform the eigenvectors of the real symmetric
  // --- tridiagonal matrix to those of the symmetric tridiagonal matrix.
  
  for(j=1;j<=m;j++){
    for(k=1;k<=n;k++){
      zr(k-1,j-1)=zr(k-1,j-1)*tau0.at(k-1);
    }
  }
  
  if(n==1) return;
  // --- recover and apply the householder matrices
  for(i=2;i<=n;i++){
    l=i-1;
    h=tau1.at(i-1);
    if(h  !=  0.0) {
      for(j=1;j<=m;j++){
	s=0.0;
	for(k=1;k<=l;k++){
	  s=s+ar(k-1,i-1)*zr(k-1,j-1) ;
	}
	
	// --------- double divisions avoid possible underflow
	s=(s/h)/h;
	smin=-s;
	
	for(k=1;k<=l;k++){
	  zr(k-1,j-1)=zr(k-1,j-1)+smin*ar(k-1,i-1) ;
	}
      }
    }
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static long imtqlv(long n,const ValueArray<double> &d,
		   const ValueArray<double> &e,
		   ValueArray<double> &e2,
		   ValueArray<double> &w,
		   ValueArray<long> &ind,
		   ValueArray<double> &rv1)
{
  /* ***********************************************************************
   *- Finds the eigenvalues of a symmetric tridiagonal matrix by the
   *- implicit QL method
   * ----------------------------------------------------------------------
   *i Inputs:
   *i   n     :order of matrix
   *i   d     :diagonal elements of matrix
   *i   e     :subdiagonal elements of matrix in its last n-1 positions
   *i   e2    :squares of the corresponding elements of e
   *o Outputs:
   *io  e2    :neglegible elements have been replaced by zero
   *i   w     :eigenvalues in ascending order
   *i   ind   :index to submatrix
   *i   return-value  :nonzero (j) if the j-th eigenvalue has not been determined
   *i          after 30 iterations
   *i   rv1   :work file
   *r Remarks:
   *r    THIS ROUTINE IS A VARIANT OF  IMTQL1  WHICH IS A TRANSLATION OF
   *r    ALGOL PROCEDURE IMTQL1, NUM. MATH. 12, 377-383(1968) BY MARTIN AND
   *r    WILKINSON, AS MODIFIED IN NUM. MATH. 15, 450(1970) BY DUBRULLE.
   *r    HANDBOOK FOR AUTO. COMP., VOL.II-LINEAR ALGEBRA, 241-248(1971).
   *r    QUESTIONS AND COMMENTS SHOULD BE DIRECTED TO BURTON S. GARBOW,
   *r    MATHEMATICS AND COMPUTER SCIENCE DIV, ARGONNE NATIONAL LABORATORY
   *r    THIS VERSION DATED AUGUST 1983.
   *r
   *r    IND CONTAINS THE SUBMATRIX INDICES ASSOCIATED WITH THE
   *r    CORRESPONDING EIGENVALUES IN W -- 1 FOR EIGENVALUES
   *r    BELONGING TO THE FIRST SUBMATRIX FROM THE TOP,
   *r    2 FOR THOSE BELONGING TO THE SECOND SUBMATRIX, ETC..
   *r
   *r    IF AN ERROR EXIT IS MADE, THE EIGENVALUES ARE CORRECT AND
   *r    ORDERED FOR INDICES 1,2,...ret-val-1, BUT MAY NOT BE
   *r    THE SMALLEST EIGENVALUES.
   * ----------------------------------------------------------------------
   */
  long i,j,k,l,m,ii,mml,tag;
  double b,c,f,g,p,r,s,tst1,tst2;
  
  k=0;
  tag=0;
  
  for(i=1;i<=n;i++){
    w[i-1]=d.at(i-1);
    if(i != 1) rv1[i-1-1]=e.at(i-1);
  }
  
  e2[1-1]=0.0;
  rv1[n-1]=0.0;
  
  for(l=1;l<=n;l++){
    j=0;
    //     .......... LOOK FOR SMALL SUB-DIAGONAL ELEMENT ..........
  l105:;
    for(m=l;m<=n;m++){
      if(m==n) goto l120;
      tst1=fabs(w[m-1])+fabs(w[m+1-1]);
      tst2=tst1+fabs(rv1[m-1]);
      if(tst2==tst1) goto l120;
      // .......... GUARD AGAINST UNDERFLOWED ELEMENT OF E2 ..........
      if(e2[m+1-1]==0.0) goto l125;
    }
    
  l120:;
    if(m <= k) goto l130;
    if(m != n) e2[m+1-1]=0.0;
  l125:;
    k=m;
    tag=tag+1;
  l130:;
    p=w[l-1];
    if(m==l) goto l215;
    if(j==30) return l;
    j=j+1;
    //     .......... FORM SHIFT ..........
    g=(w[l+1-1]-p)/(2.0*rv1[l-1]);
    r=sqrt(g*g+1.0*1.0);
    g=w[m-1]-p+rv1[l-1]/(g+dSign(r,g));
    s=1.0;
    c=1.0;
    p=0.0;
    mml=m-l;
    //     .......... FOR I=M-1 STEP -1 UNTIL L DO -- ..........
    for(ii=1;ii<=mml;ii++){
      i=m-ii;
      f=s*rv1[i-1];
      b=c*rv1[i-1];
      r=sqrt(f*f+g*g);
      rv1[i+1-1]=r;
      if(r==0.0) goto l210;
      s=f/r;
      c=g/r;
      g=w[i+1-1]-p;
      r=(w[i-1]-g)*s+2.0*c*b;
      p=s*r;
      w[i+1-1]=g+p;
      g=c*r-b;
      
    }
    
    w[l-1]=w[l-1]-p;
    rv1[l-1]=g;
    rv1[m-1]=0.0;
    goto l105;
    //     .......... RECOVER FROM UNDERFLOW ..........
  l210:;
    w[i+1-1]=w[i+1-1]-p;
    rv1[m-1]=0.0;
    goto l105;
    //     .......... ORDER EIGENVALUES ..........
  l215:;
    if(l==1) goto l250;
    //     .......... FOR I=L STEP -1 UNTIL 2 DO -- ..........
    for(ii=2;ii<=l;ii++){
      i=l+2-ii;
      if(p >= w[i-1-1]) goto l270;
      w[i-1]=w[i-1-1];
      ind[i-1]=ind[i-1-1];
    }
    
  l250:;
    i=1;
  l270:;
    w[i-1]=p;
    ind[i-1]=tag;
  }
  
  return 0;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static long binvit(long n,const ValueArray<double> &d,
		   const ValueArray<double> &e,
		   const ValueArray<double> &e2,
		   long m,
		   const ValueArray<double> &w,
		   const ValueArray<long> &ind,
		   SqMatrix<double> &z,
		   ValueArray<double> &rv1,ValueArray<double> &rv2,
		   ValueArray<double> &rv3,ValueArray<double> &rv4,
		   ValueArray<double> &rv6)
{
  /* ***********************************************************************
   *- Finds eigenvectors of a tridiagonal symmetric matrix
   * ----------------------------------------------------------------------
   *i Inputs:
   *i   n     :order of input matrix
   *i   d     :diagonal elements of input matrix
   *i   e     :subdiagonal elements of input matrix in position 2,...,n-1
   *i   e2    :square of e
   *i   m     :number of specified eigenvalues
   *i   w     :eigenvalues
   *i   ind   :index to submatrix
   *w Work arrays:
   *w   rv1   :work array
   *w   rv2   :work array
   *w   rv3   :work array
   *w   rv4   :work array
   *w   rv6   :work array
   *o Outputs:
   *o   z     :orthonormal eigenvector
   *o   ret-val  :0 for normal return
   *r Remarks:
   *r    AUTHORS -
   *r       THIS ROUTINE IS A MODIFICATION OF EISPACK ROUTINE TINVIT
   *r       WHICH IS A TRANSLATION OF THE INVERSE ITERATION TECHNIQUE
   *r       IN THE ALGOL PROCEDURE TRISTURM BY PETERS AND WILKINSON.
   *r       HANDBOOK FOR AUTO. COMP., VOL.II-LINEAR ALGEBRA, 418-439(1971).
   *r       THE EISPACK VERSION WAS MODIFIED BY C. MOLER AND D. SPANGLER
   *r       (NRCC - 1 APR 1979).  THIS VERSION IS BY S. T. ELBERT
   *r       (AMES LABORATORY - USDOE), WHO FURTHER STREAMLINED THE CODE.
   *r
   *r    PURPOSE -
   *r       THIS ROUTINE FINDS THOSE EIGENVECTORS OF A TRIDIAGONAL
   *r       SYMMETRIC MATRIX CORRESPONDING TO SPECIFIED EIGENVALUES,
   *r       USING INVERSE ITERATION.
   *r
   *r    METHOD -
   *r       THE CALCULATIONS PROCEED AS FOLLOWS.  FIRST, THE  E2  ARRAY
   *r       IS INSPECTED FOR THE PRESENCE OF A ZERO ELEMENT DEFINING A
   *r       SUBMATRIX.  THE EIGENVALUES BELONGING TO THIS SUBMATRIX ARE
   *r       IDENTIFIED BY THEIR COMMON SUBMATRIX INDEX IN  IND.
   *r
   *r       THE EIGENVECTORS OF THE SUBMATRIX ARE THEN COMPUTED BY
   *r       INVERSE ITERATION.  FIRST, THE  LU  DECOMPOSITION OF THE SUB-
   *r       MATRIX WITH AN APPROXIMATE EIGENVALUE SUBTRACTED FROM ITS
   *r       DIAGONAL ELEMENTS IS ACHIEVED BY GAUSSIAN ELIMINATION USING
   *r       PARTIAL PIVOTING.  THE MULTIPLIERS DEFINING THE LOWER
   *r       TRIANGULAR MATRIX  L  ARE STORED IN THE TEMPORARY ARRAY  RV4
   *r       AND THE UPPER TRIANGULAR MATRIX  U  IS STORED IN THE THREE
   *r       TEMPORARY ARRAYS  RV1,  RV2, AND  RV3.  SAVING THESE
   *r       QUANTITIES IN RV1,  RV2,  RV3, AND  RV4 AVOIDS REPEATING
   *r       THE  LU  DECOMPOSITION IF FURTHER ITERATIONS ARE REQUIRED.
   *r       AN APPROXIMATE VECTOR, STORED IN  RV6, IS COMPUTED STARTING
   *r       FROM AN INITIAL VECTOR, AND THE NORM OF THE APPROXIMATE
   *r       VECTOR IS COMPARED WITH A NORM OF THE SUBMATRIX TO DETERMINE
   *r       WHETHER THE GROWTH IS SUFFICIENT TO ACCEPT IT AS AN
   *r       EIGENVECTOR.  IF THIS VECTOR IS ACCEPTED, ITS EUCLIDIAN NORM
   *r       IS MADE 1.  IF THE GROWTH IS NOT SUFFICIENT, THIS VECTOR IS
   *r       USED AS THE INITIAL VECTOR IN COMPUTING THE NEXT APPROXIMATE
   *r       VECTOR.  THIS ITERATION PROCESS IS REPEATED AT MOST  5  TIMES.
   *r
   *r       EIGENVECTORS COMPUTED IN THE ABOVE WAY CORRESPONDING TO WELL-
   *r       SEPARATED EIGENVALUES OF THIS SUBMATRIX WILL BE ORTHOGONAL.
   *r       HOWEVER, EIGENVECTORS CORRESPONDING TO CLOSE EIGENVALUES OF
   *r       THIS SUBMATRIX MAY NOT BE SATISFACTORILY ORTHOGONAL.  HENCE,
   *r       TO INSURE ORTHOGONAL EIGENVECTORS, EACH APPROXIMATE VECTOR IS
   *r       MADE ORTHOGONAL TO THOSE PREVIOUSLY COMPUTED EIGENVECTORS
   *r       WHOSE EIGENVALUES ARE CLOSE TO THE CURRENT EIGENVALUE, AS
   *r       DETERMINED BY THE PARAMETER GRPTOL.
   *r       IF THE ORTHOGONALIZATION PROCESS PRODUCES A ZERO VECTOR, A
   *r       COLUMN OF THE IDENTITY MATRIX IS USED AS AN INITIAL VECTOR
   *r       FOR THE NEXT ITERATION.
   *r
   *r       IDENTICAL EIGENVALUES ARE PERTURBED SLIGHTLY IN AN ATTEMPT TO
   *r       OBTAIN INDEPENDENT EIGENVECTORS.  THESE PERTURBATIONS ARE NOT
   *r       RECORDED IN THE EIGENVALUE ARRAY  W.
   *r
   *r       THE ABOVE STEPS ARE REPEATED ON EACH SUBMATRIX UNTIL ALL THE
   *r       EIGENVECTORS ARE COMPUTED.
   *r
   *r    ON ENTRY-
   *r
   *r       N      - INTEGER
   *r                THE ORDER OF THE MATRIX.
   *r
   *r       D      - W.P. REAL (N)
   *r                THE DIAGONAL ELEMENTS OF THE INPUT MATRIX.
   *r
   *r       E      - W.P. REAL (N)
   *r                CONTAINS THE SUBDIAGONAL ELEMENTS OF THE INPUT MATRIX
   *r                IN ITS LAST N-1 POSITIONS.  E(1) IS ARBITRARY.
   *r
   *r       E2     - W.P. REAL (N)
   *r          CONTAINS THE SQUARES OF THE CORRESPONDING ELEMENTS OF E,
   *r          WITH ZEROS CORRESPONDING TO NEGLIGIBLE ELEMENTS OF E.
   *r          E(I) IS CONSIDERED NEGLIGIBLE IF IT IS NOT LARGER THAN
   *r          THE PRODUCT OF THE RELATIVE MACHINE PRECISION AND THE SUM
   *r          OF THE MAGNITUDES OF D(I) AND D(I-1).  E2(1) MUST CONTAIN
   *r          0.0 IF THE EIGENVALUES ARE IN ASCENDING ORDER, OR 2.0
   *r          IF THE EIGENVALUES ARE IN DESCENDING ORDER.  IF  BISECT,
   *r          TRIDIB, OR  EIMQLV  HAS BEEN USED TO FIND THE EIGENVALUES,
   *r          THEIR OUTPUT E2 ARRAY IS EXACTLY WHAT IS EXPECTED HERE.
   *r
   *r       M      - INTEGER
   *r                THE NUMBER OF SPECIFIED EIGENVALUES.
   *r
   *r       W      - W.P. REAL (M)
   *r                THE M EIGENVALUES IN ASCENDING OR DESCENDING ORDER.
   *r
   *r       IND    - INTEGER (M)
   *r           CONTAINS IN ITS FIRST M POSITIONS THE SUBMATRIX INDICES
   *r           ASSOCIATED WITH THE CORRESPONDING EIGENVALUES IN W --
   *r           1 FOR EIGENVALUES BELONGING TO THE FIRST SUBMATRIX FROM
   *r           TOP, 2 FOR THOSE BELONGING TO THE SECOND SUBMATRIX, ETC.
   *r
   *r    ON EXIT-
   *r
   *r       Z      - W.P. REAL (NM,M)
   *r          CONTAINS THE ASSOCIATED SET OF ORTHONORMAL EIGENVECTORS.
   *r          ANY VECTOR WHICH FAILS TO CONVERGE IS SET TO ZERO.
   *r
   *r       ret-val   - INTEGER
   *r                SET TO
   *r                ZERO   FOR NORMAL RETURN,
   *r                -R     IF THE EIGENVECTOR CORRESPONDING TO THE R-TH
   *r                       EIGENVALUE FAILS TO CONVERGE IN 5 ITERATIONS.
   *r
   *r       RV1, RV2, RV3, RV4, AND RV6  -  W.P. REAL (N)
   *r                TEMPORARY STORAGE ARRAYS.
   *r
   *r    COMPLEXITY -
   *r       O(N*M)
   *r
   *r    NOTE -
   *r    QUESTIONS AND COMMENTS ABOUT EISPACK SHOULD BE DIRECTED TO
   *r    B. S. GARBOW, APPLIED MATHEMATICS DIVISION, ARGONNE NATIONAL LAB.
   *r
   *-----------------------------------------------------------------------
   */
  long i,igroup,ip,ipm1,ipp1,iq,iqm1,iqm2,iqp1,ir,is,itag,iters;
  long j,jj,nsub,nsubup;
  double anorm,anorm2,epmach,eps2,eps3,eps4,order;
  double test,u,uk,v,xu,rtpre,rtcur;
  
  long ierr=0;
  if(m  <=  0) return 0;
  // --- EPMACH IS A MACHINE DEPENDENT PARAMETER SPECIFYING
  // --- THE RELATIVE PRECISION OF FLOATING POINT ARITHMETIC.
  epmach=::pow(0.5,20);
  for(i=1;i<=100;i++){
    epmach=epmach*0.5;
    test=1.0+epmach;
    if(test==1.0) goto l120;
  }
 l120:;
  epmach=epmach+epmach;
  epmach=epmach+epmach;
  
  itag=0;
  order=1.0-e2.at(1-1);
  ip=1;
  for(iq=1;iq<=n;iq++){
    // this a bugfix for the case when the compiler does not know
    // to stop boolean evaluation after the result is definite
    // in this case the e2.at(...-1) statement produces a segfault
    if( iq != n) {
      if(  e2.at(iq+1-1)  !=  0.0 ) continue;
    }
    // end bugfix
    //skip: if(iq==n .Or. e2.at(iq+1-1)==0.0) {
    // -- ESTABLISH AND PROCESS NEXT SUBMATRIX WITH INDEX RANGE IP TO IQ
    itag=itag+1;
    is=0;
    eps2=1.0;
    eps3=1.0;
    eps4=1.0;
    uk=1.0;
    igroup=0;
    rtcur=w.at(1-1);
    for(ir=1;ir<=m;ir++){
      if(ind.at(ir-1)==itag) {
	rtpre=rtcur;
	rtcur=w.at(ir-1);
	ipp1=ip+1;
	ipm1=ip-1;
	nsub=iq-ipm1;
	if(is==0) {
	  // ----------- CHECK FOR ISOLATED ROOT
	  xu=1.0;
	  if(ip==iq) {
	    // ------------- SET ISOLATED ROOT TO UNIT VECTOR
	    for(i=1;i<=n;i++){
	      z(i-1,ir-1)=0.0;
	    }
	    z(ip-1,ir-1)=1.0;
	    goto l920;
	  }
	  anorm=fabs(d.at(ip-1));
	  for(i=ipp1;i<=iq;i++){
	    anorm=anorm+fabs(d.at(i-1))+fabs(e.at(i-1));
	  }
	  // ----------- EPS2 IS THE CRITERION FOR GROUPING,
	  // ----------- EPS3 REPLACES ZERO PIVOTS AND EQUAL ROOTS ARE
	  // ----------- MODIFIED BY EPS3
	  // ----------- EPS4 IS TAKEN VERY SMALL TO AVOID OVERFLOW
	  uk=1.*nsub;
	  anorm2=anorm/uk;
	  eps2=1.e-5*anorm2;
	  eps3=epmach*anorm;
	  eps4=uk*eps3;
	  uk=eps4/sqrt(uk);
	  is=ip;
	  igroup=0;
	  // ----------- LOOK FOR CLOSE OR COINCIDENT ROOTS
	}else{
	  igroup=igroup+1;
	  if(fabs(rtcur-rtpre)  >=  eps2) igroup =0;
	  if(order*(rtcur-rtpre)  <=  0.0) rtcur=rtpre+order*eps3;
	  //------ ELIMINATION WITH INTERCHANGES AND INITIALIZATION OF
	  //------ VECTOR
	}
	v=0.0;
	iqm1=iq-1;
	rv6[ip-1]=uk;
	u=d.at(ip-1)-rtcur;
	if(ip  != iq) {
	  v=e.at(ipp1-1);
	  if(ipp1  <=  iqm1) {
	    for(i=ipp1;i<=iqm1;i++){
	      rv6[i-1]=uk;
	      if(fabs(e.at(i-1))  >=  fabs(u)) {
		// WARNING -- A DIVIDE CHECK MAY OCCUR HERE
		// IF E2 ARRAY HAS NOT BEEN SPECIFIED CORRECTLY
		xu=u/e.at(i-1);
		rv4[i-1]=xu;
		rv1[i-1-1]=e.at(i-1);
		rv2[i-1-1]=d.at(i-1)-rtcur;
		rv3[i-1-1]=e.at(i+1-1);
		u=v-xu*rv2[i-1-1];
		v=-xu*rv3[i-1-1];
	      }else{
		xu=e.at(i-1)/u;
		rv4[i-1]=xu;
		rv1[i-1-1]=u;
		rv2[i-1-1]=v;
		rv3[i-1-1]=0.0;
		u=d.at(i-1)-rtcur-xu*v;
		v=e.at(i+1-1);
	      }
	    }
	  }
	  rv6[iq-1]=uk;
	  rv3[iqm1-1]=0.0;
	  if(fabs(e.at(iq-1))  >=  fabs(u)) {
	    xu=u/e.at(iq-1);
	    rv4[iq-1]=xu;
	    rv1[iqm1-1]=e.at(iq-1);
	    rv2[iqm1-1]=d.at(iq-1)-rtcur;
	    u=v-xu*rv2[iqm1-1];
	    v=0.0;
	  }else{
	    xu=e.at(iq-1)/u;
	    rv4[iq-1]=xu;
	    rv1[iqm1-1]=u;
	    rv2[iqm1-1]=v;
	    u=d.at(iq-1)-rtcur-xu*v;
	  }
	}
	if(fabs(u)  <  eps3) u=eps3;
	rv1[iq-1]=u;
	rv2[iq-1]=0.0;
	rv3[iq-1]=0.0;
	// --------- BACK SUBSTITUTION FOR I=IQ STEP -1 UNTIL IP DO
	for(iters=1;iters<=5;iters++){
	  rv6[iq-1]=rv6[iq-1]/rv1[iq-1];
	  u=rv6[iq-1];
	  rv6[iqm1-1]=(rv6[iqm1-1]-u*rv2[iqm1-1])/rv1[iqm1-1];
	  v=u;
	  u=rv6[iqm1-1];
	  anorm=fabs(u)+fabs(v);
	  if(ipp1  <=  iqm1) {
	    iqm2=iqm1-1;
	    for(i=iqm2;i>=ip;i--){
	      rv6[i-1]=(rv6[i-1]-u*rv2[i-1]-v*rv3[i-1])/rv1[i-1];
	      v=u;
	      u=rv6[i-1];
	      anorm=anorm+fabs(u);
	    }
	  }
	  if(igroup  != 0) {
	    // ORTHOGONALIZE WITH RESPECT TO PREVIOUS MEMBERS OF 
	    // GROUP
	    j=ir;
	    for(jj=1;jj<=igroup;jj++){
	    l630:;
	      j=j-1;
	      if(ind.at(j-1)  != itag) goto l630;
	      xu=0.0;
	      nsubup=nsub+ip-1;
	      for(i=ip;i<=nsubup;i++){
		xu=xu+rv6[i-1]*z(i-1,j-1);
	      }
	      xu=-xu;
	      if(xu  != 0.0) {
		for(i=ip;i<=nsubup;i++){
		  rv6[i-1]=rv6[i-1]+xu*z(i-1,j-1);
		}
	      }
	    }
	    anorm=0.0;
	    for(i=ip;i<=iq;i++){
	      anorm=anorm+fabs(rv6[i-1]);
	    }
	  }
	  if(anorm  <  1.0) {
	    // ------------- FORWARD SUBSTITUTION
	    if(anorm==0.0) {
	      rv6[is-1]=eps4;
	      is=is+1;
	      if(is > iq) is=ip;
	    }else{
	      xu=eps4/anorm;
	      for(i =ip;i <=iq;i ++){
		rv6[i-1]=rv6[i-1]*xu;
	      }
	    }
	    //  ELIMINATION OPERATIONS ON NEXT VECTOR ITERATE
	    for(i=ipp1;i<=iq;i++){
	      u=rv6[i-1];
	      // -- IF RV1(I-1)==E(I), A ROW INTERCHANGE
	      // -- WAS PERFORMED EARLIER IN THE
	      // -- TRIANGULARIZATION PROCESS
	      if(rv1[i-1-1]==e.at(i-1)) {
		u=rv6[i-1-1];
		rv6[i-1-1]=rv6[i-1];
	      }
	      rv6[i-1]=u-rv4[i-1]*rv6[i-1-1];
	    }
	  }else{
	    goto l840;
	  }
	}
	// --------- SET ERROR -- NON-CONVERGED EIGENVECTOR
	ierr=-ir;
	for(i=1;i<=n;i++){
	  z(i-1,ir-1)=0.0;
	}
	goto l920;
	// -------NORMALIZE SO THAT SUM OF SQUARES IS 1 AND EXPAND 
	//TO FULL ORDER
      l840:;
	u=0.0;
	for(i=ip;i<=iq;i++){
	  u=u+rv6[i-1]*rv6[i-1];
	}
	xu=1.0/sqrt(u);
	if(ipm1 >= 1) {
	  for(i=1;i<=ipm1;i++){
	    z(i-1,ir-1)=0.0;
	  }
	}
	for(i=ip;i<=iq;i++){
	  z(i-1,ir-1)=rv6[i-1]*xu;
	}
	iqp1=iq+1;
	if(iqp1  <=  n) {
	  for(i=iqp1;i<=n;i++){
	    z(i-1,ir-1)=0.0;
	  }
	}
      }
    l920:;
    }
    ip=iq+1;
    //} skip
  }
  return ierr;
  
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void diagonalizeSymmetric(const SqMatrix<double> &A,ValueArray<double> &eband,
			  SqMatrix<double> &Z,String &msg)
{
  const char *nsubfunc="diagonalizeSymmetric: real";
  msg="";

  if(!A.isSquare())
    throw Except(Sprintf("%s: called with non-square matrix",nsubfunc));
  
  long nvdim=A.size();
  SqMatrix<double> h=A;

  ValueArray<long> iwk(nvdim,0);
  ValueArray<double> d(nvdim,0.);
  ValueArray<double> e(nvdim,0.);
  ValueArray<double> e2(nvdim,0.);
  ValueArray<double> e3(nvdim,0.);
  ValueArray<double> rv1(nvdim,0.);
  ValueArray<double> rv2(nvdim,0.);
  ValueArray<double> rv3(nvdim,0.);
  ValueArray<double> rv4(nvdim,0.);
  ValueArray<double> rv5(nvdim,0.);
  ValueArray<double> tau0(nvdim,0.);
  ValueArray<double> tau1(nvdim,0.);


  // --- Reduction of a real symmetric matrix to a real symmetric
  // --- tridiagonal matrix using unitary similarity transformations

  rbtridi(nvdim,h,d,e,e2,tau0,tau1);


  // ----- Make unit eigenvector matrix ---
  if(Z.size()!=nvdim) Z.resize(nvdim);
  if(eband.size()!=nvdim) eband.resize(nvdim);
  
  Z=0.0;
  for(long i=0;i<nvdim;i++){
    Z(i,i)=1.0;
  }
  eband=d;

  // ----- Get eigenvalues and eigenvectors of symmetric tridiag. matrix
  // ----- by implicit QL method
  e3=e;
  long ierr=imtql2(eband,e3,Z);
  

  if(ierr != 0){
    msg="error in IMTQL2: trying another subroutine.";

    ierr=imtqlv(nvdim,d,e,e2,eband,iwk,rv1);

    

    if(ierr != 0) {
      throw Except(Sprintf("%s: error in IMTQLV.",nsubfunc));
    }
    ierr=binvit(nvdim,d,e,e2,nvdim,eband,iwk,Z,rv1,rv2,rv3,rv4,rv5);
    if(ierr != 0) {
      throw Except(Sprintf("%s: error in BINVIT.",nsubfunc));
    }
  }

    
  // --- Get eigenvectors of real symmetric matrix by back transforming
  // --- those of the corresponding real symmetric tridiagonal matrix
  rbtribk(nvdim,h,tau0,tau1,nvdim,Z);


}



/*EOF*/ 
