Module bandmatrix
  Use PrecisionDefs
  Use errors
  !!
  Private
  !!@\authors{Numerical Recipies, Klaus Koepernik}
  !!@\name{module}{bandmatrix}
  !!@\synopsis{
  !!@  use this module to solve Ax=b for band matrices
  !!@}
  !!@\description{
  !!@}
  !!@\bugs{}
  !!@\seealso{}
  !!@\changes{}
  Public :: bandec,banbks,bandsolve,symtridig_solv
  Character(*),Parameter::nmodule="BANDMATRIX"  
Contains
  !!***********************************************************************
  !!@\authors{Klaus Koepernik}
  !!@\name{modulesubroutine}{bandsolve}
  !!@\synopsis{
  !!@ bandsolve(a,b,n,mleft,mright)
  !!@}
  !!@\description{
  !!@  return solution of Ax=b for a band matrix A
  !!@  wrapper around bandec and bandbks
  !!@  if only one solution is needed call bandsolve otherwise
  !!@  call bandec first and banbks subsequently.
  !!@  remark: A is changed after call to bandsolve.
  !!@}
  !!@\bugs{}
  !!@\seealso{}
  !!@\changes{}
  !!
  Subroutine bandsolve(a,b,n,mleft,mright)
    !!
    Implicit None 
    !!IO
    !!@\io{n}{row dimension}
    Integer(t_int),Intent(in)  :: n 
    !!@\io{mleft}{Number of columns left to the diagonal.}
    Integer(t_int),Intent(in)  :: mleft
    !!@\io{mright}{Number of columns right to the diagonal.}
    Integer(t_int),Intent(in)  :: mright 
    !!@\io{a}{band matrix a.}
    Real(t_real),Intent(inout),Dimension(:,:)  :: a
    !!@\io{b}{right hand site of Ax=b.}
    Real(t_real),Intent(inout),Dimension(:)  :: b
    !!LOCAL
    !Character(*),Parameter::nsubfunc="bandsolve"
    !Integer(t_int) :: ndim
    !Integer(t_int) :: mdim
    Integer(t_int),Dimension(n) :: indx
    Real(t_real),Dimension(Size(a,1),Size(a,2)) :: al
    Real(t_real) :: d
    !ndim=size(a,1)
    !mdim=size(a,2)
    call bandec(a,n,mleft,mright,al,indx,d)
    call banbks(a,n,mleft,mright,al,indx,b)
    !
  end subroutine bandsolve

  !***********************************************************************
  !                                                                       
  Subroutine banbks(a,n,m1,m2,al,indx,b) 
    !!@\authors{Numerical Recipies, Klaus Koepernik}
    !!@\name{modulesubroutine}{banbks}
    !!@\synopsis{
    !!@}
    !!@\description{
    !!@ Back substitution of the Ax=b LU decompostition of a band matrix
    !!@ with minimal pivoting                     
    !!@  A(i,j).ne.0.0  if i-m1<=j<=i+m2                                 
    !!@  stored as A(i,(1:m1,m1+1,m1+2:m1+1+m2))
    !!@                                                                   
    !!@  return x: A*x=L*U*x=b                                            
    !!@  b is replaced with c                                             
    !!@  a(i,j) contians U                                                
    !!@  al(i,j) contians L                                               
    !!@  indx contains the permutation due to pivoting                    
    !!@}
    !!@\bugs{}
    !!@\seealso{bandec}
    !!@\changes{}
    !!@\copyright{(C) Copr. 1986-92 Numerical Recipes Software}
    !!

    !!
    Implicit None 
    !!IO                                                                     
    Integer(t_int),Intent(in):: m1,m2,n
    Integer(t_int),Intent(in) ,Dimension(:):: indx
    Real(t_real),Intent(in),Dimension(:,:)  :: a
    Real(t_real),Intent(in),Dimension(:,:)  :: al
    Real(t_real),Intent(inout),Dimension(:)  :: b

    !!LOCAL                                                                  
    Character(*),Parameter::nsubfunc="banbks"
    Integer(t_int):: i,k,l,mm 
    Real(t_real) :: dum 
    !                             

    mm=m1+m2+1_t_int
    If(mm>Size(a,dim=2_t_int).Or.m1>Size(al,dim=2_t_int)&
         &.Or.n>Size(a,dim=1_t_int))Then
       Call errmsg(nmodule,nsubfunc,E_fatalerror&
            &,msg="'bad arguments, sorry!'")
    End If
    
    l=m1 
    Do k=1_t_int,n 
       i=indx(k) 
       If(i/=k)Then 
          dum=b(k) 
          b(k)=b(i) 
          b(i)=dum 
       Endif
       If(l<n)l=l+1_t_int
       Do i=k+1_t_int,l 
          b(i)=b(i)-al(k,i-k)*b(k) 
       Enddo
    Enddo
    
    l=1_t_int
    Do i=n,1_t_int,-1_t_int
       dum=b(i) 
       Do k=2_t_int,l 
          dum=dum-a(i,k)*b(k+i-1_t_int) 
       Enddo
       b(i)=dum/a(i,1_t_int) 
       If(l<mm) l=l+1_t_int 
    Enddo
    Return 
    !                                                                       
  End Subroutine banbks
  
  !!***********************************************************************
  !!                                                                       
  Subroutine bandec(a,n,m1,m2,al,indx,d) 
    !!@\authors{Numerical Recipies, Klaus Koepernik}
    !!@\name{modulesubroutine}{bandec}
    !!@\synopsis{
    !!@}
    !!@\description{
    !!@  LU decomposition of a band diagonal matrix with minimal pivoting
    !!@  A(i,j).ne.0.0  if i-m1<=j<=i+m2                                 
    !!@  stored as A(i,(1:m1,m1+1,m1+2:m1+m2+1))
    !!@  returns LU decomposition in a and al
    !!@  A(i,j) is replaced with U                                        
    !!@  al(i,j) is filled with L                                         
    !!@  indx contains the permutation due to pivoting                    
    !!@  np,mp,mpl are storage dimensions                                 
    !!@  d contains order of permutation (+-1)                            
    !!@}
    !!@\bugs{}
    !!@\seealso{banbks}
    !!@\changes{}
    !!@\copyright{(C) Copr. 1986-92 Numerical Recipes Software}
    !!

    !!
    Implicit None 
    !!IO                                                                     
    Integer(t_int),Intent(in)  :: m1,m2,n
    Integer(t_int),Intent(out),Dimension(:)  :: indx
    !                                                                       
    Real(t_real),Intent(inout),Dimension(:,:)   :: a
    Real(t_real),Intent(inout),Dimension(:,:)   :: al
    Real(t_real),Intent(out)  :: d

    !!LOCAL                                                                  
    Character(*),Parameter::nsubfunc="bandec"
    Integer(t_int) :: i,j,k,l,mm 
    Real(t_real)  :: dum 
    Real(t_real),Parameter ::  ttiny=Tiny(1.0_t_real)
    !                                                                       

    mm=m1+m2+1_t_int
    If(mm>Size(a,dim=2_t_int).Or.m1>Size(al,dim=2_t_int)&
         &.Or.n>Size(a,dim=1_t_int))Then
       Call errmsg(nmodule,nsubfunc,E_fatalerror&
            &,msg="'bad arguments, sorry!'")
    End If
    
    l=m1 
    Do i=1_t_int,m1 
       Do j=m1+2_t_int-i,mm 
          a(i,j-l)=a(i,j) 
       Enddo
       l=l-1_t_int
       Do j=mm-l,mm 
          a(i,j)=0.0_t_real
       Enddo
    Enddo
    d=1.0_t_real 
    l=m1 
    Do k=1_t_int,n 
       dum=a(k,1_t_int) 
       i=k 
       If(l<n)l=l+1_t_int 
       Do j=k+1_t_int,l 
          If(Abs(a(j,1_t_int))>Abs(dum))Then 
             dum=a(j,1_t_int) 
             i=j 
          Endif
       Enddo
       indx(k)=i 
       If(dum==0.0_t_real) Then 
          a(k,1_t_int)=ttiny
          Call errmsg(nmodule,nsubfunc,E_fatalerror&
            &,msg="'singular matrix, sorry!'")
       Endif
       If(i/=k)Then 
          d=-d 
          Do j=1_t_int,mm 
             dum=a(k,j) 
             a(k,j)=a(i,j) 
             a(i,j)=dum 
          Enddo
       Endif
       Do  i=k+1_t_int,l 
          dum=a(i,1_t_int)/a(k,1_t_int) 
          al(k,i-k)=dum 
          Do j=2_t_int,mm 
             a(i,j-1_t_int)=a(i,j)-dum*a(k,j) 
          Enddo
          a(i,mm)=0.0_t_real
       Enddo
    Enddo
    Return 
    !!                                                                       
  End Subroutine bandec
  !!***********************************************************************
  !! d(i), i=1,n --> diagonal of banded symmetric matrix
  !! c(i), i=1,n-1 --> offdiagonal of banded symmetric matrix
  !! a(i), i=1,n --> r.h.s.
  Subroutine symtridig_solv(n,d,c,a,x,z)
    !!
    Implicit None
    !!IO
    Integer(t_int),Intent(in)    :: n
    Real(t_real),Intent(in)      :: d(n),c(n-1),a(n)
    Real(t_real),Intent(out)     :: z(n),x(n)
    
    !!LOCAL
    !Character(*),Parameter::nsubfunc="symtridig_solv"
    Integer(t_int) :: i
    Real(t_real)   :: p
    !!
    !!
    x(1)=-c(1)/d(1)
    z(1)=a(1)/d(1)
    Do i=2,n
       p=d(i)+c(i-1)*x(i-1)
       x(i)=-c(i)/p
       z(i)=(a(i)-c(i-1)*z(i-1))/p
    Enddo
    !!
    !!
    x(n)=z(n)
    Do i=n-1,1,-1
       x(i)=z(i)+x(i)*x(i+1)
    Enddo
    !!
    Return
    !!
  End Subroutine symtridig_solv
  !!***********************************************************************

End Module bandmatrix
