Module bspline
  !!
  Use precisiondefs
  Use errors
!  Use memory
!  Use file_iounits
!  Use constants
 ! Use tmesh
  !!
  Implicit None 
  !!
  Private
  !!
  Character(*),Parameter::nmodule="BSPLINE"  
  !!
  Public :: bsplint,banslv,bvalue0,bvalue&
       &,set_natural_knots,set_natural_knots_x&
       &,natural_knot_index,bsp_bspvd

Contains
  !!***********************************************************************
  Subroutine bsplint(tau,t,n,k,q,bcoef,gtau)
    !!
    !!
    !! SPLINT produces the B-spline coefficients BCOEF of an interpolating 
    !! spline.
    !!
    !!
    !! Discussion:
    !!
    !!   The spline is of order K with knots T(1:N+K), and takes on the 
    !!   value GTAU(I) at TAU(I), for I = 1 to N.
    !!
    !! Method:
    !!
    !!   The I-th equation of the linear system  
    !!     A * BCOEF = B 
    !!   for the b-coefficients of the interpolant enforces interpolation
    !!   at TAU(1:N).
    !!
    !!   Hence,  b(i)=gtau(i), all i, and  a  is a band matrix with  2k-1
    !!   bands (if it is invertible).
    !!
    !!   The matrix A is generated row by row and stored, diagonal by di-
    !!   agonal, in the rows of the array  q , with the main diagonal go-
    !!   ing into row K.  see comments in the program below.
    !!
    !!   The banded system is then solved by a call to  banfac (which con-
    !!   structs the triangular factorization for  a  and stores it again in
    !!   q ), followed by a call to  banslv (which then obtains the solution
    !!   bcoef  by substitution).
    !!
    !!   BANFAC does no pivoting, since the total positivity of the matrix
    !!   A makes this unnecessary.
    !!
    !! Reference:
    !!
    !!   Carl DeBoor,
    !!   A Practical Guide to Splines,
    !!   Springer Verlag.
    !!
    !!
    Implicit None
    !!IO
    !! number of data points and dimension of spline space  s(k,t)
    Integer(t_int),Intent(in)    :: n
    !! order of spline
    Integer(t_int),Intent(in)    :: k
    !! knot sequence, of length  n+k
    Real(t_real),Intent(in)      :: t(n+k)
    !! gtau(N), the data ordinates.
    Real(t_real),Intent(in),Optional :: gtau(n)
    !! The data point abscissas. The entries in should be strictly increasing.
    Real(t_real),Intent(in)      :: tau(n)
    !!
    !!  array of size  (2*k-1)*n , containing the triangular factoriz-
    !!       ation of the coefficient matrix of the linear system for the b-
    !!       coefficients of the spline interpolant.
    !!       the b-coeffs for the interpolant of an additional data set can
    !!       be obtained without going through all the calculations in this
    !!       routine, simply by loading  htau  into  bcoef  and then execut-
    !!       ing the    call banslv ( q, 2*k-1, n, k-1, k-1, bcoef )
    Real(t_real),Intent(inout)   :: q((2*k-1)*n)
    !! the b-coefficients of the interpolant, of length  n.
    Real(t_real),Intent(out)     :: bcoef(n)
    !!LOCAL
    Character(*),Parameter::nsubfunc="bsplint"
    !!   an integer indicating success (= 1)  or failure (= 2)
    !!        the linear system to be solved is (theoretically) invertible if
    !!        and only if
    !!              t(i) < tau(i) < tau(i+k),    all i.
    !!        violation of this condition is certain to lead to  iflag=2 .
    Integer(t_int) :: iflag,i,ilp1mx,j,jj,kpkm2,left
    Real(t_real)   :: taui
    !!
    !!
    !!
    kpkm2 = 2_t_int*(k-1_t_int)
    left = k

    do i = 1, (2*k-1)*n
       q(i) = 0.0_t_real
    end do
    !!
    !! loop over i to construct the  n  interpolation equations
    !!
    Do i = 1, n

       taui = tau(i)
       ilp1mx = Min(i+k,n+1)
       !!
       !! find  left  in the closed interval (i,i+k-1) such that
       !!   t(left) <= tau(i)  < t(left+1)
       !! matrix is singular if this is not possible
       !!
       left = Max(left,i)
       
       If ( taui < t(left)) Then
          go to 70
       End If

20     Continue
       
       If ( taui < t(left+1)) Then
          go to 30
       End If
       
       left = left+1
       If ( left < ilp1mx) Then
          go to 20
       End If
       
       left = left-1
       If ( t(left+1) < taui ) Then
          go to 70
       End If
       !!
       !! The i-th equation enforces interpolation at taui, hence
       !! a(i,j)=b(j,k,t)(taui), all j. only the  k  entries with  j =
       !! left-k+1,...,left actually might be nonzero. these  k  numbers
       !! are returned, in  bcoef (used for temp.storage here), by the
       !! following
       !!
30     Continue
       
       Call bsplvb(t,k,1_t_int,taui,left,bcoef)
       !!
       !! We therefore want  bcoef(j)=b(left-k+j)(taui) to go into
       !! a(i,left-k+j), i.e., into  q(i-(left+j)+2*k,(left+j)-k) since
       !! a(i+j,j)  is to go into  q(i+k,j), all i,j,  if we consider  q
       !! as a two-dim. array , with  2*k-1  rows (see comments in
       !! banfac). in the present program, we treat  q  as an equivalent
       !! one-dimensional array (because of fortran restrictions on
       !! dimension statements) . we therefore want  bcoef(j) to go into
       !! entry
       !!   i -(left+j)+2*k + ((left+j)-k-1)*(2*k-1)
       !!  = i-left+1+(left -k)*(2*k-1) + (2*k-2)*j
       !! of  q .
       !!
       jj = i-left+1+(left-k)*(k+k-1)
       Do j = 1,k
          jj = jj+kpkm2
          q(jj) = bcoef(j)
       End Do
       
    End Do
    !!
    !! Obtain factorization of A, stored again in Q.
    !!
    Call banfac(q,k+k-1,n,k-1,k-1,iflag)
    
    If ( iflag == 2_t_int ) Then
       Call errmsg(nmodule,nsubfunc,E_fatalerror&
            &,msg='"The linear system is not invertible!?"')
    End If
    !!
    If(Present(gtau))Then
       !! Solve  a*bcoef=gtau  by backsubstitution
       !!
       bcoef(1:n) = gtau(1:n)
       
       Call banslv(q,k+k-1,n,k-1,k-1,bcoef)
    Endif
    Return
    
70  Continue
    !!
    Call errmsg(nmodule,nsubfunc,E_fatalerror&
         &,msg='"The linear system is not invertible!?"')
    !!
    Return
  End Subroutine bsplint
  !!***********************************************************************
  Subroutine banslv ( w, nroww, nrow, nbandl, nbandu, b )
    !
    !*************************************************************************
    !
    !! BANSLV solves a banded linear system X * X = B factored by BANFAC.
    !
    !
    !  Reference:
    !
    !    Carl DeBoor,
    !    A Practical Guide to Splines,
    !    Springer Verlag.
    !
    !  Parameters:
    !
    !    Input, real W(NROWW,NROW).  W contains the banded matrix,
    !    after it has been factored by BANFAC.
    !
    !    Input, integer NROWW, the row dimension of the work array W.
    !    NROWW must be at least NBANDL+1 + NBANDU.
    ! 
    !    Input, integer NROW, the number of rows in A.
    !
    !    Input, integer NBANDL, the number of bands of A below the 
    !    main diagonal.
    ! 
    !    Input, integer NBANDU, the number of bands of A above the 
    !    main diagonal.
    ! 
    !    Input/output, real B(NROW).
    !    On input, B contains the right hand side of the system to be solved.
    !    On output, B contains the solution, X.
    !!
    Implicit None
    !!IO
    Integer(t_int),Intent(in)    :: nrow
    Integer(t_int),Intent(in)    :: nroww
    Integer(t_int),Intent(in)    :: nbandl
    Integer(t_int),Intent(in)    :: nbandu
    Real(t_real),Intent(in)      :: w(nroww,nrow)
    !!
    Real(t_real),Intent(inout)   :: b(nrow)
    !!LOCAL
    Integer(t_int) :: i,j,jmax,middle
    !!
    middle = nbandu + 1_t_int

    If ( nrow == 1_t_int ) Then
       b(1) = b(1) / w(middle,1)
       Return
    End If
    !
    !  Forward pass
    !
    !  For I=1,2,...,NROW-1, subtract RHS(I)*(I-th column of L) 
    !  from the right side, below the I-th row.
    !
    If ( 0_t_int < nbandl ) Then
       Do i = 1, nrow-1
          jmax = Min ( nbandl, nrow-i )
          Do j = 1, jmax
             b(i+j) = b(i+j) - b(i) * w(middle+j,i)
          End Do
       End Do
    End If
    !
    !  Backward pass
    !
    !  For I=NROW, NROW-1,...,1, divide RHS(I) by 
    !  the I-th diagonal entry of U, then subtract 
    !  RHS(I)*(I-th column of U) from right side, above the I-th row.
    !
    Do i = nrow, 2, -1

       b(i) = b(i) / w(middle,i)
       
       Do j = 1, Min ( nbandu, i-1 )
          b(i-j) = b(i-j) - b(i) * w(middle-j,i)
       End Do
       
    End Do
    
    b(1) = b(1) / w(middle,1)
    !!
    Return
    !!
  End Subroutine banslv
  !!***********************************************************************
  Subroutine banfac ( w, nroww, nrow, nbandl, nbandu, iflag )
    !
    !*************************************************************************
    !
    !! BANFAC factors a banded matrix without pivoting.
    !
    !
    !  Discussion:
    !
    !    BANFAC returns in W the LU-factorization, without pivoting, of 
    !    the banded matrix A of order NROW with (NBANDL+1+NBANDU) bands 
    !    or diagonals in the work array W.
    ! 
    !    Gauss elimination without pivoting is used.  The routine is 
    !    intended for use with matrices A which do not require row 
    !    interchanges during factorization, especially for the totally 
    !    positive matrices which occur in spline calculations.
    !
    !  Reference:
    !
    !    Carl DeBoor,
    !    A Practical Guide to Splines,
    !    Springer Verlag.
    !
    !  Parameters:
    ! 
    !    Input/output, real W(NROWW,NROW).
    !    On input, W contains the "interesting" part of a banded 
    !    matrix A, with the diagonals or bands of A stored in the
    !    rows of W, while columns of A correspond to columns of W. 
    !
    !    This is the storage mode used in LINPACK and results in efficient 
    !    innermost loops.
    ! 
    !    Explicitly, A has 
    ! 
    !      NBANDL bands below the diagonal
    !      1     main diagonal
    !      NBANDU bands above the diagonal
    !
    !    and thus, with MIDDLE=NBANDU+1,
    !    A(I+J,J) is in W(I+MIDDLE,J) for I=-NBANDU,...,NBANDL, J=1,...,NROW.
    !
    !    For example, the interesting entries of a banded matrix
    !    matrix of order 9, with NBANDL=1, NBANDU=2:
    !
    !      11 12 13  0  0  0  0  0  0
    !      21 22 23 24  0  0  0  0  0
    !       0 32 33 34 35  0  0  0  0
    !       0  0 43 44 45 46  0  0  0
    !       0  0  0 54 55 56 57  0  0
    !       0  0  0  0 65 66 67 68  0
    !       0  0  0  0  0 76 77 78 79
    !       0  0  0  0  0  0 87 88 89
    !       0  0  0  0  0  0  0 98 99
    !
    !    would appear in the first 1+1+2=4 rows of W as follows:
    !
    !       0  0 13 24 35 46 57 68 79
    !       0 12 23 34 45 56 67 78 89
    !      11 22 33 44 55 66 77 88 99
    !      21 32 43 54 65 76 87 98  0
    ! 
    !    All other entries of W not identified in this way with an
    !    entry of A are never referenced.
    ! 
    !    On output, W contains the LU-factorization of A into a unit 
    !    lower triangular matrix L and an upper triangular matrix U 
    !    (both banded) and stored in customary fashion over the 
    !    corresponding entries of A.  
    !
    !    This makes it possible to solve any particular linear system A*X=B 
    !    for X by the call
    !
    !      call banslv(w,nroww,nrow,nbandl,nbandu,b)
    !
    !    with the solution X contained in B on return.
    ! 
    !    If IFLAG=2, then one of NROW-1, NBANDL, NBANDU failed to be nonnegative, 
    !    or else one of the potential pivots was found to be zero 
    !    indicating that A does not have an LU-factorization.  This 
    !    implies that A is singular in case it is totally positive.
    ! 
    !    Input, integer NROWW, the row dimension of the work array W.
    !    NROWW must be at least NBANDL+1 + NBANDU.
    ! 
    !    Input, integer NROW, the number of rows in A.
    !
    !    Input, integer NBANDL, the number of bands of A below the main diagonal.
    ! 
    !    Input, integer NBANDU, the number of bands of A above the main diagonal.
    ! 
    !    Output, integer IFLAG, error flag.
    !    1, success.
    !    2, failure, the matrix was not factored.
    !
    Implicit None
    !!IO
    Integer(t_int),Intent(in)    :: nrow
    Integer(t_int),Intent(in)    :: nroww
    Integer(t_int),Intent(in)    :: nbandl
    Integer(t_int),Intent(in)    :: nbandu
    Integer(t_int),Intent(out)   :: iflag
    Real(t_real),Intent(inout)   :: w(nroww,nrow)
    !!LOCAL
    !!
    Integer(t_int) :: i,j,k,middle
    Real(t_real)   :: factor
    Real(t_real)   :: pivot
    !
    iflag = 1_t_int
    middle = nbandu + 1_t_int
    !
    !  W(MIDDLE,*) contains the main diagonal of A.
    !
    If ( nrow < 1_t_int ) Then
       iflag = 2_t_int
       Return
    End If
    
    If ( nrow == 1_t_int ) Then
       If ( w(middle,nrow) == 0.0_t_real ) Then
          iflag = 2_t_int
       End If
       Return
    End If
    !
    !  A is upper triangular.  Check that the diagonal is nonzero.
    !
    If ( nbandl <= 0_t_int ) Then
       
       Do i = 1, nrow-1
          If ( w(middle,i) == 0.0_t_real ) Then
             iflag = 2_t_int
             Return
          End If
       End Do
       
       If ( w(middle,nrow) == 0.0_t_real ) Then
          iflag = 2_t_int
       End If
       
       Return
       !
       !  A is lower triangular.  Check that the diagonal is nonzero and
       !  divide each column by its diagonal.
       !
    Else If ( nbandu <= 0_t_int ) Then
       
       Do i = 1, nrow-1
          
          pivot = w(middle,i)
          
          If ( pivot == 0.0_t_real ) Then
             iflag = 2_t_int
             Return
          End If
          
          Do j = 1, Min ( nbandl, nrow-i )
             w(middle+j,i) = w(middle+j,i) / pivot
          End Do
          
       End Do
       
       Return
       
    End If
    !
    !  A is not just a triangular matrix.  
    !  Construct the LU factorization.
    !
    Do i = 1, nrow-1
       !
       !  W(MIDDLE,I) is the pivot for the I-th step.
       !
       If ( w(middle,i) == 0.0_t_real ) Then
          iflag = 2_t_int
          Write ( *, * ) ' '
          Write ( *, * ) 'BANFAC - Fatal error!'
          Write ( *, * ) '  Zero pivot encountered in column ', i
          Stop
       End If
       !
       !  Divide each entry in column I below the diagonal by PIVOT.
       !
       Do j = 1, Min ( nbandl, nrow-i )
          w(middle+j,i) = w(middle+j,i) / w(middle,i)
       End Do
       !
       !  Subtract A(I,I+K)*(I-th column) from (I+K)-th column (below
       !  row I).
       !
       Do k = 1, Min ( nbandu, nrow-i )
          factor = w(middle-k,i+k)
          Do j = 1, Min ( nbandl, nrow-i )
             w(middle-k+j,i+k) = w(middle-k+j,i+k) - w(middle+j,i) * factor
          End Do
       End Do
       
    End Do
    !
    !  Check the last diagonal entry.
    !
    If ( w(middle,nrow) == 0.0_t_real ) Then
       iflag = 2_t_int
    End If
    !!
    Return
    !!
  End Subroutine banfac
  !!***********************************************************************
  Subroutine bsp_bspvn (t, jhigh, k, index, x, ileft, BiAtX, work, iwork&
       &,bounderror)
    !!
    !! bsp_BSPVN calculates the value of all (possibly) nonzero basis ...
    !! functions at X.
    !!
    !!***LIBRARY   SLATEC
    !!***CATEGORY  E3, K6
    !!***KEYWORDS  EVALUATION OF B-SPLINE
    !!***AUTHOR  Amos, D. E., (SNLA)
    !!***DESCRIPTION
    !!
    !!    Written by Carl de Boor and modified by D. E. Amos
    !!
    !!    Abstract
    !!        bsp_BSPVN is the BSPLVN routine of the reference.
    !!
    !!        BSPVN calculates the value of all (possibly) nonzero basis
    !!        functions at X of order MAX(JHIGH,(J+1)*(INDEX-1)), where
    !!        T(K)  <=  X  <=  T(N+1) and J=IWORK is set inside the routine
    !!        on the first call when INDEX=1.  ILEFT is such that T(ILEFT)
    !!         <=  X  <  T(ILEFT+1).  A call to INTRV(T,N+1,X,ILO,ILEFT,
    !!        MFLAG) produces the proper ILEFT.  BSPVN calculates using the
    !!        basic algorithm needed in BSPVD.  If only basis functions are
    !!        desired, setting JHIGH=K and INDEX=1 can be faster than
    !!        calling BSPVD, but extra coding is required for derivatives
    !!        (INDEX=2) and BSPVD is set up for this purpose.
    !!
    !!        Left limiting values are set up as described in BSPVD.
    !!
    !!    Description of Arguments
    !!        Input
    !!         T       - knot vector of length N+K, where
    !!                   N = number of B-spline basis functions
    !!                   N = sum of knot multiplicities-K
    !!         JHIGH   - order of B-spline, 1  <=  JHIGH  <=  K
    !!         K       - highest possible order
    !!         INDEX   - INDEX = 1 gives basis functions of order JHIGH
    !!                         = 2 denotes previous entry with WORK, IWORK
    !!                             values saved for subsequent calls to
    !!                             BSPVN.
    !!         X       - argument of basis functions,
    !!                   T(K)  <=  X  <=  T(N+1)
    !!         ILEFT   - largest integer such that
    !!                   T(ILEFT)  <=  X  <  T(ILEFT+1)
    !!
    !!        Output
    !!         BIATX   - vector of length K for spline values.
    !!         WORK    - a work vector of length 2*K
    !!         IWORK   - a work parameter.  Both WORK and IWORK contain
    !!                   information necessary to continue for INDEX = 2.
    !!                   When INDEX = 1 exclusively, these are scratch
    !!                   variables and can be used for other purposes.
    !!
    !!    Error Conditions
    !!        Improper input is a fatal error.
    !!
    !!***REFERENCES  Carl de Boor, Package for calculating with B-splines,
    !!                SIAM Journal on Numerical Analysis 14, 3 (June 1977),
    !!                pp. 441-472.
    !!***REVISION HISTORY  (YYMMDD)
    !!  800901  DATE WRITTEN
    !!  890831  Modified array declarations.  (WRB)
    !!  890831  REVISION DATE from Version 3.2
    !!  891214  Prologue converted to Version 4.0 format.  (BAB)
    !!  900315  CALLs to XERROR changed to CALLs to XERMSG.  (THJ)
    !!  900326  Removed duplicate information from DESCRIPTION section.
    !!          (WRB)
    !!   920501  Reformatted the REFERENCES section.  (WRB)
    !!***END PROLOGUE  BSPVN
    !!
    Implicit None  
    !!IO
    Logical(t_log),Intent(in)    :: bounderror !! give error if out of bounds
    !! means x not in interval S_ileft, it is suitable to suppress it for
    !! the right interval point t(n+1) since due to roundoff a point may
    !! be slightly beyond  t(n+1)    
    Integer(t_int),Intent(in)    :: index
    Integer(t_int),Intent(in)    :: k
    Integer(t_int),Intent(in)    :: jhigh
    Integer(t_int),Intent(in)    :: ileft
    Integer(t_int),Intent(inout) :: iwork
    Real(t_real),Intent(in)      :: x
    Real(t_real),Intent(in)      :: t(*)
    Real(t_real),Intent(out)     :: BiAtX(*)
    Real(t_real),Intent(inout)   :: work(*)
    !!LOCAL
    Character(*),Parameter::nsubfunc="bsp_bspvn"
    Integer(t_int) :: imjp1,ipj,jp1,jp1ml,l
    Real(t_real)   ::  term,saved
    !!    content of iwork work is expected unchanged between calls.
    !!    work(i) = deltap(i), work(k+i) = deltam(i), i = 1,k
    !!
    If ( k < 1_t_int) Then
       Call errmsg(nmodule,nsubfunc,E_fatalerror&
            &,msg="'k does not satisfy k >= 1'")
    Endif
    !!
    If ( jhigh > k .Or. jhigh < 1_t_int) Then
       Call errmsg(nmodule,nsubfunc,E_fatalerror&
            &,msg="'jhigh does not satisfy 1 <= jhigh <= k'")
    Endif
    !!
    If ( index < 1_t_int .Or. index > 2_t_int) Then
       Call errmsg(nmodule,nsubfunc,E_fatalerror&
            &,msg="'index is not 1 or 2'")
    End If
    !!
    If ( bounderror .And. (x < t(ileft) .Or. x > t(ileft+1))) Then
       !Print *,'hier1'
       Call errmsg(nmodule,nsubfunc,E_fatalerror&
            &,msg="'x does not satisfy t(ileft) <= x <= t(ileft+1)'")
    Endif
    !!
    If(index==1_t_int) Then
       iwork = 1_t_int
       BiAtX(1) = 1.0_t_real
       If (iwork >= jhigh) Return
    Endif
    !!
    Do
       ipj = ileft + iwork
       work(iwork) = t(ipj) - x
       imjp1 = ileft - iwork + 1_t_int
       work(k+iwork) = x - t(imjp1)
       saved = 0.0_t_real
       jp1 = iwork + 1_t_int
       Do l=1,iwork
          jp1ml = jp1 - l
          term = BiAtX(l)/(work(l)+work(k+jp1ml))
          BiAtX(l) = term*work(l) + saved
          saved = term*work(k+jp1ml)
       Enddo
       BiAtX(jp1) = saved
       iwork = jp1
       If (iwork >= jhigh) Exit
    Enddo
    !!
    Return
    !!
  End Subroutine bsp_bspvn
  !!***********************************************************************
  Subroutine bsp_bspvd (n,t, k, nderiv, x, ileft, bdim, BiAtX, work,bounderror)
    !!
    !! bsp_bspvd calculates the value and all derivatives of order less than ...
    !! NDERIV of all basis functions which do not vanish at X.
    !!
    !! LIBRARY   SLATEC
    !! CATEGORY  E3, K6
    !! KEYWORDS  DIFFERENTIATION OF B-SPLINE, EVALUATION OF B-SPLINE
    !! AUTHOR  Amos, D. E., (SNLA)
    !! DESCRIPTION
    !!
    !!    Written by Carl de Boor and modified by D. E. Amos
    !!
    !!    Abstract
    !!        bsp_BSPVD is the BSPLVD routine of the reference.
    !!
    !!        BSPVD calculates the value and all derivatives of order
    !!        less than NDERIV of all basis functions which do not
    !!        (possibly) vanish at X.  ILEFT is input such that
    !!        T(ILEFT)  <=  X  <  T(ILEFT+1).  A call to INTRV(T,N+1,X,
    !!        ILO,ILEFT,MFLAG) will produce the proper ILEFT.  The output of
    !!        BSPVD is a matrix BIATX(I,J) of dimension at least (K,NDERIV)
    !!        whose columns contain the K nonzero basis functions and
    !!        their NDERIV-1 right derivatives at X, I=1,K, J=1,NDERIV.
    !!        These basis functions have indices ILEFT-K+I, I=1,K,
    !!        K  <=  ILEFT  <=  N. The nonzero part of the I-th basis
    !!        function lies in (T(I),T(I+K)), I=1,N.
    !!
    !!        If X=T(ILEFT+1) then BIATX contains left limiting values
    !!        (left derivatives) at T(ILEFT+1).  In particular, ILEFT = N
    !!        produces left limiting values at the right end point
    !!        X=T(N+1). To obtain left limiting values at T(I), I=K+1,N+1,
    !!        set X= next lower distinct knot, call INTRV to get ILEFT,
    !!        set X=T(I), and then call BSPVD.
    !!
    !!    Description of Arguments
    !!        Input
    !!         T       - knot vector of length N+K, where
    !!                   N = number of B-spline basis functions
    !!                   N = sum of knot multiplicities-K
    !!         K       - order of the B-spline, K  >=  1
    !!         NDERIV  - number of derivatives = NDERIV-1,
    !!                   1  <=  NDERIV  <=  K
    !!         X       - argument of basis functions,
    !!                   T(K)  <=  X  <=  T(N+1)
    !!         ILEFT   - largest integer such that
    !!                   T(ILEFT)  <=  X  <  T(ILEFT+1)
    !!         BDIM    - leading dimension of matrix BIATX
    !!
    !!        Output
    !!         BIATX   - matrix of dimension at least (K,NDERIV) contain-
    !!                   ing the nonzero basis functions at X and their
    !!                   derivatives columnwise.
    !!         WORK    - a work vector of length (K+1)*(K+2)/2
    !!
    !!    Error Conditions
    !!        Improper input is a fatal error
    !!
    !! REFERENCES  Carl de Boor, Package for calculating with B-splines,
    !!                SIAM Journal on Numerical Analysis 14, 3 (June 1977),
    !!                pp. 441-472.
    !! ROUTINES CALLED  BSPVN
    !! REVISION HISTORY  (YYMMDD)
    !!  800901  DATE WRITTEN
    !!  890531  Changed all specific intrinsics to generic.  (WRB)
    !!  890831  Modified array declarations.  (WRB)
    !!  890831  REVISION DATE from Version 3.2
    !!  891214  Prologue converted to Version 4.0 format.  (BAB)
    !!  900315  CALLs to XERROR changed to CALLs to XERMSG.  (THJ)
    !!  900326  Removed duplicate information from DESCRIPTION section.
    !!          (WRB)
    !!  920501  Reformatted the REFERENCES section.  (WRB)
    !! END PROLOGUE  BSPVD
    !!
    Implicit None 
    !!IO
    Logical(t_log),Intent(in)    :: bounderror !! give error if out of bounds
    !! means x not in interval S_ileft, it is suitable to suppress it for
    !! the right interval point t(n+1) since due to roundoff a point may
    !! be slightly beyond  t(n+1)    
    Integer(t_int),Intent(in)    :: k,n
    Integer(t_int),Intent(in)    :: nderiv
    Integer(t_int),Intent(in)    :: ileft
    Integer(t_int),Intent(in)    :: bdim
    Real(t_real),Intent(in)      :: x
    Real(t_real),Intent(in)      :: t(ileft+k) !! at least ileft+k
    Real(t_real),Intent(out)     :: work((k+1)*(k+2)/2)
    Real(t_real),Intent(out)     :: BiAtX(bdim,*)
    !!LOCAL
    Character(*),Parameter::nsubfunc="bsp_bspvd"
    Integer(t_int) :: i,ideriv,ipkmd,j,jj,jlow,jm,jp1mid,kmd,kp1,l
    Integer(t_int) :: ldummy,m,mhigh,iwork
    Real(t_real)   :: factor,fkmd,v
    !!    
    !!    a(i,j) = work(i+j*(j+1)/2),  i=1,j+1  j=1,k-1
    !!    a(i,k) = w0rk(i+k*(k-1)/2)  i=1.k
    !!    work(1) and work((k+1)*(k+2)/2) are not used.
    !!
    If(ileft<1 .Or. ileft>n)Then
       biatx(1:bdim,1:nderiv)=0.0_t_real
       Return
    Endif
    !!

    If ( k < 1_t_int) Then
       Call errmsg(nmodule,nsubfunc,E_fatalerror&
            &,msg="'k does not satisfy k >= 1'")
    Endif
    If ( nderiv < 1_t_int .Or. nderiv > k)Then
       Call errmsg(nmodule,nsubfunc,E_fatalerror&
            &,msg="'nderiv does not satisfy 1 <= nderiv <= k'")
    End If
    If ( bdim < k) Then
       Call errmsg(nmodule,nsubfunc,E_fatalerror&
            &,msg="'bdim does not satisfy bdim >= k'")
    Endif
    ideriv=nderiv
    kp1=k+1_t_int
    jj=kp1-ideriv
    Call bsp_bspvn(t, jj, k, 1_t_int, x, ileft, biatx, work, iwork,bounderror)
    If (ideriv == 1_t_int) Return
    mhigh = ideriv
    Do m=2_t_int,mhigh
       jp1mid = 1_t_int
       Do j=ideriv,k
          biatx(j,ideriv) = biatx(jp1mid,1)
          jp1mid = jp1mid + 1_t_int
       End Do
       ideriv = ideriv - 1_t_int
       jj = kp1 - ideriv
       Call bsp_bspvn(t,jj,k,2_t_int,x,ileft,biatx,work,iwork,bounderror)
    End Do
    !!   
    !!
    jm = kp1*(kp1+1_t_int)/2_t_int
    Do l = 1,jm
       work(l) = 0.0_t_real
    End Do
    !!    a(i,i) = work(i*(i+3)/2) = 1.0       i = 1,k
    l = 2_t_int
    j = 0_t_int
    Do i = 1,k
       j = j + l
       work(j) = 1.0_t_real
       l = l + 1_t_int
    End Do
    !!
    kmd = k
    Do m=2,mhigh
       kmd = kmd - 1_t_int
       fkmd = kmd
       i = ileft
       j = k
       jj = j*(j+1_t_int)/2_t_int
       jm = jj - j
       Do ldummy=1,kmd
          ipkmd = i + kmd
          factor = fkmd/(t(ipkmd)-t(i))
          Do l=1,j
             work(l+jj) = (work(l+jj)-work(l+jm))*factor
          End Do
          i = i - 1_t_int
          j = j - 1_t_int
          jj = jm
          jm = jm - j
       End Do
       !!
       Do i=1,k
          v = 0.0_t_real
          jlow = Max(i,m)
          jj = jlow*(jlow+1_t_int)/2_t_int
          Do j=jlow,k
             v = work(i+jj)*biatx(j,m) + v
             jj = jj + j + 1_t_int
          End Do
          biatx(i,m) = v
       End Do
    End Do
    !!
    Return
    !!
    !!
  End Subroutine bsp_bspvd
  !!***********************************************************************
  Subroutine bsplvb ( t, jhigh, index, x, left, biatx )
    !
    !*************************************************************************
    !
    !! BSPLVB evaluates B-splines at a point X with a given knot sequence.
    !
    !
    !  Discusion:
    !
    !    BSPLVB evaluates all possibly nonzero B-splines at X of order
    !
    !      JOUT = MAX ( JHIGH, (J+1)*(INDEX-1) ) 
    !  
    !    with knot sequence T.
    ! 
    !    The recurrence relation
    ! 
    !                     X - T(I)               T(I+J+1) - X
    !    B(I,J+1)(X) = ----------- * B(I,J)(X) + --------------- * B(I+1,J)(X)
    !                  T(I+J)-T(I)               T(I+J+1)-T(I+1)
    ! 
    !    is used to generate B(LEFT-J:LEFT,J+1)(X) from B(LEFT-J+1:LEFT,J)(X)
    !    storing the new values in BIATX over the old. 
    !
    !    The facts that 
    !
    !      B(I,1)(X) = 1  if  T(I) <= X < T(I+1)
    !
    !    and that 
    !
    !      B(I,J)(X) = 0  unless  T(I) <= X < T(I+J)
    !
    !    are used. 
    !
    !    The particular organization of the calculations follows 
    !    algorithm 8 in chapter X of the text.
    !
    !  Reference:
    !
    !    Carl DeBoor,
    !    A Practical Guide to Splines,
    !    Springer Verlag.
    !
    !  Parameters:
    !
    !    Input, real T(LEFT+JOUT), the knot sequence.  T is assumed to 
    !    be nondecreasing, and also, T(LEFT) must be strictly less than 
    !    T(LEFT+1).
    ! 
    !    Input, integer JHIGH, INDEX, determine the order 
    !    JOUT = MAX(JHIGH,(J+1)*(INDEX-1))  
    !    of the B-splines whose values at X are to be returned.  
    !    INDEX is used to avoid recalculations when several 
    !    columns of the triangular array of B-spline values are
    !    needed, for example, in BVALUE or in BSPLVD.
    ! 
    !    If INDEX = 1, the calculation starts from scratch and the entire 
    !    triangular array of B-spline values of orders
    !    1, 2, ...,JHIGH is generated order by order, i.e., 
    !    column by column.
    !   
    !    If INDEX = 2, only the B-spline values of order J+1, J+2, ..., JOUT  
    !    are generated, the assumption being that BIATX, J, 
    !    DELTAL, DELTAR are, on entry, as they were on exit 
    !    at the previous call.  In particular, if JHIGH = 0, 
    !    then JOUT = J+1, i.e., just the next column of B-spline 
    !    values is generated.
    ! 
    !    WARNING: the restriction  JOUT <= JMAX (= 20) is
    !    imposed arbitrarily by the dimension statement for DELTAL
    !    and DELTAR, but is nowhere checked for.
    ! 
    !    Input, real X, the point at which the B-splines are to be evaluated.
    ! 
    !    Input, integer LEFT, an integer chosen so that 
    !    T(LEFT) <= X <= T(LEFT+1).
    ! 
    !    Output, real BIATX(JOUT), with BIATX(I) containing the
    !    value at X of the polynomial of order JOUT which agrees 
    !    with the B-spline B(LEFT-JOUT+I,JOUT,T) on the interval 
    !    (T(LEFT),T(LEFT+1)).
    !!
    Implicit None
    !!IO
    Integer(t_int),Intent(in)    :: jhigh
    Integer(t_int),Intent(in)    :: index
    Integer(t_int),Intent(in)    :: left
    Real(t_real),Intent(in)      :: x
    Real(t_real),Intent(in)      :: t(left+jhigh)
    Real(t_real),Intent(inout)   :: biatx(jhigh)
    !!LOCAL
    Integer(t_int),Parameter :: jmax = 20_t_int
    !!
    Integer(t_int) :: i
    Integer(t_int),Save :: j
    Real(t_real),Save   :: deltal(jmax)
    Real(t_real),Save   :: deltar(jmax)
    Real(t_real)   :: saved,term
    !!
    If ( index == 1 ) Then 
       j = 1_t_int
       biatx(1) = 1.0_t_real
       If ( jhigh <= j ) Then
          Return
       End If
    End If
    
    If ( t(left+1) <= t(left) ) Then
       Write ( *, '(a)' ) ' '
       Write ( *, '(a)' ) 'BSPLVB - Fatal error!'
       Write ( *, '(a)' ) '  It is required that T(LEFT) < T(LEFT+1).'
       Write ( *, '(a,i6)' ) '  But LEFT = ', left
       Write ( *, '(a,g14.6)' ) '  T(LEFT) =   ', t(left)
       Write ( *, '(a,g14.6)' ) '  T(LEFT+1) = ', t(left+1)
       Stop
    End If
    
    Do
       
       deltar(j) = t(left+j) - x
       deltal(j) = x - t(left+1-j)
       
       saved = 0.0_t_real
       Do i = 1, j
          term = biatx(i) / ( deltar(i) + deltal(j+1-i) )
          biatx(i) = saved + deltar(i) * term
          saved = deltal(j+1-i) * term
       End Do
       
       biatx(j+1) = saved
       j = j + 1_t_int
       
       If ( jhigh <= j ) Then
          Exit
       End If
       
    End Do
    
    Return
  End Subroutine bsplvb
  !!***********************************************************************
  Subroutine interv ( xt, lxt, x, left, mflag )
    !
    !*******************************************************************************
    !
    !! INTERV brackets a real value in an ascending vector of values.
    !
    !
    !  Discussion:
    !
    !    The XT array is a set of increasing values.  The goal of the routine
    !    is to determine the largest index I so that XT(I) <= X.
    !
    !    The routine is designed to be efficient in the common situation
    !    that it is called repeatedly, with X taken from an increasing
    !    or decreasing sequence.
    !
    !    This will happen when a piecewise polynomial is to be graphed.
    !    The first guess for LEFT is therefore taken to be the value
    !    returned at the previous call and stored in the local variable ILO.
    !
    !    A first check ascertains that ILO < LXT.  This is necessary
    !    since the present call may have nothing to do with the previous
    !    call.  Then, if XT(ILO) < = X < XT(ILO+1), we set LEFT=ILO
    !    and are done after just three comparisons.
    !
    !    Otherwise, we repeatedly double the difference ISTEP=IHI-ILO
    !    while also moving ILO and IHI in the direction of X, until
    !      XT(ILO) < = X < XT(IHI)
    !    after which we use bisection to get, in addition, ILO+1=IHI.
    !    LEFT=ILO is then returned.
    !
    !  Modified:
    !
    !    05 February 2004
    !
    !  Reference:
    !
    !    Carl DeBoor,
    !    A Practical Guide to Splines,
    !    Springer Verlag.
    !
    !  Parameters:
    !
    !    Input, real XT(LXT), a nondecreasing sequence of values.
    !
    !    Input, integer LXT, the dimension of XT.
    !
    !    Input, real X, the point whose location with 
    !    respect to the sequence XT is to be determined.
    !
    !    Output, integer LEFT, the index of the bracketing value:
    !      1     if             X  <  XT(1)
    !      I     if   XT(I)  <= X  < XT(I+1)
    !      LXT   if  XT(LXT) <= X
    !
    !    Output, integer MFLAG, indicates whether X lies within the
    !    range of the data.
    !    -1:            X  <  XT(1)
    !     0: XT(I)   <= X  < XT(I+1)
    !    +1: XT(LXT) <= X
    !!
    Implicit None
    !!IO
    Integer(t_int),Intent(in)    :: lxt
    Integer(t_int),Intent(out)   :: left
    Integer(t_int),Intent(out)   :: mflag
    Real(t_real),Intent(in)      :: x
    Real(t_real),Intent(in)      :: xt(lxt)
    !!
    Integer(t_int) :: ihi,istep,middle
    integer(t_int), save :: ilo = 1_t_int
    !!
    ihi = ilo + 1_t_int

    If ( lxt <= ihi ) Then
       
       If ( xt(lxt) <= x ) Then
          go to 110
       End If
       
       If ( lxt <= 1_t_int ) Then
          mflag = -1_t_int
          left = 1_t_int
          Return
       End If
       
       ilo = lxt - 1_t_int
       ihi = lxt
       
    End If
    
    If ( xt(ihi) <= x ) Then
       go to 40
    End If
    
    If ( xt(ilo) <= x ) Then
       mflag = 0_t_int
       left = ilo
       Return
    End If
    !
    !  Now X < XT(ILO).  Decrease ILO to capture X.
    !
    istep = 1_t_int
    
31  Continue
    
    ihi = ilo
    ilo = ihi - istep
    
    If ( 1_t_int < ilo ) Then
       If ( xt(ilo) <= x ) Then
          go to 50
       End If
       istep = istep * 2_t_int
       go to 31
    End If
    
    ilo = 1_t_int
    
    If ( x < xt(1) ) Then
       mflag = -1_t_int
       left = 1_t_int
       Return
    End If
    
    go to 50
    !
    !  Now XT(IHI) <= X.  Increase IHI to capture X.
    !
40  Continue
    
    istep = 1_t_int
    
41  Continue
    
    ilo = ihi
    ihi = ilo + istep
    
    If ( ihi < lxt ) Then
       If ( x < xt(ihi) ) Then
          go to 50
       End If
       istep = istep * 2_t_int
       go to 41
    End If
    
    If ( xt(lxt) <= x ) Then
       go to 110
    End If
    !
    !  Now XT(ILO) < = X < XT(IHI).  Narrow the interval.
    !
    ihi = lxt
    
50  Continue
    
    Do
       
       middle = ( ilo + ihi ) / 2_t_int
       
       If ( middle == ilo ) Then
          mflag = 0_t_int
          left = ilo
          Return
       End If
       !
       !  It is assumed that MIDDLE = ILO in case IHI = ILO+1.
       !
       If ( xt(middle) <= x ) Then
          ilo = middle
       Else
          ihi = middle
       End If
       
    End Do
    !
    !  Set output and return.
    !
110 Continue
    
    mflag = 1_t_int

    If ( x == xt(lxt) ) Then
       mflag = 0_t_int
    End If
    
    Do left = lxt, 1, -1
       If ( xt(left) < xt(lxt) ) Then
          Return
       End If
    End Do
    
    Return
  End Subroutine interv
  !!***********************************************************************
  Real(t_real) Function bvalue ( t, bcoef, n, k, x, jderiv )
    !
    !*************************************************************************
    !
    !! BVALUE evaluates a derivative of a spline from its B-spline representation.  
    !
    !
    !  Discussion:
    !
    !    The spline is taken to be continuous from the right.
    ! 
    !    The nontrivial knot interval (T(I),T(I+1)) containing X is 
    !    located with the aid of INTERV.  The K B-spline coefficients 
    !    of F relevant for this interval are then obtained from BCOEF, 
    !    or are taken to be zero if not explicitly available, and are 
    !    then differenced JDERIV times to obtain the B-spline 
    !    coefficients of (D**JDERIV)F relevant for that interval.  
    !
    !    Precisely, with J = JDERIV, we have from X.(12) of the text that:
    ! 
    !      (D**J)F = sum ( BCOEF(.,J)*B(.,K-J,T) )
    ! 
    !    where
    !                      / BCOEF(.),                    ,  J == 0
    !                     /
    !       BCOEF(.,J) = / BCOEF(.,J-1) - BCOEF(.-1,J-1)
    !                   / -----------------------------,  0 < J
    !                  /    (T(.+K-J) - T(.))/(K-J)
    ! 
    !    Then, we use repeatedly the fact that
    ! 
    !      sum ( A(.)*B(.,M,T)(X) ) = sum ( A(.,X)*B(.,M-1,T)(X) )
    ! 
    !    with
    !                   (X - T(.))*A(.) + (T(.+M-1) - X)*A(.-1)
    !      A(.,X) =   ---------------------------------------
    !                   (X - T(.))      + (T(.+M-1) - X)
    ! 
    !    to write (D**J)F(X) eventually as a linear combination of 
    !    B-splines of order 1, and the coefficient for B(I,1,T)(X) 
    !    must then be the desired number (D**J)F(X).
    !    See x.(17)-(19) of text.
    !
    !  Reference:
    !
    !    Carl DeBoor,
    !    A Practical Guide to Splines,
    !    Springer Verlag.
    !
    !  Parameters:
    !
    !    Input, real T(N+K), the knot sequence.  T is assumed to be nondecreasing.
    !
    !    Input, real BCOEF(N), B-spline coefficient sequence.
    ! 
    !    Input, integer N, the length of BCOEF.
    ! 
    !    Input, integer K, the order of the spline.
    !
    !    Input, real X, the point at which to evaluate.
    ! 
    !    Input, integer JDERIV, the order of the derivative to 
    !    be evaluated.  JDERIV is assumed to be zero or positive.
    ! 
    !    Output, real BVALUE, the value of the (JDERIV)-th 
    !    derivative of the spline at X.
    !!
    Implicit None
    !!IO
    Integer(t_int),Intent(in)    :: n
    Integer(t_int),Intent(in)    :: k
    Integer(t_int),Intent(in)    :: jderiv
    Real(t_real),Intent(in)      :: t(n+k)
    Real(t_real),Intent(in)      :: bcoef(n)
    Real(t_real),Intent(in)      :: x
    !!LOCAL
    Integer(t_int) :: i,ilo,j,jc,jcmax,jcmin,jj,mflag
    Real(t_real)   :: aj(k)
    Real(t_real)   :: dl(k)
    Real(t_real)   :: dr(k)

    !!
    bvalue = 0.0_t_real

    If ( k <= jderiv ) Then
       Return
    End If
    !
    !  Find I so that 1 <= i < n+k and t(i) < t(i+1) and t(i) <= x < t(i+1). 
    !
    !  If no such i can be found, X lies
    !  outside the support of  the spline F and  bvalue=0.
    !  (the asymmetry in this choice of  i  makes F rightcontinuous)
    !
    Call interv ( t, n+k, x, i, mflag )
    
    If ( mflag /= 0_t_int ) Then
       Return
    End If
    !
    !  If K=1 (and jderiv = 0), bvalue = bcoef(i).
    !
    If ( k <= 1_t_int ) Then
       bvalue = bcoef(i)
       Return
    End If
    !
    !  Store the K b-spline coefficients relevant for the knot interval
    !  (t(i),t(i+1)) in aj(1),...,aj(k) and compute dl(j)=x-t(i+1-j),
    !  dr(j)=t(i+j)-x, j=1,...,k-1 . set any of the aj not obtainable
    !  from input to zero. set any t.s not obtainable equal to t(1) or
    !  to t(n+k) appropriately.
    !
    jcmin = 1_t_int
    
    If ( k <= i ) Then
       
       Do j = 1, k-1
          dl(j) = x-t(i+1-j)
       End Do
       
    Else
       
       jcmin = 1_t_int-(i-k)
       
       Do j = 1, i
          dl(j) = x-t(i+1-j)
       End Do
       
       Do j = i, k-1
          aj(k-j) = 0.0_t_real
          dl(j) = dl(i)
       End Do
       
    End If
    
    jcmax = k
    
    If ( i <= n ) Then
       go to 90
    End If
    
    jcmax = k+n-i
    Do j = 1, k+n-i
       dr(j) = t(i+j)-x
    End Do
    
    Do j = k+n-i, k-1
       aj(j+1) = 0.0_t_real
       dr(j) = dr(k+n-i)
    End Do
    
    go to 110
    
90  Continue
    
    Do j = 1, k-1
       dr(j) = t(i+j)-x
    End Do
    
110 Continue
    
    Do jc = jcmin, jcmax
       aj(jc) = bcoef(i-k+jc)
    End Do
    !
    !  Difference the coefficients JDERIV times.
    !
    Do j = 1, jderiv
       
       ilo = k-j
       Do jj = 1, k-j
          aj(jj) = ((aj(jj+1)-aj(jj))/(dl(ilo)+dr(jj)))*Real(k-j,t_real)
          ilo = ilo-1_t_int
       End Do
       
    End Do
    !
    !  Compute value at X in (t(i),t(i+1)) of jderiv-th derivative,
    !  given its relevant b-spline coeffs in aj(1),...,aj(k-jderiv).
    !
    Do j = jderiv+1, k-1
       ilo = k-j
       Do jj = 1, k-j
          aj(jj) = ( aj(jj+1) * dl(ilo) + aj(jj) * dr(jj) ) &
               / ( dl(ilo) + dr(jj) )
          ilo = ilo-1_t_int
       End Do
    End Do
    
    bvalue = aj(1)
    
    Return
  End Function bvalue
  !!***********************************************************************
  Real(t_real) Function bvalue0 (t, bcoef, n, k, x ,i)
    !
    !*************************************************************************
    !! like BVALUE but only for zero-th derivative
    !! correct interval index must be supplied
    !!
    Implicit None
    !!IO
    Integer(t_int),Intent(in)    :: n
    Integer(t_int),Intent(in)    :: k
    Integer(t_int),Intent(in)    :: i
    Real(t_real),Intent(in)      :: t(n+k)
    Real(t_real),Intent(in)      :: bcoef(n)
    !! the ordinate, where the interpolated function is required
    Real(t_real),Intent(in)      :: x
    !!LOCAL
    Integer(t_int) :: ilo,j,jc,jcmax,jcmin,jj
    Real(t_real)   :: aj(k)
    Real(t_real)   :: dl(k)
    Real(t_real)   :: dr(k)
    !!
    !!
    bvalue0 = 0.0_t_real

    !
    !  Find I so that 1 <= i < n+k and t(i) < t(i+1) and t(i) <= x < t(i+1). 
    !
    !  If no such i can be found, X lies
    !  outside the support of  the spline F and  bvalue=0.
    !  (the asymmetry in this choice of  i  makes F rightcontinuous)
    !
    !
    ! This assumes, natural parametrization
    If ( i<k .Or. i>n ) Then
       Return
    End If
    !
    !  If K=1 , bvalue = bcoef(i).
    !
    If ( k <= 1_t_int ) Then
       bvalue0 = bcoef(i)
       Return
    End If
    !
    !  Store the K b-spline coefficients relevant for the knot interval
    !  (t(i),t(i+1)) in aj(1),...,aj(k) and compute dl(j)=x-t(i+1-j),
    !  dr(j)=t(i+j)-x, j=1,...,k-1 . set any of the aj not obtainable
    !  from input to zero. set any t.s not obtainable equal to t(1) or
    !  to t(n+k) appropriately.
    !
    jcmin = 1_t_int
    
    If ( k <= i ) Then
       Do j=1,k-1
          dl(j)=x-t(i+1-j)
       End Do
    Else
       jcmin=1_t_int-(i-k)
       Do j=1,i
          dl(j)=x-t(i+1-j)
       End Do
       !!
       Do j=i,k-1
          aj(k-j)=0.0_t_real
          dl(j)=dl(i)
       End Do
    End If
    !!
    jcmax=k
    !!
    If ( i <= n ) Then
       Do j=1,k-1
          dr(j)=t(i+j)-x
       End Do
    Else
       jcmax=k+n-i
       Do j=1,k+n-i
          dr(j)=t(i+j)-x
       End Do
       
       Do j=k+n-i, k-1
          aj(j+1)=0.0_t_real
          dr(j)=dr(k+n-i)
       End Do
    Endif
    !!
    Do jc=jcmin,jcmax
       aj(jc)=bcoef(i-k+jc)
    End Do
    !
    !
    !  Compute value at X in (t(i),t(i+1)),
    !  given its relevant b-spline coeffs in aj(1),...,aj(k).
    !
    Do j=1_t_int, k-1_t_int
       ilo=k-j
       Do jj=1_t_int,k-j
          aj(jj)=(aj(jj+1)*dl(ilo)+aj(jj)*dr(jj))/(dl(ilo)+dr(jj))
          ilo=ilo-1_t_int
       End Do
    End Do
    !!
    bvalue0 = aj(1)
    !!
    Return
  End Function bvalue0
  !!***********************************************************************
  !! setup natural parametrization (knots)
  Subroutine set_natural_knots_x(n,k,x,t)
    !!
    Implicit None  
    !!IO
    !! order
    Integer(t_int),Intent(in)    :: k
    !! No. of mesh points
    Integer(t_int),Intent(in)    :: n
    !! mesh points
    Real(t_real),Intent(in)      :: x(n)
    !! knots
    Real(t_real),Intent(out)     :: t(n+k)
    !!LOCAL
    Character(*),Parameter::nsubfunc="set_natural_knots_x"
    !!
    If(n<k+1)Then
       Call terminate(TERM_CRASH&
            &,Trim(nsubfunc)//': n must be larger than k!')
    Endif
    If(Mod(k,2_t_int)/=0_t_int)Then
       Call terminate(TERM_CRASH&
            &,Trim(nsubfunc)//': order must be even!')
    Endif
    t(1:k)=x(1)
    t(k+1:n)=x(k+1-k/2:n-k/2)
    t(n+1:n+k)=x(n)
    !!
  End Subroutine set_natural_knots_x
  !!***********************************************************************
  !! setup natural parametrization (knots)
  Subroutine set_natural_knots(n,k,r,t)
    !!
    Implicit None  
    !!IO
    !! order
    Integer(t_int),Intent(in)    :: k
    !! No. of mesh points
    Integer(t_int),Intent(in)    :: n
    !! mesh points
    Real(t_real),Intent(in)      :: r(n)
    !! knots
    Real(t_real),Intent(out)     :: t(n+k)
    !!LOCAL
    Character(*),Parameter::nsubfunc="set_natural_knots"
    !!
    If(n<k+1)Then
       Call terminate(TERM_CRASH&
            &,Trim(nsubfunc)//': n must be larger than k!')
    Endif
    If(Mod(k,2_t_int)/=0_t_int)Then
       Call terminate(TERM_CRASH&
            &,Trim(nsubfunc)//': order must be even!')
    Endif
    t(1:k)=r(1)
    t(k+1:n)=r(k+1-k/2:n-k/2)
    t(n+1:n+k)=r(n)
    !!
  End Subroutine set_natural_knots
  !!***********************************************************************
  !! supposed set_natural_knots defined the knots for a paramatrized 
  !! equidistant mesh r(i)=g(x(i)), x(i)=x0+(i-1)*h, g= any usful function
  !!
  !! for the ordinate x=g^-1(r)
  !! return the index of the relevant knot interval or 
  !! zero for x<spline support or
  !! something>n  for x>spline support
  !!
  !!***********************************************************************
  Integer(t_int) Function natural_knot_index(k,n,hx,xmesh,x)
    !!
    Implicit none  
    !!IO
    !! order of b spline
    Integer(t_int),Intent(in)    :: k
    !! number of mesh points
    Integer(t_int),Intent(in)    :: n
    !! ordinate under consideration
    Real(t_real),Intent(in)      :: x
    !! step with of the x-mesh
    Real(t_real),Intent(in)      :: hx
    !! x-mesh
    Real(t_real),Intent(in)      :: xmesh(n)
    !!LOCAL
    Integer(t_int) :: index
    !!
    index=Floor((x-xmesh(1))/hx)+1
    If(index<1)Then
       natural_knot_index=0_t_int
    Else If(index>n)Then
       natural_knot_index=index
    Else
       natural_knot_index=Min(Max(k,index+k/2),n)
    Endif
    !!
  End Function natural_knot_index
  !!***********************************************************************
!!$  !!                   /       ml          |        |       mr        \
!!$  !! p_R(i,j)=        {  (d/dt)    N_i(t)  |  t^lex | (d/dt)    N_j(t) }  
!!$  !!                   \                   |        |                 /
!!$  !! with N_i being the i-th b-spline basis function
!!$  !! p_R(i,j) is is stored as banded matrix p_R(i,l)  
!!$  !!         i       \in [1,n]
!!$  !!         l=j-i+k \in [1,2*p+1 ]
!!$  !! p_R must be nullified on call and is allocated here, and must be freed
!!$  !! with pfree after final usage
!!$  !! On input nl=ml+1 and nr=mr+1
!!$  !!
!!$  !! if rightrot== .true.  then the basis functions {N_{n-1},N_{n}} are
!!$  !! transformed into {\tilde{N}_{n-1},\tilde{N}_{n}}={N_{n-1},N_{n}} U
!!$  !!   / 1  -1 \       1
!!$  !! U=|       |   --------
!!$  !!   \ 1   1 /    sqrt(2) 
!!$  !! which results in the first derivative of \tilde{N}_{n-1} at t=t_{n+1}
!!$  !! to be zero. In the transformed basis only \tilde{N}_{n} has a nonzero
!!$  !! derivative at the maximum mesh point t_{n+1}.
!!$  !! This transformation would increase the bandwidth of p_R by 1.
!!$  !! However, the flag rightrot is meant to be used in conjunction with 
!!$  !! cases, where the last transformed basis function is droped.
!!$  !! The returned matrix p_R(i,j)  is only correct for i,j<n !!!
!!$  !!
!!$  !! if leftrot== .true.  then the basis functions {N_{1},N_{2}} are
!!$  !! transformed into {\tilde{N}_{1},\tilde{N}_{2}}={N_{1},N_{2}} U
!!$  !!   / 1   1 \       1
!!$  !! U=|       |   --------
!!$  !!   \-1   1 /    sqrt(2) 
!!$  !! which results in the first derivative of \tilde{N}_{2} at t=t_{k}
!!$  !! to be zero. In the transformed basis only \tilde{N}_{1} has a nonzero
!!$  !! derivative at the minimum mesh point t_{k}.
!!$  !! This transformation would increase the bandwidth of p_R by 1.
!!$  !! However, the flag leftrot is meant to be used in conjunction with 
!!$  !! cases, where the first transformed basis function is droped.
!!$  !! The returned matrix p_R(i,j)  is only correct for i,j>1 !!!
!!$  Subroutine  bsp_quad_S(n,k,knots,nl,lex,nr,p_R,leftrot,rightrot)
!!$    !!
!!$    Implicit none  
!!$    !!IO
!!$    !!numberof basis splines
!!$    Integer(t_int),Intent(in)    :: n
!!$    !! order of bspline
!!$    Integer(t_int),Intent(in)    :: k
!!$    Real(t_real),Intent(in)      :: knots(:)
!!$    !! rotate basis such that only N_n has a nonzero
!!$    !! derivative at t=t_max, only N_{n-1} and N_n  are affected
!!$    Logical(t_log),Intent(in)    :: rightrot
!!$    !! rotate basis such that only N_1 has a nonzero
!!$    !! derivative at t=t_1, only N_{1} and N_{2}  are affected
!!$    Logical(t_log),Intent(in)    :: leftrot
!!$    !! the exponential of the kernel r^lex
!!$    Integer(t_int),Intent(in)    :: lex
!!$    !! derivative degree +1 of the left B-spline basis function
!!$    Integer(t_int),Intent(in)    ::  nl
!!$    !! derivative degree +1 of the right B-spline basis function
!!$    Integer(t_int),Intent(in)    ::  nr
!!$    Real(t_real),Pointer         :: p_R(:,:)
!!$    !!LOCAL
!!$    Character(*),Parameter::nsubfunc="bsp_quad_S"
!!$    Integer(t_int) :: p
!!$    Integer(t_int) :: ileft,iwork,li,lj,i,j,jj,l,nderiv,id
!!$    Real(t_real)   :: t,dw,di,d
!!$    Real(t_real)   :: work((k+1)*(k+2)/2),BiAtX(k,Max(nl,nr))
!!$    !! GKP
!!$    Logical(t_log) :: calc
!!$    Integer(t_int) :: count,ndim,mj
!!$    Real(t_real)   :: jac
!!$    Type(T_Quadgkp) :: gk
!!$    !!
!!$    Call init_quadgkp(gk)
!!$    !!
!!$    p=k-1
!!$    mj=k+p
!!$    !!
!!$    nderiv=Max(nl,nr)
!!$    !!
!!$    Call palloc('p_R',p_R,1_t_int,mj,1_t_int,n,nmodule,nsubfunc)
!!$    !!
!!$    ndim=k**2
!!$    !!
!!$    Call alloc_quadgkp(gk,ndim)
!!$    gk% typ=IJAC_TYPE_POW
!!$    gk% ex   =2_t_int
!!$    gk% kmin=5_t_int
!!$    gk% epsabs=0.0_t_real
!!$    gk% epsrel=1.0e-12_t_real
!!$    !!
!!$    p_R=0.0_t_real
!!$    !! forall intervals
!!$    Sloop: Do ileft=k,n
!!$       
!!$       gk% b0=knots(ileft)
!!$       gk% b1=knots(ileft+1)
!!$       !! forall gauss integration points
!!$       count=0_t_int
!!$       gk% state=QUADSTART
!!$       
!!$       loop: Do 
!!$          Call quad_gkp(gk,t,jac,calc)
!!$          If(.Not.calc) Exit loop
!!$          
!!$          !! determine all nonzero b-spline basis functions in interval 
!!$          !! S_ileft
!!$          Call bsp_bspvd(n,knots,k,nderiv,t,ileft,k,BiAtX,work,.False.)
!!$          !!
!!$          !! j=ileft-k+l = 2
!!$          !!
!!$          !! ileft==k   , j=k-k+l    l=2
!!$          !! ileft==k+1 , j=k+1-k+l  l=1
!!$          !!
!!$          If(leftrot)Then
!!$             If(ileft==k)Then
!!$                l=1
!!$                Do id=1,nderiv
!!$                   d=BiAtX(l,id)
!!$                   BiAtX(l  ,id)=( d-BiAtX(l+1,id))*sq2o2
!!$                   BiAtX(l+1,id)=( d+BiAtX(l+1,id))*sq2o2
!!$                Enddo
!!$             Endif
!!$             If(ileft==k+1)Then
!!$                l=1
!!$                Do id=1,nderiv
!!$                   BiAtX(l  ,id)=(BiAtX(l,id))*sq2o2
!!$                Enddo
!!$             Endif
!!$          Endif
!!$          !! 
!!$          !!
!!$          !! j=ileft-k+l = n-1  
!!$          !!
!!$          !! ileft==n   , j=n-k+l    l=p
!!$          !! ileft==n-1 , j=n-1-k+l  l=k
!!$          !!
!!$          If(rightrot)Then
!!$             If(ileft==n)Then
!!$                l=p
!!$                Do id=1,nderiv
!!$                   d=BiAtX(l,id)
!!$                   BiAtX(l  ,id)=( d+BiAtX(l+1,id))*sq2o2
!!$                   BiAtX(l+1,id)=(-d+BiAtX(l+1,id))*sq2o2
!!$                Enddo
!!$             Endif
!!$             If(ileft==n-1)Then
!!$                l=k
!!$                Do id=1,nderiv
!!$                   BiAtX(l  ,id)=(BiAtX(l,id))*sq2o2
!!$                Enddo
!!$             Endif
!!$          Endif
!!$          !! 
!!$          !!
!!$         dw=t**lex*jac
!!$          
!!$          !! create sub matrix
!!$          jj=0_t_int
!!$          Do li=1,k
!!$             !! first N_j in BiAtX is at j=ileft-p
!!$             di=dw*BiAtX(li,nl)
!!$             Do lj=1,k
!!$                jj=jj+1
!!$                gk% p_ff(jj)=di*BiAtX(lj,nr)
!!$             Enddo
!!$          Enddo
!!$          !!
!!$          count=count+1
!!$          
!!$       Enddo loop
!!$       If(.Not. gk% conv) Then
!!$          Write(outfile,30)Trim(nsubfunc),ileft-p,nl,lex,nr
!!$       Endif
!!$30     Format(a,': not converged in interval S_=',i4,' nl=',i3,' lex=',i3&
!!$            &,' nr=',i3) 
!!$       !Print *,ileft-p,'count=',count
!!$       
!!$       !! fill in matrix
!!$       jj=0_t_int
!!$       Do li=1,k
!!$          i=li+ileft-k
!!$          Do lj=1,k
!!$             j=lj+ileft-k
!!$             !! j  \in [i-p,i+p]
!!$             !! j=l+i-k,   l \in [1,2*p+1] 
!!$             l=j-i+k
!!$             jj=jj+1
!!$             p_R(l,i)=p_R(l,i)+gk% p_res(jj)
!!$          Enddo
!!$       Enddo
!!$       !!
!!$    Enddo Sloop
!!$    !!
!!$    Call clean_quadgkp(gk)
!!$    !!
!!$    Return
!!$    !!
!!$  End Subroutine bsp_quad_S
!!$  !!***********************************************************************
!!$  !!                  /       ml          |              |       mr        \
!!$  !! p_NNN(i,j,n)=   {  (d/dt)    N_i(t)  | N_n(t) t^rex | (d/dt)    N_j(t) }
!!$  !!                  \                   |              |                 /
!!$  !! with N_i being the i-th b-spline basis function
!!$  !! p_NNN(i,j,n) is stored as banded matrix p_NNN(ln,lj,i)
!!$  !!         i        \in [1,n]
!!$  !!         lj=j-i+k \in [1,2*p+1 ]
!!$  !!         ln=n-i+k \in [1,2*p+1 ]
!!$  !! p_V must be nullified on call and is allocated here, and must be freed
!!$  !! with pfree after final usage
!!$  !! On input nl=ml+1 and nr=mr+1
!!$  !!
!!$  !!
!!$  !! if rightrot== .true.  then the basis functions {N_{n-1},N_{n}} are
!!$  !! transformed into {\tilde{N}_{n-1},\tilde{N}_{n}}={N_{n-1},N_{n}} U
!!$  !!   / 1  -1 \       1
!!$  !! U=|       |   --------
!!$  !!   \ 1   1 /    sqrt(2) 
!!$  !! which results in the first derivative of \tilde{N}_{n-1} at t=t_{n+1}
!!$  !! to be zero. In the transformed basis only \tilde{N}_{n} has a nonzero
!!$  !! derivative at the maximum mesh point t_{n+1}.
!!$  !! This transformation would increase the bandwidth of p_R by 1.
!!$  !! However, the flag rightrot is meant to be used in conjunction with 
!!$  !! cases, where the last transformed basis function is droped.
!!$  !! The returned matrix p_R(i,j)  is only correct for i,j<n !!!
!!$  !!
!!$  !! if leftrot== .true.  then the basis functions {N_{1},N_{2}} are
!!$  !! transformed into {\tilde{N}_{1},\tilde{N}_{2}}={N_{1},N_{2}} U
!!$  !!   / 1   1 \       1
!!$  !! U=|       |   --------
!!$  !!   \-1   1 /    sqrt(2) 
!!$  !! which results in the first derivative of \tilde{N}_{2} at t=t_{k}
!!$  !! to be zero. In the transformed basis only \tilde{N}_{1} has a nonzero
!!$  !! derivative at the minimum mesh point t_{k}.
!!$  !! This transformation would increase the bandwidth of p_R by 1.
!!$  !! However, the flag leftrot is meant to be used in conjunction with 
!!$  !! cases, where the first transformed basis function is droped.
!!$  !! The returned matrix p_R(i,j)  is only correct for i,j>1 !!!
!!$  Subroutine  bsp_quad_NNN(n,k,knots,nl,rex,nr,p_NNN,leftrot,rightrot)
!!$    !!
!!$    Implicit none  
!!$    !!IO
!!$    !!numberof basis splines
!!$    Integer(t_int),Intent(in)    :: n
!!$    !! order of bspline
!!$    Integer(t_int),Intent(in)    :: k
!!$    Real(t_real),Intent(in)      :: knots(:)
!!$    !! rotate basis such that only N_1 has a nonzero
!!$    !! derivative at t=t_min, only N_{1} and N_2  are affected
!!$    Logical(t_log),Intent(in)    :: leftrot
!!$    !! rotate basis such that only N_n has a nonzero
!!$    !! derivative at t=t_max, only N_{n-1} and N_n  are affected
!!$    Logical(t_log),Intent(in)    :: rightrot
!!$    !! the exponential of the kernel r^rex
!!$    Real(t_real),Intent(in)      :: rex
!!$    !! derivative degree +1 of the left B-spline basis function
!!$    Integer(t_int),Intent(in)    ::  nl
!!$    !! derivative degree +1 of the right B-spline basis function
!!$    Integer(t_int),Intent(in)    ::  nr
!!$    Real(t_real),Pointer         :: p_NNN(:,:,:)
!!$    !!LOCAL
!!$    Character(*),Parameter::nsubfunc="bsp_quad_NNN"
!!$    Integer(t_int) :: p
!!$    Integer(t_int) :: ileft,iwork,li,lj,i,j,jj,l,nderiv,id,lm,m
!!$    Real(t_real)   :: t,dw,di,d,dj
!!$    Real(t_real)   :: work((k+1)*(k+2)/2),BiAtX(k,Max(nl,nr))
!!$    !! GKP
!!$    Logical(t_log) :: calc
!!$    Integer(t_int) :: count,ndim,mj
!!$    Real(t_real)   :: jac
!!$    Type(T_Quadgkp) :: gk
!!$    !!
!!$    Call init_quadgkp(gk)
!!$    !!
!!$    p=k-1
!!$    mj=k+p
!!$    !!
!!$    nderiv=Max(nl,nr)
!!$    !!
!!$    Call palloc('p_NNN',p_NNN,1_t_int,mj,1_t_int,mj,1_t_int,n&
!!$         &,nmodule,nsubfunc)
!!$    !!
!!$    ndim=k**3
!!$    Call alloc_quadgkp(gk,ndim)
!!$    gk% typ=IJAC_TYPE_POW
!!$    gk% ex   =2_t_int
!!$    gk% kmin=5_t_int
!!$    gk% epsabs=0.0_t_real
!!$    gk% epsrel=1.0e-12_t_real
!!$    !!
!!$    !!
!!$    p_NNN=0.0_t_real
!!$    !! forall intervals
!!$    Sloop: Do ileft=k,n
!!$       
!!$       gk% b0=knots(ileft)
!!$       gk% b1=knots(ileft+1)
!!$       !! forall gauss integration points
!!$       count=0_t_int
!!$       gk% state=QUADSTART
!!$       
!!$       loop: Do 
!!$          Call quad_gkp(gk,t,jac,calc)
!!$          If(.Not.calc) Exit loop
!!$          
!!$          !! determine all nonzero b-spline basis functions in interval 
!!$          !! S_ileft
!!$          Call bsp_bspvd(n,knots,k,nderiv,t,ileft,k,BiAtX,work,.False.)
!!$          !!
!!$          !!
!!$          !! j=ileft-k+l = 2
!!$          !!
!!$          !! ileft==k   , j=k-k+l    l=2
!!$          !! ileft==k+1 , j=k+1-k+l  l=1
!!$          !!
!!$          If(leftrot)Then
!!$             If(ileft==k)Then
!!$                l=1
!!$                Do id=1,nderiv
!!$                   d=BiAtX(l,id)
!!$                   BiAtX(l  ,id)=( d-BiAtX(l+1,id))*sq2o2
!!$                   BiAtX(l+1,id)=( d+BiAtX(l+1,id))*sq2o2
!!$                Enddo
!!$             Endif
!!$             If(ileft==k+1)Then
!!$                l=1
!!$                Do id=1,nderiv
!!$                   BiAtX(l  ,id)=(BiAtX(l,id))*sq2o2
!!$                Enddo
!!$             Endif
!!$          Endif
!!$          !! j=ileft-k+l = n-1  
!!$          !!
!!$          !! ileft==n   , j=n-k+l    l=p
!!$          !! ileft==n-1 , j=n-1-k+l  l=k
!!$          !!
!!$          If(rightrot)Then
!!$             If(ileft==n)Then
!!$                l=p
!!$                Do id=1,nderiv
!!$                   d=BiAtX(l,id)
!!$                   BiAtX(l  ,id)=( d+BiAtX(l+1,id))*sq2o2
!!$                   BiAtX(l+1,id)=(-d+BiAtX(l+1,id))*sq2o2
!!$                Enddo
!!$             Endif
!!$             If(ileft==n-1)Then
!!$                l=k
!!$                Do id=1,nderiv
!!$                   BiAtX(l  ,id)=(BiAtX(l,id))*sq2o2
!!$                Enddo
!!$             Endif
!!$          Endif
!!$          !! 
!!$          !!
!!$         dw=t**rex*jac
!!$          
!!$          !! create sub matrix
!!$          jj=0_t_int
!!$          Do li=1,k
!!$             !! first N_j in BiAtX is at j=ileft-p
!!$             di=dw*BiAtX(li,nl)
!!$             Do lj=1,k
!!$                dj=di*BiAtX(lj,nr)
!!$                Do lm=1,k
!!$                   jj=jj+1
!!$                   gk% p_ff(jj)=dj*BiAtX(lm,1)
!!$                Enddo
!!$             Enddo
!!$          Enddo
!!$          !!
!!$          count=count+1
!!$          
!!$       Enddo loop
!!$       If(.Not. gk% conv) Then
!!$          Write(outfile,30)Trim(nsubfunc),ileft-p,nl,rex,nr
!!$       Endif
!!$30     Format(a,': not converged in interval S_=',i4,' nl=',i3&
!!$            &,' lex=',g10.3,' nr=',i3) 
!!$       !Print *,ileft-p,'count=',count
!!$       
!!$       !! fill in matrix
!!$       jj=0_t_int
!!$       Do li=1,k
!!$          i=li+ileft-k
!!$          Do lj=1,k
!!$             !j=lj+ileft-k
!!$             !! j  \in [i-p,i+p]
!!$             !! j=l+i-k,   l \in [1,2*p+1] 
!!$             l=lj-li+k
!!$             Do lm=1,k
!!$                m=lm-li+k
!!$                jj=jj+1
!!$                p_NNN(m,l,i)=p_NNN(m,l,i)+gk% p_res(jj)
!!$             Enddo
!!$          Enddo
!!$       Enddo
!!$       !!
!!$    Enddo Sloop
!!$    !!
!!$    Call clean_quadgkp(gk)
!!$    !!
!!$    Return
!!$    !!
!!$  End Subroutine bsp_quad_NNN
!!$  !!***********************************************************************
!!$  !! calculate g*S*f in the transposed banded representation of S
!!$  Real(t_real) Function bsp_gSf(n,p,g,S,f)
!!$    !!
!!$    Implicit None  
!!$    !!IO
!!$    !! number of cols
!!$    Integer(t_int),Intent(in)    :: n
!!$    !! number of side diagonals
!!$    Integer(t_int),Intent(in)    :: p
!!$    Real(t_real),Intent(in)      :: g(*),f(*),S(2*p+1,*)
!!$    !!LOCAL
!!$    Character(*),Parameter::nsubfunc="bsp_gSf"
!!$    Integer(t_int) :: i,l,j,k,l1,l2
!!$    Real(t_real)   :: d,d1
!!$    !!
!!$    k=p+1
!!$    !!
!!$    d1=0.0_t_real
!!$    Do i=1,n
!!$       d=0.0_t_real
!!$       l1=Max(1,1-i+k)
!!$       l2=Min(2*p+1,n-i+k)
!!$       j=i+l1-k
!!$       Do l=l1,l2
!!$          d=d+S(l,i)*f(j)
!!$          j=j+1
!!$       Enddo
!!$       d1=d1+g(i)*d
!!$    Enddo
!!$    !!
!!$    bsp_gSf=d1
!!$    !!
!!$  End Function bsp_gSf
!!$  !!***********************************************************************
  !!***********************************************************************
End Module bspline
