!!***********************************************************************
!! linear algebra for matrix inversion and linear system solver
!!
Module matrixinvers
  !!
  Use precisiondefs
  !!
  Implicit None 
  !!
  Private
  Public :: rinv,dglei,gaussj,cinvp,cinv,dchol
  Character(*),Parameter::nmodule="MATRIXINVERS"  
  !! *********************************************************************
  Interface cinv
     Module Procedure cinv1,cinv2
  End Interface cinv
  !! *********************************************************************
Contains
  !!***********************************************************************
  !!
  Subroutine rinv(A,R,ndim,M,indef)
    !!
    Implicit None 
    !IO   A  : matrix to be inverted
    !IO   R1  : result, may be the same physical matrix A, which then
    !IO   is overwritten
    !IO   M  : matrix dimension     
    !!IO
    Integer(t_int),Intent(in):: m,ndim
    Logical(t_log),Intent(out):: indef
    Real(t_real),Intent(inout) :: a(ndim,ndim),r(ndim,ndim)
    !!LOCAL
    !Character(*),Parameter::nsubfunc="rinv"
    Real(t_real),Dimension(ndim) :: q12,q21,g12,g21
    Real(t_real) :: s22,g22
    Integer(t_int):: i,k,n
    !
    !***********************************************************************
    indef=.False.
    
    If(A(1_t_int,1_t_int)==0.0_t_real)Then
       indef=.true.
       return
    Endif
    r(1_t_int,1_t_int)=1.0_t_real/A(1_t_int,1_t_int)
    Do n=2_t_int,M
       k=n-1_t_int
       
       q12(1_t_int:k)=Matmul(r(1_t_int:k,1_t_int:k),A(1_t_int:k,n))
       q21(1_t_int:k)=Matmul(A(n,1_t_int:k),r(1_t_int:k,1_t_int:k))
       
       
       s22=Dot_product(A(n,1_t_int:k),q12(1_t_int:k))
       
       g22=(A(n,n)-s22)
       If(g22==0.0_t_real)Then
          indef=.True.
          Return
       Endif
       g22=1.0_t_real/g22
       r(n,n)=g22
       
       g21(1_t_int:k)=-g22*q21(1_t_int:k)
       r(n,1_t_int:k)=g21(1_t_int:k)
       g12(1_t_int:k)=-(q12(1_t_int:k)*g22)
       r(1_t_int:k,n)=g12(1_t_int:k)
       
       Do i=1_t_int,k
          r(i,1_t_int:k)=r(i,1_t_int:k)-(q12(i)*g21(1_t_int:k))
       End Do                ! i

    End Do                   ! n
    
    
    !
  End Subroutine rinv
  
  !
  !***********************************************************************
  
  !!***********************************************************************
  !!***********************************************************************
  !!     solution of a linear inhomogenous equation system with a
  !!     symmetric positive definite real matrix.
  !!     double precision algebra.
  !!
  !!     input:
  !!       a     : matrix of the problem
  !!       b     : inhomogenity vector
  !!       n     : order of the problem
  !!       m     : specified dimension of a(m,m)
  !!
  !!     output:
  !!       x     : solution vector
  !!***********************************************************************

  Subroutine dglei (a,b,x,dim,n,indef)
    !!
    Implicit None 
    !!IO
    Integer(t_int),Intent(in)  :: n  ,dim
    Logical(t_log),Intent(out) :: indef  
    Real(t_real),Intent(in),Dimension(dim)   :: b
    Real(t_real),Intent(inout),Dimension(dim)   :: x
    Real(t_real),Intent(inout),Dimension(dim,dim) :: a

    !!LOCAL
    !Character(*),Parameter::nsubfunc="dglei"
    Integer(t_int) :: i,i1
    Real(t_real)  :: s
    !!
    !!***********************************************************************
    
    !!
    Call dchol(a,dim,n,indef)
    If(indef) Return
    !!
    x(1_t_int:n)=b(1_t_int:n)
    
    s=x(1_t_int)/a(1_t_int,1_t_int)
    x(1_t_int)=s
    Do i=2_t_int,n
       x(i:n)=x(i:n)-s*a(i-1_t_int,i:n)
       s=x(i)/a(i,i)
       x(i)=s
    End Do
    
    s=x(n)/a(n,n)
    x(n)=s
    Do i1=2_t_int,n
       i=n+2_t_int-i1
       x(n+1_t_int-i1:1_t_int:-1_t_int)=x(n+1_t_int-i1:1_t_int:-1_t_int)&
            &-a(n+1_t_int-i1:1_t_int:-1_t_int,i)*s
       
       i=i-1_t_int
       s=x(i)/a(i,i)
       x(i)=s
    End Do
    
    Return
  End Subroutine dglei
  !!***********************************************************************
  !!     cholesky procedure for the reduction of a positive definite
  !!     symmetric real matrix into a trigonal matrix.
  !!     double precision algebra.
  !!
  !!     input:
  !!       a     : symmetric matrix
  !!       n     : order of a
  !!       m     : specified dimension of a(m,m)
  !!
  !!     output:
  !!       a     : upper trigonal matrix
  !!***********************************************************************
  
  
  Subroutine dchol (a,dim,n,indef)
    !!
    Implicit None 
    !!IO
    Integer(t_int),Intent(in)   ::n,dim
    Logical(t_log),Intent(out)  ::indef
    Real(t_real),Intent(inout),Dimension(dim,dim) ::  a
    !!LOCAL
    !Character(*),Parameter::nsubfunc="dchol"
    Integer(t_int) :: ip,kp,i
    real(t_real)  :: ap
    !!
    !!***********************************************************************
    !!
    indef=.false.
    Do ip=1_t_int,n
       ap=a(ip,ip)
       If(ap<=0.0_t_real)Then
          indef=.true.
          Exit
       End If
       ap=Sqrt(ap)
       a(ip,ip)=ap
       If(ip==n) Exit
       kp=ip+1_t_int
       a(ip,kp:n)=a(ip,kp:n)/ap
       
       Do i=kp,n
          a(i,i:n)=a(i,i:n)-a(ip,i)*a(ip,i:n)
       End Do
    End Do
    
    Return
  End Subroutine dchol
  !!***********************************************************************
  !!***********************************************************************
  !!
  !! gauss jordan elimination with full pivoting
  !! solve A x = b
  !! matrix A(n,n) is stored as A(np,np)
  !! b(n,m) is stored as b(np,mp)
  !!
  !! output: A -> A^-1 
  !!         b -> solutions
  !! 
  !!
  Subroutine gaussj(a,n,np,b,m,mp,indef)
    !!
    Implicit None 
    !!IO
    Integer(t_int),Intent(in)    :: m,mp,n,np
    Logical(t_log),Intent(out)         :: indef
    Real(t_real),Intent(inout)  :: a(np,np),b(np,mp)
    !!LOCAL
    !Character(*),Parameter::nsubfunc="gaussj"
    Integer(t_int) :: indxc(n),indxr(n),ipiv(n)
    Integer(t_int) :: i,icol,irow,j,k,l,ll
    Real(t_real)  :: big,dum,pivinv
    !!
    indef=.false.
    ipiv(1_t_int:n)=0_t_int

    Do i=1_t_int,n
       big=0.0_t_real
       icol=1
       irow=1
       Do j=1_t_int,n
          If(ipiv(j) /= 1_t_int)Then
             Do k=1_t_int,n
                If (ipiv(k) == 0_t_int) Then
                   If (Abs(a(j,k)) >= big)Then
                      big=Abs(a(j,k))
                      irow=j
                      icol=k
                   Endif
                Else If (ipiv(k) > 1_t_int) Then
                   indef=.True.
                   Return
                Endif
             End Do
          Endif
       Enddo
       ipiv(icol)=ipiv(icol)+1_t_int
       If (irow /= icol) Then
          Do l=1_t_int,n
             dum=a(irow,l)
             a(irow,l)=a(icol,l)
             a(icol,l)=dum
          End Do
          Do l=1_t_int,m
             dum=b(irow,l)
             b(irow,l)=b(icol,l)
             b(icol,l)=dum
          End Do
       Endif
       indxr(i)=irow
       indxc(i)=icol
       If (a(icol,icol) == 0.0_t_real)Then
          indef=.True.
          Return
       Endif
       pivinv=1.0_t_real/a(icol,icol)
       a(icol,icol)=1.0_t_real

       a(icol,1_t_int:n)=a(icol,1_t_int:n)*pivinv

       b(icol,1_t_int:m)=b(icol,1_t_int:m)*pivinv

       Do ll=1_t_int,n
          If(ll /= icol)Then
             dum=a(ll,icol)
             a(ll,icol)=0.0_t_real
             a(ll,1_t_int:n)=a(ll,1_t_int:n)-a(icol,1_t_int:n)*dum
             b(ll,1_t_int:m)=b(ll,1_t_int:m)-b(icol,1_t_int:m)*dum

          Endif
       Enddo
    Enddo
    Do l=n,1_t_int,-1_t_int
       If(indxr(l) /= indxc(l))Then
          Do k=1_t_int,n
             dum=a(k,indxr(l))
             a(k,indxr(l))=a(k,indxc(l))
             a(k,indxc(l))=dum
          Enddo
       Endif
    Enddo
    Return
    !!
  End Subroutine gaussj
  !!
  !!  (c) copr. 1986-92 numerical recipes software -0'.
  !!***********************************************************************
  !!***********************************************************************
  !!   invert complex matrix a by gauss elimination with partial
  !!         pivoting
  !!
  !!      important: ndim <= ntot
  !!***********************************************************************
  Subroutine cinvp(a,w,aux,ind,ndim,ntot)
    !!
    Implicit None
    !!IO
    Integer(t_int),Intent(in)    :: ndim,ntot
    Complex(t_real),Intent(inout):: a(ntot,ndim)
    Complex(t_real),Intent(out)  :: w(ntot,ndim)
    Real(t_real),Intent(out)     :: aux(ndim)
    Integer(t_int),Intent(out)   :: ind(ndim)
    !!LOCAL
    Integer(t_int) :: i,j,n,k
    Complex(t_real):: qq
    Complex(t_real),Parameter:: c0=(0.0_t_real,0.0_t_real)
    Complex(t_real),Parameter:: c1=(1.0_t_real,0.0_t_real)
    !!
    n=ndim
    !!
    w(1_t_int:n,1_t_int:n)=a(1_t_int:n,1_t_int:n)
    !!
    Do i=1_t_int,n
       aux(i:n)=Abs(w(i,i:n))
       k=idamax(n-i+1_t_int,aux(i),1_t_int)+i-1_t_int
       ind(i)=k
       If(k /= i) Then
          Do j=1_t_int,n
             qq=w(j,i)
             w(j,i)=w(j,k)
             w(j,k)=qq
          End Do
       End If
       qq= c1/w(i,i)
       w(i,i)=c1
       w(1_t_int:n,i)=qq*w(1_t_int:n,i)
       Do j=1_t_int,n
          If(j /= i) Then
             qq=w(i,j)
             w(i,j)=c0
             w(1_t_int:n,j)=w(1_t_int:n,j)-qq*w(1_t_int:n,i)
          End If
       End Do
    End Do
    
    Do i=n,1_t_int,-1_t_int
       j=ind(i)
       If(i /= j) Then
          Do k=1_t_int,n
             qq=w(i,k)
             w(i,k)=w(j,k)
             w(j,k)=qq
          End Do
       End If
    End Do
    
    a(1_t_int:n,1_t_int:n)=w(1_t_int:n,1_t_int:n)
    
    
    Return
  End Subroutine cinvp
  !!***********************************************************************
  !!
  Integer(t_int) Function idamax(n,dx,incx)
    !!
    Implicit None
    !!
    !!     finds the index of element having max. absolute value.
    !!     jack dongarra, linpack, 3/11/78.
    !!
    !!IO
    Integer(t_int),Intent(in)    :: n,incx
    Real(t_real),Intent(in)      :: dx(*)
    !!LOCAL
    Real(t_real)   ::  dmax
    Integer(t_int) ::  i,ix
    !!
    idamax = 0_t_int
    If( n < 1_t_int ) Return
    idamax = 1_t_int
    If(n == 1_t_int)Return
    If(incx == 1_t_int)go to 20
    !!
    !!        code for increment not equal to 1
    !!
    ix = 1_t_int
    dmax = Abs(dx(1_t_int))
    Do i = 2_t_int,n
       ix = ix + incx
       If(Abs(dx(ix)) <= dmax) Cycle
       idamax = i
       dmax = Abs(dx(ix))
    End Do
    
    Return
    !!
    !!        code for increment equal to 1
    !!
20  Continue
    dmax = Abs(dx(1_t_int))
    Do i = 2_t_int,n
       If(Abs(dx(i)) <= dmax) Cycle
       idamax = i
       dmax = Abs(dx(i))
    End Do
    
    Return
  End Function idamax
  !!***********************************************************************
  !!***********************************************************************
  !!     invers of a complex matrix by partitioning
  !!     + log det A = Tr Log A, in a paricular way
  !!     if one assumes that 1/A is analytic in the upper half plane
  !!     than log det A is calculated in a way that all phases of Im Tr Log A
  !!     are summed up. This is important for the integrated density of states
  !!***********************************************************************
  Subroutine cinv1(a,r,m,dim,sumlog,sing,infinite,indef,suminf)
    !!
    !!IO
    Logical(t_log),Intent(in)    :: infinite
    Logical(t_log),Intent(out)   :: sing,indef
    Integer(t_int),Intent(in)    :: m,dim
    Complex(t_real),Intent(in)   :: a(dim,dim)
    Complex(t_real),Intent(out)  :: r(dim,dim),sumlog
    Real(t_real),Intent(out)     :: suminf
    !!LOCAL
    Complex(t_real):: s22,g22,lo
    !!AUTO ARRAYS
    Complex(t_real):: q12(dim),q21(dim),g12(dim),g21(dim)
    Complex(t_real),Parameter:: c0=(0.0_t_real,0.0_t_real)
    Complex(t_real),Parameter:: c1=(1.0_t_real,0.0_t_real)
    Integer(t_int):: i,k,n
    !!
    !!*********************************************************
    !!
    sing=.False.
    indef=.False.
    If(A(1,1) == c0)Then
       sing=.True.
       Return
    Endif
    !!
    suminf=0.0_t_real
    lo=Log(a(1,1))
    sumlog=lo
    If(infinite)Then
       If( Abs(Aimag(a(1,1)))>1.e-10_t_real ) Then!! fragil
          indef=.True.
       Endif
       If( Real(a(1,1),t_real)>0.0_t_real ) Then
          suminf=suminf+1.0_t_real
       Else
          indef=.True.
       Endif
    Else
       If(Aimag(lo)<-1.e-10_t_real)Then
          indef=.True.
       Endif
    Endif
    !!
    r(1,1)=c1/a(1,1)
    !!
    Do n=2_t_int,m
       k=n-1_t_int
       !!
       !!
       q12(1:k)=Matmul(r(1:k,1:k),a(1:k,n))
       q21(1:k)=Matmul(a(n,1:k),r(1:k,1:k))
       !!
       s22=Dot_product(Conjg(a(n,1:k)),q12(1:k))
       !!
       g22=(a(n,n)-s22)
       If(g22 == c0)Then
          sing=.True.
          Return
       Endif
       !! logarithm of determinant
       lo=Log(g22)
       sumlog=sumlog+lo

       If(infinite)Then
          If( Abs(Aimag(g22))>1.e-10_t_real ) Then!! fragil
             indef=.True.
          Endif
          If( Real(g22,t_real)>0.0_t_real ) Then
             suminf=suminf+1.0_t_real
          Else
             indef=.True.
          Endif
       Else
          !!Print *,'lo=',lo
          If(Aimag(lo)<-1.e-10_t_real)Then
             indef=.True.
          Endif
       Endif
       

       g22=c1/g22

       !!
       r(n,n)=g22
       
       g21(1:k)=-g22*q21(1:k)
       r(n,1:k)=g21(1:k)
       !!
       g12(1:k)=-(q12(1:k)*g22)
       r(1:k,n)=g12(1:k)
       !!
       Do i=1,k
          r(i,1:k)=r(i,1:k)-(q12(i)*g21(1:k))
       End Do                ! i
       
    End Do
    !!
    Return
  End Subroutine cinv1
  !! *********************************************************************
  Subroutine cinv2(a,m,dim,sumlog,sing,infinite,indef,suminf)
    !!
    !!IO
    Logical(t_log),Intent(in)    :: infinite
    Logical(t_log),Intent(out)   :: sing,indef
    Integer(t_int),Intent(in)    :: m,dim
    Complex(t_real),Intent(inout):: a(dim,dim)
    Complex(t_real),Intent(out)  :: sumlog
    Real(t_real),Intent(out)     :: suminf
    !!LOCAL
    Complex(t_real):: s22,g22,lo
    !!AUTO ARRAYS
    Complex(t_real):: q12(dim),q21(dim),g12(dim),g21(dim)
    Complex(t_real),Parameter:: c0=(0.0_t_real,0.0_t_real)
    Complex(t_real),Parameter:: c1=(1.0_t_real,0.0_t_real)
    Integer(t_int):: i,k,n
    !!
    !!*********************************************************
    !!
    sing=.False.
    indef=.False.
    If(A(1,1) == c0)Then
       sing=.True.
       Return
    Endif
    !!
    suminf=0.0_t_real
    lo=Log(a(1,1))
    sumlog=lo
    If(infinite)Then
       If( Abs(Aimag(a(1,1)))>1.e-10_t_real ) Then!! fragil
          indef=.True.
       Endif
       If( Real(a(1,1),t_real)>0.0_t_real ) Then
          suminf=suminf+1.0_t_real
       Else
          indef=.True.
       Endif
    Else
       If(Aimag(lo)<-1.e-10_t_real)Then
          indef=.True.
       Endif
    Endif
    !!
    A(1,1)=c1/a(1,1)
    !!
    Do n=2_t_int,m
       k=n-1_t_int
       !!
       !!
       q12(1:k)=Matmul(A(1:k,1:k),a(1:k,n))
       q21(1:k)=Matmul(a(n,1:k),A(1:k,1:k))
       !!
       s22=Dot_product(Conjg(a(n,1:k)),q12(1:k))
       !!
       g22=(a(n,n)-s22)
       If(g22 == c0)Then
          sing=.True.
          Return
       Endif
       !! logarithm of determinant
       lo=Log(g22)
       sumlog=sumlog+lo

       If(infinite)Then
          If( Abs(Aimag(g22))>1.e-10_t_real ) Then!! fragil
             indef=.True.
          Endif
          If( Real(g22,t_real)>0.0_t_real ) Then
             suminf=suminf+1.0_t_real
          Else
             indef=.True.
          Endif
       Else
          !!Print *,'lo=',lo
          If(Aimag(lo)<-1.e-10_t_real)Then
             indef=.True.
          Endif
       Endif
       

       g22=c1/g22

       !!
       A(n,n)=g22
       
       g21(1:k)=-g22*q21(1:k)
       A(n,1:k)=g21(1:k)
       !!
       g12(1:k)=-(q12(1:k)*g22)
       A(1:k,n)=g12(1:k)
       !!
       Do i=1,k
          A(i,1:k)=A(i,1:k)-(q12(i)*g21(1:k))
       End Do                ! i
       
    End Do
    !!
    Return
  End Subroutine cinv2
  !!***********************************************************************

  
End Module matrixinvers

