Module tmesh
  !!
  Use precisiondefs
  Use errors
  Use memory
  !!
  Implicit None 
  !!
  Private
  !!
  Character(*),Parameter::nmodule="TMESH"  
  !!
  Public :: T_Mesh,init_mesh,cleanup_mesh&
       &,MESH_TYPE,LOG_MESH,LIN_MESH,ROOT_MESH,LOGZERO_MESH,POW_MESH&
       &,palloc_meshs,pfree_meshs,copy_mesh
  !!***********************************************************************
  Integer(t_int),Parameter   :: LOG_MESH     =1_t_int
  Integer(t_int),Parameter   :: LIN_MESH     =2_t_int
  Integer(t_int),Parameter   :: ROOT_MESH    =3_t_int
  Integer(t_int),Parameter   :: LOGZERO_MESH =4_t_int
  Integer(t_int),Parameter   :: POW_MESH     =5_t_int
  Character(16)              :: MESH_TYPE(5)
  Data MESH_TYPE&
       &/"Logarithmic     "&
       &,"Linear          "&
       &,"Root            "&
       &,"Logarithmic-zero"&
       &,"Power           "&
       &/

  !!***********************************************************************
  !! a mesh is defined as
  !!     r_i=g(x_i),   x_i=x0+(i-1)*hx,  i \in [1,Nr]
  !!     r_1=rmin
  !!     r_Nr=rmax
  Type T_Mesh
     !! Is the point r=0 part of the mesh (first mesh point)?
     !! This is not possible for all mesh types. Including zero requires
     !! special treatment of this point in some cases.
     !! for the logarithmic mesh it is not possible
     !! for the linear      mesh it is not possible whithout setting rmin==0
     !! for the logarithmic-zero mesh it IS possible
     !! for the root             mesh it IS possible
     !! 
     !! The mesh usually is defined by a minimal and maximal mesh point and
     !! by the number of mesh points. For some types it is possible to
     !! set r(1)=0 and r(2)=rmin maintaining the equidistancy of the x-mesh.
     !! This is done if with_zero==.True..
     Logical(t_log)        :: with_zero 
     !! type of mesh
     Integer(t_int)        :: type
     !! number of mesh points
     Integer(t_int)        :: nr
     !! minimal, maximal mesh point, step width of x-mesh, mesh-paramter
     !!   rmin is the input quantity and not the real minimum radius
     !!   for meshes, which cannot fix r(2)=rmin, rmin /= r(2)
     Real(T_real)          :: rmin,rmax,hx,alpha
     !! the radialmesh
     Real(T_real),Pointer  :: p_rmesh(:)
     !! the x-mesh
     Real(T_real),Pointer  :: p_xmesh(:)
     !! the jacobian of the mesh (d r(x)/dx)
     Real(T_real),Pointer  :: p_jmesh(:)
     !! for bsplines -------------------------------------
     Logical(t_log) :: knot_x !! take x-mesh as knot mesh
     !! order and degree of bspline
     Integer(t_int) :: k,p
     !! knot vector
     Real(T_real),Pointer  :: p_knots(:)
     !! band matrix
     Real(T_real),Pointer  :: p_q(:)
  End Type T_Mesh
  !!***********************************************************************
contains
  Subroutine init_mesh(me)
    Implicit None 
    !!IO
    Type(T_Mesh),Intent(inout) :: me
    !!LOCAL
    me% type=LOG_MESH
    me% nr=0_t_int
    me% with_zero=.True.
    me% knot_x=.False.
    Nullify(me% p_rmesh)
    Nullify(me% p_xmesh)
    Nullify(me% p_jmesh)
    me% k=4_t_int 
    Nullify(me% p_knots)
    Nullify(me% p_q)
  End Subroutine init_mesh
  !!***********************************************************************
  Subroutine cleanup_mesh(mesh)
    !!
    Implicit None
    !!IO
    Type(T_Mesh),Intent(inout) :: mesh
    !!LOCAL
    !Character(*),Parameter::nsubfunc="cleanup_mesh"
    !!
    Call pfree(mesh % p_rmesh)
    Call pfree(mesh % p_xmesh)
    Call pfree(mesh % p_jmesh)
    Call pfree(mesh % p_knots)
    Call pfree(mesh % p_q)

    !!
    Return
  End Subroutine cleanup_mesh
  !!***********************************************************************
  Subroutine palloc_meshs(name,meshs,N,nmod,nsub)
    !!
    Implicit None
    !!IO
    Integer(t_int),Intent(in)    :: N
    Type(T_Mesh),Pointer         :: meshs(:)
    Character(*),Intent(in)      :: name,nmod,nsub
    !!LOCAL
    !Character(*),Parameter::nsubfunc="palloc_meshs"
    Integer(t_int) :: i
    Integer :: icheck
    !!
    !!
    Call pfree_meshs(meshs)
    !!
    Allocate(meshs(1:N),STAT=icheck)
    If(icheck/=0)Then
       Call errmsg(nmod,nsub,E_fatalerror&
            &,msg="'Could not allocate memory for "//Trim(name)//" !'")
    Endif
    !!
    !!
    Do i=1,N
       Call init_mesh(meshs(i))
    Enddo
    !!
    !If(STATISTICS)Then
    !   Call crealmemaddress(ra,add)
    !   Call register_mem_stat(add,nmod,nsub,name,TYPE_T_REAL&
    !        &,size10,size11&
    !        &)
    !Endif
    !!
    !!
    Return
    !!
  End Subroutine palloc_meshs
  !!***********************************************************************
  Subroutine pfree_meshs(meshs)
    Implicit None 
    !!IO
    Type(T_Mesh),Pointer         :: meshs(:)
    !!LOCAL
    Integer(t_int) :: i
    External crealmemaddress
    !!
    !!
    If(Associated(meshs))Then
       !!
       !If(STATISTICS)Then
       !   Call crealmemaddress(ra,add)
       !   Call free_chunk(add)
       !End If
       !!
       Do i=1,Size(meshs)
          Call cleanup_mesh(meshs(i))
       Enddo
       !!
       Deallocate(meshs)
       Nullify(meshs)
    Endif
    !!
  End Subroutine pfree_meshs
  !!***********************************************************************
  Subroutine copy_mesh(from,to)
    !!
    Implicit None
    !!IO
    Type(T_Mesh),Intent(in)      :: from
    Type(T_Mesh),Intent(inout)   :: to
    !!LOCAL
    Character(*),Parameter::nsubfunc="copy_mesh"
    Integer(t_int) :: N
    !!
    to% with_zero=from% with_zero
    to% type=from% type
    to% nr=from% nr
    to% rmin=from% rmin
    to% rmax=from% rmax
    to% hx=from% hx
    to% alpha=from% alpha
    to% knot_x=from% knot_x
    to% k=from% k
    to% p=from% p
    !!
    N=Size(from% p_rmesh)
    Call palloc('to% p_rmesh',to% p_rmesh,1_t_int,N,nmodule,nsubfunc)
    Call palloc('to% p_xmesh',to% p_xmesh,1_t_int,N,nmodule,nsubfunc)
    Call palloc('to% p_jmesh',to% p_jmesh,1_t_int,N,nmodule,nsubfunc)
    !!
    to% p_rmesh(:)=from% p_rmesh(:)
    to% p_xmesh(:)=from% p_xmesh(:)
    to% p_jmesh(:)=from% p_jmesh(:)

    If(Associated(from% p_knots))Then
       N=Size(from% p_knots)
       Call palloc('to% p_knots',to% p_knots,1_t_int,N,nmodule,nsubfunc)
       to% p_knots(1:N)=from% p_knots(1:N)
    Endif

    If(Associated(from% p_q))Then
       N=Size(from% p_q)
       Call palloc('to% p_q',to% p_q,1_t_int,N,nmodule,nsubfunc)
       to% p_q(:)=from% p_q(:)
    Endif
    !!
    Return
    !!
  End Subroutine copy_mesh
  !!***********************************************************************


End Module tmesh
