/*************************************************************************
 * @(#)Main.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   22 May 2012
 *************************************************************************
 */

#include <stdlib.h>
#include <stdio.h>
#include "Except.hh"
#include "Main.hh"
#include "MyDebug.hh"
#include "Config.hh"
#include "OpticsControl.hh"
#include "Assorted.hh"


String startargs;

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void help()
{
  printf("\n\
This program reads inter-band output from FPLO (+imeps)\n\
and calculates some optical functions. The user can define\n\
plasmon frequencies and broadenings to include the intra-band\n\
contributions.\n\
\n\
The program also reads the file +plasmon to extract the calculated\n\
plasmon energies and the principal axes of the plasmon tensor.\n\
The user can overwrite the plasmon energies by interactive input\n\
or by the option -omega.\n\
\n\
The output functions are \n\
\tre_eps: real part of epsilon\n\
\tim_eps: imaginary part of epsilon\n\
\tre_sigma: real part of the optical conductivity in inverse Ohm*cm\n\
\tloss: the loss function\n\
For the individual spin contributions _spin1/_spin2 is appended to the \n\
file names. (Only for spin polarized cases.)\n\
If an output suffix (-os) is defined it will be appended to all file names.\n\
\n\
");
}
static void usage()
{
  printf("usage:\n\
%s [-i fplooutput_imag_part_interband_eps] ...\n\
\t[-os suffix_for_output_files] [-h]  ...\n\
\t[-gamma gamma1 [gamma2 ... gamman]] ...\n\
\t[-omega omega1 [omega2 ... omegan]] ...\n\
\t[-skip nskip] [-nointer] [-nointra] [-ni]\n\n		\
fplooutput_imag_part_interband_eps: normally +imeps\n\
omega: the plasmon frequencies for the intraband contributions in eV\n\
gamma: the broadening (life time) for the intraband contributions in eV\n\
nskip: take only every nskip data point \n\
-nointer: only take the intraband contributions\n\
-nointra: only take the interband contributions\n\
-ni = not interactively: take values from +plasmon and from -gamma and -omega\n\
but do not ask for not explicitly specified values.\n\
\n\
There are 3 gamma values: for each principal axis one.\n\
There are 3 omega values along the principal axes for each spin direction\n\
and for the total (spin sum). This is so, since the principal axes\n\
need no coincide for spin up, spin down and the spin sum.\n\
In non-spin-polarized cases ony spin1 is used and equals the spin sum.\n\
\n\
",EXECNAMESHORT());

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void processCommandLineArgs(int argc,const char** argv,OpticsControl &oc)
{

  int i=0;
  startargs="";
  while(i<argc){
    startargs+=argv[i++];
    startargs+=" ";
  }
  
  i=1;
  while(i<argc){
    String arg=argv[i++];
    //printf("arg=[%s]\n",arg());
    if(arg == "-i"){
      if(i>=argc) {
 	usage();
 	die("option -i needs an argument\n");
      }
      arg=argv[i++];
      oc.setInfile(arg);
    }else if(arg == "-os"){
      if(i>=argc) {
 	usage();
 	die("option -os needs an argument\n");
      }
      arg=argv[i++];
      oc.setOutSuffix(arg);
    }else if(arg == "-skip"){
      if(i>=argc) {
 	usage();
 	die("option -skip needs an argument\n");
      }
      arg=argv[i++];
      try{
	oc.setSkip(arg.convertToLong());
      }catch(Except &ex){
 	die(Sprintf("option -skip:%s\n",ex.msg())());
      }
    }else if(arg == "-gamma"){
      if(i>=argc || argv[i][0]=='-') {
 	usage();
 	die("option -gamma needs at least one argument\n");
      }
      long n=0;
      while(i<argc && argv[i][0]!='-' && n<3){
	arg=argv[i++];
	try{
	  oc.setGamma(n++,arg.convertToFortranReal());
	}catch(Except &ex){
	  die(Sprintf("option -gamma:%s\n",ex.msg())());
	}
      }
    }else if(arg == "-omega"){
      if(i>=argc || argv[i][0]=='-') {
 	usage();
 	die("option -omega needs at least one argument\n");
      }
      long n=0;
      long spin=-1;
      while(i<argc && argv[i][0]!='-' ){
	if(n%3==0){ spin++;n=0; }
	if(spin>2) break;
	arg=argv[i++];
	try{
	  oc.setOmega(n++,spin,arg.convertToFortranReal());
	}catch(Except &ex){
	  die(Sprintf("option -omega:%s\n",ex.msg())());
	}
      }
    }else if(arg == "-h"){
      usage();
      help();
      die("\nabort\n");
    }else if(arg == "-nointer"){
      oc.setNoInter(true);
    }else if(arg == "-nointra"){
      oc.setNoIntra(true);
    }else if(arg == "-ni"){
      oc.setNotInteractive(true);
    }else{
      usage();
      die(Sprintf("unkown option \"%s\"\nabort\n",arg())());
    }
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
int main(int argc,const char **argv)
{
  OpticsControl oc;

  vlevel=Vlevel::Info;
  
  if(vlevel>=Vlevel::All){
    printf("START\n");
    printf("verbosity level: %s\n",vlevel.name());
  }

    
  EXECNAME=argv[0];
  EXECNAMESHORT=argv[0];
  long ind;
  EXECNAMESHORT.chop("/");
  if((ind=EXECNAMESHORT.find("/",-1))>=0){
    EXECNAMESHORT=EXECNAMESHORT.subString(ind+1);
  }
  
  processCommandLineArgs(argc,argv,oc);


  try{
    oc.readInFile();
    
    oc.askForIntraData();

    oc.calcOpticsFunctions();

    oc.writeOpticalFunctions();

  }catch(Except &ex){    
    die(ex.msg());
  };

  
  if(vlevel>=Vlevel::All) 
    printf("END\n");

  return 0;
}


/*EOF*/ 
