/*************************************************************************
 * @(#)ReadOldGraph.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   07 Oct 2011
 *************************************************************************
 */

#include "Except.hh"
#include "Parser.hh"
#include "FString.hh"
#include "Attributes.hh"
#include "GraphObject.hh"
#include "FGObject.hh"
#include "Function.hh"
#include "Functions.hh"
#include "FunctionGroup.hh"
#include "FunctionGroups.hh"
#include "Graph.hh"
#include "Graphs.hh"
#include "Page.hh"
#include "Position.hh"
#include "LegendBox.hh"
#include "TextBox.hh"
#include "TicMarks.hh"
#include "World.hh"
#include "ViewBox.hh"
#include "MyMath.hh"
#include "ReadOldGraph.hh"
#include "MyDebug.hh"

inline QRgb mToRgb(const double *c){
  return
    (MyMath::max<long>(0,MyMath::min<long>(255,ulRound(255.*c[0])))<<16) |
    (MyMath::max<long>(0,MyMath::min<long>(255,ulRound(255.*c[1])))<<8) |
    (MyMath::max<long>(0,MyMath::min<long>(255,ulRound(255.*c[2]))));
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */

static void readLineStyle(const Dobj &d,LineStyle *ls)
{
  double color[3];
  d("color").getArray(3,color);
  ls->setColor(mToRgb(color));
  Xfp::LineStyles style=Xfp::getLineStyleFromString(d("style")()());;

  ls->setStyle(style);
  ls->setWidth(d("width").valueL());
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readFillStyle(const Dobj &d,FillStyle *fs)
{
  fs->setFill(d("fill").valueL());
  double color[3];
  d("color").getArray(3,color);
  fs->setColor(mToRgb(color));
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readBox(const Dobj &d,
		    bool &active,
		    FillStyle *fs,
		    LineStyle *ls,
		    double *border=0
		    )
{
  active=d("active").valueL();
  readFillStyle(d("fillstyle"),fs);
  readLineStyle(d("rimstyle"),ls);
  if(border)
    *border=d("border").valueD();

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readTextStyle(const Dobj &d,FontStyle *fs)
{
  
  fs->setSize(d("size").valueD());
  fs->setSubScriptScale(d("subscript_scale").valueD());
  double color[3];
  d("color").getArray(3,color);
  fs->setColor(mToRgb(color));
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readPlacement(const Dobj &d,Position *ps)
{
  double c[2];
  d("pos").getArray(2,c);
  ps->setPos(Point(c[0],c[1]));
  d("origin").getArray(2,c);
  ps->setOrigin(Point(c[0],c[1]));
  ps->setAngle(d("angle").valueD());
  ps->setOffset(d("offset").valueD());
  ps->setOffSetTypeFromString(d("offsettype")()());
  ps->setRestriction(d("restricted").valueL());
  Dobj da=d("addoffset");
  ps->clearAddTypes();
  for(int i=0;i<da.size();i++){
    String s=da[i]();
    for(int j=1;j<Position::numOfOffsetDescriptors();j++){
      if(s.subString(1)==Position::offsetIoOptionDescription(j) && s[0]=='+'){
	ps->appendAddType((Position::OffsetAddTypes)j);
      }
    }
  }

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readTextBox(const Dobj &d,TextBox *tb)
{
  bool active;
  double border;
  tb->setActive(d("active").valueL());
  tb->setText(d("text")());
  readBox(d("box"),active,tb->fill(),tb->rim(),&border);
  tb->setBoxIsActive(active);
  tb->setBorder(border);

  readTextStyle(d("textstyle"),tb->textStyle());
  
  readPlacement(d("placement"),tb->pos());
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readWorld(const Dobj &d,World *wd)
{
  double c[2];
  d("interval_x").getArray(2,c);
  wd->setW01(Xfp::Graph_X,c[0],c[1]);
  d("interval_y").getArray(2,c);
  wd->setW01(Xfp::Graph_Y,c[0],c[1]);
  d("autoscale_offset").getArray(2,c);
  wd->setAutoScaleOffset(Xfp::Graph_X,c[0]);
  wd->setAutoScaleOffset(Xfp::Graph_Y,c[1]);


}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readView(const Dobj &d,ViewBox *vb)
{
  double c[2];
  d("pos").getArray(2,c);
  vb->setPos(Point(c[0],c[1]));
  d("size").getArray(2,c);
  vb->setSize(Size(c[0],c[1]));
  bool active;
  readBox(d("box"),active,vb->fill(),vb->rim());
  vb->setDrawFrame(active);

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readLegendBox(const Dobj &d,LegendBox *lb)
{
  lb->setActive(d("active").valueL());
  readTextStyle(d("textstyle"),lb->textStyle());
  lb->setLineSpacing(d("linespacing").valueD());
  lb->setSymbolSpacing(d("symbols.marker.spacing").valueD());
  lb->setSymbolWidth(d("symbols.marker.width").valueD());

  bool active;
  double border;
  readBox(d("box"),active,lb->fill(),lb->rim(),&border);
  lb->setDrawFrame(active);
  lb->setBorder(border);

  Dobj dd=d("placement");
  double c[2];
  dd("pos").getArray(2,c);
  lb->setPos(Point(c[0],c[1]));
  dd("origin").getArray(2,c);
  lb->setOrigin(Point(c[0],c[1]));

  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readSymbol(const Dobj &d,SymbolStyle *sy)
{
  sy->setType(Xfp::getSymbolTypeFromString(d("type")()()));
  sy->setSize(d("size").valueD());
  sy->setExtraLineColor(d("extralinecolor").valueL());
  readLineStyle(d("linestyle"),&sy->lineStyle());
  sy->setExtraFillColor(d("extrafillcolor").valueL());
  readFillStyle(d("fillstyle"),&sy->fillStyle());

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readAttribs(const Dobj &d,FunctionAttribs *at)
{
  readLineStyle(d("linestyle"),&at->lineStyle());
  readSymbol(d("symbol"),&at->symbolStyle());
    
  at->setLegend(d("legend.text")(),d("legend.useit").valueL());
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readTics(const Dobj &d,Tic *tic)
{
  tic->setType((Xfp::XYAxisType)d("xy").valueL());
  tic->setKind((Xfp::TicKind)d("mjmn").valueL());
  tic->setPos(d("pos").valueD());
  tic->setTicSide(Xfp::getTicSiteFromString(d("tic_site")()()));
  tic->setTicLabelSide(Xfp::getTicSiteFromString(d("tic_site")()()));
  tic->setLabel(d("label")()());
  tic->setTicLength(d("tic_length").valueD());
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readUserTics(const Dobj &d,Tics *tics)
{
  tics->setDrawTics(d("drawusertics").valueL());

  Dobj da=d("tics");
      
  tics->resize(da.size(),true);

      
  for(long i=0;i<da.size();i++){
    readTics(da[i],tics->tic(i));
  }

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readRegularTicsXY(const Dobj &d,TicMarks *tm,Xfp::XYAxisType xy)
{
  
  tm->setAutoTics(xy,d("autotics").valueL());
  tm->setMajorTics(xy,d("major_spacing").valueD());
  tm->setTicLength(xy,Xfp::TKIMajor,d("major_length").valueD());
  bool drawtics=d("draw_major").valueL();

  tm->setDecimals(xy,d("decimals").valueL());
  tm->setMinorTics(xy,d("minor_subdivision").valueL());
  tm->setTicLength(xy,Xfp::TKIMinor,d("minor_length").valueD());
  tm->tics()->setDrawTics(xy,Xfp::TKIMinor,d("draw_minor").valueL());
  tm->setLabelOffset(xy,d("label_offset").valueD());

  tm->setTicSide(xy,Xfp::getTicSiteFromString(d("tic_site")()()));
  tm->setTicLabelSide(xy,Xfp::getTicSiteFromString(d("tic_site")()()));
  if(!drawtics){
    tm->setTicSide(xy,Xfp::getTicSiteFromString("None"));
    tm->setTicLabelSide(xy,Xfp::getTicSiteFromString("None"));
  }

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readRegularTics(const Dobj &d,TicMarks *tm)
{
  readRegularTicsXY(d("X"),tm,Xfp::Graph_X); 
  readRegularTicsXY(d("Y"),tm,Xfp::Graph_Y); 

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readTicMarks(const Dobj &d,TicMarks *tm)
{

  
  readRegularTics(d("regular_tics"),tm); 
  readUserTics(d("user_tics"),tm->userTics()); 

  readTextStyle(d("textstyle"),tm->textStyle(Xfp::Graph_X));; 
  readTextStyle(d("textstyle"),tm->textStyle(Xfp::Graph_Y));; 

  Dobj da=d("linestyle");
      
  for(int i=0;i<MyMath::min<long>(da.size(),2);i++){
    readLineStyle(da[i],tm->lineStyle((Xfp::XYAxisType)i,Xfp::TKIMajor));
    readLineStyle(da[i],tm->lineStyle((Xfp::XYAxisType)i,Xfp::TKIMinor));
  }

}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readFunc(const Dobj &d,Function *func)
{
  func->setActive(d("active").valueL());
  readAttribs(d("attributes"),func->attribs());
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readGroup(const Dobj &d,FunctionGroup *group)
{
  group->setActive(d("active").valueL());
  group->setUseAttribs(d("use_attributes").valueL());

  
  readAttribs(d("attributes"),group->attribs());

  readUserTics(d("user_tics"),group->userTics()); 


  Dobj da=d("funcs");

  group->funcs()->reset();
  group->funcs()->setSize(da.size());
  for(long i=0;i<da.size();i++){
    readFunc(da[i],group->funcs()->at(i));
  }

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static void readXNYData(const Dobj &d,Graph *graph)
{

  long ig=d("group").valueL();
  long jf=d("func").valueL();
  
  
  graph->groups()->insert(ig);
  FunctionGroup *group=graph->groups()->at(ig);
  group->funcs()->assureSize(jf+1);
  Function *func=group->funcs()->at(jf);

  func->adjustComp(0,0,true);

  func->setDataType(Function::dataTypeFromString(d("type")()()));
  


  

  if(func->dataType()==Function::GRIDX_Y_NZ){
    {
      Dobj dar=d("x");
      func->adjustXYLength(Xfp::Graph_X,dar.size()); 
      for(long ix=0;ix<dar.size(1);ix++){
	func->setDataXY(Xfp::Graph_X,ix,dar[ix].valueD());
      }
    }    
    
    {
      Dobj dar=d("y");
      func->adjustXYLength(Xfp::Graph_Y,dar.size()); 
      for(long iy=0;iy<dar.size(1);iy++){
	func->setDataXY(Xfp::Graph_Y,iy,dar[iy].valueD());
      }     
    }    
    func->adjustComp(1,func->xdata().size()*func->ydata().size(),true);
    
  }
  

  Dobj dac=d("components");

  func->adjustComp(dac.size(),false);
    

  for(long IC=0;IC<dac.size();IC++){
    Dobj dar=dac[IC]("array");
    func->adjustLength(dar.size(1));
    for(long I=0;I<dar.size(1);I++){
      func->setData(IC,I,dar[I].valueD());
    }     
  }


}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

void readOldGraph(Page *page,const String &filename)
{
  //myDebug() << filename();
  page->isdirty1=false;

  Parser par(1);
  par.parseFile(filename());

  String s="";
  try{
    s=par()("filetype.application")();
  }catch(Except &ex){}
  if(s!="xfpl"){
    throw Except("not an old graph file");
  }
  
  Dobj root=par();
  myDebug() << root("filetype.description")();
  if(root("filetype.description")()!="FPLO-X-Graphics") 
    throw Except("not and old graph file");

  Dobj dgraph=root("graph");
  Graph *graph=page->graphs()->at(0);
  
  readTextBox(dgraph("title"),graph->textBoxes()->at(TextBox::TBTTitle));
  readTextBox(dgraph("subtitle"),graph->textBoxes()->at(TextBox::TBTSubTitle));
  readTextBox(dgraph("xaxislabel"),graph->textBoxes()->at(TextBox::TBTXAxisLabel));
  readTextBox(dgraph("yaxislabel"),graph->textBoxes()->at(TextBox::TBTYAxisLabel));
  readTextBox(dgraph("oppxaxislabel"),graph->textBoxes()->at(TextBox::TBTOppositeXAxisLabel));
  readTextBox(dgraph("oppyaxislabel"),graph->textBoxes()->at(TextBox::TBTOppositeYAxisLabel));
  
  
  {
    Dobj da=dgraph("textboxes");
    graph->textBoxes()->adjustSizeToIndex(da.size()+TextBox::TBTGeneral,true);
    for(int i=0;i<da.size();i++){
      readTextBox(da[i],graph->textBoxes()->textBoxByType(TextBox::TBTGeneral,i));
    }
  }
  readWorld(dgraph("world"),graph->world());
  readView(dgraph("view"),graph->viewBox());
  readLegendBox(dgraph("legendbox"),graph->legendBox());
  readTicMarks(dgraph("ticmarks"),graph->ticMarks());
  
  {
    Dobj da=dgraph("groups");
    for(long i=0;i<da.size();i++){
      graph->groups()->insert(i);
      readGroup(da[i],graph->groups()->at(i));
    }
    
  }

  {
    Dobj da=root("xnydata");
    for(long i=0;i<da.size();i++){
      readXNYData(da[i],graph);
    }
  }  

}



/*EOF*/ 
