/*************************************************************************
 * @(#)SetTreeModel.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   17 May 2011
 *************************************************************************
 */

#include <QIcon>
#include "Attributes.hh"
#include "GraphObject.hh"
#include "FGObject.hh"
#include "Function.hh"
#include "Functions.hh"
#include "FunctionGroup.hh"
#include "FunctionGroups.hh"
#include "Graph.hh"
#include "SetTreeModel.hh"
#include "MyDebug.hh"


/* ***********************************************************************
 *
 * ***********************************************************************
 */
SetTreeModel::SetTreeModel(QObject *parent)
  :QAbstractItemModel(parent),pgraph(0)
{
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
SetTreeModel::~SetTreeModel()
{
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
FGObject *SetTreeModel::getFGObject(const QModelIndex &index)const
{
  if(!index.isValid())return 0;
  return static_cast<FGObject*>(index.internalPointer());
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void SetTreeModel::resetData(Graph *gr)
{
  pgraph=gr;
  reset();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void SetTreeModel::packIDs()
{
  pgraph->groups()->packIDs();
  FunctionGroup *lastgroup=pgraph->groups()->last();
  emit dataChanged(index(0,0,QModelIndex()),
		   index(lastgroup->funcs()->size()-1,0
			 ,index(pgraph->groups()->size()-1,0,QModelIndex())));
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
int SetTreeModel::rowCount(const QModelIndex &parent)const
{
  //  myDebug() << "SetTreeModel::rowCount";
  if(!pgraph) return 0;
  if(!parent.isValid()){
    // root=groups
    return pgraph->groups()->size();
  }else{
    FGObject *fgo=getFGObject(parent);
    if(fgo && fgo->type()==FGObject::FGOTGroup){
      //myDebug()<<((FunctionGroup*)fgo)->funcs()->size();
      return ((FunctionGroup*)fgo)->funcs()->size();
    }
  }

  return 0;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
int SetTreeModel::columnCount(const QModelIndex &parent)const
{
  Q_UNUSED(parent);
  return 1;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
QModelIndex SetTreeModel::index(int row, int column,
				const QModelIndex &parent)const
{
  if (parent.isValid() && parent.column() != 0) return QModelIndex();
  if (column<0 || column>= columnCount()) return QModelIndex();
  if (row<0 || row>= rowCount(parent)) return QModelIndex();

  if(!parent.isValid()){
    // root items = groups
    FunctionGroup *fg=pgraph->groups()->at(row);
    if(fg){
      return createIndex(row, column, fg);
    }else{
      return QModelIndex();
    }
  }else{
    FGObject *fgo=getFGObject(parent);
    //myDebug() <<"fgo"<<fgo;
    if(fgo && fgo->type()==FGObject::FGOTGroup){
      FunctionGroup *fg=(FunctionGroup*)fgo;
      Function *fu=fg->funcs()->at(row);
      return createIndex(row, column, fu);
    }
  }
  return QModelIndex();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
QModelIndex SetTreeModel::index(FGObject *fgo,const QModelIndex &parent)const
{
  if(!fgo) return QModelIndex();

  if(!parent.isValid()){
    for(int i=0;i<pgraph->groups()->size();i++){
      if(pgraph->groups()->at(i)==fgo){
	return createIndex(i,0,fgo);
      }
    }
  }else{
    FGObject *fgg=getFGObject(parent);
    FunctionGroup *fg=(FunctionGroup*)fgg;
    for(int i=0;i<fg->funcs()->size();i++){
      if(fg->funcs()->at(i)==fgo){
	return createIndex(i,0,fgo);
      }
    }
  }
  return QModelIndex();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
QModelIndex SetTreeModel::parent(const QModelIndex &index)const
{
  if (!index.isValid()) return QModelIndex();

  
  FGObject *fgo=getFGObject(index);
  if(fgo && fgo->type()==FGObject::FGOTGroup){
    return QModelIndex();
  }else if(fgo && fgo->type()==FGObject::FGOTFunction){
    FunctionGroup *fg=((Function*)fgo)->group();
    long ind=pgraph->groups()->indexOf(fg);
    return createIndex(ind,0,fg);
  }

  return QModelIndex();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
QVariant SetTreeModel::data(const QModelIndex &index, int role) const
{
  if(!pgraph) return QVariant();
  if (!index.isValid() || index.column() != 0) return QVariant();

  FGObject *fgo=getFGObject(index);
  if(!fgo)return QVariant();
  if (role == Qt::DisplayRole) {
    return fgo->guiName(true)();
  }else if (role == Qt::DecorationRole) {
    if(fgo->type()==FGObject::FGOTGroup){
      if(fgo->active())
	return QIcon(":/Icons/funcactive");
      else
	return QIcon(":/Icons/funcinactive");
    }else if(fgo->type()==FGObject::FGOTFunction){
      if(fgo->active())
	return QIcon(":/Icons/funcactive");
      else
	return QIcon(":/Icons/funcinactive");
    }else
      return QVariant();
    
  }
  return QVariant();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

QModelIndex SetTreeModel::insertRow(QModelIndex &parent,int position,FGObject *fgo)
{
  // this function must be called carfully, we do not check
  // if the inserted fgo is of correct type for the given parent
  beginInsertRows(parent,position,position);
  if(!parent.isValid()){
    pgraph->groups()->insert(position,(FunctionGroup*)fgo);
  }else{
    FGObject *fgg=getFGObject(parent);
    if(fgg && fgg->type()==FGObject::FGOTGroup){
      FunctionGroup*fg=((FunctionGroup*)fgg);
      Function *fu=(Function*)fgo;
      fg->funcs()->insert(position,fu);
    }
  }
  endInsertRows();
  return this->index(position,0,parent);

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
FGObject *SetTreeModel::removeRow(QModelIndex &parent,int position)
{
  FGObject *fgo=0;
  beginRemoveRows(parent,position,position);

  if(!parent.isValid()){
    fgo=pgraph->groups()->takeAt(position); 
  }else{
    FGObject *fgg=getFGObject(parent);
    if(fgg && fgg->type()==FGObject::FGOTGroup){
      fgo=((FunctionGroup*)fgg)->funcs()->takeAt(position);
    }
  }
  endRemoveRows();
  return fgo;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
QModelIndex SetTreeModel::moveRow(QModelIndex &parent,int row,int by)
{
  //myDebug()<<"moveRow"<<row<<by;
  
  if(row==0 && by<0){

    QModelIndex sibling=this->sibling(parent.row()+by,0,parent);
    if(sibling.isValid()){
      //todo: rename the set IDs appropriatly
      FGObject *fgo=removeRow(parent,row);  
      int row1=rowCount(sibling);
      return insertRow(sibling,row1,fgo);
    }
    return QModelIndex();
  }
  //  myDebug()<<rowCount(parent)-1;
  if(row==rowCount(parent)-1 && by>0){
    QModelIndex sibling=this->sibling(parent.row()+by,0,parent);
    if(sibling.isValid()){
      //todo: rename the set IDs appropriatly
      FGObject *fgo=removeRow(parent,row);  
      return insertRow(sibling,0,fgo);
    }
    return QModelIndex();
    
  }

  if(!index(row+by,0,parent).isValid()) return QModelIndex();
  //myDebug() << "here valid";


  FGObject *fgo=removeRow(parent,row);
  return insertRow(parent,MyMath::max<long>(0,row+by),fgo);
  
 
}


/*EOF*/ 
