/*************************************************************************
 * @(#)BSpline.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   15 Dec 2011
 *************************************************************************
 */

#include "BSpline.hh"
#include "Except.hh"
#include "Assorted.hh"
#include "MyMath.hh"
#include "MyDebug.hh"


/* ***********************************************************************
 *
 * ***********************************************************************
 */
BSpline::BSpline(long degree,const ValueArray<double> &x)
  :pdegree(degree),porder(degree+1),px(x),pilo(0)
{
  setKnots();
  bsplint();
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
BSpline::BSpline(long degree,const double *x,long Nx)
  :pdegree(degree),porder(degree+1),pilo(0)
{
  px.resize(Nx,true);
  px=x;
  setKnots();
  bsplint();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BSpline::writeBasisFunctions(const String &fname,
				  const ValueArray<double> &xout)
{
  FILE *FH;
  
  FH=openforwrite(fname);

  long N=px.size();
  long nout=xout.size();

  for(int j=0;j<N;j++){
    pbcoeff=0.0;
    pbcoeff[j]=1.0;
    for(int i=0;i<nout;i++){
      double a=xout.at(i);
      
      double sf=bvalue(pbcoeff,a,0);
      fprintf(FH,"%30.15e %30.15e\n",a,sf);
    }
    fprintf(FH,"\n");
  }
  for(int j=0;j<N+porder;j++){
    fprintf(FH,"%10.5f %10.5f\n",pknots[j],0.0);
    fprintf(FH,"%10.5f %10.5f\n\n",pknots[j],1.0);
  }
  for(int j=0;j<N;j++){
    fprintf(FH,"%10.5f %10.5f\n",px[j],0.0);
    fprintf(FH,"%10.5f %10.5f\n\n",px[j],0.5);
  }
  fclose(FH);

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BSpline::splint(const ValueArray<double> &y)
{
  long N=px.size();
  long p=pdegree;
  pbcoeff=y;
  banslv(pq,N,p,p,pbcoeff);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BSpline::splint(const double *y)
{
  long N=px.size();
  long p=pdegree;
  pbcoeff=y;
  banslv(pq,N,p,p,pbcoeff);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BSpline::spline(const ValueArray<double> &xout,
		     ValueArray<double> &yout,long jderiv)
{
  long nout=xout.size();
  for(int i=0;i<nout;i++){
    double a=xout.at(i);
    yout[i]=bvalue(pbcoeff,a,jderiv);
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BSpline::spline(const ValueArray<double> &xout,
		     double *yout,long jderiv)
{
  long nout=xout.size();
  for(int i=0;i<nout;i++){
    yout[i]=bvalue(pbcoeff,xout.at(i),jderiv);
  }
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BSpline::setKnots()
// px: mesh points
{
  long k=porder;
  long N=px.size();
  
  pknots.resize(N+k,true);
  pq.resize((2*k-1)*N);
  pbcoeff.resize(N);

  if(N<k){
    throw Except(Sprintf("BSpline::setKnots: number of mesh points (%ld)\n"
			 " must be larger than the order (%ld)!\n",N,k));
  }
  
  //myDebug()<<"k="<<k<<k%2;
  if((k%2)==0){
    for(long i=0;i<k;i++){
      pknots[i]=px.at(0);
    }
    for(long i=k;i<N;i++){
      pknots[i]=px.at(i-k/2);
    }
    for(long i=N;i<N+k;i++){
      pknots[i]=px.at(N-1);
    }
  }else{
    for(long i=0;i<k;i++){
      pknots[i]=px.at(0);
    }
    for(long i=0;i<N-k;i++){
      pknots[k+i]=0.5*(px.at((k-1)/2+i)+px.at((k+1)/2+i));
    }
    for(long i=N;i<N+k;i++){
      pknots[i]=px.at(N-1);
    }
  }
  //printf("knots=%s\n",printS(pknots,"%10.5f")());

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

void BSpline::bsplint()
{
  const char *ermsg="BSpline::bsplint: The linear system is not invertible!?";
  
  //  Subroutine bsplint(tau,t,n,k,q,bcoef,gtau)
  //
  //
  // SPLINT produces the B-spline coefficients BCOEF of an interpolating 
  // spline.
  //
  //
  // Discussion:
  //
  //   The spline is of order K with knots pknots(1:N+K), and takes on the 
  //   value GTAU(I) at TAU(I), for I = 1 to N.
  //
  // Method:
  //
  //   The I-th equation of the linear system  
  //     A * BCOEF = B 
  //   for the b-coefficients of the interpolant enforces interpolation
  //   at TAU(1:N).
  //
  //   Hence,  b(i)=gtau(i), all i, and  a  is a band matrix with  2k-1
  //   bands (if it is invertible).
  //
  //   The matrix A is generated row by row and stored, diagonal by di-
  //   agonal, in the rows of the array  q , with the main diagonal go-
  //   ing into row K.  see comments in the program below.
  //
  //   The banded system is then solved by a call to  banfac (which con-
  //   structs the triangular factorization for  a  and stores it again in
  //   q ), followed by a call to  banslv (which then obtains the solution
  //   bcoef  by substitution).
  //
  //   BANFAC does no pivoting, since the total positivity of the matrix
  //   A makes this unnecessary.
  //
  // Reference:
  //
  //   Carl DeBoor,
  //   A Practical Guide to Splines,
  //   Springer Verlag.
  //
  //
  long k=porder;
  long n=px.size();
  long kpkm2=2*(k-1);
  long left=k-1;

  pq.resize((2*k-1)*n,true);
  for(long i=0;i<(2*k-1)*n;i++){
    pq[i]=0.0;
  }
  //
  // loop over i to construct the  n  interpolation equations
  //

  ValueArray<double> biatx(k,0.);

  for(long i=0;i<n;i++){

    double taui=px[i];
    long ilp1mx=MyMath::min<long>(i+k,n);
    //
    // find  left  in the closed interval (i,i+k-1) such that
    //   t(left) <= tau(i)  < t(left+1)
    // matrix is singular if this is not possible
    //
    left=MyMath::max<long>(left,i);
       
    if(taui<pknots[left]) throw Except(ermsg);
    

  l20: 
       
    if(taui<pknots[left+1]){
      goto l30;
    }
       
    left++;
    if(left<ilp1mx){
      goto l20;
    }
       
    left--;
    if(pknots[left+1]<taui) throw Except(ermsg);
    //
    // The i-th equation enforces interpolation at taui, hence
    // a(i,j)=b(j,k,t)(taui), all j. only the  k  entries with  j =
    // left-k+1,...,left actually might be nonzero. these  k  numbers
    // are returned, in  bcoef (used for temp.storage here), by the
    // following
    //
  l30:
       
    bsplvb(k,1,taui,left,biatx);
    //myDebug()<<"left="<<left<<"biatx="<<biatx;
    //
    // We therefore want  bcoef(j)=b(left-k+j)(taui) to go into
    // a(i,left-k+j), i.e., into  q(i-(left+j)+2*k,(left+j)-k) since
    // a(i+j,j)  is to go into  q(i+k,j), all i,j,  if we consider  q
    // as a two-dim. array , with  2*k-1  rows (see comments in
    // banfac). in the present program, we treat  q  as an equivalent
    // one-dimensional array (because of fortran restrictions on
    // dimension statements) . we therefore want  bcoef(j) to go into
    // entry
    //   i+1 -(left+1+j+1)+2*k + ((left+1+j+1)-k-1)*(2*k-1)-1
    //  = i-left+(left+1-k)*(2*k-1) + (2*k-2)*(j+1)
    // of  q .
    //
    long jj=i-left+(left+1-k)*(k+k-1);
    for(long j=0;j<k;j++){
      jj=jj+kpkm2;
      pq[jj]=biatx[j];
    }
       
  }
  //
  // Obtain factorization of A, stored again in Q.
  //
  if(!banfac(pq,n,k-1,k-1)) throw Except(ermsg);
  
  //myDebug()<<pq;
  return;
    
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BSpline::bsplvb(long jhigh,long index,double at, long left,
		     ValueArray<double> &biatx)
{
  //
  //
  //! BSPLVB evaluates B-splines at a point X with a given knot sequence.
  //
  //
  //  Discusion:
  //
  //    BSPLVB evaluates all possibly nonzero B-splines at X of order
  //
  //      JOUT = MAX ( JHIGH, (J+1)*(INDEX-1) ) 
  //  
  //    with knot sequence T.
  // 
  //    The recurrence relation
  // 
  //                     X - T(I)               T(I+J+1) - X
  //    B(I,J+1)(X) = ----------- * B(I,J)(X) + --------------- * B(I+1,J)(X)
  //                  T(I+J)-T(I)               T(I+J+1)-T(I+1)
  // 
  //    is used to generate B(LEFT-J:LEFT,J+1)(X) from B(LEFT-J+1:LEFT,J)(X)
  //    storing the new values in BIATX over the old. 
  //
  //    The facts that 
  //
  //      B(I,1)(X) = 1  if  T(I) <= X < T(I+1)
  //
  //    and that 
  //
  //      B(I,J)(X) = 0  unless  T(I) <= X < T(I+J)
  //
  //    are used. 
  //
  //    The particular organization of the calculations follows 
  //    algorithm 8 in chapter X of the text.
  //
  //  Reference:
  //
  //    Carl DeBoor,
  //    A Practical Guide to Splines,
  //    Springer Verlag.
  //
  //  Parameters:
  //
  //    Input, real pknots(LEFT+JOUT), the knot sequence.  
  //      pknots is assumed to be nondecreasing, and also, 
  //      pknots(LEFT) must be strictly less than pknots(LEFT+1).
  // 
  //    Input, integer JHIGH, INDEX, determine the order 
  //    JOUT = MAX(JHIGH,(J+1)*(INDEX-1))  
  //    of the B-splines whose values at X are to be returned.  
  //    INDEX is used to avoid recalculations when several 
  //    columns of the triangular array of B-spline values are
  //    needed, for example, in BVALUE or in BSPLVD.
  // 
  //    If INDEX = 1, the calculation starts from scratch and the entire 
  //    triangular array of B-spline values of orders
  //    1, 2, ...,JHIGH is generated order by order, i.e., 
  //    column by column.
  //   
  //    If INDEX = 2, only the B-spline values of order J+1, J+2, ..., JOUT  
  //    are generated, the assumption being that BIATX, J, 
  //    DELTAL, DELTAR are, on entry, as they were on exit 
  //    at the previous call.  In particular, if JHIGH = 0, 
  //    then JOUT = J+1, i.e., just the next column of B-spline 
  //    values is generated.
  // 
  //    WARNING: the restriction  JOUT <= JMAX (= 20) is
  //    imposed arbitrarily by the dimension statement for DELTAL
  //    and DELTAR, but is nowhere checked for.
  // 
  //    Input, real X, the point at which the B-splines are to be evaluated.
  // 
  //    Input, integer LEFT, an integer chosen so that 
  //    T(LEFT) <= X <= T(LEFT+1).
  // 
  //    Output, real BIATX(JOUT), with BIATX(I) containing the
  //    value at X of the polynomial of order JOUT which agrees 
  //    with the B-spline B(LEFT-JOUT+I,JOUT,T) on the interval 
  //    (T(LEFT),T(LEFT+1)).
  //

  if(index==1){
    pj=0;
    biatx[0]=1.0;
    pdeltal.resize(jhigh,true);
    pdeltar.resize(jhigh,true);
    //myDebug()<<jhigh<<pj+1;
    if(jhigh<=pj+1) return;
  }else{
    pdeltal.resize(jhigh);
    pdeltar.resize(jhigh);
  }
  
  if(pknots[left+1]<=pknots[left]){
    
    throw Except(Sprintf("BSPLVB - Fatal error!\n"
			 "  It is required that T(LEFT) < T(LEFT+1).\n"
			 "  But LEFT = %ld\n"
			 "  T(LEFT)  = %g\n"
			 "  T(LEFT+1)= %g\n",
			 left,pknots[left],pknots[left+1]));
    
  }
  
  while(pj<jhigh-1){
       

    pdeltar[pj]=pknots[left+1+pj]-at;
    pdeltal[pj]=at-pknots[left-pj];
    
    //myDebug()<<"j="<<pj<<pdeltar[pj]<<pdeltal[pj];
       
    double saved=0.0;
    for(long i=0;i<=pj;i++){
      double term=biatx[i]/(pdeltar[i]+pdeltal[pj-i]);
      biatx[i]=saved+pdeltar[i]*term;
      saved=pdeltal[pj-i]*term;
    }
    
    biatx[++pj]=saved;
  }
    
  return;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

bool BSpline::banfac(ValueArray<double> &w,long N,
		     long  nbandl,long  nbandu)
{
  //NROW: the row dimension of the work array W.
  long nrow=nbandu+nbandl+1;
  //
  //  BANFAC factors a banded matrix without pivoting.
  //
  //
  //  Discussion:
  //
  //    BANFAC returns in W the LU-factorization, without pivoting, of 
  //    the banded matrix A of order N with (NBANDL+1+NBANDU) bands 
  //    or diagonals in the work array W.
  // 
  //    Gauss elimination without pivoting is used.  The routine is 
  //    intended for use with matrices A which do not require row 
  //    interchanges during factorization, especially for the totally 
  //    positive matrices which occur in spline calculations.
  //
  //  Reference:
  //
  //    Carl DeBoor,
  //    A Practical Guide to Splines,
  //    Springer Verlag.
  //
  //  Parameters:
  // 
  //    Input/output, real W(NROW*N).
  //    On input, W contains the "interesting" part of a banded 
  //    matrix A, with the diagonals or bands of A stored in the
  //    rows of W, while columns of A correspond to columns of W. 
  //
  //    This is the storage mode used in LINPACK and results in efficient 
  //    innermost loops.
  // 
  //    Explicitly, A has 
  // 
  //      NBANDL bands below the diagonal
  //      1     main diagonal
  //      NBANDU bands above the diagonal
  //
  //    and thus, with MIDDLE=NBANDU+1,
  //    A(I+J,J) is in W(I+MIDDLE,J) for I=-NBANDU,...,NBANDL, J=1,...,N.
  //
  //    For example, the interesting entries of a banded matrix
  //    matrix of order 9, with NBANDL=1, NBANDU=2:
  //
  //      11 12 13  0  0  0  0  0  0
  //      21 22 23 24  0  0  0  0  0
  //       0 32 33 34 35  0  0  0  0
  //       0  0 43 44 45 46  0  0  0
  //       0  0  0 54 55 56 57  0  0
  //       0  0  0  0 65 66 67 68  0
  //       0  0  0  0  0 76 77 78 79
  //       0  0  0  0  0  0 87 88 89
  //       0  0  0  0  0  0  0 98 99
  //
  //    would appear in the first 1+1+2=4 rows of W as follows:
  //
  //       0  0 13 24 35 46 57 68 79
  //       0 12 23 34 45 56 67 78 89
  //      11 22 33 44 55 66 77 88 99
  //      21 32 43 54 65 76 87 98  0
  // 
  //    All other entries of W not identified in this way with an
  //    entry of A are never referenced.
  // 
  //    On output, W contains the LU-factorization of A into a unit 
  //    lower triangular matrix L and an upper triangular matrix U 
  //    (both banded) and stored in customary fashion over the 
  //    corresponding entries of A.  
  //
  //    This makes it possible to solve any particular linear system A*X=B 
  //    for X by the call
  //
  //      call banslv(w,N,nbandl,nbandu,b)
  //
  //    with the solution X contained in B on return.
  // 
  //    If return=false, then one of N-1, NBANDL, NBANDU failed to be nonnegative, 
  //    or else one of the potential pivots was found to be zero 
  //    indicating that A does not have an LU-factorization.  This 
  //    implies that A is singular in case it is totally positive.
  // 
  // 
  //    Input, integer N, the number of rows in A.
  //
  //    Input, integer NBANDL, the number of bands of A below the main diagonal.
  // 
  //    Input, integer NBANDU, the number of bands of A above the main diagonal.
  // 
  //    return value:error flag.
  //    true, success.
  //    false, failure, the matrix was not factored.
  //
  long middle=nbandu;
  //
  //  W(MIDDLE,*) contains the main diagonal of A.
  //
  if(N<1){
    return false;
  }
    
  if(N==1){
    if(w[middle]==0.0 ) return false;
    return true;
  }
  //
  //  A is upper triangular.  Check that the diagonal is nonzero.
  //
  if(nbandl<=0){
    // code not tested
    for(long i=0;i<N-1;i++){
      if(w[middle+i*nrow]==0.0) return false;
    }
    
    if(w[middle+(N-1)*nrow]==0.0) return false;
    
    return true;
    //
    //  A is lower triangular.  Check that the diagonal is nonzero and
    //  divide each column by its diagonal.
    //
  }else if(nbandu<=0){
    // code not tested
       
    for(long i=0;i<N-1;i++){
          
      double pivot=w[middle+i*nrow];
          
      if(pivot==0.0) return false;
      
          
      for(long j=0;j<MyMath::min<long>(nbandl,N-1-i);j++){
	w[middle+j+1+i*nrow]/=pivot;
      }
    }     

    return true;
       
  }
  //
  //  A is not just a triangular matrix.  
  //  Construct the LU factorization.
  //
  for(long i=0;i<N-1;i++){
    //
    //  W(MIDDLE,I) is the pivot for the I-th step.
    //
    if(w[middle+i*nrow]==0.0){
      throw Except(Sprintf("BSpline::banfac: - Fatal error!"
			   "  Zero pivot encountered in column %ld", i));
    }
    //
    //  Divide each entry in column I below the diagonal by PIVOT.
    //
    for(long j=0;j<MyMath::min<long>(nbandl,N-1-i);j++){
      w[middle+j+1+i*nrow]/=w[middle+i*nrow];
    }
    //
    //  Subtract A(I,I+K)*(I-th column) from (I+K)-th column (below
    //  row I).
    //
    for(long k=0;k<MyMath::min<long>(nbandu,N-1-i);k++){
      double factor=w[middle-k-1+(i+k+1)*nrow];
      for(long j=0;j<MyMath::min<long>(nbandl,N-1-i);j++){
	w[middle-k+j+(i+k+1)*nrow]-=w[middle+j+1+i*nrow]*factor;
      }
    }
       
  }
  //
  //  Check the last diagonal entry.
  //
  if(w[middle+(N-1)*nrow]==0.0) return false;
  

  return true;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BSpline::banslv(ValueArray<double> &w,long N,
		     long nbandl,long nbandu,ValueArray<double> &b)
{
  //NROW: the row dimension of the work array W.
  long nrow=nbandu+nbandl+1;
  //
  //
  //! BANSLV solves a banded linear system X * X = B factored by BANFAC.
  //
  //
  //  Reference:
  //
  //    Carl DeBoor,
  //    A Practical Guide to Splines,
  //    Springer Verlag.
  //
  //  Parameters:
  //
  //    Input, real W(NROWW,NROW).  W contains the banded matrix,
  //    after it has been factored by BANFAC.
  //
  //    Input, integer NROWW, the row dimension of the work array W.
  //    NROWW must be at least NBANDL+1 + NBANDU.
  // 
  //    Input, integer NROW, the number of rows in A.
  //
  //    Input, integer NBANDL, the number of bands of A below the 
  //    main diagonal.
  // 
  //    Input, integer NBANDU, the number of bands of A above the 
  //    main diagonal.
  // 
  //    Input/output, real B(NROW).
  //    On input, B contains the right hand side of the system to be solved.
  //    On output, B contains the solution, X.

  long middle=nbandu;

  if(N==1){
    b[0]/=w[middle];
    return;
  }
  //
  //  Forward pass
  //
  //  For I=1,2,...,N-1, subtract RHS(I)*(I-th column of L) 
  //  from the right side, below the I-th row.
  //
  if(0<nbandl){
    for(long i=0;i<N-1;i++){
      long jmax=MyMath::min<long>(nbandl,N-i-1);
      for(long j=0;j<jmax;j++){
	b[i+j+1]-=b[i]*w[middle+j+i*nrow+1];
      }
    }
  }
  //
  //  Backward pass
  //
  //  For I=N, N-1,...,1, divide RHS(I) by 
  //  the I-th diagonal entry of U, then subtract 
  //  RHS(I)*(I-th column of U) from right side, above the I-th row.
  //
    for(long i=N-1;i>=1;i--){
      b[i]/=w[middle+i*nrow];
       
      for(long j=0;j<MyMath::min<long>(nbandu,i);j++){
	b[i-j-1]-=b[i]*w[middle-j+i*nrow-1];
      }
    }
    
    b[0]/=w[middle];

    return;
  }
/* ***********************************************************************
 *
 * ***********************************************************************
 */
double BSpline::bvalue(const ValueArray<double> &bcoef,double at,
		       long jderiv)
{
  long kk=porder;
  long p=pdegree;
  long N=px.size();
  //
  //
  // BVALUE evaluates a derivative of a spline from its B-spline representation.  
  //
  //
  //  Discussion:
  //
  //    The spline is taken to be continuous from the right.
  // 
  //    The nontrivial knot interval (T(I),T(I+1)) containing X is 
  //    located with the aid of INTERV.  The K B-spline coefficients 
  //    of F relevant for this interval are then obtained from BCOEF, 
  //    or are taken to be zero if not explicitly available, and are 
  //    then differenced JDERIV times to obtain the B-spline 
  //    coefficients of (D**JDERIV)F relevant for that interval.  
  //
  //    Precisely, with J = JDERIV, we have from X.(12) of the text that:
  // 
  //      (D**J)F = sum ( BCOEF(.,J)*B(.,K-J,T) )
  // 
  //    where
  //                      / BCOEF(.),                    ,  J == 0
  //                     /
  //       BCOEF(.,J) = / BCOEF(.,J-1) - BCOEF(.-1,J-1)
  //                   / -----------------------------,  0 < J
  //                  /    (T(.+K-J) - T(.))/(K-J)
  // 
  //    Then, we use repeatedly the fact that
  // 
  //      sum ( A(.)*B(.,M,T)(X) ) = sum ( A(.,X)*B(.,M-1,T)(X) )
  // 
  //    with
  //                   (X - T(.))*A(.) + (T(.+M-1) - X)*A(.-1)
  //      A(.,X) =   ---------------------------------------
  //                   (X - T(.))      + (T(.+M-1) - X)
  // 
  //    to write (D**J)F(X) eventually as a linear combination of 
  //    B-splines of order 1, and the coefficient for B(I,1,T)(X) 
  //    must then be the desired number (D**J)F(X).
  //    See x.(17)-(19) of text.
  //
  //  Reference:
  //
  //    Carl DeBoor,
  //    A Practical Guide to Splines,
  //    Springer Verlag.
  //
  //  Parameters:
  //
  //    Input, real T(N+K), the knot sequence.  T is assumed to be nondecreasing.
  //
  //    Input, real BCOEF(N), B-spline coefficient sequence.
  // 
  //    Input, integer N, the length of BCOEF.
  // 
  //    Input, integer K, the order of the spline.
  //
  //    Input, real X, the point at which to evaluate.
  // 
  //    Input, integer JDERIV, the order of the derivative to 
  //    be evaluated.  JDERIV is assumed to be zero or positive.
  // 
  //    Output, real BVALUE, the value of the (JDERIV)-th 
  //    derivative of the spline at X.

  ValueArray<double> dl(kk,0.);
  ValueArray<double> dr(kk,0.);
  ValueArray<double> aj(kk,0.);

  if(kk<=jderiv) return 0.;
    
  //
  //  Find I so that 1 <= i < n+k and t(i) < t(i+1) and t(i) <= x < t(i+1). 
  //
  //  If no such i can be found, X lies
  //  outside the support of  the spline F and  bvalue=0.
  //  (the asymmetry in this choice of  i  makes F rightcontinuous)
  //
  long iii;
  
  long ret=interv(at,iii);
  if(ret==-1){
    iii=kk-1;
  }else if(ret==1){
    iii=N-1;
  }else if(ret!=0){
    return 0.0;
  }
  
  //
  //  If K=1 (and jderiv = 0), bvalue = bcoef(i).
  //
  if(p<=0) return bcoef.at(iii);
  
  //
  //  Store the K b-spline coefficients relevant for the knot interval
  //  (t(i),t(i+1)) in aj(1),...,aj(k) and compute dl(j)=x-t(i+1-j),
  //  dr(j)=t(i+j)-x, j=1,...,k-1 . set any of the aj not obtainable
  //  from input to zero. set any t.s not obtainable equal to t(1) or
  //  to t(n+k) appropriately.
  //
  long jcmin=0;
    
  if(p<=iii){

    for(long j=0;j<p;j++){
      dl[j]=at-pknots[iii-j];
    }
       
  }else{
       
    jcmin=p-iii;
       
    for(long j=0;j<=iii;j++){
      dl[j]=at-pknots[iii-j];
    }
       
    for(long j=iii;j<p;j++){
      aj[p-j+1]=0.0;
      dl[j]=dl[iii];
    }
       
  }
    
  long jcmax=p;
    
  if(iii<N) goto l90;
  
    
  jcmax=p+N-iii-1;
  for(long j=0;j<p+N-iii;j++){
    dr[j]=pknots[iii+1+j]-at;
  }
    
  for(long j=p+N-iii-1;j<p;j++){
    aj[j+1]=0.0;
    dr[j]=dr[p+N-iii-1];
  }
    
  goto l110;
    
 l90:
    
  for(long j=0;j<p;j++){
    dr[j]=pknots[iii+1+j]-at;
  }
    
 l110:
    
  for(long jc=jcmin;jc<=jcmax;jc++){
    aj[jc]=bcoef.at(iii-p+jc);
  }
  //
  //  Difference the coefficients JDERIV times.
  //
  for(long j=0;j<jderiv;j++){
    long ilo=p-j-1;
    for(long jj=0;jj<p-j;jj++){
      aj[jj]=((aj[jj+1]-aj[jj])/(dl[ilo]+dr[jj]))*(p-j);
      ilo--;
    }
  }  
  //
  //  Compute value at X in (t(i),t(i+1)) of jderiv-th derivative,
  //  given its relevant b-spline coeffs in aj(1),...,aj(k-jderiv).
  //
  for(long j=jderiv;j<p;j++){
    long ilo=p-j-1;
    for(long jj=0;jj<p-j;jj++){
      aj[jj]=(aj[jj+1]*dl[ilo]+aj[jj]*dr[jj])/(dl[ilo]+dr[jj]);
      ilo--;
    }
  }
    
  return aj[0];
    
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
int BSpline::interv(double at,long &left)
{
  long npkm1=px.size()+porder-1;
  //
  // INTERV brackets a real value in an ascending vector of values.
  //
  //
  //  Discussion:
  //
  //    The PKNOTS array is a set of increasing values.  The goal of the routine
  //    is to determine the largest index I so that PKNOTS(I) <= X.
  //
  //    The routine is designed to be efficient in the common situation
  //    that it is called repeatedly, with X taken from an increasing
  //    or decreasing sequence.
  //
  //    This will happen when a piecewise polynomial is to be graphed.
  //    The first guess for LEFT is therefore taken to be the value
  //    returned at the previous call and stored in the local variable ILO.
  //
  //    A first check ascertains that ILO < NPK.  This is necessary
  //    since the present call may have nothing to do with the previous
  //    call.  Then, if PKNOTS(ILO) < = X < PKNOTS(ILO+1), we set LEFT=ILO
  //    and are done after just three comparisons.
  //
  //    Otherwise, we repeatedly double the difference ISTEP=IHI-ILO
  //    while also moving ILO and IHI in the direction of X, until
  //      PKNOTS(ILO) < = X < PKNOTS(IHI)
  //    after which we use bisection to get, in addition, ILO+1=IHI.
  //    LEFT=ILO is then returned.
  //
  //  Modified:
  //
  //    05 February 2004
  //
  //  Reference:
  //
  //    Carl DeBoor,
  //    A Practical Guide to Splines,
  //    Springer Verlag.
  //
  //  Parameters:
  //
  //    Input, real PKNOTS(NPK), a nondecreasing sequence of values.
  //
  //    Input, integer NPK, the dimension of PKNOTS.
  //
  //    Input, real X, the point whose location with 
  //    respect to the sequence PKNOTS is to be determined.
  //
  //    Output, integer LEFT, the index of the bracketing value:
  //      1     if             X  <  PKNOTS(1)
  //      I     if   PKNOTS(I)  <= X  < PKNOTS(I+1)
  //      NPK   if  PKNOTS(NPK) <= X
  //
  //    Output, integer MFLAG, indicates whether X lies within the
  //    range of the data.
  //    -1:            X  <  PKNOTS(1)
  //     0: PKNOTS(I)   <= X  < PKNOTS(I+1)
  //    +1: PKNOTS(NPK) <= X
  long ihi=pilo+1;
  long istep;
  if(npkm1<=ihi){
       
    if(pknots[npkm1]<=at) goto l110;
    
       
    if(npkm1<=0){
      left=0;
      return -1;
    }
       
    pilo=npkm1-1;
    ihi=npkm1;
  }
    
  if(pknots[ihi]<=at){
    goto l40;
  }
    
  if(pknots[pilo]<=at){
    left=pilo;
    return 0;
  }
  //
  //  Now X < PKNOTS(ILO).  Decrease ILO to capture X.
  //
  istep=1;
    
 l31:
    
  ihi=pilo;
  pilo=ihi-istep;
    
  if(0<pilo){
    if(pknots[pilo]<=at) goto l50;
    
    istep=istep*2;
    goto l31;
  }
    
  pilo=0;
    
  if(at<pknots[0]){
    left=0;
    return -1;
  }
    
  goto l50;
  //
  //  Now PKNOTS(IHI+1) <= X.  Increase IHI+1 to capture X.
  //
 l40:
    
  istep=1;
    
 l41:
    
  pilo=ihi;
  ihi=pilo+istep;
    
  if(ihi<npkm1){
    if(at<pknots[ihi]) goto l50;
    
    istep=istep*2;
    goto l41;
  }
    
  if(pknots[npkm1]<=at) goto l110;
  //
  //  Now PKNOTS(ILO) < = X < PKNOTS(IHI+1).  Narrow the interval.
  //
  ihi=npkm1;
    
 l50:
    
  while(true){
    
    long middle=(pilo+ihi)/2;
       
    if(middle==pilo){
      left=pilo;
      return 0;
    }
    //
    //  It is assumed that MIDDLE = ILO in case IHI+1 = ILO+1.
    //
    if(pknots[middle]<=at){
      pilo=middle;
    }else{
      ihi=middle;
    }
    
  }
  //
  //  Set output and return.
  //
 l110:
    
  int mflag=(at==pknots[npkm1]?0:1);

  for(long i=npkm1;i>=0;i--){
    if(pknots[i]<pknots[npkm1]){
      left=i;
      return mflag;
    }
  }
    
  return mflag;
}
/*EOF*/ 
