/*************************************************************************
 * @(#)EllipseShape.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   22 Nov 2011
 *************************************************************************
 */

#include "GraphObject.hh" 
#include "Graph.hh" 
#include "Attributes.hh" 
#include "EllipseShape.hh"
#include "ViewBox.hh"
#include "MyDebug.hh"

/* ***********************************************************************
 *
 * ***********************************************************************
 */
EllipseShape::EllipseShape(Graph *parent,long index)
  :Shape(parent,index),
   plinestyle(new LineStyle()),
   pfill(new FillStyle())
{
  ptype=Xfp::SHTEllipse;
  pname="";
  reset();
  
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
EllipseShape::~EllipseShape()
{
  if(plinestyle) delete plinestyle; plinestyle=0;
  if(pfill) delete pfill; pfill=0;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void EllipseShape::reset()
{
  pcenter=Point(0.5,0.5);
  pradii=Size(0.25,0.125);
  pangle=0;
  pcoordsystem=Xfp::CSYView;
  pextrafillcolor=true,
  plinestyle->reset();
  pfill->reset();
  pfill->setColor(0xeeeeee);
  pfill->setFill(true);
  pisCircle=false;
  setChanged(false);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool EllipseShape::changed()const
{
  return pchanged| plinestyle->changed() | pfill->changed();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void EllipseShape::setChanged(bool c)
{
  plinestyle->setChanged(c);
  pfill->setChanged(c);
  pchanged=c;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
GraphObject::GraphObjectRegion 
EllipseShape::insideRegion(const DragDescriptor &dd,bool update)
{
  
  if(!pactive) return GraphObject::GORNone;

  // work in absolute coordinates
  Point c=centerToDC()*dd.devicesize;
  Size r=radiiToDC()*dd.devicesize;
  Point pm=dd.mousepoint-c;
  Point p0=pm.rotate(-pangle);
  
  if(pisCircle)r.h=r.w;

  Point p=p0/r;
  if(norm(p)<1){
    if(update){
      pupdaterequestnumber=0;
      pupdaterequest=GOCPosition;
    }
    return GORBody;
  }
  Size ss=Size(dd.sensitivityAroundPoint,dd.sensitivityAroundPoint);
  if((p0.insideRect(Point(-r.w,-r.h)-ss,ss*2))){
    if(update){
      pupdaterequestnumber=0;
      pupdaterequest=GOCSize;
    }
    return GORUL;
  }else if((p0.insideRect(Point(r.w,-r.h)-ss,ss*2))){
    if(update){
      pupdaterequestnumber=1;
      pupdaterequest=GOCSize;
    }
    return GORUR;
  }else if((p0.insideRect(Point(r.w,r.h)-ss,ss*2))){
    if(update){
      pupdaterequestnumber=2;
      pupdaterequest=GOCSize;
    }
    return GORLR;
  }else if((p0.insideRect(Point(-r.w,r.h)-ss,ss*2))){
    if(update){
      pupdaterequestnumber=3;
      pupdaterequest=GOCSize;
    }
    return GORLL;
  }

  if(update){
    pupdaterequest=GOCNone;
  }
  return GraphObject::GORNone;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void EllipseShape::initDrag(const DragDescriptor &dd)
{
  GraphObjectRegion region=insideRegion(dd,true);

  if(region==GORNone)return;
  pmousepos=dd.mousepoint;

  pobjectpos=0;
  pdeltasize=0;
  
  
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void EllipseShape::updateDrag(const DragDescriptor &dd)
{
  if(pupdaterequest == GOCPosition){
    pobjectpos+=(dd.mousepoint-pmousepos)/dd.devicesize;
  }else{
    Size dr=((dd.mousepoint-pmousepos)/dd.devicesize).rotate(-pangle);
    if(pupdaterequestnumber==0){
      pdeltasize+=dr*Size(-1,-1);
    }else if(pupdaterequestnumber==1){
      pdeltasize+=dr*Size(1,-1);
    }else if(pupdaterequestnumber==2){
      pdeltasize+=dr;
    }else if(pupdaterequestnumber==3){
      pdeltasize+=dr*Size(-1,1);
    }
  }
  pmousepos=dd.mousepoint;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void EllipseShape::setFromDrag(DragDescriptor &)
{
  if(pupdaterequest & GOCSize){
    pradii=abs(dcToSize(radiiToDC(true),pcoordsystem));
    pdeltasize=0;
    return;
  }
  if(pupdaterequest & GOCPosition){
    pcenter=dcToPos(centerToDC(true),pcoordsystem);
  }
  pobjectpos=0;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String EllipseShape::asCommands(const String &tab)const
{
  String s;
  String name=Sprintf("%s%ld",Xfp::shapeTypeGuiName(ptype),puniqueindex+1);
  s+=Sprintf("%s%s %s\n",tab(),name(),Xfp::OnOff(active()));
  name=tab+"  "+name;
  s+=Sprintf("%s name \"%s\"\n",name(),pname());
  s+=lineStyle()->asCommands(name+" line");
  s+=fill()->asCommands(name+" fill",pextrafillcolor);
  s+=Sprintf(
	     "%s coordinate system %s\n"
 	     "%s center %g,%g\n",
	     name(),Xfp::coordinateSystemGuiName(pcoordsystem),
 	     name(),pcenter.x,pcenter.y);
  if(pisCircle){
    s+=Sprintf("%s radius %g\n",name(),pradii.w);
  }else{
    s+=Sprintf("%s radii %g,%g\n",name(),pradii.w,pradii.h);
  }
  s+=Sprintf("%s angle  %g\n",name(),pangle);
  return s;
}


/*EOF*/ 
