/*************************************************************************
 * @(#)Graph.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   21 Apr 2011
 *************************************************************************
 */

//#include <float.h>
#include <limits>
#include "Attributes.hh"
#include "PaintEngine.hh"
#include "GraphObject.hh"
#include "FGObject.hh"
#include "GraphAxis.hh"
#include "Function.hh"
#include "Functions.hh"
#include "FunctionGroup.hh"
#include "FunctionGroups.hh"
#include "Graph.hh"
#include "LegendBox.hh"
#include "MyMath.hh"
#include "Position.hh"
#include "Range.hh"
#include "TextBox.hh"
#include "TicMarks.hh"
#include "ValueArray.hh"
#include "WeightSelection.hh"
#include "OptMatelDefinition.hh"
#include "LineShape.hh"
#include "EllipseShape.hh"
#include "ViewBox.hh"
#include "World.hh"
#include "Pair.hh"
#include "MyDebug.hh"
#include "defs.hh"

#define FORMATTEXT_CALC_METRIC         0
#define FORMATTEXT_SHOW                1

#define CLIP_SEG_IS_OUTSIDE            0x00000000
#define CLIP_PART_OF_SEG_IS_INSIDE     0x00000001
#define CLIP_SECOND_POINT_IS_OUTSIDE   0x00000100
#define CLIP_FIRST_POINT_IS_OUTSIDE    0x00000200


const Graph::tsGraphID Graph::pgraphids[]={
  {Graph::NewGraph,"New graph"}
  ,{Graph::CurrentGraph,"Current graph"}
  ,{Graph::UnknownGraph,"Unknown graph"}
};

const long Graph::pmaxgraphid=sizeof(Graph::pgraphids)/sizeof(Graph::tsGraphID);


inline QRgb mToRgb(const double *c){
  return
    (MyMath::max<long>(0,MyMath::min<long>(255,ulRound(255.*c[0])))<<16) |
    (MyMath::max<long>(0,MyMath::min<long>(255,ulRound(255.*c[1])))<<8) |
    (MyMath::max<long>(0,MyMath::min<long>(255,ulRound(255.*c[2]))));
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
class Stripe{
public:
  Stripe(const Point &u,const Point &l,double v):value(v),ul(u),lr(l){};
  double value;
  Point ul,lr;
};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
class Stripes{
public: 
  Stripes(){};
  long size()const{return pstripes.size();};
  const Stripe *at(long i)const{return pstripes.at(i);};

  void print()const{
    long n=pstripes.size();
    for(long i=0;i<n;i++){
      const Stripe *si=pstripes.at(i);
      myDebug()<<i<<si->ul<<si->lr<<si->value;
    }
    
  }

  void append(Point &ul_,Point &lr_,double value_,double tol){
    double y0=ul_.y;
    double y1=lr_.y;

    if(y0>y1){ ul_.y=y1; lr_.y=y0; }

    if(pstripes.size()==0){
      pstripes.append(new Stripe(ul_,lr_,value_));
      return;
    }
    PtrArray<Stripe> padd;
    padd.append(new Stripe(ul_,lr_,value_));

    for(long ia=0;ia<padd.size();ia++){
      //myDebug()<<"padd ia="<<ia;
      Stripe *s=padd[ia];
      double dx=fabs(s->lr.x-s->ul.x);
      
      long n=pstripes.size();
      for(long i=0;i<n;i++){
	Stripe *si=pstripes[i];
	if(fabs(si->ul.x-s->ul.x)<dx*tol &&
	   fabs(si->lr.x-s->lr.x)<dx*tol ){
	  // have already one stripe at this x-range
	  
	  // supdivide the overlap and add "values"
	  
	  double y0=s->ul.y;
	  double y1=s->lr.y;
	  double Y0=si->ul.y;
	  double Y1=si->lr.y;
	  
	  if(y0<=Y0){
	    if(y1<=Y0){
	      //  DONE
	      //myDebug()<<"disjunct new above old";
	      // - y0
	      // |
	      // |
	      // - y1
	      //      - Y0
	      //      |
	      //      |
	      //      - Y1
	    }else if(y1>=Y1){
	      // DONE
	      //myDebug()<<"new envelops old";
	      // - y0
	      // |
	      // |    - Y0
	      // |    |
	      // |    |
	      // |    - Y1
	      // - y1
	      s->lr.y=Y0;
	      //myDebug()<<"si          "<<si->ul<<si->lr<<si->value;
	      //myDebug()<<"new shortend"<<s->ul<<s->lr<<s->value;
	      if(fabs(Y1-y1)>1.0e-12){
		padd.append(new Stripe(Point(s->ul.x,Y1),
				       Point(s->lr.x,y1),s->value));
		//myDebug()<<"append lower"<<s->ul<<s->lr<<s->value;
	      }
	      si->value+=s->value;
	      //myDebug()<<"old v updat "<<si->ul<<si->lr<<si->value;
	      
	    }else{
	      // DONE
	      //myDebug()<<"new partially above old";
	      // - y0
	      // |
	      // |    - Y0
	      // - y1 |
	      //      |
	      //      - Y1
	      //myDebug()<<"si          "<<si->ul<<si->lr<<si->value;
	      s->lr.y=Y0;
	      //myDebug()<<"new shortend"<<s->ul<<s->lr<<s->value;
	      si->ul.y=y1;
	      if(fabs(Y0-y1)>1.0e-12){
		pstripes.append(new Stripe(Point(s->ul.x,Y0),Point(s->lr.x,y1),
				       s->value+si->value));
	      }
	    }
	  }else if(y0>=Y1){
	    // DONE
	    //myDebug()<<"disjunct new below old";
	    //      - Y0
	    //      |
	    //      |
	    //      - Y1
	    // - y0  
	    // |     
	    // - y1     
	  }else{
	    if(y1<=Y1){
	      // 
	      //myDebug()<<"old envelops new";
	      //      - Y0
	      //      |
	      // - y0 | 
	      // |    |
	      // - y1 |
	      //      - Y1
	      si->lr.y=y0;
	      s->value+=si->value;
	      padd.append(new Stripe(Point(s->ul.x,y1),
				     Point(s->lr.x,Y1),si->value));
	    }else{
	      // DONE
	      //myDebug()<<"new partially below old";
	      //      - Y0
	      //      |
	      // - y0 | 
	      // |    |
	      // |    -Y1
	      // - y1
	      s->ul.y=Y1;
	      si->lr.y=y0;
	      pstripes.append(new Stripe(Point(s->ul.x,y0),Point(s->lr.x,Y1),
				     s->value+si->value));
	    }
	  }
	  
	  
	  
	}// same x-coordinates
      } // forall old 
      // now add the rest of this patch
      pstripes.append(padd[ia]);padd[ia]=0;
      
    }// for all to be added
    //pstripes.append(new Stripe(ul,lr,value));
    long n=pstripes.size();
    for(long i=n-1;i>=0;i--){
      Stripe *si=pstripes[i];
      if(fabs(si->ul.y-si->lr.y)<1.0e-10){
	pstripes.removeAt(i);
      }
    }

  };
  
private:
  PtrArray<Stripe> pstripes;
};


/* ***********************************************************************
 *
 * ***********************************************************************
 */
Graph::Graph():GraphObject(this)
{
  setObjectType(GraphObject::GOTGraph);
  printf("default constructor called ?!\n");
  exit(1);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
Graph::Graph(Page *parent,long uniqueid)
  :GraphObject(this),
  pviewbox(new ViewBox(this))
  ,pworld(new World(this))
  ,plegendbox(new LegendBox(this))
  ,pticmarks(new TicMarks(this))
  ,ptextboxes(new TextBoxes(this))
  ,pgroups(new FunctionGroups(this))
  ,pshapes(new Shapes(this))
  ,paxes(new GraphAxes(this))
  ,pactive(true)
  ,palloverpointscale(1.)
  ,palloverlinewidthscale(1.)
  ,ppage(parent)
  ,puniqueid(uniqueid)
{
  setObjectType(GraphObject::GOTGraph);
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
Graph::~Graph()
{
  /* NSF("clean_graph"); */

//   printf("Graph dying %p\n",this);
//   printf("Graph dying ppage=%p\n",ppage);
//   printf("Graph dying pgroups.size=%ld\n",pgroups.size());
//   printf("Graph dying pgroups[0]->parent=%p\n",pgroups[0]->parent());


  if(pviewbox)delete pviewbox;pviewbox=0;
  if(pworld)delete pworld;pworld=0;
  if(plegendbox) delete plegendbox;plegendbox=0;
  if(pticmarks)delete pticmarks;pticmarks=0;
  if(ptextboxes)delete ptextboxes;ptextboxes=0;
  if(pgroups)delete pgroups;pgroups=0;
  if(pshapes)delete pshapes;pshapes=0;
  if(paxes)delete paxes;paxes=0;
  
  
  
  
  


}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
String Graph::guiName(bool longname)const
{
  if(longname){
    return Sprintf("Graph%ld",puniqueid+1);
  }else{
    return Sprintf("G%ld",puniqueid+1);
  }
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool Graph::changed()const
{
  bool c=pchanged;
  if(pviewbox) c|=pviewbox->changed();
  if(pworld) c|=pworld->changed();
  if(plegendbox) c|=plegendbox->changed();
  if(pticmarks) c|=pticmarks->changed();
  if(ptextboxes) c|=ptextboxes->changed();
  if(pshapes) c|=pshapes->changed();
  if(paxes) c|=paxes->changed();
  if(pgroups) c|=pgroups->changed();
  
    
  return c;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::setChanged(bool c)
{
  if(pviewbox) pviewbox->setChanged(c);
  if(pworld)  pworld->setChanged(c);
  if(plegendbox) plegendbox->setChanged(c);
  if(pticmarks) pticmarks->setChanged(c);
  if(ptextboxes) ptextboxes->setChanged(c);
  if(pshapes) pshapes->setChanged(c);
  if(paxes) paxes->setChanged(c);
  if(pgroups) pgroups->setChanged(c);
  
  pchanged=c;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::changeScaling(Xfp::XYAxisType ,
			  Xfp::GraphAxisScaling ,Xfp::GraphAxisScaling )
{
  
  // pworld->setW0(xy,GraphAxes::toScaled(GraphAxes::fromScaled(pworld->w0(xy),oldsc),newsc));
  // pworld->setW1(xy,GraphAxes::toScaled(GraphAxes::fromScaled(pworld->w1(xy),oldsc),newsc));

  pgroups->invalidateRange();

}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotInit(PaintEngine *pe)
{

  pworld->checkWorld();

  pxopossitetic=false;
  pxnormsitetic=false;
  pyopossitetic=false;
  pynormsitetic=false;
  double woh=pe->paper().paperWidth()*1./pe->paper().paperHeight();
  woh=woh*(pviewbox->size().w/pviewbox->size().h);
  pticmarks->calcTicMarks(this,woh); 
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::autoScaleAll(Xfp::XYAxisType xy,bool restricted)
{



  bool haveone=false;
  double omin,omax;
  Xfp::XYAxisType othertype=Xfp::otherAxisType(xy);
  

  Range rext;
  

  omin=pworld->w0(othertype);
  omax=pworld->w1(othertype);

  for (long ig=0;ig<pgroups->size();ig++){
    FunctionGroup *group=pgroups->at(ig);
    if(group->active()){ 
      for (long j=0;j<group->funcs()->size();j++){
	Function *func=group->funcs()->at(j);
	if(func->active() && (func->length() > 0 )){ 
	  Range range;
	  func->calcRange(range,(restricted?othertype:-1),omin,omax);
	  if(!haveone){
	    rext=range;
	    haveone=true;
	  }else{
	    rext.extend(range,xy);
	  }
	} 
      }
    }
  }


  double dmin=rext.min(xy);
  double dmax=rext.max(xy);
  if(paxes->axisScaling(xy)==Xfp::GASLogarithmic){
    dmin=paxes->adjustUnscaled(xy,rext.minpos(xy));
    dmax=paxes->adjustUnscaled(xy,rext.maxpos(xy));
  }

  double dy=dmax-dmin;
  double x0=dmin-dy*pworld->autoScaleOffset(xy);
  double x1=dmax+dy*pworld->autoScaleOffset(xy);

  
  pworld->setW01(xy,x0,x1);

  pworld->checkWorld();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::autoScaleFunc(FGObject *fgo)
{
  if(!fgo) return;
  Function *func=(Function*)fgo;
  Range range;
  func->calcRange(range,-1,0.,1.);



  for(int i=Xfp::Graph_X;i<=Xfp::Graph_Y;i++){
    Xfp::XYAxisType xy=(Xfp::XYAxisType)i;
    double dmin=range.min(xy);
    double dmax=range.max(xy);
    if(paxes->axisScaling(xy)==Xfp::GASLogarithmic){
      dmin=GraphAxes::toScaled(range.minpos(xy),Xfp::GASLogarithmic);
      dmax=GraphAxes::toScaled(range.maxpos(xy),Xfp::GASLogarithmic);
    }
    double dy=dmax-dmin;
    double x0=dmin-dy*pworld->autoScaleOffset(xy);
    double x1=dmax+dy*pworld->autoScaleOffset(xy);
    pworld->setW01(xy,x0,x1);
  }
  pworld->checkWorld();
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */



void Graph::plotGraph(PaintEngine *pe)
{
  //NSF("plot_graph");

  if(!pactive) return;

  pe->setAllOverPointScale(palloverpointscale);
  pe->setAllOverLineWidthScale(palloverlinewidthscale);
  plotInit(pe);
  
  plot_viewbox(pe,pviewbox,VBPMFill);

  plotfuncs(pe);

  //  plotaxes(pe); 

  
  plot_viewbox(pe,pviewbox,VBPMFrame);

  plotticsandmarks(pe);  
 

  plotTextBox(pe,ptextboxes->textBoxByType(TextBox::TBTYAxisLabel));
  plotTextBox(pe,ptextboxes->textBoxByType(TextBox::TBTOppositeYAxisLabel));
  plotTextBox(pe,ptextboxes->textBoxByType(TextBox::TBTXAxisLabel));
  plotTextBox(pe,ptextboxes->textBoxByType(TextBox::TBTOppositeXAxisLabel));
  pe->defineValue(PaintEngine::XOppositeXALabelHeight
 		  ,PaintEngine::TextBoxHeight,1.); 
  plotTextBox(pe,ptextboxes->textBoxByType(TextBox::TBTSubTitle));
  pe->defineValue(PaintEngine::SubtitleHeight,PaintEngine::TextBoxHeight
		  ,1.); 
  plotTextBox(pe,ptextboxes->textBoxByType(TextBox::TBTTitle));
  

  plot_legend(pe,plegendbox);

  plottextboxes(pe);

  plotshapes(pe);



  //CPR("END");
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotGraphLegend(PaintEngine *pe)
{
  if(!pactive) return;
  pe->setAllOverPointScale(palloverpointscale);
  pe->setAllOverLineWidthScale(palloverlinewidthscale);
  plot_legend(pe,plegendbox);
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */



void Graph::additionalOffset(PaintEngine *pe,TextBox *tb)
{
  /* NSF("additionaloffset_graph"); */
  long i,at;
  double hx=0.,hy=0.;
  if(!tb) return;
  const Position *ps=tb->pos();
  if(!ps) return;


  /* CPRL(ps->naddtypes); */
  for(i=0;i<ps->numAddTypes();i++){
    at=ps->addType(i);
    /* CPRXL(at); */
    
    if(at == Position::XOposSiteTicLabelHeight){
      if(!pe->isPSEngine()){
	hy+=-pe->getValue(PaintEngine::XOpositeTicLabelHeight
			  ,PaintEngine::RelativeY)
	  /pviewbox->size().h;
      }else{
	pe->translate(0.,PaintEngine::Absolute
		      ,PaintEngine::XOpositeTicLabelHeight,-1.);
	
      }
    }else if(at == Position::SubTitleOffset){
      TextBox *tb=ptextboxes->textBoxByType(TextBox::TBTSubTitle);
      if(tb->active()){
	if(!pe->isPSEngine()){
	  hy+=-tb->pos()->offset();
	}else{
	  pe->translate(0.,PaintEngine::Absolute
			,-tb->pos()->offset()*pviewbox->size().h
		       ,PaintEngine::RelativeY);
	}
      }
    }else if(at == Position::SubTitleHeight){
      if(!pe->isPSEngine()){
	hy+=-pe->getValue(
			 PaintEngine::SubtitleHeight
			 ,PaintEngine::RelativeY)
	  /pviewbox->size().h;
      }else{
	pe->translate(0.,PaintEngine::Absolute
		     ,PaintEngine::SubtitleHeight,-1.);
	
      }
    }else if(at == Position::OppXaLabelOffset){
      TextBox *tb=ptextboxes->textBoxByType(TextBox::TBTOppositeXAxisLabel);
      if(tb->active()){
	if(!pe->isPSEngine()){
	  hy+=-tb->pos()->offset();
	}else{
	  pe->translate(0.,PaintEngine::Absolute
			,-tb->pos()->offset()*pviewbox->size().h
		       ,PaintEngine::RelativeY);
	}
      }
    }else if(at == Position::OppXaLabelHeight){
      if(!pe->isPSEngine()){
	hy+=-pe->getValue(
			 PaintEngine::XOppositeXALabelHeight
			 ,PaintEngine::RelativeY)
	  /pviewbox->size().h;
      }else{
	pe->translate(0.,PaintEngine::Absolute
		     ,PaintEngine::XOppositeXALabelHeight,-1.);
	
      }
    }else if(at == Position::XOposSiteTicLabelOffset){
      if(pxopossitetic){
	if(!pe->isPSEngine()){
	  hy+=-pticmarks->labelOffset(Xfp::Graph_X);
	}else{
	  pe->translate(0.,PaintEngine::Absolute
			,-pticmarks->labelOffset(Xfp::Graph_X)
		       *pviewbox->size().h
		       ,PaintEngine::RelativeY);
	  
	}
      }	/* xopossitetic */
    }else if( at == Position::XNormSiteTicLabelHeight ){
      if(!pe->isPSEngine()){
	hy+=pe->getValue(PaintEngine::XNormalTicLabelHeight
			,PaintEngine::RelativeY)
	  /pviewbox->size().h;
      }else{
	pe->translate(0.,PaintEngine::Absolute
		     ,PaintEngine::XNormalTicLabelHeight,1.);
      }
    }else if( at == Position::XNormSiteTicLabelOffset){
      if(pxnormsitetic){
	if(!pe->isPSEngine()){
	  hy+=pticmarks->labelOffset(Xfp::Graph_X);
	}else{
	  pe->translate(0.,PaintEngine::Absolute
			,pticmarks->labelOffset(Xfp::Graph_X)
		       *pviewbox->size().h
		       ,PaintEngine::RelativeY);
	  
	}
      }	/* xnormsitetic */
    }else if( at == Position::YOposSiteTicLabelWidth){
      if(!pe->isPSEngine()){
	hx+=pe->getValue(
			 PaintEngine::YOpositeTicLabelWidth
			 ,PaintEngine::RelativeX)
	  /pviewbox->size().w;
      }else{
	pe->translate(PaintEngine::YOpositeTicLabelWidth,1. 
		      ,0.,PaintEngine::Absolute);
	
      }
    }else if( at == Position::YOposSiteTicLabelOffset){
      if(pyopossitetic){
	if(!pe->isPSEngine()){
	  hx+=pticmarks->labelOffset(Xfp::Graph_Y);
	}else{
	  pe->translate(pticmarks->labelOffset(Xfp::Graph_Y) 
		       *pviewbox->size().w ,PaintEngine::RelativeX
			,0.,PaintEngine::Absolute);
	  
	}
      }	/* yopossitetic */
    }else if( at == Position::YNormSiteTicLabelWidth ){
      if(!pe->isPSEngine()){
	hx+=-pe->getValue(PaintEngine::YNormalTicLabelWidth
			     ,PaintEngine::RelativeX)
	  /pviewbox->size().w;
      }else{
	pe->translate(PaintEngine::YNormalTicLabelWidth,-1.
		      ,0.,PaintEngine::Absolute );
      }
    }else if( at == Position::YNormSiteTicLabelOffset){
      if(pynormsitetic){
	if(!pe->isPSEngine()){
	  hx+=-pticmarks->labelOffset(Xfp::Graph_Y);
	}else{
	  pe->translate(-pticmarks->labelOffset(Xfp::Graph_Y) 
		       *pviewbox->size().w,PaintEngine::RelativeX
			,0.,PaintEngine::Absolute );
	  
	}
      }	/* ynormsitetic */
    } /* at type */
  } /* all addtypes */
  


  tb->pos()->setAddPos(Point(hx,hy));
}

/* ***********************************************************************
 * ax and ay are absolute device coordinates
 * ***********************************************************************
 */
void Graph::plot_drag_textbox(PaintEngine *pe,GraphObject *go)
{
  /* NSF("plot_textbox"); */
  double originsize=4.;
  if(!go) return;
  TextBox *tb=(TextBox*)go;
  
  if(!tb) return;
  if(!tb->active()) return;
  


  {
    pe->save();
    /* shift origin of textbox to object position */
    pe->translate(go->objectPos().x,PaintEngine::Absolute
		 ,go->objectPos().y,PaintEngine::Absolute);
    /* rotate the textbox about origin */
    pe->rotate(tb->pos()->angle());
    
    
    {
      pe->save();
      /* got to UL corner */
      pe->translate(-tb->size().w*tb->pos()->origin().x,
		    PaintEngine::Absolute,
		    -tb->size().h*tb->pos()->origin().y,
		    PaintEngine::Absolute );
      pe->setPen(0xff0000,Xfp::LSTSolid,1);   
      /* draw a box */
      pe->drawRectFromValue(0.,0.,tb->size().w,tb->size().h,
			    0x0,false,PaintEngine::ULWH );
      pe->restore();
    }
    /* draw the origin cross */
    pe->setPen(0xff0000,Xfp::LSTSolid,1);   
    pe->drawLineFromValue(-originsize,-originsize,
			  +2*originsize+1,+2*originsize+1,
			  PaintEngine::ULWH );
    pe->drawLineFromValue(-originsize,
			  +originsize,
			  +2*originsize+1,
			  -2*originsize-1,
			  PaintEngine::ULWH );
    
    pe->restore();
  }

  {
    pe->save();
    /* draw the offset "spring" */
    pe->setPen(0xff0000,Xfp::LSTSolid,1);   
    
    if( tb->pos()->restriction() ){
      pe->setPen(0xff0000,Xfp::LSTSolid,1);   

      Point p=tb->pos()->pos0ToDC(this);
      pe->drawSpringFromValue(go->objectPos().x,PaintEngine::Absolute,
			      go->objectPos().y,PaintEngine::Absolute,
			      p.x,PaintEngine::RelativeX,
			      p.y,PaintEngine::RelativeY,
			      PaintEngine::ULLR );
      
      
    } /* specail treatment of spring objects */
    pe->restore();
  }  


}

/* ***********************************************************************
 *  postscript formated text
 *  special characters are:  $~ : next letter is greek
 *                           $i : switch on italic 
 *                           $n : switch on normal font 
 *                           $^ : go one superscript leve up
 *                           $_ : go one subscript level down
 *                           $. : go to normal level
 *                           $$ : a single $$
 *                           $ not followed by ~^_. : print a single
 *
 * ***********************************************************************
 */
void Graph::formText(PaintEngine *pe
		    ,const double xx0,const double yy0,const String &txt1
		    ,double const angle
		     ,const double fontsize,Xfp::TextOrientation just
		    ,const double factor
		    ,bool show)
{
  /* NSF("formtext_graph"); */
  char buf[30000];
  char *pbuf;
  long depth,i,fsz,oldfsz,len,j;
  double fontsizes[3];
  long   currentfont,metricloop,caught;
  double lfactor[3];
  double ufactor[3];
  
  
   

  /* calc local font size */


  fontsizes[0]=fontsize*1.0;
  fontsizes[1]=fontsize*factor;
  fontsizes[2]=fontsize*factor*0.857;
  

  ufactor[0]=.7;
  ufactor[1]=.7;
  ufactor[2]=.7;
  depth=fsz=0;

  lfactor[0]=0.6;
  lfactor[1]=0.6;
  lfactor[2]=0.6;



  String stxt=txt1;
  stxt.stripBlanks();
  len=stxt.length();

  pe->save();
  pe->translate(xx0,PaintEngine::RelativeX,yy0,PaintEngine::RelativeY);
  pe->rotate(angle);
  pe->moveTo(0.,0.);

  pe->textInit();
  
    

  for(metricloop=0;metricloop< (show?2:1);metricloop++){
    pe->initCalcTextWidth();
    depth=fsz=0;
    i=0;
    pbuf=buf;
    *pbuf='\0';
    currentfont=Xfp::HelveticaIsolatin1;
    
    while(i<=len){
      
      if( stxt[i]=='$'){
	caught=0;
	if( !caught && i+7<=len){
	  /* $arrowup */
	  if(stxt.subString(i+1).find("arrowup")==0){
	    /* print arrowup in symbol font */
	    /* set greek font */
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    currentfont=Xfp::Symbol;
	    buf[0]=Xfp::UpArrow;
	    buf[1]='\0';
	    if(metricloop==0){
	      pe->metricText(buf,currentfont,fontsizes[fsz]);
	    }else{
	      pe->flushText(buf,currentfont,fontsizes[fsz]);
	    }
	    pbuf=buf;buf[0]='\0';
	    currentfont=Xfp::HelveticaIsolatin1;
	    i+=8;
	    caught=1;
	    continue;
	  }
	}
	if( !caught && i+9<=len){
	  /* $arrowdown */
	  if(stxt.subString(i+1).find("arrowdown")==0){
	    /* print arrowdown in symbol font */
	    /* set greek font */
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    currentfont=Xfp::Symbol;
	    buf[0]=Xfp::DownArrow;
	    buf[1]='\0';
	    if(metricloop==0){
	      pe->metricText(buf,currentfont,fontsizes[fsz]);
	    }else{
	      pe->flushText(buf,currentfont,fontsizes[fsz]);
	    }
	    pbuf=buf;buf[0]='\0';
	    currentfont=Xfp::HelveticaIsolatin1;
	    i+=10;
	    caught=1;
	    continue;
	  }
	}
	if( !caught && i+9<=len){
	  /* $arrowright */
	  if(stxt.subString(i+1).find("arrowright")==0){
	    /* print arrowright in symbol font */
	    /* set greek font */
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    currentfont=Xfp::Symbol;
	    buf[0]=Xfp::RightArrow;
	    buf[1]='\0';
	    if(metricloop==0){
	      pe->metricText(buf,currentfont,fontsizes[fsz]);
	    }else{
	      pe->flushText(buf,currentfont,fontsizes[fsz]);
	    }
	    pbuf=buf;buf[0]='\0';
	    currentfont=Xfp::HelveticaIsolatin1;
	    i+=11;
	    caught=1;
	    continue;
	  }
	}
	if( !caught && i+9<=len){
	  /* $arrowleft */
	  if(stxt.subString(i+1).find("arrowleft")==0){
	    /* print arrowleft in symbol font */
	    /* set greek font */
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    currentfont=Xfp::Symbol;
	    buf[0]=Xfp::LeftArrow;
	    buf[1]='\0';
	    if(metricloop==0){
	      pe->metricText(buf,currentfont,fontsizes[fsz]);
	    }else{
	      pe->flushText(buf,currentfont,fontsizes[fsz]);
	    }
	    pbuf=buf;buf[0]='\0';
	    currentfont=Xfp::HelveticaIsolatin1;
	    i+=10;
	    caught=1;
	    continue;
	  }
	}

	if( !caught && i+1<=len){
	  /* $arrowleft */
	  if(stxt.subString(i+1).find("x{")==0){
	    //myDebug()<<"here";
	    /* move by some amount */
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    currentfont=Xfp::HelveticaIsolatin1;
	    double dx=0;
	    if((j=stxt.subString(i+1).find("}"))>1){
	      //myDebug()<<"got}"<<stxt.subString(i+3,j+i);
	      try{
		dx=stxt.subString(i+3,j+i).convertToReal();
	      }catch(...){dx=0;}
	      if((just & Xfp::TOCenter) == Xfp::TOCenter) dx=dx*2;
	      if(metricloop==0){
		pe->metricMove(dx,0.,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushMove(dx,0.,currentfont,fontsizes[fsz]);
	      }
	      i+=j+2;
	      caught=1;
	    }else{
	      i=i+1;
	      caught=0;
	    }
	    pbuf=buf;buf[0]='\0';
	    continue;
	  }
	}

	if( !caught && i+9<=len){
	  /* $angstroem */
	  if(stxt.subString(i+1).find("angstroem")==0){
	    /* print angstroem in Helvetica font */
	    /* set greek font */
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    currentfont=Xfp::HelveticaIsolatin1;
	    buf[0]=Xfp::Angstroem;
	    buf[1]='\0';
	    if(metricloop==0){
	      pe->metricText(buf,currentfont,fontsizes[fsz]);
	    }else{
	      pe->flushText(buf,currentfont,fontsizes[fsz]);
	    }
	    pbuf=buf;buf[0]='\0';
	    currentfont=Xfp::HelveticaIsolatin1;
	    i+=10;
	    caught=1;
	    continue;
	  }
	}
	if( !caught && i+8<=len){
	  /* $infinity */
	  if(stxt.subString(i+1).find("infinity")==0){
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    currentfont=Xfp::Symbol;
	    buf[0]=Xfp::Infinity;
	    buf[1]='\0';
	    if(metricloop==0){
	      pe->metricText(buf,currentfont,fontsizes[fsz]);
	    }else{
	      pe->flushText(buf,currentfont,fontsizes[fsz]);
	    }
	    pbuf=buf;buf[0]='\0';
	    currentfont=Xfp::HelveticaIsolatin1;
	    i+=9;
	    caught=1;
	    continue;
	  }
	}
	if( !caught && i+1<=len){
	  if( stxt[i+1]=='~'){
	    /* set greek font */
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    if(i+2<=len){
	      currentfont=Xfp::Symbol;
	      buf[0]=stxt[i+2];
	      buf[1]='\0';
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	      currentfont=Xfp::HelveticaIsolatin1;
	      i++;
	    }
	    i+=2;
	    continue;
	  }else  if( stxt[i+1]=='i'){
	    /* set italic helvetica */
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    if(i+2<=len){
	      currentfont=Xfp::HelveticaObliqueIsolatin1;
	      //buf[0]=stxt[i+2];
	      buf[0]='\0';
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	      //i++;
	    }
	    i+=2;
	    continue;
	  }else  if( stxt[i+1]=='n'){
	    /* set italic helvetica */
	    //myDebug()<<"stxt="<<stxt.subString(i);
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    if(i+2<=len){
	      currentfont=Xfp::HelveticaIsolatin1;
	      buf[0]=stxt[i+2];
	      buf[0]='\0';
	      //myDebug()<<"buf="<<buf;
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	      //i++;
	    }
	    i+=2;
	    //myDebug()<<"stxt="<<stxt.subString(i);
	    continue;
	  }else if( stxt[i+1]=='_'){
	    /* increase subscript level */
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    if(i+2<=len){
	      oldfsz=fsz;
	      depth++;
	      fsz=MyMath::min<long>(depth,2);
	      if(metricloop==0){
		pe->metricMove(0.,lfactor[oldfsz],currentfont,fontsizes[fsz]);
	      }else{
		pe->flushMove(0.,lfactor[oldfsz],currentfont,fontsizes[fsz]);
	      }
	    }
	    i+=2;
	    continue;
	  }else if( stxt[i+1]=='^'){
	    /* increase superscript level */
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    if(i+2<=len){
	      oldfsz=fsz;
	      depth++;
	      fsz=MyMath::min<long>(depth,2);
	      if(metricloop==0){
		pe->metricMove(0.,-ufactor[oldfsz],currentfont,fontsizes[oldfsz]);
	      }else{
		pe->flushMove(0.,-ufactor[oldfsz],currentfont,fontsizes[oldfsz]);
	      }
	    }
	    i+=2;
	    continue;
	  }else if( stxt[i+1]=='.'){
	    /* Back to  normal level */
	    if(pbuf!=buf){
	      if(metricloop==0){
		pe->metricText(buf,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushText(buf,currentfont,fontsizes[fsz]);
	      }
	      pbuf=buf;buf[0]='\0';
	    }
	    if(i+2<=len){
	      depth=fsz=0;
	      if(metricloop==0){
		pe->metricMove(0.,0.,currentfont,fontsizes[fsz]);
	      }else{
		pe->flushMove(0.,0.,currentfont,fontsizes[fsz]);
	      }
	      
	    }
	    i+=2;
	    continue;
	  }else if( stxt[i+1]=='$'){
	    /* print $$ in current font */
	    *pbuf++=stxt[i];*pbuf='\0';
	    i+=2;
	    continue;
	  }else{
	    /* single $ without special meaning */
	    *pbuf++=stxt[i];*pbuf='\0';
	    i++;
	    continue;
	  }
	}else{
	  /* single $ at the end*/
	  *pbuf++=stxt[i];*pbuf='\0';
	  i++;
	  continue;
	}
      }else{
	/* normal character */
	*pbuf++=stxt[i];*pbuf='\0';
	i++;
	continue;
      }
    }
    if(pbuf!=buf){
      if(metricloop==0){
	pe->metricText(buf,currentfont,fontsizes[fsz]);
      }else{
	pe->flushText(buf,currentfont,fontsizes[fsz]);
      }
      pbuf=buf;buf[0]='\0';
    }

    
    if(metricloop==0){ 
      pe->adjustText(just);
    } 
  } /* metric loop, first for centering second for printing */

  pe->textClean();
  pe->restore();
  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotTextBox(PaintEngine *pe,TextBox *tb)
{
  /* NSF("plotTextBox"); */
  
  
  if(!tb) return;

  


  pe->defineValue(PaintEngine::TextBoxHeight,0.,PaintEngine::Absolute );
  
  if(!tb->active()) return;

  pe->save();

  if(tb->pos()->restriction()){
    additionalOffset(pe,tb);  
  }
  Point q0=tb->pos()->absposToDC(this);




  formText(pe
	   ,q0.x,q0.y,tb->text(),0.
	   ,tb->textStyle()->size() , Xfp::TOTopLeft
	   ,tb->textStyle()->subScriptScale()
	   ,FORMATTEXT_CALC_METRIC);

  pe->arithValue(PaintEngine::TextHeight,
		 PaintEngine::TextHeight,1.,
		 3.,PaintEngine::Absolute,PaintEngine::Max);
  pe->arithValue(PaintEngine::TextWidth,
		 PaintEngine::TextWidth,1.,
		 3.,PaintEngine::Absolute,PaintEngine::Max);
  
  pe->arithValue(PaintEngine::TextBoxWidth,
		 PaintEngine::TextWidth,1.,
		 PaintEngine::TextHeight,tb->border()*2,
		 PaintEngine::Plus);
  pe->defineValue(PaintEngine::TextBoxHeight,PaintEngine::TextHeight
		  ,1.+tb->border()*2);
  
  pe->translate(q0.x,PaintEngine::RelativeX,q0.y,PaintEngine::RelativeY );

  pe->rotate(tb->pos()->angle());

  pe->translate(PaintEngine::TextBoxWidth,-tb->pos()->origin().x,
		PaintEngine::TextBoxHeight,-tb->pos()->origin().y);


  
  if(tb->boxIsActive()){
    pe->save();
    pe->setPen(tb->rim()->color(),tb->rim()->style(),tb->rim()->width());   
    pe->drawRectFromValue(0.,PaintEngine::RelativeX,
			  0.,PaintEngine::RelativeY,
			  PaintEngine::TextBoxWidth,1.,
			  PaintEngine::TextBoxHeight,1.,
			  tb->fill()->color(),tb->fill()->fill(),
			  PaintEngine::ULWH);
    pe->restore();
  }
  
  pe->translate(PaintEngine::TextHeight,tb->border(),
		PaintEngine::TextHeight,tb->border());
  
  pe->setPen(tb->textStyle()->color(),Xfp::LSTSolid,1);
  formText(pe
	   ,0.,0.,tb->text(),0.
	   ,tb->textStyle()->size() , Xfp::TOTopLeft
	   ,tb->textStyle()->subScriptScale()
	   ,FORMATTEXT_SHOW);
  
  if(!pe->isPSEngine()){
    tb->setSize(Size(
		     pe->getValue(PaintEngine::TextBoxWidth,PaintEngine::Absolute)
		     ,pe->getValue(PaintEngine::TextBoxHeight,PaintEngine::Absolute)));
  }
  pe->restore();
  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plot_drag_legendbox(PaintEngine *pe,GraphObject *go)
{
  /* NSF("plot_textbox"); */
  
  double originsize=4.;

  if(!go) return;
  LegendBox *lb=(LegendBox*)go;
  if(!lb) return;
  if(!lb->active()) return;  


  
  pe->save();

  pe->translate(go->objectPos().x,PaintEngine::Absolute,
		go->objectPos().y,PaintEngine::Absolute );

  /* pe->rotate(device,lb->angle); */
  
  pe->save();
  pe->translate(-lb->size().w*lb->origin().x,PaintEngine::Absolute,
		-lb->size().h*lb->origin().y,PaintEngine::Absolute);
  pe->setPen(0xff0000,Xfp::LSTSolid,1);   
  pe->drawRectFromValue(0.,0.,lb->size().w,lb->size().h,
			0x0,false,PaintEngine::ULWH );
  pe->restore();

  pe->setPen(0xff0000,Xfp::LSTSolid,1);   
  pe->drawLineFromValue(-originsize,-originsize,
			+2*originsize+1,+2*originsize+1,
			PaintEngine::ULWH);
  pe->drawLineFromValue(-originsize,+originsize,
			+2*originsize+1,-2*originsize-1,
			PaintEngine::ULWH );

  pe->restore();


  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

void Graph::legDetTextSize(PaintEngine *pe,const Point &q0,const String &text,
			const LegendBox *lb)
{
  formText(pe
	   ,q0.x,q0.y,text,0.
	   ,lb->textStyle()->size(),Xfp::TOLeft
	   ,lb->textStyle()->subScriptScale()
	   ,FORMATTEXT_CALC_METRIC) ;
  
  pe->arithValue(PaintEngine::LegendTextMaxHeight,
		 PaintEngine::LegendTextMaxHeight,1.,
		 PaintEngine::TextHeight,1.,PaintEngine::Max);
  
  pe->arithValue(PaintEngine::LegendTextMaxWidth,
		 PaintEngine::LegendTextMaxWidth,1.,
		 PaintEngine::TextWidth,1. ,PaintEngine::Max);
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::legDrawEntry(PaintEngine *pe,const FunctionAttribs *at,
			 double legsymwidth,bool adjusttosymbols,
			 const String &text,LegendBox *lb,
			 bool widthLine)
{
  {
    pe->save();
    pe->setPen( at->lineStyle().color(),
		at->lineStyle().style(),
		at->lineStyle().width() );   
    
    pe->translate(0.,PaintEngine::Absolute
		  ,PaintEngine::LegendTextMaxHeight,0.5);

    if(widthLine && at->lineStyle().style()!=Xfp::LSTNone){
      pe->drawLineFromValue(0.,PaintEngine::Absolute,
			    0.,PaintEngine::Absolute,
			    PaintEngine::LegendTextMaxHeight,
			    legsymwidth,
			    0.,PaintEngine::Absolute,
			    PaintEngine::ULWH);
    }    
    
    
    pe->drawSymbolFromValue(PaintEngine::LegendTextMaxHeight,
			    legsymwidth/2.,
			    0.,PaintEngine::Absolute,
			    at,-1. 
			    /*TODO: as option ,lb->fontsize*0.9  */
			    );
    
    pe->restore();
  }
  
  pe->save();
  
  if(adjusttosymbols){
    pe->translate(0.,PaintEngine::Absolute
		  ,PaintEngine::LegendSymbolOffset,1.);
  }
  
  pe->translate(PaintEngine::LegendTextMaxHeight
		,legsymwidth+lb->symbolSpacing()
		,0.,PaintEngine::Absolute );
  
  
  formText(pe,0.,0.,text,0.
	   ,lb->textStyle()->size(),Xfp::TOTopLeft
	   ,lb->textStyle()->subScriptScale()
	   ,FORMATTEXT_SHOW);
  pe->restore();
  
  pe->translate(0.,PaintEngine::Absolute
		,PaintEngine::LegendTextMaxHeight
		,1.+lb->lineSpacing());
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */

void Graph::plot_legend(PaintEngine *pe,LegendBox *lb)
{
  /* NSF("plotlegend_graph"); */
  long i,ig,nf;
  bool adjusttosymbols=true;
  FunctionAttribs weightat;
  FunctionGroup *group;

  if(!lb) return;


  if(!lb->active()) return;

  Point q0=pviewbox->vctodc(lb->pos());
 

  pe->save();

  weightat.lineStyle().setWidth(1);
  weightat.lineStyle().setStyle(Xfp::LSTSolid);
  weightat.symbolStyle().setSize(12.);
  


  pe->defineValue(PaintEngine::LegendTextMaxHeight,0.
		  ,PaintEngine::Absolute); 
  pe->defineValue(PaintEngine::LegendTextMaxWidth,0.
	      ,PaintEngine::Absolute); 


  /* avoid zero legendbox dimensions  */

  /* TODO */
  formText(pe
	   ,q0.x,q0.y,"a",0.
	   ,lb->textStyle()->size(),Xfp::TOLeft
	   ,lb->textStyle()->subScriptScale()
	   ,FORMATTEXT_CALC_METRIC) ;
  
  pe->arithValue(PaintEngine::LegendTextMaxHeight,
		 PaintEngine::LegendTextMaxHeight,1.,
		 PaintEngine::TextHeight,1.,PaintEngine::Max);
  
  pe->arithValue(PaintEngine::LegendTextMaxWidth,
		 PaintEngine::LegendTextMaxWidth,1.,
		 PaintEngine::TextWidth,1.,PaintEngine::Max);



  double legsymwidth=0;



  nf=0;
  

  for(ig=0;ig<pgroups->size();ig++){ 
    group=pgroups->at(ig);
    if( group->active()){
      if(group->useAttribs()){
	if(group->attribs()->useLegend()){
	  nf++;
	  legDetTextSize(pe,q0,group->attribs()->legend(),lb);
	  legsymwidth=MyMath::max<double>(legsymwidth,lb->symbolWidth());
	  
	}
	WeightSelections *wss=group->weightSelections();
	if(wss && wss->displayWeightLegend()){
	  for(long iw=0;iw<wss->size();iw++){
	    const WeightSelection *ws=wss->at(iw);
	    if(ws && ws->active()){
	      const WeightLabel *wl=ws->weightLabel();
	      if(wl){
		//myDebug() << "iw"<<iw<<wl<<wl->label()();
		nf++;
		legDetTextSize(pe,q0,wl->label(),lb);
		legsymwidth=MyMath::max<double>(legsymwidth,1);
	      }
	    }
	  }
	}
      }else{
	for(i=0;i<group->funcs()->size();i++){ 
	  Function *func=group->funcs()->at(i);
	  if(func->active()){
	    if(func->attribs()->useLegend()){
	      nf++;
	      legDetTextSize(pe,q0,func->attribs()->legend(),lb);
	      legsymwidth=MyMath::max<double>(legsymwidth,lb->symbolWidth());
	    }
	    WeightSelections *wss=func->weightSelections();
	    if(wss && wss->displayWeightLegend()){
	      for(long iw=0;iw<wss->size();iw++){
		const WeightSelection *ws=wss->at(iw);
		if(ws && ws->active()){
		  const WeightLabel *wl=ws->weightLabel();
		  if(wl){
		    //myDebug() << "iw"<<iw<<wl<<wl->label()();
		    //myDebug() << "iw"<<iw<<wl;
		    nf++;
		    legDetTextSize(pe,q0,wl->label(),lb);
		    legsymwidth=MyMath::max<double>(legsymwidth,1);
		  }
		}
	      }
	    }
	  }
	} /*  forall funcs */
      }	/* seperate attribs */
    } /* active */
  }

  if(adjusttosymbols){
    
    pe->arithValue(PaintEngine::LegendSymbolOffset,
		   PaintEngine::MaxSymbolSize,1.,
		   PaintEngine::LegendTextMaxHeight,1. ,PaintEngine::Max);
    pe->arithValue(PaintEngine::LegendSymbolOffset,
		   PaintEngine::LegendSymbolOffset,0.5,
		   PaintEngine::LegendTextMaxHeight,0.5 ,PaintEngine::Minus);
    pe->arithValue(PaintEngine::LegendTextMaxHeight,
		   PaintEngine::MaxSymbolSize,1.,
		   PaintEngine::LegendTextMaxHeight,1. ,PaintEngine::Max);
  }


  /* W=TextMaxHeigth*(border*2 +sumbolwidth+symbolspacing)+TextWidth*/
  pe->arithValue(PaintEngine::LegendBoxWidth,
		 PaintEngine::LegendTextMaxWidth,1.,
		 PaintEngine::LegendTextMaxHeight,
		 lb->border()*2+legsymwidth+lb->symbolSpacing(),
	 	 PaintEngine::Plus); 
  
 
  /* H=(2*border+n+(n-1)*(skip))*(maxheight) */
  pe->defineValue(PaintEngine::LegendBoxHeight,PaintEngine::LegendTextMaxHeight
		  ,nf+lb->lineSpacing()*(nf-1)+lb->border()*2 ); 



  pe->translate(q0.x,PaintEngine::RelativeX,q0.y,PaintEngine::RelativeY );

  pe->translate(PaintEngine::LegendBoxWidth,-lb->origin().x,
		PaintEngine::LegendBoxHeight,-lb->origin().y);


  if(lb->drawFrame()){
    pe->save();
    pe->setPen(lb->rim()->color(),lb->rim()->style(),lb->rim()->width());
    pe->drawRectFromValue(0.,PaintEngine::RelativeX,
			  0.,PaintEngine::RelativeY,
			  PaintEngine::LegendBoxWidth,1.,
			  PaintEngine::LegendBoxHeight,1.,
			  lb->fill()->color(),lb->fill()->fill()
			  ,PaintEngine::ULWH);
    pe->restore();
  }  


  /* draw the legend strings now */

  pe->setPen(lb->textStyle()->color(),Xfp::LSTSolid,1);   

  pe->translate(PaintEngine::LegendTextMaxHeight,lb->border(),
		PaintEngine::LegendTextMaxHeight,lb->border());

  

  for(ig=0;ig<pgroups->size();ig++){ 
    group=pgroups->at(ig);
    if( group->active()){
      if( group->useAttribs()){
	
	if( group->attribs()->useLegend()){      
	  
	  legDrawEntry(pe,group->attribs(),legsymwidth,adjusttosymbols,
		       group->attribs()->legend(),lb,true);
	  
	  
	}
	const WeightSelections *wss=group->weightSelections();
	if(wss && wss->displayWeightLegend()){
	  for(long iw=0;iw<wss->size();iw++){
	    const WeightSelection *ws=wss->at(iw);
	    if(ws && ws->active()){
	      const WeightLabel *wl=ws->weightLabel();
	      if(wl){
		//weightat.symbol().setSize(wss->maxWidth()*2.);
		weightat.lineStyle().setColor(ws->color());
		if(wss->style()==Xfp::WESIndividual){
		  weightat.symbolStyle().setType(ws->type());
		  weightat.symbolStyle().lineStyle().setStyle(Xfp::LSTSolid);
		  weightat.symbolStyle().lineStyle().setWidth(ws->lineWidth());
		  weightat.symbolStyle().fillStyle().setFill(ws->fill());
		}else{
		  weightat.symbolStyle().setType(Xfp::STYCircle);
		  weightat.symbolStyle().lineStyle().setStyle(Xfp::LSTNone);
		  weightat.symbolStyle().lineStyle().setWidth(1);
		  weightat.symbolStyle().fillStyle().setFill(true);
		}

		legDrawEntry(pe,&weightat,legsymwidth,adjusttosymbols,
			     wl->label(),lb,false);
	      }
	    }
	  }
	}
      }else{
	for(i=0;i<group->funcs()->size();i++){ 
	  Function *func=group->funcs()->at(i);
	  
	  if(func->active()){
	    if(func->attribs()->useLegend()){
	      
	      legDrawEntry(pe,func->attribs(),legsymwidth,adjusttosymbols,
			   func->attribs()->legend(),lb,true);
	    }      
	    WeightSelections *wss=func->weightSelections();
	    if(wss && wss->displayWeightLegend()){
	      for(long iw=0;iw<wss->size();iw++){
		const WeightSelection *ws=wss->at(iw);
		if(ws && ws->active()){
		  const WeightLabel *wl=ws->weightLabel();
		  if(wl){
		    //weightat.symbol().setSize(wss->maxWidth()*2.);
		    weightat.lineStyle().setColor(ws->color());
		    if(wss->style()==Xfp::WESIndividual){
		      weightat.symbolStyle().setType(ws->type());
		      weightat.symbolStyle().lineStyle().setStyle(Xfp::LSTSolid);
		      weightat.symbolStyle().lineStyle().setWidth(ws->lineWidth());
		      weightat.symbolStyle().fillStyle().setFill(ws->fill());
		    }else{
		      weightat.symbolStyle().setType(Xfp::STYCircle);
		      weightat.symbolStyle().lineStyle().setStyle(Xfp::LSTNone);
		      weightat.symbolStyle().lineStyle().setWidth(1);
		      weightat.symbolStyle().fillStyle().setFill(true);
		    }
		    
		    
		    legDrawEntry(pe,&weightat,legsymwidth,adjusttosymbols,
				 wl->label(),lb,false);
		  }
		}
	      }
	    }
	  }
	}
      }    
    } /* group active */
  }    

  if(!pe->isPSEngine()){
    lb->setSize(Size(pe->getValue(PaintEngine::LegendBoxWidth
				  ,PaintEngine::Absolute),
		     pe->getValue(PaintEngine::LegendBoxHeight
				  ,PaintEngine::Absolute)));
  }

  pe->restore();
  
  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plot_drag_viewbox(PaintEngine *pe,GraphObject *go)
{
  /* NSF("plot_drag_viewbox"); */

  
  if(!go) return;
  ViewBox *vb=(ViewBox*)go;
  

  
  
  pe->save();

  pe->setPen(0xff0000,Xfp::LSTSolid,1);   
  
  Point pos=vb->pos(true);
  Size  size=vb->size(true);
  pe->drawRectFromValue(pos.x,PaintEngine::RelativeX,
			pos.y,PaintEngine::RelativeY,
			size.w,PaintEngine::RelativeX,
			size.h,PaintEngine::RelativeY,
			0x0,false,PaintEngine::ULWH );

  drawCurrentMarkers(pe,true);
  
  pe->restore();


}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plot_drag_shape(PaintEngine *pe,GraphObject *go)
{
  if(!go) return;

  pe->save();
  pe->setPen(0xff0000,Xfp::LSTSolid,1);
  
  Shape *sh=(Shape *) go;
  if(sh->type()==Xfp::SHTLine){
    const LineShape *lsh=(LineShape*)sh;
      
    Point q0=lsh->startToDC(true);
    Point q1=lsh->endToDC(true);
    
    pe->drawLineFromValue(q0.x,PaintEngine::RelativeX,
			  q0.y,PaintEngine::RelativeY,
			  q1.x,PaintEngine::RelativeX,
			  q1.y,PaintEngine::RelativeY,
			  PaintEngine::ULLR);

    Size devicesize=pe->deviceSize();
    double width=0.005*devicesize.w;
    q0*=devicesize;
    q1*=devicesize;

    Size d=(q1-q0)*0.1;
    q0+=d;
    q1-=d;
    
    pe->drawRectFromValue(
			  q0.x-width,PaintEngine::Absolute,
			  q0.y-width,PaintEngine::Absolute,
			  2*width,PaintEngine::Absolute,
			  2*width,PaintEngine::Absolute,
			  0xff0000,true,PaintEngine::ULWH );

    pe->drawRectFromValue(
			  q1.x-width,PaintEngine::Absolute,
			  q1.y-width,PaintEngine::Absolute,
			  2*width,PaintEngine::Absolute,
			  2*width,PaintEngine::Absolute,
			  0xff0000,true,PaintEngine::ULWH );

      
  }else if(sh->type()==Xfp::SHTEllipse){
    EllipseShape *esh=(EllipseShape*)sh;
    pe->setDrag(true);
    pe->save();
    pe->drawEllipse(esh);
    pe->restore();
    pe->setDrag(false);
  }  
  pe->restore();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plot_viewbox(PaintEngine *pe,const ViewBox *vb,ViewBoxPlotMode  mode)
{
  NSF("plot_viewbox");
  bool fill,frame;
  
  
  if(!vb) return;
  
  if(((mode & VBPMFill) == VBPMFill)){
    pe->setClipping(true);
    pe->setClippingRectFromValue(vb->pos().x,PaintEngine::RelativeX,
				 vb->pos().y,PaintEngine::RelativeY,
				 vb->size().w,PaintEngine::RelativeX,
				 vb->size().h,PaintEngine::RelativeY,
				 PaintEngine::ULWH); 
    
  }else{
    pe->setClipping(false);
  }
  
  
  if(!vb->drawFrame()) return;


  fill=((mode & VBPMFill) == VBPMFill);
  frame=((mode & VBPMFrame) == VBPMFrame);

  pe->save();

  pe->setPen(vb->rim()->color()
	     ,(frame?vb->rim()->style():Xfp::LSTNone)
	     ,vb->rim()->width()); 
  pe->drawRectFromValue(vb->pos().x,PaintEngine::RelativeX,
			vb->pos().y,PaintEngine::RelativeY,
			vb->size().w,PaintEngine::RelativeX,
			vb->size().h,PaintEngine::RelativeY,
			vb->fill()->color(),(fill?vb->fill()->fill():0),
			PaintEngine::ULWH); 
  


  pe->restore();
  return;



}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plot_drag_world(PaintEngine *pe,GraphObject *go)
{
  /* NSF("plot_drag_world"); */

  if(!go) return;
  World *w=(World*)go;
  if(!w) return;


  pe->save();
  pe->setPen(0xff0000,Xfp::LSTSolid,1);   

  pe->drawRectFromValue(go->mousePos().x,go->mousePos().y,
			go->objectPos().x,go->objectPos().y,
			0x0,false,PaintEngine::ULLR);
  
  pe->restore();




}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotticmarks_internal(PaintEngine *pe
				  ,const Tics *tics
				  ,Xfp::XYAxisType xy
				  ,Xfp::TicSide ticside
				  ,bool show)
{
  /* NSF("plotticmarks_internal"); */
  long it;
  Xfp::TicKind mjmn;
  double pos,ticlength;
  double tol=1.e-4;

  
  if( !tics) return;

    
    
  for(it=0;it<tics->size();it++){

    const Tic *tic=tics->tic(it);

    if(tic->type()!=xy) continue;        
      

    mjmn=tic->kind();
    pos =tic->pos();
    ticlength=tic->ticLength();

    if(pos<pworld->w0(xy)-tol ||  pos>pworld->w1(xy)+tol) continue;

    if(!tics->drawTics(xy,mjmn)) continue;        
    
    bool drawTicMarks=(tic->ticSide() & ticside);
    bool drawLabel=(tic->ticLabelSide() & ticside);
   

    if( ticside == Xfp::TSNormal ){
      if(xy == Xfp::Graph_Y ){

	Point q0=pworld->wctodc(pworld->w0().x,pos);	  
	  
	if(mjmn==Xfp::TKIMajor){



	  if(drawLabel){
	    pynormsitetic=true;
	    
	    pe->save();
	    if(show){
	      pe->setPen(pticmarks->textStyle(xy)->color(),Xfp::LSTSolid,1);
	    }
	    /* translate(pe,PaintEngine::YNormalTicLabelWidth,-1. ,0,0.); */
	    
	    formText(pe
		     ,q0.x-pviewbox->size().w*pticmarks->labelOffset(xy)
		     ,q0.y,tic->label(),0.
		     ,pticmarks->textStyle(xy)->size(),Xfp::TOVCenterRight
		     ,pticmarks->textStyle(xy)->subScriptScale()
		     ,show) ;
	  
	    pe->restore();
	    
	    if(!show){
	    pe->arithValue(PaintEngine::YNormalTicLabelWidth,
			   PaintEngine::YNormalTicLabelWidth,1.,
			   PaintEngine::TextWidth,1. ,PaintEngine::Max);
	    }
	  }

	}
	if(drawTicMarks){
	  if(show){
	    pe->setPen(tic->lineStyle()->color()
		       ,tic->lineStyle()->style()
		       ,tic->lineStyle()->width());   
	    pe->drawLineRel(q0.x,q0.y
			    ,ticlength*pviewbox->size().w,0.,PaintEngine::ULWH);
	  }
	}

      }else{			/* Xfp::Graph_X */
	Point q0=pworld->wctodc(pos,pworld->w0().y);
	if(mjmn==Xfp::TKIMajor){



	  if(drawLabel){
	    
	    pxnormsitetic=true;
	    pe->save();
	    if(show){
	      pe->setPen(pticmarks->textStyle(xy)->color(),Xfp::LSTSolid,1);
	    }
	    formText(pe
		     ,q0.x,q0.y+pviewbox->size().h*pticmarks->labelOffset(xy)
		     ,tic->label(),0.
		     ,pticmarks->textStyle(xy)->size(),Xfp::TOTopCenter
		     ,pticmarks->textStyle(xy)->subScriptScale()
		     ,show) ;
	    pe->restore();
	    if(!show){
	      pe->arithValue(PaintEngine::XNormalTicLabelHeight,
			     PaintEngine::XNormalTicLabelHeight,1.,
			     PaintEngine::TextHeight,1.,PaintEngine::Max);
	    }
	  }
	    
	} /* major */
	if(drawTicMarks){
	  
	  if(show){
	    pe->setPen(tic->lineStyle()->color()
		       ,tic->lineStyle()->style()
		       ,tic->lineStyle()->width());
	    pe->drawLineRel(q0.x,q0.y
			    ,0.,-ticlength*pviewbox->size().h,PaintEngine::ULWH);
	  }
	}
      } /*  x axis */
	
    }	/* normal site */
    if( ticside == Xfp::TSOpposite ){
      if(xy == Xfp::Graph_Y ){

	Point q0=pworld->wctodc(pworld->w1().x,pos);	  

	if(mjmn==Xfp::TKIMajor){

	    
	  if(drawLabel){
	    pyopossitetic=true;   
	    pe->save();
	    pe->translate(PaintEngine::YOpositeTicLabelWidth,1.
			  ,0.,PaintEngine::Absolute);
	    if(show){
	      pe->setPen(pticmarks->textStyle(xy)->color(),Xfp::LSTSolid,1);
	    }
	    formText(pe
		     ,q0.x+pviewbox->size().w*pticmarks->labelOffset(xy)
		     ,q0.y,tic->label(),0.
		     ,pticmarks->textStyle(xy)->size(),Xfp::TOVCenterRight
		     ,pticmarks->textStyle(xy)->subScriptScale()
		     ,show) ;
	    pe->restore();
	    
	    if(!show){
	      pe->arithValue(PaintEngine::YOpositeTicLabelWidth,
			     PaintEngine::YOpositeTicLabelWidth,1.,
			     PaintEngine::TextWidth,1.,PaintEngine::Max);
	    }
	  }
	} /* major */
	if(drawTicMarks){
	  
	  if(show){
	    pe->setPen(tic->lineStyle()->color()
		       ,tic->lineStyle()->style()
		       ,tic->lineStyle()->width());
	    pe->drawLineRel(q0.x,q0.y
			    ,-ticlength*pviewbox->size().w,0.,PaintEngine::ULWH);
	  }
	}

      }else{
	  
	Point q0=pworld->wctodc(pos,pworld->w1().y);	  

	  
	if(mjmn==Xfp::TKIMajor){


	  if(drawLabel){
	    
	    pxopossitetic=true;
	    pe->save();
	    if(show){
	      pe->setPen(pticmarks->textStyle(xy)->color(),Xfp::LSTSolid,1);
	    }
	    formText(pe
		     ,q0.x,q0.y-pviewbox->size().h*pticmarks->labelOffset(xy)
		     ,tic->label(),0.
		     ,pticmarks->textStyle(xy)->size(),Xfp::TOBottomCenter
		     ,pticmarks->textStyle(xy)->subScriptScale()
		     ,show) ;
	    pe->restore();
	    
	    if(!show){
	      pe->arithValue(PaintEngine::XOpositeTicLabelHeight,
			     PaintEngine::XOpositeTicLabelHeight,1.,
			     PaintEngine::TextHeight ,1.,PaintEngine::Max);
	    }
	  }
	    
	} /* major */
	if(drawTicMarks){
	  if(show){
	    pe->setPen(tic->lineStyle()->color()
		       ,tic->lineStyle()->style()
		       ,tic->lineStyle()->width());
	    pe->drawLineRel(q0.x,q0.y
			    ,0.,ticlength*pviewbox->size().h,PaintEngine::ULWH);
	  }
	}
      } /* graph_x */
    }	/* oppisite site */
      
  } /*  for all tics */
  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotticsandmarks_internal(PaintEngine *pe
				      ,Xfp::XYAxisType xy,Xfp::TicSide ticside
				      ,bool show)
{
  /* NSF("ticmetrics"); */
  long ig;
  FunctionGroup *group;
  
  
  if(!show){
    if(xy==Xfp::Graph_Y && ticside==Xfp::TSNormal){
      pe->defineValue(PaintEngine::YNormalTicLabelWidth,0.
		  ,PaintEngine::Absolute);
    }
    if(xy==Xfp::Graph_X && ticside==Xfp::TSNormal){
      pe->defineValue(PaintEngine::XNormalTicLabelHeight,0.
		  ,PaintEngine::Absolute);
    }
    if(xy==Xfp::Graph_Y && ticside==Xfp::TSOpposite){
      pe->defineValue(PaintEngine::YOpositeTicLabelWidth,0.
		  ,PaintEngine::Absolute);
    }
    if(xy==Xfp::Graph_Y && ticside==Xfp::TSOpposite){
      pe->defineValue(PaintEngine::XOpositeTicLabelHeight,0.
		  ,PaintEngine::Absolute);
    }
  }else{    
    pe->setPen(0x00,Xfp::LSTSolid,1);
  }
  


  

  plotticmarks_internal(pe,pticmarks->tics(),xy,ticside,show);

  plotticmarks_internal(pe,pticmarks->userTics(),xy,ticside,show);

  for(ig=0;ig<pgroups->size();ig++){ 
    group=pgroups->at(ig);
    if(group){
      plotticmarks_internal(pe,group->userTics(),xy,ticside,show);
    }
    
  }


  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotticsandmarks(PaintEngine *pe)
{
  /* NSF("plotticsandmarks_graph"); */
  



  plotticsandmarks_internal(pe
			    ,Xfp::Graph_Y ,Xfp::TSNormal ,FORMATTEXT_CALC_METRIC );

  plotticsandmarks_internal(pe
			    ,Xfp::Graph_Y ,Xfp::TSOpposite ,FORMATTEXT_CALC_METRIC );

  plotticsandmarks_internal(pe
			    ,Xfp::Graph_X ,Xfp::TSNormal ,FORMATTEXT_CALC_METRIC );

  plotticsandmarks_internal(pe
			,Xfp::Graph_X ,Xfp::TSOpposite ,FORMATTEXT_CALC_METRIC );


  plotticsandmarks_internal(pe
			    ,Xfp::Graph_Y ,Xfp::TSNormal ,FORMATTEXT_SHOW );


  plotticsandmarks_internal(pe
			    ,Xfp::Graph_Y ,Xfp::TSOpposite ,FORMATTEXT_SHOW );
  

  plotticsandmarks_internal(pe
			    ,Xfp::Graph_X ,Xfp::TSNormal ,FORMATTEXT_SHOW );

  plotticsandmarks_internal(pe
			    ,Xfp::Graph_X ,Xfp::TSOpposite ,FORMATTEXT_SHOW );

  return;
  
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::produceGridFunccoordinates(ValueArray<double> &xx,
				       const double *data,long Ndata,
				       long &imin,long &imax,
				       Xfp::XYAxisType xy)
{
  //prepare the arrays, which contain the rectangle coordinates
  ViewBox *vb=pviewbox;
  long N=Ndata;
  double tol=1.e-12;

  xx.clear();
  imax=0;
  imin=N-1;

  double vmin,vmax,sign;
  if(xy==Xfp::Graph_X){
    sign=1;
    vmin=vb->pos().x;
    vmax=vb->pos().x+vb->size().w;
  }else{
    sign=-1;
    vmax=-vb->pos().y;
    vmin=-vb->pos().y-vb->size().h;
  }
  
  
  
  double qxm=sign*pworld->wctodc(data[0],xy);
  double xi=qxm;
  
  bool start=false;
  for(long ix=0;ix<N;ix++){
    double xip;
    if(ix<N-1)
      xip=sign*pworld->wctodc(data[ix+1],xy);
    else
      xip=xi;
    double qxp=(xi+xip)*0.5;
    
    if(qxp<vmin-tol){
      xi=xip;
      qxm=qxp;
      continue;
    }
    if(qxm>vmax+tol){break;}
    
    imin=MyMath::min<long>(imin,ix);
    imax=MyMath::max<long>(imax,ix);
    
    if(!start){
      start=true;
      xx.append(sign*MyMath::max<double>(vmin,qxm));
    }
    xx.append(sign*MyMath::min<double>(vmax,qxp));
    
    xi=xip;
    qxm=qxp;
  }
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotdensfunc(Function *func
			 ,const FunctionAttribs *attrib
			 ,PaintEngine *pe)
{
  /* NSF("plotdensfunc_graph"); */

  //  myDebug() << "hier";
  //  func->print();
  if( func->dataType() != Function::GRIDX_Y_NZ) return;

  bool usepixmap=false;
  //usepixmap=true;

  
  if(!func->gridIsEquidistant(1.0e-6))usepixmap=false;
  if(paxes->axisScaling(Xfp::Graph_X)!=Xfp::GASLinear||
     paxes->axisScaling(Xfp::Graph_Y)!=Xfp::GASLinear)usepixmap=false;

  long ic=0;


  func->interpolateGridFunction(ic,attrib->interpolationDepth(),
 				attrib->useSecondDerivative());
  //todo: not an attriute
  
  const double *fx=func->xdatainterpolated().constDataRef();
  const double *fy=func->ydatainterpolated().constDataRef();
  const double *fz=func->datainterpolated(ic);
  long Nx=func->xdatainterpolated().size();
  long Ny=func->ydatainterpolated().size();
  if(!func) return;
  if(!func->components()) return;
  if(!func->length()) return;
  if((!Nx) || (!Ny)) return;
  
  
  double zmax=fz[0];
  double zmin=zmax;
  for(long i=0;i<Nx*Ny;i++){
    zmax=MyMath::max<double>(fz[i],zmax);
    zmin=MyMath::min<double>(fz[i],zmin);
  }
  double znull=0.;
  zmax*=attrib->zmax();
  zmin*=0.1;
  
  const double black[]={0.0,0.0,0.0};
  const double white[]={1.0,1.0,1.0};
  const double red[]={1.0,0.0,0.0};

  const double *colmax=black;
  const double *colnull=white;
  const double *colmin=red;


  long zpower=attrib->zPower();
    
  pe->setPen(attrib->lineStyle().color(),Xfp::LSTNone,1);

  ValueArray<double> xx;
  ValueArray<double> yy;
  long ixmax,ixmin,iymax,iymin;

  produceGridFunccoordinates(xx,fx,Nx,ixmin,ixmax,Xfp::Graph_X);
  produceGridFunccoordinates(yy,fy,Ny,iymin,iymax,Xfp::Graph_Y);
  
  
  PixMap *pixmap=0;
  if(usepixmap) pixmap=pe->createPixMap(ixmax-ixmin+1,iymax-iymin+1);
  

  double dzmax=zmax-znull;
  double dzmin=zmin-znull;
  double dcolmax[3];
  double dcolmin[3];
  dcolmax[0]=colmax[0]-colnull[0];
  dcolmax[1]=colmax[1]-colnull[1];
  dcolmax[2]=colmax[2]-colnull[2];
  dcolmin[0]=colmin[0]-colnull[0];
  dcolmin[1]=colmin[1]-colnull[1];
  dcolmin[2]=colmin[2]-colnull[2];
  double x;
  for(long ix=ixmin;ix<=ixmax;ix++){
    for(long iy=iymin;iy<=iymax;iy++){
      
      double z=fz[ix*Ny+iy]-znull;

      double color[3];
      /* TODO protect against division by zero */
      for(int j=0;j<3;j++){
	if(z>=0.){
	  x=MyMath::min<double>(MyMath::max<double>(z/dzmax,0.),1.);
	  color[j]=pow(x,zpower)*dcolmax[j]+colnull[j];
	}else{
	  color[j]=MyMath::max<double>(MyMath::min<double>(z/dzmin,0.),1.)*dcolmin[j]+colnull[j];
	}
      }
      if(!usepixmap){
	pe->setPen(mToRgb(color),Xfp::LSTSolid,1);
	pe->drawRectFromValue(xx[ix-ixmin],PaintEngine::RelativeX,
			      yy[iy-iymin],PaintEngine::RelativeY,
			      xx[ix-ixmin+1],PaintEngine::RelativeX,
			      yy[iy-iymin+1],PaintEngine::RelativeY,
			      mToRgb(color),true,PaintEngine::ULLR);
      }else{
	//todo: make sure it works for non equidistant meshs too
	pixmap->setPixel(ix-ixmin,iymax-iy,mToRgb(color));
      }
      
    }

    
  }
  if(usepixmap){
    ValueArray<double> fwx(Nx,0.);
    for(long i=0;i<Nx;i++)
      fwx[i]=pworld->xwctoxdc(fx[i]);
    //todo: protect against Nx<2
    //todo: find this index out of bound
    double hx=(pworld->xwctoxdc(fx[1])-pworld->xwctoxdc(fx[0]))*0.5;
    double fxmin=pworld->xwctoxdc(fx[ixmin]);
    double fxmax=pworld->xwctoxdc(fx[ixmax]);
    double x0=fxmin-hx;
    double x1=fxmax+hx;
    double hy=-(pworld->ywctoydc(fy[1])-pworld->ywctoydc(fy[0]))*0.5;
    double fymin=pworld->ywctoydc(fy[iymin]);
    double fymax=pworld->ywctoydc(fy[iymax]);
    double y0=fymin+hy;
    double y1=fymax-hy;

    pe->drawPixMap(pixmap,
		   x0,PaintEngine::RelativeX,
		   y0,PaintEngine::RelativeY,
		   x1,PaintEngine::RelativeX,
		   y1,PaintEngine::RelativeY,PaintEngine::ULLR);
    delete pixmap;
  }


  

  
  return;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
#define RECT_INSIDE  0
#define RECT_OUTSIDE 1

static long cliprect(ViewBox *vb,Point &ul,Point &lr)
{
  
  double vxl=vb->pos().x;
  double vyu=vb->pos().y;
  double vxr=vb->pos().x+vb->size().w;
  double vyl=vb->pos().y+vb->size().h;
  //todo: make part of rect
  if(lr.x<vxl || ul.x>vxr)return RECT_OUTSIDE;
  if(lr.y<vyu || ul.y>vyl)return RECT_OUTSIDE;

  
  ul.x=MyMath::max<double>(ul.x,vxl);
  lr.x=MyMath::min<double>(lr.x,vxr);

  ul.y=MyMath::max<double>(ul.y,vyu);
  lr.y=MyMath::min<double>(lr.y,vyl);


  return RECT_INSIDE;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

static long clipline(ViewBox *vb,const Point &q0,const Point &q1
		     ,Point  &p0,Point &p1)
{
  /* NSF("clipline"); */
  long i,ret;
  double r[4],s[4],tol,lmin,lmax;

  tol=1.e-12;
  /* D=p2-p1 */
  Point d=q1-q0;


  /* n1<- x0y0,x0y1 */
  lmax=1.e30;
  lmin=-1.e30;

  Point ul=q0-vb->pos();
  Point lr=vb->pos()+vb->size()-q0;
  /* left boundary */
  r[0]=ul.x;
  s[0]=-d.x;
  /* right boundary */
  r[1]=lr.x;
  s[1]=d.x;
  /* upper boundary */
  r[2]=ul.y;
  s[2]=-d.y;
  /* lower boundary */
  r[3]=lr.y;
  s[3]=d.y;

  i=-1;
  while(++i<4){
    /* CPRD(s[i]); */
    /* CPRD(r[i]); */
    if(fabs(s[i])<tol && r[i]<0.-tol){
      /* CPR("CLIP_SEG_IS_OUTSIDE"); */
      return CLIP_SEG_IS_OUTSIDE;
    }

    if(s[i]>0.){
      lmax=MyMath::min<double>(lmax,r[i]/s[i]);
    }
    if(s[i]<0.){
      lmin=MyMath::max<double>(lmin,r[i]/s[i]);
    }
  }

  lmax=MyMath::min<double>(lmax,1.);
  lmin=MyMath::max<double>(lmin,0.);

  if(lmax>lmin){
    p1=q0+lmax*d;
    p0=q0+lmin*d;
  }else{
    return CLIP_SEG_IS_OUTSIDE;
  }
  ret=CLIP_PART_OF_SEG_IS_INSIDE;
  /* CPRD(lmax); */
  /* CPRD(lmin); */
  if(lmax<1.-tol){
    ret|= CLIP_SECOND_POINT_IS_OUTSIDE;
  }

  if(lmin>tol){
    ret|= CLIP_FIRST_POINT_IS_OUTSIDE;
  }
  return ret;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotxyfunc(Function *func
		       ,const FunctionAttribs *attrib
		       ,PaintEngine *pe,bool nowSymbols)
{
  //NSF("plotxyfunc_graph");
  long start,narray,flush,ia,ret,lastadded,j;
  

  if(!func) return;
  if(!func->length()) return;

  if( func->dataType() != Function::XY && 
      func->dataType() != Function::XYNW &&
      func->dataType() != Function::XYMatel )
    return;

  if(func->components()<2) return;
  /* CPRL(func->narray); */

  if(!func->isInRange(pworld->w0(),pworld->w1())) return;
  //myDebug()<<"in range";

  if(!nowSymbols && attrib->lineStyle().style()==Xfp::LSTNone) return;
  if(nowSymbols && attrib->symbolStyle().type()==Xfp::STYNone) return;

  ViewBox *vb=pviewbox;


  narray=func->length();

  /* init pointarray */
  ValueArray<Point> lpoints;
  ValueArray<Point> spoints;

  if(nowSymbols) 
    pe->drawComment("start function: "+func->comment()+" symbols");
  else
    pe->drawComment("start function: "+func->comment()+" lines");

  /* draw clipped line segments  */
  {
    
    start=0;
    lastadded=-1;
    ia=0;
    

    
    Point q0=pworld->wctodc(Point(func->data(0,ia),func->data(1,ia)));
    
    
    ia++;
    while(ia<(narray==1?2:narray)){
      /* CPRL(ia);       */
      j=ia;
      if(narray==1) j--;
      Point q1=pworld->wctodc(Point(func->data(0,j),func->data(1,j)));
      
      flush=0;
      Point p0,p1;
      ret=clipline(vb,q0,q1,p0,p1);
      /* CPRL(ret); */
      if(!start){
	if( (ret & CLIP_PART_OF_SEG_IS_INSIDE)==CLIP_PART_OF_SEG_IS_INSIDE){
	  
	  start=1;
	  
	  
	  
	  lpoints.append(p0);
	  lpoints.append(p1);
	  
	  
	  if((ret & CLIP_SECOND_POINT_IS_OUTSIDE)
	     ==CLIP_SECOND_POINT_IS_OUTSIDE){
	    start=0;
	    flush=1;
	  }
	}
      }else{
	
	if( (ret & CLIP_PART_OF_SEG_IS_INSIDE)==CLIP_PART_OF_SEG_IS_INSIDE){
	  
	  
	  lpoints.append(p1);
	  
	  if((ret & CLIP_SECOND_POINT_IS_OUTSIDE)
	     ==CLIP_SECOND_POINT_IS_OUTSIDE){
	    start=0;
	    flush=1;
	  }
	}else{
	  if(lpoints.size()>0){
	    start=0;
	    flush=1;
	  }
	}
      }
      if(flush){
	if(narray!=1 && !nowSymbols){
	  if(attrib->lineStyle().style()!=Xfp::LSTSolid &&
	     attrib->lineStyle().style()!=Xfp::LSTNone){
	    pe->setPen(
		      0xffffff
		      ,Xfp::LSTSolid
		      ,attrib->lineStyle().width()*1.2
		      );
	    pe->drawPolyLineRel(lpoints);
	  }
	  if(attrib->lineStyle().style()!=Xfp::LSTNone){
	    pe->setPen(
		       attrib->lineStyle().color()
		       ,attrib->lineStyle().style()
		       ,attrib->lineStyle().width()
		       );
	    pe->drawPolyLineRel(lpoints);
	  }
	}
	
	lpoints.clear();
	flush=0;
      }

      if((ret & CLIP_PART_OF_SEG_IS_INSIDE)==CLIP_PART_OF_SEG_IS_INSIDE){
	if((ret & CLIP_FIRST_POINT_IS_OUTSIDE)!=CLIP_FIRST_POINT_IS_OUTSIDE){
	  if(lastadded!=ia-1 && ia-1>=0){
	    /* add ia-1 */
	    
	    spoints.append(p0);
	    lastadded=ia-1;
	  }
	}
	if((ret & CLIP_SECOND_POINT_IS_OUTSIDE)!=CLIP_SECOND_POINT_IS_OUTSIDE){
	  if(lastadded!=ia){
	    
	    /* add ia */
	    
	    spoints.append(p1);
	    lastadded=ia;
	  }
	}
      }	/* part of segment is visible */
      q0=q1;
      ia++;
    } /* while */
    if(lpoints.size()>0){
      if(narray!=1 && !nowSymbols ){
	if(attrib->lineStyle().style()!=Xfp::LSTSolid &&
	   attrib->lineStyle().style()!=Xfp::LSTNone){
	  pe->setPen(
		    0xffffff
		    ,Xfp::LSTSolid
		    ,attrib->lineStyle().width()*1
		    );
	  pe->drawPolyLineRel(lpoints);
	}
	if(attrib->lineStyle().style()!=Xfp::LSTNone){
	  pe->setPen(
		     attrib->lineStyle().color()
		     ,attrib->lineStyle().style()
		     ,attrib->lineStyle().width()
		     );
	  
	  
	  pe->drawPolyLineRel(lpoints); 
	}
      }
      
      lpoints.clear();
    }

    if(nowSymbols){
      
      pe->drawPolySymbolRel(spoints,attrib);
      
      if(attrib->useLegend()){
	pe->arithValue(PaintEngine::MaxSymbolSize,
		       PaintEngine::MaxSymbolSize,1.,
		       PaintEngine::SymbolSize,1.,PaintEngine::Max);
	
      }
    }
    lpoints.clear();
    spoints.clear();
    
  } /* end line segments */
  if(nowSymbols){
    pe->drawComment("end function: "+func->comment()+" symbols");
  }else{
    pe->drawComment("end function: "+func->comment()+" lines");
  }

  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotxynwfunc(Function *func
			 ,const FunctionAttribs *attrib
			 ,const WeightSelections *wss
			 ,long iweightsel
			 ,PaintEngine *pe)
{
  //NSF("plotxynwfunc");
  long start,narray,flush,ia,ret,j;
  //long lastadded;
  //func->print();

  //myDebug()<<nsubfunc;
  if(!wss)return;
  if(!func) return;
  if(!func->length()) return;

  const WeightSelection *ws=wss->at(iweightsel);
  const WeightLabels *wls=func->weightLabels();
  if(!wls) return;
  long iweightdata=wls->indexOf(ws->weightLabel());
  if(iweightdata<0)return;
  
  
  if( func->dataType() != Function::XYNW) return;

  if(func->components()<2 || iweightdata+2>=func->components()) return;
  
  if(!func->isInRange(pworld->w0(),pworld->w1())) return;

  /* CPRL(func->narray); */

  ViewBox *vb=pviewbox;

  double min=0.2;
  double max=6.0;
  double factor=1.0;
  if(wss){
    min=wss->minWidth();
    max=wss->maxWidth();
    factor=wss->factor();
  }
  
  narray=func->length();
  
  
  QRgb cc=ws->color();

  /* draw clipped line segments  */
  {

    /* init pointarray */
    ValueArray<Point> lpoints;
    ValueArray<double> lradii;
    ValueArray<bool> lsymbol;
    //ValueArray<Point> spoints;
    
    start=0;
    //lastadded=-1;
    ia=0;
    

    

    
    Point q0=pworld->wctodc(Point(func->data(0,ia),func->data(1,ia)));
    double w0=func->data(iweightdata+2,ia)*factor;
    
    
    ia++;
    while(ia<(narray==1?2:narray)){
      /* CPRL(ia);       */
      j=ia;
      if(narray==1) j--;
      ;
      Point q1=pworld->wctodc(Point(func->data(0,j),func->data(1,j)));
      double w1=func->data(iweightdata+2,j)*factor;
      
      flush=0;
      Point p0,p1;
      ret=clipline(vb,q0,q1,p0,p1);
      /* CPRL(ret); */
      if(!start){
	if( (ret & CLIP_PART_OF_SEG_IS_INSIDE)==CLIP_PART_OF_SEG_IS_INSIDE){
	  
	  start=1;
	  
	  
	  //todo: interpolate w0,w1
	  lradii.append(w0);
	  lradii.append(w1);
	  
	  lpoints.append(p0);
	  lpoints.append(p1);
	  
	  bool s=(!((ret & CLIP_FIRST_POINT_IS_OUTSIDE)==CLIP_FIRST_POINT_IS_OUTSIDE)) &&
	    (((ia-1)%(ws->skip()+1))==0);
	  lsymbol.append(s);
	  s=(!((ret & CLIP_SECOND_POINT_IS_OUTSIDE)==CLIP_SECOND_POINT_IS_OUTSIDE)) &&
	    ((ia%(ws->skip()+1))==0);
	  lsymbol.append(s);

	  if((ret & CLIP_SECOND_POINT_IS_OUTSIDE)
	     ==CLIP_SECOND_POINT_IS_OUTSIDE){
	    start=0;
	    flush=1;
	  }
	}
      }else{
	
	if( (ret & CLIP_PART_OF_SEG_IS_INSIDE)==CLIP_PART_OF_SEG_IS_INSIDE){
	  
	  
	  lradii.append(w1);
	  lpoints.append(p1);
	  
	  if((ret & CLIP_SECOND_POINT_IS_OUTSIDE)
	     ==CLIP_SECOND_POINT_IS_OUTSIDE){
	    start=0;
	    flush=1;
	  }
	  lsymbol.append(!((ret & CLIP_SECOND_POINT_IS_OUTSIDE)
			    ==CLIP_SECOND_POINT_IS_OUTSIDE) &&
			 ((ia%(ws->skip()+1)))==0);
	}else{
	  if(lpoints.size()>0){
	    start=0;
	    flush=1;
	  }
	}
      }
      if(flush){
	if(narray!=1){
 	  //pe->setPen(cc,Xfp::LSTSolid,1);
	  pe->drawPolyLineRel(lpoints,lradii,lsymbol,cc,min,max,ws->symbolStyle()
			      ,wss->style());
	}
	
	lradii.clear();
	lpoints.clear();
	lsymbol.clear();
	flush=0;
      }

      // if((ret & CLIP_PART_OF_SEG_IS_INSIDE)==CLIP_PART_OF_SEG_IS_INSIDE){
      // 	if((ret & CLIP_FIRST_POINT_IS_OUTSIDE)!=CLIP_FIRST_POINT_IS_OUTSIDE){
      // 	  if(lastadded!=ia-1 && ia-1>=0){
      // 	    /* add ia-1 */
	    
      // 	    spoints.append(p0);
      // 	    lastadded=ia-1;
      // 	  }
      // 	}
      // 	if((ret & CLIP_SECOND_POINT_IS_OUTSIDE)!=CLIP_SECOND_POINT_IS_OUTSIDE){
      // 	  if(lastadded!=ia){
	    
      // 	    /* add ia */
	    
      // 	    spoints.append(p1);
      // 	    lastadded=ia;
      // 	  }
      // 	}
      // }
      /* part of segment is visible */
      q0=q1;
      ia++;
    } /* while */
    if(lpoints.size()>0){
      if(narray!=1){
	
	//pe->setPen(cc,Xfp::LSTSolid,1);
	pe->drawPolyLineRel(lpoints,lradii,lsymbol,cc,min,max,ws->symbolStyle()
			    ,wss->style());
      }
      
      lradii.clear();
      lpoints.clear();
      lsymbol.clear();
    }

    pe->setPen(
	      attrib->lineStyle().color()
	      ,Xfp::LSTSolid
	      ,attrib->lineStyle().width()
	      );
    
    //todo sets withone point ,(narray==1?MyMath::min<long>(1,spoints.size()):spoints.size())
    //myDebug()<<spoints;
    //pe->drawPolySymbolRel(spoints,attrib);
    
    if(attrib->useLegend()){
      pe->arithValue(PaintEngine::MaxSymbolSize,
		     PaintEngine::MaxSymbolSize,1.,
		     PaintEngine::SymbolSize,1.,PaintEngine::Max);

    }
    
    lradii.clear();
    lpoints.clear();
    lsymbol.clear();
    //spoints.clear();
    
  } /* end line segments */
  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotmatelfunc(const Function *func0,const Function *func1,
			  long icomp,const OptMatelDefinition *mateldef,
			  const FunctionAttribs */*attrib*/,
			  Stripes &stripes)
{
  //myDebug()<<"plotmatelfunc";
  //todo theta functions or not
  //todo intra band
  //todo: dont plot at each point
  if(!func0 || !func1) return;
  if(func0->length()<=0 || func0->length()!=func1->length()) return;

  if( func0->dataType() != Function::XYMatel ) return;
  if( func1->dataType() != Function::XYMatel ) return;
  
  if(func0->components()-2<=icomp || func1->components()-2<=icomp ) return;

  if(!func0->active() || ! func1->active() ) return;
  
  ViewBox *vb=pviewbox;

  //todo: multiple definitions, continous, blend colors, when overlapping

  double emin=mateldef->ergMin();
  double emax=mateldef->ergMax();
  double de=mateldef->ergDe();

  long narray=func0->length();
  
  /* draw clipped line segments  */
  {

    for(long ia=0;ia<narray-1;ia++){

      double xl=func0->data(0,ia);
      double xr=func0->data(0,ia+1);
      double yl=(func0->data(1,ia)+func0->data(1,ia+1))*0.5;
      double yu=(func1->data(1,ia)+func1->data(1,ia+1))*0.5;
      //if(xl<1.45) continue;
      //if(xr>1.65) continue;
      //      myDebug()<<"x"<<xl<<xr;
      double mat=(func0->data(icomp+2,ia)+func0->data(icomp+2,ia+1))*0.5;

      // theta function
      if((yl<0 && yu<0) || (yl>0 && yu>0))continue;

      Point ul=pworld->wctodc(xl,yu);
      Point lr=pworld->wctodc(xr,yl);

      if(cliprect(vb,ul,lr)==RECT_OUTSIDE) continue;
      
      double dy=fabs(yu-yl);
      //if(dy< emin || dy > emax)continue;


      //todo: negative matels;

      double fscreen=0.;
      if(emin<=dy && dy<=emax){
	fscreen=1.;
      }else if(dy>emax){
	fscreen=exp(-((dy-emax)/de)*((dy-emax)/de));
      }else if(dy<emin){
	fscreen=exp(-((dy-emin)/de)*((dy-emax)/de));
      }
      //      if(xl>1.99 && xl<2.2) myDebug()<<xl<<yl<<yu<<fscreen;

      double denom=1.;
      if(dy!=0.)denom=1./(dy*dy);
      double matel=fabs(mat)*denom*mateldef->matelScale()*fscreen;

      if(matel<=mateldef->matelMin()) {
	//myDebug()<<"matel"<<matel;
	continue;
      }
      matel=MyMath::max<double>(mateldef->matelMin(),MyMath::min<double>(matel,mateldef->matelMax()));


      //myDebug()<<"APPEND"<<ul<<lr<<matel;
      stripes.append(ul,lr,matel,1.0e-6);
      //      stripes.print();
      
    } /* for */


      
    
//     if(attrib->useLegend()){
//       pe->arithValue(PaintEngine::MaxSymbolSize,
// 		     PaintEngine::MaxSymbolSize,1.,
// 		     PaintEngine::SymbolSize,1.,PaintEngine::Max);
      
//     }


  } /* end line segments */

  //myDebug()<<"plotmatelfunc done";  
  return;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotstripes(const Stripes &stripes,PaintEngine *pe)
{
//   pe->drawComment("start function: "+func0->comment()+" "
// 		  +func1->comment()+" matrix elements");
  
  //myDebug()<<"stripes.size()"<<stripes.size();
  for(long i=0;i<stripes.size();i++){
    const Stripe *s=stripes.at(i);
    
    QRgb c=(QRgb)MyMath::max<double>(MyMath::min<double>((1-s->value)*255,255),0);
    QRgb rgb=(c<<8)+c+0xff0000;

    pe->setPen( rgb ,Xfp::LSTSolid ,1 );
    pe->drawRectFromValue(
			  s->ul.x,PaintEngine::RelativeX,
			  s->ul.y,PaintEngine::RelativeY,
			  s->lr.x,PaintEngine::RelativeX,
			  s->lr.y,PaintEngine::RelativeY,
			  rgb,true,PaintEngine::ULLR);
  }


//   pe->drawComment("end function: "+func0->comment()+" "
// 		  +func1->comment()+" matrix elements");
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotfuncs(PaintEngine *pe)
{
  /* NSF("plotfuncs_graph");  */
  long i,ig;
  const FunctionAttribs *at;
  const WeightSelections *wss=0;
 
  pe->defineValue(PaintEngine::MaxSymbolSize,0.,PaintEngine::Absolute);
  

  for(ig=0;ig<pgroups->size();ig++){ 
    FunctionGroup *group=pgroups->at(ig);
    if( !group->active()) continue;
    
    // count max number of weights to determine plot order
    long maxweights=0;
    maxweights=MyMath::max<long>(maxweights,group->weightSelections()->size());
    for(i=0;i<group->funcs()->size();i++){
      Function *func=group->funcs()->at(i);
      if(func->dataType()==Function::XYNW){
	maxweights=MyMath::max<long>(maxweights,func->weightSelections()->size());
      }
    }
    // matel plots
    long ncomp;
    if(group->allFuncsAre(Function::XYMatel) &&
       group->allFuncsHaveSaveNumberOfComponents(ncomp) &&
       ncomp>2
       ){
      Stripes stripes;
      for(i=0;i<group->funcs()->size();i++){
	//myDebug()<<"func"<<i<<"of"<<group->funcs()->size();
	Function *func0=group->funcs()->at(i);
	if(func0->dataType()!=Function::XYMatel || 
	   (!func0->active())) continue;
	
	OptMatelDefinition *mateldef=group->optMatelDefinition();
	at=func0->attribs();
	if(group->useAttribs()){
	  at=group->attribs();
	}
	//todo: if we delete or shift funcs, we need ids to identify
	//  funcs even ids, which tell me that this set of funcs forms
	// a valid set
	for(long icomp=i;icomp<ncomp-2;icomp++){
	  if(icomp>=group->funcs()->size())continue;
	  Function *func1=group->funcs()->at(icomp); //todo use ids?
	  if(!func1->active()) continue;
	  plotmatelfunc(func0,func1,icomp,mateldef,at,stripes);
	}
	
      }
      plotstripes(stripes,pe);
    }    
    //todo: avoid double ploting for dashed lines ???
    // first plot xy-funcs and xynw-funcs's underlying 
    for(i=0;i<group->funcs()->size();i++){
      Function *func=group->funcs()->at(i);
      at=func->attribs();
      if(group->useAttribs()){
	at=group->attribs();
      }
      if( func->active()){
	if(func->dataType()==Function::XY||
	   func->dataType()==Function::XYNW||
	   func->dataType()==Function::XYMatel){
	  plotxyfunc(func,at,pe,false); 
	}
      }
    }
    for(i=0;i<group->funcs()->size();i++){
      Function *func=group->funcs()->at(i);
      at=func->attribs();
      if(group->useAttribs()){
	at=group->attribs();
      }
      if( func->active()){
	if(func->dataType()==Function::XY||
	   func->dataType()==Function::XYMatel||
	   func->dataType()==Function::XYNW){
	  plotxyfunc(func,at,pe,true); 
	}
      }
    }
    // now for each weight plot the weights of all funcs
    // so that weight i+1 lays on top of weight i (if existing and active)
    for(long iweightsel=0;iweightsel<maxweights;iweightsel++){
      
      for(i=0;i<group->funcs()->size();i++){
	Function *func=group->funcs()->at(i);
	if(func->dataType()!=Function::XYNW || (!func->active())) continue;
	
	at=func->attribs();
	wss=func->weightSelections();
	if(group->useAttribs()){
	  at=group->attribs();
	  wss=group->weightSelections();
	}
	if( (!wss) ||
	    (iweightsel>=wss->size()) ||
	    (!wss->at(iweightsel)->active())) continue;
	plotxynwfunc(func,at,wss,iweightsel,pe); 
      }
    }
    // last plot densfuncs
    for(i=0;i<group->funcs()->size();i++){
      Function *func=group->funcs()->at(i);
      at=func->attribs();
      if(group->useAttribs()){
	at=group->attribs();
      }
      if( func->active() && 
	  func->dataType()==Function::GRIDX_Y_NZ){
	plotdensfunc(func,at,pe); 
      }
    }
    
  }// for all groups
  
  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plottextboxes(PaintEngine *pe)
{
  /* NSF("plottextboxes_graph"); */
  long it;


  for(it=0;it<ptextboxes->size();it++){
    TextBox *tb=ptextboxes->at(it);
    if(tb->type()!=TextBox::TBTGeneral) continue;

    pe->save();
    
    plotTextBox(pe,tb);
    pe->restore();
  }

  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotshapes(PaintEngine *pe)
{

  for(long is=0;is<pshapes->size();is++){
    const Shape *sh=pshapes->at(is);

    plotShape(pe,sh);
  }

  
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotShape(PaintEngine *pe,const Shape *sh)
{

  if(!sh)return;
  if(!sh->active())return;
  pe->save();
  if(sh->type()==Xfp::SHTLine){
    const LineShape *lsh=(LineShape*)sh;
    
    
    Point q0=lsh->startToDC();
    Point q1=lsh->endToDC();
    
    pe->drawArrow(q0.x,q0.y,q1.x,q1.y,lsh);
    
    
  }else if(sh->type()==Xfp::SHTEllipse){
    pe->drawEllipse((EllipseShape*)sh);
  }

  pe->restore();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
//GraphObject::GraphObjectRegions 
void
Graph::getObjectsAt(PtrArray<GraphObject> &objects,const DragDescriptor &dd,
		    bool update)
{
  // objects need to be cleaned before
  
  if(pworld->insideRegion(dd,update) !=GraphObject::GORNone){
    objects.append(pworld);
  }

  bool inviewbox=false;
  if(pviewbox->insideRegion(dd,update)!=GraphObject::GORNone){
    objects.append(pviewbox);
    inviewbox=true;
  }

  if(plegendbox->insideRegion(dd,update)!=GraphObject::GORNone){
    objects.append(plegendbox);
  }

  for(int i=0;i<pshapes->size();i++){
    if(pshapes->at(i)->insideRegion(dd,update)!=GraphObject::GORNone){
      objects.append(pshapes->at(i));
    }
  }

  for(int i=0;i<ptextboxes->size();i++){
    if(ptextboxes->at(i)->insideRegion(dd,update)!=GraphObject::GORNone){
      objects.append(ptextboxes->at(i));
    }
  }


  Point wp=pworld->dctowcs(dd.mousepoint/dd.devicesize);
  Size dwp=abs(pworld->dctowcs(Size(dd.sensitivityAroundPoint,
				    dd.sensitivityAroundPoint)
			       *(1.5/dd.devicesize)));

  
  if(!inviewbox) return;
  ValueArray<Pair<double,Function*> > list;
  for(long ig=0;ig<pgroups->size();ig++){ 
    FunctionGroup *group=pgroups->at(ig);
    if( !group->active()) continue;
    for(long i=0;i<group->funcs()->size();i++){
      Function *func=group->funcs()->at(i);
      if(!func->isInRange(pworld->w0(),pworld->w1())) continue;
      double dist;
      if(func->isCloseToPoint(wp,dwp,dist,paxes)){
	list.append(Pair<double,Function*>(dist,func));
      }
    }
  }
  list.qSort();
  FunctionGroup *lastgroup=0;
  for(int i=list.size()-1;i>=0;i--){
    //for(int i=0;i<list.size();i++){
    FunctionGroup *group=list.at(i).v2->group();
    if(lastgroup!=group){
      objects.append(group);
      lastgroup=group;
    }
    objects.append(list.at(i).v2);
    //myDebug()<<list.at(i);
  }
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::plotDragObject(PaintEngine *pe,GraphObject *go)
{
  if(!go) return;

  if(go->objectType() == GraphObject::GOTWorld){
    plot_drag_world(pe,go);
  }else if(go->objectType() == GraphObject::GOTViewBox){
    plot_drag_viewbox(pe,go);
  }else if(go->objectType() == GraphObject::GOTLegendBox){
    plot_drag_legendbox(pe,go);
  }else if(go->objectType() == GraphObject::GOTTextBox){
    plot_drag_textbox(pe,go);
  }else if(go->objectType() == GraphObject::GOTShape){
    plot_drag_shape(pe,go);
  }

  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Graph::drawCurrentMarkers(PaintEngine *pe,bool dragging)
{
  ViewBox *vb=pviewbox;
  pe->save();

  QRgb c=0x0;
  if(dragging) c=0xff0000;

  pe->setPen(c,Xfp::LSTSolid,1);



  Size devicesize=pe->deviceSize();
  double width=0.005*devicesize.w;
  Point q0=vb->pos(dragging)*devicesize;
  Point q1=(vb->pos(dragging)+vb->size(dragging))*devicesize;

  
  pe->drawRectFromValue(
			q0.x-width,PaintEngine::Absolute,
			q0.y-width,PaintEngine::Absolute,
			2*width,PaintEngine::Absolute,
			2*width,PaintEngine::Absolute,
			c,true,PaintEngine::ULWH );

  pe->drawRectFromValue(
			q1.x-width,PaintEngine::Absolute,
			q0.y-width,PaintEngine::Absolute,
			2*width,PaintEngine::Absolute,
			2*width,PaintEngine::Absolute,
			c,true,PaintEngine::ULWH );

  pe->drawRectFromValue(
			q1.x-width,PaintEngine::Absolute,
			q1.y-width,PaintEngine::Absolute,
			2*width,PaintEngine::Absolute,
			2*width,PaintEngine::Absolute,
			c,true,PaintEngine::ULWH );

  pe->drawRectFromValue(
			q0.x-width,PaintEngine::Absolute,
			q1.y-width,PaintEngine::Absolute,
			2*width,PaintEngine::Absolute,
			2*width,PaintEngine::Absolute,
			c,true,PaintEngine::ULWH );

  pe->restore();

}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
// long Graph::plotaxes(PaintEngine *pe)
// {
//   /* NSF("plotaxes_graph"); */
//   double qx0,qy0;
//   long ia;




//   pe->setpen(device,0x00,Xfp::LSTSolid,1);
  
  

//   for(ia=0;ia<paxes.size();ia++){
//     if(paxes[ia]->active()){      
//       if(paxes[ia]->type()==Xfp::Graph_X){
// 	wctodc(viewbox ,0.,paxes[ia]->pos() ,&qx0,&qy0);
// 	pe->drawlinerel(device,vb->pos().x,qy0 
// 		       ,vb->size().w,0.,PaintEngine::ULWH);
//       }else if(paxes[ia]->type()==Xfp::Graph_Y){
// 	wctodc(viewbox ,paxes[ia]->pos() , 0. ,&qx0,&qy0);
// 	pe->drawlinerel(device,qx0,vb->pos().y 
// 		       ,0.,vb->size().h,PaintEngine::ULWH);

//       }
//     }
//   }
  


//   return 1;
// }
























/*EOF*/ 
