/*************************************************************************
 * @(#)LineShape.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   22 Nov 2011
 *************************************************************************
 */

#include "GraphObject.hh" 
#include "Graph.hh" 
#include "Attributes.hh" 
#include "LineShape.hh"
#include "World.hh"
#include "ViewBox.hh"
#include "MyDebug.hh"

/* ***********************************************************************
 *
 * ***********************************************************************
 */
LineShape::LineShape(Graph *parent,long index)
  :Shape(parent,index),
   pposstart(Point(0.2,0.)),
   pposend(Point(0.8,1.)),
   plinestyle(new LineStyle()),
   pfill(new FillStyle()),
   pextrafillcolor(false),
   parrowposition(Xfp::ARPEnd),
   pcapposition(Xfp::ARPNone),
   parrowstyle(Xfp::ARSTClosed),
   parrowheadsize(16),
   psharpness(2.)
{
  ptype=Xfp::SHTLine;
  pname="";
  pfill->setColor(0x0);
  pfill->setFill(true);
  setChanged(false);
  
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
LineShape::~LineShape()
{
  if(plinestyle) delete plinestyle; plinestyle=0;
  if(pfill) delete pfill; pfill=0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void LineShape::reset()
{
  pextrafillcolor=false,
  pfill->setColor(0x0);
  pfill->setFill(true);
  pposstart=Point(0.2,0.);
  pposend=Point(0.8,1.);
  plinestyle->reset();
  pfill->reset();
  parrowposition=Xfp::ARPEnd;
  parrowposition=Xfp::ARPNone;
  parrowstyle=Xfp::ARSTClosed;
  parrowheadsize=16;
  psharpness=2.;
  pcoordsystem=Xfp::CSYView;
  setChanged(false);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool LineShape::changed()const
{
  return pchanged | pfill->changed() | plinestyle->changed();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void LineShape::setChanged(bool c)
{
  plinestyle->setChanged(c);
  pfill->setChanged(c);
  pchanged=c;

}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
GraphObject::GraphObjectRegion 
LineShape::insideRegion(const DragDescriptor &dd,bool update)
{
  //  myDebug() << "LineShape::insideRegion";
  
  if(!pactive) return GraphObject::GORNone;

  Point s=startToDC()*dd.devicesize;
  Point e=endToDC()*dd.devicesize;
  Point pm=dd.mousepoint-s;

  Point c=(e-s)*0.5;
  double lc=norm(c);
  Point nc=c/lc;
  Point ort=nc.ortho();
  double port=(pm-c)*ort;
  double pnorm=(pm-c)*nc;
  bool longside=(fabs(port)<dd.sensitivityAroundPoint);
  bool inrod=(fabs(pnorm)<lc+dd.sensitivityAroundPoint);

  lc*=0.8;
  bool shortsideend=(( pnorm )<lc+dd.sensitivityAroundPoint);
  bool shortsideend2=(( -pnorm )<-lc+dd.sensitivityAroundPoint);
  bool shortsidestart=(( -pnorm )<lc+dd.sensitivityAroundPoint);
  bool shortsidestart2=(( pnorm )<-lc+dd.sensitivityAroundPoint);


  if( longside && shortsidestart2 && shortsidestart ){
    if(update){
      pupdaterequest=GOCSize;
      pupdaterequestnumber=0;
    }
    return GORUL;
  }else if( longside && shortsideend && shortsideend2 ){
    if(update){
      pupdaterequest=GOCSize;
      pupdaterequestnumber=1;
    }
    return GORLR;
  }else if( longside && inrod ){ 
    if(update){
      pupdaterequestnumber=0;
      pupdaterequest=GOCPosition;
    }
    return GORBody;
  }
  
  
  if(update){
    pupdaterequest=GOCNone;
  }
  return GraphObject::GORNone;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void LineShape::initDrag(const DragDescriptor &dd)
{
  GraphObjectRegion region=insideRegion(dd,true);

  if(region==GORNone)return;
  pmousepos=dd.mousepoint;

  pobjectpos=0;
  pdeltasize=0;
  
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void LineShape::updateDrag(const DragDescriptor &dd)
{
  if(pupdaterequest == GOCPosition){
    pobjectpos+=(dd.mousepoint-pmousepos)/dd.devicesize;
    pdeltasize+=(dd.mousepoint-pmousepos)/dd.devicesize;
  }else if(pupdaterequest == GOCSize){
    if(pupdaterequestnumber==0){
      pobjectpos+=(dd.mousepoint-pmousepos)/dd.devicesize;
    }else{
      pdeltasize+=(dd.mousepoint-pmousepos)/dd.devicesize;
    }
  }

  pmousepos=dd.mousepoint;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void LineShape::setFromDrag(DragDescriptor &/*dd*/)
{
  
  pposstart=dcToPos(startToDC(true),pcoordsystem);
  pposend=dcToPos(endToDC(true),pcoordsystem);
  pobjectpos=0;
  pdeltasize=0;
  

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String LineShape::asCommands(const String &tab)const
{
  String s;
  String name=Sprintf("%s%ld",Xfp::shapeTypeGuiName(ptype),puniqueindex+1);
  s+=Sprintf("%s%s %s\n",tab(),name(),Xfp::OnOff(active()));
  name=tab+"  "+name;
  s+=Sprintf("%s name \"%s\"\n",name(),pname());
  s+=lineStyle()->asCommands(name+" line");
  s+=fill()->asCommands(name+" arrow fill",pextrafillcolor);
  s+=Sprintf(
	     "%s cap position %s\n"
	     "%s arrow position %s\n"
	     "%s arrow style %s\n"
	     "%s arrow size %g\n"
	     "%s arrow sharpness %g\n"
	     "%s coordinate system %s\n"
	     "%s start  %g,%g\n"
	     "%s end  %g,%g\n",
	     name(),Xfp::arrowPositionGuiName(pcapposition),
	     name(),Xfp::arrowPositionGuiName(parrowposition),
	     name(),Xfp::arrowStyle(parrowstyle).guiname,
	     name(),parrowheadsize,
	     name(),psharpness,
	     name(),Xfp::coordinateSystemGuiName(pcoordsystem),
	     name(),pposstart.x,pposstart.y,
	     name(),pposend.x,pposend.y
	     );
  return s;
}


/*EOF*/ 
