/* @(#)Position.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _POSITION_H
#define _POSITION_H 1


#include "ValueArray.hh"
#include "Point.hh" 
#include "FString.hh" 
#include "Graph.hh" 
#include "ViewBox.hh" 
#include "Xfp.hh"
#include "MyMath.hh" 

class Graph;

/* ***********************************************************************
 *
 * ***********************************************************************
 */

class Position{
 public:
  enum OffsetType{OSTNone=0x0,OSTPos=0x1,OSTNeg=0x2,OSTX=0x4,OSTY=0x8
		  ,OSTPosY=OSTPos|OSTY
		  ,OSTNegY=OSTNeg|OSTY
		  ,OSTPosX=OSTPos|OSTX
		  ,OSTNegX=OSTNeg|OSTX
  };
  enum OffsetAddTypes{
    None= 0,
    XOposSiteTicLabelHeight= 0x00000001,
    XOposSiteTicLabelOffset= 0x00000002,
    XNormSiteTicLabelHeight= 0x00000003,
    XNormSiteTicLabelOffset= 0x00000004,
    
    YOposSiteTicLabelWidth = 0x00000005,
    YOposSiteTicLabelOffset= 0x00000006,
    YNormSiteTicLabelWidth = 0x00000007,
    YNormSiteTicLabelOffset= 0x00000008,
    
    SubTitleOffset         = 0x00000009,
    SubTitleHeight         = 0x0000000A,
    
    OppXaLabelOffset       = 0x0000000B,
    OppXaLabelHeight       = 0x0000000C
  };

  Position():pchanged(false),pangle(0),prestriction(false),poffset(0)
    ,poffset_type(OSTNone),pcoordsystem(Xfp::CSYView){};
  ~Position(){paddtypes.clear();};

  void reset(){
    ppos=0;
    paddpos=0;
    porigin=0;
    pangle=0;
    prestriction=false;
    poffset=0;
    poffset_type=OSTNone;
    paddtypes.resize(0,true);
    pchanged=false;
  }

  const Point &pos()const{return ppos;};
  void setPos(const Point &p){pchanged|=(ppos!=p);ppos=p;};
  
  void setAddPos(const Point &p){pchanged|=(paddpos!=p);paddpos=p;};

  const Point &origin()const{return porigin;};
  void setOrigin(const Point &p){pchanged|=(porigin!=p);porigin=p;};

  double angle()const{return pangle;};
  void setAngle(double a){pchanged|=(pangle!=a);pangle=a;};

  bool restriction()const{return prestriction;};
  void setRestriction(bool r){pchanged|=(prestriction!=r);prestriction=r;};


  double offset()const{return poffset;};
  void setOffset(double offset){pchanged|=(poffset!=offset);
    poffset=MyMath::max<double>(offset,0);};


  long offsettype()const{return poffset_type;};
  void setOffsetType(OffsetType type){
    pchanged|=(poffset_type!=type);
    poffset_type=type;
    if(poffset_type==OSTNone)
      setRestriction(false);
    else
      setRestriction(true);
  };
  long offsetsign()const{return ( poffset_type & OSTNeg?-1:1);};
  void setOffSetTypeFromString(const  char *c){
    long type;
    
    type=OSTNone;
    if(c[0]=='-')type|=OSTNeg;
    if(c[0]=='+')type|=OSTPos;
    if(c[1]=='X')type|=OSTX;
    if(c[1]=='Y')type|=OSTY;

    pchanged|=(poffset_type!=type);
    poffset_type=(OffsetType)type;
  };


  String offsetTypeAsString()const{
    String b="-";
    if( (poffset_type & OSTPos) ==OSTPos ){
      b="+";
    }
    b+='X';
    if( (poffset_type & OSTY) ==OSTY ){
      b+='Y';
    }
    return b;
  }

  


  Point absposToVC(const Graph *graph)const{
    Point p=posToDC(graph,ppos,pcoordsystem);
    p=dcToPos(graph,p,Xfp::CSYView);
    if(prestriction){
      p+=paddpos;
      if((poffset_type & OSTX)){
	p.x+=(poffset_type & OSTNeg?-1.:1.)*poffset;      
      }
      if((poffset_type & OSTY) ==OSTY){
	p.y+=(poffset_type & OSTNeg?-1.:1.)*poffset;      
      }
    }
    return p;
  };


  Point absposToDC(const Graph *graph)const{
    return graph->viewBox()->vctodc(absposToVC(graph));};
  
  Point pos0ToDC(const Graph *graph)const{
    Point p=posToDC(graph,ppos,pcoordsystem);
    p=dcToPos(graph,p,Xfp::CSYView);
    if(prestriction)
      return graph->viewBox()->vctodc(p+paddpos);
    else
      return graph->viewBox()->vctodc(p);
  };
  
  
  Point posToDC(const Graph *graph,const Point &p
		 ,Xfp::CoordinateSystem currentmode)const;
  Point dcToPos(const Graph *graph,const Point &p
		 ,Xfp::CoordinateSystem currentmode)const;

  long numAddTypes()const{return paddtypes.size();};
  OffsetAddTypes addType(long i)const{return (OffsetAddTypes)paddtypes.at(i);};
  void clearAddTypes(){pchanged=true;paddtypes.clear();};
  void appendAddType(OffsetAddTypes type){pchanged=true;
    paddtypes.append(type);};
  void setAddTypes(const ValueArray<long> &ats){pchanged|=(ats!=paddtypes);
    paddtypes.clear();
    for(int i=0;i<ats.size();i++){
      paddtypes.append((ats.at(i)-1)%OppXaLabelHeight+1);
    }
  };


  Xfp::CoordinateSystem coordinateSystem()const{return pcoordsystem;};
  void setCoordinateSystem(Xfp::CoordinateSystem c){
    pchanged|=(pcoordsystem!=c);pcoordsystem=c;
  };
  
  String asCommands(const String &prefix)const{

    String s=Sprintf("%s coordinate system %s\n"
		     "%s position %g , %g\n"
		     "%s angle %g\n"
		     "%s origin %g , %g\n"
		     ,prefix(),Xfp::coordinateSystemGuiName(pcoordsystem)
		     ,prefix(),ppos.x,ppos.y,prefix(),pangle
		     ,prefix(),porigin.x,porigin.y);
    if(!prestriction){
      s+=prefix+" restriction none\n";
    }else{
      s+=Sprintf("%s restriction %c%c %g\n"
		 ,prefix()
		 ,(poffset_type & OSTPos?'+':'-')
		 ,(poffset_type & OSTX?'X':'Y')
		 ,poffset
		 );
    }
    s+=prefix+" restriction additional";
    for(int i=0;i<paddtypes.size();i++){
      s+=Sprintf(" %d",paddtypes.at(i));
    }
    s+="\n";
    
    return s;
  };


  

  bool changed()const{
    return pchanged;
  };
  void setChanged(bool c){
    pchanged=c;
  };

 private:
  struct tsOffsetDescriptor{
    OffsetAddTypes type;
    const char *gui_description;
    const char *io_option_description;
  };

 public:
  static long numOfOffsetDescriptors(){return pmax_addoffsets;};
  static const char *offsetIoOptionDescription(int i){
    return paddoffset_description[i].io_option_description;
  };
  static const char *offsetGuiDescription(int i){
    return paddoffset_description[i].gui_description;
  };
  static OffsetAddTypes offsetTypeDescription(int i){
    return paddoffset_description[i].type;
  };
  static long indexOfOffsetDescriptionBtype(OffsetAddTypes t){
    for(int i=0;i<pmax_addoffsets;i++){
      if(paddoffset_description[i].type==t)
	return i;
    }
    return 0;
  }
  
 private:
  bool pchanged;
  Point ppos;			/* pos of the origin of i.e. a box
				   in fractions of view area */
  Point paddpos;		/* additional offsets form other objects */
  
  Point porigin;		/* the origin of i.e. a box in fraction 
			   of the box size
			   the ul corner is origin 0,0
			   the lr corner is origin 1,1
			   the origin is the placed at absolutepos
			   with respect to view
			   absolutepos=pos+addpos+offsets
			*/
  
  double pangle;			/* rotation angle of the box
				   positive angle is antioclockwise
				   rotation of the box around its origin
				*/
  
  /*  some i.e. boxs have restricted position
      the offset vector of the box origin from Pos0 is
      (offset,0) if offset_type=OFFSET_X_POS 
      (-offset,0) if offset_type=OFFSET_X_NEG 
      (0,offset) if offset_type=OFFSET_Y_POS 
      (0,-offset) if offset_type=OFFSET_Y_NEG 
      objects with  offset are allowed to move 
      into offset direction only. This assures correct placement
      independend of the font size.
  */
  bool prestriction;
  double poffset;
  OffsetType poffset_type;
  ValueArray<long> paddtypes;      /* additional offsets, which
				   are active if restriction is true*/
  Xfp::CoordinateSystem pcoordsystem;

 private:
  static const tsOffsetDescriptor paddoffset_description[];
  static const long pmax_addoffsets;

};

#ifdef	__cplusplus
extern "C" {
#endif

#ifdef	__cplusplus
}
#endif


#endif /* _POSITION_H */

