/* @(#)TextBox.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _TEXTBOX_H
#define _TEXTBOX_H 1

#include "GraphObject.hh"
#include "PtrArray.hh"
#include "Point.hh"
#include "Position.hh"

#include "FString.hh"

class Position;
class LineStyle;
class FillStyle;
class FontStyle;
class TextBoxes;


/* ***********************************************************************
 *
 * ***********************************************************************
 */
class TextBox: public GraphObject{
 public:

  enum TextBoxType{TBTTitle=0,TBTSubTitle=1
		   ,TBTXAxisLabel=2,TBTYAxisLabel=3
		   ,TBTOppositeXAxisLabel=4,TBTOppositeYAxisLabel=5
		   ,TBTGeneral=6};

  TextBox(Graph *parent,TextBoxes *textboxes,
	  TextBoxType t=TBTGeneral,const String &s="new textbox");
  
  TextBox(const TextBox &);
  TextBox &operator=(const TextBox &);

  TextBox *clone()const{return new TextBox(*this);}

  ~TextBox();



  void reset();

  TextBoxType type()const{return ptype;};
  void setType(TextBoxType t){pchanged|=(ptype!=t);ptype=t;};

  bool active()const{return pactive;};
  void setActive(bool a){pchanged|=(pactive!=a);pactive=a;};

  double border()const{return pborder;};
  void setBorder(double b){pchanged|=(pborder!=b);pborder=b;};

  bool boxIsActive()const{return pboxisactive;};
  void setBoxIsActive(bool a){pchanged|=(pboxisactive!=a);pboxisactive=a;};

  LineStyle *rim(){return prim;};
  const LineStyle *rim()const{return prim;};

  FillStyle *fill(){return pfill;};
  const FillStyle *fill()const{return pfill;};

  FontStyle *textStyle(){return ptextstyle;};
  const FontStyle *textStyle()const{return ptextstyle;};

  const String &text()const{return ptext;};
  void setText(const String &a){pchanged|=(ptext!=a);ptext=a;};

  Position *pos(){return ppos;};
  const Position *pos()const{return ppos;};

  const Size &size()const{return psize;};
  void setSize(const Size &s){pchanged|=(psize!=s);psize=s;};
  
    


  virtual GraphObjectRegion 
    insideRegion(const DragDescriptor &dd,bool update);
  virtual void initDrag(const DragDescriptor &dd);
  virtual void updateDrag(const DragDescriptor &dd);
  virtual void setFromDrag(DragDescriptor &dd);

  const String guiName(bool inmenu=false)const;
  const String paramGuiName(long index)const;


  String asCommands(const String &tab)const;

  bool changed()const;
  void setChanged(bool c);

 private:
  TextBoxes *ptextboxes;
  TextBoxType ptype;
  bool pactive;			/* inactive textboxes are invisible */
  bool pboxisactive;		/* is the box visible */
  /* --- frame box --- */
  LineStyle *prim;		/* the rim style  */
  FillStyle *pfill;		/* the fill style */
  double pborder;		/* spacing between rim and text in
				   units of textheight
				 */
  /* --- text --- */
  FontStyle *ptextstyle;
  String ptext;

  Position *ppos; 
  Size  psize;			/* only used in interactive mode
				   the postscript device is write only
				   thus we never get the box size in print mode
				 */
  static const char *pguinames[];
};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
class TextBoxes
{
 public:
  TextBoxes(Graph *parent);

  void reset();

  long size()const{return ptextboxes.size();};

  void adjustSizeToIndex(long size,bool shrink=false){
    long n=ptextboxes.size();
    if((!shrink) && size<=n)return;
    pchanged=true;
    if(!shrink){
      for(long j=n;j<size;j++){
	ptextboxes.append(new TextBox(pgraph,this,
				      TextBox::TBTGeneral,"new textbox"));
      }
    }else if(shrink && size<n){
      for(long i=n-1;i>=TextBox::TBTGeneral;i--){
	ptextboxes.removeAt(i);
      }
    }
  }

  TextBox *at(long i){return ptextboxes[i];};
  const TextBox *at(long i)const{return ptextboxes.at(i);};

  TextBox *textBoxByType(TextBox::TextBoxType t,long index=0){
    if(t==TextBox::TBTGeneral){
      if(t+index<ptextboxes.size())
	return ptextboxes[t+index];
      else
	return 0;
    }else
      return ptextboxes[t];
  };
  
  inline void deleteTextBox(TextBox* tb){ptextboxes.remove(tb);};
  inline void deleteTextBox(long i){ptextboxes.removeAt(i);};

  inline void copyTextBox(TextBox* tb){appendTextBox(tb);};

  TextBox *takeAt(long i){return ptextboxes.takeAt(i);};

  TextBox *appendTextBox(TextBox *tb=0){
    TextBox *tbb;
    if(tb){
      tbb=tb->clone();
    }else{
      tbb=new TextBox(pgraph,this,TextBox::TBTGeneral,"new textbox");
    }
    ptextboxes.append(tbb);
    return tbb;
  }


  long indexOf(const TextBox *tb)const {
    return ptextboxes.indexOf(tb);
  };
  void insertTextBox(long index,TextBox *tb=0){
    if(tb)
      ptextboxes.insert(index,tb);
    else
      ptextboxes.insert(index,new TextBox(pgraph,this,TextBox::TBTGeneral,
					  "new textbox"));
  };

  String asCommands(const String &tab)const;


  bool changed()const{
    bool c=pchanged;
    for(int i=0;i<ptextboxes.size();i++){
      c|=ptextboxes.at(i)->changed();
    }
    return c;
  };
  void setChanged(bool c){pchanged=c;
    for(int i=0;i<ptextboxes.size();i++){
      ptextboxes[i]->setChanged(c);
    }
  };

 private:
  void setSpecialBoxes();
 private:
  bool pchanged;
  PtrArray<TextBox> ptextboxes;
  Graph *pgraph;

};


#ifdef	__cplusplus
extern "C" {
#endif

#ifdef	__cplusplus
}
#endif


#endif /* _TEXTBOX_H */

