/*************************************************************************
 * @(#)FString.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   03 Oct 2008
 *************************************************************************
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <math.h>
#include <ctype.h>


#include "convtests.h"
#include "ValueArray.hh"
#include "Except.hh"
#include "FString.hh"
#include "MyDebug.hh"
#include "MyMath.hh"


// const char *String::classname="String";
// const char *StringList::classname="StringList";

const String String::setaz="abcdefghijklmnopqrstuvwxyz"; 
const String String::setAZ="ABCDEFGHIJKLMNOPQRSTUVWXYZ"; 
const String String::setazAZ="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"; 

/* ***********************************************************************
 * make escape sequence out of special characters
 * ***********************************************************************
 */
static char enescape(char c)
{
  switch (c){
  case ('\0'):return '0';   break;
  case ('\a'):return 'a';   break;
  case ('\b'):return 'b';   break;    
  case ('\f'):return 'f';   break;    
  case ('\n'):return 'n';   break;    
  case ('\r'):return 'r';   break;    
  case ('\t'):return 't';   break;
  case ('\v'):return 'v';   break;
  }
  return c;
}

/* ***********************************************************************
 * translate escape character sequences from escaped form into
 * raw form 
 * ***********************************************************************
 */
static char deescape2(char c)
{
  switch (c){
  case ('0'): return '\0';
  case ('a'): return '\a';
  case ('b'): return '\b';
  case ('f'): return '\f';
  case ('n'): return '\n';
  case ('r'): return '\r';
  case ('t'): return '\t';
  case ('v'): return '\v';
  }
  return c;
}

/* ***********************************************************************
 * copy src into dest and replace special characters by their escape sequence
 * ***********************************************************************
 */
static const char *escapedecode(char *dest,const char *src)
{
  const char *ps=src;
  char *pd=dest,c;
  *pd='\0';
  while(*ps != '\0'){
    c=enescape(*ps);
    if(c=='\\'){
      *pd++='\\';
      *pd++='\\';
    }else if(c==*ps){
      *pd++=c;
    }else{
      *pd++='\\';
      *pd++=c;
    }
    ps++;
  }
  *pd='\0';
  return dest;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static const char *escapeencode(char *dest,const char *src)
{
  const char *ps=src;
  char *pd=dest;
  char c;
  *pd='\0';
  while(*ps != '\0'){
    if(*ps=='\\'){
      ps++;
      c=deescape2(*ps);
      if(c=='\\'){
	*pd++='\\';
	ps++;
      }else if(c==*ps){
	*pd++='\\';
	*pd++=*ps++;
      }else{
	*pd++=c;
	ps++;
      }
    }else{
      *pd++=*ps++;
    }
  }
  *pd='\0';
  return dest;
}



// {{{ String


/* ***********************************************************************
 *
 * ***********************************************************************
 */

String::String(char const* v):pinternalsize((v?strlen(v):0)+1),plength(pinternalsize-1),
			      pvalue(new char[pinternalsize]),pstrategy(Default)
{
//   NSF("String(char const*)");
  if(v) strcpy(pvalue, v);
  else pvalue[0]='\0';
//   printf("%s:arg=%s, pinternalsize(%ld), strlen(%ld) pvalue(%s)\n",nsubfunc,
// 	 v,pinternalsize,strlen(pvalue),pvalue);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
String::String(const char *other,size_t len)
{
  pstrategy=Default;
  size_t l=strlen(other);
  pinternalsize=(l<len?l:len)+1;
  plength=pinternalsize-1;
  pvalue=new char[pinternalsize];
  strncpy(pvalue,other,plength);
  pvalue[plength]='\0';
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String::String(String const& other)
  : pinternalsize(other.plength+1),plength(other.plength),
    pvalue(new char[pinternalsize]),pstrategy(Default)
{
//   NSF("String(String const &)");
  strcpy(pvalue, other.pvalue);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String::String(long i):pinternalsize(0),pvalue(0),pstrategy(Default)
{
  //  NSF("String(long)");
  swapAll(Sprintf("%ld",i).removeTrailingBlanks());
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
#if (defined QXFSF || defined QXFPL)
String::String(const QString &other)
  :pinternalsize(0),pvalue(0),pstrategy(Default)
{
#if QT_VERSION >= 0x050000
  String temp(other.toLatin1().constData());
#else
  String temp(other.toAscii().constData());
#endif
  swapAll(temp);
}
#endif

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String::~String()
{
  //printf("String: deleteing this %p pvalue(%p)=%s\n",this,pvalue,pvalue);
  delete [] pvalue; 
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
String const& String::operator=(String const& other)
{
  //NSF("operator=(String const&)");
  if(&other==this) return *this;
  plength=other.plength;
  if(pstrategy!=DontShrink || pinternalsize<plength+1){
    if(pvalue) delete [] pvalue;
    pinternalsize=plength+1;
    pvalue=new char[pinternalsize];
  }
  strcpy(pvalue, other.pvalue);
  return *this;

  // String temp(other);
  // swapAll(temp);
  // return *this;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
String const& String::operator=(char c)
{
  //  NSF("operator=(const char *)");
  delete [] pvalue;
  plength=1;
  pinternalsize=2;
  pvalue=new char[pinternalsize];
  pvalue[0]=c;
  pvalue[1]='\0';
  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String const& String::operator=(const char *v)
{
//   NSF("operator=(const char *)");
  if(v[0]=='\0' || v==0){
    if(pstrategy!=DontShrink){
      delete [] pvalue;
      plength=0;
      pinternalsize=1;
      pvalue=new char[pinternalsize];
    }
    pvalue[0]='\0';
    return *this;
  }
  plength=strlen(v);
  if(pstrategy!=DontShrink || pinternalsize<plength+1){
    delete [] pvalue;
    pinternalsize=plength+1;
    pvalue=new char[pinternalsize];
  }
  strcpy(pvalue, v);
  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String const& String::operator+=(char ch)
{
//   NSF("String const& String::operator+=(char ch)");
  plength+=1;
  
  // new pinternalsize must be >=plength+1
  
  if(pinternalsize>=plength+1){
    pvalue[plength-1]=ch;
    pvalue[plength]='\0';
    return *this;
  }
  unsigned long l=(unsigned long)(plength*1.5);
  pinternalsize=(l>plength+1?l:plength+1);

  char *temp=new char[pinternalsize];
  strcpy(temp,pvalue);
  temp[plength-1]=ch;
  temp[plength]='\0';

  MyMath::swap(this->pvalue,temp);
 
  delete [] temp;

  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String const& String::operator+=(String const & other)
{
  //  NSF("String::operator+=(String const&)");
  
  unsigned long newlength=plength+other.plength;


  // new pinternalsize must be >=newlength+1
  if(pinternalsize>=newlength+1){
    strcpy(pvalue+plength,other.pvalue);
    plength=newlength;
    return *this;
  }

  //let it grow a bit bigger to forsee future concatenations
  unsigned long l=(unsigned long)((newlength+1)*1.5);
  pinternalsize=(l>newlength+1?l:newlength+1);

  char *temp=new char[pinternalsize];
  strcpy(temp, pvalue);
  strcpy(temp+plength,other.pvalue);
  plength=newlength;

  MyMath::swap(this->pvalue,temp);
  delete [] temp;

  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String const& String::operator+=(const char *other)
{
  //NSF("String::operator+=(const char*)");

  unsigned long newlength=plength+strlen(other);
  
  // new pinternalsize must be >=reqlen
  if(pinternalsize>=newlength+1){
    strcpy(pvalue+plength,other);
    plength=newlength;
    return *this;
  }

  //let it grow a bit bigger to forsee future concatenations
  unsigned long l=(unsigned long)((newlength+1)*1.5);
  pinternalsize=(l>newlength+1?l:newlength+1);

  char *temp=new char[pinternalsize];
  strcpy(temp, pvalue);
  strcpy(temp+plength,other);
  plength=newlength;

  MyMath::swap(this->pvalue,temp);
  delete [] temp;

  return *this;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::set(unsigned long i,char with)
{
  if(i>plength || (i==plength && with !='\0'))
    throw Except(Sprintf("String::set(%d,%c): this=\"%s\" index out of bounds",
			i,with,pvalue));
  if(i==plength && (with=='\0')){
    return *this;
  }
  
  pvalue[i]=with;
  if(with=='\0'){
    plength=with;
  }
  

  return *this;
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void String::swapAll(String& other)
{
  MyMath::swap(this->pvalue,other.pvalue);
  MyMath::swap(this->plength,other.plength);
  MyMath::swap(this->pinternalsize,other.pinternalsize);
}







/* ***********************************************************************
 *
 * ***********************************************************************
 */
String String::escapeDecode(const char *c)
{
  //NSF("escapeDecode1");
  String s;
  s.pinternalsize=strlen(c)*3+1;
  delete [] s.pvalue;
  s.pvalue=new char[s.pinternalsize];
  escapedecode(s.pvalue,c);
  s.plength=strlen(s.pvalue);
  return s;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String String::escapeDecode(String const &c)
{
  //NSF("escapeDecode");

  String s;
  s.pinternalsize=c.plength*3+1;
  delete [] s.pvalue;
  s.pvalue=new char[s.pinternalsize];
  escapedecode(s.pvalue,c.pvalue);
  s.plength=strlen(s.pvalue);
  return s;


}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::escapeDecode()
{
  //NSF("String &String::escapeDecode()");

  pinternalsize=plength*3+1;
  char *dest=new char[pinternalsize];
  escapedecode(dest,pvalue);
  MyMath::swap(pvalue,dest);
  plength=strlen(pvalue);

  delete [] dest;
//   printf("%s:, pinternalsize(%ld), strlen(%ld) pvalue(%s)\n",nsubfunc,
// 	 pinternalsize,strlen(pvalue),pvalue);

  return *this;
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::escapeEncode()
{
  //NSF("String &String::escapeEncode()");

  pinternalsize=plength*3+1;
  char *dest=new char[pinternalsize];
  escapeencode(dest,pvalue);
  MyMath::swap(this->pvalue,dest);
  plength=strlen(pvalue);
  delete [] dest;
//   printf("%s:, pinternalsize(%ld), strlen(%ld) pvalue(%s)\n",nsubfunc,
// 	 pinternalsize,strlen(pvalue),pvalue);
  return *this;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String String::escapeEncode(const char *c)
{
  //NSF("escapeEncode1");
  String s;
  s.pinternalsize=strlen(c)*3+1;
  delete [] s.pvalue;
  s.pvalue=new char[s.pinternalsize];
  escapeencode(s.pvalue,c);
  s.plength=strlen(s.pvalue);
  return s;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String String::escapeEncode(String const &c)
{
  //NSF("escapeEncode1");

  String s;
  s.pinternalsize=c.plength*3+1;
  delete [] s.pvalue;
  s.pvalue=new char[s.pinternalsize];
  escapeencode(s.pvalue,c.pvalue);
  s.plength=strlen(s.pvalue);

  return s;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::interpolateQuotes()
{
  
  char *tmp=new char[pinternalsize];

  //strcpy(tmp,pvalue);

  const char *ps=pvalue;
  char *pd=tmp;
  
  while(*ps!='\0'){
    if(*ps=='\\' && ps[1]=='"'){
      ps++;
    }else{
      *pd++=*ps++;
    }
  }
  *pd='\0';

  MyMath::swap(tmp,pvalue);
  delete [] tmp;

  plength=strlen(pvalue);

  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::sprintf(const char *fmt,...)
{
  //NSF("sprintf");
  
  /* Guess we need no more than 100 bytes. */
  long n;
  va_list ap;
  pinternalsize=100;
  char *value;
  try{
    value= new char[pinternalsize];
  }catch(...){
    throw Except("String::sprintf: out of memory");
  }

  while (1) {
    /* Try to print in the allocated space. */
    va_start(ap, fmt);
    n = vsnprintf(value, pinternalsize, fmt, ap);
    va_end(ap);
    /* If that worked, return the string. */
    if (n > -1 && n < (long)pinternalsize)
      break;
    /* Else try again with more space. */
    if (n > -1)    /* glibc 2.1 */
      pinternalsize = n+1; /* precisely what is needed */
    else           /* glibc 2.0 */
      pinternalsize *= 2;  /* twice the old size */
    try{
      char *tmp=new char[pinternalsize];
      strcpy(tmp,value);
      MyMath::swap(tmp,value);
      delete [] tmp;
    }catch(...){
      throw Except("String::sprintf: out of memory");
    }

  }
  plength=n;
  MyMath::swap(value,pvalue);
  delete [] value;
  
  return *this;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String String::doubleToString(const String &fmt,double d,bool contract)
{
  if(contract){
    long i=fmt.find("%");
    if(i>=0 && fmt[i+1]==' '){
      String t=Sprintf(fmt(),d).stripBlanks().chop("0");
      if(t[0]!='-'){
	t=" "+t;
      }
      if(t=="-0.") t=" 0.";
      return t;
    }else{
      String t=Sprintf(fmt(),d).stripBlanks().chop("0");
      if(t=="-0.") t="0.";
      return t;
    }
  }
  return Sprintf(fmt(),d);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String Sprintf(const char *fmt,...)
{
  //NSF("Sprintf");
  
  /* Guess we need no more than 100 bytes. */
  long n;
  va_list ap;
  String s;
  s.pinternalsize=100;

  try{
    delete [] s.pvalue;
    s.pvalue= new char[s.pinternalsize];
  }catch(...){
    throw Except("String::sprintf: out of memory");
  }
  
  while (1) {
    /* Try to print in the allocated space. */
    va_start(ap, fmt);
    n = vsnprintf(s.pvalue, s.pinternalsize, fmt, ap);
    va_end(ap);
    /* If that worked, return the string. */
    if (n > -1 && n < (long)s.pinternalsize)
      break;
    /* Else try again with more space. */
    if (n > -1)    /* glibc 2.1 */
      s.pinternalsize = n+1; /* precisely what is needed */
    else           /* glibc 2.0 */
      s.pinternalsize *= 2;  /* twice the old size */
    try{
      char *tmp=new char[s.pinternalsize];
      strcpy(tmp,s.pvalue);
      MyMath::swap(tmp,s.pvalue);
      delete [] tmp;
    }catch(...){
      throw Except("String::sprintf: out of memory");
    }
  }
  s.plength=n;

  
  return s;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::cmp(String const &s1,String const &s2)
{
  return strcmp(s1.pvalue,s2.pvalue);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::cmp(const char *s1,String const &s2)
{
  return strcmp(s1,s2.pvalue);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::cmp(String const &s1,const char *s2){
  return strcmp(s1.pvalue,s2);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::cmp(const char *s1,const char *s2){
  return strcmp(s1,s2);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::cmp(String const &s1,String const &s2,long l)
{
  return strncmp(s1.pvalue,s2.pvalue,l);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::cmp(const char *s1,String const &s2,long l)
{
  return strncmp(s1,s2.pvalue,l);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::cmp(String const &s1,const char *s2,long l){
  return strncmp(s1.pvalue,s2,l);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::cmp(const char *s1,const char *s2,long l){
  return strncmp(s1,s2,l);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::toLower()
{
  char *p0=pvalue;
  
  while(*p0!='\0') { *p0=tolower(*p0); p0++; }
  
  return *this;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::toUpper()
{
  char *p0=pvalue;
  
  while(*p0!='\0') { *p0=toupper(*p0); p0++; }
  
  return *this;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::removeLeadingBlanks(bool removenewline)
{
//   NSF("String &String::removeLeadingBlanks()");
  // first nonblank is in pvalue[l];
  size_t l;
  if(removenewline)
    l=strspn(pvalue," \t\n");
  else
    l=strspn(pvalue," \t");
  if(l==0)return *this;

  char *p0=pvalue;
  char *p1=pvalue+l;

  while(*p1!='\0'){ *p0=*p1; p0++;p1++; }
  *p0=*p1;
  plength-=l;
//   printf("%s:, pinternalsize(%ld), strlen(%ld) pvalue(%s)\n",nsubfunc,
// 	 pinternalsize,strlen(pvalue),pvalue);
  return *this;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */

String &String::chop(const char *removeset)
{
  if(removeset==0) return *this;
  if(plength==0) return *this;
  char *p=pvalue+plength-1;
  char c[2]=" ";

  while(1){
    *c=*p;
    if(strstr(removeset,c)!=NULL){
      *p='\0';plength--;
    }else if(*p!='\0') {
      break;
    }
    
    if(p==pvalue)break;
    p--;
  }
  return *this;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */

String &String::removeTrailingBlanks(bool removenewline)
{
  char *p=pvalue+plength;

  while(1){
    if(*p==' ' || *p=='\t' || (removenewline && *p=='\n') ){
      *p='\0';plength--;
    }else if(*p!='\0') break;
    
    if(p==pvalue)break;
    p--;
  }
  return *this;
}

 

/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::convertToLong()const
{
  FPLOINTEGER l=0;

  if(!convertlong(pvalue,&l)) 
    throw Except(Sprintf(" invalid integer number '%s'",pvalue));

  return l;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
unsigned long String::convertToULong()const
{
  unsigned long ul;
  if(sscanf(pvalue,"%lx",&ul)-1)
    throw Except(Sprintf(" invalid unsigned long number '%s'",pvalue));

  return ul;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
double String::convertToFortranReal()
{
  FPLODOUBLEPRECISION d=0.;

  if(!convertfortranreal(pvalue,&d)) 
    throw Except(Sprintf(" invalid real number '%s'",pvalue));

  return d;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void String::convertToFortranChar(char *dest,long dest_len)const
{
  if(dest_len<(long)plength){
    String s;s.sprintf("error(String::convertToFortranChar): "
		       "string longer (%ld) than \n"
		       "fortran character variable (%ld)!",plength,dest_len);
    (void)fprintf(stderr,"fatal %s\n",s());
    (void)fprintf(stdout,"fatal %s\n",s());
    exit(2);
    

  }

  for(long i=0;i<dest_len;i++){
    if(i<(long)plength)
      dest[i]=pvalue[i];
    else
      dest[i]=' ';
  }

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
double String::convertToReal()const
{
  char *endptr;
  String s(*this);
  s.stripBlanks();
  errno=0;
  double d=strtod(pvalue,&endptr);
  
  if (errno || *endptr!='\0' || *pvalue == '\0' ) {
    throw Except(Sprintf(" invalid real number '%s'",pvalue));
  }


  return d;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

long String::islong(long *l,const char *s)
{
  char *err;
  if(s==NULL)return 0;
  *l=strtol(s,&err,10);
  return (*s!='\0' && *err=='\0');
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */

long String::islong(long *l,char c)
{
  char *err;
  char s[2]=" ";
  s[0]=c;
  if(s==NULL)return 0;
  *l=strtol(s,&err,10);
  return (*s!='\0' && *err=='\0');
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::islong(long *l,const char *s,long lowerbound,long upperbound)
{
  return (String::islong(l,s) && ( *l >= lowerbound && *l <= upperbound));
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void String::isOrbital(long maxmoment) const
{
  char p[2];
  p[0]=pvalue[0]; p[1]='\0';
  
  if( plength != 2) 
    throw Except(": state incomplete");
  
  long l;
  if (!String::islong(&l,p,1L,8L)) 
    throw Except(": unknown main qn.");
  
  if(!String::isSPDF(pvalue[1],maxmoment)) 
    throw Except(": unknown momentum");

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::isSPDF(char s,long maxmoment)
{
  char c=tolower(s);
  
  if(c=='s' && maxmoment>=0 ) return 1;
  if(c=='p' && maxmoment>=1 ) return 1;
  if(c=='d' && maxmoment>=2 ) return 1;
  if(c=='f' && maxmoment>=3 ) return 1;
  if(c=='g' && maxmoment>=4 ) return 1;
  if(c=='h' && maxmoment>=5 ) return 1;
  if(c=='i' && maxmoment>=6 ) return 1;
  return 0;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void String::orbitalToInt(long &nmain,long &l,long maxmoment)const
{
  if(plength != 2){
    throw Except(Sprintf("String::orbitalToInt:  '%s' is not an orbital"
			 ,pvalue));
  }
  char p[2];
  p[0]=pvalue[0]; p[1]='\0';
  if (!String::islong(&nmain,p,1L,8L)) 
    throw Except(Sprintf("String::orbitalToInt:  '%s': invalid main qn."
			 ,pvalue));
  if(!String::isSPDF(pvalue[1],maxmoment)) 
    throw Except(Sprintf("String::orbitalToInt:  '%s': invalid momentum"
			 ,pvalue));
  l=spdfToInt(pvalue[1]);

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::spdfToInt(char s)
{
  if(s=='s') return 0;
  if(s=='p') return 1;
  if(s=='d') return 2;
  if(s=='f') return 3;
  if(s=='g') return 4;
  if(s=='h') return 5;
  if(s=='i') return 6;
  throw Except(Sprintf("String::spdfToInt: orbital '%c' not implemented!",s));
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
char String::intToSpdf(long l)
{
  const char* spdf="spdfghi";
  if(l<0 || l>6 )
    throw Except(Sprintf("String::intToSpdf: momentum '%ld' not implemented!"
			 ,l));
  return spdf[l];
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String String::intToOrbital(long nmain,long l)
{
  return Sprintf("%1ld%1c",nmain,String::intToSpdf(l));
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String::StateLabels String::isStateLabel(const char *ss,long &nmain,long &l,
					 long &j,long &m)
{
  long maxmoment=3;
  
  //  const char *digits="0123456789";
  nmain=0;
  l=-1;
  j=0;
  m=0;
  String tmp(ss);
  tmp=tmp.stripBlanks(true).lowerCase().removeFromSet(" \t");
  if(tmp.empty() || tmp=="all")return SL_all;
  if(tmp=="allnlm")return SL_allnlm;
  


  String o=tmp.subString(0,1);

  try{
    o.orbitalToInt(nmain,l,maxmoment);
    if(tmp.length()==2) return SL_nl;
  }catch(Except ex){
    return SL_invalid;
  }
  
  tmp=tmp.subString(2);
  if(String::islong(&m,tmp(),-l,l)){
    return SL_nlm;
  }
  
  long k=tmp.find("/"); 
  
  if(k>=0){
    if(!String::islong(&j,tmp.subString(0,k-1)())){
      return SL_invalid;
    }
    if(j%2!=1||j<1) return SL_invalid;
    if(j!=2*l-1&&j!=2*l+1) return SL_invalid;
    if(tmp.pvalue[k+1]!='2') return SL_invalid;
    if(tmp.subString(k+2).length()==0)return SL_nlj;
    tmp=tmp.subString(k+2);

    long k=tmp.find("/"); 
    if(k>=0){
      if(!String::islong(&m,tmp.subString(0,k-1)())){
	return SL_invalid;
      }
      if(abs(m)%2!=1||abs(m)>j) return SL_invalid;
      if(tmp.pvalue[k+1]!='2') return SL_invalid;
      if(tmp.subString(k+2).length()==0)return SL_nljm;
    }

  }

  return SL_invalid;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
StringList String::breakup(const char *delim) const
{
  //  NSF("breakup");
  char *ptr=new char[plength+1];
  strcpy(ptr,pvalue);
  char *p,*p0;
  long slen=strlen(delim);

  StringList l;
  
  p0=ptr;
  while(p0!=NULL){
    p=strstr(p0,delim);
    if(p==NULL) { 
      l.append(p0);
      break;
    }
    *p='\0';
    l.append(p0);
    p0=p+slen;
  }
  
  delete [] ptr;

  return l;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
StringList String::split(const char *delim) const
{
  //  NSF("split");
  //myDebug()<<"String::split: this="<<pvalue;
  char *ptr=new char[plength+1];
  strcpy(ptr,pvalue);
  char *p;

  StringList l;
  
  p=strtok(ptr,delim);
  while(p!=NULL){
    l.append(p);
    p=strtok(NULL,delim);
  }
  
  delete [] ptr;

  // printf("l=");
  // l.dumpinternals();
  return l;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::expandComma()
{
//   NSF("String &String::expandComma()");
  
  unsigned long len=plength;
  char *tmp=new char[len*3+1];

  long j=0;
  for(unsigned long i=0;i<len;i++){
    if(pvalue[i]==','){
      tmp[j++]=' ';
      tmp[j++]=',';
      tmp[j++]=' ';
    }else{
      tmp[j++]=pvalue[i];
    }
  }
  tmp[j]='\0';
  
  char *p=tmp+strspn(tmp," \t");

  char *word=new char[len*3+1];
  char *val=new char[len*3+1];
  String res;
  
  val[0]='\0';
  while(sscanf(p,"%s",word)!=EOF){
    if(word[0]!=',') (void)strcpy(val,word);
    res+=" ";
    res+=val;
    p+=strcspn(p," \t");
    p+=strspn(p," \t");
  };
  delete [] val ;
  delete [] word ;
  delete [] tmp ;


  res.removeLeadingBlanks();
  swapAll(res);

//   printf("%s:, pinternalsize(%ld), strlen(%ld) pvalue(%s)\n",nsubfunc,
// 	 pinternalsize,strlen(pvalue),pvalue);
   

  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
ValueArray<long> String::splitToLongArray(const char *delim) const
{
  //  NSF("splitToIArray");
  char *ptr=new char[plength+1];
  strcpy(ptr,pvalue);
  char *p;
  
  ValueArray<long> ia;
  
  p=strtok(ptr,delim);
  while(p!=NULL){
    ia.append(String(p).convertToLong());
    p=strtok(NULL,delim);
  }
  
  delete [] ptr;

  return ia;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
ValueArray<double> String::splitToDoubleArray(const char *delim) const
{
  //  NSF("splitToDoubleArray");
  char *ptr=new char[plength+1];
  strcpy(ptr,pvalue);
  char *p;
  
  ValueArray<double> da;
  
  p=strtok(ptr,delim);
  while(p!=NULL){
    da.append(String(p).convertToFortranReal());
    p=strtok(NULL,delim);
  }
  
  delete [] ptr;

  return da;
}






/* ***********************************************************************
 *
 * ***********************************************************************
 */
String String::capitalized(bool all) const
{
  String s(*this);
  s.toLower();
  
  char *p0=s.pvalue;
  
  if(*p0!='\0') { *p0=toupper(*p0); }

  if(all){
    char *p=s.pvalue;
    while(*p!='\0'){
      p+=strspn(p," \t\n");
      if(*p!='\0'){ 
	*p=toupper(*p); 
      }else{
	break;
      }
      p+=strcspn(p," \t\n");
    }
  }


  return s;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::capitalize(bool all)
{
  this->toLower();
  
  char *p0=pvalue;
  
  if(*p0!='\0') { *p0=toupper(*p0); }
  if(all){
    char *p=pvalue;
    while(*p!='\0'){
      p+=strspn(p," \t\n");
      if(*p!='\0'){ 
	*p=toupper(*p); 
      }else{
	break;
      }
      p+=strcspn(p," \t\n");
    }
  }


  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
StringList String::wrapAround(unsigned long cols,const char *marker) const
{

  long m=cols-strlen(marker);

  if(m<=0)
    throw Except("String::wrapAround: cols-marker.length<=0!");

  const char *pp=(const char *)pvalue;

  unsigned long l=plength;

  char *p=new char[l+1];

  StringList sl;

  while(l>cols){ 
    strncpy(p,pp,(size_t)m); 
    p[cols]='\0'; 
    strcpy(&p[m],marker);
    pp+=m; 
    l-=m; 
    sl.append(p);
  } 
  sl.append(pp);

  delete [] p;p=0;

  return sl;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::shorten(long size)
{
  //  NSF("String &String::shorten(long)");
  if(size>=0 && (long)plength>size ){
    plength=size;
    pvalue[size]='\0';
  }

  return *this;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::isDigit(long c)
{
  if(c<-1 || c>255){
    return 0;
  }
  return isdigit(c);
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::isPrint(long c)
{
  if(c<-1 || c>255){
    return 0;
  }
  return isprint(c);
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::isAlpha(long c)
{
  if(c<-1 || c>255){
    return 0;
  }
  return isalpha(c);
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::contains(String const &set) const
{
  return (strcspn(pvalue,set())!=plength);
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::scan(String const &set) const
{
  size_t j=strcspn(pvalue,set());
  if(j==plength) 
    return -1;
  else
    return j;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::reject(String const &set) const
{
  size_t j=strspn(pvalue,set());
  if(j>plength) 
    return -1;
  else
    return j;
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
long String::find(String const &substring,bool fromEnd) const
{
  char *p,*pp;
  if(fromEnd){
    p=strstr(pvalue,substring());
    if(p==NULL) 
      return -1;
    while(1){
      pp=strstr(p+1,substring());
      if(pp==NULL) 
      return (p-pvalue);
      p=pp;
    }
    return (p-pvalue);
  }else{
    if((p=strstr(pvalue,substring()))!=NULL)
      return (p-pvalue);
    else
      return -1;
  }
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
String String::subString(long i)const
{
  if(i<0 || (i>(long)plength)) return String();

  String temp;
  temp.pinternalsize=plength+1;
  delete [] temp.pvalue;
  temp.pvalue=new char[temp.pinternalsize];
  unsigned long le=temp.pinternalsize-i-1;
  
  strncpy(temp.pvalue,pvalue+i,le);
  temp.pvalue[le]='\0';
  temp.plength=le;
  
  return temp;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String String::subString(long i,long j)const
{
  if(i<0) return String();
  if(j<i) return String();

  unsigned long L=plength;
  String temp;
  temp.pinternalsize=L+1;
  delete [] temp.pvalue;
  temp.pvalue=new char[temp.pinternalsize] ;
  if(j>(long)L-1)j=L-1;
  unsigned long le=j-i+1;
  
  strncpy(temp.pvalue,pvalue+i,le);
  temp.pvalue[le]='\0';
  temp.plength=le;

  return temp;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::replace(const String &substring,const String &with,
			long startat,bool all)
{
  String tmp;
  unsigned long len=substring.length();
  int i0=startat;
  if(startat){
    tmp=subString(0,startat-1);
  }
  char *p;
  
  while(1){

    if((p=strstr(pvalue+i0,substring()))!=NULL){
      int i1=(p-pvalue);
      
      tmp+=subString(i0,i1-1)+with;
      if(!all){
	tmp+=subString(i1+len);
	swapAll(tmp);
	return *this;
      }else{
	i0=i1+len;
      }
      
      
      
    }else{
      tmp+=subString(i0);
      swapAll(tmp);
      return *this;
    }
  }

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::removeSection(long i,long j)
{
//   NSF("String &String::removeSection(long,long)");

  if(j<i)
    throw Except("String::removeSection: j<i!");
  
  unsigned long len=plength;
  
  if(i<0 || i>=(long)len)
    throw Except("String::removeSection: i out of bounds!");
  
  char *v=new char[pinternalsize];
  
  strncpy(v,pvalue,i);
  plength=i;
  v[i]='\0';

  if(j<(long)len-1){
    strcat(v,pvalue+j+1);
    plength+=len-j-1;
  }
  
  MyMath::swap(this->pvalue,v);
  
  delete [] v;

//   printf("%s:args:%ld,%ld, pinternalsize(%ld), strlen(%ld) pvalue(%s)\n",
// 	 nsubfunc,i,j,pinternalsize,strlen(pvalue),pvalue);
  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::insertChar(long i,char ch)
{
//   NSF("String &String::insertChar(long i,char ch)");
  
  if(i<0 || i>(long)plength)
    throw Except("String::insertChar: i out of bounds!");
  
  unsigned long newlength=plength+1;
  
  if(pinternalsize>=newlength+1){
    
    long j;
    for(j=(long)plength;j>=i+1;j--){
      pvalue[j]=pvalue[j-1];
    }
    pvalue[i]=ch;
    pvalue[newlength]='\0';
    plength=newlength;
    
//     printf("%s:args:%ld,%c, pinternalsize(%ld), strlen(%ld) pvalue(%s)\n",
//  	   nsubfunc,i,ch,pinternalsize,strlen(pvalue),pvalue);
    return *this;
  }
  
  //let it grow a bit bigger to forsee future insertions
  unsigned long l=(unsigned long)((newlength+1)*1.5);
  pinternalsize=(l>newlength+1?l:newlength+1);
  
  char *v=new char[pinternalsize];
  
  if(i>0) strncpy(v,pvalue,i);
  v[i]=ch;
  v[i+1]='\0';
  
  if(i<(long)plength) strcat(v,pvalue+i);
  plength=newlength;
  
  MyMath::swap(this->pvalue,v);
  
  delete [] v;
  
//   printf("%s:args:%ld,%c, pinternalsize(%ld), strlen(%ld) pvalue(%s)\n",
// 	 nsubfunc,i,ch,pinternalsize,strlen(pvalue),pvalue);
  return *this;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */

String &String::leftPadd(char c,long len)
{
  //todo: make efficient and correct
  if((long)plength>len){
    return *this;
  }

  pinternalsize=len+1;
  char *temp=new char[pinternalsize];
  strcpy(temp+len-plength,pvalue);

  memset(temp,(int)c,len-plength);
  temp[len]='\0';
  plength=len;
  MyMath::swap(this->pvalue,temp);

  delete [] temp;
  
  return *this;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */

String &String::rightPadd(char c,long len)
{
  //todo: make efficient and correct
  if((long)plength>len){
    return *this;
  }

  pinternalsize=len+1;
  char *temp=new char[pinternalsize];
  strcpy(temp,pvalue);

  memset(temp+plength,(int)c,len-plength);
  temp[len]='\0';
  plength=len;
  MyMath::swap(this->pvalue,temp);

  delete [] temp;
  
  return *this;
}

/* ***********************************************************************
 * 
 * ***********************************************************************
 */
String String::removeFromSet(String const &set) const
{
  
  char *temp=new char[pinternalsize];

  char ch[2]=" ";
  
  char *pr=temp;
  
  const char *ps=pvalue;
  
  while(*ps !='\0'){
    ch[0]=*ps;
    if(strstr(set(),ch)==NULL){
      *pr++=*ps;
    }
    ps++;
  }
  *pr='\0';
  
  String res;
  res.pinternalsize=pinternalsize;
  res.plength=pr-temp;
  MyMath::swap(res.pvalue,temp);

  
  delete [] temp;

  return res;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

#ifndef RSIGN
#define RSIGN(x,y) ( (x)* ((y)>=0. ? 1.:-1.) )
#endif

void String::makeRatio(double d,long &m,long &n,long maxn,double tol)
{
  long i,s;
  double da=fabs(d);
  s=(long)RSIGN(1.,d);
  if(da<=tol){
    m=0;
    n=1;
    return;
  }
  n=m=0; 
  for(i=1;i<=maxn;i++){
    m=(((da*i))>0. ?((long)((da*i)+.5)) :((long)((da*i)-.5))) ;
    
    if(fabs(((double)m)/((double)i)-da)<tol){
      n=i;
      break;
    }
  }
  m*=s;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String String::doubleToRatio(double d,long maxn,double tol)
{
  long m,n;
  makeRatio(d,m,n,maxn,tol);
  
  if(n == 0) 
    return Sprintf("%- 30.14f",d).removeTrailingBlanks().chop("0");
  else if(n == 1)  
    return Sprintf("%- ld",m).removeTrailingBlanks();
  else
    return Sprintf("%- ld/%ld",m,n).removeTrailingBlanks();
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool String::getLine(FILE *FH,ReadBuffer &rb)
{
  if(FH==NULL) 
    throw Except("String::getLine: File handle == NULL\n");
  
  (*this)="";
  bool stat=true;

  while( 1 ){
/*     fprintf(stderr,"sb_read_buffered: start state: ptr=[%s], current=[%s], len=%ld\n" */
/* 	    ,rb->ptr,rb->current_pos,rb->len);  */
    if(rb.currentPos()[0]=='\0') {

      
      size_t result=0;
      if( ((result=fread(rb.data(),(size_t)1,rb.bufferSize(),FH))==0) &&
	  ferror(FH) ){
	throw  Except("Error reading file \n");
      }
      rb.data()[result]='\0';
      
      if(rb.data()[0]=='\0') {
	if(!this->length()){
	  
	  
	  /* printf("EOF\n"); */
	  stat=false; /* EOF situation */
	  break;
	}else{
	  /* printf("EOF+INFO\n"); */
	  stat=true; /* got some info before end of line occured */
	  break;
	}
      }
      rb.setLength(result);    rb.setCurrentPos(rb.data());
      /*       fprintf(stderr,"sb_read_buffered: read state: ptr=[%s], current=[%s], len=%ld\n" */
      /* 	    ,rb->ptr,rb->current_pos,rb->len);  */
    }
    size_t l=strcspn(rb.currentPos(),"\n");	/* where is the next newline */
    /*     fprintf(stderr,"strcspn l=%ld\n",l); */
    if(l!=rb.length()){
      /* newline in ptr */
      this->ncat(rb.currentPos(),l);
      rb.incrCurrentPos(l+1);
      stat=true; /* return line */
      break;
    }else{
      /* no newline in ptr */
      this->ncat(rb.currentPos(),l);
      rb.incrCurrentPos(l);
    }
  }

  

  return stat;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::ncat(const char *other,size_t len)
{
//   NSF("String &String::ncat(const char *,size_t)");
  size_t Lo=strlen(other);
  size_t l=(len<Lo?len:Lo);


  unsigned long newlength=plength+l;
  
  if(pinternalsize>=newlength+1){
    ::strncat(pvalue,other,l);
    plength=newlength;
    pvalue[newlength]='\0';
//     printf("%s:arg=%s,%ld pinternalsize(%ld), strlen(%ld) pvalue(%s)\n",nsubfunc,
// 	   other,len,pinternalsize,strlen(pvalue),pvalue);
    return *this;
  }

  unsigned long ll=(long)((newlength+1)*1.5);
  pinternalsize=(ll>newlength+1?ll:newlength+1);
  char *temp=new char[pinternalsize];
  strcpy(temp,pvalue);
  ::strncat(temp,other,l);
  temp[newlength]='\0';
  plength=newlength;
  MyMath::swap(this->pvalue,temp);
  delete [] temp;
//   printf("%s:arg=%s,%ld pinternalsize(%ld), strlen(%ld) pvalue(%s)\n",nsubfunc,
// 	 other,len,pinternalsize,strlen(pvalue),pvalue);
  
  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &String::ncopy(const char *other,size_t len)
{
//   NSF("String &String::ncopy(const char *,size_t)");
  size_t L=strlen(other);
  plength=(len<L?len:L);
  if(pinternalsize>=plength+1){
    strncpy(pvalue,other,plength);
    pvalue[plength]='\0';
    return *this;
  }
  pinternalsize=plength+1;
  delete [] pvalue;
  pvalue=new char[pinternalsize];
  strncpy(pvalue,other,plength);
  pvalue[plength]='\0';
  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String String::sprintRat(long nom,long den)
{
  if(nom==0)return "0";
  long D=labs(MyMath::gcd(nom,den));
  long n=nom/D*lSign(1,den);
  long m=labs(den/D);
  if(m==1) return Sprintf("%ld",n);
  return Sprintf("%ld/%ld",n,m);
}


// }}}

// {{{StringList

/* ***********************************************************************
 *
 * ***********************************************************************
 */
StringList::StringList() 
  :psize(0),internalsize(0),parray(0)
{
  //  NSF("StringList()");

}


/* ***********************************************************************
 *
 * ***********************************************************************
 */

StringList::StringList(StringList const& other)
  :psize(0),internalsize(0),parray(0)
{
  //  NSF("StringList(StringList const&)");
  
  grow(other.internalsize);
  for(long i=0;i<other.psize;i++)
    *(parray[i])=*(other.parray[i]);
  
  psize=other.psize;
  

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void StringList::clear()
{
  //printf("StringList: deleteing this %p\n",this);
  if(parray){
    for(long i=0;i<internalsize;i++){
      //printf("StringList: delete parray[%ld]=%p\n",i,parray[i]);
      if(parray[i]) delete parray[i];
    }
  }
  //printf("StringList: deleteing this->parray=%p\n",parray);
  psize=0;internalsize=0; if(parray) delete [] parray;parray=0;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void StringList::grow(long minsize)
{

  //NSF("grow(long)");
  if(minsize>internalsize){
    try{
      
      
      long isize=(internalsize?internalsize*2:1);
      while(isize<minsize){
	isize*=2;
      }

      String **temp=new String*[isize];
     
      if(psize>0){
	for(long i=0; i<psize; i++) temp[i] = parray[i];
      }
      for(long i=psize; i<isize; i++) temp[i] = new String();

      String **c=parray;
      parray=temp;
      temp=c;

      delete [] temp;
      
      internalsize=isize;
      
    }catch (...){
      throw Except("StringList: out of memeory");
    }
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String &StringList::operator[](long index) const
{
  if(index<0 || index>=psize)
    throw Except("StringList: index out of bounds");

  return *(parray[index]);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
const String &StringList::at(long index) const
{
  if(index<0 || index>=psize)
    throw Except("StringList: index out of bounds");

  return *(parray[index]);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
StringList &StringList::append(const String &s)
{
  //  NSF("append(String const&)");
  grow(psize+1);
  *(parray[psize++])=s;
  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String StringList::join(const char *delim,bool delimatend,
			bool delimatfront) const
{
  String s;
  if(delimatfront) s+=delim;
  if(!psize) return s;
  for(long i=0;i<psize;i++){
    s+=*(parray[i]);
    if(i<psize-1) s+=delim;
  }
  if(delimatend) s+=delim;
  return s;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
long StringList::allEmpty()const
{
  long ret=1;
  
  for(long i=0;i<psize;i++)
    ret&=parray[i]->empty();
  
  return ret;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
long StringList::find(String const &needle,long startfrom,bool fromback) const 
{
  if(startfrom<0) return -1;
  if(!fromback) {
    for(long i=startfrom;i<psize;i++){
      if(*(parray[i])==needle) return i;
    }
  }else if(fromback && startfrom<psize){
    for(long i=startfrom;i>=0;i--){
      if(*(parray[i])==needle) return i;
    }
  }
  return -1;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
StringList StringList::findStartWith(String const &needle,long startfrom,
				     bool fromback)const
{
  StringList ret;
  if(startfrom<0) return ret;
  if(!fromback) {
    long j0=startfrom;
    while(j0<psize){
      for(long i=j0;i<psize;i++,j0++){
	if(parray[i]->startsWith(needle)) {
	  ret<<*(parray[i]);
	  break;
	}
      }
      j0++;
    }
    
  }else if(fromback && startfrom<psize){
    long j0=startfrom;
    while(j0>=0){
      for(long i=j0;i>=0;i--,j0--){
	if(parray[i]->startsWith(needle)) {
	  ret<<*(parray[i]);
	  break;
	}
      }
      j0--;
    }
  }
  return ret;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
long StringList::remove(String const &s) 
{
  long i=find(s);

  if(i>=0){
    delete parray[i];
    for(long j=i+1;j<psize;j++){
      parray[j-1]=parray[j];
    }
    parray[psize-1]=new String();
    psize--;
    
    return 1;
  }
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
long StringList::remove(long i) 
{
  if(i>=0){
    delete parray[i];
    for(long j=i+1;j<psize;j++){
      parray[j-1]=parray[j];
    }
    parray[psize-1]=new String();
    psize--;
    
    return 1;
  }
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
StringList const& StringList::operator=(StringList const &other)
{
  //printf("other=");other.dumpinternals();
  StringList temp(other);

  //printf("temp=");temp.dumpinternals();

  String **c=this->parray;
  this->parray=temp.parray;
  temp.parray=c;
  long ps=this->psize;
  this->psize=temp.psize;
  temp.psize=ps;
  ps=this->internalsize;
  this->internalsize=temp.internalsize;
  temp.internalsize=ps;


  //printf("temp=");temp.dumpinternals();


  //printf("*this=");this->dumpinternals();

  return *this;

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
StringList const &StringList::escapeDecode()
{
  for(long i=0;i<psize;i++){
    parray[i]->escapeDecode();
  }
  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
StringList &StringList::interpolateQuotes()
{
  for(long i=-0;++i<psize;){
    parray[i]->interpolateQuotes();
  }
  return *this;
}

// }}}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void StringList::dumpinternals()const
{
  printf("StringList\n");
  printf("\tthis=%p\n",(void*)this);
  printf("\tpsize=%ld\n",psize);
  printf("\tinternalsize=%ld\n",internalsize);
  for(long i=0;i<internalsize;i++){
    printf("\t[l]=");
    parray[i]->dumpinternals();
  }
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void String::dumpinternals()const
{
  printf("String\n");
  printf("\tthis=%p\n",(void*)this);
  printf("\tpstrategy=%d\n",pstrategy);
  printf("\tpinternalsize=%lu\n",pinternalsize);
  printf("\tpvalue=%p=%s\n",pvalue,pvalue);
  printf("\tplength=%lu\n",plength);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String printS(const ValueArray<double> &va,const char *fmt,const char *delim)
{
  String xfmt=Sprintf("%s",fmt).stripBlanks();
  String s="[";
  for(long i=0;i<va.size();i++){
    s+=Sprintf(xfmt(),va.at(i)).stripBlanks();
    if(i<va.size()-1) s+=delim;
  }
  s+="]";
  return s;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String printS(const ValueArray<long> &va,const char *fmt,const char *delim)
{
  String xfmt=Sprintf("%s",fmt).stripBlanks();
  String s="[";
  for(long i=0;i<va.size();i++){
    s+=Sprintf(xfmt(),va.at(i)).stripBlanks();
    if(i<va.size()-1) s+=delim;
  }
  s+="]";
  return s;
}

/*EOF*/ 


