/* @(#)StringPtr.hh
 * author: magru@unas
 */

#ifndef _STRINGPTR_HH
#define _STRINGPTR_HH 1

#include <ctype.h>
#include "FString.hh"
#include "Except.hh"
#include "MACROS.hh"
#include "MyDebug.hh"

/* ***********************************************************************
 *
 * ***********************************************************************
 */

class StringPtr;

// ==================================================================
// use StringPtr to implement a simple lexical analyser, which
// returns Strings, keywords, punctuations/delimiters, numbers, ranges and words 
class StringPtrLexer
{
public:

  // ----------------------------------------------------------------
  enum Tokens{TNone=0,TString=1,TKeyWord=2,TRange=3,TInteger=4,TDouble=5,TWord=6,
	      TPunctuation=7,TComment=8};

  // ----------------------------------------------------------------
  class TokenValue{
  public:
    // ----------------------------------------------------------------
    inline bool isKeyword(const char *which)const
    {
      return (type==TKeyWord && sval==which);
    }
    
    // ----------------------------------------------------------------
    inline bool isKeywords(const StringList &which)const
    {
      if(type!=TKeyWord) return false;
      for(long i=-1;++i<which.size();){
	if(sval==which[i]) return true;
      }
      return false;
    }
    
    // ----------------------------------------------------------------
    inline bool isStringOrWord()const
    {
      return (type==TString || type==TWord);
    }
    // ----------------------------------------------------------------
    inline bool isComment()const
    {
      return (type==TComment);
    }

    // ----------------------------------------------------------------
    inline bool isNumber()const
    {
      return (type==TInteger || type==TDouble);
    }

    // ----------------------------------------------------------------
    inline bool isNone()const
    {
      return (type==TNone);
    }

    // ----------------------------------------------------------------
    inline bool isIntegerOrRanges()const
    {
      return (type==TInteger || type==TRange);
    }

  public:
    long lvalue,rvalue;
    double dvalue;
    String sval;
    long linenr;
    long columnnr;
    Tokens type;
  };

  // ----------------------------------------------------------------
  // ----------------------------------------------------------------
  StringPtrLexer(StringPtr &str,const StringList &keywords,const char *punctuation)
    :pkeepcomments(false)
  {
    pvalue.type=TNone;
    plookaheadvalue.type=TNone;
    pkeywords=&keywords;
    ppunctuation=punctuation;
    pstr=&str;
  }

  // ----------------------------------------------------------------
  // reached end of input
  inline bool eof() const
  {
    return (pvalue.type==TNone);
  }


  // ----------------------------------------------------------------
  void eatPunctuation(const char *what,bool all=false)
  {
    if(all){
      while(plookaheadvalue.type==TPunctuation && plookaheadvalue.sval==what)
	nextToken();
    }else{
      if(plookaheadvalue.type==TPunctuation && plookaheadvalue.sval==what)
	nextToken();
    }
  }


  // ----------------------------------------------------------------
  void nextToken()
  {
    bool prot=false;
    if(plookaheadvalue.type!=TNone){
      if(prot) myDebug()<<"have lookahead";
      pvalue=plookaheadvalue;
      nextTokenInternal(plookaheadvalue,prot);
    }else{
      if(prot) myDebug()<<"no lookahead";
      nextTokenInternal(pvalue);
      nextTokenInternal(plookaheadvalue,prot);
    }
  }
  
private:
  // ----------------------------------------------------------------
  void nextTokenInternal(TokenValue &val,bool prot=false);

public:
  static const char *tokenNames[];
private:
  StringPtr *pstr;
  const StringList *pkeywords;
  const char *ppunctuation;
  ROCPROPERTY(value,value,TokenValue);
  ROCPROPERTY(lookaheadvalue,lookAheadValue,TokenValue);
  bool pkeepcomments;
};


// ==================================================================
// take a pointer to a constant string and provide methods to easily
// disect the string using string primitives. It avoid copying and allocating
class StringPtr{
public:

  // ------------------------------------------------------------
  StringPtr():pstr(0),pend(0),plinenr(1),pcolumnnr(1){
  }
  
  // ------------------------------------------------------------
  StringPtr(const char *p):pstr(p),pend(pstr+strlen(pstr)),plinenr(1),pcolumnnr(1){
  }

  // ------------------------------------------------------------
  StringPtr(const String &s):pstr(s()),pend(pstr+strlen(pstr)),plinenr(1),pcolumnnr(1){
    //myDebug("fromString%ld,%ld <%s>",plinenr,pcolumnnr,String(pstr,20)());
  }

  // ------------------------------------------------------------
  StringPtr(const StringPtr &other)
  {
    *this=other;
  }

  // ------------------------------------------------------------
  StringPtr &operator=(const StringPtr &other)
  {
    pstr=other.pstr;
    pend=pstr+(other.pend-other.pstr);
    plinenr=other.plinenr;
    pcolumnnr=other.pcolumnnr;
    return  *this;
  }

  // ------------------------------------------------------------
  inline char operator[](long i)const{return pstr[i];}
  
  // ------------------------------------------------------------
  // two instances pointing into the same string can be substracted
  // and give the lenght of this string segment
  inline long operator-(const StringPtr &other)const
  {
    return pstr-other.pstr;
  }
  // ------------------------------------------------------------
  inline const char *operator()()const
  {
    return pstr;
  }

  // ------------------------------------------------------------
  inline long length()const{return (pend-pstr+1);}
  
  // ------------------------------------------------------------
  long lineNo()const {return plinenr;}
  // ------------------------------------------------------------
  long columnNo()const {return pcolumnnr;}
  
  // ------------------------------------------------------------
  inline bool eof()const{return (pstr==pend);}

  // ------------------------------------------------------------
  // If the forst length characters in pstr equal needle in lower
  // case return true. If length is not given, strlen is used on needle
  inline bool eqLowerCase(const char *needle,long length=-1)const
  {
    long l=(length>0?length:strlen(needle));
    if(pend-pstr+1<l)return false;
    const char *p=pstr;
    const char *end=pstr+l;
    while(p!=end) if(tolower(*p++)!=*needle++) return false;
    return true;
  }
  
  // ------------------------------------------------------------
  inline StringPtr &skipWhiteSpace()
  {
    long l;
    //myDebug("in%ld,%ld <%s>",plinenr,pcolumnnr,String(pstr,20)());
    do{
      pstr+=(l=strspn(pstr," \t\r"));
      pcolumnnr+=l;
      //myDebug("skipWhiteSpace %ld,%ld <%s>",plinenr,pcolumnnr,String(pstr,20)());
      if(*pstr=='\n'){
	pstr+=(l=1);
	plinenr++;
	pcolumnnr=1;
	//myDebug("skipWhiteSpace %ld,%ld <%s>",plinenr,pcolumnnr,String(pstr,20)());
      }
    }while(l);
    return *this;
  }

  // ------------------------------------------------------------
  inline StringPtr &skipNoneWhiteSpace(const char *skippers=" \t\n\r")
  {
    long l;
    pstr+=(l=strcspn(pstr,skippers));
    pcolumnnr+=l;
    return *this;
  }

  // ------------------------------------------------------------
  inline StringPtr &forward(long skip,bool skiptrailingwhitespace=false){
    //myDebug("forward in %ld,%ld <%s>",plinenr,pcolumnnr,String(pstr,20)());
    pstr+=skip;
    pcolumnnr+=skip;
    //myDebug("forward    %ld,%ld <%s>",plinenr,pcolumnnr,String(pstr,20)());
    if(skiptrailingwhitespace) skipWhiteSpace();
    //myDebug("forward    %ld,%ld <%s>",plinenr,pcolumnnr,String(pstr,20)());
    return *this;
  }

  // ------------------------------------------------------------
  bool punctuationAhead(long &wordlength,const char *ppunctuation)const
  {
    char c[2];c[0]=*pstr;c[1]='\0';
    if(strpbrk(ppunctuation,c)!=NULL){
      wordlength=1;
      return true;
    }
    return false;

  }
  // ------------------------------------------------------------
  bool commentAhead(long &wordlength)const
  {
    if(pstr[0]=='#'){
      long l=strcspn(pstr,"\n");
      wordlength=l;
      return true;
    }
    return false;
  }

  // ------------------------------------------------------------
  // If it is single quoted word we return true
  // and the length of the word including quotes.
  // quotes can be anything in quotes.
  // If it is not such an object wordlength is zero and we return false;
  bool quotedWordAhead(long &wordlength,const char *quotes="'\"")const
  {
    //myDebug()<<"quotedWordAhead: THIS="<<String(pstr,10)+"...";
    long lq=strlen(quotes);
    char c[2];c[1]='\0';
    for(long i=-1;++i<lq;){
      //myDebug()<<"quotedWordAhead: QUOTE="<<quotes[i];
      if(pstr[0]==quotes[i]){
	//myDebug()<<"quotedWordAhead: STARTED=";
	*c=quotes[i];
	const char *end=pstr;

	while(true){
	  const char *p=end+1;
	  end=0;
	  while(*p!='\0' && *p!='\n'){
	    if(*p==c[0]){ end=p;break; }
	    p++;
	  }
	  if(end==NULL){
	    throw Except(Sprintf("Line %ld,%ld: Missing closing quote (%c).\n",
				 plinenr,pcolumnnr,quotes[i]));
	  }
	  //myDebug()<<"quotedWordAhead: PSTR="<<String(pstr,20)+"...";
	  //myDebug()<<"quotedWordAhead: END   ="<<String(end,20)+"...";
	  if(*(end-1)=='\\'){
	    //myDebug()<<"quotedWordAhead: ESCAPED";
	  }else{
	    break;
	  }
	}
	

	wordlength=end-pstr+1;
	return true;
      }
    }
    return false;
  }

  // ------------------------------------------------------------
  // If it is a word we return true
  // and the length of the word.
  // If it is not wordlength is zero and we return false;
  inline bool wordAhead(long &wordlength)const
  {
    if((wordlength=strcspn(pstr," \t\n\r"))>0){
      return true;
    }
    return false;
  }

  // ------------------------------------------------------------
  // scan for keywords
  bool keywordAhead(const StringList &keywords,const char *punctuation,
		    long &wordlength,String &which)const
  {
    // myDebug("keywords in <%s...>",String(pstr,20).replace("\n","\\n",0,true)());
    for(long i=-1;++i<keywords.size();){
      const String &s=keywords.at(i);
      long L=s.length();
      // myDebug("keywords cmp <%s...> with <%s>",
      // 	      String(pstr,20).replace("\n","\\n",0,true)(),
      // 	      s()
      // 	      );
      if((
	  strpbrk(pstr+L,punctuation)-pstr==L || 
	  strpbrk(pstr+L," \t\r\n")-pstr==L 
	  ) && eqLowerCase(s(),L) ){
	which=s;
	wordlength=L;
	// myDebug("keywords out simple rest<%s...>",
	// 	String(pstr+wordlength,20).replace("\n","\\n",0,true)());
	return true;
      }
    }
    wordlength=0;
    // myDebug("keywords out no keyword rest<%s...>",
    //  	    String(pstr+wordlength,20).replace("\n","\\n",0,true)());

    return false;
  
  }

  // ----------------------------------------------------------------
  bool rangeAhead(long &wordlength,const char *punctuation,long &lvalue,long &rvalue)const
  {
    //myDebug("rangeAhead: p=<%s...>",String(pstr,20).replace("\n","\\n",0,true)());
    wordlength=0;
    if(!isdigit(pstr[0])) return false;
    //myDebug("rangeAhead: could be number p=<%s...>",String(pstr,20).replace("\n","\\n",0,true)());
    char *err;
    lvalue=strtol(pstr,&err,10);
    //myDebug("rangeAhead: number err<%s...>",String(err,20).replace("\n","\\n",0,true)());
    if(err==pstr || *err=='\0') return false;
    
    
    //myDebug("rangeAhead: number err<%s...>",String(err,20).replace("\n","\\n",0,true)());
    err+=strspn(err," \t\r");
    //myDebug("rangeAhead: number err<%s...>",String(err,20).replace("\n","\\n",0,true)());
    if(err[0]=='.' && (pend-err>0 && err[1]=='.'))
    {
      //myDebug("rangeAhead: number err<%s...>",String(err,20).replace("\n","\\n",0,true)());
      char *err2;
      err+=2;
      rvalue=strtol(err,&err2,10);
      //myDebug("rangeAhead: number err2<%s...>",String(err2,20).replace("\n","\\n",0,true)());
      //      myDebug()<<"err="<<err<<"err2="<<err2<<"rest="<<strpbrk(err2,punctuation)-err2;
      if(err2==err) return false;
      
      if(*err2=='\0' ||  strpbrk(err2,punctuation)-err2==0 ||
	 strpbrk(err2," \t\r\n")-err2==0 ) {
	wordlength=err2-pstr;
	if(rvalue<lvalue)
	  throw Except(Sprintf("Line %ld,%ld: range "" is invalid\n",plinenr,
			       pcolumnnr,String(pstr,wordlength)()));
	return true;
      }
    }
    return false;
  }


  // ----------------------------------------------------------------
  StringPtrLexer::Tokens numberAhead(long &wordlength,const char *punctuation,
		     double &dvalue,long &lvalue)const
  {
    //myDebug("numberAhead: p=<%s...>",String(pstr,20).replace("\n","\\n",0,true)());
    wordlength=0;
    if(!(isdigit(pstr[0]) || pstr[0]=='-' ||pstr[0]=='+' ||pstr[0]=='.' ) ) 
      return StringPtrLexer::TNone;
    //myDebug("numberAhead: could be number p=<%s...>",String(pstr,20).replace("\n","\\n",0,true)());
    char *err;
    dvalue=strtod(pstr,&err);
    //myDebug("numberAhead: err<%s...>",String(err,20).replace("\n","\\n",0,true)());
    if(err==pstr) {
      wordlength=0;
      return StringPtrLexer::TNone;
    }
    if(*err=='\0' || strpbrk(err," \t\r\n")-err==0 || strpbrk(err,punctuation)-err==0){
      char *err2;
      lvalue=strtol(pstr,&err2,10);
      wordlength=err-pstr;
      // myDebug("numberAhead: end of string and valid pstr=<%s>",
      // 	      String(pstr,wordlength)());
      if(err2==err) return StringPtrLexer::TInteger;
      return StringPtrLexer::TDouble;
    }
    return StringPtrLexer::TNone;
  }

  // ----------------------------------------------------------------
  static StringList toStringList(const String s,bool unravelprotecredquotes=true)
  {// return stringlist from String, respecting quoted stuff
    StringPtr sp(s);
    StringPtrLexer pl(sp,StringList(),"");
    StringList sl;
    do{
      pl.nextToken();
      if(!pl.eof())
	sl<<pl.value().sval;
    }while(!pl.eof());
    if(unravelprotecredquotes){
      return sl.interpolateQuotes();
    }else{
      return sl;
    }
  }





private:

private:
  const char *pstr;
  const char *pend;
  long plinenr;
  long pcolumnnr;
public:
};



// ==================================================================
inline MyDebug operator<<(MyDebug dbg,const StringPtr &p){
  dbg<<p();
  return dbg;
}


#endif /* _STRINGPTR_HH */

