/*************************************************************************
 * @(#)ParamDriver.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   04 Jun 2011
 *************************************************************************
 */

#include <iostream>
#include <sstream>
#include "Except.hh"
#include "MyQRgb.hh"
#include "WeightLabels.hh"
#include "ParamDriver.hh"
#include "ParamParser.tab.hh"
#include "Attributes.hh"
#include "FunctionGroup.hh"
#include "FunctionGroups.hh"
#include "Page.hh"
#include "Pair.hh"
#include "Graph.hh"
#include "Graphs.hh"
#include "ViewBox.hh"
#include "TextBox.hh"
#include "LegendBox.hh"
#include "TicMarks.hh"
#include "WeightSelection.hh"
//#include "iofunc.hh"
#include "graphview.hh"
#include "MyDebug.hh"





/* ***********************************************************************
 *
 * ***********************************************************************
 */
ParamDriver::ParamDriver()
  : ptracescanning(false),ptraceparsing(false),pscanner(NULL)
  ,pcurrentfontstyle(0)
  ,pcurrentlinestyle(0)
  ,pcurrentfillstyle(0)
  ,pneedrefresh(false)
  ,pfuncindex(0),pkilled(false),peof(false)
{
  Pair<String,TempSetDescriptor> *p;
  if((p=pvariables.find(Pair<String,TempSetDescriptor>("pi")))) 
    p->v2=3.14159265358979323844;
  else
    pvariables
      .append(Pair<String,
	      TempSetDescriptor>("pi",TempSetDescriptor(3.14159265358979323844)));
  ptieins.resize(TIE_DATATYPE+1);
  ptieins=false;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
ParamDriver::~ParamDriver()
{
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::resizeGraphView()
{
  pgv->updatePageSize();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
int ParamDriver::parse(const char *bufferorfilename,bool frombuffer)
{
  peof=false;
  pneedrefresh=false;

  pfrombuffer=frombuffer;
  if(!pfrombuffer){
    pfilename = bufferorfilename;
    pbuffer=0;
  }else{
    pfilename = "buffer";
    pbuffer=bufferorfilename;
  }
  
  scan_begin();
  PP::ParamParser parser(*this,pscanner);
  parser.set_debug_level(ptraceparsing);
  pkilled=false;
  String msg; 
  try{
    parser.parse();
  }catch(Except &ex){
    msg=ex.msg();
  };
  scan_end();
  
  if(pkilled){
    if(ppage->graphs()->size()==0){
      //we need at least one graph
      ppage->graphs()->getById(0,true);
    }
  }
  ppage->updateWeightLabelReferences();
  ppage->normalizeWeightSelections();


  if(msg.length()>0){
    throw Except(msg);

  }
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::error(const PP::location& l,const std::string& m)const
{
  std::ostringstream oss (std::ostringstream::out);
  oss << l << ": " << m << std::endl;
  throw  Except(Sprintf("%s",oss.str().c_str()));
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::error (const std::string& m)const
{
  throw  Except(Sprintf("%s\n",m.c_str()));
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::setCurrentWeight(const String *w)
{
  FunctionGroup *group=getGroup();
  WeightSelections *wss=0;
  if(psetd.setid<0){
    if(group) wss=group->weightSelections();
    //todo create wss if needed
  }else{
    Function *func=getSet();
    if(func) wss=func->weightSelections();
    //todo create wss if needed
  }
  if(wss){
    psetd.weightid=wss->indexOf(*w);
    if(psetd.weightid<0)
      error(ploc,Sprintf("weight with name '%s' does not exist",(*w)())());
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Graph *ParamDriver::getGraph(const SetDescriptor *sd,bool append)
{
  if(!ppage) return 0;
  const SetDescriptor *s=sd;
  if(s==0)s=&psetd;
  
  if(s->graphid<0)
    error(ploc,Sprintf("invalid graph number %ld",s->graphid+1)());
  
  Graphs *graphs=ppage->graphs();
  Graph *gr=0;
  if(!(gr=graphs->getById(s->graphid,append))){
    error(ploc,Sprintf("invalid graph number %ld",s->graphid+1)());
  }
  return gr;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
FunctionGroup *ParamDriver::getGroup(const SetDescriptor *sd,bool append)
{
  const SetDescriptor *s=sd;
  if(s==0)s=&psetd;

  Graph *graph=getGraph(s,append);
  if(!graph) return 0;

  if(s->groupid<0)
    error(ploc,Sprintf("invalid group number %ld",s->groupid+1)());

  FunctionGroups *groups=graph->groups();
  FunctionGroup  *fg=0;
  
  if(!(fg=groups->getById(s->groupid,append))){
    error(ploc,Sprintf("invalid group number %ld",s->groupid+1)());
  }
  return fg;



 }

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Function *ParamDriver::getSet(const SetDescriptor *sd,bool append,
			      bool complain)
{
  const SetDescriptor *s=sd;
  if(s==0)s=&psetd;

  FunctionGroup *group=getGroup(s,append);
  if(!group) return 0;

  if(s->setid<0){
    if(complain){
      error(ploc,Sprintf("invalid set number %ld",s->setid+1)());
    }else{
      return 0;
    }
  }

  Functions *funcs=group->funcs();
  Function  *fu=0;

  if(!(fu=funcs->getById(s->setid,append))){
    error(ploc,Sprintf("invalid set number %ld",s->setid+1)());
  }
  pcurrentfunc=fu;
  return fu;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

FunctionAttribs *ParamDriver::getFuncAttribs()
{
  FunctionGroup *group=getGroup();
  if(psetd.setid<0){
    if(group) return group->attribs();
  }else{
    Function*func=getSet();
    if(func) return func->attribs();
  }
  
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
FGObject *ParamDriver::getFGObject()
{
  FunctionGroup *group=getGroup();
  if(psetd.setid<0){
    if(group) return group;
  }else{
    Function*func=getSet();
    if(func) return func;
  }
  
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Tics *ParamDriver::getUserTics()
{
  Graph *graph=getGraph();
  if(psetd.groupid<0){
    if(graph) return graph->ticMarks()->userTics();
  }else{
    FunctionGroup *group=getGroup();
    if(group) return group->userTics();
  }
  
  return 0;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
WeightSelections *ParamDriver::
getWeightSelections(const SetDescriptor *sd)
{
  const SetDescriptor *s=sd;
  if(s==0)s=&psetd;

  FunctionGroup *group=getGroup();
  if(!group) return 0;
  if(s->setid<0){
    return group->weightSelections();
  }else{
    Function *func=group->funcs()->getById(s->setid,false);
    if(!func) return 0;
    return func->weightSelections();
  }
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
WeightSelection *ParamDriver::
getWeightSelection(const SetDescriptor *sd,bool append)
{
  const SetDescriptor *s=sd;
  if(s==0)s=&psetd;

  WeightSelections *wss=getWeightSelections(s);
  if(!wss) return 0;
  //myDebug() << "weightid "<<s->weightid<<wss->size()<<append;
  if(append){
    if(s->weightid>=wss->size()) 
      wss->resize(s->weightid+1);
  }
  
  checkCurrentWeight(s,wss);
  return (*wss)[s->weightid];
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
LineShape *ParamDriver::getLineShape(long index,bool append)
{
  Graph *graph=getGraph();
  LineShape *lsh=0;
  Shapes *shapes=graph->shapes();
  if(!(lsh=(LineShape*)shapes->getByIndex(index,Xfp::SHTLine))){
    // line does not exist
    if(!append){
      error(ploc,Sprintf("invalid line number %ld",index+1)());
    }else{
      // create one with the desired index
      lsh=new LineShape(graph,index);
      shapes->insertShape(shapes->size(),lsh);
    }
  }
  return lsh;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
EllipseShape *ParamDriver::getEllipseShape(long index,bool append)
{
  Graph *graph=getGraph();
  EllipseShape *esh=0;
  Shapes *shapes=graph->shapes();
  if(!(esh=(EllipseShape*)shapes->getByIndex(index,Xfp::SHTEllipse))){
    // ellipse does not exist
    if(!append){
      error(ploc,Sprintf("invalid ellipse number %ld",index+1)());
    }else{
      // create one with the desired index
      esh=new EllipseShape(graph,index);
      shapes->insertShape(shapes->size(),esh);
      esh->setExtraFillColor(false);
    }
  }
  return esh;

}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::readFile(const String *file,Xfp::FileType ftype,
			   const SetDescriptor *sd)
{
  if(!pgv) return;
  //printf("READ %s type %d sd=%p\n",file(),ftype,sd);
  
  if(!sd){
    // read into current graph
    getGraph();
    pgv->import(true,(*file)(),ftype,pfuncindex,psetd.graphid);
  }else{
    if(sd->groupid<0){
      getGraph(sd,true);
      pgv->import(true,(*file)(),ftype,pfuncindex,sd->graphid);
    }else{
      if(ftype==Xfp::FTBandWeight|| ftype==Xfp::FTBand){
	error(ploc,"band structure data cannot be read into existing groups");
      }
      getGroup(sd,true);
      pgv->import(true,(*file)(),ftype,pfuncindex,sd->graphid,sd->groupid);
    }
  }
    

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::newGraph()
{
  if(!ppage) return;
  ppage->graphs()->insertGraph(ppage->graphs()->size());
}




/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::checkCurrentWeight(const SetDescriptor *sd,
				     WeightSelections *wss)
{
  if(!wss) return;
  //myDebug() << "sd->weightid "<<sd->weightid<<wss->size();
  if(sd->weightid<0 || sd->weightid>=wss->size()) 
    error(ploc,Sprintf("invalid weight id number %ld",sd->weightid+1)());
}




/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::switchOnOff(bool onoff)
{
  if(psetd.weightid>=0){
    WeightSelection *ws=getWeightSelection(0,true);
    if(ws) ws->setActive(onoff);
    return;
  }


  Graph *graph=getGraph(&psetd,true);
  if(psetd.groupid<0 && psetd.setid<0){
    pfuncindex=0;
    graph->setActive(onoff);
    return;
  }

  FunctionGroup *group=getGroup(&psetd,true);
  if(psetd.setid<0){
    if(group)
      group->setActive(onoff);
  }else{
    Function *func=getSet(&psetd,true);
    if(func)
      func->setActive(onoff);
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::toggle()
{
  Graph *graph=getGraph(&psetd,true);
  if(psetd.groupid<0 && psetd.setid<0){
    pfuncindex=0;
    graph->setActive(!graph->active());
    return;
  }

  FunctionGroup *group=getGroup(&psetd,true);
  if(psetd.setid<0){
    if(group)
      group->setActive(!group->active());
  }else{
    Function *func=getSet(&psetd,true);
    if(func)
      func->setActive(!func->active());
  }
}





/* ***********************************************************************
 *
 * ***********************************************************************
 */
// void ParamDriver::setWeightFactor(double f)
// {
//   WeightSelections *wss=getWeightSelections(&psetd);
//   if(wss){
//     wss->setFactor(f);
//   }

// }


/* ***********************************************************************
 *
 * ***********************************************************************
 */
// void ParamDriver::setWeightMaxRadius(double r)
// {
//   WeightSelections *wss=getWeightSelections(&psetd);
//   if(wss){
//     wss->setMaxWidth(r);
//   }
// }


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::copyFromSetDescriptor(TempSetDescriptor *tempsd)
{
  //myDebug()<<"copyFromSetDescriptor=";
  FunctionGroup *group=getGroup();
  Function *func=getSet(0,false,false);
  if(!func){
    if(!group)return;
    //myDebug()<<"group="<<group;
    long N=group->funcs()->size();
    tempsd->grdata().resize(N,true);
    for(long i=-1;++i<N;){
      ValueArray<double> &dd=tempsd->grdata()[i];
      const Function *func=group->funcs()->at(i);
      dd.resize(func->length(),true);
      for(int j=0;j<dd.size();j++){
	dd[j]=func->data(psetd.xy,j);
      }
    }
  }else{
    tempsd->xdata().resize(func->length(),true);
    for(int i=0;i<tempsd->xdata().size();i++){
      tempsd->xdata()[i]=func->data(psetd.xy,i);
    }
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
TempSetDescriptor *ParamDriver::cloneVariable(const String *v)
{
  String lv=v->lowerCase();
  //myDebug()<<"cloneVariable"<<lv;
  Pair<String,TempSetDescriptor> *p;
  if((p=pvariables.find(Pair<String,TempSetDescriptor>(lv)))){
    //myDebug()<<"cloneVariable"<<lv<<p->v2;
    return new TempSetDescriptor(p->v2);
  }else{
    error(ploc,Sprintf("unknown variable '%s'",lv())());
  }
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
const TempSetDescriptor *ParamDriver::getVariable(const String *v)const
{
  String lv=v->lowerCase();
  Pair<String,TempSetDescriptor> *p;
  if((p=pvariables.find(Pair<String,TempSetDescriptor>(lv)))){
    return &(p->v2);
  }else{
    error(ploc,Sprintf("cannot find variable '%s'",lv())());
  }
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
double ParamDriver::getVariableValue(const String *v)
{
  String lv=v->lowerCase();
  Pair<String,TempSetDescriptor> *p;
  if((p=pvariables.find(Pair<String,TempSetDescriptor>(lv)))){
    if(p->v2.type()!=TempSetDescriptor::Value)
      error(ploc,Sprintf("variable '%s' is not scalar",lv())());
    return p->v2.value();
  }else{
    error(ploc,Sprintf("cannot find variable '%s'",lv())());
  }
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::setVariable(const String *v,const TempSetDescriptor &tempsd)
{
  String lv=v->lowerCase();
  //myDebug()<<"setVariable"<<lv<<tempsd;
  Pair<String,TempSetDescriptor> *p;
  if((p=pvariables.find(Pair<String,TempSetDescriptor>(lv)))){
    p->v2=tempsd;
    //myDebug()<<"setVariable"<<lv<<p->v2;
  }else{
    pvariables.append(Pair<String,TempSetDescriptor>(lv,tempsd));
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::setFromTemp(const TempSetDescriptor *tempsd)
{
  //myDebug()<<"setFromTemp"<<tempsd->type();
  FunctionGroup *group=getGroup();
  Function *func=getSet(0,false,false);
  //myDebug()<<"group="<<group;
  //myDebug()<<"func="<<func;
  if(!func){
    if(!group)return;
    long N=group->funcs()->size();
    //myDebug()<<"N="<<N;
    if(tempsd->type()==TempSetDescriptor::VectorSet){
      if(N!=tempsd->grdata().size())
	error(ploc,Sprintf("cannot copy/assign groups of different size")());
    }    
    for(long i=-1;++i<N;){
      Function *func=group->funcs()->at(i);
      if(tempsd->type()==TempSetDescriptor::Value){
	//myDebug()<<"func="<<i<<func<<"="<<tempsd->value();
	for(int j=0;j<func->length();j++){
	  func->setData(psetd.xy,j,tempsd->value());
	}
      }else if(tempsd->type()==TempSetDescriptor::Vector){
	//myDebug()<<"func="<<i<<func<<"=vector";

	if(func->length()!=tempsd->xdata().size()){
	  error(ploc,Sprintf("cannot copy/assign vectors of different length")());
	}
	for(int i=0;i<tempsd->xdata().size();i++){
	  func->setData(psetd.xy,i,tempsd->xdata().at(i));
	}
      }else if(tempsd->type()==TempSetDescriptor::VectorSet){
	const ValueArray<double> &dd=tempsd->grdata().at(i);
	//myDebug()<<"func="<<i<<func<<"="<<dd;
	if(func->length()!=dd.size()){
	  error(ploc,Sprintf("cannot copy/assign vectors of different length")());
	}
	for(int j=0;j<dd.size();j++){
	  func->setData(psetd.xy,j,dd.at(j));
	}
      }else{
	error(ploc,Sprintf("setFromTemp(TempSetDescriptor): type not yet implemented!")());
      }
    }
    

  }else{
    //myDebug()<< "func"<<func<<tempsd->type();
    if(tempsd->type()==TempSetDescriptor::Value){
      for(int i=0;i<func->length();i++){
	//myDebug()<< "func"<<i<<"="<<tempsd->value();
	func->setData(psetd.xy,i,tempsd->value());
      }
    }else if(tempsd->type()==TempSetDescriptor::Vector){
      if(func->length()!=tempsd->xdata().size()){
	error(ploc,Sprintf("cannot copy/assign vectors of different length")());
      }
      for(int i=0;i<tempsd->xdata().size();i++){
	func->setData(psetd.xy,i,tempsd->xdata().at(i));
      }
    }else{
      error(ploc,"cannot assign set of vectors to vector!");
    }
  }
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
// void ParamDriver::appendWeight(const SetDescriptor *sd,
// 			       const std::string *name)
// {

//   if(sd->setid<0){
//     //oops
//     error(ploc,"not implemented");
//     // group
//     FunctionGroup *group=getGroup(sd);
//     if(!group) return;
//     WeightLabels *wls=group->weightLabels();
//     if(!wls){
//       error(ploc,"no common weight definitions for this group");
//     }
//     long nf=group->funcs()->size();
//     if(!nf){
//       error(ploc,"no sets in this group: no weight added");
//     }
//     Function *func0=group->funcs()->at(0);
//     int nw=func0->components();
//     int nl=func0->length();
//     for(int iff=1;iff<nf;iff++){
//       if(group->funcs()->at(iff)->components()!=nw){
// 	error(ploc,"sets in this group do not have the same number of weights");
//       } 
//       if(group->funcs()->at(iff)->length()!=nl){
// 	error(ploc,"sets in this group do not have the same length");
//       }
//     }

//     func0->weightLabels()->append(name->c_str());
//     wls->append(new WeightDefinition(WeightDefinition::Simple,name->c_str()));
    

//     for(int iff=1;iff<nf;iff++){
//       Function *func=group->funcs()->at(iff);
//       func->setDataType(Function::XYNW);
//       func->adjustComp(nw+1,false);
//       func->weightSelections()
// 	->append(new WeightSelection(wds->last(),false,0xff00ff));
//       for(int i=0;i<func->length();i++){
// 	func->setData(nw,i,1);
//       }
//     }



//   }else{
//     Function *func=getSet(sd);
//     if(!func) return;
    
//     long iw=func->components();
    
//     if(func->dataType()==Function::XY){
//       func->setDataType(Function::XYNW);
//       func->setWeightLabels(ppage->weightLabelsCollector()->demandNew());
//       func->setWeightDefinitions(new WeightDefinitions());
//     }else if(func->dataType()==Function::XYNW){
//     }else{
//       error(ploc,"cannot append weights to a function of this type");
//     }
    
//     func->adjustComp(iw+1,false);
    
//     func->weightLabels()->append(name->c_str());
//     WeightDefinitions *wds=func->weightDefinitions();
    
//     wds->append(new WeightDefinition(WeightDefinition::Simple,name->c_str()));
    
//     WeightSelections *wss=func->weightSelections();
//     wss->append(new WeightSelection(wds->last(),true,0xff00ff));
    
//     for(int i=0;i<func->length();i++){
//       func->setData(iw,i,1);
//     }
//   }
// }







/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::checkTextBox(long tbt,bool allownew)
{
  Graph *graph=getGraph();
  if(tbt<TextBox::TBTTitle){
    error(ploc,"textbox does not exist");
  }else if(tbt>=graph->textBoxes()->size()){
    if(!allownew){
      error(ploc,Sprintf("textbox %d does not exist",
			 tbt+1-TextBox::TBTGeneral)());
    }else{
      long n=graph->textBoxes()->size();
      graph->textBoxes()->adjustSizeToIndex(tbt+1);
      for(long i=n;i<graph->textBoxes()->size();i++){
	graph->textBoxes()->at(i)->setActive(false);
      }
    }
  }
}





/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::convolute(double width)
{
  Function *func=getSet();
  if(!func) return;
  if(func->dataType()!=Function::XY){
    error(ploc,Sprintf("set must have xy data type")());
  }
  func->convolute(width);

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::resetPage()
{
  if(!ppage) return;
  pkilled=true;
  ppage->reset(false);//reset without creating graph 1
  pneedrefresh=true;
}






/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ParamDriver::copySet(bool move)
{
  SetDescriptor dest=psetd;
  popCurrentSetDescriptor();
  const SetDescriptor &src=psetd;
  if(src.isSet() && dest.isSet()){
    Function *fdest=getSet(&dest,true);
    Function *fsrc=getSet(&src,true);
    fdest->copyFrom(fsrc);

    if(move){
      FunctionGroup *group=getGroup(&src);
      group->funcs()->take(fsrc);
    }
    //fdest->reParent(getGroup());

  }else{
    error(ploc,"'copy/move set to set' needs two sets "
	  "descriptors as arguments  ");
  }
}
  
/*EOF*/ 
