/*************************************************************************
 * @(#)ObserveView.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   26 Sep 2011
 *************************************************************************
 */

#include <QComboBox>
#include <QGridLayout>
#include <QFileDialog>
#include <QHeaderView>
#include <QDoubleSpinBox>
#include <QPlainTextEdit>
#include <QSplitter>
#include <QCheckBox>
#include <QTimer>
#include <QLabel>
#include <QLineEdit>
#include <QMessageBox>
#include <QPushButton>
#include <QSpinBox>
#include <QTreeWidget>
#include "ObserveView.hh"
#include "graphview.hh"
#include "Page.hh"
#include "World.hh"
#include "Graphs.hh"
#include "GraphAxis.hh"
#include "Attributes.hh"
#include "GraphObject.hh"
#include "FGObject.hh"
#include "Function.hh"
#include "Functions.hh"
#include "FunctionGroup.hh"
#include "Settings.hh"
#include "FunctionGroups.hh"
#include "OutGrep.hh"
#include "TextBox.hh"
#include "MainWindow.hh"
#include "Except.hh"
#include "Parser.hh"
#include "Config.hh"
#include "defs.hh"
#include "MyDebug.hh"



#define VALENCEORBITAL 3
#define COREORBITAL 4
#define KINVALENCEORBITAL 5
#define KINCOREORBITAL 6

class MyHBoxLayout:public QHBoxLayout
{
public:
  MyHBoxLayout():QHBoxLayout(){};
  MyHBoxLayout(QWidget *p):QHBoxLayout(p){};
  virtual ~MyHBoxLayout(){
    myDebug() << "MyHBoxLayout dying";
  };
};


/* ***********************************************************************
 *
 * ***********************************************************************
 */
ObserveView::ObserveView(MainWindow *parent)
  :QObject(parent),p_gv(0),parentwidget(0),
   trwOrbitals(0),lSelect(0),cmbMode(0),lOutputFile(0),leOutFile(0),
   pbBrowse(0),lSite(0),cmbSite(0),pbUpdate(0),updateTimer(0),lIter0(0),
   spbxIter0(0),cbAutoUpdate(0),spbxAutoUpdateTime(0),plaintextedit(0),
   splitter(0),dspbxRmax(0),lRmax(0),rmax(0.),
   pobserveoutfile(""),
   muteIterationMode(false),pautoupdatetime(5000)
{
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
ObserveView::~ObserveView()
{
  //  myDebug() << "ObserveView::~ObserveView";
  Settings set("ObserveView");
  if(spbxAutoUpdateTime){
    set.set("autoupdate/time",spbxAutoUpdateTime->value()*1000);
  }

  cleanUp();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::setUp(QWidget *widget,GraphView *gv,Mode what)
{
  parentwidget=widget;
  pwhat=what;
  p_gv=gv;
  p_gv->killAll();

  updateTimer=new QTimer(parentwidget);
  connect(updateTimer, SIGNAL(timeout()), this, SLOT(onUpdateTimer()));

  if(parentwidget->layout()) delete parentwidget->layout();


  if(pwhat==Orbitals){
    
    QGridLayout *layout=new QGridLayout(parentwidget);
    
    int row=0;
    int col=0;
    
    
    lSelect=new QLabel(parentwidget);
    lSelect->setText("Se&lect");
    
    cmbMode=new QComboBox(parentwidget);
    lSelect->setBuddy(cmbMode);
    
    cmbMode->insertItem(0,"Valence orbitals",VALENCEORBITAL);
    cmbMode->insertItem(1,"Core orbitals",COREORBITAL);
    cmbMode->insertItem(3,"Kinetic Valence orbitals",KINVALENCEORBITAL);
    cmbMode->insertItem(3,"Kinetic Core orbitals",KINCOREORBITAL);

    connect(cmbMode,SIGNAL(currentIndexChanged(int)),
  	    this,SLOT(orbitalMode()));
    
    
    QSpacerItem *spacer=new QSpacerItem(20,20,QSizePolicy::Expanding,
					QSizePolicy::Minimum);
    
    
    QHBoxLayout *hlayout=new QHBoxLayout();
    hlayout->addWidget(lSelect);
    hlayout->addWidget(cmbMode);
    hlayout->addItem(spacer);

    lRmax=new QLabel(parentwidget);
    dspbxRmax=new QDoubleSpinBox(parentwidget);
    dspbxRmax->setValue(rmax);
    lRmax->setText("&Rmax");
    lRmax->setBuddy(dspbxRmax);
    hlayout->addWidget(lRmax);
    hlayout->addWidget(dspbxRmax);
    spacer=new QSpacerItem(20,20,QSizePolicy::Expanding,
					QSizePolicy::Minimum);
    hlayout->addItem(spacer);
    hlayout->setStretch(5,10);
    connect(dspbxRmax,SIGNAL(valueChanged(double )),
  	    this,SLOT(onDspbxRmaxValueChanged(double )));


    layout->addLayout(hlayout, row,col++,1,2);


    row++;
    col=0;
    

    trwOrbitals=new QTreeWidget(parentwidget);
    QSizePolicy pol=trwOrbitals->sizePolicy();
    pol.setHorizontalPolicy(QSizePolicy::Minimum);
    trwOrbitals->setSizePolicy(pol);
    trwOrbitals->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    trwOrbitals->setMinimumSize(QSize(150,0));
    layout->addWidget(trwOrbitals, row,col, 1, 1);
    layout->setColumnStretch (col++,1);

    connect(trwOrbitals,SIGNAL(currentItemChanged(QTreeWidgetItem *,
						  QTreeWidgetItem *)),
  	    this,SLOT(onTrwOrbitalsCurrentItemChanged(QTreeWidgetItem *,
						      QTreeWidgetItem *)));
    
    
    layout->addWidget(p_gv, row,col, 1, 1);
    layout->setColumnStretch (col++,100);

    trwOrbitals->setFocus();

    orbitalMode();
  }else if(pwhat==Iteration){

    QVBoxLayout *layout=new QVBoxLayout(parentwidget);
 

    lSelect=new QLabel(parentwidget);
    lSelect->setText("Se&lect");

    cmbMode=new QComboBox(parentwidget);
    lSelect->setBuddy(cmbMode);
    
    cmbMode->insertItem(0,"iteration",OutGrep::Iteration);
    cmbMode->insertItem(1,"iteration best",OutGrep::IterationBest);
    cmbMode->insertItem(2,OutGrep::modeName(OutGrep::TotalEnergy,true),
			OutGrep::TotalEnergy);
    cmbMode->insertItem(3,OutGrep::modeName(OutGrep::TotalSpin,true),
			OutGrep::TotalSpin);
    cmbMode->insertItem(4,OutGrep::modeName(OutGrep::CycleTime,true),
			OutGrep::CycleTime);
    cmbMode->insertItem(5,"atom spin",OutGrep::AtomSpinGros);
    cmbMode->insertItem(6,"atom occupation",OutGrep::AtomOccuGros);
    cmbMode->insertItem(7,"atom Lz",OutGrep::AtomOrbitalMoment);
    cmbMode->insertItem(8,"atom EFG",OutGrep::AtomEFG);
    cmbMode->insertItem(9,OutGrep::modeName(OutGrep::Bfsm,true),
			OutGrep::Bfsm);
    cmbMode->insertItem(10,OutGrep::modeName(OutGrep::dBfsm,true),
			OutGrep::dBfsm);
    cmbMode->insertItem(11,"force iteration",OutGrep::ForceIteration);
    connect(cmbMode,SIGNAL(currentIndexChanged(int)),
 	    this,SLOT(iterationMode()));
     
    

      
    QSpacerItem *spacer=new QSpacerItem(0,20,QSizePolicy::Minimum,
 					QSizePolicy::Minimum);


    lOutputFile=new QLabel(parentwidget);
    lOutputFile->setText("&Output file");

    leOutFile=new QLineEdit(parentwidget);
    leOutFile->setToolTip("edit and hit ENTER");

    leOutFile->setText(pobserveoutfile());
    connect(leOutFile,SIGNAL(returnPressed()),
 	    this,SLOT(onPbUpdateClicked()));
    lOutputFile->setBuddy(leOutFile);


    pbBrowse=new QPushButton(parentwidget);
    pbBrowse->setText("&Browse");
    connect(pbBrowse,SIGNAL(clicked()),this,SLOT(onPbBrowseClicked()));

    pbUpdate=new QPushButton(parentwidget);
    pbUpdate->setText("&Update");
    pbUpdate->setDefault(true);
    connect(pbUpdate,SIGNAL(clicked()),this,SLOT(onPbUpdateClicked()));


    QSpacerItem *spacer1=new QSpacerItem(0,20,QSizePolicy::Expanding,
			   QSizePolicy::Minimum);


    QHBoxLayout *hlayout=new QHBoxLayout();
    hlayout->addWidget(lSelect);
    hlayout->addWidget(cmbMode);
    hlayout->addItem(spacer);
    hlayout->addWidget(lOutputFile);
    hlayout->addWidget(leOutFile);
    hlayout->addWidget(pbBrowse);
    hlayout->addWidget(pbUpdate);
    hlayout->addItem(spacer1);
    layout->addLayout(hlayout);



    lSite=new QLabel(parentwidget);
    lSite->setText("Si&te");

    cmbSite=new QComboBox(parentwidget);
    lSite->setBuddy(cmbSite);
    connect(cmbSite,SIGNAL(currentIndexChanged(int)),
  	    this,SLOT(iterationMode()));

    lIter0=new QLabel(parentwidget);
    lIter0->setText("Sta&rt at");

    spbxIter0=new QSpinBox(parentwidget);
    lIter0->setBuddy(spbxIter0);
    spbxIter0->setMinimum(0);
    spbxIter0->setMaximum(100000);
    spbxIter0->setSingleStep(5);
    spbxIter0->setValue(0);

    connect(spbxIter0,SIGNAL(valueChanged(int)),
	    this,SLOT(onSpbxIter0ValueChanged(int)));

    cbAutoUpdate=new QCheckBox(parentwidget);
    cbAutoUpdate->setText("Auto Update");
    cbAutoUpdate->setChecked(true);
    connect(cbAutoUpdate,SIGNAL(stateChanged(int)),
	    this,SLOT(onCbAutoUpdateStatechanged(int)));

    Settings set("ObserveView");
    pautoupdatetime=set.getInt("autoupdate/time",5000);

    spbxAutoUpdateTime=new QSpinBox(parentwidget);
    //lIter0->setBuddy(spbxIter0);
    spbxAutoUpdateTime->setMinimum(1);
    spbxAutoUpdateTime->setMaximum(100000);
    spbxAutoUpdateTime->setSingleStep(3);
    spbxAutoUpdateTime->setValue(pautoupdatetime/1000);
    spbxAutoUpdateTime->setSuffix(" sec");

    connect(spbxAutoUpdateTime,SIGNAL(valueChanged(int)),
	    this,SLOT(onSpbxAutoUpdateTimeValueChanged(int)));

    QSpacerItem *spacer2=new QSpacerItem(0,0,QSizePolicy::Expanding,
					 QSizePolicy::Minimum);
    hlayout=new QHBoxLayout();
    hlayout->addWidget(lSite);
    hlayout->addWidget(cmbSite);
    hlayout->addWidget(lIter0);
    hlayout->addWidget(spbxIter0);
    hlayout->addWidget(cbAutoUpdate);
    hlayout->addWidget(spbxAutoUpdateTime);
    hlayout->addItem(spacer2);
    layout->addLayout(hlayout);

    layout->addWidget(p_gv);

//     splitter = new QSplitter(parentwidget);
//     splitter->setOrientation(Qt::Vertical);
//     layout->addWidget(splitter);
//     splitter->addWidget(p_gv);
    
    
    spacer=new QSpacerItem(20,0,QSizePolicy::Minimum,
 			   QSizePolicy::Expanding);
//     plaintextedit=new QPlainTextEdit(parentwidget);
//     splitter->addWidget(plaintextedit);

    layout->addItem(spacer);
    
    layout->setStretch (2,100);

    leOutFile->setFocus();

//     splitter->setStretchFactor(0,3);
//     splitter->setStretchFactor(1,1);


    iterationMode();
  }

  
  //myDebug() << "started";
  //updateTimer->start(25);

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::onCbAutoUpdateStatechanged(int state)
{

  if(state==Qt::Unchecked)
    updateTimer->stop();
  else 
    updateTimer->start(pautoupdatetime);
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::onSpbxIter0ValueChanged(int)
{
  //myDebug() << value;
  iterationMode();
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::onSpbxAutoUpdateTimeValueChanged(int value)
{

  //myDebug()<<"ObserveView::onSpbxAutoUpdateTimeValueChanged"<<value;
  pautoupdatetime=value*1000;
  updateTimer->stop();
  updateTimer->start(pautoupdatetime);

  Settings set("ObserveView");
  set.set("autoupdate/time",pautoupdatetime);

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::onUpdateTimer()
{
  //myDebug() << "onUpdateTimer";
  onPbUpdateClicked();

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::onPbUpdateClicked()
{
  //myDebug()<<"heir";
  pobserveoutfile="";
  iterationMode();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

void ObserveView::orbitalMode()
{
  updateTimer->stop();
  trwOrbitals->clear();
  QStringList headers;
  headers<<"Sort"<<"Ele."<<"Comp.";
  trwOrbitals->setHeaderLabels(headers);
  QHeaderView *hv=trwOrbitals->header();
  hv->setDefaultSectionSize(50);
  hv->setResizeMode(QHeaderView::Stretch);
  trwOrbitals->setRootIsDecorated(true);
  
  int ncomp=1;
  
  int relat=0;
  
  QTreeWidgetItem *item0=0;
  
  Parser par(1);
  try{
    //todo: current directory
    par.parseFile("=.in");
    String s;
    try{
      relat=par()("relativistic")("type").valueL();
      if(relat==2){
	ncomp=4;
      }else if(relat==3){
	ncomp=2;
      }else  if(relat==4){
	ncomp=3;
      }
      
      Dobj dw=par()("wyckoff_positions");
      for(int i=0;i<dw.size();i++){
	Dobj di=dw[i];
	String el=di("element")();
	for(int icomp=0;icomp<ncomp;icomp++){
	  QTreeWidgetItem *itemsort=0;
	  if(icomp==0){
	    QString s=QString("%1\n%2\n%3").arg(i+1).arg(el()).arg(icomp+1);
	    itemsort=new QTreeWidgetItem((QTreeWidget*)0,
						      s.split("\n"));
	    itemsort->setData(0,Qt::UserRole,i);
	    itemsort->setData(1,Qt::UserRole,el());
	    itemsort->setData(2,Qt::UserRole,icomp);
	    trwOrbitals->addTopLevelItem(itemsort);
	    if(i==0)item0=itemsort;
	  }else{
	    QString s=QString("\n\n%1").arg(icomp+1);
 	    QTreeWidgetItem *item=new QTreeWidgetItem(itemsort,s.split("\n"));
	    item->setData(0,Qt::UserRole,i);
 	    item->setData(1,Qt::UserRole,el());
 	    item->setData(2,Qt::UserRole,icomp);
	  }
	}
      }
    }catch(Except e){};
    
  }catch(Except &ex){
  }
  
  //myDebug() << "relat "<<relat << ncomp;
  trwOrbitals->setCurrentItem(item0);
  
  onTrwOrbitalsCurrentItemChanged(item0,0);
  
  
  
  return;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::iterationMode()
{
  if(muteIterationMode) return;
  updateTimer->stop();
  OutGrep::Mode mode=(OutGrep::Mode)cmbMode->
    itemData(cmbMode->currentIndex()).toInt();

  long site=cmbSite->itemData(cmbSite->currentIndex()).toInt();
  
  //myDebug() << "ObserveView::iterationMode" << mode;


  String fn=String(leOutFile->text());
  if(fn==""){
    fn="out";
    leOutFile->setText(fn());
  }
  
  bool newfile=(pobserveoutfile!=fn);
  try{
    if(newfile){
      poutgreper.setOutfile(fn);
      occupyCmbSite();
    }

    if(mode==OutGrep::AtomSpinGros||
       mode==OutGrep::AtomOccuGros||
       mode==OutGrep::AtomOrbitalMoment||
       mode==OutGrep::AtomEFG){
      lSite->show();
      cmbSite->show();
    }else{
      lSite->hide();
      cmbSite->hide();
    }
    
    
  }catch(Except ex) {
    QMessageBox::warning(parentwidget,IDS_APP_NAME,ex.msg()
			 ,QMessageBox::Ok);
    return;
  }

  pobserveoutfile=fn;


//   plaintextedit->clear();
//   plaintextedit->setReadOnly(true);
//   const StringList &sl=poutgreper.content();
//   myDebug() << "staert";
//   QString s;
//   for(int i=0;i<sl.size();i++){
//     //myDebug() << i<<sl[i]();
//     s+=sl[i]();
//   }
//   plaintextedit->setPlainText(s);
//   myDebug() << "end";


  ValueArray<ValueArray<double> > xs;
  ValueArray<ValueArray<double> > ys;
  ValueArray<String > legends;
  
  xs.resize(1);

  if(mode==OutGrep::AtomSpinGros||
     mode==OutGrep::AtomOccuGros||
     mode==OutGrep::AtomOrbitalMoment||
     mode==OutGrep::AtomEFG){
    ValueArray<OutGrep::SiteInfo> sites=poutgreper.extractSites();
    poutgreper.setOrbital(0);
    if(site>0){
      ys.resize(1);
      poutgreper.setSite(site);
      poutgreper.grep(mode,xs[0],ys[0]);
      String el=OutGrep::SiteInfo::getElementfromSite(sites,site);
      legends.append(Sprintf("%s(site %ld)",el(),site));
    }else{
      xs.resize(sites.size());
      ys.resize(sites.size());
      for(int i=0;i<sites.size();i++){
	poutgreper.setSite(sites.at(i).site);
	poutgreper.grep(mode,xs[i],ys[i]);
	legends.append(Sprintf("%s(%ld)",sites.at(i).el(),sites.at(i).site));
      }
    }
  }else if(mode==OutGrep::IterationBest){
    xs.resize(2);
    ys.resize(2);
    poutgreper.grep(OutGrep::IterationBest,xs[1],ys[1]);
    if(ys.at(1).size()>0){
      poutgreper.grep(OutGrep::Iteration,xs[0],ys[0]);
      legends.append("last deviation");
      legends.append("last deviation (best)");
    }else{
      xs.resize(1);
      ys.resize(1);
      poutgreper.grep(OutGrep::Iteration,xs[0],ys[0]);
      legends.append("last deviation");
    }
  }else{
    ys.resize(1);
    poutgreper.grep(mode,xs[0],ys[0]);
    switch(mode){
    case OutGrep::Iteration:
      legends.append("last deviation");
      break;
    case OutGrep::ForceIteration:
      legends.append("last Force deviation");
      break;
    case OutGrep::TotalEnergy:
      legends.append("E$_tot$.");
      break;
    case OutGrep::TotalSpin:
      legends.append("M$_tot$.");
      break;
    case OutGrep::CycleTime:
      legends.append("Cycle time");
      break;
    case OutGrep::dBfsm:
      legends.append("$~DB$_fsm$.");
      break;
    case OutGrep::Bfsm:
      legends.append("B$_fsm$.");
      break;
    default:
      legends.append(Sprintf("???"));
    }
  }


  p_gv->killAll(false);
  Graph *gr=p_gv->page()->graphs()->currentGraph();
  gr->groups()->reset();
  gr->groups()->insert(0);
  FunctionGroup *group=gr->groups()->at(0);
  group->setActive(true);
  group->funcs()->setSize(ys.size());
  double y0=0;
  if(mode==OutGrep::TotalEnergy && ys.size()>0 && ys.at(0).size()>0){
    y0=((int)(1000.*ys.at(0).at(0)))/1000.;
  }
  for(int i=0;i<ys.size();i++){
    Function *fu=group->funcs()->at(i);
    fu->setDataType(Function::XY);
    fu->setActive(true);
    if(mode==OutGrep::TotalEnergy){
      ValueArray<double> yy=ys.at(i);
      for(long j=0;j<yy.size();j++){
	yy[j]=(yy[j]-y0)*1000.;
      }
      fu->setData(xs[0],yy);
    }else{
      if(i<xs.size()){
	fu->setData(xs.at(i),ys.at(i));
      }else{
	fu->setData(xs.at(0),ys.at(i));
      }
    }
    fu->attribs()->lineStyle().setColor(Xfp::rgbColor(i%(Xfp::maxRgbColors())));
    fu->attribs()->symbolStyle().
      setType((Xfp::SymbolTypes)(i % (Xfp::maxSymbols()-2)+1));
    fu->attribs()->setLegend(legends.at(i));
  }

  

  TextBox *tbx=gr->textBoxes()->at(TextBox::TBTXAxisLabel);
  tbx->setText("iteration step");
  TextBox *tb=gr->textBoxes()->at(TextBox::TBTYAxisLabel);
  if(mode==OutGrep::Iteration||mode==OutGrep::IterationBest){
    //gr->world()->setW0(Xfp::Graph_Y,0);
    tb->setText("deviation");
    if(gr->groups()->at(0)->funcs()->at(0)->length()>1){
      gr->axes()->setScaling(Xfp::Graph_Y,Xfp::GASLogarithmic);
    }
    //gr->world()->setW0(Xfp::Graph_X,spbxIter0->value());
    //gr->autoScaleAll(Xfp::Graph_Y);
  }else if(mode==OutGrep::ForceIteration){
    gr->world()->setW0(Xfp::Graph_Y,0);
    tbx->setText("force iteration step");
    tb->setText("force deviation eV/$angstroem");
  }else if(mode==OutGrep::TotalEnergy){
    tb->setText(QString("E$_tot$.-(%1) [mHa]")
		.arg(Sprintf("%30.3f",y0).stripBlanks().chop("0")()));
  }else if(mode==OutGrep::Bfsm){
    tb->setText("B$_fsm$. [Ha]");
  }else if(mode==OutGrep::dBfsm){
    tb->setText("$~DB$_fsm$. [Ha]");
  }else if(mode==OutGrep::TotalSpin){
    tb->setText("Total gross spin [$~m$_B$.]");
  }else if(mode==OutGrep::CycleTime){
    tb->setText("Cycle time [s]");
  }else if(mode==OutGrep::AtomSpinGros){
    tb->setText("gross spin [$~m$_B$.]");
  }else if(mode==OutGrep::AtomOccuGros){
    tb->setText("gross occu [electrons]");
  }else if(mode==OutGrep::AtomOrbitalMoment){
    tb->setText("onsite L$_z$. [$~m$_B$.]");
  }else if(mode==OutGrep::AtomEFG){
    tb->setText("EFG V$_zz$. [10$^21$. V/m$^2$.]");
  }

  tb=gr->textBoxes()->at(TextBox::TBTTitle);
  tb->setText("Iteration progress");

  gr->autoScaleAll(Xfp::Graph_X);
  gr->autoScaleAll(Xfp::Graph_Y);
  gr->world()->setW0(Xfp::Graph_X,spbxIter0->value());
  gr->autoScaleAll(Xfp::Graph_Y,true);

  p_gv->refreshPage();

  if(cbAutoUpdate->isChecked()){
    updateTimer->start(pautoupdatetime);
  }

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::onTrwOrbitalsCurrentItemChanged(QTreeWidgetItem *current,
					  QTreeWidgetItem *previous)
{
  Q_UNUSED(previous);
  if(!current) return;
  long funcindex=0;
  int mode=cmbMode->itemData(cmbMode->currentIndex()).toInt();
  //myDebug()<<"ObserveView::onTrwOrbitalsCurrentItemChanged";

  int sort=current->data(0,Qt::UserRole).toInt()+1;
  String el=String(current->data(1,Qt::UserRole).toString());
  int icomp=current->data(2,Qt::UserRole).toInt()+1;

  //myDebug() << current<<    sort <<el;

  p_gv->killAll(false);
  Graph *gr=p_gv->page()->graphs()->currentGraph();
  TextBox *tb=gr->textBoxes()->at(TextBox::TBTXAxisLabel);
  tb->setText("r$_radial$. [a.u.]");
  tb=gr->textBoxes()->at(TextBox::TBTYAxisLabel);
  tb->setText("r*$~j$.(r)");
  tb=gr->textBoxes()->at(TextBox::TBTTitle);
  tb->setText(Sprintf("Sort %d %s Component %d",sort,el(),icomp)());

  String fname=Sprintf("+fcor.%03d.%d",sort,icomp);
  if(mode==VALENCEORBITAL){
    fname=Sprintf("+fval.%03d.%d",sort,icomp);
  }else if(mode==KINVALENCEORBITAL){
    fname=Sprintf("+fkval.%03d.%d",sort,icomp);
  }else if(mode==KINCOREORBITAL){
    fname=Sprintf("+fkcor.%03d.%d",sort,icomp);
  }
    
  p_gv->import(false,fname(),Xfp::FTXNY,funcindex);
  
  if(fabs(rmax)>1.0e-4){
    Graph *gr=p_gv->page()->graphs()->currentGraph();
    //    myDebug()<< "gr"<<gr;
    gr->world()->setW1(Xfp::Graph_X,rmax);
  }

  p_gv->refreshPage();



}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::cleanUp()
{
  //myDebug() << "cleanUp";
  if(!parentwidget) return;
  if(!parentwidget->layout()) return;

  if(trwOrbitals) delete trwOrbitals;trwOrbitals=0;
  if(lSelect) delete lSelect;lSelect=0;
  if(cmbMode) delete cmbMode;cmbMode=0;
  if(lOutputFile) delete lOutputFile;lOutputFile=0;
  if(leOutFile) delete leOutFile;leOutFile=0;
  if(pbBrowse) delete pbBrowse;pbBrowse=0;
  if(lSite) delete lSite;lSite=0;
  if(cmbSite) delete cmbSite;cmbSite=0;
  if(pbUpdate) delete pbUpdate;pbUpdate=0;
  if(updateTimer) delete updateTimer;updateTimer=0;
  if(cbAutoUpdate) delete cbAutoUpdate;cbAutoUpdate=0;
  if(spbxAutoUpdateTime) delete spbxAutoUpdateTime;spbxAutoUpdateTime=0;
  if(lIter0) delete lIter0;lIter0=0;
  if(spbxIter0) delete spbxIter0;spbxIter0=0;
  if(dspbxRmax) delete dspbxRmax;dspbxRmax=0;
  if(lRmax) delete lRmax;lRmax=0;
  if(splitter) delete splitter;splitter=0;
  if(plaintextedit) delete plaintextedit;plaintextedit=0;

  //if(parentwidget->layout()) delete parentwidget->layout();

  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::onPbBrowseClicked()
{
  //myDebug() << "hier";

  QFileDialog *dlg=new QFileDialog(parentwidget,"load output file","."
				   ,"Output files (out* ) ;;  All ( * )");
  
  dlg->setAcceptMode(QFileDialog::AcceptOpen);
  dlg->setFileMode(QFileDialog::ExistingFile);
#ifdef Q_OS_MAC
  dlg->setOptions(QFileDialog::DontUseNativeDialog);
#endif
  dlg->setDirectory(QDir::currentPath()+"/");

  dlg->selectFile(leOutFile->text());
  long ret;
  QString filename;

  if((ret=dlg->exec())==QDialog::Accepted){
    filename=dlg->selectedFiles().at(0);
  }
  delete (dlg);

  if(!filename.isNull() && ret==QDialog::Accepted){
    leOutFile->setText(filename);
    iterationMode();
  }

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::occupyCmbSite()
{
  muteIterationMode=true;
  long index=cmbSite->currentIndex();
  cmbSite->clear();
  ValueArray<OutGrep::SiteInfo> sites=poutgreper.extractSites();
  cmbSite->insertItem(0,"all",-1);
  for(int i=0;i<sites.size();i++){
    cmbSite->insertItem(i+1,Sprintf("%ld",sites.at(i).site)(),
			(int)sites.at(i).site);
  }
  if(index>=0 && index<cmbSite->count()){
    cmbSite->setCurrentIndex(index);
  }
  muteIterationMode=false;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ObserveView::onDspbxRmaxValueChanged(double v)
{
  rmax=v;
  QTreeWidgetItem *item0=trwOrbitals->currentItem();
  //  myDebug()<<rmax<<item0;
  if(item0){
    onTrwOrbitalsCurrentItemChanged(item0,0);
  }
}

/*EOF*/ 
