/* @(#)Array.hh
 * author: magru@unas
 */

 
#ifndef _ARRAY_HH
#define _ARRAY_HH 1

#include "Except.hh"
#include "ValueArray.hh"
#include "Types.hh"

namespace XFPLO{

  class IsoSurface;  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  
  template <class T>
  class BaseArray{
    //todo: find out, which members should be in a pure baseclass
    // and which are common to all derived classes
  protected:
    BaseArray();
    BaseArray(const BaseArray &);
    BaseArray(const ValueArray<long> &shape);
    BaseArray &operator=(const BaseArray &);
    virtual ~BaseArray();
  public:

    
    static BaseArray *construct(const ValueArray<long> &shape){
      return new BaseArray<T>(shape);
    }

    
    static void destroy(BaseArray *ar){
      if((--(ar->pusecounter))==0) delete ar;
    }
    
    
    
    inline long used()const{return (pusecounter);}
    inline void increaseUse(){pusecounter++;}

    inline long rank()const{return prank;}
    inline long size()const{return pnitems;}
    inline long shape(long rank)const{return pshape.at(rank);}
    
    void setSize(long size){
      pnitems=size;
      pdata.resize(pnitems*pdatasize,true);
    }

    
    DataTypes dataType()const{return ptype;}
    

    inline long items()const{return pnitems;}

    inline long datasize()const{return pdatasize;}


    inline const T &point(long index,long comp)const{
      return pdata.at(index*pdatasize+comp);
    }

    inline T *data(){return pdata.dataRef();} //todo: only for irregular arrays
    inline const T *data()const{return pdata.constDataRef();}
    
    inline ValueArray<T> &valueArray(){return pdata;} //todo: only for irregular arrays
    inline const ValueArray<T> &valueArray()const{return pdata;} //todo: only for irregular arrays

    virtual void debug()const;
    
    
  protected:
    
    void printValue(const T&)const;
    void setShape(const ValueArray<long> &shape){
      cleanData();
      prank=shape.size();
      pshape=shape;
      pdatasize=shape.prod();
    }
  private:
    void cleanData(){
      pnitems=0;
      pdata.resize(0,true); 
    }
    

  protected:
    String pclassname;
    DataTypes ptype;
    long prank;
    ValueArray<long> pshape;
    long pnitems;
    ValueArray<T> pdata;
    long pusecounter;
    long pdatasize;
    friend class IsoSurface;
  };

  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class BaseArrayPointer{
  private:
  public:
    BaseArrayPointer():ptype(TIunkown),array(0){}
    ~BaseArrayPointer(){clean();};

    template <class T>
    void build(const ValueArray<long> &shape){
      clean();
      ptype=TypeInfo<T>::type;
      array=BaseArray<T>::construct(shape);
    }
    
    void clean(){
      if(!array) return;
      switch(ptype){
      case TIlong:
	static_cast<BaseArray<long>*>(array)->destroy(static_cast<BaseArray<long>*>(array));
	break;
      case TIdouble:
	static_cast<BaseArray<double>*>(array)->destroy(static_cast<BaseArray<double>*>(array));
	break;
      case TICmplx:
	static_cast<BaseArray<Cmplx>*>(array)->destroy(static_cast<BaseArray<Cmplx>*>(array));
	break;
      case TIunkown:
	break;
      }
      array=0;
    };

    DataTypes dataType()const{return ptype;}

    template <class T>
    BaseArray<T> *to(){
      if(ptype!=TypeInfo<T>::type)
	throw Except(Sprintf("cannot cast BaseArrayPointer to %s",TypeInfo<T>::name));
      return static_cast<BaseArray<T>*>(array);
    }
    template <class T>
    const BaseArray<T> *to()const{
      if(ptype!=TypeInfo<T>::type)
	throw Except(Sprintf("cannot cast BaseArrayPointer to %s",TypeInfo<T>::name));
      return static_cast<const BaseArray<T>*>(array);
    }
    
  private:
    DataTypes ptype;
    void *array;
  };
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  template <class T>
  class RegularArray:public BaseArray<T>
  {
  protected:
    RegularArray();
    RegularArray(long shape,
		 const ValueArray<T> &origin,
		 const ValueArray<T> &delta,
		 long size);
    RegularArray(const RegularArray &);
    ~RegularArray();
    RegularArray &operator=(const RegularArray &);

  public:

    static RegularArray *construct(long shape,
				   const ValueArray<T> &origin,
				   const ValueArray<T> &delta,
				   long size){
      return new RegularArray<T>(shape,origin,delta,size);
    }
    

    const ValueArray<T> origin()const{return porigin;}
    const ValueArray<T> delta()const{return pdelta;}
    
    

    void debug()const;
    
  private:
    ValueArray<T> porigin;
    ValueArray<T> pdelta;
  };


  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class PathArray:public BaseArray<long>
  {
  protected:
    PathArray();
    PathArray(long size);
    PathArray(const PathArray &);
    ~PathArray();
    PathArray &operator=(const PathArray &);

  public:

    static PathArray *construct(long size){
      return new PathArray(size);
    }
    
  private:
  };

  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  template <class T>
  class ProductArray: public BaseArray<T>
  {
    // must consist of rank-1 arrays
  protected:
    ProductArray();
    ProductArray(const ProductArray<T> &);
    ~ProductArray();
    ProductArray &operator=(const ProductArray &);

  public:

    static ProductArray *construct(){
      return new ProductArray<T>();
    }

    //add existing array and optionally
    // take over possesion (no destroy needed)
    void addArray(const BaseArray<T> *ba,bool take_over=true);

    //the number of arrays==dimensionality of mesh
    inline long dimensions()const{return arrays.size();}


    const BaseArray<T> *array(long i)const{return arrays.at(i);}
    
    //this is a slow operation, and should not be used
    //for time critical operations
    // inline T &point(long index,long comp)const{
      
    //   if(pdimension==1)
    // 	return arrays.at(0)->point(index,comp);
    //   else if(pdimension==2){
    // 	long Nx=arrays.at(0)->size();
    // 	long ix=index % Nx;
    // 	long iy=(index-ix)/Nx;
    // 	return arrays.at(0)->point(ix,comp)+arrays.at(1)->point(iy,comp);
    //   }else if(pdimension==3){
    // 	long Nx=arrays.at(0)->size();
    // 	long Ny=arrays.at(1)->size();
    // 	long ix=index % Nx;
    // 	index=(index-ix)/Nx;
    // 	long iy=index%Ny;
    // 	long iz=(index-iy)/Ny;
    // 	//myDebug()<<"{"<<index<<":"<<ix<<iy<<iz<<"}";
    // 	return arrays.at(0)->point(ix,comp)+
    // 	  arrays.at(1)->point(iy,comp)+
    // 	  arrays.at(2)->point(iz,comp);
    //   }else return 0;
    // }

    void debug()const;
    
  protected:
    ValueArray<const BaseArray<T>*> arrays;
  };

  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class MeshArray: public BaseArray<long>
  {
    // must consist of rank-1 arrays
  protected:
    MeshArray();
    MeshArray(const MeshArray &);
    ~MeshArray();
    MeshArray &operator=(const MeshArray &);

  public:

    static MeshArray *construct(){
      return new MeshArray();
    }

    //add existing array and optionally
    // take over possesion (no destroy neede)
    void addArray(const PathArray *ba,bool take_over=true);

    
    //the number of arrays==dimensionality of mesh
    inline long dimensions()const{return arrays.size();}

    
    const BaseArray<long> *array(long i)const{return arrays.at(i);}
    
    //this is a slow operation, and should not be used
    //for time critical operations
    // inline long &point(long index,long comp)const{
      
    //   }else if(pdimension==2){
    // 	long Nx=arrays.at(0)->size();
    // 	long ix=index % Nx;
    //    	long iy=(index-ix)/Nx;
    // 	long ic=comp%2;
    // 	long jc=(comp-ic)/2;
    // 	long i=arrays.at(0)->point(ix,ic);
    // 	long j=arrays.at(1)->point(iy,jc);
    // 	return i+(Nx+1)*j;
    //   }else if(pdimension==3){
    //    	long Nx=arrays.at(0)->size();
    // 	long Ny=arrays.at(1)->size();

    // 	long ix=index % Nx;
    // 	index=(index-ix)/Nx;
    // 	long iy=index%Ny;
    // 	long iz=(index-iy)/Ny;

    // 	long ic=comp%2;
    // 	comp=(comp-ic)/2;
    // 	long jc=comp%2;
    // 	long kc=(comp-jc)/2;

    // 	long i=arrays.at(0)->point(ix,ic);
    // 	long j=arrays.at(1)->point(iy,jc);
    // 	long k=arrays.at(2)->point(iz,kc);
    // 	return i+(Nx+1)*(j+(Ny+1)*k);

    //   }else return 0;
      
    // }

    void debug()const;
    
  private:
    ValueArray<const PathArray*> arrays;
  };

  
}




#endif /* _ARRAY_HH */
  

// template <class T>
// class Object1
// {
// private:
//   T a1;
//   T a2;
// public:
//   T getA1() {return a1;}
//   typedef T type;
// };

// template <class Object>
// void foo(Object *o1, Object *o2)
// {
//   typename Object::type x = o1.getA1();
  
// }
  
