/*************************************************************************
 * @(#)Axes.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   14 Aug 2009
 *************************************************************************
 */

#include "Parser.hh"
#include "Except.hh"
#include "Axes.hh"
#include "Symmetry.hh"
#include "constants.h"
#include "Xf.hh"
#include "MyDebug.hh"


using namespace XFPLO;

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Axes::Axes(Xf::AxesType t) :ptype(t),pchanged(false),pcalced(false),
			    pvisible(false)
{
  if(ptype==Xf::ATCartesian || ptype==Xf::ATPrimitive
     || ptype==Xf::ATConventional){
    pdefaultrgb[0]=0xff000000;
    pdefaultrgb[1]=0xffff0000;
    pdefaultrgb[2]=0xff00ff00;
    pname=Xf::axesTypeName(ptype);
    pid="axes "+pname;
  }

  prgb[0]=pdefaultrgb[0];
  prgb[1]=pdefaultrgb[1];
  prgb[2]=pdefaultrgb[2];

  setDefaultAxes();
  setDefaultRod();
  setDefaultTip();
  setDefaultOrigin();
  setDefaultLabel();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Axes::Axes(const Axes &o)
{
  setAxesType(o.ptype);
  setID(o.pid);
  setName(o.pname);
  setVisible(o.pvisible);
  setColor(o.prgb);
  setDefaultColor(o.pdefaultrgb);
  setRodRadius(o.prodradius);
  setRodRadiusUnits(o.prodradiusunits);
  setTipRadius(o.ptipradius);
  setTipRadiusUnits(o.ptipradiusunits);
  setTipLength(o.ptiplength);
  setTipLengthUnits(o.ptiplengthunits);
  setHasTip(o.phastip);
  setAxisLength(o.paxeslength);
  setAxesLengthUnits(o.paxeslengthunits);
  setAxisOrigin(o.paxesorigin);
  setOriginUnits(o.poriginunits);
  setLabel(o.plabel);
  setFontShift(o.pfontshift);
  setFontScale(o.pfontscale);
  setFontColor(o.pfontcolor);}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
const Axes &Axes::operator=(const Axes &o)
{
  setAxesType(o.ptype);
  setID(o.pid);
  setName(o.pname);
  setVisible(o.pvisible);
  setColor(o.prgb);
  setDefaultColor(o.pdefaultrgb);
  setRodRadius(o.prodradius);
  setRodRadiusUnits(o.prodradiusunits);
  setTipRadius(o.ptipradius);
  setTipRadiusUnits(o.ptipradiusunits);
  setTipLength(o.ptiplength);
  setTipLengthUnits(o.ptiplengthunits);
  setHasTip(o.phastip);
  setAxisLength(o.paxeslength);
  setAxesLengthUnits(o.paxeslengthunits);
  setAxisOrigin(o.paxesorigin);
  setOriginUnits(o.poriginunits);
  setLabel(o.plabel);
  setFontShift(o.pfontshift);
  setFontScale(o.pfontscale);
  setFontColor(o.pfontcolor);
  return *this;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Axes::print()
{
  printf("Axes\n"
	 "\ttype: %d\n"
	 "\tvisible: %d\n"
	 "\tdefault colors: %x %x %x\n"
	 "\tcolors: %x %x %x\n"
	 "\tRod radius: %s %s\n"
	 "\tTip radius: %s %s\n"
	 "\tTip length: %s %s\n"
	 "\tx-Axes length: %s %s\n"
	 "\ty-Axes length: %s %s\n"
	 "\tz-Axes length: %s %s\n"
	 "\torigin : %s %s %s %s\n"
	 "\tchanged: %d\n"
	 ,ptype,pvisible
	 ,pdefaultrgb[0],pdefaultrgb[1],pdefaultrgb[2]
	 ,prgb[0],prgb[1],prgb[2]
	 ,prodradius.S(),Xf::sceeneUnitName(prodradiusunits)
	 ,ptipradius.S(),Xf::sceeneUnitName(ptipradiusunits)
	 ,ptiplength.S(),Xf::sceeneUnitName(ptiplengthunits)
	 ,paxeslength[0].S(),Xf::sceeneUnitName(paxeslengthunits)
	 ,paxeslength[1].S(),Xf::sceeneUnitName(paxeslengthunits)
	 ,paxeslength[2].S(),Xf::sceeneUnitName(paxeslengthunits)
	 ,paxesorigin[0].S() ,paxesorigin[1].S()
	 ,paxesorigin[2].S() ,Xf::sceeneUnitName(poriginunits)
	 ,pchanged
	 );
}

/* ***********************************************************************
 * 
 * ***********************************************************************
 */
void Axes::calculateAxes(Xf::LatticeType latticetype
			 ,double alloverscale,const Symmetry *sym)
{
  NSF("Axes::calculateAxes");

  double u=constant_aBtoAng;
  if(latticetype==Xf::LTReciprocial) u=1/u;

  if(ptype==Xf::ATConventional){
    if(latticetype==Xf::LTDirect) pvectors=sym->ccell();
    else pvectors=sym->crcell();
  }else if(ptype==Xf::ATPrimitive){
    if(latticetype==Xf::LTDirect) pvectors=sym->cell();
    else pvectors=sym->rcell();
  }else if(ptype==Xf::ATCartesian){
    Vec3 length;
    length[0]=paxeslength[0].D;
    length[1]=paxeslength[1].D;
    length[2]=paxeslength[2].D;

    if(paxeslengthunits == Xf::SUBoundary) length*=(alloverscale*2);
    
    
    if(paxeslengthunits == Xf::SUAngstroem){
      if(sym->dataLengthUnit()==Xf::BohrRadius)
	length/=u;
    }else if(paxeslengthunits == Xf::SUBohrRadius){
      if(sym->dataLengthUnit()==Xf::Angstroem)
	length*=u;
    }
    
    
    pvectors=Matrix3x3::unitMatrix();
    pvectors.data()[0]*=length[0];
    pvectors.data()[4]*=length[1];
    pvectors.data()[8]*=length[2];

  }else{
    myDebug() << nsubfunc<< "Oops, unknown AxesType";
  }

  
  porigin[0]=paxesorigin[0].D;
  porigin[1]=paxesorigin[1].D;
  porigin[2]=paxesorigin[2].D;

  if(poriginunits== Xf::SUConvCell &&
     sym->dataStructureType() == Xf::STCrystal){
    if(latticetype==Xf::LTReciprocial)
      porigin=sym->crcell()*porigin;
    else
      porigin=sym->ccell()*porigin;
  }else if(poriginunits== Xf::SUPrimCell&&
	   sym->dataStructureType() == Xf::STCrystal){
    if(latticetype==Xf::LTReciprocial)
      porigin=sym->rcell()*porigin;
    else
      porigin=sym->cell()*porigin;
  }else if(poriginunits== Xf::SUAngstroem){
    if(sym->dataLengthUnit()==Xf::BohrRadius)
      porigin/=u;
  }else if(poriginunits== Xf::SUBohrRadius){
    if(sym->dataLengthUnit()==Xf::Angstroem)
      porigin*=u;
  }
  
  prr=prodradius.D;
  if(prodradiusunits == Xf::SUBoundary){
    prr*=(alloverscale*2);
  }else if(prodradiusunits == Xf::SUAngstroem){
    if(sym->dataLengthUnit()==Xf::BohrRadius)
      prr/=u;
  }else if(prodradiusunits == Xf::SUBohrRadius){
    if(sym->dataLengthUnit()==Xf::Angstroem)
      prr*=u;
  }
  

  ptr=ptipradius.D;
  if(ptipradiusunits == Xf::SUBoundary){
    ptr*=(alloverscale*2);
  }else if(ptipradiusunits == Xf::SURodRadius){
    ptr*=prr;
  }else if(ptipradiusunits == Xf::SUAngstroem){
    if(sym->dataLengthUnit()==Xf::BohrRadius)
      ptr/=u;
  }else if(ptipradiusunits == Xf::SUBohrRadius){
    if(sym->dataLengthUnit()==Xf::Angstroem) 
      ptr*=u;
  }
  
  
  ptl=ptiplength.D;
  if(ptiplengthunits == Xf::SUBoundary){
    ptl*=(alloverscale*2);
  }else if(ptiplengthunits == Xf::SUTipRadius){
    ptl*=ptr;
  }else if(ptiplengthunits == Xf::SURodRadius){
    ptl*=prr;
  }else if(ptiplengthunits == Xf::SUAngstroem){
    if(sym->dataLengthUnit()==Xf::BohrRadius)
      ptl/=u;
  }else if(ptiplengthunits == Xf::SUBohrRadius){
    if(sym->dataLengthUnit()==Xf::Angstroem)
      ptl*=u;
  }

  
  if(!phastip){  //todo dont use arrow
    ptr=prr;
    ptl=0;
  }

  pshow= ( sym->dataStructureType() == Xf::STCrystal ||
	   (ptype==Xf::ATCartesian) );
    
  
  pcalced=true;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Axess::insert(int at,Axes *a,bool createuniquename)
{
  if(createuniquename) a->setName(pnamecache.makeUniqueName("Axes "));
  a->setID(pidcache.makeUniqueName("axes "));
  paxes.insert(at,a);
  pchanged=true;
  pcalced=false;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Axes *Axess::findByType(Xf::AxesType type)
{
  //myDebug()<<"Axess"<<(void*)this;
  Axes *ax;
  for(long i=-1;++i<paxes.size();){
    ax=paxes[i];
    //myDebug()<<(void*)paxes.at(i)<<ax;
    if(ax && ax->axesType()==type)
      return ax;
    
  }
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Axes *Axess::findByID(const String &id)
{
  Axes *ax;
  for(long i=-1;++i<paxes.size();){
    ax=paxes[i];
    if(ax && id.startsWith(ax->id())){
      return ax;
    }
  }
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Axess::readAxesPTSectionFromFile(Parser *par)
{
  //return;
  Dobj d;
  try{
    d=(*par)()("axessystems");
  }catch(Except &ex){ return;};

  
  
  clear();
  Axes *a;
  for(int ia=0;ia<d.size();ia++){
    Dobj da=d[ia];
    Xf::AxesType t=(Xf::AxesType)da("axes_type.type").valueL();

    insert(ia,a=new Axes(t),false);

    
    
    a->setAxesType(t);
    
    try{
      bool t=da("visible").valueL();
      a->setVisible(t);
    }catch(Except &ex){};
    
    try{
      String r[3];
      da("color").getStringArray(3,r);
      QRgb c[3];
      c[0]=r[0].convertToULong()|0xff000000;
      c[1]=r[1].convertToULong()|0xff000000;
      c[2]=r[2].convertToULong()|0xff000000;
      a->setColor(c);
    }catch(Except &ex){};
    
    try{
      SReal t;t=da("rod.radius")();
      a->setRodRadius(t);
    }catch(Except &ex){};
    
    try{
      Xf::SceeneUnits t=(Xf::SceeneUnits)da("rod.radius_unit.type").valueL();
      a->setRodRadiusUnits(t);
    }catch(Except &ex){};
    
    try{
      SReal t;t=da("tip.radius")();
      a->setTipRadius(t);
    }catch(Except &ex){};
    
    try{
      Xf::SceeneUnits t=(Xf::SceeneUnits)da("tip.radius_unit.type").valueL();
      a->setTipRadiusUnits(t);
    }catch(Except &ex){};
    
    try{
      SReal t;t=da("tip.length")();
      a->setTipLength(t);
    }catch(Except &ex){};
    
    try{
      Xf::SceeneUnits t=(Xf::SceeneUnits)da("tip.length_unit.type").valueL();
      a->setTipLengthUnits(t);
    }catch(Except &ex){};
    
    try{
      bool t=da("tip.visible").valueL();
      a->setHasTip(t);
    }catch(Except &ex){};
    
    try{
      SReal t[3];
      t[0]=da("axislength.length")[0]();
      t[1]=da("axislength.length")[1]();
      t[2]=da("axislength.length")[2]();
      a->setAxisLength(t);
    }catch(Except &ex){};
    
    try{
      Xf::SceeneUnits t=(Xf::SceeneUnits)da("axislength.length_unit.type").valueL();
      a->setAxesLengthUnits(t);
    }catch(Except &ex){};


    try{
      SReal t[3];
      t[0]=da("axisorigin.origin")[0]();
      t[1]=da("axisorigin.origin")[1]();
      t[2]=da("axisorigin.origin")[2]();
      a->setAxisOrigin(t);
    }catch(Except &ex){};
    
    try{
      Xf::SceeneUnits t=(Xf::SceeneUnits)da("axisorigin.origin_unit.type").valueL();
      a->setOriginUnits(t);
    }catch(Except &ex){};

    for(int ia=-1;++ia<3;){
      try{
	String t;
	t=da("axislabels")[ia]("label")();
	a->setLabel(ia,t);
      }catch(Except &ex){
	//	myDebug()<<"oops"<<ex.msg();
      };
      try{
	SReal3 t;
	t[0]=da("axislabels")[ia]("shift")[0]();
	t[1]=da("axislabels")[ia]("shift")[1]();
	t[2]=da("axislabels")[ia]("shift")[2]();
	a->setFontShift(ia,t);
      }catch(Except &ex){
	//	myDebug()<<"oops"<<ex.msg();
      };
    }

    
    
    try{
      SReal t;t=da("fontscale")();
      a->setFontScale(t);
    }catch(Except &ex){
      //      myDebug()<<"oops"<<ex.msg();
    };
    
    try{
      QRgb t;t=da("fontcolor")().convertToULong();
      a->setFontColor(t);
    }catch(Except &ex){
      //      myDebug()<<"oops"<<ex.msg();
    };
    
  }
  

  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Axess::axesPTSectionToFile(FILE *FH)
{

#define axesfmt "       {\n"						\
    "         {%d,\"%s\"},\n"						\
    "         %c,{%s,%s,%s},{%s,{%d,\"%s\"}},\n"			\
    "         {%c,%s,{%d,\"%s\"},%s,{%d,\"%s\"}},\n"			\
    "         {{%s,%s,%s},{%d,\"%s\"}},\n"				\
    "         {{%s,%s,%s},{%d,\"%s\"}},\n"				\
    "         {{\"%s\",{%s,%s,%s}},{\"%s\",{%s,%s,%s}},{\"%s\",{%s,%s,%s}}},%s,\"%x\"\n" \
    "       }%c\n" 

  fprintf(FH,
	  "  struct { \n"
	  "       struct  {int type;char[*] description;} axes_type; \n"
	  "       logical visible;\n"
	  "       char[*] color[3]; \n"
	  "       struct{real radius;\n"
	  "         struct  {int type;char[*] description;} radius_unit; \n"
	  "       }rod;\n"
	  "       struct{logical visible;real radius;\n"
	  "         struct  {int type;char[*] description;} radius_unit; \n"
	  "         real length;\n"
	  "         struct  {int type;char[*] description;} length_unit; \n"
	  "       }tip;\n"
	  "       struct {real length[3];\n"
	  "         struct  {int type;char[*] description;} length_unit; \n"
	  "       } axislength;\n"
	  "       struct {real origin[3];\n"
	  "         struct  {int type;char[*] description;} origin_unit; \n"
	  "       } axisorigin;\n"
	  "       struct{char[*]label;real shift[3];}axislabels[3];\n"
	  "       real fontscale;char[*] fontcolor;\n" 
	  "       } axessystems[*]\n"
	  "       ={\n");
  for(long i=-1;++i<paxes.size();){
      Axes *ax=paxes[i];
      fprintf(FH,axesfmt
	      ,ax->ptype
	      ,Xf::axesTypeName(ax->ptype)
	      ,(ax->pvisible?'t':'f')
	      ,Sprintf("\"%lx\"",0xffffff&ax->prgb[0])() 
	      ,Sprintf("\"%lx\"",0xffffff&ax->prgb[1])() 
	      ,Sprintf("\"%lx\"",0xffffff&ax->prgb[2])() 
	      ,ax->prodradius.S()
	      ,ax->prodradiusunits,Xf::sceeneUnitName(ax->prodradiusunits)
	      
	      ,(ax->hasTip()?'t':'f')
	      ,ax->ptipradius.S()
	      ,ax->ptipradiusunits,Xf::sceeneUnitName(ax->ptipradiusunits)
	      
	      ,ax->ptiplength.S()
	      ,ax->ptiplengthunits,Xf::sceeneUnitName(ax->ptiplengthunits)
	      
	      ,ax->paxeslength[0].S(),ax->paxeslength[1].S()
	      ,ax->paxeslength[2].S()
	      ,ax->paxeslengthunits,Xf::sceeneUnitName(ax->paxeslengthunits)
	      
	      ,ax->paxesorigin[0].S(),ax->paxesorigin[1].S()
	      ,ax->paxesorigin[2].S()
	      ,ax->poriginunits,Xf::sceeneUnitName(ax->poriginunits),
	      ax->plabel[0](),
	      ax->fontShift(0)[0].S(),ax->fontShift(0)[1].S(),ax->fontShift(0)[2].S(),
	      ax->plabel[1](),
	      ax->fontShift(1)[0].S(),ax->fontShift(1)[1].S(),ax->fontShift(1)[2].S(),
	      ax->plabel[2](),
	      ax->fontShift(2)[0].S(),ax->fontShift(2)[1].S(),ax->fontShift(2)[2].S(),
	      ax->fontScale().S(),
	      ax->pfontcolor,
	      (i<paxes.size()-1 ?  ',' : ' ' )
	      );
  }
  fprintf(FH,"       };\n\n");


}

/*EOF*/ 
