/*************************************************************************
 * @(#)BBins.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   06 Mar 2011
 *************************************************************************
 */

#include "BBins.hh"
#include "BondDefs.hh"
#include "MyMath.hh"


ValueArray<Pair<Bin::BinType,const char *> > BBins::binTypeName=
    (ValueArray<Pair<Bin::BinType,const char *> >()
     <<Pair<Bin::BinType,const char *>(Bin::BTRoot,"Root")
     <<Pair<Bin::BinType,const char *>(Bin::BTWhatEver,"Distance")
     <<Pair<Bin::BinType,const char *>(Bin::BTType,"User type")
     <<Pair<Bin::BinType,const char *>(Bin::BTElement,"Element")
     <<Pair<Bin::BinType,const char *>(Bin::BTWyck,"Wyckoff position")
     <<Pair<Bin::BinType,const char *>(Bin::BTSite,"Site")
     );


ValueArray<Pair<BBins::BinsCenterType,const char *> > BBins::binsCenterTypeName=
       (ValueArray<Pair<BinsCenterType,const char *> >()
	<<Pair<BBins::BinsCenterType,const char *>(BBins::BCWyck,"Wyckoff position")
	<<Pair<BBins::BinsCenterType,const char *>(BBins::BCType,"User type")
	<<Pair<BBins::BinsCenterType,const char *>(BBins::BCElement,"Element")
	<<Pair<BBins::BinsCenterType,const char *>(BBins::BCSite,"Site")
	<<Pair<BBins::BinsCenterType,const char *>(BBins::BCIndividual,"Individual atom")
	);

ValueArray<Pair<BBins::BinsTopLevelGrouping,const char *> > BBins::binsTopLevelGroupingName=
     (ValueArray<Pair<BBins::BinsTopLevelGrouping,const char *> >()
      <<Pair<BBins::BinsTopLevelGrouping,const char *>(BBins::BTLElement,"Element")
      <<Pair<BBins::BinsTopLevelGrouping,const char *>(BBins::BTLType,"User type")
      <<Pair<BBins::BinsTopLevelGrouping,const char *>(BBins::BTLDistance,"Distance")
      );




/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Bin::print(int depth)const
{
  printf(" %*s-> %s: %s\n",depth," ",BBins::binTypeName.at(pbintype).v2,
	 toString()()); 
  for(int i=-1;++i<pbins.size();){
    const Bin *b=pbins.at(i);
    b->print(depth+5);
  }

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String Bin::toString(bool showDist)const
{
  String txt;
  if(pbintype==BTRoot){
    txt="root";
  }else if(pbintype==BTWhatEver){
    if(fabs(rmin() - rmax())>1.0e-6){
      txt=Sprintf("[%.4f,%.4f] %ld neighbours"
		  ,rmin() ,prmax,neighbours());
    }else{
      txt=Sprintf("[%.4f] %ld neighbours"
		  ,(rmin()+rmax())*0.5,neighbours() );
    }
  }else if(pbintype==BTType){
    if(fabs(rmin() - rmax())>1.0e-6){
      txt=Sprintf("T%ld [%.4f,%.4f] %ld neighbours"
		  ,type(),rmin() ,prmax,neighbours());
    }else{
      txt=Sprintf("T%ld [%.4f] %ld neighbours"
		  ,type(),(rmin()+rmax())*0.5,neighbours() );
    }
  }else if(pbintype==BTElement){
    if(fabs(rmin() - rmax())>1.0e-6){
      txt=Sprintf("%s, [%.4f,%.4f] %ld neighbours"
		  ,element()(),rmin() ,prmax,neighbours());
    }else{
      txt=Sprintf("%s, [%.4f] %ld neighbours"
		  ,element()(),(rmin()+rmax())*0.5,neighbours() );
    }
  }else if(pbintype==BTWyck){
    txt=Sprintf("%s(W%ld) [%.4f] %ld neighbours"
		,element()(),sort()+1,pabs(),neighbours()
		);
  }else if(pbintype==BTSite){
    if(showDist){
      txt=Sprintf("%s(S%ld,W%ld) %g (%- f , %- f , %- f)" 
 		  ,ppair.el(),ppair.site+1,ppair.sort+1,ppair.pabs
 		  ,ppair.p[0],ppair.p[1],ppair.p[2]);
    }else{
       txt=Sprintf("%s(S%ld,W%ld) (%- f , %- f , %- f)" 
 		  ,ppair.el(),ppair.site+1,ppair.sort+1
 		  ,ppair.p[0],ppair.p[1],ppair.p[2]);
    }

  }
  return txt;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Bin::setCheckStateInternal(int state)
{
  int st=state;
  if(st!=Qt::Unchecked)st=Qt::Checked;
  
  for(int i=-1;++i<pbins.size();){
    Bin *b=pbins[i];
    b->setCheckStateInternal(st); // give fully (un)checked  to all minions
  }
  pcheckstate=st;
  //myDebug() << printS()()<<pcheckstate;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Bin::updateCheckStateInternal()
{
  if(pbintype==BTRoot) return;
  //myDebug() << "updateCheckStateInternal"<<printS()();
  
  bool some=false;
  bool all=true;
  
  for(int i=-1;++i<pbins.size();){
    const Bin *b=pbins.at(i);
    int state=b->pcheckstate;
    //myDebug() << b->printS()()<<b->pcheckstate;

    if(state==Qt::Checked){
      some=true;
    }else if(state==Qt::PartiallyChecked){
      some=true;
      all=false;
    }else if(state==Qt::Unchecked){
      all=false;
    }
  }
  //myDebug() << all<<some;

  if(some && !all) {
    pcheckstate=Qt::PartiallyChecked;
  }else if(all){
    pcheckstate=Qt::Checked;
  }else{
    pcheckstate=Qt::Unchecked;
  }
  //myDebug() << pcheckstate;
  
  if(pparent)
    pparent->updateCheckStateInternal();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Bin::add(BinType bt,const PairVec &pair,double bintol,const String &prefix)
{
  bool prot=false;
  bintol=MyMath::max<double>(bintol,1.0e-6);
  if(prot){
    myDebug()<<prefix<<"pbintype="<<BBins::binTypeName.at(pbintype).v2;
    myDebug()<<prefix<<"bt      ="<<BBins::binTypeName.at(bt).v2;
    myDebug()<<prefix<<"bintol  ="<<bintol;
  }
  if(pbintype==BTRoot){
    
    if(bt==BTElement){
      

      if(prot){
	myDebug()<<prefix<<"pbins.size()"<<pbins.size();
	myDebug()<<prefix<<"pair.el"<<pair.el;
	if(pbins.size()) myDebug()<<prefix<<"pbins.last()->element()"
				  <<pbins.last()->element();
	myDebug()<<prefix<<"pair.pabs"<<pair.pabs;
	if(pbins.size()) myDebug()<<prefix<<"pbins.last()->pabs()"<<pbins.last()->pabs();
	if(pbins.size()) myDebug()<<prefix<<"dr"<<pair.pabs-pbins.last()->pabs();
	if(pbins.size()) myDebug()<<prefix<<"tests"<<(!pbins.size())
				  <<(pair.el !=pbins.last()->element())
				  <<(fabs(pair.pabs-pbins.last()->pabs())>=bintol);
      }
      if((!pbins.size())||
	 (pair.el !=pbins.last()->element())||
	 (fabs(pair.pabs-pbins.last()->pabs())>=bintol)){
	if(prot){myDebug()<<prefix<<"append new "<<BBins::binTypeName.at(bt).v2;}
	pbins.append(new Bin(this,bt,pair));
      }else{
	if(prot){myDebug()<<prefix<<"add  BTWyck";}
	pbins.last()->add(BTWyck,pair,1.0e-6,prefix+"  ");
      }
    }else if(bt==BTWhatEver){
      
      if(
	 (!pbins.size())||
	 (fabs(pair.pabs-pbins.last()->pabs())>=bintol) ){
	pbins.append(new Bin(this,bt,pair));
      }else{
	pbins.last()->add(BTElement,pair,bintol,prefix+"  ");
      }
    }else if(bt==BTType){
      
      if(
	 (!pbins.size())||
	 (pair.type !=pbins.last()->type())||
	 (fabs(pair.pabs-pbins.last()->pabs())>=bintol) ){
	pbins.append(new Bin(this,bt,pair));
      }else{
	pbins.last()->add(BTElement,pair,bintol,prefix+"  ");
      }
      
    }
  }else if(pbintype==BTWhatEver){
    prmin=MyMath::min<double>(prmin,pair.pabs);
    prmax=MyMath::max<double>(prmax,pair.pabs);
    
    if((!pbins.size())||
       (pair.el !=pbins.last()->element())||
       (fabs(pair.pabs-pbins.last()->pabs())>=bintol)){
      pbins.append(new Bin(this,BTElement,pair));
    }else{
      pbins.last()->add(BTWyck,pair,1.0e-6,prefix+"  ");
    }
    
  }else if(pbintype==BTType){
    prmin=MyMath::min<double>(prmin,pair.pabs);
    prmax=MyMath::max<double>(prmax,pair.pabs);
    
    if((!pbins.size())||
       (pair.el !=pbins.last()->element())||
       (fabs(pair.pabs-pbins.last()->pabs())>=bintol)){
      pbins.append(new Bin(this,BTElement,pair));
    }else{
      pbins.last()->add(BTWyck,pair,1.0e-6,prefix+"  ");
    }
    
  }else if(pbintype==BTElement){
    prmin=MyMath::min<double>(prmin,pair.pabs);
    prmax=MyMath::max<double>(prmax,pair.pabs);
    if(prot){
      myDebug()<<prefix<<"pbins.size()"<<pbins.size();
      myDebug()<<prefix<<"pair.sort"<<pair.sort;
      myDebug()<<prefix<<"pbins.last()->sort()"<<pbins.last()->sort();
      myDebug()<<prefix<<"pair.pabs"<<pair.pabs;
      myDebug()<<prefix<<"pbins.last()->pabs()"<<pbins.last()->pabs();
    }
    if( (!pbins.size()) ||
 	(pair.sort !=pbins.last()->sort())||
 	(fabs(pair.pabs-pbins.last()->pabs())>=1.0e-6) ) {
      if(prot){myDebug()<<prefix<<"append new Wyck";}
      pbins.append(new Bin(this,BTWyck,pair));
    }else{
      if(prot){
	myDebug()<<prefix<<"add Site";
	pbins.last()->print(0);
      }
      pbins.last()->add(BTSite,pair,bintol,prefix+"  ");
      if(prot){
	pbins.last()->print(0);
      }
    }
  }else if(pbintype==BTWyck){
    pbins.append(new Bin(this,BTSite,pair));
   }
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */

void Bin::initializeCheckState(long binssort
			       ,const BondDefGroup *bdg)
{
  for(int i=-1;++i<pbins.size();){
    Bin *b=pbins[i];
    if(b->pbintype==BTWyck){
      if(bdg->permits(binssort,b->sort(),b->pabs()))
	b->setCheckState(Qt::Checked);
      else
	b->setCheckState(0);
      
    }else if(b->pbintype!=BTSite){
      b->initializeCheckState(binssort,bdg);
    }
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
// void Bin::getWyckPosListInternal(QList<long> &res)
// {
//   foreach(Bin *b,pbins){
//     if(b->pbintype==BTWyck){
//       int sort=b->sort();
//       if(!res.contains(sort)) res.append(sort);
//     }else{
//       b->getWyckPosListInternal(res);
//     }
//   }
// }
/* ***********************************************************************
 *
 * ***********************************************************************
 */
// QList<long> Bin::getWyckPosList()
// {
//   if(pbintype!=BTWhatEver) return QList<long>();
//   QList<long> res;
  
//   getWyckPosListInternal(res);
//   return res;
// }


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BBins::calculateBins(BinsCenterType bct,BinsTopLevelGrouping btlg
			  ,long sort,long type,long site
			  ,const ValueArray<PtrArray<PairVec> > &pairs,double bintol)
{
  NSF("BBins::calculateBins()");
  // myDebug() << nsubfunc; 
  // myDebug() << "sort,site,tol: "<< sort<<site<<bintol; 

  pbinscentertype=bct;
  pbinstoplevelgrouping=btlg;
  psort=sort;
  ptype=type;
  psite=site;

  long nn=pairs.at(psite).size();

  prootbin->clear();
  
  if(!nn) return;

  for(int ipair=0;ipair<nn;ipair++){
    if(pbinscentertype==BCWyck){
      if(pbinstoplevelgrouping==BTLElement){
	prootbin->add(Bin::BTElement,*(pairs.at(psite).at(ipair)),bintol,"");
      }else if(pbinstoplevelgrouping==BTLDistance){
	prootbin->add(Bin::BTWhatEver,*(pairs.at(psite).at(ipair)),bintol,"");
      }else if(pbinstoplevelgrouping==BTLType){
	prootbin->add(Bin::BTType,*(pairs.at(psite).at(ipair)),bintol,"");
      }else{
	myDebug()<< "BinsTopLevelGrouping \"" 
		<< binsTopLevelGroupingName.at(pbinstoplevelgrouping).v2
		<< "\" not implemented"; 
	exit(1);
      }
    }else{
      myDebug()<< "BinsCenterType \"" 
	      << binsCenterTypeName.at(pbinscentertype).v2
	      << "\" not implemented"; 
      exit(1);
    }
  }

//   QList<long> wl=bin->getWyckPosList();
//   myDebug() << "site list";
//   foreach(long issa,wl){
//     const QList<long> &sl=psym->siteListBySort(issa);
//     foreach(long isa,sl){
//       //psites.at(site)
//       psym->pairs(isa).pabs()
// 	myDebug() << issa<<isa;
//     }
//   }
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void BBins::reportBins(const Symmetry *sym)
{
  NSF("BBins::reportBins()");
  //myDebug() << nsubfunc; 

  const ValueArray<Site> &sites=sym->sites();
  printf("\n\n\n\n"); 
  printf("Bins: around %s %s (W%ld,S%ld), grouped by %s\n"
	 ,binsCenterTypeName.at(pbinscentertype).v2
	 ,sites.at(psite).el()(),psort+1,psite+1
	 ,binsTopLevelGroupingName.at(pbinstoplevelgrouping).v2
	 ); 
  int depth=0;
  prootbin->print(depth);
  printf("\n"); 
}




/*EOF*/ 
