/* @(#)BBins.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _BBINS_H
#define _BBINS_H 1

#include "MyMath.hh"
#include "Symmetry.hh"
#include "PtrArray.hh"
#include "MyDebug.hh"


class BinModel;
class BondModel;
class BondsDlg;

using namespace XFPLO;

namespace XFPLO{

  class BondDefs;
  class BondDefGroup;
  class BBins;
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  
  class Bin{
  public: 
    
    // bin tree level types
    enum BinType{BTRoot=0,BTWhatEver=1,BTType=2,BTElement=3,BTWyck=4,BTSite=5};
    
    Bin(Bin *parent,BBins *bbins):pparent(parent),pbintype(BTRoot),pcheckstate(false),
				 pbbins(bbins){};
    Bin(Bin *parent,BinType bt,const PairVec &pair)
      :pparent(parent),pbintype(bt),pcheckstate(false),ppair(pair),pbbins(0){
      if(bt==BTWhatEver || bt==BTType ){
	prmin=pair.pabs;
	prmax=pair.pabs;
	pbins.append(new Bin(this,BTElement,pair));
      }else if(bt==BTElement ){
	prmin=pair.pabs;
	prmax=pair.pabs;
	//myDebug()<<"construct new Element and inserting new Wyck";
	pbins.append(new Bin(this,BTWyck,pair));
      }else if(bt==BTWyck){
	//myDebug()<<"construct new Wyck and inserting new site";
	pbins.append(new Bin(this,BTSite,pair));
      }
    };
    Bin(const Bin &other); //disable copy constructor
    Bin &operator=(const Bin &other); //disable operator=
    ~Bin(){
      //myDebug()<<"Bin "<<binTypeName()<<"dies!";
      clear();
    };
    
    inline BinType binType()const {return pbintype;};
    String toString(bool showDist=false)const;
    inline long size()const{return pbins.size();};
    inline const Bin *bin(long i)const{return pbins.at(i);};
    inline long sort()const{return ppair.sort;};
    inline double pabs()const{
      if(pbintype>=BTWyck) 
	return ppair.pabs;
      else if(pbintype==BTElement||pbintype==BTType||pbintype==BTWhatEver)  
	return pbins.last()->pabs();
      else{
	myDebug()<<"BinType"<<pbintype<<" not implemented";
	exit(0);
      }
      
    };
    //    ValueArray<long> getWyckPosList();
  private:
    BBins *bbins()const {return pbbins;}
    inline bool isCheckable()const { 
      if(pbintype==BTSite) return false; else return true; };
    inline void setCheckState(int state){
      setCheckStateInternal(state); parent()->updateCheckStateInternal();};
    inline int checkState()const {return pcheckstate;};
    inline long site()const{return ppair.site;};
    inline Bin *parent()const{return pparent;};
    inline long binNumber()const{
      if(pparent){
	return pparent->pbins.indexOf(const_cast<Bin*>(this));
      }else{
	return 0;
      }
    };
    inline long levels()const{
      if(pbins.size()==0) return 0;
      long l=0;
      for(int i=-1;++i<pbins.size();){
	const Bin *b=pbins.at(i);
	l=MyMath::max<long>(l,b->levels());
      }
      return l+1;
    };
    //    void getWyckPosListInternal(ValueArray<long> &res);
    
    
    
    inline void clear(){
      pbins.clear();
    };
    inline const String &element()const{return ppair.el;};
    void updateCheckStateInternal();
    void setCheckStateInternal(int state);
    inline long type()const{return ppair.type;};
    inline double rmin()const{ 
      if(pbintype<BTWyck) return prmin; else return ppair.pabs; };
    inline double rmax()const{ 
      if(pbintype<BTWyck) return prmax; else return ppair.pabs; };
    
    
    inline long neighbours()const { 
      long n=0;
      if( (pbintype<=BTWyck) ){
	for(int i=0;i<pbins.size();i++){
	  n+=pbins.at(i)->neighbours();
	};
      }else if(pbintype==BTSite){
	n=1;
      }
      return n;
    };
    
    
    void add(BinType bt,const PairVec &pair,double bintol,const String &prefix);
    
    void initializeCheckState(long binssort,const BondDefGroup *bdg);
    
    void print(int depth)const;
    
  private:
    Bin *pparent;
    BinType pbintype;
    int pcheckstate;
    double prmin,prmax;
    PairVec ppair;
    PtrArray<Bin> pbins;
    BBins *pbbins;
    friend class BBins;
    friend class ::BinModel;
    friend class ::BondModel;
    friend class ::BondsDlg;
    friend MyDebug operator<<(MyDebug dbg, const Bin &b);

  };
  
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class BBins{
    
  public:
    
    // what defines the first atom of the bond ?
    enum BinsCenterType{BCWyck=0,BCType=1,BCElement=2,BCSite=3,BCIndividual=4};
    // what coarsness does the bin tree have?
    enum BinsTopLevelGrouping{BTLElement=0,BTLType=1,BTLDistance=2};
    
    
    
    /////////////////////////////////
    
    BBins():prootbin(new Bin(0,this)){};
    ~BBins(){ delete prootbin;};
    BBins(const BBins &other); //disable copy constructor
    BBins &operator=(const BBins &other); //disable operator=
    
    void clearRoot(){prootbin->clear();};
    
    void calculateBins(BinsCenterType bct,BinsTopLevelGrouping btlg
		       ,long sort,long type,long site
		       ,const ValueArray<PtrArray<PairVec> > &pairs,double bintol);
    inline void initializeCheckState(const BondDefGroup *bdg){  
      prootbin->initializeCheckState(psort,bdg);
    };
    void reportBins(const Symmetry *sym);
    
    inline long sort()const{return psort;};
    
    inline long levels()const{ return prootbin->levels(); };
    Bin *rootBin()const{return prootbin;};
    
    
    
  public:
    static ValueArray<Pair<Bin::BinType,const char *> > binTypeName;
    static ValueArray<Pair<BinsCenterType,const char *> > binsCenterTypeName;
    static ValueArray<Pair<BinsTopLevelGrouping,const char *> > binsTopLevelGroupingName;
    
  private:
    BinsCenterType pbinscentertype;
    BinsTopLevelGrouping pbinstoplevelgrouping;
    long psort; // wyckoff position number
    long ptype; // user defined type (nuclear charge by default)
    long psite; // site number
    Bin  *prootbin;

  };
  
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  inline MyDebug operator<<(MyDebug dbg, const Bin &b)
  {
    dbg << b.toString()<<"checkstate="<<b.checkState();
    
    return dbg;
  }
}
#ifdef	__cplusplus
extern "C" {
#endif
  
#ifdef	__cplusplus
}
#endif


#endif /* _BBINS_H */

