/* @(#)BondDefs.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _BONDDEFS_H
#define _BONDDEFS_H 1

#include "MyQRgb.hh"
#include "SReal.hh"
#include "Polyhedron.hh"
#include "IDCache.hh"
#include "PtrArray.hh"
#include "MACROS.hh"
#include "MyDebug.hh"
#include "BBins.hh"

class Dobj;



namespace XFPLO{
  class Symmetry;

  // ----------------------------------------------------------------
  // Helper to define various modes, defining how bonds are used.
  class BondDefGroupEnum{
  public:
    enum CutOffMode{Boundary=0,AddMissingAtoms=1
		    ,RemoveIncomplete=2,RemoveLoner=3};


    inline static ValueArray<Pair<CutOffMode,const char *> > &cutoffModeNames(){
      return pcutoffmodename;
    }


    inline static const char *cutOffModeName(CutOffMode mode)
    {
      return pcutoffmodename.find(Pair<CutOffMode,const char *>(mode,""))->v2;
    }

  private:
    static ValueArray<Pair<CutOffMode,const char *> > pcutoffmodename;

  };

  /* ***********************************************************************
   * A bond def consists of two sorts and a min and max distance
   * defining possible pairs of atoms to be connected by bonds..
   * It depends on sorts, because then there is invariance with respect
   * to local lattice rotations. In fact I have to check the use of
   * bonds to determine, if we can go to sites-resolved bonds. Is this invariance
   * still important?
   * ***********************************************************************
   */
  class BondDef{
  public:
    // --------------------------------------------------------------
    // Create an empty Bond, needed for PtrArray
    BondDef():psorta(-1),psortb(-1),ppabsmin(-2.),ppabsmax(-1.){};

    // --------------------------------------------------------------
    // Create a bond. Sorts are sorted, which makes tasks easier.
    BondDef(long issa,long issb,double pabsmin,double pabsmax)
      :psorta(issa),psortb(issb),ppabsmin(pabsmin),ppabsmax(pabsmax)
    {
      if(psorta>psortb) MyMath::swap(psorta,psortb);
    };

    // --------------------------------------------------------------
    // Create a bond which has zero bound spread  from a sort and a 
    // corresponding bin. Use bbins to set appropriate bounds.
    // Sorts are sorted, which makes tasks easier.
    BondDef(long sort,const Bin *bin,const BBins *bbins)
      :psorta(sort),psortb(bin->sort()),
       ppabsmin(bin->pabs()),ppabsmax(bin->pabs())
    {
      if(psorta>psortb) MyMath::swap(psorta,psortb);
      findBondDefBounds(bbins);
    }
    
    // --------------------------------------------------------------
    // clone is needed by PtrArray
    BondDef *clone()const {return new BondDef(*this);};
    
    // --------------------------------------------------------------
    // Check if two bonds are of equal sort pattern. Sorted sorts assumed!
    inline bool eqSorts(const BondDef &o) const{
      if(psorta==o.psorta && psortb==o.psortb ){ 
	return true; 
      } 
      return false;
    };

    // --------------------------------------------------------------
    // Does the current bond allow a pair with sorta, sortb and distance pabs
    // under the tolerance tol?
    //
    //              pabs
    // input   sa     |      sb
    // this:  psa  |------|  psb
    //
    //        -tol         +tol
    // check  psa {|------|} psb
    //
    //
    inline bool permits(long sorta,long sortb,double pabs,double tol=1.0e-3)const{
      if(sorta>sortb) MyMath::swap(sorta,sortb);
      if(psorta==sorta && psortb==sortb && 
	 ppabsmin-tol<pabs && pabs<ppabsmax+tol){ 
	return true; 
      } 
      return false;
    };
    
    // --------------------------------------------------------------

    // In/Out handling.
    static void PTHeaderToFile(FILE *FH,const char *prefix,const char *tail);
    void PTBodyToFile(FILE *FH,const char *prefix,const char *tail)const;
    void readPTSectionFromFile(const Dobj &root);

  private:
    
    // --------------------------------------------------------------
    // If we have a bdef and a Bins structure look up which are the closest
    // bins of the same sort-pattern as bdef in order to extract bounds
    // for the look up of neighbours.
    // We assume that bdef.psorta or bdef.psortb == bins.psort already!
    // A Bins structure belongs to a particular sort and contains a tree of
    // neighbor atoms, sorted in a particular way. E.g. toplevel:element
    // nextlevel: WyckPos, nectlevel: sites.
    void findBondDefBounds(const BBins *bins);
    
  private:
    PROPERTY(sorta,sorta,Sorta,long)
    PROPERTY(sortb,sortb,Sortb,long)
    PROPERTY(pabsmin,pabsmin,Pabsmin,double)
    PROPERTY(pabsmax,pabsmax,Pabsmax,double)
  };
  
  /* ***********************************************************************
   * collect the bond/poly properties of a particular sort
   * ***********************************************************************
   */
  class BondDefSortProperties{
  public:  
    
    BondDefSortProperties()
      :patomcutmode(BondDefGroupEnum::Boundary)
      ,ppolyprops(defaultPolyProps){
    };
    
    
    
    
    static void PTHeaderToFile(FILE *FH,const char *prefix,const char *tail);
    void PTBodyToFile(FILE *FH,const char *prefix,const char *tail)const;
    void readPTSectionFromFile(const Dobj &root);
    
    static void setUpDefaultPolyProps();
    static PolyhedronProperties defaultPolyProps;
  private:
    PROPERTY(atomcutmode,cutOffMode,CutOffMode,BondDefGroupEnum::CutOffMode);
    CMODPROPERTY(polyprops,polyProperties,setPolyProperties,PolyhedronProperties);
  };
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class BondDefGroup{
  public:
    
    
    BondDefGroup(QRgb col);
    ~BondDefGroup(){
      pbonddeflist.clear();
    };
    
    BondDefGroup(const BondDefGroup &other)
    :pactive(other.pactive),pcolor(other.pcolor)
    ,pname(other.pname)
    ,pvisible(other.pvisible),pbintol(other.pbintol),
     ptoplevelgrouping(other.ptoplevelgrouping),pchanged(other.pchanged),
     pbonddeflist(other.pbonddeflist),pa1properties(other.pa1properties)
    {
    }
    const BondDefGroup &operator=(const BondDefGroup &other){
      MY_UNUSED(other);
      myDebug() << "BondDefGroup operator=";
      exit(0);
      return *this;
    }
    
    BondDefGroup *clone(){return new BondDefGroup(*this);};
    
    void setDefaultFromSymmetry(const Symmetry *sym);

    inline bool hasBonds(long sort)const{
      for(int i=-1;++i<pbonddeflist.size();){
	const BondDef* bd=pbonddeflist.at(i);
	if(bd->sorta()==sort||bd->sortb()==sort)
	  return true;
      }
      return false;
    };
    
    inline       void resize(long newsize,bool shrink=false){
      pbonddeflist.resize(newsize,shrink);};
    inline       long size()            const {return pbonddeflist.size();};
    inline       BondDef *bond(long i)       {return pbonddeflist[i];    }; 
    inline const BondDef *bond(long i) const {return pbonddeflist.at(i); }; 
    
    
    inline void removeAt(int i){BondDef *b=pbonddeflist.takeAt(i); 
      if(b) delete b;pchanged=true;}; 
    
    inline bool permits(long sorta,long sortb,double pabs)const
     {
       long N=pbonddeflist.size();
       BondDef* *bd=pbonddeflist.dataRef();
       for(int i=-1;++i<N;){
     	if((*bd)->permits(sorta,sortb,pabs)) return true;
     	bd++;
       }
       return false;
     }
    
    void setFromBinSelection(const BBins *bbins,const Bin *bin,bool checked);
    
    
    const PolyhedronProperties &polyProperties(long sort)const{
      //myDebug() << "here const";
      if(sort<0 || sort >=pa1properties.size()) 
	return BondDefSortProperties::defaultPolyProps;
      return pa1properties.at(sort).polyProperties();
    }
    
    PolyhedronProperties &polyProperties(long sort){
      //myDebug() << "here non const";
      if(sort>=pa1properties.size()){
	for(int i=pa1properties.size();i<=sort;i++){
	  pa1properties.append(BondDefSortProperties());
	}
      }
      return pa1properties[sort].polyProperties();
      pchanged=true;
    }
    
    
    BondDefGroupEnum::CutOffMode atomCutOffMode(long sort)const{
      if(sort<0 || sort >=pa1properties.size()) return BondDefGroupEnum::Boundary;
      return pa1properties.at(sort).cutOffMode();
    };
    
    
    void setAtomCutOffMode(long sort,BondDefGroupEnum::CutOffMode mode){
      if(sort>=pa1properties.size()){
	for(int i=pa1properties.size();i<=sort;i++){
	  pa1properties.append(BondDefSortProperties());
	}
      }
      pa1properties[sort].setCutOffMode(mode);
      pchanged=true;
    };
    
    const ValueArray<BondDefSortProperties> &a1Properties()const{
      return pa1properties;
    }
    void resizeA1Properties(long newsize,bool shrink=false){
      pa1properties.resize(newsize,shrink);
    }
    
    
    
    static void PTHeaderToFile(FILE *FH,const char *prefix,const char *tail);
    void PTBodyToFile(FILE *FH,const char *prefix,const char *tail)const;
    void readPTSectionFromFile(const Dobj &root);
    
    
  private:
    void add(const BondDef &b);
    void remove(const BondDef &b,const BBins *bin);
    void fuse();
    
  private:
    PROPERTY(active,active,Active,bool);
    CPROPERTY(color,color,Color,QRgb);
    CPROPERTY(name,name,Name,String);
    PROPERTY(visible,visible,Visible,bool);
    PROPERTY_CHANGED(bintol,binTol,BinTol,double);
    PROPERTY(toplevelgrouping,topLevelGrouping,TopLevelGrouping,
	     BBins::BinsTopLevelGrouping);
    PROPERTY(changed,changed,Changed,bool);
    PtrArray<BondDef> pbonddeflist;
    ValueArray<BondDefSortProperties> pa1properties;
  };
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class BondDefGroups{
  public:
    BondDefGroups():pchanged(false){};
    ~BondDefGroups(){ clear(); };
    
    BondDefGroups(const BondDefGroups &other)
    :pchanged(other.pchanged),pidcache(other.pidcache)
    {
      myDebug() << "BondDefGroups copy constructor";
      exit(0);
    }
    inline void clear(){
      pidcache.clear();
      pbonddefgrouplist.clear();
      pchanged=true;
    }; 
    
    const BondDefGroups &operator=(const BondDefGroups &other){
      clear();
      pchanged=other.pchanged;
      pidcache=other.pidcache;
      pbonddefgrouplist=other.pbonddefgrouplist;
      return *this;
    };
    
    inline       long size()                const{return pbonddefgrouplist.size();}; 
    inline       BondDefGroup *group(long i)     {return pbonddefgrouplist[i];}; 
    inline const BondDefGroup *group(long i)const{return pbonddefgrouplist.at(i);};
    

    inline void insert(int at,BondDefGroup *b,bool createuniquename){
      if(createuniquename) b->setName(pidcache.makeUniqueName("Bond group "));
      pbonddefgrouplist.insert(at,b);
      pchanged=true;
    }
    inline void append(BondDefGroup *b,bool createuniquename){
      if(createuniquename) b->setName(pidcache.makeUniqueName("Bond group "));
      pbonddefgrouplist.append(b);
      pchanged=true;
    }
    
    inline BondDefGroup *removeAt(int i){
      BondDefGroup *bdg=pbonddefgrouplist.takeAt(i);
      pidcache.remove(bdg->name());
      pchanged=true;
      return bdg;
    } 
    
    
    bool changed()const{ 
      bool c=pchanged; 
      for(int i=-1;++i<pbonddefgrouplist.size();){
	const BondDefGroup* bdg=pbonddefgrouplist.at(i);
	c|=bdg->changed();
      }
      return c;
    }
    
    void setChanged(bool s){pchanged=s;
      for(int i=-1;++i<pbonddefgrouplist.size();){
	BondDefGroup* bdg=pbonddefgrouplist[i];
	bdg->setChanged(s);
      }
    }
    
    
    
    
    static void PTHeaderToFile(FILE *FH,const char *prefix,const char *tail);
    void PTBodyToFile(FILE *FH,const char *prefix,const char *tail)const;
    void readPTSectionFromFile(const Dobj &root);
    
    
  private:
    
  private:
    bool pchanged;
    PtrArray<BondDefGroup> pbonddefgrouplist;
    IDCache pidcache;
    
  };
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  
  class BondDefs{
  public:
    
    
    BondDefs();
    
    void clear(){pbonddefgroups.clear();};
    
    void setDefaultFromSymmetry(const Symmetry *sym)
    {
      pbonddefgroups.clear();
      BondDefGroup *bdg=new BondDefGroup(0x9f9f9f);
      pbonddefgroups.append(bdg,true);
      bdg->setDefaultFromSymmetry(sym);
    }

    // make sure that on symmetry update no old crap is lurking in the
    // wings
    void cleanBonds(const Symmetry *sym);
    
    inline BondDefGroups *bondDefGroups(){return &pbonddefgroups;};
    inline const BondDefGroups *bondDefGroups()const{return &pbonddefgroups;};
    
    inline bool permits(long sorta,long sortb,double pabs)const{
      for(int i=0;i<pbonddefgroups.size();i++){
	const BondDefGroup *bdl=pbonddefgroups.group(i);
	for(int i=0;i<bdl->size();i++){
	  if(bdl->bond(i)->permits(sorta,sortb,pabs)) return true;
	}
      }
      return false;
    };
    
    bool changed(){return pchanged | pbonddefgroups.changed();};
    void setChanged(bool s){pchanged=s;};
    
    void PTSectionToFile(FILE *FH,const SReal &ballscale,
			       const SReal &stickballratio)const;
    void readPTSectionFromFile(const Dobj &root);
    
  private:
    bool pchanged;
    BondDefGroups pbonddefgroups;
    
  };
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  inline MyDebug operator<<(MyDebug dbg, const BondDef &bd)
  {
    dbg << Sprintf("W%ld--W%ld |p| in [%g,%g]"
		   ,bd.sorta()+1,bd.sortb()+1,bd.pabsmin(),bd.pabsmax());
    
    return dbg;
  }
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  inline MyDebug operator<<(MyDebug dbg, const BondDefGroup &bdg)
  {
    
    dbg << "     name="<<bdg.name()<<"\n";
    dbg << "     modes=\n";
    for(int i=0;i<bdg.a1Properties().size();i++){
      dbg << "     Props("<<i<<"): "
	  <<bdg.a1Properties().at(i).cutOffMode()
	  <<" "<<Sprintf("%x",bdg.a1Properties().at(i).polyProperties()
			 .faceMaterial().color())
	  <<"\n";
    }
    
    for(int i=0;i<bdg.size();i++){
      const BondDef *bd=bdg.bond(i);
      dbg << "     b["<<i<<" ]:"<<(*bd)<<"\n";
    }
    return dbg;
  }
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  inline MyDebug operator<<(MyDebug dbg, const BondDefGroups &bdgs)
  {
    for(int i=0;i<bdgs.size();i++){
      const BondDefGroup *bdg=bdgs.group(i);
      dbg << " bdg["<<i<<" ]:\n"<<*bdg;
    }
    return dbg;
  }
}
#ifdef	__cplusplus
extern "C" {
#endif

#ifdef	__cplusplus
}
#endif


#endif /* _BONDDEFS_H */

