/* @(#)Boundary.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */
#ifndef BOUNDARY_H
#define BOUNDARY_H

#include "MyDebug.hh"
#include "FString.hh"
#include "Plane.hh"
#include "Pair.hh"
#include "ValueArray.hh"
#include "Xf.hh"

class Vlevel;
class Dobj;


namespace XFPLO{
  class Symmetry;
  class Polyhedron;
}

using namespace XFPLO;

class Face{
public:
  Face():distD(0),distS("0"){pS[0]="0";pS[1]="0";pS[2]="0";}
  Face(const String p[3],const String d);
  Face(const Face &other){*this=other;}

  Face &operator=(const Face &other){
    pD=other.pD;
    distD=other.distD;
    distS=other.distS;
    pS[0]=other.pS[0];
    pS[1]=other.pS[1];
    pS[2]=other.pS[2];
    return *this;
  }
  //
  Vec3 pD;
  double distD;
  String pS[3];
  String distS;
};



class Boundary
{
public:
  Boundary();
  ~Boundary();
  const Boundary &operator=(const Boundary &other);

  enum CellType{ Cell=0,WignerSeitz=1,Faces=2};
  enum BasisType{ Primitive=0,Conventional=1,Cartesian=2,Box=3};
  enum CenteringType{ None=0,Centered=1 };


  bool changed();
  void setChanged(bool s);

  bool calculated();
  void setCalculated(bool s);

  BasisType cellBasisType() const { return pcellbasistype;};
  void setCellBasisType(BasisType);
  static const char *cellBasisTypeName(BasisType i){ return pcellbasistypename[i];};
  
  CellType cellType() const { return pcelltype;};
  void setCellType(CellType);

  const String *boxVecS(int i) const {return pboxS+3*i;};
  void setBox(const String box[9]);
  void setBox(int i,int j,const String &v);

  const Pair<String,String> *boxDistanceS(int i) const {return pboxdistS+i;};
  void setBoxDistance(int i,int fs,const String&);
  void setBoxDistances(const Pair<String,String> d[3]);

  void getBoxCell(Vec3 &origin,Matrix3x3 &box,const Symmetry *sym)const;

  CenteringType centeringType() const {return pcentered;};
  void setCentering(CenteringType);

  Xf::LatticeType latticeType() const {return plattice;};
  void setLattice(Xf::LatticeType);

  void setOrigin(const String o[3],BasisType);
  const String *originS() const{return poriginS;};
  void setOriginBasis(BasisType os) {poriginbasis=os;};
  BasisType originBasis() const {return poriginbasis;};


  inline const Vec3 &midPoint() const {return pmid;};
  inline double radius() const {return pradius;};
  // to read old files
  static CellType cellTypefromCompressed(long );
  static BasisType cellBasisTypefromCompressed(long );


  const ValueArray<Face> &faces() const {return pfaces;};
  void setFaces(const ValueArray<Face> &f);
  void removeFace(int i);
  void setDefaultFaces(const String &mode,const Symmetry *sym,BasisType b);
  void setDefault(Xf::LatticeType lt,Xf::DisplayCellType dct);
  void modifyFace(int i,int which,const String &s);
  void insertFace(long i,const Face &face);

  void copyPolyPropertiesFrom(const Boundary *other);

  inline XFPLO::Polyhedron *polyhedron(){return ppolyhedron;};
  inline const XFPLO::Polyhedron *polyhedron()const{return ppolyhedron;};


  static void PTHeaderToFile(FILE *FH,const char *prefix
				     ,const char *tail);

  void PTBodyToFile(FILE *FH,const char *prefix,const char *tail);
  void readPTSectionFromFile(const Dobj &root);
  
  String makeBoundaryCell(const Symmetry *sym,double tol,const Vlevel &vlevel);
  static void hklEnclSphere(const Vec3 mid,double radius
			    ,const Matrix3x3 &cell
			    ,long h1[2],long h2[2],long h3[2]);
  void hklEnclSphere(const Vec3 &tau,
		     const Matrix3x3 &rcell,
		     long h1[2],long h2[2],long h3[2])const;


  bool isInBoundary(const Vec3 &x,double tol) const;
  bool isInBoundary(double x,double y,double z,double tol) const;
  
  static void prepWSCellPlanes(const Matrix3x3 &primcell,ValueArray<Plane>&pla);
  
private:
  void clean();

  void initCell(ValueArray<Plane> &planes,const Symmetry *sym);
  void calcCellSphere();
  void calcCellExtend(const Matrix3x3 &rcell);
  String reportBoundary(const Vlevel &vlevel);

  static void cellReduce(Matrix3x3 &aa,double tol,bool prot);
  static void minl123(Matrix3x3 &aa,long i1,long i2,long i3,double &smin,
		      bool prot);

private:
  
  CellType       pcelltype; /* which boundary type */
  BasisType       pcellbasistype; /* which type of basis reference */
  CenteringType  pcentered; /* which centering */
  Xf::LatticeType    plattice; /* which lattice */
  BasisType      poriginbasis;	/* in which basis is porigin */
  Matrix3x3      pboxD;		/* the three vectors defining a box in
				 the basis of pcellbasistype
				 (not used in WigSei)
				*/
  String pboxS[9];		/* V1=0-2 V2=3-5 V3=6-8 */
  Pair<String,String> pboxdistS[3]; /* the box's planes are positioned
					at 0 and 1 in units of the box 
					vectors, pboxdist can change that
				      */
  Pair<double,double> pboxdistD[3];

  ValueArray<Face> pfaces;

  Vec3 poriginD;		/* the cell origin */
  String poriginS[3];		/* the cell origin */
  Vec3 porigincartesian;	/* the cell origin in cartesian */
  static const char *pcelltypename[];
  static const char *pcellbasistypename[];
  static const char *pcenteringtypename[];
  double pradius;		/* circumscribing sphere radius */
  Vec3 pmid;		/* circumscribing sphere midpoint */
  Vec3 pRADxGneg;
  Vec3 pRADxGpos;
  XFPLO::Polyhedron *ppolyhedron;
  bool pchanged,pcalced;

public:
  static const Pair<String,String> defaultBoxDistS;
  static ValueArray<Face> userFaces;
  static void setDefaultUserFaces();

};

inline MyDebug operator<<(MyDebug dbg, const Face &f)
{
  dbg<<(void*)&f<<f.pD<<f.pS[0]<<f.pS[1]<<f.pS[2]<<f.distD<<f.distS;
  return dbg.nospace();
}
  
#endif // BOUNDARY_H
