/*************************************************************************
 * @(#)Cluster.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   04 Jul 2009
 *************************************************************************
 */

#include <QApplication>
#include "Cluster.hh"
#include "Symmetry.hh"
#include "Boundary.hh"
#include "MachineTime.hh"
#include "Status.hh"
#include "MyMath.hh"
#include "BondDefs.hh"
#include "Interrupt.hh"




/* ***********************************************************************
 *
 * ***********************************************************************
 */
Cluster::Cluster()
  : pradius(0.),pcalced(false)
{
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Cluster::~Cluster()
{
  ppolyhedrons.clear();
}

/* ***********************************************************************
 * calculate all atoms within a certain boundary
 * add bonds and polyhedra
 * ***********************************************************************
 */

String Cluster::create(const Symmetry *sym,const Boundary *bound
			,const BondDefs *bonddefs
			,Status *status)
{
  NSF("Cluster::create");
  //  myDebug() << nsubfunc;
  MachineTime time;
     
  double tol=1.0e-3;

  ppolyhedrons.clear();
  pbonds.clear();
  patoms.clear();
  const ValueArray<Site> &psites=sym->sites();
  
  double pc=status->getCurrentProgress();
  double alloted=status->allotedPercent();
  double pf=pc+alloted*(1-pc);
  double pf1=pc+0.5*(pf-pc);
  double D=(pf1-pc)/psites.size();
  
  //myDebug()<< "have"  << pc << pc+D*psites.size() << pf1 << pf;
  
  
 if(sym->structureType()== Xf::STMolecule){
    for(long is=0;is<psites.size();is++){
      patoms.append(Atom(is,psites.at(is),is));
    }
  }else{
    

   Matrix3x3 cell=sym->cell();
   Matrix3x3 rcell=sym->rcell();
      
    long h1[2],h2[2],h3[2];

    
    //myDebug() << "collect atoms";
    long ica=0;
    /* for all sites in unit cell */
    for(long is=0;is<psites.size();is++){
      const Vec3 &tau=psites.at(is).tauD();
      //      printf("p; %g %g %g\n",p[0],p[1],p[2]);     
      
      status->setAllotedPercent(D/ (1-pc-D*is));
      

      bound->hklEnclSphere(tau,rcell,h1,h2,h3);
      
      //myDebug()<<h1[0]<<h1[1]<<h2[0]<<h2[1]<<h3[0]<<h3[1];

      /* for all lattice vectors enclosing the boundary */
      long nn=0;
      double dt=(h1[1]-h1[0])*(h2[1]-h2[0])*(h3[1]-h3[0]);
      
      double pp=status->getCurrentProgress(); 
      double pf=pp+status->allotedPercent()*(1-pp);
      dt=(pf-pp)/dt;
      //myDebug() << is<< pp << pf;

      for(long h=h1[0];h<h1[1];h++){
	for(long k=h2[0];k<h2[1];k++){
	  for(long l=h3[0];l<h3[1];l++){

	    status->setProgress(pp+dt*nn);
	    //myDebug()<< "have"  << pp+dt*nn;
	    nn++;
	    //myDebug() << (nn-1)/dt;
	    if((nn)%100 ==0 ) if(interrupt.requested()) return ""; 
	    

	    /* translate site */
	    Vec3 x=tau+cell*Vec3(h,k,l);
	    
	    //printf("x; %g %g %g in=%d\n",x[0],x[1],x[2],bound->isInBoundary(x,tol)); 
	    
	    if(bound->isInBoundary(x,tol)){
	      /* site is within bounary take it */
	      
	      patoms.append(Atom(ica,psites.at(is),is,x) );
	      ica++;
	      /* printf("x; %g %g %g\n",x[0],x[1],x[2]); */
	    } /* isinboundary */
	  } /* l */
	} /* k */
      } /* h */
      status->setProgress(pf);
    } /* is */

    

    
    
    
  }
  status->setProgress(pf1);
  status->setAllotedPercent((pf-pf1)/(1-pf1));

  //myDebug() << "PF1" <<pf1;
  //myDebug() << "PF" <<pf;

  //pnatom=patoms.size();


  time.print(String(nsubfunc)+" atoms");

  createBonds(sym,bonddefs,status,bound->latticeType(),bound->radius());
  
    

  pmid=0;
  for(long i=0;i<patoms.size();i++){
    pmid+=patoms.at(i).tauD;
  }
  if(patoms.size()) pmid/=patoms.size();
  pradius=0.;
  for(long i=0;i<patoms.size();i++){
    pradius=MyMath::max<double>(pradius,norm(patoms.at(i).tauD-pmid)
				+patoms.at(i).NN());
  }
  if(patoms.size()==1){
    pradius=patoms.at(0).atomicRadius();
  }
  
  //myDebug() << "mid" << pmid.printS()() << pradius;
  //  myDebug() << "END" << nsubfunc;

  pcalced=true;

  time.print(String(nsubfunc)+"with bonds");
  
  return reportCluster();

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
class RelativeBond
{
public:
  RelativeBond(){}
  RelativeBond(long ia,long ib,const Vec3 &p,const QRgb &rgb,bool v
	       ,BondDefGroupEnum::CutOffMode cma
	       ,BondDefGroupEnum::CutOffMode cmb
	       )
    :isa(ia),isb(ib),pBA(p),color(rgb),visible(v)
    ,cutmodea(cma),cutmodeb(cmb){ 
    if(isa>isb) { MyMath::swap(isa,isb); pBA=-pBA; MyMath::swap(cutmodea,cutmodeb);}
  };
  
  bool operator==(const RelativeBond &o)
  {
    if(isa!=isb){
      return ((isa==o.isa) && ( isb==o.isb)
	      && ( norm(pBA-o.pBA)< bondtol ));
    }else{
      if(o.isa!= isa || o.isa!= o.isb) return false; 
      return (norm(pBA-o.pBA)< bondtol || norm(pBA+o.pBA)< bondtol);
    }
  };
  
  RelativeBond operator-() const {
    return RelativeBond(isa,isb,-pBA,color,visible,cutmodea,cutmodeb);};
  
  void updateFrom(const RelativeBond &other){
    color=other.color;
    visible=other.visible;
    if(isa!=isb){
      if((isa==other.isa) && (isb==other.isb)){
	cutmodea=other.cutmodea;
	cutmodeb=other.cutmodeb;
      }else if((isa==other.isb) && (isb==other.isa)){
	cutmodea=other.cutmodeb;
	cutmodeb=other.cutmodea;
      }
    }else{
      cutmodea=other.cutmodea;
      cutmodeb=other.cutmodeb;
    }
  }

  
  long isa,isb;
  Vec3 pBA;  //vector from site A to site B
  QRgb color;
  bool visible;
  BondDefGroupEnum::CutOffMode cutmodea;
  BondDefGroupEnum::CutOffMode cutmodeb;
  static double bondtol;
};

double RelativeBond::bondtol=1.0e-4;


/* ***********************************************************************
 *
 * ***********************************************************************
 */
inline MyDebug operator<<(MyDebug dbg, const RelativeBond &rb)
{
  dbg << Sprintf(" relbond: %ld->%ld [%s] col(0x%08x),vis(%d) cuta(%d) cutb(%d)",
		 rb.isa+1,rb.isb+1,
		 rb.pBA.printS("%12.5f",",")(),
		 rb.color,rb.visible,
		 rb.cutmodea,
		 rb.cutmodeb) ;
  return dbg;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static bool hasCutMode(const ValueArray<RelativeBond> &relativebonds
		       ,BondDefGroupEnum::CutOffMode mode)
{
  for(long i=-1;++i<relativebonds.size();){ 
    const RelativeBond &b=relativebonds.at(i); 
    if(b.cutmodea==mode||b.cutmodeb==mode) return true;
  }
  return false;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Cluster::createBonds(const Symmetry *sym
			  ,const BondDefs *bonddefs
			  ,Status *status
			  ,Xf::LatticeType latticetype
			  ,double alloverscale)
{
  NSF("Cluster::createBonds");

  bool prot_bonds=false;
  double tol=1.0e-4;

  MachineTime time;
 
  long natom=patoms.size();
  
  ValueArray<RelativeBond> relativebonds;

  const BondDefGroups *bdgs=bonddefs->bondDefGroups();

  double pc=status->getCurrentProgress();
  double alloted=status->allotedPercent();
  double pf=pc+alloted*(1-pc);
  
  const ValueArray<PtrArray<PairVec> > &pairs=sym->pairs();

  PtrArray<AtomPolyhedron> polytmp;

  //myDebug() << "pc" <<pc;
  
  double dt=patoms.size()*(patoms.size()-1)/2;
  dt=(pf-pc)/dt;

  // make unique Bond objects for one unit cell
  // this way we are much faster for large boundary cells
  // since locality is used
  
  for(long ibdefs=0;ibdefs<bdgs->size();ibdefs++){
    const BondDefGroup *bdg=bdgs->group(ibdefs);
    if(!bdg->active()) continue;
    //myDebug()<<"BondDefGruop"<<ibdefs<<*bdg;
    
    AtomPolyhedron *pol=0;
    for(long site=0;site<sym->nsite();site++){
      //myDebug()<<"site="<<site+1;
      long sort=sym->sortBySite(site);

      pol=0;
      const PolyhedronProperties &polyprop=bdg->polyProperties(sort);
      if(polyprop.isVisible()){
	polytmp.append(pol=new AtomPolyhedron());
	pol->relatedsite=site;
	pol->properties()=polyprop;
      }
      ValueArray<Vec3> neighbours;
      neighbours << Vec3();
      
      for(long ibd=0;ibd<bdg->size();ibd++){
	long sorta=bdg->bond(ibd)->sorta();
	long sortb=bdg->bond(ibd)->sortb();
	if(sort==sorta){
	}else if(sort==sortb){
	  MyMath::swap(sorta,sortb);
	}else{
	  continue;
	}
	double pabsmin=bdg->bond(ibd)->pabsmin();
	double pabsmax=bdg->bond(ibd)->pabsmax();
	
	for(long ip=0;ip<pairs.at(site).size();ip++){
	  const PairVec *pa=pairs.at(site).at(ip);
	  //myDebug()<<site<<pa.site<<pa.p;
	  if(pa->pabs>pabsmax+tol) break;
	  if(pa->sort!=sortb) continue;
	  if( pabsmin<=pa->pabs && pa->pabs <=pabsmax ){
	    // myDebug()<<"pair"<<ip<<"site="<<pa->site+1<<"p="<<pa->p<<"sort="<<sort+1
	    // 	     <<"pasort="<<sym->sortBySite(pa->site)+1
	    // 	     <<"cuta="<<bdg->atomCutOffMode(sort)
	    // 	     <<"cutb="<<bdg->atomCutOffMode(sym->sortBySite(pa->site))
	      ;
	    RelativeBond b(site,pa->site,pa->p,bdg->color(),bdg->visible()
			   ,bdg->atomCutOffMode(sort)
			   ,bdg->atomCutOffMode(sym->sortBySite(pa->site))
			   );
	    long indx=-1;
	    if((indx=relativebonds.indexOf(b))<0){
	      relativebonds.append(b);
	    }else{
	      // myDebug()<<indx<<relativebonds.at(indx);
	      // myDebug()<<"b"<<b;
	      relativebonds[indx].updateFrom(b);
	      // myDebug()<<"n"<<relativebonds.at(indx);
	      // myDebug();
	    }
	    if(polyprop.isVisible()&& ( sorta==sort)){
	      neighbours.append(pa->p);
	      //myDebug()<<"add"<<pa->site;
	    }
	    
	  }
	}
      }
      if(pol){
	pol->calculateFaces(neighbours,10.);//todo theta0
	pol->properties().setCalculated(false);
	pol->properties().calculate(latticetype,alloverscale,sym->dataLengthUnit());
	//myDebug()<<"poly around"<<site<<pol->properties();
      }
    }
    
    
    
    
    
    if(0||prot_bonds){
      printf("Bonds\n");
      for(long ib=0;ib<relativebonds.size();ib++){
	printf("bond %3ld: %s(W%ld,T%ld) %3ld -> %s(W%ld,T%ld) %3ld , "
	       "(%s) |p|=%- 10.5g\n",ib
	       ,sym->sites().at(relativebonds[ib].isa).el()()
	       ,sym->sites().at(relativebonds[ib].isa).sort()+1
	       ,sym->sites().at(relativebonds[ib].isa).type()
	       ,relativebonds[ib].isa+1
	       ,sym->sites().at(relativebonds[ib].isb).el()()
	       ,sym->sites().at(relativebonds[ib].isb).sort()+1
	       ,sym->sites().at(relativebonds[ib].isb).type()
	       ,relativebonds[ib].isb+1
	       ,relativebonds[ib].pBA.printS("%- 10.5g")()
	       ,norm(relativebonds[ib].pBA)
	       );
      }
    }
    
    
    
    
    // delete atoms
    for(long ica=0;ica<patoms.size();ica++){
      const Atom &atom=patoms[ica];
      long sort=atom.sort;
      if(
	 bdg->atomCutOffMode(sort)!=BondDefGroupEnum::RemoveIncomplete&&
	 bdg->atomCutOffMode(sort)!=BondDefGroupEnum::RemoveLoner )continue;
      bool incomplete=false;
      bool hasbonds=false;
      bool hassomebonds=false;
      
      for(long ib=0;ib<relativebonds.size();ib++){
	
	if(relativebonds[ib].isa==atom.site){
	  
	  hasbonds=true;
	  Vec3 pb=atom.tauD+relativebonds[ib].pBA;
	  if(findByPosition(pb)<0)
	    incomplete=true;
	  else
	    hassomebonds=true;
	  
	  
	}
	if(relativebonds[ib].isb==atom.site){
	  
	  hasbonds=true;
	  Vec3 pb=atom.tauD-relativebonds[ib].pBA;
	  if(findByPosition(pb)<0)
	    incomplete=true;
	  else
	    hassomebonds=true;
	  
	}
      }
      if(hasbonds && incomplete){
	if(bdg->atomCutOffMode(sort)==BondDefGroupEnum::RemoveIncomplete ){
	  patoms[ica].hidden=true;
	}
	if( !hassomebonds &&
	    bdg->atomCutOffMode(sort)==BondDefGroupEnum::RemoveLoner ){
	  patoms[ica].hidden=true;
	}
      }      
    }
    
  }// for all bondgroups
  
  time.print(String(nsubfunc)+" after all bondgroups");

  for(long ica=patoms.size()-1;ica>=0;ica--){
    if(patoms.at(ica).hidden)
      patoms.removeAt(ica);
  }
  for(long ica=patoms.size()-1;ica>=0;ica--){
    patoms[ica].index=ica;
  }
  
  natom=patoms.size();
  
  // add atoms where one of a bond is missing
  long nn=0;
  if(hasCutMode(relativebonds,BondDefGroupEnum::AddMissingAtoms)){
    //myDebug() << "add atoms";
    for(long ib=0;ib<relativebonds.size();ib++){
      long isa=relativebonds[ib].isa;
      BondDefGroupEnum::CutOffMode cutmodea=relativebonds[ib].cutmodea;
      long isb=relativebonds[ib].isb;
      BondDefGroupEnum::CutOffMode cutmodeb=relativebonds[ib].cutmodeb;
      
      
      
      for(long ica=0;ica<natom;ica++){
	
	
	if((nn)%500 ==0 ) if(interrupt.requested()) return; 
	nn++;
	
	
	//todo: remove unbonded but not  in bonds, it is a global property
	//todo: remove all atoms bonded to incomplete atom
	const Atom atom=patoms[ica];
	if(atom.hidden) continue;
	if(atom.site == isa && 
	   (cutmodea==BondDefGroupEnum::AddMissingAtoms)){
	  Vec3 pb=atom.tauD+relativebonds[ib].pBA;
	  int ind=findByPositionSite(pb,isb);
	  if(ind<0){
	    Site ssite=sym->sites().at(isb);
	    ssite.setTauD(pb);
	    patoms.append(Atom(patoms.size(),ssite,isb));
	  }
	}
	//continue;
	if(atom.site == isb && 
	   (cutmodeb==BondDefGroupEnum::AddMissingAtoms) ){
	  Vec3 pb=atom.tauD-relativebonds[ib].pBA;
	  int ind=findByPositionSite(pb,isa);
	  if(ind<0){
	    Site ssite=sym->sites().at(isa);
	    ssite.setTauD(pb);
	    patoms.append(Atom(patoms.size(),ssite,isa));
	  }
	}
	
	
	
      }
      // myDebug() << ppolyhedrons.size() << ica << icb;
      
    }
  }
  time.print(String(nsubfunc)+" after adding missing");

  
  
  // perpetuate the Bonds onto the cluster
  nn=0;
  for(long ib=0;ib<relativebonds.size();ib++){
    long isa=relativebonds[ib].isa;
    long isb=relativebonds[ib].isb;
    for(long ica=0;ica<patoms.size();ica++){
      
      status->setProgress(pc+dt*nn);
      //myDebug() << "pc" <<pc+dt*nn;
      
      if((nn)%500 ==0 ) if(interrupt.requested()) return; 
      nn++;
      const Atom &atom=patoms[ica];
      //myDebug()<<"atom"<<atom;
      if(atom.hidden) continue;
      if(atom.site != isa ) continue;
      Vec3 pb=atom.tauD+relativebonds[ib].pBA;
      int ind=findByPositionSite(pb,isb);
      if(ind>=0){
	double radius=(atom.NN()+patoms[ind].NN())*0.5;
	pbonds.append(Bond(atom.tauD,pb,relativebonds[ib].color,1.
			   ,radius,relativebonds[ib].visible));
      }
      // myDebug() << ppolyhedrons.size() << ica << icb;
      
    }
  }
  time.print(String(nsubfunc)+" after putting bonds on cluster");
  
  //put polys on cluster
  for(long ip=0;ip<polytmp.size();ip++){
    long site=polytmp[ip]->relatedsite;
    if(polytmp[ip]->polyIndices().size()==0) continue;
    for(long ica=0;ica<patoms.size();ica++){
      
      const Atom &atom=patoms[ica];
      if(atom.hidden) continue;
      if(atom.site != site ) continue;
      AtomPolyhedron *pp=new AtomPolyhedron();
      *pp=*(polytmp.at(ip));
      pp->position=atom.tauD;
      ppolyhedrons.append(pp);
    }
  }
  
  
  
  time.print(String(nsubfunc)+"after putting polys");
  
  
  if(0){
    myDebug()<<"polies"<<ppolyhedrons.size();
    for(long ip=0;ip<ppolyhedrons.size();ip++){
      myDebug("%ld: %s",ip+1,ppolyhedrons[ip]->printS()());
    }
  }
  
  if(0){
    myDebug()<<"bond"<<pbonds.size();
    for(long ip=0;ip<pbonds.size();ip++){
      myDebug("%ld: %s",ip+1,pbonds[ip].printS()());
    }
  }
  
  
  
  status->setProgress(pf);
  //myDebug() << "pf" <<pf;
  
  
  //  myDebug() << "END Cluster::createBonds";
    
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String Cluster::reportCluster()
{
  NSF("Cluster::reportCluster()");
  String s;

  MachineTime time;


  const char *line70="--------------------------------------"
    "--------------------------------\n";

  s+="\n";
  s+=line70;
  s+="                     Report Atoms in Boundary    \n";
  s+=line70;
  
  /* sites positions */
  s+="\n                   ATOMS\n\n"; 
  s+=Sprintf("Number of atoms : %ld\n",patoms.size()); 
  s+="No.   WPS   SITE  Element  Type    X               Y    "
    "           Z  \n";
  for(long i=0;i<patoms.size();i++){
    s+=Sprintf(" %-4ld %-4ld  %-4ld  %-4s     %-4ld"
	       "   %-15.5f %-15.5f %-15.5f\n"
	       ,i+1,patoms[i].sort+1,patoms[i].site+1,patoms[i].el()
	       ,patoms[i].type,patoms[i].tauD[0],patoms[i].tauD[1]
	       ,patoms[i].tauD[2]);
  }


  s+="\n";
  s+=line70;
  s+="\n";

  time.print(nsubfunc);
  return s;
}





