/* @(#)DMDoc.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _DMDOC_H
#define _DMDOC_H 1

#include "DMQuantAxis.hh"
#include "FString.hh"
#include "SqMatrix.hh"
#include "Cmplx.hh"
#include "ValueArray.hh"
#include "MACROS.hh"
#include "Vec3.hh"





/* ***********************************************************************
 *
 * ***********************************************************************
 */
class Occu
{
  // hold the data for a LSDA+U occupation number matrix
  // provide transformation tools to create different representations
 public:

  // ----------------------------------------------------------------
  enum YlmRepresentation{RealYlm=0,ComplexYlm=1,SphericalSpinor=2};

  // ----------------------------------------------------------------
  class Props{
    // properties resulting from data analysis
  public:
    Props():N(0.){}
    void reset(){N=0.;L=0.;S=0.;J=0.;}
  public:
    double N;
    Vec3 L,S,J;
  };

  // ----------------------------------------------------------------
  Occu() : pel("??"),pl(-1),pnmain(-1),psite(-1),pdegl(0),currentaxes(0),
   	   pcurrentylmrep(RealYlm){ 
    localaxes.append(DMQuantAxis());
  };

  Occu(const Occu&); // forbidden

  // allow inplicit Occu &operator=(const Occu&);

  // ----------------------------------------------------------------
  ~Occu(){};
  
  // ----------------------------------------------------------------
  // the angular momentum of the shell
  long l()const{return pl;};

  // ----------------------------------------------------------------
  // a string containing something like 2p,5d, 4f
  String orbital() const{ return String::intToOrbital(pnmain,pl);};

  // ----------------------------------------------------------------
  bool operator==(const Occu &o)const
  {
    // this does not consider representations, but identifies an nl shell of
    // a particular atom
    return (pel==o.pel && psite==o.psite && pnmain==o.pnmain && pl==o.pl);
  }

  // ----------------------------------------------------------------
  void fromName(const String &name);

  // ----------------------------------------------------------------
  // transform to temporary representation, using the global spin quantization axis
  // and return data analysis
  void toRepresentation(YlmRepresentation rep,const Vec3 &spinzaxis,Props &pr);

  // ----------------------------------------------------------------
  // put current temporary representation back into default rep
  void fromRepresentation(const Vec3 &spinzaxis,Props &pr);

  // ----------------------------------------------------------------
  // 
  void flipSpin();// in current rep
  // 
  void flipLz();// in current rep

  // ----------------------------------------------------------------
  void setCurrentAxes(long ca){currentaxes=ca;};

  // ----------------------------------------------------------------
  long currentAxes()const {return currentaxes;};
  
  // ----------------------------------------------------------------
  // prepare the axis transformations
  void calcTrans(bool needDj);

  // ----------------------------------------------------------------
  void setL(long ll){
    pl=ll;
    pdegl=2*pl+1;
    //myDebug() << "l=" << pl<< "pdegl="<<pdegl;
    sp.resize(2*pdegl);
    
    // trans from RealYlm to ComplexYlm
    pU.resize(2*pdegl);
    pU.copyBlock(0,0,makeYlmUMatrix(pl),0,0,pdegl,pdegl);
    pU.copyBlock(pdegl,pdegl,pU,0,0,pdegl,pdegl);

    // vec{sigma}
    Cmplx ci=Cmplx(0.,1.);
    SqMatrix<Cmplx> sigma[3];
    sigma[0].resize(2*pdegl);
    sigma[1].resize(2*pdegl);
    sigma[2].resize(2*pdegl);
    for(int i=0;i<pdegl;i++){
      sigma[0](i+pdegl,i)=1.;
      sigma[0](i,i+pdegl)=1.;
      sigma[1](i+pdegl,i)=ci;
      sigma[1](i,i+pdegl)=-ci;
      sigma[2](i,i)=1.;
      sigma[2](i+pdegl,i+pdegl)=-1.;
    }



    for(int k=-1;++k<3;){
      pLxyz[ComplexYlm][k]=LxyzYlmChi(pl,k);
      pLxyz[RealYlm][k]=pU*pLxyz[ComplexYlm][k]*pU.hermiteConjugated();
      pLxyz[SphericalSpinor][k]=LxyzSphericalSpinorLShell(pl,k);
      psigma[RealYlm][k]=sigma[k];
      psigma[ComplexYlm][k]=sigma[k];
      psigma[SphericalSpinor][k]=SigmaxyzSphericalSpinorLShell(pl,k);
    }

    
    ptheta[RealYlm]=timeReversalRealYlmChi(pl);
    ptheta[ComplexYlm]=timeReversalYlmChi(pl);
    ptheta[SphericalSpinor]=timeReversalSphericalSpinorLShell(pl);

    // myDebug()<<"Lx";
    // printSqMatrix(pLxyz[0],"(%6.3f, %6.3f) ");

    // myDebug()<<"Ly";
    // printSqMatrix(pLxyz[1],"(%6.3f, %6.3f) ");
    
    // myDebug()<<"Lz";
    // printSqMatrix(pLxyz[2],"(%6.3f, %6.3f) ");


  };


  // ----------------------------------------------------------------
  inline String name()const{
    return Sprintf("%2s(%ld) %2s",pel(),psite,orbital()());};

  // ----------------------------------------------------------------
  inline String printS() const
  {
    return Sprintf("Element=%s site=%d nmain=%d l=%d "
		   ,pel(),psite,pnmain,pl);
  };

  // ----------------------------------------------------------------
  const Cmplx &currentRepMatElement(int ms,int i,int j) const
  {
    // return matel of current representation
    int si,sj; getSpinOffset(ms,si,sj);
    return trsp(i+si,j+sj);
  }

  // ----------------------------------------------------------------
  void currentRepMatSet(int ms,int i,int j,const Cmplx &c)
  {
    // set matel of current representation
    int si,sj; getSpinOffset(ms,si,sj);
    
    trsp(i+si,j+sj)=c;
    trsp(j+sj,i+si)=conjg(c);
    // myDebug()<<"trsp("<<i+si<<","<<j+sj<<")="<<c;
    // myDebug()<<"trsp("<<j+sj<<","<<i+si<<")="<<conjg(c);
    
  }

  // ----------------------------------------------------------------
  void currentRepMatSetToDiag(int ms,const Cmplx &c)
  {
    // set some default diagonal matrices
    int si,sj; getSpinOffset(ms,si,sj);
    int dim=pdegl;
    if(pcurrentylmrep==SphericalSpinor){
      si=sj=0;
      dim=2*pdegl;
    }
    
    for(int ii=0;ii<dim;ii++){
      for(int jj=0;jj<dim;jj++){
	if(ii==jj){
	  trsp(ii+si,ii+sj)=c;
	  if(si!=sj) trsp(ii+sj,ii+si)=conjg(c);
	}else{
	  trsp(ii+si,jj+sj)=Cmplx();
	  if(si!=sj) trsp(jj+sj,ii+si)=Cmplx();
	}
      }
    }

    //myDebug()<<"norm trsp="<<trsp.Fnorm();
    //printSqMatrix(trsp,"(%6.3f, %6.3f) ");
  }
  

  // ----------------------------------------------------------------
private:
  
  // ----------------------------------------------------------------
  void analyseProps(Props &pr)
  {
    pr.reset();
    
    pr.N+=trsp.trace().real();
    pr.S=SVec(trsp);
    pr.L=LVec(trsp);
    pr.J=pr.L+pr.S;
  }

  // ----------------------------------------------------------------
  void getSpinOffset(int ms,int &si, int &sj)const
  {
    // get index offset for spin sub matrices, if applicable
    if(ms==0 || (pcurrentylmrep==SphericalSpinor)){
      si=sj=0;
    }else if(ms==1){
      si=sj=pdegl;
    }else{
      si=0;
      sj=pdegl;
    }
  }
  


  // ----------------------------------------------------------------
  Vec3 SVec(const SqMatrix<Cmplx> a)const
  {
    Vec3 res;
    res[0]=trace(a,psigma[pcurrentylmrep][0]).real()*0.5;
    res[1]=trace(a,psigma[pcurrentylmrep][1]).real()*0.5;
    res[2]=trace(a,psigma[pcurrentylmrep][2]).real()*0.5;
    return res;
  }

  // ----------------------------------------------------------------
  Vec3 LVec(const SqMatrix<Cmplx> a)const
  {
    Vec3 res;
    res[0]=trace(a,pLxyz[pcurrentylmrep][0]).real();
    res[1]=trace(a,pLxyz[pcurrentylmrep][1]).real();
    res[2]=trace(a,pLxyz[pcurrentylmrep][2]).real();
    return res;
  }

  // ----------------------------------------------------------------
  // return a transformation matrix, which flips the spin vector and not 
  // the orbital moment vector
  SqMatrix<Cmplx> getSpinFlipMatrix(const SqMatrix<Cmplx> &a,
				    YlmRepresentation ylmrep)const;

  // ----------------------------------------------------------------
  
public:
  QList<DMQuantAxis> localaxes;


  // ----------------------------------------------------------------
private:
  CPROPERTY(el,el,El,String);
  long pl;
  ROPROPERTY(nmain,nmain,long);
  PROPERTY(site,site,Site,long);
  ROPROPERTY(degl,degL,long);
  long currentaxes;
  PROPERTY(currentylmrep,currentYlmRep,CurrentYlmRep,YlmRepresentation);
  SqMatrix<Cmplx> pU; // CYlm=RYlm*pU
  SqMatrix<Cmplx> pLxyz[3][3]; // \vec{L} for each rep
  SqMatrix<Cmplx> psigma[3][3]; // \vec{sigma} for each rep
  SqMatrix<Cmplx> ptheta[3]; // time reversal for each rep



  // the occupation number matrix
  SqMatrix <Cmplx> sp; // in global real-Ylm spin-quantaxis settings
  SqMatrix <Cmplx> trsp; // in temporary representation
  
  friend class DMDoc;
};




/* ***********************************************************************
 *
 * ***********************************************************************
 */
class DMDoc
{
 public:

  // ----------------------------------------------------------------
  DMDoc(const String &mode);

  // ----------------------------------------------------------------
  ~DMDoc();

  // ----------------------------------------------------------------
  bool load();
  bool save();

  // ----------------------------------------------------------------
  void saveAxes();
  void loadAxes();
  
  // ----------------------------------------------------------------
  ValueArray<Occu> &occus(){return poccus;};
  
  // ----------------------------------------------------------------
  const String &mode()const{return pmode;};

  // ----------------------------------------------------------------
 private:
  ValueArray<Occu> poccus;
  String pmode;

  ROCPROPERTY(filename,fileName,String);
  PROPERTY(frel,isFREL,FREL,bool);
  CPROPERTY(globalspinzaxis,globalSpinZaxis,GlobalSpinZaxis,Vec3)

};


#ifdef	__cplusplus
extern "C" {
#endif

#ifdef	__cplusplus
}
#endif


#endif /* _DMDOC_H */

