/*************************************************************************
 * @(#)DisplayCells.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   09 Mar 2011
 *************************************************************************
 */

#include "Boundary.hh"
#include "DisplayCells.hh"
#include "Parser.hh"
#include "Except.hh"
#include "Polyhedron.hh"
#include "Vlevel.hh"
#include "MyDebug.hh"
#include "Symmetry.hh"


using namespace XFPLO;

/* ***********************************************************************
 *
 * ***********************************************************************
 */
DisplayCell::DisplayCell()
  :ptype(Xf::DCTGeneral),pchanged(false),pcalced(false)
{
  pboundary=new Boundary();
  pname=Xf::displayCellTypeName(ptype);
  pid="Display cells "+pname;
  setChanged(false);
  setCalculated(false);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
DisplayCell::DisplayCell(Xf::DisplayCellType type,Xf::LatticeType lt)
  :ptype(type),pchanged(false),pcalced(false)
{


  pboundary=new Boundary();
  pboundary->setDefault(lt,type);

  pname=Xf::displayCellTypeName(ptype);
  pid="Display cells "+pname;
  if(ptype==Xf::DCTMainBoundary) {
    pboundary->polyhedron()->properties().setVisible(true);
    pboundary->polyhedron()->properties().setRimMaterial(Material(0x0,50.));
    pboundary->polyhedron()->properties().setRimRadius(SReal("0.01"));
  }else if(ptype==Xf::DCTConventional) {
    pboundary->polyhedron()->properties().setVisible(false);
    pboundary->polyhedron()->properties().setRimMaterial(Material(0xff7f00,50.));
    pboundary->polyhedron()->properties().setRimRadius(SReal("0.015"));
  }else if(ptype==Xf::DCTPrimitive)    {
    pboundary->polyhedron()->properties().setVisible(false);
    pboundary->polyhedron()->properties().setRimMaterial(Material(0x0C00FF,50.));
    pboundary->polyhedron()->properties().setRimRadius(SReal("0.015"));
  }else{
    pboundary->polyhedron()->properties().setRimRadius(SReal("0.02"));
  }
  setChanged(false);
  setCalculated(false);

  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
DisplayCell::DisplayCell(const DisplayCell &o)
  :ptype(o.ptype),pid(o.pid),pname(o.pname)
{
  pboundary=new Boundary();
  *pboundary=*o.pboundary;
  pchanged=o.pchanged;
  pcalced=o.pcalced;
};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
const DisplayCell &DisplayCell::operator=(const DisplayCell &)
{
  myDebug()<<"DisplayCell::operator=";
  exit(1);
  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
DisplayCell::~DisplayCell()
{
  delete pboundary;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool DisplayCell::isVisible()const
{
  return pboundary->polyhedron()->properties().isVisible();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCell::setVisible(bool v)const
{
  pboundary->polyhedron()->properties().setVisible(v);
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool DisplayCell::changed()const
{
  bool c=pchanged;
  if(pboundary)
    c|=pboundary->changed();
  return c;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCell::setChanged(bool c)
{
  pchanged=c;
  if(pboundary)
    pboundary->setChanged(c);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool DisplayCell::calculated()
{
  bool c=pcalced;
  if(pboundary)
    c&=pboundary->calculated();
  return c;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCell::setCalculated(bool s)
{
  if(pboundary)
    pboundary->setCalculated(s);
  pcalced=s;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String DisplayCell::calculate(Xf::LatticeType latticetype
			      ,double alloverscale,const Symmetry *sym
			      ,double tol,const Vlevel &level)
{
  NSF("DisplayCell::calculate");
  //myDebug()<<nsubfunc;

  String ret="";
  if(calculated())return ret;
  //myDebug()<<"props"<<pproperties.changed()<<pproperties.calculated();
  double scale=alloverscale;
  if(pboundary){
    //myDebug()<<"props"<<pboundary->changed()<<pboundary->calculated();
    if(ptype==Xf::DCTMainBoundary){
      ret=pboundary->makeBoundaryCell(sym,tol,level);
      scale=pboundary->radius();
    }else{
      pboundary->makeBoundaryCell(sym,tol,Vlevel::Silent);
    }
    pboundary->polyhedron()->properties().calculate(latticetype,scale,
						    sym->dataLengthUnit());
  }
  pchanged=false;
  pcalced=true;
  return ret;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCell::PTHeaderToFile(FILE *FH,const char *prefix,const char *tail)
{
  String s=String(prefix)+"       ";
  const char *ps=s();
  fprintf(FH,
	  "%1$sstruct{\n"
	  "%1$s       char[*] name;\n"
	  "%1$s       struct  {int type;char[*] description;} celltype;\n"
	  ,prefix);


  Boundary::PTHeaderToFile(FH,ps,";\n");
  
  PolyhedronProperties::PTHeaderToFile(FH,prefix,";\n");


  fprintf(FH,"%s}displayCell%s",prefix,tail);

}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCell::PTBodyToFile(FILE *FH,const char *prefix
					  ,const char *tail)const
{
  String s=String(prefix)+"       ";
  const char *ps=s();
  fprintf(FH,"%s{\"%s\",{%d,\"%s\"},\n"
	  ,prefix,pname(),ptype
	  ,Xf::displayCellTypeName(ptype)
	  );

  pboundary->PTBodyToFile(FH,ps,",\n");
  pboundary->polyhedron()->properties().PTBodyToFile(FH,ps,"\n");

  fprintf(FH,"%s}%s",prefix,tail);
  
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCell::readPTSectionFromFile(const Dobj &root)
{
  pname=root("name")()();
  ptype=(Xf::DisplayCellType)root("celltype.type").valueL();
  pboundary->readPTSectionFromFile(root("boundary"));

  pboundary->polyhedron()->properties().readPTSectionFromFile(root);

  setChanged(false);
  setCalculated(false);
}




























/* ***********************************************************************
 *
 * ***********************************************************************
 */
DisplayCells::DisplayCells()
  :pchanged(false),pcalced(false)
{
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
DisplayCells::DisplayCells(const DisplayCells &)
{
  myDebug()<<"DisplayCells copy constructor";
  exit(1);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
const DisplayCells &DisplayCells::operator=(const DisplayCells &other)
{
  clear();
  foreach(DisplayCell *c,other.pcells){
    pcells.append(new DisplayCell(*c));
  }
  pidcache=other.pidcache;
  pnamecache=other.pnamecache;
  pchanged=other.pchanged;
  pcalced=other.pcalced;
  return *this;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCells::copyFrom(const DisplayCells *other)
{
  for(int i=pcells.size()-1;i>=0;i--){
    DisplayCell *dc=pcells[i];
    if(dc->type()!=Xf::DCTMainBoundary&&
       dc->type()!=Xf::DCTConventional&&
       dc->type()!=Xf::DCTPrimitive){
      pidcache.remove(pcells.at(i)->id());
      pnamecache.remove(pcells.at(i)->name());
      delete pcells.takeAt(i);
    }
  }
  foreach(DisplayCell *dc,other->pcells){
    if(dc->type()!=Xf::DCTMainBoundary&&
       dc->type()!=Xf::DCTConventional&&
       dc->type()!=Xf::DCTPrimitive){
      pcells.append(new DisplayCell(*dc));
    }else{
      foreach(DisplayCell *c,pcells){
	if(dc->type()==c->type()){
	  c->boundary()->copyPolyPropertiesFrom(dc->boundary());
	}
      }
    }
  }
  pidcache=other->pidcache;
  pnamecache=other->pnamecache;

  pchanged=other->pchanged;
  pcalced=other->pcalced;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCells::setDefault(Xf::LatticeType lt)
{
  platticetype=lt;
  insert(0,new DisplayCell(Xf::DCTMainBoundary,lt),false);
  insert(1,new DisplayCell(Xf::DCTConventional,lt),false);
  insert(2,new DisplayCell(Xf::DCTPrimitive,lt),false);

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
DisplayCells::~DisplayCells()
{
  clear();
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool DisplayCells::changed()const
{
  bool c=pchanged; 
  foreach(DisplayCell *dc,pcells) c|=dc->changed();
  return c;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCells::setChanged(bool c)
{
  pchanged=c;
  foreach(DisplayCell *dc,pcells) dc->setChanged(c);
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool DisplayCells::calculated()
{
  bool c=pcalced;
  foreach(DisplayCell *dc,pcells) c&=dc->calculated();
  return c;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCells::setCalculated(bool s)
{
  pcalced=s;
  foreach(DisplayCell *dc,pcells) dc->setCalculated(s);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCells::insert(int at,DisplayCell *c,bool createuniquename)
{
  if(createuniquename) c->setName(pnamecache.makeUniqueName("Cell "));
  else pnamecache.insert(c->name());
  // if(c->type()!=Xf::DCTMainBoundary&&
  //    c->type()!=Xf::DCTConventional&&
  //    c->type()!=Xf::DCTPrimitive){
  c->setId(pidcache.makeUniqueName("Display cells cell "));
  //  }
  pcells.insert(at,c);
  pchanged=true;
  pcalced=false;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCells::remove(DisplayCell *dc)
{
  int index=pcells.indexOf(dc);
  if(index>=0&&index<pcells.size()){
    pidcache.remove(pcells.at(index)->id());
    pnamecache.remove(pcells.at(index)->name());
    delete pcells.takeAt(index);
  }
  pchanged=true;
  pcalced=false;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCells::clear()
{
  pidcache.clear();
  pnamecache.clear();
  qDeleteAll(pcells.begin(), pcells.end());
  pcells.clear();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String DisplayCells::calculate(Xf::LatticeType latticetype
			       ,const Symmetry *sym
			       ,double tol,const Vlevel &level)
{
  NSF("DisplayCells::calculate");
  //myDebug()<<nsubfunc<<calculated();
  if(calculated())return"";
  setCalculated(false); //propagate the changes trough all properties
  String ret="";
  double scale;
  foreach(DisplayCell *dc,pcells){
    if(dc){
      //myDebug()<<"dc"<<dc<<dc->changed()<<dc->calculated();
      if(dc->type()==Xf::DCTMainBoundary){
	ret=dc->calculate(latticetype,0,sym,tol,level);
	scale=dc->boundary()->radius();
      }else{
	dc->calculate(latticetype,scale,sym,tol,level);
      }
      //myDebug()<<"dc"<<dc<<dc->changed()<<dc->calculated();
    }
  }
  pchanged=false;
  pcalced=true;
  return ret;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
DisplayCell *DisplayCells::findByType(Xf::DisplayCellType type)
{
  DisplayCell *dc;
  foreach(dc,pcells){
    if(dc && dc->type()==type)
      return dc;
      
  }
  return 0;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCells::PTSectionToFile(FILE *FH)const
{

  fprintf(FH,"struct{\n");
  DisplayCell::PTHeaderToFile(FH,"       ","[*];}displayCells={{\n");
  for(int i=0;i<pcells.size();i++){
    DisplayCell *dc=pcells.at(i);
    dc->PTBodyToFile(FH,"       ",(i<pcells.size()-1?",\n":""));
  }
  fprintf(FH,"}};\n\n\n");
  
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCells::readPTSectionFromFile(const Dobj &root)
{
  clear();
  Dobj ddcsa=root("displayCell");
  for(int idc=0;idc<ddcsa.size();idc++){
    DisplayCell *dc=new DisplayCell();
    dc->readPTSectionFromFile(ddcsa[idc]);
    dc->boundary()->setLattice(platticetype);
    insert(idc,dc,false);
  }
  setChanged(false);
  setCalculated(false);
}






/*EOF*/ 
