/*************************************************************************
 * @(#)Light.cpp
 * author: magru@unas
 * date:   12 Jul 2013
 *************************************************************************
 */

#include "Light.hh"
#include "Parser.hh"
#include "Except.hh"
#include "MyMath.hh"
#include "Settings.hh"
#include "MyDebug.hh"

using namespace XFPLO;

namespace XFPLO{

  ValueArray<Pair<Lights::LightTypes,const char *> > Lights::lighttypes=
    (ValueArray<Pair<Lights::LightTypes,const char *> >()
     <<Pair<Lights::LightTypes,const char *>(Lights::Directional,"directional")
     <<Pair<Lights::LightTypes,const char *>(Lights::Positional,"positional")
     <<Pair<Lights::LightTypes,const char *>(Lights::Spot,"spotlight")
     );

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Lights::Light::PTSectionToFile(FILE *FH,bool cont)const
{
  fprintf(FH,"      {{%d,\"%s\"},\"%s\",%s,\"%x\",\"%x\",\"%x\","
	  "{%g,%g,%g,%g},{%g,%g,%g},%g,%g,%g,%g,%g}%s\n",
	  ptype,lighttypes.find(makePair<Lights::LightTypes,const char *>(ptype,""))->v2,
	  pname(),
	  (pactive?"t":"f"),
	  pdiffuse,
	  pambient,
	  pspecular,
	  pposition.at(0),pposition.at(1),pposition.at(2),
	  pposition.at(3),
	  pspotdirection.at(0),pspotdirection.at(1),pspotdirection.at(2),
	  pspotexponent,pspotcutoff,pconstantattenuation,
	  plinearattenuation,pquadraticattenuation,
	  (cont?",":"")
	  );
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Vec3 Lights::Light::realPosition()const
{
  if(ptype==Directional){
    return Vec3(pposition.constDataRef());
  }else{
    double w=pposition.at(3);
    if(fabs(w)>1.0e-10){
      return Vec3(pposition.constDataRef())/w;
    }else{
      return Vec3(pposition.constDataRef());
    }
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String Lights::Light::PTHeader()
{
  
  return Sprintf(
		 "            struct {struct  {int type;char[*] description;} type;\n"
		 "                    char[*] name;logical active;\n"
		 "                    char[*] diffuse;char[*] ambient;\n"
		 "                    char [*] specular; real position[4];\n"
		 "                    real spotdirection[3];real spotexponent;\n"
		 "                    real spotcutoff;real constantattenuation;\n"
		 "                    real linearattenuation;real quadraticattenuation;\n"
		 "            }lights[*];"
		 );
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Lights::Light::readPTSectionFromFile(const Dobj &root)
{
  try{
    pname=root("name")();
    pactive=root("active").valueL();
    pdiffuse=root("diffuse")().convertToULong();
    pambient=root("ambient")().convertToULong();
    pspecular=root("specular")().convertToULong();
    pposition.clear();
    pposition<<root("position")[0].valueD();
    pposition<<root("position")[1].valueD();
    pposition<<root("position")[2].valueD();
    pposition<<root("position")[3].valueD();
    pspotdirection.clear();
    pspotdirection<<root("spotdirection")[0].valueD();
    pspotdirection<<root("spotdirection")[1].valueD();
    pspotdirection<<root("spotdirection")[2].valueD();
    pspotexponent=root("spotexponent").valueD();
    pspotcutoff=root("spotcutoff").valueD();
    pconstantattenuation=root("constantattenuation").valueD();
    plinearattenuation=root("linearattenuation").valueD();
    pquadraticattenuation=root("quadraticattenuation").valueD();
    ptype=(LightTypes)root("type.type").valueL();
  }catch(Except e){
    myDebug()<<"oops"<<e.msg();
  };
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Lights::PTSectionToFile(FILE *FH)const
{
  fprintf(FH,"section Lights{\n\n");

  fprintf(FH,
	  "    struct {char[*] background;char[*] globalambient;\n"
	  "            real globalambientfactor;\n%s\n"
	  "    }lights\n"
	  "    ={%s,%s,%g,{\n",
	  Light::PTHeader()(),
	  Sprintf("\"%x\"",pbackgroundcolor)(),
	  Sprintf("\"%x\"",pglobalambient)(),
	  pglobalambientfactor
	  );
  long N=plights.size();
  for(long i=-1;++i<N;){
    plights.at(i)->PTSectionToFile(FH,(i<N-1));
  }
  fprintf(FH,"    }};\n");

  /* end section Light  */  
  fprintf(FH,"};\n\n\n");

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Lights::readPTSectionFromFile(const Dobj &root)
{
  try{
    Dobj o=root("lights");
    pbackgroundcolor=o("background")().convertToULong();
    pglobalambient=o("globalambient")().convertToULong();
    pglobalambientfactor=o("globalambientfactor").valueD();

    Dobj da=o("lights");
    long N=da.size();
    plights.resize(N,true);
    for(long i=-1;++i<N;){
      plights[i]->readPTSectionFromFile(da[i]);
    }

  }catch(Except e){
    //myDebug()<<"oops"<<e.msg();
  };
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Lights::insertLight(long i,Light *li)
{
  if(i<0) return ;
  if(li==0){
    if(plights.size()==0||i==0){
      plights.insert(MyMath::min<long>(i,plights.size()),new Light()); 
    }else{
      plights.insert(MyMath::min<long>(i,plights.size()),plights.at(i-1)->clone());
    }
  }else{
    plights.insert(MyMath::min<long>(i,plights.size()),li);
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Lights::Light *Lights::removeLightAt(long pos)
{
  if(pos<0 || pos>= plights.size()) return new Light();
  return plights.takeAt(pos);

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void LightThemes::loadSettings()
{
  //myDebug()<<"loadSettings";
  pthemes.clear();
  Settings set(Settings::lightSettingsFile,"Lights Themes");
  long ntheme=set.beginReadArray("Theme");
  for(int ic=0;ic<ntheme;ic++){
    set.setArrayIndex(ic);
    pthemes<<makePair(String(set.getString("Name")),Lights());
    Lights &lis=pthemes.last().v2;
    long nli=set.beginReadArray("lights");
    lis.clear();
    lis.setGlobalAmbient(set.getQRgb("globalambient"));
    lis.setGlobalAmbientFactor(set.getDouble("globalambientfactor"));
    lis.setBackgroundColor(set.getQRgb("background"));
    
    //myDebug()<<pthemes;

    for(int i=0;i<nli;i++){
      set.setArrayIndex(i);
      Lights::Light *li=new Lights::Light();
      lis.insertLight(i,li);
      li->setType((Lights::LightTypes)set.getInt("type"));
      li->setName(set.getString("name"));
      li->setActive(set.getBool("active"));
      li->setDiffuse(set.getQRgb("diffuse"));
      li->setAmbient(set.getQRgb("ambient"));
      li->setSpecular(set.getQRgb("specular"));
      li->setPosition(set.getValueArray<double>("position"));
      //myDebug()<<"li->position"<<li->position();
      li->setSpotDirection(set.getValueArray<double>("spotdirection"));
      li->setSpotExponent(set.getDouble("spotexponent"));
      li->setSpotCutoff(set.getDouble("spotcutoff"));
      li->setConstantAttenuation(set.getDouble("constantattenuation"));
      li->setLinearAttenuation(set.getDouble("linearattenuation"));
      li->setQuadraticAttenuation(set.getDouble("quadraticattenuation"));
    }
    set.endArray();
    
  }
  set.endArray();

  //myDebug()<<"end loadSettings"<<pthemes;

  pcurrenttheme=set.getInt("currentTheme");
  if(pcurrenttheme<0 || pcurrenttheme>=pthemes.size())pcurrenttheme=0;
  
  if(pthemes.size()<1){
    pthemes<<makePair(String("default"),Lights());
    pcurrenttheme=0;
  }

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void LightThemes::saveSettings()
{
  Settings set(Settings::lightSettingsFile,"Lights Themes");
  set.beginWriteArray("Theme");
  for(int ic=0;ic<pthemes.size();ic++){
    set.setArrayIndex(ic);
    set.set("Name",pthemes.at(ic).v1);
    const Lights &lis=pthemes.at(ic).v2;
    set.beginWriteArray("lights");
    set.set("globalambient",lis.globalAmbient());
    set.set("globalambientfactor",lis.globalAmbientFactor());
    set.set("background",lis.backgroundColor());
    for(int i=-1;++i<lis.size();){
      const Lights::Light *li=lis.at(i);
      set.setArrayIndex(i);
      set.set("type",li->type());
      set.set("name",li->name());
      set.set("active",li->active());
      set.set("diffuse",li->diffuse());
      set.set("ambient",li->ambient());
      set.set("specular",li->specular());
      set.set("position",li->position());
      set.set("spotdirection",li->spotDirection());
      set.set("spotexponent",li->spotExponent());
      set.set("spotcutoff",li->spotCutoff());
      set.set("constantattenuation",li->constantAttenuation());
      set.set("linearattenuation",li->linearAttenuation());
      set.set("quadraticattenuation",li->quadraticAttenuation());
    }
    set.endArray();
  }
  set.endArray();

  set.set("currentTheme",(int)pcurrenttheme);



}


/*EOF*/ 
