/*************************************************************************
 * @(#)Position.cpp
 * author: magru@unas
 * date:   20 Jun 2013
 *************************************************************************
 */

#include "Array.hh"
#include "Array_cpp.hh"
#include "Position.hh"
#include "BoundingBox.hh"
#include "DataStatistics.hh"
#include "MyDebug.hh"

using namespace XFPLO;

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Positions::Positions()
 :Module("Positions"),pusecounter(1),pdata(0),pbbox(new BoundingBox())
{
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Positions::~Positions()
{
  clearData();
  delete pbbox;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Positions::setGridD(long vectordimension,
			 const ValueArray<long> &gridsize, // Nx x Ny ( x Nz)
			 const ValueArray<double> &origin, // one origin for all
			 const ValueArray<ValueArray<double> > &deltas
			 )
{
  //printf("%s::setGridD\n",id()());
  clearData();
  pdata=ProductArray<double>::construct();
  ValueArray<double> o0(vectordimension,0.);
  for(int i=-1;++i<gridsize.size();){
    ((ProductArray<double>*)pdata)
      ->addArray(RegularArray<double>::construct(vectordimension,
						 (i==0?origin:o0),
						 deltas.at(i),gridsize.at(i)));
    
  }
  
  pisdirty=true;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool Positions::cascadingUpdate(const String &prefix)
{
  if(Module::debugmode) printf("%s%s::cascadingUpdate\n",prefix(),id()());
  Module::cascadingUpdate(prefix+"  ");
  showDirt(prefix,Sprintf("%s must react to",id()()));
  bool ret=false;
  if(pisdirty){
    calculateBoundingBox();
    sendDirt("changed");
    ret=true;
  }
  resetDirt();
  if(Module::debugmode) printf("%sret=%d\n",prefix(),ret);
  return ret;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Positions::calculateBoundingBox()
{
  Vec3 ori;
  Vec3 v[3];

  ProductArray<double> *pa=0;
  BaseArray<double> *ba=0;

  if((pa=dynamic_cast<ProductArray<double>*>(pdata))!=0){
    long D=pa->dimensions();
    long d=embeddingDimension();

    for(long i=-1;++i<D;){
      const RegularArray<double> *ra;
      if((ra=dynamic_cast<const RegularArray<double>* >(pa->array(i)))==0)
	throw Except("do not know how to calculateBoundingBox with non-RegularArrays");
      
      for(int c=-1;++c<d;){
	ori[c]+=ra->origin().at(c);
	v[i][c]=ra->delta().at(c)*(ra->size()-1);
      }
      
    }
    if(D==1){
      double radius=norm(v[0])*0.01;
      v[0].findPerpDirections(v[1],v[2]);
      v[1]*=radius;
      v[2]*=radius;
      ori-=(v[2]+v[1])*0.5;
    }else if(D==2){
      double radius=(norm(v[0])+norm(v[1]))*0.005;
      v[2]=v[0].cross(v[1]).normalize()*radius;
      ori-=v[2]*0.5;
    }
    
    if(D>3){
      *pbbox=BoundingBox();
    }else{
      pbbox->setOrigin(ori);
      pbbox->setCell(Matrix3x3(v[0],v[1],v[2]));
    }
    
    
  }else  if((ba=dynamic_cast<BaseArray<double>*>(pdata))!=0){
    long d=embeddingDimension();

    DataStatistics dstat;
    dstat.calculate<double>(ba);
    const ValueArray<double> &dmin=dstat.data<double>()->minValues;
    const ValueArray<double> &dmax=dstat.data<double>()->maxValues;
    // myDebug()<< "Positions are enclosed between"
    // 	     <<dmin<<"and"<<dmax;
    d=dmin.size();
    if(d==1){
      ori[0]=dmin.at(0);
      v[0]=Vec3(dmax.at(0)-ori[0],0,0);
      v[1]=0;
      v[2]=0;
    }else if(d==2){
      ori[0]=dmin.at(0);
      ori[1]=dmin.at(1);
      v[0]=Vec3(dmax.at(0)-ori[0],0,0);
      v[1]=Vec3(0,dmax.at(1)-ori[1],0);
      v[2]=0;
    }else if(d==3){
      ori[0]=dmin.at(0);
      ori[1]=dmin.at(1);
      ori[2]=dmin.at(2);
      v[0]=Vec3(dmax.at(0)-ori[0],0,0);
      v[1]=Vec3(0,dmax.at(1)-ori[1],0);
      v[2]=Vec3(0,0,dmax.at(2)-ori[2]);
    }

    if(d>3||d<1){
      *pbbox=BoundingBox();
    }else{
      pbbox->setOrigin(ori);
      pbbox->setCell(Matrix3x3(v[0],v[1],v[2]));
    }
    
    
  }else{
    throw Except("do not know how to calculateBoundingBox on this object");
  }


  
	
  // myDebug()<<"ori="<<ori;
  // myDebug()<<"v0="<<v[0];
  // myDebug()<<"v1="<<v[1];
  // myDebug()<<"v2="<<v[2];
  
  
  //  myDebug()<<pbbox->toString();

}

/*EOF*/ 
