/* @(#)PropertyTable.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _PROPERTYTABLE_H
#define _PROPERTYTABLE_H 1

#include <QList> 
#include "Material.hh"
#include "FString.hh"
#include "constants.h"
#include "Pair.hh"
#include "ValueArray.hh"
#include "PtrArray.hh"
#include "MACROS.hh"
#include "Xf.hh"

namespace XFPLO{
  
  class AtomicRadii{
  public:

    enum AtomicRadiusType{ARCovalent=0,ARIonic=1,ARvdW=2,ARMetallic=3,ARCrystal=4};
    
    AtomicRadii(): pcovalent(0),pionic(0),pvdW(0),pmetallic(0),pcrystal(0),pnn(1),
		   pcurrenttype(ARCovalent){};
    AtomicRadii(double c,double i,double v,double m,double cr)
      :pcovalent(c),pionic(i),pvdW(v),pmetallic(m),pcrystal(cr),pnn(1),
       pcurrenttype(ARCovalent){;};
    
    bool operator!=(const AtomicRadii &other)const{
      return ( 
	      (pcurrenttype!=other.pcurrenttype)||
	      (pcovalent!=other.pcovalent)||
	      (pionic!=other.pionic)||
	      (pvdW!=other.pvdW)||
	      (pmetallic!=other.pmetallic)||
	      (pcrystal!=other.pcrystal)||
	      (pnn!=other.pnn) );
    }

    void setCurrent(double r){
      switch(pcurrenttype){
      case ARCovalent:
      default:
	pcovalent=r; break;
      case ARIonic:
	pionic=r; break;
      case ARvdW:
	pvdW=r; break;
      case ARMetallic:
	pmetallic=r; break;
      case ARCrystal:
	pcrystal=r; break;
      }
    }

    double current(Xf::LengthUnits lengthunit)const{
      double unit=1.;
      if(lengthunit==Xf::BohrRadius){
	unit=1./constant_aBtoAng;
      }
      switch(pcurrenttype){
      case ARCovalent:
      default:
	return pcovalent*unit;
      case ARIonic:
	return pionic*unit;
      case ARvdW:
	return pvdW*unit;
      case ARMetallic:
	return pmetallic*unit;
      case ARCrystal:
	return pcrystal*unit;
      }
    }
    
  private:
    PROPERTY(covalent,covalent,Covalent,double);
    PROPERTY(ionic,ionic,Ionic,double);
    PROPERTY(vdW,vdW,vdW,double);
    PROPERTY(metallic,metallic,Metallic,double);
    PROPERTY(crystal,crystal,Crystal,double);
    CMODPROPERTY(nn,NN,NN,double);
    PROPERTY(currenttype,currentType,CurrentType,AtomicRadiusType);

    static ValueArray<Xf::PairIntCharP > radiinames;

  };
  
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class AtomProperty{
  public:
    
    AtomProperty():pmaterial(Material(0xC8C8C8,50.)){};
    AtomProperty(const Material &m,const AtomicRadii &rad)
      :pmaterial(Material(m)),patomicradii(rad){};
    
    bool operator!=(const AtomProperty &other)const{
      return ( 
	      (pmaterial!=other.pmaterial)||
	      (patomicradii!=other.patomicradii)
	       );
    }
    
    
  private:
    CMODPROPERTY(material,material,Material,Material);
    CMODPROPERTY(atomicradii,atomicRadii,AtomicRadii,AtomicRadii);
  };
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class AtomPropertyTable
  {
  public:
    AtomPropertyTable():ptheme("default"){loadDefault();};
    AtomPropertyTable(const String &theme):ptheme(theme){loadDefault();};
    AtomPropertyTable(const AtomPropertyTable &other):ptheme(other.ptheme){
      list=other.list;
    };
    AtomPropertyTable(const String &theme,const AtomPropertyTable &other)
      :ptheme(theme){    list=other.list;};

    AtomPropertyTable *clone()const{ return new AtomPropertyTable(*this);};
    
    void loadDefault();
    
    const AtomProperty &find(const String &e) const;
    AtomProperty &find(const String &e);

    long indexOf(const String &e)const;
    
    inline long size()const{return list.size();};
    inline const AtomProperty &at(int i)const{return list.at(i).v2;};
    inline AtomProperty &operator[](int i){return list[i].v2;};
    
    const String &element(long i)const{return list.at(i).v1;}
    
    
    static const StringList allowedElements;
    
    
  private:
    ValueArray<Pair<String,AtomProperty> > list;
    AtomProperty defaultProperty;
    CPROPERTY(theme,theme,Theme,String);
  };
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class AtomPropertyTables
  {
  public:
    AtomPropertyTables();
    ~AtomPropertyTables();
    
    AtomPropertyTables(const AtomPropertyTables &other);
    AtomPropertyTables &operator=(const AtomPropertyTables &other);

    
    StringList themes()const;
    
    long size()const{return list.size();};
    const AtomPropertyTable *at(int i)const{return list.at(i);};
    AtomPropertyTable *operator[](int i){return list[i];};
    
    const AtomPropertyTable *current()const {return list.at(pcurrentindex);};
    AtomPropertyTable *current(){return list[pcurrentindex];};
    
    const AtomPropertyTable *find(const String &theme){
      for(int i=0;i<list.size();i++){
	if(list.at(i)->theme()==theme){
	  return list.at(i);
	}
      }
      return list.at(0);
    };
    
    
    void removeAt(int i){
      if(i>=0 && i<list.size())
	list.removeAt(i);
    };
    void insertNewAt(int i,const String &theme,const AtomPropertyTable *copyfrom){
      if(i>=list.size()){
	if(copyfrom)
	  list.append(new AtomPropertyTable(theme,*copyfrom));
	else
	  list.append(new AtomPropertyTable(theme));
      }else{
	if(copyfrom)
	  list.insert(i,new AtomPropertyTable(theme,*copyfrom));
	else
	  list.insert(i,new AtomPropertyTable(theme));
      }
    };
    
    void saveSettings();
    void loadSettings();
    
  private:
    PtrArray<AtomPropertyTable> list;
    PROPERTY(currentindex,currentIndex,CurrentIndex,long);
  };
  
  
  extern AtomPropertyTables *globalproperties;
  
}
#ifdef	__cplusplus
extern "C" {
#endif
  
#ifdef	__cplusplus
}
#endif


#endif /* _PROPERTYTABLE_H */

