/* @(#)RegularMesh.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _REGULARMESH_H
#define _REGULARMESH_H 1

#include "MeshData.hh"
#include "Vec3.hh"

class TriangleMesh;
class Vlevel;
class MeshSymmetry;
class Boundary;
class Plane;


class RegularMesh
{
 public:
  RegularMesh(const QString &id);
  ~RegularMesh();

  const QString &id()const{return pid;};
  
  void setCell(const Matrix3x3 &m);
  void setCell(const Vec3 &c0,const Vec3 &c1,const Vec3 &c2);

  void setPeriodic(bool p);

  void setOrigin(const Vec3 &o);
  void setSubdivisions(long Nx,long Ny,long Nz);

  void setXSubdivision(long Nz);
  void setYSubdivision(long Nz);
  void setZSubdivision(long Nz);
  
  String setSymmetricMeshSubdivision(const MeshSymmetry &sym
				      ,long nxyz,const Vlevel &vlevel);

  void calcPoints(const MeshSymmetry &sym,bool useRotSymmetry);
  void calcMicroCells();

  void printPoints();
  void printData();
  QString reportKPoints(const Vlevel &vlevel);


  void meshDataMinMax(long idata,double &min,double &max)const;
  inline QList<MeshData*> &meshData(){return pmeshdata;};
  void appendMeshData(long n);
  void clearMeshData();
  inline void setMeshData(long id,long ipi,double d){
    pmeshdata[id]->parray[ipi]=d;
  };
  inline void setMeshExtraData(long id,long index,long iw,double d){
    pmeshdata[id]->pextradata[iw][index]=d;
  };

  void calculateGradient();
  void createIsoSurface(TriangleMesh *triamesh,const Boundary *boundary
			,const QList<Plane> *clippingplanes
			,double isolevel,long idata,bool clip);
  
  inline bool meshIsReady()const{return ppointecalced;};
  inline bool microCellsAreReady()const{return pmicrocellscalced;};

  inline const Vec3 &origin()const{return porigin;};
  inline const Matrix3x3 &cell()const{return pcell;};


  inline long numPoints()const {return pnpoints;};
  inline long numIrredPoints()const {return pnpointsirred;};
  inline long numIrredRotPoints()const {return pnpointsirredrot;};
  inline const Vec3 &point(long i)const { return ppoints[i];};
  inline const Vec3 &irredPoint(long ipi)const { return ppoints[ieqi[ipi]];};
  inline const Vec3 &irredRotPoint(long ipi)const { return ppoints[ieqirot[ipi]];};
  inline long RequivI_Rot(long ipi)const{return ieqirot[ipi];}; 
  inline long IequivR(long ip)const{return ieqir[ip];}; 

  inline long NMicroCells()const { return pnmicrocells;};
  inline const long *microCell(long ic)const { return pmicrocells+8*ic;};

  inline long NTetrahedra()const { return pntetras;};
  inline const long *tetrahedron(long it)const { return ptetras+4*it;};

  // we read in rotational irred points from FPLO
  // and have to fill the translation irreds fro smooth gradients
  void fillIrredFromIrredRot();

 private:

  
 private:
  Vec3 porigin;    // the origin
  Matrix3x3 pcell; // which are the directions
  Matrix3x3 prcell; // which are the reciprocial directions
  long pNx,pNy,pNz; // How many sibdivisions
  bool pperiodic;  // is the mesh  periodically continued
  Matrix3x3 pprimcell; //  primitive cell 
  Matrix3x3 prprimcell; //  reciprocial of ppcell
  bool phexsubdivision;
  bool pirred;
                       
  //dependend data
  bool ppointecalced;
  long pnpoints;
  long pnpointsirred;
  long pnpointsirredrot;
  Vec3 *ppoints; // the mesh points
  bool pmicrocellscalced;
  long pnmicrocells;
  long *pmicrocells;    // the microcells of the regular mesh
  long pntetras;
  long *ptetras;        // the tetrahedron subdivision of the microcells
  long *ieq,*ieqrot;    // each point ip in the full mesh 
  //                       has a point ieq[ip] in the full mesh, which
  //                       it is equivalent to
  //                       either with respect to translation or trns+rot
  long *ieqi,*ieqirot;  // the irreducible meshpoints ipi is ip=ieqi[ipi]
  //                       in the full mesh
  long *ieqir,*ieqirrot;// the point ip of the full mesh is the irreducible
  //                       point ieqir[ip]

  QList<MeshData*> pmeshdata;
  QString pid;
};

#ifdef	__cplusplus
extern "C" {
#endif

#ifdef	__cplusplus
}
#endif


#endif /* _REGULARMESH_H */

