/*************************************************************************
 * @(#)Symmetry.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   04 Jul 2009
 *************************************************************************
 */
// #include <math.h>
// #include <stdlib.h>

#include "merror.h" 
#include "cbuffer.h"
#include "constants.h"
#include "fsymmetry.h"
#include "generatetau.h"
#include "Except.hh"
#include "Assorted.hh"
#include "MyMath.hh"
#include "PropertyTable.hh"
#include "Boundary.hh"
#include "MachineTime.hh"
#include "Symmetry.hh"
#include "IntegerTricks.hh"
#include "Parser.hh"
//#include "SpaceGroupOperation.hh"
#include "MyDebug.hh"
#include "qt5compat.hh"



const char *Symmetry::pstructuretypes[]={
  "Unknown unit",
  "Crystal",
  "Molecule"
};


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void WyckPos::print()const
{
  NSF("WyckPos::print");
  myDebug() << nsubfunc;
  myDebug()<<"this="<<(void*)this;
  myDebug() << "el=" << pel;
  myDebug() << "type=" << ptype;
  myDebug() << "tau="<<ptauD.printS()<<"\n";

}




/* ***********************************************************************
 *
 * ***********************************************************************
 */
String Site::printS()const
{
  NSF("Site::print");
  String s;
  s=Sprintf("%s(W%ld,T%ld): %s",pel(),psort+1,ptype,ptauD.printS()());

  return s;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Symmetry::Symmetry()
  :pspacegroup(221),plengthunit(Xf::BohrRadius),pstructuretype(Xf::STCrystal),
   //   pspacegroupoperations(new SpaceGroupOperations()),
   sym(0),pchanged(true),pcalced(false)
{
  palatD=5.;
  palatS[0]=palatS[1]=palatS[2]="5.";
  panglesD=90.;
  panglesS[0]=panglesS[1]=panglesS[2]="90.";

  sym=new tssymmetry;
  sym->nugen=0;
  
  
  pwyckpos.append(WyckPos("Al","0","0","0"
			  ,globalproperties->current()->find("Al"),true));


}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
Symmetry::Symmetry(const Symmetry &other)
  :pspacegroup(221),plengthunit(Xf::BohrRadius),pstructuretype(Xf::STCrystal),
   //   pspacegroupoperations(new SpaceGroupOperations()),
   sym(0),
   pchanged(true),pcalced(false)
{
  palatD=5.;
  palatS[0]=palatS[1]=palatS[2]="5.";
  panglesD=90.;
  panglesS[0]=panglesS[1]=panglesS[2]="90.";

  sym=new tssymmetry;
  sym->nugen=0;
  
  
  pwyckpos.append(WyckPos("Al","0","0","0",globalproperties
			  ->current()->find("Al"),true));

  setSpacegroup(other.pspacegroup);
  setLengthunit(other.plengthunit);
  setLatticeConstants(other.palatS);
  setAxesAngles(other.panglesS);

  setStructureType(other.pstructuretype);

  setSubGroupGenerators(other.psubgroupgenerators);

  setWyckoffPositions(other.pwyckpos);

  //  *pspacegroupoperations=*(other.pspacegroupoperations);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
const Symmetry &Symmetry::operator=(const Symmetry &other)
{
  setSpacegroup(other.pspacegroup);
  setLengthunit(other.plengthunit);
  setLatticeConstants(other.palatS);
  setAxesAngles(other.panglesS);

  setStructureType(other.pstructuretype);

  setSubGroupGenerators(other.psubgroupgenerators);

  setWyckoffPositions(other.pwyckpos);

  //  *(pspacegroupoperations)=*(other.pspacegroupoperations);

//   if(sym && other.sym){
//     myDebug()<<"operator=() copySymmetry ";
//     copySymmetry(other.sym,sym);
//   }

  return *this;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
Symmetry::~Symmetry()
{
  //  NSF("Symmetry::~Symmetry");
  if(sym) delete sym; sym=0;
  //  if(pspacegroupoperations)delete pspacegroupoperations;pspacegroupoperations=0;
  ppairs.clear();
  pwyckpos.clear();
  psites.clear();

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::setLengthunit(Xf::LengthUnits s)
{
  pchanged|=(plengthunit!=s);
  if(pchanged) pcalced =false;
  plengthunit=s;
}

void Symmetry::setStructureType(Xf::StructureType t)
{
  pchanged|=(pstructuretype!=t);
  if(pchanged) pcalced =false;
  pstructuretype=t;
}

void Symmetry::setSpacegroup(long s)
{
  pchanged|=(pspacegroup != s);
  if(pchanged) pcalced =false;
  pspacegroup=s;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::setLatticeConstants(const Vec3 &a)
{
  //NSF("Symmetry::setLatticeConstants");

  pchanged|=palatD.neq(a,1.0e-10);

  if(pchanged) pcalced =false;

  palatD=a;
  palatS[0]=Sprintf("%30.15f",a[0]).stripBlanks().chop("0");
  palatS[1]=Sprintf("%30.15f",a[1]).stripBlanks().chop("0");
  palatS[2]=Sprintf("%30.15f",a[2]).stripBlanks().chop("0");

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::setLatticeConstants(const String a[3])
{
  //NSF("Symmetry::setLatticeConstants");
  //myDebug() << nsubfunc ;


  Vec3 pa(a);

  //  myDebug() << "pa=" << pa[0]  << pa[1] << pa[2] ;

  pchanged|=palatD.neq(pa,1.0e-10);

  if(pchanged) pcalced =false;

  palatD=pa;
  palatS[0]=a[0];
  palatS[1]=a[1];
  palatS[2]=a[2];

  //  myDebug() << "palatS=" << palatS[0]()  << palatS[1]() << palatS[2]() ;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::setLatticeConstants(const StringList &sl)
{
  //NSF("Symmetry::setLatticeConstants");
  //myDebug() << nsubfunc ;


  Vec3 pa(sl);

  //  myDebug() << "pa=" << pa[0]  << pa[1] << pa[2] ;

  pchanged|=palatD.neq(pa,1.0e-10);

  if(pchanged) pcalced =false;

  palatD=pa;
  palatS[0]=sl[0];
  palatS[1]=sl[1];
  palatS[2]=sl[2];

  //  myDebug() << "palatS=" << palatS[0]()  << palatS[1]() << palatS[2]() ;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::setAxesAngles(const Vec3 &a)
{

  pchanged|=panglesD.neq(a);
  if(pchanged) pcalced =false;

  panglesD=a;
  
  panglesS[0]=Sprintf("%30.15f",a[0]).stripBlanks().chop("0");
  panglesS[1]=Sprintf("%30.15f",a[1]).stripBlanks().chop("0");
  panglesS[2]=Sprintf("%30.15f",a[2]).stripBlanks().chop("0");
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::setAxesAngles(const String a[3])
{
  Vec3 pa(a);

  pchanged|=panglesD.neq(pa);
  if(pchanged) pcalced =false;

  panglesD=pa;
  
  panglesS[0]=a[0];
  panglesS[1]=a[1];
  panglesS[2]=a[2];
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::setAxesAngles(const StringList sl)
{
  Vec3 pa(sl);

  pchanged|=panglesD.neq(pa);
  if(pchanged) pcalced =false;

  panglesD=pa;
  
  panglesS[0]=sl[0];
  panglesS[1]=sl[1];
  panglesS[2]=sl[2];
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::setSubGroupGenerators(ValueArray<int> l)
{
  pchanged |= (psubgroupgenerators.size() != l.size());

  for(int i=0;i<MyMath::min<long>(l.size(),psubgroupgenerators.size());i++){
    pchanged |= (psubgroupgenerators[i] != l[i]);
  }
  if(pchanged) pcalced =false;

  psubgroupgenerators.clear();  
  for(int i=0;i<l.size();i++){
    psubgroupgenerators.append(l[i]);
  }
  
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::setWyckoffPositions(const ValueArray<WyckPos> &l)
{
  pchanged |= (pwyckpos.size() != l.size());

  for(int i=0;i<MyMath::min<long>(l.size(),pwyckpos.size());i++){
    const WyckPos &w=l.at(i);
    pchanged |= (pwyckpos[i] != w);
  }
  if(pchanged) pcalced =false;

  pwyckpos.clear();  
  for(int i=0;i<l.size();i++){
    const WyckPos &w=l.at(i);
    // we do that to get the real tauD set from tauS
    pwyckpos.append(WyckPos(w.el(),w.type(),w.tauS(),w.property(),w.visible()));
 }
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::setWyckoffPosition(long i,const WyckPos &pos)
{
  if(i<0 || i>=pwyckpos.size()) return;

  pchanged |= (pwyckpos[i] != pos);
  
  if(pchanged) pcalced =false;

  pwyckpos.replace(i,pos);
  
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
WyckPos Symmetry::removeWyckoffPosition(long pos)
{
  if(pos<0 || pos>= pwyckpos.size()) 
    return WyckPos("Al","0","0","0",globalproperties
		   ->current()->find("Al"),true);
  pchanged=true;
  if(pchanged) pcalced =false;
  return pwyckpos.takeAt(pos);

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::insertWyckoffPosition(long i,const WyckPos *wpos)
{
  if(i<0) return ;

  if(wpos==0){
    if(pwyckpos.size()==0||i==0){
      pwyckpos.insert(MyMath::min<long>(i,pwyckpos.size())
		       ,WyckPos("Al","0","0","0"
				 ,globalproperties
				->current()->find("Al"),true)); 
    }else{
      pwyckpos.insert(MyMath::min<long>(i,pwyckpos.size()),pwyckpos.at(i-1)); 
    }
  }else
    pwyckpos.insert(MyMath::min<long>(i,pwyckpos.size()),*wpos);

  pchanged=true;
  if(pchanged) pcalced =false;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Matrix3x3 Symmetry::cell() const { return Matrix3x3(sym->cell); }

Matrix3x3 Symmetry::ccell() const { return Matrix3x3(sym->ccell); }

Matrix3x3 Symmetry::rcell() const { return Matrix3x3(sym->rcell); }

Matrix3x3 Symmetry::crcell() const { return Matrix3x3(sym->crcell); }

double *Symmetry::opmat() const { return sym->opmat; }

double *Symmetry::optau() const { return sym->v; }

long Symmetry::numOps() const { return sym->ng; }


long Symmetry::iTable(long ig) const {return sym->itable[ig];}

long Symmetry::bravais() const { return sym->bravais; }

long Symmetry::crystal() const { return sym->crystal; }

double Symmetry::vuc() const { return sym->vuc; }


/* ***********************************************************************
 *
 * ***********************************************************************
 */

void Symmetry::symPTSectionToFile(FILE *FH,bool full)
{
  //  NSF("Symmetry::symPTSectionToFile");
  //  myDebug() << nsubfunc;
  fprintf(FH,"section symmetry{\n\n");

  fprintf(FH,"    struct {int number;char[*] symbol;}spacegroup\n");
  fprintf(FH,"    ={%ld,\"%s\"};\n\n",pspacegroup
	  ,spcgrp_name(getSpacegroupDescriptor("",&pspacegroup)));


  fprintf(FH,"    struct {int type;char[*] description;} lengthunit\n");
  fprintf(FH,"    ={%d,\"%s\"};\n\n",plengthunit,Xf::lengthUnitName(plengthunit));


  fprintf(FH,"    real lattice_constants[3]\n");
  fprintf(FH,"    ={%s,%s,%s};\n\n" ,palatS[0](),palatS[1](),palatS[2]());
  fprintf(FH,"    real axis_angles[3]\n");
  fprintf(FH,"    ={%s,%s,%s};\n\n",panglesS[0](),panglesS[1](),panglesS[2]());
  
  if(full){
    fprintf(FH,"    struct {int type;char[*] description;} structure_type\n");
    fprintf(FH,"    ={%d,\"%s\"};\n\n",pstructuretype
	    ,structureTypes(pstructuretype) );


    fprintf(FH,"    int subgroupgenerators[*]\n"
	    "    ={");
    for(int i=0;i<psubgroupgenerators.size();i++){
      fprintf(FH,"%d",psubgroupgenerators[i]);
      if(i<psubgroupgenerators.size()-1) fprintf(FH,",");
    }
    fprintf(FH,"};\n");


    //todo nsort wyckoffpos
    fprintf(FH,"    struct {char[2] element;int type;real tau[3];\n"
	    "          struct {char[*] atomcolor;char[*] specularcolor;\n"
	    "                  real shininess;real rcovalent;real rionic;\n"
	    "                  real rvdw;real rmetallic;real rcrystal;\n"
	    "                  int rcurrenttype;}properties;\n"
	    "          logical visible;"
	    "       } "
	    "wyckoff_positions[*]\n");
    fprintf(FH,"    ={\n");
    for(int i=0;i<pwyckpos.size();i++){
      fprintf(FH,"     {\"%s\",%ld,{%s,%s,%s},{%s,%s,%g,%g,%g,%g,%g,%g,%d},%c}"
	      ,pwyckpos.at(i).el()()
	      ,pwyckpos.at(i).type()
	      ,pwyckpos.at(i).tauS(0)()
	      ,pwyckpos.at(i).tauS(1)()
	      ,pwyckpos.at(i).tauS(2)() 
	      ,Sprintf("\"%x\"",pwyckpos.at(i).property().material().color())()
	      ,Sprintf("\"%x\"",pwyckpos.at(i).property().material().specularColor())()
	      ,pwyckpos.at(i).property().material().shininess()
	      ,pwyckpos.at(i).property().atomicRadii().covalent()
	      ,pwyckpos.at(i).property().atomicRadii().ionic()
	      ,pwyckpos.at(i).property().atomicRadii().vdW()
	      ,pwyckpos.at(i).property().atomicRadii().metallic()
	      ,pwyckpos.at(i).property().atomicRadii().crystal()
	      ,pwyckpos.at(i).property().atomicRadii().currentType()
	      ,(pwyckpos.at(i).visible()?'t':'f')
	      );
      if(i<pwyckpos.size()-1) fprintf(FH,",");
      fprintf(FH,"\n");
    }
    fprintf(FH,"    };\n\n");



  }

  /* end section symmetry  */  
  fprintf(FH,"};\n\n\n");
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::readSymmetryPTSectionFromFile(Parser *par,Symmetry &tmp
					     ,bool full)
{

  long num=(*par)()("spacegroup.number").valueL();
  tmp.setSpacegroup(num);

  try{
    Xf::LengthUnits un=(Xf::LengthUnits)(*par)()("lengthunit.type").valueL();
    tmp.setLengthunit(un);
  }catch(Except e){};

  
  String t_alat[3];
  t_alat[0]=t_alat[1]=t_alat[2]="5.";
  (*par)()("lattice_constants").getArray(3,t_alat);
  tmp.setLatticeConstants(t_alat);

  String t_angle[3];    
  t_angle[0]=t_angle[1]=t_angle[2]="90.";
  (*par)()("axis_angles").getArray(3,t_angle);
  tmp.setAxesAngles(t_angle);

  if(!full) return;

  try{
    Xf::StructureType st=
      (Xf::StructureType)(*par)()("structure_type.type").valueL();
    tmp.setStructureType(st);
  }catch(Except e){};
  try{
    Dobj d=(*par)()("wyckoff_positions");
    int nsort=d.size();
    tmp.pwyckpos.clear();
    for(int i=0;i<nsort;i++){
      String el=d[i]("element")();
      String tauS[3];
      d[i]("tau").getArray(3,tauS);
      long ty=d[i]("type").valueL();
      bool vis=true;
      try{
	vis=d[i]("visible").valueL();
      }catch(Except e){}

      AtomProperty pr=globalproperties->current()->find(el);
      try{
	QRgb ac=d[i]("properties.atomcolor")().convertToULong();
	pr.material().setColor(ac);
      }catch(Except e){}
      
      try{
	QRgb ac=d[i]("properties.specularcolor")().convertToULong();
	pr.material().setSpecularColor(ac);
	double dd=d[i]("properties.shininess").valueD();
	pr.material().setShininess(dd);
	dd=d[i]("properties.rcovalent").valueD();
	pr.atomicRadii().setCovalent(dd);
	dd=d[i]("properties.rionic").valueD();
	pr.atomicRadii().setIonic(dd);
	dd=d[i]("properties.rvdw").valueD();
	pr.atomicRadii().setvdW(dd);
	dd=d[i]("properties.rmetallic").valueD();
	pr.atomicRadii().setMetallic(dd);
	dd=d[i]("properties.rcrystal").valueD();
	pr.atomicRadii().setCrystal(dd);
	long ll=d[i]("properties.rcurrenttype").valueL();
	pr.atomicRadii().setCurrentType((AtomicRadii::AtomicRadiusType)ll);
      }catch(Except e){
	//myDebug()<<"oops"<<e.msg();
      }


      tmp.pwyckpos.append(WyckPos(el,ty,tauS,pr,vis));
    }
  }catch(Except e){};


  tmp.psubgroupgenerators.clear();
  try{
    Dobj d=(*par)()("subgroupgenerators");
    for(int i=0;i<d.size();i++){
      tmp.psubgroupgenerators.append(d[i].valueL());
    }
  }catch(Except e){};


  tmp.pchanged=true;
  tmp.pcalced=false;

}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
const char *Symmetry::spacegroupName(long spacegroup){
  return spcgrp_name(getSpacegroupDescriptor("",&spacegroup));
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */

void Symmetry::readSymInfo(const String &fname)
{
  
  String t_alat[3];
  String t_angle[3];

  Parser symparser(1);
  symparser.parseFile(fname);

  String vm=symparser()("version")("mainversion")();
  String vs=symparser()("version")("subversion")();

  long t_num=symparser()("spacegroup")("number")().convertToLong();
  Xf::LengthUnits t_unit=(Xf::LengthUnits)
    (symparser()("lengthunit")("type")().convertToLong());



  Xf::StructureType t_strtype=Xf::STCrystal;
  try{
    t_strtype=(Xf::StructureType)
      (symparser()("structure_type")("type")().convertToLong());
  }catch(Except){}

  
  symparser()("lattice_constants").getArray(3,t_alat);
  symparser()("axis_angles").getArray(3,t_angle);

  //  myDebug() << t_alat[0]()  << t_alat[1]() << t_alat[2]() ;
  //  myDebug() << t_angle[0]()  << t_angle[1]() << t_angle[2]() ;
  
  ValueArray<WyckPos> tmp;

  Dobj d=symparser()("wyckoff_positions");
  long nsort=d.size();
  for(int i=0;i<nsort;i++){
    Dobj ds=d[i];
    String el=ds("element")();
    tmp.append(WyckPos(el()
			,ds("tau")[0]()(), ds("tau")[1]()()
			 ,ds("tau")[2]()(),globalproperties
		       ->current()->find(el()),true ));
  }

  ValueArray<int> l;
  try{
    d=symparser()("subgroupgenerators");
    for(int i=0;i<d.size();i++){
      l.append(d[i].valueL());
    }
  }catch(Except e){};


  /* defered final assignement in order to allow exceptions */

  setSpacegroup(t_num);
  setLengthunit(t_unit);
  setLatticeConstants(t_alat);
  setAxesAngles(t_angle);

  setStructureType(t_strtype);
  //  myDebug() << "t_strtype=" <<t_strtype;

  
  setWyckoffPositions(tmp);

  
  
  setSubGroupGenerators(l);

  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::calcSymFromGroupInput(String &out,bool calcAtoms,bool giveinfo,
				     bool calcpairinfo)
{
  NSF("Symmetry::calcSymFromGroupInput");
  //  myDebug() << "START" << nsubfunc ;
  /* catch errors into buffer */
  seterrorhandling(ERR_NOCATCH | ERR_BUFFERED |ERR_RESETBUFFER);

  MachineTime time;

  /* calc symmetry info */

  sym->nugen=psubgroupgenerators.size();
  for(int i=0;i<psubgroupgenerators.size();i++){
    sym->iugen[i]=psubgroupgenerators[i];
  }
  
  long err=calcsym(sym,pspacegroup,"",palatD(),panglesD(),1L
		   ,pstructuretype);


  
  time.print(String(nsubfunc)+" calc group");
  
  if( !giveinfo &&  err==INFORMATION ){
    err=0;
  };
  /* catch errors  */
  seterrorhandling(ERR_CATCH );
  String msg;
  if(err) {
    msg+=errorgetbuffer();
  }
  //myDebug() <<" err=" <<err ;
  if(err>WARNING) {
    //msg+="\nSorry ABORT!!!\n ";
    throw Except(msg,"Symmetry calculation",1);
    flusherror(stdout);
  }

  pcalced=true;
  pchanged=false;

  /* collect output */
  tsCBUFFER bs;
  CBInit(&bs,50000,10000);
  CBStart(&bs);
  reportcalcsym(&bs,sym);

  out+=CBGetBuffer(&bs);

  CBEnd(&bs);

  double fac=constant_aBtoAng*constant_aBtoAng*constant_aBtoAng;
  if(lengthUnit()==Xf::BohrRadius){
    out+=Sprintf("Volume (real cell) [aB^3] : %- 22.15f\n",vuc())();
    out+=Sprintf("Volume (real cell) [Ang^3]: %- 22.15f\n",vuc()*fac)();
  }else{
    out+=Sprintf("Volume (real cell) [aB^3] : %- 22.15f\n",vuc()/fac)();
    out+=Sprintf("Volume (real cell) [Ang^3]: %- 22.15f\n",vuc())();
  }


  pdatalengthunit=plengthunit;
  pdatastructuretype=pstructuretype;

  if(calcAtoms){
    try{
      calcAtomsFromGroup();
      
      time.print(String(nsubfunc)+" calcAtomsFromGroup");
            
      out+=reportCalcStructure();

      time.print(String(nsubfunc)+" reportCalcStructure");
      
    }catch(Except ex){
      msg+=ex.msg();
      if(ex.stop()==2) err=MyMath::max<long>(err,SEVERE_ERROR);
      if(ex.stop()==1) err=MyMath::max<long>(err,ERROR);
      if(ex.stop()==0) err=MyMath::max<long>(err,WARNING);
    }

    if(calcpairinfo){

      // combiShells();
      
      // time.print(String(nsubfunc)+" combiShells");

      String m;
      long ret=checkSites(m);
      time.print(String(nsubfunc)+" checkSites");
      
      if(ret>0){
	err=MyMath::max<long>(err,ret);
	msg+=m;
      }
      
      
      try{
	calcAutoRadii();
      }catch(Except ex){
	msg+=ex.msg();
	if(ex.stop()==2) err=MyMath::max<long>(err,SEVERE_ERROR);
	if(ex.stop()==1) err=MyMath::max<long>(err,ERROR);
	if(ex.stop()==0) err=MyMath::max<long>(err,WARNING);
      }
      
      time.print(String(nsubfunc)+" calcAutoRadii");
      
      determineShells();
      
      time.print(String(nsubfunc)+" determineShells");

    }
  }
  
  
  
  if(err) {

    //myDebug() <<" msg=" <<msg ;
  
    if(( err == INFORMATION) ||
       err == WARNING ){
      throw Except(msg,"Symmetry calculation",0);
    }else if(err == ERROR ){
      throw Except(msg,"Symmetry calculation",1);
    }else if(err == SEVERE_ERROR ){
      msg+="\nSorry ABORT!!!\n ";
      throw Except(msg,"Symmetry calculation",2);
      flusherror(stdout);
    }
  }

}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::calcAtomsFromGroup()
{
  NSF("Symmetry::calcAtomsFromGroup");
  //myDebug() << nsubfunc;
//   long nsort,i,is,is_crystal;
//   double *wtau;
//   char *p=NULL;

  double tol=1.0e-4;

  seterrorhandling(ERR_NOCATCH | ERR_BUFFERED |ERR_RESETBUFFER);
  const char *title="Site calculation from group";

  /* generate sites from WyckPos */
  long nsort=pwyckpos.size();

  double *wtau=(double*)malloc(sizeof(double)*3* (nsort));

  for(long is=0;is<nsort;is++){
    for(long i=0;i<3;i++){
      wtau[i+3*is]=pwyckpos[is].tauD()[i];
    }
  }
  double *taus=(double*)malloc(sizeof(double)*3* (nsort)*48);
  long *isrt=(long*)malloc(sizeof(long)*3*(nsort)*48);

  long nsite;
  if(pstructuretype == Xf::STCrystal){
    generatetau(sym,&nsite,nsort,wtau,taus,isrt,tol);
  }else if(pstructuretype == Xf::STMolecule){
    generatetau_mol(sym,&nsite,nsort,wtau,taus,isrt,tol);
  }else{
    throw Except(
		 Sprintf("Structure type '%s' not implemented"
			 ,structureTypes(pstructuretype))
		 ,title,1);
  }

  seterrorhandling(ERR_CATCH );
  
  long err=severitylevel();
  
  String msg;
  if(err) {
    msg+=errorgetbuffer();
  }
  
  psites.clear();
  psitebysort.clear();
  psitelistbysort.clear();
  for(long issa=0;issa<nsort;issa++) {
    psitebysort.append(-1);
    psitelistbysort.append(ValueArray<long>());
  }
  
  for(long isa=0;isa<nsite;isa++){
    long issa=isrt[isa];
    psites.append(Site(pwyckpos[issa].el(),pwyckpos[issa].type(),issa,taus+3*isa
		       ,pdatalengthunit,pwyckpos[issa].property()
		       ,pwyckpos[issa].visible()));
    if(psitebysort[issa]<0) psitebysort[issa]=isa;
    psitelistbysort[issa].append(isa);
  }
  
  free(wtau);
  free(taus);
  free(isrt);

  if(err) {

    //myDebug() <<" msg=" <<msg ;
  
    if(err == INFORMATION ||
       err == WARNING ){
      throw Except(msg,title,0);
    }else if(err == ERROR ){
      throw Except(msg,title,1);
    }else if(err == SEVERE_ERROR ){
      //      msg+="\nSorry ABORT!!!\n ";
      throw Except(msg,title,2);
      flusherror(stdout);
    }
  }

  return ;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::applyTheme(const AtomPropertyTable *ptable)
{
  long nsort=pwyckpos.size();
  for(long is=0;is<nsort;is++){
    for(long i=0;i<3;i++){
      pwyckpos[is].setProperty(ptable->find(pwyckpos[is].el()));
    }
  }
  for(int i=0;i<psites.size();i++){
    long issa=psites[i].sort();
    psites[i].setProperty(pwyckpos[issa].property());
  }
  setChanged(true);

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
String Symmetry::reportCalcStructure()
{

  const char *line70="--------------------------------"
    "--------------------------------------\n";

  long nsite=psites.size();

  String s;

  s+="\n";
  s+=line70;
  s+="                     Report of  Unit Cell Creation   \n";
  s+=line70;

   /* Wyckoff positions */
  s+="\n                   WYCKHOFF POSITIONS\n\n";

  if(sym->RTG  && pstructuretype == Xf::STCrystal)
    s+="RTG as simple bravais lattice\n\n";

  s+=Sprintf("Number of Wyckhoff positions : %ld\n",pwyckpos.size());
  s+="No.   Element  Type    X               Y               Z  \n";
  for(long ig=0;ig<pwyckpos.size();ig++){
    s+=Sprintf(" %-4ld    %-4s    %-4ld  %-16s %-16s %-16s\n"
			 ,ig+1,pwyckpos[ig].el()()
			 ,pwyckpos[ig].type()
			 ,pwyckpos[ig].tauS(0)()
			 ,pwyckpos[ig].tauS(1)()
			 ,pwyckpos[ig].tauS(2)() );
  }

  if(sym->RTG && pstructuretype == Xf::STCrystal ){
    s+="\nRTG as nonsimple STG-bravais lattice\n\n";

    s+=Sprintf("Number of Wyckhoff positions : %ld\n",pwyckpos.size());
    s+="No.   Element  Type    X               Y               Z  \n";
    for(long ig=0;ig<pwyckpos.size();ig++){
      double dum[3];
      ConvertVecRTGtoSTG(dum,pwyckpos[ig].tauD()());
      String st[3];
      for(int i=0;i<3;i++)
	st[i]=String::doubleToRatio(dum[i],20,1.e-5);
      s+=Sprintf(" %-4ld    %-4s    %-4ld  %-15s %-15s %-15s\n"
		 ,ig+1,pwyckpos[ig].el()(),pwyckpos[ig].type()
		 ,st[0]() ,st[1]() ,st[2]());
    }
  }
  /* sites positions */
  s+="\n                   ATOM SITES\n\n";
  s+=Sprintf("Number of sites : %ld\n",nsite);
  s+="No.   WPS   Element  Type     X               Y               Z  \n";
  for(long i=0;i<nsite;i++){
    s+=Sprintf(" %-4ld %-4ld   %-4s     %-4ld   %- 15.5f %- 15.5f %- 15.5f\n"
	       ,i+1 ,psites[i].sort()+1 ,psites[i].el()() ,psites[i].type()
	       ,psites[i].tauD()[0] ,psites[i].tauD()[1] ,psites[i].tauD()[2] );
  }

  s+="\n";
  s+=line70;
  s+="\n";

  /* atom concentrations */
  s+="\n                   NOMINAL CONCENTRATION\n\n";
  typedef Pair<String,long> PSL;
  ValueArray<PSL> data;
  for(long i=0;i<nsite;i++){
    String el=psites.at(i).el();
    
    PSL *p=data.find(PSL(el));
    if(p){
      p->v2++;
    }else{
      data.append(PSL(el,1)); 
    }
  }
  pformula="";
  for(long i=0;i<data.size();i++){
    const PSL &psl=data.at(i);
    pformula+=Sprintf("%s%ld",psl.v1(),psl.v2);
  }
  s+="Element  Multiplicity Concentration\n";

  for(long i=0;i<data.size();i++){
    const PSL &psl=data.at(i);
    s+=Sprintf(" %2s  %5ld %g%%\n",psl.v1(),psl.v2,100.*psl.v2/nsite);
  }
  s+="\n";
  s+=line70;
  s+="\n";

  return s;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
long Symmetry::checkSites(String &msg)
{

  double tol=1.0e-4;
  double tol2=tol*tol;
  msg="";

  
  long nsite=psites.size();
  // make relative
  ValueArray<Vec3> taus;
  taus.resize(nsite);
  for(long j=-1;++j<nsite;){
    if(pstructuretype == Xf::STCrystal){
      taus[j]=psites.at(j).tauD()*rcell();
    }else{
      taus[j]=psites.at(j).tauD();
    }
  }


  /* for all WPS */
  long nsort=pwyckpos.size();
  for(long issa=-1;++issa<nsort;){

    /* j is first occurance of WPS in sites */
    long isa0=siteBySort(issa);
    if(isa0 == -1 ){
      msg.sprintf("Heavy! Cant find sort %ld \n",issa);
      return 2; // fatal
    }
    const Vec3 &tau0=taus[isa0];
    
    for(long j=-1;++j<nsite;){
      if(psites[j].sort()<=issa) continue;
    
      Vec3 angrel=tau0-taus[j];
      

      double equiv;
      if(pstructuretype == Xf::STCrystal){
	equiv=normsq(angrel.backfold());
      }else{
	equiv=normsq(angrel);
      }
      
      if(equiv < tol2 ) {
	Site &sj=psites[j];
	msg.sprintf("Equivalent Wyckoff positions!\n"
		    "\tatom:sort %ld type %ld %s at %g %g %g and \n"
		    "\tatom: sort %ld type %ld %s at %g %g %g"
		    ,psites[isa0].sort()+1,psites[isa0].type(),psites[isa0].el()()
		    ,tau0[0] ,tau0[1] ,tau0[2]
		    ,sj.sort()+1,sj.type(),sj.el()()
		    ,sj.tauD()[0] ,sj.tauD()[1] ,sj.tauD()[2]
		    );
	
	
	
	return 1; // non fatal
      }
    }
  }
  return 0;
}

/* ***********************************************************************
 * auto atom radii
 * ***********************************************************************
 */
void Symmetry::calcAutoRadii()
{
  NSF("Symmetry::calcAutoRadii");

  double tol=1.e-5;
  MachineTime time;
  /* calculate the max site distance*/

  long nsite=psites.size();
  if(!nsite) return;
  
  
  if(pstructuretype == Xf::STMolecule){
    if(nsite==1){
      long is=0;
      psites[is].setNNDist(1.);
      //printf("isite %ld: rmin=%g\n",is,psites[is].nndist); 
      return;
    }
    double tolmin=0.1;
    bool close=false;
    for(long is=0;is<nsite;is++){
      double rmin=norm(psites[is].tauD()-psites[(is+1)%nsite].tauD()); 

      for(long js=0;js<nsite;js++){
	if(is==js) continue;
	Vec3 p=psites[is].tauD()-psites[js].tauD();
	double d=norm(p);
	if(d > tol)
	  rmin=MyMath::min<double>(rmin,d);
	if(d<tolmin && is!=js){
	  close=true;
	}
	
      } /* js */
      psites[is].setNNDist(rmin);
      //printf("isite %ld: rmin=%g\n",is,psites[is].nndist); 
    }

    if(close)
      throw Except("Atoms are very close. \n"
		   "Check symmetry group and positions."
		   ,"Symmetry calculation",0);
    
    return;
  }


  Matrix3x3 lcell(sym->cell);
  Matrix3x3 lrcell(sym->rcell);

  // make relative
  double tausrel[nsite*3];
  Vec3 v;
  for(long j=-1;++j<nsite;){
    v=psites.at(j).tauD();
    if(pstructuretype == Xf::STCrystal) v=v*lrcell;
    tausrel[0+3*j]=v[0];
    tausrel[1+3*j]=v[1];
    tausrel[2+3*j]=v[2];
  }

  time.print(String(nsubfunc)+" make relative");
  
  double ga=norm(lrcell(0));
  double gb=norm(lrcell(1));
  double gc=norm(lrcell(2));
  //double ga2=ga*ga;
  //double gab=lrcell(0)*lrcell(1)/ga2;
  // double fab2=normsq(lrcell(1)-gab*lrcell(0));
  // double fab=sqrt(fab2);


  /* for all sites in unit cell */
  bool compare=false;
  long newcount=0;
  long skip0=0;
  long skip1=0;
  long skip2=0;
  for(long is=0;is<nsite;is++){
    double rmin; 
    double rmin1; //todo:remove
    if(nsite==1){
      rmin=norm(lcell(0));
    }else{
      rmin=norm(psites[is].tauD()-psites[(is+1)%nsite].tauD()); 
    }
    rmin1=rmin;//todo:remove

    for(long js=0;js<nsite;js++){
      double pp[3];
      pp[0]=tausrel[0+3*js]-tausrel[0+3*is];
      pp[1]=tausrel[1+3*js]-tausrel[1+3*is];
      pp[2]=tausrel[2+3*js]-tausrel[2+3*is];
      
      /* for if p is in the first cell then the surounding cells
	 can make it shorter
      */

      if(compare){
	Vec3 p=Vec3(pp).backfold();
	double pp[3];
	pp[0]=p[0];
	pp[1]=p[1];
	pp[2]=p[2];
	for(int h=-1;h<=1;h++){
	  for(int k=-1;k<=1;k++){
	    for(int l=-1;l<=1;l++){
	      /* translate site */
	      double d=sqrt(lcell.dot3normsq(pp[0]+h,pp[1]+k,pp[2]+l));
	      if(d > tol)
		rmin=(rmin<d?rmin:d);
	    } /* l */
	  } /* k */
	} /* h */
      }

      {
	long N0min=(long)ceil (-rmin1*ga-pp[0]-0.001);
	long N0max=(long)floor( rmin1*ga-pp[0]+0.001);
	
	if(N0max<N0min) { skip0++; continue; }

	for(int h=N0min;h<=N0max;h++){
	  double D0=pp[0]+h;
	  long N1min=(long)ceil (-rmin1*gb-pp[1]-0.01);
	  long N1max=(long)floor( rmin1*gb-pp[1]+0.01);

	  if(N1max<N1min) { skip1++; continue; }


	  for(int k=N1min;k<=N1max;k++){
	    long N2min=(long)ceil (-rmin1*gc-pp[2]-0.01);
	    long N2max=(long)floor( rmin1*gc-pp[2]+0.01);
	    
	    if(N2max<N2min) { skip2++; continue; }

	    for(int l=N2min-1;++l<=N2max;){
	      newcount++;
	      /* translate site */
	      double d=sqrt(lcell.dot3normsq(D0,pp[1]+k,pp[2]+l));
	      if(d > tol)
		rmin1=(rmin1<d?rmin1:d);
	    } /* l */
	  } /* k */
	} /* h */
	
      }

      if(compare){
	if(fabs(rmin-rmin1)>tol) throw Except("algo error");
      }
      

      
      

      
    } /* js */
    psites[is].setNNDist(rmin1);
    if(compare) printf("isite %ld: rmin=%g\n",is,psites[is].nndist()); 
    //printf("isite %ld: rmin=%g\n",is,psites[is].nndist()); 
  } /* is */
  myDebug()<<"total ="<<nsite*nsite*9;
  myDebug()<<"actual="<<newcount;
  myDebug()<<"skip0 ="<<skip0;
  myDebug()<<"skip1 ="<<skip1;
  myDebug()<<"skip2 ="<<skip2;
  
  time.print(String(nsubfunc)+" big loop");

}






/* ***********************************************************************
 * sort pair vectors
 * ***********************************************************************
 */
static int sortPairVec(const PairVec *a,const PairVec *b)
{
  double tol=1.0e-8;
  if(a->pabs<b->pabs-tol )return -1;
  else if(a->pabs<b->pabs+tol ){
    if(  strcmp(a->el(),b->el())<0 ) return -1;
    if(  strcmp(a->el(),b->el())==0 ) {
      if(  a->sort<b->sort ) return -1;
      if(  a->sort>b->sort ) return 1;
      return 0;
    }
    return 1;
  }else{
    return 1;
  }
  
}

/* ***********************************************************************
 * for PtrArray
 * ***********************************************************************
 */
static int sortPairVecPtr(const PairVec **a,const PairVec **b)
{
  double tol=1.0e-8;
  if((*a)->pabs<(*b)->pabs-tol )return -1;
  else if((*a)->pabs<(*b)->pabs+tol ){
    if(  strcmp((*a)->el(),(*b)->el())<0 ) return -1;
    if(  strcmp((*a)->el(),(*b)->el())==0 ) {
      if(  (*a)->sort<(*b)->sort ) return -1;
      if(  (*a)->sort>(*b)->sort ) return 1;
      return 0;
    }
    return 1;
  }else{
    return 1;
  }
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

void Symmetry::combiShells()
{
  NSF("Symmetry::combiShells");

  long nsite=psites.size();
  myDebug()<<"nsite="<<nsite;

  //todo: molecule
  if(pstructuretype != Xf::STCrystal){
    throw Except(String(nsubfunc)+"Molecules not impl");
  }
  
  Matrix3x3 lcell(cell());
  Matrix3x3 lrcell(lcell.inverseT());
  
  double Gx=sqrt(lrcell(0)*lrcell(0));
  double Gy=sqrt(lrcell(1)*lrcell(1));
  double Gz=sqrt(lrcell(2)*lrcell(2));
  double ax=sqrt(lcell(0)*lcell(0));
  double ay=sqrt(lcell(1)*lcell(1));
  double az=sqrt(lcell(2)*lcell(2));
  //todo: make the cell as isotropic as possible (angles)!

  myDebug()<<"axyz="<<ax<<ay<<az;


  long N0=15;
  long Lx=MyMath::max<long>(1,dRound(pow((ax*ax)/(ay*az)*nsite/N0,1./3.)));
  long Ly=dRound(Lx*(ay/ax));
  long Lz=dRound(Lx*(az/ax));
  myDebug()<<"N0="<<N0;
  myDebug()<<"L="<<Lx<<Ly<<Lz;

  
  // make relative
  ValueArray<Vec3> taus;
  taus.resize(nsite);
  long *ijk=new long[nsite*3];
  
  for(long ii=-1;++ii<nsite;){
    taus[ii]=(psites.at(ii).tauD()*rcell()).firstOctant();
    if(taus[ii][0]<0. || taus[ii][1]<0. || taus[ii][2]<0.){
      throw Except("firstOctant does not work as expected");
    }
    //myDebug()<<taus[ii].printS();

    
    ijk[0+ii*3]=(long)floor(taus[ii][0]*Lx);
    ijk[1+ii*3]=(long)floor(taus[ii][1]*Ly);
    ijk[2+ii*3]=(long)floor(taus[ii][2]*Lz);
    //myDebug()<<"ijk="<<ijk[0+ii*3]<<ijk[1+ii*3]<<ijk[2+ii*3];
    
  }

  long LLL=Lx*Ly*Lz;

  ValueArray<ValueArray<long> > cellcontent;
  cellcontent.resize(LLL);

  for(long isa=-1;++isa<nsite;){
    long i=ijk[0+isa*3];
    long j=ijk[1+isa*3];
    long k=ijk[2+isa*3];
    long ijk=i+Lx*(j+k*Ly);
    //myDebug()<<"ijk="<<i<<j<<k<<ijk<<taus[isa].printS("%30.15f");
    if(ijk>=LLL) throw Except(Sprintf("ijk=%ld index > LLL=%ld",ijk,LLL));
    
    cellcontent[ijk]<<isa;
  }
  

  // now, for each cell do a separate O(N0^2) job to determine some
  // meaninfull radius and the starting set of shells.
  // todo: if the cell has only one atom, special code is needed

  double tol=1.0e-4;
  double tol2=tol*tol;
  ValueArray<double> rmin(nsite,-1.);
  
  for(long icell=-1;++icell<LLL;){
    const ValueArray<long> &cel=cellcontent.at(icell);
    myDebug()<<cel;
    long n=cel.size();
    if(n==1) throw Except("todo: one-atom cell");
    
    for(long ic=-1;++ic<n;){
      long isa=cel.at(ic);
      for(long jc=-1;++jc<n;){
	if(jc==ic) continue;
	long jsa=cel.at(jc);
	Vec3 dt=taus[jsa]-taus[isa];
	Vec3 t=lcell*dt;
	double d=normsq(t);
	if(d<tol2) throw Except("wyckoffs sit on top of each other");//todo: better message
	rmin[isa]=(rmin[isa]<0?d:(d<rmin[isa]?d:rmin[isa]));
	
      }
      rmin[isa]=sqrt(rmin[isa]);
      myDebug()<<"rmin="<<ic<<rmin[isa];
    }
    
  }
  
  double nfac=3;
  double rhotol=0.01;
  // look for surrounding cells, which could contribute and complete the job
  
  for(long icell=-1;++icell<LLL;){
    const ValueArray<long> &cel=cellcontent.at(icell);
    long n=cel.size();
    for(long ic=-1;++ic<n;){
      long isa=cel.at(ic);
      double rho=rmin[isa]*nfac;
      // find my own cell
      long i0=ijk[0+isa*3];
      long j0=ijk[1+isa*3];
      long k0=ijk[2+isa*3];

      // // min[i|(i/Lx*R1-tau)*n1>rho]
      // // min[ i>irho=Lx(rho*|G1|+tau*G1) 
      double x=Lx*((rho+rhotol)*Gx+taus[isa][0]);
      long imax=(long)ceil(x)-1;
      x=Lx*(-(rho+rhotol)*Gx+taus[isa][0]);
      long imin=(long)floor(x);

      x=Ly*((rho+rhotol)*Gy+taus[isa][1]);
      long jmax=(long)ceil(x)-1;
      x=Ly*(-(rho+rhotol)*Gy+taus[isa][1]);
      long jmin=(long)floor(x);

      x=Lz*((rho+rhotol)*Gz+taus[isa][2]);
      long kmax=(long)ceil(x)-1;
      x=Lz*(-(rho+rhotol)*Gz+taus[isa][2]);
      long kmin=(long)floor(x);

      

      myDebug("cidx[%ld],site[%ld]: ijk=%ld,%ld,%ld ijk in [%ld:%ld,%ld:%ld,%ld:%ld]\n",
	      icell,isa,i0,j0,k0,imin,imax,jmin,jmax,kmin,kmax);

#define mymod(i,Lx,i1,h)			\
      long i1=i%Lx;				\
      long h=(i-i1)/Lx;				\
      if(i1<0){					\
	i1+=Lx; h--;				\
      }						
    						

      for(long i=imin-1;++i<=imax;){
	mymod(i,Lx,i1,H);
	for(long j=jmin-1;++j<=jmax;){
	  mymod(j,Ly,j1,K);
	  for(long k=kmin-1;++k<=kmax;){
	    mymod(k,Lz,k1,L);
	    myDebug("ijk=(%ld,%ld,%ld)=R1*(%ld,%ld,%ld)"
		    "+cell(%ld,%ld,%ld)",i,j,k,H,K,L,i1,j1,k1);
	  }
	}
      }

      
    }
  }



  // cleanup/throw-away the shells on the now available radius information

  
  delete [] ijk;
#undef mymod

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::determineShells()
{
  NSF("Symmetry::determineShells");
  //myDebug() << nsubfunc;

  double nfac=3;

  double tol=1.0e-4;  

  long h1[2],h2[2],h3[2];

  ppairs.clear();
  if(psites.size()>0)
    ppairs.setSize(psites.size());

  for(long isa=0;isa<psites.size();isa++){

    double radius=MyMath::max<double>(psites[isa].nndist(),1)*nfac;
    //myDebug()<<"radius="<<radius;
    
    for(long isb=0;isb<psites.size();isb++){

      Vec3 v=psites[isa].tauD()-psites[isb].tauD();

      if(pstructuretype == Xf::STCrystal){
	Boundary::hklEnclSphere(v,radius,rcell(),h1,h2,h3);
	
	for(long h=h1[0];h<h1[1];h++){
	  for(long k=h2[0];k<h2[1];k++){
	    for(long l=h3[0];l<h3[1];l++){
	      
	      Vec3 x=cell()*Vec3(h,k,l)-v;
	      double xa=norm(x);
	      
	      if(xa<=radius+tol && xa>tol)
		ppairs[isa].append(new PairVec(psites[isb].el(),isb
					       ,psites[isb].sort(),psites[isb].type()
					       ,x));
	      
	    }
	  }
	}
      }else{
	
	double xa=norm(v);
	if(xa<=radius+tol && xa>tol){
	  ppairs[isa].append(new PairVec(psites[isb].el(),isb
					 ,psites[isb].sort(),psites[isb].type()
					 ,-v));
	}
	      
      }

    }
    ppairs[isa].qSort(sortPairVecPtr);
  } //isa

  return;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::getGroupOperations(long num,Xf::StructureType str_type
				  ,ValueArray<Operation> &ops)
{
  tsSYMMETRY tsym;

  ops.clear();

  seterrorhandling(ERR_NOCATCH | ERR_BUFFERED |ERR_RESETBUFFER);

  long err=get_group_operations(num,&tsym);

  seterrorhandling(ERR_CATCH );


  String msg;
  if(err) {
    msg+=errorgetbuffer();
  }
  if(err>WARNING) {
    return ;
  }  

  
  for(int ig=0;ig<tsym.ng;ig++){
    
    int j=tsym.itable[ig];
    long ix=operationindex(tsym.opmat+9*j,tsym.hex);
    String st[3];
    for(int i=0;i<3;i++)
      st[i]=String::doubleToRatio(tsym.v[i+3*j],6,1.e-5);
    
    
    if(str_type==Xf::STCrystal)
      
      ops.append( Operation(j, Sprintf("(%-4s,%-4s,%-4s) + (%-4s,%-4s,%-4s)" 
				       ,operationXYZ(tsym.opmat+9*j)
				       ,operationXYZ(tsym.opmat+9*j+3)
				       ,operationXYZ(tsym.opmat+9*j+6)
				       ,st[0](),st[1](),st[2]()
				       )
			    ,operationsymbols[ix+1]
			    ) );
    else
      
      ops.append( Operation(j, Sprintf("(%-4s,%-4s,%-4s)" 
				       ,operationXYZ(tsym.opmat+9*j)
				       ,operationXYZ(tsym.opmat+9*j+3)
				       ,operationXYZ(tsym.opmat+9*j+6)
				       )
			    ,operationsymbols[ix+1]
			    ) );
    
  }
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
// void Symmetry::calcSymmetryOperations()
// {
//   bool prot=false;

//   ValueArray<Pair<String,bool> >previousdata;
//   for(int ig=0;ig<pspacegroupoperations->size();ig++){
//     const SpaceGroupOperationBase *spop=pspacegroupoperations->at(ig);
//     previousdata.append(makePair(spop->symbol(),spop->isVisible()));
//   }

//   pspacegroupoperations->clear();
//   pspacegroupoperations->setCell(cell());
  
//   long ng=numOps();
//   for(long ig=0;ig<ng;ig++){
//     bool hex=(crystal() == SYM_HEXAGONAL || 
// 	      crystal() == SYM_TRIGONAL );
//     long it=iTable(ig);
//     long ix=operationindex(opmat()+9*it,hex);
//     double dum[9];
//     Matrix3x3 Gen=Matrix3x3(transbtoa(dum,opmat()+9*it,hex));//*prprimcell;
//     Vec3 gtau=ccell()*Vec3(optau()+3*it);
//     if(prot){
//       printf("g(%ld): it=%ld \nR=\n%s\ntau=%s\nTau=%s\ng 100=%s\ng 010=%s\n",ig,it
// 	     ,Gen.printS("%10.5f")()
// 	     ,Vec3(optau()+3*it).printS("%10.5f")()
// 	     ,gtau.printS("%10.5f")()
// 	     ,(Vec3(1,0,0)*Gen).printS("%10.5f")()
// 	     ,(Vec3(0,1,0)*Gen).printS("%10.5f")()
// 	     );
//     }    
    
//     pspacegroupoperations->append(Gen,gtau,prot);
    
//     SpaceGroupOperationBase *spop=pspacegroupoperations->last();

//     for(long ii=-1;++ii<previousdata.size();){
//       const Pair<String,bool> &p=previousdata.at(ii);
//       if(p.v1==spop->symbol()()){
// 	spop->setVisible(p.v2);
// 	break;
//       }
//     }

    
//     if(prot){
//       printf("Symbol=%s\n      =%s\n",operationsymbols[ix+1],
// 	     pspacegroupoperations->last()->symbol()());
//     }

//   }
// }


/* ***********************************************************************
 *
 * ***********************************************************************
 */
Matrix3x3 Symmetry::convertToSimple(const Matrix3x3 &cell,long bravais
				   ,Xf::LatticeType lattype)
{
  Matrix3x3 m;
  if(lattype==Xf::LTDirect){
    if(bravais == SYM_SIMPLE ){
      converttosimple(cell(),m.data(),SYM_SIMPLE ); 
    }else if(bravais == SYM_FACECENTERED ){
      converttosimple(cell(),m.data(),SYM_FACECENTERED ); 
    }else if(bravais == SYM_BODYCENTERED ){   
      converttosimple(cell(),m.data(),SYM_BODYCENTERED ); 
    }else if(bravais == SYM_CBASECENTERED ){   
      converttosimple(cell(),m.data(),SYM_CBASECENTERED ); 
    }else if(bravais == SYM_ABASECENTERED ){   
      converttosimple(cell(),m.data(),SYM_ABASECENTERED ); 
    }else if(bravais == SYM_RHOMBOHEDRAL ){   
      converttosimple(cell(),m.data(),SYM_RHOMBOHEDRAL ); 
    }
    
  }else{
    if(bravais == SYM_SIMPLE ){
      converttosimple(cell(),m.data(),SYM_SIMPLE); 
    }else if(bravais == SYM_FACECENTERED ){
      converttosimple(cell(),m.data(),SYM_BODYCENTERED ); 
    }else if(bravais == SYM_BODYCENTERED ){   
      converttosimple(cell(),m.data(),SYM_FACECENTERED ); 
    }else if(bravais == SYM_CBASECENTERED ){   
      converttosimple(cell(),m.data(),SYM_CBASECENTERED ); 
    }else if(bravais == SYM_ABASECENTERED ){   
      converttosimple(cell(),m.data(),SYM_ABASECENTERED ); 
    }else if(bravais == SYM_RHOMBOHEDRAL ){   
      converttosimple(cell(),m.data(),SYM_RHOMBOHEDRAL ); 
    }      
  }

  return m;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::convertTo(const String &mode)
{
  if(mode=="P1"){
    myDebug()<<"Symmetry::convertTo(): mode=P1";
  }else if(mode=="ConvCell"){
    //myDebug()<<"Symmetry::convertTo(): mode=ConvCell";
    increaseCell(Matrix3x3(),true);
  }else if(mode=="RemoveGenerators"){
    setCalculated(false);
    String out;
    calcSymFromGroupInput(out,true,false,false);
    long nsite=psites.size();
    ValueArray<int> l;
    l<<0;
    setSubGroupGenerators(l);
    pwyckpos.clear();
    for(long isa=0;isa<nsite;isa++){
      const Site &site=psites.at(isa);
      Vec3 tau=site.tauD();
      if(sym->RTG){
	tau=tau*cell().inverseT();
      }else{
	tau=tau*ccell().inverseT();
      }
      pwyckpos.append(WyckPos(site.el(),site.type(),tau.toString(12)
			      ,site.property(),site.visible()));
    }
  }else if(mode=="DetermineSymmetry"){
    setCalculated(false);
    String out;
    calcSymFromGroupInput(out,true,false,false);

    FILE *FO;
    double tol=1.0e-6;
    FO=openforwrite("=.str");
    fprintf(FO,"%g\n",tol);
    fprintf(FO,"%ld\n",pspacegroup);
    for(int i=0;i<3;i++){
      fprintf(FO,"%30.15g %30.15g %30.15g\n",
	      cell()(i)[0],
	      cell()(i)[1],
	      cell()(i)[2]
	      );
    }
    long nsite=psites.size();
    fprintf(FO,"%ld\n",nsite);
    fprintf(FO,"f\n");
    for(long isa=0;isa<nsite;isa++){
      const Site &site=psites.at(isa);
      Vec3 tau=site.tauD();
      // tau=tau*rcell();
      fprintf(FO,"%s %ld %30.15g %30.15g %30.15g\n",
	      site.el()(),site.type(),tau[0],tau[1],tau[2]);
    }
    fprintf(FO,"T\n");
    fprintf(FO,"1 0 0\n");
    fprintf(FO,"0 1 0\n");
    fprintf(FO,"0 0 1\n");
    fclose(FO);
    String serr,sout;
    int ret=mySystem("str",StringList(),sout,serr);
    //printf("ret: %d",ret);
    if(ret){
      printf("stdout: %s",sout());
      printf("stderr: %s",serr());
      throw Except("Could not launch program 'str', or error therein\n");
    }

    // now read back +atom
    FILE *FI=openforread("+atom");
    String line;
    while(getLine(FI,line)){
      line.stripBlanks(true);
      if(line.find(":SYMMETRY:")>=0){
	break;
      }
    }
    if(!getLine(FI,line))throw Except("missing info in +atom\n");
    line.stripBlanks(true);
    pspacegroup=line.convertToLong();

    if(!getLine(FI,line))throw Except("missing info in +atom\n");
    line.stripBlanks(true);
    setLatticeConstants(Vec3(line.split(" \t")).toString(1));

    if(!getLine(FI,line))throw Except("missing info in +atom\n");
    line.stripBlanks(true);
    setAxesAngles(Vec3(line.split(" \t")).toString(1));

    if(!getLine(FI,line))throw Except("missing info in +atom\n");
    line.stripBlanks(true);
    long nsort=line.convertToLong();
    pwyckpos.clear();
    for(long issa=0;issa<nsort;issa++){
      if(!getLine(FI,line))throw Except("missing info in +atom\n");
      line.stripBlanks(true);
      StringList sl=line.split(" \t");
      Vec3 v(sl[2],sl[3],sl[4]);
      pwyckpos.append(WyckPos(sl[0],0,v.toString(12),globalproperties
			      ->current()->find(sl[0]),true));
      pwyckpos.last().setDefaultType();
    }
    fclose(FI);
    setSubGroupGenerators(ValueArray<int>());

  }else{
    throw Except("Symmetry::convertTo(): unkown mode");
  }

  

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void enlargeCell(const Matrix3x3 &oldcell,const Matrix3x3 &newcell,
		 const ValueArray<Site> &oldsites,ValueArray<Site> &sites,
		 long prot=0)
// Matrices in column arrangement: cell=A^T=(a1,a2,a3).
{
  NSF("enlargeCell");
  double tol=1.0e-6;
  MachineTime time;
  
  if(prot>=1) myDebug()<<nsubfunc<<"START";
  
  if(prot>=1){
    myDebug()<<"oldcell=\n"<<oldcell;
    myDebug()<<"newcell=\n"<<newcell;
  }
  
  double d=Fnorm(oldcell-newcell)/9;
  if(prot>=10) myDebug("Fnorm(oldcell-newcell)=%30.15f tol=%30.15f\n",d,tol);
  if(d<tol){
    sites=oldsites;
    return;
  }


  // A'=U*A
  // In column arrangement: A=cell^T
  // newcell^T=U*oldcell^T
  // U=newcell^T*oldcell.inverse() // we need U in row arrangement
  // for enlarged_lattice_site_list
  Matrix3x3 U=newcell.transposed()*oldcell.inverseT();
  if(prot >= 1) myDebug() <<"U=(u1,u2,u3)^T=\n"<<U;


  if(prot>=1) printf("start cell enlargement\n");
  

  if(prot>=10){
    myDebug()<<"old sites=\n";
    for(long isa=-1;++isa<oldsites.size();) myDebug()<<oldsites.at(isa).printS();
  }
  
  ValueArray<Vec3> mappedsites;
  IntegerTricks::enlarged_lattice_site_list(U.data(),mappedsites);
  if(prot>=2){
    myDebug()<<"maps of (000) in newcell coordinates=\n";
    for(long isa=-1;++isa<mappedsites.size();) myDebug()<<mappedsites.at(isa);
  }


  long N=mappedsites.size();
  for(long isa=-1;++isa<N;){
    // s=A'^T*mappedsites
    mappedsites[isa]=newcell*mappedsites.at(isa);// expressed in absolute coordinates
  }

  if(prot>=2){
    myDebug()<<"maps of (000) in absolute coordinates=\n";
    for(long isa=-1;++isa<mappedsites.size();) myDebug()<<mappedsites.at(isa);
  }

  sites.clear();
  long oldnsite=oldsites.size();
  sites.resize(oldnsite*N);
  long n=0;
  for(long isa=-1;++isa<oldnsite;){
    for(long j=-1;++j<N;){
      Vec3 t=oldsites.at(isa).tauD()+mappedsites.at(j); // sites are absolute
      sites[n]=oldsites.at(isa);
      sites[n].setTauD(t);
      n++;
    }
  }


  if(prot>=1) myDebug()<<nsubfunc<<"END";
  
  time.print(String(nsubfunc)+" all");
}

// EOF



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::increaseCell(const Matrix3x3 &incrmat,bool toconventional,long prot)
// Make a new lattice A'=U*A and fill/transform sites.
// Input incrmat is in column arrangement incrmat=U^T=(I1,I2,I3).
{
  NSF("Symmetry::increaseCell");
  MachineTime time;
  if(prot>=1) myDebug()<<nsubfunc<<"START";

  setCalculated(false);
  String out;
  calcSymFromGroupInput(out,true,false,false);
  
  if(prot>=1) myDebug()<<"current A=(a1,a2,a3)=\n"<<cell();
  
  Matrix3x3 newcell;
  if(toconventional){
    if(prot>=1) myDebug()<<"toconventional\n";
    newcell=convertToSimple(cell(),bravais(),Xf::LTDirect);
  }else{
    if(prot>=1) myDebug()<<"incrmat=\n"<<incrmat;
    newcell=cell()*incrmat;
  }
  if(prot>=1) myDebug()<<"result=\n"<<newcell;

  double det=newcell.determinant();
  
  if(det<0){
    newcell.scaleColumn(0,-1);
  }
  
  double a=norm(newcell(0));
  double b=norm(newcell(1));
  double c=norm(newcell(2));
  
  double tol=1.0e-6;
  if(a<tol || b<tol || c<tol|| fabs(det)<tol*a*b*c){
    throw Except("increased cell seems to be coplanar\n");
  }
  
  double al=acos(newcell(1)*newcell(2)/(b*c))*180./Pi;
  double be=acos(newcell(0)*newcell(2)/(a*c))*180./Pi;
  double ga=acos(newcell(0)*newcell(1)/(a*b))*180./Pi;
  
  
  ValueArray<Site> newsites;
  enlargeCell(cell(),newcell,psites,newsites,prot);
  
  // repackage as P1
  pspacegroup=1;
  setSubGroupGenerators(ValueArray<int>());
  setLatticeConstants(Vec3(a,b,c).toString(1));
  setAxesAngles(Vec3(al,be,ga).toString(1));
  
  long newnsite=newsites.size();
  Matrix3x3 newg=newcell.inverseT();
  pwyckpos.resize(newnsite,true);
  for(long isa=0;isa<newnsite;isa++){
    const Site &site=newsites.at(isa);
    Vec3 tau=site.tauD();
    tau=tau*newg;
    pwyckpos[isa]=WyckPos(site.el(),site.type(),tau.toString(12)
			  ,site.property(),site.visible());
  }

  if(prot>=1) myDebug()<<nsubfunc<<"END";
  
  time.print(String(nsubfunc)+" all");
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Symmetry::insertVacuum(const Vec3 &/*axis*/,double amount,double z0)
{
  //todo: make a full generic algorithm
  //todo: where to split

  //myDebug()<<axis<<amount;

  setCalculated(false);
  String out;
  calcSymFromGroupInput(out,true,false,false);

  Matrix3x3 newcell=cell();
  double det=newcell.determinant();


  double c=norm(newcell(2));
  newcell.scaleColumn(2,(c+amount)/c);
  double a=norm(newcell(0));
  double b=norm(newcell(1));
  c=norm(newcell(2));
  
  double tol=1.0e-6;
  if(a<tol || b<tol || c<tol|| fabs(det)<tol*a*b*c){
    throw Except("increased cell seems to be coplanar\n");
  }
  
  double al=acos(newcell(1)*newcell(2)/(b*c))*180./Pi;
  double be=acos(newcell(0)*newcell(2)/(a*c))*180./Pi;
  double ga=acos(newcell(0)*newcell(1)/(a*b))*180./Pi;

  pspacegroup=1;
  setSubGroupGenerators(ValueArray<int>());
  setLatticeConstants(Vec3(a,b,c).toString(1));
  setAxesAngles(Vec3(al,be,ga).toString(1));
  
  long newnsite=psites.size();
  Matrix3x3 newg=cell().inverseT();
  ValueArray<Site> tmpsites;
  for(long isa=0;isa<newnsite;isa++){
    Site &site=psites[isa];
    Vec3 tau=site.tauD();
    tau=tau*newg;
    tau[2]-=z0;
    while(tau[2]<0.){
      tau[2]+=1.;
    }
    tau=cell()*tau;
    site.setTauD(tau);
    tmpsites.append(site);
  }


  newnsite=tmpsites.size();
  newg=newcell.inverseT();
  pwyckpos.resize(newnsite,true);
  for(long isa=0;isa<newnsite;isa++){
    const Site &site=tmpsites.at(isa);
    Vec3 tau=site.tauD();
    tau=tau*newg;
    pwyckpos[isa]=WyckPos(site.el(),site.type(),tau.toString(12)
			  ,site.property(),site.visible());
  }



}

