/*************************************************************************
 * @(#)View.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   26 Jun 2009
 *************************************************************************
 */

#include <string.h> 
#include <math.h> 
#include <stdio.h> 
#include "View.hh"
#include "MyDebug.hh"
#include "Parser.hh"
#include "Except.hh"
#include "Settings.hh"
#include "Vec3.hh"








/* ***********************************************************************
 *
 * ***********************************************************************
 */
View::View()
  :plrim(0),prrim(0),ptrim(0),pbrim(0)
{
  setDefaultXYZRot();
  pfocus=Vec3();
  pradius=1.;
  pscale=1.;
  ptheta=10.;
  plambda=1.00;

  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void View::setZ(const Vec3 &vv)
{
  pz=vv;
  pz.normalize();
  pup=(pup-pz*(pz*pup)).normalize();
  px=pup.cross(pz);
}




/* ***********************************************************************
 *
 * ***********************************************************************
 */
void View::setUp(const Vec3 &vv)
{
  pup=(vv-pz*(pz*vv)).normalize();
  px=pup.cross(pz);
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void View::setDefaultXYZRot()
{


  setZ(Vec3(0.899871625622465,0.339982035955932,
	    0.2732104204).normalize());

  setUp(Vec3(-0.254888676746669,-0.0983821993269905
	     ,0.9619523424).normalize());


  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void View::setGLViewTransformation(double near,double far)
{
  double rl=plambda*pradius;
  double fz=pfocus*pz;
  double muf=0.1;
  double d=(far-near);
  double dn=near-d*muf;
  double df=far+d*muf;

  if(pperspective){
    pf=rl/(pscale*ptantheta);
    pN=MyMath::max<double>(minNear(),pf+dn+fz);
    pT=pN*ptantheta;
  }else{
    pT=rl/pscale;
    pN=0.;
    pf=-dn-fz;
  }
  pF=df-dn+pN;
  pshift=(pfocus+pf*pz);
  
  if(fabs(pN-pF)<1.0e-10)
    pF=1.0e-10+pN;

}





/* ***********************************************************************
 *
 * ***********************************************************************
 */
void View::print() const
{
  (void)printf("View Axes\n"
	       "up     %g %g %g\n"
	       "z      %g %g %g\n\n"
	       "focus  %g %g %g\n"
	       "radius %g \n"
	       "scale  %g \n"
	       "theta  %g \n"
	       ,pup [0],pup [1],pup [2]
	       ,pz  [0],pz  [1],pz  [2] 
	       ,pfocus[0],pfocus[1],pfocus[2] 
	       ,pradius,pscale,ptheta
	       );
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void View::PTSectionToFile(FILE *FH)
{
  fprintf(FH,
	  "    struct {real up[3];real z[3];real focus[3];real radius;\n"
	  "            real scale;real theta;logical perspective;}view\n"
	  "    ={{%s},\n"
	  "      {%s},\n"
	  "      {%s},\n"
	  "      %-30.15g,%-30.15g,%-30.15g,%c"
	  "    };\n\n"
	  ,pup.printS("%-30.15g",",")()
	  ,pz.printS("%-30.15g",",")()
	  ,pfocus.printS("%-30.15g",",")()
	  ,pradius,pscale,ptheta
	  ,(pperspective?'t':'f')
	  );
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void View::readPTSectionFromFile(Parser *par)
{
  
  try{ 
    pperspective=(bool)(*par)()("view.perspective").valueL();
  }catch(Except){};




  try{ 
    double t[3];
    (*par)()("view.z").getArray(3,t);
    setZ(t);
  }catch(Except){};

  try{ 
    double t[3];
    (*par)()("view.up").getArray(3,t);
    setUp(t);
  }catch(Except){};

  
  try{
    double t_mid[3];
    (*par)()("view.focus").getArray(3,t_mid);
    setFocus(t_mid);
  }catch(Except){};

  try{
    double t_radius;
    t_radius=(*par)()("view.radius").valueD();
    setRadius(t_radius);
  }catch(Except){};
  

  try{
    double t_scale;
    t_scale=(*par)()("view.scale").valueD();
    setScale(t_scale);
  }catch(Except){};

  try{
    double t_theta;
    t_theta=(*par)()("view.theta").valueD();
    setTheta(t_theta);
  }catch(Except){};
  

}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void View::saveCameraToSettings()
{
  //myDebug() <<"save";

  print();
  Settings set("CameraExchange");
  set.set("axes/z",pz);
  set.set("axes/up",pup);
  set.set("axes/focus",pfocus);


  set.set("axes/theta",ptheta);

  set.set("axes/scale",pscale);


  set.set("axes/radius",pradius);


  set.set("axes/perspective",pperspective);

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void View::loadCameraFromSettings()
{
  //myDebug() <<"load";
  Settings set("CameraExchange");

  setZ(set.getVec3("axes/z",pz));
  setUp(set.getVec3("axes/up",pup));
  setFocus(set.getVec3("axes/focus",pfocus));


  setTheta(set.getDouble("axes/theta",ptheta));
  setScale(set.getDouble("axes/scale",pscale));
  setRadius(set.getDouble("axes/radius",pradius));
  setPerspective(set.getBool("axes/perspective",pperspective));

  print();

}



/*EOF*/ 
