/*************************************************************************
 * @(#)DisplayCellWidget.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   09 Mar 2011
 *************************************************************************
 */

#include <QMessageBox>
#include "Boundary.hh"
#include "DisplayCells.hh"
#include "DisplayCellWidget.hh"
#include "Polyhedron.hh"
#include "Valid.hh"
#include "Config.hh"
#include "MyDebug.hh"


using namespace XFPLO;

/* ***********************************************************************
 *
 * ***********************************************************************
 */
DisplayCellWidget::DisplayCellWidget(QWidget *parent)
  :Sheet(parent),t_dc(0)
{
  setupUi(this);
  setObjectName("DisplayCellWidget");

  QSizePolicy sizePolicy1(QSizePolicy::Expanding, QSizePolicy::Expanding);
  sizePolicy1.setHorizontalStretch(3);
  sizePolicy1.setVerticalStretch(0);
  sizePolicy1.setHeightForWidth(sizePolicy().hasHeightForWidth());
  setSizePolicy(sizePolicy1);

  rimRmapper <<Xf::SUBoundary<<Xf::SUWorld<<Xf::SUAngstroem
	     <<Xf::SUBohrRadius;

  cmbRimUnits->clear();
  for(int i=0;i<rimRmapper.size();i++){
    cmbRimUnits->
      insertItem(i,Xf::sceeneUnitName(rimRmapper[i]),rimRmapper[i]);
  }


  new RatValidator(this,leRimRadius,"The rim radius ",1.e-10,1.e10);

  connectChanged(this);

  connect(grpbxRim,SIGNAL(toggled(bool)),this,SLOT(onGrpbxToggled()));
  connect(grpbxFaces,SIGNAL(toggled(bool)),this,SLOT(onGrpbxToggled()));

}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
DisplayCellWidget::~DisplayCellWidget()
{
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCellWidget::setData(DisplayCell *dc,const Symmetry *sym)
{
  t_dc=dc;
  //myDebug() <<"dc"<<dc;
  const PolyhedronProperties &polprops=t_dc->boundary()->polyhedron()
    ->properties();


  //cbActive->setChecked(polprops.isVisible());
  
  grpbxRim->setChecked(polprops.drawRim());
  grpbxFaces->setChecked(polprops.drawFaces());

  cmbRimUnits->setCurrentIndex(
			       rimRmapper
			       .indexOf(polprops.rimRadiusUnits())
			       );


  tbRimColor->setColor(polprops.rimMaterial().color());
  QRgb c=polprops.faceMaterial().color();
  tbFaceColor->setColor(c&0xffffff);
  dspbxFaceOpacity->setValue(qAlpha(c)/255.);
  dspbxFaceShine->setValue(polprops.faceMaterial().shininess());
  tbFaceSpecular->setColor(polprops.faceMaterial().specularColor());


  leRimRadius->setText(polprops.rimRadius().S());

  if(dc->type()==Xf::DCTGeneral){
    wBoundDlg->setVisible(true);
    wBoundDlg->setData(t_dc->boundary(),sym,false);
  }else{
    wBoundDlg->setVisible(false);
  }


  onGrpbxToggled();
  onActiveToggled(t_dc);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool DisplayCellWidget::isValid()
{
  if(!t_dc) return true;

  QString s,txt="";

  ((RatValidator*)leRimRadius->validator())->check(txt);

  
  if(t_dc->type()==Xf::DCTGeneral){
    txt+=wBoundDlg->dataAcceptable();
  }

  if(txt.length()){
    QMessageBox::warning ( this,IDS_APP_NAME+ " " + FPLOMAINVERSION()
			   , txt, 1, 0,0 ) ;

    return false;
  }
  
  return true;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool DisplayCellWidget::commit()
{
  NSF("DisplayCellWidget::commit");
  //myDebug() << nsubfunc<<t_dc->name()<<changed();
  if(!t_dc) return true;
  if(!isValid()) return false;

  QString t;

  PolyhedronProperties &polprops=t_dc->boundary()->polyhedron()->properties();

  //  polprops.setVisible(cbActive->isChecked());

  polprops.setDrawRim(grpbxRim->isChecked());
  polprops.setDrawFaces(grpbxFaces->isChecked());

  polprops.setRimMaterial(Material(tbRimColor->color(),50.));
  
  unsigned long a=(unsigned long)(dspbxFaceOpacity->value()*255);
  polprops.setFaceMaterial(Material(tbFaceColor->color()| a*0x1000000,
				    tbFaceSpecular->color(),
				    dspbxFaceShine->value()));

  t=leRimRadius->text(); IntValidator::stripAllWhiteSpace(t);
  polprops.setRimRadius(t);

  polprops.setRimRadiusUnits((Xf::SceeneUnits)cmbRimUnits->
		      itemData(cmbRimUnits->currentIndex(),Qt::UserRole)
		      .toInt());

  if(t_dc->type()==Xf::DCTGeneral){
    bool chn;
    wBoundDlg->getData(t_dc->boundary(),chn);
  }

  Sheet::commit();

  return true;

}





/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCellWidget::onActiveToggled(const DisplayCell *dc)
{
  if(!dc || (dc!=t_dc)) return;
  bool on=t_dc->isVisible();
  grpbxRim->setEnabled(on);
  grpbxFaces->setEnabled(on);
  if(t_dc->type()==Xf::DCTGeneral){
    wBoundDlg->setEnabled(on);
  }
  lbMsg->setEnabled(on);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void DisplayCellWidget::onGrpbxToggled()
{
  if(!(grpbxRim->isChecked() || grpbxFaces->isChecked()))
    lbMsg->setText("no rim, no faces: cell will be invisible");
  else
    lbMsg->setText("");
}



/*EOF*/ 
