/*************************************************************************
 * @(#)HsvSampler.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   20 Jun 2012
 *************************************************************************
 */

#include <QVBoxLayout>
#include <QPainter>
#include <QMouseEvent>
#include "HsvSampler.hh"
#include "MyMath.hh"
#include "Vec3.hh"
#include "MyDebug.hh"


/* ***********************************************************************
 *
 * ***********************************************************************
 */
HsvSampler::HsvSampler(QWidget *parent,int xmode,int ymode,bool circular)
  :QWidget(parent),pointradius(8),Rmin(0),
   background(0),needbackgroundupdate(false),nodelist(0),currentnode(0),
   pcircular(circular),pinterpolinspherical(true)
{
  if(xmode==Hue && ymode==Saturation){
    setMinimumSize(QSize(120,120));
  }else{
    setMinimumSize(QSize(60,120));
  }

  pmode[0]=xmode;
  pmode[1]=ymode;
  setMouseTracking(true);
  setCursor(Qt::CrossCursor);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
HsvSampler::~HsvSampler()
{
  delete background;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void HsvSampler::setNodeList(NodeList *nl)
{
  nodelist=nl;
  nodelist->recalculateIminImax();
  update();
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void HsvSampler::NodeList::recalculateIminImax()
{
  imin=0;
  imax=0;

  for(int i=0;i<size();i++){
    int h=lRound(at(i)->rawHue());
    int m=h%360,d=h/360; if(m<0){m+=360;d--;}
    if(i==0){ 
      imin=imax=d; 
    }else{
      imin=MyMath::min<long>(imin,d);
      imax=MyMath::max<long>(imax,d);
    }
  }

};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void HsvSampler::paintEvent(QPaintEvent*event)
{
  //myDebug()<<"HsvSampler::paintEvent";
  QWidget::paintEvent( event );
  int ww=width();
  int hh=height();
  int ww1=ww-1;
  int hh1=hh-1;
  double R=MyMath::min<double>(ww1*0.5,hh1*0.5);

  int N=nodelist->size();


  QPainter wp(this);
  wp.setRenderHint(QPainter::Antialiasing,true);

  bool newsize=(!background) || 
    (background->width() != ww) ||
    (background->height() != hh);

  needbackgroundupdate=
    (pmode[1]==HsvSampler::Parameter && pmode[0]==HsvSampler::None) ||
    (pmode[0]==HsvSampler::Hue);


  if(newsize ||needbackgroundupdate ){

    if(newsize){
      if(background) delete background; 
      background=new QImage(ww,hh,QImage::Format_ARGB32_Premultiplied);
    }
    
    if(pcircular){
      QRgb bgc=this->palette().color(QPalette::Button).rgb();
      for(int i=0;i<ww;i++){
	for(int j=0;j<hh;j++){
	  double x=i-ww1*0.5;
	  double y=j-hh1*0.5;
	  double r=(sqrt(x*x+y*y)-Rmin)/(R-Rmin);
	  double phi=atan2(y,x)/Pi*180.;
	  if(phi<0)phi+=360.;
	  double hue=0;
	  double sat=255.;
	  double val=255.;
	  if(r>1){
	    background->setPixel(i,j,bgc);
	  }else if(r<0){
	    if(pmode[1]==Value)
	      background->setPixel(i,j,0xff000000);
	    else if(pmode[1]==Saturation)
	      background->setPixel(i,j,0xffffffff);
	  }else{
	    if(pmode[0]==Hue) hue=MyMath::min<double>(phi,359);
	    if(pmode[1]==Saturation) sat=255.*r;
	    if(pmode[1]==Value) val=255.*r;
	    background->setPixel(i,j,QColor::fromHsv(hue,sat,val).rgb());
	  }
	}
      }
      
    }else{

      double hue=0;
      double sat=255.;
      double val=255.;
      for(int i=0;i<ww;i++){
	for(int j=0;j<hh;j++){
	  switch(pmode[0])
	    {
	    case Hue:
	      hue=359.*i/ww1*(nodelist->imax-nodelist->imin+1);
	      hue=lRound(hue)%360;
	      break;
	    case Value: val=255.*i/ww1; break;
	    case Saturation: sat=255.*i/ww1; break;
	    }
	  switch(pmode[1])
	    {
	    case Hue: hue=359.*j/hh1; break;
	    case Saturation: sat=255.*j/hh1; break;
	    case Value: val=255.*j/hh1; break;
	    case Parameter: if(pmode[0]==Value) sat=0.; break;
	    }
	  background->setPixel(i,j,QColor::fromHsv(hue,sat,val).rgb());
	}
      }
      
      if(pmode[1]==Parameter && pmode[0]==None){
	if(N<1){
	  background->fill(Qt::black);
	}else if(N==1){
	  background->fill(nodelist->at(0)->color().rgb());
	}else{
	  int j0=0;
	  double s0=0;
	  const Node *n0=nodelist->at(0);
	  for(int k=1;k<N;k++){
	    const Node *n1=nodelist->at(k);
	    double s1=MyMath::max<double>(s0,n1->parameter());
	    int j1=MyMath::max<long>(j0,s1*hh1);
	    
	    for(int j=j0;j<=MyMath::min<long>(hh1,j1);j++){
	      double s=0.;
	      if(j1>j0){
		s=1.*(j-j0)/(j1-j0);
	      }
	      //myDebug()<<pinterpolinspherical;
	      Node n=Node::interpolateNodes(*n0,*n1,s, pinterpolinspherical);
	      
	      QRgb col=n.color().rgb();
	      for(int i=0;i<ww;i++){
		background->setPixel(i,j,col);
	      }
	    }
	    j0=j1+1;
	    s0=s1;
	    n0=n1;
	  }
	}
      }
    }
  }
  wp.drawImage(0,0,*background);
  if(false&& pcircular){
    wp.setBrush(Qt::NoBrush);
    if(pmode[1]==Saturation)
      wp.setPen(Qt::black);
    else
      wp.setPen(Qt::white);
    QRectF r(ww1*0.5-Rmin,hh1*0.5-Rmin,2*Rmin,2*Rmin); 
    wp.drawEllipse(r);
    r.adjust(1,1,-1,-1);
    wp.drawEllipse(r);
  }

  if(!nodelist)return;

  QPainterPath curve;;
  if(N>1){
    if(pcircular || pinterpolinspherical){
      curve.moveTo(nodelist->at(0)->toDevice(this,true));
      for(int i=0;i<N-1;i++){
	Node *n1=(*nodelist)[i];
	Node *n2=(*nodelist)[i+1];
	double s1=n1->parameter();
	double s2=n2->parameter();
	int Nsub=MyMath::max<long>(1,lRound(s2-s1))/0.2*5*(nodelist->imax-nodelist->imin+1);
	for(int j=0;j<=Nsub;j++){
	  double x=j/(Nsub+1.);
	  double s=(s2-s1)*x+s1;
	  Node n=Node::interpolateNodes(*n1,*n2,x,pinterpolinspherical);
	  n.setParameter(s);
	  curve.lineTo(n.toDevice(this,true));
	}
      }
      curve.lineTo(nodelist->at(N-1)->toDevice(this,true));
      
    }else{
      curve.moveTo(nodelist->at(0)->toDevice(this,true));
      for(int i=0;i<N-1;i++){
	curve.lineTo(nodelist->at(i+1)->toDevice(this,true));
      }
    }
  }else if(N==1 && (pmode[1]==Parameter&& pmode[0]==None)){
    curve.moveTo(0.5*ww1,0);
    curve.lineTo(0.5*ww1,hh1);
  }

  wp.setPen(QPen(Qt::black,2));
  wp.setBrush(Qt::NoBrush);
  wp.drawPath(curve);

  for(int i=0;i<N;i++){
    nodelist->at(i)->paint(this,wp,true);
  }

  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void HsvSampler::mousePressEvent(QMouseEvent *event)
{
  if(!nodelist)return;

  //QWidget::mousePressEvent(event);
  int hh1=height()-1;
  QPoint pos=event->pos();

  currentnode=0;
  for(int i=0;i<nodelist->size();i++){
    Node *node=(*nodelist)[i];
    QPoint pt=node->toDevice(this,true).toPoint();
    delta=pos-pt;
    if(node->underCursor(this,pos)){
      lastdpos=pos-delta;
      currentnode=node;
      break;
    }
  }  

    
  if(event->button()!=Qt::RightButton){
    return;
  }
  
  if(event->button()==Qt::RightButton){
    if(currentnode){
      nodelist->removeOne(currentnode);
      currentnode=0;
      snapNodes();
      emit registerUndo();

      emit nodeChanged();
	
      return;
    }
  }

  //   if(pmode & HsvSampler::Parameter){
  //     return;
  //   }

  if(!currentnode){
    int N=nodelist->size();
    if(N==0){
      currentnode=new Node(0.5,0,255,255);
      nodelist->append(currentnode);
      delta=QPoint(0,0);
      currentnode->setPoint(this,pos,pos,true);
      currentnode->highlight=true;
      emit nodeChanged();
      return;

    }else{
      
      currentnode=new Node(1,240,255,255);
      currentnode->setPoint(this,pos,pos,true);
      currentnode->highlight=true;
      delta=QPoint(0,0);
      QPointF pt=currentnode->toDevice(this,true);
      Vec3 c(pt.x(),pt.y(),0);
      

      if(N==1 && pmode[1]==Parameter){
	double s=nodelist->at(0)->parameter();
	if(1.*pos.y()/hh1<s){
	  nodelist->prepend(currentnode);
	}else{
	  nodelist->append(currentnode);
	}
	emit nodeChanged();
	return;
      }

      // find closest
      Node *ninsert=0;
      Node *closest=0;

      if(pmode[0]!=Parameter && pmode[1]!=Parameter){
	double dist=-1;
	
	for(int i=0;i<N;i++){
	  Node *node=(*nodelist)[i];
	  QPointF pt=node->toDevice(this,true);
	  Vec3 n(pt.x(),pt.y(),0);
	  double dd=norm(n-c);
	  if((i==0||i==N-1)&&( dist<0 || dd<dist)){
	    dist=dd;
	    closest=node;
	  }
	  
	  if(i<N-1){
	    Node *nodep=(*nodelist)[i+1];
	    QPointF pt=nodep->toDevice(this,true);
	    Vec3 dnnp=Vec3(pt.x(),pt.y(),0)-n;
	    double dn=norm(dnnp);
	    Vec3 v=dnnp/dn;
	    Vec3 vp=v.cross(Vec3(0,0,1));
	    
	    double d=v*(c-n);
	    if(fabs(vp*(c-n))<=pointradius && 0<=d && d<=dn){
	      ninsert=node;
	      double s=d/MyMath::max<double>(1.,dn);
	      currentnode->setParameter((nodep->parameter()-node->parameter())*s
					+node->parameter());
	    }
	    
	  }
	}
      }else{
	// no parameter
	for(int i=0;i<N-1;i++){
	  Node *node=(*nodelist)[i];
	  Node *nodep=(*nodelist)[i+1];
	  double s1=node->parameter();
	  double s2=nodep->parameter();
	  double s=1.*pos.y()/hh1;
	  if(s1<s && s<s2){
	    double x=(s-s1)/(s2-s1);
	    ninsert=node;
	    *currentnode=Node::interpolateNodes(*node,*nodep,x,
						pinterpolinspherical);
	    QPoint p=QPoint(pos.x(),currentnode->yy(this,true));
	    currentnode->setPoint(this,p,p,true);
	    
	    currentnode->setParameter(s);
	  }
	}
      }
      

      if(ninsert){	
	nodelist->insert(nodelist->indexOf(ninsert)+1,currentnode);
      }else{
	if(closest && closest==nodelist->last()){
	  if(N>1){
	    double s0=nodelist->at(N-2)->parameter();
	    double s1=nodelist->at(N-1)->parameter();
	    (*nodelist)[N-1]->setParameter((s1-s0)*0.5+s0);
	  }
	  nodelist->append(currentnode);
	}else if(closest && closest==nodelist->at(0)){
	  if(N>1){
	    double s0=nodelist->at(0)->parameter();
	    double s1=nodelist->at(1)->parameter();
	    (*nodelist)[0]->setParameter((s1-s0)*0.5+s0);
	    currentnode->setParameter(0);
	  }
	  nodelist->prepend(currentnode);
	}
      }
      emit nodeChanged();
      return;
    }
  }


}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void HsvSampler::mouseReleaseEvent(QMouseEvent *event)
{
  if(!currentnode)return;
  QPoint dpos=event->pos()-delta;
  currentnode->setPoint(this,dpos,lastdpos,false);
  lastdpos=dpos;

  snapNodes();

  emit registerUndo();
  
  emit nodeChanged();
  currentnode=0;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void HsvSampler::mouseMoveEvent(QMouseEvent *event)
{
  //myDebug()<<"HsvSampler::mouseMoveEvent";
  //QWidget::mouseMoveEvent(event);
  if(!nodelist)return;

  int ww1=width()-1;
  QPoint pos=event->pos();
  QPoint dpos=pos-delta;

  
  if(currentnode){ // move the current mode
    
    dpos.rx()=MyMath::max<long>(0,MyMath::min<long>(dpos.rx(),ww1));
    
    currentnode->setPoint(this,dpos,lastdpos,false);
    lastdpos=dpos;

    if(pmode[1]==Parameter){ // enforce monoonous parameter
      bool smaller=true;
      for(int i=0;i<nodelist->size();i++){
	Node *node=(*nodelist)[i];
	if(node==currentnode){smaller=false;continue;}
	if(smaller){
	  node->setParameter(MyMath::min<double>(node->parameter(),
				  currentnode->parameter()));
	}else{
	  node->setParameter(MyMath::max<double>(node->parameter(),
				  currentnode->parameter()));
	}
      }
    }
    emit nodeChanged();
    return;
  }


  {// set marker 
    // unmark all
    bool needupdate=false;
    for(int i=0;i<nodelist->size();i++){
      Node *node=(*nodelist)[i];
      if(node->highlight){
	node->highlight=false;
	needupdate|=true;
      }
    }
    
    // mark under cursor
    for(int i=0;i<nodelist->size();i++){
      Node *node=(*nodelist)[i];
      if(node->underCursor(this,pos)){
	node->highlight=true;
	needupdate|=true;
      }
    }
    
    if(needupdate){
      emit nodeChanged();
    }
    return ;
  }  
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void HsvSampler::leaveEvent(QEvent *event)
{
  QWidget::leaveEvent(event);
  if(!nodelist)return;

  if(currentnode)return;
  bool needupdate=false;
  for(int i=0;i<nodelist->size();i++){
    Node *node=(*nodelist)[i];
    if(node->highlight){
      node->highlight=false;
      needupdate|=true;
    }
  }

  if(needupdate){
    emit nodeChanged();
  }
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void HsvSampler::snapNodes()
{
  //  if(pmode & HsvSampler::Parameter){
    if(nodelist->size()==1){
      (*nodelist)[0]->setParameter(0.5);
    }else if(nodelist->size()>=2){
      (*nodelist)[0]->setParameter(0);
      (*nodelist)[nodelist->size()-1]->setParameter(1);
    }
    //}

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
QRectF HsvSampler::Node::boundingRect(HsvSampler *parent,
					   const QPointF &pt)const
{
  if(parent->pmode[0]!=HsvSampler::Parameter && 
     parent->pmode[1]!=HsvSampler::Parameter){
    return QRectF(pt.x()-parent->pointradius,pt.y()-parent->pointradius,
		  2.*parent->pointradius, 2.*parent->pointradius);
  }
  int ps=lRound(MyMath::max<double>(parent->pointradius*0.5,6));
  if(parent->pmode[0]==HsvSampler::None && 
     parent->pmode[1]==HsvSampler::Parameter){
    return QRectF(pt.x()-parent->pointradius,pt.y()-ps,
		  2.*parent->pointradius,2*ps);
  }else{
    return QRectF(pt.x()-ps,pt.y()-parent->pointradius,2*ps,
		  2.*parent->pointradius);
  }

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void HsvSampler::Node::paint(HsvSampler *parent,QPainter &wp,bool raw)const
{
  
  wp.setPen(Qt::NoPen);
  
  QPointF pt=toDevice(parent,raw);
  
  int dr=(highlight?2:1);

  QRectF r=boundingRect(parent,pt);
  
  if(parent->pmode[0]!=HsvSampler::Parameter && 
     parent->pmode[1]!=HsvSampler::Parameter){
    wp.setBrush(QBrush(Qt::black));wp.drawEllipse(r);
    r.adjust(dr,dr,-dr,-dr);
    wp.setBrush(QBrush(Qt::white));wp.drawEllipse(r);
    r.adjust(1,1,-1,-1);
    wp.setBrush(QBrush(color()));wp.drawEllipse(r);
  }else{
    wp.fillRect(r,QBrush(Qt::black));
    r.adjust(dr,dr,-dr,-dr);
    wp.fillRect(r,QBrush(Qt::white));
    r.adjust(1,1,-1,-1);
    wp.fillRect(r,QBrush(color()));
  }
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool HsvSampler::Node::underCursor(HsvSampler *parent,QPoint pos)const
{
  QPointF pt=toDevice(parent,true);
  
  if(parent->pmode[0]!=HsvSampler::Parameter && 
     parent->pmode[1]!=HsvSampler::Parameter){
    double dx=(pos.x()-pt.x());
    double dy=(pos.y()-pt.y());
    return (dx*dx+dy*dy<=parent->pointradius*parent->pointradius);
  }

  QRectF r=boundingRect(parent,pt);
  
  return r.contains(pos);
      
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void HsvSampler::Node::setPoint(HsvSampler *parent,QPoint pos,
				     QPoint oldpos,bool /*isnew*/)
{
  int ww1=parent->width()-1;
  int hh1=parent->height()-1;

  double xx,yy;
  if(parent->pcircular){
    QPointF c=QPoint(ww1*0.5,hh1*0.5);
    //myDebug()<<"setPoint,pos"<<pos-c;
    double R=MyMath::min<double>(ww1*0.5,hh1*0.5);
    xx=pos.x()-ww1*0.5;
    yy=pos.y()-hh1*0.5;
    //myDebug()<<"setPoint,xx,yy"<<xx<<yy;
    double r=(sqrt(xx*xx+yy*yy)-parent->Rmin)/(R-parent->Rmin);
    double phi=atan2(yy,xx)/Pi*0.5;
    if(phi<0)phi+=1.;
    xx=phi;
    yy=r;
    //myDebug()<<"setPoint phi,r"<<xx<<yy;
    

    //myDebug()<<"setPoint,oldpos"<<oldpos-c;
    QPoint dp=pos-oldpos;
    if(dp.y()!=0){
      double u=-(oldpos.y()-hh1*0.5)/dp.y();
      //myDebug()<<"setPoint,u"<<u;
      double x=dp.x()*u+oldpos.x()-ww1*0.5;
      if(x>0){
	if( ((oldpos-c).y()<0 && (pos-c).y() >=0) ){
	  //myDebug()<<"crossing to positive"<<phi;
	  //myDebug()<<"crossing to positive from"<<oldpos-c<<"to"<<pos-c;
	  phueinterval++;
	}
	if( ((oldpos-c).y()>=0 && (pos-c).y() <0) ){
	  //myDebug()<<"crossing to negative"<<phi;
	  phueinterval--;
	}
      }
    }

    xx=MyMath::min<double>(1.,MyMath::max<double>(0.,xx));
    yy=MyMath::min<double>(1.,MyMath::max<double>(0.,yy));
  }else{
    xx=1.*pos.x()/ww1;
    yy=1.*pos.y()/hh1;
    if(parent->pmode[0]==HsvSampler::Hue){

      xx=xx*(parent->nodelist->imax-parent->nodelist->imin+1)
	+parent->nodelist->imin;
      int m=(int)xx;
      if(xx<0)m--;
      //myDebug()<<xx<<m<<xx-m;
      phueinterval=m;
      xx=xx-m;
    }
    xx=MyMath::min<double>(1.,MyMath::max<double>(0.,xx));
    yy=MyMath::min<double>(1.,MyMath::max<double>(0.,yy));
  }    


  if(parent->pmode[1]==HsvSampler::Parameter){
    //if(mode[1]==HsvSampler::Parameter && mode[0]==HsvSampler::None){
    ps=yy;
    //    return;
  }
  
  if(parent->pcircular || parent->pmode[0]==HsvSampler::Hue){
    parent->nodelist->imin=MyMath::min<long>(parent->nodelist->imin,phueinterval);
    parent->nodelist->imax=MyMath::max<long>(parent->nodelist->imax,phueinterval);
    parent->nodelist->recalculateIminImax();

    phue=MyMath::max<long>(0,MyMath::min<long>(359,lRound(xx*360)));
  }
  if(parent->pmode[0]==HsvSampler::Value){
//     if(isnew){
//       psat=255;
//     }
    pval=MyMath::max<long>(0,MyMath::min<long>(255,lRound(xx*255)));
  }
  if(parent->pmode[0]==HsvSampler::Saturation){
//     if(isnew){
//       pval=255;
//     }
    psat=MyMath::max<long>(0,MyMath::min<long>(255,lRound(xx*255)));
  }
  if(parent->pmode[1]==HsvSampler::Saturation){
//     if(isnew){
//       pval=255;
//     }
    psat=MyMath::max<long>(0,MyMath::min<long>(255,lRound(yy*255)));

  }
  if(parent->pmode[1]==HsvSampler::Value){
//     if(isnew){
//       psat=255;
//     }
    pval=MyMath::max<long>(0,MyMath::min<long>(255,lRound(yy*255)));

  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
QPointF HsvSampler::Node::toDevice(HsvSampler *parent,bool raw)const
{
  int ww1=parent->width()-1;
  int hh1=parent->height()-1;
  if(parent->pcircular){
    double R=MyMath::min<double>(ww1*0.5,hh1*0.5);
    
    double phi=xx(parent,raw)*2*Pi;
    double r  =yy(parent)*(R-parent->Rmin)+parent->Rmin;
    return QPointF(r*cos(phi)+ww1*0.5,r*sin(phi)+hh1*0.5);
  }else{
    double x;
    if(parent->pmode[0]==HsvSampler::Hue){
      x=(xx(parent,raw)-parent->nodelist->imin)
	/(parent->nodelist->imax-parent->nodelist->imin+1);
    }else{
      x=xx(parent,raw);
    }
    return QPointF(x*ww1,yy(parent)*hh1);
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
double HsvSampler::Node::xx(HsvSampler *parent,bool raw)const
{
  if(parent->pmode[0]==HsvSampler::Hue){
    if(raw){
      return phue/360.+phueinterval;
    }else{
      return phue/360.;
    }
  }
  if(parent->pmode[0]==HsvSampler::Value){
    return pval/255.;
  }
  if(parent->pmode[0]==HsvSampler::Saturation){
    return psat/255.;
  }
  if(parent->pmode[0]==HsvSampler::Parameter){
    return ps;
  }
  if(parent->pmode[0]==HsvSampler::None){
    return 0.5;
  }
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
double HsvSampler::Node::yy(HsvSampler *parent,bool raw)const
{
  if(parent->pmode[1]==HsvSampler::Saturation){
    return psat/255.;
  }
  if(parent->pmode[1]==HsvSampler::Hue){
    if(raw){
      return phue/360.+phueinterval;
    }else{
      return phue/360.;
    }
  }
  if(parent->pmode[1]==HsvSampler::Value){
    return pval/255.;
  }
  if(parent->pmode[1]==HsvSampler::Parameter){
    return ps;
  }
  return 0;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
HsvSampler::Node HsvSampler::Node
::interpolateNodes(const Node &n1,const Node &n2,double s,bool inspherical)
{
  if(inspherical){
    double tol=1.e-3;
    
    double phi1=n1.rawHue()/180.*Pi;
    double phi2=n2.rawHue()/180.*Pi;
    double r1=n1.psat/255.;
    double r2=n2.psat/255.;

    //myDebug()<<"phi12s"<<n1.rawHue()<<n2.rawHue()<<s;
    double x1=r1*cos(phi1),y1=r1*sin(phi1);
    double x2=r2*cos(phi2),y2=r2*sin(phi2);
    //myDebug()<<"v1="<<x1<<y1;
    //myDebug()<<"v2="<<x2<<y2;

    double x=(x2-x1)*s+x1;
    double y=(y2-y1)*s+y1;
    //myDebug()<<"v ="<<x<<y;
    double phi=atan2(y,x)/Pi*180;
    //myDebug()<<"phi0"<<phi;
    while(phi<phi1)phi+=360.;
    while(phi>phi2)phi-=360.;
    
    if(fabs(s)<tol && fabs(phi-phi2) <tol )phi=phi1;
    if(fabs(1.-s)<tol && fabs(phi-phi1) <tol )phi=phi2;
    

    double r=sqrt(x*x+y*y)*255.;
    //myDebug()<<"phi"<<phi<<"r"<<r;


    int rawHue=lRound(phi);
    int sat=MyMath::max<long>(0,MyMath::min<long>(255,lRound(r)));
    int val=MyMath::max<long>(0,MyMath::min<long>(255,lRound((n2.pval-n1.pval)*s+n1.pval)));
    return Node(s,rawHue,sat,val);
  }
  
  int rawhue=lRound((n2.rawHue()-n1.rawHue())*s+n1.rawHue());

  int sat=MyMath::max<long>(0,MyMath::min<long>(255,lRound((n2.psat-n1.psat)*s+n1.psat)));
  int val=MyMath::max<long>(0,MyMath::min<long>(255,lRound((n2.pval-n1.pval)*s+n1.pval)));
    
  return Node(s,rawhue,sat,val);
}


/*EOF*/ 
