/*************************************************************************
 * @(#)MeshSymDlg.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   16 Aug 2010
 *************************************************************************
 */

#include <QMessageBox>
#include "fsymmetry.h"
#include "MeshSymDlg.hh"
#include "MeshSymmetry.hh"
#include "Valid.hh"
#include "Config.hh"
#include "MyDebug.hh"
#include "qt5compat.hh"



/* ***********************************************************************
 *
 * ***********************************************************************
 */
MeshSymDlg::MeshSymDlg(QWidget *parent ,QWidget *base,const char *name)
  : BrainDialog(parent,base,name,false)
{

  setupUi(this);
  setWindowTitle(IDS_APP_NAME + ": Hand made mesh symmetry settings");

  
  lesC[0][0]=leC00;
  lesC[0][1]=leC01;
  lesC[0][2]=leC02;
  
  lesC[1][0]=leC10;
  lesC[1][1]=leC11;
  lesC[1][2]=leC12;

  lesC[2][0]=leC20;
  lesC[2][1]=leC21;
  lesC[2][2]=leC22;
  

  lesOp[0][0]=leOp00;
  lesOp[0][1]=leOp01;
  lesOp[0][2]=leOp02;
  
  lesOp[1][0]=leOp10;
  lesOp[1][1]=leOp11;
  lesOp[1][2]=leOp12;

  lesOp[2][0]=leOp20;
  lesOp[2][1]=leOp21;
  lesOp[2][2]=leOp22;
  
  
  for(int r=0;r<3;r++){
    for(int c=0;c<3;c++){
      new RatValidator(this,lesC[r][c],QString().sprintf("C%1d%1d",r,c)
		       ,-1.e10,1.e10);
    }
  }
  changed=false;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void MeshSymDlg::setData(const MeshSymmetry * meshsym)
{
  
  long csystem=SYM_ORTHORHOMBIC;
  cmbBravais->clear();
  for(int i=0;i<numberofbravaistypesofcryssystem(csystem);i++){
    int btype=bravaistypesofcryssystem(csystem,i);
    cmbBravais->insertItem(i,QString()
			   .sprintf("%-12s",
				    bravaistypes_longname(btype)),QVariant(btype));
  }
  long cs=SYM_ORTHORHOMBIC;
  long bt=meshsym->userBravais();
  myDebug() << "bravais"<< bt;
  for(int i=0;i<numberofbravaistypesofcryssystem(cs);i++){
    int btype=bravaistypesofcryssystem(cs,i);
    if(btype==bt){
      cmbBravais->setCurrentIndex(i);
    }
  }
  
  for(int r=0;r<3;r++){
    for(int c=0;c<3;c++){
      t_C[r][c]=QString().setNum(meshsym->convCell()(c,r),'g',15);
      lesC[r][c]->setText(t_C[r][c]);
    }
  }

  for(int ig=0;ig<meshsym->userNumOps();ig++){
    QString op[3][3];
    const double *opmat=meshsym->userOpmat(ig);
    for(int r=0;r<3;r++){
      for(int c=0;c<3;c++){
	double d=opmat[r+3*c];
	QString t=QString().setNum(d,'g',15);
	//todo record changes t_Op[r][c]=t;
	lesOp[r][c]->setText(t);
	op[r][c]=t;
      }
    }
    SymOp sop=SymOp(meshsym->userOpSymbol(ig),op);
    symops.insert(ig,sop);
    liwiOps->insertItem(ig,sop.printQ());
  }
  liwiOps->setCurrentRow(0);

  

  
  t_handmade=meshsym->userDefinedSymmetry();
  cbHandMade->setChecked(t_handmade);

  t_hex=meshsym->hexagonal();
  cbHex->setChecked(t_hex);

  t_rtg=meshsym->rhombohedral();
  cbRTG->setChecked(t_rtg);

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void MeshSymDlg::getData(MeshSymmetry * meshsym)
{
  QString t;
  
  bool chn=0;

  Matrix3x3 cell;
  for(int c=0;c<3;c++){
    String s[3];
    for(int r=0;r<3;r++){
      t=lesC[c][r]->text();
      chn|=( t != t_C[c][r] );
      IntValidator::stripAllWhiteSpace(t);
      s[r]=t.toAscii().constData();
    }
    cell.setColumn(c,Vec3(s));
  }
  myDebug() << cell.printS()();
  meshsym->setConvCell(cell);

  double opmat[9];
  meshsym->setUserNumOps(liwiOps->count());
  for(int ig=0;ig<liwiOps->count();ig++){
    const SymOp &sop=symops.at(ig);
    myDebug() << sop.printQ();
    for(int r=0;r<3;r++){
      for(int c=0;c<3;c++){
	t=sop.op[c][r].simplified();
	opmat[r+3*c]=String(t.toAscii().constData()).convertToFortranReal();
      }
    }
    meshsym->setUserOpSymbol(ig,sop.symbol);
    meshsym->setUserOpmat(ig,opmat);
  }

  bool b=cbHandMade->isChecked();
  chn |= ( b != t_handmade);
  meshsym->setUserDefinedSymmetry(b);

  b=cbHex->isChecked();
  chn |= ( b != t_hex);
  meshsym->setHexagonal(b);

  b=cbRTG->isChecked();
  chn |= ( b != t_rtg);
  meshsym->setRTG(b);

  long bt=cmbBravais->itemData(cmbBravais->currentIndex()
			       ,Qt::UserRole).toInt();

  meshsym->setUserBravais(bt);

  meshsym->setUserDefinedStuffChanged(chn||changed);

}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void MeshSymDlg::accept()
{
  QString txt="";

  for(int r=0;r<3;r++){
    for(int c=0;c<3;c++){
      ((RatValidator*)lesC[r][c]->validator())->check(txt);
    }
  }      

  if(txt.length()){
    QMessageBox::warning ( this,IDS_APP_NAME, txt, 1, 0,0 ) ;
  }else{
    QDialog::accept();
  }  

}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void MeshSymDlg::on_pbAdd_clicked()
{
  myDebug() << "MeshSymDlg::on_pbAdd_clicked()";
  myDebug() << liwiOps->currentRow();
  int currentrow=liwiOps->currentRow();
  int row=currentrow+1;
  if(currentrow<0) row=liwiOps->count();
  QString s=leSymbol->text();
  s=s.simplified();
  if(s.isEmpty()){
    s="noname";
  }


  QString op[3][3];
  for(int c=0;c<3;c++){
    for(int r=0;r<3;r++){
      QString t=lesOp[r][c]->text();
      if(t.isEmpty()){
	if(c==r){
	  t="1";
	}else{
	  t="0";
	}
      }
      t=t.simplified();
      op[r][c]=t;
    }
  }

  
  SymOp sop=SymOp(s,op);
  symops.insert(row,sop);
  

  liwiOps->insertItem(row,sop.printQ());
  
  //on_liwiOps_currentRowChanged(row);
  liwiOps->setCurrentRow(row);
  
//   foreach(SymOp sop,symops){
//     myDebug()<< "symop "<<sop.symbol;
//     myDebug()<< "symop "<<sop.op[0][0]<<sop.op[0][1]<<sop.op[0][2];
//     myDebug()<< "symop "<<sop.op[1][0]<<sop.op[1][1]<<sop.op[1][2];
//     myDebug()<< "symop "<<sop.op[2][0]<<sop.op[2][1]<<sop.op[2][2];
//   }
  changed=true;

}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void MeshSymDlg::on_pbDelete_clicked()
{
  myDebug() <<"MeshSymDlg::on_pbDelete_clicked()";
  myDebug() << liwiOps->currentRow();
  int currentrow=liwiOps->currentRow();
  if(currentrow<0 && liwiOps->count()){
    QMessageBox::information ( this,IDS_APP_NAME
			   , "select a list item for deletion", 1, 0,0 ) ;
    return;
  }
  liwiOps->takeItem(currentrow);

  symops.removeAt(currentrow);

//   foreach(SymOp sop,symops){
//     myDebug()<< "symop "<<sop.symbol;
//     myDebug()<< "symop "<<sop.op[0][0]<<sop.op[0][1]<<sop.op[0][2];
//     myDebug()<< "symop "<<sop.op[1][0]<<sop.op[1][1]<<sop.op[1][2];
//     myDebug()<< "symop "<<sop.op[2][0]<<sop.op[2][1]<<sop.op[2][2];
//   }
  changed=true;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void MeshSymDlg::on_pbTake_clicked()
{
  myDebug()<<"MeshSymDlg::on_pbTake_clicked()";
  int currentrow=liwiOps->currentRow();
  if(currentrow<0){
    QMessageBox::information ( this,IDS_APP_NAME
			   , "select a list item for changes", 1, 0,0 ) ;
    return;
  }
  myDebug() <<currentrow;
  if(currentrow>=0){
    symops[currentrow].symbol=leSymbol->text();
    for(int c=0;c<3;c++){
      for(int r=0;r<3;r++){
	QString t=lesOp[c][r]->text();
	if(t.isEmpty()){
	  if(c==r){
	    t="1";
	  }else{
	    t="0";
	  }
	}
	t=t.simplified();
	symops[currentrow].op[c][r]=t;
      }
    }
    liwiOps->item(currentrow)->setText(symops[currentrow].printQ());

    
  }
  changed=true;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void MeshSymDlg::on_liwiOps_currentRowChanged(int currentrow)
{
  if(currentrow<0){
    leSymbol->setText("");
    return;
  }
  leSymbol->setText(symops.at(currentrow).symbol);
  for(int c=0;c<3;c++){
    for(int r=0;r<3;r++){
      lesOp[r][c]->setText(symops.at(currentrow).op[r][c]);
    }
  }
}


/*EOF*/ 
