/*************************************************************************
 * @(#)ToolBox.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   02 May 2010
 *************************************************************************
 */

#include <QRadioButton>
#include <QToolTip>
#include <QToolButton>
#include <QDoubleSpinBox>
#include <QMenu>
#include <QSignalMapper>
#include "AntiAliasing.hh"
#include "ControlBase.hh"
#include "ToolBox.hh"
#include "MyDebug.hh"
#include "DocBase.hh"
#include "STDoc.hh"
#include "Symmetry.hh"
#include "Boundary.hh"
#include "OGLView.hh"
#include "Axes.hh"
#include "Polyhedron.hh"
#include "DisplayCells.hh"
#include "STR.hh"
#include "AnnotationDlg.hh"
#include "Settings.hh"


const int CAM_SAVE=0;
const int CAM_LOAD=1;


/* ***********************************************************************
 *
 * ***********************************************************************
 */
ToolBox::ToolBox(ControlBase *parent,DocBase *d)
  :QDockWidget(0),prbPersp(0),prbOrtho(0),ptbResetCamera(0),ptbCamSaveLoad(0)
  ,ptbCart(0),ptbConv(0),ptbPrim(0)
  ,ptbXYView(0)
  ,ptbYXView(0)
  ,ptbXZView(0)
  ,ptbZXView(0)
  ,ptbYZView(0)
  ,ptbZYView(0)
  ,ptbXYZView(0)
  ,ptbWireFrame(0)
  ,ptbBoundOnRotate(0)
  ,ptbAA(0)
  ,ptbBoundCell(0)
  ,ptbPrimCell(0)
  ,ptbConvCell(0)
  ,pdspbxEye(0)
  ,muted(true)
  ,pparent(parent),pdoc(d)
  ,resetKeyAction(0)
  ,resetaction(0),centeraction(0),rescaleaction(0)
  ,camsaveaction(0),camloadaction(0)
  ,signalMapper(0)
  ,camSignalMapper(0)
  ,showhideAxesAction(0)
  ,showhideCellsAction(0)
  ,showHideAxesMenu(0)
  ,showHideCellsMenu(0)
  ,resetViewMenu(0)
  ,camSaveLoadMenu(0)
{
  //this->resize(40,40);
  sizePolicy().setHorizontalStretch(1);

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
ToolBox::~ToolBox()
{
  if(camsaveaction) delete camsaveaction;
  if(camloadaction) delete camloadaction;
  
  if(resetaction) delete resetaction;
  if(centeraction) delete centeraction;
  if(rescaleaction) delete rescaleaction;
  if(signalMapper) delete signalMapper;
  if(camSignalMapper) delete camSignalMapper;
  if(resetKeyAction) delete resetKeyAction;

  if(resetViewMenu) delete resetViewMenu;
  if(camSaveLoadMenu) delete camSaveLoadMenu;

  foreach (QObject *o, axesMenuActions) delete o;
  foreach (QObject *o, cellMenuActions) delete o;


}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::init()
{
  connect(prbPersp,SIGNAL(toggled(bool)),this,SLOT(setPerspective(bool)));
  connect(prbOrtho,SIGNAL(toggled(bool)),this,SLOT(setPerspective(bool)));

  connect(ptbCart,SIGNAL(toggled(bool)),this,SLOT(ontbAxesToggled(bool)));
  connect(ptbConv,SIGNAL(toggled(bool)),this,SLOT(ontbAxesToggled(bool)));
  connect(ptbPrim,SIGNAL(toggled(bool)),this,SLOT(ontbAxesToggled(bool)));


  connect(ptbBoundCell,SIGNAL(toggled(bool))
	  ,this,SLOT(ontbDisplayCellsToggled(bool)));
  connect(ptbConvCell,SIGNAL(toggled(bool))
	  ,this,SLOT(ontbDisplayCellsToggled(bool)));
  connect(ptbPrimCell,SIGNAL(toggled(bool))
	  ,this,SLOT(ontbDisplayCellsToggled(bool)));

  connect(ptbXYView,SIGNAL(clicked()),this,SLOT(ontbXYViewclicked()));
  connect(ptbYXView,SIGNAL(clicked()),this,SLOT(ontbYXViewclicked()));
  connect(ptbXZView,SIGNAL(clicked()),this,SLOT(ontbXZViewclicked()));
  connect(ptbZXView,SIGNAL(clicked()),this,SLOT(ontbZXViewclicked()));
  connect(ptbYZView,SIGNAL(clicked()),this,SLOT(ontbYZViewclicked()));
  connect(ptbZYView,SIGNAL(clicked()),this,SLOT(ontbZYViewclicked()));
  connect(ptbXYZView,SIGNAL(clicked()),this,SLOT(ontbXYZViewclicked()));

  connect(ptbWireFrame,SIGNAL(toggled(bool))
	  ,this,SLOT(ontbWireFrameToggled(bool)));
  connect(ptbBoundOnRotate,SIGNAL(toggled(bool))
	  ,this,SLOT(onTbBoundOnRotateToggled(bool)));

  connect(ptbAA,SIGNAL(clicked())
	  ,this,SLOT(onTbAAClicked()));

  connect(pdspbxEye,SIGNAL(valueChanged(double))
	  ,this,SLOT(ondspbxEyeValueChanged(double)));



  QAction *a;
  QMenu *m=resetViewMenu=new QMenu(ptbResetCamera);
  a=resetaction=m->addAction("&Reset");  
  const char *tip="Reset/Center/Re-scale camera. (Ctrl-R)";
  a->setToolTip(tip);
  a->setStatusTip(tip);
  a=centeraction=m->addAction("&Center");  
  a->setToolTip(tip);
  a->setStatusTip(tip);
  a=rescaleaction=m->addAction("Re-&Scale");  
  a->setToolTip(tip);
  a->setStatusTip(tip);


  a=resetKeyAction=new QAction(this);
  a->setShortcut(QKeySequence(Qt::CTRL+Qt::Key_R));
  connect(a,SIGNAL(triggered()),this,SLOT(onResetKeyAction()));
  MASSERT("Oops",pparent);
  pparent->addAction(a);


  signalMapper = new QSignalMapper(this);
  signalMapper->setMapping(resetaction, OGLView::Reset);
  signalMapper->setMapping(centeraction, OGLView::Center);
  signalMapper->setMapping(rescaleaction, OGLView::Rescale);

  ptbResetCamera->setMenu(m);
  ptbResetCamera->setDefaultAction(resetaction);

  connect(resetaction,SIGNAL(triggered()),signalMapper,SLOT(map()));
  connect(centeraction,SIGNAL(triggered()),signalMapper,SLOT(map()));
  connect(rescaleaction,SIGNAL(triggered()),signalMapper,SLOT(map()));

  connect(signalMapper,SIGNAL(mapped(int)),
	  this,SLOT(onTbResetCameraClicked(int)));


  showHideAxesMenu=m=new QMenu(this);
  m->setTitle("Show axis");
  a=m->addAction("Cartesian a&xis");
  const char *tip1="Show/hide axes.";
  a->setToolTip(tip1);
  a->setStatusTip(tip1);
  a->setCheckable(true);
  connect(a,SIGNAL(triggered()),ptbCart,SLOT(toggle()));
  axesMenuActions.insert(ptbCart,a);

  a=m->addAction("&Conventional axis");
  a->setToolTip(tip1);
  a->setStatusTip(tip1);
  a->setCheckable(true);
  connect(a,SIGNAL(triggered()),ptbConv,SLOT(toggle()));
  axesMenuActions.insert(ptbConv,a);

  a=m->addAction("&Primitive axis");
  a->setToolTip(tip1);
  a->setStatusTip(tip1);
  a->setCheckable(true);
  connect(a,SIGNAL(triggered()),ptbPrim,SLOT(toggle()));
  axesMenuActions.insert(ptbPrim,a);


  
  a=showhideAxesAction=new QAction(this);
  a->setShortcut(QKeySequence(Qt::Key_A));
  connect(a,SIGNAL(triggered()),this,SLOT(onShowhideAxesAction()));
  pparent->addAction(a);




  {
    m=new QMenu(ptbCamSaveLoad);
    a=camsaveaction=m->addAction("Save");  
    const char *tip2="Save/Load camera";
    a->setToolTip(tip2);
    a->setStatusTip(tip2);
    a=camloadaction=m->addAction("Load");  
    a->setToolTip(tip2);
    a->setStatusTip(tip2);
    
    
    
    
    camSignalMapper = new QSignalMapper(this);
    camSignalMapper->setMapping(camsaveaction, CAM_SAVE);
    camSignalMapper->setMapping(camloadaction, CAM_LOAD);
    
    ptbCamSaveLoad->setMenu(m);
    ptbCamSaveLoad->setDefaultAction(camsaveaction);
    
    connect(camsaveaction,SIGNAL(triggered()),camSignalMapper,SLOT(map()));
    connect(camloadaction,SIGNAL(triggered()),camSignalMapper,SLOT(map()));
    
    connect(camSignalMapper,SIGNAL(mapped(int)),
	    this,SLOT(onTbCamSaveLoadCameraClicked(int)));
    
  }



  {


    a=showhideCellsAction=new QAction(this);
    a->setShortcut(QKeySequence(Qt::Key_B));
    connect(a,SIGNAL(triggered()),this,SLOT(onShowhideCellsAction()));
    pparent->addAction(a);

    showHideCellsMenu=m=new QMenu(this);
    m->setTitle("Show cells");
    a=m->addAction("Main &boundary");
    const char *tip1="Show/hide cell.";
    a->setToolTip(tip1);
    a->setStatusTip(tip1);
    a->setCheckable(true);
    connect(a,SIGNAL(triggered()),ptbBoundCell,SLOT(toggle()));
    cellMenuActions.insert(ptbBoundCell,a);

    a=m->addAction("&Conventional cell");
    a->setToolTip(tip1);
    a->setStatusTip(tip1);
    a->setCheckable(true);
    connect(a,SIGNAL(triggered()),ptbConvCell,SLOT(toggle()));
    cellMenuActions.insert(ptbConvCell,a);
    
    a=m->addAction("&Primitive cell");
    a->setToolTip(tip1);
    a->setStatusTip(tip1);
    a->setCheckable(true);
    connect(a,SIGNAL(triggered()),ptbPrimCell,SLOT(toggle()));
    cellMenuActions.insert(ptbPrimCell,a);
    
  }
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::onTbResetCameraClicked(int mode)
{
  //myDebug("STToolBox::onTbResetCameraClicked");
  if(muted) return;
  if(!pparent) return;
  if(!pparent->oglView()) return;
  if(!pdoc) return;


  pdoc->setModified();

  if(!pdoc->allcalced) 
    pparent->updateData();
    

  ptbResetCamera->
    setDefaultAction(qobject_cast<QAction*>(signalMapper->mapping(mode)));


  pparent->oglView()->resetPanZoom((OGLView::CameraResetMode)mode,true);
  
 
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::onResetKeyAction()
{

  QPoint pos;
  if(isVisible()){
    pos=ptbResetCamera->mapToGlobal(QPoint(0,ptbResetCamera->height()));
  }else{
    pos=QCursor::pos();
  }

  ptbResetCamera->menu()->popup(pos);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::setPerspective(bool onoff)
{
  //myDebug("setPerspective");
  //myDebug() << onoff << prbPersp->isChecked();
  //myDebug() <<pparent << pdoc << muted <<onoff; 
  if(muted) return;
  if(!pparent) return;
  if(!pdoc) return;

  if(!onoff)return;
    
  if(prbPersp->isChecked()){
    pdoc->view().setPerspective(true);
  }else{
    pdoc->view().setPerspective(false);
  }
  pdoc->setModified();


  
  if(!pdoc->allcalced) 
    pparent->updateData();
  else{

    pparent->oglView()->doRendering();
    


  }

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::onShowhideAxesAction()
{
  //  myDebug() << "STToolBox::onShowhideAxesAction";
  
  
  QPoint pos;
  if(isVisible()){
    pos=ptbCart->mapToGlobal(QPoint(0,ptbCart->height()));
  }else{
    pos=QCursor::pos();
  }
  
  showHideAxesMenu->popup(pos);
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::onShowhideCellsAction()
{
  //  myDebug() << "STToolBox::onShowhideCellsAction";
  
  
  QPoint pos;
  if(isVisible()){
    pos=ptbBoundCell->mapToGlobal(QPoint(0,ptbBoundCell->height()));
  }else{
    pos=QCursor::pos();
  }
  
  showHideCellsMenu->popup(pos);
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::updateCellButtons()
{
  STDoc *stdoc=dynamic_cast<STDoc*>(pdoc);
  //bool oldmuted=muted;
  //  muted=true;
  bool iscrystal=true;
  if(stdoc){
    iscrystal=(stdoc->symmetry->dataStructureType()==Xf::STCrystal);
  }
  
  foreach(DisplayCell *dc,pdoc->displayCells()->cells()){
    if(dc){
      bool on=dc->boundary()->polyhedron()->properties().isVisible();
      bool enab=false;
      QToolButton *tb=0;
      if(dc->type()==Xf::DCTMainBoundary){
	tb=ptbBoundCell;
	enab=true;
      }else if(dc->type()==Xf::DCTConventional){
	tb=ptbConvCell;
      }else if(dc->type()==Xf::DCTPrimitive){
	tb=ptbPrimCell;
      }
      if(tb){
	tb->setChecked(on);
	tb->setEnabled(enab|iscrystal);
        cellMenuActions[tb]->setChecked(on);
        cellMenuActions[tb]->setEnabled(enab|iscrystal);
      }
    }
  }
  //  muted=oldmuted;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::updateAxesButtons()
{
  //myDebug() << "ToolBox::updateAxesButtons";
  //myDebug() << doc->symmetry->dataStructureType();
  
  STDoc *stdoc=dynamic_cast<STDoc*>(pdoc);
  
  bool iscrystal=true;
  if(stdoc){
    iscrystal=(stdoc->symmetry->dataStructureType()==Xf::STCrystal);
  }

  ptbCart->setChecked(pdoc->axes()->findByType(Xf::ATCartesian)->isVisible());
  axesMenuActions[ptbCart]->setChecked(pdoc->axes()->findByType(Xf::ATCartesian)->isVisible());
  if(iscrystal){
    ptbConv->setEnabled(true);
    ptbPrim->setEnabled(true);
    ptbConv->setChecked(pdoc->axes()->findByType(Xf::ATConventional)->isVisible());
    ptbPrim->setChecked(pdoc->axes()->findByType(Xf::ATPrimitive)->isVisible());

    axesMenuActions[ptbConv]->setEnabled(true);
    axesMenuActions[ptbPrim]->setEnabled(true);
    axesMenuActions[ptbConv]->setChecked(pdoc->axes()->findByType(Xf::ATConventional)->isVisible());
    axesMenuActions[ptbPrim]->setChecked(pdoc->axes()->findByType(Xf::ATPrimitive)->isVisible());
  }else{
    ptbConv->setEnabled(false);
    ptbPrim->setEnabled(false);
    axesMenuActions[ptbConv]->setEnabled(false);
    axesMenuActions[ptbPrim]->setEnabled(false);
  }
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::ontbAxesToggled(bool onoff)
{
//  myDebug("on_tbConv_toggled");
  if(muted) return;
  if(!pparent) return;
  if(!pdoc) return;

  Xf::AxesType type;
  if(sender()==ptbConv){
    type=Xf::ATConventional;
  }else if(sender()==ptbCart){
    type=Xf::ATCartesian;
  }else if(sender()==ptbPrim){
    type=Xf::ATPrimitive;
  }else{
    return;
  }

  Axes *ax=pdoc->axes()->findByType(type);

  ax->setVisible(onoff);
  pdoc->setModified();

  if(pparent->annotationDlg())
    pparent->annotationDlg()->refreshAxesActive(type);

  pparent->oglView()->setVisible(ax->id(),onoff);

  if(!pdoc->allcalced) 
    pparent->updateData();
  else{

    pparent->oglView()->doRendering();
  }

  axesMenuActions[sender()]->setChecked(onoff);

}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::ontbDisplayCellsToggled(bool on)
{
  if(muted) return;
  if(!pparent) return;
  if(!pdoc) return;
  if(!ptbConvCell) return;

  Xf::DisplayCellType type;
  if(sender()==ptbBoundCell){
    type=Xf::DCTMainBoundary;
  }else if(sender()==ptbConvCell){
    type=Xf::DCTConventional;
  }else if(sender()==ptbPrimCell){
    type=Xf::DCTPrimitive;
  }else
    return;
  DisplayCell *dc=pdoc->displayCells()->findByType(type);
  if(dc){
    dc->boundary()->polyhedron()->properties().setVisible(on);
    pdoc->setModified();
    
    if(pparent->annotationDlg())
      pparent->annotationDlg()->refreshDisplayCellsActive(type,on);

    pparent->oglView()->setVisible(dc->id(),on);
    if(!pdoc->allcalced) 
      pparent->updateData();
    else{
      
      pparent->oglView()->doRendering();
    }

    cellMenuActions[sender()]->setChecked(on);
    
  }
}




/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::ontbXYViewclicked( )
{
  //myDebug("ontbXYViewclicked");
  
  if(muted) return;
  if(!pparent) return;
  if(!pdoc) return;

  pdoc->setModified();
  
  pdoc->view().setXYView();

  if(!pdoc->allcalced) {
    pparent->updateData();
  
  }else{
    
    pparent->oglView()->doRendering();
  } 

 
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::ontbXZViewclicked( )
{
  //myDebug("ontbXZViewclicked");
  
  if(muted) return;
  if(!pparent) return;
  if(!pdoc) return;

  
  pdoc->setModified();
  
  pdoc->view().setXZView();

  
  if(!pdoc->allcalced) {
    pparent->updateData();
  
  }else{
    pparent->oglView()->doRendering();
  }
  

  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::ontbYXViewclicked( )
{
  //myDebug("ontbYXViewclicked");
  
  if(muted) return;
  if(!pparent) return;
  if(!pdoc) return;

  pdoc->setModified();
  
  pdoc->view().setYXView();

  
  if(!pdoc->allcalced) {
    pparent->updateData();
  
  
  }else{

    pparent->oglView()->doRendering();
  }
  
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::ontbZXViewclicked( )
{
  //myDebug("ontbZXViewclicked");
  
  if(muted) return;
  if(!pparent) return;
  if(!pdoc) return;

  
  pdoc->setModified();
  
  pdoc->view().setZXView();

  if(!pdoc->allcalced) {
    pparent->updateData();
    
  }else{
  
    pparent->oglView()->doRendering();
  }


}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::ontbYZViewclicked( )
{
  //myDebug("ontbYZViewclicked");
  
  if(muted) return;
  if(!pparent) return;
  if(!pdoc) return;

  pdoc->setModified();

  pdoc->view().setYZView();

  
  
  if(!pdoc->allcalced) {
    pparent->updateData();
  
  }else{
  
    pparent->oglView()->doRendering();
  }
  


}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::ontbZYViewclicked( )
{
  //myDebug("ontbZYViewclicked");
  
  if(muted) return;
  if(!pparent) return;
  if(!pdoc) return;

  pdoc->setModified();
  
  pdoc->view().setZYView();

  if(!pdoc->allcalced) {
    pparent->updateData();
  
  }else{
    pparent->oglView()->doRendering();
  }

  
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::ontbXYZViewclicked( )
{
  //myDebug("ontbXYZViewclicked");
  
  if(muted) return;
  if(!pparent) return;
  if(!pdoc) return;

  pdoc->setModified();
  
  pdoc->view().setDefaultXYZRot();

  if(!pdoc->allcalced) {
    pparent->updateData();
  
  }else{
    pparent->oglView()->doRendering();
  }

  
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::ontbWireFrameToggled(bool onoff)
{
  //myDebug("ontbWireFrametoggled %d\n",onoff);
  if(!pparent) return;
  Settings set("application");
  set.set("graphics/wireFrameOnMove",onoff);
  pparent->oglView()->setWireFrameOnMove(onoff);
  
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::onTbBoundOnRotateToggled(bool on )
{
  if(!pparent) return;
  Settings set("application");
  set.set("graphics/reducedModelOnMove",on);
  pparent->oglView()->setReducedModelOnMove(on);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::onTbAAClicked( )
{
  //myDebug() << "ToolBox::onTbAAClicked start";
  if(!pparent) return;
  bool on=ptbAA->isChecked();
  antiAliasing.setAntiAliasing(on);
  //antiAliasing.setAntiAliasingForPrintOnly(false);
  antiAliasing.writeSettings();

  QPoint pos=pparent->oglView()->mapToGlobal(QPoint(0,0));

  OGLView::AntiAliasingMode aamode=pparent->oglView()->currentAntiAliasingMode();
  
  QString text;
  if(on){
    if(aamode==OGLView::Hardware){
      text="Hardware antialiasing";
    }else if(aamode==OGLView::Software){
      text="Software antialiasing";
    }else if(aamode==OGLView::None){
      text="No antialiasing due to missing hardware features";
    }
    if(antiAliasing.forPrintingOnly()){
      text+="\nAA is currently set for printing\nonly in preferences.";
    }
    QToolTip::showText(pos,text,pparent->oglView());
  }

  //if(muted) return;
  //myDebug() << "ToolBox::onTbAAToggled after muted";

  if(!pdoc->allcalced) 
    pparent->updateData();
  else{

    pparent->oglView()->doRendering();
    


  }
  //myDebug() << "ToolBox::onTbAAToggled end";

  //  pparent->oglView()->setReducedModelOnMove(on);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::ondspbxEyeValueChanged(double value)
{
  //myDebug()<<"ToolBox::ondspbxEyeValueChanged"<<value;
  if(muted) return;
  if(!pparent) return;
  if(!pdoc) return;

    
  pdoc->view().setTheta(value);
  pdoc->setModified();


  
  if(!pdoc->allcalced) 
    pparent->updateData();
  else{
    pparent->oglView()->doRendering();
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::updateToolBox()
{
  //myDebug() << "ToolBox::updateToolBox";

  if(!pparent) return;
  if(!pdoc) return;

  muted=true;

  if(prbPersp)
    prbPersp->setChecked(pdoc->view().isPerspective());
  if(prbOrtho)
    prbOrtho->setChecked(!pdoc->view().isPerspective());

  if(pdspbxEye)
    pdspbxEye->setValue(pdoc->view().theta());

  

//   if(ptbBoundCell) {
//     DisplayCell *dc=pdoc->displayCells()->findByType(Xf::DCTMainBoundary);
//     if(dc)
//       ptbBoundCell->setChecked(dc->boundary()->polyhedron()
// 			       ->properties().isVisible());
//   }

//   if(ptbPrimCell){
//     DisplayCell *dc=pdoc->displayCells()->findByType(Xf::DCTPrimitive);
//     if(dc)
//       ptbPrimCell->setChecked(dc->boundary()->polyhedron()
// 			      ->properties().isVisible());
//   }
//   if(ptbConvCell){
//     DisplayCell *dc=pdoc->displayCells()->findByType(Xf::DCTConventional);
//     ptbConvCell->setChecked(dc->boundary()->polyhedron()
// 			    ->properties().isVisible());
//   }

  updateAxesButtons();
  updateCellButtons();


  //myDebug() << "muted "<<muted;



  muted=false;

  Settings set("application");
  if(ptbWireFrame)
    ptbWireFrame->setChecked(set.getBool("graphics/wireFrameOnMove",false));
  if(ptbBoundOnRotate)
    ptbBoundOnRotate->setChecked(set.getBool("graphics/reducedModelOnMove",false));

  if(ptbAA)
    ptbAA->setChecked(antiAliasing.antiAliasing());

  muted=true;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void ToolBox::onTbCamSaveLoadCameraClicked(int mode)
{
  //myDebug("STToolBox::onTbCamSaveLoadCameraClicked");
  if(muted) return;
  if(!pparent) return;
  if(!pparent->oglView()) return;
  if(!pdoc) return;


  pdoc->setModified();

    

  ptbCamSaveLoad->
    setDefaultAction(qobject_cast<QAction*>(camSignalMapper->mapping(mode)));


  
  if(mode==CAM_SAVE){
    pdoc->view().saveCameraToSettings();
    return;
  }else{
    pdoc->view().loadCameraFromSettings();
  }

  if(prbPersp)
    prbPersp->setChecked(pdoc->view().isPerspective());
  if(prbOrtho)
    prbOrtho->setChecked(!pdoc->view().isPerspective());


  if(!pdoc->allcalced) 
    pparent->updateData();
  else
    pparent->oglView()->doRendering();

 
}


/*EOF*/ 
