/*************************************************************************
 * @(#)UnfoldDlg.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   02 Jul 2012
 *************************************************************************
 */

#include <QMessageBox>
#include <QFileDialog>
#include "UnfoldDlg.hh"
#include "Symmetry.hh"
#include "Settings.hh"
#include "Assorted.hh"
#include "MWBase.hh"
#include "STR.hh"
#include "DocBase.hh"
#include "SReal.hh"
#include "Vec3.hh"
#include "Valid.hh"
#include "Config.hh"
#include "MyDebug.hh"

bool prot=false;


class SiteClassList{
private:
  
  // ================================================================
  class Site{
  public:
    Site(long isa,const String &el,const Vec3 &tauD)
      :pisa(isa),pel(el.lowerCase()),ptauD(tauD){};
    inline long isa()const{return pisa;};
    inline const String el()const{return pel;};
    inline const Vec3 &tauD()const{return ptauD;};
  private:
    long pisa;
    String pel;
    Vec3 ptauD;
  };
  // ================================================================

  // ================================================================
  class SiteClass{
  public:
  
    // --------------------------------------------------------------
    SiteClass(const Site&site)
    {
      list.append(site);
    }

    // --------------------------------------------------------------
    const String element()const
    {
      if(list.size()<1) return "??";
      return list.at(0).el().capitalized();
    }
    
    // --------------------------------------------------------------
    bool contains(const Site &site,const Matrix3x3 &NCrcell,double tol)
    {
      if(prot)myDebug()<<"\t\t\tSiteClass::contains Site"<<site.isa()<<site.tauD();
      
      for(long i=0;i<list.size();i++){
	if(prot)myDebug()<<"\t\t\t\tcheck site "<<i<<list.at(i).isa()
			 <<list.at(i).tauD();
	Vec3 x=(list.at(i).tauD()-site.tauD())*NCrcell;
	if(prot)myDebug()<<"\t\t\t\tdif"<<(list.at(i).tauD()-site.tauD());
	if(prot)myDebug()<<"\t\t\t\tx"<<x;
	if(site.el() != list.at(i).el()) continue;
	if(fabs(dRound(x[0])-x[0])>tol) continue;
	if(fabs(dRound(x[1])-x[1])>tol) continue;
	if(fabs(dRound(x[2])-x[2])>tol) continue;
	return true;
	
      }
      
      return false;
    }

    // --------------------------------------------------------------
    inline void append(const Site &site)
    {
      list.append(site);
    }

    // --------------------------------------------------------------
    String printS()
    {
      String ret;
      for(long i=0;i<list.size();i++){
	const Site &site=list.at(i);
	ret+=Sprintf("\tsite %ld: %ld %s\n",i+1,site.isa()+1,
		     site.tauD().printS()());
      }
      return ret;
    }

    // --------------------------------------------------------------
    String toUnfold()
    {
      String ret;
      for(long i=0;i<list.size();i++){
	const Site &site=list.at(i);
	ret+=Sprintf(" %ld",site.isa()+1);
      }
      return ret;
    }

    // --------------------------------------------------------------
    // --------------------------------------------------------------
    // --------------------------------------------------------------

  private:
    QList<Site> list;
  };
  // ================================================================

public:
  // --------------------------------------------------------------
  void append(long isa,const String &el,const Vec3 tauD,
	      const Matrix3x3 &NCrcell,double tol)
  {
    Site site(isa,el,tauD);
    if(prot)myDebug()<<"\tappend Site"<<site.isa()<<site.tauD();
    if(classes.size()<1){
      classes.append(SiteClass(site));
      return;
    }
    for(long i=0;i<classes.size();i++){
      SiteClass &siteclass=classes[i];
      if(prot)myDebug()<<"\t\tsiteclass "<<i<<"of"<<classes.size();
      
      if(siteclass.contains(site,NCrcell,tol)){
	if(prot)myDebug()<<"\t\tappend to siteclass "<<i;
	siteclass.append(site);
	return;
      }
    }
    if(prot)myDebug()<<"\t\tcreate new site class ";
    classes.append(SiteClass(site));
    return;
  }
 
    // --------------------------------------------------------------
  String printS()
  {
    String ret;
    for(long i=0;i<classes.size();i++){
      ret+=Sprintf("class %ld\n",i+1);
      ret+=classes[i].printS();
    }
    return ret;
  }

 
  // --------------------------------------------------------------
  String toUnfold()
  {
    String ret;
    String last="xx";
    for(long i=0;i<classes.size();i++){
      const String &el=classes[i].element();
      if(last !=el){
	ret+=Sprintf("# %s\n",el());
	last=el;
      }
      ret+=Sprintf("%ld    ",i+1);
      ret+=classes[i].toUnfold()+"\n";
    }
    return ret;
  }

 
private:
  QList<SiteClass> classes;
};


/* ***********************************************************************
 *
 * ***********************************************************************
 */
UnfoldDlg::UnfoldDlg(QWidget *parent,QWidget *base,const char *name)
  : ProtoDlg(parent,base,name)
{
  MWBase *mwbase=dynamic_cast<MWBase*>(base);
  if(mwbase){
    pstr=dynamic_cast<STR*>(mwbase->control());
  }

  setupUi(this);
  setUp(0,pbClose,0,pbHelp,"UnfoldEditor",": Unfold Editor");
  xyzvalid=new RatValidator(this,0,"",-1.e10,1.e10);
  //xyzvalid=new IntValidator(this,0,"",-100000,100000);


  // cmbUnits->insertItem(0,"large cell in small cell units","small");
  // cmbUnits->insertItem(1,"small cell in large cell units","large");

  unitHelper
    =new UCH(cmbUnits,
	     ValueArray<UCH::Type>() 
	     <<UCH::Type("small","large cell in small cell units")
	     <<UCH::Type("large","small cell in large cell units")
	     );
  
  unitHelper->populate();
  cbChange->setChecked(false);
  loadSettings();
  plTxEdList->setPlainText("#NCsite SCsites\n");
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
UnfoldDlg::~UnfoldDlg()
{
  delete unitHelper;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void UnfoldDlg::setData(const Symmetry *sym)
{
  t_sym=sym;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool UnfoldDlg::dataAcceptable()
{

  try{
    Matrix3x3 SCMat=readTwBasis();
    saveSettings(SCMat);
  }catch(QString &txt){
    return ProtoDlg::dataAcceptable(txt);
  }
  
  return true;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void UnfoldDlg::on_pbApplySCBasis_clicked()
{

  Matrix3x3 SCMat;
  try{
    SCMat=readTwBasis();
  }catch(QString &ex){
    QMessageBox::warning ( this,IDS_APP_NAME,ex, 1, 0,0 ) ;
    return;
  }
  
  saveSettings(SCMat);

  if(currentunits=="large"){
    SCMat=SCMat.inverse();
  }
  // SCMat produces the current larger cell out of the imagined
  // smaller cell
  if(prot){
    myDebug()<<"SCMat";
    myDebug()<<SCMat(0);
    myDebug()<<SCMat(1);
    myDebug()<<SCMat(2);
  }

  // the relation is largercell=smallercell*B
  // or              largercell*B^{-1}=smallercell
  Matrix3x3 Binv=SCMat.inverse();
//   myDebug()<<Binv(0);
//   myDebug()<<Binv(1);
//   myDebug()<<Binv(2);

  const Matrix3x3 &cell=t_sym->cell();

  if(prot){
    myDebug()<<"SCcell";
    myDebug()<<cell(0);
    myDebug()<<cell(1);
    myDebug()<<cell(2);
  }
  Matrix3x3 NCcell=cell*Binv;

  if(prot){  
    myDebug()<<"NCcell";
    myDebug()<<NCcell(0);
    myDebug()<<NCcell(1);
    myDebug()<<NCcell(2);
  }
  Matrix3x3 NCrcell=NCcell.inverse().transpose();
  if(prot){
    myDebug()<<"NCrcell";
    myDebug()<<NCrcell(0);
    myDebug()<<NCrcell(1);
    myDebug()<<NCrcell(2);
  }
  // now try to find unfolding 
  // site mod NCcell lready there: we haveit

  
  double tol=dspbxTolerance->value();
  long scnsite=t_sym->nsite();
  const ValueArray<Site> &scsites=t_sym->sites();

  SiteClassList siteclasslist;
  for(long isa=0;isa<scnsite;isa++){
    if(prot)myDebug()<<"SCsite"<<isa;
    siteclasslist.append(isa,scsites.at(isa).el(),
			 scsites.at(isa).tauD(),NCrcell,tol);
  }
  
  if(prot)  myDebug()<<"siteclasslist";
  //  siteclasslist.print();
  plTxEdList->setPlainText(("#NCsite SCsites\n"+siteclasslist.toUnfold())());


}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void UnfoldDlg::on_cmbUnits_currentIndexChanged(int index)
{
  Q_UNUSED(index);
  QString units=unitHelper->data();
  if(currentunits!=units){
    //myDebug()<<"current index"<<index<<units;
    
    try{
      Matrix3x3 mat=readTwBasis().inverse();
      if(cbChange->isChecked()){
	setTwBasis(mat);
      }
  }catch(QString &msg){
      currentunits=units;
      return;
    }
    
    currentunits=units;
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Matrix3x3 UnfoldDlg::readTwBasis()
{
  QString m,s,txt;
  Matrix3x3 mat;
  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      s=twBasis->item(i,j)->data(Qt::DisplayRole).toString();
      IntValidator::stripAllWhiteSpace(s);m="";
      long err=!xyzvalid->check(m,s);
      if(err)
	txt+=QString("element(%1,%2): %3").arg(i+1).arg(j+1).arg(m);
      else
	mat.data()[j+3*i]=SReal(s).D;
    }
  }
  if(txt.length()) throw txt;

  return mat;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void UnfoldDlg::setTwBasis(const Matrix3x3 &mat)
{
  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      twBasis->item(i,j)->setData(Qt::DisplayRole | Qt::EditRole,
				  QString(Sprintf("%g",mat(i)[j])()));
    }
  }    
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void UnfoldDlg::loadSettings()
{
  Settings set("./unfold.ini","Controls");
  currentunits=set.getString("units","small");
  unitHelper->setIndexFromData(currentunits);
  setTwBasis(set.getMatrix3x3("matrix",Matrix3x3::unitMatrix()));
  cbChange->setChecked(set.getBool("changeonunits",false));
  dspbxTolerance->setValue(set.getDouble("tolerance",1.0e-2));
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void UnfoldDlg::saveSettings(const Matrix3x3 &mat)
{
  Settings set("./unfold.ini","Controls");
  set.set("units",cmbUnits->itemData(cmbUnits->currentIndex()).toString());
  set.set("matrix",mat);
  set.set("changeonunits",cbChange->isChecked()>0);
  set.set("tolerance",dspbxTolerance->value());
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void UnfoldDlg::on_pbLoad_clicked()
{
  
  QString filename="=.unfold";
  QString fn;
  QString workdir=".";
  if(pstr){
    workdir=pstr->document()->currentDir()+"/";
  }
  filename=workdir+filename;


#ifdef Q_OS_MAC
  fn = QFileDialog::getOpenFileName(this,"Save File",filename);
#else
  fn = QFileDialog::getOpenFileName(this,"Save File",filename
				    , " FPLO files(=.*) ;; All files(*)"
				    ,0,0);
#endif
  if(fn.isEmpty()) return;


  String txt;
  FILE *FH=0;
  try{
    FH=openforread(String(fn));
    String line;
    while (getLine(FH,line)){
      txt+=line;
    }   
    fclose(FH);FH=0;
  }catch(Except &ex) {
    QMessageBox::warning(this,IDS_APP_NAME+ " " + FPLOMAINVERSION()
			 ,QString("While opening %1 for read!: %2").arg(fn)
			 .arg(ex.msg())
			 ,QMessageBox::Ok|QMessageBox::Default
			 );
  }
  if(FH) fclose(FH);
  plTxEdList->setPlainText(txt());
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void UnfoldDlg::on_pbSave_clicked()
{
  QString filename="=.unfold";
  QString fn;
  QString workdir=".";
  if(pstr){
    workdir=pstr->document()->currentDir()+"/";
  }
  filename=workdir+filename;

#ifdef Q_OS_MAC
  fn = QFileDialog::getSaveFileName(this,"Save File",filename);
#else
  fn = QFileDialog::getSaveFileName(this,"Save File",filename
				    , " FPLO files(=.*) ;; All files(*)"
				    ,0,0);
#endif
  if(fn.isEmpty()) return;

  FILE *FH=0;
  try{
    FH=openforwrite(String(fn));
    fprintf(FH,"%s\n",plTxEdList->toPlainText().toAscii().constData());
    fclose(FH);FH=0;
  }catch(Except &ex) {
    QMessageBox::warning(this,IDS_APP_NAME+ " " + FPLOMAINVERSION()
			 ,QString("While opening %1 for write!: %2").arg(fn)
			 .arg(ex.msg())
			 ,QMessageBox::Ok|QMessageBox::Default
			 );
  }

  if(FH) fclose(FH);
}

/*EOF*/ 
