/* @(#)Fractional.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _FRACTIONAL_HH
#define _FRACTIONAL_HH 1


#include "FString.hh"
#include "Except.hh"
#include "MyMath.hh"
#include "MyDebug.hh"

/* ***********************************************************************
 *
 * ***********************************************************************
 */
class Fractional;
MyDebug operator<<(MyDebug dbg,const Fractional &f);

class Fractional{
public:
  Fractional():pn(0),pm(1){};
  Fractional(const Fractional &other):pn(other.pn),pm(other.pm){};
  Fractional(long n):pn(n),pm(1){};
  Fractional(long n,long m):pn(n),pm(m){regularize();};
  Fractional &operator=(long n){pn=n;pm=1;return *this;};
  Fractional &operator=(const Fractional &other){
    pn=other.pn;pm=other.pm;return *this;};

  bool isInteger()const{return pm==1;};


  Fractional &regularize(){
    if(pm<0){pn=-pn;pm=-pm;} 
    long D=labs(MyMath::gcd(pn,pm));pn/=D;pm/=D;
    return *this;}

  Fractional operator-()const{ return Fractional(-pn,pm); }

  Fractional &inv(){long l=pn;pn=pm;pm=l;if(pm<0){pm=-pm;pn=-pn;}return *this;}

  inline Fractional &abs(){ 
    pn=labs(pn);
    pm=labs(pm);
    return *this;
  }

  Fractional &operator*=(const Fractional &other){
    pn*=other.pn;
    pm*=other.pm;
    regularize();
    return *this;
  }
  Fractional &operator*=(long s){
    long D=labs(MyMath::gcd(pm,s));
    pn*=(s/D);
    pm/=D;
    return *this;
  }
  Fractional &operator/=(const Fractional &other){
    pn*=other.pm;
    pm*=other.pn;
    regularize();
    return *this;
  }
  Fractional &operator+=(const Fractional &other){
    //to reduce overflow, we do it in two steps
    long D=labs(MyMath::gcd(pm,other.pm));
    long m=pm/D;
    long om=other.pm/D;
    pn=pn*om+other.pn*m;
    pm=om*m*D;
    regularize();
    return *this;
  };
  Fractional &operator-=(const Fractional &other){
    //to reduce overflow, we do it in two steps
    long D=labs(MyMath::gcd(pm,other.pm));
    long m=pm/D;
    long om=other.pm/D;
    pn=pn*om-other.pn*m;
    pm=om*m*D;
    regularize();
    return *this;
  };

  long sign()const{ return lSign(1,pn*pm); }

  Fractional operator*(const Fractional &other)const{
    return Fractional(pn*other.pn,pm*other.pm);
  }
  Fractional operator*(long other)const{
    return Fractional(pn*other,pm);
  }
  Fractional operator/(const Fractional &other)const{
    return Fractional(pn*other.pm,pm*other.pn);
  }
  Fractional operator+(const Fractional &other)const{
    long D=labs(MyMath::gcd(pm,other.pm));
    long m=pm/D;
    long om=other.pm/D;
    return Fractional(pn*om+other.pn*m,om*m*D);
  }
  Fractional operator-(const Fractional &other)const{
    long D=labs(MyMath::gcd(pm,other.pm));
    long m=pm/D;
    long om=other.pm/D;
    return Fractional(pn*om-other.pn*m,om*m*D);
  }

  bool operator<=(const Fractional &o)const{return (pn*o.pm<=pm*o.pn);};
  bool operator<(const Fractional &o)const{return (pn*o.pm<pm*o.pn);};
  bool operator>(const Fractional &o)const{return (pn*o.pm>pm*o.pn);};
  bool operator>=(const Fractional &o)const{return (pn*o.pm>=pm*o.pn);};
  bool operator!=(long n)const{return (pn!=pm*n);};
  bool operator==(const Fractional &o)const{return (pn*o.pm==pm*o.pn);};
  bool operator!=(const Fractional &o)const{return (pn*o.pm!=pm*o.pn);};


  long backfolder(const Fractional &o)const{
    // return k such that *this+k*o gives smallest positive
    // nominator
    long a=pn*o.pm;
    long b=pm*o.pn;
    // (a+k*b)=r >=0 ;
    long r=labs(a%b);
    return (r-a)/b;
  }

  Fractional &regularizeInterval(bool centered){
    pn-=(pn/pm)*pm;
    if(centered){
      while(2*pn<=-pm) pn+=pm;
      while(2*pn>  pm) pn-=pm;
    }else{
      while(pn<0) pn+=pm;
      while(pn>=pm) pn-=pm;
    }
    return *this;
  };

  long toLong()const{
    if(pm!=1) throw Except("cannot convert non-integer Fractional to long");
    return pn;
  }


  long round()const{
    long n=pn*2+pm;
    long m=pm*2;
    long d=n/m;
    /*
      // to make r/m positive
      long r=n%m;
      if(r*m<0){ d--; r=lSign(labs(m)-labs(r),m); }
    */

    if((n%m)*m<0) d--;

    return d;
  }
    
  double toDouble()const{
    return ((double)pn)/((double)pm);
  }


  String printS(bool withsign=false)const{
    if(withsign){
      //if(pn==0) return "+0";
      if(pm==1) return Sprintf("%+ld",pn);
      return Sprintf("%+ld/%ld",pn,pm);
    }else{
      //if(pn==0) return "0";
      if(pm==1) return Sprintf("%ld",pn);
      return Sprintf("%ld/%ld",pn,pm);
    }
  };

  String printS(const char *xyz,bool withsign=false)const{
    if(withsign){
      if(pm==1) {
	if(pn==1) 
	  return Sprintf("+%s",xyz);
	else if(pn==-1)
	  return Sprintf("-%s",xyz);
	else
	  return Sprintf("%+ld%s",pn,xyz);
      }else{
	if(pn==1) 
	  return Sprintf("+%s/%ld",xyz,pm);
	else if(pn==-1)
	  return Sprintf("-%s/%ld",xyz,pm);
	else
	  return Sprintf("%+ld%s/%ld",pn,xyz,pm);
      }
    }else{
      if(pm==1) {
	if(pn==1) 
	  return Sprintf("%s",xyz);
	else if(pn==-1)
	  return Sprintf("-%s",xyz);
	else
	  return Sprintf("%ld%s",pn,xyz);
      }else{
	if(pn==1) 
	  return Sprintf("%s/%ld",xyz,pm);
	else if(pn==-1)
	  return Sprintf("-%s/%ld",xyz,pm);
	else
	  return Sprintf("%ld%s/%ld",pn,xyz,pm);
      }
    }
  };
  
  inline long n()const{return pn;};
  inline long m()const{return pm;};
private:
  long pn,pm;

  friend inline Fractional abs(const Fractional &a);
};


inline Fractional abs(const Fractional &a){ 
  return Fractional(a).abs();
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */

inline MyDebug operator<<(MyDebug dbg,const Fractional &f){
  if(f.m()==1){
    fprintf(dbg.FH," %ld ",f.n());
  }else{
    fprintf(dbg.FH," %ld/%ld ",f.n(),f.m());
  }
  return dbg;
}



#endif /* _FRACTIONAL_HH */

