/* @(#)MolecularOrbital.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _MOLECULARORBITAL_H
#define _MOLECULARORBITAL_H 1

#include "FString.hh"
#include "Cmplx.hh"
#include "Vec3.hh"
#include "PtrArray.hh"
#include "ValueArray.hh"
#include "SqMatrix.hh"
#include "MACROS.hh"

namespace XFPLO{

  class CoeffBasisInfo;
  class MolecularOrbitals;  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class MolecularOrbitalData
  {
  public:
    
    // ----------------------------------------------------------------
    MolecularOrbitalData():psite(0),porbital("all"),porbtype(String::SL_all),
			   pxaxis(Vec3(1.,0.,0.)),
			   pzaxis(Vec3(0.,0.,1.)),pspinzaxis(Vec3(0.,0.,1.)),
			   puseglobalspinzaxis(true),pfactor(1.)
    {
    }
    

    // ----------------------------------------------------------------
    MolecularOrbitalData(const MolecularOrbitalData &other)
    {
      *this=other;
    }
    // ----------------------------------------------------------------
    MolecularOrbitalData &operator=(const MolecularOrbitalData &other)
    {
      psite=other.psite;
      porbital=other.porbital;
      porbtype=other.porbtype;
      pnmain=other.pnmain;
      pl=other.pl;
      pj=other.pj;
      pm=other.pm;
      pposition=other.pposition;
      pxaxis=other.pxaxis;
      pzaxis=other.pzaxis;
      pspinzaxis=other.pspinzaxis;
      puseglobalspinzaxis=other.puseglobalspinzaxis;
      pfactor=other.pfactor;
      return *this;
    }
    
    // ----------------------------------------------------------------
    MolecularOrbitalData *clone()
    {
      return new MolecularOrbitalData(*this);
    }
    
    // ----------------------------------------------------------------
    ~MolecularOrbitalData()
    {
    }
    
    // --------------------------------------------------------------
    void setOrbital(const String &o){
      porbital=o;
      porbtype=String::isStateLabel(porbital,pnmain,pl,pj,pm);
    }
    
    // ----------------------------------------------------------------
    void resetTrans()
    {
      pDjmm.clear();
      pDlms.clear();
      pDlm.clear();
    }

    // ----------------------------------------------------------------
    SqMatrix<Cmplx> &Djmm(long j)
    {
      if(j>=pDjmm.size()){
	pDjmm.resize(j+1);
      }
      return pDjmm[j];
    }

    // ----------------------------------------------------------------
    SqMatrix<Cmplx> &Dlms(long l)
    {
      if(l>=pDlms.size()){
	pDlms.resize(l+1);
      }
      return pDlms[l];
    }

    // ----------------------------------------------------------------
    SqMatrix<Cmplx> &Dlm(long l)
    {
      if(l>=pDlm.size()){
	pDlm.resize(l+1);
      }
      return pDlm[l];
    }

    // ----------------------------------------------------------------
    String printS(const String &el,bool molecular)const;
    
  private:
    PROPERTY(site,site,Site,long);
    ROCPROPERTY(orbital,orbital,String);
    PROPERTY(orbtype,orbtype,Orbtype,String::StateLabels);
    PROPERTY(nmain,nmain,Nmain,long);
    PROPERTY(l,l,L,long);
    PROPERTY(j,j,J,long);
    PROPERTY(m,m,M,long);
    CMODPROPERTY(position,position,Position,Vec3);
    CMODPROPERTY(xaxis,xaxis,Xaxis,Vec3);
    CMODPROPERTY(zaxis,zaxis,Zaxis,Vec3);
    CMODPROPERTY(spinzaxis,spinZaxis,SpinZaxis,Vec3);
    PROPERTY(useglobalspinzaxis,useGlobalSpinZaxis,UseGlobalSpinZaxis,bool)
    CPROPERTY(factor,factor,Factor,Cmplx);
    ValueArray<SqMatrix<Cmplx> > pDjmm;
    ValueArray<SqMatrix<Cmplx> > pDlms;
    ValueArray<SqMatrix<Cmplx> > pDlm;
    friend class MolecularOrbitals;
  };
  
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  
  class MolecularOrbital{
  private:
    
  public:
    // ----------------------------------------------------------------
    MolecularOrbital()
      :pname("MolecularOrbital"),pactive(true),pismolecular(false)
    {
    }
    
    // ----------------------------------------------------------------
    MolecularOrbital(MolecularOrbitalData *md)
    :pname("MolecularOrbital"),pactive(true),pismolecular(false)
    {
      pmolorbcontriblist.append(md);
    }
    
    // ----------------------------------------------------------------
    MolecularOrbital(const MolecularOrbital &other)
    {
      pname=other.pname;
      pactive=other.pactive;
      pismolecular=other.pismolecular;
      pmolorbcontriblist=other.pmolorbcontriblist;
    }
    
    // ----------------------------------------------------------------
    MolecularOrbital(const String &n) 
    :pname(n),pactive(true),pismolecular(true)
    {
    }
    
    // ----------------------------------------------------------------
    ~MolecularOrbital();
    
    // ----------------------------------------------------------------
    inline long numOfContribs()const
    {
      return pmolorbcontriblist.size();
    }
    
    // ----------------------------------------------------------------
    inline bool isMolecular()const
    {
      return pismolecular;
    }
    
    // ----------------------------------------------------------------
    inline MolecularOrbitalData *operator[](long index)
    {
      return pmolorbcontriblist[index];
    }
    
    // ----------------------------------------------------------------
    inline const MolecularOrbitalData *at(long index) const
    {
      return pmolorbcontriblist.at(index);
    }
    
    // ----------------------------------------------------------------
    inline void append(MolecularOrbitalData *md){
      pmolorbcontriblist.append(md);
    }
    
    // ----------------------------------------------------------------
    void insert(int where,MolecularOrbitalData *md)
    {
      pmolorbcontriblist
	.insert(MyMath::min<long>(where,pmolorbcontriblist.size()),md);
    }
    
    // ----------------------------------------------------------------
    inline MolecularOrbitalData *takeAt(int where)
    {
      return pmolorbcontriblist.takeAt(where);
    }
    
    // ----------------------------------------------------------------
    void resetTrans()
    {
      long N=pmolorbcontriblist.size();
      for(int im=0;im<N;im++) pmolorbcontriblist[im]->resetTrans();
    }

    // ----------------------------------------------------------------
    void writeToFile(FILE *FH)const;
    
  private:
    
    CPROPERTY(name,name,Name,String);
    PROPERTY(active,isActive,Active,bool);
    bool pismolecular;
    
    PtrArray<MolecularOrbitalData> pmolorbcontriblist;
    friend class MolecularOrbitals;
  };
  
  
  /* ***********************************************************************
   *
   * ***********************************************************************
   */
  class MolecularOrbitals{
  public:
    // --------------------------------------------------------------
    MolecularOrbitals();

    // --------------------------------------------------------------
    ~MolecularOrbitals();
    
    // --------------------------------------------------------------
    void clean();
    
    // --------------------------------------------------------------
    inline long size()const
    {
      return pmolorblist.size();
    }
    
    // --------------------------------------------------------------
    inline MolecularOrbital *operator[](long index)
    {
      return pmolorblist[index];
    }
    
    // --------------------------------------------------------------
    inline long indexOf(MolecularOrbital *mo)const
    {
      return pmolorblist.indexOf(mo);
    }
    
    
    // --------------------------------------------------------------
    void insert(int where,MolecularOrbital *mo){
      pmolorblist.insert(MyMath::min<long>(where,pmolorblist.size()),mo);
    }
    
    // --------------------------------------------------------------
    MolecularOrbital *takeAt(int where){
      return pmolorblist.takeAt(where);
    }
    
    
    // --------------------------------------------------------------
    void buildMatrix(const Vec3 &kpoint,const CoeffBasisInfo &cbi);
    
    
    
    
    
    // --------------------------------------------------------------
    inline const Cmplx &U(long iz,long iw)const
    {
      return p_U[iz+pnvdim*iw];
    }

    // --------------------------------------------------------------
    inline const Cmplx *U()const{
      return p_U;
    }
    // --------------------------------------------------------------
    inline const String &name(long iw)const
    {
      return pnames[iw];
    }
    
    // --------------------------------------------------------------
    inline long numWeights()const
    {
      return pnumweights;
    }
    
    // --------------------------------------------------------------
    void writeToFile(FILE *FH);

    // --------------------------------------------------------------
    String readFromFile(FILE *FH,String &coeffilename,String &bwfilename,
			bool prot,StringList &output);

    // --------------------------------------------------------------
    void setup(const CoeffBasisInfo &cbi,bool prot,StringList &output);

    // ----------------------------------------------------------------
    void bcast_data();
    
  private:
    // --------------------------------------------------------------
    void rotatedUnsrel(long l,long mv0,long num,long iz
		       ,long &iw,long pnvdim,const String &el
		       ,MolecularOrbitalData *md,const Cmplx &fac
		       ,bool molcontrib);

    // --------------------------------------------------------------
    void rotatedUfrel(long j,long mu0,long num,long iz,
		      long &iw,long pnvdim,const String &el,
		      MolecularOrbitalData *md,const Cmplx &fac,
		      bool molcontrib);

    // --------------------------------------------------------------
    void rotatedUfrelNL(long l,long mv0,
			long num,long iz,
			long &iw,long pnvdim,
			const String &el,
			MolecularOrbitalData *md,
			const Vec3 &globalspinzaxis,
			const Cmplx &fac);

    // ----------------------------------------------------------------
    void resetTrans()
    {
      long N=pmolorblist.size();
      for(int ic=0;ic<N;ic++) pmolorblist[ic]->resetTrans();
    }

    
  private:
    long pnumweights;
    long pnvdim;
    Cmplx *p_U;
    StringList pnames;
    
    PtrArray<MolecularOrbital> pmolorblist;
    CPROPERTY(globalspinzaxis,GlobalSpinZaxis,GlobalSpinZaxis,Vec3)
  };
  
}
#ifdef	__cplusplus
extern "C" {
#endif

#ifdef	__cplusplus
}
#endif


#endif /* _MOLECULARORBITAL_H */

