/*************************************************************************
 * @(#)SqMatrixYlm.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   05 Mar 2010
 *************************************************************************
 */

#include "SqMatrix.hh"
#include "Vec3.hh"
#include "MyMath.hh"
#include "Cmplx.hh"
#include "MyDebug.hh"

static const double sq2o2=.7071067811865475244008443621048490392848;
static const double sq2=1.4142135623730950488016887242096980785696;




/* ***********************************************************************
 *
 *  make U such that \cal{Y}=Y U
 *  where \cal{Y} are complex spherical harmonics and Y are real ones
 *
 * ***********************************************************************
 */

SqMatrix<Cmplx> makeYlmUMatrix(long l){
  if(l<0){
    fprintf(stderr,"argument l=%ld  <  0 !\n",l);
    exit(2);
  }
  //myDebug() << "makeYlmUMatrix: create temp";
  long deg=2*l+1;
  SqMatrix<Cmplx> temp(deg);

  //temp.printSInfo("makeYlmUMatrix::temp");

  Cmplx c;
  for(long ma=-l;ma<=l;ma++){
    double onema=1.;
    if(ma%2) onema=-1.;
    if(ma<0){
      c=Cmplx(0.,-sq2o2);
    }else if(ma>0){
      c=Cmplx(onema*sq2o2);
    }else{
      c=Cmplx(1.);
    }
    temp(ma+l,ma+l)=c;
    if(ma>0){
      c=Cmplx(sq2o2);
    }else if(ma<0){
      c=Cmplx(0.,onema*sq2o2);
    }else{
      c=Cmplx(1.);
    }
    temp(ma+l,-ma+l)=c;
  }

  return temp;
}

/* ***********************************************************************
 * return the transformation from pseudo non rel complex spin-harmonics to 
 * spherical spinors
 *
 * basis spinors are |s> = \chi_{s} (eigen functions of \sigma_z)
 * complex spherical harmonics are |lm> = \cal{Y}_{lm}
 * complex spin-spherical harmonics are |lms> = \chi_{s}*\cal{Y}_{lm}
 * spherical spinors are |\kappa\mu> = \chi_{\kappa\mu}
 *
 * |\kappa\mu> = sum_{s} c^s_{\kappa\mu} \chi_{s}*\cal{Y}_{l,\mu-s/2}
 *             = |lms> T^l_{ms,\kappa\mu}
 *
 * T^l_{ms,\kappa\mu} = c^s_{\kappa\mu} *\delta_{m,\mu-s/2}
 *
 * where c^s_{\kappa\mu} are the Clebsch-Gordon coefficients for
 * adding momenta l and s=1/2
 *
 * the T -matrix is returned such that s=1/2,-1/2 runs first (first is majority spin)
 * and m runs second, which gives a matrix dimension 2*deg_l
 * the kappa,mu side is organized such that the outer loop is kappa: 
 * kappa=l comes first and kappa=-l-1 later
 * the inner loop is the mu loop
 * ***********************************************************************
 */
SqMatrix<Cmplx> relTOpseudoCmplx(const Vec3 &spinzaxis,long l)
{
  long degl=2*l+1;
  long dim=2*degl;
  SqMatrix<Cmplx> temp(dim);


  SqMatrix<Cmplx> D1o2=wignerDChijm(Vec3::eulerAngles(spinzaxis),1,true);


  long i=0;
  for(long s=1;s>=-1;s-=2){
    for(long m=-l;m<=l;m++){
      long j=0;
      for(long kap=l;kap>=-l-1;kap-=degl){
	if(!kap) continue;
	long J=2*abs(kap)-1;

	for(long t=-1;t<=1;t+=2){

	  long mu=2*m+t;
	  if(MyMath::Abs(mu)<=J){
	    temp(i,j+(mu+J)/2)+=D1o2((1+s)/2,(1+t)/2)
	      *MyMath::ClebschGordonChi(kap,mu,t);
	  }
	}
	j+=J+1;
      }
      i++;
    }
  }


  return temp;
  
}

/* ***********************************************************************
 * same as relTOpseudoCmplx but transforming to real spherical harmonics
 * ***********************************************************************
 */
SqMatrix<Cmplx> relTOpseudo(const Vec3 &spinzaxis,long l)
{
  long degl=2*l+1;
  long dim=2*degl;

  SqMatrix<Cmplx> temp(dim);
  SqMatrix<Cmplx> T=relTOpseudoCmplx(spinzaxis,l);
  SqMatrix<Cmplx> U=makeYlmUMatrix(l);
  
  
  for(long j=-1;++j<dim;){
    for(long ms=-1;++ms<2;){
      for(long m=-1;++m<degl;){
	Cmplx c;
	for(long n=-1;++n<degl;){
	  c+=U(m,n)*T(n+degl*ms,j);
	}
	temp(m+degl*ms,j)=c;
      }
    }
  }
  
  return temp;

}

/* ***********************************************************************
 *
 !! this function computes dlmm1 matrices as defined in Varshalovich & Co
 !! page 68 4.3.3
 !!
 !! sb=sin(beta/2)
 !! cb=cos(beta/2)
 !!
 !!                     attention!
 !!     0 to the 0 power is ZERO under NDP FORTRAN
 !!
 *
 * ***********************************************************************
 */
static double dlmm1(long l,long m,long m1,double sb,double cb,MyMath::dFactorial &fac)
{

  double sum=0.;
  for(long k=MyMath::max<long>(0,m+m1);k<=MyMath::min<long>(l+m,l+m1);k++){
    double c=(k%2?-1.:1.)/(fac(k)*fac(l+m-k)*fac(l+m1-k)*fac(k-m-m1));
    long k1=2*k-m-m1;
    long k2=2*l-k1;
    double cbk=1.,sbk=1.;
    if(k1) cbk=pow(cb,k1);
    if(k2) sbk=pow(sb,k2);
    
    sum+=c*cbk*sbk;
  }

  return ((l+m)%2?-1.:1.)*sqrt(fac(l+m)*fac(l-m)*fac(l+m1)*fac(l-m1))*sum;

}

/* ***********************************************************************
 *
 !! This routine returns the Wigner D-functions for real spherical harmonics
 !! for even Integer J = l
 !! The input are the euler angles and and inv
 !! for inv==0 a rotation is assumed
 !! for inv==1 a roto-inversion is assumed
 !! The array dr returns the rotation matrix
 !! which is defined as
 !!
 !!  Y_{lm}(a_{ij} r_j) = \sum_m1 \tilde{D}^l_{mm1} Y_{lm1}(r_i)
 !!
 !!  (if trans == .false.)
 !!  here a is the matrix, which transforms the old into the
 !!  new axes via \vec{e}^\prime_i=\alpha_{ij} \vec{e}_j
 !!
 !!  The real wigner function returned here is related to Wigner function
 !!  of the complex spherical harmonics 
 !!
 !!    D^l_mm1=<lm|\hat{D}(\alpha,\beta,\gamma)|lm1>
 !!
 !! via
 !!   \tilde{D}=UD^{+}U^{+}
 !! where U is the transformation of the real (Y) into the complex ({\cal Y})
 !! spherical harmonics
 !!
 !!  {\cal Y}^{T}= Y^{T} U
 !!  
 *
 * ***********************************************************************
 */
SqMatrix<Cmplx> wignerDRealYlm(const Vec3 &eul,bool inv,long l,bool trans)
{

  MyMath::dFactorial fac;


  //
  // l,m in new coordinates, l,m1 in old ones
  //
    
  double sb=sin(eul[1]*0.5);
  double cb=cos(eul[1]*0.5);
  
  //  for inversion
  double si;
  if((!inv) ||  (!(l%2))){
    si=1.;
  }else{
    si=-1.;
  }
  //
  // for m1=0
  //

  SqMatrix<Cmplx> dr(2*l+1);

  
  dr(l,l)=si*dlmm1(l,0,0,sb,cb,fac);
  for(int m=1;m<=l;m++){
    double dp=sq2*si*dlmm1(l,0,m,sb,cb,fac);
    dr(l,l-m)=dp*sin(m*eul[0]);
    dr(l,l+m)=dp*cos(m*eul[0]);
  }
    
  //
  // for m1>0
  //
  for(int m1=1;m1<=l;m1++){

    double dp=sq2*si*dlmm1(l,m1,0,sb,cb,fac);
    dr(l-m1,l)=-dp*sin(m1*eul[2]);
    dr(l+m1,l)= dp*cos(m1*eul[2]);

    for(int m=1;m<=l;m++){

      double sapg=sin(m*eul[0]+m1*eul[2]);
      double capg=cos(m*eul[0]+m1*eul[2]);
      double samg=sin(m*eul[0]-m1*eul[2]);
      double camg=cos(m*eul[0]-m1*eul[2]);
      double dp=si*dlmm1(l,m1, m,sb,cb,fac);
      double dm=si*(m%2?-1.:1.)*dlmm1(l,m1,-m,sb,cb,fac);
      dr(l-m1,l-m)= dp*capg-dm*camg;
      dr(l-m1,l+m)=-(dp*sapg)+dm*samg;
      dr(l+m1,l-m)= dp*sapg+dm*samg;
      dr(l+m1,l+m)= dp*capg+dm*camg;
    }    
  }    
  

  if(trans){
    return dr.transposed();
  }

  return dr;
}

/* ***********************************************************************
   !! returns the dj,mm'(beta) functions
   !!
   !! The input quantum numbers are J=2j, M=2m, MP=2m'
   !! with |m|,|m'|<=j
   !! with j,m,m' half integer
   * ***********************************************************************
   */
static double djmm(long J,long M,long MP,double sb,double cb,MyMath::dFactorial &fac)
{
  // J  equals 2*j
  // M,MP equal 2*m,2*mp
  long l=(M+J)/2;
  long lp=(MP+J)/2;
  // M=(-J+2*l)
  // MP=(-J+2*lp)
  //
  
  double d=0.0;
  for(long k=MyMath::max<long>(-J+l+lp,0);k<=MyMath::min<long>(l,lp);k++){
    double f=pow(-1.,k)/(fac(k)*fac(l-k)*fac(lp-k)*fac(k+J-l-lp));
    long k1=2*k+J-l-lp;
    long k2=J-k1;
    double ck,sk;
    if(k1 == 0){
      ck=1.0;
    }else{
      ck=pow(cb,k1);
    }
    if(k2 == 0){
      sk=1.0;
    }else{
      sk=pow(sb,k2);
    }
    d+=f*ck*sk;
  }

  return d*pow(-1.,l)*sqrt(fac(l)*fac(J-l)*fac(lp)*fac(J-lp));
  
      
}
/* ***********************************************************************
 !! This routine returns the Wigner D-functions for spherical spinors
 !! for half Integer J 
 !! The input are the euler angles, J  and and trans
 !! for trans==true the hermitian conjugate matrix is returned
 !! The array returns the rotation matrix
 !! which is defined as
 !!
 !!  \hat{D} \chi_{j\mu} = \sum_{mu_1}  \chi_{j\mu_1} D^j_{\mu_1\mu}
 !!
 !! the matrix index refers to the order:  -m to m , m half integer
 !! j must be half integer
 !! on input J=2j
 !!
 !!
 !!  
 * ***********************************************************************
 */
SqMatrix<Cmplx> wignerDChijm(const Vec3 &eul,long J,bool trans)
{

  SqMatrix<Cmplx> D(J+1);
  
  MyMath::dFactorial fac;
  
  double cb=cos(eul[1]*0.5);
  double sb=sin(eul[1]*0.5);
  
  for(int l=0;l<=J;l++){
    for(int lp=0;lp<=J;lp++){
      long M=(-J+2*l);
      long MP=(-J+2*lp);
      if(trans){
	D(lp,l)=exp(Cmplx::ci*M*eul[0]*0.5)
	  *djmm(J,M,MP,sb,cb,fac)
	  *exp(Cmplx::ci*MP*eul[2]*0.5);
      }else{
	D(l,lp)=exp(-Cmplx::ci*M*eul[0]*0.5)
	  *djmm(J,M,MP,sb,cb,fac)
	  *exp(-Cmplx::ci*MP*eul[2]*0.5);
      }
    }
  }

  return D;
}

/* ***********************************************************************
 * return the time reversal matrix such that
 * 
 * \theta Y_{lm} =sum_{n} Y_{ln} D^{\theta}_{nm} *K_0
 * 
 * where K_0 is the complex conjugation
 * 
 * and Y_{lm} are real spherical harmonics and the matrix index order is
 * 
 * m=-l,-l+1,..,0,1,..,l-1,l
 * 
 * ***********************************************************************
 */
SqMatrix<Cmplx> timeReversalRealYlm(long l)
{
  long degl=2*l+1;
  SqMatrix<Cmplx> res(degl);
  for(long i=-1;++i<degl;){
    res(i,i)=1.;
  }
  return res;
  
}

/* ***********************************************************************
 * return the time reversal matrix such that
 * 
 * \theta Y_{lm}\chi_{s} =sum_{nt} Y_{ln}\chi_{t} D^{\theta}_{nt,ms} *K_0
 * 
 * where K_0 is the complex conjugation
 * 
 * and Y_{lm} are real spherical harmonics and \chi_{s} are spin eigen 
 * functions. The matrix index order is
 * 
 * index(ms)=-l up,..,l up,-l down,..,l down
 * 
 * ***********************************************************************
 */
SqMatrix<Cmplx> timeReversalRealYlmChi(long l)
{
  long degl=2*l+1;
  SqMatrix<Cmplx> res(2*degl);
  for(long i=-1;++i<degl;){
    res(i,i+degl)=-1.;
    res(i+degl,i)=1.;
  }
  return res;
  
}

/* ***********************************************************************
 * return the time reversal matrix such that
 * 
 * \theta Y_{lm} =sum_{n} Y_{ln} D^{\theta}_{nm} *K_0
 * 
 * where K_0 is the complex conjugation
 * 
 * and Y_{lm} are complex spherical harmonics and the matrix index order is
 * 
 * m=-l,-l+1,..,0,1,..,l-1,l
 * 
 * ***********************************************************************
 */
SqMatrix<Cmplx> timeReversalYlm(long l)
{
  long degl=2*l+1;
  SqMatrix<Cmplx> res(degl);
  double s=1.;
  if(l%2==1) s=-1.;
  for(int i=-1;++i<degl;){
    res(i,degl-1-i)=s;
    s*=-1.;
  }
  //printSqMatrix(res,"(%2.0f, %2.0f) ");
  return res;
}


/* ***********************************************************************
 * return the time reversal matrix such that
 * 
 * \theta Y_{lm}\chi_{s} =sum_{nt} Y_{ln}\chi_{t} D^{\theta}_{nt,ms} *K_0
 * 
 * where K_0 is the complex conjugation
 * 
 * and Y_{lm} are complex spherical harmonics and \chi_{s} are spin eigen 
 * functions. The matrix index order is
 * 
 * index(ms)=-l up,..,l up,-l down,..,l down
 * 
 * ***********************************************************************
 */

SqMatrix<Cmplx> timeReversalYlmChi(long l)
{
  long degl=2*l+1;
  SqMatrix<Cmplx> res(2*degl);
  double s=-1.;
  if(l%2==1) s=1.;
  for(int i=-1;++i<2*degl;){
    res(i,2*degl-1-i)=s;
    s*=-1.;
  }
  //printSqMatrix(res,"(%2.0f, %2.0f) ");
  return res;
}


/* ***********************************************************************
 * return the time reversal matrix such that
 * 
 * \theta \chi_{\kappa\mu} =sum_{nt} \chi_{\kappa\nu} D^{\theta}_{\nu,\mu} *K_0
 * 
 * where K_0 is the complex conjugation
 * 
 * and \chi_{\kappa\mu} are spherical spinors. The matrix index order is
 * 
 * index(mu)=-j,..,j
 * 
 * The j input is actually 2*j!
 * 
 * ***********************************************************************
 */

SqMatrix<Cmplx> timeReversalSphericalSpinor(long l,long j)
{
  long degj=j+1;
  SqMatrix<Cmplx> res(degj);
  int s=1;
  if(l%2==1)s=-1;
  for(int i=-1;++i<degj;){
    res(i,degj-1-i)=s;
    s*=-1;
  }
  //printSqMatrix(res,"(%2.0f, %2.0f) ");
  return res;
}

/* ***********************************************************************
 * return the time reversal matrix such that
 * 
 * \theta \chi_{\kappa\mu} =sum_{nt} \chi_{\kappa\nu} D^{\theta}_{\nu,\mu} *K_0
 * 
 * where K_0 is the complex conjugation
 * 
 * and \chi_{\kappa\mu} are spherical spinors. The matrix index order is
 * 
 * index('kappa\mu)=l -(l-1/2),..,l (l-1/2), -l-1 -(l+1/2),.., -l-1 +(l+1/2)
 * 
 * in other words first the j=l-1/2 block then the j=l+1/2 block
 * 
 * 
 * ***********************************************************************
 */

SqMatrix<Cmplx> timeReversalSphericalSpinorLShell(long l)
{
  long degl=2*l+1;
  SqMatrix<Cmplx> res(2*degl);
  int s=1;
  if(l%2==1)s=-1;
  for(int i=-1;++i<degl-1;){
    res(i,degl-2-i)=s;
    s*=-1;
  }
  for(int i=-1;++i<degl+1;){
    res(i+degl-1,2*degl-1-i)=s;
    s*=-1;
  }
  //printSqMatrix(res,"(%2.0f, %2.0f) ");
  return res;
}

/* ***********************************************************************
 * return the Ls matrix <ln|L_s|lm> such that
 * 
 *  L_0=L_z
 *  L_s=L_x+i*s*L_y for s=+/-1
 *
 * where |lm> are complex harmonics. The matrix index order is
 * m=-l,-l+1,..,0,1,..,l-1,l
 * 
 * 
 * ***********************************************************************
 */

SqMatrix<Cmplx> LsYlm(long l,long s)
{
  long degl=2*l+1;
  SqMatrix<Cmplx> res(degl);
  
  for(int i=-1;++i<degl;){
    if(s==0){
      res(i,i)=(i-l);
    }else{
      if(abs(i-s-l)<=l){
	res(i,i-s)=sqrt(l*(l+1.)-1.*(i-s-l)*(i-l));
      }
    }
  }

  //myDebug();
  //printSqMatrix(res,"(%4.2f, %4.2f) ");
  
  return res;
}

/* ***********************************************************************
 * spin-complex harmonics
 * ***********************************************************************
 */
SqMatrix<Cmplx> LsYlmChi(long l,long s)
{
  long degl=2*l+1;
  SqMatrix<Cmplx> res(2*degl);
  
  for(int i=-1;++i<degl;){
    if(s==0){
      res(i,i)=(i-l);
      res(i+degl,i+degl)=(i-l);
    }else{
      double C=sqrt(l*(l+1.)-1.*(i-s-l)*(i-l));
      if(abs(i-s-l)<=l){
	res(i,i-s)=C;
	res(i+degl,i-s+degl)=C;
      }
    }
  }

  //myDebug();
  //printSqMatrix(res,"(%4.2f, %4.2f) ");
  
  return res;
}

/* ***********************************************************************
 * Lx,Ly,Lz in complex harmonics
 * ***********************************************************************
 */
SqMatrix<Cmplx> LxyzYlm(long l,long k)
{
  long degl=2*l+1;
  SqMatrix<Cmplx> res(degl);
  
  for(int i=-1;++i<degl;){
    if(k==2){
      res(i,i)=(i-l);
    }else if(k==0){

      double C=sqrt(l*(l+1.)-(i-1.-l)*(i-l))*0.5;
      if(abs(i-1-l)<=l){
	res(i,i-1)=C;
      }
      C=sqrt(l*(l+1.)-(i+1.-l)*(i-l))*0.5;
      if(abs(i+1-l)<=l){
	res(i,i+1)=C;
      }
    

    }else if(k==1){
    
      double C=-sqrt(l*(l+1.)-(i-1.-l)*(i-l))*0.5;
      if(abs(i-1-l)<=l){
       	res(i,i-1)=Cmplx(0.,C);
      }
    
      C=-sqrt(l*(l+1.)-(i+1.-l)*(i-l))*0.5;
      if(abs(i+1-l)<=l){
       	res(i,i+1)=Cmplx(0.,-C);
      }
    

    }
  }

  // myDebug()<<"k="<<k;
  // printSqMatrix(res,"(%4.2f, %4.2f) ");
  
  return res;
}
/* ***********************************************************************
 * Lx,Ly,Lz in spin-complex harmonics
 * ***********************************************************************
 */
SqMatrix<Cmplx> LxyzYlmChi(long l,long k)
{
  long degl=2*l+1;
  SqMatrix<Cmplx> res(2*degl);
  
  for(int i=-1;++i<degl;){
    if(k==2){
      res(i,i)=(i-l);
      res(i+degl,i+degl)=(i-l);
    }else if(k==0){

      double C=sqrt(l*(l+1.)-(i-1.-l)*(i-l))*0.5;
      if(abs(i-1-l)<=l){
	res(i,i-1)=C;
	res(i+degl,i-1+degl)=C;
      }
      C=sqrt(l*(l+1.)-(i+1.-l)*(i-l))*0.5;
      if(abs(i+1-l)<=l){
	res(i,i+1)=C;
	res(i+degl,i+1+degl)=C;
      }
    

    }else if(k==1){
    
      double C=-sqrt(l*(l+1.)-(i-1.-l)*(i-l))*0.5;
      if(abs(i-1-l)<=l){
       	res(i,i-1)=Cmplx(0.,C);
       	res(i+degl,i-1+degl)=Cmplx(0.,C);
      }
    
      C=-sqrt(l*(l+1.)-(i+1.-l)*(i-l))*0.5;
      if(abs(i+1-l)<=l){
       	res(i,i+1)=Cmplx(0.,-C);
       	res(i+degl,i+1+degl)=Cmplx(0.,-C);
      }
    

    }
  }

  //myDebug();
  //printSqMatrix(res,"(%4.2f, %4.2f) ");
  
  return res;
}


/* ***********************************************************************
 * Lxyz matrices in spherical spinors for a whole l-shell
 * ***********************************************************************
 */
SqMatrix<Cmplx> LxyzSphericalSpinorLShell(long l,long k)
{
  int degl=2*l+1;
  SqMatrix<Cmplx> res(2*degl);
  

  double c1[2];
  double c2[2];

  Cmplx ci(0.,1.);

  int ind1=0;
  for(int kap1=l;kap1>=-l-1;kap1-=degl){
    int degj1=2*abs(kap1);
    for(int i1=-1;++i1<degj1;){
      int mu21=2*i1-(degj1-1);
      int ind2=0;
      c1[0]=dSign(1.,-kap1)*sqrt(0.5*(1.-mu21/(2.*kap1+1)));//s=1
      c1[1]=sqrt(0.5*(1.+mu21/(2.*kap1+1)));//s=-1
      for(int kap2=l;kap2>=-l-1;kap2-=degl){
	int degj2=2*abs(kap2);
	for(int i2=-1;++i2<degj2;){
	  int mu22=2*i2-(degj2-1);

	  c2[0]=dSign(1.,-kap2)*sqrt(0.5*(1.-mu22/(2.*kap2+1)));
	  c2[1]=sqrt(0.5*(1.+mu22/(2.*kap2+1)));

	  //myDebug()<<"i,j="<<ind1<<ind2<<"qn="<<(degj1-1)<<mu21<<","<<(degj2-1)<<mu22;

	  Cmplx lp,lm;
	  for(int s=-1;s<2;s+=2){
	    if(mu21==mu22 && k==2){
	      res(ind1,ind2)+=c1[(1-s)/2]*c2[(1-s)/2]*(mu22-s)*0.5;
	    }
	    int q=1;
	    double m=(mu22-s)*0.5;
	    if(mu21==mu22+2*q){
	      lp+=c1[(1-s)/2]*c2[(1-s)/2]*sqrt(l*(l+1.)-m*(m+q));
	    }
	    q=-1;
	    if(mu21==mu22+2*q){
	      lm+=c1[(1-s)/2]*c2[(1-s)/2]*sqrt(l*(l+1.)-m*(m+q));
	    }
	  }
	  if(k==0) res(ind1,ind2)=(lp+lm)*0.5;
	  if(k==1) res(ind1,ind2)=-ci*(lp-lm)*0.5;

	  ind2++;
	}
      }
      ind1++;
    }
  }



  return res;
}


/* ***********************************************************************
 * Sigma-xyz matrices in spherical spinors for a whole l-shell
 * ***********************************************************************
 */

SqMatrix<Cmplx> SigmaxyzSphericalSpinorLShell(long l,long k)
{
  int degl=2*l+1;
  SqMatrix<Cmplx> res(2*degl);

  double c1[2];
  double c2[2];

  Cmplx ci(0.,1.);

  int ind1=0;
  for(int kap1=l;kap1>=-l-1;kap1-=degl){
    int degj1=2*abs(kap1);
    for(int i1=-1;++i1<degj1;){
      int mu21=2*i1-(degj1-1);
      int ind2=0;
      c1[0]=dSign(1.,-kap1)*sqrt(0.5*(1.-mu21/(2.*kap1+1)));//s=1
      c1[1]=sqrt(0.5*(1.+mu21/(2.*kap1+1)));//s=-1
      for(int kap2=l;kap2>=-l-1;kap2-=degl){
	int degj2=2*abs(kap2);
	for(int i2=-1;++i2<degj2;){
	  int mu22=2*i2-(degj2-1);

	  c2[0]=dSign(1.,-kap2)*sqrt(0.5*(1.-mu22/(2.*kap2+1)));
	  c2[1]=sqrt(0.5*(1.+mu22/(2.*kap2+1)));

	  //myDebug()<<"i,j="<<ind1<<ind2<<"qn="<<(degj1-1)<<mu21<<","<<(degj2-1)<<mu22;

	  for(int s=-1;s<2;s+=2){
	    if(mu21==mu22 && k==2){
	      res(ind1,ind2)+=c1[(1-s)/2]*c2[(1-s)/2]*s;
	    }
	    if(mu21+2*s==mu22){
	      if(k==0) res(ind1,ind2)+=c1[(1+s)/2]*c2[(1-s)/2];
	      if(k==1) res(ind1,ind2)+=c1[(1+s)/2]*c2[(1-s)/2]*s*ci;
	    }
	  }

	  ind2++;
	}
      }
      ind1++;
    }
  }

  // myDebug()<<"LxJM";
  // printSqMatrix(pLxyz[SphericalSpinor][0],"(%6.3f, %6.3f) ");



  return res;

}

/*EOF*/ 
