/************************************************************************
 *
 * graphic package for FPLO
 * (C) Klaus Koepernik April '98
 *
 * **********************************************************************
 */
#include <stdlib.h>
#include <string.h>
#include <math.h> 
#include "tools.h"   
#include "fsymmetry.h" 
#include "strtype.h"   
#include "cbuffer.h" 
#include "merror.h" 


#ifndef LSIGN
#define LSIGN(x,y) ( (x)* ((y)>=0 ? 1:-1) )
#endif

#define VEC(a,i) ((a)+3*(i))
#define VECN(a,dim,i) ((a)+(dim)*(i))



/* ***********************************************************************
 *
 *    i,j: integer
 *    gcd: largest common divisor
 *    if(i == 0 && j==0) gcd=0 by definition, to allow iterative usage
 *
 * ***********************************************************************
 */

long gcd(long i,long j)
{
  long n,m,l;
  m=0;
  if(i==0 && j==0) {
    return 0;
  }
  m=i;
  n=j;
  if(abs(m)<abs(n)) {
    m=j;
    n=i;
  }
 
  while(n != 0) {
    
    l=n;
    n=m%n;
    m=l;
  }
  return m;
  
}
/* ***********************************************************************
 * make ratio (m/n) "teilerfremd", put sign onto m
 * ***********************************************************************
 */
static void normalize_ratio(long *m,long *n)
{
  long D;
  *m*=LSIGN(1,*n);
  *n=abs(*n);
  D=abs(gcd(*m,*n));
  *m/=D;
  *n/=D;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static long LMAX(long i,long j)
{
  return ( ((i)>j?i:j) );
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static long ROUND(double x)
{
  return ((x)>0. ?((long)(x+.5)) :((long)(x-.5)));
}
/* ***********************************************************************
 * v0=v1+v2*g mod unit cell
 * ***********************************************************************
 */
extern double firstcube(double *p,const double *cell,const double *rcell);

#define VECVEC(v0,v1,v2,g,cell,rcell)		\
  vec1((v0),(v1),(v2),(g));			\
  (void)firstcube((v0),(cell),(rcell)); 

/* #include "symmetryinternal.h"  */


#define  CPi  3.1415926535897932384626433832795028841968
#define  CPio2 1.5707963267948966192313216916397514420984
#define  CPi3o2 4.7123889803846898576939650749192543262952
#define  CPi5o3 5.2359877559829887307710723054658381403280
#define  CPi4o3 4.1887902047863909846168578443726705122624
#define  CPi2o3 2.0943951023931954923084289221863352561312
#define  CPio3 1.0471975511965977461542144610931676280656



/* forward prototypes */


static void extractgroupinfo(const char *sginfo,long num,tsSYMMETRY *sym );
static void makeunitcell(tsSYMMETRY *sym);
static void rotations(double *gen,long hex);
static void eulerangles(double *euler,long *eulerinv);
static void hexatob(double *b,const double *a);
static void hexbtoa(double *a,const double *b);
static void groupgeneration(tsSYMMETRY *sym,long usergen);


static long operationXYZindex(double *u);

static  const double pi180=.0174532925199432957692369076848861271344;
/* cartesian to hex sheer matrix */
/* hexagonal lattice unit vectors: */
/* a1[k]=carttohex[0+3*k] */
/* a2[k]=carttohex[1+3*k] */
/* a3[k]=carttohex[2+3*k] */


static  const double carttohex[]={
  .8660254037844386467,0.,0.,
  -0.5,1.0,0.,
  0.,0.,1.
};
/* invers of carttohex */
static  const double hextocart[]={
  1.15470053837925152902,0.,0.,
  .57735026918962576451,1.,0.,
  0.,0.,1.
};


/* table of XYZ operations symbols */
static const char *operationXYZsymbol[]={
  "-X-Y-Z","-Y-Z","+X-Y-Z","-X-Z","-Z",
  "+X-Z","-X+Y-Z","+Y-Z","+X+Y-Z","",
  "-X-Y","-Y","+X-Y","-X","",
  "+X","-X+Y","+Y","+X+Y","",
  "-X-Y+Z","-Y+Z","+X-Y+Z","-X+Z","+Z",
  "+X+Z","-X+Y+Z","+Y+Z","+X+Y+Z"
  
};
/* table of operations symbols */

const char *operationsymbols[]={
  /* -1 */
  "unknown operation",
  /* 0 */
  "E","C2(y)","C2(x)","C2(z)","C1/3(-111)","C1/3(1-11)","C2/3(-1-11)",
  /* 7 */
  "C2/3(111)","C2/3(-111)","C1/3(111)","C2/3(1-11)","C1/3(-1-11)",
  /* 12 */
  "C2(xy)","C3/4(z)","C1/4(z)","C2(x-y)","C2(xz)","C2(x-z)","C3/4(y)",
  /* 19 */
  "C1/4(y)","C2(y-z)","C3/4(x)","C2(yz)","C1/4(x)",
  /* 24 */
  "C5/6(z)","C2/3(z)","C1/3(z)","C1/6(z)","C2(xy[30])","C2(xy[60])",
  /* 30 */
  "C2(xy[-60])","C2(xy[-30])",
  /* 32 */
  "I","s(y)","s(x)","s(z)","S5/6(-111)","S5/6(1-11)","S1/6(-1-11)",
  /* 39 */
  "S1/6(111)","S1/6(-111)","S5/6(111)","S1/6(1-11)","S5/6(-1-11)",
  /* 44 */
  "s(xy)","S1/4(z)","S3/4(z)","s(x-y)","s(xz)","s(x-z)","S1/4(y)",
  /* 51 */
  "S3/4(y)","s(y-z)","S1/4(x)","s(yz)","S3/4(x)",
  "S1/3(z)","S1/6(z)","S5/6(z)","S5/3(z)","s(xy[30])","s(xy[60])",
  /* 62 */
  "s(xy[-60])","s(xy[-30])","unknown operation",
  /* 65 */
  "C2(90)","C2(0)","unknown operation","unknown operation","unknown operation",
  /* 70 */
  "unknown operation","unknown operation","unknown operation",
  "unknown operation","unknown operation","unknown operation",
  "unknown operation","unknown operation","unknown operation",
  "unknown operation",
  /* 80 */
  "unknown operation","unknown operation","unknown operation",
  "unknown operation","unknown operation","unknown operation",
  "unknown operation","unknown operation","unknown operation",
  "unknown operation",
  /* 90 */
  "unknown operation","unknown operation","unknown operation",
  "unknown operation","unknown operation","unknown operation",
  "unknown operation","s(90)","s(0)"
};

/* bravais is the trafo form primitive (p) to bravais lattice (b)*/
/* p(i)[k]=p[i+3*k] is i-th vector k-th cartesian coordinate */
/* b(1)[k]=sum(i) bravais[0+3*i+9*type] *p(i)[k] */
/* b(2)[k]=sum(i) bravais[1+3*i+9*type] *p(i)[k] */
/* b(3)[k]=sum(i) bravais[2+3*i+9*type] *p(i)[k] */

static const long ibravais[]={
/* simple */
  12 ,  0,  0
  , 0, 12,  0
  , 0,  0, 12
  /* body centered */
  ,-6,  6,  6
  , 6, -6,  6
  , 6,  6, -6
  /* face centered */
  , 0,  6,  6
  , 6,  0,  6
  , 6,  6,  0
  /* a-base centered */
  , 0,  0, 12
  , 6,  6,  0
  ,-6,  6,  0
  /* c-base centered */
  , 6, -6,  0
  , 6,  6,  0
  , 0,  0, 12
  /* rhomboedric */
  , 8, -4, -4
  , 4,  4, -8
  , 4,  4,  4
  /* b-base centered */
  ,-6,  6,  0
  , 0,  0, 12 
  , 6,  6,  0 
};

/* bravais types */
const char* bravaistypes1[]={"S","BC","FC","AFC","CFC","R","BFC","unknown"};


const char* bravaistypes_longname1[]={"Simple","Body-centered","Face-centered","A-Face-centered"
				      ,"C-Face-centered","Rhombohedral","B-Face-centered","unknown"};

static const long bravaistypes_size=sizeof(bravaistypes_longname1)/sizeof(char*);

static const long numberofbravaistypesofcryssystem1[]={3,1,2,2,5,2,1};
static const long bravaistypesofcryssystem1[7][5]={
  {0,1,2}, {0}, {0,1}, {0,5}, {0,1,2,3,4}, {0,4}, {0}, };

/* crystal systems */
static const char* crystalsystems[]={"C","H","T","TG","O","M","TC","unknown"};
static const long crystalsystem_size=sizeof(crystalsystems)/sizeof(char*);
/* crystal systems */
const char* crystalsystems_longname1[]={"Cubic","Hexagonal"
				       ,"Tetragonal","Trigonal","Orthorhombic"
				       ,"Monoclinic","Triclinic","Unknown"};


/* pointgroups */
static const char *pointgroups[]={
  "Unknown","C1","CI","C2","CS","C2H","D2","C2V","D2H","C4"
  ,"S4","C4H","D4","C4V","D2D(a)","D2D(b)","D4H","C3","C3I"
  ,"D3(a)","D3(b)","C3V(a)","C3V(b)","D3D(a)"
  ,"D3D(b)","C6","C3H","C6H","D6","C6V","D3H(a)","D3H(b)","D6H","T","TH"
  ,"O","TD","OH"
};
const long pointgroups_size=sizeof(pointgroups)/sizeof(char*);


/* multiplication table */
static const long in[64*64]={ 
  1,2,3,4,5,6,7,8,9,10,
  11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,
  33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,
  55,56,57,58,59,60,61,62,63,64,2,1,4,3,8,7,6,5,11,12,9,10,15,16,13,
  14,18,17,20,19,24,23,22,21,31,32,29,30,27,28,25,26,34,33,36,35,40,
  39,38,37,43,44,41,42,47,48,45,46,50,49,52,51,56,55,54,53,63,64,61,
  62,59,60,57,58,3,4,1,2,6,5,8,7,12,11,10,9,14,13,16,15,20,19,18,17,
  23,24,21,22,29,30,31,32,25,26,27,28,35,36,33,34,38,37,40,39,44,43,
  42,41,46,45,48,47,52,51,50,49,55,56,53,54,61,62,63,64,57,58,59,60,
  4,3,2,1,7,8,5,6,10,9,12,11,16,15,14,13,19,20,17,18,22,21,24,23,27,
  28,25,26,31,32,29,30,36,35,34,33,39,40,37,38,42,41,44,43,48,47,46,
  45,51,52,49,50,54,53,56,55,59,60,57,58,63,64,61,62,5,6,7,8,9,10,
  11,12,1,2,3,4,21,22,23,24,13,14,15,16,17,18,19,20,0,0,0,0,0,0,0,0,
  37,38,39,40,41,42,43,44,33,34,35,36,53,54,55,56,45,46,47,48,49,50,
  51,52,0,0,0,0,0,0,0,0,6,5,8,7,12,11,10,9,3,4,1,2,23,24,21,22,14,
  13,16,15,20,19,18,17,0,0,0,0,0,0,0,0,38,37,40,39,44,43,42,41,35,
  36,33,34,55,56,53,54,46,45,48,47,52,51,50,49,0,0,0,0,0,0,0,0,7,8,
  5,6,10,9,12,11,4,3,2,1,22,21,24,23,16,15,14,13,19,20,17,18,0,0,0,
  0,0,0,0,0,39,40,37,38,42,41,44,43,36,35,34,33,54,53,56,55,48,47,
  46,45,51,52,49,50,0,0,0,0,0,0,0,0,8,7,6,5,11,12,9,10,2,1,4,3,24,
  23,22,21,15,16,13,14,18,17,20,19,0,0,0,0,0,0,0,0,40,39,38,37,43,
  44,41,42,34,33,36,35,56,55,54,53,47,48,45,46,50,49,52,51,0,0,0,0,
  0,0,0,0,9,10,11,12,1,2,3,4,5,6,7,8,17,18,19,20,21,22,23,24,13,14,
  15,16,0,0,0,0,0,0,0,0,41,42,43,44,33,34,35,36,37,38,39,40,49,50,
  51,52,53,54,55,56,45,46,47,48,0,0,0,0,0,0,0,0,10,9,12,11,4,3,2,1,
  7,8,5,6,19,20,17,18,22,21,24,23,16,15,14,13,0,0,0,0,0,0,0,0,42,41,
  44,43,36,35,34,33,39,40,37,38,51,52,49,50,54,53,56,55,48,47,46,45,
  0,0,0,0,0,0,0,0,11,12,9,10,2,1,4,3,8,7,6,5,18,17,20,19,24,23,22,
  21,15,16,13,14,0,0,0,0,0,0,0,0,43,44,41,42,34,33,36,35,40,39,38,
  37,50,49,52,51,56,55,54,53,47,48,45,46,0,0,0,0,0,0,0,0,12,11,10,9,
  3,4,1,2,6,5,8,7,20,19,18,17,23,24,21,22,14,13,16,15,0,0,0,0,0,0,0,
  0,44,43,42,41,35,36,33,34,38,37,40,39,52,51,50,49,55,56,53,54,46,
  45,48,47,0,0,0,0,0,0,0,0,13,14,15,16,17,18,19,20,21,22,23,24,1,2,
  3,4,5,6,7,8,9,10,11,12,0,0,0,0,0,0,0,0,45,46,47,48,49,50,51,52,53,
  54,55,56,33,34,35,36,37,38,39,40,41,42,43,44,0,0,0,0,0,0,0,0,14,
  13,16,15,20,19,18,17,23,24,21,22,3,4,1,2,6,5,8,7,12,11,10,9,0,0,0,
  0,0,0,0,0,46,45,48,47,52,51,50,49,55,56,53,54,35,36,33,34,38,37,
  40,39,44,43,42,41,0,0,0,0,0,0,0,0,15,16,13,14,18,17,20,19,24,23,
  22,21,2,1,4,3,8,7,6,5,11,12,9,10,0,0,0,0,0,0,0,0,47,48,45,46,50,
  49,52,51,56,55,54,53,34,33,36,35,40,39,38,37,43,44,41,42,0,0,0,0,
  0,0,0,0,16,15,14,13,19,20,17,18,22,21,24,23,4,3,2,1,7,8,5,6,10,9,
  12,11,0,0,0,0,0,0,0,0,48,47,46,45,51,52,49,50,54,53,56,55,36,35,
  34,33,39,40,37,38,42,41,44,43,0,0,0,0,0,0,0,0,17,18,19,20,21,22,
  23,24,13,14,15,16,9,10,11,12,1,2,3,4,5,6,7,8,0,0,0,0,0,0,0,0,49,
  50,51,52,53,54,55,56,45,46,47,48,41,42,43,44,33,34,35,36,37,38,39,
  40,0,0,0,0,0,0,0,0,18,17,20,19,24,23,22,21,15,16,13,14,11,12,9,10,
  2,1,4,3,8,7,6,5,0,0,0,0,0,0,0,0,50,49,52,51,56,55,54,53,47,48,45,
  46,43,44,41,42,34,33,36,35,40,39,38,37,0,0,0,0,0,0,0,0,19,20,17,
  18,22,21,24,23,16,15,14,13,10,9,12,11,4,3,2,1,7,8,5,6,0,0,0,0,0,0,
  0,0,51,52,49,50,54,53,56,55,48,47,46,45,42,41,44,43,36,35,34,33,
  39,40,37,38,0,0,0,0,0,0,0,0,20,19,18,17,23,24,21,22,14,13,16,15,
  12,11,10,9,3,4,1,2,6,5,8,7,0,0,0,0,0,0,0,0,52,51,50,49,55,56,53,
  54,46,45,48,47,44,43,42,41,35,36,33,34,38,37,40,39,0,0,0,0,0,0,0,
  0,21,22,23,24,13,14,15,16,17,18,19,20,5,6,7,8,9,10,11,12,1,2,3,4,
  0,0,0,0,0,0,0,0,53,54,55,56,45,46,47,48,49,50,51,52,37,38,39,40,
  41,42,43,44,33,34,35,36,0,0,0,0,0,0,0,0,22,21,24,23,16,15,14,13,
  19,20,17,18,7,8,5,6,10,9,12,11,4,3,2,1,0,0,0,0,0,0,0,0,54,53,56,
  55,48,47,46,45,51,52,49,50,39,40,37,38,42,41,44,43,36,35,34,33,0,
  0,0,0,0,0,0,0,23,24,21,22,14,13,16,15,20,19,18,17,6,5,8,7,12,11,
  10,9,3,4,1,2,0,0,0,0,0,0,0,0,55,56,53,54,46,45,48,47,52,51,50,49,
  38,37,40,39,44,43,42,41,35,36,33,34,0,0,0,0,0,0,0,0,24,23,22,21,
  15,16,13,14,18,17,20,19,8,7,6,5,11,12,9,10,2,1,4,3,0,0,0,0,0,0,0,
  0,56,55,54,53,47,48,45,46,50,49,52,51,40,39,38,37,43,44,41,42,34,
  33,36,35,0,0,0,0,0,0,0,0,25,30,32,27,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,26,4,28,1,3,29,2,31,57,62,64,59,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,58,36,60,33,35,61,34,63,26,29,31,28,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,27,1,25,32,3,30,2,58,61,63,60,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,36,59,33,57,64,35,62,34,27,32,
  30,25,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,28,1,26,4,2,31,3,29,
  59,64,62,57,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,60,33,58,36,
  34,63,35,61,28,31,29,26,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,
  25,4,27,30,2,32,3,60,63,61,58,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,33,57,36,59,62,34,64,35,29,26,28,31,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,30,2,32,3,1,25,4,27,61,58,60,63,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,62,34,64,35,33,57,36,59,30,25,27,32,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,31,3,29,28,1,26,4,62,57,59,64,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,34,63,35,61,60,33,58,36,31,28,
  26,29,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,3,30,2,4,27,1,25,
  63,60,58,61,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,35,62,34,
  36,59,33,57,32,27,25,30,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,3,
  29,2,31,26,4,28,1,64,59,57,62,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,35,61,34,63,58,36,60,33,33,34,35,36,37,38,39,40,41,42,43,44,
  45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,1,2,3,
  4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,
  28,29,30,31,32,34,33,36,35,40,39,38,37,43,44,41,42,47,48,45,46,50,
  49,52,51,56,55,54,53,63,64,61,62,59,60,57,58,2,1,4,3,8,7,6,5,11,
  12,9,10,15,16,13,14,18,17,20,19,24,23,22,21,31,32,29,30,27,28,25,
  26,35,36,33,34,38,37,40,39,44,43,42,41,46,45,48,47,52,51,50,49,55,
  56,53,54,61,62,63,64,57,58,59,60,3,4,1,2,6,5,8,7,12,11,10,9,14,13,
  16,15,20,19,18,17,23,24,21,22,29,30,31,32,25,26,27,28,36,35,34,33,
  39,40,37,38,42,41,44,43,48,47,46,45,51,52,49,50,54,53,56,55,59,60,
  57,58,63,64,61,62,4,3,2,1,7,8,5,6,10,9,12,11,16,15,14,13,19,20,17,
  18,22,21,24,23,27,28,25,26,31,32,29,30,37,38,39,40,41,42,43,44,33,
  34,35,36,53,54,55,56,45,46,47,48,49,50,51,52,0,0,0,0,0,0,0,0,5,6,
  7,8,9,10,11,12,1,2,3,4,21,22,23,24,13,14,15,16,17,18,19,20,0,0,0,
  0,0,0,0,0,38,37,40,39,44,43,42,41,35,36,33,34,55,56,53,54,46,45,
  48,47,52,51,50,49,0,0,0,0,0,0,0,0,6,5,8,7,12,11,10,9,3,4,1,2,23,
  24,21,22,14,13,16,15,20,19,18,17,0,0,0,0,0,0,0,0,39,40,37,38,42,
  41,44,43,36,35,34,33,54,53,56,55,48,47,46,45,51,52,49,50,0,0,0,0,
  0,0,0,0,7,8,5,6,10,9,12,11,4,3,2,1,22,21,24,23,16,15,14,13,19,20,
  17,18,0,0,0,0,0,0,0,0,40,39,38,37,43,44,41,42,34,33,36,35,56,55,
  54,53,47,48,45,46,50,49,52,51,0,0,0,0,0,0,0,0,8,7,6,5,11,12,9,10,
  2,1,4,3,24,23,22,21,15,16,13,14,18,17,20,19,0,0,0,0,0,0,0,0,41,42,
  43,44,33,34,35,36,37,38,39,40,49,50,51,52,53,54,55,56,45,46,47,48,
  0,0,0,0,0,0,0,0,9,10,11,12,1,2,3,4,5,6,7,8,17,18,19,20,21,22,23,
  24,13,14,15,16,0,0,0,0,0,0,0,0,42,41,44,43,36,35,34,33,39,40,37,
  38,51,52,49,50,54,53,56,55,48,47,46,45,0,0,0,0,0,0,0,0,10,9,12,11,
  4,3,2,1,7,8,5,6,19,20,17,18,22,21,24,23,16,15,14,13,0,0,0,0,0,0,0,
  0,43,44,41,42,34,33,36,35,40,39,38,37,50,49,52,51,56,55,54,53,47,
  48,45,46,0,0,0,0,0,0,0,0,11,12,9,10,2,1,4,3,8,7,6,5,18,17,20,19,
  24,23,22,21,15,16,13,14,0,0,0,0,0,0,0,0,44,43,42,41,35,36,33,34,
  38,37,40,39,52,51,50,49,55,56,53,54,46,45,48,47,0,0,0,0,0,0,0,0,
  12,11,10,9,3,4,1,2,6,5,8,7,20,19,18,17,23,24,21,22,14,13,16,15,0,
  0,0,0,0,0,0,0,45,46,47,48,49,50,51,52,53,54,55,56,33,34,35,36,37,
  38,39,40,41,42,43,44,0,0,0,0,0,0,0,0,13,14,15,16,17,18,19,20,21,
  22,23,24,1,2,3,4,5,6,7,8,9,10,11,12,0,0,0,0,0,0,0,0,46,45,48,47,
  52,51,50,49,55,56,53,54,35,36,33,34,38,37,40,39,44,43,42,41,0,0,0,
  0,0,0,0,0,14,13,16,15,20,19,18,17,23,24,21,22,3,4,1,2,6,5,8,7,12,
  11,10,9,0,0,0,0,0,0,0,0,47,48,45,46,50,49,52,51,56,55,54,53,34,33,
  36,35,40,39,38,37,43,44,41,42,0,0,0,0,0,0,0,0,15,16,13,14,18,17,
  20,19,24,23,22,21,2,1,4,3,8,7,6,5,11,12,9,10,0,0,0,0,0,0,0,0,48,
  47,46,45,51,52,49,50,54,53,56,55,36,35,34,33,39,40,37,38,42,41,44,
  43,0,0,0,0,0,0,0,0,16,15,14,13,19,20,17,18,22,21,24,23,4,3,2,1,7,
  8,5,6,10,9,12,11,0,0,0,0,0,0,0,0,49,50,51,52,53,54,55,56,45,46,47,
  48,41,42,43,44,33,34,35,36,37,38,39,40,0,0,0,0,0,0,0,0,17,18,19,
  20,21,22,23,24,13,14,15,16,9,10,11,12,1,2,3,4,5,6,7,8,0,0,0,0,0,0,
  0,0,50,49,52,51,56,55,54,53,47,48,45,46,43,44,41,42,34,33,36,35,
  40,39,38,37,0,0,0,0,0,0,0,0,18,17,20,19,24,23,22,21,15,16,13,14,
  11,12,9,10,2,1,4,3,8,7,6,5,0,0,0,0,0,0,0,0,51,52,49,50,54,53,56,
  55,48,47,46,45,42,41,44,43,36,35,34,33,39,40,37,38,0,0,0,0,0,0,0,
  0,19,20,17,18,22,21,24,23,16,15,14,13,10,9,12,11,4,3,2,1,7,8,5,6,
  0,0,0,0,0,0,0,0,52,51,50,49,55,56,53,54,46,45,48,47,44,43,42,41,
  35,36,33,34,38,37,40,39,0,0,0,0,0,0,0,0,20,19,18,17,23,24,21,22,
  14,13,16,15,12,11,10,9,3,4,1,2,6,5,8,7,0,0,0,0,0,0,0,0,53,54,55,
  56,45,46,47,48,49,50,51,52,37,38,39,40,41,42,43,44,33,34,35,36,0,
  0,0,0,0,0,0,0,21,22,23,24,13,14,15,16,17,18,19,20,5,6,7,8,9,10,11,
  12,1,2,3,4,0,0,0,0,0,0,0,0,54,53,56,55,48,47,46,45,51,52,49,50,39,
  40,37,38,42,41,44,43,36,35,34,33,0,0,0,0,0,0,0,0,22,21,24,23,16,
  15,14,13,19,20,17,18,7,8,5,6,10,9,12,11,4,3,2,1,0,0,0,0,0,0,0,0,
  55,56,53,54,46,45,48,47,52,51,50,49,38,37,40,39,44,43,42,41,35,36,
  33,34,0,0,0,0,0,0,0,0,23,24,21,22,14,13,16,15,20,19,18,17,6,5,8,7,
  12,11,10,9,3,4,1,2,0,0,0,0,0,0,0,0,56,55,54,53,47,48,45,46,50,49,
  52,51,40,39,38,37,43,44,41,42,34,33,36,35,0,0,0,0,0,0,0,0,24,23,
  22,21,15,16,13,14,18,17,20,19,8,7,6,5,11,12,9,10,2,1,4,3,0,0,0,0,
  0,0,0,0,57,62,64,59,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,58,36,
  60,33,35,61,34,63,25,30,32,27,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,26,4,28,1,3,29,2,31,58,61,63,60,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,36,59,33,57,64,35,62,34,26,29,31,28,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,4,27,1,25,32,3,30,2,59,64,62,57,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,60,33,58,36,34,63,35,61,27,32,30,25,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,28,1,26,4,2,31,3,29,60,63,61,
  58,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,33,57,36,59,62,34,64,
  35,28,31,29,26,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,25,4,27,
  30,2,32,3,61,58,60,63,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,62,
  34,64,35,33,57,36,59,29,26,28,31,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,30,2,32,3,1,25,4,27,62,57,59,64,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,34,63,35,61,60,33,58,36,30,25,27,32,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,2,31,3,29,28,1,26,4,63,60,58,61,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,35,62,34,36,59,33,57,31,28,26,29,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,3,30,2,4,27,1,25,64,59,
  57,62,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,35,61,34,63,58,36,
  60,33,32,27,25,30,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,3,29,2,
  31,26,4,28,1 
};


/* all posible point group rotations coded as euler angles 
   must be consistent with pgrotations
 */

static const double  pgeuler[3*32]={
  0.0,0.0,0.0, 0.0,CPi,0.0, 0.0,CPi,CPi, CPi,0.0,0.0, CPi,CPio2,CPio2, 0.0,CPio2,CPi3o2, CPi,CPio2,CPi3o2, 0.0,CPio2,CPio2,
  CPio2,CPio2,0.0, CPio2 ,CPio2 ,CPi, CPi3o2,CPio2,CPi, CPi3o2,CPio2,0.0, CPi3o2,CPi,0.0, CPio2,0.0,0.0, CPi3o2,0.0,0.0,
  CPio2,CPi,0.0, 0.0,CPio2,CPi, CPi,CPio2,0.0, 0.0,CPio2,0.0, CPi,CPio2,CPi, CPi3o2,CPio2,CPi3o2, CPi3o2,CPio2,CPio2, CPio2,CPio2,CPio2,
  CPio2,CPio2,CPi3o2,
  CPio3,0.0,0.0, CPi2o3,0.0,0.0, CPi4o3,0.0,0.0, CPi5o3,0.0,0.0, CPi4o3,CPi,0.0, CPi5o3,CPi,0.0, CPio3,CPi,0.0, CPi2o3,CPi,0.0
};
/* all posible point group rotations */
static const double  pgrotations[]=
{
  /* 1 */
  1.,0.,0.,       0.,1.,0.,  0.,0.,1.,
  -1.0,0.0,0.0,    0.0,1.0,0.0,    0.0,0.0,-1.0,
  1.0,0.0,0.0,    0.0,-1.0,0.0,    0.0,0.0,-1.0,
  -1.0,0.0,0.0,    0.0,-1.0,0.0,    0.0,0.0,1.0,
  0.0,-1.0,0.0,    0.0,0.0,1.0,    -1.0,0.0,0.0,
  0.0,-1.0,0.0,    0.0,0.0,-1.0,    1.0,0.0,0.0,
  0.0,1.0,0.0,    0.0,0.0,-1.0,    -1.0,0.0,0.0,
  0.0,1.0,0.0,    0.0,0.0,1.0,      1.0,0.0,0.0,
  0.0,0.0,-1.0,    -1.0,0.0,0.0,    0.0,1.0,0.0,
  0.0,0.0,1.0,    1.0,0.0,0.0,    0.0,1.0,0.0,
  /* 11 */
  0.0,0.0,1.0,    -1.0,0.0,0.0,    0.0,-1.0,0.0,
  0.0,0.0,-1.0,    1.0,0.0,0.0,    0.0,-1.0,0.0,
  0.0,1.0,0.0,    1.0,0.0,0.0,    0.0,0.0,-1.0,
  0.0,1.0,0.0,    -1.0,0.0,0.0,    0.0,0.0,1.0,
  0.0,-1.0,0.0,    1.0,0.0,0.0,    0.0,0.0,1.0,
  0.0,-1.0,0.0,    -1.0,0.0,0.0,    0.0,0.0,-1.0,
  0.0,0.0,1.0,    0.0,-1.0,0.0,    1.0,0.0,0.0,
  0.0,0.0,-1.0,    0.0,-1.0,0.0,    -1.0,0.0,0.0,
  0.0,0.0,-1.0,    0.0,1.0,0.0,    1.0,0.0,0.0,
  0.0,0.0,1.0,    0.0,1.0,0.0,    -1.0,0.0,0.0,
  /* 21 */
  -1.0,0.0,0.0,    0.0,0.0,-1.0,    0.0,-1.0,0.0,
  1.0,0.0,0.0,    0.0,0.0,1.0,    0.0,-1.0,0.0,
  -1.0,0.0,0.0,    0.0,0.0,1.0,    0.0,1.0,0.0,
  1.0,0.0,0.0,    0.0,0.0,-1.0,    0.0,1.0,0.0,
  /* HEX 25 */
  .50,.866025403784438646763723170750,0.0,  
  -.866025403784438646763723170750,0.50,0.0, 
  0.0,0.0,1.0,
  -.50,.866025403784438646763723170750,0.0, 
  -.866025403784438646763723170750,-0.50,0.0,
  0.0,0.0,1.0,
  -.50,-.866025403784438646763723170750,0.0, 
  .866025403784438646763723170750,-0.50,0.0, 
  0.0,0.0,1.0,
  .50,-.866025403784438646763723170750,0.0,  
  .866025403784438646763723170750,0.50,0.0, 
  0.0,0.0,1.0,
  .50,.866025403784438646763723170750,0.0,   
  .866025403784438646763723170750,-0.50,0.0,  
  0.0,0.0,-1.0,
  -.50,.866025403784438646763723170750,0.0, 
  .866025403784438646763723170750,0.50,0.0,  
  0.0,0.0,-1.0,
  /* 31 */
  -.50,-.866025403784438646763723170750,0.0,
  -.866025403784438646763723170750,0.50,0.0, 
  0.0,0.0,-1.0,
  .50,-.866025403784438646763723170750,0.0,  
  -.866025403784438646763723170750,-0.50,0.0,  
  0.0,0.0,-1.0
};

/* 
 * which operations are in the pointgroups (calculated by 
 * getpointgroupmasksinternal())
 *
*/

/* ***********************************************************************
 *
 * ***********************************************************************
 */
static const char *spcgrp_num(const char* sginfo) { return sginfo+0; }
const char *spcgrp_name(const char* sginfo) { return sginfo+5; }
static const char *spcgrp_pgrname(const char* sginfo) { return sginfo+15; }
static const char *spcgrp_schoenfliesnum(const char* sginfo) { return sginfo+23; }
/* static const char *spcgrp_order(const char* sginfo) { return sginfo+27; } */
static const char *spcgrp_pgrnum(const char* sginfo) { return sginfo+31; }
static const char *spcgrp_inversion(const char* sginfo) { return sginfo+35; }
/* static const char *spcgrp_?(const char* sginfo) { return sginfo+38; } */
const char *spcgrp_crystallsystem(const char* sginfo) { return sginfo+42; }
static const char *spcgrp_gennum(const char* sginfo) { return sginfo+45; }
static const char *spcgrp_gen(const char* sginfo,long i) { return sginfo+48+i*8; }

#define SPGR_DESCRIPTOR_LENGTH 73
static const char spacegrouptable[230][SPGR_DESCRIPTOR_LENGTH]={
  /* num spgr       pgr  index of  spgr in this pgr
   *                      Schoenflies number           
   *                          number of operations  
   *                                 number of pgr including orientation type
   *                                 for pgrs with different orientations in 
   *                                 different spgr: is counted differently 
   *                                 ouside
   *                                     inversion(0: yes, 1: no)
   *                                         ?
   *                                              crystall system
   *                                                  number of generators
   *                                                      gen1   gen2   gen3
   *                    gen=index(2 byte) 12*tx(1 byte) 12*ty(1 byte) 12*tz(1)
   * tx,ty,tz being the nontrivial translations (nonsymmorphic groups)
   * in unit coordinates
   */

  "1\0   P1\0       C1\0     1\0  1\0  1\0  1\0 0\0  7\0 1\0 010000\0 \0       \0       ",
  "2\0   PB1\0      CI\0     1\0  2\0  2\0  0\0 0\0  7\0 1\0 330000\0 \0       \0       ",
  "3\0   P2\0       C2\0     1\0  2\0  3\0  1\0 0\0  6\0 1\0 020000\0 \0       \0       ",
  "4\0   P21\0      C2\0     2\0  2\0  3\0  1\0 0\0  6\0 1\0 020600\0 \0       \0       ",
  "5\0   C2\0       C2\0     3\0  2\0  3\0  1\0 0\0  6\0 1\0 020000\0 \0       \0       ",

  "6\0   PM\0       CS\0     1\0  2\0  4\0  1\0 0\0  6\0 1\0 340000\0 \0       \0       ",
  "7\0   PC\0       CS\0     2\0  2\0  4\0  1\0 0\0  6\0 1\0 340006\0 \0       \0       ",
  "8\0   CM\0       CS\0     3\0  2\0  4\0  1\0 0\0  6\0 1\0 340000\0 \0       \0       ",
  "9\0   CC\0       CS\0     4\0  2\0  4\0  1\0 0\0  6\0 1\0 340006\0 \0       \0       ",
  "10\0  P2/M\0     C2H\0    1\0  4\0  5\0  0\0 0\0  6\0 2\0 020000\0 330000\0 \0       ",
  "11\0  P21/M\0    C2H\0    2\0  4\0  5\0  0\0 0\0  6\0 2\0 020600\0 330000\0 \0       ",
  "12\0  C2/M\0     C2H\0    3\0  4\0  5\0  0\0 0\0  6\0 2\0 020000\0 330000\0 \0       ",
  "13\0  P2/C\0     C2H\0    4\0  4\0  5\0  0\0 0\0  6\0 2\0 020006\0 330000\0 \0       ",
  "14\0  P21/C\0    C2H\0    5\0  4\0  5\0  0\0 0\0  6\0 2\0 020606\0 330000\0 \0       ",
  "15\0  C2/C\0     C2H\0    6\0  4\0  5\0  0\0 0\0  6\0 2\0 020006\0 330000\0 \0       ",
  "16\0  P222\0     D2\0     1\0  4\0  6\0  1\0 0\0  5\0 2\0 040000\0 020000\0 \0       ",
  "17\0  P2221\0    D2\0     2\0  4\0  6\0  1\0 0\0  5\0 2\0 040006\0 020006\0 \0       ",
  "18\0  P21212\0   D2\0     3\0  4\0  6\0  1\0 0\0  5\0 2\0 040000\0 026600\0 \0       ",
  "19\0  P212121\0  D2\0     4\0  4\0  6\0  1\0 0\0  5\0 2\0 046006\0 020606\0 \0       ",
  "20\0  C2221\0    D2\0     5\0  4\0  6\0  1\0 0\0  5\0 2\0 040006\0 020006\0 \0       ",
  "21\0  C222\0     D2\0     6\0  4\0  6\0  1\0 0\0  5\0 2\0 040000\0 020000\0 \0       ",
  "22\0  F222\0     D2\0     7\0  4\0  6\0  1\0 0\0  5\0 2\0 040000\0 020000\0 \0       ",
  "23\0  I222\0     D2\0     8\0  4\0  6\0  1\0 0\0  5\0 2\0 040000\0 020000\0 \0       ",
  "24\0  I212121\0  D2\0     9\0  4\0  6\0  1\0 0\0  5\0 2\0 046006\0 026000\0 \0       ",
  "25\0  PMM2\0     C2V\0    1\0  4\0  7\0  1\0 0\0  5\0 2\0 040000\0 340000\0 \0       ",
  "26\0  PMC21\0    C2V\0    2\0  4\0  7\0  1\0 0\0  5\0 2\0 040006\0 340006\0 \0       ",
  "27\0  PCC2\0     C2V\0    3\0  4\0  7\0  1\0 0\0  5\0 2\0 040000\0 340006\0 \0       ",
  "28\0  PMA2\0     C2V\0    4\0  4\0  7\0  1\0 0\0  5\0 2\0 040000\0 346000\0 \0       ",
  "29\0  PCA21\0    C2V\0    5\0  4\0  7\0  1\0 0\0  5\0 2\0 040006\0 346000\0 \0       ",
  "30\0  PNC2\0     C2V\0    6\0  4\0  7\0  1\0 0\0  5\0 2\0 040000\0 340606\0 \0       ",
  "31\0  PMN21\0    C2V\0    7\0  4\0  7\0  1\0 0\0  5\0 2\0 046006\0 346006\0 \0       ",
  "32\0  PBA2\0     C2V\0    8\0  4\0  7\0  1\0 0\0  5\0 2\0 040000\0 346600\0 \0       ",
  "33\0  PNA21\0    C2V\0    9\0  4\0  7\0  1\0 0\0  5\0 2\0 040006\0 346600\0 \0       ",
  "34\0  PNN2\0     C2V\0    10\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 346606\0 \0       ",
  "35\0  CMM2\0     C2V\0    11\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 340000\0 \0       ",
  "36\0  CMC21\0    C2V\0    12\0 4\0  7\0  1\0 0\0  5\0 2\0 040006\0 340006\0 \0       ",
  "37\0  CCC2\0     C2V\0    13\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 340006\0 \0       ",
  "38\0  AMM2\0     C2V\0    14\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 340000\0 \0       ",
  "39\0  ABM2\0     C2V\0    15\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 340600\0 \0       ",
  "40\0  AMA2\0     C2V\0    16\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 346000\0 \0       ",
  "41\0  ABA2\0     C2V\0    17\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 346600\0 \0       ",
  "42\0  FMM2\0     C2V\0    18\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 340000\0 \0       ",
  "43\0  FDD2\0     C2V\0    19\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 343303\0 \0       ",
  "44\0  IMM2\0     C2V\0    20\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 340000\0 \0       ",
  "45\0  IBA2\0     C2V\0    21\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 340006\0 \0       ",
  "46\0  IMA2\0     C2V\0    22\0 4\0  7\0  1\0 0\0  5\0 2\0 040000\0 346000\0 \0       ",
  "47\0  PMMM\0     D2H\0    1\0  8\0  8\0  0\0 0\0  5\0 3\0 040000\0 020000\0 330000\0 ",
  "48\0  PNNN\0     D2H\0    2\0  8\0  8\0  0\0 0\0  5\0 3\0 046600\0 026006\0 330000\0 ",
  "49\0  PCCM\0     D2H\0    3\0  8\0  8\0  0\0 0\0  5\0 3\0 040000\0 020006\0 330000\0 ",
  "50\0  PBAN\0     D2H\0    4\0  8\0  8\0  0\0 0\0  5\0 3\0 046600\0 026000\0 330000\0 ",
  "51\0  PMMA\0     D2H\0    5\0  8\0  8\0  0\0 0\0  5\0 3\0 046000\0 020000\0 330000\0 ",
  "52\0  PNNA\0     D2H\0    6\0  8\0  8\0  0\0 0\0  5\0 3\0 046000\0 026606\0 330000\0 ",
  "53\0  PMNA\0     D2H\0    7\0  8\0  8\0  0\0 0\0  5\0 3\0 046006\0 026006\0 330000\0 ",
  "54\0  PCCA\0     D2H\0    8\0  8\0  8\0  0\0 0\0  5\0 3\0 046000\0 020006\0 330000\0 ",
  "55\0  PBAM\0     D2H\0    9\0  8\0  8\0  0\0 0\0  5\0 3\0 040000\0 026600\0 330000\0 ",
  "56\0  PCCN\0     D2H\0    10\0 8\0  8\0  0\0 0\0  5\0 3\0 046600\0 020606\0 330000\0 ",
  "57\0  PBCM\0     D2H\0    11\0 8\0  8\0  0\0 0\0  5\0 3\0 040006\0 020606\0 330000\0 ",
  "58\0  PNNM\0     D2H\0    12\0 8\0  8\0  0\0 0\0  5\0 3\0 040000\0 026606\0 330000\0 ",
  "59\0  PMMN\0     D2H\0    13\0 8\0  8\0  0\0 0\0  5\0 3\0 046600\0 020600\0 330000\0 ",
  "60\0  PBCN\0     D2H\0    14\0 8\0  8\0  0\0 0\0  5\0 3\0 046606\0 020006\0 330000\0 ",
  "61\0  PBCA\0     D2H\0    15\0 8\0  8\0  0\0 0\0  5\0 3\0 046006\0 020606\0 330000\0 ",
  "62\0  PNMA\0     D2H\0    16\0 8\0  8\0  0\0 0\0  5\0 3\0 046006\0 020600\0 330000\0 ",
  "63\0  CMCM\0     D2H\0    17\0 8\0  8\0  0\0 0\0  5\0 3\0 040006\0 020006\0 330000\0 ",
  "64\0  CMCA\0     D2H\0    18\0 8\0  8\0  0\0 0\0  5\0 3\0 040606\0 020606\0 330000\0 ",
  "65\0  CMMM\0     D2H\0    19\0 8\0  8\0  0\0 0\0  5\0 3\0 040000\0 020000\0 330000\0 ",
  "66\0  CCCM\0     D2H\0    20\0 8\0  8\0  0\0 0\0  5\0 3\0 040000\0 020006\0 330000\0 ",
  "67\0  CMMA\0     D2H\0    21\0 8\0  8\0  0\0 0\0  5\0 3\0 046000\0 026000\0 330000\0 ",
  "68\0  CCCA\0     D2H\0    22\0 8\0  8\0  0\0 0\0  5\0 3\0 040600\0 026606\0 330000\0 ",
  "69\0  FMMM\0     D2H\0    23\0 8\0  8\0  0\0 0\0  5\0 3\0 040000\0 020000\0 330000\0 ",
  "70\0  FDDD\0     D2H\0    24\0 8\0  8\0  0\0 0\0  5\0 3\0 043300\0 023003\0 330000\0 ",
  "71\0  IMMM\0     D2H\0    25\0 8\0  8\0  0\0 0\0  5\0 3\0 040000\0 020000\0 330000\0 ",
  "72\0  IBAM\0     D2H\0    26\0 8\0  8\0  0\0 0\0  5\0 3\0 040000\0 020006\0 330000\0 ",
  "73\0  IBCA\0     D2H\0    27\0 8\0  8\0  0\0 0\0  5\0 3\0 046006\0 026000\0 330000\0 ",
  "74\0  IMMA\0     D2H\0    28\0 8\0  8\0  0\0 0\0  5\0 3\0 040600\0 020600\0 330000\0 ",
  "75\0  P4\0       C4\0     1\0  4\0  9\0  1\0 48\0 3\0 1\0 150000\0 \0 \0             ",
  "76\0  P41\0      C4\0     2\0  4\0  9\0  1\0 48\0 3\0 1\0 150003\0 \0 \0             ",
  "77\0  P42\0      C4\0     3\0  4\0  9\0  1\0 48\0 3\0 1\0 150006\0 \0 \0             ",
  "78\0  P43\0      C4\0     4\0  4\0  9\0  1\0 48\0 3\0 1\0 150009\0 \0 \0             ",
  "79\0  I4\0       C4\0     5\0  4\0  9\0  1\0 48\0 3\0 1\0 150000\0 \0 \0             ",
  "80\0  I41\0      C4\0     6\0  4\0  9\0  1\0 48\0 3\0 1\0 150603\0 \0 \0             ",
  "81\0  PB4\0      S4\0     1\0  4\0  10\0 1\0 48\0 3\0 1\0 470000\0 \0 \0             ",
  "82\0  IB4\0      S4\0     2\0  4\0  10\0 1\0 48\0 3\0 1\0 470000\0 \0 \0             ",
  "83\0  P4/M\0     C4H\0    1\0  8\0  11\0 0\0 48\0 3\0 2\0 150000\0 330000\0 \0       ",
  "84\0  P42/M\0    C4H\0    2\0  8\0  11\0 0\0 48\0 3\0 2\0 150006\0 330000\0 \0       ",
  "85\0  P4/N\0     C4H\0    3\0  8\0  11\0 0\0 48\0 3\0 2\0 156000\0 330000\0 \0       ",
  "86\0  P42/N\0    C4H\0    4\0  8\0  11\0 0\0 48\0 3\0 2\0 150606\0 330000\0 \0       ",
  "87\0  I4/M\0     C4H\0    5\0  8\0  11\0 0\0 48\0 3\0 2\0 150000\0 330000\0 \0       ",
  "88\0  I41/A\0    C4H\0    6\0  8\0  11\0 0\0 48\0 3\0 2\0 159303\0 330000\0 \0       ",
  "89\0  P422\0     D4\0     1\0  8\0  12\0 1\0 0\0  3\0 2\0 150000\0 030000\0 \0       ",
  "90\0  P4212\0    D4\0     2\0  8\0  12\0 1\0 0\0  3\0 2\0 156600\0 036600\0 \0       ",
  "91\0  P4122\0    D4\0     3\0  8\0  12\0 1\0 0\0  3\0 2\0 150003\0 030006\0 \0       ",
  /*  " 92\0 P41212\0   D4\0  4\0 8\0 12\0 1\0  0\0 3\0   2\0 156603\0 036606\0 \0      ", */
  /*      this seems to be incorrect                              ^ */
  /*      So, I correted it !                                    /|\ */
  /*                                                          ____| */
  "92\0  P41212\0   D4\0     4\0  8\0  12\0 1\0 0\0  3\0 2\0 156603\0 036609\0 \0       ",
  "93\0  P4222\0    D4\0     5\0  8\0  12\0 1\0 0\0  3\0 2\0 150006\0 030000\0 \0       ",
  "94\0  P42212\0   D4\0     6\0  8\0  12\0 1\0 0\0  3\0 2\0 156606\0 036606\0 \0       ",
  "95\0  P4322\0    D4\0     7\0  8\0  12\0 1\0 0\0  3\0 2\0 150009\0 030006\0 \0       ",
  "96\0  P43212\0   D4\0     8\0  8\0  12\0 1\0 0\0  3\0 2\0 156609\0 036603\0 \0       ",
  "97\0  I422\0     D4\0     9\0  8\0  12\0 1\0 0\0  3\0 2\0 150000\0 030000\0 \0       ",
  "98\0  I4122\0    D4\0     10\0 8\0  12\0 1\0 0\0  3\0 2\0 150603\0 030603\0 \0       ",
  "99\0  P4MM\0     C4V\0    1\0  8\0  13\0 1\0 0\0  3\0 2\0 150000\0 350000\0 \0       ",
  "100\0 P4BM\0     C4V\0    2\0  8\0  13\0 1\0 0\0  3\0 2\0 150000\0 356600\0 \0       ",
  "101\0 P42CM\0    C4V\0    3\0  8\0  13\0 1\0 0\0  3\0 2\0 150006\0 350006\0 \0       ",
  "102\0 P42NM\0    C4V\0    4\0  8\0  13\0 1\0 0\0  3\0 2\0 156606\0 356606\0 \0       ",
  "103\0 P4CC\0     C4V\0    5\0  8\0  13\0 1\0 0\0  3\0 2\0 150000\0 350006\0 \0       ",
  "104\0 P4NC\0     C4V\0    6\0  8\0  13\0 1\0 0\0  3\0 2\0 150000\0 356606\0 \0       ",
  "105\0 P42MC\0    C4V\0    7\0  8\0  13\0 1\0 0\0  3\0 2\0 150006\0 350000\0 \0       ",
  "106\0 P42BC\0    C4V\0    8\0  8\0  13\0 1\0 0\0  3\0 2\0 150006\0 356600\0 \0       ",
  "107\0 I4MM\0     C4V\0    9\0  8\0  13\0 1\0 0\0  3\0 2\0 150000\0 350000\0 \0       ",
  "108\0 I4CM\0     C4V\0    10\0 8\0  13\0 1\0 0\0  3\0 2\0 150000\0 356600\0 \0       ",
  "109\0 I41MD\0    C4V\0    11\0 8\0  13\0 1\0 0\0  3\0 2\0 150603\0 350000\0 \0       ",
  "110\0 I41CD\0    C4V\0    12\0 8\0  13\0 1\0 0\0  3\0 2\0 150603\0 350006\0 \0       ",
  "111\0 PB42M\0    D2D(a)\0 1\0  8\0  14\0 1\0 0\0  3\0 2\0 470000\0 030000\0 \0       ",
  "112\0 PB42C\0    D2D(a)\0 2\0  8\0  14\0 1\0 0\0  3\0 2\0 470000\0 030006\0 \0       ",
  "113\0 PB421M\0   D2D(a)\0 3\0  8\0  14\0 1\0 0\0  3\0 2\0 470000\0 036600\0 \0       ",
  "114\0 PB421C\0   D2D(a)\0 4\0  8\0  14\0 1\0 0\0  3\0 2\0 470000\0 036606\0 \0       ",
  "115\0 PB4M2\0    D2D(b)\0 5\0  8\0  15\0 1\0 0\0  3\0 2\0 470000\0 350000\0 \0       ",
  "116\0 PB4C2\0    D2D(b)\0 6\0  8\0  15\0 1\0 0\0  3\0 2\0 470000\0 350006\0 \0       ",
  "117\0 PB4B2\0    D2D(b)\0 7\0  8\0  15\0 1\0 0\0  3\0 2\0 470000\0 356600\0 \0       ",
  "118\0 PB4N2\0    D2D(b)\0 8\0  8\0  15\0 1\0 0\0  3\0 2\0 470000\0 356606\0 \0       ",
  "119\0 IB4M2\0    D2D(b)\0 9\0  8\0  15\0 1\0 0\0  3\0 2\0 470000\0 350000\0 \0       ",
  "120\0 IB4C2\0    D2D(b)\0 10\0 8\0  15\0 1\0 0\0  3\0 2\0 470000\0 350006\0 \0       ",
  "121\0 IB42M\0    D2D(a)\0 11\0 8\0  14\0 1\0 0\0  3\0 2\0 470000\0 030000\0 \0       ",
  "122\0 IB42D\0    D2D(a)\0 12\0 8\0  14\0 1\0 0\0  3\0 2\0 470000\0 030603\0 \0       ",
  "123\0 P4/MMM\0   D4H\0    1\0  16\0 16\0 0\0 0\0  3\0 3\0 150000\0 350000\0 330000\0 ",
  "124\0 P4/MCC\0   D4H\0    2\0  16\0 16\0 0\0 0\0  3\0 3\0 150000\0 350006\0 330000\0 ",
  "125\0 P4/NBM\0   D4H\0    3\0  16\0 16\0 0\0 0\0  3\0 3\0 156000\0 350600\0 330000\0 ",
  "126\0 P4/NNC\0   D4H\0    4\0  16\0 16\0 0\0 0\0  3\0 3\0 156000\0 350606\0 330000\0 ",
  "127\0 P4/MBM\0   D4H\0    5\0  16\0 16\0 0\0 0\0  3\0 3\0 150000\0 356600\0 330000\0 ",
  "128\0 P4/MNC\0   D4H\0    6\0  16\0 16\0 0\0 0\0  3\0 3\0 150000\0 356606\0 330000\0 ",
  "129\0 P4/NMM\0   D4H\0    7\0  16\0 16\0 0\0 0\0  3\0 3\0 156000\0 356000\0 330000\0 ",
  "130\0 P4/NCC\0   D4H\0    8\0  16\0 16\0 0\0 0\0  3\0 3\0 156000\0 356006\0 330000\0 ",
  "131\0 P42/MMC\0  D4H\0    9\0  16\0 16\0 0\0 0\0  3\0 3\0 150006\0 350000\0 330000\0 ",
  "132\0 P42/MCM\0  D4H\0    10\0 16\0 16\0 0\0 0\0  3\0 3\0 150006\0 350006\0 330000\0 ",
  "133\0 P42/NBC\0  D4H\0    11\0 16\0 16\0 0\0 0\0  3\0 3\0 156006\0 350600\0 330000\0 ",
  "134\0 P42/NNM\0  D4H\0    12\0 16\0 16\0 0\0 0\0  3\0 3\0 156006\0 350606\0 330000\0 ",
  "135\0 P42/MBC\0  D4H\0    13\0 16\0 16\0 0\0 0\0  3\0 3\0 150006\0 356600\0 330000\0 ",
  "136\0 P42/MNM\0  D4H\0    14\0 16\0 16\0 0\0 0\0  3\0 3\0 156606\0 356606\0 330000\0 ",
  "137\0 P42/NMC\0  D4H\0    15\0 16\0 16\0 0\0 0\0  3\0 3\0 156006\0 356000\0 330000\0 ",
  "138\0 P42/NCM\0  D4H\0    16\0 16\0 16\0 0\0 0\0  3\0 3\0 156006\0 356006\0 330000\0 ",
  "139\0 I4/MMM\0   D4H\0    17\0 16\0 16\0 0\0 0\0  3\0 3\0 150000\0 350000\0 330000\0 ",
  "140\0 I4/MCM\0   D4H\0    18\0 16\0 16\0 0\0 0\0  3\0 3\0 150000\0 350006\0 330000\0 ",
  "141\0 I41/AMD\0  D4H\0    19\0 16\0 16\0 0\0 0\0  3\0 3\0 159309\0 350000\0 330000\0 ",
  "142\0 I41/ACD\0  D4H\0    20\0 16\0 16\0 0\0 0\0  3\0 3\0 159309\0 350006\0 330000\0 ",
  "143\0 P3\0       C3\0     1\0  3\0  17\0 1\0 34\0 4\0 1\0 270000\0 \0       \0       ",
  "144\0 P31\0      C3\0     2\0  3\0  17\0 1\0 34\0 4\0 1\0 270004\0 \0       \0       ",
  "145\0 P32\0      C3\0     3\0  3\0  17\0 1\0 34\0 4\0 1\0 270008\0 \0       \0       ",
  "146\0 R3\0       C3\0     4\0  3\0  17\0 1\0 34\0 4\0 1\0 270000\0 \0       \0       ",
  "147\0 PB3\0      C3I\0    1\0  6\0  18\0 0\0 34\0 4\0 1\0 590000\0 \0       \0       ",
  "148\0 RB3\0      C3I\0    2\0  6\0  18\0 0\0 34\0 4\0 1\0 590000\0 \0       \0       ",
  "149\0 P312\0     D3(a)\0  1\0  6\0  19\0 1\0 0\0  4\0 2\0 270000\0 030000\0 \0       ",
  "150\0 P321\0     D3(b)\0  2\0  6\0  20\0 1\0 0\0  4\0 2\0 270000\0 020000\0 \0       ",
  "151\0 P3112\0    D3(a)\0  3\0  6\0  19\0 1\0 0\0  4\0 2\0 270004\0 030000\0 \0       ",
  "152\0 P3121\0    D3(b)\0  4\0  6\0  20\0 1\0 0\0  4\0 2\0 270004\0 020004\0 \0       ",
  "153\0 P3212\0    D3(a)\0  5\0  6\0  19\0 1\0 0\0  4\0 2\0 270008\0 030000\0 \0       ",
  "154\0 P3221\0    D3(b)\0  6\0  6\0  20\0 1\0 0\0  4\0 2\0 270008\0 020008\0 \0       ",
  "155\0 R32\0      D3(b)\0  7\0  6\0  20\0 1\0 0\0  4\0 2\0 270000\0 020000\0 \0       ",
  "156\0 P3M1\0     C3V(a)\0 1\0  6\0  21\0 1\0 0\0  4\0 2\0 270000\0 340000\0 \0       ",
  "157\0 P31M\0     C3V(b)\0 2\0  6\0  22\0 1\0 0\0  4\0 2\0 270000\0 350000\0 \0       ",
  "158\0 P3C1\0     C3V(a)\0 3\0  6\0  21\0 1\0 0\0  4\0 2\0 270000\0 340006\0 \0       ",
  "159\0 P31C\0     C3V(b)\0 4\0  6\0  22\0 1\0 0\0  4\0 2\0 270000\0 350006\0 \0       ",
  "160\0 R3M\0      C3V(a)\0 5\0  6\0  21\0 1\0 0\0  4\0 2\0 270000\0 340000\0 \0       ",
  "161\0 R3C\0      C3V(a)\0 6\0  6\0  21\0 1\0 0\0  4\0 2\0 270000\0 340006\0 \0       ",
  "162\0 PB31M\0    D3D(a)\0 1\0  12\0 23\0 0\0 0\0  4\0 2\0 590000\0 350000\0 \0       ",
  "163\0 PB31C\0    D3D(a)\0 2\0  12\0 23\0 0\0 0\0  4\0 2\0 590000\0 350006\0 \0       ",
  "164\0 PB3M1\0    D3D(b)\0 3\0  12\0 24\0 0\0 0\0  4\0 2\0 590000\0 340000\0 \0       ",
  "165\0 PB3C1\0    D3D(b)\0 4\0  12\0 24\0 0\0 0\0  4\0 2\0 590000\0 340006\0 \0       ",
  "166\0 RB3M\0     D3D(b)\0 5\0  12\0 24\0 0\0 0\0  4\0 2\0 590000\0 340000\0 \0       ",
  "167\0 RB3C\0     D3D(b)\0 6\0  12\0 24\0 0\0 0\0  4\0 2\0 590000\0 340006\0 \0       ",
  "168\0 P6\0       C6\0     1\0  6\0  25\0 1\0 34\0 2\0 1\0 280000\0 \0       \0       ",
  "169\0 P61\0      C6\0     2\0  6\0  25\0 1\0 34\0 2\0 1\0 280002\0 \0       \0       ",
  "170\0 P65\0      C6\0     3\0  6\0  25\0 1\0 34\0 2\0 1\0 280010\0 \0       \0       ",
  "171\0 P62\0      C6\0     4\0  6\0  25\0 1\0 34\0 2\0 1\0 280004\0 \0       \0       ",
  "172\0 P64\0      C6\0     5\0  6\0  25\0 1\0 34\0 2\0 1\0 280008\0 \0       \0       ",
  "173\0 P63\0      C6\0     6\0  6\0  25\0 1\0 34\0 2\0 1\0 280006\0 \0       \0       ",
  "174\0 PB6\0      C3H\0    1\0  6\0  26\0 1\0 34\0 2\0 1\0 600000\0 \0       \0       ",
  "175\0 P6/M\0     C6H\0    1\0  12\0 27\0 0\0 34\0 2\0 2\0 280000\0 330000\0 \0       ",
  "176\0 P63/M\0    C6H\0    2\0  12\0 27\0 0\0 34\0 2\0 2\0 280006\0 330000\0 \0       ",
  "177\0 P622\0     D6\0     1\0  12\0 28\0 1\0 0\0  2\0 2\0 280000\0 020000\0 \0       ",
  "178\0 P6122\0    D6\0     2\0  12\0 28\0 1\0 0\0  2\0 2\0 280002\0 020008\0 \0       ",
  "179\0 P6522\0    D6\0     3\0  12\0 28\0 1\0 0\0  2\0 2\0 280010\0 020004\0 \0       ",
  "180\0 P6222\0    D6\0     4\0  12\0 28\0 1\0 0\0  2\0 2\0 280004\0 020004\0 \0       ",
  "181\0 P6422\0    D6\0     5\0  12\0 28\0 1\0 0\0  2\0 2\0 280008\0 020008\0 \0       ",
  "182\0 P6322\0    D6\0     6\0  12\0 28\0 1\0 0\0  2\0 2\0 280006\0 020000\0 \0       ",
  "183\0 P6MM\0     C6V\0    1\0  12\0 29\0 1\0 0\0  2\0 2\0 280000\0 340000\0 \0       ",
  "184\0 P6CC\0     C6V\0    2\0  12\0 29\0 1\0 0\0  2\0 2\0 280000\0 340006\0 \0       ",
  "185\0 P63CM\0    C6V\0    3\0  12\0 29\0 1\0 0\0  2\0 2\0 280006\0 340006\0 \0       ",
  "186\0 P63MC\0    C6V\0    4\0  12\0 29\0 1\0 0\0  2\0 2\0 280006\0 340000\0 \0       ",
  "187\0 PB6M2\0    D3H(a)\0 1\0  12\0 30\0 1\0 0\0  2\0 2\0 600000\0 340000\0 \0       ",
  "188\0 PB6C2\0    D3H(a)\0 2\0  12\0 30\0 1\0 0\0  2\0 2\0 600006\0 340006\0 \0       ",
  "189\0 PB62M\0    D3H(b)\0 3\0  12\0 31\0 1\0 0\0  2\0 2\0 600000\0 020000\0 \0       ",
  "190\0 PB62C\0    D3H(b)\0 4\0  12\0 31\0 1\0 0\0  2\0 2\0 600006\0 020000\0 \0       ",
  "191\0 P6/MMM\0   D6H\0    1\0  24\0 32\0 0\0 0\0  2\0 3\0 280000\0 020000\0 330000\0 ",
  "192\0 P6/MCC\0   D6H\0    2\0  24\0 32\0 0\0 0\0  2\0 3\0 280000\0 020006\0 330000\0 ",
  "193\0 P63/MCM\0  D6H\0    3\0  24\0 32\0 0\0 0\0  2\0 3\0 280006\0 020006\0 330000\0 ",
  "194\0 P63/MMC\0  D6H\0    4\0  24\0 32\0 0\0 0\0  2\0 3\0 280006\0 020000\0 330000\0 ",
  "195\0 P23\0      T\0      1\0  12\0 33\0 1\0 48\0 1\0 2\0 040000\0 100000\0 \0       ",
  "196\0 F23\0      T\0      2\0  12\0 33\0 1\0 48\0 1\0 2\0 040000\0 100000\0 \0       ",
  "197\0 I23\0      T\0      3\0  12\0 33\0 1\0 48\0 1\0 2\0 040000\0 100000\0 \0       ",
  "198\0 P213\0     T\0      4\0  12\0 33\0 1\0 48\0 1\0 2\0 046006\0 100000\0 \0       ",
  "199\0 I213\0     T\0      5\0  12\0 33\0 1\0 48\0 1\0 2\0 040600\0 100000\0 \0       ",
  "200\0 PM3\0      TH\0     1\0  24\0 34\0 0\0 48\0 1\0 2\0 360000\0 420000\0 \0       ",
  "201\0 PN3\0      TH\0     2\0  24\0 34\0 0\0 48\0 1\0 2\0 366600\0 420000\0 \0       ",
  "202\0 FM3\0      TH\0     3\0  24\0 34\0 0\0 48\0 1\0 2\0 360000\0 420000\0 \0       ",
  "203\0 FD3\0      TH\0     4\0  24\0 34\0 0\0 48\0 1\0 2\0 363300\0 420000\0 \0       ",
  "204\0 IM3\0      TH\0     5\0  24\0 34\0 0\0 48\0 1\0 2\0 360000\0 420000\0 \0       ",
  "205\0 PA3\0      TH\0     6\0  24\0 34\0 0\0 48\0 1\0 2\0 366006\0 420000\0 \0       ",
  "206\0 IA3\0      TH\0     7\0  24\0 34\0 0\0 48\0 1\0 2\0 360600\0 420000\0 \0       ",
  "207\0 P432\0     O\0      1\0  24\0 35\0 1\0 0\0  1\0 2\0 150000\0 100000\0 \0       ",
  "208\0 P4232\0    O\0      2\0  24\0 35\0 1\0 0\0  1\0 2\0 156606\0 100000\0 \0       ",
  "209\0 F432\0     O\0      3\0  24\0 35\0 1\0 0\0  1\0 2\0 150000\0 100000\0 \0       ",
  "210\0 F4132\0    O\0      4\0  24\0 35\0 1\0 0\0  1\0 2\0 153303\0 100000\0 \0       ",
  "211\0 I432\0     O\0      5\0  24\0 35\0 1\0 0\0  1\0 2\0 150000\0 100000\0 \0       ",
  "212\0 P4332\0    O\0      6\0  24\0 35\0 1\0 0\0  1\0 2\0 159309\0 100000\0 \0       ",
  "213\0 P4132\0    O\0      7\0  24\0 35\0 1\0 0\0  1\0 2\0 153903\0 100000\0 \0       ",
  "214\0 I4132\0    O\0      8\0  24\0 35\0 1\0 0\0  1\0 2\0 159309\0 100000\0 \0       ",
  "215\0 PB43M\0    TD\0     1\0  24\0 36\0 1\0 0\0  1\0 2\0 470000\0 100000\0 \0       ",
  "216\0 FB43M\0    TD\0     2\0  24\0 36\0 1\0 0\0  1\0 2\0 470000\0 100000\0 \0       ",
  "217\0 IB43M\0    TD\0     3\0  24\0 36\0 1\0 0\0  1\0 2\0 470000\0 100000\0 \0       ",
  "218\0 PB43N\0    TD\0     4\0  24\0 36\0 1\0 0\0  1\0 2\0 476606\0 100000\0 \0       ",
  "219\0 FB43C\0    TD\0     5\0  24\0 36\0 1\0 0\0  1\0 2\0 476000\0 100000\0 \0       ",
  "220\0 IB43D\0    TD\0     6\0  24\0 36\0 1\0 0\0  1\0 2\0 479309\0 100000\0 \0       ",
  "221\0 PM3M\0     OH\0     1\0  48\0 37\0 0\0 0\0  1\0 3\0 150000\0 100000\0 330000\0 ",
  "222\0 PN3N\0     OH\0     2\0  48\0 37\0 0\0 0\0  1\0 3\0 156000\0 100000\0 330000\0 ",
  "223\0 PM3N\0     OH\0     3\0  48\0 37\0 0\0 0\0  1\0 3\0 156606\0 100000\0 330000\0 ",
  "224\0 PN3M\0     OH\0     4\0  48\0 37\0 0\0 0\0  1\0 3\0 150606\0 100000\0 330000\0 ",
  "225\0 FM3M\0     OH\0     5\0  48\0 37\0 0\0 0\0  1\0 3\0 150000\0 100000\0 330000\0 ",
  "226\0 FM3C\0     OH\0     6\0  48\0 37\0 0\0 0\0  1\0 3\0 150006\0 100000\0 330000\0 ",
  "227\0 FD3M\0     OH\0     7\0  48\0 37\0 0\0 0\0  1\0 3\0 150303\0 100000\0 330000\0 ",
  "228\0 FD3C\0     OH\0     8\0  48\0 37\0 0\0 0\0  1\0 3\0 156303\0 100000\0 330000\0 ",
  "229\0 IM3M\0     OH\0     9\0  48\0 37\0 0\0 0\0  1\0 3\0 150000\0 100000\0 330000\0 ",
  "230\0 IA3D\0     OH\0     10\0 48\0 37\0 0\0 0\0  1\0 3\0 159309\0 100000\0 330000\0 ",
  										   
};


/* ***********************************************************************
 * space group definitions
 * give valid name in buf or valid num
 * return NULL on failure , pointer to spacegroup descriptor otherwise
 * ***********************************************************************
 */
const char *getSpacegroupDescriptor(const char *name,long *num)
{
  char *subfunc="getSpacegroupDescriptor";
  long i,lb,lt;
  const char *p;
  if(*num<1){
    if(*name!='\0'){
      for(i=0;i<230;i++){
	p=spcgrp_name(spacegrouptable[i]);
	lb=strcspn(name," ");
	lt=strlen(p);
	
	if(lt==lb && strncmp(p,name,lt)==0 ){ 
	  *num=i+1; 
	  return p; 
	  
	} 
      }
      (void)sprintf(errormsg,"Such group is absent!\nspacegroup '%s'\n",name);
      errmsg(subfunc,errormsg,ERROR);
    }else{
      (void)sprintf(errormsg,"Such group is absent!\nspacegroup number %ld",*num);
      errmsg(subfunc,errormsg,ERROR);
    }
  }else if(*num<231){
    return spacegrouptable[*num-1];
  }else{
    (void)sprintf(errormsg,"Such group is absent!\nspacegroup number %ld",*num);
    errmsg(subfunc,errormsg,ERROR);
  }
  return NULL;
}

/* ***********************************************************************
 * sginfo contains all spacegroup info
 * num must have the right spacegroiup number ( this is cheked)
 * ***********************************************************************
 */
static void extractgroupinfo(const char *sginfo,long num,tsSYMMETRY *sym )
{
  char *subfunc="extractgroupinfo";
  char pgr[100];
  long i,n=0,h,k,l,j;
  long ng;


  /* group number, for consistency check */
  n+=sscanf(spcgrp_num(sginfo),"%ld",&(sym->num))-1;
  /* group symbol */
  (void)strcpy(sym->group,spcgrp_name(sginfo));
  /* point group symbol */
  (void)strcpy(pgr,spcgrp_pgrname(sginfo));
  /* Schoenflies number of spacegroup */
  n+=sscanf(spcgrp_schoenfliesnum(sginfo),"%ld",&(sym->schoenflies))-1;
  /* number of pointgroup */
  n+=sscanf(spcgrp_pgrnum(sginfo),"%ld",&(sym->npgr))-1;
  /* inversion */
  n+=sscanf(spcgrp_inversion(sginfo),"%ld",&(ng))-1;
  sym->inversion=!ng;
  /* crystalsystem */
  n+=sscanf(spcgrp_crystallsystem(sginfo),"%ld",&(ng))-1;
  sym->crystal=ng-1;
  /* number of generators */
  n+=sscanf(spcgrp_gennum(sginfo),"%ld",&(sym->ngen))-1;


  /* generators , symmorphic */
  sym->nonsym=0;
  for(i=0;i<sym->ngen;i++){
    n+=sscanf(spcgrp_gen(sginfo,i)
	      ,"%2ld%1ld%1ld%2ld",&j ,&h,&k,&l)-4;
    j--;
    sym->igen[i]=j;
    sym->v[0+3*j]=h/12.;
    sym->v[1+3*j]=k/12.;
    sym->v[2+3*j]=l/12.;
    sym->nonsym|=h || k|| l;
  }

  sym->bravais=get_bravais_from_sginfo(sginfo);


  /* check bravais type pgr */
  if(sym->bravais == SYM_BRAVAISUNKNOWN ) {
    (void)sprintf(errormsg,"Error in spacegroup name '%s'!",sym->group);
    errmsg(subfunc,errormsg,SEVERE_ERROR);
    return;
  }
  /* check translation pgr */
  /* i=strcspn( pointgroups[sym->npgr],"(" );  */
  if(strcmp(pgr ,pointgroups[sym->npgr])     ) n++;
  
  /* lattice type */
  (void)sprintf(sym->lattice,"%s%s",bravaistype(sym->bravais)
	  ,crystalsystem(sym->crystal));
  /* cubic or hex */
  sym->hex=(sym->crystal == SYM_HEXAGONAL ) || (sym->crystal == SYM_TRIGONAL);
  sym->RTG=(sym->bravais == SYM_RHOMBOHEDRAL );


  if( (sym->num != num) || n  )
    errmsg(subfunc,"Inconsistency in space group table!"
	   ,SEVERE_ERROR);


}

/* ***********************************************************************
 * check lattice constants and angles
 * make sheer operations , lattice vetors
 * ***********************************************************************
 */
static void makeunitcell(tsSYMMETRY *sym)

{ 
  const char *nsubfunc="makeunitcell";
  long i,j,k,RTG;
  const long *C;
  double rtgtol=1.0e-14;
  double dum[9],det,ang1=0.,ang2=0.,boa,coa;
  char *invalid_alpha="First axis angle !=  90 degree!\n"
    "I will correct it !" ;
  char *invalid_beta="Second axis angle !=  90 degree!\n"
    "I will correct it !" ;
  char *invalid_gamma="Third axis angle !=  90 degree!\n"
    "I will correct it !" ;
  RTG=sym->RTG;

  /* check lattice constants and angles */
  if(sym->alat[0]<=0.) 
    {errmsg(nsubfunc,"nonpositive first lattice constant !"
	    ,ERROR);return;}
  if(sym->alat[1]<=0.) 
    {errmsg(nsubfunc,"nonpositive second lattice constant !"
	    ,ERROR);return;}
  if(sym->alat[2]<=0.) 
    {errmsg(nsubfunc,"nonpositive third lattice constant !"
	    ,ERROR);return;}
  boa=sym->alat[1]/sym->alat[0];
  coa=sym->alat[2]/sym->alat[0];
  /* usheer = unit matrix  */
  dinit(sym->usheer,0.,9);
  sym->usheer[0+3*0]=1.;
  sym->usheer[1+3*1]=1.;
  sym->usheer[2+3*2]=1.;

  /* cubic */
  if(sym->crystal == SYM_CUBIC ){
    if(boa != 1.0){
      errmsg(nsubfunc,"Noncubic 2. lattice constant detected!\nI "
	     "will use the first !",INFORMATION);
      boa=1.0;
    }
    if(coa != 1.0){
      coa=1.0;
      errmsg(nsubfunc,"Noncubic 3. lattice constant detected!\nI will"
	     " use the first !",INFORMATION);
    }
    /* tetragonal */
  }else if(sym->crystal == SYM_TETRAGONAL){
    if(boa != 1.0){
      boa=1.0;
      errmsg(nsubfunc,"Nontetragonal 2. lattice constant detected"
	     "!\nI will use the first as \'b\'!",INFORMATION);
    }
    /* hexagonal */
  }else if(sym->crystal == SYM_HEXAGONAL){
    if(boa != 1.0){
      boa=1.0;
      errmsg(nsubfunc,"Nonhexagonal 2. lattice constant detected!"
	     "\nI will use the first as \'b\'!",INFORMATION);
    }
    /* trigonal */
  }else if(sym->crystal == SYM_TRIGONAL){
    if(fabs(boa-1.0)>rtgtol){
      if(RTG){
	errmsg(nsubfunc,"Nonrhomboedric 2. lattice constant "
	       "detected!\nI will use the first as \'b\'!",INFORMATION);
      }else{
	errmsg(nsubfunc,"Nontrigonal 2. lattice constant detected!\n"
	       "I will use the first as \'b\'!",INFORMATION);
      }
      boa=1.0;
    }
    if(RTG && (fabs(coa-1.0)>rtgtol) ){
      coa=1.0;
      errmsg(nsubfunc,"Nonrhomboedric 3. lattice constant detected!\n"
	     "I will use the first as \'c\'!",INFORMATION);
    }
  }
  /* cubic tetragonal orthorhombic */
  if(sym->crystal == SYM_CUBIC 
     || sym->crystal == SYM_TETRAGONAL 
     || sym->crystal == SYM_ORTHORHOMBIC){
    if(sym->angle[0] != 90.0) {
      sym->angle[0]=90.0;
      errmsg(nsubfunc,invalid_alpha,INFORMATION);
    }
    if(sym->angle[1] != 90.0) {
      sym->angle[1]=90.0;
      errmsg(nsubfunc,invalid_beta,INFORMATION);
    }
    if(sym->angle[2] != 90.0) {
      sym->angle[2]=90.0;
      errmsg(nsubfunc,invalid_gamma,INFORMATION);
    }
  }
/*   if(sym->bravais == SYM_ABASECENTERED  */
/*      &&  sym->crystal == SYM_ORTHORHOMBIC){ */
/*     errmsg(nsubfunc,"A-base centered orthorhombic lattice:\n" */
/* 	   "The monoclinic axis is the x-axis !" */
/* 	   ,INFORMATION); */
/*   } */
  /* hexagonal/simple-trigonal */
  if(sym->crystal == SYM_HEXAGONAL 
     ||  ((! RTG) && (sym->crystal == SYM_TRIGONAL))){
    if(sym->angle[0] != 90.0 ){
      sym->angle[0]=90.0;
      errmsg(nsubfunc,invalid_alpha,INFORMATION);
    }
    if(sym->angle[1] != 90.0){
      sym->angle[1]=90.0;
      errmsg(nsubfunc,invalid_beta,INFORMATION);
    }
    if(sym->angle[2] != 120.0){
      sym->angle[2]=120.0;
      errmsg(nsubfunc,"Third hexagonal axis angle != 120 degree!\n"
	     "I will correct it !",INFORMATION);
    }
    (void)memcpy(sym->usheer,carttohex,sizeof(double)*9);
  }
  /* trigonal-rhomboedric */
  if(RTG){
    if(sym->angle[1] != sym->angle[0]){
      errmsg(nsubfunc,"Rhomboedric second axis angle different from"
	     " first!\nI will use the first !",INFORMATION);
      sym->angle[1]=sym->angle[0];
    }
    if(sym->angle[2] != sym->angle[0]){
      sym->angle[2]=sym->angle[0];
      errmsg(nsubfunc,"Rhomboedric third axis angle different " 
	     "from first!\nI will use the first !",INFORMATION);
    }
    if((sym->angle[0]>120.0-0.001) || (sym->angle[0]<0.001)){
      errmsg(nsubfunc,"Rhomboedric angle gives coplanar axis!( " 
	     "0<alpha<120 !!)\nPlease new input data !",ERROR);
      return;
    }

    /* I treat RTG as nonsimple bravais of stg */
    (void)memcpy(sym->usheer,carttohex,sizeof(double)*9);
    /* this are the special elements in the sheer of RTG treated as 
       simple lattice , also use to convert lattice constants to STG*/
    ang1=((1.0-cos(sym->angle[0]*pi180))/1.50);
    ang2=sqrt(1.0-ang1);
    ang1=sqrt(ang1);
  }
  /* monoclinic */
  if(sym->crystal == SYM_MONOCLINIC){
    if(sym->angle[0] != 90.0){
      sym->angle[0]=90.0;
      errmsg(nsubfunc,invalid_alpha,INFORMATION);
    }
    if(sym->angle[2] != 90.0){
      sym->angle[2]=90.0;
      errmsg(nsubfunc,invalid_gamma,INFORMATION);
    }
    if(sym->angle[1] < 0.001){
      errmsg(nsubfunc,"Monoclinic angle (beta) < 0.001 degree!\n"
	     "Please new input data !",ERROR);
      return;
    }
    if(sym->angle[1] > 179.999){
      errmsg(nsubfunc,"Monoclinic angle (beta) > 179.999 degree!"
	     "\nPlease new input data !",ERROR);
      return;
    }
    errmsg(nsubfunc,"Monoclinic axis is y-axis in this code !"
	   ,INFORMATION);
    sym->usheer[0+3*0]=sin(sym->angle[1]*pi180);
    sym->usheer[0+3*2]=cos(sym->angle[1]*pi180);
  }
  /* triclinic */
  if(sym->crystal ==  SYM_TRICLINIC){
    if(sym->angle[0] < 0.001) {
      errmsg(nsubfunc,"Triclinic angle (alpha) < 0.001 degree!\n"
	     "Please new input data !",ERROR);
      return;
    }
    if(sym->angle[0] > 179.999){
      errmsg(nsubfunc,"Triclinic angle (alpha) > 179.999 degree!"
	     "\nPlease new input data !",ERROR);
      return;
    }
    if(sym->angle[1] <0.001){
      errmsg(nsubfunc,"Triclinic angle (beta) < 0.001 degree!"
	     "\nPlease new input data !",ERROR);
      return;
    }
    if(sym->angle[1] > 179.999){
      errmsg(nsubfunc,"Triclinic angle (beta) > 179.999 degree!"
	     "\nPlease new input data !",ERROR);
      return;
    }
    if(sym->angle[2]<0.001){
      errmsg(nsubfunc,"Triclinic angle (gamma) < 0.001 degree!"
	     "\nPlease new input data !",ERROR);
      return;
    }
    if(sym->angle[2] > 179.999){
      errmsg(nsubfunc,"Triclinic angle (gamma) > 179.999 degree!"
	     "\nPlease new input data !",ERROR);
      return;
    }
    ang2=(cos(sym->angle[2]*pi180)-cos(sym->angle[0]*pi180)
	  *cos(sym->angle[1]*pi180)) 
      /sin(sym->angle[0]*pi180); 
    ang1=pow(sin(sym->angle[1]*pi180),2)-ang2*ang2; 
    if(ang1 <0.001) { 
      errmsg(nsubfunc,"Triclinic angles give coplanar or unpossible "
	     "axis!\nPlease new input data !",ERROR); 
      return; 
    } 
    ang1=sqrt(ang1); 
    sym->usheer[0+3*0]=ang1;
    sym->usheer[0+3*1]=ang2;
    sym->usheer[0+3*2]=cos(sym->angle[1]*pi180);
    sym->usheer[1+3*1]=sin(sym->angle[0]*pi180);
    sym->usheer[1+3*2]=cos(sym->angle[0]*pi180);
    (void)dinv33(sym->usheer,1,dum,&det);
    if(fabs(det) <0.0001) { 
      errmsg(nsubfunc,"Triclinic angles give coplanar or unpossible "
	     "axis!\nPlease new input data !",ERROR); 
      return ;
    } 
  }

  /*   lattice vectors 
       b(i)[k]=  bravais[i,j]*abclat[j]*usheer[j,k] 
       
       lattice vectors are saved as b(i)[k]=rb[k+3*i] !!!
       cartesian runs first
       
   */

  sym->alat[1]=sym->alat[0]*boa;
  sym->alat[2]=sym->alat[0]*coa;
  /* lattice constant scaling (is for rhomboeric) */
  sym->scal[0]=1.;
  sym->scal[1]=1.;
  sym->scal[2]=1.;

  if(RTG){
    /* this is RTG as nonsimple bravais of TG */
    sym->scal[0]=ang1*1.73205080756887729352;
    sym->scal[1]=ang1*1.73205080756887729352;
    sym->scal[2]=ang2*3;
  }
  dinit(sym->cell,0.,9);
  C=bravais_centering(sym->bravais);
  if(C==NULL){
    errmsg(nsubfunc,"Unknown bravais type!",SEVERE_ERROR);
    return;
  }
  
  for(i=0;i<3;i++){
    for(k=0;k<3;k++){
      for(j=0;j<3;j++){
	sym->cell[k+3*i]+=
	  C[i+3*j]/12.
	  *sym->alat[j]
	  *sym->scal[j]
	  *sym->usheer[j+3*k];
      }
    }
  }

  if(!dinv33(sym->cell,1,sym->rcell,&sym->vuc)){
    errmsg(nsubfunc,"Lattice vectors are coplanar?!",SEVERE_ERROR);
    return;
  }

  /* convert  rec cell to rec conventional cell */
  if(sym->bravais == SYM_SIMPLE ){
    converttosimple(sym->rcell,sym->crcell,SYM_SIMPLE ); 
  }else if(sym->bravais == SYM_FACECENTERED ){
    converttosimple(sym->rcell,sym->crcell,SYM_BODYCENTERED ); 
  }else if(sym->bravais == SYM_BODYCENTERED ){   
    converttosimple(sym->rcell,sym->crcell,SYM_FACECENTERED ); 
  }else if(sym->bravais == SYM_CBASECENTERED ){   
    converttosimple(sym->rcell,sym->crcell,SYM_CBASECENTERED ); 
  }else if(sym->bravais == SYM_ABASECENTERED ){   
    converttosimple(sym->rcell,sym->crcell,SYM_ABASECENTERED ); 
  }else if(sym->bravais == SYM_RHOMBOHEDRAL ){   
    converttosimple(sym->rcell,sym->crcell,SYM_RHOMBOHEDRAL ); 
  }else{
    errmsg(nsubfunc,"convert to conventional cell: "
	   "bravais not yet implemeted\n",SEVERE_ERROR);
    return;
  }

  /* convert  rec cell to rec conventional cell */
  if(sym->bravais == SYM_SIMPLE ){
    converttosimple(sym->cell,sym->ccell,SYM_SIMPLE ); 
  }else if(sym->bravais == SYM_FACECENTERED ){
    converttosimple(sym->cell,sym->ccell,SYM_FACECENTERED ); 
  }else if(sym->bravais == SYM_BODYCENTERED ){   
    converttosimple(sym->cell,sym->ccell,SYM_BODYCENTERED ); 
  }else if(sym->bravais == SYM_CBASECENTERED ){   
    converttosimple(sym->cell,sym->ccell,SYM_CBASECENTERED ); 
  }else if(sym->bravais == SYM_ABASECENTERED ){   
    converttosimple(sym->cell,sym->ccell,SYM_ABASECENTERED ); 
  }else if(sym->bravais == SYM_RHOMBOHEDRAL ){   
    converttosimple(sym->cell,sym->ccell,SYM_RHOMBOHEDRAL ); 
  }else{
    errmsg(nsubfunc,"convert to conventional cell: "
	   "bravais not yet implemeted\n",SEVERE_ERROR);
    return;
  }


}


/* ***********************************************************************
 * the rotation matrices gen[i,j,ig]=gen[i+3*j+9*ig]
 * this matrices rotate the lattice vetors not the cartesian !!!!
 *
 * gen=b=u*a*1/u, with a=cartesian transfo, u=sheer operation
 *
 * if !hex return cubic subgroups
 * if hex return hxagonal subgroups
 * 
 * 
 * ***********************************************************************
 */

static void rotations(double *gen,long hex)
{
  long j,ig;
  double dum[9];
  /* cubic sub groups */
  for(ig=0;ig<32;ig++){
    transatob(dum,pgrotations+9*ig,hex);
    for(j=0;j<9;j++){
      gen[j+9*ig]=dum[j];
      gen[j+9*(ig+32)]=-gen[j+9*ig];
    }
  }
}
/* ***********************************************************************
 * get the euler angles which would generate pgrotations
 * ***********************************************************************
 */
static void eulerangles(double *euler,long *eulerinv)
{
  long i;
  (void)memcpy(euler,pgeuler,sizeof(double)*32*3);
  (void)memcpy(euler+3*32,pgeuler,sizeof(double)*32*3);
  for(i=0;i<32;i++){
    eulerinv[i]=0;
    eulerinv[i+32]=1;
  }
}

/* ***********************************************************************
 * the invers of above, make cartes rot out of basis vector rot
 * ***********************************************************************
 */
void rothextocart(double *gout,const double *gin,long hex)
{
  long j,ig;
  double dum[9];
  /* cubic sub groups */
  for(ig=0;ig<64;ig++){
    transbtoa(dum,gin+9*ig,hex);
    for(j=0;j<9;j++){
      gout[j+9*ig]=dum[j];
    }
  }
}

/* ***********************************************************************

 * ***********************************************************************
 */
void transatob(double *b,const double *a,long hex)
{
  if(hex){
    hexatob(b,a);
  }else{
    (void)memcpy(b,a,sizeof(double)*9);
  }
}

/* ***********************************************************************
 * enscapsulate hexbtoa
 * ***********************************************************************
 */
const double *transbtoa(double *a,const double *b,long hex)
{
  if(hex){
    hexbtoa(a,b);
  }else{
    (void)memcpy(a,b,sizeof(double)*9);
  }
  return a;
}

/* ***********************************************************************
 * transform hexagonal cartesian transformations into transformations
 * of the lattice vectors: b=uhex * a * 1/uhex
 * ***********************************************************************
 */

static void hexatob(double *b,const double *a)
{
  double dum[9];
  matmult(dum,carttohex,a,3);
  matmult(b,dum,hextocart,3);
}
static void hexbtoa(double *a,const double *b)
{
  double dum[9];
  matmult(dum,hextocart,b,3);
  matmult(a,dum,carttohex,3);
}
/* ***********************************************************************
 * generate all operations by multiplication table
 * ***********************************************************************
 */

static void groupgeneration(tsSYMMETRY *sym,long usergen)
{
  const char *nsubfunc="groupgeneration";
  long i,j,k,n1,go,l,mw,ii,p,ng;
  const long *C;
  double cell[9],rcell[9],det,glm;


  for(i=0;i<64;i++){
    sym->itable[i]=0;
  }

  if(usergen){
    sym->ng=sym->nugen;
    for(i=0;i<sym->nugen;i++){
      sym->itable[i]=sym->iugen[i];
    }
  }else{
    sym->ng=sym->ngen;
    for(i=0;i<sym->ngen;i++){
      sym->itable[i]=sym->igen[i];
    }
  }  
  
  /* convert bravaistype lattice to form, used by firstcell*/
  
  C=bravais_centering(sym->bravais);
  if(C==NULL){
    errmsg(nsubfunc,"Unknown bravais type!",SEVERE_ERROR);
    return;
  }
  for(i=0;i<9;i++){
    cell[i]=C[i]/12.;
  }

  transpose(cell,3);

  if(!dinv33(cell,1,rcell,&det)){
    errmsg("dinv33","Vanishing determinant of matrix!",SEVERE_ERROR);
    return;
  }

  /* generator multiplications */
  
  n1=ng=sym->ng;ng=0;
  while( ng != n1){
    ng=n1;
    for(i=0;i<ng;i++){
      for(j=0;j<ng;j++){
	go=0;
	for(k=0;k<n1;k++){
	  go = go || (in[sym->itable[i]+sym->itable[j]*64]==sym->itable[k]+1);
	}
	if(!go){
	  sym->itable[n1]=in[sym->itable[i]+sym->itable[j]*64]-1;
	  VECVEC(&(sym->v[sym->itable[n1]*3])
		 ,&(sym->v[sym->itable[j]*3])
		 ,&(sym->v[sym->itable[i]*3])
		 ,&(sym->opmat[sym->itable[j]*9]),cell,rcell);
	  if(erroroccured())return;
	  n1++;
	  for(l=0;l<3;l++){
	    for(mw=0;mw<3;mw++){
	      glm=0.0;
	      for(k=0;k<3;k++){
		glm+=sym->opmat[l+(k+sym->itable[i]*3)*3]
		  *sym->opmat[k+(mw+sym->itable[j]*3)*3];
	      }
	      if(fabs(glm-sym->opmat[l+(mw+sym->itable[n1-1]*3)*3]) > 1.e-5){
		errmsg(nsubfunc,"Group generation via multiplication "
		       "table failed\nSorry!",SEVERE_ERROR);
		return;
		
	      }
	    }
	  }
	}
      }
    }
  }
  
  /* sort operations */
  for(ii=1;ii<ng;ii++){
    i=ii-1;
    k=i;			
    p=sym->itable[i];
    for(j=ii;j<ng;j++){
      if(sym->itable[j] < p){
	k=j;
	p=sym->itable[j];
      }
    }
    
    if(k != i){
      sym->itable[k]=sym->itable[i];
      sym->itable[i]=p;
    }
  }
  sym->ng=ng;
  

}

/* ***********************************************************************
 * index of the operation u
 * if !hex:
 *         index is index into pgrotations
 *         index+1 is index into operationsymbol
 *         ret = -1 : this operation is not in cubic subgroups
 * if hex :
 *         index%64 is index into pgrotations
 *         index+1 is index into operationsymbol
 *         ret = -1 : this operation is not in hexagonal subgroups
 *        
 * ***********************************************************************
 */

long operationindex(double *u,long hex)
{
  long ixx,iyy,izz,ret;

  ixx=operationXYZindex(u);
  iyy=operationXYZindex(u+3);
  izz=operationXYZindex(u+6);

  ret=-1;
  if(ixx == 15 ){
    if(iyy == 17 ){
      if(izz == 24 ){
	ret=0;
      }else if(izz == 4 ){
	ret=35;
      }
    }else if(iyy == 11 ){
      if(izz == 4 ){
	ret=2;
      }else if(izz == 24 ){
	ret=33;
      }
    }else if(iyy == 12 ){
      if(izz == 4 ){
	ret=-1;
	if(hex)ret=66;
      }else if(izz == 24 ){
	ret=-1;
	if(hex)ret=97;
      }
    }else if(iyy == 24 ){
      if(izz == 11 ){
	ret=21;
      }else if(izz == 17 ){
	ret=52;
      }
    }else if(iyy == 4 ){
      if(izz == 17 ){
	ret=23;
      }else if(izz == 11 ){
	ret=54;
      }
    }
  }else if(ixx == 13 ){
    if(iyy == 17 ){
      if(izz == 4 ){
	ret=1;
      }else if(izz == 24 ){
	ret=34;
      }
    }else if(iyy == 11 ){
      if(izz == 24 ){
	ret=3;
      }else if(izz == 4 ){
	ret=32;
      }
    }else if(iyy == 4 ){
      if(izz == 11 ){
	ret=20;
      }else if(izz == 17 ){
	ret=53;
      }
    }else if(iyy == 24 ){
      if(izz == 17 ){
	ret=22;
      }else if(izz == 11 ){
	ret=55;
      }
    }else if(iyy == 16 ){
      if(izz == 4 ){
	ret=-1;
	if(hex)ret=65;
      }else if(izz == 24 ){
	ret=-1;
	if(hex)ret=98;
      }
    }
  }else if(ixx == 11 ){
    if(iyy == 24 ){
      if(izz == 13 ){
	ret=4;
      }else if(izz == 15 ){
	ret=38;
      }
    }else if(iyy == 4 ){
      if(izz == 15 ){
	ret=5;
      }else if(izz == 13 ){
	ret=39;
      }
    }else if(iyy == 12 ){
      if(izz == 24 ){
	ret=26;
      }else if(izz == 4 ){
	ret=56;
      }
    }else if(iyy == 15 ){
      if(izz == 24 ){
	ret=14;
      }else if(izz == 4 ){
	ret=45;
      }
    }else if(iyy == 13 ){
      if(izz == 4 ){
	ret=15;
	if(hex)ret=30;
      }else if(izz == 24 ){
	ret=44;
	if(hex)ret=60;
      }
    }
  }else if(ixx == 17 ){
    if(iyy == 4 ){
      if(izz == 13 ){
	ret=6;
      }else if(izz == 15 ){
	ret=36;
      }
    }else if(iyy == 24 ){
      if(izz == 15 ){
	ret=7;
      }else if(izz == 13 ){
	ret=37;
      }
    }else if(iyy == 15 ){
      if(izz == 4 ){
	ret=12;
	if(hex)ret=28;
      }else if(izz == 24 ){
	ret=47;
	if(hex)ret=62;
      }
    }else if(iyy == 16 ){
      if(izz == 24 ){
	ret=24;
      }else if(izz == 4 ){
	ret=58;
      }
    }else if(iyy == 13 ){
      if(izz == 24 ){
	ret=13;
      }else if(izz == 4 ){
	ret=46;
      }
    }
    
  }else if(ixx == 4 ){
    if(iyy == 13 ){
      if(izz == 17 ){
	ret=8;
      }else if(izz == 11 ){
	ret=41;
      }
    }else if(iyy == 15 ){
      if(izz == 11 ){
	ret=11;
      }else if(izz == 17 ){
	ret=42;
      }
    }else if(iyy == 11 ){
      if(izz == 13 ){
	ret=17;
      }else if(izz == 15 ){
	ret=51;
      }
    }else if(iyy == 17 ){
      if(izz == 15 ){
	ret=18;
      }else if(izz == 13 ){
	ret=48;
      }
    }
  }else if(ixx == 24 ){
    if(iyy == 15 ){
      if(izz == 17 ){
	ret=9;
      }else if(izz == 11 ){
	ret=40;
      }
    }else if(iyy == 13 ){
      if(izz == 11 ){
	ret=10;
      }else if(izz == 17 ){
	ret=43;
      }
    }else if(iyy == 11 ){
      if(izz == 15 ){
	ret=16;
      }else if(izz == 13 ){
	ret=50;
      }
    }else if(iyy == 17 ){
      if(izz == 13 ){
	ret=19;
      }else if(izz == 15 ){
	ret=49;
      }
    }
  }else if(ixx == 16 ){
    if(iyy == 13 ){
      if(izz == 24 ){
	ret=25;
      }else if(izz == 4 ){
	ret=59;
      }
    }else if(iyy == 17 ){
      if(izz == 4 ){
	ret=29;
      }else if(izz == 24 ){
	ret=63;
      }
    }
  }else if(ixx == 12 ){
    if(iyy == 15 ){
      if(izz == 24 ){
	ret=27;
      }else if(izz == 4 ){
	ret=57;
      }
    }else if(iyy == 11 ){
      if(izz == 4 ){
	ret=31;
      }else if(izz == 24 ){
	ret=61;
      }
    }
  }
  
  return ret;

}
/* ***********************************************************************
 * symbol of axis permutations from rotation matrices 
 * ***********************************************************************
 */
const char *operationXYZ(double *u){
  return operationXYZsymbol[operationXYZindex(u)];
}

static long operationXYZindex(double *u)
{
  long ret;
  ret=ROUND(u[0])+(3*ROUND(u[1])+4)+ROUND(u[2])*10+10;
  return ret;
}               

/* ***********************************************************************
 * formated output of report to CBUFFER
 * ***********************************************************************
 */
void reportcalcsym(tsCBUFFER *b,tsSYMMETRY *sym)
{
  long ix,i,j,m,n,ig,k;
  char r[3][100];
  double dum[9],d;
  const char *INVERSION[]={"no ","yes"};  
  const long *C;

  (void)CBAppend(b,0,"\n                   SYMMETRYGROUP      \n\n"); 
  (void)CBAppend(b,100,"Space group      : %3d - %s\n",sym->num,sym->group ); 
  (void)CBAppend(b,100,"Point group      : %3d - %s\n",sym->npgr
	   ,pointgroups[sym->npgr]); 
  (void)CBAppend(b,100,"Inversion        : %s\n",INVERSION[sym->inversion]); 
  if(sym->str_type==STRTYPE_CRYSTAL)
    (void)CBAppend(b,100,"Symmorphic       : %s\n",INVERSION[!sym->nonsym]); 

  /* generators */
  (void)CBAppend(b,0,"\n                   OPERATIONS      \n\n"); 
  (void)CBAppend(b,0,"\n(The XYZ-operation symbols \"(A,B,C)+(t1,t2,t3)\" mean:\n"
	   " if a vector is given relative to the lattice basis vectors \n"
	   " r=X*a1+Y*a2+Z*a3 then it is transformed into\n"
	   " r\'=(A+t1)*a1+(B+t2)*a2+(C+t3)*a2,\n"
	   " for RTG lattices the operations refer to the corresponding\n"
	   " STG unit cell!)\n\n"); 


  
  if(sym->nugen>0){
    (void)CBAppend(b,0,"\nSub Group Generators  : %ld\n",sym->nugen); 
    if(sym->str_type==STRTYPE_CRYSTAL)
      (void)CBAppend(b,100,"Index   rotation          translation      "
		     ": symbol\n" );
    else 
      (void)CBAppend(b,100,"Index   rotation          : symbol\n" );
     
    for(ig=0;ig<sym->nugen;ig++){
      j=sym->iugen[ig];
      ix=operationindex(sym->opmat+9*j,sym->hex);
      for(i=0;i<3;i++)
	doubletoratio(r[i],sym->v[i+3*j],6,1.e-5);

      if(sym->str_type==STRTYPE_CRYSTAL)
	      
	(void)CBAppend(b,100,"%2d:   (%-4s,%-4s,%-4s) + (%-4s,%-4s,%-4s)"
		       "  :   %s \n" 
		       ,j
		       ,operationXYZ(sym->opmat+9*j)
		       ,operationXYZ(sym->opmat+9*j+3)
		       ,operationXYZ(sym->opmat+9*j+6)
		       ,r[0],r[1],r[2]
		       ,operationsymbols[ix+1]
		       ); 
      else
	(void)CBAppend(b,100,"%2d:   (%-4s,%-4s,%-4s) "
		       "  :   %s \n" 
		       ,j
		       ,operationXYZ(sym->opmat+9*j)
		       ,operationXYZ(sym->opmat+9*j+3)
		       ,operationXYZ(sym->opmat+9*j+6)
		       ,operationsymbols[ix+1]
		       ); 
    }
  }


  if(sym->nugen>0) {
    groupgeneration(sym,1L);
    
    
    (void)CBAppend(b,0,"\nSub Group Operations  : %ld\n",sym->ng); 
    if(sym->str_type==STRTYPE_CRYSTAL)
      (void)CBAppend(b,100,"Index   rotation          translation      "
		     ": symbol\n" );
    else
      (void)CBAppend(b,100,"Index   rotation          : symbol\n" );

    for(ig=0;ig<sym->ng;ig++){
      j=sym->itable[ig];
      ix=operationindex(sym->opmat+9*j,sym->hex);
      for(i=0;i<3;i++)
	doubletoratio(r[i],sym->v[i+3*j],6,1.e-5);

      if(sym->str_type==STRTYPE_CRYSTAL)
	
	(void)CBAppend(b,100,"%2d:   (%-4s,%-4s,%-4s) + (%-4s,%-4s,%-4s)  "
		       ":   %s \n" 
		       ,j
		       ,operationXYZ(sym->opmat+9*j)
		       ,operationXYZ(sym->opmat+9*j+3)
		       ,operationXYZ(sym->opmat+9*j+6)
		       ,r[0],r[1],r[2]
		       ,operationsymbols[ix+1]
		       ); 
      else
	(void)CBAppend(b,100,"%2d:   (%-4s,%-4s,%-4s) :   %s \n" 
		       ,j
		       ,operationXYZ(sym->opmat+9*j)
		       ,operationXYZ(sym->opmat+9*j+3)
		       ,operationXYZ(sym->opmat+9*j+6)
		       ,operationsymbols[ix+1]
		       ); 
      
    }

    groupgeneration(sym,0L);
  }
 

  if(sym->nugen>0) 
    (void)CBAppend(b,0,"\nFull Group Generators : %ld --> not used\n",sym->ngen); 
  else
    (void)CBAppend(b,0,"\nGroup Generators      : %ld\n",sym->ngen); 


  if(sym->str_type==STRTYPE_CRYSTAL)
    (void)CBAppend(b,100,"Index   rotation          translation      "
		   ": symbol\n" );
  else
    (void)CBAppend(b,100,"Index   rotation       : symbol\n" );

  for(ig=0;ig<sym->ngen;ig++){
    j=sym->igen[ig];
    ix=operationindex(sym->opmat+9*j,sym->hex);
    for(i=0;i<3;i++)
      doubletoratio(r[i],sym->v[i+3*j],6,1.e-5);

  if(sym->str_type==STRTYPE_CRYSTAL)
    (void)CBAppend(b,100,"%2d:   (%-4s,%-4s,%-4s) + (%-4s,%-4s,%-4s)  "
		   ":   %s \n" 
		   ,j
		   ,operationXYZ(sym->opmat+9*j)
		   ,operationXYZ(sym->opmat+9*j+3)
		   ,operationXYZ(sym->opmat+9*j+6)
		   ,r[0],r[1],r[2]
		   ,operationsymbols[ix+1]
		   ); 
  else
    (void)CBAppend(b,100,"%2d:   (%-4s,%-4s,%-4s) :   %s \n" 
		   ,j
		   ,operationXYZ(sym->opmat+9*j)
		   ,operationXYZ(sym->opmat+9*j+3)
		   ,operationXYZ(sym->opmat+9*j+6)
		   ,operationsymbols[ix+1]
		   ); 
  }

  if(sym->nugen>0)
    (void)CBAppend(b,0,"\nFull Group Operations : %ld --> not used\n",sym->ng); 
  else
    (void)CBAppend(b,0,"\nFull Group Operations : %ld\n",sym->ng); 

  if(sym->str_type==STRTYPE_CRYSTAL)
    (void)CBAppend(b,100,"Index   rotation          translation      "
		   ": symbol\n" );
  else
    (void)CBAppend(b,100,"Index   rotation       : symbol\n" );

  for(ig=0;ig<sym->ng;ig++){
    j=sym->itable[ig];
    ix=operationindex(sym->opmat+9*j,sym->hex);
    for(i=0;i<3;i++)
      doubletoratio(r[i],sym->v[i+3*j],6,1.e-5);

    if(sym->str_type==STRTYPE_CRYSTAL)
      (void)CBAppend(b,100,"%2d:   (%-4s,%-4s,%-4s) + (%-4s,%-4s,%-4s)  "
		     ":   %s \n" 
		     ,j
		     ,operationXYZ(sym->opmat+9*j)
		     ,operationXYZ(sym->opmat+9*j+3)
		     ,operationXYZ(sym->opmat+9*j+6)
		     ,r[0],r[1],r[2]
		     ,operationsymbols[ix+1]
		     ); 
    else
      (void)CBAppend(b,100,"%2d:   (%-4s,%-4s,%-4s) :   %s \n" 
		     ,j
		     ,operationXYZ(sym->opmat+9*j)
		     ,operationXYZ(sym->opmat+9*j+3)
		     ,operationXYZ(sym->opmat+9*j+6)
		     ,operationsymbols[ix+1]
		     ); 
  }
  


  if(sym->nugen>0) {
    groupgeneration(sym,1L);
  }

  if(sym->str_type!=STRTYPE_CRYSTAL) goto ende;


  
  (void)CBAppend(b,0,"\n                   TRANSLATION     \n\n"); 
  if(sym->RTG)
     (void)CBAppend(b,0,"RTG as nonsimple STG-bravais lattice\n\n"); 
  (void)CBAppend(b,100,"lattice constants: %- 22.15f %- 22.15f %- 22.15f\n"
	   ,sym->alat[0]*sym->scal[0]
	   ,sym->alat[1]*sym->scal[1]
	   ,sym->alat[2]*sym->scal[2]);

  if(sym->RTG){
    (void)CBAppend(b,100,"axis angles      : %- 22.15f %- 22.15f %- 22.15f\n"
	     ,90.,90.,120.); 
  }else{
    (void)CBAppend(b,100,"axis angles      : %- 22.15f %- 22.15f %- 22.15f\n"
	     ,sym->angle[0],sym->angle[1],sym->angle[2]); 
  }
  
  (void)CBAppend(b,100,"bravais lattice  : %s \n",sym->lattice); 
  /* the bravais lattice transformation */
  (void)CBAppend(b,100,"primitive to bravais transformation\n");

  C=bravais_centering(sym->bravais);
  if(C==NULL){
    C=bravais_centering(SYM_SIMPLE);
  }
  for(i=0;i<3;i++){
    for(j=0;j<3;j++){
      m=VEC(C,j)[i]; n=12;
      normalize_ratio(&m,&n);
      if(m ==0 ){
	sprintf(r[j],"%ld",0L);
      }else if(n==1){
	sprintf(r[j],"%ld",m);
      }else {
	sprintf(r[j],"%ld/%ld",m,n);
      }
    }
    (void)CBAppend(b,100,"      b%1d  : %-4s %-4s %-4s\n"
		   ,i+1,r[0],r[1],r[2]);
  }
  /* the unit vectors of the primitive lattice */
  (void)CBAppend(b,100,"unit vectors of primitive lattice (S%s)\n"
	   ,crystalsystem(sym->crystal));
  for(i=0;i<3;i++){
    (void)CBAppend(b,100,"      u%1d  : %- 22.15f %- 22.15f %- 22.15f\n"
	     ,i+1,sym->usheer[i+3*0],sym->usheer[i+3*1],sym->usheer[i+3*2]); 
  }


  /* once more for RTG case as primitive bravais lattice */
  if(sym->RTG){
    (void)CBAppend(b,0,"\nRTG as simple bravais lattice\n\n"); 
    (void)CBAppend(b,100,"lattice constants: %- 22.15f %- 22.15f %- 22.15f\n"
	     ,sym->alat[0]
	     ,sym->alat[1]
	     ,sym->alat[2]);
    
    (void)CBAppend(b,100,"axis angles      : %- 22.15f %- 22.15f %- 22.15f\n"
	     ,sym->angle[0],sym->angle[1],sym->angle[2]); 
    
    
    (void)CBAppend(b,100,"bravais lattice  : %s \n",sym->lattice); 
    /* the bravais lattice transformation */
    (void)CBAppend(b,100,"primitive to bravais transformation\n");

    C=bravais_centering(SYM_SIMPLE);
    for(i=0;i<3;i++){
      for(j=0;j<3;j++){
	m=VEC(C,j)[i]; n=12;
	normalize_ratio(&m,&n);
	if(m ==0 ){
	  sprintf(r[j],"%ld",0L);
	}else if(n==1){
	  sprintf(r[j],"%ld",m);
	}else {
	  sprintf(r[j],"%ld/%ld",m,n);
	}
      }
      (void)CBAppend(b,100,"      b%1d  : %-4s %-4s %-4s\n"
	       ,i+1,r[0],r[1],r[2]);
    }
    /* the unit vectors of the primitive lattice */
    C=bravais_centering(SYM_RHOMBOHEDRAL);
    for(i=0;i<3;i++){
      for(k=0;k<3;k++){
	d=0.;
	for(j=0;j<3;j++){
	  d+=C[i+3*j]/12.*sym->scal[j]
	    *sym->usheer[j+3*k];
	}
	dum[i+3*k]=d;
      }
    }
    (void)CBAppend(b,100,"unit vectors of primitive lattice (R%s)\n"
	     ,crystalsystem(sym->crystal));
    for(i=0;i<3;i++){
      (void)CBAppend(b,100,"      u%1d  : %- 22.15f %- 22.15f %- 22.15f\n"
	       ,i+1,dum[i+3*0],dum[i+3*1],dum[i+3*2]); 
    }
  }
  /* the lattice vectors */
  (void)CBAppend(b,100,"lattice vectors\n");
  for(i=0;i<3;i++){
    (void)CBAppend(b,100,"      a%1d  : %- 22.15f %- 22.15f %- 22.15f\n"
	     ,i+1,sym->cell[0+3*i],sym->cell[1+3*i],sym->cell[2+3*i]);
  }
  (void)CBAppend(b,100,"reciprocial lattice vectors / 2*Pi\n");
  for(i=0;i<3;i++){
    (void)CBAppend(b,100,"      g%1d  : %- 22.15f %- 22.15f %- 22.15f\n"
	     ,i+1,sym->rcell[0+3*i],sym->rcell[1+3*i],sym->rcell[2+3*i]);
  }
  

 ende:
  (void)CBAppend(b,1000,"\n"); 

}

/* ***********************************************************************
 * make symmetry from spacgroup info
 * ***********************************************************************
 */
long calcsym(tsSYMMETRY *sym
	    ,long num		/* spacegroup number */
	    ,const char *spgrname	/* alternative give group name */
	    ,double alat[3]	/* lattice parameters */
	    ,double angle[3]	/* axis angles */
            ,long complain_subgroups /* bitch around about subgroup generator table ?
				        it makes no sense to bitch during updates*/
	     ,long str_type
	     )
{
  long            s=0;		/* severity level of failure */
  const char *sginfo;	/* spacegroup info */
  long i,j,ix,err,found,k;
  char buf[5000];
  char *subfunc="calcsym";
  /* get spacegroup description from table */

  sginfo=getSpacegroupDescriptor(spgrname,&num);
  if((s=LMAX(s,severitylevel()))>WARNING) goto clean2;

  /* extract info from sginfo */
  extractgroupinfo(sginfo,num,sym);
  if((s=LMAX(s,severitylevel()))>WARNING) goto clean2;


  /* check and adjust lattice parameter */
  sym->alat[0]=alat[0];
  sym->alat[1]=alat[1];
  sym->alat[2]=alat[2];

  sym->angle[0]=angle[0];
  sym->angle[1]=angle[1];
  sym->angle[2]=angle[2];

  sym->str_type=str_type;
  if(str_type==STRTYPE_CRYSTAL){
    /* printf("being crystal\n"); */
    makeunitcell(sym);
    if((s=LMAX(s,severitylevel()))>WARNING) goto clean2;
  }


  /* def rotation matrices */
  rotations(sym->opmat,sym->hex);  
  eulerangles(sym->eulerangles,sym->eulerinv);

  /* generate spacegroup */
  groupgeneration(sym,0L);
  if((s=LMAX(s,severitylevel()))>WARNING) goto clean2;


  /* manipulate the generator table according to user whish
     and recalculate the symmetry table */
/*   goto clean2; */
  if(sym->nugen>0){

    err=0;
    (void)sprintf(errormsg,"The following subgroup generators are invalid\n"
		  "in this spacegroup:");
    for(i=0;i<sym->nugen;i++){
      j=sym->iugen[i];
      found=0;
      for(k=0;k<sym->ng;k++){
	if(j == sym->itable[k]){
	  found=1;
	  break;
	}
      }
      if(!found){
	err=1;
	(void)sprintf(buf," %ld",j);
	strcat(errormsg,buf);
      }
    }
    if(err){
      if(complain_subgroups)
	errmsg(subfunc,errormsg,SEVERE_ERROR);
      else
	errmsg(subfunc,errormsg,WARNING);
    }
    if((s=LMAX(s,severitylevel()))>WARNING) {
      goto clean2;
    }
    
    /* re-generate modified spacegroup */
    groupgeneration(sym,1L);
    if((s=LMAX(s,severitylevel()))>WARNING) goto clean2;
    
    
    sym->inversion=0;
    for(i=0;i<sym->ng;i++){
      j=sym->itable[i];
      ix=operationindex(sym->opmat+9*j,sym->hex);
      if(strcmp(operationsymbols[ix+1],"I")==0){
	sym->inversion=1;
      }
    }
    
  }    
 clean2:

  return s;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
long get_group_operations(long num,tsSYMMETRY *sym)
{
  long s=0;		/* severity level of failure */
  const char *sginfo;	/* spacegroup info */
    
  sginfo=getSpacegroupDescriptor("",&num);
  if((s=LMAX(s,severitylevel()))>WARNING) goto clean2;

  /* extract info from sginfo */
  extractgroupinfo(sginfo,num,sym);
  if((s=LMAX(s,severitylevel()))>WARNING) goto clean2;


  /* def rotation matrices */
  rotations(sym->opmat,sym->hex);  
  eulerangles(sym->eulerangles,sym->eulerinv);

  /* generate spacegroup */
  groupgeneration(sym,0L);


 clean2:

  return s;
}

/* ***********************************************************************
 * return bravaislattice
 * ***********************************************************************
 */
const long *bravais_centering(long brav)
{
  if(brav<0 || brav >=SYM_BRAVAISUNKNOWN )
    return NULL;
  else
    return VECN(ibravais,9,brav);

}

const char *bravaistype(long brav)
{
  if(brav<0 || brav >SYM_BRAVAISUNKNOWN )
    return bravaistypes1[SYM_BRAVAISUNKNOWN];
  else
    return bravaistypes1[brav];
}

const char *crystalsystem(long crys)
{
  if(crys<0 || crys >=crystalsystem_size )
    return crystalsystems[SYM_CRYSTALUNKNOWN];
  else
    return crystalsystems[crys];
}

long numberofbravaistypes()
{  
  return bravaistypes_size-1;
}
long numberofcrystalsystem()
{
  return crystalsystem_size-1;
}

const char *bravaistypes_longname(long bravais)
{
  if(bravais<0 || bravais >=bravaistypes_size )
    return bravaistypes_longname1[SYM_BRAVAISUNKNOWN];
  else
    return bravaistypes_longname1[bravais];
}


const char *crystalsystems_longname(long crys)
{
  if(crys<0 || crys >=crystalsystem_size )
    return crystalsystems_longname1[SYM_CRYSTALUNKNOWN];
  else
    return crystalsystems_longname1[crys];
}

const char *pointgroup(long pgr)
{
  if(pgr<1 || pgr >=pointgroups_size )
    return pointgroups[0];
  else
    return pointgroups[pgr];
}

/* ***********************************************************************
 * makes corresponding simple cell out off bravais cell
 * cell[i+3*j] is i-th componet of j-th lattice vector
 * ***********************************************************************
 */
void converttosimple(const double *cell,double *scell,long bravais)
{
  
  if(bravais == SYM_SIMPLE ){
    (void)memcpy(scell,cell,sizeof(double)*9);
    return;
  }else if(bravais == SYM_FACECENTERED  ){
    scell[0+3*0]=(cell[0+3*1]+cell[0+3*2]-cell[0+3*0]);
    scell[1+3*0]=(cell[1+3*1]+cell[1+3*2]-cell[1+3*0]);
    scell[2+3*0]=(cell[2+3*1]+cell[2+3*2]-cell[2+3*0]);
    scell[0+3*1]=(cell[0+3*0]+cell[0+3*2]-cell[0+3*1]);
    scell[1+3*1]=(cell[1+3*0]+cell[1+3*2]-cell[1+3*1]);
    scell[2+3*1]=(cell[2+3*0]+cell[2+3*2]-cell[2+3*1]);
    scell[0+3*2]=(cell[0+3*0]+cell[0+3*1]-cell[0+3*2]);
    scell[1+3*2]=(cell[1+3*0]+cell[1+3*1]-cell[1+3*2]);
    scell[2+3*2]=(cell[2+3*0]+cell[2+3*1]-cell[2+3*2]);
    return;
  }else if(bravais == SYM_BODYCENTERED ){
    scell[0+3*0]=(cell[0+3*1]+cell[0+3*2]);
    scell[1+3*0]=(cell[1+3*1]+cell[1+3*2]);
    scell[2+3*0]=(cell[2+3*1]+cell[2+3*2]);
    scell[0+3*1]=(cell[0+3*0]+cell[0+3*2]);
    scell[1+3*1]=(cell[1+3*0]+cell[1+3*2]);
    scell[2+3*1]=(cell[2+3*0]+cell[2+3*2]);
    scell[0+3*2]=(cell[0+3*0]+cell[0+3*1]);
    scell[1+3*2]=(cell[1+3*0]+cell[1+3*1]);
    scell[2+3*2]=(cell[2+3*0]+cell[2+3*1]);
    return;
  }else if(bravais == SYM_CBASECENTERED ){
    scell[0+3*0]=(cell[0+3*0]-cell[0+3*1]);
    scell[1+3*0]=(cell[1+3*0]-cell[1+3*1]);
    scell[2+3*0]=(cell[2+3*0]-cell[2+3*1]);
    scell[0+3*1]=(cell[0+3*0]+cell[0+3*1]);
    scell[1+3*1]=(cell[1+3*0]+cell[1+3*1]);
    scell[2+3*1]=(cell[2+3*0]+cell[2+3*1]);

    (void)memcpy(scell+6,cell+6,sizeof(double)*3);
    return;
  }else if(bravais == SYM_ABASECENTERED ){
    scell[0+3*2]=(cell[0+3*1]-cell[0+3*0]);
    scell[1+3*2]=(cell[1+3*1]-cell[1+3*0]);
    scell[2+3*2]=(cell[2+3*1]-cell[2+3*0]);
    scell[0+3*1]=(cell[0+3*1]+cell[0+3*0]);
    scell[1+3*1]=(cell[1+3*1]+cell[1+3*0]);
    scell[2+3*1]=(cell[2+3*1]+cell[2+3*0]);

    (void)memcpy(scell+0,cell+6,sizeof(double)*3);
    return;
  }else if(bravais == SYM_RHOMBOHEDRAL ){
    scell[0+3*0]=(cell[0+3*0]-cell[0+3*1]);
    scell[1+3*0]=(cell[1+3*0]-cell[1+3*1]);
    scell[2+3*0]=(cell[2+3*0]-cell[2+3*1]);
    scell[0+3*1]=(cell[0+3*1]-cell[0+3*2]);
    scell[1+3*1]=(cell[1+3*1]-cell[1+3*2]);
    scell[2+3*1]=(cell[2+3*1]-cell[2+3*2]);
    scell[0+3*2]=(cell[0+3*0]+cell[0+3*1]+cell[0+3*2]);
    scell[1+3*2]=(cell[1+3*0]+cell[1+3*1]+cell[1+3*2]);
    scell[2+3*2]=(cell[2+3*0]+cell[2+3*1]+cell[2+3*2]);
    return;
  }
  (void)printf("converttosimple: %ld %s bot yet impl\n",bravais,bravaistype(bravais));
  exit(1);
}

/* ***********************************************************************
 * copy s1 onto s2
 * ***********************************************************************
 */
void copySymmetry(const tsSYMMETRY *s1,tsSYMMETRY *s2)
{
  *s2=*s1;
  memcpy(s2->itable     ,s1->itable     ,sizeof(long)*64);
  memcpy(s2->eulerinv   ,s1->eulerinv   ,sizeof(long)*64);
  memcpy(s2->igen       ,s1->igen       ,sizeof(long)*48);
  memcpy(s2->iugen      ,s1->iugen      ,sizeof(long)*48);

  memcpy(s2->v          ,s1->v          ,sizeof(double)*3*64);
  memcpy(s2->opmat      ,s1->opmat      ,sizeof(double)*9*64);
  memcpy(s2->eulerangles,s1->eulerangles,sizeof(double)*3*64);
  memcpy(s2->scal       ,s1->scal       ,sizeof(double)*3);
  memcpy(s2->alat       ,s1->alat       ,sizeof(double)*3);
  memcpy(s2->angle      ,s1->angle      ,sizeof(double)*3);
  memcpy(s2->usheer     ,s1->usheer     ,sizeof(double)*9);
  memcpy(s2->cell       ,s1->cell       ,sizeof(double)*9);
  memcpy(s2->rcell      ,s1->rcell      ,sizeof(double)*9);
  memcpy(s2->crcell     ,s1->crcell     ,sizeof(double)*9);




  strcpy(s2->group,s1->group);
  strcpy(s2->lattice,s1->lattice);

}



/* ***********************************************************************
 * the rotation matrices gen[i,j,ig]=gen[i+3*j+9*ig]
 * this matrices rotate the  cartesian !!!!
 *
 * ***********************************************************************
 */

void getrotations(double *gen)
{
  long j,ig;
  double a;

  for(ig=0;ig<32;ig++){
    for(j=0;j<9;j++){
      a=gen[j+9*ig]=pgrotations[j+9*ig];
      gen[j+9*(ig+32)]=-a;
    }
  }
}




/* ***********************************************************************
 *
 * ***********************************************************************
 */

const char *operationsymbol(long ig)
{
  return operationsymbols[ig];
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
/* void getpointgroupmasks(long *pg) */
/* { */
/*   (void)memcpy(pg,pointgroupmask,sizeof(long)*MAXPOINTGRP*64); */
/* } */ 

/* ***********************************************************************
 * get maximal pointgroup for lattice type
 * 
 * ***********************************************************************
 */
/* long getMaxPointGroupForLattice(long crystal,long bravais) */
/* { */
/*   long ig0=0; */


/*   if(crystal==SYM_HEXAGONAL || crystal==SYM_TRIGONAL)ig0=31;  */
/*   if(crystal==SYM_TRIGONAL && bravais==SYM_RHOMBOHEDRAL)ig0=23;  */
/*   if(crystal==SYM_TRICLINIC)ig0=1;		 */
/*   if(crystal==SYM_MONOCLINIC)ig0=4;		 */
/*   if(crystal==SYM_ORTHORHOMBIC)ig0=7;		 */
/*   if(crystal==SYM_TETRAGONAL)ig0=15;		 */
/*   if(crystal==SYM_CUBIC)ig0=36;		 */

/*   if(! ig0){ */
/*     errmsg("getMaxPointGroupForLattice","unknown lattice type",SEVERE_ERROR); */
/*   } */
/*   return ig0; */
/* } */







/* ***********************************************************************
 * pointgroup masks, with rotations indicces
 * ***********************************************************************
 */
/* index in help array */
/* #define nu1unity      0 */
/* #define nu1xAxis      1 */
/* #define nu1yAxis      2 */
/* #define nu1zAxis      3 */
/* #define nu1xyAxis     4 */
/* #define nu1xmyAxis    5 */
/* #define nu1xzAxis     6 */
/* #define nu1xmzAxis    7 */
/* #define nu1yzAxis     8 */
/* #define nu1ymzAxis    9 */
/* #define nu130Axis     10 */
/* #define nu160Axis     11 */
/* #define nu1m30Axis    12 */
/* #define nu1m60Axis    13 */
/* #define nu1m111Axis   14 */
/* #define nu11m11Axis   15 */
/* #define nu1111Axis    16 */
/* #define nu1m1m11Axis  17 */
/* #define nu1inv        18 */


/* #define sqrt2  .70710678118654752440 */
/* #define sqrt32 .866025403784438646763723170750 */
/* #define sqrt33 .577350269189625764509148780501 */

/* const double rotationhelper[3*18]={ */
/*   0.0,0.0,0.0 */
/*   / *x,y,z axis * / */
/*   ,1.0,0.0,0.0 */
/*   ,0.0,1.0,0.0 */
/*   ,0.0,0.0,1.0 */
/*   / * xy * / */
/*   ,sqrt2,sqrt2,0.0  */
/*   / * x-y * / */
/*   ,sqrt2,-sqrt2,0.0  */
/*   / * xz * / */
/*   ,sqrt2,0.0,sqrt2  */
/*   / * x-z * / */
/*   ,sqrt2,0.0,-sqrt2  */
/*   / * yz * / */
/*   ,0.0,sqrt2,sqrt2  */
/*   / * y-z * / */
/*   ,0.0,sqrt2,-sqrt2  */
/*   / * 30* / */
/*   ,sqrt32,0.5,0.0  */
/*   / * 60* / */
/*   ,0.5,sqrt32,0.0  */
/*   / * -30* / */
/*   ,sqrt32,-0.5,0.0  */
/*   / * -60* / */
/*   ,0.5,-sqrt32,0.0  */
/*   / *-111* / */
/*   ,-sqrt33,sqrt33,sqrt33 */
/*   / *1-11 * / */
/*   ,sqrt33,-sqrt33,sqrt33 */
/*   / * 111* / */
/*   ,sqrt33,sqrt33,sqrt33 */
/*   /  -1-11* / */
/*   ,-sqrt33,-sqrt33,sqrt33 */
/* }; */

/* take all operations, in contrary to RTG->MC where z has to be excludeed
 */
/* static const long mmask[64]={ */
/*   1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1 */
/*   ,1,1,1,1,1,1,1,1 */
/*   ,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1 */
/*   ,1,1,1,1,1,1,1,1 */
/* }; */
/* static const long rtgmask[]={ */
/*   1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1 */
/*   ,1,1,1,1,1,0,0,1 */
/*   ,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1 */
/*   ,1,1,1,1,1,0,0,1 */
/* }; */


/* #include "masks.h" */

/* ***********************************************************************
 * get group analysis helper
 * ***********************************************************************
 */
/* void getanalysishelper(long c4[64],long s4[64],long c2[64],long s2[64] */
/* 		       ,long c3[64],long s3[64] */
/* 		       ,long c6[64],long s6[64],long opi[64],long ope[64] */
/* 		       ,long mm[64],long rm[64] */
/* 		       ,double u1[3*17] */
/* 		       ,long *xaxis,long *yaxis,long *zaxis */
/* 		       ,long *c2x,long *c2y,long *c2z */
/* 		       ,long *sx,long *sy */
/* 		       ,long *c14y,long *inv */
/* 		       ,long *c13z */
/* 		       ,long *c14z */
/* 		       ,long *s14z */
/* 		       ) */
/* { */

/*   (void)memcpy(c4,maskc4,sizeof(long)*64); */
/*   (void)memcpy(s4,masks4,sizeof(long)*64); */
/*   (void)memcpy(c2,maskc2,sizeof(long)*64); */
/*   (void)memcpy(s2,masks2,sizeof(long)*64); */
/*   (void)memcpy(c3,maskc3,sizeof(long)*64);  */
/*   (void)memcpy(s3,masks3,sizeof(long)*64);  */
/*   (void)memcpy(c6,maskc6,sizeof(long)*64);  */
/*   (void)memcpy(s6,masks6,sizeof(long)*64);  */
/*   (void)memcpy(opi,maskopi,sizeof(long)*64);  */
/*   (void)memcpy(ope,maskope,sizeof(long)*64);  */
/*   (void)memcpy(mm,mmask,sizeof(long)*64); */
/*   (void)memcpy(rm,rtgmask,sizeof(long)*64); */
/*   *xaxis=nu1xAxis; */
/*   *yaxis=nu1yAxis; */
/*   *zaxis=nu1zAxis; */
/*   *c2x=nc2x; */
/*   *c2y=nc2y; */
/*   *c2z=nc2z; */
/*   *sx=nsx; */
/*   *sy=nsy; */
/*   *c14y=nc14y; */
/*   *c14z=nc14z; */
/*   *s14z=ns14z; */
/*   *inv=ninv; */
/*   *c13z=nc13z; */
/*   (void)memcpy(u1,rotationhelper, sizeof(rotationhelper) ); */

/*   ** initializedata();   ** */
/* } */


/* ***********************************************************************
 * helper for calculation of static const data arrays
 * ***********************************************************************
 */
/* static long initializedata() */
/* { */
/*   long pg[64*MAXPOINTGRP]; */
/*   getpointgroupmasksinternal(pg); */

/*   getmasksinternal(operationsymbols); */

/*   return 1; */
/* } */

/* ***********************************************************************
 * convert i.e. Wyckhoff positions of simple RTG into WPS of nonsimple STG
 * ***********************************************************************
 */
void ConvertVecRTGtoSTG(double *result,const double *v)
{
  long i;
  double vzero[3]={0.,0.,0.};
  double m[9];
  for(i=0;i<9;i++){
    m[i]=bravais_centering(SYM_RHOMBOHEDRAL)[i]/12.;
  }
  vec1(result,vzero,v,m); 
}


/* ***********************************************************************
 * get tau of generator from table
 * ***********************************************************************
 */

long get_gentau_from_sginfo(const char *sginfo,long igen
		,long *j
		,long iv[3]
		)
{
  long n,ngen;
  n=sscanf(spcgrp_gennum(sginfo),"%ld",&ngen)-1;
  if(igen<0 || igen>=ngen) return n+1;
  n+=sscanf(spcgrp_gen(sginfo,igen)
	    ,"%2ld%1ld%1ld%2ld",j ,iv,iv+1,iv+2)-4;
  return n;
}

/* ***********************************************************************
 * get pgrnum from spcegroup info string
 * return 0 on error
 * ***********************************************************************
 */
long get_pgrnum_from_sginfo(const char *sginfo)
{
  long n,npgr;
  n=sscanf(spcgrp_pgrnum(sginfo),"%ld",&npgr)-1;
  if(n)return 0;
  return npgr;
}

/* ***********************************************************************
 * get bravais type form sginfo 
 * return SYM_BRAVAISUNKNOWN on error
 * ***********************************************************************
 */
long get_bravais_from_sginfo(const char *sginfo)
{
  switch (spcgrp_name(sginfo)[0]) {
  case 'P' : return SYM_SIMPLE ;break;
  case 'I' : return SYM_BODYCENTERED ;break;
  case 'F' : return SYM_FACECENTERED ;break;
  case 'A' : return SYM_ABASECENTERED ;break;
  case 'C' : return SYM_CBASECENTERED ;break;
  case 'R' : return SYM_RHOMBOHEDRAL ;break;
  default: return SYM_BRAVAISUNKNOWN;break;
  }
  
}

long numberofbravaistypesofcryssystem(long csystem){
  if(csystem <0 || csystem >=crystalsystem_size )
    return 0;
  return numberofbravaistypesofcryssystem1[csystem];
}

long bravaistypesofcryssystem(long csystem,long index)
{
  if(csystem <0 || csystem >=crystalsystem_size )
    return SYM_BRAVAISUNKNOWN;

  if(index <0 || index >=numberofbravaistypesofcryssystem1[csystem] )
    return SYM_BRAVAISUNKNOWN;

  return bravaistypesofcryssystem1[csystem][index];
}

long numberofspacegroupsof_c_b(long csystem,long btype)
{
  const char *sginfo;
  long c,i,b,count=0;

  if(csystem <0 || csystem >=crystalsystem_size )
    return 0;
  if(btype<0 || btype >=bravaistypes_size )
    return 0;

  for(i=0;i<230;i++){
    sginfo=spacegrouptable[i];
    spcgrp_crystallsystem(sginfo);
    c=get_crystal_system_from_num(i+1);
/*     if(i<2){ */
/*       printf("i=%ld  C=%ld csystem=%ld\n",i,c,csystem);  */
/*     } */
  
    b=get_bravais_from_sginfo(sginfo);
    if(csystem==c && b==btype){
      /* printf("n=%ld i=%ld\n",n,i); */
      count++;
    }

  }

  return count;
}




long spacegroupsof_c_b(long csystem,long btype,long index)
{
  const char *sginfo;
  long count=0,i,c,b;

  if(csystem <0 || csystem >=crystalsystem_size )
    return -1;
  if(btype<0 || btype >=bravaistypes_size )
    return -1;

  for(i=0;i<230;i++){
    sginfo=spacegrouptable[i];
    spcgrp_crystallsystem(sginfo);
    c=get_crystal_system_from_num(i+1);
    b=get_bravais_from_sginfo(sginfo);

    if(csystem==c && b==btype){
      /* printf("n=%ld i=%ld\n",n,i); */
      if(count==index)
	return i+1;
      count++;
    }

  }

  return -1;
}

long get_crystal_system_from_num(long num)
{
  char c;
  
/*   printf("num=%ld\n",num); */

  if(num<1 || num>230) return -1;
  
  c=spcgrp_crystallsystem(spacegrouptable[num-1])[0];
  /*   printf("c=%c\n",c); */
/*   if(num<=2){ */
/*     printf("num=%ld  C=%c \n",num,c);  */
/*   } */
  
  switch (c) {
  case '1': return 0; break;
  case '2': return 1; break;
  case '3': return 2; break;
  case '4': return 3; break;
  case '5': return 4; break;
  case '6': return 5; break;
  case '7': return 6; break;
  default: return -1; break;
  }

}

long get_bravais_from_num(long num)
{
  if(num<1 || num>230) return -1;
  
  return get_bravais_from_sginfo(spacegrouptable[num-1]);
  
}


long get_pointgroup_from_num(long num)
{
  long n,pgr;
  if(num<1 || num>230) return -1;
  n=sscanf(spcgrp_pgrnum(spacegrouptable[num-1]),"%ld",&pgr)-1;
  if(n) return -1;
  
  return pgr;
  
}


long get_spacegroup_of_from_pointgroup(long pgrnum)
{
  long i,j,nonsym,ngen,n,h,k,l,jj;
  
  for(i=0;i<230;i++){
    if( get_pointgroup_from_num(i+1) != pgrnum ) continue;
    if(get_bravais_from_num(i+1) != SYM_SIMPLE ) continue;
    /* generators , symmorphic */
    n+=sscanf(spcgrp_gennum(spacegrouptable[i]),"%ld",&ngen)-1;
    nonsym=0;
    for(j=0;j<ngen;j++){
      n+=sscanf(spcgrp_gen(spacegrouptable[i],j)
		,"%2ld%1ld%1ld%2ld",&jj ,&h,&k,&l)-4;
      nonsym |= h || k|| l;
    }
    if(nonsym) continue;

    /* printf("i %ld pgr %ld\n",i+1,pgrnum); */
    
    return i+1;
  }  
  /*   printf("\n"); */
  return -1;
}
/* EOF */

