/* @(#)AtomPropsModel.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _ATOMPROPSMODEL_H
#define _ATOMPROPSMODEL_H 1

#include <QAbstractTableModel>
#include <QToolTip>
#include "Symmetry.hh"
#include "PropertyTable.hh"
#include "Settings.hh"

using namespace XFPLO;

class AtomPropsModel: public QAbstractTableModel
{
public:

  enum SetDefaultsActions{ColorByZ=0};
  enum Columns{El=0,Visible=1,AtomColor=2,AtomShininess=3,SpecularColor=4,
	       RadiusType=5,Radius=6
  };


  // ----------------------------------------------------------------
  AtomPropsModel( QObject * parent=0,XFPLO::AtomPropertyTable *ptable=0,Symmetry *sym=0)
    : QAbstractTableModel(parent),pptable(ptable),psym(sym)
  {
    
    pparent=dynamic_cast<QWidget*>(parent);

    pheader<<"El";
    pheader<<"Vis";
    pheader<<"Color";
    pheader<<"Shine";
    pheader<<"Specular";
    pheader<<"RadType";
    pheader<<"Radius";
  }

  // ----------------------------------------------------------------
  ~AtomPropsModel()
  {
  }

  // ----------------------------------------------------------------
  void reload(XFPLO::AtomPropertyTable *t=0)
  {
    beginResetModel();
    if(pptable && t) pptable=t;
    endResetModel();
  }

  // ----------------------------------------------------------------
  int rowCount(const QModelIndex & parent=QModelIndex()) const
  {
    Q_UNUSED(parent);
    if(pptable) return pptable->size();
    if(psym) return psym->nsort();
    return 0;
  }

  // ----------------------------------------------------------------
  int columnCount(const QModelIndex &parent=QModelIndex()) const
  {
    Q_UNUSED(parent);
    return pheader.size();
  }

  // ----------------------------------------------------------------
  QVariant headerData(int section, Qt::Orientation orientation, int role) const
  {
    if (role != Qt::DisplayRole)
      return QVariant();
    if (orientation == Qt::Horizontal) {
      return pheader.at(section);
    }else{
      return section+1;
    }
    return QVariant();
  }

  // ----------------------------------------------------------------
  Qt::ItemFlags flags(const QModelIndex &index) const
  {
    if (!index.isValid()) return Qt::ItemIsEnabled;
    
    int row=index.row();
    int col=index.column();
    
    if (row >= rowCount()    || row < 0) return QAbstractTableModel::flags(index);
    if (col >= columnCount() || col < 0) return QAbstractTableModel::flags(index);
    
    
    if(psym && col==Visible){
      return QAbstractTableModel::flags(index)  | Qt::ItemIsUserCheckable;
    }
    if(col!=El){
      return QAbstractTableModel::flags(index) | Qt::ItemIsEditable;
    }
    
    return QAbstractTableModel::flags(index);
  }

  // ----------------------------------------------------------------
  QVariant data(const QModelIndex &index, int role) const
  {
    NSF("SpspModel::data");
    if (!index.isValid()) return QVariant();
    
    int row=index.row();
    int col=index.column();
    
    if (row >= rowCount()    || row < 0) return QVariant();
    if (col >= columnCount() || col < 0) return QVariant();
    
    const AtomProperty *p=0;
    if(pptable) p=&(pptable->at(row));
    if(psym) p=&(psym->wyckoffPosition(row).property());
    if(!p) return QVariant();

    if (role == Qt::DisplayRole || role == Qt::EditRole) {
      switch(col){
      case El:
	if(pptable) return pptable->element(row)();
	if(psym) return psym->wyckoffPosition(row).el()();
      case AtomColor:
	return p->material().color();
      case AtomShininess:
	return p->material().shininess();
      case SpecularColor:
	return p->material().specularColor();
      case RadiusType:
	if(role == Qt::DisplayRole)
	  return AtomicRadii::radiinames
	    .find(Pair<int,const char*>(p->atomicRadii().currentType(),""))->v2;
	else if(role == Qt::EditRole)
	  return p->atomicRadii().currentType();
      case Radius:
	return p->atomicRadii().current(Xf::Angstroem);
      default:
	return QVariant();
      }
    }else if( role == Qt::CheckStateRole ){
      if(col==Visible && psym){
	return (psym->wyckoffPosition(row).visible()?2:0);
      }
    }
    return QVariant();
    
  }
  
  // ----------------------------------------------------------------
  bool setData(const QModelIndex &index, const QVariant &value
	       , int role=Qt::EditRole)
  {
    NSF("SpspModel::setData"); 
    //  myDebug() << nsubfunc;
    
    if (!index.isValid()) return false;
    
    int row = index.row();
    int col = index.column();
  
    AtomProperty *p=0;
    if(pptable) p=&((*pptable)[row]);
    if(psym) p=&(psym->wyckoffPosition(row).property());

    if(!p) return false;
    if ( role == Qt::EditRole) {
      
      QModelIndex index1;      
      switch(col)
	{
	case AtomColor:
	  p->material().setColor(value.toUInt());
	  break;
	case SpecularColor:
	  p->material().setSpecularColor(value.toUInt());
	  break;
	case RadiusType:
	  p->atomicRadii().setCurrentType((AtomicRadii::AtomicRadiusType)value.toInt());
	  index1=this->index(row,Radius,QModelIndex());
	  emit(dataChanged(index1, index1));
	  break;
	case Radius:
	  p->atomicRadii().setCurrent(value.toDouble());
	  break;
	case AtomShininess:
	  p->material().setShininess(value.toDouble());
	  break;
	default:
	  return false;
	}
      emit(dataChanged(index, index));
      
      return true;
    }else if ( role == Qt::CheckStateRole) {
      if(col==Visible && psym){
	if(!value.toInt() && Settings("structure").getBool("Atoms/ShowHidden",false)){
	  if(pparent){
	    QToolTip::showText(pparent->mapToGlobal(QPoint(0,0)),
			       "Currently the \"show hidden atoms\""
			       "\nbutton is activeated.",
			       pparent);
	  }
	}
	psym->wyckoffPosition(row).setVisible(value.toInt());
	emit(dataChanged(index, index));
      }
    }
    
    return false;
    
  }

  // ----------------------------------------------------------------
  void setDefaults(SetDefaultsActions action,const QString &theme)
  {
    if(!psym) return;
    if(action==ColorByZ){
      beginResetModel();
      psym->applyTheme(globalproperties->find(theme));
      endResetModel();
      
    }else{
      myDebug() <<"unknown SetDefaultsActions"<<action;
    }
    
  }

  
private:
  XFPLO::AtomPropertyTable *pptable;
  Symmetry *psym;
  QStringList pheader;
  QWidget *pparent;
};


#ifdef	__cplusplus
extern "C" {
#endif

#ifdef	__cplusplus
}
#endif


#endif /* _ATOMPROPSMODEL_H */

