/* @(#)BondModel.hh
 * author: magru@unas
 */

#ifndef _BONDMODEL_HH
#define _BONDMODEL_HH 1

#include <QAbstractItemModel>
#include <QDebug>
#include "Symmetry.hh"
#include "BondDefs.hh"
#include "IconCache.hh"
#include "BBins.hh"
#include "MyVariant.hh"

class BondModel : public QAbstractItemModel
{
  Q_OBJECT

public:

  // ================================================================
  class PrimaryAtomData{
  public:
    
    // --------------------------------------------------------------
    PrimaryAtomData(PolyhedronProperties *pp,BondDefGroupEnum::CutOffMode m)
      :polyprops(pp),mode(m)
    {
    }
    
    PolyhedronProperties *polyprops;
    BondDefGroupEnum::CutOffMode mode;
  };

  // ================================================================
public:
  enum Items{PrimaryAtom=0};
  enum Columns{Primary=0};

  // ----------------------------------------------------------------
  BondModel( QObject * parent=0)
    : QAbstractItemModel(parent),ppropagatechanges(true),psym(0),pmaxlevels(0)
  {
    //pheader<<"";
    pheader<<"Bonds";
  }

  // ----------------------------------------------------------------
  ~BondModel()
  {
  }

  // ----------------------------------------------------------------
  void resetModel(const Symmetry *sym,BondDefGroup *bdg,BBins::BinsCenterType bct,
		  bool keep=false)
  {
    beginResetModel();
    if(!keep) {
      psym=sym; 
      pbdg=bdg;
    }
    pmaxlevels=1;
    pbins.clear();
    if(psym && pbdg){
      long nsort=psym->nsort();
      for(long sort=-1;++sort<nsort;){
	long site=psym->siteBySort(sort);
	long type=psym->wyckoffPosition(sort).type();
	BBins *bbins=new BBins();
	bbins->calculateBins(bct,pbdg->topLevelGrouping()
			     ,sort,type,site,psym->pairs(),pbdg->binTol());
	bbins->initializeCheckState(pbdg);
	pbins.append(bbins);
	//pbins.last()->reportBins(psym);
	pmaxlevels=MyMath::max<long>(pmaxlevels,bbins->levels());
      }
    }
    endResetModel();
    if(pbdg) myDebug()<<"BondModel::resetModel"<<*pbdg;
    pmaxlevels++;
  }

  // ----------------------------------------------------------------
  Bin *getBin(const QModelIndex &index) const
  {
    //myDebug()<<"Bin *getBin(const QModelIndex) const";
    //qDebug()<<"index="<<index;
    if (index.isValid() && (index.internalId()!=PrimaryAtom)) {
      //qDebug()<<"index.internalId()="<<index.internalId();
      Bin *bin=static_cast<Bin*>(index.internalPointer());
      //myDebug()<<"bin="<<bin;
      if (bin) return bin;
      //else return 0;
    }else if(index.internalId()==PrimaryAtom){
      return pbins.at(index.row())->rootBin();
    }
    return 0;
  }

  // ----------------------------------------------------------------
  int rowCount(const QModelIndex & parent=QModelIndex()) const
  {
    Q_UNUSED(parent);
    
    if(psym&&pbdg) {
      if(!parent.isValid()){
	// root 
	return psym->nsort();
      }
      return getBin(parent)->size();
    }
    return 0;
  }

  // ----------------------------------------------------------------
  int columnCount(const QModelIndex &parent=QModelIndex()) const
  {
    Q_UNUSED(parent);
    return pheader.size();
  }

  // ----------------------------------------------------------------
  QVariant headerData(int section, Qt::Orientation orientation, int role) const
  {
    if (role != Qt::DisplayRole)
      return QVariant();
    if (orientation == Qt::Horizontal) {
      return pheader.at(section);
    }else{
      return section+1;
    }
    return QVariant();
  }

  // ----------------------------------------------------------------
  Qt::ItemFlags flags(const QModelIndex &index) const
  {
    if (!index.isValid()) return 0;
    
    int row=index.row();
    int col=index.column();
    

    if (row >= rowCount(index.parent())    || row < 0) return 0;
    if (col >= columnCount() || col < 0) return 0;

    
    Bin *bin=getBin(index);
    if(col==Primary){
      if(index.internalId()==PrimaryAtom){
	return   Qt::ItemIsEnabled | Qt::ItemIsSelectable /* | Qt::ItemIsEditable*/ ;
      }else{
	if(bin->isCheckable()){
	  return   Qt::ItemIsEnabled | Qt::ItemIsSelectable 
	    | Qt::ItemIsUserCheckable;
	}else{
	  return   Qt::ItemIsEnabled | Qt::ItemIsSelectable ;
	}
      }
    }else{
      return   Qt::ItemIsEnabled | Qt::ItemIsSelectable ;
    }
    
    return 0;

  }
  // ----------------------------------------------------------------
  QModelIndex index(int row, int column,
		    const QModelIndex &parent = QModelIndex()) const
  {
    //myDebug()<<"BondModel::index()";
    //  if (parent.isValid() && parent.column() != 0) return QModelIndex();
    if (column<0 || column>= columnCount()) return QModelIndex();
    if (row<0 || row>= rowCount(parent)) return QModelIndex();
    if (!hasIndex(row, column, parent)) return QModelIndex();

    QModelIndex rindex=QModelIndex();

    if(!parent.isValid()){
      rindex=createIndex(row,column,PrimaryAtom);
    }else{
      if(parent.internalId()==PrimaryAtom){
	long sort=parent.row();
	const Bin *bin=pbins.at(sort)->rootBin()->bin(row);
	return createIndex(row, column, (void*)bin);
	
      }else{
	Bin *parentBin=getBin(parent);
	const Bin *bin=parentBin->bin(row);
	if (bin){
	  return createIndex(row, column, (void*)bin);
	}
      }
    }
    return rindex;
  }

  // ----------------------------------------------------------------
  QModelIndex parent(const QModelIndex &index=QModelIndex()) const
  {
    //qDebug()<<"index="<<index;
    if (!index.isValid()) return QModelIndex();

    //qDebug()<<"index.internalId()="<<index.internalId();
    if(index.internalId()==PrimaryAtom){
      //root items do not have parents
      return QModelIndex();
    }else{

      Bin *bin=getBin(index);
      Bin *parentBin=bin->parent();

      if (parentBin->binType()==Bin::BTRoot) 
	return createIndex(parentBin->bbins()->sort(), Primary, PrimaryAtom);
      
      return createIndex(parentBin->binNumber(), Primary, (void*)parentBin);

    }
    //  todo: make the root item clickable instead of bondicon?
  }

  // ----------------------------------------------------------------
  QVariant data(const QModelIndex &index, int role) const
  {
    if(!index.isValid() || !psym || ! pbdg ) return QVariant();

    long row=index.row();
    long col=index.column();


    
    if(index.internalId()==PrimaryAtom){
      
      if (row >= psym->nsort() || row < 0) return QVariant();
      
      const WyckPos &pos= psym->wyckoffPosition(row);
      if (role == Qt::DisplayRole) {
	if (col==Primary) {
	  return Sprintf("%- 2s (T%ld) %s %s %s",pos.el()(),pos.type()
			 ,pos.tauS(0)() ,pos.tauS(1)() ,pos.tauS(2)()
			 //,pos.tauD().printS("%20.5f"," ",true)()
			 )();
	}
      // }else if (role == Qt::SizeHintRole && col==Primary) {
      //  	return QSize(-1,20);
      }else if(role == Qt::DecorationRole) {
	if(col==Primary){
	  //myDebug()<<row<<pbdg->hasBonds(row);
	  const PolyhedronProperties & polyprops=pbdg->polyProperties(row);
	  return iconcache.getBondAtomIcon(psym->sites().at(psym->siteBySort(row))
					   .property().material().color(),
					   pbdg->hasBonds(row),
					   polyprops,
					   pbdg->atomCutOffMode(row))
	    .pixmap(QSize(48,16));
	}
      }else if(role == Qt::EditRole) {
	//note: the receiver has to delete this pointer, this means
	// it is not suited for standard model/view editing!
	return toQVariant(new PrimaryAtomData(&(pbdg->polyProperties(row)),
					      pbdg->atomCutOffMode(row)
					      ));
      }
      
    }else{
      if(col==Primary){
	Bin *bin=getBin(index);
	if (role == Qt::DisplayRole) {
	  return bin->toString()();
	}else if( role == Qt::DecorationRole ){
	  if(
	     bin->binType()==Bin::BTWhatEver||
	     bin->binType()==Bin::BTType
	     ){
	    return QIcon(":/Icons/ball164colors");
	  }else{
	    long site=bin->site();
	    if(0<=site && site<psym->nsite()){
	      return iconcache.getAtomIcon(psym->sites().at(site).property()
					       .material().color());
	    }else{
	      return 0x00;
	    }
	  }
	}else if( role == Qt::CheckStateRole ){
	  if( bin->isCheckable()) return bin->checkState();
	}//role
      }
      
    }


    return QVariant();
    
  }

  // ----------------------------------------------------------------
  bool setData(const QModelIndex &index, const QVariant &value,int role)
  {
    //myDebug() << "setData"<<index << role << value;
    
    if(index.internalId()!=PrimaryAtom){
      if (role == Qt::CheckStateRole){
	
	Bin *bin=getBin(index);
	bin->setCheckState(value.toInt());
	
	
	bool checked=value.toInt()!=Qt::Unchecked;
	// propagateDataChangeDown(index);
	// propagateDataChangeUp(index);
	myDebug()<<"setData: index="<<index.row();
	pbdg->setFromBinSelection(bbins(index),bin,checked);
	if(ppropagatechanges){
	  propagateDataChange();
	}
	return true;
      }      
    }else{
      if( role == Qt::EditRole ){
	long sort=index.row();
	BondModel::PrimaryAtomData *data=fromQVariant<BondModel::PrimaryAtomData>(value);
	PolyhedronProperties *pp=data->polyprops;
	pbdg->setAtomCutOffMode(sort,data->mode);
	PolyhedronProperties &pro=pbdg->polyProperties(sort);
	pro.setVisible(pp->isVisible());
	pro.setDrawFaces(pp->drawFaces());
	pro.setFaceMaterial(pp->faceMaterial());
	pro.setDrawRim(pp->drawRim());
	pro.setRimMaterial(pp->rimMaterial());
	pro.setRimRadiusUnits(pp->rimRadiusUnits());
	pro.setRimRadius(pp->rimRadius());


	emit dataChanged(index,index);
	delete data;

      }      
    }
    return false;
  }

  // ----------------------------------------------------------------
  void propagateDataChangeDown(const QModelIndex &index)
  {
    if(!index.isValid())return;
    //give state change down the hierarchy
    //myDebug()<<*(getBin(index));
    for(int i=0;i<rowCount(index);i++){
      QModelIndex childIndex=this->index(i,0,index);
      propagateDataChangeDown(childIndex);
    }
    emit dataChanged(index, index);
    
  }

  // ----------------------------------------------------------------
  void propagateDataChangeUp(const QModelIndex &index)
  {
    if(!index.isValid())return;
    //give state change up the hierarchy
    QModelIndex parentIndex=parent(index);
    if(parentIndex.isValid()){
      emit dataChanged(parentIndex,parentIndex);
      propagateDataChangeUp(parentIndex);
    }
  }

  // ----------------------------------------------------------------
  void propagateDataChange()
  {
    for(int i=-1;++i<rowCount(QModelIndex());){
      QModelIndex index=this->index(i,0,QModelIndex());
      //myDebug()<<"propagateDataChange"<<*getBin(index);
      pbins[i]->initializeCheckState(pbdg);
      propagateDataChangeDown(index);    
    }
  }

  // ----------------------------------------------------------------
  BBins *bbins(const QModelIndex &current)
  {
    if(!current.isValid())return 0;
    if(current.internalId()==PrimaryAtom) return 0;

    Bin *bin=getBin(current);
    while(bin->binType()!=Bin::BTRoot){
      bin=bin->parent();
    }
    return bin->bbins();
  }

  // ----------------------------------------------------------------
  long currentPrimaryIndex(const QModelIndex &current)
  {
    if(!current.isValid())return -1;
    //myDebug()<<(long)current.internalId();
    if(current.internalId()==PrimaryAtom){
      return current.row();
    }else{
      return bbins(current)->sort();
    }
  }
  
  // ----------------------------------------------------------------
  int maxLevels()const 
  {
    return pmaxlevels;
  }


private:
  PROPERTY(propagatechanges,propagateChanges,PropagateChanges,bool);
  const Symmetry *psym;
  PtrArray<BBins> pbins;
  BondDefGroup *pbdg;
  QStringList pheader;
  int pmaxlevels;

};

#endif /* _BONDMODEL_HH */

