/* @(#)CombiModel.hh
 * author: magru@unas
 */

#ifndef _COMBIMODEL_HH
#define _COMBIMODEL_HH 1

#include <QAbstractItemModel>
#include <QBrush>
#include <QStringList>
#include "MolecularOrbital.hh"
#include "CoeffBasisInfo.hh"
#include "MyVariant.hh"

using namespace XFPLO;

/* ***********************************************************************
 *
 * ***********************************************************************
 */

class CombiModel : public QAbstractItemModel
{
  Q_OBJECT;
public:

  enum Columns{Name=0,Site=1,Orbital=2,Position=3,Factor=4,Xaxis=5,Zaxis=6};

  // ----------------------------------------------------------------
  CombiModel( QObject * parent=0)
    : QAbstractItemModel(parent),pmolorbs(0),pcbi(0)
  {
    pheader<<"Name"<<"Site"<<"Orbital"<<"Position"<<"Factor"<<"Xaxis"<<"Zaxis";
  }

  // ----------------------------------------------------------------
  ~CombiModel()
  {
  }

  // ----------------------------------------------------------------
  void resetModel(MolecularOrbitals *mos=0,const CoeffBasisInfo *cbi=0)
  {
    beginResetModel();
    if(mos) pmolorbs=mos;
    if(cbi) pcbi=cbi;
    endResetModel();
  }


  // ----------------------------------------------------------------
  QVariant headerData(int section, Qt::Orientation orientation, int role) const
  {
    if (role != Qt::DisplayRole)
      return QVariant();
    if (orientation == Qt::Horizontal) {
      return pheader.at(section);
    }else{
      return section+1;
    }
    return QVariant();
  }
  
  // ----------------------------------------------------------------
  inline bool isTopLevelItem(const QModelIndex &index)const
  {
    return (index.internalPointer()==pmolorbs);
  }
  
  // ----------------------------------------------------------------
  MolecularOrbital *molOrb(const QModelIndex &index) const
  {
    if(!index.isValid() || !pmolorbs) return 0;

    if(isTopLevelItem(index)) return (*pmolorbs)[index.row()];
    
    return static_cast<MolecularOrbital*>(index.internalPointer());
  }
  // ----------------------------------------------------------------
  int rowCount(const QModelIndex &parent) const
  {
    if(!pmolorbs) return 0;
    if(!parent.isValid()){
      //root
      return pmolorbs->size();
    }
    if(parent.internalPointer()==pmolorbs){
      MolecularOrbital *mo=molOrb(parent);
      if(mo->isMolecular()){
	return mo->numOfContribs();
      }else{
	return 0;
      }
    }
    return 0;
    
  }

  // ----------------------------------------------------------------
  int columnCount(const QModelIndex &parent=QModelIndex()) const
  {
    Q_UNUSED(parent);
    return pheader.size();
  }

  // ----------------------------------------------------------------
  QModelIndex index(int row, int column, const QModelIndex&parent) const
  {
    if (column<0 || column>= columnCount() || !pmolorbs) return QModelIndex();
    if (row<0 || row>= rowCount(parent)) return QModelIndex();
    if (!hasIndex(row, column, parent)) return QModelIndex();

    if(!parent.isValid()){
      return createIndex(row, column, pmolorbs);
    }else{
      return createIndex(row, column, (*pmolorbs)[parent.row()]);
    }


    return QModelIndex();
  }

  // ----------------------------------------------------------------
  QModelIndex parent(const QModelIndex &index) const
  {
    Q_UNUSED(index);

    if(!index.isValid() || isTopLevelItem(index) || !pmolorbs) return QModelIndex();

    return createIndex(pmolorbs->indexOf(molOrb(index)),0,pmolorbs);
    

    return QModelIndex();
  }

  // ----------------------------------------------------------------
  Qt::ItemFlags flags(const QModelIndex &index) const
  {
    if (!index.isValid()) return Qt::ItemIsEnabled;

    if(isTopLevelItem(index)){
      return QAbstractItemModel::flags(index)  | Qt::ItemIsUserCheckable;
    }else{
      return QAbstractItemModel::flags(index);
    }
  }

  // ----------------------------------------------------------------
  QModelIndex topLevelIndex(const QModelIndex &index)const
  {
    if(!index.isValid()) return QModelIndex();
    if(!index.parent().isValid()){
      return this->index(index.row(),0,QModelIndex());
    }else{
      return index.parent();
    }
  }
  
  // ----------------------------------------------------------------
  QVariant data(const QModelIndex &index, int role) const
  {
    if(!index.isValid() || !pmolorbs ) return QVariant();

    long row=index.row();
    long col=index.column();

    bool iscontrib=false;
    bool ismol=false;
    const MolecularOrbital *mo=molOrb(index);
    QString name;
    long isa=0;
    Cmplx c;
    const MolecularOrbitalData *md=0;
    if(isTopLevelItem(index)){
      if(mo->isMolecular()){
	ismol=true;
	name=QString("%1: %2").arg(row+1).arg(mo->name()());
      }else{
	md=mo->at(0);
	name=QString("%1: auto").arg(row+1);
	isa=md->site();
      }
    }else{
      iscontrib=true;
      md=mo->at(row);
      name="";
      isa=md->site();
      c=md->factor();
    }

    
    if (role == Qt::DisplayRole) {
      switch(col){
      case Name:
	return name;
      case Site:
	if(!ismol){
	  if(pcbi->isReady())
	    return Sprintf("%2s(%03ld)",pcbi->element(isa)(),isa+1)();
	  else
	    return Sprintf("%03ld",isa+1)();
	}
	break;
      case Orbital:
	if(!ismol){
	  return md->orbital()();
	}
	break;
      case Position:
	if(iscontrib){
	  return md->position().printS("%16.8f"," , ",true).chop(0)();
	}
	break;
      case Factor:
	if(iscontrib){
	  if(c.isReal()){
	    return Sprintf("%32.8g",c.real()).stripBlanks()();
	  }else{
	    return (Sprintf("%32.8g",c.real()).stripBlanks()+"+i*"
		    +Sprintf("%32.8g",c.imag()).stripBlanks())();
	  }
	}
	break;
      case Xaxis:
	if(!ismol){
	  return md->xaxis().printS("%16.8f"," , ",true).chop(0)();
	}
	break;
      case Zaxis:
	if(!ismol){
	  return md->zaxis().printS("%16.8f"," , ",true).chop(0)();
	}
	break;
      }
      
    }else if(role == Qt::CheckStateRole && col==Name) {
      if(isTopLevelItem(index)){
	return (mo->isActive()?2:0);
      }
    }else if(role == Qt::BackgroundRole) {
      if(ismol){
	return QBrush(0xe0ffe0);
      }else if(!iscontrib){
	return QBrush(0xffe0e0);
      }

    }


    return QVariant();
  }

  // ----------------------------------------------------------------
  bool setData(const QModelIndex &index, const QVariant &value,int role)
  {
    if(!index.isValid() || !pmolorbs ) return false;
    
    if( role == Qt::EditRole ){
      MolecularOrbital *mo=molOrb(index);
      if(isTopLevelItem(index)){
	if(mo->isMolecular()){
	  MolecularOrbital *moin=fromQVariant<MolecularOrbital>(value);
	  mo->setName(moin->name());
	  emit dataChanged(index,index);
	}else{
	  MolecularOrbitalData *mdin=fromQVariant<MolecularOrbitalData>(value);
	  (*(*mo)[0])=*mdin;
	}
      }
    }else if(role == Qt::CheckStateRole && index.column()==Name) {
      MolecularOrbital *mo=molOrb(index);
      if(isTopLevelItem(index)){
	mo->setActive(value.toBool());
	emit dataChanged(index,index);
      }
    }

    return false;
  }
  // ----------------------------------------------------------------
  QModelIndex insertRow(const QModelIndex &index,MolecularOrbital *mo,
			bool prepend=false)
  {
    long row=0;
    long col=0;
    if(index.isValid()){
      col=index.column();
    }
    QModelIndex pindex=topLevelIndex(index);
    if(pindex.isValid()) row=pindex.row()+1;
    if(prepend) row=0;
    beginInsertRows(QModelIndex(),row,row);
    pmolorbs->insert(row,mo);
    endInsertRows();
    return this->index(row,col,QModelIndex());
  }
  
  // ----------------------------------------------------------------
  QModelIndex insertRow(const QModelIndex &index,MolecularOrbitalData *md,
			bool prepend=false)
  {
    if(!index.isValid()) return QModelIndex();
    QModelIndex pindex=topLevelIndex(index);
    MolecularOrbital *mo=molOrb(pindex);
    long row;
    if(isTopLevelItem(index)){
      // root item
      if(!mo->isMolecular()) {
	return QModelIndex();
      }
      row=rowCount(pindex);
      if(prepend)row=0;
    }else{
      row=index.row()+1;
    }
    beginInsertRows(pindex,row,row);
    mo->insert(row,md);
    endInsertRows();
    return this->index(row,index.column(),pindex);
  }

  // ----------------------------------------------------------------
  void *removeRow(const QModelIndex &index)
  {
    if(!index.isValid()) return 0;
    long row=index.row();
    if(isTopLevelItem(index)){
      //root item
      
      beginRemoveRows(index.parent(),row,row);
      MolecularOrbital *mo=pmolorbs->takeAt(row);
      endRemoveRows();
      return mo;
      
    }else{
      // contrib
      beginRemoveRows(index.parent(),row,row);
      MolecularOrbital *mo=molOrb(index);
      MolecularOrbitalData *md=mo->takeAt(row);
      endRemoveRows();
      return md;
    }
  }

  // ----------------------------------------------------------------
  QModelIndex nextMolecularIndex(const QModelIndex &index,int direction)const
  {
    QModelIndex pindex=topLevelIndex(index);
    
    while(pindex.row()+direction>=0 && pindex.row()+direction<=pmolorbs->size()-1){
      pindex=this->index(pindex.row()+direction,0,pindex.parent());
      if(molOrb(pindex)->isMolecular()){
	return pindex;
      }
    }

    return QModelIndex();
  }

  // ----------------------------------------------------------------
  QModelIndex moveRow(const QModelIndex &index,int by)
  {
    if(!index.isValid()) return QModelIndex();

    long row=index.row();
    if(isTopLevelItem(index)){
      //root item
      int lrow=rowCount(index.parent())-1;
      long newrow=row+by;
      if(newrow<0) return QModelIndex();
      if(newrow>lrow) return QModelIndex();
      newrow--;
      bool prepend=false;
      if(newrow<0) {
	prepend=true;
	newrow=0;
      }
      QModelIndex newindex=this->index(newrow,index.column(),index.parent());
      return insertRow(newindex,static_cast<MolecularOrbital*>(removeRow(index)),prepend);
    }else{
      // contrib item
      QModelIndex pindex=index.parent();
      int lrow=rowCount(pindex)-1;
      long newrow=row+by;
      if(newrow<0) {
	if(!(pindex=nextMolecularIndex(index,by)).isValid()) return QModelIndex();
	newrow=rowCount(pindex);
      }else if(row+by>lrow){
	if(!(pindex=nextMolecularIndex(index,by)).isValid()) return QModelIndex();
	newrow=0;
      }
      QModelIndex newindex;
      MolecularOrbitalData* md=static_cast<MolecularOrbitalData*>(removeRow(index));
      QModelIndex rindex;
      if(newrow<=0){
	newindex=pindex;
      }else{
	newindex=this->index(newrow-1,0,pindex);
      }
      insertRow(newindex,md,true);
      rindex=this->index(newrow,index.column(),pindex);
      return rindex;
    }
    

    return QModelIndex();
  }

  // ----------------------------------------------------------------
  QModelIndex copyRow(const QModelIndex &index)
  {
    if(isTopLevelItem(index)){
      return insertRow(index,new MolecularOrbital(*molOrb(index)));
    }else{
      return insertRow(index,new MolecularOrbitalData(*(*molOrb(index))[index.row()]));
    }
    return QModelIndex();
  }


  
protected:
private:
  QStringList pheader;
  MolecularOrbitals *pmolorbs;
  const CoeffBasisInfo *pcbi;
};


#endif /* _COMBIMODEL_HH */

