/*************************************************************************
 * @(#)SpspModel.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   08 Apr 2011
 *************************************************************************
 */

#include "MyMath.hh"
#include "SpspModel.hh"
#include "Spsp.hh"
#include "MyDebug.hh"
#include "Vec3.hh"
#include "Valid.hh"

 
/* ***********************************************************************
 *
 * ***********************************************************************
 */
SpspModel::SpspModel( QObject * parent,SpecialSymPoints *spsp)
  : QAbstractTableModel(parent),pspsp(spsp),psym(0)
{
  NSF("SpspModel::SpspModel");   
  MASSERT("parameter SpecialSymPoints *spsp",spsp);
  pheader=QString("blah Name x y z Tdepth point PSize font TScale").split(" ");

  pheader[Active]="";
  pheader[PointColor]="Pcol";
  pheader[TextColor]="Tcol";
  pheader[PointRadius]="PSize";
  pheader[FontScale]="FScale";
  xyzvalid=new RatValidator(this,0,"x,y,z",-1e10,1e10);
  sfxyzvalid=new RatValidator(this,0,"sx,sy,sz",-1e10,1e10);
  psizevalid=new RatValidator(this,0,"point size",0.0001,1e10);
  fscalevalid=new RatValidator(this,0,"text scale",0.0001,1e10);

  for(int i=0;i<pheader.size();i++){
    columnEnabled.append(true);
  }

}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
SpspModel::~SpspModel()
{
  delete xyzvalid;
  delete sfxyzvalid;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
int SpspModel::rowCount(const QModelIndex & parent ) const
{
  Q_UNUSED(parent);
  if(!pspsp) return 0;
  return pspsp->size();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
int SpspModel::columnCount(const QModelIndex & parent ) const
{
  Q_UNUSED(parent);
  return pheader.size();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Qt::ItemFlags SpspModel::flags(const QModelIndex &index) const
{
  if (!index.isValid())
    return Qt::ItemIsEnabled;

  int row=index.row();
  int col=index.column();

  if (row >= pspsp->size() || row < 0) return QAbstractTableModel::flags(index);
  if (col >= columnCount() || col < 0) return QAbstractTableModel::flags(index);

  if(!columnEnabled[index.column()]) return 0;

  const SpecialPoint *p=pspsp->at(row);
  if(p->type()==Xf::SPTJump||(p->type()==Xf::SPTBreak)){
    if(col==Active)
      return QAbstractTableModel::flags(index) |
	Qt::ItemIsUserCheckable;
    else// if(col==Name)
      return QAbstractTableModel::flags(index);
      //     else
      //return Qt::NoItemFlags;
  }

  if(col==Active)
    return QAbstractTableModel::flags(index) |
      Qt::ItemIsUserCheckable;
  else
    return QAbstractTableModel::flags(index) | Qt::ItemIsEditable;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
QVariant SpspModel::headerData(int section, Qt::Orientation orientation
			       , int role) const
{
  if (role != Qt::DisplayRole)
    return QVariant();
  if (orientation == Qt::Horizontal) {
    return pheader.at(section);
  }
  return QVariant();

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
QVariant SpspModel::data(const QModelIndex &index, int role) const
{
  NSF("SpspModel::data");
  if (!pspsp) return QVariant();
  if (!index.isValid()) return QVariant();

  int row=index.row();
  int col=index.column();

  if (row >= pspsp->size() || row < 0) return QVariant();
  if (col >= columnCount() || col < 0) return QVariant();
  
  const SpecialPoint *p=pspsp->at(row);
  if (role == Qt::DisplayRole || role == Qt::EditRole) {
    if(p->type()==Xf::SPTJump){
      if(col==Name){
	return "Jump";
      }else
	return QVariant();
    }else if(p->type()==Xf::SPTBreak){
      if(col==Name){
	return "Break";
      }else
	return QVariant();
    }

    //    myDebug() << nsubfunc<<col<<p->textColor();;
    switch(col){
    case Name:
      return p->name();
    case PointPosX:
      return p->point()[0].S();
    case PointPosY:
      return p->point()[1].S();
    case PointPosZ:
      return p->point()[2].S();
    case TextDepth:
      return p->shift()[2].S();
    case PointColor:
      return p->pointColor();
    case PointRadius:
      return p->pointRadius().S();
    case TextColor:
      return p->textColor();
    case FontScale:
      return p->fontScale().S();
    default:
      return QVariant();
    }
  }else if( role == Qt::CheckStateRole ){
    if(col==Active) 
      return (p->isVisible()?2:0);
    else
      return QVariant();
  }
  return QVariant();

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool SpspModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
  NSF("SpspModel::setData"); 
  //  myDebug() << nsubfunc;

  if(!pspsp) return false;
  if (!index.isValid()) return false;

  int row = index.row();
  int col = index.column();
  SpecialPoint *p=(*pspsp)[row];
  
  if (role == Qt::CheckStateRole){
    if(col==Active){
      p->setVisible(value.toBool());
      emit(dataChanged(index, index));
      return true;
    }else
      return false;
  }else if ( role == Qt::EditRole) {
    if(p->type()!=Xf::SPTPoint) return false;
      
    SReal3 point=p->point();
    SReal3 shift=p->shift();
    //myDebug() << nsubfunc<<col<<value;
    switch(col){
    case Name:
      p->setName(value.toString());
      break;
    case PointPosX:
      point[0]=value.toString();
      break;
    case PointPosY:
      point[1]=value.toString();
      break;
    case PointPosZ:
      point[2]=value.toString();
      break;
    case TextDepth:
      shift[2]=value.toString();
      break;
    case PointColor:
      p->setPointColor(value.toUInt());
      break;
    case PointRadius:
      p->setPointRadius(value.toString());
      break;
    case TextColor:
      p->setTextColor(value.toUInt());
      break;
    case FontScale:
      p->setFontScale(value.toString());
      break;
    default:
      return false;
    }
    p->setPoint(point);
    p->setShift(shift);
    emit(dataChanged(index, index));

    return true;
  }
  
  return false;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void SpspModel::validateData(QString &txt)
{
  NSF("SpspModel::validateData");
  //  myDebug() << nsubfunc << "start";
  QModelIndex index;
  for(int row=0; row < rowCount(); row++){
    if(pspsp->at(row)->type()!=Xf::SPTPoint) continue;
    long err=0;
    QString msg;
    {
      index=this->index(row,PointPosX,QModelIndex());
      QString s=data(index,Qt::DisplayRole).toString();
      err=!xyzvalid->check(msg,s);
    }
    {
      index=this->index(row,PointPosY,QModelIndex());
      QString s=data(index,Qt::DisplayRole).toString();
      err|=!xyzvalid->check(msg,s);
    }
    {
      index=this->index(row,PointPosZ,QModelIndex());
      QString s=data(index,Qt::DisplayRole).toString();
      err|=!xyzvalid->check(msg,s);
    }
    {
      index=this->index(row,TextDepth,QModelIndex());
      QString s=data(index,Qt::DisplayRole).toString();
      err|=!sfxyzvalid->check(msg,s);
    }
    {
      index=this->index(row,PointRadius,QModelIndex());
      QString s=data(index,Qt::DisplayRole).toString();
      err|=!psizevalid->check(msg,s);
    }
    {
      index=this->index(row,FontScale,QModelIndex());
      QString s=data(index,Qt::DisplayRole).toString();
      err|=!fscalevalid->check(msg,s);
    }

    if(err)
      txt+=QString("Symmetry point %1:").arg(row+1)+msg;
  }
  //myDebug() << nsubfunc << "end"<<txt;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void SpspModel::setDefaultForColumn(Columns action,const QVariant &value)
{
  //myDebug()<<action;
  for(int row=0; row < rowCount(); row++){
    QModelIndex index;
    index=this->index(row,action,QModelIndex());
    if(index.isValid()){
      setData(index,value,Qt::DisplayRole|Qt::EditRole);
    }
  }
  //reset();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void SpspModel::setColumnsEnabled(bool en)
{
  columnEnabled[PointRadius]=en;
  columnEnabled[PointColor]=en;
  columnEnabled[TextColor]=en;
  columnEnabled[FontScale]=en;
  beginResetModel();
  endResetModel();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

bool SpspModel::insertRow(int position,Xf::SpecialPointType type,SpecialPoint *p)
{
  beginInsertRows(QModelIndex(),position,position);
  pspsp->insertNewPoint(position,type,p);
  pspsp->setChanged(true);
  endInsertRows();
  return true;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
SpecialPoint *SpspModel::removeRow(int position)
{
  SpecialPoint *p;
  beginRemoveRows(QModelIndex(), position, position);
  p=pspsp->removePoint(position);
  pspsp->setChanged(true);
  endRemoveRows();
  return p;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool SpspModel::moveRow(int row,int by)
{

  if(!index(row+by,0,QModelIndex()).isValid()) return false;
  

  SpecialPoint *p=removeRow(row);
  insertRow(MyMath::max<long>(0,row+by),p->type(),p);
  pspsp->setChanged(true);

  return true;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void SpspModel::setPoint(int row,const Vec3 &v)
{
  
  (*pspsp)[row]->setPoint(SReal3(v[0],v[1],v[2]));
  emit(dataChanged(this->index(row,PointPosX), this->index(row,PointPosZ)));
}


/*EOF*/ 
