/*************************************************************************
 * @(#)WyckPosModel.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   04 Jul 2009
 *************************************************************************
 */
#include <QIcon>
#include "WyckPosModel.hh"
#include "MyMath.hh"
#include "IconCache.hh"
#include "PropertyTable.hh"
#include "Symmetry.hh"
#include "MyDebug.hh"

/* ***********************************************************************
 *
 * ***********************************************************************
 */
WyckPosModel::WyckPosModel( QObject * parent ,Symmetry *sym)
  : QAbstractTableModel(parent),psym(sym),pallowDefaultOnSetData(false)
{
  NSF("WyckPosModel::WyckPosModel");   
  //  myDebug() << nsubfunc;
  MASSERT("parameter Symmetry *sym",sym);
  //  myDebug("sym(%p)\n",sym);
  //  myDebug("psym(%p)\n",psym);
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
WyckPosModel::~WyckPosModel()
{
  NSF("WyckPosModel::~WyckPosModel"); Q_UNUSED(nsubfunc);
  //  myDebug() << nsubfunc;
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
int WyckPosModel::rowCount(const QModelIndex & parent ) const
{
  Q_UNUSED(parent);
  if(!psym) return 0;
  return psym->nsort();
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
int WyckPosModel::columnCount(const QModelIndex & parent ) const
{
  Q_UNUSED(parent);
  return Z+1;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
QVariant WyckPosModel::data(const QModelIndex &index, int role) const
{
  //  myDebug()<<"WyckPosModel::data";
  
  if (!psym) return QVariant();
  if (!index.isValid()) return QVariant();
  
  

  if (index.row() >= psym->nsort() || index.row() < 0) return QVariant();


  int c=index.column();

  const WyckPos &pos= psym->wyckoffPosition(index.row());
  if (role == Qt::DisplayRole || role == Qt::EditRole) {
    
    if (c == El) return QString(pos.el()());
    else if (c == Type) return (int)pos.type();
    else if (X <= c && c<= Z) return pos.tauS(c-X)();
  }
  return QVariant();
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
QVariant WyckPosModel::headerData(int section, Qt::Orientation orientation, int role) const
{
  if (role != Qt::DisplayRole)
    return QVariant();
  
  if (orientation == Qt::Horizontal) {
    switch (section) {
    case El:
      return "Ele.";
    case Type:
      return "Type";
      
    case X:
      return "x";
    case Y:
      return "y";
    case Z:
      return "z";
      
    default:
      return QVariant();
    }
  }else{
    return section+1;
  }
  return QVariant();
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Qt::ItemFlags WyckPosModel::flags(const QModelIndex &index) const
{
  if (!index.isValid())
    return Qt::ItemIsEnabled;

  //  int c=index.column();
  return QAbstractTableModel::flags(index) | Qt::ItemIsEditable;
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool WyckPosModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
  NSF("WyckPosModel::setData"); 
  //myDebug() << nsubfunc;
  
  //  myDebug("pwyckpos(%p)\n",psym);

  if(!psym) return false;

  

  if (index.isValid() && role == Qt::EditRole) {
    int row = index.row();
    
    WyckPos pos = psym->wyckoffPosition(row);
    int c=index.column();
    bool updateprops=false;
    if (c <= Z){
      String s(value.toString());
      if(c==El){ 
	pos.setElement(s);
	if(pallowDefaultOnSetData){
	  pos.setProperty(globalproperties->current()->find(s));
	  pos.setDefaultType();
	  updateprops=true;
	}
      }else if(c==Type) pos.setType( value.toInt());
      else if (c <= Z)
        pos.setTauS(c-X,s);

    }else
      return false;
    psym->setWyckoffPosition(row,pos);
    QModelIndex index1=index;
    if(updateprops){
      index1=this->index(row,Type,QModelIndex());
    }
    emit(dataChanged(index, index1));
    
    return true;
  }
  
  return false;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
WyckPos WyckPosModel::removeRow(int position)
{
  
  beginRemoveRows(QModelIndex(), position, position);
  WyckPos w=psym->removeWyckoffPosition(position);
  endRemoveRows();
  return w;
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */

bool WyckPosModel::insertRow(int position,const WyckPos *w)
{
  beginInsertRows(QModelIndex(),position,position);
  psym->insertWyckoffPosition(position,w);
  endInsertRows();
  return true;
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
// const WyckPos *WyckPosModel::getData(int row) const
// {
  
//   if (!psym) return 0;
//   if (row >= psym->nsort() || row < 0) return 0;
  
//   return &psym->wyckoffPosition(row);
    
// }

/* ***********************************************************************
 *
 * ***********************************************************************
 */
bool WyckPosModel::moveRow(int row,int by)
{

  if(!index(row+by,0,QModelIndex()).isValid()) return false;
  

  WyckPos w=removeRow(row);
  insertRow(MyMath::max<long>(0,row+by),&w);

  return true;
  
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void WyckPosModel::setDefaults(SetDefaultsActions action
			       ,const QMap<QString,int>&map)
{
  if(action==TypeByZ){
    for(int row=0; row < rowCount(); row++){
      QModelIndex index;
      index=this->index(row,Type,QModelIndex());
      if(index.isValid()){
	WyckPos pos = psym->wyckoffPosition(row);
	pos.setDefaultType();
	setData(index,(int)pos.type(),Qt::DisplayRole|Qt::EditRole);
      }
      
    }
  }else if(action==TypePerElement){
    for(int row=0; row < rowCount(); row++){
      QModelIndex index;
      index=this->index(row,Type,QModelIndex());
      if(index.isValid()){
	WyckPos pos = psym->wyckoffPosition(row);
	if(map.contains(pos.el()())){
	  setData(index,map[pos.el()()],Qt::DisplayRole|Qt::EditRole);
	}
      }
      
    }
  }else{
    myDebug() <<"unknown SetDefaultsActions"<<action;
  }
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
QMap<QString,int> WyckPosModel::getElementTypeList()
{
  QMap<QString,int> map;
  for(int row=0; row < rowCount(); row++){
    WyckPos pos = psym->wyckoffPosition(row);
    map[pos.el()()]=pos.type();
  } 
  return map;
}

