/*************************************************************************
 * @(#)OGLDataCache.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   28 Apr 2010
 *************************************************************************
 */

#include "OGLDataCache.hh"
#include "OGLMeshPoints.hh"
#include "OGLMeshCells.hh"
#include "OGLTriangleMesh.hh"
#include "OGLLegend.hh"
#include "MyDebug.hh"




/* ***********************************************************************
 *
 * ***********************************************************************
 */
OGLDataCacheItem::OGLDataCacheItem(const QString &isa)
  :pisa(isa),pid(0),pnused(0),pconstructed(false)
  ,pdispListName(0),pverticessize(0),pindicessize(0)
  ,pnormals(0),pvertices(0),pcolors(0),pindices(0),pndefs(0)
  ,pdefs(0),pcull(true),pimage(0)
{
  pid=new char[1];
  pid[0]='\0';
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
OGLDataCacheItem::~OGLDataCacheItem()
{
  NSF("OGLDataCacheItem::~OGLDataCacheItem");
  //myDebug() << nsubfunc;
  if(pdispListName)
    glDeleteLists(pdispListName,1);

  if(pid) delete [] pid;

  cleanInternal();
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
void OGLDataCacheItem::cleanInternal()
{
  if(pdefs){ delete [] pdefs;pdefs=0;pndefs=0;};

  if(pnormals){ delete [] pnormals; pnormals=0;pverticessize=0; }
  if(pvertices){ delete [] pvertices; pvertices=0;pverticessize=0; }
  if(pcolors){ delete [] pcolors; pcolors=0; }
  if(pindices){ delete [] pindices; pindices=0;pindicessize=0; }
  if(pimage){ delete [] pimage; pimage=0; }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */

bool OGLDataCacheItem::operator==(const OGLDataCacheItem &other)
{
  NSF("OGLDataCacheItem::operator==");
  
  if(strcmp(pid,other.pid)==0) return true;
  return false;
  
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
QString OGLDataCacheItem::printSpecQ()const
{
  QString s;

  s.sprintf(
 	    "               verticessize=%ld\n"
 	    "               indicessize=%ld\n"
 	    "               displaylist=%d\n"
 	    "               used=%d times\n"
 	    ,pverticessize,pindicessize,pdispListName,pnused);
   

  return s;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void OGLDataCacheItem::makeRodMesh(long &nvert,long &idef,long &nn0,
				   long pnphi,
				   double radius,double L,
				   double z0,const Matrix3x3 &rot,
				   RodEnding left,RodEnding right,
				   bool initsize)
{
  // upper lid, pnphi+2 vertices, 1 def
  // rod, 2*(pnphi+1) vertices, 1 def
  // lower lid, pnphi+2 vertices, 1 def
  // cap (2*pnphi+2)*(pntheta-2)+pnphi+2 vertices, (pntheta-1) defs

  int pntheta=MyMath::max<long>(1,pnphi/2)+1;

  if(initsize){
    pverticessize+=2*pnphi+2; // rod
    pndefs++;
    if(left==Flat){
      pverticessize+=pnphi+2;
      pndefs++;
    }
    if(right==Flat){
      pverticessize+=pnphi+2;
      pndefs++;
    }
    if(left==Round){
      pverticessize+=(2*pnphi+2)*(pntheta-2)+pnphi+2;
      pndefs+=(pntheta-1);
    }
    if(right==Round){
      pverticessize+=(2*pnphi+2)*(pntheta-2)+pnphi+2;
      pndefs+=(pntheta-1);
    }
    return;
  }

  double *phis=new double[2*pnphi];
  double *ptheta=new double[2*pntheta];
  { //prepare the cos(phi),sin(phi) array
    double cp=1.;
    double sp=0.;
    double c=cos(2.*Pi/pnphi);
    double s=sin(2.*Pi/pnphi);
    phis[0]=cp; 
    phis[1]=sp;
    for(int j=1;j<pnphi;j++){
      double d=cp;
      cp=cp*c-sp*s;
      sp=sp*c+s*d;
      phis[2*j]=cp;
      phis[2*j+1]=sp;
    }
  }
  { //prepare the cos(theta) array
    double cp=1.;
    double sp=0.;
    double c=cos(Pi/2./(pntheta-1));
    double s=sin(Pi/2./(pntheta-1));
    ptheta[0]=cp; 
    ptheta[1]=sp;
    for(int j=1;j<pntheta;j++){
      double d=cp;
      cp=cp*c-sp*s;
      sp=sp*c+s*d;
      ptheta[2*j]=cp;
      ptheta[2*j+1]=sp;
    }

  }
  
  // rod, 2*(pnphi+1) vertices, 1 def
  for(int j=0;j<=pnphi;j++){
    long jj=(j+1)%pnphi;
    double c=phis[2*jj];
    double s=phis[2*jj+1];

    
    Vec3 v=rot*Vec3(c,s,0);
    pnormals[nvert*3+0]=v[0];
    pnormals[nvert*3+1]=v[1];
    pnormals[nvert*3+2]=v[2];
    
    pnormals[nvert*3+3]=v[0];
    pnormals[nvert*3+4]=v[1];
    pnormals[nvert*3+5]=v[2];
    
    v=rot*Vec3(radius*c,radius*s,L*0.5);
    pvertices[nvert*3+0]=v[0];
    pvertices[nvert*3+1]=v[1];
    pvertices[nvert*3+2]=v[2]+z0;
    
    nvert++;
    
    v=rot*Vec3(radius*c,radius*s,-L*0.5);
    pvertices[nvert*3+0]=v[0];
    pvertices[nvert*3+1]=v[1];
    pvertices[nvert*3+2]=v[2]+z0;
    
    nvert++;
    
  }
  
  pdefs[idef].mode=GL_QUAD_STRIP;
  pdefs[idef].count=2*(pnphi+1);
  pdefs[idef].start=nn0;
  pdefs[idef].useindex=false;
  idef++;
  
  nn0+=2*(pnphi+1);

  Vec3 n=rot*Vec3(0,0,-1);
  if(left==Flat){
    // lower lid, pnphi+2 vertices, 1 def
    
    pnormals[nvert*3+0]=n[0];
    pnormals[nvert*3+1]=n[1];
    pnormals[nvert*3+2]=n[2];
    
    Vec3 v=rot*Vec3(0,0,-L*0.5);
    pvertices[nvert*3+0]=v[0];
    pvertices[nvert*3+1]=v[1];
    pvertices[nvert*3+2]=v[2]+z0;
    
    
    nvert++;
    
    
    for(int j=pnphi;j>=0;j--){
      long jj=(j+1)%pnphi;
      
      double c=phis[2*jj];
      double s=phis[2*jj+1];
      pnormals[nvert*3+0]=n[0];
      pnormals[nvert*3+1]=n[1];
      pnormals[nvert*3+2]=n[2];
      v=rot*Vec3(radius*c,radius*s,-L*0.5);
      pvertices[nvert*3+0]=v[0];
      pvertices[nvert*3+1]=v[1];
      pvertices[nvert*3+2]=v[2]+z0;
      nvert++;
    }
    
    pdefs[idef].mode=GL_TRIANGLE_FAN;
    pdefs[idef].count=pnphi+2;
    pdefs[idef].start=nn0;
    pdefs[idef].useindex=false;
    idef++;
    
    nn0+=pnphi+2;
  }
  if(right==Flat){
    //upper lid, pnphi+2 vertices, 1 def
    pnormals[nvert*3+0]=-n[0];
    pnormals[nvert*3+1]=-n[1];
    pnormals[nvert*3+2]=-n[2];
    
    Vec3 v=rot*Vec3(0,0,L*0.5);
    pvertices[nvert*3+0]=v[0];
    pvertices[nvert*3+1]=v[1];
    pvertices[nvert*3+2]=v[2]+z0;
    nvert++;
    
    
    for(int j=0;j<=pnphi;j++){
      long jj=(j+1)%pnphi;
      
      double c=phis[2*jj];
      double s=phis[2*jj+1];
      pnormals[nvert*3+0]=-n[0];
      pnormals[nvert*3+1]=-n[1];
      pnormals[nvert*3+2]=-n[2];
      v=rot*Vec3(radius*c,radius*s,L*0.5);
      pvertices[nvert*3+0]=v[0];
      pvertices[nvert*3+1]=v[1];
      pvertices[nvert*3+2]=v[2]+z0;
      nvert++;
    }
    pdefs[idef].mode=GL_TRIANGLE_FAN;
    pdefs[idef].count=pnphi+2;
    pdefs[idef].start=nn0;
    pdefs[idef].useindex=false;
    
    idef++;
    nn0+=pnphi+2;
    
  }
  if(left==Round){
    //lower cap, ?? vertices, pntheta-1 defs

    for(int itheta=pntheta-1;itheta>1;itheta--){
      double c0=ptheta[2*itheta];
      double s0=ptheta[2*itheta+1];
      double c1=ptheta[2*itheta-2];
      double s1=ptheta[2*itheta-1];
      
      for(int j=pnphi;j>=0;j--){
	long jj=(j+1)%pnphi;
	
	double c=phis[2*jj];
	double s=phis[2*jj+1];
	Vec3 v=rot*Vec3(s1*c,s1*s,-c1);
	pnormals[nvert*3+0]=v[0];
	pnormals[nvert*3+1]=v[1];
	pnormals[nvert*3+2]=v[2];
	v=rot*Vec3(radius*s1*c,radius*s1*s,-L*0.5-radius*c1);
	pvertices[nvert*3+0]=v[0];
	pvertices[nvert*3+1]=v[1];
	pvertices[nvert*3+2]=v[2]+z0;
	nvert++;
	v=rot*Vec3(s0*c,s0*s,-c0);
	pnormals[nvert*3+0]=v[0];
	pnormals[nvert*3+1]=v[1];
	pnormals[nvert*3+2]=v[2];
	v=rot*Vec3(radius*s0*c,radius*s0*s,-L*0.5-radius*c0);
	pvertices[nvert*3+0]=v[0];
	pvertices[nvert*3+1]=v[1];
	pvertices[nvert*3+2]=v[2]+z0;
	nvert++;
      }
      pdefs[idef].mode=GL_QUAD_STRIP;
      pdefs[idef].count=2*pnphi+2;
      pdefs[idef].start=nn0;
      pdefs[idef].useindex=false;
      
      idef++;
      nn0+=2*pnphi+2;
    }

    double c0=ptheta[2];
    double s0=ptheta[3];

    Vec3 v=rot*Vec3(0,0,-1);
    pnormals[nvert*3+0]=v[0];
    pnormals[nvert*3+1]=v[1];
    pnormals[nvert*3+2]=v[2];
    v=rot*Vec3(0,0,-L*0.5-radius);
    pvertices[nvert*3+0]=v[0];
    pvertices[nvert*3+1]=v[1];
    pvertices[nvert*3+2]=v[2]+z0;
    nvert++;


    for(int j=pnphi;j>=0;j--){
      long jj=(j+1)%pnphi;
      
      double c=phis[2*jj];
      double s=phis[2*jj+1];
      
      v=rot*Vec3(s0*c,s0*s,-c0);
      pnormals[nvert*3+0]=v[0];
      pnormals[nvert*3+1]=v[1];
      pnormals[nvert*3+2]=v[2];
      v=rot*Vec3(radius*s0*c,radius*s0*s,-L*0.5-radius*c0);
      pvertices[nvert*3+0]=v[0];
      pvertices[nvert*3+1]=v[1];
      pvertices[nvert*3+2]=v[2]+z0;
      nvert++;
      
    }
    
    pdefs[idef].mode=GL_TRIANGLE_FAN;
    pdefs[idef].count=(pnphi+2);
    pdefs[idef].start=nn0;
    pdefs[idef].useindex=false;
    idef++;
    nn0+=pnphi+2;
    
  }

  if(right==Round){
    //upper cap, ?? vertices, pntheta-1 defs

    for(int itheta=pntheta-1;itheta>1;itheta--){
      double c0=ptheta[2*itheta];
      double s0=ptheta[2*itheta+1];
      double c1=ptheta[2*itheta-2];
      double s1=ptheta[2*itheta-1];
      
      for(int j=0;j<=pnphi;j++){
	long jj=(j+1)%pnphi;
	
	double c=phis[2*jj];
	double s=phis[2*jj+1];
	Vec3 v=rot*Vec3(s1*c,s1*s,c1);
	pnormals[nvert*3+0]=v[0];
	pnormals[nvert*3+1]=v[1];
	pnormals[nvert*3+2]=v[2];
	v=rot*Vec3(radius*s1*c,radius*s1*s,L*0.5+radius*c1);
	pvertices[nvert*3+0]=v[0];
	pvertices[nvert*3+1]=v[1];
	pvertices[nvert*3+2]=v[2]+z0;
	nvert++;
	v=rot*Vec3(s0*c,s0*s,c0);
	pnormals[nvert*3+0]=v[0];
	pnormals[nvert*3+1]=v[1];
	pnormals[nvert*3+2]=v[2];
	v=rot*Vec3(radius*s0*c,radius*s0*s,L*0.5+radius*c0);
	pvertices[nvert*3+0]=v[0];
	pvertices[nvert*3+1]=v[1];
	pvertices[nvert*3+2]=v[2]+z0;
	nvert++;
      }
      pdefs[idef].mode=GL_QUAD_STRIP;
      pdefs[idef].count=2*pnphi+2;
      pdefs[idef].start=nn0;
      pdefs[idef].useindex=false;
      
      idef++;
      nn0+=2*pnphi+2;
    }

    double c0=ptheta[2];
    double s0=ptheta[3];

    Vec3 v=rot*Vec3(0,0,1);
    pnormals[nvert*3+0]=v[0];
    pnormals[nvert*3+1]=v[1];
    pnormals[nvert*3+2]=v[2];
    v=rot*Vec3(0,0,L*0.5+radius);
    pvertices[nvert*3+0]=v[0];
    pvertices[nvert*3+1]=v[1];
    pvertices[nvert*3+2]=v[2]+z0;
    nvert++;


    for(int j=0;j<=pnphi;j++){
      long jj=(j+1)%pnphi;
      
      double c=phis[2*jj];
      double s=phis[2*jj+1];
      
      v=rot*Vec3(s0*c,s0*s,c0);
      pnormals[nvert*3+0]=v[0];
      pnormals[nvert*3+1]=v[1];
      pnormals[nvert*3+2]=v[2];
      v=rot*Vec3(radius*s0*c,radius*s0*s,L*0.5+radius*c0);
      pvertices[nvert*3+0]=v[0];
      pvertices[nvert*3+1]=v[1];
      pvertices[nvert*3+2]=v[2]+z0;
      nvert++;
      
    }
    
    pdefs[idef].mode=GL_TRIANGLE_FAN;
    pdefs[idef].count=(pnphi+2);
    pdefs[idef].start=nn0;
    pdefs[idef].useindex=false;
    idef++;
    nn0+=pnphi+2;
    
  }
  delete [] phis;
  delete [] ptheta;

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void OGLDataCacheItem::paint(bool transparent) const
{
  NSF("OGLDataCacheItem::paint");
  //  myDebug() << nsubfunc << this<<transparent;
  
  if(pimage){
    
    glDrawPixels(pimageNx,pimageNy,GL_RGBA,GL_UNSIGNED_BYTE,pimage);

    ogl.reportGLError(nsubfunc);
    return;
  }

  if(pcull) {
    glEnable(GL_CULL_FACE);
  }else{
    glDisable(GL_CULL_FACE);
  }
  if(transparent){
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  }
  if(pdispListName){
    glCallList(pdispListName);
  }else{
    OGLIssueVertexArray(0,pvertices,pnormals,pcolors,pindices,pndefs,pdefs);
  }
  if(transparent){
    glDepthMask(GL_TRUE);
    glDisable(GL_BLEND);
  }
  glDisable(GL_CULL_FACE);

  ogl.reportGLError(nsubfunc);

  return;
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
OGLDataCache::~OGLDataCache()
{
  NSF("OGLDataCache::~OGLDataCache");
  //myDebug() << nsubfunc;
  cleanCache();
}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
void OGLDataCache::cleanCache(const QString &what)
{
  NSF("OGLDataCache::cleanCache");
  //myDebug() << nsubfunc << what;
  for(int i=plist.size()-1;i>=0;i--){
    OGLDataCacheItem *item=plist.at(i);
    if(item && (QString(item->id()).startsWith(what) || what=="")){
      //myDebug() << "delete " << item->printSpecQ();
      delete item;
      plist.removeAt(i);
    }
  }
  //plist.clear();
}




/* ***********************************************************************
 *
 * ***********************************************************************
 */
OGLDataCacheItem *OGLDataCache::append(OGLDataCacheItem *d,bool transparent,
				       bool forceconstruct)
{
  NSF("OGLDataCache::append");
  //myDebug() << nsubfunc;
  int ind=indexOf(d);
  if(ind<0){
    // not existing yet
    //myDebug() << "append " << d->printSpecQ();
    d->increaseUsed();
    d->construct(transparent);
    plist.append(d);
    //myDebug() << "appended " << d->printSpecQ();
    return d;
  }else{
    OGLDataCacheItem *item=plist.at(ind);
    item->increaseUsed();
    if(!item->isConstructed() || forceconstruct) {
      item->construct(transparent);
    }
    //myDebug() << "not append " << plist.at(ind)->printSpecQ();
    return item;
  }
}

// /* ***********************************************************************
//  *
//  * ***********************************************************************
//  */
void OGLDataCache::markAsOld()
{
  NSF("OGLDataCache::markAsOld");
  //myDebug() << nsubfunc;
  for(int i=0;i<plist.size();i++){
    OGLDataCacheItem *item=plist.at(i);
    if(item){
      item->setUnUsed();
    } 
  } 
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void OGLDataCache::cleanOldCacheData()
{
  NSF("OGLDataCache::cleanOldCacheData");
  //myDebug() << nsubfunc;
  for(int i=plist.size()-1;i>=0;i--){
    OGLDataCacheItem *item=plist.at(i);
    if(item && (item->used()==0)){
      //      myDebug()<< nsubfunc<< "delete" << item->id();
      delete item;
      plist.removeAt(i);
    } 
  } 
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
int OGLDataCache::indexOf(const OGLDataCacheItem *d)const
{
  for(int i=0;i<plist.size();i++){
    if(*plist.at(i)==*d){
      return i;
    }
  }
  return -1;
}

/* ***********************************************************************
 *
 * remove all cache objects which refer to this mesh;
 *
 * ***********************************************************************
 */
void OGLDataCache::invalidateMesh(const RegularMesh *mesh)
{
  NSF("OGLDataCache::invalidateMesh");
  //myDebug() << nsubfunc << mesh;
  
  for(int i=plist.size()-1;i>=0;i--){
    OGLDataCacheItem *item=plist.at(i);
    //myDebug() << item->objectType();
    if(item->isA("OGLDataCacheMeshPointsItem")){
      //
      OGLDataCacheMeshPointsItem *mi
	=dynamic_cast<OGLDataCacheMeshPointsItem*>(item);
      if(mi && mi->pmesh==mesh){
	//myDebug()<< nsubfunc<< "delete" << item->id();
	delete item;
	plist.removeAt(i);
	continue;
      }
      
    }
    if(item->isA("OGLDataCacheMeshCellsItem")){
      //
      OGLDataCacheMeshCellsItem *mi
	=dynamic_cast<OGLDataCacheMeshCellsItem*>(item);
      if(mi && mi->pmesh==mesh){
	//myDebug()<< nsubfunc <<"delete" << item->id();
	delete item;
	plist.removeAt(i);
      }
      
    }
  }
}

/* ***********************************************************************
 *
 * remove all cache objects which refer to this mesh;
 *
 * ***********************************************************************
 */
void OGLDataCache::invalidateMesh(const TriangleMesh *mesh)
{
  NSF("OGLDataCache::invalidateMesh");
  //myDebug() << nsubfunc << mesh;
  
  for(int i=plist.size()-1;i>=0;i--){
    OGLDataCacheItem *item=plist.at(i);
    //myDebug() << item->objectType();
    if(item->isA("OGLDataCacheTriangleMeshItem")){
      //
      OGLDataCacheTriangleMeshItem *mi
	=dynamic_cast<OGLDataCacheTriangleMeshItem*>(item);
      if(mi && mi->pmesh==mesh){
	//myDebug()<<nsubfunc << "delete"<< item->id();
	delete item;
	plist.removeAt(i);
	continue;
      }
      
    }

  }
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void OGLDataCache::invalidateLegend(const ColorModel *cmod)
{
  NSF("OGLDataCache::invalidateLegend");
  //myDebug() << nsubfunc << mesh;
  
  for(int i=plist.size()-1;i>=0;i--){
    OGLDataCacheItem *item=plist.at(i);
    //myDebug() << item->objectType();
    if(item->isA("OGLDataCacheLegendItem")){
      //
      OGLDataCacheLegendItem *li
	=dynamic_cast<OGLDataCacheLegendItem*>(item);
      if(li && li->pcolormodel==cmod){
	//myDebug()<< nsubfunc<< "delete" << item->id();
	delete item;
	plist.removeAt(i);
	continue;
      }
    }
  }
}


/*EOF*/ 
