/*************************************************************************
 * @(#)Master.cpp
 * author: Klaus.Koepernik@ifw-dresden.de
 * date:   03 Jul 2009
 *************************************************************************
 */
#include <QApplication>
#include <QMessageBox>
#include <QFileInfo>
#include <QTimer>
#include "Parser.hh" 
#include "Coeffs.hh" 
#include "BWFromCoeffs.hh" 
#include "Except.hh" 
#include "Assorted.hh" 
#include "Master.hh"
#include "Config.hh"
#include "Activity.hh"
#include "MWBase.hh"
#include "MyDebug.hh"

using namespace XFPLO;

/* ***********************************************************************
 *
 * ***********************************************************************
 */
Master::Master(QApplication *a)
  :timer(0),showoutwin(false)
{
  
  MASSERT("Master called with ",a);
  arg=a->arguments();

  //todo  BrainDialog::setMainWidget(base);


  Activity::ActivityType into=Activity::None;
  int i=arg.size()-1;
  while(i>=0){
    //myDebug()<<i<<arg[i];
    if(arg[i]=="-fs"){
      into=Activity::FermiSurface;
      arg.removeAt(i);
    }else if(arg[i]=="-str"){
      into=Activity::Structure;
      arg.removeAt(i);
    }else if(arg[i]=="-bw"){
      into=Activity::BandWeightCreate;
      arg.removeAt(i);
    }else if(arg[i]=="-o"){
      showoutwin=true;
      arg.removeAt(i);
    }else if(arg[i].startsWith("-")){
      arg.removeAt(i);
    }
    i--;
  }

  QString file="";
  {
    QString ft=fileType(getArgument(1),true);
    if(
       ft=="xstr"||
       ft=="in"||
       ft=="cif"||
       ft=="sym" ){
      into=Activity::Structure;
      file=getArgument(1);
    }else if(ft=="xef"){
      into=Activity::FermiSurface;
      file=getArgument(1);
    }else if(ft=="coeff"){
      into=Activity::BandWeightCreate;
      file=getArgument(1);
    }else if(ft=="bwdef"){
      into=Activity::BandWeightCreate;
      file=getArgument(1);
    }
    //myDebug() <<"ft" << ft<< "into"<< into;
    
  }
  
  QString path=".";
  if(!file.isEmpty()){
    QFileInfo fi(file);
    path=fi.absolutePath();
    file=fi.fileName();
  }
//   QMessageBox::information(0,"Check", "path="+path
// 			   ,QMessageBox::Ok);


  MWBase *base=new MWBase(this,Activity(into,path,file),true);
  base->show();

  // for(i=0;i<arg.size();i++){
  //     myDebug()<<i<<arg[i];
  //}
  timer= new QTimer(this);
  connect(timer,SIGNAL(timeout()),this,SLOT(checkActs()));
  timer->start(1*1000);


}



/* ***********************************************************************
 *
 * ***********************************************************************
 */
Master::~Master()
{
  if(timer) delete timer;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Master::checkActs()
{
  //  myDebug() << "timer timeout";
  for(int i=0;i<activities.size();i++){
    //     myDebug() << "base="<<activities[i]
    // 	     <<" closed="<< activities[i]->isVisible()
    // 	     <<" hiddden="<< activities[i]->isHidden();
    if(activities[i]->isHidden()){
      //activities[i]->close();
      activities[i]->deleteLater();
      activities.removeAt(i);
      break;
    }
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Master::registerActivity(MWBase *base)
{
  if(!activities.contains(base)){
    //myDebug()<<"add activity "<<base;
    activities.append(base);
  }
}
/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Master::unRegisterActivity(MWBase *base)
{
  int ind;
  if((ind=activities.indexOf(base))>-1){
    //myDebug()<<"unregister activity "<<ind << " " << base;
    //base->close();
    base->deleteLater();
    activities.removeAt(ind);
  }
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
void Master::requireQuit()
{
  int i=activities.size()-1;
  while(i>=0){
    MWBase *base=activities[i];
    if(base->requireQuit()){
      unRegisterActivity(base);
    }
    i--;
  }

}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
long  Master::numOfActivitiesOfType(Activity act)
{
  int count=0;
  int i=activities.size()-1;
  while(i>=0){
    MWBase *base=activities[i];
    if(base->isA(act.type()))
      count++;
    i--;
  }

  return count;
}

/* ***********************************************************************
 *
 * ***********************************************************************
 */
QString Master::getArgument(int i)
{
  if(i<arg.size()){
    return arg[i];
  }else
    return QString();
}


/* ***********************************************************************
 *
 * ***********************************************************************
 */
QString Master::fileType(QString file,bool complain)
{
  if(file.isEmpty()) return QString();
  QFileInfo fi(file);
  if(fi.isDir()) return "directory";
  if(!fi.isFile()) return QString();

  if(fi.suffix()=="cif") return "cif";
  if(fi.suffix()=="dic") return "cif";

  try{
    if(Coeffs::isCoeffFile(String(file)))
      return QString("coeff");
  }catch(Except ex){}

  try{
    if(BWFromCoeffs::isBWDefFile(file))
      return QString("bwdef");
  }catch(Except ex){}

  String fname(file);

  if(pathExistence(fname()) == DoesNotExist) return QString();


  Parser par(1);
  try{
    par.parseFile(fname());
  }catch(Except &e){
    if(complain)
      QMessageBox::warning(0,"Open"
			   ,Sprintf("cannot read file \"%s\"\n(%s)"
				    ,fname(),e.msg())()
			   ,QMessageBox::Ok);
    
    return QString();
  }

  String s;

  // is it a newer file
  try{
     s=par()("filetype.application")();
     return QString(s());
   }catch(Except e){};

  // is it a fermi surface file
  try{
    par()("empty_lattice");
    return QString("xef");
  }catch(Except e){};


  // is it an =.in file
  try{
    par()("initial_spin_split");
    return QString("in");
  }catch(Except e){};

  // is it an =.sym file
  try{
    par()("axis_angles");
    return QString("sym");
  }catch(Except e){};
  

  par.cleanup();

  return QString();
}


//EOF
