/* @(#)SpaceGroupOperation.hh
 * author: Klaus.Koepernik@ifw-dresden.de
 */

#ifndef _SPACEGROUPOPERATION_H
#define _SPACEGROUPOPERATION_H 1

#include "FString.hh"
#include "PtrArray.hh"
#include "ValueArray.hh"
#include "Vec3.hh"
#include "IVec3.hh"
#include "MyDebug.hh"
#include "MyQRgb.hh"

class IVec3;
class IMatrix3x3;

/* ***********************************************************************
 *
 * ***********************************************************************
 */
class SpaceGroupOperationBase
{
 public:
  enum SpaceGroupOperationType{Identity=0,Inversion=1,Mirror=2,
			       Rotation=3,RotoInversion=4};

  SpaceGroupOperationBase():ptype(Identity),
    pissymmorphic(false),pvisible(false),psymbol("E"){};

  virtual ~SpaceGroupOperationBase(){}

  virtual SpaceGroupOperationBase *clone()const=0;

  // generate a spacegroup operation from {alpa|tau} and the lattice
  static SpaceGroupOperationBase *generateOperation(const Matrix3x3 &opmat,
						    const Vec3 &optau,
						    const Matrix3x3 &cell,
						    bool prot);

  SpaceGroupOperationType type()const {return ptype;};
  inline String symbol()const{return psymbol;};
  inline bool isSymmorphic()const {return pissymmorphic;};

  inline bool isVisible()const {return pvisible;};
  inline void setVisible(bool v){pvisible=v;};

  virtual void expand(const Matrix3x3 &){};//todo: pure

  virtual String printS()const=0;
  
 protected:
  struct OperationLatticeHelper{
    IMatrix3x3 iopmat; /* operation in terms of lattice vectors cell */
    IVec3 RNMIN,GNMIN; /* cell*RNMIN and cell*GNMIN are the shortest vectors
			  in axis direction, RNMIN*GNMIN=PMIN_GC*/
    IMatrix3x3 RN,GN;  /* cell*RN[0,1] spans the perpendicular lattice plane
			  RN[2] points to the first lattice plane in axis 
			  direction,but need not be parallel, GN*RN=1 */ 
    long PMIN_GC;      /* RNMIN*GNMIN=|cell*RNMIN|*|rcell*GNMIN|=PMIN_GC,
			  if PMIN_GC=1, we have a simple situation, 
			  where cell*RNMIN \parrallel axis*/
    double tauG0,tauG1;
    double tauGmin;    /* tau=(tau*(rcell*GNMIN))*(cell*PNMIN)
			  tau=(tau*\hat{rcell*GNMIN})*\hat{cell*PNMIN}/PMIN_GC
			  tau=tauGmin*\hat{cell*PNMIN}
			  tauGmin=(tau*\hat{rcell*GNMIN})/PMIN_GC
			  is the parallel tau in units of the shortest 
			  lattice vector PMin.
		       */
    long NP0,NP1;
  };

  void prepareSetUpOrigin(const Vec3 &axis,long order,const Matrix3x3 &pcell,
			  OperationLatticeHelper &olh);

  static String printAxis(const Vec3 &axis,bool isC2);
  static void subLatticeForAxis(const Vec3 &axis,long n,
				const Matrix3x3 &A,
				const Matrix3x3 &G,
				OperationLatticeHelper &olh,
				bool prot
				);
  static void RAlongAxis(const Vec3 &axis,long n,
			 const Matrix3x3 &A
			 ,const Matrix3x3 &G,
			 IVec3 &N,double tol);
  static void latticePlane(const IVec3 &MG,IMatrix3x3 &RN,IMatrix3x3 &GN,
			   bool prot);
  static void cellReduce2D(const Matrix3x3 &A,IMatrix3x3 &RN,double tol,
			   bool prot);
  static void minl12(const Matrix3x3 &A,IMatrix3x3 &RN,long i1,long i2,
		     double &smin,bool prot);
  static void minimizeVectorIn2D(const Matrix3x3 &A,
				 const IMatrix3x3 &RN,
				 long &it0,long &it1,long n,bool prot);
 
 protected:
  SpaceGroupOperationType ptype;
  bool pissymmorphic;
  bool pvisible;
  Matrix3x3 popmat;
  Vec3 poptau;
  String psymbol;
};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
class SpaceGroupIdentity:public SpaceGroupOperationBase
{
 public:
  SpaceGroupIdentity():SpaceGroupOperationBase(){};
//   SpaceGroupIdentity &operator=(const SpaceGroupIdentity&other){
//     *this=other;
//     return *this;
//   };

  SpaceGroupIdentity *clone()const{
    //myDebug()<<"SpaceGroupIdentity::clone";
    return new SpaceGroupIdentity();};

  String printS()const{return " sym   : E";}
};

/* ***********************************************************************
 *
 * ***********************************************************************
 */
class SpaceGroupInversion:public SpaceGroupOperationBase
{
 public:
  SpaceGroupInversion(const Matrix3x3 &opmat,
		      const Vec3 &optau,const Matrix3x3 &pcell)
    :SpaceGroupOperationBase(){ 
    ptype=Inversion;popmat=opmat;poptau=optau;psymbol="I";
    setUpOrigin(pcell);
  }
  SpaceGroupInversion *clone()const{
    //myDebug()<<"SpaceGroupInversion::clone";
    return new SpaceGroupInversion(*this);
  };

  String printS()const;

 private:
  void setUpOrigin(const Matrix3x3 &pcell);
 private:
  Vec3 porigin;//origin in absolute units
  Vec3 pshiftperiods[3];//periods in absolute units

};
/* ***********************************************************************
 *
 * ***********************************************************************
 */
class SpaceGroupMirror:public SpaceGroupOperationBase
{
 public:
  SpaceGroupMirror(const Matrix3x3 &opmat,
		   const Vec3 &optau,const Vec3 &axis,const Matrix3x3 &pcell)
    :SpaceGroupOperationBase(),paxis(axis){
    ptype=Mirror;
    popmat=opmat;poptau=optau;
    psymbol.sprintf("s(%s)",printAxis(paxis,true)());
    setUpOrigin(pcell);  
  }
  SpaceGroupMirror *clone()const{
    //myDebug()<<"SpaceGroupMirror::clone";
    return new SpaceGroupMirror(*this);
  };

  String printS()const;

 private:
 void setUpOrigin(const Matrix3x3 &pcell);
 private:
  Vec3 paxis;
  // at origin of coordinatesystem Sigma axis points perp to plane
  // mirror sits at origin and has non-symmorphic tau
  ValueArray<Vec3> porigins;//origin in absolute units
  ValueArray<Vec3> ptaus;
  Vec3 pshiftperiod;//period along axis in absolute units
};
/* ***********************************************************************
 *
 * ***********************************************************************
 */
class SpaceGroupRotation:public SpaceGroupOperationBase
{
 public:
  SpaceGroupRotation(const Matrix3x3 &opmat,
		     const Vec3 &optau,const Vec3 &axis,long n,long m,
		     const Matrix3x3 &pcell)
    :SpaceGroupOperationBase(),paxis(axis),pn(n),pm(m){
    ptype=Rotation;
    popmat=opmat;poptau=optau;
    if(pn==2){
      psymbol.sprintf("C2(%s)",printAxis(paxis,true)());
    }else{
      if(pm==1){
	psymbol.sprintf("C%ld(%s)",pn,printAxis(paxis,false)());
      }else{
	psymbol.sprintf("C%ld/%ld(%s)",pm,pn,printAxis(paxis,false)());
      }
    }
    setUpOrigin(pcell);  
  }
  SpaceGroupRotation *clone()const{
    return new SpaceGroupRotation(*this);
  };
  inline long exponent()const{return pm;};

  virtual void expand(const Matrix3x3 &cell);

  String printS()const;

private:
  void setUpOrigin(const Matrix3x3 &);
 private:
  Vec3 paxis;
  long pn,pm;
  // sits at origin from global porigins[i], with a tau ptaus[i]/pn in
  // units of the shortest lattice vector PMin sitting on the axis
  ValueArray<Vec3> porigins;
  ValueArray<long> pmtaus;
  IVec3 pdelta[2]; // in units of the 12*SpaceGroupOperations.pcell
  OperationLatticeHelper polh;
};
/* ***********************************************************************
 *
 * ***********************************************************************
 */
class SpaceGroupRotoInversion:public SpaceGroupOperationBase
{
 public:
  SpaceGroupRotoInversion(const Matrix3x3 &opmat,
			  const Vec3 &optau,const Vec3 &axis,long n,long m,
			  const Matrix3x3 &pcell)
    :SpaceGroupOperationBase(),paxis(axis),pn(n),pm(m){
    ptype=RotoInversion;
    popmat=opmat;poptau=optau;
    if(pm==1){
      psymbol.sprintf("-%ld(%s)",pn,printAxis(paxis,false)());
    }else{
      psymbol.sprintf("-%ld^%ld(%s)",pn,pm,printAxis(paxis,false)());
    }
    setUpOrigin(pcell);
  }
  SpaceGroupRotoInversion *clone()const{
    //myDebug()<<"SpaceGroupRotoInversion::clone";
    return new SpaceGroupRotoInversion(*this);
  };

  inline long exponent()const{return pm;};
  String printS()const;
 private:
  void setUpOrigin(const Matrix3x3 &);
 private:
  Vec3 paxis;
  long pn,pm;
  ValueArray<Vec3> porigins;
  Vec3 pshiftperiods[2];
  // the shortest lattice vector along the axis
  Vec3 ppmin;
};


/* ***********************************************************************
 * now the collection of all ops
 * ***********************************************************************
 */
class SpaceGroupOperations
{
 public:
  SpaceGroupOperations():
    psymircolor(0x22ccccff),pnonsymircolor(0x22ff9999),
    pmirrorsvisible(false),protationsvisible(false),
    pmirrorrotationsvisible(false),pinversionvisible(false),
    pallvisible(false),
    pshowbottom(true),
    pshowtop(true),
    pinversionsymbolsize(2.){};
  ~SpaceGroupOperations(){};


  inline void clear(){pspacegroupoperations.clear();};

  inline void setCell(const Matrix3x3 &cell){pcell=cell;};

  inline void append(const Matrix3x3 &opalpha,const Vec3 &optau,bool prot){
    //append an operation given by {alpha|tau}
    pspacegroupoperations.append(SpaceGroupOperationBase
				 ::generateOperation(opalpha,optau,pcell,prot));
  };

  void unfold();//make each operation with a different origin an individual
  // operation

  inline long size()const{return pspacegroupoperations.size();};

  inline const SpaceGroupOperationBase *at(long i)const{
    return pspacegroupoperations.at(i);};

  inline  SpaceGroupOperationBase *operator[](long i){
    return pspacegroupoperations[i];};

  inline const SpaceGroupOperationBase *last()const{
    return pspacegroupoperations.last();};

  inline SpaceGroupOperationBase *last(){
    return pspacegroupoperations.last();};


  bool hasType(SpaceGroupOperationBase::SpaceGroupOperationType t)const;


  inline  bool mirrorsAreVisible()const{return pmirrorsvisible;};
  inline  void setMirrorsAreVisible(bool a){pmirrorsvisible=a;};

  inline  bool inversionIsVisible()const{return pinversionvisible;};
  inline  void setInversionIsVisible(bool a){pinversionvisible=a;};

  inline  bool rotationsAreVisible()const{return protationsvisible;};
  inline  void setRotationsAreVisible(bool a){protationsvisible=a;};

  inline  bool mirrorRotationsAreVisible()const{
    return pmirrorrotationsvisible;};
  inline  void setMirrorRotationsAreVisible(bool a){
    pmirrorrotationsvisible=a;};

  inline  bool allAreVisible()const{return pallvisible;};
  inline  void setAllAreVisible(bool a){pallvisible=a;};

  inline  bool showBottom()const{return pshowbottom;};
  inline  void setShowBottom(bool a){pshowbottom=a;};
  inline  bool showTop()const{return pshowtop;};
  inline  void setShowTop(bool a){pshowtop=a;};


  inline double inversionSymbolSize()const{
    return pinversionsymbolsize;};
  inline void setInversionSymbolSize(double value){
    pinversionsymbolsize=value;
  };

  inline QRgb symmorphicColor()const{return psymircolor;};
  inline QRgb nonsymmorphicColor()const{return pnonsymircolor;};


 private:
  Matrix3x3 pcell;
  PtrArray<SpaceGroupOperationBase> pspacegroupoperations;
  QRgb psymircolor;
  QRgb pnonsymircolor;
  bool pmirrorsvisible;
  bool protationsvisible;
  bool pmirrorrotationsvisible;
  bool pinversionvisible;
  bool pallvisible;
  bool pshowbottom;
  bool pshowtop;
  double pinversionsymbolsize;
};




#ifdef	__cplusplus
extern "C" {
#endif

#ifdef	__cplusplus
}
#endif


#endif /* _SPACEGROUPOPERATION_H */

