Module vecharm
  !!
  Use precisiondefs
  Use errors
  Use constants
  Use matserv
  Use diracdefs
  !!
  Implicit None 
  !!
  Private
  !!
  Character(*),Parameter::nmodule="VECHARM"  
  !!
  Public :: calc_real_vec_Ylm,calc_real_vec_Ylm_and_Ylm&
       &,calc_real_vec_Ylm_opt,calc_real_quad_harmonics_opt&
       &,calc_real_vec_Ylm_coefs
  !!
  !!***********************************************************************
Contains

  !!***********************************************************************
  !!
  !!
  !! BEWARE that the functions defined here are for numerical
  !! efficiency and are not normalized vector/tensor harmonics
  !!
  !! create the rvylm and qvylm functions from ylm
  !! ylm is the real harmonics
  !! rvylm is some real vector harmonics
  !! qvylm is some real tensor harmonics
  !! 
  !! 
  !! the gradient of f*Ylm(ilm) is
  !!      (df+(l+1)/r*f)*rvylm(:,1,ilm)/deg+(df-l/r*f)*rvylm(:,2,ilm)/deg
  !!
  !!   r*grad Ylm= (l+1)*rvylm(:,1,ilm)-l*rvylm(:,2,ilm)
  !!
  !!   where  rvylm is defined via (deg=2*l+1)
  !!
  !!   rvylm(:,1,ilm)= deg*R^{l-1}_{lm} sqrt(l/(2l+1))
  !!   rvylm(:,2,ilm)=-deg*R^{l+1}_{lm} sqrt((l+1)/(2l+1))
  !!
  !!   with the R^{l+/-1}_{lm} being the normalized real vector harmonics
  !!
  !! we have the following identity 
  !!
  !!
  !! rvylm(:,1,ilm)=deg*vec{e}_r*Ylm-rvylm(:,2,ilm)
  !!
  !!    vec{e}_r  is the radial unit vector vec{r}/|r|.
  !!
  !! can be transformed into (absorbing a factor 2*l+1)
  !!
  !!      grad[f*Ylm(ilm)]=
  !!      f/r*rvylm(:,1,ilm)+(df-l/r*f)*vec{e}_r*Ylm
  !!
  !!
  !! which is cheaper since we only need R^(l-1)_l which on top is
  !! cheaper than R^(l+1)_l
  !! so rvylm returns R^(l-1)_l 
  !!
  !! furthermore the second term can be summed up over lm as a scalar
  !! and this removes one vector*scalar operation in the lm loop
  !!
  !! for the symmetric tensor harmonics we return the Q^1 only and of this
  !! the lower left triangle
  !!
  !! grad \circ grad (f Y_lm(ilm)) =
  !!
  !!     D(-l)f/r*ylm(ilm)*unit_matrix
  !!    +D(++,l)f*ylm(ilm) *vec{e}_r \circ vec{e}_r
  !!    +D(-l)f/r *( rvylm(:,ilm) \circ vec{e}_r + tranposed) 
  !!    + f/r**2 * qvylm(:,:,ilm)
  !!                        (beware only the LL triangle is returned)
  !!   
  !!     D(-l)=d/dr-l/r 
  !!     D(++,l)=d2/dr^2-deg/r *d/dr+ l(l+2)/r**2
  !!   
  !!   \circ is the dyadic product
  !!   
  !!   the LL triangle of qvylm(:,:,ilm) is saved in linear order
  !!   n=0
  !!   do j=1,3
  !!     do i=j,3
  !!       n=n+1
  !!       Q(i,j)=qvylm(n)
  !!   
  !!  the term qvylm(n==6) is not stored since Q is traceless
  !!   
  !!***********************************************************************
  Subroutine calc_real_quad_harmonics_opt(maxL,ylm,rvylm,qvylm)
    !! created by dovecharm.pl
    !!
    Implicit None
    !!IO
    Integer(t_int),Intent(in)    :: maxL
    Real(t_real),Intent(in)      :: ylm((maxL+1)**2)
    Real(t_real),Intent(out)     :: rvylm(3,((maxL+1)**2))
    Real(t_real),Intent(out)     :: qvylm(5,((maxL+1)**2))
    !!LOCAL
    Character(*),Parameter::nsubfunc="calc_real_quad_harmonics_opt"
    !!
    If(maxL>12_t_int)Then
3      Format('''maxl > 12 not implemented! ''')
       Write(errbuf,3) 
       Call errmsg(nmodule,nsubfunc,E_fatalerror,msg=errbuf)
    Endif
    !!
    !!
    !! XXXXXXXXXXXXXXXXXXX
    !!..............................................
    !! L=0
    !!..............................................
    !! L-1
    rvylm(1,1)= 0.0_t_real
    rvylm(2,1)= 0.0_t_real
    rvylm(3,1)= 0.0_t_real
    !!
    If(maxL<=0) Goto 1000
    !!
    !!..............................................
    !! L=1
    !!..............................................
    !! L-1
    rvylm(1,2)= 0.0_t_real
    rvylm(2,2)= 0.48860251190291992157_t_real
    rvylm(3,2)= 0.0_t_real
    rvylm(1,3)= 0.0_t_real
    rvylm(2,3)= 0.0_t_real
    rvylm(3,3)= 0.48860251190291992158_t_real
    rvylm(1,4)= 0.48860251190291992157_t_real
    rvylm(2,4)= 0.0_t_real
    rvylm(3,4)= 0.0_t_real
    !!
    If(maxL<=1) Goto 1000
    !!
    !!..............................................
    !! L=2
    !!..............................................
    !! L-1
    rvylm(1,5)=2.23606797749978969630_t_real*ylm(2)
    rvylm(2,5)=2.23606797749978969630_t_real*ylm(4)
    rvylm(3,5)= 0.0_t_real
    rvylm(1,6)= 0.0_t_real
    rvylm(2,6)=2.23606797749978969630_t_real*ylm(3)
    rvylm(3,6)=2.23606797749978969630_t_real*ylm(2)
    rvylm(1,7)=-1.29099444873580562829_t_real*ylm(4)
    rvylm(2,7)=-1.29099444873580562829_t_real*ylm(2)
    rvylm(3,7)=2.58198889747161125670_t_real*ylm(3)
    rvylm(1,8)=2.23606797749978969630_t_real*ylm(3)
    rvylm(2,8)= 0.0_t_real
    rvylm(3,8)=2.23606797749978969630_t_real*ylm(4)
    rvylm(1,9)=2.23606797749978969630_t_real*ylm(4)
    rvylm(2,9)=-2.23606797749978969630_t_real*ylm(2)
    rvylm(3,9)= 0.0_t_real
    !!
    If(maxL<=2) Goto 1000
    !!
    !!..............................................
    !! L=3
    !!..............................................
    !! L-1
    rvylm(1,10)=3.24037034920393011516_t_real*ylm(5)
    rvylm(2,10)=3.24037034920393011516_t_real*ylm(9)
    rvylm(3,10)= 0.0_t_real
    rvylm(1,11)=2.64575131106459059024_t_real*ylm(6)
    rvylm(2,11)=2.64575131106459059024_t_real*ylm(8)
    rvylm(3,11)=2.64575131106459059024_t_real*ylm(5)
    rvylm(1,12)=-0.83666002653407554786_t_real*ylm(5)
    rvylm(2,12)=2.89827534923788771438_t_real*ylm(7)&
         &      +0.83666002653407554786_t_real*ylm(9)
    rvylm(3,12)=3.34664010613630219159_t_real*ylm(6)
    rvylm(1,13)=-2.04939015319191967638_t_real*ylm(8)
    rvylm(2,13)=-2.04939015319191967638_t_real*ylm(6)
    rvylm(3,13)=3.54964786985976962523_t_real*ylm(7)
    rvylm(1,14)=2.89827534923788771438_t_real*ylm(7)&
         &      -0.83666002653407554786_t_real*ylm(9)
    rvylm(2,14)=-0.83666002653407554786_t_real*ylm(5)
    rvylm(3,14)=3.34664010613630219159_t_real*ylm(8)
    rvylm(1,15)=2.64575131106459059024_t_real*ylm(8)
    rvylm(2,15)=-2.64575131106459059024_t_real*ylm(6)
    rvylm(3,15)=2.64575131106459059024_t_real*ylm(9)
    rvylm(1,16)=3.24037034920393011516_t_real*ylm(9)
    rvylm(2,16)=-3.24037034920393011516_t_real*ylm(5)
    rvylm(3,16)= 0.0_t_real
    !!
    If(maxL<=3) Goto 1000
    !!
    !!..............................................
    !! L=4
    !!..............................................
    !! L-1
    rvylm(1,17)=4.24264068711928514617_t_real*ylm(10)
    rvylm(2,17)=4.24264068711928514617_t_real*ylm(16)
    rvylm(3,17)= 0.0_t_real
    rvylm(1,18)=3.67423461417476714708_t_real*ylm(11)
    rvylm(2,18)=3.67423461417476714708_t_real*ylm(15)
    rvylm(3,18)=2.99999999999999999985_t_real*ylm(10)
    rvylm(1,19)=-0.80178372573727315397_t_real*ylm(10)&
         &      +3.10529501704059397988_t_real*ylm(12)
    rvylm(2,19)=3.10529501704059397988_t_real*ylm(14)&
         &      +0.80178372573727315397_t_real*ylm(16)
    rvylm(3,19)=3.92792202424786286258_t_real*ylm(11)
    rvylm(1,20)=-1.38873014965882719226_t_real*ylm(11)
    rvylm(2,20)=3.58568582800318091960_t_real*ylm(13)&
         &      +1.38873014965882719226_t_real*ylm(15)
    rvylm(3,20)=4.39155032826839930683_t_real*ylm(12)
    rvylm(1,21)=-2.77746029931765438447_t_real*ylm(14)
    rvylm(2,21)=-2.77746029931765438447_t_real*ylm(12)
    rvylm(3,21)=4.53557367611072672635_t_real*ylm(13)
    rvylm(1,22)=3.58568582800318091960_t_real*ylm(13)&
         &      -1.38873014965882719226_t_real*ylm(15)
    rvylm(2,22)=-1.38873014965882719226_t_real*ylm(11)
    rvylm(3,22)=4.39155032826839930683_t_real*ylm(14)
    rvylm(1,23)=3.10529501704059397988_t_real*ylm(14)&
         &      -0.80178372573727315397_t_real*ylm(16)
    rvylm(2,23)=-0.80178372573727315397_t_real*ylm(10)&
         &      -3.10529501704059397988_t_real*ylm(12)
    rvylm(3,23)=3.92792202424786286258_t_real*ylm(15)
    rvylm(1,24)=3.67423461417476714708_t_real*ylm(15)
    rvylm(2,24)=-3.67423461417476714708_t_real*ylm(11)
    rvylm(3,24)=2.99999999999999999985_t_real*ylm(16)
    rvylm(1,25)=4.24264068711928514617_t_real*ylm(16)
    rvylm(2,25)=-4.24264068711928514617_t_real*ylm(10)
    rvylm(3,25)= 0.0_t_real
    !!
    If(maxL<=4) Goto 1000
    !!
    !!..............................................
    !! L=5
    !!..............................................
    !! L-1
    rvylm(1,26)=5.24404424085075773482_t_real*ylm(17)
    rvylm(2,26)=5.24404424085075773482_t_real*ylm(25)
    rvylm(3,26)= 0.0_t_real
    rvylm(1,27)=4.69041575982342955444_t_real*ylm(18)
    rvylm(2,27)=4.69041575982342955444_t_real*ylm(24)
    rvylm(3,27)=3.31662479035539984906_t_real*ylm(17)
    rvylm(1,28)=-0.78173595997057159239_t_real*ylm(17)&
         &      +4.13655788199695219863_t_real*ylm(19)
    rvylm(2,28)=4.13655788199695219863_t_real*ylm(23)&
         &      +0.78173595997057159239_t_real*ylm(25)
    rvylm(3,28)=4.42216638714053313209_t_real*ylm(18)
    rvylm(1,29)=-1.35400640077266006004_t_real*ylm(18)&
         &      +3.58236421003411283978_t_real*ylm(20)
    rvylm(2,29)=3.58236421003411283978_t_real*ylm(22)&
         &      +1.35400640077266006004_t_real*ylm(24)
    rvylm(3,29)=5.06622805119022121044_t_real*ylm(19)
    rvylm(1,30)=-1.91485421551267621986_t_real*ylm(19)
    rvylm(2,30)=4.28174419288837633985_t_real*ylm(21)&
         &      +1.91485421551267621986_t_real*ylm(23)
    rvylm(3,30)=5.41602560309064024018_t_real*ylm(20)
    rvylm(1,31)=-3.49602949390050515650_t_real*ylm(22)
    rvylm(2,31)=-3.49602949390050515650_t_real*ylm(20)
    rvylm(3,31)=5.52770798392566641511_t_real*ylm(21)
    rvylm(1,32)=4.28174419288837633985_t_real*ylm(21)&
         &      -1.91485421551267621986_t_real*ylm(23)
    rvylm(2,32)=-1.91485421551267621986_t_real*ylm(19)
    rvylm(3,32)=5.41602560309064024018_t_real*ylm(22)
    rvylm(1,33)=3.58236421003411283978_t_real*ylm(22)&
         &      -1.35400640077266006004_t_real*ylm(24)
    rvylm(2,33)=-1.35400640077266006004_t_real*ylm(18)&
         &      -3.58236421003411283978_t_real*ylm(20)
    rvylm(3,33)=5.06622805119022121044_t_real*ylm(23)
    rvylm(1,34)=4.13655788199695219863_t_real*ylm(23)&
         &      -0.78173595997057159239_t_real*ylm(25)
    rvylm(2,34)=-0.78173595997057159239_t_real*ylm(17)&
         &      -4.13655788199695219863_t_real*ylm(19)
    rvylm(3,34)=4.42216638714053313209_t_real*ylm(24)
    rvylm(1,35)=4.69041575982342955444_t_real*ylm(24)
    rvylm(2,35)=-4.69041575982342955444_t_real*ylm(18)
    rvylm(3,35)=3.31662479035539984906_t_real*ylm(25)
    rvylm(1,36)=5.24404424085075773482_t_real*ylm(25)
    rvylm(2,36)=-5.24404424085075773482_t_real*ylm(17)
    rvylm(3,36)= 0.0_t_real
    !!
    If(maxL<=5) Goto 1000
    !!
    !!..............................................
    !! L=6
    !!..............................................
    !! L-1
    rvylm(1,37)=6.24499799839839820498_t_real*ylm(26)
    rvylm(2,37)=6.24499799839839820498_t_real*ylm(36)
    rvylm(3,37)= 0.0_t_real
    rvylm(1,38)=5.70087712549568989494_t_real*ylm(27)
    rvylm(2,38)=5.70087712549568989494_t_real*ylm(35)
    rvylm(3,38)=3.60555127546398929258_t_real*ylm(26)
    rvylm(1,39)=-0.76870611478580740429_t_real*ylm(26)&
         &      +5.15663738214246464010_t_real*ylm(28)
    rvylm(2,39)=5.15663738214246464010_t_real*ylm(34)&
         &      +0.76870611478580740429_t_real*ylm(36)
    rvylm(3,39)=4.86172434804397732390_t_real*ylm(27)
    rvylm(1,40)=-1.33143804689789180092_t_real*ylm(27)&
         &      +4.61223668871484442618_t_real*ylm(29)
    rvylm(2,40)=4.61223668871484442618_t_real*ylm(33)&
         &      +1.33143804689789180092_t_real*ylm(35)
    rvylm(3,40)=5.64881323014763067168_t_real*ylm(28)
    rvylm(1,41)=-1.88293774338254355719_t_real*ylm(28)&
         &      +4.06761042203583521229_t_real*ylm(30)
    rvylm(2,41)=4.06761042203583521229_t_real*ylm(32)&
         &      +1.88293774338254355719_t_real*ylm(34)
    rvylm(3,41)=6.14964891828645923485_t_real*ylm(29)
    rvylm(1,42)=-2.43086217402198866191_t_real*ylm(29)
    rvylm(2,42)=4.98178500320736625169_t_real*ylm(31)&
         &      +2.43086217402198866191_t_real*ylm(33)
    rvylm(3,42)=6.43145678393599746847_t_real*ylm(30)
    rvylm(1,43)=-4.21037679160342219218_t_real*ylm(32)
    rvylm(2,43)=-4.21037679160342219218_t_real*ylm(30)
    rvylm(3,43)=6.52268767805530801070_t_real*ylm(31)
    rvylm(1,44)=4.98178500320736625169_t_real*ylm(31)&
         &      -2.43086217402198866191_t_real*ylm(33)
    rvylm(2,44)=-2.43086217402198866191_t_real*ylm(29)
    rvylm(3,44)=6.43145678393599746847_t_real*ylm(32)
    rvylm(1,45)=4.06761042203583521229_t_real*ylm(32)&
         &      -1.88293774338254355719_t_real*ylm(34)
    rvylm(2,45)=-1.88293774338254355719_t_real*ylm(28)&
         &      -4.06761042203583521229_t_real*ylm(30)
    rvylm(3,45)=6.14964891828645923485_t_real*ylm(33)
    rvylm(1,46)=4.61223668871484442618_t_real*ylm(33)&
         &      -1.33143804689789180092_t_real*ylm(35)
    rvylm(2,46)=-1.33143804689789180092_t_real*ylm(27)&
         &      -4.61223668871484442618_t_real*ylm(29)
    rvylm(3,46)=5.64881323014763067168_t_real*ylm(34)
    rvylm(1,47)=5.15663738214246464010_t_real*ylm(34)&
         &      -0.76870611478580740429_t_real*ylm(36)
    rvylm(2,47)=-0.76870611478580740429_t_real*ylm(26)&
         &      -5.15663738214246464010_t_real*ylm(28)
    rvylm(3,47)=4.86172434804397732390_t_real*ylm(35)
    rvylm(1,48)=5.70087712549568989494_t_real*ylm(35)
    rvylm(2,48)=-5.70087712549568989494_t_real*ylm(27)
    rvylm(3,48)=3.60555127546398929258_t_real*ylm(36)
    rvylm(1,49)=6.24499799839839820498_t_real*ylm(36)
    rvylm(2,49)=-6.24499799839839820498_t_real*ylm(26)
    rvylm(3,49)= 0.0_t_real
    !!
    If(maxL<=6) Goto 1000
    !!
    !!..............................................
    !! L=7
    !!..............................................
    !! L-1
    rvylm(1,50)=7.24568837309471928553_t_real*ylm(37)
    rvylm(2,50)=7.24568837309471928553_t_real*ylm(49)
    rvylm(3,50)= 0.0_t_real
    rvylm(1,51)=6.70820393249936908803_t_real*ylm(38)
    rvylm(2,51)=6.70820393249936908803_t_real*ylm(48)
    rvylm(3,51)=3.87298334620741688447_t_real*ylm(37)
    rvylm(1,52)=-0.75955452531274996207_t_real*ylm(37)&
         &      +6.17065013405581737244_t_real*ylm(39)
    rvylm(2,52)=6.17065013405581737244_t_real*ylm(47)&
         &      +0.75955452531274996207_t_real*ylm(49)
    rvylm(3,52)=5.26234811584217528733_t_real*ylm(38)
    rvylm(1,53)=-1.31558702896054382183_t_real*ylm(38)&
         &      +5.63300712149108069607_t_real*ylm(40)
    rvylm(2,53)=5.63300712149108069607_t_real*ylm(46)&
         &      +1.31558702896054382183_t_real*ylm(48)
    rvylm(3,53)=6.17065013405581737244_t_real*ylm(39)
    rvylm(1,54)=-1.86052101883812684248_t_real*ylm(39)&
         &      +5.09524665365068087684_t_real*ylm(41)
    rvylm(2,54)=5.09524665365068087684_t_real*ylm(45)&
         &      +1.86052101883812684248_t_real*ylm(47)
    rvylm(3,54)=6.79366220486757450242_t_real*ylm(40)
    rvylm(1,55)=-2.40192230707630700172_t_real*ylm(40)&
         &      +4.55732715187649977242_t_real*ylm(42)
    rvylm(2,55)=4.55732715187649977242_t_real*ylm(44)&
         &      +2.40192230707630700172_t_real*ylm(46)
    rvylm(3,55)=7.20576692122892100538_t_real*ylm(41)
    rvylm(1,56)=-2.94174202707276047827_t_real*ylm(41)
    rvylm(2,56)=5.68398560058805107456_t_real*ylm(43)&
         &      +2.94174202707276047827_t_real*ylm(45)
    rvylm(3,56)=7.44208407535250737003_t_real*ylm(42)
    rvylm(1,57)=-4.92247592485420194143_t_real*ylm(44)
    rvylm(2,57)=-4.92247592485420194143_t_real*ylm(42)
    rvylm(3,57)=7.51920617741404524399_t_real*ylm(43)
    rvylm(1,58)=5.68398560058805107456_t_real*ylm(43)&
         &      -2.94174202707276047827_t_real*ylm(45)
    rvylm(2,58)=-2.94174202707276047827_t_real*ylm(41)
    rvylm(3,58)=7.44208407535250737003_t_real*ylm(44)
    rvylm(1,59)=4.55732715187649977242_t_real*ylm(44)&
         &      -2.40192230707630700172_t_real*ylm(46)
    rvylm(2,59)=-2.40192230707630700172_t_real*ylm(40)&
         &      -4.55732715187649977242_t_real*ylm(42)
    rvylm(3,59)=7.20576692122892100538_t_real*ylm(45)
    rvylm(1,60)=5.09524665365068087684_t_real*ylm(45)&
         &      -1.86052101883812684248_t_real*ylm(47)
    rvylm(2,60)=-1.86052101883812684248_t_real*ylm(39)&
         &      -5.09524665365068087684_t_real*ylm(41)
    rvylm(3,60)=6.79366220486757450242_t_real*ylm(46)
    rvylm(1,61)=5.63300712149108069607_t_real*ylm(46)&
         &      -1.31558702896054382183_t_real*ylm(48)
    rvylm(2,61)=-1.31558702896054382183_t_real*ylm(38)&
         &      -5.63300712149108069607_t_real*ylm(40)
    rvylm(3,61)=6.17065013405581737244_t_real*ylm(47)
    rvylm(1,62)=6.17065013405581737244_t_real*ylm(47)&
         &      -0.75955452531274996207_t_real*ylm(49)
    rvylm(2,62)=-0.75955452531274996207_t_real*ylm(37)&
         &      -6.17065013405581737244_t_real*ylm(39)
    rvylm(3,62)=5.26234811584217528733_t_real*ylm(48)
    rvylm(1,63)=6.70820393249936908803_t_real*ylm(48)
    rvylm(2,63)=-6.70820393249936908803_t_real*ylm(38)
    rvylm(3,63)=3.87298334620741688447_t_real*ylm(49)
    rvylm(1,64)=7.24568837309471928553_t_real*ylm(49)
    rvylm(2,64)=-7.24568837309471928553_t_real*ylm(37)
    rvylm(3,64)= 0.0_t_real
    !!
    If(maxL<=7) Goto 1000
    !!
    !!..............................................
    !! L=8
    !!..............................................
    !! L-1
    rvylm(1,65)=8.24621125123532109794_t_real*ylm(50)
    rvylm(2,65)=8.24621125123532109794_t_real*ylm(64)
    rvylm(3,65)= 0.0_t_real
    rvylm(1,66)=7.71362431027075624291_t_real*ylm(51)
    rvylm(2,66)=7.71362431027075624291_t_real*ylm(63)
    rvylm(3,66)=4.12310562561766054891_t_real*ylm(50)
    rvylm(1,67)=-0.75277265270908099474_t_real*ylm(50)&
         &      +7.18099343173816492009_t_real*ylm(52)
    rvylm(2,67)=7.18099343173816492009_t_real*ylm(62)&
         &      +0.75277265270908099474_t_real*ylm(64)
    rvylm(3,67)=5.63323471314069544695_t_real*ylm(51)
    rvylm(1,68)=-1.30384048104052974265_t_real*ylm(51)&
         &      +6.64830805543786455952_t_real*ylm(53)
    rvylm(2,68)=6.64830805543786455952_t_real*ylm(61)&
         &      +1.30384048104052974265_t_real*ylm(63)
    rvylm(3,68)=6.64830805543786455952_t_real*ylm(52)
    rvylm(1,69)=-1.84390889145857746161_t_real*ylm(52)&
         &      +6.11555394056826220916_t_real*ylm(54)
    rvylm(2,69)=6.11555394056826220916_t_real*ylm(60)&
         &      +1.84390889145857746161_t_real*ylm(62)
    rvylm(3,69)=7.37563556583430984644_t_real*ylm(53)
    rvylm(1,70)=-2.38047614284761666544_t_real*ylm(53)&
         &      +5.58271140814807537673_t_real*ylm(55)
    rvylm(2,70)=5.58271140814807537673_t_real*ylm(59)&
         &      +2.38047614284761666544_t_real*ylm(61)
    rvylm(3,70)=7.89514618821800748719_t_real*ylm(54)
    rvylm(1,71)=-2.91547594742265023479_t_real*ylm(54)&
         &      +5.04975246918103897561_t_real*ylm(56)
    rvylm(2,71)=5.04975246918103897561_t_real*ylm(58)&
         &      +2.91547594742265023479_t_real*ylm(60)
    rvylm(3,71)=8.24621125123532109794_t_real*ylm(55)
    rvylm(1,72)=-3.44963766213206803729_t_real*ylm(55)
    rvylm(2,72)=6.38748776906852479812_t_real*ylm(57)&
         &      +3.44963766213206803729_t_real*ylm(59)
    rvylm(3,72)=8.44985206971104317037_t_real*ylm(56)
    rvylm(1,73)=-5.63323471314069544683_t_real*ylm(58)
    rvylm(2,73)=-5.63323471314069544683_t_real*ylm(56)
    rvylm(3,73)=8.51665035875803306448_t_real*ylm(57)
    rvylm(1,74)=6.38748776906852479812_t_real*ylm(57)&
         &      -3.44963766213206803729_t_real*ylm(59)
    rvylm(2,74)=-3.44963766213206803729_t_real*ylm(55)
    rvylm(3,74)=8.44985206971104317037_t_real*ylm(58)
    rvylm(1,75)=5.04975246918103897561_t_real*ylm(58)&
         &      -2.91547594742265023479_t_real*ylm(60)
    rvylm(2,75)=-2.91547594742265023479_t_real*ylm(54)&
         &      -5.04975246918103897561_t_real*ylm(56)
    rvylm(3,75)=8.24621125123532109794_t_real*ylm(59)
    rvylm(1,76)=5.58271140814807537673_t_real*ylm(59)&
         &      -2.38047614284761666544_t_real*ylm(61)
    rvylm(2,76)=-2.38047614284761666544_t_real*ylm(53)&
         &      -5.58271140814807537673_t_real*ylm(55)
    rvylm(3,76)=7.89514618821800748719_t_real*ylm(60)
    rvylm(1,77)=6.11555394056826220916_t_real*ylm(60)&
         &      -1.84390889145857746161_t_real*ylm(62)
    rvylm(2,77)=-1.84390889145857746161_t_real*ylm(52)&
         &      -6.11555394056826220916_t_real*ylm(54)
    rvylm(3,77)=7.37563556583430984644_t_real*ylm(61)
    rvylm(1,78)=6.64830805543786455952_t_real*ylm(61)&
         &      -1.30384048104052974265_t_real*ylm(63)
    rvylm(2,78)=-1.30384048104052974265_t_real*ylm(51)&
         &      -6.64830805543786455952_t_real*ylm(53)
    rvylm(3,78)=6.64830805543786455952_t_real*ylm(62)
    rvylm(1,79)=7.18099343173816492009_t_real*ylm(62)&
         &      -0.75277265270908099474_t_real*ylm(64)
    rvylm(2,79)=-0.75277265270908099474_t_real*ylm(50)&
         &      -7.18099343173816492009_t_real*ylm(52)
    rvylm(3,79)=5.63323471314069544695_t_real*ylm(63)
    rvylm(1,80)=7.71362431027075624291_t_real*ylm(63)
    rvylm(2,80)=-7.71362431027075624291_t_real*ylm(51)
    rvylm(3,80)=4.12310562561766054891_t_real*ylm(64)
    rvylm(1,81)=8.24621125123532109794_t_real*ylm(64)
    rvylm(2,81)=-8.24621125123532109794_t_real*ylm(50)
    rvylm(3,81)= 0.0_t_real
    !!
    If(maxL<=8) Goto 1000
    !!
    !!..............................................
    !! L=9
    !!..............................................
    !! L-1
    rvylm(1,82)=9.24662100445346467442_t_real*ylm(65)
    rvylm(2,82)=9.24662100445346467442_t_real*ylm(81)
    rvylm(3,82)= 0.0_t_real
    rvylm(1,83)=8.71779788708134710358_t_real*ylm(66)
    rvylm(2,83)=8.71779788708134710358_t_real*ylm(80)
    rvylm(3,83)=4.35889894354067355172_t_real*ylm(65)
    rvylm(1,84)=-0.74754500159640202466_t_real*ylm(65)&
         &      +8.18894520249169604204_t_real*ylm(67)
    rvylm(2,84)=8.18894520249169604204_t_real*ylm(79)&
         &      +0.74754500159640202466_t_real*ylm(81)
    rvylm(3,84)=5.98036001277121619814_t_real*ylm(66)
    rvylm(1,85)=-1.29478592370912579251_t_real*ylm(66)&
         &      +7.66005682669752092386_t_real*ylm(68)
    rvylm(2,85)=7.66005682669752092386_t_real*ylm(78)&
         &      +1.29478592370912579251_t_real*ylm(80)
    rvylm(3,85)=7.09183457555651276193_t_real*ylm(67)
    rvylm(1,86)=-1.83110381367922125644_t_real*ylm(67)&
         &      +7.13112481846101107951_t_real*ylm(69)
    rvylm(2,86)=7.13112481846101107951_t_real*ylm(77)&
         &      +1.83110381367922125644_t_real*ylm(79)
    rvylm(3,86)=7.91127267221384849204_t_real*ylm(68)
    rvylm(1,87)=-2.36394485851883758726_t_real*ylm(68)&
         &      +6.60213869091809120630_t_real*ylm(70)
    rvylm(2,87)=6.60213869091809120630_t_real*ylm(76)&
         &      +2.36394485851883758726_t_real*ylm(78)
    rvylm(3,87)=8.52332439975913457832_t_real*ylm(69)
    rvylm(1,88)=-2.89522934172346191113_t_real*ylm(69)&
         &      +6.07308430216282034611_t_real*ylm(71)
    rvylm(2,88)=6.07308430216282034611_t_real*ylm(75)&
         &      +2.89522934172346191113_t_real*ylm(77)
    rvylm(3,88)=8.97054001915682429734_t_real*ylm(70)
    rvylm(1,89)=-3.42568155520139653580_t_real*ylm(70)&
         &      +5.54394210987516163984_t_real*ylm(72)
    rvylm(2,89)=5.54394210987516163984_t_real*ylm(74)&
         &      +3.42568155520139653580_t_real*ylm(76)
    rvylm(3,89)=9.27678950550306302935_t_real*ylm(71)
    rvylm(1,90)=-3.95563633610692424602_t_real*ylm(71)
    rvylm(2,90)=7.09183457555651276193_t_real*ylm(73)&
         &      +3.95563633610692424602_t_real*ylm(75)
    rvylm(3,90)=9.45577943407535034933_t_real*ylm(72)
    rvylm(1,91)=-6.34312967845109239606_t_real*ylm(74)
    rvylm(2,91)=-6.34312967845109239606_t_real*ylm(72)
    rvylm(3,91)=9.51469451767663859429_t_real*ylm(73)
    rvylm(1,92)=7.09183457555651276193_t_real*ylm(73)&
         &      -3.95563633610692424602_t_real*ylm(75)
    rvylm(2,92)=-3.95563633610692424602_t_real*ylm(71)
    rvylm(3,92)=9.45577943407535034933_t_real*ylm(74)
    rvylm(1,93)=5.54394210987516163984_t_real*ylm(74)&
         &      -3.42568155520139653580_t_real*ylm(76)
    rvylm(2,93)=-3.42568155520139653580_t_real*ylm(70)&
         &      -5.54394210987516163984_t_real*ylm(72)
    rvylm(3,93)=9.27678950550306302935_t_real*ylm(75)
    rvylm(1,94)=6.07308430216282034611_t_real*ylm(75)&
         &      -2.89522934172346191113_t_real*ylm(77)
    rvylm(2,94)=-2.89522934172346191113_t_real*ylm(69)&
         &      -6.07308430216282034611_t_real*ylm(71)
    rvylm(3,94)=8.97054001915682429734_t_real*ylm(76)
    rvylm(1,95)=6.60213869091809120630_t_real*ylm(76)&
         &      -2.36394485851883758726_t_real*ylm(78)
    rvylm(2,95)=-2.36394485851883758726_t_real*ylm(68)&
         &      -6.60213869091809120630_t_real*ylm(70)
    rvylm(3,95)=8.52332439975913457832_t_real*ylm(77)
    rvylm(1,96)=7.13112481846101107951_t_real*ylm(77)&
         &      -1.83110381367922125644_t_real*ylm(79)
    rvylm(2,96)=-1.83110381367922125644_t_real*ylm(67)&
         &      -7.13112481846101107951_t_real*ylm(69)
    rvylm(3,96)=7.91127267221384849204_t_real*ylm(78)
    rvylm(1,97)=7.66005682669752092386_t_real*ylm(78)&
         &      -1.29478592370912579251_t_real*ylm(80)
    rvylm(2,97)=-1.29478592370912579251_t_real*ylm(66)&
         &      -7.66005682669752092386_t_real*ylm(68)
    rvylm(3,97)=7.09183457555651276193_t_real*ylm(79)
    rvylm(1,98)=8.18894520249169604204_t_real*ylm(79)&
         &      -0.74754500159640202466_t_real*ylm(81)
    rvylm(2,98)=-0.74754500159640202466_t_real*ylm(65)&
         &      -8.18894520249169604204_t_real*ylm(67)
    rvylm(3,98)=5.98036001277121619814_t_real*ylm(80)
    rvylm(1,99)=8.71779788708134710358_t_real*ylm(80)
    rvylm(2,99)=-8.71779788708134710358_t_real*ylm(66)
    rvylm(3,99)=4.35889894354067355172_t_real*ylm(81)
    rvylm(1,100)=9.24662100445346467442_t_real*ylm(81)
    rvylm(2,100)=-9.24662100445346467442_t_real*ylm(65)
    rvylm(3,100)= 0.0_t_real
    !!
    If(maxL<=9) Goto 1000
    !!
    !!..............................................
    !! L=10
    !!..............................................
    !! L-1
    rvylm(1,101)=10.24695076595959838021_t_real*ylm(82)
    rvylm(2,101)=10.24695076595959838021_t_real*ylm(100)
    rvylm(3,101)= 0.0_t_real
    rvylm(1,102)=9.72111104761179034356_t_real*ylm(83)
    rvylm(2,102)=9.72111104761179034356_t_real*ylm(99)
    rvylm(3,102)=4.58257569495584000524_t_real*ylm(82)
    rvylm(1,103)=-0.74339194167502812698_t_real*ylm(82)&
         &      +9.19525049027743280505_t_real*ylm(84)
    rvylm(2,103)=9.19525049027743280505_t_real*ylm(98)&
         &      +0.74339194167502812698_t_real*ylm(100)
    rvylm(3,103)=6.30788979645416176110_t_real*ylm(83)
    rvylm(1,104)=-1.28759261291842819510_t_real*ylm(83)&
         &      +8.66936530184546469713_t_real*ylm(85)
    rvylm(2,104)=8.66936530184546469713_t_real*ylm(97)&
         &      +1.28759261291842819510_t_real*ylm(99)
    rvylm(3,104)=7.50789058608552039380_t_real*ylm(84)
    rvylm(1,105)=-1.82093093600065198186_t_real*ylm(84)&
         &      +8.14345071045955373554_t_real*ylm(86)
    rvylm(2,105)=8.14345071045955373554_t_real*ylm(96)&
         &      +1.82093093600065198186_t_real*ylm(98)
    rvylm(3,105)=8.41051972860554901480_t_real*ylm(85)
    rvylm(1,106)=-2.35081172990813626463_t_real*ylm(85)&
         &      +7.61750062615512255368_t_real*ylm(87)
    rvylm(2,106)=7.61750062615512255368_t_real*ylm(95)&
         &      +2.35081172990813626463_t_real*ylm(97)
    rvylm(3,106)=9.10465468000325991007_t_real*ylm(86)
    rvylm(1,107)=-2.87914460981217932131_t_real*ylm(86)&
         &      +7.09150715181268968924_t_real*ylm(88)
    rvylm(2,107)=7.09150715181268968924_t_real*ylm(94)&
         &      +2.87914460981217932131_t_real*ylm(96)
    rvylm(3,107)=9.63546082256359416530_t_real*ylm(87)
    rvylm(1,108)=-3.40664984374601330048_t_real*ylm(87)&
         &      +6.56545985882898661229_t_real*ylm(89)
    rvylm(2,108)=6.56545985882898661229_t_real*ylm(93)&
         &      +3.40664984374601330048_t_real*ylm(95)
    rvylm(3,108)=10.02890559175930484906_t_real*ylm(88)
    rvylm(1,109)=-3.93366040864311465542_t_real*ylm(88)&
         &      +6.03934468386482439955_t_real*ylm(90)
    rvylm(2,109)=6.03934468386482439955_t_real*ylm(92)&
         &      +3.93366040864311465542_t_real*ylm(94)
    rvylm(3,109)=10.30074090334742556083_t_real*ylm(89)
    rvylm(1,110)=-4.46035165005016876262_t_real*ylm(89)
    rvylm(2,110)=7.79676046087158729237_t_real*ylm(91)&
         &      +4.46035165005016876262_t_real*ylm(93)
    rvylm(3,110)=10.46045183687487503868_t_real*ylm(90)
    rvylm(1,111)=-7.05243518972440879420_t_real*ylm(92)
    rvylm(2,111)=-7.05243518972440879420_t_real*ylm(90)
    rvylm(3,111)=10.51314966075693626850_t_real*ylm(91)
    rvylm(1,112)=7.79676046087158729237_t_real*ylm(91)&
         &      -4.46035165005016876262_t_real*ylm(93)
    rvylm(2,112)=-4.46035165005016876262_t_real*ylm(89)
    rvylm(3,112)=10.46045183687487503868_t_real*ylm(92)
    rvylm(1,113)=6.03934468386482439955_t_real*ylm(92)&
         &      -3.93366040864311465542_t_real*ylm(94)
    rvylm(2,113)=-3.93366040864311465542_t_real*ylm(88)&
         &      -6.03934468386482439955_t_real*ylm(90)
    rvylm(3,113)=10.30074090334742556083_t_real*ylm(93)
    rvylm(1,114)=6.56545985882898661229_t_real*ylm(93)&
         &      -3.40664984374601330048_t_real*ylm(95)
    rvylm(2,114)=-3.40664984374601330048_t_real*ylm(87)&
         &      -6.56545985882898661229_t_real*ylm(89)
    rvylm(3,114)=10.02890559175930484906_t_real*ylm(94)
    rvylm(1,115)=7.09150715181268968924_t_real*ylm(94)&
         &      -2.87914460981217932131_t_real*ylm(96)
    rvylm(2,115)=-2.87914460981217932131_t_real*ylm(86)&
         &      -7.09150715181268968924_t_real*ylm(88)
    rvylm(3,115)=9.63546082256359416530_t_real*ylm(95)
    rvylm(1,116)=7.61750062615512255368_t_real*ylm(95)&
         &      -2.35081172990813626463_t_real*ylm(97)
    rvylm(2,116)=-2.35081172990813626463_t_real*ylm(85)&
         &      -7.61750062615512255368_t_real*ylm(87)
    rvylm(3,116)=9.10465468000325991007_t_real*ylm(96)
    rvylm(1,117)=8.14345071045955373554_t_real*ylm(96)&
         &      -1.82093093600065198186_t_real*ylm(98)
    rvylm(2,117)=-1.82093093600065198186_t_real*ylm(84)&
         &      -8.14345071045955373554_t_real*ylm(86)
    rvylm(3,117)=8.41051972860554901480_t_real*ylm(97)
    rvylm(1,118)=8.66936530184546469713_t_real*ylm(97)&
         &      -1.28759261291842819510_t_real*ylm(99)
    rvylm(2,118)=-1.28759261291842819510_t_real*ylm(83)&
         &      -8.66936530184546469713_t_real*ylm(85)
    rvylm(3,118)=7.50789058608552039380_t_real*ylm(98)
    rvylm(1,119)=9.19525049027743280505_t_real*ylm(98)&
         &      -0.74339194167502812698_t_real*ylm(100)
    rvylm(2,119)=-0.74339194167502812698_t_real*ylm(82)&
         &      -9.19525049027743280505_t_real*ylm(84)
    rvylm(3,119)=6.30788979645416176110_t_real*ylm(99)
    rvylm(1,120)=9.72111104761179034356_t_real*ylm(99)
    rvylm(2,120)=-9.72111104761179034356_t_real*ylm(83)
    rvylm(3,120)=4.58257569495584000524_t_real*ylm(100)
    rvylm(1,121)=10.24695076595959838021_t_real*ylm(100)
    rvylm(2,121)=-10.24695076595959838021_t_real*ylm(82)
    rvylm(3,121)= 0.0_t_real
    !!
    If(maxL<=10) Goto 1000
    !!
    !!..............................................
    !! L=11
    !!..............................................
    !! L-1
    rvylm(1,122)=11.24722187920199251574_t_real*ylm(101)
    rvylm(2,122)=11.24722187920199251574_t_real*ylm(121)
    rvylm(3,122)= 0.0_t_real
    rvylm(1,123)=10.72380529476360830159_t_real*ylm(102)
    rvylm(2,123)=10.72380529476360830159_t_real*ylm(120)
    rvylm(3,123)=4.79583152331271954017_t_real*ylm(101)
    rvylm(1,124)=-0.74001286990095491087_t_real*ylm(101)&
         &      +10.20037347588896117257_t_real*ylm(103)
    rvylm(2,124)=10.20037347588896117257_t_real*ylm(119)&
         &      +0.74001286990095491087_t_real*ylm(121)
    rvylm(3,124)=6.61887632529297298225_t_real*ylm(102)
    rvylm(1,125)=-1.28173988892331148207_t_real*ylm(102)&
         &      +9.67692395045332278867_t_real*ylm(104)
    rvylm(2,125)=9.67692395045332278867_t_real*ylm(118)&
         &      +1.28173988892331148207_t_real*ylm(120)
    rvylm(3,125)=7.90117531944276177215_t_real*ylm(103)
    rvylm(1,126)=-1.81265393434993148228_t_real*ylm(103)&
         &      +9.15345368075429142629_t_real*ylm(105)
    rvylm(2,126)=9.15345368075429142629_t_real*ylm(117)&
         &      +1.81265393434993148228_t_real*ylm(119)
    rvylm(3,126)=8.88015443881145893082_t_real*ylm(104)
    rvylm(1,127)=-2.34012616672487900012_t_real*ylm(104)&
         &      +8.62995889191776371504_t_real*ylm(106)
    rvylm(2,127)=8.62995889191776371504_t_real*ylm(116)&
         &      +2.34012616672487900012_t_real*ylm(118)
    rvylm(3,127)=9.64858736267844004767_t_real*ylm(105)
    rvylm(1,128)=-2.86605752110555420381_t_real*ylm(105)&
         &      +8.10643483377777571872_t_real*ylm(107)
    rvylm(2,128)=8.10643483377777571872_t_real*ylm(115)&
         &      +2.86605752110555420381_t_real*ylm(117)
    rvylm(3,128)=10.25391911138649185674_t_real*ylm(106)
    rvylm(1,129)=-3.39116499156263406841_t_real*ylm(106)&
         &      +7.58287544405155055191_t_real*ylm(108)
    rvylm(2,129)=7.58287544405155055191_t_real*ylm(114)&
         &      +3.39116499156263406841_t_real*ylm(116)
    rvylm(3,129)=10.72380529476360830159_t_real*ylm(107)
    rvylm(1,130)=-3.91578004149024352770_t_real*ylm(107)&
         &      +7.05927286151579023264_t_real*ylm(109)
    rvylm(2,130)=7.05927286151579023264_t_real*ylm(113)&
         &      +3.91578004149024352770_t_real*ylm(115)
    rvylm(3,130)=11.07549848389076646685_t_real*ylm(108)
    rvylm(1,131)=-4.44007721940572946541_t_real*ylm(108)&
         &      +6.53561670497021376997_t_real*ylm(110)
    rvylm(2,131)=6.53561670497021376997_t_real*ylm(112)&
         &      +4.44007721940572946541_t_real*ylm(114)
    rvylm(3,131)=11.32002019180430361392_t_real*ylm(109)
    rvylm(1,132)=-4.96415724396972973668_t_real*ylm(109)
    rvylm(2,132)=8.50210058078086286687_t_real*ylm(111)&
         &      +4.96415724396972973668_t_real*ylm(113)
    rvylm(3,132)=11.46423008442221681558_t_real*ylm(110)
    rvylm(1,133)=-7.76132045711908728798_t_real*ylm(112)
    rvylm(2,133)=-7.76132045711908728798_t_real*ylm(110)
    rvylm(3,133)=11.51189860639023886521_t_real*ylm(111)
    rvylm(1,134)=8.50210058078086286687_t_real*ylm(111)&
         &      -4.96415724396972973668_t_real*ylm(113)
    rvylm(2,134)=-4.96415724396972973668_t_real*ylm(109)
    rvylm(3,134)=11.46423008442221681558_t_real*ylm(112)
    rvylm(1,135)=6.53561670497021376997_t_real*ylm(112)&
         &      -4.44007721940572946541_t_real*ylm(114)
    rvylm(2,135)=-4.44007721940572946541_t_real*ylm(108)&
         &      -6.53561670497021376997_t_real*ylm(110)
    rvylm(3,135)=11.32002019180430361392_t_real*ylm(113)
    rvylm(1,136)=7.05927286151579023264_t_real*ylm(113)&
         &      -3.91578004149024352770_t_real*ylm(115)
    rvylm(2,136)=-3.91578004149024352770_t_real*ylm(107)&
         &      -7.05927286151579023264_t_real*ylm(109)
    rvylm(3,136)=11.07549848389076646685_t_real*ylm(114)
    rvylm(1,137)=7.58287544405155055191_t_real*ylm(114)&
         &      -3.39116499156263406841_t_real*ylm(116)
    rvylm(2,137)=-3.39116499156263406841_t_real*ylm(106)&
         &      -7.58287544405155055191_t_real*ylm(108)
    rvylm(3,137)=10.72380529476360830159_t_real*ylm(115)
    rvylm(1,138)=8.10643483377777571872_t_real*ylm(115)&
         &      -2.86605752110555420381_t_real*ylm(117)
    rvylm(2,138)=-2.86605752110555420381_t_real*ylm(105)&
         &      -8.10643483377777571872_t_real*ylm(107)
    rvylm(3,138)=10.25391911138649185674_t_real*ylm(116)
    rvylm(1,139)=8.62995889191776371504_t_real*ylm(116)&
         &      -2.34012616672487900012_t_real*ylm(118)
    rvylm(2,139)=-2.34012616672487900012_t_real*ylm(104)&
         &      -8.62995889191776371504_t_real*ylm(106)
    rvylm(3,139)=9.64858736267844004767_t_real*ylm(117)
    rvylm(1,140)=9.15345368075429142629_t_real*ylm(117)&
         &      -1.81265393434993148228_t_real*ylm(119)
    rvylm(2,140)=-1.81265393434993148228_t_real*ylm(103)&
         &      -9.15345368075429142629_t_real*ylm(105)
    rvylm(3,140)=8.88015443881145893082_t_real*ylm(118)
    rvylm(1,141)=9.67692395045332278867_t_real*ylm(118)&
         &      -1.28173988892331148207_t_real*ylm(120)
    rvylm(2,141)=-1.28173988892331148207_t_real*ylm(102)&
         &      -9.67692395045332278867_t_real*ylm(104)
    rvylm(3,141)=7.90117531944276177215_t_real*ylm(119)
    rvylm(1,142)=10.20037347588896117257_t_real*ylm(119)&
         &      -0.74001286990095491087_t_real*ylm(121)
    rvylm(2,142)=-0.74001286990095491087_t_real*ylm(101)&
         &      -10.20037347588896117257_t_real*ylm(103)
    rvylm(3,142)=6.61887632529297298225_t_real*ylm(120)
    rvylm(1,143)=10.72380529476360830159_t_real*ylm(120)
    rvylm(2,143)=-10.72380529476360830159_t_real*ylm(102)
    rvylm(3,143)=4.79583152331271954017_t_real*ylm(121)
    rvylm(1,144)=11.24722187920199251574_t_real*ylm(121)
    rvylm(2,144)=-11.24722187920199251574_t_real*ylm(101)
    rvylm(3,144)= 0.0_t_real
    !!
    If(maxL<=11) Goto 1000
    !!
    !!..............................................
    !! L=12
    !!..............................................
    !! L-1
    rvylm(1,145)=12.24744871391589049060_t_real*ylm(122)
    rvylm(2,145)=12.24744871391589049060_t_real*ylm(144)
    rvylm(3,145)= 0.0_t_real
    rvylm(1,146)=11.72603939955857388608_t_real*ylm(123)
    rvylm(2,146)=11.72603939955857388608_t_real*ylm(143)
    rvylm(3,146)=4.99999999999999999974_t_real*ylm(122)
    rvylm(1,147)=-0.73720978077448566723_t_real*ylm(122)&
         &      +11.20461861291447818860_t_real*ylm(124)
    rvylm(2,147)=11.20461861291447818860_t_real*ylm(142)&
         &      +0.73720978077448566723_t_real*ylm(144)
    rvylm(3,147)=6.91564074808124624060_t_real*ylm(123)
    rvylm(1,148)=-1.27688479613812288973_t_real*ylm(123)&
         &      +10.68318467417879543822_t_real*ylm(125)
    rvylm(2,148)=10.68318467417879543822_t_real*ylm(141)&
         &      +1.27688479613812288973_t_real*ylm(143)
    rvylm(3,148)=8.27515926551055676937_t_real*ylm(124)
    rvylm(1,149)=-1.80578779628653801127_t_real*ylm(124)&
         &      +10.16173555871325512508_t_real*ylm(126)
    rvylm(2,149)=10.16173555871325512508_t_real*ylm(140)&
         &      +1.80578779628653801127_t_real*ylm(142)
    rvylm(3,149)=9.32504808240313765608_t_real*ylm(125)
    rvylm(1,150)=-2.33126202060078441389_t_real*ylm(125)&
         &      +9.64026880375727138200_t_real*ylm(127)
    rvylm(2,150)=9.64026880375727138200_t_real*ylm(139)&
         &      +2.33126202060078441389_t_real*ylm(141)
    rvylm(3,150)=10.16173555871325512508_t_real*ylm(126)
    rvylm(1,151)=-2.85520120360080372741_t_real*ylm(126)&
         &      +9.11878138311493232910_t_real*ylm(128)
    rvylm(2,151)=9.11878138311493232910_t_real*ylm(138)&
         &      +2.85520120360080372741_t_real*ylm(140)
    rvylm(3,151)=10.83472677771922806832_t_real*ylm(127)
    rvylm(1,152)=-3.37831962346088111590_t_real*ylm(127)&
         &      +8.59726953621095152654_t_real*ylm(129)
    rvylm(2,152)=8.59726953621095152654_t_real*ylm(137)&
         &      +3.37831962346088111590_t_real*ylm(139)
    rvylm(3,152)=11.37311857350289484499_t_real*ylm(128)
    rvylm(1,153)=-3.90094748802746972907_t_real*ylm(128)&
         &      +8.07572853087248259575_t_real*ylm(130)
    rvylm(2,153)=8.07572853087248259575_t_real*ylm(136)&
         &      +3.90094748802746972907_t_real*ylm(138)
    rvylm(3,153)=11.79535649239177067632_t_real*ylm(129)
    rvylm(1,154)=-4.42325868464691400360_t_real*ylm(129)&
         &      +7.55415232778002351432_t_real*ylm(131)
    rvylm(2,154)=7.55415232778002351432_t_real*ylm(135)&
         &      +4.42325868464691400360_t_real*ylm(137)
    rvylm(3,154)=12.11359279630872389380_t_real*ylm(130)
    rvylm(1,155)=-4.94535355046840253523_t_real*ylm(130)&
         &      +7.03253309548774429223_t_real*ylm(132)
    rvylm(2,155)=7.03253309548774429223_t_real*ylm(134)&
         &      +4.94535355046840253523_t_real*ylm(136)
    rvylm(3,155)=12.33587909487922392313_t_real*ylm(131)
    rvylm(1,156)=-5.46729406085186595524_t_real*ylm(131)
    rvylm(2,156)=9.20774721067276987093_t_real*ylm(133)&
         &      +5.46729406085186595524_t_real*ylm(135)
    rvylm(3,156)=12.46734865994753722431_t_real*ylm(132)
    rvylm(1,157)=-8.46989553859919860879_t_real*ylm(134)
    rvylm(2,157)=-8.46989553859919860879_t_real*ylm(132)
    rvylm(3,157)=12.51086484342448576045_t_real*ylm(133)
    rvylm(1,158)=9.20774721067276987093_t_real*ylm(133)&
         &      -5.46729406085186595524_t_real*ylm(135)
    rvylm(2,158)=-5.46729406085186595524_t_real*ylm(131)
    rvylm(3,158)=12.46734865994753722431_t_real*ylm(134)
    rvylm(1,159)=7.03253309548774429223_t_real*ylm(134)&
         &      -4.94535355046840253523_t_real*ylm(136)
    rvylm(2,159)=-4.94535355046840253523_t_real*ylm(130)&
         &      -7.03253309548774429223_t_real*ylm(132)
    rvylm(3,159)=12.33587909487922392313_t_real*ylm(135)
    rvylm(1,160)=7.55415232778002351432_t_real*ylm(135)&
         &      -4.42325868464691400360_t_real*ylm(137)
    rvylm(2,160)=-4.42325868464691400360_t_real*ylm(129)&
         &      -7.55415232778002351432_t_real*ylm(131)
    rvylm(3,160)=12.11359279630872389380_t_real*ylm(136)
    rvylm(1,161)=8.07572853087248259575_t_real*ylm(136)&
         &      -3.90094748802746972907_t_real*ylm(138)
    rvylm(2,161)=-3.90094748802746972907_t_real*ylm(128)&
         &      -8.07572853087248259575_t_real*ylm(130)
    rvylm(3,161)=11.79535649239177067632_t_real*ylm(137)
    rvylm(1,162)=8.59726953621095152654_t_real*ylm(137)&
         &      -3.37831962346088111590_t_real*ylm(139)
    rvylm(2,162)=-3.37831962346088111590_t_real*ylm(127)&
         &      -8.59726953621095152654_t_real*ylm(129)
    rvylm(3,162)=11.37311857350289484499_t_real*ylm(138)
    rvylm(1,163)=9.11878138311493232910_t_real*ylm(138)&
         &      -2.85520120360080372741_t_real*ylm(140)
    rvylm(2,163)=-2.85520120360080372741_t_real*ylm(126)&
         &      -9.11878138311493232910_t_real*ylm(128)
    rvylm(3,163)=10.83472677771922806832_t_real*ylm(139)
    rvylm(1,164)=9.64026880375727138200_t_real*ylm(139)&
         &      -2.33126202060078441389_t_real*ylm(141)
    rvylm(2,164)=-2.33126202060078441389_t_real*ylm(125)&
         &      -9.64026880375727138200_t_real*ylm(127)
    rvylm(3,164)=10.16173555871325512508_t_real*ylm(140)
    rvylm(1,165)=10.16173555871325512508_t_real*ylm(140)&
         &      -1.80578779628653801127_t_real*ylm(142)
    rvylm(2,165)=-1.80578779628653801127_t_real*ylm(124)&
         &      -10.16173555871325512508_t_real*ylm(126)
    rvylm(3,165)=9.32504808240313765608_t_real*ylm(141)
    rvylm(1,166)=10.68318467417879543822_t_real*ylm(141)&
         &      -1.27688479613812288973_t_real*ylm(143)
    rvylm(2,166)=-1.27688479613812288973_t_real*ylm(123)&
         &      -10.68318467417879543822_t_real*ylm(125)
    rvylm(3,166)=8.27515926551055676937_t_real*ylm(142)
    rvylm(1,167)=11.20461861291447818860_t_real*ylm(142)&
         &      -0.73720978077448566723_t_real*ylm(144)
    rvylm(2,167)=-0.73720978077448566723_t_real*ylm(122)&
         &      -11.20461861291447818860_t_real*ylm(124)
    rvylm(3,167)=6.91564074808124624060_t_real*ylm(143)
    rvylm(1,168)=11.72603939955857388608_t_real*ylm(143)
    rvylm(2,168)=-11.72603939955857388608_t_real*ylm(123)
    rvylm(3,168)=4.99999999999999999974_t_real*ylm(144)
    rvylm(1,169)=12.24744871391589049060_t_real*ylm(144)
    rvylm(2,169)=-12.24744871391589049060_t_real*ylm(122)
    rvylm(3,169)= 0.0_t_real
    !!
    If(maxL<=12) Goto 1000
    !!

    !! XXXXXXXXXXXXXXXXXXX
    !!
    !!
    !!
1000 Continue
    !!
    !Return
    !!
    !!
    !!
    !!XXXXXXXXXXXXXXXXX
    !!..............................................
    !! L=0
    !!..............................................
    !! I1
    qvylm(1,1)= 0.0_t_real
    qvylm(2,1)= 0.0_t_real
    qvylm(3,1)= 0.0_t_real
    qvylm(4,1)= 0.0_t_real
    qvylm(5,1)= 0.0_t_real
    !!
    If(maxL<=0) Goto 2000
    !!
    !!..............................................
    !! L=1
    !!..............................................
    !! I1
    qvylm(1,2)= 0.0_t_real
    qvylm(2,2)= 0.0_t_real
    qvylm(3,2)= 0.0_t_real
    qvylm(4,2)= 0.0_t_real
    qvylm(5,2)= 0.0_t_real
    qvylm(1,3)= 0.0_t_real
    qvylm(2,3)= 0.0_t_real
    qvylm(3,3)= 0.0_t_real
    qvylm(4,3)= 0.0_t_real
    qvylm(5,3)= 0.0_t_real
    qvylm(1,4)= 0.0_t_real
    qvylm(2,4)= 0.0_t_real
    qvylm(3,4)= 0.0_t_real
    qvylm(4,4)= 0.0_t_real
    qvylm(5,4)= 0.0_t_real
    !!
    If(maxL<=1) Goto 2000
    !!
    !!..............................................
    !! L=2
    !!..............................................
    !! I1
    qvylm(1,5)= 0.0_t_real
    qvylm(2,5)=1.09254843059207907046_t_real
    qvylm(3,5)= 0.0_t_real
    qvylm(4,5)= 0.0_t_real
    qvylm(5,5)= 0.0_t_real
    qvylm(1,6)= 0.0_t_real
    qvylm(2,6)= 0.0_t_real
    qvylm(3,6)= 0.0_t_real
    qvylm(4,6)= 0.0_t_real
    qvylm(5,6)=1.09254843059207907046_t_real
    qvylm(1,7)=-0.63078313050504001199_t_real
    qvylm(2,7)= 0.0_t_real
    qvylm(3,7)= 0.0_t_real
    qvylm(4,7)=-0.63078313050504001199_t_real
    qvylm(5,7)= 0.0_t_real
    qvylm(1,8)= 0.0_t_real
    qvylm(2,8)= 0.0_t_real
    qvylm(3,8)=1.09254843059207907046_t_real
    qvylm(4,8)= 0.0_t_real
    qvylm(5,8)= 0.0_t_real
    qvylm(1,9)=1.09254843059207907046_t_real
    qvylm(2,9)= 0.0_t_real
    qvylm(3,9)= 0.0_t_real
    qvylm(4,9)=-1.09254843059207907046_t_real
    qvylm(5,9)= 0.0_t_real
    !!
    If(maxL<=2) Goto 2000
    !!
    !!..............................................
    !! L=3
    !!..............................................
    !! I1
    qvylm(1,10)=7.24568837309471928578_t_real*ylm(2)
    qvylm(2,10)=7.24568837309471928578_t_real*ylm(4)
    qvylm(3,10)= 0.0_t_real
    qvylm(4,10)=-7.24568837309471928578_t_real*ylm(2)
    qvylm(5,10)= 0.0_t_real
    qvylm(1,11)= 0.0_t_real
    qvylm(2,11)=5.91607978309961604169_t_real*ylm(3)
    qvylm(3,11)=5.91607978309961604169_t_real*ylm(2)
    qvylm(4,11)= 0.0_t_real
    qvylm(5,11)=5.91607978309961604169_t_real*ylm(4)
    qvylm(1,12)=-1.87082869338697069243_t_real*ylm(2)
    qvylm(2,12)=-1.87082869338697069238_t_real*ylm(4)
    qvylm(3,12)= 0.0_t_real
    qvylm(4,12)=-5.61248608016091207724_t_real*ylm(2)
    qvylm(5,12)=7.48331477354788277008_t_real*ylm(3)
    qvylm(1,13)=-4.58257569495584000577_t_real*ylm(3)
    qvylm(2,13)= 0.0_t_real
    qvylm(3,13)=-4.58257569495584000582_t_real*ylm(4)
    qvylm(4,13)=-4.58257569495584000577_t_real*ylm(3)
    qvylm(5,13)=-4.58257569495584000582_t_real*ylm(2)
    qvylm(1,14)=-5.61248608016091207724_t_real*ylm(4)
    qvylm(2,14)=-1.87082869338697069243_t_real*ylm(2)
    qvylm(3,14)=7.48331477354788277008_t_real*ylm(3)
    qvylm(4,14)=-1.87082869338697069243_t_real*ylm(4)
    qvylm(5,14)= 0.0_t_real
    qvylm(1,15)=5.91607978309961604169_t_real*ylm(3)
    qvylm(2,15)= 0.0_t_real
    qvylm(3,15)=5.91607978309961604169_t_real*ylm(4)
    qvylm(4,15)=-5.91607978309961604169_t_real*ylm(3)
    qvylm(5,15)=-5.91607978309961604169_t_real*ylm(2)
    qvylm(1,16)=7.24568837309471928578_t_real*ylm(4)
    qvylm(2,16)=-7.24568837309471928578_t_real*ylm(2)
    qvylm(3,16)= 0.0_t_real
    qvylm(4,16)=-7.24568837309471928578_t_real*ylm(4)
    qvylm(5,16)= 0.0_t_real
    !!
    If(maxL<=3) Goto 2000
    !!
    !!..............................................
    !! L=4
    !!..............................................
    !! I1
    qvylm(1,17)=13.74772708486752001763_t_real*ylm(5)
    qvylm(2,17)=13.74772708486752001763_t_real*ylm(9)
    qvylm(3,17)= 0.0_t_real
    qvylm(4,17)=-13.74772708486752001763_t_real*ylm(5)
    qvylm(5,17)= 0.0_t_real
    qvylm(1,18)=9.72111104761179034491_t_real*ylm(6)
    qvylm(2,18)=9.72111104761179034491_t_real*ylm(8)
    qvylm(3,18)=9.72111104761179034499_t_real*ylm(5)
    qvylm(4,18)=-9.72111104761179034491_t_real*ylm(6)
    qvylm(5,18)=9.72111104761179034499_t_real*ylm(9)
    qvylm(1,19)=-5.19615242270663187951_t_real*ylm(5)
    qvylm(2,19)=8.99999999999999999828_t_real*ylm(7)&
         &      +0.00000000000000000001_t_real*ylm(9)
    qvylm(3,19)=10.39230484541326375958_t_real*ylm(6)
    qvylm(4,19)=-5.19615242270663187951_t_real*ylm(5)
    qvylm(5,19)=10.39230484541326375958_t_real*ylm(8)
    qvylm(1,20)=-3.67423461417476714669_t_real*ylm(6)
    qvylm(2,20)=-3.67423461417476714632_t_real*ylm(8)
    qvylm(3,20)=-3.67423461417476714655_t_real*ylm(5)
    qvylm(4,20)=-11.02270384252430143970_t_real*ylm(6)
    qvylm(5,20)=12.72792206135785543685_t_real*ylm(7)&
         &      +3.67423461417476714655_t_real*ylm(9)
    qvylm(1,21)=-8.04984471899924290539_t_real*ylm(7)&
         &      +2.32379000772445013058_t_real*ylm(9)
    qvylm(2,21)=2.32379000772445013058_t_real*ylm(5)
    qvylm(3,21)=-9.29516003089780052277_t_real*ylm(8)
    qvylm(4,21)=-8.04984471899924290539_t_real*ylm(7)&
         &      -2.32379000772445013058_t_real*ylm(9)
    qvylm(5,21)=-9.29516003089780052277_t_real*ylm(6)
    qvylm(1,22)=-11.02270384252430143970_t_real*ylm(8)
    qvylm(2,22)=-3.67423461417476714669_t_real*ylm(6)
    qvylm(3,22)=12.72792206135785543685_t_real*ylm(7)&
         &      -3.67423461417476714655_t_real*ylm(9)
    qvylm(4,22)=-3.67423461417476714669_t_real*ylm(8)
    qvylm(5,22)=-3.67423461417476714655_t_real*ylm(5)
    qvylm(1,23)=8.99999999999999999828_t_real*ylm(7)&
         &      -5.19615242270663187951_t_real*ylm(9)
    qvylm(2,23)=-0.00000000000000000001_t_real*ylm(5)
    qvylm(3,23)=10.39230484541326375958_t_real*ylm(8)
    qvylm(4,23)=-5.19615242270663187951_t_real*ylm(9)&
         &      -8.99999999999999999828_t_real*ylm(7)
    qvylm(5,23)=-10.39230484541326375958_t_real*ylm(6)
    qvylm(1,24)=9.72111104761179034491_t_real*ylm(8)
    qvylm(2,24)=-9.72111104761179034491_t_real*ylm(6)
    qvylm(3,24)=9.72111104761179034499_t_real*ylm(9)
    qvylm(4,24)=-9.72111104761179034491_t_real*ylm(8)
    qvylm(5,24)=-9.72111104761179034499_t_real*ylm(5)
    qvylm(1,25)=13.74772708486752001763_t_real*ylm(9)
    qvylm(2,25)=-13.74772708486752001763_t_real*ylm(5)
    qvylm(3,25)= 0.0_t_real
    qvylm(4,25)=-13.74772708486752001763_t_real*ylm(9)
    qvylm(5,25)= 0.0_t_real
    !!
    If(maxL<=4) Goto 2000
    !!
    !!..............................................
    !! L=5
    !!..............................................
    !! I1
    qvylm(1,26)=22.24859546128698884431_t_real*ylm(10)
    qvylm(2,26)=22.24859546128698884431_t_real*ylm(16)
    qvylm(3,26)= 0.0_t_real
    qvylm(4,26)=-22.24859546128698884431_t_real*ylm(10)
    qvylm(5,26)= 0.0_t_real
    qvylm(1,27)=17.23368793961408597807_t_real*ylm(11)
    qvylm(2,27)=17.23368793961408597807_t_real*ylm(15)
    qvylm(3,27)=14.07124727947028866268_t_real*ylm(10)
    qvylm(4,27)=-17.23368793961408597807_t_real*ylm(11)
    qvylm(5,27)=14.07124727947028866268_t_real*ylm(16)
    qvylm(1,28)=-6.63324958071079969747_t_real*ylm(10)&
         &      +12.84523257866512901902_t_real*ylm(12)
    qvylm(2,28)=12.84523257866512901902_t_real*ylm(14)&
         &      +0.00000000000000000007_t_real*ylm(16)
    qvylm(3,28)=16.24807680927192071973_t_real*ylm(11)
    qvylm(4,28)=-6.63324958071079969747_t_real*ylm(10)&
         &      -12.84523257866512901902_t_real*ylm(12)
    qvylm(5,28)=16.24807680927192071973_t_real*ylm(15)
    qvylm(1,29)=-9.94987437106619954645_t_real*ylm(11)
    qvylm(2,29)=12.84523257866512901868_t_real*ylm(13)&
         &      +0.00000000000000000003_t_real*ylm(15)
    qvylm(3,29)=-4.06201920231798017971_t_real*ylm(10)&
         &      +15.73213272255227320348_t_real*ylm(12)
    qvylm(4,29)=-9.94987437106619954645_t_real*ylm(11)
    qvylm(5,29)=15.73213272255227320348_t_real*ylm(14)&
         &      +4.06201920231798017971_t_real*ylm(16)
    qvylm(1,30)=1.53529894715747693114_t_real*ylm(10)&
         &      -5.94618725379068911936_t_real*ylm(12)
    qvylm(2,30)=-5.94618725379068911919_t_real*ylm(14)&
         &      -1.53529894715747693114_t_real*ylm(16)
    qvylm(3,30)=-7.52139804633610462254_t_real*ylm(11)
    qvylm(4,30)=-17.83856176137206735791_t_real*ylm(12)&
         &      -1.53529894715747693114_t_real*ylm(10)
    qvylm(5,30)=19.42016624910448965065_t_real*ylm(13)&
         &      +7.52139804633610462254_t_real*ylm(15)
    qvylm(1,31)=-12.53566341056017437044_t_real*ylm(13)&
         &      +4.85504156227612241268_t_real*ylm(15)
    qvylm(2,31)=4.85504156227612241268_t_real*ylm(11)
    qvylm(3,31)=-15.35298947157476931224_t_real*ylm(14)
    qvylm(4,31)=-12.53566341056017437044_t_real*ylm(13)&
         &      -4.85504156227612241268_t_real*ylm(15)
    qvylm(5,31)=-15.35298947157476931224_t_real*ylm(12)
    qvylm(1,32)=-17.83856176137206735791_t_real*ylm(14)&
         &      +1.53529894715747693114_t_real*ylm(16)
    qvylm(2,32)=1.53529894715747693114_t_real*ylm(10)&
         &      -5.94618725379068911936_t_real*ylm(12)
    qvylm(3,32)=19.42016624910448965065_t_real*ylm(13)&
         &      -7.52139804633610462254_t_real*ylm(15)
    qvylm(4,32)=-5.94618725379068911936_t_real*ylm(14)&
         &      -1.53529894715747693114_t_real*ylm(16)
    qvylm(5,32)=-7.52139804633610462254_t_real*ylm(11)
    qvylm(1,33)=12.84523257866512901868_t_real*ylm(13)&
         &      -9.94987437106619954645_t_real*ylm(15)
    qvylm(2,33)=-0.00000000000000000003_t_real*ylm(11)
    qvylm(3,33)=15.73213272255227320348_t_real*ylm(14)&
         &      -4.06201920231798017971_t_real*ylm(16)
    qvylm(4,33)=-9.94987437106619954645_t_real*ylm(15)&
         &      -12.84523257866512901868_t_real*ylm(13)
    qvylm(5,33)=-4.06201920231798017971_t_real*ylm(10)&
         &      -15.73213272255227320348_t_real*ylm(12)
    qvylm(1,34)=12.84523257866512901902_t_real*ylm(14)&
         &      -6.63324958071079969747_t_real*ylm(16)
    qvylm(2,34)=-0.00000000000000000007_t_real*ylm(10)&
         &      -12.84523257866512901902_t_real*ylm(12)
    qvylm(3,34)=16.24807680927192071973_t_real*ylm(15)
    qvylm(4,34)=-6.63324958071079969747_t_real*ylm(16)&
         &      -12.84523257866512901902_t_real*ylm(14)
    qvylm(5,34)=-16.24807680927192071973_t_real*ylm(11)
    qvylm(1,35)=17.23368793961408597807_t_real*ylm(15)
    qvylm(2,35)=-17.23368793961408597807_t_real*ylm(11)
    qvylm(3,35)=14.07124727947028866268_t_real*ylm(16)
    qvylm(4,35)=-17.23368793961408597807_t_real*ylm(15)
    qvylm(5,35)=-14.07124727947028866268_t_real*ylm(10)
    qvylm(1,36)=22.24859546128698884431_t_real*ylm(16)
    qvylm(2,36)=-22.24859546128698884431_t_real*ylm(10)
    qvylm(3,36)= 0.0_t_real
    qvylm(4,36)=-22.24859546128698884431_t_real*ylm(16)
    qvylm(5,36)= 0.0_t_real
    !!
    If(maxL<=5) Goto 2000
    !!
    !!..............................................
    !! L=6
    !!..............................................
    !! I1
    qvylm(1,37)=32.74904578762562968312_t_real*ylm(17)
    qvylm(2,37)=32.74904578762562968312_t_real*ylm(25)
    qvylm(3,37)= 0.0_t_real
    qvylm(4,37)=-32.74904578762562968312_t_real*ylm(17)
    qvylm(5,37)= 0.0_t_real
    qvylm(1,38)=26.73948391424187528116_t_real*ylm(18)
    qvylm(2,38)=26.73948391424187528116_t_real*ylm(24)
    qvylm(3,38)=18.90767040118903701306_t_real*ylm(17)
    qvylm(4,38)=-26.73948391424187528116_t_real*ylm(18)
    qvylm(5,38)=18.90767040118903701306_t_real*ylm(25)
    qvylm(1,39)=-8.06225774829854965058_t_real*ylm(17)&
         &      +21.33072900770154174743_t_real*ylm(19)
    qvylm(2,39)=21.33072900770154174743_t_real*ylm(23)&
         &      -0.00000000000000000024_t_real*ylm(25)
    qvylm(3,39)=22.80350850198275957877_t_real*ylm(18)
    qvylm(4,39)=-8.06225774829854965058_t_real*ylm(17)&
         &      -21.33072900770154174743_t_real*ylm(19)
    qvylm(5,39)=22.80350850198275957877_t_real*ylm(24)
    qvylm(1,40)=-12.48999599679679640950_t_real*ylm(18)&
         &      +16.52271164185830605925_t_real*ylm(20)
    qvylm(2,40)=16.52271164185830605925_t_real*ylm(22)&
         &      -0.00000000000000000024_t_real*ylm(24)
    qvylm(3,40)=-4.41588043316392342660_t_real*ylm(17)&
         &      +23.36664289109584521786_t_real*ylm(19)
    qvylm(4,40)=-12.48999599679679640950_t_real*ylm(18)&
         &      -16.52271164185830605925_t_real*ylm(20)
    qvylm(5,40)=23.36664289109584521786_t_real*ylm(23)&
         &      +4.41588043316392342660_t_real*ylm(25)
    qvylm(1,41)=1.47196014438797447550_t_real*ylm(17)&
         &      -15.57776192739723014479_t_real*ylm(19)
    qvylm(2,41)=17.41646730348417509475_t_real*ylm(21)&
         &      +0.00000000000000000013_t_real*ylm(23)&
         &      -1.47196014438797447550_t_real*ylm(25)
    qvylm(3,41)=-8.32666399786453093975_t_real*ylm(18)&
         &      +22.03028218914440807880_t_real*ylm(20)
    qvylm(4,41)=-15.57776192739723014479_t_real*ylm(19)&
         &      -1.47196014438797447550_t_real*ylm(17)
    qvylm(5,41)=22.03028218914440807880_t_real*ylm(22)&
         &      +8.32666399786453093975_t_real*ylm(24)
    qvylm(1,42)=3.29140294302191650228_t_real*ylm(18)&
         &      -8.70823365174208754732_t_real*ylm(20)
    qvylm(2,42)=-8.70823365174208754740_t_real*ylm(22)&
         &      -3.29140294302191650228_t_real*ylm(24)
    qvylm(3,42)=-12.31530213460744399491_t_real*ylm(19)
    qvylm(4,42)=-26.12470095522626264204_t_real*ylm(20)&
         &      -3.29140294302191650228_t_real*ylm(18)
    qvylm(5,42)=27.53785273643051009796_t_real*ylm(21)&
         &      +12.31530213460744399491_t_real*ylm(23)
    qvylm(1,43)=-18.02775637731994646220_t_real*ylm(21)&
         &      +8.06225774829854965082_t_real*ylm(23)
    qvylm(2,43)=8.06225774829854965082_t_real*ylm(19)
    qvylm(3,43)=-22.80350850198275957893_t_real*ylm(22)
    qvylm(4,43)=-18.02775637731994646220_t_real*ylm(21)&
         &      -8.06225774829854965082_t_real*ylm(23)
    qvylm(5,43)=-22.80350850198275957893_t_real*ylm(20)
    qvylm(1,44)=-26.12470095522626264204_t_real*ylm(22)&
         &      +3.29140294302191650228_t_real*ylm(24)
    qvylm(2,44)=3.29140294302191650228_t_real*ylm(18)&
         &      -8.70823365174208754732_t_real*ylm(20)
    qvylm(3,44)=27.53785273643051009796_t_real*ylm(21)&
         &      -12.31530213460744399491_t_real*ylm(23)
    qvylm(4,44)=-8.70823365174208754732_t_real*ylm(22)&
         &      -3.29140294302191650228_t_real*ylm(24)
    qvylm(5,44)=-12.31530213460744399491_t_real*ylm(19)
    qvylm(1,45)=17.41646730348417509475_t_real*ylm(21)&
         &      -15.57776192739723014479_t_real*ylm(23)&
         &      +1.47196014438797447550_t_real*ylm(25)
    qvylm(2,45)=1.47196014438797447550_t_real*ylm(17)&
         &      -0.00000000000000000013_t_real*ylm(19)
    qvylm(3,45)=22.03028218914440807880_t_real*ylm(22)&
         &      -8.32666399786453093975_t_real*ylm(24)
    qvylm(4,45)=-15.57776192739723014479_t_real*ylm(23)&
         &      -1.47196014438797447550_t_real*ylm(25)&
         &      -17.41646730348417509475_t_real*ylm(21)
    qvylm(5,45)=-8.32666399786453093975_t_real*ylm(18)&
         &      -22.03028218914440807880_t_real*ylm(20)
    qvylm(1,46)=16.52271164185830605925_t_real*ylm(22)&
         &      -12.48999599679679640950_t_real*ylm(24)
    qvylm(2,46)= 0.00000000000000000024_t_real*ylm(18)&
         &      -16.52271164185830605925_t_real*ylm(20)
    qvylm(3,46)=23.36664289109584521786_t_real*ylm(23)&
         &      -4.41588043316392342660_t_real*ylm(25)
    qvylm(4,46)=-12.48999599679679640950_t_real*ylm(24)&
         &      -16.52271164185830605925_t_real*ylm(22)
    qvylm(5,46)=-4.41588043316392342660_t_real*ylm(17)&
         &      -23.36664289109584521786_t_real*ylm(19)
    qvylm(1,47)=21.33072900770154174743_t_real*ylm(23)&
         &      -8.06225774829854965058_t_real*ylm(25)
    qvylm(2,47)= 0.00000000000000000024_t_real*ylm(17)&
         &      -21.33072900770154174743_t_real*ylm(19)
    qvylm(3,47)=22.80350850198275957877_t_real*ylm(24)
    qvylm(4,47)=-8.06225774829854965058_t_real*ylm(25)&
         &      -21.33072900770154174743_t_real*ylm(23)
    qvylm(5,47)=-22.80350850198275957877_t_real*ylm(18)
    qvylm(1,48)=26.73948391424187528116_t_real*ylm(24)
    qvylm(2,48)=-26.73948391424187528116_t_real*ylm(18)
    qvylm(3,48)=18.90767040118903701306_t_real*ylm(25)
    qvylm(4,48)=-26.73948391424187528116_t_real*ylm(24)
    qvylm(5,48)=-18.90767040118903701306_t_real*ylm(17)
    qvylm(1,49)=32.74904578762562968312_t_real*ylm(25)
    qvylm(2,49)=-32.74904578762562968312_t_real*ylm(17)
    qvylm(3,49)= 0.0_t_real
    qvylm(4,49)=-32.74904578762562968312_t_real*ylm(25)
    qvylm(5,49)= 0.0_t_real
    !!
    If(maxL<=6) Goto 2000
    !!
    !!..............................................
    !! L=7
    !!..............................................
    !! I1
    qvylm(1,50)=45.24930938699506824419_t_real*ylm(26)
    qvylm(2,50)=45.24930938699506824419_t_real*ylm(36)
    qvylm(3,50)= 0.0_t_real
    qvylm(4,50)=-45.24930938699506824419_t_real*ylm(26)
    qvylm(5,50)= 0.0_t_real
    qvylm(1,51)=38.24264635194588621341_t_real*ylm(27)
    qvylm(2,51)=38.24264635194588621341_t_real*ylm(35)
    qvylm(3,51)=24.18677324489564894931_t_real*ylm(26)
    qvylm(4,51)=-38.24264635194588621341_t_real*ylm(27)
    qvylm(5,51)=24.18677324489564894931_t_real*ylm(36)
    qvylm(1,52)=-9.48683298050513799255_t_real*ylm(26)&
         &      +31.81980515339463858782_t_real*ylm(28)
    qvylm(2,52)=31.81980515339463858782_t_real*ylm(34)&
         &      +0.00000000000000000039_t_real*ylm(36)
    qvylm(3,52)=29.99999999999999999045_t_real*ylm(27)
    qvylm(4,52)=-9.48683298050513799255_t_real*ylm(26)&
         &      -31.81980515339463858782_t_real*ylm(28)
    qvylm(5,52)=29.99999999999999999045_t_real*ylm(35)
    qvylm(1,53)=-14.99999999999999999491_t_real*ylm(27)&
         &      +25.98076211353315939461_t_real*ylm(29)
    qvylm(2,53)=25.98076211353315939461_t_real*ylm(33)&
         &      +0.00000000000000000029_t_real*ylm(35)
    qvylm(3,53)=-4.74341649025256899608_t_real*ylm(26)&
         &      +31.81980515339463858782_t_real*ylm(28)
    qvylm(4,53)=-14.99999999999999999491_t_real*ylm(27)&
         &      -25.98076211353315939461_t_real*ylm(29)
    qvylm(5,53)=31.81980515339463858782_t_real*ylm(34)&
         &      +4.74341649025256899608_t_real*ylm(36)
    qvylm(1,54)=1.43019388386838847256_t_real*ylm(26)&
         &      -19.18806447200493908042_t_real*ylm(28)&
         &      +20.72547839123272312753_t_real*ylm(30)
    qvylm(2,54)=20.72547839123272312753_t_real*ylm(32)&
         &      -0.00000000000000000016_t_real*ylm(34)&
         &      -1.43019388386838847256_t_real*ylm(36)
    qvylm(3,54)=-9.04534033733290867623_t_real*ylm(27)&
         &      +31.33397807202561086213_t_real*ylm(29)
    qvylm(4,54)=-19.18806447200493908042_t_real*ylm(28)&
         &      -20.72547839123272312753_t_real*ylm(30)&
         &      -1.43019388386838847256_t_real*ylm(26)
    qvylm(5,54)=31.33397807202561086213_t_real*ylm(33)&
         &      +9.04534033733290867623_t_real*ylm(35)
    qvylm(1,55)=3.19801074533415651323_t_real*ylm(27)&
         &      -22.15646837627989188446_t_real*ylm(29)
    qvylm(2,55)=22.70362405992808572358_t_real*ylm(31)&
         &      -0.00000000000000000004_t_real*ylm(33)&
         &      -3.19801074533415651323_t_real*ylm(35)
    qvylm(3,55)=-13.56801050599936301452_t_real*ylm(28)&
         &      +29.31025262755183231663_t_real*ylm(30)
    qvylm(4,55)=-22.15646837627989188446_t_real*ylm(29)&
         &      -3.19801074533415651323_t_real*ylm(27)
    qvylm(5,55)=29.31025262755183231663_t_real*ylm(32)&
         &      +13.56801050599936301452_t_real*ylm(34)
    qvylm(1,56)=5.53911709406997297110_t_real*ylm(28)&
         &      -11.96586052826198462376_t_real*ylm(30)
    qvylm(2,56)=-11.96586052826198462391_t_real*ylm(32)&
         &      -5.53911709406997297110_t_real*ylm(34)
    qvylm(3,56)=-18.09068067466581735250_t_real*ylm(29)
    qvylm(4,56)=-35.89758158478595387143_t_real*ylm(30)&
         &      -5.53911709406997297110_t_real*ylm(28)
    qvylm(5,56)=37.07486283919948023393_t_real*ylm(31)&
         &      +18.09068067466581735250_t_real*ylm(33)
    qvylm(1,57)=-24.52271674108797357492_t_real*ylm(31)&
         &      +11.96586052826198462325_t_real*ylm(33)
    qvylm(2,57)=11.96586052826198462325_t_real*ylm(29)
    qvylm(3,57)=-31.65869118066518036774_t_real*ylm(32)
    qvylm(4,57)=-24.52271674108797357492_t_real*ylm(31)&
         &      -11.96586052826198462325_t_real*ylm(33)
    qvylm(5,57)=-31.65869118066518036774_t_real*ylm(30)
    qvylm(1,58)=-35.89758158478595387143_t_real*ylm(32)&
         &      +5.53911709406997297110_t_real*ylm(34)
    qvylm(2,58)=5.53911709406997297110_t_real*ylm(28)&
         &      -11.96586052826198462376_t_real*ylm(30)
    qvylm(3,58)=37.07486283919948023393_t_real*ylm(31)&
         &      -18.09068067466581735250_t_real*ylm(33)
    qvylm(4,58)=-11.96586052826198462376_t_real*ylm(32)&
         &      -5.53911709406997297110_t_real*ylm(34)
    qvylm(5,58)=-18.09068067466581735250_t_real*ylm(29)
    qvylm(1,59)=22.70362405992808572358_t_real*ylm(31)&
         &      -22.15646837627989188446_t_real*ylm(33)&
         &      +3.19801074533415651323_t_real*ylm(35)
    qvylm(2,59)=3.19801074533415651323_t_real*ylm(27)&
         &      +0.00000000000000000004_t_real*ylm(29)
    qvylm(3,59)=29.31025262755183231663_t_real*ylm(32)&
         &      -13.56801050599936301452_t_real*ylm(34)
    qvylm(4,59)=-22.15646837627989188446_t_real*ylm(33)&
         &      -3.19801074533415651323_t_real*ylm(35)&
         &      -22.70362405992808572358_t_real*ylm(31)
    qvylm(5,59)=-13.56801050599936301452_t_real*ylm(28)&
         &      -29.31025262755183231663_t_real*ylm(30)
    qvylm(1,60)=20.72547839123272312753_t_real*ylm(32)&
         &      -19.18806447200493908042_t_real*ylm(34)&
         &      +1.43019388386838847256_t_real*ylm(36)
    qvylm(2,60)=1.43019388386838847256_t_real*ylm(26)&
         &      +0.00000000000000000016_t_real*ylm(28)&
         &      -20.72547839123272312753_t_real*ylm(30)
    qvylm(3,60)=31.33397807202561086213_t_real*ylm(33)&
         &      -9.04534033733290867623_t_real*ylm(35)
    qvylm(4,60)=-19.18806447200493908042_t_real*ylm(34)&
         &      -1.43019388386838847256_t_real*ylm(36)&
         &      -20.72547839123272312753_t_real*ylm(32)
    qvylm(5,60)=-9.04534033733290867623_t_real*ylm(27)&
         &      -31.33397807202561086213_t_real*ylm(29)
    qvylm(1,61)=25.98076211353315939461_t_real*ylm(33)&
         &      -14.99999999999999999491_t_real*ylm(35)
    qvylm(2,61)=-0.00000000000000000029_t_real*ylm(27)&
         &      -25.98076211353315939461_t_real*ylm(29)
    qvylm(3,61)=31.81980515339463858782_t_real*ylm(34)&
         &      -4.74341649025256899608_t_real*ylm(36)
    qvylm(4,61)=-14.99999999999999999491_t_real*ylm(35)&
         &      -25.98076211353315939461_t_real*ylm(33)
    qvylm(5,61)=-4.74341649025256899608_t_real*ylm(26)&
         &      -31.81980515339463858782_t_real*ylm(28)
    qvylm(1,62)=31.81980515339463858782_t_real*ylm(34)&
         &      -9.48683298050513799255_t_real*ylm(36)
    qvylm(2,62)=-0.00000000000000000039_t_real*ylm(26)&
         &      -31.81980515339463858782_t_real*ylm(28)
    qvylm(3,62)=29.99999999999999999045_t_real*ylm(35)
    qvylm(4,62)=-9.48683298050513799255_t_real*ylm(36)&
         &      -31.81980515339463858782_t_real*ylm(34)
    qvylm(5,62)=-29.99999999999999999045_t_real*ylm(27)
    qvylm(1,63)=38.24264635194588621341_t_real*ylm(35)
    qvylm(2,63)=-38.24264635194588621341_t_real*ylm(27)
    qvylm(3,63)=24.18677324489564894931_t_real*ylm(36)
    qvylm(4,63)=-38.24264635194588621341_t_real*ylm(35)
    qvylm(5,63)=-24.18677324489564894931_t_real*ylm(26)
    qvylm(1,64)=45.24930938699506824419_t_real*ylm(36)
    qvylm(2,64)=-45.24930938699506824419_t_real*ylm(26)
    qvylm(3,64)= 0.0_t_real
    qvylm(4,64)=-45.24930938699506824419_t_real*ylm(36)
    qvylm(5,64)= 0.0_t_real
    !!
    If(maxL<=7) Goto 2000
    !!
    !!..............................................
    !! L=8
    !!..............................................
    !! I1
    qvylm(1,65)=59.74947698515862320431_t_real*ylm(37)
    qvylm(2,65)=59.74947698515862320431_t_real*ylm(49)
    qvylm(3,65)= 0.0_t_real
    qvylm(4,65)=-59.74947698515862320431_t_real*ylm(37)
    qvylm(5,65)= 0.0_t_real
    qvylm(1,66)=51.74456493198102055052_t_real*ylm(38)
    qvylm(2,66)=51.74456493198102055052_t_real*ylm(48)
    qvylm(3,66)=29.87473849257931160172_t_real*ylm(37)
    qvylm(4,66)=-51.74456493198102055052_t_real*ylm(38)
    qvylm(5,66)=29.87473849257931160172_t_real*ylm(49)
    qvylm(1,67)=-10.90871211463571440658_t_real*ylm(37)&
         &      +44.31139808220905140193_t_real*ylm(39)
    qvylm(2,67)=44.31139808220905140193_t_real*ylm(47)&
         &      -0.00000000000000000072_t_real*ylm(49)
    qvylm(3,67)=37.78888725538236854780_t_real*ylm(38)
    qvylm(4,67)=-10.90871211463571440658_t_real*ylm(37)&
         &      -44.31139808220905140193_t_real*ylm(39)
    qvylm(5,67)=37.78888725538236854780_t_real*ylm(48)
    qvylm(1,68)=-17.49285568453590140575_t_real*ylm(38)&
         &      +37.44996662214800958436_t_real*ylm(40)
    qvylm(2,68)=37.44996662214800958436_t_real*ylm(46)&
         &      +0.00000000000000000017_t_real*ylm(48)
    qvylm(3,68)=-5.04975246918103897468_t_real*ylm(37)&
         &      +41.02438299353202945970_t_real*ylm(39)
    qvylm(4,68)=-17.49285568453590140575_t_real*ylm(38)&
         &      -37.44996662214800958436_t_real*ylm(40)
    qvylm(5,68)=41.02438299353202945970_t_real*ylm(47)&
         &      +5.04975246918103897468_t_real*ylm(49)
    qvylm(1,69)=1.40054934277177879689_t_real*ylm(37)&
         &      -22.75623329653116923637_t_real*ylm(39)&
         &      +31.16025575090067293966_t_real*ylm(41)
    qvylm(2,69)=31.16025575090067293966_t_real*ylm(45)&
         &      +0.00000000000000000017_t_real*ylm(47)&
         &      -1.40054934277177879689_t_real*ylm(49)
    qvylm(3,69)=-9.70329048075167920614_t_real*ylm(38)&
         &      +41.54700766786756391874_t_real*ylm(40)
    qvylm(4,69)=-22.75623329653116923637_t_real*ylm(39)&
         &      -31.16025575090067293966_t_real*ylm(41)&
         &      -1.40054934277177879689_t_real*ylm(37)
    qvylm(5,69)=41.54700766786756391874_t_real*ylm(46)&
         &      +9.70329048075167920614_t_real*ylm(48)
    qvylm(1,70)=3.13172353628035111779_t_real*ylm(38)&
         &      -26.81847813040048755100_t_real*ylm(40)&
         &      +25.44224228144391182149_t_real*ylm(42)
    qvylm(2,70)=25.44224228144391182149_t_real*ylm(44)&
         &      -3.13172353628035111779_t_real*ylm(48)
    qvylm(3,70)=-14.68908542997932084164_t_real*ylm(39)&
         &      +40.22771719560073132796_t_real*ylm(41)
    qvylm(4,70)=-26.81847813040048755100_t_real*ylm(40)&
         &      -25.44224228144391182149_t_real*ylm(42)&
         &      -3.13172353628035111779_t_real*ylm(38)
    qvylm(5,70)=40.22771719560073132796_t_real*ylm(45)&
         &      +14.68908542997932084164_t_real*ylm(47)
    qvylm(1,71)=5.42430428009684234122_t_real*ylm(39)&
         &      -29.71013812980861406194_t_real*ylm(41)
    qvylm(2,71)=28.70272032135898169616_t_real*ylm(43)&
         &      +0.00000000000000000020_t_real*ylm(45)&
         &      -5.42430428009684234122_t_real*ylm(47)
    qvylm(3,71)=-19.80675875320574270758_t_real*ylm(40)&
         &      +37.58068243536421341488_t_real*ylm(42)
    qvylm(4,71)=-29.71013812980861406194_t_real*ylm(41)&
         &      -5.42430428009684234122_t_real*ylm(39)
    qvylm(5,71)=37.58068243536421341488_t_real*ylm(44)&
         &      +19.80675875320574270758_t_real*ylm(46)
    qvylm(1,72)=8.28576165200557490582_t_real*ylm(40)&
         &      -15.72112738177024483991_t_real*ylm(42)
    qvylm(2,72)=-15.72112738177024483823_t_real*ylm(44)&
         &      -8.28576165200557490582_t_real*ylm(46)
    qvylm(3,72)=-24.85728495601672471819_t_real*ylm(41)
    qvylm(4,72)=-47.16338214531073451805_t_real*ylm(42)&
         &      -8.28576165200557490582_t_real*ylm(40)
    qvylm(5,72)=48.02883749133671013082_t_real*ylm(43)&
         &      +24.85728495601672471819_t_real*ylm(45)
    qvylm(1,73)=-32.01922499422447342007_t_real*ylm(43)&
         &      +16.57152330401114981188_t_real*ylm(45)
    qvylm(2,73)=16.57152330401114981188_t_real*ylm(41)
    qvylm(3,73)=-41.92300635138731957243_t_real*ylm(44)
    qvylm(4,73)=-32.01922499422447342007_t_real*ylm(43)&
         &      -16.57152330401114981188_t_real*ylm(45)
    qvylm(5,73)=-41.92300635138731957243_t_real*ylm(42)
    qvylm(1,74)=-47.16338214531073451805_t_real*ylm(44)&
         &      +8.28576165200557490582_t_real*ylm(46)
    qvylm(2,74)=8.28576165200557490582_t_real*ylm(40)&
         &      -15.72112738177024483991_t_real*ylm(42)
    qvylm(3,74)=48.02883749133671013082_t_real*ylm(43)&
         &      -24.85728495601672471819_t_real*ylm(45)
    qvylm(4,74)=-15.72112738177024483991_t_real*ylm(44)&
         &      -8.28576165200557490582_t_real*ylm(46)
    qvylm(5,74)=-24.85728495601672471819_t_real*ylm(41)
    qvylm(1,75)=28.70272032135898169616_t_real*ylm(43)&
         &      -29.71013812980861406194_t_real*ylm(45)&
         &      +5.42430428009684234122_t_real*ylm(47)
    qvylm(2,75)=5.42430428009684234122_t_real*ylm(39)&
         &      -0.00000000000000000020_t_real*ylm(41)
    qvylm(3,75)=37.58068243536421341488_t_real*ylm(44)&
         &      -19.80675875320574270758_t_real*ylm(46)
    qvylm(4,75)=-29.71013812980861406194_t_real*ylm(45)&
         &      -5.42430428009684234122_t_real*ylm(47)&
         &      -28.70272032135898169616_t_real*ylm(43)
    qvylm(5,75)=-19.80675875320574270758_t_real*ylm(40)&
         &      -37.58068243536421341488_t_real*ylm(42)
    qvylm(1,76)=25.44224228144391182149_t_real*ylm(44)&
         &      -26.81847813040048755100_t_real*ylm(46)&
         &      +3.13172353628035111779_t_real*ylm(48)
    qvylm(2,76)=3.13172353628035111779_t_real*ylm(38)&
         &      -25.44224228144391182149_t_real*ylm(42)
    qvylm(3,76)=40.22771719560073132796_t_real*ylm(45)&
         &      -14.68908542997932084164_t_real*ylm(47)
    qvylm(4,76)=-26.81847813040048755100_t_real*ylm(46)&
         &      -3.13172353628035111779_t_real*ylm(48)&
         &      -25.44224228144391182149_t_real*ylm(44)
    qvylm(5,76)=-14.68908542997932084164_t_real*ylm(39)&
         &      -40.22771719560073132796_t_real*ylm(41)
    qvylm(1,77)=31.16025575090067293966_t_real*ylm(45)&
         &      -22.75623329653116923637_t_real*ylm(47)&
         &      +1.40054934277177879689_t_real*ylm(49)
    qvylm(2,77)=1.40054934277177879689_t_real*ylm(37)&
         &      -0.00000000000000000017_t_real*ylm(39)&
         &      -31.16025575090067293966_t_real*ylm(41)
    qvylm(3,77)=41.54700766786756391874_t_real*ylm(46)&
         &      -9.70329048075167920614_t_real*ylm(48)
    qvylm(4,77)=-22.75623329653116923637_t_real*ylm(47)&
         &      -1.40054934277177879689_t_real*ylm(49)&
         &      -31.16025575090067293966_t_real*ylm(45)
    qvylm(5,77)=-9.70329048075167920614_t_real*ylm(38)&
         &      -41.54700766786756391874_t_real*ylm(40)
    qvylm(1,78)=37.44996662214800958436_t_real*ylm(46)&
         &      -17.49285568453590140575_t_real*ylm(48)
    qvylm(2,78)=-0.00000000000000000017_t_real*ylm(38)&
         &      -37.44996662214800958436_t_real*ylm(40)
    qvylm(3,78)=41.02438299353202945970_t_real*ylm(47)&
         &      -5.04975246918103897468_t_real*ylm(49)
    qvylm(4,78)=-17.49285568453590140575_t_real*ylm(48)&
         &      -37.44996662214800958436_t_real*ylm(46)
    qvylm(5,78)=-5.04975246918103897468_t_real*ylm(37)&
         &      -41.02438299353202945970_t_real*ylm(39)
    qvylm(1,79)=44.31139808220905140193_t_real*ylm(47)&
         &      -10.90871211463571440658_t_real*ylm(49)
    qvylm(2,79)= 0.00000000000000000072_t_real*ylm(37)&
         &      -44.31139808220905140193_t_real*ylm(39)
    qvylm(3,79)=37.78888725538236854780_t_real*ylm(48)
    qvylm(4,79)=-10.90871211463571440658_t_real*ylm(49)&
         &      -44.31139808220905140193_t_real*ylm(47)
    qvylm(5,79)=-37.78888725538236854780_t_real*ylm(38)
    qvylm(1,80)=51.74456493198102055052_t_real*ylm(48)
    qvylm(2,80)=-51.74456493198102055052_t_real*ylm(38)
    qvylm(3,80)=29.87473849257931160172_t_real*ylm(49)
    qvylm(4,80)=-51.74456493198102055052_t_real*ylm(48)
    qvylm(5,80)=-29.87473849257931160172_t_real*ylm(37)
    qvylm(1,81)=59.74947698515862320431_t_real*ylm(49)
    qvylm(2,81)=-59.74947698515862320431_t_real*ylm(37)
    qvylm(3,81)= 0.0_t_real
    qvylm(4,81)=-59.74947698515862320431_t_real*ylm(49)
    qvylm(5,81)= 0.0_t_real
    !!
    If(maxL<=8) Goto 2000
    !!
    !!..............................................
    !! L=9
    !!..............................................
    !! I1
    qvylm(1,82)=76.24959016283300651113_t_real*ylm(50)
    qvylm(2,82)=76.24959016283300651113_t_real*ylm(64)
    qvylm(3,82)= 0.0_t_real
    qvylm(4,82)=-76.24959016283300651113_t_real*ylm(50)
    qvylm(5,82)= 0.0_t_real
    qvylm(1,83)=67.24581771381771216807_t_real*ylm(51)
    qvylm(2,83)=67.24581771381771216807_t_real*ylm(63)
    qvylm(3,83)=35.94440151122285690351_t_real*ylm(50)
    qvylm(4,83)=-67.24581771381771216807_t_real*ylm(51)
    qvylm(5,83)=35.94440151122285690351_t_real*ylm(64)
    qvylm(1,84)=-12.32882800593795289499_t_real*ylm(50)&
         &      +58.80476171195662619095_t_real*ylm(52)
    qvylm(2,84)=58.80476171195662619095_t_real*ylm(62)&
         &      -0.00000000000000000019_t_real*ylm(64)
    qvylm(3,84)=46.13025037868318354254_t_real*ylm(51)
    qvylm(4,84)=-12.32882800593795289499_t_real*ylm(50)&
         &      -58.80476171195662619095_t_real*ylm(52)
    qvylm(5,84)=46.13025037868318354254_t_real*ylm(63)
    qvylm(1,85)=-19.97498435543817890876_t_real*ylm(51)&
         &      +50.92641750604493481494_t_real*ylm(53)
    qvylm(2,85)=50.92641750604493481494_t_real*ylm(61)&
         &      -0.00000000000000000072_t_real*ylm(63)
    qvylm(3,85)=-5.33853912601565560303_t_real*ylm(50)&
         &      +50.92641750604493481544_t_real*ylm(52)
    qvylm(4,85)=-19.97498435543817890876_t_real*ylm(51)&
         &      -50.92641750604493481494_t_real*ylm(53)
    qvylm(5,85)=50.92641750604493481544_t_real*ylm(62)&
         &      +5.33853912601565560303_t_real*ylm(64)
    qvylm(1,86)=1.37840487520902217618_t_real*ylm(50)&
         &      -26.29828891772238476802_t_real*ylm(52)&
         &      +43.61077848422336978681_t_real*ylm(54)
    qvylm(2,86)=43.61077848422336978681_t_real*ylm(60)&
         &      -0.00000000000000000004_t_real*ylm(62)&
         &      -1.37840487520902217618_t_real*ylm(64)
    qvylm(3,86)=-10.31503756658210139816_t_real*ylm(51)&
         &      +52.59657783544476953604_t_real*ylm(53)
    qvylm(4,86)=-26.29828891772238476802_t_real*ylm(52)&
         &      -43.61077848422336978681_t_real*ylm(54)&
         &      -1.37840487520902217618_t_real*ylm(50)
    qvylm(5,86)=52.59657783544476953604_t_real*ylm(61)&
         &      +10.31503756658210139816_t_real*ylm(63)
    qvylm(1,87)=3.08220700148448822408_t_real*ylm(51)&
         &      -31.43246729100342194956_t_real*ylm(53)&
         &      +36.85783498796422794487_t_real*ylm(55)
    qvylm(2,87)=36.85783498796422794487_t_real*ylm(59)&
         &      +0.00000000000000000008_t_real*ylm(61)&
         &      -3.08220700148448822408_t_real*ylm(63)
    qvylm(3,87)=-15.71623364550171097514_t_real*ylm(52)&
         &      +52.12485011968859367422_t_real*ylm(54)
    qvylm(4,87)=-31.43246729100342194956_t_real*ylm(53)&
         &      -36.85783498796422794487_t_real*ylm(55)&
         &      -3.08220700148448822408_t_real*ylm(51)
    qvylm(5,87)=52.12485011968859367422_t_real*ylm(60)&
         &      +15.71623364550171097514_t_real*ylm(62)
    qvylm(1,88)=5.33853912601565560353_t_real*ylm(52)&
         &      -35.41186241925154660577_t_real*ylm(54)&
         &      +30.66757245039130904362_t_real*ylm(56)
    qvylm(2,88)=30.66757245039130904362_t_real*ylm(58)&
         &      -0.00000000000000000017_t_real*ylm(60)&
         &      -5.33853912601565560353_t_real*ylm(62)
    qvylm(3,88)=-21.35415650406262241453_t_real*ylm(53)&
         &      +50.07993610219565763861_t_real*ylm(55)
    qvylm(4,88)=-35.41186241925154660577_t_real*ylm(54)&
         &      -30.66757245039130904362_t_real*ylm(56)&
         &      -5.33853912601565560353_t_real*ylm(52)
    qvylm(5,88)=50.07993610219565763861_t_real*ylm(59)&
         &      +21.35415650406262241453_t_real*ylm(61)
    qvylm(1,89)=8.15475321515004523523_t_real*ylm(53)&
         &      -38.24918299781055452921_t_real*ylm(55)
    qvylm(2,89)=35.41186241925154660524_t_real*ylm(57)&
         &      +0.00000000000000000035_t_real*ylm(59)&
         &      -8.15475321515004523523_t_real*ylm(61)
    qvylm(3,89)=-27.04625667259704165949_t_real*ylm(54)&
         &      +46.84549071148684208854_t_real*ylm(56)
    qvylm(4,89)=-38.24918299781055452921_t_real*ylm(55)&
         &      -8.15475321515004523523_t_real*ylm(53)
    qvylm(5,89)=46.84549071148684208854_t_real*ylm(58)&
         &      +27.04625667259704165949_t_real*ylm(60)
    qvylm(1,90)=11.53256259467079588562_t_real*ylm(54)&
         &      -19.97498435543817890950_t_real*ylm(56)
    qvylm(2,90)=-19.97498435543817890868_t_real*ylm(58)&
         &      -11.53256259467079588562_t_real*ylm(60)
    qvylm(3,90)=-32.61901286060018094179_t_real*ylm(55)
    qvylm(4,90)=-59.92495306631453672768_t_real*ylm(56)&
         &      -11.53256259467079588562_t_real*ylm(54)
    qvylm(5,90)=60.39867548216599755752_t_real*ylm(57)&
         &      +32.61901286060018094179_t_real*ylm(59)
    qvylm(1,91)=-40.51666323872191722518_t_real*ylm(57)&
         &      +21.88149903457256284139_t_real*ylm(59)
    qvylm(2,91)=21.88149903457256284139_t_real*ylm(55)
    qvylm(3,91)=-53.59850744190550683512_t_real*ylm(58)
    qvylm(4,91)=-40.51666323872191722518_t_real*ylm(57)&
         &      -21.88149903457256284139_t_real*ylm(59)
    qvylm(5,91)=-53.59850744190550683512_t_real*ylm(56)
    qvylm(1,92)=-59.92495306631453672768_t_real*ylm(58)&
         &      +11.53256259467079588562_t_real*ylm(60)
    qvylm(2,92)=11.53256259467079588562_t_real*ylm(54)&
         &      -19.97498435543817890950_t_real*ylm(56)
    qvylm(3,92)=60.39867548216599755752_t_real*ylm(57)&
         &      -32.61901286060018094179_t_real*ylm(59)
    qvylm(4,92)=-19.97498435543817890950_t_real*ylm(58)&
         &      -11.53256259467079588562_t_real*ylm(60)
    qvylm(5,92)=-32.61901286060018094179_t_real*ylm(55)
    qvylm(1,93)=35.41186241925154660524_t_real*ylm(57)&
         &      -38.24918299781055452921_t_real*ylm(59)&
         &      +8.15475321515004523523_t_real*ylm(61)
    qvylm(2,93)=8.15475321515004523523_t_real*ylm(53)&
         &      -0.00000000000000000035_t_real*ylm(55)
    qvylm(3,93)=46.84549071148684208854_t_real*ylm(58)&
         &      -27.04625667259704165949_t_real*ylm(60)
    qvylm(4,93)=-38.24918299781055452921_t_real*ylm(59)&
         &      -8.15475321515004523523_t_real*ylm(61)&
         &      -35.41186241925154660524_t_real*ylm(57)
    qvylm(5,93)=-27.04625667259704165949_t_real*ylm(54)&
         &      -46.84549071148684208854_t_real*ylm(56)
    qvylm(1,94)=30.66757245039130904362_t_real*ylm(58)&
         &      -35.41186241925154660577_t_real*ylm(60)&
         &      +5.33853912601565560353_t_real*ylm(62)
    qvylm(2,94)=5.33853912601565560353_t_real*ylm(52)&
         &      +0.00000000000000000017_t_real*ylm(54)&
         &      -30.66757245039130904362_t_real*ylm(56)
    qvylm(3,94)=50.07993610219565763861_t_real*ylm(59)&
         &      -21.35415650406262241453_t_real*ylm(61)
    qvylm(4,94)=-35.41186241925154660577_t_real*ylm(60)&
         &      -5.33853912601565560353_t_real*ylm(62)&
         &      -30.66757245039130904362_t_real*ylm(58)
    qvylm(5,94)=-21.35415650406262241453_t_real*ylm(53)&
         &      -50.07993610219565763861_t_real*ylm(55)
    qvylm(1,95)=36.85783498796422794487_t_real*ylm(59)&
         &      -31.43246729100342194956_t_real*ylm(61)&
         &      +3.08220700148448822408_t_real*ylm(63)
    qvylm(2,95)=3.08220700148448822408_t_real*ylm(51)&
         &      -0.00000000000000000008_t_real*ylm(53)&
         &      -36.85783498796422794487_t_real*ylm(55)
    qvylm(3,95)=52.12485011968859367422_t_real*ylm(60)&
         &      -15.71623364550171097514_t_real*ylm(62)
    qvylm(4,95)=-31.43246729100342194956_t_real*ylm(61)&
         &      -3.08220700148448822408_t_real*ylm(63)&
         &      -36.85783498796422794487_t_real*ylm(59)
    qvylm(5,95)=-15.71623364550171097514_t_real*ylm(52)&
         &      -52.12485011968859367422_t_real*ylm(54)
    qvylm(1,96)=43.61077848422336978681_t_real*ylm(60)&
         &      -26.29828891772238476802_t_real*ylm(62)&
         &      +1.37840487520902217618_t_real*ylm(64)
    qvylm(2,96)=1.37840487520902217618_t_real*ylm(50)&
         &      +0.00000000000000000004_t_real*ylm(52)&
         &      -43.61077848422336978681_t_real*ylm(54)
    qvylm(3,96)=52.59657783544476953604_t_real*ylm(61)&
         &      -10.31503756658210139816_t_real*ylm(63)
    qvylm(4,96)=-26.29828891772238476802_t_real*ylm(62)&
         &      -1.37840487520902217618_t_real*ylm(64)&
         &      -43.61077848422336978681_t_real*ylm(60)
    qvylm(5,96)=-10.31503756658210139816_t_real*ylm(51)&
         &      -52.59657783544476953604_t_real*ylm(53)
    qvylm(1,97)=50.92641750604493481494_t_real*ylm(61)&
         &      -19.97498435543817890876_t_real*ylm(63)
    qvylm(2,97)= 0.00000000000000000072_t_real*ylm(51)&
         &      -50.92641750604493481494_t_real*ylm(53)
    qvylm(3,97)=50.92641750604493481544_t_real*ylm(62)&
         &      -5.33853912601565560303_t_real*ylm(64)
    qvylm(4,97)=-19.97498435543817890876_t_real*ylm(63)&
         &      -50.92641750604493481494_t_real*ylm(61)
    qvylm(5,97)=-5.33853912601565560303_t_real*ylm(50)&
         &      -50.92641750604493481544_t_real*ylm(52)
    qvylm(1,98)=58.80476171195662619095_t_real*ylm(62)&
         &      -12.32882800593795289499_t_real*ylm(64)
    qvylm(2,98)= 0.00000000000000000019_t_real*ylm(50)&
         &      -58.80476171195662619095_t_real*ylm(52)
    qvylm(3,98)=46.13025037868318354254_t_real*ylm(63)
    qvylm(4,98)=-12.32882800593795289499_t_real*ylm(64)&
         &      -58.80476171195662619095_t_real*ylm(62)
    qvylm(5,98)=-46.13025037868318354254_t_real*ylm(51)
    qvylm(1,99)=67.24581771381771216807_t_real*ylm(63)
    qvylm(2,99)=-67.24581771381771216807_t_real*ylm(51)
    qvylm(3,99)=35.94440151122285690351_t_real*ylm(64)
    qvylm(4,99)=-67.24581771381771216807_t_real*ylm(63)
    qvylm(5,99)=-35.94440151122285690351_t_real*ylm(50)
    qvylm(1,100)=76.24959016283300651113_t_real*ylm(64)
    qvylm(2,100)=-76.24959016283300651113_t_real*ylm(50)
    qvylm(3,100)= 0.0_t_real
    qvylm(4,100)=-76.24959016283300651113_t_real*ylm(64)
    qvylm(5,100)= 0.0_t_real
    !!
    If(maxL<=9) Goto 2000
    !!
    !!..............................................
    !! L=10
    !!..............................................
    !! I1
    qvylm(1,101)=94.74967018412254079073_t_real*ylm(65)
    qvylm(2,101)=94.74967018412254079073_t_real*ylm(81)
    qvylm(3,101)= 0.0_t_real
    qvylm(4,101)=-94.74967018412254079073_t_real*ylm(65)
    qvylm(5,101)= 0.0_t_real
    qvylm(1,102)=84.74668135095320648067_t_real*ylm(66)
    qvylm(2,102)=84.74668135095320648067_t_real*ylm(80)
    qvylm(3,102)=42.37334067547660324043_t_real*ylm(65)
    qvylm(4,102)=-84.74668135095320648067_t_real*ylm(66)
    qvylm(5,102)=42.37334067547660324043_t_real*ylm(81)
    qvylm(1,103)=-13.74772708486752001220_t_real*ylm(65)&
         &      +75.29940238806679928942_t_real*ylm(67)
    qvylm(2,103)=75.29940238806679928942_t_real*ylm(79)&
         &      -0.00000000000000000016_t_real*ylm(81)
    qvylm(3,103)=54.99090833947008005791_t_real*ylm(66)
    qvylm(4,103)=-13.74772708486752001220_t_real*ylm(65)&
         &      -75.29940238806679928942_t_real*ylm(67)
    qvylm(5,103)=54.99090833947008005791_t_real*ylm(80)
    qvylm(1,104)=-22.44994432064364830371_t_real*ylm(66)&
         &      +66.40783086353596594501_t_real*ylm(68)
    qvylm(2,104)=66.40783086353596594501_t_real*ylm(78)&
         &      +0.00000000000000000079_t_real*ylm(80)
    qvylm(3,104)=-5.61248608016091207533_t_real*ylm(65)&
         &      +61.48170459575759027572_t_real*ylm(67)
    qvylm(4,104)=-22.44994432064364830371_t_real*ylm(66)&
         &      -66.40783086353596594501_t_real*ylm(68)
    qvylm(5,104)=61.48170459575759027572_t_real*ylm(79)&
         &      +5.61248608016091207533_t_real*ylm(81)
    qvylm(1,105)=1.36122781945954521877_t_real*ylm(65)&
         &      -29.82300730486250530052_t_real*ylm(67)&
         &      +58.07196346927207683167_t_real*ylm(69)
    qvylm(2,105)=58.07196346927207683167_t_real*ylm(77)&
         &      -0.00000000000000000096_t_real*ylm(79)&
         &      -1.36122781945954521877_t_real*ylm(81)
    qvylm(3,105)=-10.88982255567636175201_t_real*ylm(66)&
         &      +64.42505906317911668353_t_real*ylm(68)
    qvylm(4,105)=-29.82300730486250530052_t_real*ylm(67)&
         &      -58.07196346927207683167_t_real*ylm(69)&
         &      -1.36122781945954521877_t_real*ylm(65)
    qvylm(5,105)=64.42505906317911668353_t_real*ylm(78)&
         &      +10.88982255567636175201_t_real*ylm(80)
    qvylm(1,106)=3.04379793717535414962_t_real*ylm(66)&
         &      -36.01470287992685583343_t_real*ylm(68)&
         &      +50.29179561203152089220_t_real*ylm(70)
    qvylm(2,106)=50.29179561203152089220_t_real*ylm(76)&
         &      -0.00000000000000000047_t_real*ylm(78)&
         &      -3.04379793717535414962_t_real*ylm(80)
    qvylm(3,106)=-16.67156790678633906496_t_real*ylm(67)&
         &      +64.92642895208844174819_t_real*ylm(69)
    qvylm(4,106)=-36.01470287992685583343_t_real*ylm(68)&
         &      -50.29179561203152089220_t_real*ylm(70)&
         &      -3.04379793717535414962_t_real*ylm(66)
    qvylm(5,106)=64.92642895208844174819_t_real*ylm(77)&
         &      +16.67156790678633906496_t_real*ylm(79)
    qvylm(1,107)=5.27201267516105498843_t_real*ylm(67)&
         &      -41.06307916593975168274_t_real*ylm(69)&
         &      +43.06732076234901824502_t_real*ylm(71)
    qvylm(2,107)=43.06732076234901824502_t_real*ylm(75)&
         &      +0.00000000000000000024_t_real*ylm(77)&
         &      -5.27201267516105498843_t_real*ylm(79)
    qvylm(3,107)=-22.77769807095889805037_t_real*ylm(68)&
         &      +63.61464870147256187402_t_real*ylm(70)
    qvylm(4,107)=-41.06307916593975168274_t_real*ylm(69)&
         &      -43.06732076234901824502_t_real*ylm(71)&
         &      -5.27201267516105498843_t_real*ylm(67)
    qvylm(5,107)=63.61464870147256187402_t_real*ylm(76)&
         &      +22.77769807095889805037_t_real*ylm(78)
    qvylm(1,108)=8.05313238289738958523_t_real*ylm(68)&
         &      -44.98234947961124841740_t_real*ylm(70)&
         &      +36.39852938205705292602_t_real*ylm(72)
    qvylm(2,108)=36.39852938205705292602_t_real*ylm(74)&
         &      +0.00000000000000000028_t_real*ylm(76)&
         &      -8.05313238289738958523_t_real*ylm(78)
    qvylm(3,108)=-29.03598173463603841471_t_real*ylm(69)&
         &      +60.90638911718636472081_t_real*ylm(71)
    qvylm(4,108)=-44.98234947961124841740_t_real*ylm(70)&
         &      -36.39852938205705292602_t_real*ylm(72)&
         &      -8.05313238289738958523_t_real*ylm(68)
    qvylm(5,108)=60.90638911718636472081_t_real*ylm(75)&
         &      +29.03598173463603841471_t_real*ylm(77)
    qvylm(1,109)=11.38884903547944902497_t_real*ylm(69)&
         &      -47.77890255553976936732_t_real*ylm(71)
    qvylm(2,109)=42.83003344273597869329_t_real*ylm(73)&
         &      -0.00000000000000000046_t_real*ylm(75)&
         &      -11.38884903547944902497_t_real*ylm(77)
    qvylm(3,109)=-35.28705811750584703457_t_real*ylm(70)&
         &      +57.10671125698130492396_t_real*ylm(72)
    qvylm(4,109)=-47.77890255553976936732_t_real*ylm(71)&
         &      -11.38884903547944902497_t_real*ylm(69)
    qvylm(5,109)=57.10671125698130492396_t_real*ylm(74)&
         &      +35.28705811750584703457_t_real*ylm(76)
    qvylm(1,110)=15.27974437728897732551_t_real*ylm(70)&
         &      -24.72793133756429122996_t_real*ylm(72)
    qvylm(2,110)=-24.72793133756429122904_t_real*ylm(74)&
         &      -15.27974437728897732551_t_real*ylm(76)
    qvylm(3,110)=-41.37774337803867631414_t_real*ylm(71)
    qvylm(4,110)=-74.18379401269287368896_t_real*ylm(72)&
         &      -15.27974437728897732551_t_real*ylm(70)
    qvylm(5,110)=74.18379401269287369089_t_real*ylm(73)&
         &      +41.37774337803867631414_t_real*ylm(75)
    qvylm(1,111)=-50.01470372035901719390_t_real*ylm(73)&
         &      +27.89686889451300156772_t_real*ylm(75)
    qvylm(2,111)=27.89686889451300156772_t_real*ylm(71)
    qvylm(3,111)=-66.68627162714535625968_t_real*ylm(74)
    qvylm(4,111)=-50.01470372035901719390_t_real*ylm(73)&
         &      -27.89686889451300156772_t_real*ylm(75)
    qvylm(5,111)=-66.68627162714535625968_t_real*ylm(72)
    qvylm(1,112)=-74.18379401269287368896_t_real*ylm(74)&
         &      +15.27974437728897732551_t_real*ylm(76)
    qvylm(2,112)=15.27974437728897732551_t_real*ylm(70)&
         &      -24.72793133756429122996_t_real*ylm(72)
    qvylm(3,112)=74.18379401269287369089_t_real*ylm(73)&
         &      -41.37774337803867631414_t_real*ylm(75)
    qvylm(4,112)=-24.72793133756429122996_t_real*ylm(74)&
         &      -15.27974437728897732551_t_real*ylm(76)
    qvylm(5,112)=-41.37774337803867631414_t_real*ylm(71)
    qvylm(1,113)=42.83003344273597869329_t_real*ylm(73)&
         &      -47.77890255553976936732_t_real*ylm(75)&
         &      +11.38884903547944902497_t_real*ylm(77)
    qvylm(2,113)=11.38884903547944902497_t_real*ylm(69)&
         &      +0.00000000000000000046_t_real*ylm(71)
    qvylm(3,113)=57.10671125698130492396_t_real*ylm(74)&
         &      -35.28705811750584703457_t_real*ylm(76)
    qvylm(4,113)=-47.77890255553976936732_t_real*ylm(75)&
         &      -11.38884903547944902497_t_real*ylm(77)&
         &      -42.83003344273597869329_t_real*ylm(73)
    qvylm(5,113)=-35.28705811750584703457_t_real*ylm(70)&
         &      -57.10671125698130492396_t_real*ylm(72)
    qvylm(1,114)=36.39852938205705292602_t_real*ylm(74)&
         &      -44.98234947961124841740_t_real*ylm(76)&
         &      +8.05313238289738958523_t_real*ylm(78)
    qvylm(2,114)=8.05313238289738958523_t_real*ylm(68)&
         &      -0.00000000000000000028_t_real*ylm(70)&
         &      -36.39852938205705292602_t_real*ylm(72)
    qvylm(3,114)=60.90638911718636472081_t_real*ylm(75)&
         &      -29.03598173463603841471_t_real*ylm(77)
    qvylm(4,114)=-44.98234947961124841740_t_real*ylm(76)&
         &      -8.05313238289738958523_t_real*ylm(78)&
         &      -36.39852938205705292602_t_real*ylm(74)
    qvylm(5,114)=-29.03598173463603841471_t_real*ylm(69)&
         &      -60.90638911718636472081_t_real*ylm(71)
    qvylm(1,115)=43.06732076234901824502_t_real*ylm(75)&
         &      -41.06307916593975168274_t_real*ylm(77)&
         &      +5.27201267516105498843_t_real*ylm(79)
    qvylm(2,115)=5.27201267516105498843_t_real*ylm(67)&
         &      -0.00000000000000000024_t_real*ylm(69)&
         &      -43.06732076234901824502_t_real*ylm(71)
    qvylm(3,115)=63.61464870147256187402_t_real*ylm(76)&
         &      -22.77769807095889805037_t_real*ylm(78)
    qvylm(4,115)=-41.06307916593975168274_t_real*ylm(77)&
         &      -5.27201267516105498843_t_real*ylm(79)&
         &      -43.06732076234901824502_t_real*ylm(75)
    qvylm(5,115)=-22.77769807095889805037_t_real*ylm(68)&
         &      -63.61464870147256187402_t_real*ylm(70)
    qvylm(1,116)=50.29179561203152089220_t_real*ylm(76)&
         &      -36.01470287992685583343_t_real*ylm(78)&
         &      +3.04379793717535414962_t_real*ylm(80)
    qvylm(2,116)=3.04379793717535414962_t_real*ylm(66)&
         &      +0.00000000000000000047_t_real*ylm(68)&
         &      -50.29179561203152089220_t_real*ylm(70)
    qvylm(3,116)=64.92642895208844174819_t_real*ylm(77)&
         &      -16.67156790678633906496_t_real*ylm(79)
    qvylm(4,116)=-36.01470287992685583343_t_real*ylm(78)&
         &      -3.04379793717535414962_t_real*ylm(80)&
         &      -50.29179561203152089220_t_real*ylm(76)
    qvylm(5,116)=-16.67156790678633906496_t_real*ylm(67)&
         &      -64.92642895208844174819_t_real*ylm(69)
    qvylm(1,117)=58.07196346927207683167_t_real*ylm(77)&
         &      -29.82300730486250530052_t_real*ylm(79)&
         &      +1.36122781945954521877_t_real*ylm(81)
    qvylm(2,117)=1.36122781945954521877_t_real*ylm(65)&
         &      +0.00000000000000000096_t_real*ylm(67)&
         &      -58.07196346927207683167_t_real*ylm(69)
    qvylm(3,117)=64.42505906317911668353_t_real*ylm(78)&
         &      -10.88982255567636175201_t_real*ylm(80)
    qvylm(4,117)=-29.82300730486250530052_t_real*ylm(79)&
         &      -1.36122781945954521877_t_real*ylm(81)&
         &      -58.07196346927207683167_t_real*ylm(77)
    qvylm(5,117)=-10.88982255567636175201_t_real*ylm(66)&
         &      -64.42505906317911668353_t_real*ylm(68)
    qvylm(1,118)=66.40783086353596594501_t_real*ylm(78)&
         &      -22.44994432064364830371_t_real*ylm(80)
    qvylm(2,118)=-0.00000000000000000079_t_real*ylm(66)&
         &      -66.40783086353596594501_t_real*ylm(68)
    qvylm(3,118)=61.48170459575759027572_t_real*ylm(79)&
         &      -5.61248608016091207533_t_real*ylm(81)
    qvylm(4,118)=-22.44994432064364830371_t_real*ylm(80)&
         &      -66.40783086353596594501_t_real*ylm(78)
    qvylm(5,118)=-5.61248608016091207533_t_real*ylm(65)&
         &      -61.48170459575759027572_t_real*ylm(67)
    qvylm(1,119)=75.29940238806679928942_t_real*ylm(79)&
         &      -13.74772708486752001220_t_real*ylm(81)
    qvylm(2,119)= 0.00000000000000000016_t_real*ylm(65)&
         &      -75.29940238806679928942_t_real*ylm(67)
    qvylm(3,119)=54.99090833947008005791_t_real*ylm(80)
    qvylm(4,119)=-13.74772708486752001220_t_real*ylm(81)&
         &      -75.29940238806679928942_t_real*ylm(79)
    qvylm(5,119)=-54.99090833947008005791_t_real*ylm(66)
    qvylm(1,120)=84.74668135095320648067_t_real*ylm(80)
    qvylm(2,120)=-84.74668135095320648067_t_real*ylm(66)
    qvylm(3,120)=42.37334067547660324043_t_real*ylm(81)
    qvylm(4,120)=-84.74668135095320648067_t_real*ylm(80)
    qvylm(5,120)=-42.37334067547660324043_t_real*ylm(65)
    qvylm(1,121)=94.74967018412254079073_t_real*ylm(81)
    qvylm(2,121)=-94.74967018412254079073_t_real*ylm(65)
    qvylm(3,121)= 0.0_t_real
    qvylm(4,121)=-94.74967018412254079073_t_real*ylm(81)
    qvylm(5,121)= 0.0_t_real
    !!
    If(maxL<=10) Goto 2000
    !!
    !!..............................................
    !! L=11
    !!..............................................
    !! I1
    qvylm(1,122)=115.24972885000641069583_t_real*ylm(82)
    qvylm(2,122)=115.24972885000641069583_t_real*ylm(100)
    qvylm(3,122)= 0.0_t_real
    qvylm(4,122)=-115.24972885000641069583_t_real*ylm(82)
    qvylm(5,122)= 0.0_t_real
    qvylm(1,123)=104.24730212336432443972_t_real*ylm(83)
    qvylm(2,123)=104.24730212336432443972_t_real*ylm(99)
    qvylm(3,123)=49.14264950122245898790_t_real*ylm(82)
    qvylm(4,123)=-104.24730212336432443972_t_real*ylm(83)
    qvylm(5,123)=49.14264950122245898790_t_real*ylm(100)
    qvylm(1,124)=-15.16575088810310109793_t_real*ylm(82)&
         &      +93.79498920518089163363_t_real*ylm(84)
    qvylm(2,124)=93.79498920518089163363_t_real*ylm(98)&
         &      +0.00000000000000000099_t_real*ylm(100)
    qvylm(3,124)=64.34283176858164979025_t_real*ylm(83)
    qvylm(4,124)=-15.16575088810310109793_t_real*ylm(82)&
         &      -93.79498920518089163363_t_real*ylm(84)
    qvylm(5,124)=64.34283176858164979025_t_real*ylm(99)
    qvylm(1,125)=-24.91987158875422454264_t_real*ylm(83)&
         &      +83.89278872465737738116_t_real*ylm(85)
    qvylm(2,125)=83.89278872465737738116_t_real*ylm(97)&
         &      -0.00000000000000000030_t_real*ylm(99)
    qvylm(3,125)=-5.87367006223536528902_t_real*ylm(82)&
         &      +72.65328622987400694438_t_real*ylm(84)
    qvylm(4,125)=-24.91987158875422454264_t_real*ylm(83)&
         &      -83.89278872465737738116_t_real*ylm(85)
    qvylm(5,125)=72.65328622987400694438_t_real*ylm(98)&
         &      +5.87367006223536528902_t_real*ylm(100)
    qvylm(1,126)=1.34751232784127452799_t_real*ylm(82)&
         &      -33.33561395706904991813_t_real*ylm(84)&
         &      +74.54069887969715168305_t_real*ylm(86)
    qvylm(2,126)=74.54069887969715168305_t_real*ylm(96)&
         &      +0.00000000000000000085_t_real*ylm(98)&
         &      -1.34751232784127452799_t_real*ylm(100)
    qvylm(3,126)=-11.43402125698842479371_t_real*ylm(83)&
         &      +76.98530276686104681911_t_real*ylm(85)
    qvylm(4,126)=-33.33561395706904991813_t_real*ylm(84)&
         &      -74.54069887969715168305_t_real*ylm(86)&
         &      -1.34751232784127452799_t_real*ylm(82)
    qvylm(5,126)=76.98530276686104681911_t_real*ylm(97)&
         &      +11.43402125698842479371_t_real*ylm(99)
    qvylm(1,127)=3.01312916557207228875_t_real*ylm(83)&
         &      -40.57481718349060175463_t_real*ylm(85)&
         &      +65.73871726287653270130_t_real*ylm(87)
    qvylm(2,127)=65.73871726287653270130_t_real*ylm(95)&
         &      +0.00000000000000000107_t_real*ylm(97)&
         &      -3.01312916557207228875_t_real*ylm(99)
    qvylm(3,127)=-17.56941121740611400695_t_real*ylm(84)&
         &      +78.57279561353481447309_t_real*ylm(86)
    qvylm(4,127)=-40.57481718349060175463_t_real*ylm(85)&
         &      -65.73871726287653270130_t_real*ylm(87)&
         &      -3.01312916557207228875_t_real*ylm(83)
    qvylm(5,127)=78.57279561353481447309_t_real*ylm(96)&
         &      +17.56941121740611400695_t_real*ylm(98)
    qvylm(1,128)=5.21889280453844518865_t_real*ylm(84)&
         &      -46.67919631292994561215_t_real*ylm(86)&
         &      +57.48684059943860886008_t_real*ylm(88)
    qvylm(2,128)=57.48684059943860886008_t_real*ylm(94)&
         &      -0.00000000000000000005_t_real*ylm(96)&
         &      -5.21889280453844518865_t_real*ylm(98)
    qvylm(3,128)=-24.10503332457657830862_t_real*ylm(85)&
         &      +78.10923525153057956444_t_real*ylm(87)
    qvylm(4,128)=-46.67919631292994561215_t_real*ylm(86)&
         &      -57.48684059943860886008_t_real*ylm(88)&
         &      -5.21889280453844518865_t_real*ylm(84)
    qvylm(5,128)=78.10923525153057956444_t_real*ylm(95)&
         &      +24.10503332457657830862_t_real*ylm(97)
    qvylm(1,129)=7.97199044021926611433_t_real*ylm(85)&
         &      -51.66440289324739181647_t_real*ylm(87)&
         &      +49.78506434242048225700_t_real*ylm(89)
    qvylm(2,129)=49.78506434242048225700_t_real*ylm(93)&
         &      -0.00000000000000000095_t_real*ylm(95)&
         &      -7.97199044021926611433_t_real*ylm(97)
    qvylm(3,129)=-30.87538621109395167707_t_real*ylm(86)&
         &      +76.04794194246291711776_t_real*ylm(88)
    qvylm(4,129)=-51.66440289324739181647_t_real*ylm(87)&
         &      -49.78506434242048225700_t_real*ylm(89)&
         &      -7.97199044021926611433_t_real*ylm(85)
    qvylm(5,129)=76.04794194246291711776_t_real*ylm(94)&
         &      +30.87538621109395167707_t_real*ylm(96)
    qvylm(1,130)=11.27409699966674655537_t_real*ylm(86)&
         &      -55.53756433830690547727_t_real*ylm(88)&
         &      +42.63338202814661447575_t_real*ylm(90)
    qvylm(2,130)=42.63338202814661447575_t_real*ylm(92)&
         &      -0.00000000000000000035_t_real*ylm(94)&
         &      -11.27409699966674655537_t_real*ylm(96)
    qvylm(3,130)=-37.73034517955568679187_t_real*ylm(87)&
         &      +72.71574071250612686253_t_real*ylm(89)
    qvylm(4,130)=-55.53756433830690547727_t_real*ylm(88)&
         &      -42.63338202814661447575_t_real*ylm(90)&
         &      -11.27409699966674655537_t_real*ylm(86)
    qvylm(5,130)=72.71574071250612686253_t_real*ylm(93)&
         &      +37.73034517955568679187_t_real*ylm(95)
    qvylm(1,131)=15.12578836570876149747_t_real*ylm(87)&
         &      -58.30229750821867998555_t_real*ylm(89)
    qvylm(2,131)=50.95663791272360866730_t_real*ylm(91)&
         &      -0.00000000000000000001_t_real*ylm(93)&
         &      -15.12578836570876149747_t_real*ylm(95)
    qvylm(3,131)=-44.52911525354122609508_t_real*ylm(88)&
         &      +68.36550376661579087251_t_real*ylm(90)
    qvylm(4,131)=-58.30229750821867998555_t_real*ylm(89)&
         &      -15.12578836570876149747_t_real*ylm(87)
    qvylm(5,131)=68.36550376661579087251_t_real*ylm(92)&
         &      +44.52911525354122609508_t_real*ylm(94)
    qvylm(1,132)=19.52730881288264489104_t_real*ylm(88)&
         &      -29.98025666123764540595_t_real*ylm(90)
    qvylm(2,132)=-29.98025666123764540403_t_real*ylm(92)&
         &      -19.52730881288264489104_t_real*ylm(94)
    qvylm(3,132)=-51.13449757360742030775_t_real*ylm(89)
    qvylm(4,132)=-89.94076998371293621593_t_real*ylm(90)&
         &      -19.52730881288264489104_t_real*ylm(88)
    qvylm(5,132)=89.38385583655767927164_t_real*ylm(91)&
         &      +51.13449757360742030775_t_real*ylm(93)
    qvylm(1,133)=-60.51315646421989356064_t_real*ylm(91)&
         &      +34.61821850747925107506_t_real*ylm(93)
    qvylm(2,133)=34.61821850747925107506_t_real*ylm(89)
    qvylm(3,133)=-81.18691883224590142753_t_real*ylm(92)
    qvylm(4,133)=-60.51315646421989356064_t_real*ylm(91)&
         &      -34.61821850747925107506_t_real*ylm(93)
    qvylm(5,133)=-81.18691883224590142753_t_real*ylm(90)
    qvylm(1,134)=-89.94076998371293621593_t_real*ylm(92)&
         &      +19.52730881288264489104_t_real*ylm(94)
    qvylm(2,134)=19.52730881288264489104_t_real*ylm(88)&
         &      -29.98025666123764540595_t_real*ylm(90)
    qvylm(3,134)=89.38385583655767927164_t_real*ylm(91)&
         &      -51.13449757360742030775_t_real*ylm(93)
    qvylm(4,134)=-29.98025666123764540595_t_real*ylm(92)&
         &      -19.52730881288264489104_t_real*ylm(94)
    qvylm(5,134)=-51.13449757360742030775_t_real*ylm(89)
    qvylm(1,135)=50.95663791272360866730_t_real*ylm(91)&
         &      -58.30229750821867998555_t_real*ylm(93)&
         &      +15.12578836570876149747_t_real*ylm(95)
    qvylm(2,135)=15.12578836570876149747_t_real*ylm(87)&
         &      +0.00000000000000000001_t_real*ylm(89)
    qvylm(3,135)=68.36550376661579087251_t_real*ylm(92)&
         &      -44.52911525354122609508_t_real*ylm(94)
    qvylm(4,135)=-58.30229750821867998555_t_real*ylm(93)&
         &      -15.12578836570876149747_t_real*ylm(95)&
         &      -50.95663791272360866730_t_real*ylm(91)
    qvylm(5,135)=-44.52911525354122609508_t_real*ylm(88)&
         &      -68.36550376661579087251_t_real*ylm(90)
    qvylm(1,136)=42.63338202814661447575_t_real*ylm(92)&
         &      -55.53756433830690547727_t_real*ylm(94)&
         &      +11.27409699966674655537_t_real*ylm(96)
    qvylm(2,136)=11.27409699966674655537_t_real*ylm(86)&
         &      +0.00000000000000000035_t_real*ylm(88)&
         &      -42.63338202814661447575_t_real*ylm(90)
    qvylm(3,136)=72.71574071250612686253_t_real*ylm(93)&
         &      -37.73034517955568679187_t_real*ylm(95)
    qvylm(4,136)=-55.53756433830690547727_t_real*ylm(94)&
         &      -11.27409699966674655537_t_real*ylm(96)&
         &      -42.63338202814661447575_t_real*ylm(92)
    qvylm(5,136)=-37.73034517955568679187_t_real*ylm(87)&
         &      -72.71574071250612686253_t_real*ylm(89)
    qvylm(1,137)=49.78506434242048225700_t_real*ylm(93)&
         &      -51.66440289324739181647_t_real*ylm(95)&
         &      +7.97199044021926611433_t_real*ylm(97)
    qvylm(2,137)=7.97199044021926611433_t_real*ylm(85)&
         &      +0.00000000000000000095_t_real*ylm(87)&
         &      -49.78506434242048225700_t_real*ylm(89)
    qvylm(3,137)=76.04794194246291711776_t_real*ylm(94)&
         &      -30.87538621109395167707_t_real*ylm(96)
    qvylm(4,137)=-51.66440289324739181647_t_real*ylm(95)&
         &      -7.97199044021926611433_t_real*ylm(97)&
         &      -49.78506434242048225700_t_real*ylm(93)
    qvylm(5,137)=-30.87538621109395167707_t_real*ylm(86)&
         &      -76.04794194246291711776_t_real*ylm(88)
    qvylm(1,138)=57.48684059943860886008_t_real*ylm(94)&
         &      -46.67919631292994561215_t_real*ylm(96)&
         &      +5.21889280453844518865_t_real*ylm(98)
    qvylm(2,138)=5.21889280453844518865_t_real*ylm(84)&
         &      +0.00000000000000000005_t_real*ylm(86)&
         &      -57.48684059943860886008_t_real*ylm(88)
    qvylm(3,138)=78.10923525153057956444_t_real*ylm(95)&
         &      -24.10503332457657830862_t_real*ylm(97)
    qvylm(4,138)=-46.67919631292994561215_t_real*ylm(96)&
         &      -5.21889280453844518865_t_real*ylm(98)&
         &      -57.48684059943860886008_t_real*ylm(94)
    qvylm(5,138)=-24.10503332457657830862_t_real*ylm(85)&
         &      -78.10923525153057956444_t_real*ylm(87)
    qvylm(1,139)=65.73871726287653270130_t_real*ylm(95)&
         &      -40.57481718349060175463_t_real*ylm(97)&
         &      +3.01312916557207228875_t_real*ylm(99)
    qvylm(2,139)=3.01312916557207228875_t_real*ylm(83)&
         &      -0.00000000000000000107_t_real*ylm(85)&
         &      -65.73871726287653270130_t_real*ylm(87)
    qvylm(3,139)=78.57279561353481447309_t_real*ylm(96)&
         &      -17.56941121740611400695_t_real*ylm(98)
    qvylm(4,139)=-40.57481718349060175463_t_real*ylm(97)&
         &      -3.01312916557207228875_t_real*ylm(99)&
         &      -65.73871726287653270130_t_real*ylm(95)
    qvylm(5,139)=-17.56941121740611400695_t_real*ylm(84)&
         &      -78.57279561353481447309_t_real*ylm(86)
    qvylm(1,140)=74.54069887969715168305_t_real*ylm(96)&
         &      -33.33561395706904991813_t_real*ylm(98)&
         &      +1.34751232784127452799_t_real*ylm(100)
    qvylm(2,140)=1.34751232784127452799_t_real*ylm(82)&
         &      -0.00000000000000000085_t_real*ylm(84)&
         &      -74.54069887969715168305_t_real*ylm(86)
    qvylm(3,140)=76.98530276686104681911_t_real*ylm(97)&
         &      -11.43402125698842479371_t_real*ylm(99)
    qvylm(4,140)=-33.33561395706904991813_t_real*ylm(98)&
         &      -1.34751232784127452799_t_real*ylm(100)&
         &      -74.54069887969715168305_t_real*ylm(96)
    qvylm(5,140)=-11.43402125698842479371_t_real*ylm(83)&
         &      -76.98530276686104681911_t_real*ylm(85)
    qvylm(1,141)=83.89278872465737738116_t_real*ylm(97)&
         &      -24.91987158875422454264_t_real*ylm(99)
    qvylm(2,141)= 0.00000000000000000030_t_real*ylm(83)&
         &      -83.89278872465737738116_t_real*ylm(85)
    qvylm(3,141)=72.65328622987400694438_t_real*ylm(98)&
         &      -5.87367006223536528902_t_real*ylm(100)
    qvylm(4,141)=-24.91987158875422454264_t_real*ylm(99)&
         &      -83.89278872465737738116_t_real*ylm(97)
    qvylm(5,141)=-5.87367006223536528902_t_real*ylm(82)&
         &      -72.65328622987400694438_t_real*ylm(84)
    qvylm(1,142)=93.79498920518089163363_t_real*ylm(98)&
         &      -15.16575088810310109793_t_real*ylm(100)
    qvylm(2,142)=-0.00000000000000000099_t_real*ylm(82)&
         &      -93.79498920518089163363_t_real*ylm(84)
    qvylm(3,142)=64.34283176858164979025_t_real*ylm(99)
    qvylm(4,142)=-15.16575088810310109793_t_real*ylm(100)&
         &      -93.79498920518089163363_t_real*ylm(98)
    qvylm(5,142)=-64.34283176858164979025_t_real*ylm(83)
    qvylm(1,143)=104.24730212336432443972_t_real*ylm(99)
    qvylm(2,143)=-104.24730212336432443972_t_real*ylm(83)
    qvylm(3,143)=49.14264950122245898790_t_real*ylm(100)
    qvylm(4,143)=-104.24730212336432443972_t_real*ylm(99)
    qvylm(5,143)=-49.14264950122245898790_t_real*ylm(82)
    qvylm(1,144)=115.24972885000641069583_t_real*ylm(100)
    qvylm(2,144)=-115.24972885000641069583_t_real*ylm(82)
    qvylm(3,144)= 0.0_t_real
    qvylm(4,144)=-115.24972885000641069583_t_real*ylm(100)
    qvylm(5,144)= 0.0_t_real
    !!
    If(maxL<=11) Goto 2000
    !!
    !!..............................................
    !! L=12
    !!..............................................
    !! I1
    qvylm(1,145)=137.74977313955910826876_t_real*ylm(101)
    qvylm(2,145)=137.74977313955910826876_t_real*ylm(121)
    qvylm(3,145)= 0.0_t_real
    qvylm(4,145)=-137.74977313955910826876_t_real*ylm(101)
    qvylm(5,145)= 0.0_t_real
    qvylm(1,146)=125.74776339959291693290_t_real*ylm(102)
    qvylm(2,146)=125.74776339959291693290_t_real*ylm(120)
    qvylm(3,146)=56.23610939600996257577_t_real*ylm(101)
    qvylm(4,146)=-125.74776339959291693290_t_real*ylm(102)
    qvylm(5,146)=56.23610939600996257577_t_real*ylm(121)
    qvylm(1,147)=-16.58312395177699923939_t_real*ylm(101)&
         &      +114.29129450662460665879_t_real*ylm(103)
    qvylm(2,147)=114.29129450662460665879_t_real*ylm(119)&
         &      -0.00000000000000000003_t_real*ylm(121)
    qvylm(3,147)=74.16198487095662946311_t_real*ylm(102)
    qvylm(4,147)=-16.58312395177699923939_t_real*ylm(101)&
         &      -114.29129450662460665879_t_real*ylm(103)
    qvylm(5,147)=74.16198487095662946311_t_real*ylm(120)
    qvylm(1,148)=-27.38612787525830566231_t_real*ylm(102)&
         &      +103.38036564067666322663_t_real*ylm(104)
    qvylm(2,148)=103.38036564067666322663_t_real*ylm(118)&
         &      -0.00000000000000000109_t_real*ylm(120)
    qvylm(3,148)=-6.12372435695794524318_t_real*ylm(101)&
         &      +84.40971508067066088663_t_real*ylm(103)
    qvylm(4,148)=-27.38612787525830566231_t_real*ylm(102)&
         &      -103.38036564067666322663_t_real*ylm(104)
    qvylm(5,148)=84.40971508067066088663_t_real*ylm(119)&
         &      +6.12372435695794524318_t_real*ylm(121)
    qvylm(1,149)=1.33630620956212192285_t_real*ylm(101)&
         &      -36.83941988065036213354_t_real*ylm(103)&
         &      +93.01497575275561119760_t_real*ylm(105)
    qvylm(2,149)=93.01497575275561119760_t_real*ylm(117)&
         &      -0.00000000000000000048_t_real*ylm(119)&
         &      -1.33630620956212192285_t_real*ylm(121)
    qvylm(3,149)=-11.95228609334393639549_t_real*ylm(102)&
         &      +90.23778112773575314046_t_real*ylm(104)
    qvylm(4,149)=-36.83941988065036213354_t_real*ylm(103)&
         &      -93.01497575275561119760_t_real*ylm(105)&
         &      -1.33630620956212192285_t_real*ylm(101)
    qvylm(5,149)=90.23778112773575314046_t_real*ylm(118)&
         &      +11.95228609334393639549_t_real*ylm(120)
    qvylm(1,150)=2.98807152333598409870_t_real*ylm(102)&
         &      -45.11889056386787656898_t_real*ylm(104)&
         &      +83.19512348346248728031_t_real*ylm(106)
    qvylm(2,150)=83.19512348346248728031_t_real*ylm(116)&
         &      -0.00000000000000000128_t_real*ylm(118)&
         &      -2.98807152333598409870_t_real*ylm(120)
    qvylm(3,150)=-18.41970994032518106701_t_real*ylm(103)&
         &      +93.01497575275561119760_t_real*ylm(105)
    qvylm(4,150)=-45.11889056386787656898_t_real*ylm(104)&
         &      -83.19512348346248728031_t_real*ylm(106)&
         &      -2.98807152333598409870_t_real*ylm(102)
    qvylm(5,150)=93.01497575275561119760_t_real*ylm(117)&
         &      +18.41970994032518106701_t_real*ylm(119)
    qvylm(1,151)=5.17549169506765663143_t_real*ylm(103)&
         &      -52.26990393278771983565_t_real*ylm(105)&
         &      +73.92080704568717221915_t_real*ylm(107)
    qvylm(2,151)=73.92080704568717221915_t_real*ylm(115)&
         &      -0.00000000000000000001_t_real*ylm(117)&
         &      -5.17549169506765663143_t_real*ylm(119)
    qvylm(3,151)=-25.35462764185549731720_t_real*ylm(104)&
         &      +93.50324669687755206881_t_real*ylm(106)
    qvylm(4,151)=-52.26990393278771983565_t_real*ylm(105)&
         &      -73.92080704568717221915_t_real*ylm(107)&
         &      -5.17549169506765663143_t_real*ylm(103)
    qvylm(5,151)=93.50324669687755206881_t_real*ylm(116)&
         &      +25.35462764185549731720_t_real*ylm(118)
    qvylm(1,152)=7.90569415042094832747_t_real*ylm(104)&
         &      -58.30951894845300468904_t_real*ylm(106)&
         &      +65.19202405202648712398_t_real*ylm(108)
    qvylm(2,152)=65.19202405202648712398_t_real*ylm(114)&
         &      +0.00000000000000000110_t_real*ylm(116)&
         &      -7.90569415042094832747_t_real*ylm(118)
    qvylm(3,152)=-32.59601202601324356161_t_real*ylm(105)&
         &      +92.19544457292887307035_t_real*ylm(107)
    qvylm(4,152)=-58.30951894845300468904_t_real*ylm(106)&
         &      -65.19202405202648712398_t_real*ylm(108)&
         &      -7.90569415042094832747_t_real*ylm(104)
    qvylm(5,152)=92.19544457292887307035_t_real*ylm(115)&
         &      +32.59601202601324356161_t_real*ylm(117)
    qvylm(1,153)=11.18033988749894847780_t_real*ylm(105)&
         &      -63.24555320336758661705_t_real*ylm(107)&
         &      +57.00877125495689893754_t_real*ylm(109)
    qvylm(2,153)=57.00877125495689893754_t_real*ylm(113)&
         &      +0.00000000000000000049_t_real*ylm(115)&
         &      -11.18033988749894847780_t_real*ylm(117)
    qvylm(3,153)=-39.99999999999999998569_t_real*ylm(106)&
         &      +89.44271909999158782630_t_real*ylm(108)
    qvylm(4,153)=-63.24555320336758661705_t_real*ylm(107)&
         &      -57.00877125495689893754_t_real*ylm(109)&
         &      -11.18033988749894847780_t_real*ylm(105)
    qvylm(5,153)=89.44271909999158782630_t_real*ylm(114)&
         &      +39.99999999999999998569_t_real*ylm(116)
    qvylm(1,154)=14.99999999999999999456_t_real*ylm(106)&
         &      -67.08203932499369086950_t_real*ylm(108)&
         &      +49.37104414532874752656_t_real*ylm(110)
    qvylm(2,154)=49.37104414532874752656_t_real*ylm(112)&
         &      -0.00000000000000000008_t_real*ylm(114)&
         &      -14.99999999999999999456_t_real*ylm(116)
    qvylm(3,154)=-47.43416490252568996310_t_real*ylm(107)&
         &      +85.51315688243534840755_t_real*ylm(109)
    qvylm(4,154)=-67.08203932499369086950_t_real*ylm(108)&
         &      -49.37104414532874752656_t_real*ylm(110)&
         &      -14.99999999999999999456_t_real*ylm(106)
    qvylm(5,154)=85.51315688243534840755_t_real*ylm(113)&
         &      +47.43416490252568996310_t_real*ylm(115)
    qvylm(1,155)=19.36491673103708441891_t_real*ylm(107)&
         &      -69.82120021884470582613_t_real*ylm(109)
    qvylm(2,155)=59.79130371550699008416_t_real*ylm(111)&
         &      +0.00000000000000000037_t_real*ylm(113)&
         &      -19.36491673103708441891_t_real*ylm(115)
    qvylm(3,155)=-54.77225575051661132801_t_real*ylm(108)&
         &      +80.62257748298549649710_t_real*ylm(110)
    qvylm(4,155)=-69.82120021884470582613_t_real*ylm(109)&
         &      -19.36491673103708441891_t_real*ylm(107)
    qvylm(5,155)=80.62257748298549649710_t_real*ylm(112)&
         &      +54.77225575051661132801_t_real*ylm(114)
    qvylm(1,156)=24.27520781138061205805_t_real*ylm(108)&
         &      -35.73213839508789158894_t_real*ylm(110)
    qvylm(2,156)=-35.73213839508789158688_t_real*ylm(112)&
         &      -24.27520781138061205805_t_real*ylm(114)
    qvylm(3,156)=-61.88987916337486964517_t_real*ylm(109)
    qvylm(4,156)=-107.19641518526367476476_t_real*ylm(110)&
         &      -24.27520781138061205805_t_real*ylm(108)
    qvylm(5,156)=105.99865228253746862125_t_real*ylm(111)&
         &      +61.88987916337486964517_t_real*ylm(113)
    qvylm(1,157)=-72.01190377787748579168_t_real*ylm(111)&
         &      +42.04589329360410741817_t_real*ylm(113)
    qvylm(2,157)=42.04589329360410741817_t_real*ylm(109)
    qvylm(3,157)=-97.10083124552244823177_t_real*ylm(112)
    qvylm(4,157)=-72.01190377787748579168_t_real*ylm(111)&
         &      -42.04589329360410741817_t_real*ylm(113)
    qvylm(5,157)=-97.10083124552244823177_t_real*ylm(110)
    qvylm(1,158)=-107.19641518526367476476_t_real*ylm(112)&
         &      +24.27520781138061205805_t_real*ylm(114)
    qvylm(2,158)=24.27520781138061205805_t_real*ylm(108)&
         &      -35.73213839508789158894_t_real*ylm(110)
    qvylm(3,158)=105.99865228253746862125_t_real*ylm(111)&
         &      -61.88987916337486964517_t_real*ylm(113)
    qvylm(4,158)=-35.73213839508789158894_t_real*ylm(112)&
         &      -24.27520781138061205805_t_real*ylm(114)
    qvylm(5,158)=-61.88987916337486964517_t_real*ylm(109)
    qvylm(1,159)=59.79130371550699008416_t_real*ylm(111)&
         &      -69.82120021884470582613_t_real*ylm(113)&
         &      +19.36491673103708441891_t_real*ylm(115)
    qvylm(2,159)=19.36491673103708441891_t_real*ylm(107)&
         &      -0.00000000000000000037_t_real*ylm(109)
    qvylm(3,159)=80.62257748298549649710_t_real*ylm(112)&
         &      -54.77225575051661132801_t_real*ylm(114)
    qvylm(4,159)=-69.82120021884470582613_t_real*ylm(113)&
         &      -19.36491673103708441891_t_real*ylm(115)&
         &      -59.79130371550699008416_t_real*ylm(111)
    qvylm(5,159)=-54.77225575051661132801_t_real*ylm(108)&
         &      -80.62257748298549649710_t_real*ylm(110)
    qvylm(1,160)=49.37104414532874752656_t_real*ylm(112)&
         &      -67.08203932499369086950_t_real*ylm(114)&
         &      +14.99999999999999999456_t_real*ylm(116)
    qvylm(2,160)=14.99999999999999999456_t_real*ylm(106)&
         &      +0.00000000000000000008_t_real*ylm(108)&
         &      -49.37104414532874752656_t_real*ylm(110)
    qvylm(3,160)=85.51315688243534840755_t_real*ylm(113)&
         &      -47.43416490252568996310_t_real*ylm(115)
    qvylm(4,160)=-67.08203932499369086950_t_real*ylm(114)&
         &      -14.99999999999999999456_t_real*ylm(116)&
         &      -49.37104414532874752656_t_real*ylm(112)
    qvylm(5,160)=-47.43416490252568996310_t_real*ylm(107)&
         &      -85.51315688243534840755_t_real*ylm(109)
    qvylm(1,161)=57.00877125495689893754_t_real*ylm(113)&
         &      -63.24555320336758661705_t_real*ylm(115)&
         &      +11.18033988749894847780_t_real*ylm(117)
    qvylm(2,161)=11.18033988749894847780_t_real*ylm(105)&
         &      -0.00000000000000000049_t_real*ylm(107)&
         &      -57.00877125495689893754_t_real*ylm(109)
    qvylm(3,161)=89.44271909999158782630_t_real*ylm(114)&
         &      -39.99999999999999998569_t_real*ylm(116)
    qvylm(4,161)=-63.24555320336758661705_t_real*ylm(115)&
         &      -11.18033988749894847780_t_real*ylm(117)&
         &      -57.00877125495689893754_t_real*ylm(113)
    qvylm(5,161)=-39.99999999999999998569_t_real*ylm(106)&
         &      -89.44271909999158782630_t_real*ylm(108)
    qvylm(1,162)=65.19202405202648712398_t_real*ylm(114)&
         &      -58.30951894845300468904_t_real*ylm(116)&
         &      +7.90569415042094832747_t_real*ylm(118)
    qvylm(2,162)=7.90569415042094832747_t_real*ylm(104)&
         &      -0.00000000000000000110_t_real*ylm(106)&
         &      -65.19202405202648712398_t_real*ylm(108)
    qvylm(3,162)=92.19544457292887307035_t_real*ylm(115)&
         &      -32.59601202601324356161_t_real*ylm(117)
    qvylm(4,162)=-58.30951894845300468904_t_real*ylm(116)&
         &      -7.90569415042094832747_t_real*ylm(118)&
         &      -65.19202405202648712398_t_real*ylm(114)
    qvylm(5,162)=-32.59601202601324356161_t_real*ylm(105)&
         &      -92.19544457292887307035_t_real*ylm(107)
    qvylm(1,163)=73.92080704568717221915_t_real*ylm(115)&
         &      -52.26990393278771983565_t_real*ylm(117)&
         &      +5.17549169506765663143_t_real*ylm(119)
    qvylm(2,163)=5.17549169506765663143_t_real*ylm(103)&
         &      +0.00000000000000000001_t_real*ylm(105)&
         &      -73.92080704568717221915_t_real*ylm(107)
    qvylm(3,163)=93.50324669687755206881_t_real*ylm(116)&
         &      -25.35462764185549731720_t_real*ylm(118)
    qvylm(4,163)=-52.26990393278771983565_t_real*ylm(117)&
         &      -5.17549169506765663143_t_real*ylm(119)&
         &      -73.92080704568717221915_t_real*ylm(115)
    qvylm(5,163)=-25.35462764185549731720_t_real*ylm(104)&
         &      -93.50324669687755206881_t_real*ylm(106)
    qvylm(1,164)=83.19512348346248728031_t_real*ylm(116)&
         &      -45.11889056386787656898_t_real*ylm(118)&
         &      +2.98807152333598409870_t_real*ylm(120)
    qvylm(2,164)=2.98807152333598409870_t_real*ylm(102)&
         &      +0.00000000000000000128_t_real*ylm(104)&
         &      -83.19512348346248728031_t_real*ylm(106)
    qvylm(3,164)=93.01497575275561119760_t_real*ylm(117)&
         &      -18.41970994032518106701_t_real*ylm(119)
    qvylm(4,164)=-45.11889056386787656898_t_real*ylm(118)&
         &      -2.98807152333598409870_t_real*ylm(120)&
         &      -83.19512348346248728031_t_real*ylm(116)
    qvylm(5,164)=-18.41970994032518106701_t_real*ylm(103)&
         &      -93.01497575275561119760_t_real*ylm(105)
    qvylm(1,165)=93.01497575275561119760_t_real*ylm(117)&
         &      -36.83941988065036213354_t_real*ylm(119)&
         &      +1.33630620956212192285_t_real*ylm(121)
    qvylm(2,165)=1.33630620956212192285_t_real*ylm(101)&
         &      +0.00000000000000000048_t_real*ylm(103)&
         &      -93.01497575275561119760_t_real*ylm(105)
    qvylm(3,165)=90.23778112773575314046_t_real*ylm(118)&
         &      -11.95228609334393639549_t_real*ylm(120)
    qvylm(4,165)=-36.83941988065036213354_t_real*ylm(119)&
         &      -1.33630620956212192285_t_real*ylm(121)&
         &      -93.01497575275561119760_t_real*ylm(117)
    qvylm(5,165)=-11.95228609334393639549_t_real*ylm(102)&
         &      -90.23778112773575314046_t_real*ylm(104)
    qvylm(1,166)=103.38036564067666322663_t_real*ylm(118)&
         &      -27.38612787525830566231_t_real*ylm(120)
    qvylm(2,166)= 0.00000000000000000109_t_real*ylm(102)&
         &      -103.38036564067666322663_t_real*ylm(104)
    qvylm(3,166)=84.40971508067066088663_t_real*ylm(119)&
         &      -6.12372435695794524318_t_real*ylm(121)
    qvylm(4,166)=-27.38612787525830566231_t_real*ylm(120)&
         &      -103.38036564067666322663_t_real*ylm(118)
    qvylm(5,166)=-6.12372435695794524318_t_real*ylm(101)&
         &      -84.40971508067066088663_t_real*ylm(103)
    qvylm(1,167)=114.29129450662460665879_t_real*ylm(119)&
         &      -16.58312395177699923939_t_real*ylm(121)
    qvylm(2,167)= 0.00000000000000000003_t_real*ylm(101)&
         &      -114.29129450662460665879_t_real*ylm(103)
    qvylm(3,167)=74.16198487095662946311_t_real*ylm(120)
    qvylm(4,167)=-16.58312395177699923939_t_real*ylm(121)&
         &      -114.29129450662460665879_t_real*ylm(119)
    qvylm(5,167)=-74.16198487095662946311_t_real*ylm(102)
    qvylm(1,168)=125.74776339959291693290_t_real*ylm(120)
    qvylm(2,168)=-125.74776339959291693290_t_real*ylm(102)
    qvylm(3,168)=56.23610939600996257577_t_real*ylm(121)
    qvylm(4,168)=-125.74776339959291693290_t_real*ylm(120)
    qvylm(5,168)=-56.23610939600996257577_t_real*ylm(101)
    qvylm(1,169)=137.74977313955910826876_t_real*ylm(121)
    qvylm(2,169)=-137.74977313955910826876_t_real*ylm(101)
    qvylm(3,169)= 0.0_t_real
    qvylm(4,169)=-137.74977313955910826876_t_real*ylm(121)
    qvylm(5,169)= 0.0_t_real
    !!
    If(maxL<=12) Goto 2000
    !!

    !!XXXXXXXXXXXXXXXXX
    !!
2000 Continue
    !!
    Return
    !!
  End Subroutine calc_real_quad_harmonics_opt
  !!***********************************************************************
  Subroutine  calc_real_vec_Ylm_opt(maxL,ylm,rvylm)
    !!
    Implicit None  
    !!IO
    Integer(t_int),Intent(in)    :: maxL
    Real(t_real),Intent(in)      :: ylm((maxL+1)**2)
    Real(t_real),Intent(out)     :: rvylm(3,((maxL+1)**2))
    !!LOCAL
    Character(*),Parameter::nsubfunc="calc_real_vec_Ylm_opt"
    !!
    If(maxL>12_t_int)Then
3      Format('''maxl > 12 not implemented! ''')
       Write(errbuf,3) 
       Call errmsg(nmodule,nsubfunc,E_fatalerror,msg=errbuf)
    Endif
    !!
    !! XXXXXXXXXXXXXXXXXXXXXX
    !!..............................................
    !! L=0
    !!..............................................
    !! L-1
    rvylm(1,1)= 0.0_t_real
    rvylm(2,1)= 0.0_t_real
    rvylm(3,1)= 0.0_t_real
    !!
    If(maxL<=0) Goto 1000
    !!
    !!..............................................
    !! L=1
    !!..............................................
    !! L-1
    rvylm(1,2)= 0.0_t_real
    rvylm(2,2)= 0.48860251190291992157_t_real
    rvylm(3,2)= 0.0_t_real
    rvylm(1,3)= 0.0_t_real
    rvylm(2,3)= 0.0_t_real
    rvylm(3,3)= 0.48860251190291992158_t_real
    rvylm(1,4)= 0.48860251190291992157_t_real
    rvylm(2,4)= 0.0_t_real
    rvylm(3,4)= 0.0_t_real
    !!
    If(maxL<=1) Goto 1000
    !!
    !!..............................................
    !! L=2
    !!..............................................
    !! L-1
    rvylm(1,5)=2.23606797749978969630_t_real*ylm(2)
    rvylm(2,5)=2.23606797749978969630_t_real*ylm(4)
    rvylm(3,5)= 0.0_t_real
    rvylm(1,6)= 0.0_t_real
    rvylm(2,6)=2.23606797749978969630_t_real*ylm(3)
    rvylm(3,6)=2.23606797749978969630_t_real*ylm(2)
    rvylm(1,7)=-1.29099444873580562829_t_real*ylm(4)
    rvylm(2,7)=-1.29099444873580562829_t_real*ylm(2)
    rvylm(3,7)=2.58198889747161125670_t_real*ylm(3)
    rvylm(1,8)=2.23606797749978969630_t_real*ylm(3)
    rvylm(2,8)= 0.0_t_real
    rvylm(3,8)=2.23606797749978969630_t_real*ylm(4)
    rvylm(1,9)=2.23606797749978969630_t_real*ylm(4)
    rvylm(2,9)=-2.23606797749978969630_t_real*ylm(2)
    rvylm(3,9)= 0.0_t_real
    !!
    If(maxL<=2) Goto 1000
    !!
    !!..............................................
    !! L=3
    !!..............................................
    !! L-1
    rvylm(1,10)=3.24037034920393011516_t_real*ylm(5)
    rvylm(2,10)=3.24037034920393011516_t_real*ylm(9)
    rvylm(3,10)= 0.0_t_real
    rvylm(1,11)=2.64575131106459059024_t_real*ylm(6)
    rvylm(2,11)=2.64575131106459059024_t_real*ylm(8)
    rvylm(3,11)=2.64575131106459059024_t_real*ylm(5)
    rvylm(1,12)=-0.83666002653407554786_t_real*ylm(5)
    rvylm(2,12)=2.89827534923788771438_t_real*ylm(7)&
         &      +0.83666002653407554786_t_real*ylm(9)
    rvylm(3,12)=3.34664010613630219159_t_real*ylm(6)
    rvylm(1,13)=-2.04939015319191967638_t_real*ylm(8)
    rvylm(2,13)=-2.04939015319191967638_t_real*ylm(6)
    rvylm(3,13)=3.54964786985976962523_t_real*ylm(7)
    rvylm(1,14)=2.89827534923788771438_t_real*ylm(7)&
         &      -0.83666002653407554786_t_real*ylm(9)
    rvylm(2,14)=-0.83666002653407554786_t_real*ylm(5)
    rvylm(3,14)=3.34664010613630219159_t_real*ylm(8)
    rvylm(1,15)=2.64575131106459059024_t_real*ylm(8)
    rvylm(2,15)=-2.64575131106459059024_t_real*ylm(6)
    rvylm(3,15)=2.64575131106459059024_t_real*ylm(9)
    rvylm(1,16)=3.24037034920393011516_t_real*ylm(9)
    rvylm(2,16)=-3.24037034920393011516_t_real*ylm(5)
    rvylm(3,16)= 0.0_t_real
    !!
    If(maxL<=3) Goto 1000
    !!
    !!..............................................
    !! L=4
    !!..............................................
    !! L-1
    rvylm(1,17)=4.24264068711928514617_t_real*ylm(10)
    rvylm(2,17)=4.24264068711928514617_t_real*ylm(16)
    rvylm(3,17)= 0.0_t_real
    rvylm(1,18)=3.67423461417476714708_t_real*ylm(11)
    rvylm(2,18)=3.67423461417476714708_t_real*ylm(15)
    rvylm(3,18)=2.99999999999999999985_t_real*ylm(10)
    rvylm(1,19)=-0.80178372573727315397_t_real*ylm(10)&
         &      +3.10529501704059397988_t_real*ylm(12)
    rvylm(2,19)=3.10529501704059397988_t_real*ylm(14)&
         &      +0.80178372573727315397_t_real*ylm(16)
    rvylm(3,19)=3.92792202424786286258_t_real*ylm(11)
    rvylm(1,20)=-1.38873014965882719226_t_real*ylm(11)
    rvylm(2,20)=3.58568582800318091960_t_real*ylm(13)&
         &      +1.38873014965882719226_t_real*ylm(15)
    rvylm(3,20)=4.39155032826839930683_t_real*ylm(12)
    rvylm(1,21)=-2.77746029931765438447_t_real*ylm(14)
    rvylm(2,21)=-2.77746029931765438447_t_real*ylm(12)
    rvylm(3,21)=4.53557367611072672635_t_real*ylm(13)
    rvylm(1,22)=3.58568582800318091960_t_real*ylm(13)&
         &      -1.38873014965882719226_t_real*ylm(15)
    rvylm(2,22)=-1.38873014965882719226_t_real*ylm(11)
    rvylm(3,22)=4.39155032826839930683_t_real*ylm(14)
    rvylm(1,23)=3.10529501704059397988_t_real*ylm(14)&
         &      -0.80178372573727315397_t_real*ylm(16)
    rvylm(2,23)=-0.80178372573727315397_t_real*ylm(10)&
         &      -3.10529501704059397988_t_real*ylm(12)
    rvylm(3,23)=3.92792202424786286258_t_real*ylm(15)
    rvylm(1,24)=3.67423461417476714708_t_real*ylm(15)
    rvylm(2,24)=-3.67423461417476714708_t_real*ylm(11)
    rvylm(3,24)=2.99999999999999999985_t_real*ylm(16)
    rvylm(1,25)=4.24264068711928514617_t_real*ylm(16)
    rvylm(2,25)=-4.24264068711928514617_t_real*ylm(10)
    rvylm(3,25)= 0.0_t_real
    !!
    If(maxL<=4) Goto 1000
    !!
    !!..............................................
    !! L=5
    !!..............................................
    !! L-1
    rvylm(1,26)=5.24404424085075773482_t_real*ylm(17)
    rvylm(2,26)=5.24404424085075773482_t_real*ylm(25)
    rvylm(3,26)= 0.0_t_real
    rvylm(1,27)=4.69041575982342955444_t_real*ylm(18)
    rvylm(2,27)=4.69041575982342955444_t_real*ylm(24)
    rvylm(3,27)=3.31662479035539984906_t_real*ylm(17)
    rvylm(1,28)=-0.78173595997057159239_t_real*ylm(17)&
         &      +4.13655788199695219863_t_real*ylm(19)
    rvylm(2,28)=4.13655788199695219863_t_real*ylm(23)&
         &      +0.78173595997057159239_t_real*ylm(25)
    rvylm(3,28)=4.42216638714053313209_t_real*ylm(18)
    rvylm(1,29)=-1.35400640077266006004_t_real*ylm(18)&
         &      +3.58236421003411283978_t_real*ylm(20)
    rvylm(2,29)=3.58236421003411283978_t_real*ylm(22)&
         &      +1.35400640077266006004_t_real*ylm(24)
    rvylm(3,29)=5.06622805119022121044_t_real*ylm(19)
    rvylm(1,30)=-1.91485421551267621986_t_real*ylm(19)
    rvylm(2,30)=4.28174419288837633985_t_real*ylm(21)&
         &      +1.91485421551267621986_t_real*ylm(23)
    rvylm(3,30)=5.41602560309064024018_t_real*ylm(20)
    rvylm(1,31)=-3.49602949390050515650_t_real*ylm(22)
    rvylm(2,31)=-3.49602949390050515650_t_real*ylm(20)
    rvylm(3,31)=5.52770798392566641511_t_real*ylm(21)
    rvylm(1,32)=4.28174419288837633985_t_real*ylm(21)&
         &      -1.91485421551267621986_t_real*ylm(23)
    rvylm(2,32)=-1.91485421551267621986_t_real*ylm(19)
    rvylm(3,32)=5.41602560309064024018_t_real*ylm(22)
    rvylm(1,33)=3.58236421003411283978_t_real*ylm(22)&
         &      -1.35400640077266006004_t_real*ylm(24)
    rvylm(2,33)=-1.35400640077266006004_t_real*ylm(18)&
         &      -3.58236421003411283978_t_real*ylm(20)
    rvylm(3,33)=5.06622805119022121044_t_real*ylm(23)
    rvylm(1,34)=4.13655788199695219863_t_real*ylm(23)&
         &      -0.78173595997057159239_t_real*ylm(25)
    rvylm(2,34)=-0.78173595997057159239_t_real*ylm(17)&
         &      -4.13655788199695219863_t_real*ylm(19)
    rvylm(3,34)=4.42216638714053313209_t_real*ylm(24)
    rvylm(1,35)=4.69041575982342955444_t_real*ylm(24)
    rvylm(2,35)=-4.69041575982342955444_t_real*ylm(18)
    rvylm(3,35)=3.31662479035539984906_t_real*ylm(25)
    rvylm(1,36)=5.24404424085075773482_t_real*ylm(25)
    rvylm(2,36)=-5.24404424085075773482_t_real*ylm(17)
    rvylm(3,36)= 0.0_t_real
    !!
    If(maxL<=5) Goto 1000
    !!
    !!..............................................
    !! L=6
    !!..............................................
    !! L-1
    rvylm(1,37)=6.24499799839839820498_t_real*ylm(26)
    rvylm(2,37)=6.24499799839839820498_t_real*ylm(36)
    rvylm(3,37)= 0.0_t_real
    rvylm(1,38)=5.70087712549568989494_t_real*ylm(27)
    rvylm(2,38)=5.70087712549568989494_t_real*ylm(35)
    rvylm(3,38)=3.60555127546398929258_t_real*ylm(26)
    rvylm(1,39)=-0.76870611478580740429_t_real*ylm(26)&
         &      +5.15663738214246464010_t_real*ylm(28)
    rvylm(2,39)=5.15663738214246464010_t_real*ylm(34)&
         &      +0.76870611478580740429_t_real*ylm(36)
    rvylm(3,39)=4.86172434804397732390_t_real*ylm(27)
    rvylm(1,40)=-1.33143804689789180092_t_real*ylm(27)&
         &      +4.61223668871484442618_t_real*ylm(29)
    rvylm(2,40)=4.61223668871484442618_t_real*ylm(33)&
         &      +1.33143804689789180092_t_real*ylm(35)
    rvylm(3,40)=5.64881323014763067168_t_real*ylm(28)
    rvylm(1,41)=-1.88293774338254355719_t_real*ylm(28)&
         &      +4.06761042203583521229_t_real*ylm(30)
    rvylm(2,41)=4.06761042203583521229_t_real*ylm(32)&
         &      +1.88293774338254355719_t_real*ylm(34)
    rvylm(3,41)=6.14964891828645923485_t_real*ylm(29)
    rvylm(1,42)=-2.43086217402198866191_t_real*ylm(29)
    rvylm(2,42)=4.98178500320736625169_t_real*ylm(31)&
         &      +2.43086217402198866191_t_real*ylm(33)
    rvylm(3,42)=6.43145678393599746847_t_real*ylm(30)
    rvylm(1,43)=-4.21037679160342219218_t_real*ylm(32)
    rvylm(2,43)=-4.21037679160342219218_t_real*ylm(30)
    rvylm(3,43)=6.52268767805530801070_t_real*ylm(31)
    rvylm(1,44)=4.98178500320736625169_t_real*ylm(31)&
         &      -2.43086217402198866191_t_real*ylm(33)
    rvylm(2,44)=-2.43086217402198866191_t_real*ylm(29)
    rvylm(3,44)=6.43145678393599746847_t_real*ylm(32)
    rvylm(1,45)=4.06761042203583521229_t_real*ylm(32)&
         &      -1.88293774338254355719_t_real*ylm(34)
    rvylm(2,45)=-1.88293774338254355719_t_real*ylm(28)&
         &      -4.06761042203583521229_t_real*ylm(30)
    rvylm(3,45)=6.14964891828645923485_t_real*ylm(33)
    rvylm(1,46)=4.61223668871484442618_t_real*ylm(33)&
         &      -1.33143804689789180092_t_real*ylm(35)
    rvylm(2,46)=-1.33143804689789180092_t_real*ylm(27)&
         &      -4.61223668871484442618_t_real*ylm(29)
    rvylm(3,46)=5.64881323014763067168_t_real*ylm(34)
    rvylm(1,47)=5.15663738214246464010_t_real*ylm(34)&
         &      -0.76870611478580740429_t_real*ylm(36)
    rvylm(2,47)=-0.76870611478580740429_t_real*ylm(26)&
         &      -5.15663738214246464010_t_real*ylm(28)
    rvylm(3,47)=4.86172434804397732390_t_real*ylm(35)
    rvylm(1,48)=5.70087712549568989494_t_real*ylm(35)
    rvylm(2,48)=-5.70087712549568989494_t_real*ylm(27)
    rvylm(3,48)=3.60555127546398929258_t_real*ylm(36)
    rvylm(1,49)=6.24499799839839820498_t_real*ylm(36)
    rvylm(2,49)=-6.24499799839839820498_t_real*ylm(26)
    rvylm(3,49)= 0.0_t_real
    !!
    If(maxL<=6) Goto 1000
    !!
    !!..............................................
    !! L=7
    !!..............................................
    !! L-1
    rvylm(1,50)=7.24568837309471928553_t_real*ylm(37)
    rvylm(2,50)=7.24568837309471928553_t_real*ylm(49)
    rvylm(3,50)= 0.0_t_real
    rvylm(1,51)=6.70820393249936908803_t_real*ylm(38)
    rvylm(2,51)=6.70820393249936908803_t_real*ylm(48)
    rvylm(3,51)=3.87298334620741688447_t_real*ylm(37)
    rvylm(1,52)=-0.75955452531274996207_t_real*ylm(37)&
         &      +6.17065013405581737244_t_real*ylm(39)
    rvylm(2,52)=6.17065013405581737244_t_real*ylm(47)&
         &      +0.75955452531274996207_t_real*ylm(49)
    rvylm(3,52)=5.26234811584217528733_t_real*ylm(38)
    rvylm(1,53)=-1.31558702896054382183_t_real*ylm(38)&
         &      +5.63300712149108069607_t_real*ylm(40)
    rvylm(2,53)=5.63300712149108069607_t_real*ylm(46)&
         &      +1.31558702896054382183_t_real*ylm(48)
    rvylm(3,53)=6.17065013405581737244_t_real*ylm(39)
    rvylm(1,54)=-1.86052101883812684248_t_real*ylm(39)&
         &      +5.09524665365068087684_t_real*ylm(41)
    rvylm(2,54)=5.09524665365068087684_t_real*ylm(45)&
         &      +1.86052101883812684248_t_real*ylm(47)
    rvylm(3,54)=6.79366220486757450242_t_real*ylm(40)
    rvylm(1,55)=-2.40192230707630700172_t_real*ylm(40)&
         &      +4.55732715187649977242_t_real*ylm(42)
    rvylm(2,55)=4.55732715187649977242_t_real*ylm(44)&
         &      +2.40192230707630700172_t_real*ylm(46)
    rvylm(3,55)=7.20576692122892100538_t_real*ylm(41)
    rvylm(1,56)=-2.94174202707276047827_t_real*ylm(41)
    rvylm(2,56)=5.68398560058805107456_t_real*ylm(43)&
         &      +2.94174202707276047827_t_real*ylm(45)
    rvylm(3,56)=7.44208407535250737003_t_real*ylm(42)
    rvylm(1,57)=-4.92247592485420194143_t_real*ylm(44)
    rvylm(2,57)=-4.92247592485420194143_t_real*ylm(42)
    rvylm(3,57)=7.51920617741404524399_t_real*ylm(43)
    rvylm(1,58)=5.68398560058805107456_t_real*ylm(43)&
         &      -2.94174202707276047827_t_real*ylm(45)
    rvylm(2,58)=-2.94174202707276047827_t_real*ylm(41)
    rvylm(3,58)=7.44208407535250737003_t_real*ylm(44)
    rvylm(1,59)=4.55732715187649977242_t_real*ylm(44)&
         &      -2.40192230707630700172_t_real*ylm(46)
    rvylm(2,59)=-2.40192230707630700172_t_real*ylm(40)&
         &      -4.55732715187649977242_t_real*ylm(42)
    rvylm(3,59)=7.20576692122892100538_t_real*ylm(45)
    rvylm(1,60)=5.09524665365068087684_t_real*ylm(45)&
         &      -1.86052101883812684248_t_real*ylm(47)
    rvylm(2,60)=-1.86052101883812684248_t_real*ylm(39)&
         &      -5.09524665365068087684_t_real*ylm(41)
    rvylm(3,60)=6.79366220486757450242_t_real*ylm(46)
    rvylm(1,61)=5.63300712149108069607_t_real*ylm(46)&
         &      -1.31558702896054382183_t_real*ylm(48)
    rvylm(2,61)=-1.31558702896054382183_t_real*ylm(38)&
         &      -5.63300712149108069607_t_real*ylm(40)
    rvylm(3,61)=6.17065013405581737244_t_real*ylm(47)
    rvylm(1,62)=6.17065013405581737244_t_real*ylm(47)&
         &      -0.75955452531274996207_t_real*ylm(49)
    rvylm(2,62)=-0.75955452531274996207_t_real*ylm(37)&
         &      -6.17065013405581737244_t_real*ylm(39)
    rvylm(3,62)=5.26234811584217528733_t_real*ylm(48)
    rvylm(1,63)=6.70820393249936908803_t_real*ylm(48)
    rvylm(2,63)=-6.70820393249936908803_t_real*ylm(38)
    rvylm(3,63)=3.87298334620741688447_t_real*ylm(49)
    rvylm(1,64)=7.24568837309471928553_t_real*ylm(49)
    rvylm(2,64)=-7.24568837309471928553_t_real*ylm(37)
    rvylm(3,64)= 0.0_t_real
    !!
    If(maxL<=7) Goto 1000
    !!
    !!..............................................
    !! L=8
    !!..............................................
    !! L-1
    rvylm(1,65)=8.24621125123532109794_t_real*ylm(50)
    rvylm(2,65)=8.24621125123532109794_t_real*ylm(64)
    rvylm(3,65)= 0.0_t_real
    rvylm(1,66)=7.71362431027075624291_t_real*ylm(51)
    rvylm(2,66)=7.71362431027075624291_t_real*ylm(63)
    rvylm(3,66)=4.12310562561766054891_t_real*ylm(50)
    rvylm(1,67)=-0.75277265270908099474_t_real*ylm(50)&
         &      +7.18099343173816492009_t_real*ylm(52)
    rvylm(2,67)=7.18099343173816492009_t_real*ylm(62)&
         &      +0.75277265270908099474_t_real*ylm(64)
    rvylm(3,67)=5.63323471314069544695_t_real*ylm(51)
    rvylm(1,68)=-1.30384048104052974265_t_real*ylm(51)&
         &      +6.64830805543786455952_t_real*ylm(53)
    rvylm(2,68)=6.64830805543786455952_t_real*ylm(61)&
         &      +1.30384048104052974265_t_real*ylm(63)
    rvylm(3,68)=6.64830805543786455952_t_real*ylm(52)
    rvylm(1,69)=-1.84390889145857746161_t_real*ylm(52)&
         &      +6.11555394056826220916_t_real*ylm(54)
    rvylm(2,69)=6.11555394056826220916_t_real*ylm(60)&
         &      +1.84390889145857746161_t_real*ylm(62)
    rvylm(3,69)=7.37563556583430984644_t_real*ylm(53)
    rvylm(1,70)=-2.38047614284761666544_t_real*ylm(53)&
         &      +5.58271140814807537673_t_real*ylm(55)
    rvylm(2,70)=5.58271140814807537673_t_real*ylm(59)&
         &      +2.38047614284761666544_t_real*ylm(61)
    rvylm(3,70)=7.89514618821800748719_t_real*ylm(54)
    rvylm(1,71)=-2.91547594742265023479_t_real*ylm(54)&
         &      +5.04975246918103897561_t_real*ylm(56)
    rvylm(2,71)=5.04975246918103897561_t_real*ylm(58)&
         &      +2.91547594742265023479_t_real*ylm(60)
    rvylm(3,71)=8.24621125123532109794_t_real*ylm(55)
    rvylm(1,72)=-3.44963766213206803729_t_real*ylm(55)
    rvylm(2,72)=6.38748776906852479812_t_real*ylm(57)&
         &      +3.44963766213206803729_t_real*ylm(59)
    rvylm(3,72)=8.44985206971104317037_t_real*ylm(56)
    rvylm(1,73)=-5.63323471314069544683_t_real*ylm(58)
    rvylm(2,73)=-5.63323471314069544683_t_real*ylm(56)
    rvylm(3,73)=8.51665035875803306448_t_real*ylm(57)
    rvylm(1,74)=6.38748776906852479812_t_real*ylm(57)&
         &      -3.44963766213206803729_t_real*ylm(59)
    rvylm(2,74)=-3.44963766213206803729_t_real*ylm(55)
    rvylm(3,74)=8.44985206971104317037_t_real*ylm(58)
    rvylm(1,75)=5.04975246918103897561_t_real*ylm(58)&
         &      -2.91547594742265023479_t_real*ylm(60)
    rvylm(2,75)=-2.91547594742265023479_t_real*ylm(54)&
         &      -5.04975246918103897561_t_real*ylm(56)
    rvylm(3,75)=8.24621125123532109794_t_real*ylm(59)
    rvylm(1,76)=5.58271140814807537673_t_real*ylm(59)&
         &      -2.38047614284761666544_t_real*ylm(61)
    rvylm(2,76)=-2.38047614284761666544_t_real*ylm(53)&
         &      -5.58271140814807537673_t_real*ylm(55)
    rvylm(3,76)=7.89514618821800748719_t_real*ylm(60)
    rvylm(1,77)=6.11555394056826220916_t_real*ylm(60)&
         &      -1.84390889145857746161_t_real*ylm(62)
    rvylm(2,77)=-1.84390889145857746161_t_real*ylm(52)&
         &      -6.11555394056826220916_t_real*ylm(54)
    rvylm(3,77)=7.37563556583430984644_t_real*ylm(61)
    rvylm(1,78)=6.64830805543786455952_t_real*ylm(61)&
         &      -1.30384048104052974265_t_real*ylm(63)
    rvylm(2,78)=-1.30384048104052974265_t_real*ylm(51)&
         &      -6.64830805543786455952_t_real*ylm(53)
    rvylm(3,78)=6.64830805543786455952_t_real*ylm(62)
    rvylm(1,79)=7.18099343173816492009_t_real*ylm(62)&
         &      -0.75277265270908099474_t_real*ylm(64)
    rvylm(2,79)=-0.75277265270908099474_t_real*ylm(50)&
         &      -7.18099343173816492009_t_real*ylm(52)
    rvylm(3,79)=5.63323471314069544695_t_real*ylm(63)
    rvylm(1,80)=7.71362431027075624291_t_real*ylm(63)
    rvylm(2,80)=-7.71362431027075624291_t_real*ylm(51)
    rvylm(3,80)=4.12310562561766054891_t_real*ylm(64)
    rvylm(1,81)=8.24621125123532109794_t_real*ylm(64)
    rvylm(2,81)=-8.24621125123532109794_t_real*ylm(50)
    rvylm(3,81)= 0.0_t_real
    !!
    If(maxL<=8) Goto 1000
    !!
    !!..............................................
    !! L=9
    !!..............................................
    !! L-1
    rvylm(1,82)=9.24662100445346467442_t_real*ylm(65)
    rvylm(2,82)=9.24662100445346467442_t_real*ylm(81)
    rvylm(3,82)= 0.0_t_real
    rvylm(1,83)=8.71779788708134710358_t_real*ylm(66)
    rvylm(2,83)=8.71779788708134710358_t_real*ylm(80)
    rvylm(3,83)=4.35889894354067355172_t_real*ylm(65)
    rvylm(1,84)=-0.74754500159640202466_t_real*ylm(65)&
         &      +8.18894520249169604204_t_real*ylm(67)
    rvylm(2,84)=8.18894520249169604204_t_real*ylm(79)&
         &      +0.74754500159640202466_t_real*ylm(81)
    rvylm(3,84)=5.98036001277121619814_t_real*ylm(66)
    rvylm(1,85)=-1.29478592370912579251_t_real*ylm(66)&
         &      +7.66005682669752092386_t_real*ylm(68)
    rvylm(2,85)=7.66005682669752092386_t_real*ylm(78)&
         &      +1.29478592370912579251_t_real*ylm(80)
    rvylm(3,85)=7.09183457555651276193_t_real*ylm(67)
    rvylm(1,86)=-1.83110381367922125644_t_real*ylm(67)&
         &      +7.13112481846101107951_t_real*ylm(69)
    rvylm(2,86)=7.13112481846101107951_t_real*ylm(77)&
         &      +1.83110381367922125644_t_real*ylm(79)
    rvylm(3,86)=7.91127267221384849204_t_real*ylm(68)
    rvylm(1,87)=-2.36394485851883758726_t_real*ylm(68)&
         &      +6.60213869091809120630_t_real*ylm(70)
    rvylm(2,87)=6.60213869091809120630_t_real*ylm(76)&
         &      +2.36394485851883758726_t_real*ylm(78)
    rvylm(3,87)=8.52332439975913457832_t_real*ylm(69)
    rvylm(1,88)=-2.89522934172346191113_t_real*ylm(69)&
         &      +6.07308430216282034611_t_real*ylm(71)
    rvylm(2,88)=6.07308430216282034611_t_real*ylm(75)&
         &      +2.89522934172346191113_t_real*ylm(77)
    rvylm(3,88)=8.97054001915682429734_t_real*ylm(70)
    rvylm(1,89)=-3.42568155520139653580_t_real*ylm(70)&
         &      +5.54394210987516163984_t_real*ylm(72)
    rvylm(2,89)=5.54394210987516163984_t_real*ylm(74)&
         &      +3.42568155520139653580_t_real*ylm(76)
    rvylm(3,89)=9.27678950550306302935_t_real*ylm(71)
    rvylm(1,90)=-3.95563633610692424602_t_real*ylm(71)
    rvylm(2,90)=7.09183457555651276193_t_real*ylm(73)&
         &      +3.95563633610692424602_t_real*ylm(75)
    rvylm(3,90)=9.45577943407535034933_t_real*ylm(72)
    rvylm(1,91)=-6.34312967845109239606_t_real*ylm(74)
    rvylm(2,91)=-6.34312967845109239606_t_real*ylm(72)
    rvylm(3,91)=9.51469451767663859429_t_real*ylm(73)
    rvylm(1,92)=7.09183457555651276193_t_real*ylm(73)&
         &      -3.95563633610692424602_t_real*ylm(75)
    rvylm(2,92)=-3.95563633610692424602_t_real*ylm(71)
    rvylm(3,92)=9.45577943407535034933_t_real*ylm(74)
    rvylm(1,93)=5.54394210987516163984_t_real*ylm(74)&
         &      -3.42568155520139653580_t_real*ylm(76)
    rvylm(2,93)=-3.42568155520139653580_t_real*ylm(70)&
         &      -5.54394210987516163984_t_real*ylm(72)
    rvylm(3,93)=9.27678950550306302935_t_real*ylm(75)
    rvylm(1,94)=6.07308430216282034611_t_real*ylm(75)&
         &      -2.89522934172346191113_t_real*ylm(77)
    rvylm(2,94)=-2.89522934172346191113_t_real*ylm(69)&
         &      -6.07308430216282034611_t_real*ylm(71)
    rvylm(3,94)=8.97054001915682429734_t_real*ylm(76)
    rvylm(1,95)=6.60213869091809120630_t_real*ylm(76)&
         &      -2.36394485851883758726_t_real*ylm(78)
    rvylm(2,95)=-2.36394485851883758726_t_real*ylm(68)&
         &      -6.60213869091809120630_t_real*ylm(70)
    rvylm(3,95)=8.52332439975913457832_t_real*ylm(77)
    rvylm(1,96)=7.13112481846101107951_t_real*ylm(77)&
         &      -1.83110381367922125644_t_real*ylm(79)
    rvylm(2,96)=-1.83110381367922125644_t_real*ylm(67)&
         &      -7.13112481846101107951_t_real*ylm(69)
    rvylm(3,96)=7.91127267221384849204_t_real*ylm(78)
    rvylm(1,97)=7.66005682669752092386_t_real*ylm(78)&
         &      -1.29478592370912579251_t_real*ylm(80)
    rvylm(2,97)=-1.29478592370912579251_t_real*ylm(66)&
         &      -7.66005682669752092386_t_real*ylm(68)
    rvylm(3,97)=7.09183457555651276193_t_real*ylm(79)
    rvylm(1,98)=8.18894520249169604204_t_real*ylm(79)&
         &      -0.74754500159640202466_t_real*ylm(81)
    rvylm(2,98)=-0.74754500159640202466_t_real*ylm(65)&
         &      -8.18894520249169604204_t_real*ylm(67)
    rvylm(3,98)=5.98036001277121619814_t_real*ylm(80)
    rvylm(1,99)=8.71779788708134710358_t_real*ylm(80)
    rvylm(2,99)=-8.71779788708134710358_t_real*ylm(66)
    rvylm(3,99)=4.35889894354067355172_t_real*ylm(81)
    rvylm(1,100)=9.24662100445346467442_t_real*ylm(81)
    rvylm(2,100)=-9.24662100445346467442_t_real*ylm(65)
    rvylm(3,100)= 0.0_t_real
    !!
    If(maxL<=9) Goto 1000
    !!
    !!..............................................
    !! L=10
    !!..............................................
    !! L-1
    rvylm(1,101)=10.24695076595959838021_t_real*ylm(82)
    rvylm(2,101)=10.24695076595959838021_t_real*ylm(100)
    rvylm(3,101)= 0.0_t_real
    rvylm(1,102)=9.72111104761179034356_t_real*ylm(83)
    rvylm(2,102)=9.72111104761179034356_t_real*ylm(99)
    rvylm(3,102)=4.58257569495584000524_t_real*ylm(82)
    rvylm(1,103)=-0.74339194167502812698_t_real*ylm(82)&
         &      +9.19525049027743280505_t_real*ylm(84)
    rvylm(2,103)=9.19525049027743280505_t_real*ylm(98)&
         &      +0.74339194167502812698_t_real*ylm(100)
    rvylm(3,103)=6.30788979645416176110_t_real*ylm(83)
    rvylm(1,104)=-1.28759261291842819510_t_real*ylm(83)&
         &      +8.66936530184546469713_t_real*ylm(85)
    rvylm(2,104)=8.66936530184546469713_t_real*ylm(97)&
         &      +1.28759261291842819510_t_real*ylm(99)
    rvylm(3,104)=7.50789058608552039380_t_real*ylm(84)
    rvylm(1,105)=-1.82093093600065198186_t_real*ylm(84)&
         &      +8.14345071045955373554_t_real*ylm(86)
    rvylm(2,105)=8.14345071045955373554_t_real*ylm(96)&
         &      +1.82093093600065198186_t_real*ylm(98)
    rvylm(3,105)=8.41051972860554901480_t_real*ylm(85)
    rvylm(1,106)=-2.35081172990813626463_t_real*ylm(85)&
         &      +7.61750062615512255368_t_real*ylm(87)
    rvylm(2,106)=7.61750062615512255368_t_real*ylm(95)&
         &      +2.35081172990813626463_t_real*ylm(97)
    rvylm(3,106)=9.10465468000325991007_t_real*ylm(86)
    rvylm(1,107)=-2.87914460981217932131_t_real*ylm(86)&
         &      +7.09150715181268968924_t_real*ylm(88)
    rvylm(2,107)=7.09150715181268968924_t_real*ylm(94)&
         &      +2.87914460981217932131_t_real*ylm(96)
    rvylm(3,107)=9.63546082256359416530_t_real*ylm(87)
    rvylm(1,108)=-3.40664984374601330048_t_real*ylm(87)&
         &      +6.56545985882898661229_t_real*ylm(89)
    rvylm(2,108)=6.56545985882898661229_t_real*ylm(93)&
         &      +3.40664984374601330048_t_real*ylm(95)
    rvylm(3,108)=10.02890559175930484906_t_real*ylm(88)
    rvylm(1,109)=-3.93366040864311465542_t_real*ylm(88)&
         &      +6.03934468386482439955_t_real*ylm(90)
    rvylm(2,109)=6.03934468386482439955_t_real*ylm(92)&
         &      +3.93366040864311465542_t_real*ylm(94)
    rvylm(3,109)=10.30074090334742556083_t_real*ylm(89)
    rvylm(1,110)=-4.46035165005016876262_t_real*ylm(89)
    rvylm(2,110)=7.79676046087158729237_t_real*ylm(91)&
         &      +4.46035165005016876262_t_real*ylm(93)
    rvylm(3,110)=10.46045183687487503868_t_real*ylm(90)
    rvylm(1,111)=-7.05243518972440879420_t_real*ylm(92)
    rvylm(2,111)=-7.05243518972440879420_t_real*ylm(90)
    rvylm(3,111)=10.51314966075693626850_t_real*ylm(91)
    rvylm(1,112)=7.79676046087158729237_t_real*ylm(91)&
         &      -4.46035165005016876262_t_real*ylm(93)
    rvylm(2,112)=-4.46035165005016876262_t_real*ylm(89)
    rvylm(3,112)=10.46045183687487503868_t_real*ylm(92)
    rvylm(1,113)=6.03934468386482439955_t_real*ylm(92)&
         &      -3.93366040864311465542_t_real*ylm(94)
    rvylm(2,113)=-3.93366040864311465542_t_real*ylm(88)&
         &      -6.03934468386482439955_t_real*ylm(90)
    rvylm(3,113)=10.30074090334742556083_t_real*ylm(93)
    rvylm(1,114)=6.56545985882898661229_t_real*ylm(93)&
         &      -3.40664984374601330048_t_real*ylm(95)
    rvylm(2,114)=-3.40664984374601330048_t_real*ylm(87)&
         &      -6.56545985882898661229_t_real*ylm(89)
    rvylm(3,114)=10.02890559175930484906_t_real*ylm(94)
    rvylm(1,115)=7.09150715181268968924_t_real*ylm(94)&
         &      -2.87914460981217932131_t_real*ylm(96)
    rvylm(2,115)=-2.87914460981217932131_t_real*ylm(86)&
         &      -7.09150715181268968924_t_real*ylm(88)
    rvylm(3,115)=9.63546082256359416530_t_real*ylm(95)
    rvylm(1,116)=7.61750062615512255368_t_real*ylm(95)&
         &      -2.35081172990813626463_t_real*ylm(97)
    rvylm(2,116)=-2.35081172990813626463_t_real*ylm(85)&
         &      -7.61750062615512255368_t_real*ylm(87)
    rvylm(3,116)=9.10465468000325991007_t_real*ylm(96)
    rvylm(1,117)=8.14345071045955373554_t_real*ylm(96)&
         &      -1.82093093600065198186_t_real*ylm(98)
    rvylm(2,117)=-1.82093093600065198186_t_real*ylm(84)&
         &      -8.14345071045955373554_t_real*ylm(86)
    rvylm(3,117)=8.41051972860554901480_t_real*ylm(97)
    rvylm(1,118)=8.66936530184546469713_t_real*ylm(97)&
         &      -1.28759261291842819510_t_real*ylm(99)
    rvylm(2,118)=-1.28759261291842819510_t_real*ylm(83)&
         &      -8.66936530184546469713_t_real*ylm(85)
    rvylm(3,118)=7.50789058608552039380_t_real*ylm(98)
    rvylm(1,119)=9.19525049027743280505_t_real*ylm(98)&
         &      -0.74339194167502812698_t_real*ylm(100)
    rvylm(2,119)=-0.74339194167502812698_t_real*ylm(82)&
         &      -9.19525049027743280505_t_real*ylm(84)
    rvylm(3,119)=6.30788979645416176110_t_real*ylm(99)
    rvylm(1,120)=9.72111104761179034356_t_real*ylm(99)
    rvylm(2,120)=-9.72111104761179034356_t_real*ylm(83)
    rvylm(3,120)=4.58257569495584000524_t_real*ylm(100)
    rvylm(1,121)=10.24695076595959838021_t_real*ylm(100)
    rvylm(2,121)=-10.24695076595959838021_t_real*ylm(82)
    rvylm(3,121)= 0.0_t_real
    !!
    If(maxL<=10) Goto 1000
    !!
    !!..............................................
    !! L=11
    !!..............................................
    !! L-1
    rvylm(1,122)=11.24722187920199251574_t_real*ylm(101)
    rvylm(2,122)=11.24722187920199251574_t_real*ylm(121)
    rvylm(3,122)= 0.0_t_real
    rvylm(1,123)=10.72380529476360830159_t_real*ylm(102)
    rvylm(2,123)=10.72380529476360830159_t_real*ylm(120)
    rvylm(3,123)=4.79583152331271954017_t_real*ylm(101)
    rvylm(1,124)=-0.74001286990095491087_t_real*ylm(101)&
         &      +10.20037347588896117257_t_real*ylm(103)
    rvylm(2,124)=10.20037347588896117257_t_real*ylm(119)&
         &      +0.74001286990095491087_t_real*ylm(121)
    rvylm(3,124)=6.61887632529297298225_t_real*ylm(102)
    rvylm(1,125)=-1.28173988892331148207_t_real*ylm(102)&
         &      +9.67692395045332278867_t_real*ylm(104)
    rvylm(2,125)=9.67692395045332278867_t_real*ylm(118)&
         &      +1.28173988892331148207_t_real*ylm(120)
    rvylm(3,125)=7.90117531944276177215_t_real*ylm(103)
    rvylm(1,126)=-1.81265393434993148228_t_real*ylm(103)&
         &      +9.15345368075429142629_t_real*ylm(105)
    rvylm(2,126)=9.15345368075429142629_t_real*ylm(117)&
         &      +1.81265393434993148228_t_real*ylm(119)
    rvylm(3,126)=8.88015443881145893082_t_real*ylm(104)
    rvylm(1,127)=-2.34012616672487900012_t_real*ylm(104)&
         &      +8.62995889191776371504_t_real*ylm(106)
    rvylm(2,127)=8.62995889191776371504_t_real*ylm(116)&
         &      +2.34012616672487900012_t_real*ylm(118)
    rvylm(3,127)=9.64858736267844004767_t_real*ylm(105)
    rvylm(1,128)=-2.86605752110555420381_t_real*ylm(105)&
         &      +8.10643483377777571872_t_real*ylm(107)
    rvylm(2,128)=8.10643483377777571872_t_real*ylm(115)&
         &      +2.86605752110555420381_t_real*ylm(117)
    rvylm(3,128)=10.25391911138649185674_t_real*ylm(106)
    rvylm(1,129)=-3.39116499156263406841_t_real*ylm(106)&
         &      +7.58287544405155055191_t_real*ylm(108)
    rvylm(2,129)=7.58287544405155055191_t_real*ylm(114)&
         &      +3.39116499156263406841_t_real*ylm(116)
    rvylm(3,129)=10.72380529476360830159_t_real*ylm(107)
    rvylm(1,130)=-3.91578004149024352770_t_real*ylm(107)&
         &      +7.05927286151579023264_t_real*ylm(109)
    rvylm(2,130)=7.05927286151579023264_t_real*ylm(113)&
         &      +3.91578004149024352770_t_real*ylm(115)
    rvylm(3,130)=11.07549848389076646685_t_real*ylm(108)
    rvylm(1,131)=-4.44007721940572946541_t_real*ylm(108)&
         &      +6.53561670497021376997_t_real*ylm(110)
    rvylm(2,131)=6.53561670497021376997_t_real*ylm(112)&
         &      +4.44007721940572946541_t_real*ylm(114)
    rvylm(3,131)=11.32002019180430361392_t_real*ylm(109)
    rvylm(1,132)=-4.96415724396972973668_t_real*ylm(109)
    rvylm(2,132)=8.50210058078086286687_t_real*ylm(111)&
         &      +4.96415724396972973668_t_real*ylm(113)
    rvylm(3,132)=11.46423008442221681558_t_real*ylm(110)
    rvylm(1,133)=-7.76132045711908728798_t_real*ylm(112)
    rvylm(2,133)=-7.76132045711908728798_t_real*ylm(110)
    rvylm(3,133)=11.51189860639023886521_t_real*ylm(111)
    rvylm(1,134)=8.50210058078086286687_t_real*ylm(111)&
         &      -4.96415724396972973668_t_real*ylm(113)
    rvylm(2,134)=-4.96415724396972973668_t_real*ylm(109)
    rvylm(3,134)=11.46423008442221681558_t_real*ylm(112)
    rvylm(1,135)=6.53561670497021376997_t_real*ylm(112)&
         &      -4.44007721940572946541_t_real*ylm(114)
    rvylm(2,135)=-4.44007721940572946541_t_real*ylm(108)&
         &      -6.53561670497021376997_t_real*ylm(110)
    rvylm(3,135)=11.32002019180430361392_t_real*ylm(113)
    rvylm(1,136)=7.05927286151579023264_t_real*ylm(113)&
         &      -3.91578004149024352770_t_real*ylm(115)
    rvylm(2,136)=-3.91578004149024352770_t_real*ylm(107)&
         &      -7.05927286151579023264_t_real*ylm(109)
    rvylm(3,136)=11.07549848389076646685_t_real*ylm(114)
    rvylm(1,137)=7.58287544405155055191_t_real*ylm(114)&
         &      -3.39116499156263406841_t_real*ylm(116)
    rvylm(2,137)=-3.39116499156263406841_t_real*ylm(106)&
         &      -7.58287544405155055191_t_real*ylm(108)
    rvylm(3,137)=10.72380529476360830159_t_real*ylm(115)
    rvylm(1,138)=8.10643483377777571872_t_real*ylm(115)&
         &      -2.86605752110555420381_t_real*ylm(117)
    rvylm(2,138)=-2.86605752110555420381_t_real*ylm(105)&
         &      -8.10643483377777571872_t_real*ylm(107)
    rvylm(3,138)=10.25391911138649185674_t_real*ylm(116)
    rvylm(1,139)=8.62995889191776371504_t_real*ylm(116)&
         &      -2.34012616672487900012_t_real*ylm(118)
    rvylm(2,139)=-2.34012616672487900012_t_real*ylm(104)&
         &      -8.62995889191776371504_t_real*ylm(106)
    rvylm(3,139)=9.64858736267844004767_t_real*ylm(117)
    rvylm(1,140)=9.15345368075429142629_t_real*ylm(117)&
         &      -1.81265393434993148228_t_real*ylm(119)
    rvylm(2,140)=-1.81265393434993148228_t_real*ylm(103)&
         &      -9.15345368075429142629_t_real*ylm(105)
    rvylm(3,140)=8.88015443881145893082_t_real*ylm(118)
    rvylm(1,141)=9.67692395045332278867_t_real*ylm(118)&
         &      -1.28173988892331148207_t_real*ylm(120)
    rvylm(2,141)=-1.28173988892331148207_t_real*ylm(102)&
         &      -9.67692395045332278867_t_real*ylm(104)
    rvylm(3,141)=7.90117531944276177215_t_real*ylm(119)
    rvylm(1,142)=10.20037347588896117257_t_real*ylm(119)&
         &      -0.74001286990095491087_t_real*ylm(121)
    rvylm(2,142)=-0.74001286990095491087_t_real*ylm(101)&
         &      -10.20037347588896117257_t_real*ylm(103)
    rvylm(3,142)=6.61887632529297298225_t_real*ylm(120)
    rvylm(1,143)=10.72380529476360830159_t_real*ylm(120)
    rvylm(2,143)=-10.72380529476360830159_t_real*ylm(102)
    rvylm(3,143)=4.79583152331271954017_t_real*ylm(121)
    rvylm(1,144)=11.24722187920199251574_t_real*ylm(121)
    rvylm(2,144)=-11.24722187920199251574_t_real*ylm(101)
    rvylm(3,144)= 0.0_t_real
    !!
    If(maxL<=11) Goto 1000
    !!
    !!..............................................
    !! L=12
    !!..............................................
    !! L-1
    rvylm(1,145)=12.24744871391589049060_t_real*ylm(122)
    rvylm(2,145)=12.24744871391589049060_t_real*ylm(144)
    rvylm(3,145)= 0.0_t_real
    rvylm(1,146)=11.72603939955857388608_t_real*ylm(123)
    rvylm(2,146)=11.72603939955857388608_t_real*ylm(143)
    rvylm(3,146)=4.99999999999999999974_t_real*ylm(122)
    rvylm(1,147)=-0.73720978077448566723_t_real*ylm(122)&
         &      +11.20461861291447818860_t_real*ylm(124)
    rvylm(2,147)=11.20461861291447818860_t_real*ylm(142)&
         &      +0.73720978077448566723_t_real*ylm(144)
    rvylm(3,147)=6.91564074808124624060_t_real*ylm(123)
    rvylm(1,148)=-1.27688479613812288973_t_real*ylm(123)&
         &      +10.68318467417879543822_t_real*ylm(125)
    rvylm(2,148)=10.68318467417879543822_t_real*ylm(141)&
         &      +1.27688479613812288973_t_real*ylm(143)
    rvylm(3,148)=8.27515926551055676937_t_real*ylm(124)
    rvylm(1,149)=-1.80578779628653801127_t_real*ylm(124)&
         &      +10.16173555871325512508_t_real*ylm(126)
    rvylm(2,149)=10.16173555871325512508_t_real*ylm(140)&
         &      +1.80578779628653801127_t_real*ylm(142)
    rvylm(3,149)=9.32504808240313765608_t_real*ylm(125)
    rvylm(1,150)=-2.33126202060078441389_t_real*ylm(125)&
         &      +9.64026880375727138200_t_real*ylm(127)
    rvylm(2,150)=9.64026880375727138200_t_real*ylm(139)&
         &      +2.33126202060078441389_t_real*ylm(141)
    rvylm(3,150)=10.16173555871325512508_t_real*ylm(126)
    rvylm(1,151)=-2.85520120360080372741_t_real*ylm(126)&
         &      +9.11878138311493232910_t_real*ylm(128)
    rvylm(2,151)=9.11878138311493232910_t_real*ylm(138)&
         &      +2.85520120360080372741_t_real*ylm(140)
    rvylm(3,151)=10.83472677771922806832_t_real*ylm(127)
    rvylm(1,152)=-3.37831962346088111590_t_real*ylm(127)&
         &      +8.59726953621095152654_t_real*ylm(129)
    rvylm(2,152)=8.59726953621095152654_t_real*ylm(137)&
         &      +3.37831962346088111590_t_real*ylm(139)
    rvylm(3,152)=11.37311857350289484499_t_real*ylm(128)
    rvylm(1,153)=-3.90094748802746972907_t_real*ylm(128)&
         &      +8.07572853087248259575_t_real*ylm(130)
    rvylm(2,153)=8.07572853087248259575_t_real*ylm(136)&
         &      +3.90094748802746972907_t_real*ylm(138)
    rvylm(3,153)=11.79535649239177067632_t_real*ylm(129)
    rvylm(1,154)=-4.42325868464691400360_t_real*ylm(129)&
         &      +7.55415232778002351432_t_real*ylm(131)
    rvylm(2,154)=7.55415232778002351432_t_real*ylm(135)&
         &      +4.42325868464691400360_t_real*ylm(137)
    rvylm(3,154)=12.11359279630872389380_t_real*ylm(130)
    rvylm(1,155)=-4.94535355046840253523_t_real*ylm(130)&
         &      +7.03253309548774429223_t_real*ylm(132)
    rvylm(2,155)=7.03253309548774429223_t_real*ylm(134)&
         &      +4.94535355046840253523_t_real*ylm(136)
    rvylm(3,155)=12.33587909487922392313_t_real*ylm(131)
    rvylm(1,156)=-5.46729406085186595524_t_real*ylm(131)
    rvylm(2,156)=9.20774721067276987093_t_real*ylm(133)&
         &      +5.46729406085186595524_t_real*ylm(135)
    rvylm(3,156)=12.46734865994753722431_t_real*ylm(132)
    rvylm(1,157)=-8.46989553859919860879_t_real*ylm(134)
    rvylm(2,157)=-8.46989553859919860879_t_real*ylm(132)
    rvylm(3,157)=12.51086484342448576045_t_real*ylm(133)
    rvylm(1,158)=9.20774721067276987093_t_real*ylm(133)&
         &      -5.46729406085186595524_t_real*ylm(135)
    rvylm(2,158)=-5.46729406085186595524_t_real*ylm(131)
    rvylm(3,158)=12.46734865994753722431_t_real*ylm(134)
    rvylm(1,159)=7.03253309548774429223_t_real*ylm(134)&
         &      -4.94535355046840253523_t_real*ylm(136)
    rvylm(2,159)=-4.94535355046840253523_t_real*ylm(130)&
         &      -7.03253309548774429223_t_real*ylm(132)
    rvylm(3,159)=12.33587909487922392313_t_real*ylm(135)
    rvylm(1,160)=7.55415232778002351432_t_real*ylm(135)&
         &      -4.42325868464691400360_t_real*ylm(137)
    rvylm(2,160)=-4.42325868464691400360_t_real*ylm(129)&
         &      -7.55415232778002351432_t_real*ylm(131)
    rvylm(3,160)=12.11359279630872389380_t_real*ylm(136)
    rvylm(1,161)=8.07572853087248259575_t_real*ylm(136)&
         &      -3.90094748802746972907_t_real*ylm(138)
    rvylm(2,161)=-3.90094748802746972907_t_real*ylm(128)&
         &      -8.07572853087248259575_t_real*ylm(130)
    rvylm(3,161)=11.79535649239177067632_t_real*ylm(137)
    rvylm(1,162)=8.59726953621095152654_t_real*ylm(137)&
         &      -3.37831962346088111590_t_real*ylm(139)
    rvylm(2,162)=-3.37831962346088111590_t_real*ylm(127)&
         &      -8.59726953621095152654_t_real*ylm(129)
    rvylm(3,162)=11.37311857350289484499_t_real*ylm(138)
    rvylm(1,163)=9.11878138311493232910_t_real*ylm(138)&
         &      -2.85520120360080372741_t_real*ylm(140)
    rvylm(2,163)=-2.85520120360080372741_t_real*ylm(126)&
         &      -9.11878138311493232910_t_real*ylm(128)
    rvylm(3,163)=10.83472677771922806832_t_real*ylm(139)
    rvylm(1,164)=9.64026880375727138200_t_real*ylm(139)&
         &      -2.33126202060078441389_t_real*ylm(141)
    rvylm(2,164)=-2.33126202060078441389_t_real*ylm(125)&
         &      -9.64026880375727138200_t_real*ylm(127)
    rvylm(3,164)=10.16173555871325512508_t_real*ylm(140)
    rvylm(1,165)=10.16173555871325512508_t_real*ylm(140)&
         &      -1.80578779628653801127_t_real*ylm(142)
    rvylm(2,165)=-1.80578779628653801127_t_real*ylm(124)&
         &      -10.16173555871325512508_t_real*ylm(126)
    rvylm(3,165)=9.32504808240313765608_t_real*ylm(141)
    rvylm(1,166)=10.68318467417879543822_t_real*ylm(141)&
         &      -1.27688479613812288973_t_real*ylm(143)
    rvylm(2,166)=-1.27688479613812288973_t_real*ylm(123)&
         &      -10.68318467417879543822_t_real*ylm(125)
    rvylm(3,166)=8.27515926551055676937_t_real*ylm(142)
    rvylm(1,167)=11.20461861291447818860_t_real*ylm(142)&
         &      -0.73720978077448566723_t_real*ylm(144)
    rvylm(2,167)=-0.73720978077448566723_t_real*ylm(122)&
         &      -11.20461861291447818860_t_real*ylm(124)
    rvylm(3,167)=6.91564074808124624060_t_real*ylm(143)
    rvylm(1,168)=11.72603939955857388608_t_real*ylm(143)
    rvylm(2,168)=-11.72603939955857388608_t_real*ylm(123)
    rvylm(3,168)=4.99999999999999999974_t_real*ylm(144)
    rvylm(1,169)=12.24744871391589049060_t_real*ylm(144)
    rvylm(2,169)=-12.24744871391589049060_t_real*ylm(122)
    rvylm(3,169)= 0.0_t_real
    !!
    If(maxL<=12) Goto 1000
    !!
    !! XXXXXXXXXXXXXXXXXXXXXX
    !!
1000 Continue
    !!
    Return
    !!
  End Subroutine calc_real_vec_Ylm_opt
  !!***********************************************************************
  !! 
  !!  ATTENTION: not efficient, just for developement purposes
  !!
  !!
  Subroutine  calc_real_vec_Ylm_and_Ylm(r,maxj,ylm,rvylm)
    !!
    Implicit None  
    !!IO
    Integer(t_int),Intent(in)    :: maxj
    Real(t_real),Intent(in)      :: r(3)
    Real(t_real),Intent(out)     :: ylm(((maxj+2)**2))
    Real(t_real),Intent(out)     :: rvylm(3,2,((maxj+1)**2))
    !!LOCAL
    Character(*),Parameter::nsubfunc="calc_real_vec_Ylm_and_Ylm"
    Integer(t_int) :: l,j,m,k,i,la
    Real(t_real)   :: rd(3),del,sm,delk,delmk,c,cm1,c1,denom
    Real(t_real),Parameter :: sq2=1.41421356237309504880_t_real
    Real(t_real),Parameter :: sq3o10=0.54772255750516611345_t_real
    Real(t_real),Parameter :: sq5o21=0.48795003647426658966_t_real
    Real(t_real),Parameter :: sq8o21=0.61721339984836764104_t_real
    Real(t_real),Parameter :: sq3o7 =0.65465367070797714379_t_real
    Real(t_real),Parameter :: sq5o14=0.59761430466719681997_t_real
    Real(t_real),Parameter :: sq1o42=0.15430334996209191024_t_real
    Real(t_real),Parameter :: sq1o7 =0.37796447300922722721_t_real
    Real(t_real),Parameter :: sq2o7 =0.53452248382484876936_t_real
    !Real(t_real),Parameter :: sq1o14=0.26726124191242438468_t_real
    Real(t_real),Parameter :: sq3o5 =0.77459666924148337703_t_real
    Real(t_real),Parameter :: sq8o15=0.73029674334022148460_t_real
    Real(t_real),Parameter :: sq5o12=0.64549722436790281419_t_real
    Real(t_real),Parameter :: sq7o36=0.44095855184409843174_t_real
    Real(t_real),Parameter :: sq1o5 =0.44721359549995793928_t_real
    Real(t_real),Parameter :: sq2o5 =0.63245553203367586639_t_real
    Real(t_real),Parameter :: sq1o30=0.18257418583505537114_t_real
    Real(t_real),Parameter :: sq1o72=0.11785113019775792069_t_real
    Real(t_real),Parameter :: sq7o18=0.62360956446232356425_t_real
    Real(t_real),Parameter :: sq7o24=0.54006172486732168590_t_real
    Real(t_real),Parameter :: sq1o24=0.20412414523193150816_t_real
    Real(t_real),Parameter :: sq5o18=0.52704627669472988865_t_real
    Real(t_real),Parameter :: sq5o24=0.45643546458763842787_t_real
    Real(t_real),Parameter :: sq15o28=0.73192505471139988451_t_real
    Real(t_real),Parameter :: sq4o7=0.75592894601845445442_t_real
    Real(t_real),Parameter :: sq3o14=0.46291004988627573077_t_real
    Real(t_real),Parameter :: sq3o8=0.61237243569579452454_t_real
    Real(t_real),Parameter :: sq1o56=0.13363062095621219233_t_real
    Real(t_real),Parameter :: sq15o56=0.51754916950676566334_t_real
    Real(t_real),Parameter :: sq3o56=0.23145502494313786537_t_real
    Real(t_real),Parameter :: sq9o55=0.40451991747794525174_t_real
    Real(t_real),Parameter :: sq16o55=0.53935988997059366899_t_real
    Real(t_real),Parameter :: sq21o55=0.61791438065332467076_t_real
    Real(t_real),Parameter :: sq24o55=0.66057825907581636336_t_real
    Real(t_real),Parameter :: sq5o11=0.67419986246324208624_t_real
    Real(t_real),Parameter :: sq3o11=0.52223296786709351452_t_real
    Real(t_real),Parameter :: sq21o110=0.43693144875265148289_t_real
    Real(t_real),Parameter :: sq2o11=0.42640143271122086859_t_real
    Real(t_real),Parameter :: sq3o55=0.23354968324845689126_t_real
    Real(t_real),Parameter :: sq14o55=0.50452497910951301624_t_real
    Real(t_real),Parameter :: sq3o110=0.16514456476895409081_t_real
    Real(t_real),Parameter :: sq18o55=0.57207755354735538926_t_real
    Real(t_real),Parameter :: sq9o22=0.63960214906683130288_t_real
    Real(t_real),Parameter :: sq1o110=0.09534625892455923149_t_real
    !!
    If(maxj>12_t_int)Then
3      Format('''maxl > 12 not implemented! ''')
       Write(errbuf,3) 
       Call errmsg(nmodule,nsubfunc,E_fatalerror,msg=errbuf)
    Endif
    !!
    !!
    Call ylms13(ylm,r(1),r(2),r(3),maxj+1)
    !!
    rvylm(:,:,:)=0.0_t_real
    !!
    Do j=5,maxj
       Do m=-j,j
          !!
          del=0.0_t_real
          If(m==0) del=1.0_t_real
          sm=Real(Sign(1_t_int,m),t_real)
          !!
          Do l=j-1,j+1,2
             If(l<0) Cycle
             !!
             !!
             la=l-j

             denom=1.0_t_real/Sqrt(2.0_t_real*(2*j+1+la)*(2*j+1+2*la))
             !!
             Do k=-1,1,2

                cm1=Sqrt(((j+la*k*m+la)*(j+la*k*m+1+la))*1.0_t_real)*denom
                c1=Sqrt(((j-la*k*m+la)*(j-la*k*m+1+la))*1.0_t_real)*denom

                

                delk=0.0_t_real
                If(m==k) delk=1.0_t_real
                delmk=0.0_t_real
                If(m==-k) delmk=1.0_t_real
                
                i=(3-k)/2
                rd(i)=-sm*(&
                     & (1.0_t_real-delmk)&
                     &*(1.0_t_real+(Sqrt(1.0_t_real+k)-1)*del)&
                     &*cm1*ylm(l*(l+1)+1+k*m+1)&
                     &-k*((1.0_t_real+delk*(sq2-1.0_t_real))&
                     &   *(1.0_t_real-del)+Sqrt(1.0_t_real-k)*del)&
                     &*c1*ylm(l*(l+1)+1+k*m-1)&
                     &)
             Enddo



             c=-la*Sqrt(((2*j-2*m+1+la)*(2*j+2*m+1+la))*1.0_t_real)*denom
             rd(3)=c*ylm(l*(l+1)+1+m)

             rvylm(:,(l-j+3)/2,j*(j+1)+1+m)=rd(:)
          Enddo
       Enddo
    Enddo
    !!
    !! j=0
    !! l=1
    rvylm(1:3,1,1)=0.0_t_real
    rvylm(1,2,1)=-sq3o3*ylm(4)
    rvylm(2,2,1)=-sq3o3*ylm(2)
    rvylm(3,2,1)=-sq3o3*ylm(3)
    If(maxj<=0) goto 1000

    !!j=1
    !!    l=0
    rvylm(1, 1,4)=ylm(1)
    rvylm(2, 1,2)=ylm(1)
    rvylm(3, 1,3)=ylm(1)
    !! j=1
    !!    l=2
    rvylm(1,2,4)=(ylm(7)*sq3o3-ylm(9))*sq3o10
    rvylm(2,2,4)=-ylm(5)*sq3o10
    rvylm(3,2,4)=-ylm(8)*sq3o10
    rvylm(1,2,2)=-ylm(5)*sq3o10
    rvylm(2,2,2)=(ylm(7)*sq3o3+ylm(9))*sq3o10
    rvylm(3,2,2)=-ylm(6)*sq3o10
    rvylm(1,2,3)=-ylm(8)*sq3o10
    rvylm(2,2,3)=-ylm(6)*sq3o10
    rvylm(3,2,3)=-ylm(7)*sq3o10*sq3o3*2.0_t_real
    If(maxj<=1) goto 1000
    !! j=2
    !!    l=1
    rvylm(3, 1,5)=0.0_t_real
    rvylm(3, 1,6)=ylm(2)*sq2o2
    rvylm(3, 1,7)=ylm(3)*sq2*sq3o3
    rvylm(3, 1,8)=ylm(4)*sq2o2
    rvylm(3, 1,9)=0.0_t_real
    !!
    rvylm(1, 1,5)= ylm(2)*sq2o2
    rvylm(1, 1,6)= 0.0_t_real
    rvylm(1, 1,7)=-ylm(4)*sq2o2*sq3o3
    rvylm(1, 1,8)= ylm(3)*sq2o2
    rvylm(1, 1,9)= ylm(4)*sq2o2
    !!
    rvylm(2, 1,5)= ylm(4)*sq2o2
    rvylm(2, 1,6)= ylm(3)*sq2o2
    rvylm(2, 1,7)=-ylm(2)*sq2o2*sq3o3
    rvylm(2, 1,8)= 0.0_t_real
    rvylm(2, 1,9)=-ylm(2)*sq2o2
    !! j=2
    !!    l=3
    !!
    rvylm(1,2,5)= ylm(12)*sq1o42-ylm(10)*sq5o14
    rvylm(1,2,6)=-ylm(11)*sq5o21
    rvylm(1,2,7)=-ylm(14)*sq2o7
    rvylm(1,2,8)=-ylm(15)*sq5o21+ylm(13)*sq1o7
    rvylm(1,2,9)=-ylm(16)*sq5o14+ylm(14)*sq1o42
    !!
    rvylm(2,2,5)= ylm(16)*sq5o14+ylm(14)*sq1o42
    rvylm(2,2,6)= ylm(15)*sq5o21+ylm(13)*sq1o7
    rvylm(2,2,7)=-ylm(12)*sq2o7
    rvylm(2,2,8)=-ylm(11)*sq5o21
    rvylm(2,2,9)=-ylm(10)*sq5o14 -ylm(12)*sq1o42
    !!
    rvylm(3,2,5)=-ylm(11)*sq5o21
    rvylm(3,2,6)=-ylm(12)*sq8o21
    rvylm(3,2,7)=-ylm(13)*sq3o7
    rvylm(3,2,8)=-ylm(14)*sq8o21
    rvylm(3,2,9)=-ylm(15)*sq5o21
    If(maxj<=2) goto 1000
    !!
    !! j=3
    !!    l=2
    !!
    rvylm(1, 1,10)= ylm(5)*sq2o2
    rvylm(1, 1,11)= ylm(6)*sq3o3
    rvylm(1, 1,12)=-ylm(5)*sq1o30
    rvylm(1, 1,13)=-ylm(8)*sq1o5
    rvylm(1, 1,14)=-ylm(9)*sq1o30+ylm(7)*sq2o5
    rvylm(1, 1,15)= ylm(8)*sq3o3
    rvylm(1, 1,16)= ylm(9)*sq2o2
    !!
    rvylm(3, 1,10)= 0.0_t_real
    rvylm(3, 1,11)= ylm(5)*sq3o3
    rvylm(3, 1,12)= ylm(6)*sq8o15
    rvylm(3, 1,13)= ylm(7)*sq3o5
    rvylm(3, 1,14)= ylm(8)*sq8o15
    rvylm(3, 1,15)= ylm(9)*sq3o3
    rvylm(3, 1,16)= 0.0_t_real
    !!
    rvylm(2, 1,10)= ylm(9)*sq2o2
    rvylm(2, 1,11)= ylm(8)*sq3o3
    rvylm(2, 1,12)= ylm(7)*sq2o5 +ylm(9)*sq1o30
    rvylm(2, 1,13)=-ylm(6)*sq1o5
    rvylm(2, 1,14)=-ylm(5)*sq1o30
    rvylm(2, 1,15)=-ylm(6)*sq3o3
    rvylm(2, 1,16)=-ylm(5)*sq2o2
    !!
    !! j=3
    !!    l=4
    !!
    rvylm(1, 2,10)= ylm(19)*sq1o72-ylm(17)*sq7o18
    rvylm(1, 2,11)= ylm(20)*sq1o24-ylm(18)*sq7o24
    rvylm(1, 2,12)=-ylm(19)*sq5o24
    rvylm(1, 2,13)=-ylm(22)*sq5o18
    rvylm(1, 2,14)=-ylm(23)*sq5o24+ylm(21)*sq2o2*sq3o3
    rvylm(1, 2,15)=+ylm(22)*sq1o24-ylm(24)*sq7o24
    rvylm(1, 2,16)=+ylm(23)*sq1o72-ylm(25)*sq7o18
    !!
    rvylm(2, 2,10)=+ylm(25)*sq7o18+ylm(23)*sq1o72
    rvylm(2, 2,11)=+ylm(24)*sq7o24+ylm(22)*sq1o24
    rvylm(2, 2,12)=+ylm(23)*sq5o24+ylm(21)*sq2o2*sq3o3
    rvylm(2, 2,13)=-ylm(20)*sq5o18
    rvylm(2, 2,14)=-ylm(19)*sq5o24
    rvylm(2, 2,15)=-ylm(20)*sq1o24-ylm(18)*sq7o24
    rvylm(2, 2,16)=-ylm(19)*sq1o72-ylm(17)*sq7o18
    !!
    rvylm(3, 2,10)=-ylm(18)*sq7o36
    rvylm(3, 2,11)=-ylm(19)*sq3o3
    rvylm(3, 2,12)=-ylm(20)*sq5o12
    rvylm(3, 2,13)=-ylm(21)*2.0_t_real/3.0_t_real
    rvylm(3, 2,14)=-ylm(22)*sq5o12
    rvylm(3, 2,15)=-ylm(23)*sq3o3
    rvylm(3, 2,16)=-ylm(24)*sq7o36
    If(maxj<=3) Goto 1000
    !!
    !! j=4
    !!    l=3
    !!
    rvylm(1,1,17)= ylm(10)*sq2o2
    rvylm(1,1,18)= ylm(11)*sq3o8
    rvylm(1,1,19)= ylm(12)*sq15o56 -ylm(10)*sq1o56
    rvylm(1,1,20)=-ylm(11)*sq3o56
    rvylm(1,1,21)=-ylm(14)*sq3o14 
    rvylm(1,1,22)=-ylm(15)*sq3o56 +ylm(13)*sq5o14
    rvylm(1,1,23)=-ylm(16)*sq1o56 +ylm(14)*sq15o56
    rvylm(1,1,24)=ylm(15)*sq3o8
    rvylm(1,1,25)=ylm(16)*sq2o2
    !!
    rvylm(2,1,17)= ylm(16)*sq2o2
    rvylm(2,1,18)= ylm(15)*sq3o8
    rvylm(2,1,19)= ylm(16)*sq1o56 +ylm(14)*sq15o56
    rvylm(2,1,20)= ylm(15)*sq3o56 +ylm(13)*sq5o14
    rvylm(2,1,21)=-ylm(12)*sq3o14
    rvylm(2,1,22)=-ylm(11)*sq3o56
    rvylm(2,1,23)=-ylm(12)*sq15o56 -ylm(10)*sq1o56 
    rvylm(2,1,24)=-ylm(11)*sq3o8
    rvylm(2,1,25)=-ylm(10)*sq2o2
    !!
    rvylm(3,1,21-4)=0.0_t_real
    rvylm(3,1,21-3)=ylm(10)*0.5_t_real
    rvylm(3,1,21-2)=ylm(11)*sq3o7
    rvylm(3,1,21-1)=ylm(12)*sq15o28
    rvylm(3,1,21+0)=ylm(13)*sq4o7
    rvylm(3,1,21+1)=ylm(14)*sq15o28
    rvylm(3,1,21+2)=ylm(15)*sq3o7
    rvylm(3,1,21+3)=ylm(16)*0.5_t_real
    rvylm(3,1,21+4)=0.0_t_real
    !!
    !! j=4
    !!    l=5
    !!
    rvylm(1,2,17)= ylm(28)*sq1o110 -ylm(26)*sq9o22
    rvylm(1,2,18)= ylm(29)*sq3o110 -ylm(27)*sq18o55
    rvylm(1,2,19)= ylm(30)*sq3o55 -ylm(28)*sq14o55
    rvylm(1,2,20)=-ylm(29)*sq21o110
    rvylm(1,2,21)=-ylm(32)*sq3o11
    rvylm(1,2,22)=-ylm(33)*sq21o110 +ylm(31)*sq2o11
    rvylm(1,2,23)=-ylm(34)*sq14o55 +ylm(32)*sq3o55
    rvylm(1,2,24)=-ylm(35)*sq18o55 +ylm(33)*sq3o110
    rvylm(1,2,25)=-ylm(36)*sq9o22 +ylm(34)*sq1o110
    !!
    rvylm(2,2,17)= ylm(36)*sq9o22 +ylm(34)*sq1o110
    rvylm(2,2,18)=ylm(35)*sq18o55 +ylm(33)*sq3o110
    rvylm(2,2,19)= ylm(34)*sq14o55 +ylm(32)*sq3o55
    rvylm(2,2,20)= ylm(33)*sq21o110 +ylm(31)*sq2o11
    rvylm(2,2,21)=-ylm(30)*sq3o11
    rvylm(2,2,22)=-ylm(29)*sq21o110 
    rvylm(2,2,23)=-ylm(30)*sq3o55 -ylm(28)*sq14o55
    rvylm(2,2,24)=-ylm(29)*sq3o110 -ylm(27)*sq18o55
    rvylm(2,2,25)=-ylm(28)*sq1o110 -ylm(26)*sq9o22
    !!
    rvylm(3,2,17)=-ylm(27)*sq9o55
    rvylm(3,2,18)=-ylm(28)*sq16o55
    rvylm(3,2,19)=-ylm(29)*sq21o55
    rvylm(3,2,20)=-ylm(30)*sq24o55
    rvylm(3,2,21)=-ylm(31)*sq5o11
    rvylm(3,2,22)=-ylm(32)*sq24o55
    rvylm(3,2,23)=-ylm(33)*sq21o55
    rvylm(3,2,24)=-ylm(34)*sq16o55
    rvylm(3,2,25)=-ylm(35)*sq9o55
    !!
1000 Continue
    !!
    Return
    !!
  End Subroutine calc_real_vec_Ylm_and_Ylm
  !!***********************************************************************
  !! 
  !!  ATTENTION: not efficient, just for developement purposes
  !!
  !!
  Subroutine  calc_real_vec_Ylm(r,maxj,rvylm)
    !!
    Implicit None  
    !!IO
    Integer(t_int),Intent(in)    :: maxj
    Real(t_real),Intent(in)      :: r(3)
    Real(t_real),Intent(out)     :: rvylm(3,2,((maxj+1)**2))
    !!LOCAL
    Character(*),Parameter::nsubfunc="calc_real_vec_Ylm"
    Integer(t_int) :: l,j,m,k,i,la
    Real(t_real)   :: ylm((maxj+2)**2),rd(3),del,sm,delk,delmk,c,cm1,c1,denom
    Real(t_real),Parameter :: sq2=1.41421356237309504880_t_real
    Real(t_real),Parameter :: sq3o10=0.54772255750516611345_t_real
    Real(t_real),Parameter :: sq5o21=0.48795003647426658966_t_real
    Real(t_real),Parameter :: sq8o21=0.61721339984836764104_t_real
    Real(t_real),Parameter :: sq3o7 =0.65465367070797714379_t_real
    Real(t_real),Parameter :: sq5o14=0.59761430466719681997_t_real
    Real(t_real),Parameter :: sq1o42=0.15430334996209191024_t_real
    Real(t_real),Parameter :: sq1o7 =0.37796447300922722721_t_real
    Real(t_real),Parameter :: sq2o7 =0.53452248382484876936_t_real
    !Real(t_real),Parameter :: sq1o14=0.26726124191242438468_t_real
    Real(t_real),Parameter :: sq3o5 =0.77459666924148337703_t_real
    Real(t_real),Parameter :: sq8o15=0.73029674334022148460_t_real
    Real(t_real),Parameter :: sq5o12=0.64549722436790281419_t_real
    Real(t_real),Parameter :: sq7o36=0.44095855184409843174_t_real
    Real(t_real),Parameter :: sq1o5 =0.44721359549995793928_t_real
    Real(t_real),Parameter :: sq2o5 =0.63245553203367586639_t_real
    Real(t_real),Parameter :: sq1o30=0.18257418583505537114_t_real
    Real(t_real),Parameter :: sq1o72=0.11785113019775792069_t_real
    Real(t_real),Parameter :: sq7o18=0.62360956446232356425_t_real
    Real(t_real),Parameter :: sq7o24=0.54006172486732168590_t_real
    Real(t_real),Parameter :: sq1o24=0.20412414523193150816_t_real
    Real(t_real),Parameter :: sq5o18=0.52704627669472988865_t_real
    Real(t_real),Parameter :: sq5o24=0.45643546458763842787_t_real
    Real(t_real),Parameter :: sq15o28=0.73192505471139988451_t_real
    Real(t_real),Parameter :: sq4o7=0.75592894601845445442_t_real
    Real(t_real),Parameter :: sq3o14=0.46291004988627573077_t_real
    Real(t_real),Parameter :: sq3o8=0.61237243569579452454_t_real
    Real(t_real),Parameter :: sq1o56=0.13363062095621219233_t_real
    Real(t_real),Parameter :: sq15o56=0.51754916950676566334_t_real
    Real(t_real),Parameter :: sq3o56=0.23145502494313786537_t_real
    Real(t_real),Parameter :: sq9o55=0.40451991747794525174_t_real
    Real(t_real),Parameter :: sq16o55=0.53935988997059366899_t_real
    Real(t_real),Parameter :: sq21o55=0.61791438065332467076_t_real
    Real(t_real),Parameter :: sq24o55=0.66057825907581636336_t_real
    Real(t_real),Parameter :: sq5o11=0.67419986246324208624_t_real
    Real(t_real),Parameter :: sq3o11=0.52223296786709351452_t_real
    Real(t_real),Parameter :: sq21o110=0.43693144875265148289_t_real
    Real(t_real),Parameter :: sq2o11=0.42640143271122086859_t_real
    Real(t_real),Parameter :: sq3o55=0.23354968324845689126_t_real
    Real(t_real),Parameter :: sq14o55=0.50452497910951301624_t_real
    Real(t_real),Parameter :: sq3o110=0.16514456476895409081_t_real
    Real(t_real),Parameter :: sq18o55=0.57207755354735538926_t_real
    Real(t_real),Parameter :: sq9o22=0.63960214906683130288_t_real
    Real(t_real),Parameter :: sq1o110=0.09534625892455923149_t_real
    !!
    If(maxj>12_t_int)Then
3      Format('''maxl > 12 not implemented! ''')
       Write(errbuf,3) 
       Call errmsg(nmodule,nsubfunc,E_fatalerror,msg=errbuf)
    Endif
    !!
    !!
    Call ylms13(ylm,r(1),r(2),r(3),maxj+1)
    !!
    rvylm(:,:,:)=0.0_t_real
    !!
    Do j=5,maxj
       Do m=-j,j
          !!
          del=0.0_t_real
          If(m==0) del=1.0_t_real
          sm=Real(Sign(1_t_int,m),t_real)
          !!
          Do l=j-1,j+1,2
             If(l<0) Cycle
             !!
             !!
             la=l-j

             denom=1.0_t_real/Sqrt(2.0_t_real*(2*j+1+la)*(2*j+1+2*la))
             !!
             Do k=-1,1,2

                cm1=Sqrt(((j+la*k*m+la)*(j+la*k*m+1+la))*1.0_t_real)*denom
                c1=Sqrt(((j-la*k*m+la)*(j-la*k*m+1+la))*1.0_t_real)*denom

                

                delk=0.0_t_real
                If(m==k) delk=1.0_t_real
                delmk=0.0_t_real
                If(m==-k) delmk=1.0_t_real
                
                i=(3-k)/2
                rd(i)=-sm*(&
                     & (1.0_t_real-delmk)&
                     &*(1.0_t_real+(Sqrt(1.0_t_real+k)-1)*del)&
                     &*cm1*ylm(l*(l+1)+1+k*m+1)&
                     &-k*((1.0_t_real+delk*(sq2-1.0_t_real))&
                     &   *(1.0_t_real-del)+Sqrt(1.0_t_real-k)*del)&
                     &*c1*ylm(l*(l+1)+1+k*m-1)&
                     &)
             Enddo



             c=-la*Sqrt(((2*j-2*m+1+la)*(2*j+2*m+1+la))*1.0_t_real)*denom
             rd(3)=c*ylm(l*(l+1)+1+m)

             rvylm(:,(l-j+3)/2,j*(j+1)+1+m)=rd(:)
          Enddo
       Enddo
    Enddo
    !!
    !! j=0
    !! l=1
    rvylm(1:3,1,1)=0.0_t_real
    rvylm(1,2,1)=-sq3o3*ylm(4)
    rvylm(2,2,1)=-sq3o3*ylm(2)
    rvylm(3,2,1)=-sq3o3*ylm(3)
    If(maxj<=0) goto 1000

    !!j=1
    !!    l=0
    rvylm(1, 1,4)=ylm(1)
    rvylm(2, 1,2)=ylm(1)
    rvylm(3, 1,3)=ylm(1)
    !! j=1
    !!    l=2
    rvylm(1,2,4)=(ylm(7)*sq3o3-ylm(9))*sq3o10
    rvylm(2,2,4)=-ylm(5)*sq3o10
    rvylm(3,2,4)=-ylm(8)*sq3o10
    rvylm(1,2,2)=-ylm(5)*sq3o10
    rvylm(2,2,2)=(ylm(7)*sq3o3+ylm(9))*sq3o10
    rvylm(3,2,2)=-ylm(6)*sq3o10
    rvylm(1,2,3)=-ylm(8)*sq3o10
    rvylm(2,2,3)=-ylm(6)*sq3o10
    rvylm(3,2,3)=-ylm(7)*sq3o10*sq3o3*2.0_t_real
    If(maxj<=1) goto 1000
    !! j=2
    !!    l=1
    rvylm(3, 1,5)=0.0_t_real
    rvylm(3, 1,6)=ylm(2)*sq2o2
    rvylm(3, 1,7)=ylm(3)*sq2*sq3o3
    rvylm(3, 1,8)=ylm(4)*sq2o2
    rvylm(3, 1,9)=0.0_t_real
    !!
    rvylm(1, 1,5)= ylm(2)*sq2o2
    rvylm(1, 1,6)= 0.0_t_real
    rvylm(1, 1,7)=-ylm(4)*sq2o2*sq3o3
    rvylm(1, 1,8)= ylm(3)*sq2o2
    rvylm(1, 1,9)= ylm(4)*sq2o2
    !!
    rvylm(2, 1,5)= ylm(4)*sq2o2
    rvylm(2, 1,6)= ylm(3)*sq2o2
    rvylm(2, 1,7)=-ylm(2)*sq2o2*sq3o3
    rvylm(2, 1,8)= 0.0_t_real
    rvylm(2, 1,9)=-ylm(2)*sq2o2
    !! j=2
    !!    l=3
    !!
    rvylm(1,2,5)= ylm(12)*sq1o42-ylm(10)*sq5o14
    rvylm(1,2,6)=-ylm(11)*sq5o21
    rvylm(1,2,7)=-ylm(14)*sq2o7
    rvylm(1,2,8)=-ylm(15)*sq5o21+ylm(13)*sq1o7
    rvylm(1,2,9)=-ylm(16)*sq5o14+ylm(14)*sq1o42
    !!
    rvylm(2,2,5)= ylm(16)*sq5o14+ylm(14)*sq1o42
    rvylm(2,2,6)= ylm(15)*sq5o21+ylm(13)*sq1o7
    rvylm(2,2,7)=-ylm(12)*sq2o7
    rvylm(2,2,8)=-ylm(11)*sq5o21
    rvylm(2,2,9)=-ylm(10)*sq5o14 -ylm(12)*sq1o42
    !!
    rvylm(3,2,5)=-ylm(11)*sq5o21
    rvylm(3,2,6)=-ylm(12)*sq8o21
    rvylm(3,2,7)=-ylm(13)*sq3o7
    rvylm(3,2,8)=-ylm(14)*sq8o21
    rvylm(3,2,9)=-ylm(15)*sq5o21
    If(maxj<=2) goto 1000
    !!
    !! j=3
    !!    l=2
    !!
    rvylm(1, 1,10)= ylm(5)*sq2o2
    rvylm(1, 1,11)= ylm(6)*sq3o3
    rvylm(1, 1,12)=-ylm(5)*sq1o30
    rvylm(1, 1,13)=-ylm(8)*sq1o5
    rvylm(1, 1,14)=-ylm(9)*sq1o30+ylm(7)*sq2o5
    rvylm(1, 1,15)= ylm(8)*sq3o3
    rvylm(1, 1,16)= ylm(9)*sq2o2
    !!
    rvylm(3, 1,10)= 0.0_t_real
    rvylm(3, 1,11)= ylm(5)*sq3o3
    rvylm(3, 1,12)= ylm(6)*sq8o15
    rvylm(3, 1,13)= ylm(7)*sq3o5
    rvylm(3, 1,14)= ylm(8)*sq8o15
    rvylm(3, 1,15)= ylm(9)*sq3o3
    rvylm(3, 1,16)= 0.0_t_real
    !!
    rvylm(2, 1,10)= ylm(9)*sq2o2
    rvylm(2, 1,11)= ylm(8)*sq3o3
    rvylm(2, 1,12)= ylm(7)*sq2o5 +ylm(9)*sq1o30
    rvylm(2, 1,13)=-ylm(6)*sq1o5
    rvylm(2, 1,14)=-ylm(5)*sq1o30
    rvylm(2, 1,15)=-ylm(6)*sq3o3
    rvylm(2, 1,16)=-ylm(5)*sq2o2
    !!
    !! j=3
    !!    l=4
    !!
    rvylm(1, 2,10)= ylm(19)*sq1o72-ylm(17)*sq7o18
    rvylm(1, 2,11)= ylm(20)*sq1o24-ylm(18)*sq7o24
    rvylm(1, 2,12)=-ylm(19)*sq5o24
    rvylm(1, 2,13)=-ylm(22)*sq5o18
    rvylm(1, 2,14)=-ylm(23)*sq5o24+ylm(21)*sq2o2*sq3o3
    rvylm(1, 2,15)=+ylm(22)*sq1o24-ylm(24)*sq7o24
    rvylm(1, 2,16)=+ylm(23)*sq1o72-ylm(25)*sq7o18
    !!
    rvylm(2, 2,10)=+ylm(25)*sq7o18+ylm(23)*sq1o72
    rvylm(2, 2,11)=+ylm(24)*sq7o24+ylm(22)*sq1o24
    rvylm(2, 2,12)=+ylm(23)*sq5o24+ylm(21)*sq2o2*sq3o3
    rvylm(2, 2,13)=-ylm(20)*sq5o18
    rvylm(2, 2,14)=-ylm(19)*sq5o24
    rvylm(2, 2,15)=-ylm(20)*sq1o24-ylm(18)*sq7o24
    rvylm(2, 2,16)=-ylm(19)*sq1o72-ylm(17)*sq7o18
    !!
    rvylm(3, 2,10)=-ylm(18)*sq7o36
    rvylm(3, 2,11)=-ylm(19)*sq3o3
    rvylm(3, 2,12)=-ylm(20)*sq5o12
    rvylm(3, 2,13)=-ylm(21)*2.0_t_real/3.0_t_real
    rvylm(3, 2,14)=-ylm(22)*sq5o12
    rvylm(3, 2,15)=-ylm(23)*sq3o3
    rvylm(3, 2,16)=-ylm(24)*sq7o36
    If(maxj<=3) Goto 1000
    !!
    !! j=4
    !!    l=3
    !!
    rvylm(1,1,17)= ylm(10)*sq2o2
    rvylm(1,1,18)= ylm(11)*sq3o8
    rvylm(1,1,19)= ylm(12)*sq15o56 -ylm(10)*sq1o56
    rvylm(1,1,20)=-ylm(11)*sq3o56
    rvylm(1,1,21)=-ylm(14)*sq3o14 
    rvylm(1,1,22)=-ylm(15)*sq3o56 +ylm(13)*sq5o14
    rvylm(1,1,23)=-ylm(16)*sq1o56 +ylm(14)*sq15o56
    rvylm(1,1,24)=ylm(15)*sq3o8
    rvylm(1,1,25)=ylm(16)*sq2o2
    !!
    rvylm(2,1,17)= ylm(16)*sq2o2
    rvylm(2,1,18)= ylm(15)*sq3o8
    rvylm(2,1,19)= ylm(16)*sq1o56 +ylm(14)*sq15o56
    rvylm(2,1,20)= ylm(15)*sq3o56 +ylm(13)*sq5o14
    rvylm(2,1,21)=-ylm(12)*sq3o14
    rvylm(2,1,22)=-ylm(11)*sq3o56
    rvylm(2,1,23)=-ylm(12)*sq15o56 -ylm(10)*sq1o56 
    rvylm(2,1,24)=-ylm(11)*sq3o8
    rvylm(2,1,25)=-ylm(10)*sq2o2
    !!
    rvylm(3,1,21-4)=0.0_t_real
    rvylm(3,1,21-3)=ylm(10)*0.5_t_real
    rvylm(3,1,21-2)=ylm(11)*sq3o7
    rvylm(3,1,21-1)=ylm(12)*sq15o28
    rvylm(3,1,21+0)=ylm(13)*sq4o7
    rvylm(3,1,21+1)=ylm(14)*sq15o28
    rvylm(3,1,21+2)=ylm(15)*sq3o7
    rvylm(3,1,21+3)=ylm(16)*0.5_t_real
    rvylm(3,1,21+4)=0.0_t_real
    !!
    !! j=4
    !!    l=5
    !!
    rvylm(1,2,17)= ylm(28)*sq1o110 -ylm(26)*sq9o22
    rvylm(1,2,18)= ylm(29)*sq3o110 -ylm(27)*sq18o55
    rvylm(1,2,19)= ylm(30)*sq3o55 -ylm(28)*sq14o55
    rvylm(1,2,20)=-ylm(29)*sq21o110
    rvylm(1,2,21)=-ylm(32)*sq3o11
    rvylm(1,2,22)=-ylm(33)*sq21o110 +ylm(31)*sq2o11
    rvylm(1,2,23)=-ylm(34)*sq14o55 +ylm(32)*sq3o55
    rvylm(1,2,24)=-ylm(35)*sq18o55 +ylm(33)*sq3o110
    rvylm(1,2,25)=-ylm(36)*sq9o22 +ylm(34)*sq1o110
    !!
    rvylm(2,2,17)= ylm(36)*sq9o22 +ylm(34)*sq1o110
    rvylm(2,2,18)=ylm(35)*sq18o55 +ylm(33)*sq3o110
    rvylm(2,2,19)= ylm(34)*sq14o55 +ylm(32)*sq3o55
    rvylm(2,2,20)= ylm(33)*sq21o110 +ylm(31)*sq2o11
    rvylm(2,2,21)=-ylm(30)*sq3o11
    rvylm(2,2,22)=-ylm(29)*sq21o110 
    rvylm(2,2,23)=-ylm(30)*sq3o55 -ylm(28)*sq14o55
    rvylm(2,2,24)=-ylm(29)*sq3o110 -ylm(27)*sq18o55
    rvylm(2,2,25)=-ylm(28)*sq1o110 -ylm(26)*sq9o22
    !!
    rvylm(3,2,17)=-ylm(27)*sq9o55
    rvylm(3,2,18)=-ylm(28)*sq16o55
    rvylm(3,2,19)=-ylm(29)*sq21o55
    rvylm(3,2,20)=-ylm(30)*sq24o55
    rvylm(3,2,21)=-ylm(31)*sq5o11
    rvylm(3,2,22)=-ylm(32)*sq24o55
    rvylm(3,2,23)=-ylm(33)*sq21o55
    rvylm(3,2,24)=-ylm(34)*sq16o55
    rvylm(3,2,25)=-ylm(35)*sq9o55
    !!
1000 Continue
    !!
    Return
    !!
  End Subroutine calc_real_vec_Ylm
  !!***********************************************************************
  !! 
  !!  ATTENTION: not efficient, just for developement purposes
  !!
  !! \vec{Y}_{jm}^l=\sum_{s=-1}^{1}C^{jm}_{lm-s,1s}\cal{Y}_{lm-s}\vec{e}_{s}
  !! for j=0..maxj, l=j-1..j+1, 0<= |m| <=j
  !!
#ifdef my_compile_unused
  Subroutine  calc_spherical_vec_Ylm(r,maxj,cvylm)
    !!
    Implicit None  
    !!IO
    Integer(t_int),Intent(in)    :: maxj
    Real(t_real),Intent(in)      :: r(3)
    Complex(t_real),Intent(out)  :: cvylm(3,-1:1,((maxj+1)**2))
    !!LOCAL
    Integer(t_int) :: l,j,is,m
    Real(t_real)   :: theta,phi,ra
    Complex(t_real):: cyp((maxj+2)**2),esph(3,-1:1),cd(3)
    !!
    Complex(t_real),Parameter:: c0=(0.0_t_real,0.0_t_real)
    Complex(t_real),Parameter:: c1=(1.0_t_real,0.0_t_real)
    Complex(t_real),Parameter:: ci=(0.0_t_real,1.0_t_real)
    !!
    Call get_spherical_coordinates(r,theta,phi,ra)
    !! spin=1 eigenfunctions == spherical basis vectors
    esph(:, 1)=-sq2o2*(/ c1, ci, c0 /)
    esph(:, 0)=       (/ c0, c0, c1 /)
    esph(:,-1)= sq2o2*(/ c1,-ci, c0 /)
    !!
    Call cxylm(maxj+1,cyp,theta,phi)
    !!
    cvylm(:,:,:)=c0
    Do j=0,maxj
       Do m=-j,j
          Do l=Max(0_t_int,j-1),j+1
             cd(:)=c0
             Do is=-1,1
                !Print *,l,m,lm,is,clebshS1(l,m,lm,is)
                cd(:)=cd(:)&
                     &+clebshS1(j,m,l,is)*cyp(l*(l+1)+1+m-is)*esph(:,is) 
                !Print *,j,l,m-is,(l*(l+1)+1+m-is),(maxj+2)**2
             Enddo
             cvylm(:,l-j,j*(j+1)+1+m)=cd(:)
          Enddo
       Enddo
    Enddo
    !!
    !!
    Return
    !!
  End Subroutine calc_spherical_vec_Ylm
#endif
  !!***********************************************************************
  !! 
  !!  returns spin=1 Clebsch Gordans C^{j,m}_{l,m-ms,1,ms}
  !!  ATTENTION: not efficient, just for developement purposes
  !!
  Real(t_real) Function clebshS1(j,m,l,ms)
    !!
    Implicit None  
    !!IO
    Integer(t_int),Intent(in)    :: j,m,l,ms
    !!LOCAL
    Real(t_real)   :: denom
    !!
    !!
    clebshS1=0.0_t_real
    !!
    If(m<-j .Or. m> j .Or. l<0) Return
    !!
    If(j==l+1)Then
       denom=1.0_t_real/((2*j-1)*2*j)
       If(ms==1) Then
          clebshS1=Sqrt((j+m-1)*(j+m)*denom)
       Else If(ms==0) Then
          clebshS1=Sqrt(2*(j-m)*(j+m)*denom)
       Else If(ms==-1) Then
          clebshS1=Sqrt((j-m-1)*(j-m)*denom)
       Endif
    Else If(j==l)Then
       denom=1.0_t_real/((j+1)*2*j)
       If(ms==1) Then
          clebshS1=-Sqrt((j-m+1)*(j+m)*denom)
       Else If(ms==0) Then
          clebshS1=m*Sqrt(2*denom)
       Else If(ms==-1) Then
          clebshS1=Sqrt((j+m+1)*(j-m)*denom)
       Endif
    Else If(j==l-1.And.j>=0)Then
       denom=1.0_t_real/((2*j+2)*(2*j+3))
       If(ms==1) Then
          clebshS1=Sqrt((j-m+1)*(j-m+2)*denom)
       Else If(ms==0) Then
          clebshS1=-Sqrt(2*(j+m+1)*(j-m+1)*denom)
       Else If(ms==-1) Then
          clebshS1=Sqrt((j+m+1)*(j+m+2)*denom)
       Endif
    Endif
    !!
    Return
    !!
  End Function clebshS1
  !!***********************************************************************
  !! the clebsch Gordon coefficients for the real vector harmoncis
  !!
  !! R_{LM}^(l+lam)= \vec{e}_k \sum_{m_1=-1}^1 C^{lam k}_{LMm_1} 
  !!                                         Y_{L+lam,ki*M+m_1}
  !! R_{lm}^{l+\lambda}=\sum_{i}\sum_{M=-1}^{1}C_{lmM}^{\lambda i}
  !!                            Y_{l+\lambda,k_{i}m+M}
  !!                            \delta_{\left|k_{i}m+M\right|\leq l+\lambda}
  !!                            \vec{e}_{i}^{c}
  !!                  k_{i}=2\left|i-2\right|-1
  !!                  k_{1,2,3}=1,-1,1
  !! with \lambda=-1,+1 mapped to ilam=1,2
  !! ccc(k,ilam,m1+2,ilm)  ilm=L*L+L+1+M
  !! beware that one needs Ylms up to maxL+1 since l+lam can be l+1
  !! and for quad_ylms up to maxL+2
  !!
#ifdef my_compile_unused
  Subroutine  calc_real_vec_ccoefs(ccc,maxL,clpm)
    !!
    !!
    Implicit None  
    !!IO
    Integer(t_int),Intent(in)    :: maxL
    Real(t_real),Intent(out)     :: clpm(2,0:maxL)
    Real(t_real),Intent(out)     :: ccc(3,2,1:3,((maxL+1)**2))
    !!LOCAL
    !Character(*),Parameter::nsubfunc="calc_real_vec_ccoefs"
    Integer(t_int) :: L,M,ilm,la,ilam,ilam0
    Real(t_real)   :: sm,c1,cm1,denom,deg
    Real(t_real),Parameter :: sq2=1.41421356237309504880_t_real
    !!
    !!
    Do l=0,maxL
       deg=2.0_t_real*l+1.0_t_real
       clpm(1,l)= Sqrt(Real(l,t_real)/deg)
       clpm(2,l)=-Sqrt(Real(l+1,t_real)/deg)
    Enddo
    !!
    !!
    ccc(:,:,:,:)=0.0_t_real
    !!
    Do L=14,maxL
       Do M=-L,L
          !!
          ilm=L*L+L+1+M
          !!
          sm=Real(Sign(1_t_int,M),t_real)
          !!
          ilam0=1
          If(L==0)ilam0=2
          Do ilam=ilam0,2
             !!
             !!
             la=2*ilam-3
             !!

             denom=1.0_t_real/Sqrt(2.0_t_real*(2*L+1+la)*(2*L+1+2*la))
             !!
             !!
             
             cm1=Sqrt(((L-la*M+la)*(L-la*M+1+la))*1.0_t_real)*denom
             c1=Sqrt(((L+la*M+la)*(L+la*M+1+la))*1.0_t_real)*denom
             

             If(M /= 0 .And. M /= 1) ccc(2,ilam,3,ilm)=-sm*cm1
             ccc(2,ilam,1,ilm)=-sm*c1
             If(M==-1 .Or. M==0) ccc(2,ilam,1,ilm)=ccc(2,ilam,1,ilm)*sq2
             

             If(M /= 0) ccc(1,ilam,1,ilm)=sm*cm1
             If(M == 1) ccc(1,ilam,1,ilm)=ccc(1,ilam,1,ilm)*sq2
             If(M /=-1) ccc(1,ilam,3,ilm)=-sm*c1
             If(M == 0) ccc(1,ilam,3,ilm)=ccc(1,ilam,3,ilm)*sq2
                          
             
             
             
             ccc(3,ilam,2,ilm)=&
                  &-la*Sqrt(((2*L-2*M+1+la)*(2*L+2*M+1+la))*1.0_t_real)*denom

          Enddo
       Enddo
    Enddo
    !!
    !!..............................................
    !! L=0
    !!..............................................
    !! M=0 L+1
    ccc(1,2,3,1)=-0.57735026918962576450_t_real
    ccc(2,2,1,1)=-0.57735026918962576450_t_real
    ccc(3,2,2,1)=-0.57735026918962576450_t_real
    !!
    If(maxL<=0) Return
    !!
    !!..............................................
    !! L=1
    !!..............................................
    !!
    !! M=-1
    !! L-1
    ccc(2,1,1,2)= 1.0_t_real
    !!
    !! L+1
    ccc(1,2,1,2)=-0.54772255750516611345_t_real
    ccc(2,2,1,2)= 0.31622776601683793319_t_real
    ccc(2,2,3,2)= 0.54772255750516611345_t_real
    ccc(3,2,2,2)=-0.54772255750516611345_t_real
    !!
    !!
    !! M==0
    !! L-1
    ccc(3,1, 2,3)=1.0_t_real
    !! L+1
    ccc(1,2,3,3)=-0.54772255750516611344_t_real
    ccc(2,2,1,3)=-0.54772255750516611344_t_real
    ccc(3,2,2,3)=-0.63245553203367586640_t_real
    !!
    !! M=1
    !! L-1
    ccc(1,1,1,4)=1.0_t_real
    !! L+1
    ccc(1,2,1,4)= 0.31622776601683793319_t_real
    ccc(1,2,3,4)=-0.54772255750516611345_t_real
    ccc(2,2,1,4)=-0.54772255750516611345_t_real
    ccc(3,2,2,4)=-0.54772255750516611345_t_real
    !!
    If(maxL<=1) Return
    !!
    !!..............................................
    !! L=2
    !!..............................................
    !!
    !!
    !! M=-2
    !! L-1
    ccc(1,1,3,5)= 0.70710678118654752440_t_real
    ccc(2,1,1,5)= 0.70710678118654752440_t_real
    !!
    !! L-1
    ccc(1,2,1,5)=-0.59761430466719681998_t_real
    ccc(1,2,3,5)= 0.15430334996209191026_t_real
    ccc(2,2,1,5)= 0.15430334996209191026_t_real
    ccc(2,2,3,5)= 0.59761430466719681998_t_real
    ccc(3,2,2,5)=-0.48795003647426658967_t_real
    !!
    !! M=-1
    !! L-1
    ccc(2,1,1,6)= 0.70710678118654752438_t_real
    ccc(3,1,2,6)= 0.70710678118654752440_t_real
    !! L+1
    ccc(1,2,1,6)=-0.48795003647426658967_t_real
    ccc(2,2,1,6)= 0.37796447300922722720_t_real
    ccc(2,2,3,6)= 0.48795003647426658967_t_real
    ccc(3,2,2,6)=-0.61721339984836764104_t_real
    !!
    !! M=0
    !! L-1
    ccc(1,1,3,7)=-0.40824829046386301635_t_real
    ccc(2,1,1,7)=-0.40824829046386301635_t_real
    ccc(3,1,2,7)= 0.81649658092772603273_t_real
    !! L+1
    ccc(2,2,1,7)=-0.53452248382484876936_t_real
    ccc(1,2,3,7)=-0.53452248382484876936_t_real
    ccc(3,2,2,7)=-0.65465367070797714379_t_real
    !!
    !!
    !!
    !! M=1
    !! L-1
    ccc(1,1,1,8)= 0.70710678118654752438_t_real
    ccc(3,1,2,8)= 0.70710678118654752440_t_real
    !!
    !! L+1
    ccc(1,2,1,8)= 0.37796447300922722720_t_real
    ccc(1,2,3,8)=-0.48795003647426658967_t_real
    ccc(2,2,1,8)=-0.48795003647426658967_t_real
    ccc(3,2,2,8)=-0.61721339984836764104_t_real
    !!
    !!
    !! M=2
    !! L-1
    ccc(1,1,1,9)= 0.70710678118654752440_t_real
    ccc(2,1,3,9)=-0.70710678118654752440_t_real
    !!
    !! L+1
    ccc(1,2,1,9)= 0.15430334996209191026_t_real
    ccc(1,2,3,9)=-0.59761430466719681998_t_real
    ccc(2,2,1,9)=-0.59761430466719681998_t_real
    ccc(2,2,3,9)=-0.15430334996209191026_t_real
    ccc(3,2,2,9)=-0.48795003647426658967_t_real
    !!
    If(maxL<=2) Return
    !!
    !!..............................................
    !! L=3
    !!..............................................
    !!
    !!
    !! M=-3
    !! L-1
    ccc(1,1,3,10)= 0.70710678118654752440_t_real
    ccc(2,1,1,10)= 0.70710678118654752440_t_real
    !! L+1
    ccc(1,2,1,10)=-0.62360956446232356426_t_real
    ccc(1,2,3,10)= 0.11785113019775792073_t_real
    ccc(2,2,1,10)= 0.11785113019775792073_t_real
    ccc(2,2,3,10)= 0.62360956446232356426_t_real
    ccc(3,2,2,10)=-0.44095855184409843175_t_real
    !! M=-2
    !! L-1
    ccc(1,1,3,11)= 0.57735026918962576450_t_real
    ccc(2,1,1,11)= 0.57735026918962576450_t_real
    ccc(3,1,2,11)= 0.57735026918962576450_t_real
    !!
    !! L+1
    ccc(1,2,1,11)=-0.54006172486732168591_t_real
    ccc(1,2,3,11)= 0.20412414523193150818_t_real
    ccc(2,2,1,11)= 0.20412414523193150818_t_real
    ccc(2,2,3,11)= 0.54006172486732168591_t_real
    ccc(3,2,2,11)=-0.57735026918962576450_t_real
    !!
    !! M-1
    !! L-1
    ccc(1,1,1,12)=-0.18257418583505537115_t_real
    ccc(2,1,1,12)= 0.63245553203367586639_t_real
    ccc(2,1,3,12)= 0.18257418583505537115_t_real
    ccc(3,1,2,12)= 0.73029674334022148460_t_real
    !! L+1
    ccc(1,2,1,12)=-0.45643546458763842788_t_real
    ccc(2,2,1,12)= 0.40824829046386301635_t_real
    ccc(2,2,3,12)= 0.45643546458763842788_t_real
    ccc(3,2,2,12)=-0.64549722436790281419_t_real
    !! M=0
    !! L-1
    ccc(1,1,3,13)=-0.44721359549995793926_t_real
    ccc(2,1,1,13)=-0.44721359549995793926_t_real
    ccc(3,1,2,13)= 0.77459666924148337703_t_real
    !! L+1
    ccc(1,2,3,13)=-0.52704627669472988866_t_real
    ccc(2,2,1,13)=-0.52704627669472988866_t_real
    ccc(3,2,2,13)=-0.66666666666666666666_t_real
    !!
    !! M=1
    !! L-1
    ccc(1,1,1,14)= 0.63245553203367586639_t_real
    ccc(1,1,3,14)=-0.18257418583505537115_t_real
    ccc(2,1,1,14)=-0.18257418583505537115_t_real
    ccc(3,1,2,14)= 0.73029674334022148460_t_real
    !! L+1
    ccc(1,2,1,14)= 0.40824829046386301635_t_real
    ccc(1,2,3,14)=-0.45643546458763842788_t_real
    ccc(2,2,1,14)=-0.45643546458763842788_t_real
    ccc(3,2,2,14)=-0.64549722436790281419_t_real
    !!
    !!
    !! M=2
    !! L-1
    ccc(1,1,1,15)= 0.57735026918962576450_t_real
    ccc(2,1,3,15)=-0.57735026918962576450_t_real
    ccc(3,1,2,15)= 0.57735026918962576450_t_real
    !! L+1
    ccc(1,2,1,15)= 0.20412414523193150818_t_real
    ccc(1,2,3,15)=-0.54006172486732168591_t_real
    ccc(2,2,1,15)=-0.54006172486732168591_t_real
    ccc(2,2,3,15)=-0.20412414523193150818_t_real
    ccc(3,2,2,15)=-0.57735026918962576450_t_real
    !!
    !! M=3
    !! L-1
    ccc(1,1,1,16)= 0.70710678118654752440_t_real
    ccc(2,1,3,16)=-0.70710678118654752440_t_real
    !!
    !! L+1
    ccc(1,2,1,16)= 0.11785113019775792073_t_real
    ccc(1,2,3,16)=-0.62360956446232356426_t_real
    ccc(2,2,1,16)=-0.62360956446232356426_t_real
    ccc(2,2,3,16)=-0.11785113019775792073_t_real
    ccc(3,2,2,16)=-0.44095855184409843175_t_real
    !!
    If(maxL<=3) Return
    !!
    !!..............................................
    !! L=4
    !!..............................................
    ! M=-4
    !! L-1
    ccc(1,1,3,17)= 0.70710678118654752440_t_real
    ccc(2,1,1,17)= 0.70710678118654752440_t_real
    !! L+1
    ccc(1,2,1,17)=-0.63960214906683130289_t_real
    ccc(1,2,3,17)= 0.09534625892455923154_t_real
    ccc(2,2,1,17)= 0.09534625892455923154_t_real
    ccc(2,2,3,17)= 0.63960214906683130289_t_real
    ccc(3,2,2,17)=-0.40451991747794525174_t_real
    !! M=-3
    !! L-1
    ccc(1,1,3,18)= 0.61237243569579452454_t_real
    ccc(2,1,1,18)= 0.61237243569579452454_t_real
    ccc(3,1,2,18)= 0.5_t_real
    !! L+1
    ccc(1,2,1,18)=-0.57207755354735538926_t_real
    ccc(1,2,3,18)= 0.16514456476895409084_t_real
    ccc(2,2,1,18)= 0.16514456476895409084_t_real
    ccc(2,2,3,18)= 0.57207755354735538926_t_real
    ccc(3,2,2,18)=-0.53935988997059366899_t_real
    !! M=-2
    !! L-1
    ccc(1,1,1,19)=-0.13363062095621219234_t_real
    ccc(1,1,3,19)= 0.51754916950676566334_t_real
    ccc(2,1,1,19)= 0.51754916950676566334_t_real
    ccc(2,1,3,19)= 0.13363062095621219234_t_real
    ccc(3,1,2,19)= 0.65465367070797714379_t_real
    !! L+1
    ccc(1,2,1,19)=-0.50452497910951301624_t_real
    ccc(1,2,3,19)= 0.23354968324845689127_t_real
    ccc(2,2,1,19)= 0.23354968324845689127_t_real
    ccc(2,2,3,19)= 0.50452497910951301624_t_real
    ccc(3,2,2,19)=-0.61791438065332467076_t_real
    !! M=-1
    !! L-1
    ccc(1,1,1,20)=-0.23145502494313786539_t_real
    ccc(2,1,1,20)= 0.59761430466719681997_t_real
    ccc(2,1,3,20)= 0.23145502494313786539_t_real
    ccc(3,1,2,20)= 0.73192505471139988451_t_real
    !! L+1
    ccc(1,2,1,20)=-0.43693144875265148289_t_real
    ccc(2,2,1,20)= 0.42640143271122086858_t_real
    ccc(2,2,3,20)= 0.43693144875265148289_t_real
    ccc(3,2,2,20)=-0.66057825907581636336_t_real
    !! M=0
    !! L-1
    ccc(1,1,3,21)=-0.46291004988627573076_t_real
    ccc(2,1,1,21)=-0.46291004988627573076_t_real
    ccc(3,1,2,21)= 0.75592894601845445442_t_real
    !! L+1
    ccc(1,2,3,21)=-0.52223296786709351452_t_real
    ccc(2,2,1,21)=-0.52223296786709351452_t_real
    ccc(3,2,2,21)=-0.67419986246324208624_t_real
    !! M=1
    !! L-1
    ccc(1,1,1,22)= 0.59761430466719681997_t_real
    ccc(1,1,3,22)=-0.23145502494313786539_t_real
    ccc(2,1,1,22)=-0.23145502494313786539_t_real
    ccc(3,1,2,22)= 0.73192505471139988451_t_real
    !! L+1
    ccc(1,2,1,22)= 0.42640143271122086858_t_real
    ccc(1,2,3,22)=-0.43693144875265148289_t_real
    ccc(2,2,1,22)=-0.43693144875265148289_t_real
    ccc(3,2,2,22)=-0.66057825907581636336_t_real
    !!M=2
    !! L-1
    ccc(1,1,1,23)= 0.51754916950676566334_t_real
    ccc(1,1,3,23)=-0.13363062095621219234_t_real
    ccc(2,1,1,23)=-0.13363062095621219234_t_real
    ccc(2,1,3,23)=-0.51754916950676566334_t_real
    ccc(3,1,2,23)= 0.65465367070797714379_t_real
    !! L+1
    ccc(1,2,1,23)= 0.23354968324845689127_t_real
    ccc(1,2,3,23)=-0.50452497910951301624_t_real
    ccc(2,2,1,23)=-0.50452497910951301624_t_real
    ccc(2,2,3,23)=-0.23354968324845689127_t_real
    ccc(3,2,2,23)=-0.61791438065332467076_t_real
    !! M=3
    !! L-1
    ccc(1,1,1,24)= 0.61237243569579452454_t_real
    ccc(2,1,3,24)=-0.61237243569579452454_t_real
    ccc(3,1,2,24)= 0.5_t_real
    !! L+1
    ccc(1,2,1,24)= 0.16514456476895409084_t_real
    ccc(1,2,3,24)=-0.57207755354735538926_t_real
    ccc(2,2,1,24)=-0.57207755354735538926_t_real
    ccc(2,2,3,24)=-0.16514456476895409084_t_real
    ccc(3,2,2,24)=-0.53935988997059366899_t_real
    !! M=4
    !! L+1
    ccc(1,1,1,25)= 0.70710678118654752440_t_real
    ccc(2,1,3,25)=-0.70710678118654752440_t_real
    !! L-1
    ccc(1,2,1,25)= 0.09534625892455923154_t_real
    ccc(1,2,3,25)=-0.63960214906683130289_t_real
    ccc(2,2,1,25)=-0.63960214906683130289_t_real
    ccc(2,2,3,25)=-0.09534625892455923154_t_real
    ccc(3,2,2,25)=-0.40451991747794525174_t_real
    !!
    If(maxL<=4) Return
    !!
    !!..............................................
    !! L=5
    !!..............................................
    !! M=-5
    !! L-1
    ccc(1,1,3,26)= 0.70710678118654752439_t_real
    ccc(2,1,1,26)= 0.70710678118654752439_t_real
    !! L+1
    ccc(1,2,1,26)=-0.65044363558799088069_t_real
    ccc(1,2,3,26)= 0.08006407690254356674_t_real
    ccc(2,2,1,26)= 0.08006407690254356674_t_real
    ccc(2,2,3,26)= 0.65044363558799088069_t_real
    ccc(3,2,2,26)=-0.37553380809940537980_t_real
    !! M=-4
    !! L-1
    ccc(1,1,3,27)= 0.63245553203367586639_t_real
    ccc(2,1,1,27)= 0.63245553203367586639_t_real
    ccc(3,1,2,27)= 0.44721359549995793928_t_real
    !! L+1
    ccc(1,2,1,27)=-0.59377108599535441179_t_real
    ccc(1,2,3,27)= 0.13867504905630728050_t_real
    ccc(2,2,1,27)= 0.13867504905630728050_t_real
    ccc(2,2,3,27)= 0.59377108599535441179_t_real
    ccc(3,2,2,27)=-0.50636968354183330813_t_real
    !! M=-3
    !! L-1
    ccc(1,1,1,28)=-0.10540925533894597773_t_real
    ccc(1,1,3,28)= 0.55777335102271703198_t_real
    ccc(2,1,1,28)= 0.55777335102271703198_t_real
    ccc(2,1,3,28)= 0.10540925533894597773_t_real
    ccc(3,1,2,28)= 0.59628479399994391904_t_real
    !! L+1
    ccc(1,2,1,28)=-0.53708615552957466037_t_real
    ccc(1,2,3,28)= 0.19611613513818403192_t_real
    ccc(2,2,1,28)= 0.19611613513818403192_t_real
    ccc(2,2,3,28)= 0.53708615552957466037_t_real
    ccc(3,2,2,28)=-0.58834840541455209576_t_real
    !! M=-2
    !! L-1
    ccc(1,1,1,29)=-0.18257418583505537115_t_real
    ccc(1,1,3,29)= 0.48304589153964795245_t_real
    ccc(2,1,1,29)= 0.48304589153964795245_t_real
    ccc(2,1,3,29)= 0.18257418583505537115_t_real
    ccc(3,1,2,29)= 0.68313005106397322554_t_real
    !! L+1
    ccc(1,2,1,29)=-0.48038446141526140044_t_real
    ccc(1,2,3,29)= 0.25318484177091665406_t_real
    ccc(2,2,1,29)= 0.25318484177091665406_t_real
    ccc(2,2,3,29)= 0.48038446141526140044_t_real
    ccc(3,2,2,29)=-0.64051261522034853392_t_real
    !! M=-1
    !! L-1
    ccc(1,1,1,30)=-0.25819888974716112567_t_real
    ccc(2,1,1,30)= 0.57735026918962576449_t_real
    ccc(2,1,3,30)= 0.25819888974716112567_t_real
    ccc(3,1,2,30)= 0.73029674334022148460_t_real
    !! L+1
    ccc(1,2,1,30)=-0.42365927286816169388_t_real
    ccc(2,2,1,30)= 0.43852900965351460735_t_real
    ccc(2,2,3,30)= 0.42365927286816169388_t_real
    ccc(3,2,2,30)=-0.66986412705708365760_t_real
    !! M=0
    !! L-1
    ccc(1,1,3,31)=-0.47140452079103168292_t_real
    ccc(2,1,1,31)=-0.47140452079103168292_t_real
    ccc(3,1,2,31)= 0.74535599249992989880_t_real
    !! L+1
    ccc(1,2,3,31)=-0.51887452166277082558_t_real
    ccc(2,2,1,31)=-0.51887452166277082558_t_real
    ccc(3,2,2,31)=-0.67936622048675745036_t_real
    !! M=1
    !! L-1
    ccc(1,1,1,32)= 0.57735026918962576449_t_real
    ccc(1,1,3,32)=-0.25819888974716112567_t_real
    ccc(2,1,1,32)=-0.25819888974716112567_t_real
    ccc(3,1,2,32)= 0.73029674334022148460_t_real
    !! L+1
    ccc(1,2,1,32)= 0.43852900965351460735_t_real
    ccc(1,2,3,32)=-0.42365927286816169388_t_real
    ccc(2,2,1,32)=-0.42365927286816169388_t_real
    ccc(3,2,2,32)=-0.66986412705708365760_t_real
    !! M=2
    !! L-1
    ccc(1,1,1,33)= 0.48304589153964795245_t_real
    ccc(1,1,3,33)=-0.18257418583505537115_t_real
    ccc(2,1,1,33)=-0.18257418583505537115_t_real
    ccc(2,1,3,33)=-0.48304589153964795245_t_real
    ccc(3,1,2,33)= 0.68313005106397322554_t_real
    !! L+1
    ccc(1,2,1,33)= 0.25318484177091665406_t_real
    ccc(1,2,3,33)=-0.48038446141526140044_t_real
    ccc(2,2,1,33)=-0.48038446141526140044_t_real
    ccc(2,2,3,33)=-0.25318484177091665406_t_real
    ccc(3,2,2,33)=-0.64051261522034853392_t_real
    !! M=3
    !! L-1
    ccc(1,1,1,34)= 0.55777335102271703198_t_real
    ccc(1,1,3,34)=-0.10540925533894597773_t_real
    ccc(2,1,1,34)=-0.10540925533894597773_t_real
    ccc(2,1,3,34)=-0.55777335102271703198_t_real
    ccc(3,1,2,34)= 0.59628479399994391904_t_real
    !! L+1
    ccc(1,2,1,34)= 0.19611613513818403192_t_real
    ccc(1,2,3,34)=-0.53708615552957466037_t_real
    ccc(2,2,1,34)=-0.53708615552957466037_t_real
    ccc(2,2,3,34)=-0.19611613513818403192_t_real
    ccc(3,2,2,34)=-0.58834840541455209576_t_real
    !! M=4
    !! L-1
    ccc(1,1,1,35)= 0.63245553203367586639_t_real
    ccc(2,1,3,35)=-0.63245553203367586639_t_real
    ccc(3,1,2,35)= 0.44721359549995793928_t_real
    !! L+1
    ccc(1,2,1,35)= 0.13867504905630728050_t_real
    ccc(1,2,3,35)=-0.59377108599535441179_t_real
    ccc(2,2,1,35)=-0.59377108599535441179_t_real
    ccc(2,2,3,35)=-0.13867504905630728050_t_real
    ccc(3,2,2,35)=-0.50636968354183330813_t_real
    !! M=5
    !! L-1
    ccc(1,1,1,36)= 0.70710678118654752439_t_real
    ccc(2,1,3,36)=-0.70710678118654752439_t_real
    !! L+1
    ccc(1,2,1,36)= 0.08006407690254356674_t_real
    ccc(1,2,3,36)=-0.65044363558799088069_t_real
    ccc(2,2,1,36)=-0.65044363558799088069_t_real
    ccc(2,2,3,36)=-0.08006407690254356674_t_real
    ccc(3,2,2,36)=-0.37553380809940537980_t_real
    !!
    If(maxL<=5) Return
    !!
    !!..............................................
    !! L=6
    !!..............................................
    !! M=-6
    !! L-1
    ccc(1,1,3,37)= 0.70710678118654752431_t_real
    ccc(2,1,1,37)= 0.70710678118654752431_t_real
    !! L+1
    ccc(1,2,1,37)=-0.65828058860438330047_t_real
    ccc(1,2,3,37)= 0.06900655593423542176_t_real
    ccc(2,2,1,37)= 0.06900655593423542176_t_real
    ccc(2,2,3,37)= 0.65828058860438330047_t_real
    ccc(3,2,2,37)=-0.35186577527449839986_t_real
    !! M=-5
    !! L-1
    ccc(1,1,3,38)= 0.64549722436790281412_t_real
    ccc(2,1,1,38)= 0.64549722436790281412_t_real
    ccc(3,1,2,38)= 0.40824829046386301631_t_real
    !! L+1
    ccc(1,2,1,38)=-0.60944940022004404992_t_real
    ccc(1,2,3,38)= 0.11952286093343936397_t_real
    ccc(2,2,1,38)= 0.11952286093343936397_t_real
    ccc(2,2,3,38)= 0.60944940022004404992_t_real
    ccc(3,2,2,38)=-0.47809144373375745591_t_real
    !! M=-4
    !! L-1
    ccc(1,1,1,39)=-0.08703882797784891907_t_real
    ccc(1,1,3,39)= 0.58387420812114222811_t_real
    ccc(2,1,1,39)= 0.58387420812114222811_t_real
    ccc(2,1,3,39)= 0.08703882797784891907_t_real
    ccc(3,1,2,39)= 0.55048188256318030273_t_real
    !! L+1
    ccc(1,2,1,39)=-0.56061191058138809901_t_real
    ccc(1,2,3,39)= 0.16903085094570331547_t_real
    ccc(2,2,1,39)= 0.16903085094570331547_t_real
    ccc(2,2,3,39)= 0.56061191058138809901_t_real
    ccc(3,2,2,39)=-0.56061191058138809901_t_real
    !! M=-3
    !! L-1
    ccc(1,1,1,40)=-0.15075567228888181130_t_real
    ccc(1,1,3,40)= 0.52223296786709351447_t_real
    ccc(2,1,1,40)= 0.52223296786709351447_t_real
    ccc(2,1,3,40)= 0.15075567228888181130_t_real
    ccc(3,1,2,40)= 0.63960214906683130282_t_real
    !! L+1
    ccc(1,2,1,40)=-0.51176631571915897704_t_real
    ccc(1,2,3,40)= 0.21821789023599238123_t_real
    ccc(2,2,1,40)= 0.21821789023599238123_t_real
    ccc(2,2,3,40)= 0.51176631571915897704_t_real
    ccc(3,2,2,40)=-0.61721339984836764094_t_real
    !! M=-2
    !! L-1
    ccc(1,1,1,41)=-0.21320071635561043427_t_real
    ccc(1,1,3,41)= 0.46056618647183829181_t_real
    ccc(2,1,1,41)= 0.46056618647183829181_t_real
    ccc(2,1,3,41)= 0.21320071635561043427_t_real
    ccc(3,1,2,41)= 0.69631062382279135262_t_real
    !! L+1
    ccc(1,2,1,41)=-0.46291004988627573070_t_real
    ccc(1,2,3,41)= 0.26726124191242438464_t_real
    ccc(2,2,1,41)= 0.26726124191242438464_t_real
    ccc(2,2,3,41)= 0.46291004988627573070_t_real
    ccc(3,2,2,41)=-0.65465367070797714369_t_real
    !! M=-1
    !! L-1
    ccc(1,1,1,42)=-0.27524094128159015136_t_real
    ccc(2,1,1,42)= 0.56407607481776620884_t_real
    ccc(2,1,3,42)= 0.27524094128159015136_t_real
    ccc(3,1,2,42)= 0.72821908125441913807_t_real
    !! L+1
    ccc(1,2,1,42)=-0.41403933560541253061_t_real
    ccc(2,2,1,42)= 0.44721359549995793919_t_real
    ccc(2,2,3,42)= 0.41403933560541253061_t_real
    ccc(3,2,2,42)=-0.67612340378281326190_t_real
    !! M=0
    !! L-1
    ccc(1,1,3,43)=-0.47673129462279615766_t_real
    ccc(2,1,1,43)=-0.47673129462279615766_t_real
    ccc(3,1,2,43)= 0.73854894587599639632_t_real
    !! L+1
    ccc(1,2,3,43)=-0.51639777949432225126_t_real
    ccc(2,2,1,43)=-0.51639777949432225126_t_real
    ccc(3,2,2,43)=-0.68313005106397322544_t_real
    !! M=1
    !! L-1
    ccc(1,1,1,44)= 0.56407607481776620884_t_real
    ccc(1,1,3,44)=-0.27524094128159015136_t_real
    ccc(2,1,1,44)=-0.27524094128159015136_t_real
    ccc(3,1,2,44)= 0.72821908125441913807_t_real
    !! L+1
    ccc(1,2,1,44)= 0.44721359549995793919_t_real
    ccc(1,2,3,44)=-0.41403933560541253061_t_real
    ccc(2,2,1,44)=-0.41403933560541253061_t_real
    ccc(3,2,2,44)=-0.67612340378281326190_t_real
    !! M=2
    !! L-1
    ccc(1,1,1,45)= 0.46056618647183829181_t_real
    ccc(1,1,3,45)=-0.21320071635561043427_t_real
    ccc(2,1,1,45)=-0.21320071635561043427_t_real
    ccc(2,1,3,45)=-0.46056618647183829181_t_real
    ccc(3,1,2,45)= 0.69631062382279135262_t_real
    !! L+1
    ccc(1,2,1,45)= 0.26726124191242438464_t_real
    ccc(1,2,3,45)=-0.46291004988627573070_t_real
    ccc(2,2,1,45)=-0.46291004988627573070_t_real
    ccc(2,2,3,45)=-0.26726124191242438464_t_real
    ccc(3,2,2,45)=-0.65465367070797714369_t_real
    !! M=3
    !! L-1
    ccc(1,1,1,46)= 0.52223296786709351447_t_real
    ccc(1,1,3,46)=-0.15075567228888181130_t_real
    ccc(2,1,1,46)=-0.15075567228888181130_t_real
    ccc(2,1,3,46)=-0.52223296786709351447_t_real
    ccc(3,1,2,46)= 0.63960214906683130282_t_real
    !! L+1
    ccc(1,2,1,46)= 0.21821789023599238123_t_real
    ccc(1,2,3,46)=-0.51176631571915897704_t_real
    ccc(2,2,1,46)=-0.51176631571915897704_t_real
    ccc(2,2,3,46)=-0.21821789023599238123_t_real
    ccc(3,2,2,46)=-0.61721339984836764094_t_real
    !! M=4
    !! L-1
    ccc(1,1,1,47)= 0.58387420812114222811_t_real
    ccc(1,1,3,47)=-0.08703882797784891907_t_real
    ccc(2,1,1,47)=-0.08703882797784891907_t_real
    ccc(2,1,3,47)=-0.58387420812114222811_t_real
    ccc(3,1,2,47)= 0.55048188256318030273_t_real
    !! L+1
    ccc(1,2,1,47)= 0.16903085094570331547_t_real
    ccc(1,2,3,47)=-0.56061191058138809901_t_real
    ccc(2,2,1,47)=-0.56061191058138809901_t_real
    ccc(2,2,3,47)=-0.16903085094570331547_t_real
    ccc(3,2,2,47)=-0.56061191058138809901_t_real
    !! M=5
    !! L-1
    ccc(1,1,1,48)= 0.64549722436790281412_t_real
    ccc(2,1,3,48)=-0.64549722436790281412_t_real
    ccc(3,1,2,48)= 0.40824829046386301631_t_real
    !! L+1
    ccc(1,2,1,48)= 0.11952286093343936397_t_real
    ccc(1,2,3,48)=-0.60944940022004404992_t_real
    ccc(2,2,1,48)=-0.60944940022004404992_t_real
    ccc(2,2,3,48)=-0.11952286093343936397_t_real
    ccc(3,2,2,48)=-0.47809144373375745591_t_real
    !! M=6
    !! L-1
    ccc(1,1,1,49)= 0.70710678118654752431_t_real
    ccc(2,1,3,49)=-0.70710678118654752431_t_real
    !! L+1
    ccc(1,2,1,49)= 0.06900655593423542176_t_real
    ccc(1,2,3,49)=-0.65828058860438330047_t_real
    ccc(2,2,1,49)=-0.65828058860438330047_t_real
    ccc(2,2,3,49)=-0.06900655593423542176_t_real
    ccc(3,2,2,49)=-0.35186577527449839986_t_real
    !!
    If(maxL<=6) Return
    !!
    !!..............................................
    !! L=7
    !!..............................................
    !! M=-7
    !! L-1
    ccc(1,1,3,50)= 0.70710678118654752428_t_real
    ccc(2,1,1,50)= 0.70710678118654752428_t_real
    !! L+1
    ccc(1,2,1,50)=-0.66421116415507146609_t_real
    ccc(1,2,3,50)= 0.06063390625908324337_t_real
    ccc(2,2,1,50)= 0.06063390625908324337_t_real
    ccc(2,2,3,50)= 0.66421116415507146609_t_real
    ccc(3,2,2,50)=-0.33210558207753573304_t_real
    !! M=-6
    !! L-1
    ccc(1,1,3,51)= 0.65465367070797714369_t_real
    ccc(2,1,1,51)= 0.65465367070797714369_t_real
    ccc(3,1,2,51)= 0.37796447300922722715_t_real
    !! L+1
    ccc(1,2,1,51)=-0.62131265218463552729_t_real
    ccc(1,2,3,51)= 0.10502100630210073525_t_real
    ccc(2,2,1,51)= 0.10502100630210073525_t_real
    ccc(2,2,3,51)= 0.62131265218463552729_t_real
    ccc(3,2,2,51)=-0.45374260648651507315_t_real
    !! M=-5
    !! L-1
    ccc(1,1,1,52)=-0.07412493166611011793_t_real
    ccc(1,1,3,52)= 0.60219379159649482162_t_real
    ccc(2,1,1,52)= 0.60219379159649482162_t_real
    ccc(2,1,3,52)= 0.07412493166611011793_t_real
    ccc(3,1,2,52)= 0.51355259101309549928_t_real
    !! L+1
    ccc(1,2,1,52)=-0.57841060115579811581_t_real
    ccc(1,2,3,52)= 0.14852213144650114635_t_real
    ccc(2,2,1,52)= 0.14852213144650114635_t_real
    ccc(2,2,3,52)= 0.57841060115579811581_t_real
    ccc(3,2,2,52)=-0.53550416047156248131_t_real
    !! M=-4
    !! L-1
    ccc(1,1,1,53)=-0.12838814775327387482_t_real
    ccc(1,1,3,53)= 0.54972520607827525530_t_real
    ccc(2,1,1,53)= 0.54972520607827525530_t_real
    ccc(2,1,3,53)= 0.12838814775327387482_t_real
    ccc(3,1,2,53)= 0.60219379159649482162_t_real
    !! L+1
    ccc(1,2,1,53)=-0.53550416047156248131_t_real
    ccc(1,2,3,53)= 0.19174124721184260923_t_real
    ccc(2,2,1,53)= 0.19174124721184260923_t_real
    ccc(2,2,3,53)= 0.53550416047156248131_t_real
    ccc(3,2,2,53)=-0.59408852578600458542_t_real
    !! M=-3
    !! L-1
    ccc(1,1,1,54)=-0.18156825980064072592_t_real
    ccc(1,1,3,54)= 0.49724515809884690386_t_real
    ccc(2,1,1,54)= 0.49724515809884690386_t_real
    ccc(2,1,3,54)= 0.18156825980064072592_t_real
    ccc(3,1,2,54)= 0.66299354413179587181_t_real
    !! L+1
    ccc(1,2,1,54)=-0.49259218307188900387_t_real
    ccc(1,2,3,54)= 0.23483410915693105878_t_real
    ccc(2,2,1,54)= 0.23483410915693105878_t_real
    ccc(2,2,3,54)= 0.49259218307188900387_t_real
    ccc(3,2,2,54)=-0.63593377383646038968_t_real
    !! M=-2
    !! L-1
    ccc(1,1,1,55)=-0.23440361546924771150_t_real
    ccc(1,1,3,55)= 0.44474958999666070758_t_real
    ccc(2,1,1,55)= 0.44474958999666070758_t_real
    ccc(2,1,3,55)= 0.23440361546924771150_t_real
    ccc(3,1,2,55)= 0.70321084640774313452_t_real
    !! L+1
    ccc(1,2,1,55)=-0.44967308386531339777_t_real
    ccc(1,2,3,55)= 0.27785946511310565116_t_real
    ccc(2,2,1,55)= 0.27785946511310565116_t_real
    ccc(2,2,3,55)= 0.44967308386531339777_t_real
    ccc(3,2,2,55)=-0.66421116415507146609_t_real
    !! M=-1
    !! L-1
    ccc(1,1,1,56)=-0.28708462588160728178_t_real
    ccc(2,1,1,56)= 0.55470019622522912192_t_real
    ccc(2,1,3,56)= 0.28708462588160728178_t_real
    ccc(3,1,2,56)= 0.72627303920256290369_t_real
    !! L+1
    ccc(1,2,1,56)=-0.40674460840998032245_t_real
    ccc(2,2,1,56)= 0.45374260648651507315_t_real
    ccc(2,2,3,56)= 0.40674460840998032245_t_real
    ccc(3,2,2,56)=-0.68061390972977260973_t_real
    !! M=0
    !! L-1
    ccc(1,1,3,57)=-0.48038446141526140035_t_real
    ccc(2,1,1,57)=-0.48038446141526140035_t_real
    ccc(3,1,2,57)= 0.73379938570534280692_t_real
    !! L+1
    ccc(1,2,3,57)=-0.51449575542752651206_t_real
    ccc(2,2,1,57)=-0.51449575542752651206_t_real
    ccc(3,2,2,57)=-0.68599434057003534944_t_real
    !! M=1
    !! L-1
    ccc(1,1,1,58)= 0.55470019622522912192_t_real
    ccc(1,1,3,58)=-0.28708462588160728178_t_real
    ccc(2,1,1,58)=-0.28708462588160728178_t_real
    ccc(3,1,2,58)= 0.72627303920256290369_t_real
    !! L+1
    ccc(1,2,1,58)= 0.45374260648651507315_t_real
    ccc(1,2,3,58)=-0.40674460840998032245_t_real
    ccc(2,2,1,58)=-0.40674460840998032245_t_real
    ccc(3,2,2,58)=-0.68061390972977260973_t_real
    !! M=2
    !! L-1
    ccc(1,1,1,59)= 0.44474958999666070758_t_real
    ccc(1,1,3,59)=-0.23440361546924771150_t_real
    ccc(2,1,1,59)=-0.23440361546924771150_t_real
    ccc(2,1,3,59)=-0.44474958999666070758_t_real
    ccc(3,1,2,59)= 0.70321084640774313452_t_real
    !! L+1
    ccc(1,2,1,59)= 0.27785946511310565116_t_real
    ccc(1,2,3,59)=-0.44967308386531339777_t_real
    ccc(2,2,1,59)=-0.44967308386531339777_t_real
    ccc(2,2,3,59)=-0.27785946511310565116_t_real
    ccc(3,2,2,59)=-0.66421116415507146609_t_real
    !! M=3
    !! L-1
    ccc(1,1,1,60)= 0.49724515809884690386_t_real
    ccc(1,1,3,60)=-0.18156825980064072592_t_real
    ccc(2,1,1,60)=-0.18156825980064072592_t_real
    ccc(2,1,3,60)=-0.49724515809884690386_t_real
    ccc(3,1,2,60)= 0.66299354413179587181_t_real
    !! L+1
    ccc(1,2,1,60)= 0.23483410915693105878_t_real
    ccc(1,2,3,60)=-0.49259218307188900387_t_real
    ccc(2,2,1,60)=-0.49259218307188900387_t_real
    ccc(2,2,3,60)=-0.23483410915693105878_t_real
    ccc(3,2,2,60)=-0.63593377383646038968_t_real
    !! M=4
    !! L-1
    ccc(1,1,1,61)= 0.54972520607827525530_t_real
    ccc(1,1,3,61)=-0.12838814775327387482_t_real
    ccc(2,1,1,61)=-0.12838814775327387482_t_real
    ccc(2,1,3,61)=-0.54972520607827525530_t_real
    ccc(3,1,2,61)= 0.60219379159649482162_t_real
    !! L+1
    ccc(1,2,1,61)= 0.19174124721184260923_t_real
    ccc(1,2,3,61)=-0.53550416047156248131_t_real
    ccc(2,2,1,61)=-0.53550416047156248131_t_real
    ccc(2,2,3,61)=-0.19174124721184260923_t_real
    ccc(3,2,2,61)=-0.59408852578600458542_t_real
    !! M=5
    !! L-1
    ccc(1,1,1,62)= 0.60219379159649482162_t_real
    ccc(1,1,3,62)=-0.07412493166611011793_t_real
    ccc(2,1,1,62)=-0.07412493166611011793_t_real
    ccc(2,1,3,62)=-0.60219379159649482162_t_real
    ccc(3,1,2,62)= 0.51355259101309549928_t_real
    !! L+1
    ccc(1,2,1,62)= 0.14852213144650114635_t_real
    ccc(1,2,3,62)=-0.57841060115579811581_t_real
    ccc(2,2,1,62)=-0.57841060115579811581_t_real
    ccc(2,2,3,62)=-0.14852213144650114635_t_real
    ccc(3,2,2,62)=-0.53550416047156248131_t_real
    !! M=6
    !! L-1
    ccc(1,1,1,63)= 0.65465367070797714369_t_real
    ccc(2,1,3,63)=-0.65465367070797714369_t_real
    ccc(3,1,2,63)= 0.37796447300922722715_t_real
    !! L+1
    ccc(1,2,1,63)= 0.10502100630210073525_t_real
    ccc(1,2,3,63)=-0.62131265218463552729_t_real
    ccc(2,2,1,63)=-0.62131265218463552729_t_real
    ccc(2,2,3,63)=-0.10502100630210073525_t_real
    ccc(3,2,2,63)=-0.45374260648651507315_t_real
    !! M=7
    !! L-1
    ccc(1,1,1,64)= 0.70710678118654752428_t_real
    ccc(2,1,3,64)=-0.70710678118654752428_t_real
    !! L+1
    ccc(1,2,1,64)= 0.06063390625908324337_t_real
    ccc(1,2,3,64)=-0.66421116415507146609_t_real
    ccc(2,2,1,64)=-0.66421116415507146609_t_real
    ccc(2,2,3,64)=-0.06063390625908324337_t_real
    ccc(3,2,2,64)=-0.33210558207753573304_t_real
    !!
    If(maxL<=7) Return
    !!
    !!..............................................
    !! L=8
    !!..............................................
    !! M=-8
    !! L-1
    ccc(1,1,3,65)= 0.70710678118654752427_t_real
    ccc(2,1,1,65)= 0.70710678118654752427_t_real
    !! L+1
    ccc(1,2,1,65)=-0.66885605405993865379_t_real
    ccc(1,2,3,65)= 0.05407380704358751271_t_real
    ccc(2,2,1,65)= 0.05407380704358751271_t_real
    ccc(2,2,3,65)= 0.66885605405993865379_t_real
    ccc(3,2,2,65)=-0.31530176764230576240_t_real
    !! M=-7
    !! L-1
    ccc(1,1,3,66)= 0.66143782776614764750_t_real
    ccc(2,1,1,66)= 0.66143782776614764750_t_real
    ccc(3,1,2,66)= 0.35355339059327376213_t_real
    !! L+1
    ccc(1,2,1,66)=-0.63060353528461152480_t_real
    ccc(1,2,3,66)= 0.09365858115816939658_t_real
    ccc(2,2,1,66)= 0.09365858115816939658_t_real
    ccc(2,2,3,66)= 0.63060353528461152480_t_real
    ccc(3,2,2,66)=-0.43259045634870010175_t_real
    !! M=-6
    !! L-1
    ccc(1,1,1,67)=-0.06454972243679028140_t_real
    ccc(1,1,3,67)= 0.61576510673037219975_t_real
    ccc(2,1,1,67)= 0.61576510673037219975_t_real
    ccc(2,1,3,67)= 0.06454972243679028140_t_real
    ccc(3,1,2,67)= 0.48304589153964795237_t_real
    !! L+1
    ccc(1,2,1,67)=-0.59234887775909235728_t_real
    ccc(1,2,3,67)= 0.13245323570650438064_t_real
    ccc(2,2,1,67)= 0.13245323570650438064_t_real
    ccc(2,2,3,67)= 0.59234887775909235728_t_real
    ccc(3,2,2,67)=-0.51298917604257704771_t_real
    !! M=-5
    !! L-1
    ccc(1,1,1,68)=-0.11180339887498948480_t_real
    ccc(1,1,3,68)= 0.57008771254956898946_t_real
    ccc(2,1,1,68)= 0.57008771254956898946_t_real
    ccc(2,1,3,68)= 0.11180339887498948480_t_real
    ccc(3,1,2,68)= 0.57008771254956898946_t_real
    !! L+1
    ccc(1,2,1,68)=-0.55409163850364058962_t_real
    ccc(1,2,3,68)= 0.17099639201419234923_t_real
    ccc(2,2,1,68)= 0.17099639201419234923_t_real
    ccc(2,2,3,68)= 0.55409163850364058962_t_real
    ccc(3,2,2,68)=-0.57226338351930142019_t_real
    !! M=-4
    !! L-1
    ccc(1,1,1,69)=-0.15811388300841896657_t_real
    ccc(1,1,3,69)= 0.52440442408507577340_t_real
    ccc(2,1,1,69)= 0.52440442408507577340_t_real
    ccc(2,1,3,69)= 0.15811388300841896657_t_real
    ccc(3,1,2,69)= 0.63245553203367586628_t_real
    !! L+1
    ccc(1,2,1,69)=-0.51583124308733882637_t_real
    ccc(1,2,3,69)= 0.20942695414584775348_t_real
    ccc(2,2,1,69)= 0.20942695414584775348_t_real
    ccc(2,2,3,69)= 0.51583124308733882637_t_real
    ccc(3,2,2,69)=-0.61653625932651153794_t_real
    !! M=-3
    !! L-1
    ccc(1,1,1,70)=-0.20412414523193150814_t_real
    ccc(1,1,3,70)= 0.47871355387816905490_t_real
    ccc(2,1,1,70)= 0.47871355387816905490_t_real
    ccc(2,1,3,70)= 0.20412414523193150814_t_real
    ccc(3,1,2,70)= 0.67700320038633002991_t_real
    !! L+1
    ccc(1,2,1,70)=-0.47756693294091927864_t_real
    ccc(1,2,3,70)= 0.24779731389167604243_t_real
    ccc(2,2,1,70)= 0.24779731389167604243_t_real
    ccc(2,2,3,70)= 0.47756693294091927864_t_real
    ccc(3,2,2,70)=-0.64888568452305015263_t_real
    !! M=-2
    !! L-1
    ccc(1,1,1,71)=-0.24999999999999999995_t_real
    ccc(1,1,3,71)= 0.43301270189221932330_t_real
    ccc(2,1,1,71)= 0.43301270189221932330_t_real
    ccc(2,1,3,71)= 0.24999999999999999995_t_real
    ccc(3,1,2,71)= 0.70710678118654752427_t_real
    !! L+1
    ccc(1,2,1,71)=-0.43929768510697945308_t_real
    ccc(1,2,3,71)= 0.28613169175965071009_t_real
    ccc(2,2,1,71)= 0.28613169175965071009_t_real
    ccc(2,2,3,71)= 0.43929768510697945308_t_real
    ccc(3,2,2,71)=-0.67103829820720271123_t_real
    !! M=-1
    !! L-1
    ccc(1,1,1,72)=-0.29580398915498080207_t_real
    ccc(2,1,1,72)= 0.54772255750516611334_t_real
    ccc(2,1,3,72)= 0.29580398915498080207_t_real
    ccc(3,1,2,72)= 0.72456883730947192855_t_real
    !! L+1
    ccc(1,2,1,72)=-0.40102208598815651471_t_real
    ccc(2,2,1,72)= 0.45883146774112353178_t_real
    ccc(2,2,3,72)= 0.40102208598815651471_t_real
    ccc(3,2,2,72)=-0.68398556805676939694_t_real
    !! M=0
    !! L-1
    ccc(1,1,3,73)=-0.48304589153964795236_t_real
    ccc(2,1,1,73)=-0.48304589153964795236_t_real
    ccc(3,1,2,73)= 0.73029674334022148448_t_real
    !! L+1
    ccc(1,2,3,73)=-0.51298917604257704770_t_real
    ccc(2,2,1,73)=-0.51298917604257704770_t_real
    ccc(3,2,2,73)=-0.68824720161168529770_t_real
    !! M=1
    !! L-1
    ccc(1,1,1,74)= 0.54772255750516611334_t_real
    ccc(1,1,3,74)=-0.29580398915498080207_t_real
    ccc(2,1,1,74)=-0.29580398915498080207_t_real
    ccc(3,1,2,74)= 0.72456883730947192855_t_real
    !! L+1
    ccc(1,2,1,74)= 0.45883146774112353178_t_real
    ccc(1,2,3,74)=-0.40102208598815651471_t_real
    ccc(2,2,1,74)=-0.40102208598815651471_t_real
    ccc(3,2,2,74)=-0.68398556805676939694_t_real
    !! M=2
    !! L-1
    ccc(1,1,1,75)= 0.43301270189221932330_t_real
    ccc(1,1,3,75)=-0.24999999999999999995_t_real
    ccc(2,1,1,75)=-0.24999999999999999995_t_real
    ccc(2,1,3,75)=-0.43301270189221932330_t_real
    ccc(3,1,2,75)= 0.70710678118654752427_t_real
    !! L+1
    ccc(1,2,1,75)= 0.28613169175965071009_t_real
    ccc(1,2,3,75)=-0.43929768510697945308_t_real
    ccc(2,2,1,75)=-0.43929768510697945308_t_real
    ccc(2,2,3,75)=-0.28613169175965071009_t_real
    ccc(3,2,2,75)=-0.67103829820720271123_t_real
    !! M=3
    !! L-1
    ccc(1,1,1,76)= 0.47871355387816905490_t_real
    ccc(1,1,3,76)=-0.20412414523193150814_t_real
    ccc(2,1,1,76)=-0.20412414523193150814_t_real
    ccc(2,1,3,76)=-0.47871355387816905490_t_real
    ccc(3,1,2,76)= 0.67700320038633002991_t_real
    !! L+1
    ccc(1,2,1,76)= 0.24779731389167604243_t_real
    ccc(1,2,3,76)=-0.47756693294091927864_t_real
    ccc(2,2,1,76)=-0.47756693294091927864_t_real
    ccc(2,2,3,76)=-0.24779731389167604243_t_real
    ccc(3,2,2,76)=-0.64888568452305015263_t_real
    !! M=4
    !! L-1
    ccc(1,1,1,77)= 0.52440442408507577340_t_real
    ccc(1,1,3,77)=-0.15811388300841896657_t_real
    ccc(2,1,1,77)=-0.15811388300841896657_t_real
    ccc(2,1,3,77)=-0.52440442408507577340_t_real
    ccc(3,1,2,77)= 0.63245553203367586628_t_real
    !! L+1
    ccc(1,2,1,77)= 0.20942695414584775348_t_real
    ccc(1,2,3,77)=-0.51583124308733882637_t_real
    ccc(2,2,1,77)=-0.51583124308733882637_t_real
    ccc(2,2,3,77)=-0.20942695414584775348_t_real
    ccc(3,2,2,77)=-0.61653625932651153794_t_real
    !! M=5
    !! L-1
    ccc(1,1,1,78)= 0.57008771254956898946_t_real
    ccc(1,1,3,78)=-0.11180339887498948480_t_real
    ccc(2,1,1,78)=-0.11180339887498948480_t_real
    ccc(2,1,3,78)=-0.57008771254956898946_t_real
    ccc(3,1,2,78)= 0.57008771254956898946_t_real
    !! L+1
    ccc(1,2,1,78)= 0.17099639201419234923_t_real
    ccc(1,2,3,78)=-0.55409163850364058962_t_real
    ccc(2,2,1,78)=-0.55409163850364058962_t_real
    ccc(2,2,3,78)=-0.17099639201419234923_t_real
    ccc(3,2,2,78)=-0.57226338351930142019_t_real
    !! M=6
    !! L-1
    ccc(1,1,1,79)= 0.61576510673037219975_t_real
    ccc(1,1,3,79)=-0.06454972243679028140_t_real
    ccc(2,1,1,79)=-0.06454972243679028140_t_real
    ccc(2,1,3,79)=-0.61576510673037219975_t_real
    ccc(3,1,2,79)= 0.48304589153964795237_t_real
    !! L+1
    ccc(1,2,1,79)= 0.13245323570650438064_t_real
    ccc(1,2,3,79)=-0.59234887775909235728_t_real
    ccc(2,2,1,79)=-0.59234887775909235728_t_real
    ccc(2,2,3,79)=-0.13245323570650438064_t_real
    ccc(3,2,2,79)=-0.51298917604257704771_t_real
    !! M=7
    !! L-1
    ccc(1,1,1,80)= 0.66143782776614764750_t_real
    ccc(2,1,3,80)=-0.66143782776614764750_t_real
    ccc(3,1,2,80)= 0.35355339059327376213_t_real
    !! L+1
    ccc(1,2,1,80)= 0.09365858115816939658_t_real
    ccc(1,2,3,80)=-0.63060353528461152480_t_real
    ccc(2,2,1,80)=-0.63060353528461152480_t_real
    ccc(2,2,3,80)=-0.09365858115816939658_t_real
    ccc(3,2,2,80)=-0.43259045634870010175_t_real
    !! M=8
    !! L-1
    ccc(1,1,1,81)= 0.70710678118654752427_t_real
    ccc(2,1,3,81)=-0.70710678118654752427_t_real
    !! L+1
    ccc(1,2,1,81)= 0.05407380704358751271_t_real
    ccc(1,2,3,81)=-0.66885605405993865379_t_real
    ccc(2,2,1,81)=-0.66885605405993865379_t_real
    ccc(2,2,3,81)=-0.05407380704358751271_t_real
    ccc(3,2,2,81)=-0.31530176764230576240_t_real
    !!
    If(maxL<=8) Return
    !!
    !!..............................................
    !! L=9
    !!..............................................
    !! M=-9
    !! L-1
    ccc(1,1,3,82)= 0.70710678118654752434_t_real
    ccc(2,1,1,82)= 0.70710678118654752434_t_real
    !! L+1
    ccc(1,2,1,82)=-0.67259270913454925786_t_real
    ccc(1,2,3,82)= 0.04879500364742665895_t_real
    ccc(2,2,1,82)= 0.04879500364742665895_t_real
    ccc(2,2,3,82)= 0.67259270913454925786_t_real
    ccc(3,2,2,82)=-0.30079260375911917714_t_real
    !! M=-8
    !! L-1
    ccc(1,1,3,83)= 0.66666666666666666661_t_real
    ccc(2,1,1,83)= 0.66666666666666666661_t_real
    ccc(3,1,2,83)= 0.33333333333333333330_t_real
    !! L+1
    ccc(1,2,1,83)=-0.63807746954649412901_t_real
    ccc(1,2,3,83)= 0.08451542547285165772_t_real
    ccc(2,2,1,83)= 0.08451542547285165772_t_real
    ccc(2,2,3,83)= 0.63807746954649412901_t_real
    ccc(3,2,2,83)=-0.41403933560541253056_t_real
    !! M=-7
    !! L-1
    ccc(1,1,1,84)=-0.05716619504750294578_t_real
    ccc(1,1,3,84)= 0.62622429108514949011_t_real
    ccc(2,1,1,84)= 0.62622429108514949011_t_real
    ccc(2,1,3,84)= 0.05716619504750294578_t_real
    ccc(3,1,2,84)= 0.45732956038002356630_t_real
    !! L+1
    ccc(1,2,1,84)=-0.60356086212221736926_t_real
    ccc(1,2,3,84)= 0.11952286093343936396_t_real
    ccc(2,2,1,84)= 0.11952286093343936396_t_real
    ccc(2,2,3,84)= 0.60356086212221736926_t_real
    ccc(3,2,2,84)=-0.49280538030458114815_t_real
    !! M=-6
    !! L-1
    ccc(1,1,1,85)=-0.09901475429766743090_t_real
    ccc(1,1,3,85)= 0.58577918612900611002_t_real
    ccc(2,1,1,85)= 0.58577918612900611002_t_real
    ccc(2,1,3,85)= 0.09901475429766743090_t_real
    ccc(3,1,2,85)= 0.54232614454664042995_t_real
    !! L+1
    ccc(1,2,1,85)=-0.56904263795389164962_t_real
    ccc(1,2,3,85)= 0.15430334996209191021_t_real
    ccc(2,2,1,85)= 0.15430334996209191021_t_real
    ccc(2,2,3,85)= 0.56904263795389164962_t_real
    ccc(3,2,2,85)=-0.55205244747388337408_t_real
    !! M=-5
    !! L-1
    ccc(1,1,1,86)=-0.14002800840280098034_t_real
    ccc(1,1,3,86)= 0.54533074451660313454_t_real
    ccc(2,1,1,86)= 0.54533074451660313454_t_real
    ccc(2,1,3,86)= 0.14002800840280098034_t_real
    ccc(3,1,2,86)= 0.60499014198202009756_t_real
    !! L+1
    ccc(1,2,1,86)=-0.53452248382484876921_t_real
    ccc(1,2,3,86)= 0.18898223650461361355_t_real
    ccc(2,2,1,86)= 0.18898223650461361355_t_real
    ccc(2,2,3,86)= 0.53452248382484876921_t_real
    ccc(3,2,2,86)=-0.59761430466719681981_t_real
    !! M=-4
    !! L-1
    ccc(1,1,1,87)=-0.18077538151554680998_t_real
    ccc(1,1,3,87)= 0.50487816429740128487_t_real
    ccc(2,1,1,87)= 0.50487816429740128487_t_real
    ccc(2,1,3,87)= 0.18077538151554680998_t_real
    ccc(3,1,2,87)= 0.65179490739586907454_t_real
    !! L+1
    ccc(1,2,1,87)=-0.49999999999999999985_t_real
    ccc(1,2,3,87)= 0.22360679774997896957_t_real
    ccc(2,2,1,87)= 0.22360679774997896957_t_real
    ccc(2,2,3,87)= 0.49999999999999999985_t_real
    ccc(3,2,2,87)=-0.63245553203367586621_t_real
    !! M=-3
    !! L-1
    ccc(1,1,1,88)=-0.22140372138502382203_t_real
    ccc(1,1,3,88)= 0.46442036401282396989_t_real
    ccc(2,1,1,88)= 0.46442036401282396989_t_real
    ccc(2,1,3,88)= 0.22140372138502382203_t_real
    ccc(3,1,2,88)= 0.68599434057003534946_t_real
    !! L+1
    ccc(1,2,1,88)=-0.46547466812563137219_t_real
    ccc(1,2,3,88)= 0.25819888974716112560_t_real
    ccc(2,2,1,88)= 0.25819888974716112560_t_real
    ccc(2,2,3,88)= 0.46547466812563137219_t_real
    ccc(3,2,2,88)=-0.65828058860438330039_t_real
    !! M=-2
    !! L-1
    ccc(1,1,1,89)=-0.26196841599779191101_t_real
    ccc(1,1,3,89)= 0.42395584922430692646_t_real
    ccc(2,1,1,89)= 0.42395584922430692646_t_real
    ccc(2,1,3,89)= 0.26196841599779191101_t_real
    ccc(3,1,2,89)= 0.70941382412257033084_t_real
    !! L+1
    ccc(1,2,1,89)=-0.43094580368566731662_t_real
    ccc(1,2,3,89)= 0.29277002188455995372_t_real
    ccc(2,2,1,89)= 0.29277002188455995372_t_real
    ccc(2,2,3,89)= 0.43094580368566731662_t_real
    ccc(3,2,2,89)=-0.67612340378281326181_t_real
    !! M=-1
    !! L-1
    ccc(1,1,1,90)=-0.30249507099101004878_t_real
    ccc(2,1,1,90)= 0.54232614454664042995_t_real
    ccc(2,1,3,90)= 0.30249507099101004878_t_real
    ccc(3,1,2,90)= 0.72310152606218723994_t_real
    !! L+1
    ccc(1,2,1,90)=-0.39641248358604594122_t_real
    ccc(2,2,1,90)= 0.46291004988627573064_t_real
    ccc(2,2,3,90)= 0.39641248358604594122_t_real
    ccc(3,2,2,90)=-0.68660656232559518721_t_real
    !! M=0
    !! L-1
    ccc(1,1,3,91)=-0.48507125007266594699_t_real
    ccc(2,1,1,91)=-0.48507125007266594699_t_real
    ccc(3,1,2,91)= 0.72760687510899892050_t_real
    !! L+1
    ccc(1,2,3,91)=-0.51176631571915897696_t_real
    ccc(2,2,1,91)=-0.51176631571915897696_t_real
    ccc(3,2,2,91)=-0.69006555934235421760_t_real
    !! M=1
    !! L-1
    ccc(1,1,1,92)= 0.54232614454664042995_t_real
    ccc(1,1,3,92)=-0.30249507099101004878_t_real
    ccc(2,1,1,92)=-0.30249507099101004878_t_real
    ccc(3,1,2,92)= 0.72310152606218723994_t_real
    !! L+1
    ccc(1,2,1,92)= 0.46291004988627573064_t_real
    ccc(1,2,3,92)=-0.39641248358604594122_t_real
    ccc(2,2,1,92)=-0.39641248358604594122_t_real
    ccc(3,2,2,92)=-0.68660656232559518721_t_real
    !! M=2
    !! L-1
    ccc(1,1,1,93)= 0.42395584922430692646_t_real
    ccc(1,1,3,93)=-0.26196841599779191101_t_real
    ccc(2,1,1,93)=-0.26196841599779191101_t_real
    ccc(2,1,3,93)=-0.42395584922430692646_t_real
    ccc(3,1,2,93)= 0.70941382412257033084_t_real
    !! L+1
    ccc(1,2,1,93)= 0.29277002188455995372_t_real
    ccc(1,2,3,93)=-0.43094580368566731662_t_real
    ccc(2,2,1,93)=-0.43094580368566731662_t_real
    ccc(2,2,3,93)=-0.29277002188455995372_t_real
    ccc(3,2,2,93)=-0.67612340378281326181_t_real
    !! M=3
    !! L-1
    ccc(1,1,1,94)= 0.46442036401282396989_t_real
    ccc(1,1,3,94)=-0.22140372138502382203_t_real
    ccc(2,1,1,94)=-0.22140372138502382203_t_real
    ccc(2,1,3,94)=-0.46442036401282396989_t_real
    ccc(3,1,2,94)= 0.68599434057003534946_t_real
    !! L+1
    ccc(1,2,1,94)= 0.25819888974716112560_t_real
    ccc(1,2,3,94)=-0.46547466812563137219_t_real
    ccc(2,2,1,94)=-0.46547466812563137219_t_real
    ccc(2,2,3,94)=-0.25819888974716112560_t_real
    ccc(3,2,2,94)=-0.65828058860438330039_t_real
    !! M=4
    !! L-1
    ccc(1,1,1,95)= 0.50487816429740128487_t_real
    ccc(1,1,3,95)=-0.18077538151554680998_t_real
    ccc(2,1,1,95)=-0.18077538151554680998_t_real
    ccc(2,1,3,95)=-0.50487816429740128487_t_real
    ccc(3,1,2,95)= 0.65179490739586907454_t_real
    !! L+1
    ccc(1,2,1,95)= 0.22360679774997896957_t_real
    ccc(1,2,3,95)=-0.49999999999999999985_t_real
    ccc(2,2,1,95)=-0.49999999999999999985_t_real
    ccc(2,2,3,95)=-0.22360679774997896957_t_real
    ccc(3,2,2,95)=-0.63245553203367586621_t_real
    !! M=5
    !! L-1
    ccc(1,1,1,96)= 0.54533074451660313454_t_real
    ccc(1,1,3,96)=-0.14002800840280098034_t_real
    ccc(2,1,1,96)=-0.14002800840280098034_t_real
    ccc(2,1,3,96)=-0.54533074451660313454_t_real
    ccc(3,1,2,96)= 0.60499014198202009756_t_real
    !! L+1
    ccc(1,2,1,96)= 0.18898223650461361355_t_real
    ccc(1,2,3,96)=-0.53452248382484876921_t_real
    ccc(2,2,1,96)=-0.53452248382484876921_t_real
    ccc(2,2,3,96)=-0.18898223650461361355_t_real
    ccc(3,2,2,96)=-0.59761430466719681981_t_real
    !! M=6
    !! L-1
    ccc(1,1,1,97)= 0.58577918612900611002_t_real
    ccc(1,1,3,97)=-0.09901475429766743090_t_real
    ccc(2,1,1,97)=-0.09901475429766743090_t_real
    ccc(2,1,3,97)=-0.58577918612900611002_t_real
    ccc(3,1,2,97)= 0.54232614454664042995_t_real
    !! L+1
    ccc(1,2,1,97)= 0.15430334996209191021_t_real
    ccc(1,2,3,97)=-0.56904263795389164962_t_real
    ccc(2,2,1,97)=-0.56904263795389164962_t_real
    ccc(2,2,3,97)=-0.15430334996209191021_t_real
    ccc(3,2,2,97)=-0.55205244747388337408_t_real
    !! M=7
    !! L-1
    ccc(1,1,1,98)= 0.62622429108514949011_t_real
    ccc(1,1,3,98)=-0.05716619504750294578_t_real
    ccc(2,1,1,98)=-0.05716619504750294578_t_real
    ccc(2,1,3,98)=-0.62622429108514949011_t_real
    ccc(3,1,2,98)= 0.45732956038002356630_t_real
    !! L+1
    ccc(1,2,1,98)= 0.11952286093343936396_t_real
    ccc(1,2,3,98)=-0.60356086212221736926_t_real
    ccc(2,2,1,98)=-0.60356086212221736926_t_real
    ccc(2,2,3,98)=-0.11952286093343936396_t_real
    ccc(3,2,2,98)=-0.49280538030458114815_t_real
    !! M=8
    !! L-1
    ccc(1,1,1,99)= 0.66666666666666666661_t_real
    ccc(2,1,3,99)=-0.66666666666666666661_t_real
    ccc(3,1,2,99)= 0.33333333333333333330_t_real
    !! L+1
    ccc(1,2,1,99)= 0.08451542547285165772_t_real
    ccc(1,2,3,99)=-0.63807746954649412901_t_real
    ccc(2,2,1,99)=-0.63807746954649412901_t_real
    ccc(2,2,3,99)=-0.08451542547285165772_t_real
    ccc(3,2,2,99)=-0.41403933560541253056_t_real
    !! M=9
    !! L-1
    ccc(1,1,1,100)= 0.70710678118654752434_t_real
    ccc(2,1,3,100)=-0.70710678118654752434_t_real
    !! L+1
    ccc(1,2,1,100)= 0.04879500364742665895_t_real
    ccc(1,2,3,100)=-0.67259270913454925786_t_real
    ccc(2,2,1,100)=-0.67259270913454925786_t_real
    ccc(2,2,3,100)=-0.04879500364742665895_t_real
    ccc(3,2,2,100)=-0.30079260375911917714_t_real
    !!
    If(maxL<=9) Return
    !!
    !!..............................................
    !! L=10
    !!..............................................
    !! M=-10
    !! L-1
    ccc(1,1,3,101)= 0.70710678118654752420_t_real
    ccc(2,1,1,101)= 0.70710678118654752420_t_real
    !! L+1
    ccc(1,2,1,101)=-0.67566392469217622316_t_real
    ccc(1,2,3,101)= 0.04445542244743870560_t_real
    ccc(2,2,1,101)= 0.04445542244743870560_t_real
    ccc(2,2,3,101)= 0.67566392469217622316_t_real
    ccc(3,2,2,101)=-0.28810406552003038415_t_real
    !! M=-9
    !! L-1
    ccc(1,1,3,102)= 0.67082039324993690873_t_real
    ccc(2,1,1,102)= 0.67082039324993690873_t_real
    ccc(3,1,2,102)= 0.31622776601683793311_t_real
    !! L+1
    ccc(1,2,1,102)=-0.64422027509684123750_t_real
    ccc(1,2,3,102)= 0.07699905035090180553_t_real
    ccc(2,2,1,102)= 0.07699905035090180553_t_real
    ccc(2,2,3,102)= 0.64422027509684123750_t_real
    ccc(3,2,2,102)=-0.39762138624377206946_t_real
    !! M=-8
    !! L-1
    ccc(1,1,1,103)=-0.05129891760425770475_t_real
    ccc(1,1,3,103)= 0.63453256728663535228_t_real
    ccc(2,1,1,103)= 0.63453256728663535228_t_real
    ccc(2,1,3,103)= 0.05129891760425770475_t_real
    ccc(3,1,2,103)= 0.43528575006600700308_t_real
    !! L+1
    ccc(1,2,1,103)=-0.61277571031026112242_t_real
    ccc(1,2,3,103)= 0.10889310129609415676_t_real
    ccc(2,2,1,103)= 0.10889310129609415676_t_real
    ccc(2,2,3,103)= 0.61277571031026112242_t_real
    ccc(3,2,2,103)=-0.47465402419841236983_t_real
    !! M=-7
    !! L-1
    ccc(1,1,1,104)=-0.08885233166386385357_t_real
    ccc(1,1,3,104)= 0.59824304161611881046_t_real
    ccc(2,1,1,104)= 0.59824304161611881046_t_real
    ccc(2,1,3,104)= 0.08885233166386385357_t_real
    ccc(3,1,2,104)= 0.51809367167683002613_t_real
    !! L+1
    ccc(1,2,1,104)=-0.58133008182238475428_t_real
    ccc(1,2,3,104)= 0.14058038927888331729_t_real
    ccc(2,2,1,104)= 0.14058038927888331729_t_real
    ccc(2,2,3,104)= 0.58133008182238475428_t_real
    ccc(3,2,2,104)=-0.53346506936926446727_t_real
    !! M=-6
    !! L-1
    ccc(1,1,1,105)=-0.12565617248750865205_t_real
    ccc(1,1,3,105)= 0.56195148694901637936_t_real
    ccc(2,1,1,105)= 0.56195148694901637936_t_real
    ccc(2,1,3,105)= 0.12565617248750865205_t_real
    ccc(3,1,2,105)= 0.58038100008800933744_t_real
    !! L+1
    ccc(1,2,1,105)=-0.54988320714673246380_t_real
    ccc(1,2,3,105)= 0.17217511078754547242_t_real
    ccc(2,2,1,105)= 0.17217511078754547242_t_real
    ccc(2,2,3,105)= 0.54988320714673246380_t_real
    ccc(3,2,2,105)=-0.57962779388728445979_t_real
    !! M=-5
    !! L-1
    ccc(1,1,1,106)=-0.16222142113076253811_t_real
    ccc(1,1,3,106)= 0.52565748303784681342_t_real
    ccc(2,1,1,106)= 0.52565748303784681342_t_real
    ccc(2,1,3,106)= 0.16222142113076253811_t_real
    ccc(3,1,2,106)= 0.62828086243754326030_t_real
    !! L+1
    ccc(1,2,1,106)=-0.51843485951208205086_t_real
    ccc(1,2,3,106)= 0.20372033841662687615_t_real
    ccc(2,2,1,106)= 0.20372033841662687615_t_real
    ccc(2,2,3,106)= 0.51843485951208205086_t_real
    ccc(3,2,2,106)=-0.61599240280721444426_t_real
    !! M=-4
    !! L-1
    ccc(1,1,1,107)=-0.19867985355975657092_t_real
    ccc(1,1,3,107)= 0.48936048492959289577_t_real
    ccc(2,1,1,107)= 0.48936048492959289577_t_real
    ccc(2,1,3,107)= 0.19867985355975657092_t_real
    ccc(3,1,2,107)= 0.66490996620436870738_t_real
    !! L+1
    ccc(1,2,1,107)=-0.48698475355767397830_t_real
    ccc(1,2,3,107)= 0.23523598444848237187_t_real
    ccc(2,2,1,107)= 0.23523598444848237187_t_real
    ccc(2,2,3,107)= 0.48698475355767397830_t_real
    ccc(3,2,2,107)=-0.64422027509684123750_t_real
    !! M=-3
    !! L-1
    ccc(1,1,1,108)=-0.23508117299081362648_t_real
    ccc(1,1,3,108)= 0.45305977298225986065_t_real
    ccc(2,1,1,108)= 0.45305977298225986065_t_real
    ccc(2,1,3,108)= 0.23508117299081362648_t_real
    ccc(3,1,2,108)= 0.69206023467690486234_t_real
    !! L+1
    ccc(1,2,1,108)=-0.45553252509883956821_t_real
    ccc(1,2,3,108)= 0.26673253468463223363_t_real
    ccc(2,2,1,108)= 0.26673253468463223363_t_real
    ccc(2,2,3,108)= 0.45553252509883956821_t_real
    ccc(3,2,2,108)=-0.66534783913046048344_t_real
    !! M=-2
    !! L-1
    ccc(1,1,1,109)=-0.27144835701531845785_t_real
    ccc(1,1,3,109)= 0.41675437673324534550_t_real
    ccc(2,1,1,109)= 0.41675437673324534550_t_real
    ccc(2,1,3,109)= 0.27144835701531845785_t_real
    ccc(3,1,2,109)= 0.71081865331091082856_t_real
    !! L+1
    ccc(1,2,1,109)=-0.42407770188162638293_t_real
    ccc(1,2,3,109)= 0.29821603968282905209_t_real
    ccc(2,2,1,109)= 0.29821603968282905209_t_real
    ccc(2,2,3,109)= 0.42407770188162638293_t_real
    ccc(3,2,2,109)=-0.68003719963350203039_t_real
    !! M=-1
    !! L-1
    ccc(1,1,1,110)=-0.30779350562554622855_t_real
    ccc(2,1,1,110)= 0.53802758684897033411_t_real
    ccc(2,1,3,110)= 0.30779350562554622855_t_real
    ccc(3,1,2,110)= 0.72183975477868172631_t_real
    !! L+1
    ccc(1,2,1,110)=-0.39261966026736167253_t_real
    ccc(2,2,1,110)= 0.46625240412015688279_t_real
    ccc(2,2,3,110)= 0.39261966026736167253_t_real
    ccc(3,2,2,110)=-0.68870044315018188969_t_real
    !! M=0
    !! L-1
    ccc(1,1,3,111)=-0.48666426339228761435_t_real
    ccc(2,1,1,111)=-0.48666426339228761435_t_real
    ccc(3,1,2,111)= 0.72547625011001167180_t_real
    !! L+1
    ccc(1,2,3,111)=-0.51075391845524915591_t_real
    ccc(2,2,1,111)=-0.51075391845524915591_t_real
    ccc(3,2,2,111)=-0.69156407480812462404_t_real
    !! M=1
    !! L-1
    ccc(1,1,1,112)= 0.53802758684897033411_t_real
    ccc(1,1,3,112)=-0.30779350562554622855_t_real
    ccc(2,1,1,112)=-0.30779350562554622855_t_real
    ccc(3,1,2,112)= 0.72183975477868172631_t_real
    !! L+1
    ccc(1,2,1,112)= 0.46625240412015688279_t_real
    ccc(1,2,3,112)=-0.39261966026736167253_t_real
    ccc(2,2,1,112)=-0.39261966026736167253_t_real
    ccc(3,2,2,112)=-0.68870044315018188969_t_real
    !! M=2
    !! L-1
    ccc(1,1,1,113)= 0.41675437673324534550_t_real
    ccc(1,1,3,113)=-0.27144835701531845785_t_real
    ccc(2,1,1,113)=-0.27144835701531845785_t_real
    ccc(2,1,3,113)=-0.41675437673324534550_t_real
    ccc(3,1,2,113)= 0.71081865331091082856_t_real
    !! L+1
    ccc(1,2,1,113)= 0.29821603968282905209_t_real
    ccc(1,2,3,113)=-0.42407770188162638293_t_real
    ccc(2,2,1,113)=-0.42407770188162638293_t_real
    ccc(2,2,3,113)=-0.29821603968282905209_t_real
    ccc(3,2,2,113)=-0.68003719963350203039_t_real
    !! M=3
    !! L-1
    ccc(1,1,1,114)= 0.45305977298225986065_t_real
    ccc(1,1,3,114)=-0.23508117299081362648_t_real
    ccc(2,1,1,114)=-0.23508117299081362648_t_real
    ccc(2,1,3,114)=-0.45305977298225986065_t_real
    ccc(3,1,2,114)= 0.69206023467690486234_t_real
    !! L+1
    ccc(1,2,1,114)= 0.26673253468463223363_t_real
    ccc(1,2,3,114)=-0.45553252509883956821_t_real
    ccc(2,2,1,114)=-0.45553252509883956821_t_real
    ccc(2,2,3,114)=-0.26673253468463223363_t_real
    ccc(3,2,2,114)=-0.66534783913046048344_t_real
    !! M=4
    !! L-1
    ccc(1,1,1,115)= 0.48936048492959289577_t_real
    ccc(1,1,3,115)=-0.19867985355975657092_t_real
    ccc(2,1,1,115)=-0.19867985355975657092_t_real
    ccc(2,1,3,115)=-0.48936048492959289577_t_real
    ccc(3,1,2,115)= 0.66490996620436870738_t_real
    !! L+1
    ccc(1,2,1,115)= 0.23523598444848237187_t_real
    ccc(1,2,3,115)=-0.48698475355767397830_t_real
    ccc(2,2,1,115)=-0.48698475355767397830_t_real
    ccc(2,2,3,115)=-0.23523598444848237187_t_real
    ccc(3,2,2,115)=-0.64422027509684123750_t_real
    !! M=5
    !! L-1
    ccc(1,1,1,116)= 0.52565748303784681342_t_real
    ccc(1,1,3,116)=-0.16222142113076253811_t_real
    ccc(2,1,1,116)=-0.16222142113076253811_t_real
    ccc(2,1,3,116)=-0.52565748303784681342_t_real
    ccc(3,1,2,116)= 0.62828086243754326030_t_real
    !! L+1
    ccc(1,2,1,116)= 0.20372033841662687615_t_real
    ccc(1,2,3,116)=-0.51843485951208205086_t_real
    ccc(2,2,1,116)=-0.51843485951208205086_t_real
    ccc(2,2,3,116)=-0.20372033841662687615_t_real
    ccc(3,2,2,116)=-0.61599240280721444426_t_real
    !! M=6
    !! L-1
    ccc(1,1,1,117)= 0.56195148694901637936_t_real
    ccc(1,1,3,117)=-0.12565617248750865205_t_real
    ccc(2,1,1,117)=-0.12565617248750865205_t_real
    ccc(2,1,3,117)=-0.56195148694901637936_t_real
    ccc(3,1,2,117)= 0.58038100008800933744_t_real
    !! L+1
    ccc(1,2,1,117)= 0.17217511078754547242_t_real
    ccc(1,2,3,117)=-0.54988320714673246380_t_real
    ccc(2,2,1,117)=-0.54988320714673246380_t_real
    ccc(2,2,3,117)=-0.17217511078754547242_t_real
    ccc(3,2,2,117)=-0.57962779388728445979_t_real
    !! M=7
    !! L-1
    ccc(1,1,1,118)= 0.59824304161611881046_t_real
    ccc(1,1,3,118)=-0.08885233166386385357_t_real
    ccc(2,1,1,118)=-0.08885233166386385357_t_real
    ccc(2,1,3,118)=-0.59824304161611881046_t_real
    ccc(3,1,2,118)= 0.51809367167683002613_t_real
    !! L+1
    ccc(1,2,1,118)= 0.14058038927888331729_t_real
    ccc(1,2,3,118)=-0.58133008182238475428_t_real
    ccc(2,2,1,118)=-0.58133008182238475428_t_real
    ccc(2,2,3,118)=-0.14058038927888331729_t_real
    ccc(3,2,2,118)=-0.53346506936926446727_t_real
    !! M=8
    !! L-1
    ccc(1,1,1,119)= 0.63453256728663535228_t_real
    ccc(1,1,3,119)=-0.05129891760425770475_t_real
    ccc(2,1,1,119)=-0.05129891760425770475_t_real
    ccc(2,1,3,119)=-0.63453256728663535228_t_real
    ccc(3,1,2,119)= 0.43528575006600700308_t_real
    !! L+1
    ccc(1,2,1,119)= 0.10889310129609415676_t_real
    ccc(1,2,3,119)=-0.61277571031026112242_t_real
    ccc(2,2,1,119)=-0.61277571031026112242_t_real
    ccc(2,2,3,119)=-0.10889310129609415676_t_real
    ccc(3,2,2,119)=-0.47465402419841236983_t_real
    !! M=9
    !! L-1
    ccc(1,1,1,120)= 0.67082039324993690873_t_real
    ccc(2,1,3,120)=-0.67082039324993690873_t_real
    ccc(3,1,2,120)= 0.31622776601683793311_t_real
    !! L+1
    ccc(1,2,1,120)= 0.07699905035090180553_t_real
    ccc(1,2,3,120)=-0.64422027509684123750_t_real
    ccc(2,2,1,120)=-0.64422027509684123750_t_real
    ccc(2,2,3,120)=-0.07699905035090180553_t_real
    ccc(3,2,2,120)=-0.39762138624377206946_t_real
    !! M=10
    !! L-1
    ccc(1,1,1,121)= 0.70710678118654752420_t_real
    ccc(2,1,3,121)=-0.70710678118654752420_t_real
    !! L+1
    ccc(1,2,1,121)= 0.04445542244743870560_t_real
    ccc(1,2,3,121)=-0.67566392469217622316_t_real
    ccc(2,2,1,121)=-0.67566392469217622316_t_real
    ccc(2,2,3,121)=-0.04445542244743870560_t_real
    ccc(3,2,2,121)=-0.28810406552003038415_t_real
    !!
    If(maxL<=10) Return
    !!
    !!..............................................
    !! L=11
    !!..............................................
    !! M=-11
    !! L-1
    ccc(1,1,3,122)= 0.70710678118654752420_t_real
    ccc(2,1,1,122)= 0.70710678118654752420_t_real
    !! L+1
    ccc(1,2,1,122)=-0.67823299831252681377_t_real
    ccc(1,2,3,122)= 0.04082482904638630162_t_real
    ccc(2,2,1,122)= 0.04082482904638630162_t_real
    ccc(2,2,3,122)= 0.67823299831252681377_t_real
    ccc(3,2,2,122)=-0.27688746209726916170_t_real
    !! M=-10
    !! L-1
    ccc(1,1,3,123)= 0.67419986246324208605_t_real
    ccc(2,1,1,123)= 0.67419986246324208605_t_real
    ccc(3,1,2,123)= 0.30151134457776362256_t_real
    !! L+1
    ccc(1,2,1,123)=-0.64935865795927189324_t_real
    ccc(1,2,3,123)= 0.07071067811865475242_t_real
    ccc(2,2,1,123)= 0.07071067811865475242_t_real
    ccc(2,2,3,123)= 0.64935865795927189324_t_real
    ccc(3,2,2,123)=-0.38297084310253524391_t_real
    !! M=-9
    !! L-1
    ccc(1,1,1,124)=-0.04652421051992354336_t_real
    ccc(1,1,3,124)= 0.64129198595913488553_t_real
    ccc(2,1,1,124)= 0.64129198595913488553_t_real
    ccc(2,1,3,124)= 0.04652421051992354336_t_real
    ccc(3,1,2,124)= 0.41612518928823950740_t_real
    !! L+1
    ccc(1,2,1,124)=-0.62048368229954282968_t_real
    ccc(1,2,3,124)= 0.09999999999999999998_t_real
    ccc(2,2,1,124)= 0.09999999999999999998_t_real
    ccc(2,2,3,124)= 0.62048368229954282968_t_real
    ccc(3,2,2,124)=-0.45825756949558400057_t_real
    !! M=-8
    !! L-1
    ccc(1,1,1,125)=-0.08058229640253802982_t_real
    ccc(1,1,3,125)= 0.60838299625307587130_t_real
    ccc(2,1,1,125)= 0.60838299625307587130_t_real
    ccc(2,1,3,125)= 0.08058229640253802982_t_real
    ccc(3,1,2,125)= 0.49674263633520200689_t_real
    !! L+1
    ccc(1,2,1,125)=-0.59160797830996160414_t_real
    ccc(1,2,3,125)= 0.12909944487358056281_t_real
    ccc(2,2,1,125)= 0.12909944487358056281_t_real
    ccc(2,2,3,125)= 0.59160797830996160414_t_real
    ccc(3,2,2,125)=-0.51639777949432225125_t_real
    !! M=-7
    !! L-1
    ccc(1,1,1,126)=-0.11396057645963794880_t_real
    ccc(1,1,3,126)= 0.57547270236635131696_t_real
    ccc(2,1,1,126)= 0.57547270236635131696_t_real
    ccc(2,1,3,126)= 0.11396057645963794880_t_real
    ccc(3,1,2,126)= 0.55829052623908252034_t_real
    !! L+1
    ccc(1,2,1,126)=-0.56273143387113773933_t_real
    ccc(1,2,3,126)= 0.15811388300841896656_t_real
    ccc(2,2,1,126)= 0.15811388300841896656_t_real
    ccc(2,2,3,126)= 0.56273143387113773933_t_real
    ccc(3,2,2,126)=-0.56273143387113773933_t_real
    !! M=-6
    !! L-1
    ccc(1,1,1,127)=-0.14712247158412492157_t_real
    ccc(1,1,3,127)= 0.54256086697465969392_t_real
    ccc(2,1,1,127)= 0.54256086697465969392_t_real
    ccc(2,1,3,127)= 0.14712247158412492157_t_real
    ccc(3,1,2,127)= 0.60660149024327987152_t_real
    !! L+1
    ccc(1,2,1,127)=-0.53385391260156556043_t_real
    ccc(1,2,3,127)= 0.18708286933869706924_t_real
    ccc(2,2,1,127)= 0.18708286933869706924_t_real
    ccc(2,2,3,127)= 0.53385391260156556043_t_real
    ccc(3,2,2,127)=-0.59999999999999999988_t_real
    !! M=-5
    !! L-1
    ccc(1,1,1,128)=-0.18018749253911179146_t_real
    ccc(1,1,3,128)= 0.50964719143762554449_t_real
    ccc(2,1,1,128)= 0.50964719143762554449_t_real
    ccc(2,1,3,128)= 0.18018749253911179146_t_real
    ccc(3,1,2,128)= 0.64465837122030423857_t_real
    !! L+1
    ccc(1,2,1,128)=-0.50497524691810389757_t_real
    ccc(1,2,3,128)= 0.21602468994692867432_t_real
    ccc(2,2,1,128)= 0.21602468994692867432_t_real
    ccc(2,2,3,128)= 0.50497524691810389757_t_real
    ccc(3,2,2,128)=-0.62981478758970614258_t_real
    !! M=-4
    !! L-1
    ccc(1,1,1,129)=-0.21320071635561043423_t_real
    ccc(1,1,3,129)= 0.47673129462279615758_t_real
    ccc(2,1,1,129)= 0.47673129462279615758_t_real
    ccc(2,1,3,129)= 0.21320071635561043423_t_real
    ccc(3,1,2,129)= 0.67419986246324208605_t_real
    !! L+1
    ccc(1,2,1,129)=-0.47609522856952333310_t_real
    ccc(1,2,3,129)= 0.24494897427831780977_t_real
    ccc(2,2,1,129)= 0.24494897427831780977_t_real
    ccc(2,2,3,129)= 0.47609522856952333310_t_real
    ccc(3,2,2,129)=-0.65319726474218082606_t_real
    !! M=-3
    !! L-1
    ccc(1,1,1,130)=-0.24618298195866546539_t_real
    ccc(1,1,3,130)= 0.44381268229929726692_t_real
    ccc(2,1,1,130)= 0.44381268229929726692_t_real
    ccc(2,1,3,130)= 0.24618298195866546539_t_real
    ccc(3,1,2,130)= 0.69631062382279135251_t_real
    !! L+1
    ccc(1,2,1,130)=-0.44721359549995793919_t_real
    ccc(1,2,3,130)= 0.27386127875258305667_t_real
    ccc(2,2,1,130)= 0.27386127875258305667_t_real
    ccc(2,2,3,130)= 0.44721359549995793919_t_real
    ccc(3,2,2,130)=-0.67082039324993690879_t_real
    !! M=-2
    !! L-1
    ccc(1,1,1,131)=-0.27914526311954126017_t_real
    ccc(1,1,3,131)= 0.41089070180665907966_t_real
    ccc(2,1,1,131)= 0.41089070180665907966_t_real
    ccc(2,1,3,131)= 0.27914526311954126017_t_real
    ccc(3,1,2,131)= 0.71168357188676660728_t_real
    !! L+1
    ccc(1,2,1,131)=-0.41833001326703777390_t_real
    ccc(1,2,3,131)= 0.30276503540974916648_t_real
    ccc(2,2,1,131)= 0.30276503540974916648_t_real
    ccc(2,2,3,131)= 0.41833001326703777390_t_real
    ccc(3,2,2,131)=-0.68313005106397322541_t_real
    !! M=-1
    !! L-1
    ccc(1,1,1,132)=-0.31209389196617963055_t_real
    ccc(2,1,1,132)= 0.53452248382484876920_t_real
    ccc(2,1,3,132)= 0.31209389196617963055_t_real
    ccc(3,1,2,132)= 0.72074997015644716586_t_real
    !! L+1
    ccc(1,2,1,132)=-0.38944404818493075361_t_real
    ccc(2,2,1,132)= 0.46904157598234295535_t_real
    ccc(2,2,3,132)= 0.38944404818493075361_t_real
    ccc(3,2,2,132)=-0.69041050590693263760_t_real
    !! M=0
    !! L-1
    ccc(1,1,3,133)=-0.48795003647426658953_t_real
    ccc(2,1,1,133)=-0.48795003647426658953_t_real
    ccc(3,1,2,133)= 0.72374686445574588702_t_real
    !! L+1
    ccc(1,2,3,133)=-0.50990195135927848290_t_real
    ccc(2,2,1,133)=-0.50990195135927848290_t_real
    ccc(3,2,2,133)=-0.69282032302755091728_t_real
    !! M=1
    !! L-1
    ccc(1,1,1,134)= 0.53452248382484876920_t_real
    ccc(1,1,3,134)=-0.31209389196617963055_t_real
    ccc(2,1,1,134)=-0.31209389196617963055_t_real
    ccc(3,1,2,134)= 0.72074997015644716586_t_real
    !! L+1
    ccc(1,2,1,134)= 0.46904157598234295535_t_real
    ccc(1,2,3,134)=-0.38944404818493075361_t_real
    ccc(2,2,1,134)=-0.38944404818493075361_t_real
    ccc(3,2,2,134)=-0.69041050590693263760_t_real
    !! M=2
    !! L-1
    ccc(1,1,1,135)= 0.41089070180665907966_t_real
    ccc(1,1,3,135)=-0.27914526311954126017_t_real
    ccc(2,1,1,135)=-0.27914526311954126017_t_real
    ccc(2,1,3,135)=-0.41089070180665907966_t_real
    ccc(3,1,2,135)= 0.71168357188676660728_t_real
    !! L+1
    ccc(1,2,1,135)= 0.30276503540974916648_t_real
    ccc(1,2,3,135)=-0.41833001326703777390_t_real
    ccc(2,2,1,135)=-0.41833001326703777390_t_real
    ccc(2,2,3,135)=-0.30276503540974916648_t_real
    ccc(3,2,2,135)=-0.68313005106397322541_t_real
    !! M=3
    !! L-1
    ccc(1,1,1,136)= 0.44381268229929726692_t_real
    ccc(1,1,3,136)=-0.24618298195866546539_t_real
    ccc(2,1,1,136)=-0.24618298195866546539_t_real
    ccc(2,1,3,136)=-0.44381268229929726692_t_real
    ccc(3,1,2,136)= 0.69631062382279135251_t_real
    !! L+1
    ccc(1,2,1,136)= 0.27386127875258305667_t_real
    ccc(1,2,3,136)=-0.44721359549995793919_t_real
    ccc(2,2,1,136)=-0.44721359549995793919_t_real
    ccc(2,2,3,136)=-0.27386127875258305667_t_real
    ccc(3,2,2,136)=-0.67082039324993690879_t_real
    !! M=4
    !! L-1
    ccc(1,1,1,137)= 0.47673129462279615758_t_real
    ccc(1,1,3,137)=-0.21320071635561043423_t_real
    ccc(2,1,1,137)=-0.21320071635561043423_t_real
    ccc(2,1,3,137)=-0.47673129462279615758_t_real
    ccc(3,1,2,137)= 0.67419986246324208605_t_real
    !! L+1
    ccc(1,2,1,137)= 0.24494897427831780977_t_real
    ccc(1,2,3,137)=-0.47609522856952333310_t_real
    ccc(2,2,1,137)=-0.47609522856952333310_t_real
    ccc(2,2,3,137)=-0.24494897427831780977_t_real
    ccc(3,2,2,137)=-0.65319726474218082606_t_real
    !! M=5
    !! L-1
    ccc(1,1,1,138)= 0.50964719143762554449_t_real
    ccc(1,1,3,138)=-0.18018749253911179146_t_real
    ccc(2,1,1,138)=-0.18018749253911179146_t_real
    ccc(2,1,3,138)=-0.50964719143762554449_t_real
    ccc(3,1,2,138)= 0.64465837122030423857_t_real
    !! L+1
    ccc(1,2,1,138)= 0.21602468994692867432_t_real
    ccc(1,2,3,138)=-0.50497524691810389757_t_real
    ccc(2,2,1,138)=-0.50497524691810389757_t_real
    ccc(2,2,3,138)=-0.21602468994692867432_t_real
    ccc(3,2,2,138)=-0.62981478758970614258_t_real
    !! M=6
    !! L-1
    ccc(1,1,1,139)= 0.54256086697465969392_t_real
    ccc(1,1,3,139)=-0.14712247158412492157_t_real
    ccc(2,1,1,139)=-0.14712247158412492157_t_real
    ccc(2,1,3,139)=-0.54256086697465969392_t_real
    ccc(3,1,2,139)= 0.60660149024327987152_t_real
    !! L+1
    ccc(1,2,1,139)= 0.18708286933869706924_t_real
    ccc(1,2,3,139)=-0.53385391260156556043_t_real
    ccc(2,2,1,139)=-0.53385391260156556043_t_real
    ccc(2,2,3,139)=-0.18708286933869706924_t_real
    ccc(3,2,2,139)=-0.59999999999999999988_t_real
    !! M=7
    !! L-1
    ccc(1,1,1,140)= 0.57547270236635131696_t_real
    ccc(1,1,3,140)=-0.11396057645963794880_t_real
    ccc(2,1,1,140)=-0.11396057645963794880_t_real
    ccc(2,1,3,140)=-0.57547270236635131696_t_real
    ccc(3,1,2,140)= 0.55829052623908252034_t_real
    !! L+1
    ccc(1,2,1,140)= 0.15811388300841896656_t_real
    ccc(1,2,3,140)=-0.56273143387113773933_t_real
    ccc(2,2,1,140)=-0.56273143387113773933_t_real
    ccc(2,2,3,140)=-0.15811388300841896656_t_real
    ccc(3,2,2,140)=-0.56273143387113773933_t_real
    !! M=8
    !! L-1
    ccc(1,1,1,141)= 0.60838299625307587130_t_real
    ccc(1,1,3,141)=-0.08058229640253802982_t_real
    ccc(2,1,1,141)=-0.08058229640253802982_t_real
    ccc(2,1,3,141)=-0.60838299625307587130_t_real
    ccc(3,1,2,141)= 0.49674263633520200689_t_real
    !! L+1
    ccc(1,2,1,141)= 0.12909944487358056281_t_real
    ccc(1,2,3,141)=-0.59160797830996160414_t_real
    ccc(2,2,1,141)=-0.59160797830996160414_t_real
    ccc(2,2,3,141)=-0.12909944487358056281_t_real
    ccc(3,2,2,141)=-0.51639777949432225125_t_real
    !! M=9
    !! L-1
    ccc(1,1,1,142)= 0.64129198595913488553_t_real
    ccc(1,1,3,142)=-0.04652421051992354336_t_real
    ccc(2,1,1,142)=-0.04652421051992354336_t_real
    ccc(2,1,3,142)=-0.64129198595913488553_t_real
    ccc(3,1,2,142)= 0.41612518928823950740_t_real
    !! L+1
    ccc(1,2,1,142)= 0.09999999999999999998_t_real
    ccc(1,2,3,142)=-0.62048368229954282968_t_real
    ccc(2,2,1,142)=-0.62048368229954282968_t_real
    ccc(2,2,3,142)=-0.09999999999999999998_t_real
    ccc(3,2,2,142)=-0.45825756949558400057_t_real
    !! M=10
    !! L-1
    ccc(1,1,1,143)= 0.67419986246324208605_t_real
    ccc(2,1,3,143)=-0.67419986246324208605_t_real
    ccc(3,1,2,143)= 0.30151134457776362256_t_real
    !! L+1
    ccc(1,2,1,143)= 0.07071067811865475242_t_real
    ccc(1,2,3,143)=-0.64935865795927189324_t_real
    ccc(2,2,1,143)=-0.64935865795927189324_t_real
    ccc(2,2,3,143)=-0.07071067811865475242_t_real
    ccc(3,2,2,143)=-0.38297084310253524391_t_real
    !! M=11
    !! L-1
    ccc(1,1,1,144)= 0.70710678118654752420_t_real
    ccc(2,1,3,144)=-0.70710678118654752420_t_real
    !! L+1
    ccc(1,2,1,144)= 0.04082482904638630162_t_real
    ccc(1,2,3,144)=-0.67823299831252681377_t_real
    ccc(2,2,1,144)=-0.67823299831252681377_t_real
    ccc(2,2,3,144)=-0.04082482904638630162_t_real
    ccc(3,2,2,144)=-0.27688746209726916170_t_real
    !!
    If(maxL<=11) Return
    !!
    !!..............................................
    !! L=12
    !!..............................................
    !! M=-12
    !! L-1
    ccc(1,1,3,145)= 0.70710678118654752438_t_real
    ccc(2,1,1,145)= 0.70710678118654752438_t_real
    !! L+1
    ccc(1,2,1,145)=-0.68041381743977169384_t_real
    ccc(1,2,3,145)= 0.03774256780481985834_t_real
    ccc(2,2,1,145)= 0.03774256780481985834_t_real
    ccc(2,2,3,145)= 0.68041381743977169384_t_real
    ccc(3,2,2,145)=-0.26688025634181188910_t_real
    !! M=-11
    !! L-1
    ccc(1,1,3,146)= 0.67700320038633003002_t_real
    ccc(2,1,1,146)= 0.67700320038633003002_t_real
    ccc(3,1,2,146)= 0.28867513459481288224_t_real
    !! L+1
    ccc(1,2,1,146)=-0.65372045046061343965_t_real
    ccc(1,2,3,146)= 0.06537204504606134396_t_real
    ccc(2,2,1,146)= 0.06537204504606134396_t_real
    ccc(2,2,3,146)= 0.65372045046061343965_t_real
    ccc(3,2,2,146)=-0.36980013081681941462_t_real
    !! M=-10
    !! L-1
    ccc(1,1,1,147)=-0.04256282653793742966_t_real
    ccc(1,1,3,147)= 0.64689895723332652267_t_real
    ccc(2,1,1,147)= 0.64689895723332652267_t_real
    ccc(2,1,3,147)= 0.04256282653793742966_t_real
    ccc(3,1,2,147)= 0.39927470475234524148_t_real
    !! L+1
    ccc(1,2,1,147)=-0.62702662875064019268_t_real
    ccc(1,2,3,147)= 0.09245003270420485365_t_real
    ccc(2,2,1,147)= 0.09245003270420485365_t_real
    ccc(2,2,3,147)= 0.62702662875064019268_t_real
    ccc(3,2,2,147)=-0.44337478117411750366_t_real
    !! M=-9
    !! L-1
    ccc(1,1,1,148)=-0.07372097807744856672_t_real
    ccc(1,1,3,148)= 0.61679395474396119616_t_real
    ccc(2,1,1,148)= 0.61679395474396119616_t_real
    ccc(2,1,3,148)= 0.07372097807744856672_t_real
    ccc(3,1,2,148)= 0.47776654295295457753_t_real
    !! L+1
    ccc(1,2,1,148)=-0.60033229165093259936_t_real
    ccc(1,2,3,148)= 0.11935247900657214673_t_real
    ccc(2,2,1,148)= 0.11935247900657214673_t_real
    ccc(2,2,3,148)= 0.60033229165093259936_t_real
    ccc(3,2,2,148)=-0.50071174413254050633_t_real
    !! M=-8
    !! L-1
    ccc(1,1,1,149)=-0.10425720702853738133_t_real
    ccc(1,1,3,149)= 0.58668807602568900144_t_real
    ccc(2,1,1,149)= 0.58668807602568900144_t_real
    ccc(2,1,3,149)= 0.10425720702853738133_t_real
    ccc(3,1,2,149)= 0.53838190205816550639_t_real
    !! L+1
    ccc(1,2,1,149)=-0.57363736721018188890_t_real
    ccc(1,2,3,149)= 0.14617633655117153576_t_real
    ccc(2,2,1,149)= 0.14617633655117153576_t_real
    ccc(2,2,3,149)= 0.57363736721018188890_t_real
    ccc(3,2,2,149)=-0.54694176942824466019_t_real
    !! M=-7
    !! L-1
    ccc(1,1,1,150)=-0.13459547551454137659_t_real
    ccc(1,1,3,150)= 0.55658117889096121851_t_real
    ccc(2,1,1,150)= 0.55658117889096121851_t_real
    ccc(2,1,3,150)= 0.13459547551454137659_t_real
    ccc(3,1,2,150)= 0.58668807602568900144_t_real
    !! L+1
    ccc(1,2,1,150)=-0.54694176942824466019_t_real
    ccc(1,2,3,150)= 0.17295817388759027517_t_real
    ccc(2,2,1,150)= 0.17295817388759027517_t_real
    ccc(2,2,3,150)= 0.54694176942824466019_t_real
    ccc(3,2,2,150)=-0.58470534620468614306_t_real
    !! M=-6
    !! L-1
    ccc(1,1,1,151)=-0.16484511834894675117_t_real
    ccc(1,1,3,151)= 0.52647308862227541293_t_real
    ccc(2,1,1,151)= 0.52647308862227541293_t_real
    ccc(2,1,3,151)= 0.16484511834894675117_t_real
    ccc(3,1,2,151)= 0.62554324217122428802_t_real
    !! L+1
    ccc(1,2,1,151)=-0.52024539465070774924_t_real
    ccc(1,2,3,151)= 0.19971489650509269417_t_real
    ccc(2,2,1,151)= 0.19971489650509269417_t_real
    ccc(2,2,3,151)= 0.52024539465070774924_t_real
    ccc(3,2,2,151)=-0.61556265230874664986_t_real
    !! M=-5
    !! L-1
    ccc(1,1,1,152)=-0.19504737440137348646_t_real
    ccc(1,1,3,152)= 0.49636358810271619131_t_real
    ccc(2,1,1,152)= 0.49636358810271619131_t_real
    ccc(2,1,3,152)= 0.19504737440137348646_t_real
    ccc(3,1,2,152)= 0.65662730699374289154_t_real
    !! L+1
    ccc(1,2,1,152)=-0.49354811679282455719_t_real
    ccc(1,2,3,152)= 0.22645540682891915008_t_real
    ccc(2,2,1,152)= 0.22645540682891915008_t_real
    ccc(2,2,3,152)= 0.49354811679282455719_t_real
    ccc(3,2,2,152)=-0.64051261522034853384_t_real
    !! M=-4
    !! L-1
    ccc(1,1,1,153)=-0.22522130823072540772_t_real
    ccc(1,1,3,153)= 0.46625240412015688280_t_real
    ccc(2,1,1,153)= 0.46625240412015688280_t_real
    ccc(2,1,3,153)= 0.22522130823072540772_t_real
    ccc(3,1,2,153)= 0.68100522460699887459_t_real
    !! L+1
    ccc(1,2,1,153)=-0.46684978092392626574_t_real
    ccc(1,2,3,153)= 0.25318484177091665403_t_real
    ccc(2,2,1,153)= 0.25318484177091665403_t_real
    ccc(2,2,3,153)= 0.46684978092392626574_t_real
    ccc(3,2,2,153)=-0.66022529177352475710_t_real
    !! M=-3
    !! L-1
    ccc(1,1,1,154)=-0.25537695922762457797_t_real
    ccc(1,1,3,154)= 0.43613918799432346592_t_real
    ccc(2,1,1,154)= 0.43613918799432346592_t_real
    ccc(2,1,3,154)= 0.25537695922762457797_t_real
    ccc(3,1,2,154)= 0.69937860618023532421_t_real
    !! L+1
    ccc(1,2,1,154)=-0.44015019451568317140_t_real
    ccc(1,2,3,154)= 0.27990637425321051011_t_real
    ccc(2,2,1,154)= 0.27990637425321051011_t_real
    ccc(2,2,3,154)= 0.44015019451568317140_t_real
    ccc(3,2,2,154)=-0.67515957805577774408_t_real
    !! M=-2
    !! L-1
    ccc(1,1,1,155)=-0.28552012036008037275_t_real
    ccc(1,1,3,155)= 0.40602348757648013180_t_real
    ccc(2,1,1,155)= 0.40602348757648013180_t_real
    ccc(2,1,3,155)= 0.28552012036008037275_t_real
    ccc(3,1,2,155)= 0.71221231161191969588_t_real
    !! L+1
    ccc(1,2,1,155)=-0.41344911529736152057_t_real
    ccc(1,2,3,155)= 0.30662207033593328254_t_real
    ccc(2,2,1,155)= 0.30662207033593328254_t_real
    ccc(2,2,3,155)= 0.41344911529736152057_t_real
    ccc(3,2,2,155)=-0.68562779267286859696_t_real
    !! M=-1
    !! L-1
    ccc(1,1,1,156)=-0.31565436977716669945_t_real
    ccc(2,1,1,156)= 0.53160953307119494602_t_real
    ccc(2,1,3,156)= 0.31565436977716669945_t_real
    ccc(3,1,2,156)= 0.71980271049016306634_t_real
    !! L+1
    ccc(1,2,1,156)=-0.38674623407688092523_t_real
    ccc(2,2,1,156)= 0.47140452079103168285_t_real
    ccc(2,2,3,156)= 0.38674623407688092523_t_real
    ccc(3,2,2,156)=-0.69183269555036110069_t_real
    !! M=0
    !! L-1
    ccc(1,1,3,157)=-0.48900964692182576168_t_real
    ccc(2,1,1,157)=-0.48900964692182576168_t_real
    ccc(3,1,2,157)= 0.72231511851461520456_t_real
    !! L+1
    ccc(1,2,3,157)=-0.50917507721731555620_t_real
    ccc(2,2,1,157)=-0.50917507721731555620_t_real
    ccc(3,2,2,157)=-0.69388866648871091166_t_real
    !! M=1
    !! L-1
    ccc(1,1,1,158)= 0.53160953307119494602_t_real
    ccc(1,1,3,158)=-0.31565436977716669945_t_real
    ccc(2,1,1,158)=-0.31565436977716669945_t_real
    ccc(3,1,2,158)= 0.71980271049016306634_t_real
    !! L+1
    ccc(1,2,1,158)= 0.47140452079103168285_t_real
    ccc(1,2,3,158)=-0.38674623407688092523_t_real
    ccc(2,2,1,158)=-0.38674623407688092523_t_real
    ccc(3,2,2,158)=-0.69183269555036110069_t_real
    !! M=2
    !! L-1
    ccc(1,1,1,159)= 0.40602348757648013180_t_real
    ccc(1,1,3,159)=-0.28552012036008037275_t_real
    ccc(2,1,1,159)=-0.28552012036008037275_t_real
    ccc(2,1,3,159)=-0.40602348757648013180_t_real
    ccc(3,1,2,159)= 0.71221231161191969588_t_real
    !! L+1
    ccc(1,2,1,159)= 0.30662207033593328254_t_real
    ccc(1,2,3,159)=-0.41344911529736152057_t_real
    ccc(2,2,1,159)=-0.41344911529736152057_t_real
    ccc(2,2,3,159)=-0.30662207033593328254_t_real
    ccc(3,2,2,159)=-0.68562779267286859696_t_real
    !! M=3
    !! L-1
    ccc(1,1,1,160)= 0.43613918799432346592_t_real
    ccc(1,1,3,160)=-0.25537695922762457797_t_real
    ccc(2,1,1,160)=-0.25537695922762457797_t_real
    ccc(2,1,3,160)=-0.43613918799432346592_t_real
    ccc(3,1,2,160)= 0.69937860618023532421_t_real
    !! L+1
    ccc(1,2,1,160)= 0.27990637425321051011_t_real
    ccc(1,2,3,160)=-0.44015019451568317140_t_real
    ccc(2,2,1,160)=-0.44015019451568317140_t_real
    ccc(2,2,3,160)=-0.27990637425321051011_t_real
    ccc(3,2,2,160)=-0.67515957805577774408_t_real
    !! M=4
    !! L-1
    ccc(1,1,1,161)= 0.46625240412015688280_t_real
    ccc(1,1,3,161)=-0.22522130823072540772_t_real
    ccc(2,1,1,161)=-0.22522130823072540772_t_real
    ccc(2,1,3,161)=-0.46625240412015688280_t_real
    ccc(3,1,2,161)= 0.68100522460699887459_t_real
    !! L+1
    ccc(1,2,1,161)= 0.25318484177091665403_t_real
    ccc(1,2,3,161)=-0.46684978092392626574_t_real
    ccc(2,2,1,161)=-0.46684978092392626574_t_real
    ccc(2,2,3,161)=-0.25318484177091665403_t_real
    ccc(3,2,2,161)=-0.66022529177352475710_t_real
    !! M=5
    !! L-1
    ccc(1,1,1,162)= 0.49636358810271619131_t_real
    ccc(1,1,3,162)=-0.19504737440137348646_t_real
    ccc(2,1,1,162)=-0.19504737440137348646_t_real
    ccc(2,1,3,162)=-0.49636358810271619131_t_real
    ccc(3,1,2,162)= 0.65662730699374289154_t_real
    !! L+1
    ccc(1,2,1,162)= 0.22645540682891915008_t_real
    ccc(1,2,3,162)=-0.49354811679282455719_t_real
    ccc(2,2,1,162)=-0.49354811679282455719_t_real
    ccc(2,2,3,162)=-0.22645540682891915008_t_real
    ccc(3,2,2,162)=-0.64051261522034853384_t_real
    !! M=6
    !! L-1
    ccc(1,1,1,163)= 0.52647308862227541293_t_real
    ccc(1,1,3,163)=-0.16484511834894675117_t_real
    ccc(2,1,1,163)=-0.16484511834894675117_t_real
    ccc(2,1,3,163)=-0.52647308862227541293_t_real
    ccc(3,1,2,163)= 0.62554324217122428802_t_real
    !! L+1
    ccc(1,2,1,163)= 0.19971489650509269417_t_real
    ccc(1,2,3,163)=-0.52024539465070774924_t_real
    ccc(2,2,1,163)=-0.52024539465070774924_t_real
    ccc(2,2,3,163)=-0.19971489650509269417_t_real
    ccc(3,2,2,163)=-0.61556265230874664986_t_real
    !! M=7
    !! L-1
    ccc(1,1,1,164)= 0.55658117889096121851_t_real
    ccc(1,1,3,164)=-0.13459547551454137659_t_real
    ccc(2,1,1,164)=-0.13459547551454137659_t_real
    ccc(2,1,3,164)=-0.55658117889096121851_t_real
    ccc(3,1,2,164)= 0.58668807602568900144_t_real
    !! L+1
    ccc(1,2,1,164)= 0.17295817388759027517_t_real
    ccc(1,2,3,164)=-0.54694176942824466019_t_real
    ccc(2,2,1,164)=-0.54694176942824466019_t_real
    ccc(2,2,3,164)=-0.17295817388759027517_t_real
    ccc(3,2,2,164)=-0.58470534620468614306_t_real
    !! M=8
    !! L-1
    ccc(1,1,1,165)= 0.58668807602568900144_t_real
    ccc(1,1,3,165)=-0.10425720702853738133_t_real
    ccc(2,1,1,165)=-0.10425720702853738133_t_real
    ccc(2,1,3,165)=-0.58668807602568900144_t_real
    ccc(3,1,2,165)= 0.53838190205816550639_t_real
    !! L+1
    ccc(1,2,1,165)= 0.14617633655117153576_t_real
    ccc(1,2,3,165)=-0.57363736721018188890_t_real
    ccc(2,2,1,165)=-0.57363736721018188890_t_real
    ccc(2,2,3,165)=-0.14617633655117153576_t_real
    ccc(3,2,2,165)=-0.54694176942824466019_t_real
    !! M=9
    !! L-1
    ccc(1,1,1,166)= 0.61679395474396119616_t_real
    ccc(1,1,3,166)=-0.07372097807744856672_t_real
    ccc(2,1,1,166)=-0.07372097807744856672_t_real
    ccc(2,1,3,166)=-0.61679395474396119616_t_real
    ccc(3,1,2,166)= 0.47776654295295457753_t_real
    !! L+1
    ccc(1,2,1,166)= 0.11935247900657214673_t_real
    ccc(1,2,3,166)=-0.60033229165093259936_t_real
    ccc(2,2,1,166)=-0.60033229165093259936_t_real
    ccc(2,2,3,166)=-0.11935247900657214673_t_real
    ccc(3,2,2,166)=-0.50071174413254050633_t_real
    !! M=10
    !! L-1
    ccc(1,1,1,167)= 0.64689895723332652267_t_real
    ccc(1,1,3,167)=-0.04256282653793742966_t_real
    ccc(2,1,1,167)=-0.04256282653793742966_t_real
    ccc(2,1,3,167)=-0.64689895723332652267_t_real
    ccc(3,1,2,167)= 0.39927470475234524148_t_real
    !! L+1
    ccc(1,2,1,167)= 0.09245003270420485365_t_real
    ccc(1,2,3,167)=-0.62702662875064019268_t_real
    ccc(2,2,1,167)=-0.62702662875064019268_t_real
    ccc(2,2,3,167)=-0.09245003270420485365_t_real
    ccc(3,2,2,167)=-0.44337478117411750366_t_real
    !! M=11
    !! L-1
    ccc(1,1,1,168)= 0.67700320038633003002_t_real
    ccc(2,1,3,168)=-0.67700320038633003002_t_real
    ccc(3,1,2,168)= 0.28867513459481288224_t_real
    !! L+1
    ccc(1,2,1,168)= 0.06537204504606134396_t_real
    ccc(1,2,3,168)=-0.65372045046061343965_t_real
    ccc(2,2,1,168)=-0.65372045046061343965_t_real
    ccc(2,2,3,168)=-0.06537204504606134396_t_real
    ccc(3,2,2,168)=-0.36980013081681941462_t_real
    !! M=12
    !! L-1
    ccc(1,1,1,169)= 0.70710678118654752438_t_real
    ccc(2,1,3,169)=-0.70710678118654752438_t_real
    !! L+1
    ccc(1,2,1,169)= 0.03774256780481985834_t_real
    ccc(1,2,3,169)=-0.68041381743977169384_t_real
    ccc(2,2,1,169)=-0.68041381743977169384_t_real
    ccc(2,2,3,169)=-0.03774256780481985834_t_real
    ccc(3,2,2,169)=-0.26688025634181188910_t_real
    !!
    If(maxL<=12) Return
    !!
    !!..............................................
    !! L=13
    !!..............................................
    !! M=-13
    !! L-1
    ccc(1,1,3,170)= 0.70710678118654752437_t_real
    ccc(2,1,1,170)= 0.70710678118654752437_t_real
    !! L+1
    ccc(1,2,1,170)=-0.68228823922101306055_t_real
    ccc(1,2,3,170)= 0.03509312031717982061_t_real
    ccc(2,2,1,170)= 0.03509312031717982061_t_real
    ccc(2,2,3,170)= 0.68228823922101306055_t_real
    ccc(3,2,2,170)=-0.25788071477756376056_t_real
    !! M=-12
    !! L-1
    ccc(1,1,3,171)= 0.67936622048675745034_t_real
    ccc(2,1,1,171)= 0.67936622048675745034_t_real
    ccc(3,1,2,171)= 0.27735009811261456100_t_real
    !! L+1
    ccc(1,2,1,171)=-0.65746939841502642270_t_real
    ccc(1,2,3,171)= 0.06078306738548308356_t_real
    ccc(2,2,1,171)= 0.06078306738548308356_t_real
    ccc(2,2,3,171)= 0.65746939841502642270_t_real
    ccc(3,2,2,171)=-0.35788101058031864758_t_real
    !! M=-11
    !! L-1
    ccc(1,1,1,172)=-0.03922322702763680638_t_real
    ccc(1,1,3,172)= 0.65162518721684217549_t_real
    ccc(2,1,1,172)= 0.65162518721684217549_t_real
    ccc(2,1,3,172)= 0.03922322702763680638_t_real
    ccc(3,1,2,172)= 0.38430756913220912034_t_real
    !! L+1
    ccc(1,2,1,172)=-0.63265022359809469351_t_real
    ccc(1,2,3,172)= 0.08596023825918792018_t_real
    ccc(2,2,1,172)= 0.08596023825918792018_t_real
    ccc(2,2,3,172)= 0.63265022359809469351_t_real
    ccc(3,2,2,172)=-0.42980119129593960094_t_real
    !! M=-10
    !! L-1
    ccc(1,1,1,173)=-0.06793662204867574503_t_real
    ccc(1,1,3,173)= 0.62388361833820354826_t_real
    ccc(2,1,1,173)= 0.62388361833820354826_t_real
    ccc(2,1,3,173)= 0.06793662204867574503_t_real
    ccc(3,1,2,173)= 0.46076858867298268520_t_real
    !! L+1
    ccc(1,2,1,173)=-0.60783067385483083563_t_real
    ccc(1,2,3,173)= 0.11097419040461881708_t_real
    ccc(2,2,1,173)= 0.11097419040461881708_t_real
    ccc(2,2,3,173)= 0.60783067385483083563_t_real
    ccc(3,2,2,173)=-0.48626453908386466850_t_real
    !! M=-9
    !! L-1
    ccc(1,1,1,174)=-0.09607689228305228008_t_real
    ccc(1,1,3,174)= 0.59614143907684809140_t_real
    ccc(2,1,1,174)= 0.59614143907684809140_t_real
    ccc(2,1,3,174)= 0.09607689228305228008_t_real
    ccc(3,1,2,174)= 0.52035490847039270454_t_real
    !! L+1
    ccc(1,2,1,174)=-0.58301070130188203540_t_real
    ccc(1,2,3,174)= 0.13591507055489058862_t_real
    ccc(2,2,1,174)= 0.13591507055489058862_t_real
    ccc(2,2,3,174)= 0.58301070130188203540_t_real
    ccc(3,2,2,174)=-0.53221352061657884571_t_real
    !! M=-8
    !! L-1
    ccc(1,1,1,175)=-0.12403473458920845618_t_real
    ccc(1,1,3,175)= 0.56839856005880510754_t_real
    ccc(2,1,1,175)= 0.56839856005880510754_t_real
    ccc(2,1,3,175)= 0.12403473458920845618_t_real
    ccc(3,1,2,175)= 0.56839856005880510754_t_real
    !! L+1
    ccc(1,2,1,175)=-0.55819024953845526273_t_real
    ccc(1,2,3,175)= 0.16081688022566922492_t_real
    ccc(2,2,1,175)= 0.16081688022566922492_t_real
    ccc(2,2,3,175)= 0.55819024953845526273_t_real
    ccc(3,2,2,175)=-0.57019571439055871414_t_real
    !! M=-7
    !! L-1
    ccc(1,1,1,176)=-0.15191090506254999243_t_real
    ccc(1,1,3,176)= 0.54065487356324857942_t_real
    ccc(2,1,1,176)= 0.54065487356324857942_t_real
    ccc(2,1,3,176)= 0.15191090506254999243_t_real
    ccc(3,1,2,176)= 0.60764362025019996974_t_real
    !! L+1
    ccc(1,2,1,176)=-0.53336925166406964067_t_real
    ccc(1,2,3,176)= 0.18569533817705186314_t_real
    ccc(2,2,1,176)= 0.18569533817705186314_t_real
    ccc(2,2,3,176)= 0.53336925166406964067_t_real
    ccc(3,2,2,176)=-0.60172166781431544132_t_real
    !! M=-6
    !! L-1
    ccc(1,1,1,177)=-0.17974340685458342477_t_real
    ccc(1,1,3,177)= 0.51291024855906620586_t_real
    ccc(2,1,1,177)= 0.51291024855906620586_t_real
    ccc(2,1,3,177)= 0.17974340685458342477_t_real
    ccc(3,1,2,177)= 0.63971147342436279829_t_real
    !! L+1
    ccc(1,2,1,177)=-0.50854762771560778738_t_real
    ccc(1,2,3,177)= 0.21055872190307892367_t_real
    ccc(2,2,1,177)= 0.21055872190307892367_t_real
    ccc(2,2,3,177)= 0.50854762771560778738_t_real
    ccc(3,2,2,177)=-0.62776482057434447816_t_real
    !! M=-5
    !! L-1
    ccc(1,1,1,178)=-0.20754980866510833023_t_real
    ccc(1,1,3,178)= 0.48516452403758396048_t_real
    ccc(2,1,1,178)= 0.48516452403758396048_t_real
    ccc(2,1,3,178)= 0.20754980866510833023_t_real
    ccc(3,1,2,178)= 0.66564023547027494640_t_real
    !! L+1
    ccc(1,2,1,178)=-0.48372528131497451382_t_real
    ccc(1,2,3,178)= 0.23541180771537917931_t_real
    ccc(2,2,1,178)= 0.23541180771537917931_t_real
    ccc(2,2,3,178)= 0.48372528131497451382_t_real
    ccc(3,2,2,178)=-0.64898556687329512346_t_real
    !! M=-4
    !! L-1
    ccc(1,1,1,179)=-0.23533936216582083830_t_real
    ccc(1,1,3,179)= 0.45741749991749247760_t_real
    ccc(2,1,1,179)= 0.45741749991749247760_t_real
    ccc(2,1,3,179)= 0.23533936216582083830_t_real
    ccc(3,1,2,179)= 0.68612624987623871640_t_real
    !! L+1
    ccc(1,2,1,179)=-0.45890209522830260048_t_real
    ccc(1,2,3,179)= 0.26025754580373505710_t_real
    ccc(2,2,1,179)= 0.26025754580373505710_t_real
    ccc(2,2,3,179)= 0.45890209522830260048_t_real
    ccc(3,2,2,179)=-0.66584514242771290249_t_real
    !! M=-3
    !! L-1
    ccc(1,1,1,180)=-0.26311740579210876440_t_real
    ccc(1,1,3,180)= 0.42966892442365972828_t_real
    ccc(2,1,1,180)= 0.42966892442365972828_t_real
    ccc(2,1,3,180)= 0.26311740579210876440_t_real
    ccc(3,1,2,180)= 0.70164641544562337175_t_real
    !! L+1
    ccc(1,2,1,180)=-0.43407792540072461547_t_real
    ccc(1,2,3,180)= 0.28509785719527935707_t_real
    ccc(2,2,1,180)= 0.28509785719527935707_t_real
    ccc(2,2,3,180)= 0.43407792540072461547_t_real
    ccc(3,2,2,180)=-0.67866864743033923974_t_real
    !! M=-2
    !! L-1
    ccc(1,1,1,181)=-0.29088723694136980024_t_real
    ccc(1,1,3,181)= 0.40191847623425019455_t_real
    ccc(2,1,1,181)= 0.40191847623425019455_t_real
    ccc(2,1,3,181)= 0.29088723694136980024_t_real
    ccc(3,1,2,181)= 0.71252530319442529414_t_real
    !! L+1
    ccc(1,2,1,181)=-0.40925259281898757373_t_real
    ccc(1,2,3,181)= 0.30993404669460341628_t_real
    ccc(2,2,1,181)= 0.30993404669460341628_t_real
    ccc(2,2,3,181)= 0.40925259281898757373_t_real
    ccc(3,2,2,181)=-0.68768190607350336151_t_real
    !! M=-1
    !! L-1
    ccc(1,1,1,182)=-0.31865100272627660202_t_real
    ccc(2,1,1,182)= 0.52915026221291811807_t_real
    ccc(2,1,3,182)= 0.31865100272627660202_t_real
    ccc(3,1,2,182)= 0.71897362741833369909_t_real
    !! L+1
    ccc(1,2,1,182)=-0.38442587221924475148_t_real
    ccc(2,2,1,182)= 0.47343207647399933921_t_real
    ccc(2,2,3,182)= 0.38442587221924475148_t_real
    ccc(3,2,2,182)=-0.69303359695072724098_t_real
    !! M=0
    !! L-1
    ccc(1,1,3,183)=-0.48989794855663561961_t_real
    ccc(2,1,1,183)=-0.48989794855663561961_t_real
    ccc(3,1,2,183)= 0.72111025509279785860_t_real
    !! L+1
    ccc(1,2,3,183)=-0.50854762771560778737_t_real
    ccc(2,2,1,183)=-0.50854762771560778737_t_real
    ccc(3,2,2,183)=-0.69480833377965118676_t_real
    !! M=1
    !! L-1
    ccc(1,1,1,184)= 0.52915026221291811807_t_real
    ccc(1,1,3,184)=-0.31865100272627660202_t_real
    ccc(2,1,1,184)=-0.31865100272627660202_t_real
    ccc(3,1,2,184)= 0.71897362741833369909_t_real
    !! L+1
    ccc(1,2,1,184)= 0.47343207647399933921_t_real
    ccc(1,2,3,184)=-0.38442587221924475148_t_real
    ccc(2,2,1,184)=-0.38442587221924475148_t_real
    ccc(3,2,2,184)=-0.69303359695072724098_t_real
    !! M=2
    !! L-1
    ccc(1,1,1,185)= 0.40191847623425019455_t_real
    ccc(1,1,3,185)=-0.29088723694136980024_t_real
    ccc(2,1,1,185)=-0.29088723694136980024_t_real
    ccc(2,1,3,185)=-0.40191847623425019455_t_real
    ccc(3,1,2,185)= 0.71252530319442529414_t_real
    !! L+1
    ccc(1,2,1,185)= 0.30993404669460341628_t_real
    ccc(1,2,3,185)=-0.40925259281898757373_t_real
    ccc(2,2,1,185)=-0.40925259281898757373_t_real
    ccc(2,2,3,185)=-0.30993404669460341628_t_real
    ccc(3,2,2,185)=-0.68768190607350336151_t_real
    !! M=3
    !! L-1
    ccc(1,1,1,186)= 0.42966892442365972828_t_real
    ccc(1,1,3,186)=-0.26311740579210876440_t_real
    ccc(2,1,1,186)=-0.26311740579210876440_t_real
    ccc(2,1,3,186)=-0.42966892442365972828_t_real
    ccc(3,1,2,186)= 0.70164641544562337175_t_real
    !! L+1
    ccc(1,2,1,186)= 0.28509785719527935707_t_real
    ccc(1,2,3,186)=-0.43407792540072461547_t_real
    ccc(2,2,1,186)=-0.43407792540072461547_t_real
    ccc(2,2,3,186)=-0.28509785719527935707_t_real
    ccc(3,2,2,186)=-0.67866864743033923974_t_real
    !! M=4
    !! L-1
    ccc(1,1,1,187)= 0.45741749991749247760_t_real
    ccc(1,1,3,187)=-0.23533936216582083830_t_real
    ccc(2,1,1,187)=-0.23533936216582083830_t_real
    ccc(2,1,3,187)=-0.45741749991749247760_t_real
    ccc(3,1,2,187)= 0.68612624987623871640_t_real
    !! L+1
    ccc(1,2,1,187)= 0.26025754580373505710_t_real
    ccc(1,2,3,187)=-0.45890209522830260048_t_real
    ccc(2,2,1,187)=-0.45890209522830260048_t_real
    ccc(2,2,3,187)=-0.26025754580373505710_t_real
    ccc(3,2,2,187)=-0.66584514242771290249_t_real
    !! M=5
    !! L-1
    ccc(1,1,1,188)= 0.48516452403758396048_t_real
    ccc(1,1,3,188)=-0.20754980866510833023_t_real
    ccc(2,1,1,188)=-0.20754980866510833023_t_real
    ccc(2,1,3,188)=-0.48516452403758396048_t_real
    ccc(3,1,2,188)= 0.66564023547027494640_t_real
    !! L+1
    ccc(1,2,1,188)= 0.23541180771537917931_t_real
    ccc(1,2,3,188)=-0.48372528131497451382_t_real
    ccc(2,2,1,188)=-0.48372528131497451382_t_real
    ccc(2,2,3,188)=-0.23541180771537917931_t_real
    ccc(3,2,2,188)=-0.64898556687329512346_t_real
    !! M=6
    !! L-1
    ccc(1,1,1,189)= 0.51291024855906620586_t_real
    ccc(1,1,3,189)=-0.17974340685458342477_t_real
    ccc(2,1,1,189)=-0.17974340685458342477_t_real
    ccc(2,1,3,189)=-0.51291024855906620586_t_real
    ccc(3,1,2,189)= 0.63971147342436279829_t_real
    !! L+1
    ccc(1,2,1,189)= 0.21055872190307892367_t_real
    ccc(1,2,3,189)=-0.50854762771560778738_t_real
    ccc(2,2,1,189)=-0.50854762771560778738_t_real
    ccc(2,2,3,189)=-0.21055872190307892367_t_real
    ccc(3,2,2,189)=-0.62776482057434447816_t_real
    !! M=7
    !! L-1
    ccc(1,1,1,190)= 0.54065487356324857942_t_real
    ccc(1,1,3,190)=-0.15191090506254999243_t_real
    ccc(2,1,1,190)=-0.15191090506254999243_t_real
    ccc(2,1,3,190)=-0.54065487356324857942_t_real
    ccc(3,1,2,190)= 0.60764362025019996974_t_real
    !! L+1
    ccc(1,2,1,190)= 0.18569533817705186314_t_real
    ccc(1,2,3,190)=-0.53336925166406964067_t_real
    ccc(2,2,1,190)=-0.53336925166406964067_t_real
    ccc(2,2,3,190)=-0.18569533817705186314_t_real
    ccc(3,2,2,190)=-0.60172166781431544132_t_real
    !! M=8
    !! L-1
    ccc(1,1,1,191)= 0.56839856005880510754_t_real
    ccc(1,1,3,191)=-0.12403473458920845618_t_real
    ccc(2,1,1,191)=-0.12403473458920845618_t_real
    ccc(2,1,3,191)=-0.56839856005880510754_t_real
    ccc(3,1,2,191)= 0.56839856005880510754_t_real
    !! L+1
    ccc(1,2,1,191)= 0.16081688022566922492_t_real
    ccc(1,2,3,191)=-0.55819024953845526273_t_real
    ccc(2,2,1,191)=-0.55819024953845526273_t_real
    ccc(2,2,3,191)=-0.16081688022566922492_t_real
    ccc(3,2,2,191)=-0.57019571439055871414_t_real
    !! M=9
    !! L-1
    ccc(1,1,1,192)= 0.59614143907684809140_t_real
    ccc(1,1,3,192)=-0.09607689228305228008_t_real
    ccc(2,1,1,192)=-0.09607689228305228008_t_real
    ccc(2,1,3,192)=-0.59614143907684809140_t_real
    ccc(3,1,2,192)= 0.52035490847039270454_t_real
    !! L+1
    ccc(1,2,1,192)= 0.13591507055489058862_t_real
    ccc(1,2,3,192)=-0.58301070130188203540_t_real
    ccc(2,2,1,192)=-0.58301070130188203540_t_real
    ccc(2,2,3,192)=-0.13591507055489058862_t_real
    ccc(3,2,2,192)=-0.53221352061657884571_t_real
    !! M=10
    !! L-1
    ccc(1,1,1,193)= 0.62388361833820354826_t_real
    ccc(1,1,3,193)=-0.06793662204867574503_t_real
    ccc(2,1,1,193)=-0.06793662204867574503_t_real
    ccc(2,1,3,193)=-0.62388361833820354826_t_real
    ccc(3,1,2,193)= 0.46076858867298268520_t_real
    !! L+1
    ccc(1,2,1,193)= 0.11097419040461881708_t_real
    ccc(1,2,3,193)=-0.60783067385483083563_t_real
    ccc(2,2,1,193)=-0.60783067385483083563_t_real
    ccc(2,2,3,193)=-0.11097419040461881708_t_real
    ccc(3,2,2,193)=-0.48626453908386466850_t_real
    !! M=11
    !! L-1
    ccc(1,1,1,194)= 0.65162518721684217549_t_real
    ccc(1,1,3,194)=-0.03922322702763680638_t_real
    ccc(2,1,1,194)=-0.03922322702763680638_t_real
    ccc(2,1,3,194)=-0.65162518721684217549_t_real
    ccc(3,1,2,194)= 0.38430756913220912034_t_real
    !! L+1
    ccc(1,2,1,194)= 0.08596023825918792018_t_real
    ccc(1,2,3,194)=-0.63265022359809469351_t_real
    ccc(2,2,1,194)=-0.63265022359809469351_t_real
    ccc(2,2,3,194)=-0.08596023825918792018_t_real
    ccc(3,2,2,194)=-0.42980119129593960094_t_real
    !! M=12
    !! L-1
    ccc(1,1,1,195)= 0.67936622048675745034_t_real
    ccc(2,1,3,195)=-0.67936622048675745034_t_real
    ccc(3,1,2,195)= 0.27735009811261456100_t_real
    !! L+1
    ccc(1,2,1,195)= 0.06078306738548308356_t_real
    ccc(1,2,3,195)=-0.65746939841502642270_t_real
    ccc(2,2,1,195)=-0.65746939841502642270_t_real
    ccc(2,2,3,195)=-0.06078306738548308356_t_real
    ccc(3,2,2,195)=-0.35788101058031864758_t_real
    !! M=13
    !! L-1
    ccc(1,1,1,196)= 0.70710678118654752437_t_real
    ccc(2,1,3,196)=-0.70710678118654752437_t_real
    !! L+1
    ccc(1,2,1,196)= 0.03509312031717982061_t_real
    ccc(1,2,3,196)=-0.68228823922101306055_t_real
    ccc(2,2,1,196)=-0.68228823922101306055_t_real
    ccc(2,2,3,196)=-0.03509312031717982061_t_real
    ccc(3,2,2,196)=-0.25788071477756376056_t_real
    !!
    If(maxL<=13) Return
    !!
    !!
    Return
    !!
  End Subroutine calc_real_vec_ccoefs
#endif

  !!***********************************************************************
  !! calculate ylms
  !! real vec harmonics rvylms
  !! and  real vec vec harmonics qvylms
  !!
  !! ylm=Y_{lm}
  !! rvylm=R^{l+lambda}_{lm}
  !! qvylm=expansion functions for dyadic grad \circ grad 
  !!
#ifdef my_compile_unused
  Subroutine calc_real_quad_harmonics(r,ccc,maxL,clpm,ylm,rvylm,qvylm)
    !!
    Implicit None
    !!IO
    Integer(t_int),Intent(in)    :: maxL
    Real(t_real),Intent(in)      :: r(3)
    Real(t_real),Intent(in)      :: clpm(2,0:(maxL+1))
    Real(t_real),Intent(in)      :: ccc(3,2,1:3,(((maxL+1)+1)**2))
    Real(t_real),Intent(out)     :: ylm(((maxL+2)+1)**2)
    Real(t_real),Intent(out)     :: rvylm(3,2,(((maxL+1)+1)**2))
    Real(t_real),Intent(out)     :: qvylm(3,3,3,(((maxL)+1)**2))
    !!LOCAL
    !Character(*),Parameter::nsubfunc="calc_real_quad_harmonics"
    Integer(t_int) :: l,m,ilam0,ilam,ilm,lpm,i,m1,jlm,ki
    Real(t_real)   :: rd(3,3)
    !!
    Call ylms14(ylm,r(1),r(2),r(3),maxL+2)
    !!
    rvylm(:,:,:)=0.0_t_real
    qvylm(:,:,:,:)=0.0_t_real
    !!
    Do l=14,maxL+1
       Do m=-l,l
          ilm=l*l+l+1+m
    
          ilam0=1
          If(l==0) ilam0=2
          Do ilam=ilam0,2
             !!
             !!
             lpm=l+2*ilam-3
             !!
             jlm=lpm*lpm+lpm+1+m-1
             !!
             !!If( Abs(m-1) <=lpm ) Then
             !! see below
             If( jlm > 0 ) Then
                rvylm(1,ilam,ilm)=rvylm(1,ilam,ilm)&
                     &+ccc(1,ilam,1,ilm)*ylm(jlm)
             Endif


             jlm=jlm+2
             !!
             !! If( Abs(m+1) <=lpm ) Then
             !! see below
             rvylm(1,ilam,ilm)=rvylm(1,ilam,ilm)&
                  &+ccc(1,ilam,3,ilm)*ylm(jlm)
             
             
             jlm=jlm-2-2*m

             
             !!
             !!If( Abs(-m-1) <=lpm ) Then
             !! see below
             If( jlm > 0 ) Then
                rvylm(2,ilam,ilm)=rvylm(2,ilam,ilm)&
                     &+ccc(2,ilam,1,ilm)*ylm(jlm)
             Endif
             
             jlm=jlm+2
             !!
             !! If( Abs(-m+1) <=lpm ) Then
             !! see below
             rvylm(2,ilam,ilm)=rvylm(2,ilam,ilm)&
                  &+ccc(2,ilam,3,ilm)*ylm(jlm)
             
             
             !! the rigorous test would be
             !! jlm=lpm*lpm+lpm+1+m
             !! If( Abs(m) <=lpm ) Then
             !!   ...
             !! Endif
             !! however ccc is well defined and only memory access
             !! of ylm beyond its limits must be avoided
             !! for this it is enough to require 1<= jlm <= (((maxL+2)+1)**2)
             !!
             !! jlm=lpm*lpm+lpm+1+m < (maxL+2)(maxL+4) < (maxL+3)**2
             !! thus the upper bound is never touched
             !! the case lpm=l+1 gives no probelms for the lower bound
             !! for lpm=l-1
             !! m=-l
             !! jlm=lpm*lpm+lpm+1+m > (l-1)**2 >= 0
             !! we get problems for l=1, m=-1
             !! m=-l+n, n>0
             !! jlm=lpm*lpm+lpm+1+m = (l-1)**2+n > 1 
             !! no further problemes
             !!
             jlm=lpm*lpm+lpm+1+m
             !If( jlm > 0 ) Then
             If( .Not. (l==1 .And. m==-1 .And. ilam ==1) ) Then
                rvylm(3,ilam,ilm)=rvylm(3,ilam,ilm)&
                     &+ccc(3,ilam,2,ilm)*ylm(jlm)
             Endif
             
          Enddo
       Enddo ! m
    Enddo ! l
    !!
    !!..............................................
    !! L=0
    !!..............................................
    !! L+1
    rvylm(1,2,1)=-0.57735026918962576448_t_real*ylm(4)
    rvylm(2,2,1)=-0.57735026918962576448_t_real*ylm(2)
    rvylm(3,2,1)=-0.57735026918962576450_t_real*ylm(3)
    !!
    If(maxL+1<=0) Goto 1000
    !!
    !!..............................................
    !! L=1
    !!..............................................
    !! L-1
    rvylm(2,1,2)= ylm(1)
    rvylm(3,1,3)= ylm(1)
    rvylm(1,1,4)= ylm(1)
    !! L+1
    rvylm(1,2,2)=-0.54772255750516611345_t_real*ylm(5)
    rvylm(2,2,2)= 0.31622776601683793319_t_real*ylm(7)&
      &+.54772255750516611345_t_real*ylm(9)
    rvylm(3,2,2)=-0.54772255750516611345_t_real*ylm(6)
    rvylm(1,2,3)=-0.54772255750516611344_t_real*ylm(8)
    rvylm(2,2,3)=-0.54772255750516611344_t_real*ylm(6)
    rvylm(3,2,3)=-0.63245553203367586640_t_real*ylm(7)
    rvylm(1,2,4)= 0.31622776601683793319_t_real*ylm(7)&
       &-0.54772255750516611345_t_real*ylm(9)
    rvylm(2,2,4)=-0.54772255750516611345_t_real*ylm(5)
    rvylm(3,2,4)=-0.54772255750516611345_t_real*ylm(8)
    !!
    If(maxL+1<=1) Goto 1000
    !!
    !!..............................................
    !! L=2
    !!..............................................
    !! L-1
    rvylm(1,1,5)= 0.70710678118654752438_t_real*ylm(2)
    rvylm(2,1,5)= 0.70710678118654752438_t_real*ylm(4)
    rvylm(2,1,6)= 0.70710678118654752438_t_real*ylm(3)
    rvylm(3,1,6)= 0.70710678118654752438_t_real*ylm(2)
    rvylm(1,1,7)=-0.40824829046386301634_t_real*ylm(4)
    rvylm(2,1,7)=-0.40824829046386301634_t_real*ylm(2)
    rvylm(3,1,7)= 0.81649658092772603272_t_real*ylm(3)
    rvylm(1,1,8)= 0.70710678118654752438_t_real*ylm(3)
    rvylm(3,1,8)= 0.70710678118654752438_t_real*ylm(4)
    rvylm(1,1,9)= 0.70710678118654752438_t_real*ylm(4)
    rvylm(2,1,9)=-0.70710678118654752438_t_real*ylm(2)
    !! L+1
    rvylm(1,2,5)=-0.59761430466719681996_t_real*ylm(10)&
       &+.15430334996209191025_t_real*ylm(12)
    rvylm(2,2,5)= 0.15430334996209191025_t_real*ylm(14)&
      &+.59761430466719681996_t_real*ylm(16)
    rvylm(3,2,5)=-0.48795003647426658966_t_real*ylm(11)
    rvylm(1,2,6)=-0.48795003647426658966_t_real*ylm(11)
    rvylm(2,2,6)= 0.37796447300922722719_t_real*ylm(13)&
      &+.48795003647426658966_t_real*ylm(15)
    rvylm(3,2,6)=-0.61721339984836764102_t_real*ylm(12)
    rvylm(1,2,7)=-0.53452248382484876934_t_real*ylm(14)
    rvylm(2,2,7)=-0.53452248382484876934_t_real*ylm(12)
    rvylm(3,2,7)=-0.65465367070797714378_t_real*ylm(13)
    rvylm(1,2,8)= 0.37796447300922722719_t_real*ylm(13)&
       &-0.48795003647426658966_t_real*ylm(15)
    rvylm(2,2,8)=-0.48795003647426658966_t_real*ylm(11)
    rvylm(3,2,8)=-0.61721339984836764102_t_real*ylm(14)
    rvylm(1,2,9)= 0.15430334996209191025_t_real*ylm(14)&
       &-0.59761430466719681996_t_real*ylm(16)
    rvylm(2,2,9)=-0.59761430466719681996_t_real*ylm(10)&
      &-0.15430334996209191025_t_real*ylm(12)
    rvylm(3,2,9)=-0.48795003647426658966_t_real*ylm(15)
    !!
    If(maxL+1<=2) Goto 1000
    !!
    !!..............................................
    !! L=3
    !!..............................................
    !! L-1
    rvylm(1,1,10)= 0.70710678118654752434_t_real*ylm(5)
    rvylm(2,1,10)= 0.70710678118654752434_t_real*ylm(9)
    rvylm(1,1,11)= 0.57735026918962576446_t_real*ylm(6)
    rvylm(2,1,11)= 0.57735026918962576446_t_real*ylm(8)
    rvylm(3,1,11)= 0.57735026918962576446_t_real*ylm(5)
    rvylm(1,1,12)=-0.18257418583505537113_t_real*ylm(5)
    rvylm(2,1,12)= 0.63245553203367586633_t_real*ylm(7)&
      &+.18257418583505537113_t_real*ylm(9)
    rvylm(3,1,12)= 0.73029674334022148455_t_real*ylm(6)
    rvylm(1,1,13)=-0.44721359549995793923_t_real*ylm(8)
    rvylm(2,1,13)=-0.44721359549995793923_t_real*ylm(6)
    rvylm(3,1,13)= 0.77459666924148337698_t_real*ylm(7)
    rvylm(1,1,14)= 0.63245553203367586633_t_real*ylm(7)&
       &-0.18257418583505537113_t_real*ylm(9)
    rvylm(2,1,14)=-0.18257418583505537113_t_real*ylm(5)
    rvylm(3,1,14)= 0.73029674334022148455_t_real*ylm(8)
    rvylm(1,1,15)= 0.57735026918962576446_t_real*ylm(8)
    rvylm(2,1,15)=-0.57735026918962576446_t_real*ylm(6)
    rvylm(3,1,15)= 0.57735026918962576446_t_real*ylm(9)
    rvylm(1,1,16)= 0.70710678118654752434_t_real*ylm(9)
    rvylm(2,1,16)=-0.70710678118654752434_t_real*ylm(5)
    !! L+1
    rvylm(1,2,10)=-0.62360956446232356423_t_real*ylm(17)&
       &+.11785113019775792072_t_real*ylm(19)
    rvylm(2,2,10)= 0.11785113019775792072_t_real*ylm(23)&
      &+.62360956446232356423_t_real*ylm(25)
    rvylm(3,2,10)=-0.44095855184409843173_t_real*ylm(18)
    rvylm(1,2,11)=-0.54006172486732168589_t_real*ylm(18)&
       &+.20412414523193150817_t_real*ylm(20)
    rvylm(2,2,11)= 0.20412414523193150817_t_real*ylm(22)&
      &+.54006172486732168589_t_real*ylm(24)
    rvylm(3,2,11)=-0.57735026918962576448_t_real*ylm(19)
    rvylm(1,2,12)=-0.45643546458763842786_t_real*ylm(19)
    rvylm(2,2,12)= 0.40824829046386301634_t_real*ylm(21)&
      &+.45643546458763842786_t_real*ylm(23)
    rvylm(3,2,12)=-0.64549722436790281417_t_real*ylm(20)
    rvylm(1,2,13)=-0.52704627669472988863_t_real*ylm(22)
    rvylm(2,2,13)=-0.52704627669472988863_t_real*ylm(20)
    rvylm(3,2,13)=-0.66666666666666666664_t_real*ylm(21)
    rvylm(1,2,14)= 0.40824829046386301634_t_real*ylm(21)&
       &-0.45643546458763842786_t_real*ylm(23)
    rvylm(2,2,14)=-0.45643546458763842786_t_real*ylm(19)
    rvylm(3,2,14)=-0.64549722436790281417_t_real*ylm(22)
    rvylm(1,2,15)= 0.20412414523193150817_t_real*ylm(22)&
       &-0.54006172486732168589_t_real*ylm(24)
    rvylm(2,2,15)=-0.54006172486732168589_t_real*ylm(18)&
      &-0.20412414523193150817_t_real*ylm(20)
    rvylm(3,2,15)=-0.57735026918962576448_t_real*ylm(23)
    rvylm(1,2,16)= 0.11785113019775792072_t_real*ylm(23)&
       &-0.62360956446232356423_t_real*ylm(25)
    rvylm(2,2,16)=-0.62360956446232356423_t_real*ylm(17)&
      &-0.11785113019775792072_t_real*ylm(19)
    rvylm(3,2,16)=-0.44095855184409843173_t_real*ylm(24)
    !!
    If(maxL+1<=3) Goto 1000
    !!..............................................
    !! L=4
    !!..............................................
    !! L-1
    rvylm(1,1,17)= 0.70710678118654752437_t_real*ylm(10)
    rvylm(2,1,17)= 0.70710678118654752437_t_real*ylm(16)
    rvylm(1,1,18)= 0.61237243569579452452_t_real*ylm(11)
    rvylm(2,1,18)= 0.61237243569579452452_t_real*ylm(15)
    rvylm(3,1,18)= 0.49999999999999999998_t_real*ylm(10)
    rvylm(1,1,19)=-0.13363062095621219233_t_real*ylm(10)&
       &+.51754916950676566332_t_real*ylm(12)
    rvylm(2,1,19)= 0.51754916950676566332_t_real*ylm(14)&
      &+.13363062095621219233_t_real*ylm(16)
    rvylm(3,1,19)= 0.65465367070797714377_t_real*ylm(11)
    rvylm(1,1,20)=-0.23145502494313786538_t_real*ylm(11)
    rvylm(2,1,20)= 0.59761430466719681994_t_real*ylm(13)&
      &+.23145502494313786538_t_real*ylm(15)
    rvylm(3,1,20)= 0.73192505471139988448_t_real*ylm(12)
    rvylm(1,1,21)=-0.46291004988627573075_t_real*ylm(14)
    rvylm(2,1,21)=-0.46291004988627573075_t_real*ylm(12)
    rvylm(3,1,21)= 0.75592894601845445440_t_real*ylm(13)
    rvylm(1,1,22)= 0.59761430466719681994_t_real*ylm(13)&
       &-0.23145502494313786538_t_real*ylm(15)
    rvylm(2,1,22)=-0.23145502494313786538_t_real*ylm(11)
    rvylm(3,1,22)= 0.73192505471139988448_t_real*ylm(14)
    rvylm(1,1,23)= 0.51754916950676566332_t_real*ylm(14)&
       &-0.13363062095621219233_t_real*ylm(16)
    rvylm(2,1,23)=-0.13363062095621219233_t_real*ylm(10)&
      &-0.51754916950676566332_t_real*ylm(12)
    rvylm(3,1,23)= 0.65465367070797714377_t_real*ylm(15)
    rvylm(1,1,24)= 0.61237243569579452452_t_real*ylm(15)
    rvylm(2,1,24)=-0.61237243569579452452_t_real*ylm(11)
    rvylm(3,1,24)= 0.49999999999999999998_t_real*ylm(16)
    rvylm(1,1,25)= 0.70710678118654752437_t_real*ylm(16)
    rvylm(2,1,25)=-0.70710678118654752437_t_real*ylm(10)
    !! L+1
    rvylm(1,2,17)=-0.63960214906683130285_t_real*ylm(26)&
       &+.09534625892455923153_t_real*ylm(28)
    rvylm(2,2,17)= 0.09534625892455923153_t_real*ylm(34)&
      &+.63960214906683130285_t_real*ylm(36)
    rvylm(3,2,17)=-0.40451991747794525172_t_real*ylm(27)
    rvylm(1,2,18)=-0.57207755354735538922_t_real*ylm(27)&
       &+.16514456476895409082_t_real*ylm(29)
    rvylm(2,2,18)= 0.16514456476895409082_t_real*ylm(33)&
      &+.57207755354735538922_t_real*ylm(35)
    rvylm(3,2,18)=-0.53935988997059366896_t_real*ylm(28)
    rvylm(1,2,19)=-0.50452497910951301620_t_real*ylm(28)&
       &+.23354968324845689125_t_real*ylm(30)
    rvylm(2,2,19)= 0.23354968324845689125_t_real*ylm(32)&
      &+.50452497910951301620_t_real*ylm(34)
    rvylm(3,2,19)=-0.61791438065332467072_t_real*ylm(29)
    rvylm(1,2,20)=-0.43693144875265148286_t_real*ylm(29)
    rvylm(2,2,20)= 0.42640143271122086855_t_real*ylm(31)&
      &+.43693144875265148286_t_real*ylm(33)
    rvylm(3,2,20)=-0.66057825907581636331_t_real*ylm(30)
    rvylm(1,2,21)=-0.52223296786709351448_t_real*ylm(32)
    rvylm(2,2,21)=-0.52223296786709351448_t_real*ylm(30)
    rvylm(3,2,21)=-0.67419986246324208620_t_real*ylm(31)
    rvylm(1,2,22)= 0.42640143271122086855_t_real*ylm(31)&
       &-0.43693144875265148286_t_real*ylm(33)
    rvylm(2,2,22)=-0.43693144875265148286_t_real*ylm(29)
    rvylm(3,2,22)=-0.66057825907581636331_t_real*ylm(32)
    rvylm(1,2,23)= 0.23354968324845689125_t_real*ylm(32)&
       &-0.50452497910951301620_t_real*ylm(34)
    rvylm(2,2,23)=-0.50452497910951301620_t_real*ylm(28)&
      &-0.23354968324845689125_t_real*ylm(30)
    rvylm(3,2,23)=-0.61791438065332467072_t_real*ylm(33)
    rvylm(1,2,24)= 0.16514456476895409082_t_real*ylm(33)&
       &-0.57207755354735538922_t_real*ylm(35)
    rvylm(2,2,24)=-0.57207755354735538922_t_real*ylm(27)&
      &-0.16514456476895409082_t_real*ylm(29)
    rvylm(3,2,24)=-0.53935988997059366896_t_real*ylm(34)
    rvylm(1,2,25)= 0.09534625892455923153_t_real*ylm(34)&
       &-0.63960214906683130285_t_real*ylm(36)
    rvylm(2,2,25)=-0.63960214906683130285_t_real*ylm(26)&
      &-0.09534625892455923153_t_real*ylm(28)
    rvylm(3,2,25)=-0.40451991747794525172_t_real*ylm(35)
    !!
    If(maxL+1<=4) goto 1000
    !!
    !!..............................................
    !! L=5
    !!..............................................
    !! L-1
    rvylm(1,1,26)= 0.70710678118654752439_t_real*ylm(17)
    rvylm(2,1,26)= 0.70710678118654752439_t_real*ylm(25)
    rvylm(1,1,27)= 0.63245553203367586639_t_real*ylm(18)
    rvylm(2,1,27)= 0.63245553203367586639_t_real*ylm(24)
    rvylm(3,1,27)= 0.44721359549995793928_t_real*ylm(17)
    rvylm(1,1,28)=-0.10540925533894597773_t_real*ylm(17)&
       &+.55777335102271703198_t_real*ylm(19)
    rvylm(2,1,28)= 0.55777335102271703198_t_real*ylm(23)&
      &+.10540925533894597773_t_real*ylm(25)
    rvylm(3,1,28)= 0.59628479399994391904_t_real*ylm(18)
    rvylm(1,1,29)=-0.18257418583505537115_t_real*ylm(18)&
       &+.48304589153964795245_t_real*ylm(20)
    rvylm(2,1,29)= 0.48304589153964795245_t_real*ylm(22)&
      &+.18257418583505537115_t_real*ylm(24)
    rvylm(3,1,29)= 0.68313005106397322554_t_real*ylm(19)
    rvylm(1,1,30)=-0.25819888974716112567_t_real*ylm(19)
    rvylm(2,1,30)= 0.57735026918962576449_t_real*ylm(21)&
      &+.25819888974716112567_t_real*ylm(23)
    rvylm(3,1,30)= 0.73029674334022148460_t_real*ylm(20)
    rvylm(1,1,31)=-0.47140452079103168292_t_real*ylm(22)
    rvylm(2,1,31)=-0.47140452079103168292_t_real*ylm(20)
    rvylm(3,1,31)= 0.74535599249992989880_t_real*ylm(21)
    rvylm(1,1,32)= 0.57735026918962576449_t_real*ylm(21)&
       &-0.25819888974716112567_t_real*ylm(23)
    rvylm(2,1,32)=-0.25819888974716112567_t_real*ylm(19)
    rvylm(3,1,32)= 0.73029674334022148460_t_real*ylm(22)
    rvylm(1,1,33)= 0.48304589153964795245_t_real*ylm(22)&
       &-0.18257418583505537115_t_real*ylm(24)
    rvylm(2,1,33)=-0.18257418583505537115_t_real*ylm(18)&
      &-0.48304589153964795245_t_real*ylm(20)
    rvylm(3,1,33)= 0.68313005106397322554_t_real*ylm(23)
    rvylm(1,1,34)= 0.55777335102271703198_t_real*ylm(23)&
       &-0.10540925533894597773_t_real*ylm(25)
    rvylm(2,1,34)=-0.10540925533894597773_t_real*ylm(17)&
      &-0.55777335102271703198_t_real*ylm(19)
    rvylm(3,1,34)= 0.59628479399994391904_t_real*ylm(24)
    rvylm(1,1,35)= 0.63245553203367586639_t_real*ylm(24)
    rvylm(2,1,35)=-0.63245553203367586639_t_real*ylm(18)
    rvylm(3,1,35)= 0.44721359549995793928_t_real*ylm(25)
    rvylm(1,1,36)= 0.70710678118654752439_t_real*ylm(25)
    rvylm(2,1,36)=-0.70710678118654752439_t_real*ylm(17)
    !! L+1
    rvylm(1,2,26)=-0.65044363558799088069_t_real*ylm(37)&
       &+.08006407690254356674_t_real*ylm(39)
    rvylm(2,2,26)= 0.08006407690254356674_t_real*ylm(47)&
      &+.65044363558799088069_t_real*ylm(49)
    rvylm(3,2,26)=-0.37553380809940537980_t_real*ylm(38)
    rvylm(1,2,27)=-0.59377108599535441179_t_real*ylm(38)&
       &+.13867504905630728050_t_real*ylm(40)
    rvylm(2,2,27)= 0.13867504905630728050_t_real*ylm(46)&
      &+.59377108599535441179_t_real*ylm(48)
    rvylm(3,2,27)=-0.50636968354183330813_t_real*ylm(39)
    rvylm(1,2,28)=-0.53708615552957466037_t_real*ylm(39)&
       &+.19611613513818403192_t_real*ylm(41)
    rvylm(2,2,28)= 0.19611613513818403192_t_real*ylm(45)&
      &+.53708615552957466037_t_real*ylm(47)
    rvylm(3,2,28)=-0.58834840541455209576_t_real*ylm(40)
    rvylm(1,2,29)=-0.48038446141526140044_t_real*ylm(40)&
       &+.25318484177091665406_t_real*ylm(42)
    rvylm(2,2,29)= 0.25318484177091665406_t_real*ylm(44)&
      &+.48038446141526140044_t_real*ylm(46)
    rvylm(3,2,29)=-0.64051261522034853392_t_real*ylm(41)
    rvylm(1,2,30)=-0.42365927286816169388_t_real*ylm(41)
    rvylm(2,2,30)= 0.43852900965351460735_t_real*ylm(43)&
      &+.42365927286816169388_t_real*ylm(45)
    rvylm(3,2,30)=-0.66986412705708365760_t_real*ylm(42)
    rvylm(1,2,31)=-0.51887452166277082558_t_real*ylm(44)
    rvylm(2,2,31)=-0.51887452166277082558_t_real*ylm(42)
    rvylm(3,2,31)=-0.67936622048675745036_t_real*ylm(43)
    rvylm(1,2,32)= 0.43852900965351460735_t_real*ylm(43)&
       &-0.42365927286816169388_t_real*ylm(45)
    rvylm(2,2,32)=-0.42365927286816169388_t_real*ylm(41)
    rvylm(3,2,32)=-0.66986412705708365760_t_real*ylm(44)
    rvylm(1,2,33)= 0.25318484177091665406_t_real*ylm(44)&
       &-0.48038446141526140044_t_real*ylm(46)
    rvylm(2,2,33)=-0.48038446141526140044_t_real*ylm(40)&
      &-0.25318484177091665406_t_real*ylm(42)
    rvylm(3,2,33)=-0.64051261522034853392_t_real*ylm(45)
    rvylm(1,2,34)= 0.19611613513818403192_t_real*ylm(45)&
       &-0.53708615552957466037_t_real*ylm(47)
    rvylm(2,2,34)=-0.53708615552957466037_t_real*ylm(39)&
      &-0.19611613513818403192_t_real*ylm(41)
    rvylm(3,2,34)=-0.58834840541455209576_t_real*ylm(46)
    rvylm(1,2,35)= 0.13867504905630728050_t_real*ylm(46)&
       &-0.59377108599535441179_t_real*ylm(48)
    rvylm(2,2,35)=-0.59377108599535441179_t_real*ylm(38)&
      &-0.13867504905630728050_t_real*ylm(40)
    rvylm(3,2,35)=-0.50636968354183330813_t_real*ylm(47)
    rvylm(1,2,36)= 0.08006407690254356674_t_real*ylm(47)&
       &-0.65044363558799088069_t_real*ylm(49)
    rvylm(2,2,36)=-0.65044363558799088069_t_real*ylm(37)&
      &-0.08006407690254356674_t_real*ylm(39)
    rvylm(3,2,36)=-0.37553380809940537980_t_real*ylm(48)
    !!
    If(maxL+1<=5) goto 1000
    !!
    !!..............................................
    !! L=6
    !!..............................................
    !! L-1
    rvylm(1,1,37)= 0.70710678118654752431_t_real*ylm(26)
    rvylm(2,1,37)= 0.70710678118654752431_t_real*ylm(36)
    rvylm(1,1,38)= 0.64549722436790281412_t_real*ylm(27)
    rvylm(2,1,38)= 0.64549722436790281412_t_real*ylm(35)
    rvylm(3,1,38)= 0.40824829046386301631_t_real*ylm(26)
    rvylm(1,1,39)=-0.08703882797784891907_t_real*ylm(26)&
       &+.58387420812114222811_t_real*ylm(28)
    rvylm(2,1,39)= 0.58387420812114222811_t_real*ylm(34)&
      &+.08703882797784891907_t_real*ylm(36)
    rvylm(3,1,39)= 0.55048188256318030273_t_real*ylm(27)
    rvylm(1,1,40)=-0.15075567228888181130_t_real*ylm(27)&
       &+.52223296786709351447_t_real*ylm(29)
    rvylm(2,1,40)= 0.52223296786709351447_t_real*ylm(33)&
      &+.15075567228888181130_t_real*ylm(35)
    rvylm(3,1,40)= 0.63960214906683130282_t_real*ylm(28)
    rvylm(1,1,41)=-0.21320071635561043427_t_real*ylm(28)&
       &+.46056618647183829181_t_real*ylm(30)
    rvylm(2,1,41)= 0.46056618647183829181_t_real*ylm(32)&
      &+.21320071635561043427_t_real*ylm(34)
    rvylm(3,1,41)= 0.69631062382279135262_t_real*ylm(29)
    rvylm(1,1,42)=-0.27524094128159015136_t_real*ylm(29)
    rvylm(2,1,42)= 0.56407607481776620884_t_real*ylm(31)&
      &+.27524094128159015136_t_real*ylm(33)
    rvylm(3,1,42)= 0.72821908125441913807_t_real*ylm(30)
    rvylm(1,1,43)=-0.47673129462279615766_t_real*ylm(32)
    rvylm(2,1,43)=-0.47673129462279615766_t_real*ylm(30)
    rvylm(3,1,43)= 0.73854894587599639632_t_real*ylm(31)
    rvylm(1,1,44)= 0.56407607481776620884_t_real*ylm(31)&
       &-0.27524094128159015136_t_real*ylm(33)
    rvylm(2,1,44)=-0.27524094128159015136_t_real*ylm(29)
    rvylm(3,1,44)= 0.72821908125441913807_t_real*ylm(32)
    rvylm(1,1,45)= 0.46056618647183829181_t_real*ylm(32)&
       &-0.21320071635561043427_t_real*ylm(34)
    rvylm(2,1,45)=-0.21320071635561043427_t_real*ylm(28)&
      &-0.46056618647183829181_t_real*ylm(30)
    rvylm(3,1,45)= 0.69631062382279135262_t_real*ylm(33)
    rvylm(1,1,46)= 0.52223296786709351447_t_real*ylm(33)&
       &-0.15075567228888181130_t_real*ylm(35)
    rvylm(2,1,46)=-0.15075567228888181130_t_real*ylm(27)&
      &-0.52223296786709351447_t_real*ylm(29)
    rvylm(3,1,46)= 0.63960214906683130282_t_real*ylm(34)
    rvylm(1,1,47)= 0.58387420812114222811_t_real*ylm(34)&
       &-0.08703882797784891907_t_real*ylm(36)
    rvylm(2,1,47)=-0.08703882797784891907_t_real*ylm(26)&
      &-0.58387420812114222811_t_real*ylm(28)
    rvylm(3,1,47)= 0.55048188256318030273_t_real*ylm(35)
    rvylm(1,1,48)= 0.64549722436790281412_t_real*ylm(35)
    rvylm(2,1,48)=-0.64549722436790281412_t_real*ylm(27)
    rvylm(3,1,48)= 0.40824829046386301631_t_real*ylm(36)
    rvylm(1,1,49)= 0.70710678118654752431_t_real*ylm(36)
    rvylm(2,1,49)=-0.70710678118654752431_t_real*ylm(26)
    !! L+1
    rvylm(1,2,37)=-0.65828058860438330047_t_real*ylm(50)&
       &+.06900655593423542176_t_real*ylm(52)
    rvylm(2,2,37)= 0.06900655593423542176_t_real*ylm(62)&
      &+.65828058860438330047_t_real*ylm(64)
    rvylm(3,2,37)=-0.35186577527449839986_t_real*ylm(51)
    rvylm(1,2,38)=-0.60944940022004404992_t_real*ylm(51)&
       &+.11952286093343936397_t_real*ylm(53)
    rvylm(2,2,38)= 0.11952286093343936397_t_real*ylm(61)&
      &+.60944940022004404992_t_real*ylm(63)
    rvylm(3,2,38)=-0.47809144373375745591_t_real*ylm(52)
    rvylm(1,2,39)=-0.56061191058138809901_t_real*ylm(52)&
       &+.16903085094570331547_t_real*ylm(54)
    rvylm(2,2,39)= 0.16903085094570331547_t_real*ylm(60)&
      &+.56061191058138809901_t_real*ylm(62)
    rvylm(3,2,39)=-0.56061191058138809901_t_real*ylm(53)
    rvylm(1,2,40)=-0.51176631571915897704_t_real*ylm(53)&
       &+.21821789023599238123_t_real*ylm(55)
    rvylm(2,2,40)= 0.21821789023599238123_t_real*ylm(59)&
      &+.51176631571915897704_t_real*ylm(61)
    rvylm(3,2,40)=-0.61721339984836764094_t_real*ylm(54)
    rvylm(1,2,41)=-0.46291004988627573070_t_real*ylm(54)&
       &+.26726124191242438464_t_real*ylm(56)
    rvylm(2,2,41)= 0.26726124191242438464_t_real*ylm(58)&
      &+.46291004988627573070_t_real*ylm(60)
    rvylm(3,2,41)=-0.65465367070797714369_t_real*ylm(55)
    rvylm(1,2,42)=-0.41403933560541253061_t_real*ylm(55)
    rvylm(2,2,42)= 0.44721359549995793919_t_real*ylm(57)&
      &+.41403933560541253061_t_real*ylm(59)
    rvylm(3,2,42)=-0.67612340378281326190_t_real*ylm(56)
    rvylm(1,2,43)=-0.51639777949432225126_t_real*ylm(58)
    rvylm(2,2,43)=-0.51639777949432225126_t_real*ylm(56)
    rvylm(3,2,43)=-0.68313005106397322544_t_real*ylm(57)
    rvylm(1,2,44)= 0.44721359549995793919_t_real*ylm(57)&
       &-0.41403933560541253061_t_real*ylm(59)
    rvylm(2,2,44)=-0.41403933560541253061_t_real*ylm(55)
    rvylm(3,2,44)=-0.67612340378281326190_t_real*ylm(58)
    rvylm(1,2,45)= 0.26726124191242438464_t_real*ylm(58)&
       &-0.46291004988627573070_t_real*ylm(60)
    rvylm(2,2,45)=-0.46291004988627573070_t_real*ylm(54)&
      &-0.26726124191242438464_t_real*ylm(56)
    rvylm(3,2,45)=-0.65465367070797714369_t_real*ylm(59)
    rvylm(1,2,46)= 0.21821789023599238123_t_real*ylm(59)&
       &-0.51176631571915897704_t_real*ylm(61)
    rvylm(2,2,46)=-0.51176631571915897704_t_real*ylm(53)&
      &-0.21821789023599238123_t_real*ylm(55)
    rvylm(3,2,46)=-0.61721339984836764094_t_real*ylm(60)
    rvylm(1,2,47)= 0.16903085094570331547_t_real*ylm(60)&
       &-0.56061191058138809901_t_real*ylm(62)
    rvylm(2,2,47)=-0.56061191058138809901_t_real*ylm(52)&
      &-0.16903085094570331547_t_real*ylm(54)
    rvylm(3,2,47)=-0.56061191058138809901_t_real*ylm(61)
    rvylm(1,2,48)= 0.11952286093343936397_t_real*ylm(61)&
       &-0.60944940022004404992_t_real*ylm(63)
    rvylm(2,2,48)=-0.60944940022004404992_t_real*ylm(51)&
      &-0.11952286093343936397_t_real*ylm(53)
    rvylm(3,2,48)=-0.47809144373375745591_t_real*ylm(62)
    rvylm(1,2,49)= 0.06900655593423542176_t_real*ylm(62)&
       &-0.65828058860438330047_t_real*ylm(64)
    rvylm(2,2,49)=-0.65828058860438330047_t_real*ylm(50)&
      &-0.06900655593423542176_t_real*ylm(52)
    rvylm(3,2,49)=-0.35186577527449839986_t_real*ylm(63)
    !!
    If(maxL+1<=6) goto 1000
    !!
    !!..............................................
    !! L=7
    !!..............................................
    !! L-1
    rvylm(1,1,50)= 0.70710678118654752428_t_real*ylm(37)
    rvylm(2,1,50)= 0.70710678118654752428_t_real*ylm(49)
    rvylm(1,1,51)= 0.65465367070797714369_t_real*ylm(38)
    rvylm(2,1,51)= 0.65465367070797714369_t_real*ylm(48)
    rvylm(3,1,51)= 0.37796447300922722715_t_real*ylm(37)
    rvylm(1,1,52)=-0.07412493166611011793_t_real*ylm(37)&
       &+.60219379159649482162_t_real*ylm(39)
    rvylm(2,1,52)= 0.60219379159649482162_t_real*ylm(47)&
      &+.07412493166611011793_t_real*ylm(49)
    rvylm(3,1,52)= 0.51355259101309549928_t_real*ylm(38)
    rvylm(1,1,53)=-0.12838814775327387482_t_real*ylm(38)&
       &+.54972520607827525530_t_real*ylm(40)
    rvylm(2,1,53)= 0.54972520607827525530_t_real*ylm(46)&
      &+.12838814775327387482_t_real*ylm(48)
    rvylm(3,1,53)= 0.60219379159649482162_t_real*ylm(39)
    rvylm(1,1,54)=-0.18156825980064072592_t_real*ylm(39)&
       &+.49724515809884690386_t_real*ylm(41)
    rvylm(2,1,54)= 0.49724515809884690386_t_real*ylm(45)&
      &+.18156825980064072592_t_real*ylm(47)
    rvylm(3,1,54)= 0.66299354413179587181_t_real*ylm(40)
    rvylm(1,1,55)=-0.23440361546924771150_t_real*ylm(40)&
       &+.44474958999666070758_t_real*ylm(42)
    rvylm(2,1,55)= 0.44474958999666070758_t_real*ylm(44)&
      &+.23440361546924771150_t_real*ylm(46)
    rvylm(3,1,55)= 0.70321084640774313452_t_real*ylm(41)
    rvylm(1,1,56)=-0.28708462588160728178_t_real*ylm(41)
    rvylm(2,1,56)= 0.55470019622522912192_t_real*ylm(43)&
      &+.28708462588160728178_t_real*ylm(45)
    rvylm(3,1,56)= 0.72627303920256290369_t_real*ylm(42)
    rvylm(1,1,57)=-0.48038446141526140035_t_real*ylm(44)
    rvylm(2,1,57)=-0.48038446141526140035_t_real*ylm(42)
    rvylm(3,1,57)= 0.73379938570534280692_t_real*ylm(43)
    rvylm(1,1,58)= 0.55470019622522912192_t_real*ylm(43)&
       &-0.28708462588160728178_t_real*ylm(45)
    rvylm(2,1,58)=-0.28708462588160728178_t_real*ylm(41)
    rvylm(3,1,58)= 0.72627303920256290369_t_real*ylm(44)
    rvylm(1,1,59)= 0.44474958999666070758_t_real*ylm(44)&
       &-0.23440361546924771150_t_real*ylm(46)
    rvylm(2,1,59)=-0.23440361546924771150_t_real*ylm(40)&
      &-0.44474958999666070758_t_real*ylm(42)
    rvylm(3,1,59)= 0.70321084640774313452_t_real*ylm(45)
    rvylm(1,1,60)= 0.49724515809884690386_t_real*ylm(45)&
       &-0.18156825980064072592_t_real*ylm(47)
    rvylm(2,1,60)=-0.18156825980064072592_t_real*ylm(39)&
      &-0.49724515809884690386_t_real*ylm(41)
    rvylm(3,1,60)= 0.66299354413179587181_t_real*ylm(46)
    rvylm(1,1,61)= 0.54972520607827525530_t_real*ylm(46)&
       &-0.12838814775327387482_t_real*ylm(48)
    rvylm(2,1,61)=-0.12838814775327387482_t_real*ylm(38)&
      &-0.54972520607827525530_t_real*ylm(40)
    rvylm(3,1,61)= 0.60219379159649482162_t_real*ylm(47)
    rvylm(1,1,62)= 0.60219379159649482162_t_real*ylm(47)&
       &-0.07412493166611011793_t_real*ylm(49)
    rvylm(2,1,62)=-0.07412493166611011793_t_real*ylm(37)&
      &-0.60219379159649482162_t_real*ylm(39)
    rvylm(3,1,62)= 0.51355259101309549928_t_real*ylm(48)
    rvylm(1,1,63)= 0.65465367070797714369_t_real*ylm(48)
    rvylm(2,1,63)=-0.65465367070797714369_t_real*ylm(38)
    rvylm(3,1,63)= 0.37796447300922722715_t_real*ylm(49)
    rvylm(1,1,64)= 0.70710678118654752428_t_real*ylm(49)
    rvylm(2,1,64)=-0.70710678118654752428_t_real*ylm(37)
    !! L+1
    rvylm(1,2,50)=-0.66421116415507146609_t_real*ylm(65)&
       &+.06063390625908324337_t_real*ylm(67)
    rvylm(2,2,50)= 0.06063390625908324337_t_real*ylm(79)&
      &+.66421116415507146609_t_real*ylm(81)
    rvylm(3,2,50)=-0.33210558207753573304_t_real*ylm(66)
    rvylm(1,2,51)=-0.62131265218463552729_t_real*ylm(66)&
       &+.10502100630210073525_t_real*ylm(68)
    rvylm(2,2,51)= 0.10502100630210073525_t_real*ylm(78)&
      &+.62131265218463552729_t_real*ylm(80)
    rvylm(3,2,51)=-0.45374260648651507315_t_real*ylm(67)
    rvylm(1,2,52)=-0.57841060115579811581_t_real*ylm(67)&
       &+.14852213144650114635_t_real*ylm(69)
    rvylm(2,2,52)= 0.14852213144650114635_t_real*ylm(77)&
      &+.57841060115579811581_t_real*ylm(79)
    rvylm(3,2,52)=-0.53550416047156248131_t_real*ylm(68)
    rvylm(1,2,53)=-0.53550416047156248131_t_real*ylm(68)&
       &+.19174124721184260923_t_real*ylm(70)
    rvylm(2,2,53)= 0.19174124721184260923_t_real*ylm(76)&
      &+.53550416047156248131_t_real*ylm(78)
    rvylm(3,2,53)=-0.59408852578600458542_t_real*ylm(69)
    rvylm(1,2,54)=-0.49259218307188900387_t_real*ylm(69)&
       &+.23483410915693105878_t_real*ylm(71)
    rvylm(2,2,54)= 0.23483410915693105878_t_real*ylm(75)&
      &+.49259218307188900387_t_real*ylm(77)
    rvylm(3,2,54)=-0.63593377383646038968_t_real*ylm(70)
    rvylm(1,2,55)=-0.44967308386531339777_t_real*ylm(70)&
       &+.27785946511310565116_t_real*ylm(72)
    rvylm(2,2,55)= 0.27785946511310565116_t_real*ylm(74)&
      &+.44967308386531339777_t_real*ylm(76)
    rvylm(3,2,55)=-0.66421116415507146609_t_real*ylm(71)
    rvylm(1,2,56)=-0.40674460840998032245_t_real*ylm(71)
    rvylm(2,2,56)= 0.45374260648651507315_t_real*ylm(73)&
      &+.40674460840998032245_t_real*ylm(75)
    rvylm(3,2,56)=-0.68061390972977260973_t_real*ylm(72)
    rvylm(1,2,57)=-0.51449575542752651206_t_real*ylm(74)
    rvylm(2,2,57)=-0.51449575542752651206_t_real*ylm(72)
    rvylm(3,2,57)=-0.68599434057003534944_t_real*ylm(73)
    rvylm(1,2,58)= 0.45374260648651507315_t_real*ylm(73)&
       &-0.40674460840998032245_t_real*ylm(75)
    rvylm(2,2,58)=-0.40674460840998032245_t_real*ylm(71)
    rvylm(3,2,58)=-0.68061390972977260973_t_real*ylm(74)
    rvylm(1,2,59)= 0.27785946511310565116_t_real*ylm(74)&
       &-0.44967308386531339777_t_real*ylm(76)
    rvylm(2,2,59)=-0.44967308386531339777_t_real*ylm(70)&
      &-0.27785946511310565116_t_real*ylm(72)
    rvylm(3,2,59)=-0.66421116415507146609_t_real*ylm(75)
    rvylm(1,2,60)= 0.23483410915693105878_t_real*ylm(75)&
       &-0.49259218307188900387_t_real*ylm(77)
    rvylm(2,2,60)=-0.49259218307188900387_t_real*ylm(69)&
      &-0.23483410915693105878_t_real*ylm(71)
    rvylm(3,2,60)=-0.63593377383646038968_t_real*ylm(76)
    rvylm(1,2,61)= 0.19174124721184260923_t_real*ylm(76)&
       &-0.53550416047156248131_t_real*ylm(78)
    rvylm(2,2,61)=-0.53550416047156248131_t_real*ylm(68)&
      &-0.19174124721184260923_t_real*ylm(70)
    rvylm(3,2,61)=-0.59408852578600458542_t_real*ylm(77)
    rvylm(1,2,62)= 0.14852213144650114635_t_real*ylm(77)&
       &-0.57841060115579811581_t_real*ylm(79)
    rvylm(2,2,62)=-0.57841060115579811581_t_real*ylm(67)&
      &-0.14852213144650114635_t_real*ylm(69)
    rvylm(3,2,62)=-0.53550416047156248131_t_real*ylm(78)
    rvylm(1,2,63)= 0.10502100630210073525_t_real*ylm(78)&
       &-0.62131265218463552729_t_real*ylm(80)
    rvylm(2,2,63)=-0.62131265218463552729_t_real*ylm(66)&
      &-0.10502100630210073525_t_real*ylm(68)
    rvylm(3,2,63)=-0.45374260648651507315_t_real*ylm(79)
    rvylm(1,2,64)= 0.06063390625908324337_t_real*ylm(79)&
       &-0.66421116415507146609_t_real*ylm(81)
    rvylm(2,2,64)=-0.66421116415507146609_t_real*ylm(65)&
      &-0.06063390625908324337_t_real*ylm(67)
    rvylm(3,2,64)=-0.33210558207753573304_t_real*ylm(80)
    !!
    If(maxL+1<=7) goto 1000
    !!
    !!..............................................
    !! L=8
    !!..............................................
    !! L-1
    rvylm(1,1,65)= 0.70710678118654752427_t_real*ylm(50)
    rvylm(2,1,65)= 0.70710678118654752427_t_real*ylm(64)
    rvylm(1,1,66)= 0.66143782776614764750_t_real*ylm(51)
    rvylm(2,1,66)= 0.66143782776614764750_t_real*ylm(63)
    rvylm(3,1,66)= 0.35355339059327376213_t_real*ylm(50)
    rvylm(1,1,67)=-0.06454972243679028140_t_real*ylm(50)&
       &+.61576510673037219975_t_real*ylm(52)
    rvylm(2,1,67)= 0.61576510673037219975_t_real*ylm(62)&
      &+.06454972243679028140_t_real*ylm(64)
    rvylm(3,1,67)= 0.48304589153964795237_t_real*ylm(51)
    rvylm(1,1,68)=-0.11180339887498948480_t_real*ylm(51)&
       &+.57008771254956898946_t_real*ylm(53)
    rvylm(2,1,68)= 0.57008771254956898946_t_real*ylm(61)&
      &+.11180339887498948480_t_real*ylm(63)
    rvylm(3,1,68)= 0.57008771254956898946_t_real*ylm(52)
    rvylm(1,1,69)=-0.15811388300841896657_t_real*ylm(52)&
       &+.52440442408507577340_t_real*ylm(54)
    rvylm(2,1,69)= 0.52440442408507577340_t_real*ylm(60)&
      &+.15811388300841896657_t_real*ylm(62)
    rvylm(3,1,69)= 0.63245553203367586628_t_real*ylm(53)
    rvylm(1,1,70)=-0.20412414523193150814_t_real*ylm(53)&
       &+.47871355387816905490_t_real*ylm(55)
    rvylm(2,1,70)= 0.47871355387816905490_t_real*ylm(59)&
      &+.20412414523193150814_t_real*ylm(61)
    rvylm(3,1,70)= 0.67700320038633002991_t_real*ylm(54)
    rvylm(1,1,71)=-0.24999999999999999995_t_real*ylm(54)&
       &+.43301270189221932330_t_real*ylm(56)
    rvylm(2,1,71)= 0.43301270189221932330_t_real*ylm(58)&
      &+.24999999999999999995_t_real*ylm(60)
    rvylm(3,1,71)= 0.70710678118654752427_t_real*ylm(55)
    rvylm(1,1,72)=-0.29580398915498080207_t_real*ylm(55)
    rvylm(2,1,72)= 0.54772255750516611334_t_real*ylm(57)&
      &+.29580398915498080207_t_real*ylm(59)
    rvylm(3,1,72)= 0.72456883730947192855_t_real*ylm(56)
    rvylm(1,1,73)=-0.48304589153964795236_t_real*ylm(58)
    rvylm(2,1,73)=-0.48304589153964795236_t_real*ylm(56)
    rvylm(3,1,73)= 0.73029674334022148448_t_real*ylm(57)
    rvylm(1,1,74)= 0.54772255750516611334_t_real*ylm(57)&
       &-0.29580398915498080207_t_real*ylm(59)
    rvylm(2,1,74)=-0.29580398915498080207_t_real*ylm(55)
    rvylm(3,1,74)= 0.72456883730947192855_t_real*ylm(58)
    rvylm(1,1,75)= 0.43301270189221932330_t_real*ylm(58)&
       &-0.24999999999999999995_t_real*ylm(60)
    rvylm(2,1,75)=-0.24999999999999999995_t_real*ylm(54)&
      &-0.43301270189221932330_t_real*ylm(56)
    rvylm(3,1,75)= 0.70710678118654752427_t_real*ylm(59)
    rvylm(1,1,76)= 0.47871355387816905490_t_real*ylm(59)&
       &-0.20412414523193150814_t_real*ylm(61)
    rvylm(2,1,76)=-0.20412414523193150814_t_real*ylm(53)&
      &-0.47871355387816905490_t_real*ylm(55)
    rvylm(3,1,76)= 0.67700320038633002991_t_real*ylm(60)
    rvylm(1,1,77)= 0.52440442408507577340_t_real*ylm(60)&
       &-0.15811388300841896657_t_real*ylm(62)
    rvylm(2,1,77)=-0.15811388300841896657_t_real*ylm(52)&
      &-0.52440442408507577340_t_real*ylm(54)
    rvylm(3,1,77)= 0.63245553203367586628_t_real*ylm(61)
    rvylm(1,1,78)= 0.57008771254956898946_t_real*ylm(61)&
       &-0.11180339887498948480_t_real*ylm(63)
    rvylm(2,1,78)=-0.11180339887498948480_t_real*ylm(51)&
      &-0.57008771254956898946_t_real*ylm(53)
    rvylm(3,1,78)= 0.57008771254956898946_t_real*ylm(62)
    rvylm(1,1,79)= 0.61576510673037219975_t_real*ylm(62)&
       &-0.06454972243679028140_t_real*ylm(64)
    rvylm(2,1,79)=-0.06454972243679028140_t_real*ylm(50)&
      &-0.61576510673037219975_t_real*ylm(52)
    rvylm(3,1,79)= 0.48304589153964795237_t_real*ylm(63)
    rvylm(1,1,80)= 0.66143782776614764750_t_real*ylm(63)
    rvylm(2,1,80)=-0.66143782776614764750_t_real*ylm(51)
    rvylm(3,1,80)= 0.35355339059327376213_t_real*ylm(64)
    rvylm(1,1,81)= 0.70710678118654752427_t_real*ylm(64)
    rvylm(2,1,81)=-0.70710678118654752427_t_real*ylm(50)
    !! L+1
    rvylm(1,2,65)=-0.66885605405993865379_t_real*ylm(82)&
       &+.05407380704358751271_t_real*ylm(84)
    rvylm(2,2,65)= 0.05407380704358751271_t_real*ylm(98)&
      &+.66885605405993865379_t_real*ylm(100)
    rvylm(3,2,65)=-0.31530176764230576240_t_real*ylm(83)
    rvylm(1,2,66)=-0.63060353528461152480_t_real*ylm(83)&
       &+.09365858115816939658_t_real*ylm(85)
    rvylm(2,2,66)= 0.09365858115816939658_t_real*ylm(97)&
      &+.63060353528461152480_t_real*ylm(99)
    rvylm(3,2,66)=-0.43259045634870010175_t_real*ylm(84)
    rvylm(1,2,67)=-0.59234887775909235728_t_real*ylm(84)&
       &+.13245323570650438064_t_real*ylm(86)
    rvylm(2,2,67)= 0.13245323570650438064_t_real*ylm(96)&
      &+.59234887775909235728_t_real*ylm(98)
    rvylm(3,2,67)=-0.51298917604257704771_t_real*ylm(85)
    rvylm(1,2,68)=-0.55409163850364058962_t_real*ylm(85)&
       &+.17099639201419234923_t_real*ylm(87)
    rvylm(2,2,68)= 0.17099639201419234923_t_real*ylm(95)&
      &+.55409163850364058962_t_real*ylm(97)
    rvylm(3,2,68)=-0.57226338351930142019_t_real*ylm(86)
    rvylm(1,2,69)=-0.51583124308733882637_t_real*ylm(86)&
       &+.20942695414584775348_t_real*ylm(88)
    rvylm(2,2,69)= 0.20942695414584775348_t_real*ylm(94)&
      &+.51583124308733882637_t_real*ylm(96)
    rvylm(3,2,69)=-0.61653625932651153794_t_real*ylm(87)
    rvylm(1,2,70)=-0.47756693294091927864_t_real*ylm(87)&
       &+.24779731389167604243_t_real*ylm(89)
    rvylm(2,2,70)= 0.24779731389167604243_t_real*ylm(93)&
      &+.47756693294091927864_t_real*ylm(95)
    rvylm(3,2,70)=-0.64888568452305015263_t_real*ylm(88)
    rvylm(1,2,71)=-0.43929768510697945308_t_real*ylm(88)&
       &+.28613169175965071009_t_real*ylm(90)
    rvylm(2,2,71)= 0.28613169175965071009_t_real*ylm(92)&
      &+.43929768510697945308_t_real*ylm(94)
    rvylm(3,2,71)=-0.67103829820720271123_t_real*ylm(89)
    rvylm(1,2,72)=-0.40102208598815651471_t_real*ylm(89)
    rvylm(2,2,72)= 0.45883146774112353178_t_real*ylm(91)&
      &+.40102208598815651471_t_real*ylm(93)
    rvylm(3,2,72)=-0.68398556805676939694_t_real*ylm(90)
    rvylm(1,2,73)=-0.51298917604257704770_t_real*ylm(92)
    rvylm(2,2,73)=-0.51298917604257704770_t_real*ylm(90)
    rvylm(3,2,73)=-0.68824720161168529770_t_real*ylm(91)
    rvylm(1,2,74)= 0.45883146774112353178_t_real*ylm(91)&
       &-0.40102208598815651471_t_real*ylm(93)
    rvylm(2,2,74)=-0.40102208598815651471_t_real*ylm(89)
    rvylm(3,2,74)=-0.68398556805676939694_t_real*ylm(92)
    rvylm(1,2,75)= 0.28613169175965071009_t_real*ylm(92)&
       &-0.43929768510697945308_t_real*ylm(94)
    rvylm(2,2,75)=-0.43929768510697945308_t_real*ylm(88)&
      &-0.28613169175965071009_t_real*ylm(90)
    rvylm(3,2,75)=-0.67103829820720271123_t_real*ylm(93)
    rvylm(1,2,76)= 0.24779731389167604243_t_real*ylm(93)&
       &-0.47756693294091927864_t_real*ylm(95)
    rvylm(2,2,76)=-0.47756693294091927864_t_real*ylm(87)&
      &-0.24779731389167604243_t_real*ylm(89)
    rvylm(3,2,76)=-0.64888568452305015263_t_real*ylm(94)
    rvylm(1,2,77)= 0.20942695414584775348_t_real*ylm(94)&
       &-0.51583124308733882637_t_real*ylm(96)
    rvylm(2,2,77)=-0.51583124308733882637_t_real*ylm(86)&
      &-0.20942695414584775348_t_real*ylm(88)
    rvylm(3,2,77)=-0.61653625932651153794_t_real*ylm(95)
    rvylm(1,2,78)= 0.17099639201419234923_t_real*ylm(95)&
       &-0.55409163850364058962_t_real*ylm(97)
    rvylm(2,2,78)=-0.55409163850364058962_t_real*ylm(85)&
      &-0.17099639201419234923_t_real*ylm(87)
    rvylm(3,2,78)=-0.57226338351930142019_t_real*ylm(96)
    rvylm(1,2,79)= 0.13245323570650438064_t_real*ylm(96)&
       &-0.59234887775909235728_t_real*ylm(98)
    rvylm(2,2,79)=-0.59234887775909235728_t_real*ylm(84)&
      &-0.13245323570650438064_t_real*ylm(86)
    rvylm(3,2,79)=-0.51298917604257704771_t_real*ylm(97)
    rvylm(1,2,80)= 0.09365858115816939658_t_real*ylm(97)&
       &-0.63060353528461152480_t_real*ylm(99)
    rvylm(2,2,80)=-0.63060353528461152480_t_real*ylm(83)&
      &-0.09365858115816939658_t_real*ylm(85)
    rvylm(3,2,80)=-0.43259045634870010175_t_real*ylm(98)
    rvylm(1,2,81)= 0.05407380704358751271_t_real*ylm(98)&
       &-0.66885605405993865379_t_real*ylm(100)
    rvylm(2,2,81)=-0.66885605405993865379_t_real*ylm(82)&
      &-0.05407380704358751271_t_real*ylm(84)
    rvylm(3,2,81)=-0.31530176764230576240_t_real*ylm(99)
    !!
    If(maxL+1<=8) goto 1000
    !!
    !!..............................................
    !! L=9
    !!..............................................
    !! L-1
    rvylm(1,1,82)= 0.70710678118654752434_t_real*ylm(65)
    rvylm(2,1,82)= 0.70710678118654752434_t_real*ylm(81)
    rvylm(1,1,83)= 0.66666666666666666661_t_real*ylm(66)
    rvylm(2,1,83)= 0.66666666666666666661_t_real*ylm(80)
    rvylm(3,1,83)= 0.33333333333333333330_t_real*ylm(65)
    rvylm(1,1,84)=-0.05716619504750294578_t_real*ylm(65)&
       &+.62622429108514949011_t_real*ylm(67)
    rvylm(2,1,84)= 0.62622429108514949011_t_real*ylm(79)&
      &+.05716619504750294578_t_real*ylm(81)
    rvylm(3,1,84)= 0.45732956038002356630_t_real*ylm(66)
    rvylm(1,1,85)=-0.09901475429766743090_t_real*ylm(66)&
       &+.58577918612900611002_t_real*ylm(68)
    rvylm(2,1,85)= 0.58577918612900611002_t_real*ylm(78)&
      &+.09901475429766743090_t_real*ylm(80)
    rvylm(3,1,85)= 0.54232614454664042995_t_real*ylm(67)
    rvylm(1,1,86)=-0.14002800840280098034_t_real*ylm(67)&
       &+.54533074451660313454_t_real*ylm(69)
    rvylm(2,1,86)= 0.54533074451660313454_t_real*ylm(77)&
      &+.14002800840280098034_t_real*ylm(79)
    rvylm(3,1,86)= 0.60499014198202009756_t_real*ylm(68)
    rvylm(1,1,87)=-0.18077538151554680998_t_real*ylm(68)&
       &+.50487816429740128487_t_real*ylm(70)
    rvylm(2,1,87)= 0.50487816429740128487_t_real*ylm(76)&
      &+.18077538151554680998_t_real*ylm(78)
    rvylm(3,1,87)= 0.65179490739586907454_t_real*ylm(69)
    rvylm(1,1,88)=-0.22140372138502382203_t_real*ylm(69)&
       &+.46442036401282396989_t_real*ylm(71)
    rvylm(2,1,88)= 0.46442036401282396989_t_real*ylm(75)&
      &+.22140372138502382203_t_real*ylm(77)
    rvylm(3,1,88)= 0.68599434057003534946_t_real*ylm(70)
    rvylm(1,1,89)=-0.26196841599779191101_t_real*ylm(70)&
       &+.42395584922430692646_t_real*ylm(72)
    rvylm(2,1,89)= 0.42395584922430692646_t_real*ylm(74)&
      &+.26196841599779191101_t_real*ylm(76)
    rvylm(3,1,89)= 0.70941382412257033084_t_real*ylm(71)
    rvylm(1,1,90)=-0.30249507099101004878_t_real*ylm(71)
    rvylm(2,1,90)= 0.54232614454664042995_t_real*ylm(73)&
      &+.30249507099101004878_t_real*ylm(75)
    rvylm(3,1,90)= 0.72310152606218723994_t_real*ylm(72)
    rvylm(1,1,91)=-0.48507125007266594699_t_real*ylm(74)
    rvylm(2,1,91)=-0.48507125007266594699_t_real*ylm(72)
    rvylm(3,1,91)= 0.72760687510899892050_t_real*ylm(73)
    rvylm(1,1,92)= 0.54232614454664042995_t_real*ylm(73)&
       &-0.30249507099101004878_t_real*ylm(75)
    rvylm(2,1,92)=-0.30249507099101004878_t_real*ylm(71)
    rvylm(3,1,92)= 0.72310152606218723994_t_real*ylm(74)
    rvylm(1,1,93)= 0.42395584922430692646_t_real*ylm(74)&
       &-0.26196841599779191101_t_real*ylm(76)
    rvylm(2,1,93)=-0.26196841599779191101_t_real*ylm(70)&
      &-0.42395584922430692646_t_real*ylm(72)
    rvylm(3,1,93)= 0.70941382412257033084_t_real*ylm(75)
    rvylm(1,1,94)= 0.46442036401282396989_t_real*ylm(75)&
       &-0.22140372138502382203_t_real*ylm(77)
    rvylm(2,1,94)=-0.22140372138502382203_t_real*ylm(69)&
      &-0.46442036401282396989_t_real*ylm(71)
    rvylm(3,1,94)= 0.68599434057003534946_t_real*ylm(76)
    rvylm(1,1,95)= 0.50487816429740128487_t_real*ylm(76)&
       &-0.18077538151554680998_t_real*ylm(78)
    rvylm(2,1,95)=-0.18077538151554680998_t_real*ylm(68)&
      &-0.50487816429740128487_t_real*ylm(70)
    rvylm(3,1,95)= 0.65179490739586907454_t_real*ylm(77)
    rvylm(1,1,96)= 0.54533074451660313454_t_real*ylm(77)&
       &-0.14002800840280098034_t_real*ylm(79)
    rvylm(2,1,96)=-0.14002800840280098034_t_real*ylm(67)&
      &-0.54533074451660313454_t_real*ylm(69)
    rvylm(3,1,96)= 0.60499014198202009756_t_real*ylm(78)
    rvylm(1,1,97)= 0.58577918612900611002_t_real*ylm(78)&
       &-0.09901475429766743090_t_real*ylm(80)
    rvylm(2,1,97)=-0.09901475429766743090_t_real*ylm(66)&
      &-0.58577918612900611002_t_real*ylm(68)
    rvylm(3,1,97)= 0.54232614454664042995_t_real*ylm(79)
    rvylm(1,1,98)= 0.62622429108514949011_t_real*ylm(79)&
       &-0.05716619504750294578_t_real*ylm(81)
    rvylm(2,1,98)=-0.05716619504750294578_t_real*ylm(65)&
      &-0.62622429108514949011_t_real*ylm(67)
    rvylm(3,1,98)= 0.45732956038002356630_t_real*ylm(80)
    rvylm(1,1,99)= 0.66666666666666666661_t_real*ylm(80)
    rvylm(2,1,99)=-0.66666666666666666661_t_real*ylm(66)
    rvylm(3,1,99)= 0.33333333333333333330_t_real*ylm(81)
    rvylm(1,1,100)= 0.70710678118654752434_t_real*ylm(81)
    rvylm(2,1,100)=-0.70710678118654752434_t_real*ylm(65)
    !! L+1
    rvylm(1,2,82)=-0.67259270913454925786_t_real*ylm(101)&
       &+.04879500364742665895_t_real*ylm(103)
    rvylm(2,2,82)= 0.04879500364742665895_t_real*ylm(119)&
      &+.67259270913454925786_t_real*ylm(121)
    rvylm(3,2,82)=-0.30079260375911917714_t_real*ylm(102)
    rvylm(1,2,83)=-0.63807746954649412901_t_real*ylm(102)&
       &+.08451542547285165772_t_real*ylm(104)
    rvylm(2,2,83)= 0.08451542547285165772_t_real*ylm(118)&
      &+.63807746954649412901_t_real*ylm(120)
    rvylm(3,2,83)=-0.41403933560541253056_t_real*ylm(103)
    rvylm(1,2,84)=-0.60356086212221736926_t_real*ylm(103)&
       &+.11952286093343936396_t_real*ylm(105)
    rvylm(2,2,84)= 0.11952286093343936396_t_real*ylm(117)&
      &+.60356086212221736926_t_real*ylm(119)
    rvylm(3,2,84)=-0.49280538030458114815_t_real*ylm(104)
    rvylm(1,2,85)=-0.56904263795389164962_t_real*ylm(104)&
       &+.15430334996209191021_t_real*ylm(106)
    rvylm(2,2,85)= 0.15430334996209191021_t_real*ylm(116)&
      &+.56904263795389164962_t_real*ylm(118)
    rvylm(3,2,85)=-0.55205244747388337408_t_real*ylm(105)
    rvylm(1,2,86)=-0.53452248382484876921_t_real*ylm(105)&
       &+.18898223650461361355_t_real*ylm(107)
    rvylm(2,2,86)= 0.18898223650461361355_t_real*ylm(115)&
      &+.53452248382484876921_t_real*ylm(117)
    rvylm(3,2,86)=-0.59761430466719681981_t_real*ylm(106)
    rvylm(1,2,87)=-0.49999999999999999985_t_real*ylm(106)&
       &+.22360679774997896957_t_real*ylm(108)
    rvylm(2,2,87)= 0.22360679774997896957_t_real*ylm(114)&
      &+.49999999999999999985_t_real*ylm(116)
    rvylm(3,2,87)=-0.63245553203367586621_t_real*ylm(107)
    rvylm(1,2,88)=-0.46547466812563137219_t_real*ylm(107)&
       &+.25819888974716112560_t_real*ylm(109)
    rvylm(2,2,88)= 0.25819888974716112560_t_real*ylm(113)&
      &+.46547466812563137219_t_real*ylm(115)
    rvylm(3,2,88)=-0.65828058860438330039_t_real*ylm(108)
    rvylm(1,2,89)=-0.43094580368566731662_t_real*ylm(108)&
       &+.29277002188455995372_t_real*ylm(110)
    rvylm(2,2,89)= 0.29277002188455995372_t_real*ylm(112)&
      &+.43094580368566731662_t_real*ylm(114)
    rvylm(3,2,89)=-0.67612340378281326181_t_real*ylm(109)
    rvylm(1,2,90)=-0.39641248358604594122_t_real*ylm(109)
    rvylm(2,2,90)= 0.46291004988627573064_t_real*ylm(111)&
      &+.39641248358604594122_t_real*ylm(113)
    rvylm(3,2,90)=-0.68660656232559518721_t_real*ylm(110)
    rvylm(1,2,91)=-0.51176631571915897696_t_real*ylm(112)
    rvylm(2,2,91)=-0.51176631571915897696_t_real*ylm(110)
    rvylm(3,2,91)=-0.69006555934235421760_t_real*ylm(111)
    rvylm(1,2,92)= 0.46291004988627573064_t_real*ylm(111)&
       &-0.39641248358604594122_t_real*ylm(113)
    rvylm(2,2,92)=-0.39641248358604594122_t_real*ylm(109)
    rvylm(3,2,92)=-0.68660656232559518721_t_real*ylm(112)
    rvylm(1,2,93)= 0.29277002188455995372_t_real*ylm(112)&
       &-0.43094580368566731662_t_real*ylm(114)
    rvylm(2,2,93)=-0.43094580368566731662_t_real*ylm(108)&
      &-0.29277002188455995372_t_real*ylm(110)
    rvylm(3,2,93)=-0.67612340378281326181_t_real*ylm(113)
    rvylm(1,2,94)= 0.25819888974716112560_t_real*ylm(113)&
       &-0.46547466812563137219_t_real*ylm(115)
    rvylm(2,2,94)=-0.46547466812563137219_t_real*ylm(107)&
      &-0.25819888974716112560_t_real*ylm(109)
    rvylm(3,2,94)=-0.65828058860438330039_t_real*ylm(114)
    rvylm(1,2,95)= 0.22360679774997896957_t_real*ylm(114)&
       &-0.49999999999999999985_t_real*ylm(116)
    rvylm(2,2,95)=-0.49999999999999999985_t_real*ylm(106)&
      &-0.22360679774997896957_t_real*ylm(108)
    rvylm(3,2,95)=-0.63245553203367586621_t_real*ylm(115)
    rvylm(1,2,96)= 0.18898223650461361355_t_real*ylm(115)&
       &-0.53452248382484876921_t_real*ylm(117)
    rvylm(2,2,96)=-0.53452248382484876921_t_real*ylm(105)&
      &-0.18898223650461361355_t_real*ylm(107)
    rvylm(3,2,96)=-0.59761430466719681981_t_real*ylm(116)
    rvylm(1,2,97)= 0.15430334996209191021_t_real*ylm(116)&
       &-0.56904263795389164962_t_real*ylm(118)
    rvylm(2,2,97)=-0.56904263795389164962_t_real*ylm(104)&
      &-0.15430334996209191021_t_real*ylm(106)
    rvylm(3,2,97)=-0.55205244747388337408_t_real*ylm(117)
    rvylm(1,2,98)= 0.11952286093343936396_t_real*ylm(117)&
       &-0.60356086212221736926_t_real*ylm(119)
    rvylm(2,2,98)=-0.60356086212221736926_t_real*ylm(103)&
      &-0.11952286093343936396_t_real*ylm(105)
    rvylm(3,2,98)=-0.49280538030458114815_t_real*ylm(118)
    rvylm(1,2,99)= 0.08451542547285165772_t_real*ylm(118)&
       &-0.63807746954649412901_t_real*ylm(120)
    rvylm(2,2,99)=-0.63807746954649412901_t_real*ylm(102)&
      &-0.08451542547285165772_t_real*ylm(104)
    rvylm(3,2,99)=-0.41403933560541253056_t_real*ylm(119)
    rvylm(1,2,100)= 0.04879500364742665895_t_real*ylm(119)&
       &-0.67259270913454925786_t_real*ylm(121)
    rvylm(2,2,100)=-0.67259270913454925786_t_real*ylm(101)&
      &-0.04879500364742665895_t_real*ylm(103)
    rvylm(3,2,100)=-0.30079260375911917714_t_real*ylm(120)
    !!
    If(maxL+1<=9) goto 1000
    !!
    !!..............................................
    !! L=10
    !!..............................................
    !! L-1
    rvylm(1,1,101)= 0.70710678118654752420_t_real*ylm(82)
    rvylm(2,1,101)= 0.70710678118654752420_t_real*ylm(100)
    rvylm(1,1,102)= 0.67082039324993690873_t_real*ylm(83)
    rvylm(2,1,102)= 0.67082039324993690873_t_real*ylm(99)
    rvylm(3,1,102)= 0.31622776601683793311_t_real*ylm(82)
    rvylm(1,1,103)=-0.05129891760425770475_t_real*ylm(82)&
       &+.63453256728663535228_t_real*ylm(84)
    rvylm(2,1,103)= 0.63453256728663535228_t_real*ylm(98)&
      &+.05129891760425770475_t_real*ylm(100)
    rvylm(3,1,103)= 0.43528575006600700308_t_real*ylm(83)
    rvylm(1,1,104)=-0.08885233166386385357_t_real*ylm(83)&
       &+.59824304161611881046_t_real*ylm(85)
    rvylm(2,1,104)= 0.59824304161611881046_t_real*ylm(97)&
      &+.08885233166386385357_t_real*ylm(99)
    rvylm(3,1,104)= 0.51809367167683002613_t_real*ylm(84)
    rvylm(1,1,105)=-0.12565617248750865205_t_real*ylm(84)&
       &+.56195148694901637936_t_real*ylm(86)
    rvylm(2,1,105)= 0.56195148694901637936_t_real*ylm(96)&
      &+.12565617248750865205_t_real*ylm(98)
    rvylm(3,1,105)= 0.58038100008800933744_t_real*ylm(85)
    rvylm(1,1,106)=-0.16222142113076253811_t_real*ylm(85)&
       &+.52565748303784681342_t_real*ylm(87)
    rvylm(2,1,106)= 0.52565748303784681342_t_real*ylm(95)&
      &+.16222142113076253811_t_real*ylm(97)
    rvylm(3,1,106)= 0.62828086243754326030_t_real*ylm(86)
    rvylm(1,1,107)=-0.19867985355975657092_t_real*ylm(86)&
       &+.48936048492959289577_t_real*ylm(88)
    rvylm(2,1,107)= 0.48936048492959289577_t_real*ylm(94)&
      &+.19867985355975657092_t_real*ylm(96)
    rvylm(3,1,107)= 0.66490996620436870738_t_real*ylm(87)
    rvylm(1,1,108)=-0.23508117299081362648_t_real*ylm(87)&
       &+.45305977298225986065_t_real*ylm(89)
    rvylm(2,1,108)= 0.45305977298225986065_t_real*ylm(93)&
      &+.23508117299081362648_t_real*ylm(95)
    rvylm(3,1,108)= 0.69206023467690486234_t_real*ylm(88)
    rvylm(1,1,109)=-0.27144835701531845785_t_real*ylm(88)&
       &+.41675437673324534550_t_real*ylm(90)
    rvylm(2,1,109)= 0.41675437673324534550_t_real*ylm(92)&
      &+.27144835701531845785_t_real*ylm(94)
    rvylm(3,1,109)= 0.71081865331091082856_t_real*ylm(89)
    rvylm(1,1,110)=-0.30779350562554622855_t_real*ylm(89)
    rvylm(2,1,110)= 0.53802758684897033411_t_real*ylm(91)&
      &+.30779350562554622855_t_real*ylm(93)
    rvylm(3,1,110)= 0.72183975477868172631_t_real*ylm(90)
    rvylm(1,1,111)=-0.48666426339228761435_t_real*ylm(92)
    rvylm(2,1,111)=-0.48666426339228761435_t_real*ylm(90)
    rvylm(3,1,111)= 0.72547625011001167180_t_real*ylm(91)
    rvylm(1,1,112)= 0.53802758684897033411_t_real*ylm(91)&
       &-0.30779350562554622855_t_real*ylm(93)
    rvylm(2,1,112)=-0.30779350562554622855_t_real*ylm(89)
    rvylm(3,1,112)= 0.72183975477868172631_t_real*ylm(92)
    rvylm(1,1,113)= 0.41675437673324534550_t_real*ylm(92)&
       &-0.27144835701531845785_t_real*ylm(94)
    rvylm(2,1,113)=-0.27144835701531845785_t_real*ylm(88)&
      &-0.41675437673324534550_t_real*ylm(90)
    rvylm(3,1,113)= 0.71081865331091082856_t_real*ylm(93)
    rvylm(1,1,114)= 0.45305977298225986065_t_real*ylm(93)&
       &-0.23508117299081362648_t_real*ylm(95)
    rvylm(2,1,114)=-0.23508117299081362648_t_real*ylm(87)&
      &-0.45305977298225986065_t_real*ylm(89)
    rvylm(3,1,114)= 0.69206023467690486234_t_real*ylm(94)
    rvylm(1,1,115)= 0.48936048492959289577_t_real*ylm(94)&
       &-0.19867985355975657092_t_real*ylm(96)
    rvylm(2,1,115)=-0.19867985355975657092_t_real*ylm(86)&
      &-0.48936048492959289577_t_real*ylm(88)
    rvylm(3,1,115)= 0.66490996620436870738_t_real*ylm(95)
    rvylm(1,1,116)= 0.52565748303784681342_t_real*ylm(95)&
       &-0.16222142113076253811_t_real*ylm(97)
    rvylm(2,1,116)=-0.16222142113076253811_t_real*ylm(85)&
      &-0.52565748303784681342_t_real*ylm(87)
    rvylm(3,1,116)= 0.62828086243754326030_t_real*ylm(96)
    rvylm(1,1,117)= 0.56195148694901637936_t_real*ylm(96)&
       &-0.12565617248750865205_t_real*ylm(98)
    rvylm(2,1,117)=-0.12565617248750865205_t_real*ylm(84)&
      &-0.56195148694901637936_t_real*ylm(86)
    rvylm(3,1,117)= 0.58038100008800933744_t_real*ylm(97)
    rvylm(1,1,118)= 0.59824304161611881046_t_real*ylm(97)&
       &-0.08885233166386385357_t_real*ylm(99)
    rvylm(2,1,118)=-0.08885233166386385357_t_real*ylm(83)&
      &-0.59824304161611881046_t_real*ylm(85)
    rvylm(3,1,118)= 0.51809367167683002613_t_real*ylm(98)
    rvylm(1,1,119)= 0.63453256728663535228_t_real*ylm(98)&
       &-0.05129891760425770475_t_real*ylm(100)
    rvylm(2,1,119)=-0.05129891760425770475_t_real*ylm(82)&
      &-0.63453256728663535228_t_real*ylm(84)
    rvylm(3,1,119)= 0.43528575006600700308_t_real*ylm(99)
    rvylm(1,1,120)= 0.67082039324993690873_t_real*ylm(99)
    rvylm(2,1,120)=-0.67082039324993690873_t_real*ylm(83)
    rvylm(3,1,120)= 0.31622776601683793311_t_real*ylm(100)
    rvylm(1,1,121)= 0.70710678118654752420_t_real*ylm(100)
    rvylm(2,1,121)=-0.70710678118654752420_t_real*ylm(82)
    !! L+1
    rvylm(1,2,101)=-0.67566392469217622316_t_real*ylm(122)&
       &+.04445542244743870560_t_real*ylm(124)
    rvylm(2,2,101)= 0.04445542244743870560_t_real*ylm(142)&
      &+.67566392469217622316_t_real*ylm(144)
    rvylm(3,2,101)=-0.28810406552003038415_t_real*ylm(123)
    rvylm(1,2,102)=-0.64422027509684123750_t_real*ylm(123)&
       &+.07699905035090180553_t_real*ylm(125)
    rvylm(2,2,102)= 0.07699905035090180553_t_real*ylm(141)&
      &+.64422027509684123750_t_real*ylm(143)
    rvylm(3,2,102)=-0.39762138624377206946_t_real*ylm(124)
    rvylm(1,2,103)=-0.61277571031026112242_t_real*ylm(124)&
       &+.10889310129609415676_t_real*ylm(126)
    rvylm(2,2,103)= 0.10889310129609415676_t_real*ylm(140)&
      &+.61277571031026112242_t_real*ylm(142)
    rvylm(3,2,103)=-0.47465402419841236983_t_real*ylm(125)
    rvylm(1,2,104)=-0.58133008182238475428_t_real*ylm(125)&
       &+.14058038927888331729_t_real*ylm(127)
    rvylm(2,2,104)= 0.14058038927888331729_t_real*ylm(139)&
      &+.58133008182238475428_t_real*ylm(141)
    rvylm(3,2,104)=-0.53346506936926446727_t_real*ylm(126)
    rvylm(1,2,105)=-0.54988320714673246380_t_real*ylm(126)&
       &+.17217511078754547242_t_real*ylm(128)
    rvylm(2,2,105)= 0.17217511078754547242_t_real*ylm(138)&
      &+.54988320714673246380_t_real*ylm(140)
    rvylm(3,2,105)=-0.57962779388728445979_t_real*ylm(127)
    rvylm(1,2,106)=-0.51843485951208205086_t_real*ylm(127)&
       &+.20372033841662687615_t_real*ylm(129)
    rvylm(2,2,106)= 0.20372033841662687615_t_real*ylm(137)&
      &+.51843485951208205086_t_real*ylm(139)
    rvylm(3,2,106)=-0.61599240280721444426_t_real*ylm(128)
    rvylm(1,2,107)=-0.48698475355767397830_t_real*ylm(128)&
       &+.23523598444848237187_t_real*ylm(130)
    rvylm(2,2,107)= 0.23523598444848237187_t_real*ylm(136)&
      &+.48698475355767397830_t_real*ylm(138)
    rvylm(3,2,107)=-0.64422027509684123750_t_real*ylm(129)
    rvylm(1,2,108)=-0.45553252509883956821_t_real*ylm(129)&
       &+.26673253468463223363_t_real*ylm(131)
    rvylm(2,2,108)= 0.26673253468463223363_t_real*ylm(135)&
      &+.45553252509883956821_t_real*ylm(137)
    rvylm(3,2,108)=-0.66534783913046048344_t_real*ylm(130)
    rvylm(1,2,109)=-0.42407770188162638293_t_real*ylm(130)&
       &+.29821603968282905209_t_real*ylm(132)
    rvylm(2,2,109)= 0.29821603968282905209_t_real*ylm(134)&
      &+.42407770188162638293_t_real*ylm(136)
    rvylm(3,2,109)=-0.68003719963350203039_t_real*ylm(131)
    rvylm(1,2,110)=-0.39261966026736167253_t_real*ylm(131)
    rvylm(2,2,110)= 0.46625240412015688279_t_real*ylm(133)&
      &+.39261966026736167253_t_real*ylm(135)
    rvylm(3,2,110)=-0.68870044315018188969_t_real*ylm(132)
    rvylm(1,2,111)=-0.51075391845524915591_t_real*ylm(134)
    rvylm(2,2,111)=-0.51075391845524915591_t_real*ylm(132)
    rvylm(3,2,111)=-0.69156407480812462404_t_real*ylm(133)
    rvylm(1,2,112)= 0.46625240412015688279_t_real*ylm(133)&
       &-0.39261966026736167253_t_real*ylm(135)
    rvylm(2,2,112)=-0.39261966026736167253_t_real*ylm(131)
    rvylm(3,2,112)=-0.68870044315018188969_t_real*ylm(134)
    rvylm(1,2,113)= 0.29821603968282905209_t_real*ylm(134)&
       &-0.42407770188162638293_t_real*ylm(136)
    rvylm(2,2,113)=-0.42407770188162638293_t_real*ylm(130)&
      &-0.29821603968282905209_t_real*ylm(132)
    rvylm(3,2,113)=-0.68003719963350203039_t_real*ylm(135)
    rvylm(1,2,114)= 0.26673253468463223363_t_real*ylm(135)&
       &-0.45553252509883956821_t_real*ylm(137)
    rvylm(2,2,114)=-0.45553252509883956821_t_real*ylm(129)&
      &-0.26673253468463223363_t_real*ylm(131)
    rvylm(3,2,114)=-0.66534783913046048344_t_real*ylm(136)
    rvylm(1,2,115)= 0.23523598444848237187_t_real*ylm(136)&
       &-0.48698475355767397830_t_real*ylm(138)
    rvylm(2,2,115)=-0.48698475355767397830_t_real*ylm(128)&
      &-0.23523598444848237187_t_real*ylm(130)
    rvylm(3,2,115)=-0.64422027509684123750_t_real*ylm(137)
    rvylm(1,2,116)= 0.20372033841662687615_t_real*ylm(137)&
       &-0.51843485951208205086_t_real*ylm(139)
    rvylm(2,2,116)=-0.51843485951208205086_t_real*ylm(127)&
      &-0.20372033841662687615_t_real*ylm(129)
    rvylm(3,2,116)=-0.61599240280721444426_t_real*ylm(138)
    rvylm(1,2,117)= 0.17217511078754547242_t_real*ylm(138)&
       &-0.54988320714673246380_t_real*ylm(140)
    rvylm(2,2,117)=-0.54988320714673246380_t_real*ylm(126)&
      &-0.17217511078754547242_t_real*ylm(128)
    rvylm(3,2,117)=-0.57962779388728445979_t_real*ylm(139)
    rvylm(1,2,118)= 0.14058038927888331729_t_real*ylm(139)&
       &-0.58133008182238475428_t_real*ylm(141)
    rvylm(2,2,118)=-0.58133008182238475428_t_real*ylm(125)&
      &-0.14058038927888331729_t_real*ylm(127)
    rvylm(3,2,118)=-0.53346506936926446727_t_real*ylm(140)
    rvylm(1,2,119)= 0.10889310129609415676_t_real*ylm(140)&
       &-0.61277571031026112242_t_real*ylm(142)
    rvylm(2,2,119)=-0.61277571031026112242_t_real*ylm(124)&
      &-0.10889310129609415676_t_real*ylm(126)
    rvylm(3,2,119)=-0.47465402419841236983_t_real*ylm(141)
    rvylm(1,2,120)= 0.07699905035090180553_t_real*ylm(141)&
       &-0.64422027509684123750_t_real*ylm(143)
    rvylm(2,2,120)=-0.64422027509684123750_t_real*ylm(123)&
      &-0.07699905035090180553_t_real*ylm(125)
    rvylm(3,2,120)=-0.39762138624377206946_t_real*ylm(142)
    rvylm(1,2,121)= 0.04445542244743870560_t_real*ylm(142)&
       &-0.67566392469217622316_t_real*ylm(144)
    rvylm(2,2,121)=-0.67566392469217622316_t_real*ylm(122)&
      &-0.04445542244743870560_t_real*ylm(124)
    rvylm(3,2,121)=-0.28810406552003038415_t_real*ylm(143)
    !!
    If(maxL+1<=10) goto 1000
    !!
    !!..............................................
    !! L=11
    !!..............................................
    !! L-1
    rvylm(1,1,122)= 0.70710678118654752420_t_real*ylm(101)
    rvylm(2,1,122)= 0.70710678118654752420_t_real*ylm(121)
    rvylm(1,1,123)= 0.67419986246324208605_t_real*ylm(102)
    rvylm(2,1,123)= 0.67419986246324208605_t_real*ylm(120)
    rvylm(3,1,123)= 0.30151134457776362256_t_real*ylm(101)
    rvylm(1,1,124)=-0.04652421051992354336_t_real*ylm(101)&
       &+.64129198595913488553_t_real*ylm(103)
    rvylm(2,1,124)= 0.64129198595913488553_t_real*ylm(119)&
      &+.04652421051992354336_t_real*ylm(121)
    rvylm(3,1,124)= 0.41612518928823950740_t_real*ylm(102)
    rvylm(1,1,125)=-0.08058229640253802982_t_real*ylm(102)&
       &+.60838299625307587130_t_real*ylm(104)
    rvylm(2,1,125)= 0.60838299625307587130_t_real*ylm(118)&
      &+.08058229640253802982_t_real*ylm(120)
    rvylm(3,1,125)= 0.49674263633520200689_t_real*ylm(103)
    rvylm(1,1,126)=-0.11396057645963794880_t_real*ylm(103)&
       &+.57547270236635131696_t_real*ylm(105)
    rvylm(2,1,126)= 0.57547270236635131696_t_real*ylm(117)&
      &+.11396057645963794880_t_real*ylm(119)
    rvylm(3,1,126)= 0.55829052623908252034_t_real*ylm(104)
    rvylm(1,1,127)=-0.14712247158412492157_t_real*ylm(104)&
       &+.54256086697465969392_t_real*ylm(106)
    rvylm(2,1,127)= 0.54256086697465969392_t_real*ylm(116)&
      &+.14712247158412492157_t_real*ylm(118)
    rvylm(3,1,127)= 0.60660149024327987152_t_real*ylm(105)
    rvylm(1,1,128)=-0.18018749253911179146_t_real*ylm(105)&
       &+.50964719143762554449_t_real*ylm(107)
    rvylm(2,1,128)= 0.50964719143762554449_t_real*ylm(115)&
      &+.18018749253911179146_t_real*ylm(117)
    rvylm(3,1,128)= 0.64465837122030423857_t_real*ylm(106)
    rvylm(1,1,129)=-0.21320071635561043423_t_real*ylm(106)&
       &+.47673129462279615758_t_real*ylm(108)
    rvylm(2,1,129)= 0.47673129462279615758_t_real*ylm(114)&
      &+.21320071635561043423_t_real*ylm(116)
    rvylm(3,1,129)= 0.67419986246324208605_t_real*ylm(107)
    rvylm(1,1,130)=-0.24618298195866546539_t_real*ylm(107)&
       &+.44381268229929726692_t_real*ylm(109)
    rvylm(2,1,130)= 0.44381268229929726692_t_real*ylm(113)&
      &+.24618298195866546539_t_real*ylm(115)
    rvylm(3,1,130)= 0.69631062382279135251_t_real*ylm(108)
    rvylm(1,1,131)=-0.27914526311954126017_t_real*ylm(108)&
       &+.41089070180665907966_t_real*ylm(110)
    rvylm(2,1,131)= 0.41089070180665907966_t_real*ylm(112)&
      &+.27914526311954126017_t_real*ylm(114)
    rvylm(3,1,131)= 0.71168357188676660728_t_real*ylm(109)
    rvylm(1,1,132)=-0.31209389196617963055_t_real*ylm(109)
    rvylm(2,1,132)= 0.53452248382484876920_t_real*ylm(111)&
      &+.31209389196617963055_t_real*ylm(113)
    rvylm(3,1,132)= 0.72074997015644716586_t_real*ylm(110)
    rvylm(1,1,133)=-0.48795003647426658953_t_real*ylm(112)
    rvylm(2,1,133)=-0.48795003647426658953_t_real*ylm(110)
    rvylm(3,1,133)= 0.72374686445574588702_t_real*ylm(111)
    rvylm(1,1,134)= 0.53452248382484876920_t_real*ylm(111)&
       &-0.31209389196617963055_t_real*ylm(113)
    rvylm(2,1,134)=-0.31209389196617963055_t_real*ylm(109)
    rvylm(3,1,134)= 0.72074997015644716586_t_real*ylm(112)
    rvylm(1,1,135)= 0.41089070180665907966_t_real*ylm(112)&
       &-0.27914526311954126017_t_real*ylm(114)
    rvylm(2,1,135)=-0.27914526311954126017_t_real*ylm(108)&
      &-0.41089070180665907966_t_real*ylm(110)
    rvylm(3,1,135)= 0.71168357188676660728_t_real*ylm(113)
    rvylm(1,1,136)= 0.44381268229929726692_t_real*ylm(113)&
       &-0.24618298195866546539_t_real*ylm(115)
    rvylm(2,1,136)=-0.24618298195866546539_t_real*ylm(107)&
      &-0.44381268229929726692_t_real*ylm(109)
    rvylm(3,1,136)= 0.69631062382279135251_t_real*ylm(114)
    rvylm(1,1,137)= 0.47673129462279615758_t_real*ylm(114)&
       &-0.21320071635561043423_t_real*ylm(116)
    rvylm(2,1,137)=-0.21320071635561043423_t_real*ylm(106)&
      &-0.47673129462279615758_t_real*ylm(108)
    rvylm(3,1,137)= 0.67419986246324208605_t_real*ylm(115)
    rvylm(1,1,138)= 0.50964719143762554449_t_real*ylm(115)&
       &-0.18018749253911179146_t_real*ylm(117)
    rvylm(2,1,138)=-0.18018749253911179146_t_real*ylm(105)&
      &-0.50964719143762554449_t_real*ylm(107)
    rvylm(3,1,138)= 0.64465837122030423857_t_real*ylm(116)
    rvylm(1,1,139)= 0.54256086697465969392_t_real*ylm(116)&
       &-0.14712247158412492157_t_real*ylm(118)
    rvylm(2,1,139)=-0.14712247158412492157_t_real*ylm(104)&
      &-0.54256086697465969392_t_real*ylm(106)
    rvylm(3,1,139)= 0.60660149024327987152_t_real*ylm(117)
    rvylm(1,1,140)= 0.57547270236635131696_t_real*ylm(117)&
       &-0.11396057645963794880_t_real*ylm(119)
    rvylm(2,1,140)=-0.11396057645963794880_t_real*ylm(103)&
      &-0.57547270236635131696_t_real*ylm(105)
    rvylm(3,1,140)= 0.55829052623908252034_t_real*ylm(118)
    rvylm(1,1,141)= 0.60838299625307587130_t_real*ylm(118)&
       &-0.08058229640253802982_t_real*ylm(120)
    rvylm(2,1,141)=-0.08058229640253802982_t_real*ylm(102)&
      &-0.60838299625307587130_t_real*ylm(104)
    rvylm(3,1,141)= 0.49674263633520200689_t_real*ylm(119)
    rvylm(1,1,142)= 0.64129198595913488553_t_real*ylm(119)&
       &-0.04652421051992354336_t_real*ylm(121)
    rvylm(2,1,142)=-0.04652421051992354336_t_real*ylm(101)&
      &-0.64129198595913488553_t_real*ylm(103)
    rvylm(3,1,142)= 0.41612518928823950740_t_real*ylm(120)
    rvylm(1,1,143)= 0.67419986246324208605_t_real*ylm(120)
    rvylm(2,1,143)=-0.67419986246324208605_t_real*ylm(102)
    rvylm(3,1,143)= 0.30151134457776362256_t_real*ylm(121)
    rvylm(1,1,144)= 0.70710678118654752420_t_real*ylm(121)
    rvylm(2,1,144)=-0.70710678118654752420_t_real*ylm(101)
    !! L+1
    rvylm(1,2,122)=-0.67823299831252681377_t_real*ylm(145)&
       &+.04082482904638630162_t_real*ylm(147)
    rvylm(2,2,122)= 0.04082482904638630162_t_real*ylm(167)&
      &+.67823299831252681377_t_real*ylm(169)
    rvylm(3,2,122)=-0.27688746209726916170_t_real*ylm(146)
    rvylm(1,2,123)=-0.64935865795927189324_t_real*ylm(146)&
       &+.07071067811865475242_t_real*ylm(148)
    rvylm(2,2,123)= 0.07071067811865475242_t_real*ylm(166)&
      &+.64935865795927189324_t_real*ylm(168)
    rvylm(3,2,123)=-0.38297084310253524391_t_real*ylm(147)
    rvylm(1,2,124)=-0.62048368229954282968_t_real*ylm(147)&
       &+.09999999999999999998_t_real*ylm(149)
    rvylm(2,2,124)= 0.09999999999999999998_t_real*ylm(165)&
      &+.62048368229954282968_t_real*ylm(167)
    rvylm(3,2,124)=-0.45825756949558400057_t_real*ylm(148)
    rvylm(1,2,125)=-0.59160797830996160414_t_real*ylm(148)&
       &+.12909944487358056281_t_real*ylm(150)
    rvylm(2,2,125)= 0.12909944487358056281_t_real*ylm(164)&
      &+.59160797830996160414_t_real*ylm(166)
    rvylm(3,2,125)=-0.51639777949432225125_t_real*ylm(149)
    rvylm(1,2,126)=-0.56273143387113773933_t_real*ylm(149)&
       &+.15811388300841896656_t_real*ylm(151)
    rvylm(2,2,126)= 0.15811388300841896656_t_real*ylm(163)&
      &+.56273143387113773933_t_real*ylm(165)
    rvylm(3,2,126)=-0.56273143387113773933_t_real*ylm(150)
    rvylm(1,2,127)=-0.53385391260156556043_t_real*ylm(150)&
       &+.18708286933869706924_t_real*ylm(152)
    rvylm(2,2,127)= 0.18708286933869706924_t_real*ylm(162)&
      &+.53385391260156556043_t_real*ylm(164)
    rvylm(3,2,127)=-0.59999999999999999988_t_real*ylm(151)
    rvylm(1,2,128)=-0.50497524691810389757_t_real*ylm(151)&
       &+.21602468994692867432_t_real*ylm(153)
    rvylm(2,2,128)= 0.21602468994692867432_t_real*ylm(161)&
      &+.50497524691810389757_t_real*ylm(163)
    rvylm(3,2,128)=-0.62981478758970614258_t_real*ylm(152)
    rvylm(1,2,129)=-0.47609522856952333310_t_real*ylm(152)&
       &+.24494897427831780977_t_real*ylm(154)
    rvylm(2,2,129)= 0.24494897427831780977_t_real*ylm(160)&
      &+.47609522856952333310_t_real*ylm(162)
    rvylm(3,2,129)=-0.65319726474218082606_t_real*ylm(153)
    rvylm(1,2,130)=-0.44721359549995793919_t_real*ylm(153)&
       &+.27386127875258305667_t_real*ylm(155)
    rvylm(2,2,130)= 0.27386127875258305667_t_real*ylm(159)&
      &+.44721359549995793919_t_real*ylm(161)
    rvylm(3,2,130)=-0.67082039324993690879_t_real*ylm(154)
    rvylm(1,2,131)=-0.41833001326703777390_t_real*ylm(154)&
       &+.30276503540974916648_t_real*ylm(156)
    rvylm(2,2,131)= 0.30276503540974916648_t_real*ylm(158)&
      &+.41833001326703777390_t_real*ylm(160)
    rvylm(3,2,131)=-0.68313005106397322541_t_real*ylm(155)
    rvylm(1,2,132)=-0.38944404818493075361_t_real*ylm(155)
    rvylm(2,2,132)= 0.46904157598234295535_t_real*ylm(157)&
      &+.38944404818493075361_t_real*ylm(159)
    rvylm(3,2,132)=-0.69041050590693263760_t_real*ylm(156)
    rvylm(1,2,133)=-0.50990195135927848290_t_real*ylm(158)
    rvylm(2,2,133)=-0.50990195135927848290_t_real*ylm(156)
    rvylm(3,2,133)=-0.69282032302755091728_t_real*ylm(157)
    rvylm(1,2,134)= 0.46904157598234295535_t_real*ylm(157)&
       &-0.38944404818493075361_t_real*ylm(159)
    rvylm(2,2,134)=-0.38944404818493075361_t_real*ylm(155)
    rvylm(3,2,134)=-0.69041050590693263760_t_real*ylm(158)
    rvylm(1,2,135)= 0.30276503540974916648_t_real*ylm(158)&
       &-0.41833001326703777390_t_real*ylm(160)
    rvylm(2,2,135)=-0.41833001326703777390_t_real*ylm(154)&
      &-0.30276503540974916648_t_real*ylm(156)
    rvylm(3,2,135)=-0.68313005106397322541_t_real*ylm(159)
    rvylm(1,2,136)= 0.27386127875258305667_t_real*ylm(159)&
       &-0.44721359549995793919_t_real*ylm(161)
    rvylm(2,2,136)=-0.44721359549995793919_t_real*ylm(153)&
      &-0.27386127875258305667_t_real*ylm(155)
    rvylm(3,2,136)=-0.67082039324993690879_t_real*ylm(160)
    rvylm(1,2,137)= 0.24494897427831780977_t_real*ylm(160)&
       &-0.47609522856952333310_t_real*ylm(162)
    rvylm(2,2,137)=-0.47609522856952333310_t_real*ylm(152)&
      &-0.24494897427831780977_t_real*ylm(154)
    rvylm(3,2,137)=-0.65319726474218082606_t_real*ylm(161)
    rvylm(1,2,138)= 0.21602468994692867432_t_real*ylm(161)&
       &-0.50497524691810389757_t_real*ylm(163)
    rvylm(2,2,138)=-0.50497524691810389757_t_real*ylm(151)&
      &-0.21602468994692867432_t_real*ylm(153)
    rvylm(3,2,138)=-0.62981478758970614258_t_real*ylm(162)
    rvylm(1,2,139)= 0.18708286933869706924_t_real*ylm(162)&
       &-0.53385391260156556043_t_real*ylm(164)
    rvylm(2,2,139)=-0.53385391260156556043_t_real*ylm(150)&
      &-0.18708286933869706924_t_real*ylm(152)
    rvylm(3,2,139)=-0.59999999999999999988_t_real*ylm(163)
    rvylm(1,2,140)= 0.15811388300841896656_t_real*ylm(163)&
       &-0.56273143387113773933_t_real*ylm(165)
    rvylm(2,2,140)=-0.56273143387113773933_t_real*ylm(149)&
      &-0.15811388300841896656_t_real*ylm(151)
    rvylm(3,2,140)=-0.56273143387113773933_t_real*ylm(164)
    rvylm(1,2,141)= 0.12909944487358056281_t_real*ylm(164)&
       &-0.59160797830996160414_t_real*ylm(166)
    rvylm(2,2,141)=-0.59160797830996160414_t_real*ylm(148)&
      &-0.12909944487358056281_t_real*ylm(150)
    rvylm(3,2,141)=-0.51639777949432225125_t_real*ylm(165)
    rvylm(1,2,142)= 0.09999999999999999998_t_real*ylm(165)&
       &-0.62048368229954282968_t_real*ylm(167)
    rvylm(2,2,142)=-0.62048368229954282968_t_real*ylm(147)&
      &-0.09999999999999999998_t_real*ylm(149)
    rvylm(3,2,142)=-0.45825756949558400057_t_real*ylm(166)
    rvylm(1,2,143)= 0.07071067811865475242_t_real*ylm(166)&
       &-0.64935865795927189324_t_real*ylm(168)
    rvylm(2,2,143)=-0.64935865795927189324_t_real*ylm(146)&
      &-0.07071067811865475242_t_real*ylm(148)
    rvylm(3,2,143)=-0.38297084310253524391_t_real*ylm(167)
    rvylm(1,2,144)= 0.04082482904638630162_t_real*ylm(167)&
       &-0.67823299831252681377_t_real*ylm(169)
    rvylm(2,2,144)=-0.67823299831252681377_t_real*ylm(145)&
      &-0.04082482904638630162_t_real*ylm(147)
    rvylm(3,2,144)=-0.27688746209726916170_t_real*ylm(168)
    !!
    If(maxL+1<=11) goto 1000
    !!
    !!..............................................
    !! L=12
    !!..............................................
    !! L-1
    rvylm(1,1,145)= 0.70710678118654752438_t_real*ylm(122)
    rvylm(2,1,145)= 0.70710678118654752438_t_real*ylm(144)
    rvylm(1,1,146)= 0.67700320038633003002_t_real*ylm(123)
    rvylm(2,1,146)= 0.67700320038633003002_t_real*ylm(143)
    rvylm(3,1,146)= 0.28867513459481288224_t_real*ylm(122)
    rvylm(1,1,147)=-0.04256282653793742966_t_real*ylm(122)&
       &+.64689895723332652267_t_real*ylm(124)
    rvylm(2,1,147)= 0.64689895723332652267_t_real*ylm(142)&
      &+.04256282653793742966_t_real*ylm(144)
    rvylm(3,1,147)= 0.39927470475234524148_t_real*ylm(123)
    rvylm(1,1,148)=-0.07372097807744856672_t_real*ylm(123)&
       &+.61679395474396119616_t_real*ylm(125)
    rvylm(2,1,148)= 0.61679395474396119616_t_real*ylm(141)&
      &+.07372097807744856672_t_real*ylm(143)
    rvylm(3,1,148)= 0.47776654295295457753_t_real*ylm(124)
    rvylm(1,1,149)=-0.10425720702853738133_t_real*ylm(124)&
       &+.58668807602568900144_t_real*ylm(126)
    rvylm(2,1,149)= 0.58668807602568900144_t_real*ylm(140)&
      &+.10425720702853738133_t_real*ylm(142)
    rvylm(3,1,149)= 0.53838190205816550639_t_real*ylm(125)
    rvylm(1,1,150)=-0.13459547551454137659_t_real*ylm(125)&
       &+.55658117889096121851_t_real*ylm(127)
    rvylm(2,1,150)= 0.55658117889096121851_t_real*ylm(139)&
      &+.13459547551454137659_t_real*ylm(141)
    rvylm(3,1,150)= 0.58668807602568900144_t_real*ylm(126)
    rvylm(1,1,151)=-0.16484511834894675117_t_real*ylm(126)&
       &+.52647308862227541293_t_real*ylm(128)
    rvylm(2,1,151)= 0.52647308862227541293_t_real*ylm(138)&
      &+.16484511834894675117_t_real*ylm(140)
    rvylm(3,1,151)= 0.62554324217122428802_t_real*ylm(127)
    rvylm(1,1,152)=-0.19504737440137348646_t_real*ylm(127)&
       &+.49636358810271619131_t_real*ylm(129)
    rvylm(2,1,152)= 0.49636358810271619131_t_real*ylm(137)&
      &+.19504737440137348646_t_real*ylm(139)
    rvylm(3,1,152)= 0.65662730699374289154_t_real*ylm(128)
    rvylm(1,1,153)=-0.22522130823072540772_t_real*ylm(128)&
       &+.46625240412015688280_t_real*ylm(130)
    rvylm(2,1,153)= 0.46625240412015688280_t_real*ylm(136)&
      &+.22522130823072540772_t_real*ylm(138)
    rvylm(3,1,153)= 0.68100522460699887459_t_real*ylm(129)
    rvylm(1,1,154)=-0.25537695922762457797_t_real*ylm(129)&
       &+.43613918799432346592_t_real*ylm(131)
    rvylm(2,1,154)= 0.43613918799432346592_t_real*ylm(135)&
      &+.25537695922762457797_t_real*ylm(137)
    rvylm(3,1,154)= 0.69937860618023532421_t_real*ylm(130)
    rvylm(1,1,155)=-0.28552012036008037275_t_real*ylm(130)&
       &+.40602348757648013180_t_real*ylm(132)
    rvylm(2,1,155)= 0.40602348757648013180_t_real*ylm(134)&
      &+.28552012036008037275_t_real*ylm(136)
    rvylm(3,1,155)= 0.71221231161191969588_t_real*ylm(131)
    rvylm(1,1,156)=-0.31565436977716669945_t_real*ylm(131)
    rvylm(2,1,156)= 0.53160953307119494602_t_real*ylm(133)&
      &+.31565436977716669945_t_real*ylm(135)
    rvylm(3,1,156)= 0.71980271049016306634_t_real*ylm(132)
    rvylm(1,1,157)=-0.48900964692182576168_t_real*ylm(134)
    rvylm(2,1,157)=-0.48900964692182576168_t_real*ylm(132)
    rvylm(3,1,157)= 0.72231511851461520456_t_real*ylm(133)
    rvylm(1,1,158)= 0.53160953307119494602_t_real*ylm(133)&
       &-0.31565436977716669945_t_real*ylm(135)
    rvylm(2,1,158)=-0.31565436977716669945_t_real*ylm(131)
    rvylm(3,1,158)= 0.71980271049016306634_t_real*ylm(134)
    rvylm(1,1,159)= 0.40602348757648013180_t_real*ylm(134)&
       &-0.28552012036008037275_t_real*ylm(136)
    rvylm(2,1,159)=-0.28552012036008037275_t_real*ylm(130)&
      &-0.40602348757648013180_t_real*ylm(132)
    rvylm(3,1,159)= 0.71221231161191969588_t_real*ylm(135)
    rvylm(1,1,160)= 0.43613918799432346592_t_real*ylm(135)&
       &-0.25537695922762457797_t_real*ylm(137)
    rvylm(2,1,160)=-0.25537695922762457797_t_real*ylm(129)&
      &-0.43613918799432346592_t_real*ylm(131)
    rvylm(3,1,160)= 0.69937860618023532421_t_real*ylm(136)
    rvylm(1,1,161)= 0.46625240412015688280_t_real*ylm(136)&
       &-0.22522130823072540772_t_real*ylm(138)
    rvylm(2,1,161)=-0.22522130823072540772_t_real*ylm(128)&
      &-0.46625240412015688280_t_real*ylm(130)
    rvylm(3,1,161)= 0.68100522460699887459_t_real*ylm(137)
    rvylm(1,1,162)= 0.49636358810271619131_t_real*ylm(137)&
       &-0.19504737440137348646_t_real*ylm(139)
    rvylm(2,1,162)=-0.19504737440137348646_t_real*ylm(127)&
      &-0.49636358810271619131_t_real*ylm(129)
    rvylm(3,1,162)= 0.65662730699374289154_t_real*ylm(138)
    rvylm(1,1,163)= 0.52647308862227541293_t_real*ylm(138)&
       &-0.16484511834894675117_t_real*ylm(140)
    rvylm(2,1,163)=-0.16484511834894675117_t_real*ylm(126)&
      &-0.52647308862227541293_t_real*ylm(128)
    rvylm(3,1,163)= 0.62554324217122428802_t_real*ylm(139)
    rvylm(1,1,164)= 0.55658117889096121851_t_real*ylm(139)&
       &-0.13459547551454137659_t_real*ylm(141)
    rvylm(2,1,164)=-0.13459547551454137659_t_real*ylm(125)&
      &-0.55658117889096121851_t_real*ylm(127)
    rvylm(3,1,164)= 0.58668807602568900144_t_real*ylm(140)
    rvylm(1,1,165)= 0.58668807602568900144_t_real*ylm(140)&
       &-0.10425720702853738133_t_real*ylm(142)
    rvylm(2,1,165)=-0.10425720702853738133_t_real*ylm(124)&
      &-0.58668807602568900144_t_real*ylm(126)
    rvylm(3,1,165)= 0.53838190205816550639_t_real*ylm(141)
    rvylm(1,1,166)= 0.61679395474396119616_t_real*ylm(141)&
       &-0.07372097807744856672_t_real*ylm(143)
    rvylm(2,1,166)=-0.07372097807744856672_t_real*ylm(123)&
      &-0.61679395474396119616_t_real*ylm(125)
    rvylm(3,1,166)= 0.47776654295295457753_t_real*ylm(142)
    rvylm(1,1,167)= 0.64689895723332652267_t_real*ylm(142)&
       &-0.04256282653793742966_t_real*ylm(144)
    rvylm(2,1,167)=-0.04256282653793742966_t_real*ylm(122)&
      &-0.64689895723332652267_t_real*ylm(124)
    rvylm(3,1,167)= 0.39927470475234524148_t_real*ylm(143)
    rvylm(1,1,168)= 0.67700320038633003002_t_real*ylm(143)
    rvylm(2,1,168)=-0.67700320038633003002_t_real*ylm(123)
    rvylm(3,1,168)= 0.28867513459481288224_t_real*ylm(144)
    rvylm(1,1,169)= 0.70710678118654752438_t_real*ylm(144)
    rvylm(2,1,169)=-0.70710678118654752438_t_real*ylm(122)
    !! L+1
    rvylm(1,2,145)=-0.68041381743977169384_t_real*ylm(170)&
       &+.03774256780481985834_t_real*ylm(172)
    rvylm(2,2,145)= 0.03774256780481985834_t_real*ylm(194)&
      &+.68041381743977169384_t_real*ylm(196)
    rvylm(3,2,145)=-0.26688025634181188910_t_real*ylm(171)
    rvylm(1,2,146)=-0.65372045046061343965_t_real*ylm(171)&
       &+.06537204504606134396_t_real*ylm(173)
    rvylm(2,2,146)= 0.06537204504606134396_t_real*ylm(193)&
      &+.65372045046061343965_t_real*ylm(195)
    rvylm(3,2,146)=-0.36980013081681941462_t_real*ylm(172)
    rvylm(1,2,147)=-0.62702662875064019268_t_real*ylm(172)&
       &+.09245003270420485365_t_real*ylm(174)
    rvylm(2,2,147)= 0.09245003270420485365_t_real*ylm(192)&
      &+.62702662875064019268_t_real*ylm(194)
    rvylm(3,2,147)=-0.44337478117411750366_t_real*ylm(173)
    rvylm(1,2,148)=-0.60033229165093259936_t_real*ylm(173)&
       &+.11935247900657214673_t_real*ylm(175)
    rvylm(2,2,148)= 0.11935247900657214673_t_real*ylm(191)&
      &+.60033229165093259936_t_real*ylm(193)
    rvylm(3,2,148)=-0.50071174413254050633_t_real*ylm(174)
    rvylm(1,2,149)=-0.57363736721018188890_t_real*ylm(174)&
       &+.14617633655117153576_t_real*ylm(176)
    rvylm(2,2,149)= 0.14617633655117153576_t_real*ylm(190)&
      &+.57363736721018188890_t_real*ylm(192)
    rvylm(3,2,149)=-0.54694176942824466019_t_real*ylm(175)
    rvylm(1,2,150)=-0.54694176942824466019_t_real*ylm(175)&
       &+.17295817388759027517_t_real*ylm(177)
    rvylm(2,2,150)= 0.17295817388759027517_t_real*ylm(189)&
      &+.54694176942824466019_t_real*ylm(191)
    rvylm(3,2,150)=-0.58470534620468614306_t_real*ylm(176)
    rvylm(1,2,151)=-0.52024539465070774924_t_real*ylm(176)&
       &+.19971489650509269417_t_real*ylm(178)
    rvylm(2,2,151)= 0.19971489650509269417_t_real*ylm(188)&
      &+.52024539465070774924_t_real*ylm(190)
    rvylm(3,2,151)=-0.61556265230874664986_t_real*ylm(177)
    rvylm(1,2,152)=-0.49354811679282455719_t_real*ylm(177)&
       &+.22645540682891915008_t_real*ylm(179)
    rvylm(2,2,152)= 0.22645540682891915008_t_real*ylm(187)&
      &+.49354811679282455719_t_real*ylm(189)
    rvylm(3,2,152)=-0.64051261522034853384_t_real*ylm(178)
    rvylm(1,2,153)=-0.46684978092392626574_t_real*ylm(178)&
       &+.25318484177091665403_t_real*ylm(180)
    rvylm(2,2,153)= 0.25318484177091665403_t_real*ylm(186)&
      &+.46684978092392626574_t_real*ylm(188)
    rvylm(3,2,153)=-0.66022529177352475710_t_real*ylm(179)
    rvylm(1,2,154)=-0.44015019451568317140_t_real*ylm(179)&
       &+.27990637425321051011_t_real*ylm(181)
    rvylm(2,2,154)= 0.27990637425321051011_t_real*ylm(185)&
      &+.44015019451568317140_t_real*ylm(187)
    rvylm(3,2,154)=-0.67515957805577774408_t_real*ylm(180)
    rvylm(1,2,155)=-0.41344911529736152057_t_real*ylm(180)&
       &+.30662207033593328254_t_real*ylm(182)
    rvylm(2,2,155)= 0.30662207033593328254_t_real*ylm(184)&
      &+.41344911529736152057_t_real*ylm(186)
    rvylm(3,2,155)=-0.68562779267286859696_t_real*ylm(181)
    rvylm(1,2,156)=-0.38674623407688092523_t_real*ylm(181)
    rvylm(2,2,156)= 0.47140452079103168285_t_real*ylm(183)&
      &+.38674623407688092523_t_real*ylm(185)
    rvylm(3,2,156)=-0.69183269555036110069_t_real*ylm(182)
    rvylm(1,2,157)=-0.50917507721731555620_t_real*ylm(184)
    rvylm(2,2,157)=-0.50917507721731555620_t_real*ylm(182)
    rvylm(3,2,157)=-0.69388866648871091166_t_real*ylm(183)
    rvylm(1,2,158)= 0.47140452079103168285_t_real*ylm(183)&
       &-0.38674623407688092523_t_real*ylm(185)
    rvylm(2,2,158)=-0.38674623407688092523_t_real*ylm(181)
    rvylm(3,2,158)=-0.69183269555036110069_t_real*ylm(184)
    rvylm(1,2,159)= 0.30662207033593328254_t_real*ylm(184)&
       &-0.41344911529736152057_t_real*ylm(186)
    rvylm(2,2,159)=-0.41344911529736152057_t_real*ylm(180)&
      &-0.30662207033593328254_t_real*ylm(182)
    rvylm(3,2,159)=-0.68562779267286859696_t_real*ylm(185)
    rvylm(1,2,160)= 0.27990637425321051011_t_real*ylm(185)&
       &-0.44015019451568317140_t_real*ylm(187)
    rvylm(2,2,160)=-0.44015019451568317140_t_real*ylm(179)&
      &-0.27990637425321051011_t_real*ylm(181)
    rvylm(3,2,160)=-0.67515957805577774408_t_real*ylm(186)
    rvylm(1,2,161)= 0.25318484177091665403_t_real*ylm(186)&
       &-0.46684978092392626574_t_real*ylm(188)
    rvylm(2,2,161)=-0.46684978092392626574_t_real*ylm(178)&
      &-0.25318484177091665403_t_real*ylm(180)
    rvylm(3,2,161)=-0.66022529177352475710_t_real*ylm(187)
    rvylm(1,2,162)= 0.22645540682891915008_t_real*ylm(187)&
       &-0.49354811679282455719_t_real*ylm(189)
    rvylm(2,2,162)=-0.49354811679282455719_t_real*ylm(177)&
      &-0.22645540682891915008_t_real*ylm(179)
    rvylm(3,2,162)=-0.64051261522034853384_t_real*ylm(188)
    rvylm(1,2,163)= 0.19971489650509269417_t_real*ylm(188)&
       &-0.52024539465070774924_t_real*ylm(190)
    rvylm(2,2,163)=-0.52024539465070774924_t_real*ylm(176)&
      &-0.19971489650509269417_t_real*ylm(178)
    rvylm(3,2,163)=-0.61556265230874664986_t_real*ylm(189)
    rvylm(1,2,164)= 0.17295817388759027517_t_real*ylm(189)&
       &-0.54694176942824466019_t_real*ylm(191)
    rvylm(2,2,164)=-0.54694176942824466019_t_real*ylm(175)&
      &-0.17295817388759027517_t_real*ylm(177)
    rvylm(3,2,164)=-0.58470534620468614306_t_real*ylm(190)
    rvylm(1,2,165)= 0.14617633655117153576_t_real*ylm(190)&
       &-0.57363736721018188890_t_real*ylm(192)
    rvylm(2,2,165)=-0.57363736721018188890_t_real*ylm(174)&
      &-0.14617633655117153576_t_real*ylm(176)
    rvylm(3,2,165)=-0.54694176942824466019_t_real*ylm(191)
    rvylm(1,2,166)= 0.11935247900657214673_t_real*ylm(191)&
       &-0.60033229165093259936_t_real*ylm(193)
    rvylm(2,2,166)=-0.60033229165093259936_t_real*ylm(173)&
      &-0.11935247900657214673_t_real*ylm(175)
    rvylm(3,2,166)=-0.50071174413254050633_t_real*ylm(192)
    rvylm(1,2,167)= 0.09245003270420485365_t_real*ylm(192)&
       &-0.62702662875064019268_t_real*ylm(194)
    rvylm(2,2,167)=-0.62702662875064019268_t_real*ylm(172)&
      &-0.09245003270420485365_t_real*ylm(174)
    rvylm(3,2,167)=-0.44337478117411750366_t_real*ylm(193)
    rvylm(1,2,168)= 0.06537204504606134396_t_real*ylm(193)&
       &-0.65372045046061343965_t_real*ylm(195)
    rvylm(2,2,168)=-0.65372045046061343965_t_real*ylm(171)&
      &-0.06537204504606134396_t_real*ylm(173)
    rvylm(3,2,168)=-0.36980013081681941462_t_real*ylm(194)
    rvylm(1,2,169)= 0.03774256780481985834_t_real*ylm(194)&
       &-0.68041381743977169384_t_real*ylm(196)
    rvylm(2,2,169)=-0.68041381743977169384_t_real*ylm(170)&
      &-0.03774256780481985834_t_real*ylm(172)
    rvylm(3,2,169)=-0.26688025634181188910_t_real*ylm(195)
    !!
    If(maxL+1<=12) goto 1000
    !!
    !!..............................................
    !! L=13
    !!..............................................
    !! L-1
    rvylm(1,1,170)= 0.70710678118654752437_t_real*ylm(145)
    rvylm(2,1,170)= 0.70710678118654752437_t_real*ylm(169)
    rvylm(1,1,171)= 0.67936622048675745034_t_real*ylm(146)
    rvylm(2,1,171)= 0.67936622048675745034_t_real*ylm(168)
    rvylm(3,1,171)= 0.27735009811261456100_t_real*ylm(145)
    rvylm(1,1,172)=-0.03922322702763680638_t_real*ylm(145)&
       &+.65162518721684217549_t_real*ylm(147)
    rvylm(2,1,172)= 0.65162518721684217549_t_real*ylm(167)&
      &+.03922322702763680638_t_real*ylm(169)
    rvylm(3,1,172)= 0.38430756913220912034_t_real*ylm(146)
    rvylm(1,1,173)=-0.06793662204867574503_t_real*ylm(146)&
       &+.62388361833820354826_t_real*ylm(148)
    rvylm(2,1,173)= 0.62388361833820354826_t_real*ylm(166)&
      &+.06793662204867574503_t_real*ylm(168)
    rvylm(3,1,173)= 0.46076858867298268520_t_real*ylm(147)
    rvylm(1,1,174)=-0.09607689228305228008_t_real*ylm(147)&
       &+.59614143907684809140_t_real*ylm(149)
    rvylm(2,1,174)= 0.59614143907684809140_t_real*ylm(165)&
      &+.09607689228305228008_t_real*ylm(167)
    rvylm(3,1,174)= 0.52035490847039270454_t_real*ylm(148)
    rvylm(1,1,175)=-0.12403473458920845618_t_real*ylm(148)&
       &+.56839856005880510754_t_real*ylm(150)
    rvylm(2,1,175)= 0.56839856005880510754_t_real*ylm(164)&
      &+.12403473458920845618_t_real*ylm(166)
    rvylm(3,1,175)= 0.56839856005880510754_t_real*ylm(149)
    rvylm(1,1,176)=-0.15191090506254999243_t_real*ylm(149)&
       &+.54065487356324857942_t_real*ylm(151)
    rvylm(2,1,176)= 0.54065487356324857942_t_real*ylm(163)&
      &+.15191090506254999243_t_real*ylm(165)
    rvylm(3,1,176)= 0.60764362025019996974_t_real*ylm(150)
    rvylm(1,1,177)=-0.17974340685458342477_t_real*ylm(150)&
       &+.51291024855906620586_t_real*ylm(152)
    rvylm(2,1,177)= 0.51291024855906620586_t_real*ylm(162)&
      &+.17974340685458342477_t_real*ylm(164)
    rvylm(3,1,177)= 0.63971147342436279829_t_real*ylm(151)
    rvylm(1,1,178)=-0.20754980866510833023_t_real*ylm(151)&
       &+.48516452403758396048_t_real*ylm(153)
    rvylm(2,1,178)= 0.48516452403758396048_t_real*ylm(161)&
      &+.20754980866510833023_t_real*ylm(163)
    rvylm(3,1,178)= 0.66564023547027494640_t_real*ylm(152)
    rvylm(1,1,179)=-0.23533936216582083830_t_real*ylm(152)&
       &+.45741749991749247760_t_real*ylm(154)
    rvylm(2,1,179)= 0.45741749991749247760_t_real*ylm(160)&
      &+.23533936216582083830_t_real*ylm(162)
    rvylm(3,1,179)= 0.68612624987623871640_t_real*ylm(153)
    rvylm(1,1,180)=-0.26311740579210876440_t_real*ylm(153)&
       &+.42966892442365972828_t_real*ylm(155)
    rvylm(2,1,180)= 0.42966892442365972828_t_real*ylm(159)&
      &+.26311740579210876440_t_real*ylm(161)
    rvylm(3,1,180)= 0.70164641544562337175_t_real*ylm(154)
    rvylm(1,1,181)=-0.29088723694136980024_t_real*ylm(154)&
       &+.40191847623425019455_t_real*ylm(156)
    rvylm(2,1,181)= 0.40191847623425019455_t_real*ylm(158)&
      &+.29088723694136980024_t_real*ylm(160)
    rvylm(3,1,181)= 0.71252530319442529414_t_real*ylm(155)
    rvylm(1,1,182)=-0.31865100272627660202_t_real*ylm(155)
    rvylm(2,1,182)= 0.52915026221291811807_t_real*ylm(157)&
      &+.31865100272627660202_t_real*ylm(159)
    rvylm(3,1,182)= 0.71897362741833369909_t_real*ylm(156)
    rvylm(1,1,183)=-0.48989794855663561961_t_real*ylm(158)
    rvylm(2,1,183)=-0.48989794855663561961_t_real*ylm(156)
    rvylm(3,1,183)= 0.72111025509279785860_t_real*ylm(157)
    rvylm(1,1,184)= 0.52915026221291811807_t_real*ylm(157)&
       &-0.31865100272627660202_t_real*ylm(159)
    rvylm(2,1,184)=-0.31865100272627660202_t_real*ylm(155)
    rvylm(3,1,184)= 0.71897362741833369909_t_real*ylm(158)
    rvylm(1,1,185)= 0.40191847623425019455_t_real*ylm(158)&
       &-0.29088723694136980024_t_real*ylm(160)
    rvylm(2,1,185)=-0.29088723694136980024_t_real*ylm(154)&
      &-0.40191847623425019455_t_real*ylm(156)
    rvylm(3,1,185)= 0.71252530319442529414_t_real*ylm(159)
    rvylm(1,1,186)= 0.42966892442365972828_t_real*ylm(159)&
       &-0.26311740579210876440_t_real*ylm(161)
    rvylm(2,1,186)=-0.26311740579210876440_t_real*ylm(153)&
      &-0.42966892442365972828_t_real*ylm(155)
    rvylm(3,1,186)= 0.70164641544562337175_t_real*ylm(160)
    rvylm(1,1,187)= 0.45741749991749247760_t_real*ylm(160)&
       &-0.23533936216582083830_t_real*ylm(162)
    rvylm(2,1,187)=-0.23533936216582083830_t_real*ylm(152)&
      &-0.45741749991749247760_t_real*ylm(154)
    rvylm(3,1,187)= 0.68612624987623871640_t_real*ylm(161)
    rvylm(1,1,188)= 0.48516452403758396048_t_real*ylm(161)&
       &-0.20754980866510833023_t_real*ylm(163)
    rvylm(2,1,188)=-0.20754980866510833023_t_real*ylm(151)&
      &-0.48516452403758396048_t_real*ylm(153)
    rvylm(3,1,188)= 0.66564023547027494640_t_real*ylm(162)
    rvylm(1,1,189)= 0.51291024855906620586_t_real*ylm(162)&
       &-0.17974340685458342477_t_real*ylm(164)
    rvylm(2,1,189)=-0.17974340685458342477_t_real*ylm(150)&
      &-0.51291024855906620586_t_real*ylm(152)
    rvylm(3,1,189)= 0.63971147342436279829_t_real*ylm(163)
    rvylm(1,1,190)= 0.54065487356324857942_t_real*ylm(163)&
       &-0.15191090506254999243_t_real*ylm(165)
    rvylm(2,1,190)=-0.15191090506254999243_t_real*ylm(149)&
      &-0.54065487356324857942_t_real*ylm(151)
    rvylm(3,1,190)= 0.60764362025019996974_t_real*ylm(164)
    rvylm(1,1,191)= 0.56839856005880510754_t_real*ylm(164)&
       &-0.12403473458920845618_t_real*ylm(166)
    rvylm(2,1,191)=-0.12403473458920845618_t_real*ylm(148)&
      &-0.56839856005880510754_t_real*ylm(150)
    rvylm(3,1,191)= 0.56839856005880510754_t_real*ylm(165)
    rvylm(1,1,192)= 0.59614143907684809140_t_real*ylm(165)&
       &-0.09607689228305228008_t_real*ylm(167)
    rvylm(2,1,192)=-0.09607689228305228008_t_real*ylm(147)&
      &-0.59614143907684809140_t_real*ylm(149)
    rvylm(3,1,192)= 0.52035490847039270454_t_real*ylm(166)
    rvylm(1,1,193)= 0.62388361833820354826_t_real*ylm(166)&
       &-0.06793662204867574503_t_real*ylm(168)
    rvylm(2,1,193)=-0.06793662204867574503_t_real*ylm(146)&
      &-0.62388361833820354826_t_real*ylm(148)
    rvylm(3,1,193)= 0.46076858867298268520_t_real*ylm(167)
    rvylm(1,1,194)= 0.65162518721684217549_t_real*ylm(167)&
       &-0.03922322702763680638_t_real*ylm(169)
    rvylm(2,1,194)=-0.03922322702763680638_t_real*ylm(145)&
      &-0.65162518721684217549_t_real*ylm(147)
    rvylm(3,1,194)= 0.38430756913220912034_t_real*ylm(168)
    rvylm(1,1,195)= 0.67936622048675745034_t_real*ylm(168)
    rvylm(2,1,195)=-0.67936622048675745034_t_real*ylm(146)
    rvylm(3,1,195)= 0.27735009811261456100_t_real*ylm(169)
    rvylm(1,1,196)= 0.70710678118654752437_t_real*ylm(169)
    rvylm(2,1,196)=-0.70710678118654752437_t_real*ylm(145)
    !! L+1
    rvylm(1,2,170)=-0.68228823922101306055_t_real*ylm(197)&
       &+.03509312031717982061_t_real*ylm(199)
    rvylm(2,2,170)= 0.03509312031717982061_t_real*ylm(223)&
      &+.68228823922101306055_t_real*ylm(225)
    rvylm(3,2,170)=-0.25788071477756376056_t_real*ylm(198)
    rvylm(1,2,171)=-0.65746939841502642270_t_real*ylm(198)&
       &+.06078306738548308356_t_real*ylm(200)
    rvylm(2,2,171)= 0.06078306738548308356_t_real*ylm(222)&
      &+.65746939841502642270_t_real*ylm(224)
    rvylm(3,2,171)=-0.35788101058031864758_t_real*ylm(199)
    rvylm(1,2,172)=-0.63265022359809469351_t_real*ylm(199)&
       &+.08596023825918792018_t_real*ylm(201)
    rvylm(2,2,172)= 0.08596023825918792018_t_real*ylm(221)&
      &+.63265022359809469351_t_real*ylm(223)
    rvylm(3,2,172)=-0.42980119129593960094_t_real*ylm(200)
    rvylm(1,2,173)=-0.60783067385483083563_t_real*ylm(200)&
       &+.11097419040461881708_t_real*ylm(202)
    rvylm(2,2,173)= 0.11097419040461881708_t_real*ylm(220)&
      &+.60783067385483083563_t_real*ylm(222)
    rvylm(3,2,173)=-0.48626453908386466850_t_real*ylm(201)
    rvylm(1,2,174)=-0.58301070130188203540_t_real*ylm(201)&
       &+.13591507055489058862_t_real*ylm(203)
    rvylm(2,2,174)= 0.13591507055489058862_t_real*ylm(219)&
      &+.58301070130188203540_t_real*ylm(221)
    rvylm(3,2,174)=-0.53221352061657884571_t_real*ylm(202)
    rvylm(1,2,175)=-0.55819024953845526273_t_real*ylm(202)&
       &+.16081688022566922492_t_real*ylm(204)
    rvylm(2,2,175)= 0.16081688022566922492_t_real*ylm(218)&
      &+.55819024953845526273_t_real*ylm(220)
    rvylm(3,2,175)=-0.57019571439055871414_t_real*ylm(203)
    rvylm(1,2,176)=-0.53336925166406964067_t_real*ylm(203)&
       &+.18569533817705186314_t_real*ylm(205)
    rvylm(2,2,176)= 0.18569533817705186314_t_real*ylm(217)&
      &+.53336925166406964067_t_real*ylm(219)
    rvylm(3,2,176)=-0.60172166781431544132_t_real*ylm(204)
    rvylm(1,2,177)=-0.50854762771560778738_t_real*ylm(204)&
       &+.21055872190307892367_t_real*ylm(206)
    rvylm(2,2,177)= 0.21055872190307892367_t_real*ylm(216)&
      &+.50854762771560778738_t_real*ylm(218)
    rvylm(3,2,177)=-0.62776482057434447816_t_real*ylm(205)
    rvylm(1,2,178)=-0.48372528131497451382_t_real*ylm(205)&
       &+.23541180771537917931_t_real*ylm(207)
    rvylm(2,2,178)= 0.23541180771537917931_t_real*ylm(215)&
      &+.48372528131497451382_t_real*ylm(217)
    rvylm(3,2,178)=-0.64898556687329512346_t_real*ylm(206)
    rvylm(1,2,179)=-0.45890209522830260048_t_real*ylm(206)&
       &+.26025754580373505710_t_real*ylm(208)
    rvylm(2,2,179)= 0.26025754580373505710_t_real*ylm(214)&
      &+.45890209522830260048_t_real*ylm(216)
    rvylm(3,2,179)=-0.66584514242771290249_t_real*ylm(207)
    rvylm(1,2,180)=-0.43407792540072461547_t_real*ylm(207)&
       &+.28509785719527935707_t_real*ylm(209)
    rvylm(2,2,180)= 0.28509785719527935707_t_real*ylm(213)&
      &+.43407792540072461547_t_real*ylm(215)
    rvylm(3,2,180)=-0.67866864743033923974_t_real*ylm(208)
    rvylm(1,2,181)=-0.40925259281898757373_t_real*ylm(208)&
       &+.30993404669460341628_t_real*ylm(210)
    rvylm(2,2,181)= 0.30993404669460341628_t_real*ylm(212)&
      &+.40925259281898757373_t_real*ylm(214)
    rvylm(3,2,181)=-0.68768190607350336151_t_real*ylm(209)
    rvylm(1,2,182)=-0.38442587221924475148_t_real*ylm(209)
    rvylm(2,2,182)= 0.47343207647399933921_t_real*ylm(211)&
      &+.38442587221924475148_t_real*ylm(213)
    rvylm(3,2,182)=-0.69303359695072724098_t_real*ylm(210)
    rvylm(1,2,183)=-0.50854762771560778737_t_real*ylm(212)
    rvylm(2,2,183)=-0.50854762771560778737_t_real*ylm(210)
    rvylm(3,2,183)=-0.69480833377965118676_t_real*ylm(211)
    rvylm(1,2,184)= 0.47343207647399933921_t_real*ylm(211)&
       &-0.38442587221924475148_t_real*ylm(213)
    rvylm(2,2,184)=-0.38442587221924475148_t_real*ylm(209)
    rvylm(3,2,184)=-0.69303359695072724098_t_real*ylm(212)
    rvylm(1,2,185)= 0.30993404669460341628_t_real*ylm(212)&
       &-0.40925259281898757373_t_real*ylm(214)
    rvylm(2,2,185)=-0.40925259281898757373_t_real*ylm(208)&
      &-0.30993404669460341628_t_real*ylm(210)
    rvylm(3,2,185)=-0.68768190607350336151_t_real*ylm(213)
    rvylm(1,2,186)= 0.28509785719527935707_t_real*ylm(213)&
       &-0.43407792540072461547_t_real*ylm(215)
    rvylm(2,2,186)=-0.43407792540072461547_t_real*ylm(207)&
      &-0.28509785719527935707_t_real*ylm(209)
    rvylm(3,2,186)=-0.67866864743033923974_t_real*ylm(214)
    rvylm(1,2,187)= 0.26025754580373505710_t_real*ylm(214)&
       &-0.45890209522830260048_t_real*ylm(216)
    rvylm(2,2,187)=-0.45890209522830260048_t_real*ylm(206)&
      &-0.26025754580373505710_t_real*ylm(208)
    rvylm(3,2,187)=-0.66584514242771290249_t_real*ylm(215)
    rvylm(1,2,188)= 0.23541180771537917931_t_real*ylm(215)&
       &-0.48372528131497451382_t_real*ylm(217)
    rvylm(2,2,188)=-0.48372528131497451382_t_real*ylm(205)&
      &-0.23541180771537917931_t_real*ylm(207)
    rvylm(3,2,188)=-0.64898556687329512346_t_real*ylm(216)
    rvylm(1,2,189)= 0.21055872190307892367_t_real*ylm(216)&
       &-0.50854762771560778738_t_real*ylm(218)
    rvylm(2,2,189)=-0.50854762771560778738_t_real*ylm(204)&
      &-0.21055872190307892367_t_real*ylm(206)
    rvylm(3,2,189)=-0.62776482057434447816_t_real*ylm(217)
    rvylm(1,2,190)= 0.18569533817705186314_t_real*ylm(217)&
       &-0.53336925166406964067_t_real*ylm(219)
    rvylm(2,2,190)=-0.53336925166406964067_t_real*ylm(203)&
      &-0.18569533817705186314_t_real*ylm(205)
    rvylm(3,2,190)=-0.60172166781431544132_t_real*ylm(218)
    rvylm(1,2,191)= 0.16081688022566922492_t_real*ylm(218)&
       &-0.55819024953845526273_t_real*ylm(220)
    rvylm(2,2,191)=-0.55819024953845526273_t_real*ylm(202)&
      &-0.16081688022566922492_t_real*ylm(204)
    rvylm(3,2,191)=-0.57019571439055871414_t_real*ylm(219)
    rvylm(1,2,192)= 0.13591507055489058862_t_real*ylm(219)&
       &-0.58301070130188203540_t_real*ylm(221)
    rvylm(2,2,192)=-0.58301070130188203540_t_real*ylm(201)&
      &-0.13591507055489058862_t_real*ylm(203)
    rvylm(3,2,192)=-0.53221352061657884571_t_real*ylm(220)
    rvylm(1,2,193)= 0.11097419040461881708_t_real*ylm(220)&
       &-0.60783067385483083563_t_real*ylm(222)
    rvylm(2,2,193)=-0.60783067385483083563_t_real*ylm(200)&
      &-0.11097419040461881708_t_real*ylm(202)
    rvylm(3,2,193)=-0.48626453908386466850_t_real*ylm(221)
    rvylm(1,2,194)= 0.08596023825918792018_t_real*ylm(221)&
       &-0.63265022359809469351_t_real*ylm(223)
    rvylm(2,2,194)=-0.63265022359809469351_t_real*ylm(199)&
      &-0.08596023825918792018_t_real*ylm(201)
    rvylm(3,2,194)=-0.42980119129593960094_t_real*ylm(222)
    rvylm(1,2,195)= 0.06078306738548308356_t_real*ylm(222)&
       &-0.65746939841502642270_t_real*ylm(224)
    rvylm(2,2,195)=-0.65746939841502642270_t_real*ylm(198)&
      &-0.06078306738548308356_t_real*ylm(200)
    rvylm(3,2,195)=-0.35788101058031864758_t_real*ylm(223)
    rvylm(1,2,196)= 0.03509312031717982061_t_real*ylm(223)&
       &-0.68228823922101306055_t_real*ylm(225)
    rvylm(2,2,196)=-0.68228823922101306055_t_real*ylm(197)&
      &-0.03509312031717982061_t_real*ylm(199)
    rvylm(3,2,196)=-0.25788071477756376056_t_real*ylm(224)
    !!
    If(maxL+1<=13) goto 1000
    !!

1000 Continue
    !!
    Do l=0,maxL
       Do m=-l,l
          ilm=l*l+l+1+m
          !!
          !!
          !!
          lpm=l-1
          If(lpm>=0)Then
             !!
             rd(:,:)=0.0_t_real
             Do i=1,3
                ki=2*Abs(i-2)-1
                Do m1=-1,1
                   jlm=lpm*lpm+lpm+1+ki*m+m1
                   
                   
                   If( jlm>0 ) Then
                      rd(i,:)=rd(i,:) +&
                           & ccc(i,1,m1+2,ilm)&
                           &*clpm(1,lpm)&
                           &*rvylm(:,1,jlm)&
                           &*clpm(1,l) 
                   Endif
                Enddo
             Enddo
             qvylm(:,:,1,ilm)=rd(:,:)
          Endif
          
          lpm=l+1
          !!
          rd(:,:)=0.0_t_real
          Do i=1,3
             ki=2*Abs(i-2)-1
             Do m1=-1,1
                jlm=lpm*lpm+lpm+1+ki*m+m1
                
                
                If( jlm>0 ) Then
                   rd(i,:)=rd(i,:) +&
                        & ccc(i,2,m1+2,ilm)&
                        &*clpm(2,lpm)&
                        &*rvylm(:,2,jlm)&
                        &*clpm(2,l) 
                Endif
             Enddo
          Enddo
          qvylm(:,:,2,ilm)=rd(:,:)

          
          lpm=l+1
          !!
          rd(:,:)=0.0_t_real
          Do i=1,3
             ki=2*Abs(i-2)-1
             Do m1=-1,1
                jlm=lpm*lpm+lpm+1+ki*m+m1
                
                
                If( jlm>0 ) Then
                   rd(i,:)=rd(i,:) +&
                        & ccc(i,2,m1+2,ilm)&
                        &*rvylm(:,1,jlm)&
                        &*clpm(1,l+1)&
                        &*clpm(2,l) 
                Endif
             Enddo
          Enddo
          !!
          !!
          lpm=l-1
          If(lpm>=0)Then
             !!
             Do i=1,3
                ki=2*Abs(i-2)-1
                Do m1=-1,1
                   jlm=lpm*lpm+lpm+1+ki*m+m1
                   
                   
                   If( jlm>0 ) Then
                      rd(i,:)=rd(i,:) +&
                           & ccc(i,1,m1+2,ilm)&
                           &*rvylm(:,2,jlm)&
                           &*clpm(2,l-1)&
                           &*clpm(1,l) 
                   Endif
                Enddo
             Enddo
          Endif
          qvylm(:,:,3,ilm)=rd(:,:)
          
          
          
          
          
          !!

       Enddo
    Enddo
    !!
    Return
    !!
  End Subroutine calc_real_quad_harmonics
#endif
  !!***********************************************************************
  !! R_{l,m}^{la}=sum_{k,nu}Y_{la,nu} \vec{e}_k*T_{la,nu,1,k}^{l,m}
  !! k=1,3
  !! T_{la,nu,1,k}^{l,m}=[
  !!                klj=(la-l+3)/2
  !!                T(1)= ccc(1,klj,1,l*(l+1)+1+m)*delta(nu,m+1)&
  !!                     &+ccc(1,klj,-1,l*(l+1)+1+m)*delta(nu,m-1)
  !!                T(2)= ccc(2,klj,1,l*(l+1)+1+m)*delta(nu,-m+1)&
  !!                     &+ccc(2,klj,-1,l*(l+1)+1+m)*delta(nu,-m-1)
  !!                T(3)= ccc(3,klj,0,l*(l+1)+1+m)*delta(nu,m)
  !! ]

  Subroutine  calc_real_vec_Ylm_coefs(maxj,ccc)
    !!
    Implicit None  
    !!IO
    Integer(t_int),Intent(in)    :: maxj
    Real(t_real),Intent(out)     :: ccc(3,2,-1:1,((maxj+1)**2))
    !!LOCAL
    !Character(*),Parameter::nsubfunc="calc_real_vec_Ylm_coefs"
    Integer(t_int) :: j,m,l,la,ilm,klj
    Real(t_real)   :: sm,denom,cm1,c1
    Real(t_real),Parameter :: sq2=1.41421356237309504880_t_real
    !!
    !!
    !!
    !!
    ccc(:,:,-1:1,:)=0.0_t_real
    !!
    Do j=0,maxj
       Do m=-j,j
          !!
          sm=Real(Sign(1_t_int,m),t_real)
          !!
          Do l=j-1,j+1,2
             If(l<0) Cycle
             !!
             !!
             la=l-j

             denom=1.0_t_real/Sqrt(2.0_t_real*(2*j+1+la)*(2*j+1+2*la))
             !!

             
             cm1=Sqrt(((j-la*m+la)*(j-la*m+1+la))*1.0_t_real)*denom
             c1=Sqrt(((j+la*m+la)*(j+la*m+1+la))*1.0_t_real)*denom
             

             ilm=j*(j+1)+1+m
             klj=(l-j+3)/2
             
             
             ccc(2,klj,-1,ilm)=-sm*c1
             If(m==0 .Or. m==-1) ccc(2,klj,-1,ilm)=ccc(2,klj,-1,ilm)*sq2
             If(m/=0 .And. m/=1) ccc(2,klj,1,ilm)=-sm*cm1
             
             
             
             If(m /=  0) ccc(1,klj,-1,ilm)= sm *cm1
             If(m ==  1) ccc(1,klj,-1,ilm)= ccc(1,klj,-1,ilm)*sq2
             If(m /= -1) ccc(1,klj,1,ilm) =-sm*c1
             If(m ==  0) ccc(1,klj,1,ilm) = ccc(1,klj,1,ilm) *sq2
             


             ccc(3,klj,0,ilm)=&
                  &-la*Sqrt(((2*j-2*m+1+la)*(2*j+2*m+1+la))*1.0_t_real)*denom
             
             
          Enddo
       Enddo
    Enddo
    !!
    !!
    !!
    Return
    !!
  End Subroutine calc_real_vec_Ylm_coefs
  !!***********************************************************************
  
End Module vecharm
  
